#include "gcc-plugin.h"

#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "tree.h"
#include "intl.h"
#include "tm.h"

#include "gimple.h"
#include "tree-pass.h"
#include "plugin-version.h"
#include "langhooks.h"
#include <string.h>
#include "tree-flow.h"
#include "tree-dump.h"
#include "tree-pass.h"
#include "basic-block.h"
#include "toplev.h"
#include "tree-inline.h"
#include "tree-iterator.h"
#include "real.h"
#include "fixed-value.h"
#include "tree-ssa-operands.h"
#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
#include "c-family/c-common.h"
#else
#include "c-common.h"
#endif
#include <assert.h>


int plugin_is_GPL_compatible;

int serialize_state;
int GimpleSSA_dump_nr;
static FILE *GimpleSSA_dump_file;
/* Information about a node to be dumped.  */
typedef struct serialize_node_info
{
  /* The index for the node.  */
  unsigned int index;
  /* Nonzero if the node is a binfo.  */
  unsigned int binfo_p : 1;
  unsigned int gimple_p : 1;
  unsigned int statement_p : 1;
} *serialize_node_info_p;

/* A serialize_queue is a link in the queue of things to be dumped.  */
typedef struct serialize_queue
{
  /* The queued tree node.  */
  splay_tree_node node;
  /* The next node in the queue.  */
  struct serialize_queue *next;
} *serialize_queue_p;

/* A serialize_info gives information about how we should perform the serialization
   and about the current state of the serialization.  */
struct serialize_info
{
  /* The stream on which to dump the information.  */
  FILE *stream;
  /* The original node.  */
  const_tree node;
  /* The next unused node index.  */
  unsigned int index;
  /* The next column.  */
  unsigned int column;
  /* The first node in the queue of nodes to be written out.  */
  serialize_queue_p queue;
  /* The last node in the queue.  */
  serialize_queue_p queue_end;
  /* Free queue nodes.  */
  serialize_queue_p free_list;
  /* The tree nodes which we have already written out.  The
     keys are the addresses of the nodes; the values are the integer
     indices we assigned them.  */
  splay_tree nodes;
};

typedef struct serialize_info *serialize_info_p;

/* Serialize the CHILD and its children.  */
#define serialize_child(field, child) \
  queue_and_serialize_index (di, field, child, DUMP_NONE)
/* Serialize the GIMPLE CHILD and its children.  */
#define serialize_gimple_child(field, child) \
  queue_and_serialize_gimple_index (di, field, child, DUMP_NONE)
/* Serialize the STATEMENT CHILD and its children.  */
#define serialize_statement_child(field, child) \
  queue_and_serialize_statement_index (di, field, child, DUMP_NONE)

#define LOCATION_COLUMN(LOC) ((expand_location (LOC)).column)
#define EXPR_COLUMNNO(NODE) LOCATION_COLUMN (EXPR_CHECK (NODE)->exp.locus)


static unsigned int queue (serialize_info_p, const_tree, int);
static unsigned int queue_gimple (serialize_info_p di, gimple t, int flags);
static unsigned int queue_statement (serialize_info_p di, struct control_flow_graph * cfg, int flags);
static void serialize_index (serialize_info_p, unsigned int);
static void queue_and_serialize_index (serialize_info_p di, const char *field, const_tree t, int);
static void queue_and_serialize_gimple_index (serialize_info_p di, const char *field, gimple t, int);
static void queue_and_serialize_statement_index (serialize_info_p di, const char *field, struct control_flow_graph * cfg, int);
static void queue_and_serialize_type (serialize_info_p di, const_tree t);
static void dequeue_and_serialize (serialize_info_p);
static void dequeue_and_serialize_gimple (serialize_info_p);
static void dequeue_and_serialize_statement (serialize_info_p);
static void serialize_new_line (serialize_info_p);
static void serialize_maybe_newline (serialize_info_p);
static void serialize_pointer (serialize_info_p di, const char *field, void *ptr);
static void serialize_int (serialize_info_p di, const char *field, int i);
static void serialize_wide_int (serialize_info_p di, const char *field, HOST_WIDE_INT i);
static void serialize_real (serialize_info_p di, tree t);
static int serialize_with_double_quote(const char * input, FILE * output, int length);
static int serialize_with_escape(const char * input, FILE * output, int length);
static void serialize_fixed (serialize_info_p di, const char *field, const FIXED_VALUE_TYPE *f);
static void serialize_string (serialize_info_p di, const char *string);
static void serialize_string_field (serialize_info_p di, const char *field, const char *str);
static void serialize_symbols (serialize_info_p di, const char * field, bitmap syms);
static void serialize_vops (serialize_info_p di, gimple stmt);
static void serialize_node (const_tree t, FILE *stream);
static void serialize_function_to_file (tree fn, FILE *file);
static tree gimple_call_expr_arglist (gimple g);
static tree build_custom_function_call_expr (location_t loc, tree fn, tree arglist);

struct serialize_info di_local = {0, 0, 0, 0, 0, 0, 0, 0};

/**
 * Dump the current version of gcc + current version of plugin
 */
void DumpVersion(FILE * stream)
{
   char * panda_plugin_version = PANDA_PLUGIN_VERSION;
   int version = __GNUC__, minor = __GNUC_MINOR__, patchlevel = __GNUC_PATCHLEVEL__;
   fprintf(stream, "GCC_VERSION: \"%d.%d.%d\"\n", version, minor, patchlevel);
   fprintf(stream, "PLUGIN_VERSION: \"%s\"\n", panda_plugin_version);
}

/* Begin a tree serialization and returns a stream to write to.
   Multiple calls will reopen and append to the dump file.  */
static FILE *
serialize_begin (char * name)
{
  FILE *stream;

  stream = fopen (name, serialize_state < 0 ? "w" : "a");
  if (stream && serialize_state < 0)
  {
     DumpVersion(stream);
  }
  if (!stream)
    error ("could not open serialize file %qs: %m", name);
  else
    serialize_state = 1;
  free (name);
   /* initialization of di_local */
   if (!di_local.nodes)
   {
      di_local.index = 0;
      di_local.nodes = splay_tree_new (splay_tree_compare_pointers, 0,
                                       (splay_tree_delete_value_fn) &free);
   }

  return stream;
}

/* Finish a tree serialize. STREAM is the stream created by
   serialize_begin.  */

static void
serialize_end (FILE *stream)
{
  fclose (stream);
}

unsigned int
gimplePssa (void)
{
   FILE *stream;
   /* dummy pass */
   if (!(cfun->curr_properties & PROP_cfg))
      fatal_error("PROP_cfg marked in the pass descriptor but not found");
   if (!(cfun->curr_properties & PROP_ssa))
      fatal_error("PROP_ssa marked in the pass descriptor but not found");
   if(cfun->cfg && basic_block_info)
   {
      struct dump_file_info *dfi = get_dump_file_info (GimpleSSA_dump_nr);
      char * name = concat (dump_base_name, dfi->suffix, NULL);
      stream = serialize_begin (name);
      if (stream)
        {
          serialize_function_to_file (cfun->decl, stream);
          serialize_end (stream);
        }
    }
 return 0;
}

/* Add T to the end of the queue of nodes to serialize.  Returns the index
   assigned to T.  */
static unsigned int
queue (serialize_info_p di, const_tree t, int flags)
{
  serialize_queue_p dq;
  serialize_node_info_p dni;
  unsigned int index;

  splay_tree_node n;

//  assert (!splay_tree_lookup (di_local.nodes, (splay_tree_key) t));
  /* Assign the next available index to T.  */
  index = ++di_local.index;

  /* Obtain a new queue node.  */
  if (di->free_list)
    {
      dq = di->free_list;
      di->free_list = dq->next;
    }
  else
    dq = XNEW (struct serialize_queue);

  /* Create a new entry in the splay-tree.  */
  dni = XNEW (struct serialize_node_info);
  dni->index = index;
  dni->binfo_p = ((flags & DUMP_BINFO) != 0);
  dni->gimple_p = 0;
  dni->statement_p = 0;
  dq->node = splay_tree_insert (di_local.nodes, (splay_tree_key) t,
				(splay_tree_value) dni);

  /* Add it to the end of the queue.  */
  dq->next = 0;
  if (!di->queue_end)
    di->queue = dq;
  else
    di->queue_end->next = dq;
  di->queue_end = dq;

  /* Return the index.  */
  return index;
}

/* Add G to the end of the queue of nodes to serialize.  Returns the index
   assigned to G.  */
static unsigned int
queue_gimple (serialize_info_p di, gimple g, int flags)
{
  serialize_queue_p dq;
  serialize_node_info_p dni;
  unsigned int index;

  splay_tree_node n;

  assert (!splay_tree_lookup (di_local.nodes, (splay_tree_key) g));
  
  /* Assign the next available index to G.  */
  index = ++di_local.index;

  /* Obtain a new queue node.  */
  if (di->free_list)
    {
      dq = di->free_list;
      di->free_list = dq->next;
    }
  else
    dq = XNEW (struct serialize_queue);

  /* Create a new entry in the splay-tree.  */
  dni = XNEW (struct serialize_node_info);
  dni->index = index;
  dni->binfo_p = ((flags & DUMP_BINFO) != 0);
  dni->gimple_p = 1;
  dni->statement_p = 0;
  dq->node = splay_tree_insert (di_local.nodes, (splay_tree_key) g,
				(splay_tree_value) dni);

  /* Add it to the end of the queue.  */
  dq->next = 0;
  if (!di->queue_end)
    di->queue = dq;
  else
    di->queue_end->next = dq;
  di->queue_end = dq;

  /* Return the index.  */
  return index;
}

/* Add CFG to the end of the queue of nodes to serialize.  Returns the index
   assigned to CFG.  */
static unsigned int
queue_statement (serialize_info_p di, struct control_flow_graph * cfg, int flags)
{
  serialize_queue_p dq;
  serialize_node_info_p dni;
  unsigned int index;

  splay_tree_node n;

  assert (!splay_tree_lookup (di_local.nodes, (splay_tree_key) cfg));
  /* Assign the next available index to G.  */
  index = ++di_local.index;

  /* Obtain a new queue node.  */
  if (di->free_list)
    {
      dq = di->free_list;
      di->free_list = dq->next;
    }
  else
    dq = XNEW (struct serialize_queue);

  /* Create a new entry in the splay-tree.  */
  dni = XNEW (struct serialize_node_info);
  dni->index = index;
  dni->binfo_p = ((flags & DUMP_BINFO) != 0);
  dni->gimple_p = 0;
  dni->statement_p = 1;
  dq->node = splay_tree_insert (di_local.nodes, (splay_tree_key) cfg,
				(splay_tree_value) dni);

  /* Add it to the end of the queue.  */
  dq->next = 0;
  if (!di->queue_end)
    di->queue = dq;
  else
    di->queue_end->next = dq;
  di->queue_end = dq;

  /* Return the index.  */
  return index;
}

static void
serialize_index (serialize_info_p di, unsigned int index)
{
  fprintf (di->stream, "@%-6u ", index);
  di->column += 8;
}

/* If T has not already been output, queue it for subsequent output.
   FIELD is a string to print before printing the index.  Then, the
   index of T is printed.  */
static void
queue_and_serialize_index (serialize_info_p di, const char *field, const_tree t, int flags)
{
  unsigned int index;
  splay_tree_node n;

  /* If there's no node, just return.  This makes for fewer checks in
     our callers.  */
  if (!t)
    return;

  /* See if we've already queued or dumped this node.  */
  n = splay_tree_lookup(di_local.nodes, (splay_tree_key) t);
  if (n)
  {
     index = ((serialize_node_info_p) n->value)->index;
      /*fprintf(di_local.stream, "Yet dumped as %u\n", index);*/
  }
  else
  {
    /* If we haven't, add it to the queue.  */
    index = queue (di, t, flags);
      /*fprintf(di_local.stream, "NOT Yet dumped as %u\n", index);*/
  }

  /* Print the index of the node.  */
  serialize_maybe_newline (di);
  fprintf (di->stream, "%-4s: ", field);
  di->column += 6;
  serialize_index (di, index);
}

/* If G has not already been output, queue it for subsequent output.
   FIELD is a string to print before printing the index.  Then, the
   index of G is printed.  */
static void
queue_and_serialize_gimple_index (serialize_info_p di, const char *field, gimple g, int flags)
{
  unsigned int index;
  splay_tree_node n;

  /* If there's no node, just return.  This makes for fewer checks in
     our callers.  */
  if (!g)
    return;

  /* See if we've already queued or dumped this node.  */
  n = splay_tree_lookup(di_local.nodes, (splay_tree_key) g);
  if (n)
  {
     index = ((serialize_node_info_p) n->value)->index;
      /*fprintf(di_local.stream, "Yet dumped as %u\n", index);*/
  }
  else
  {
    /* If we haven't, add it to the queue.  */
    index = queue_gimple (di, g, flags);
      /*fprintf(di_local.stream, "NOT Yet dumped as %u\n", index);*/
  }

  /* Print the index of the node.  */
  serialize_maybe_newline (di);
  fprintf (di->stream, "%-4s: ", field);
  di->column += 6;
  serialize_index (di, index);
}

/* If T has not already been output, queue it for subsequent output.
   FIELD is a string to print before printing the index.  Then, the
   index of T is printed.  */
static void
queue_and_serialize_statement_index (serialize_info_p di, const char *field, struct control_flow_graph * cfg, int flags)
{
  unsigned int index;
  splay_tree_node n;

  /* If there's no node, just return.  This makes for fewer checks in
     our callers.  */
  if (!cfg)
    return;

  /* See if we've already queued or dumped this node.  */
  n = splay_tree_lookup(di_local.nodes, (splay_tree_key) cfg);
  if (n)
  {
     index = ((serialize_node_info_p) n->value)->index;
      /*fprintf(di_local.stream, "Yet dumped as %u\n", index);*/
  }
  else
  {
    /* If we haven't, add it to the queue.  */
    index = queue_statement (di, cfg, flags);
      /*fprintf(di_local.stream, "NOT Yet dumped as %u\n", index);*/
  }

  /* Print the index of the node.  */
  serialize_maybe_newline (di);
  fprintf (di->stream, "%-4s: ", field);
  di->column += 6;
  serialize_index (di, index);
}

/* Serialize the type of T.  */
static void
queue_and_serialize_type (serialize_info_p di, const_tree t)
{
  queue_and_serialize_index (di, "type", TREE_TYPE (t), DUMP_NONE);
}

/* Serialize column control */
#define SOL_COLUMN 25		/* Start of line column.  */
#define EOL_COLUMN 55		/* End of line column.  */
#define COLUMN_ALIGNMENT 15	/* Alignment.  */

/* Insert a new line in the serialize output, and indent to an appropriate
   place to start printing more fields.  */

static void
serialize_new_line (serialize_info_p di)
{
  fprintf (di->stream, "\n%*s", SOL_COLUMN, "");
  di->column = SOL_COLUMN;
}

/* If necessary, insert a new line.  */
static void
serialize_maybe_newline (serialize_info_p di)
{
  int extra;

  /* See if we need a new line.  */
  if (di->column > EOL_COLUMN)
    serialize_new_line (di);
  /* See if we need any padding.  */
  else if ((extra = (di->column - SOL_COLUMN) % COLUMN_ALIGNMENT) != 0)
    {
      fprintf (di->stream, "%*s", COLUMN_ALIGNMENT - extra, "");
      di->column += COLUMN_ALIGNMENT - extra;
    }
}

/* Serialize pointer PTR using FIELD to identify it.  */
static void
serialize_pointer (serialize_info_p di, const char *field, void *ptr)
{
  serialize_maybe_newline (di);
  fprintf (di->stream, "%-4s: %-8lx ", field, (unsigned long) ptr);
  di->column += 15;
}

/* Serialize integer I using FIELD to identify it.  */
static void
serialize_int (serialize_info_p di, const char *field, int i)
{
  serialize_maybe_newline (di);
  fprintf (di->stream, "%-4s: %-7d ", field, i);
  di->column += 14;
}

/* Serialize wide integer i using FIELD to identify it.  */
static void
serialize_wide_int (serialize_info_p di, const char *field, HOST_WIDE_INT i)
{
   serialize_maybe_newline (di);
   fprintf (di->stream, "%-4s: "HOST_WIDE_INT_PRINT_DEC, field, i);
   di->column += 21;
}

/* Serialize real r using FIELD to identify it.  */
static void
serialize_real (serialize_info_p di, tree t)
{
#if !defined(REAL_IS_NOT_DOUBLE) || defined(REAL_ARITHMETIC)
   REAL_VALUE_TYPE d;
#endif
   static char string[1024];
   serialize_maybe_newline (di);
   /* Code copied from print_node.  */
   if (TREE_OVERFLOW (t))
   {
      fprintf (di->stream, "%s ", "overflow");
      di->column += 8;
   }
#if !defined(REAL_IS_NOT_DOUBLE) || defined(REAL_ARITHMETIC)
   d = TREE_REAL_CST (t);
   if (REAL_VALUE_ISINF (d))
      fprintf (di->stream, "valr: %-7s ", "\"Inf\"");
   else if (REAL_VALUE_ISNAN (d))
      fprintf (di->stream, "valr: %-7s ", "\"Nan\"");
   else
   {
      real_to_decimal (string, &d, sizeof (string), 0, 1);
      fprintf (di->stream, "valr: \"%s\" ", string);
   }
#endif
   {
      fprintf (di->stream, "valx: \"");
      real_to_hexadecimal(string, &d, sizeof (string), 0, 0);
      fprintf (di->stream, "%s", string);
      fprintf (di->stream, "\"");
   }
   di->column += 21;
}

/* Serialize a string and add double quotes */
static int
serialize_with_double_quote(const char * input, FILE * output, int length)
{
   int new_length;
   fprintf(output, "\"");
   new_length = serialize_with_escape(input, output, length);
   fprintf(output, "\"");
   return new_length + 2;
}


/* Add a backslash before an escape sequence to serialize the string
   with the escape sequence */
static int
serialize_with_escape(const char * input, FILE * output, int length)
{
   int i;
   int k = 0;
   for (i = 0; i < length; i++)
   {
      switch (input[i])
      {
         case '\n':
         {
            /* new line*/
            fprintf(output, "\\");
            fprintf(output, "n");
            k += 2;
            break;
         }
         case '\t':
         {
            /* horizontal tab */
            fprintf(output, "\\");
            fprintf(output, "t");
            k += 2;
            break;
         }
         case '\v':
         {
            /* vertical tab */
            fprintf(output, "\\");
            fprintf(output, "v");
            k += 2;
            break;
         }
         case '\b':
         {
            /* backspace */
            fprintf(output, "\\");
            fprintf(output, "b");
            k += 2;
            break;
         }
         case '\r':
         {
            /* carriage return */
            fprintf(output, "\\");
            fprintf(output, "r");
            k += 2;
            break;
         }
         case '\f':
         {
            /* jump page */
            fprintf(output, "\\");
            fprintf(output, "f");
            k += 2;
            break;
         }
         case '\a':
         {
            /* alarm */
            fprintf(output, "\\");
            fprintf(output, "a");
            k += 2;
            break;
         }
         case '\\':
         {
            /* backslash */
            fprintf(output, "\\");
            fprintf(output, "\\");
            k += 2;
            break;
         }
         case '\"':
         {
            /* double quote */
            fprintf(output, "\\");
            fprintf(output, "\"");
            k += 2;
            break;
         }
         case '\'':
         {
            /* quote */
            fprintf(output, "\\");
            fprintf(output, "\'");
            k += 2;
            break;
         }
         case '\0':
         {
            /* null */
            fprintf(output, "\\");
            fprintf(output, "0");
            k += 2;
            break;
         }
         default:
         {
            fprintf(output, "%c", input[i]);
            k++;
         }
      }
   }
   return k;
}

/* Serialize the fixed-point value F, using FIELD to identify it.  */
static void
serialize_fixed (serialize_info_p di, const char *field, const FIXED_VALUE_TYPE *f)
{
  char buf[32];
  fixed_to_decimal (buf, f, sizeof (buf));
  serialize_maybe_newline (di);
  fprintf (di->stream, "%-4s: %s ", field, buf);
  di->column += strlen (buf) + 7;
}

/* Serialize the string S.  */
static void
serialize_string (serialize_info_p di, const char *string)
{
  serialize_maybe_newline (di);
  fprintf (di->stream, "%-13s ", string);
  if (strlen (string) > 13)
    di->column += strlen (string) + 1;
  else
    di->column += 14;
}

/* Serialize the string field S.  */
static void
serialize_string_field (serialize_info_p di, const char *field, const char *str)
{
   int length;
   serialize_maybe_newline (di);
   fprintf (di->stream, "%-4s: ", field);
   length = strlen(str);
   length = serialize_with_double_quote(str, di->stream, length);
   if (length > 7)
      di->column += 6 + length + 1;
   else
      di->column += 14;
}

/* Serialize symbols */
static void
serialize_symbols (serialize_info_p di, const char * field, bitmap syms)
{
   unsigned i;
   bitmap_iterator bi;

   if (syms == NULL)
      serialize_string_field (di, field, "NIL");
   else
   {

      EXECUTE_IF_SET_IN_BITMAP (syms, 0, i, bi)
      {
#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
         tree sym = referenced_var_lookup (cfun, i);
#else
         tree sym = referenced_var_lookup (i);
#endif
         serialize_child (field, sym);
      }
   }
}

/* Serialize virtual operands */
static void
serialize_vops (serialize_info_p di, gimple gs)
{

// dump_dfa_stats (stderr);
// dump_referenced_vars (stderr);
// dump_alias_info (stderr);

tree vdef = gimple_vdef (gs);
tree vuse = gimple_vuse (gs);

if (!ssa_operands_active () || !gimple_references_memory_p (gs))
  return;

if (vdef != NULL_TREE)
  {
    serialize_child ("vdef", vdef);
    serialize_child ("vdef_op", vuse);
    serialize_child ("stores", gimple_vop (cfun));
  }
else if (vuse != NULL_TREE)
  {
    serialize_string (di, "vuse");
    serialize_child ("vuse_op", vuse);
    // serialize_child ("loads", get_virtual_var (vuse));
    serialize_child ("loads", gimple_vop (cfun));
  }

}

static void
serialize_string_cst (serialize_info_p di, const char *field, const char *str, int length, unsigned int precision)
{
   int new_length;
   serialize_maybe_newline (di);
   if (precision == 8)
   {
      fprintf (di->stream, "%-4s: ", field);
      new_length = serialize_with_double_quote(str, di->stream, length - 1);
      if (new_length > 7)
         di->column += 6 + new_length + 1;
  else
    di->column += 14;
      serialize_int (di, "lngt", length);
   }
   else
   {
      const unsigned int * string = (const unsigned int *) str;
      unsigned int i, lngt = length / 4 - 1;
      fprintf (di->stream, "%-4s: \"", field);
      for (i = 0; i < lngt; i++)
         fprintf (di->stream, "\\x%x", string[i]);
      fprintf (di->stream, "\" ");
      di->column += 7 + lngt;
      serialize_int (di, "lngt", lngt + 1);
   }
}


/* Serialize the next node in the queue.  */
static void
dequeue_and_serialize (serialize_info_p di)
{
  serialize_queue_p dq;
  splay_tree_node stn;
  serialize_node_info_p dni;
  tree t, arg;
  unsigned int index;
  enum tree_code code;
  enum tree_code_class code_class;
  const char* code_name;

  /* Get the next node from the queue.  */
  dq = di->queue;
  stn = dq->node;
  dni = (serialize_node_info_p) stn->value;
  if(dni->gimple_p)
  {
    dequeue_and_serialize_gimple (di);
    return;
  }
  else if(dni->statement_p)
  {
    dequeue_and_serialize_statement (di);
    return;
  }

  t = (tree) stn->key;
  index = dni->index;

  /* Remove the node from the queue, and put it on the free list.  */
  di->queue = dq->next;
  if (!di->queue)
    di->queue_end = 0;
  dq->next = di->free_list;
  di->free_list = dq;

  /* Print the node index.  */
  serialize_index (di, index);

  /* And the type of node this is.  */
  if (dni->binfo_p)
    code_name = "binfo";
#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6 && __GNUC_PATCHLEVEL__>= 1)
  else if(TREE_CODE (t) == TARGET_MEM_REF)
    code_name = "target_mem_ref461";
#endif
  else
    code_name = tree_code_name[(int) TREE_CODE (t)];
  fprintf (di->stream, "%-16s ", code_name);
  di->column = 25;

  /* Figure out what kind of node this is.  */
  code = TREE_CODE (t);
  code_class = TREE_CODE_CLASS (code);

   /* We use has_stmt_ann because CALL_EXPR can be both an expression
      and a statement, and we have no guarantee that it will have a
      stmt_ann when it is used as an RHS expression.  stmt_ann will assert
      if you call it on something with a non-stmt annotation attached.  */

  // if (code != ERROR_MARK
  //       && is_gimple_stmt (t)
  //       && has_stmt_ann (t)
  //       && code != PHI_NODE)
  // {
  //    serialize_vops(di, t);
  // }

  /* Although BINFOs are TREE_VECs, we serialize them specially so as to be
     more informative.  */
  if (dni->binfo_p)
  {
    unsigned ix;
    tree base;
    VEC(tree,gc) *accesses = BINFO_BASE_ACCESSES (t);

    serialize_child ("type", BINFO_TYPE (t));

    if (BINFO_VIRTUAL_P (t))
      serialize_string (di, "virt");

    serialize_int (di, "bases", BINFO_N_BASE_BINFOS (t));
    for (ix = 0; BINFO_BASE_ITERATE (t, ix, base); ix++)
    {
      tree access = (accesses ? VEC_index (tree, accesses, ix) : access_public_node);
      const char *string = NULL;

      if (access == access_public_node)
        string = "pub";
      else if (access == access_protected_node)
        string = "prot";
      else if (access == access_private_node)
        string = "priv";
      else
        gcc_unreachable ();

      serialize_string (di, string);
      queue_and_serialize_index (di, "binf", base, DUMP_BINFO);
    }

    goto done;
  }

  /* We can knock off a bunch of expression nodes in exactly the same
     way.  */
  if (IS_EXPR_CODE_CLASS (code_class))
  {
    expanded_location xloc;
    /* If we're serializing children, serialize them now.  */
    if (!(code == MODIFY_EXPR))
      queue_and_serialize_type (di, t);
      if (EXPR_HAS_LOCATION(t))
      {
        serialize_maybe_newline (di);
        fprintf (di->stream, "srcp: \"%s\":%-d:%-6d ", EXPR_FILENAME(t), EXPR_LINENO(t), EXPR_COLUMNNO(t));
        di->column += 12 + strlen(EXPR_FILENAME(t)) + 8;
     }

     switch (code_class)
     {
       case tcc_unary:
         serialize_child ("op", TREE_OPERAND (t, 0));
         break;

       case tcc_binary:
       case tcc_comparison:
         serialize_child ("op", TREE_OPERAND (t, 0));
         serialize_child ("op", TREE_OPERAND (t, 1));
         break;

       case tcc_expression:
       case tcc_reference:
       case tcc_statement:
       case tcc_vl_exp:
         /* These nodes are handled explicitly below.  */
         break;

       default:
         gcc_unreachable ();
    }
  }
  else if (DECL_P (t))
  {
    expanded_location xloc;
    /* All declarations have names.  */
    if (DECL_NAME (t))
      serialize_child ("name", DECL_NAME (t));
    if (DECL_ASSEMBLER_NAME_SET_P (t) && DECL_ASSEMBLER_NAME (t) != DECL_NAME (t))
      serialize_child ("mngl", DECL_ASSEMBLER_NAME (t));
    if (DECL_ABSTRACT_ORIGIN (t))
      serialize_child ("orig", DECL_ABSTRACT_ORIGIN (t));
    /* And types.  */
    queue_and_serialize_type (di, t);
    serialize_child ("scpe", DECL_CONTEXT (t));
    if (!DECL_SOURCE_LOCATION (t))
    {
      serialize_maybe_newline (di);
      fprintf (di->stream, "srcp: \"<built-in>\":0:0 ");
      di->column += 12 + strlen ("<built-in>") + 8;
    }
    else
    {
       /* And a source position.  */
       xloc = expand_location (DECL_SOURCE_LOCATION (t));
       if (xloc.file)
       {
          serialize_maybe_newline (di);
          fprintf (di->stream, "srcp: \"%s\":%-d:%-6d ", xloc.file,
                xloc.line, xloc.column);
          di->column += 12 + strlen (xloc.file) + 8;
       }
    }
    /* And any declaration can be compiler-generated.  */
    if (CODE_CONTAINS_STRUCT (TREE_CODE (t), TS_DECL_COMMON) && DECL_ARTIFICIAL (t))
       serialize_string (di, "artificial");
    if (code == LABEL_DECL && LABEL_DECL_UID(t) != -1)
       serialize_int(di, "uid", LABEL_DECL_UID (t));
    else
       serialize_int(di, "uid", DECL_UID(t));
    /*if (DECL_ATTRIBUTES(t) && !MTAG_P(t))
      {
      serialize_child("attributes", DECL_ATTRIBUTES(t));
      }*/
  }
  else if (code_class == tcc_type)
  {
     /* All types have qualifiers.  */
     int quals = lang_hooks.tree_dump.type_quals (t);

     if (quals != TYPE_UNQUALIFIED)
     {
        if(quals & TYPE_QUAL_VOLATILE)
           fprintf (di->stream, "qual: %c%c%c     ",
                 (quals & TYPE_QUAL_CONST) ? 'c' : ' ',
                 'v',
                 (quals & TYPE_QUAL_RESTRICT) ? 'r' : ' ');
        else
           fprintf (di->stream, "qual: %c%c     ",
                 (quals & TYPE_QUAL_CONST) ? 'c' : ' ',
                 (quals & TYPE_QUAL_RESTRICT) ? 'r' : ' ');
        di->column += 14;
     }

     /* All types have associated declarations.  */
     serialize_child ("name", TYPE_NAME (t));

     /* All types have a main variant.  */
     if (TYPE_MAIN_VARIANT (t) != t)
        serialize_child ("unql", TYPE_MAIN_VARIANT (t));

     /* And sizes.  */
     serialize_child ("size", TYPE_SIZE (t));

     /* All types have context.  */ /*NB TBC*/
     serialize_child ("scpe", TYPE_CONTEXT (t));/*NB TBC*/

     /* All types have alignments.  */
     serialize_int (di, "algn", TYPE_ALIGN (t));
  }
  else if (code_class == tcc_constant)
     /* All constants can have types.  */
     queue_and_serialize_type (di, t);

  /* Give the language-specific code a chance to print something.  If
     it's completely taken care of things, don't bother printing
     anything more ourselves.
     FIXME this language specific part cannot be used since the type of the parameters used by this plugin is different from the GCC one.
     FIXME Currently, only C is supported correctly.
     if (lang_hooks.tree_dump.dump_tree (di, t))
     goto done;
     */
  /* Now handle the various kinds of nodes.  */
  switch (code)
    {
      int i;

    case IDENTIFIER_NODE:
      serialize_string_field (di, "strg", IDENTIFIER_POINTER (t));
      serialize_int (di, "lngt", IDENTIFIER_LENGTH (t));
      break;

    case TREE_LIST:
      serialize_child ("purp", TREE_PURPOSE (t));
      serialize_child ("valu", TREE_VALUE (t));
      serialize_child ("chan", TREE_CHAIN (t));
      break;

    case STATEMENT_LIST:
      {
	tree_stmt_iterator it;
	for (i = 0, it = tsi_start (t); !tsi_end_p (it); tsi_next (&it), i++)
	  {
	    char buffer[32];
	    sprintf (buffer, "%u", i);
	    serialize_child (buffer, tsi_stmt (it));
	  }
      }
      break;

    case TREE_VEC:
      serialize_int (di, "lngt", TREE_VEC_LENGTH (t));
      for (i = 0; i < TREE_VEC_LENGTH (t); ++i)
	{
	  serialize_child ("op", TREE_VEC_ELT (t, i));
	}
      break;

    case INTEGER_TYPE:
    case ENUMERAL_TYPE:
      serialize_int (di, "prec", TYPE_PRECISION (t));
      if (TYPE_UNSIGNED (t))
        serialize_string (di, "unsigned");
      serialize_child ("min", TYPE_MIN_VALUE (t));
      serialize_child ("max", TYPE_MAX_VALUE (t));

      if (code == ENUMERAL_TYPE)
	serialize_child ("csts", TYPE_VALUES (t));
      break;

    case COMPLEX_TYPE:
      if (TYPE_UNSIGNED (t))
        serialize_string (di, "unsigned");
      break;

    case REAL_TYPE:
      serialize_int (di, "prec", TYPE_PRECISION (t));
      break;

    case FIXED_POINT_TYPE:
      serialize_int (di, "prec", TYPE_PRECISION (t));
      serialize_string_field (di, "sign", TYPE_UNSIGNED (t) ? "unsigned": "signed");
      serialize_string_field (di, "saturating",
			 TYPE_SATURATING (t) ? "saturating": "non-saturating");
      break;

    case POINTER_TYPE:
      serialize_child ("ptd", TREE_TYPE (t));
      break;

    case REFERENCE_TYPE:
      serialize_child ("refd", TREE_TYPE (t));
      break;

    case METHOD_TYPE:
      serialize_child ("retn", TREE_TYPE (t));
      serialize_child ("prms", TYPE_ARG_TYPES (t));
      serialize_child ("clas", TYPE_METHOD_BASETYPE (t));
      break;

    case FUNCTION_TYPE:
      serialize_child ("retn", TREE_TYPE (t));
      serialize_child ("prms", TYPE_ARG_TYPES (t));
      if (stdarg_p(t))
        serialize_string(di, "varargs");
      break;

    case ARRAY_TYPE:
      serialize_child ("elts", TREE_TYPE (t));
      serialize_child ("domn", TYPE_DOMAIN (t));
      break;

    case VECTOR_TYPE:
      serialize_child ("elts", TREE_TYPE (t));
      break;

    case RECORD_TYPE:
    case UNION_TYPE:
    {
      tree op;
      if (TREE_CODE (t) == RECORD_TYPE)
	serialize_string (di, "struct");
      else
	serialize_string (di, "union");
      if (TYPE_FIELDS(t))
         for ( op = TYPE_FIELDS (t); op; op = TREE_CHAIN(op))
         {
            serialize_child ("flds", op);
         }

      if (TYPE_METHODS(t))
         for ( op = TYPE_METHODS (t); op; op = TREE_CHAIN(op))
         {
            serialize_child ("fncs", op);
         }

      /*serialize_child ("flds", TYPE_FIELDS (t));
      serialize_child ("fncs", TYPE_METHODS (t));*/
      queue_and_serialize_index (di, "binf", TYPE_BINFO (t),
			    DUMP_BINFO);
      break;
    }
    case CONST_DECL:
      serialize_child ("cnst", DECL_INITIAL (t));
      break;
    case SSA_NAME:
      serialize_child ("var", SSA_NAME_VAR (t));
      serialize_int (di, "vers", SSA_NAME_VERSION (t));
      // if (SSA_NAME_PTR_INFO (t) && SSA_NAME_PTR_INFO (t)->name_mem_tag)
      //  serialize_child ("ptr_info", SSA_NAME_PTR_INFO (t)->name_mem_tag);
      if (TREE_THIS_VOLATILE (t))
        serialize_string(di, "volatile");
      else 
        serialize_gimple_child("def_stmt", SSA_NAME_DEF_STMT(t));
      if (!is_gimple_reg (SSA_NAME_VAR (t)))
        serialize_string(di, "virtual");
      if (SSA_NAME_IS_DEFAULT_DEF(t))
         serialize_string(di, "default");
      break;

    case DEBUG_EXPR_DECL:
      serialize_int (di, "-uid", DEBUG_TEMP_UID (t));
      /* Fall through.  */

    case VAR_DECL:
    case PARM_DECL:
    case FIELD_DECL:
    case RESULT_DECL:
      if (TREE_CODE (t) == FIELD_DECL && DECL_C_BIT_FIELD (t))
         serialize_string (di, "bitfield");
      if(TREE_CODE (t) == VAR_DECL)
      {
         if (DECL_EXTERNAL (t))
            serialize_string (di, "extern");
         else if (!TREE_PUBLIC (t) && TREE_STATIC (t))
            serialize_string (di, "static");
      }
      if (TREE_CODE (t) == PARM_DECL)
        serialize_child ("argt", DECL_ARG_TYPE (t));
      else
        serialize_child ("init", DECL_INITIAL (t));
      serialize_child ("size", DECL_SIZE (t));
      serialize_int (di, "algn", DECL_ALIGN (t));

      if (TREE_CODE (t) == FIELD_DECL)
      {
        if (DECL_FIELD_OFFSET (t))
          serialize_child ("bpos", bit_position (t));
      }
      else if (TREE_CODE (t) == VAR_DECL || TREE_CODE (t) == PARM_DECL)
      {
        serialize_int (di, "used", TREE_USED (t));
        if (DECL_REGISTER (t))
          serialize_string (di, "register");
      }
      break;

    case FUNCTION_DECL:
      arg = DECL_ARGUMENTS (t);
      while (arg)
        {
          serialize_child ("arg", arg);
          arg = TREE_CHAIN (arg);
        }
      if (DECL_EXTERNAL (t))
         serialize_string (di, "undefined");
      if (!TREE_PUBLIC (t))
         serialize_string (di, "static");
      if (gimple_has_body_p (t) && t == cfun->decl)
         serialize_statement_child("body", cfun->cfg);
      break;

    case INTEGER_CST:
      serialize_wide_int(di, "value", TREE_INT_CST_LOW(t));
      break;

    case STRING_CST:
      if (TREE_TYPE (t))
        serialize_string_cst(di, "strg" , TREE_STRING_POINTER (t), TREE_STRING_LENGTH (t), TYPE_ALIGN(TREE_TYPE (t)));
      else
        serialize_string_cst(di, "strg" , TREE_STRING_POINTER (t), TREE_STRING_LENGTH (t), 8);
      break;

    case REAL_CST:
      serialize_real (di, t);
      break;

    case COMPLEX_CST:
      serialize_child ("real", TREE_REALPART (t));
      serialize_child ("imag", TREE_IMAGPART (t));
      break;

    case FIXED_CST:
      serialize_fixed (di, "valu", TREE_FIXED_CST_PTR (t));
      break;

    case VECTOR_CST:
      {
         tree elt;

         for (elt = TREE_VECTOR_CST_ELTS (t); elt; elt = TREE_CHAIN (elt))
         {
            serialize_child ("valu", TREE_VALUE (elt));
         }
      }
      break;

    case TRUTH_NOT_EXPR:
    case ADDR_EXPR:
    case INDIRECT_REF:
#if __GNUC__ == 4 && __GNUC_MINOR__ ==  5
    case ALIGN_INDIRECT_REF:
    case MISALIGNED_INDIRECT_REF:
#endif
    case CLEANUP_POINT_EXPR:
    case SAVE_EXPR:
    case REALPART_EXPR:
    case IMAGPART_EXPR:
    case VIEW_CONVERT_EXPR:
      /* These nodes are unary, but do not have code class `1'.  */
      serialize_child ("op", TREE_OPERAND (t, 0));
      break;

#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
    case MEM_REF:
#endif
    case TRUTH_ANDIF_EXPR:
    case TRUTH_ORIF_EXPR:
    case TRUTH_AND_EXPR:
    case TRUTH_OR_EXPR:
    case TRUTH_XOR_EXPR:
    case INIT_EXPR:
    case MODIFY_EXPR:
      serialize_child ("op", TREE_OPERAND (t, 0));
      serialize_child ("op", TREE_OPERAND (t, 1));
      break;
    case COMPOUND_EXPR:
    case PREDECREMENT_EXPR:
    case PREINCREMENT_EXPR:
    case POSTDECREMENT_EXPR:
    case POSTINCREMENT_EXPR:
    case WITH_SIZE_EXPR:
      /* These nodes are binary, but do not have code class `2'.  */
      serialize_child ("op", TREE_OPERAND (t, 0));
      serialize_child ("op", TREE_OPERAND (t, 1));
      break;

    case COMPONENT_REF:
      serialize_child ("op", TREE_OPERAND (t, 0));
      serialize_child ("op", TREE_OPERAND (t, 1));
      serialize_child ("op", TREE_OPERAND (t, 2));
      break;

    case ARRAY_REF:
    case ARRAY_RANGE_REF:
      serialize_child ("op", TREE_OPERAND (t, 0));
      serialize_child ("op", TREE_OPERAND (t, 1));
      serialize_child ("op", TREE_OPERAND (t, 2));
      serialize_child ("op", TREE_OPERAND (t, 3));
      break;

    case COND_EXPR:
      serialize_child ("op", TREE_OPERAND (t, 0));
      serialize_child ("op", TREE_OPERAND (t, 1));
      serialize_child ("op", TREE_OPERAND (t, 2));
      break;

    case BIT_FIELD_REF:
      serialize_child ("op", TREE_OPERAND (t, 0));
      serialize_child ("op", TREE_OPERAND (t, 1));
      serialize_child ("op", TREE_OPERAND (t, 2));
      /*in case t is the last statement of a basic block we have to dump the true and the false edge*/
      /*since it is more easy to compute that edge during the basic block that dump is performed there*/
      break;

    case TRY_FINALLY_EXPR:
      serialize_child ("op", TREE_OPERAND (t, 0));
      serialize_child ("op", TREE_OPERAND (t, 1));
      break;

    case CALL_EXPR:
      serialize_child ("fn", TREE_OPERAND (t, 1));
      tree arg;
      call_expr_arg_iterator iter;
      FOR_EACH_CALL_EXPR_ARG (arg, iter, t)
      {
         serialize_child ("arg", arg);
      }
      break;

    case CONSTRUCTOR:
      {
	unsigned HOST_WIDE_INT cnt;
	tree index, value;
	FOR_EACH_CONSTRUCTOR_ELT (CONSTRUCTOR_ELTS (t), cnt, index, value)
	  {
	    serialize_child ("idx", index);
	    serialize_child ("valu", value);
	  }
      }
      break;

    case BIND_EXPR:
      serialize_child ("vars", TREE_OPERAND (t, 0));
      serialize_child ("body", TREE_OPERAND (t, 1));
      break;

    case LOOP_EXPR:
    case EXIT_EXPR:
    case RETURN_EXPR:
      serialize_child ("op", TREE_OPERAND (t, 0));
      break;

    case TARGET_EXPR:
      serialize_child ("decl", TREE_OPERAND (t, 0));
      serialize_child ("init", TREE_OPERAND (t, 1));
      serialize_child ("clnp", TREE_OPERAND (t, 2));
      /* There really are two possible places the initializer can be.
	 After RTL expansion, the second operand is moved to the
	 position of the fourth operand, and the second operand
	 becomes NULL.  */
      serialize_child ("init", TREE_OPERAND (t, 3));
      break;

    case CASE_LABEL_EXPR:
         if (CASE_LOW (t) && CASE_HIGH (t))
         {
            serialize_child ("op", CASE_LOW (t));
            serialize_child ("op", CASE_HIGH (t));
            serialize_child ("goto", CASE_LABEL (t));
         }
         else if (CASE_LOW (t))
         {
            serialize_child ("op", CASE_LOW (t));
            serialize_child ("goto", CASE_LABEL (t));
         }
         else
         {
            serialize_string (di, "default");
            serialize_child ("goto", CASE_LABEL (t));
         }
      break;
    case LABEL_EXPR:
      serialize_child ("name", TREE_OPERAND (t,0));
      break;
    case GOTO_EXPR:
      serialize_child ("labl", TREE_OPERAND (t, 0));
      break;
    case SWITCH_EXPR:
      serialize_child ("cond", TREE_OPERAND (t, 0));
      serialize_child ("body", TREE_OPERAND (t, 1));
      if (TREE_OPERAND (t, 2))
        {
      	  serialize_child ("labl", TREE_OPERAND (t,2));
        }
      break;
    case OMP_CLAUSE:
      {
	int i;
	fprintf (di->stream, "%s\n", omp_clause_code_name[OMP_CLAUSE_CODE (t)]);
	for (i = 0; i < omp_clause_num_ops[OMP_CLAUSE_CODE (t)]; i++)
	  serialize_child ("op: ", OMP_CLAUSE_OPERAND (t, i));
      }
      break;
      case TARGET_MEM_REF:
#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6 && __GNUC_PATCHLEVEL__>= 1)
         serialize_child ("base", TMR_BASE(t));
         serialize_child ("offset", TMR_OFFSET(t));
         serialize_child ("idx", TMR_INDEX(t));
         serialize_child ("step", TMR_STEP(t));
         serialize_child ("idx2", TMR_INDEX2(t));
#else
         serialize_child ("symbol", TMR_SYMBOL(t));
         serialize_child ("base", TMR_BASE(t));
         serialize_child ("idx", TMR_INDEX(t));
         serialize_child ("step", TMR_STEP(t));
         serialize_child ("offset", TMR_OFFSET(t));
         serialize_child ("orig", TMR_ORIGINAL(t));
#endif
         break;
    default:
      /* There are no additional fields to print.  */
      break;
    }

 done:

  /* Terminate the line.  */
  fprintf (di->stream, "\n");
}

/* Serialize the next gimple node in the queue.  */
static void
dequeue_and_serialize_gimple (serialize_info_p di)
{
  serialize_queue_p dq;
  splay_tree_node stn;
  serialize_node_info_p dni;
  gimple g;
  unsigned int index;
  enum tree_code code;
  enum tree_code_class code_class;
  const char* code_name;
  expanded_location xloc;
  int i;
  tree t;
  
  /* Get the next node from the queue.  */
  dq = di->queue;
  stn = dq->node;
  dni = (serialize_node_info_p) stn->value;
  g = (gimple) stn->key;
  index = dni->index;

  /* Remove the node from the queue, and put it on the free list.  */
  di->queue = dq->next;
  if (!di->queue)
    di->queue_end = 0;
  dq->next = di->free_list;
  di->free_list = dq;

  di->column = 0;
  /* Print the node index.  */
  serialize_index (di, index);
  if(gimple_code (g) == GIMPLE_CALL && gimple_call_lhs (g))
     code_name = "gimple_assign";
  else
     code_name = gimple_code_name[(int) gimple_code(g)];
  fprintf (di->stream, "%-16s ", code_name);

  di->column = 25;

  // print_gimple_stmt (stderr, g, 0, 0);  

   if (gimple_has_mem_ops (g))
     serialize_vops (di, g);

  if(gimple_has_location (g))
  {
    xloc = expand_location (gimple_location (g));
    serialize_maybe_newline (di);
    fprintf (di->stream, "srcp: \"%s\":%-d:%-6d ", xloc.file,
                  xloc.line, xloc.column);
    di->column += 12 + strlen(xloc.file) + 8;
  }

  switch (gimple_code (g))
    {

    case GIMPLE_ASM:
      {
        int i, n;
        if (gimple_asm_volatile_p (g))
          serialize_string (di, "volatile");  
        serialize_string_field (di, "str", gimple_asm_string (g));
        n = gimple_asm_noutputs (g);
        if (n > 0)
          {
            tree arglist = NULL_TREE;
            for (i = n - 1; i >= 0; i--)
            {
              tree item = gimple_asm_output_op (g, i);
              arglist = tree_cons (TREE_PURPOSE (item), TREE_VALUE (item), arglist);
            }
            serialize_child ("out", arglist);
          }
        n = gimple_asm_ninputs (g);
        if (n > 0)
          {
            tree arglist = NULL_TREE;
            for (i = n - 1; i >= 0; i--)
            {
              tree item = gimple_asm_input_op (g, i);
              arglist = tree_cons (TREE_PURPOSE (item), TREE_VALUE (item), arglist);
            }

            serialize_child ("in", arglist);
          }
        n = gimple_asm_nclobbers (g); 
        if (n > 0)
          {
            tree arglist = NULL_TREE;
            for (i = n - 1; i >= 0; i--)
            {
              tree item = gimple_asm_clobber_op (g, i);
              arglist = tree_cons (TREE_PURPOSE (item), TREE_VALUE (item), arglist);
            }

            serialize_child ("clob", arglist);
          }
      }
      break;

    /* For instance, given the GENERIC expression, a = b + c, its tree representation is:
    MODIFY_EXPR <VAR_DECL  <a>, PLUS_EXPR <VAR_DECL <b>, VAR_DECL <c>>>
    In this case, the GIMPLE form for this statement is logically identical to its GENERIC form but in GIMPLE, 
    the PLUS_EXPR on the RHS of the assignment is not represented as a tree, instead the two operands are taken 
    out of the PLUS_EXPR sub-tree and flattened into the GIMPLE tuple as follows:
    GIMPLE_ASSIGN <PLUS_EXPR, VAR_DECL <a>, VAR_DECL <b>, VAR_DECL <c>> */
    
    /* The first operand is what to set; the second, the new value - consistent with olderversions */

    case GIMPLE_ASSIGN:
      if (get_gimple_rhs_class (gimple_expr_code (g)) == GIMPLE_BINARY_RHS) 
        {
          serialize_child ("op", gimple_assign_lhs (g)); 
          serialize_child ("op", build2 (gimple_assign_rhs_code (g), TREE_TYPE (gimple_assign_lhs (g)), gimple_assign_rhs1 (g), gimple_assign_rhs2 (g))); 
        }
      else if (get_gimple_rhs_class (gimple_expr_code (g)) == GIMPLE_UNARY_RHS)
        {
          serialize_child ("op", gimple_assign_lhs (g)); 
          serialize_child ("op", build1 (gimple_assign_rhs_code (g), TREE_TYPE (gimple_assign_lhs (g)), gimple_assign_rhs1 (g))); 
        }
      else if (get_gimple_rhs_class (gimple_expr_code (g)) == GIMPLE_SINGLE_RHS)
        {
          t = gimple_assign_rhs1 (g);
          serialize_child ("op", gimple_assign_lhs (g)); 
          serialize_child ("op", gimple_assign_rhs1 (g)); 
        }
      else
        gcc_unreachable ();

      break;

    case GIMPLE_BIND:
      {
        tree op;
        if (gimple_bind_vars(g))
          for ( op = gimple_bind_vars (g); op; op = TREE_CHAIN (op))
            {
              serialize_child ("vars", op);
            }
        serialize_child ("body", gimple_op (g, 1));
      }
      break;

    case GIMPLE_CALL:
      /* Function call.  Operand 0 is the function.
      * Operand 1 is the argument list, a list of expressions made out of a chain of TREE_LIST nodes.
      * Operand 2 is the static chain argument, or NULL. */
      {
        size_t i;
        bool builtin_p;
        tree lhs = gimple_call_lhs (g);
        if(lhs)
        {
           tree arg_list = gimple_call_expr_arglist (g);
           t = build_custom_function_call_expr (UNKNOWN_LOCATION, gimple_call_fn (g), arg_list);
           serialize_child ("op",lhs); 
           serialize_child ("op", t); 
        }
        else
        {
           serialize_child ("fn", gimple_call_fn (g));
           unsigned int arg_index;
           for(arg_index = 0; arg_index < gimple_call_num_args(g); arg_index++)
           {
              serialize_child ("arg", gimple_call_arg(g, arg_index));
           }
        }
      }
      break;

    case GIMPLE_COND:
      GIMPLE_CHECK (g, GIMPLE_COND);
      t = build2 (gimple_cond_code (g), boolean_type_node, gimple_cond_lhs (g), gimple_cond_rhs (g));
      serialize_child ("op", t);
      break;

    case GIMPLE_LABEL:
      serialize_child ("op", gimple_label_label (g));
      break;

    case GIMPLE_GOTO:
      serialize_child ("op", gimple_goto_dest (g));
      break;

    case GIMPLE_NOP:
      break;

    case GIMPLE_RETURN:
      serialize_child ("op", gimple_return_retval (g));
      break;

    case GIMPLE_SWITCH:
      {
         tree label_vec;
         size_t i, num_elmts = gimple_switch_num_labels (g);

         label_vec = make_tree_vec (num_elmts);


         for (i = 0; i < num_elmts; i++)
         {
            TREE_VEC_ELT (label_vec, i) = gimple_switch_label (g, i);
         }
         serialize_child ("op", gimple_switch_index (g));
         serialize_child ("op", label_vec);
      }
      break;

    case GIMPLE_TRY:
      error ("Encountered gimple instruction set GIMPLE_TRY");
      break;

    case GIMPLE_PHI:
      serialize_child ("res", gimple_phi_result (g));
      for (i = 0; i < gimple_phi_num_args (g); i++)
        {
          serialize_child ("def", gimple_phi_arg_def (g, i));
          serialize_int (di, "edge", gimple_phi_arg_edge (g, i)->src->index);
        }
      if(!is_gimple_reg (gimple_phi_result (g)))
        serialize_string(di, "virtual");
      break;

    case GIMPLE_OMP_PARALLEL:
      error ("Encountered gimple instruction set GIMPLE_OMP_PARALLEL");
      break;

    case GIMPLE_OMP_TASK:
      error ("Encountered gimple instruction set GIMPLE_TASK");
      break;

    case GIMPLE_OMP_ATOMIC_LOAD:
      error ("Encountered gimple instruction set GIMPLE_OMP_ATOMIC_LOAD");
      break;

    case GIMPLE_OMP_ATOMIC_STORE:
      error ("Encountered gimple instruction set GIMPLE_OMP_ATOMIC_STORE");
      break;

    case GIMPLE_OMP_FOR:
      error ("Encountered gimple instruction set GIMPLE_OMP_FOR");
      break;

    case GIMPLE_OMP_CONTINUE:
      error ("Encountered gimple instruction set GIMPLE_OMP_CONTINUE");
      break;

    case GIMPLE_OMP_SINGLE:
      error ("Encountered gimple instruction set GIMPLE_OMP_SINGLE");
      break;

    case GIMPLE_OMP_RETURN:
      error ("Encountered gimple instruction set GIMPLE_OMP_RETURN");
      break;

    case GIMPLE_OMP_SECTIONS:
      error ("Encountered gimple instruction set GIMPLE_OMP_SECTIONS");
      break;

    case GIMPLE_OMP_SECTIONS_SWITCH:
      error ("Encountered gimple instruction set GIMPLE_OMP_SECTIONS_SWITCH");
      break;

    case GIMPLE_OMP_MASTER:
      error ("Encountered gimple instruction set GIMPLE_OMP_MASTER");
      break;

    case GIMPLE_OMP_ORDERED:
      error ("Encountered gimple instruction set GIMPLE_OMP_ORDERED");
      break;

    case GIMPLE_OMP_SECTION:
      error ("Encountered gimple instruction set GIMPLE_OMP_SECTION");
      break;

    case GIMPLE_OMP_CRITICAL:
      error ("Encountered gimple instruction set GIMPLE_OMP_CRITICAL");
      break;

    case GIMPLE_CATCH:
      error ("Encountered gimple instruction set GIMPLE_CATCH");
      break;

    case GIMPLE_EH_FILTER:
      error ("Encountered gimple instruction set GIMPLE_EH_FILTER");
      break;

    case GIMPLE_EH_MUST_NOT_THROW:
      error ("Encountered gimple instruction set GIMPLE_EH_MUST_NOT_THROW");
      break;

    case GIMPLE_RESX:
      error ("Encountered gimple instruction set GIMPLE_RESX");
      break;

    case GIMPLE_EH_DISPATCH:
      error ("Encountered gimple instruction set GIMPLE_EH_DISPATCH");
      break;

    case GIMPLE_DEBUG:
      if (gimple_debug_bind_p (g)) 
        {
          serialize_child ("var", gimple_debug_bind_get_var (g));
          serialize_child ("val", gimple_debug_bind_get_value (g));
        }
      break;

    case GIMPLE_PREDICT:
      error ("Encountered gimple instruction set GIMPLE_PREDICT");
      // t = build1 (NOP_EXPR, void_type_node, size_zero_node);
      break;

    default:
      error ("Encountered UNKNOWN gimple instruction set");
      break;
    }
  fprintf (di->stream, "\n");

}

/* Serialize the next statement node in the queue.  */
static void
dequeue_and_serialize_statement (serialize_info_p di)
{
  serialize_queue_p dq;
  splay_tree_node stn;
  serialize_node_info_p dni;
  unsigned int index;
  enum tree_code code;
  enum tree_code_class code_class;
  const char* code_name;
  basic_block bb;

  /* Get the next node from the queue.  */
  dq = di->queue;
  stn = dq->node;
  dni = (serialize_node_info_p) stn->value;
  // cfg = (tree) stn->key;
  index = dni->index;

  /* Remove the node from the queue, and put it on the free list.  */
  di->queue = dq->next;
  if (!di->queue)
    di->queue_end = 0;
  dq->next = di->free_list;
  di->free_list = dq;

  /* Print the node index.  */
  serialize_index (di, index);

  /* And the type of node this is.  */
  if (dni->binfo_p)
    code_name = "binfo";
  else
    code_name = "statement_list";
  fprintf (di->stream, "%-16s ", code_name);
  di->column = 25;

  /* In case of basic blocks the function print:
                 + first a list of all statements
                 +  then for each basic block it prints:
                    - the number of the basic block
                    - the predecessor basic block
                    - the successor basic block
                    - list of statement
                 + otherwise it prints only the list of statements */
  FOR_EACH_BB (bb)
    {
      gimple_stmt_iterator gsi;
      gimple stmt;
      edge e;
      edge_iterator ei;
      const char *field;

      if (bb->index != 0)
        serialize_new_line (di);
      serialize_int (di, "bloc", bb->index );

      /*if(bb->flags & BB_PARALLELIZABLE_LOOP)
        serialize_string(di, "hpl");
      */

      FOR_EACH_EDGE (e, ei, bb->preds)
        if (e->src == ENTRY_BLOCK_PTR)
          {
            serialize_maybe_newline (di);
            field = "pred: ENTRY";
            fprintf (di->stream, "%-4s ", field);
            di->column += 14;
	  }
        else
          serialize_int (di, "pred", e->src->index);
      FOR_EACH_EDGE (e, ei, bb->succs)
        if (e->dest == EXIT_BLOCK_PTR)
          {
            serialize_maybe_newline (di);
            field = "succ: EXIT";
            fprintf (di->stream, "%-4s ", field);
            di->column += 14;
          }
        else
          serialize_int (di, "succ", e->dest->index);
      /* in case the last statement is a cond_expr we print also the type of the outgoing edges*/
      stmt = last_stmt (bb);
      if (stmt && gimple_code (stmt) == GIMPLE_COND)
        {
          edge true_edge, false_edge;
          /* When we are emitting the code or changing CFG, it is possible that
          the edges are not yet created.  When we are using debug_bb in such
          a situation, we do not want it to crash.  */
          if (EDGE_COUNT (bb->succs) == 2)
            {
              extract_true_false_edges_from_block (bb, &true_edge, &false_edge);
              serialize_int (di, "true_edge", true_edge->dest->index);
              serialize_int (di, "false_edge", false_edge->dest->index);
            }
        }
      /* dump phi information*/
      {
        gimple_stmt_iterator i;
        for (i = gsi_start_phis (bb); !gsi_end_p (i); gsi_next (&i))
          {
            gimple phi = gsi_stmt (i);
            serialize_gimple_child("phi", phi);
          }
      }

      for (gsi = gsi_start_bb (bb); !gsi_end_p (gsi); gsi_next (&gsi))
        {
          stmt = gsi_stmt (gsi);
          serialize_gimple_child("stmt", stmt);
        }
    }
  /* Terminate the line.  */
  fprintf (di->stream, "\n");
}

/* Serialize T, and all its children, on STREAM.  */

static void
serialize_node (const_tree t, FILE *stream)
{
  struct serialize_info di;
  serialize_queue_p dq;
  serialize_queue_p next_dq;
  splay_tree_node n;

  /* Initialize the dump-information structure.  */
  di_local.stream = stream;
  di_local.column = 0;
  di_local.queue = 0;
  di_local.queue_end = 0;
  di_local.free_list = 0;
  di_local.node = t;

   /* Queue up the first node when not yet considered.  */
   n = splay_tree_lookup(di_local.nodes, (splay_tree_key) t);
   if (!n)
      queue (&di_local, t, DUMP_NONE);
   else if (FUNCTION_DECL == TREE_CODE(t))
   {
      queue (&di_local, t, DUMP_NONE);
   }


  /* Until the queue is empty, keep dumping nodes.  */
  while (di_local.queue)
    dequeue_and_serialize (&di_local);

  /* Now, clean up.  */
  for (dq = di_local.free_list; dq; dq = next_dq)
    {
      next_dq = dq->next;
      free (dq);
    }
  /*splay_tree_delete (di.nodes);*/
}

/* Serialize FUNCTION_DECL FN to file FILE
   */
static void
serialize_function_to_file (tree fn, FILE *file)
{
  tree arg;
  fprintf (file, "\n;; Function %s (%s)\n\n", lang_hooks.decl_printable_name (fn, 2), lang_hooks.decl_printable_name (fn, 2)); 
  fprintf (file, ";; %s (", lang_hooks.decl_printable_name (fn, 2));
  arg = DECL_ARGUMENTS (fn);
  while (arg)
    {
      print_generic_expr (file, TREE_TYPE (arg), TDF_SLIM);
      fprintf (file, " ");
      print_generic_expr (file, arg, TDF_SLIM);
      if (TREE_CHAIN (arg))
        fprintf (file, ", ");
      arg = TREE_CHAIN (arg);
    }
  fprintf (file, ")\n");

  serialize_node (fn, file);
}

tree
gimple_call_expr_arglist (gimple g)
{
  tree arglist = NULL_TREE;
  int i;
  for (i = gimple_call_num_args (g) - 1; i >= 0; i--)
    arglist = tree_cons (NULL_TREE, gimple_call_arg (g, i), arglist);
  return arglist;
}

tree
build_custom_function_call_expr (location_t loc, tree fn, tree arglist)
{
  tree fntype = TREE_TYPE (fn);
  int n = list_length (arglist);
  tree *argarray = (tree *) alloca (n * sizeof (tree));
  int i;

  for (i = 0; i < n; i++, arglist = TREE_CHAIN (arglist))
    argarray[i] = TREE_VALUE (arglist);
  return fold_builtin_call_array (loc, TREE_TYPE (fntype), fn, n, argarray);
}

/**
 * Serialize globlal variable on files
 */
void serialize_global_variable()
{
   struct varpool_node *vnode;
   char * name = concat (dump_base_name, ".001t.tu", NULL);
   FILE * stream = serialize_begin(name);
   FOR_EACH_STATIC_VARIABLE (vnode)
   {
      tree var = vnode->decl;

      if (TREE_CODE (var) == VAR_DECL)
      {
         serialize_node(var, stream);
      }
   }
   serialize_end(stream);
}


