/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file BambuParameter.cpp
 * @brief This file contains the implementation of some methods for parameter parsing in Bambu tool
 *
 *
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10670 $
 * $Date: 2012-03-20 16:59:29 +0100 (Tue, 20 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/

///Autoheader
#include "config_BAMBU_DYNAMIC_GENERATOR_DIR.hpp"
#include "config_HAVE_ALTERA.hpp"
#include "config_HAVE_BEAGLE.hpp"
#include "config_HAVE_COIN_OR.hpp"
#include "config_HAVE_CUDD.hpp"
#include "config_HAVE_EXPERIMENTAL.hpp"
#include "config_HAVE_ICARUS.hpp"
#include "config_HAVE_ILP_BUILT.hpp"
#include "config_HAVE_LP_SOLVE.hpp"
#include "config_HAVE_MODELSIM.hpp"
#include "config_HAVE_XILINX.hpp"
#include "config_SKIP_WARNING_SECTIONS.hpp"
#include "config_HAVE_MAPPING_BUILT.hpp"
#include "config_RELEASE.hpp"
#include "config_REVISION.hpp"

///Header include
#include "BambuParameter.hpp"

///Frontend include
#include "FrontendFlow.hpp"

///HLS include
#if HAVE_EXPERIMENTAL
#include "evaluation.hpp"
#endif
#include "hls_constraints.hpp"
#include "HLS_constraintsIO.hpp"
#if HAVE_ILP_BUILT
#include "meilp_solver.hpp"
#endif
#include "memory_allocation.hpp"
#include "scheduling.hpp"
#include "parametric_list_based.hpp"
#include "STG_creator.hpp"
#include "liveness_computer.hpp"
#include "storage_value_insertion.hpp"
#include "reg_binding_creator.hpp"
#include "fu_binding_creator.hpp"
#include "conn_binding_creator.hpp"
#if HAVE_EXPERIMENTAL
#include "module_interface.hpp"
#endif
///creation of the target architecture
#include "controller_creator.hpp"
#include "datapath_creator.hpp"
#include "chaining.hpp"

#include "hls_flow.hpp"
#if HAVE_BEAGLE
#if SKIP_WARNING_SECTIONS
#pragma GCC diagnostic ignored "-Wshadow"
#pragma GCC diagnostic ignored "-Wsign-conversion"
#pragma GCC diagnostic ignored "-Wunused-parameter"
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Woverloaded-virtual"
#pragma GCC diagnostic ignored "-Wtype-limits"
#pragma GCC diagnostic ignored "-Wold-style-cast"
#pragma GCC diagnostic ignored "-Wextra"
#pragma GCC diagnostic ignored "-Wignored-qualifiers"
#pragma GCC diagnostic ignored "-Wsign-compare"
#pragma GCC diagnostic ignored "-Wsign-promo"
#endif
#include "FitnessFunction.hpp"
#include "dse_hls.hpp"
#endif

///Constant option include
#include "constant_strings.hpp"

///STD include
#include <cstring>
#include <iostream>
#include <string.h>

///Technology include
#include "language_writer.hpp"
#include "parse_technology.hpp"
#include "technology_manager.hpp"
#include "target_device.hpp"
#include "technology_builtin.hpp"

///Utility include
#include <boost/filesystem.hpp>
#include <boost/lexical_cast.hpp>
#include "cpu_time.hpp"
#include "dbgPrintHelper.hpp"
#include "fileIO.hpp"
#include <getopt.h>
#include "global_enums.hpp"
#include "utility.hpp"

///Wrapper include
#include "gcc_wrapper.hpp"

///=========== PARAMETERS STUFF =================///
///Design Space Exploration
#define OPT_DSE 284
#define OPT_MAX_EVALUATIONS 256
//ACO
#define OPT_ACO_FLOW 257
#define OPT_ACO_GENERATIONS 258
//Genetic Algorithm
#if HAVE_BEAGLE
#define OPT_MAX_INHERITANCE 259
#define OPT_MIN_INHERITANCE 260
#define OPT_GENERATION 261
#define OPT_TIME_WEIGHT 262
#define OPT_AREA_WEIGHT 263
#endif
//Simulated Annealing
#define OPT_MOSA_FLOW 281

///Scheduling
#define OPT_FIXED_SCHED 264
#define OPT_LIST_BASED 265
///Finite state machine
#define OPT_STG 267
///Estimation
#define OPT_EVALUTION 268
#define OPT_EVALUTION_MODEL 269
#define OPT_AREA_RESULT 270
#define OPT_DEVICE_NAME 275
#define OPT_PERIOD_CLOCK 276
#define OPT_POWER_OPTIMIZATION 277

///Datapath
#define OPT_DATAPATH 278
///Controller
#define OPT_CONTROLLER 279
///Output
#define OPT_PRETTY_PRINT 280

#define OPT_INSERT_VERIFICATION_OPERATION 282

#define OPT_DUMP_CONSTRAINTS 283

#define OPT_INSERT_VERIFICATION_OPERATION 282

#define OPT_TESTBENCH 285

#define OPT_GENERATE_VCD 286

///Parallel controller
#define OPT_PARALLEL_CONTROLLER 287
#define OPT_LOGICAL_OPTIMIZATION 288

#define OPT_ENABLE_IOB 289
#define OPT_DYNAMIC_GENERATORS_DIR 290
#define OPT_REGISTER_ALLOCATION 291
#define OPT_INSERT_MEMORY_PROFILE 293

///XML configuration
#define OPT_XML_CONFIG 294


/// constant correspond to the "parametric list based option"
#define PAR_LIST_BASED_OPT "parametric-list-based"
/// constant correspond to the fixed scheduling option
#define FIXED_SCHEDULING_OPT "fixed-scheduling"

void BambuParameter::PrintUsage(std::ostream &os)
{
   PrintVersion(os);
   os
   << "Usage:\n"
   << "       " << getOption<std::string>(OPT_program_name) << " [Options] -c <function_name> <source_file> [<constraints_file>] [<technology_file>]\n"
#if !RELEASE
   << "       " << getOption<std::string>(OPT_program_name) << " [Options] -c <function_name> <raw_file> [<constraints_file>] [<technology_file>] --use-raw\n"
#endif
   << "Options:\n"
   << "\n";
   PrintGeneralOptionsUsage(os);
   os
   << "    --top-fname, -c <fun_name>      Define the top function to be synthesized.\n"
   << "    --xml-config <file>             Define the path to the XML configuration file for the synthesis.\n";
#if HAVE_ILP_BUILT
   os
   << "    --time, -t <time>               Set maximun execution time (in seconds) for ILP formulations. (infinite)\n";
#endif
   PrintGccOptionsUsage(os);
   os << "\n"
   << "  Target:\n"
   << "    --target-device-type=<type>     Specify the target technology.\n"
   << "                                       FPGA - Field Programmable Gate Array (default)\n"
   << "                                       ASIC - Application Specific Integrated Circuit\n"
   << "    --target-file=file, -b<file>    Specify an XML description of the target device.\n";
#if HAVE_EXPERIMENTAL
   os
   << "    --edk-config <file>             Specify the configuration file for Xilinx EDK.\n"
   << "\n"
   << "  Frontend analysis:\n"
   << "    --pdg-reduction[=<algorithm>]   Perform Program Dependence Graph reduction.\n"
   << "                                    The reduced PDG is used to build the parallel controller, if enabled\n"
   << "                                       GP - Girkar-Polychronopoulos algorithm (default)\n"
   << "                                       P  - Polimi algorithm (not supported yet)\n";
#endif
   os
   << "\n"
   << "  Scheduling algorithms:\n"
   << "    --parametric-list-based[=<type] Perform scheduling with priority list-based based (on):\n"
   << "                                        0 - Dynamic mobility (default)\n"
   << "                                        1 - Static mobility\n"
   << "                                        2 - Priority-fixed mobility\n";
#if HAVE_EXPERIMENTAL
#if HAVE_ILP_BUILT
   os << "    --ilp=[<solver>]             Perform scheduling by using the integer linear programming\n"
   << "                                     formulation (Default option is <solver>=glpk) (off):\n"
   << "                                        G - Solve the ilp problem using glpk solver. (default)\n";
#if HAVE_COIN_OR
   os
   << "                                        C - Solve the ilp problem using the COIN-OR solver.\n";
#endif
#if HAVE_LP_SOLVE
   os
   << "                                        L - Solve the ilp problem using lp_solve solver.\n";
#endif
   os << "    --ilp-newform=[<solver>]     Perform scheduling by using the integer linear programming\n"
   << "                                     with the new formulation (Default option is <solver>=glpk) (off):\n"
   << "                                        G - Solve the ilp problem using glpk solver. (default)\n";
#if HAVE_COIN_OR
   os
   << "                                        C - Solve the ilp problem using the COIN-OR solver.\n";
#endif
#if HAVE_LP_SOLVE
   os
   << "                                        L - Solve the ilp problem using lp_solve solver.\n";
#endif
   os << "    --silp=[<solver>]            Perform scheduling by using the scheduling and allocation with ILP\n"
   << "                                     formulation (Default option is <solver>=glpk) (off):\n"
   << "                                        G - Solve the silp problem using glpk solver. (default)\n";
#if HAVE_COIN_OR
   os
   << "                                        C - Solve the silp problem using the COIN-OR solver.\n";
#endif
#if HAVE_LP_SOLVE
   os
   << "                                        L - Solve the silp problem using lp_solve solver.\n";
#endif
//   << "    --ilp-loop=[<solver>]        Perform loops scheduling by using the integer linear programming\n"
//   << "                                     formulation (Default option is <solver>=glpk) (off):\n"
//   << "                                        G - Solve the silp problem using glpk solver. (default)\n"
//#if HAVE_COIN_OR
//   << "                                        C - Solve the ilp problem using the COIN-OR solver.\n"
//#endif
//#if HAVE_LP_SOLVE
//   << "                                        L - Solve the ilp problem using lp_solve solver.\n"
//#endif
#endif
//   << "    --speculative, -S               Perform scheduling by using speculative computation (only list\n"
//   << "                                     and ILP scheduling have been affected by this option) (off).\n"
#endif
   os
   << "    --fixed-scheduling=<file>       Provide scheduling as XML file (off).\n"
   << "    --no-chaining                   Deactivate the chaining optimization (off).\n"
   << "\n"
   << "  State Transition Graph:\n"
   << "    --stg[=<type>]                  Create the state transition graph:\n"
   << "                                        0 - Basic Block-based (exploiting scheduling information) (default)\n"
   << "\n"
   << "  Binding:\n"
   << "    --storage-value-insertion       Define the storage value insertion algorithm\n"
   << "                                        0 - VARIABLE/VALUES (default)\n";
#if HAVE_EXPERIMENTAL
   os
   << "                                        1 - VALUES_INSTANCES\n"
   << "                                        2 - LIMITED_VALUES_INSTANCES\n" 
   << "                                        3 - PAULIN_SCHEME\n";
#endif

#if HAVE_EXPERIMENTAL
   os
   << "    --explore-fu-reg=[string]       Perform simultaneous FU/reg binding\n";
#endif
   os
   << "    --register-allocation=type      Perform register allocation with one of the following algorithm\n"
   << "                                        WEIGHTED_COLORING - use weighted coloring algorithm (default)\n"
   << "                                        COLORING - use simple coloring algorithm\n"
   << "                                        CHORDAL_COLORING - use chordal coloring algorithm\n"
   << "                                        BIPARTITE_MATCHING - use bipartite matching algorithm\n"
   << "                                        TTT_CLIQUE_COVERING - use a weighted clique covering algorithm\n"
   << "                                        UNIQUE_BINDING - unique binding algorithm\n";
#if HAVE_EXPERIMENTAL
   os
   << "                                        K_COFAMILY - use k_cofamily algorithm\n"
   << "                                        LEFT_EDGE - use left edge algorithm\n"
   << "                                        CYCLIC_ALLOCATION - use cyclic allocation algorithm\n";
#endif
   os
   << "    --module-binding=type            Perform module binding with one of the following algorithm\n"
   << "                                        WEIGHTED_COLORING  - solve the weighted clique covering problem performing a coloring on the conflict graph (default)\n"
   << "                                        COLORING  - solve the unweighted clique covering problem performing a coloring on the conflict graph\n"
   << "                                        TTT_FAST  - use Tomita, A. Tanaka, H. Takahashi maximal weighted cliques heuristic to solve the clique covering problem \n"
   << "                                        TTT_FULL  - use Tomita, A. Tanaka, H. Takahashi maximal weighted cliques algorithm to solve the clique covering problem\n"
   << "                                        BIPARTITE_MATCHING - solve the weighted clique covering problem exploiting the bipartite matching approach\n"
   << "                                        UNIQUE - use a 1-to-1 binding algorithm\n"
   << "\n"
   << "  Memory allocation:\n"
   << "    --memory-allocation=type        Perform memory allocation with one of the following algorithms\n"
   << "                                        GSS      - all global variables, static variables and strings are allocated on BRAMs (default)\n"
   << "                                        ALL_BRAM - all objects that need to be stored in memory are allocated on BRAMs\n"
   << "                                        NO_BRAM  - all objects that need to be stored in memory are allocated on an external memory\n"
   << "\n"
   << "  Interconnection:\n"
   << "    --interconnection, -C[<type>]   Perform interconnection binding\n"
   << "                                        M - mux-based architecture (default)\n"
   << "\n"
   << "  Datapath creation:\n"
   << "    --no-datapath                   Does not create the structural description of the datapath (off).\n"
   << "\n"
   << "  Controller creation:\n"
   << "    --no-controller                 Does not create the structural description of the controller (off).\n";
#if HAVE_EXPERIMENTAL
   os
   << "    --parallel-controller           Create the parallel controller\n"
   << "    --logical-optimization=<level>  Enable logic optimization on EPDG\n"
   << "                                        1  - enable optimizations based on CONDITIONAL instructions information\n"
   << "                                        2  - enable optimizations based on ACTIVATION PATH information\n"
   << "                                        3  - enable CONDITIONAL and ACTIVATION PATH optimizations\n"
   << "                                        4  - enable DATA_TRANSITIVITY optimizations\n"
   << "                                        5  - enable CONDITIONAL and DATA_TRANSITIVITY optimizations\n"
   << "                                        6  - enable ACTIVATION PATH and DATA_TRANSITIVITY optimizations\n"
   << "                                        7  - enable CONDITIONAL, ACTIVATION PATH and DATA_TRANSITIVITY optimizations\n"
   << "                                        8  - enable CONTROL_DATA_TRANSITIVITY optimizations\n"
   << "                                        9  - enable CONTROL_DATA_TRANSITIVITY and CONDITIONAL instructions optimizations\n"
   << "                                        16 - enable CONTROL_DATA_CONTROL_FLOW_TRANSITIVITY and CONDITIONAL instructions optimizations\n"
   << "\n";
#if HAVE_BEAGLE
   os
   << "  High-Level Synthesis Design Space Exploration:\n"
   << "    --dse=<type>                    Perform design space exploration (off)\n"
   << "                                      MOGA  - Multi-objective Genetic Algorithm with NSGA-II\n"
   << "                                      MOSA  - Multi-objective Simulated Annealing\n"
   << "                                      MOTS  - Multi-objective Tabu Search\n"
   << "                                      ACO   - Multi-objective Ant Colony Optimization\n"
   << "                                      GRASP - Grasp\n"
   << "    --max-evalutions=<num>          Perform no more than the specified number of evalutions.\n"
   << "    --aco-flow[=<number of ants>]   Use Ant Colony Optimization - (off)\n"
   << "    --generations[=<num>]           Number ant colony's generations\n"
   << "    --mixed_synthesis, -H[<type>]   Determine the exploration space\n"
   << "                                        0 - based on operation binding information (default)\n"
   << "                                        1 - based on scheduling priority information\n"
   << "                                        2 - based on binding and scheduling information\n"
   << "\n";
#endif
   os
   << "  Evaluation of HLS results:\n"
   << "    --evaluation                    Perform evaluation of the results (off).\n"
   << "                                      EXACT: based on actual synthesis\n"
   << "                                      LINEAR: based on linear regression (default)\n"
   << "                                      STATISTICAL: based on statistical information\n"
   << "    --objective=<obj1,obj2>         Select the objective to be analyzed\n"
   << "                                      AREA         - Area occupation\n"
   << "                                      TIME         - Latency of the entire computation\n"
   << "                                      POWER        - Power consumption\n"
   << "                                      FREQUENCY    - Maximum target frequency\n"
   << "                                      CLOCK_SLACK  - Slack between computation and period\n"
   << "                                      NUM_AF_EDGES    - Number of incoming edges in the FAFG\n"
   << "                                      EDGES_REDUCTION - Performance evaluation for dependence reduction techniques\n"
   << "    --evaluation-model=<model>      Set XML evaluation model\n"
   << "    --area_result=<variable>        Select the variable to be shown for area evaluation\n"
   << "    --timing-simulation             Perform a simulation considering the timing delays (off)\n";
#endif
   os
   << "\n";
#if HAVE_EXPERIMENTAL || HAVE_MODELSIM
   os
   << "  Checks and debugging:\n"
#if HAVE_EXPERIMENTAL
   << "    --vcd-debug                     Enable debugging based on vcd files\n"
#endif
#if HAVE_MODELSIM
   << "    --assert-debug                  Enable assertion debugging performed by modelsim\n"
#endif
   << "\n";
#endif
   os
   << "  RTL synthesis:\n"
   << "    --perform-synthesis             Enables the synthesis (off)\n"
   << "    --clock-period=value            Specify the period of the clock signal (default = 10ns)\n"
   << "    --device-name=value             Specify the name of the device. Two different cases are foreseen:\n"
   << "                                    - Xilinx: comma separated string specifying device, speed grade and package (e.g.,: \"xc5vlx50,-3,ff1153\")\n"
   << "                                    - Altera: a string defining the device name (e.g., CycloneII,EP2C70F896C6)\n"
   << "    --power-optimization            Enable the Xilinx power based optimization (default no)\n"
   << "    --no-connect-iob                Disconnect primary ports from the IOB (default primary ports are connected to IOBs)\n"
   << "\n";
   PrintOutputOptionsUsage(os);
   os
   << "    --pretty-print=<file>           Pretty print of current raw file (off).\n"
#if HAVE_EXPERIMENTAL
   << "    --insert-memory-profile         Insert additional instruction to the pretty print file to profile memory operations\n"
   << "    --parse-memory-profile          Parses the access xml file produced from the execution of the C file (created with --insert-memory-profile option)\n" 
   << "                                    to produce the read/write-weights xml file\n"
#endif
   << "    --writer,-w<language>           Represent the produced description using the specified language\n"
   << "                                        V - Verilog (default)\n"
   #if HAVE_EXPERIMENTAL
   << "                                        S - SystemC\n"
   << "                                        H - VHDL\n"
   #endif
   << "    --generate-tb=<file>            Generate testbench for the input values defined in the specified XML file\n"
   << "    --generate-vcd                  Enable .vcd output file generation for waveform visualization (requires testbench generation)\n"
#if HAVE_EXPERIMENTAL
   << "    --generate-interface=<type>     Wrap the top level with an interface\n"
   << "                                        FSL (interface to the FSL bus)\n"
#endif
 #if HAVE_ICARUS || HAVE_XILINX || HAVE_MODELSIM || HAVE_ALTERA
   << "    --simulate=<file>               Simulate the resulting implementation\n"
   << "    --simulator=<string>            Use the specified simulator\n"
#if HAVE_MODELSIM
   << "                                        MODELSIM - Mentor Modelsim\n"
#endif
#if HAVE_XILINX
   << "                                        ISIM - Xilinx iSim\n"
#endif
#if HAVE_ICARUS
   << "                                        ICARUS - Verilog Icarus project\n"
#endif
#endif
   << std::endl;
}

BambuParameter::BambuParameter(const std::string _program_name) :
   Parameter(_program_name)
{
   SetDefaults();
}

int BambuParameter::exec(int argc, char* argv[])
{
   exit_code = PARAMETER_NOTPARSED;

   /// flag to check if scheduling algorithm has been already chosen
   bool scheduling_set_p = false;
   /// variable used into option parsing
   int option_index;
   int next_option;

   // Bambu short option. An option character in this string can be followed by a colon (`:') to indicate that it
   // takes a required argument. If an option character is followed by two colons (`::'), its argument is optional;
   // this is a GNU extension.
   const char* const short_options = COMMON_SHORT_OPTIONS_STRING
                                     "c:t:U:H:SC::b:w:"
                                     GCC_SHORT_OPTIONS_STRING;

   const struct option long_options[] =
      {
         COMMON_LONG_OPTIONS,
         ///General options
         {"top-fname",                  required_argument, 0, 'c'},
         {"xml-config",                 required_argument, 0, OPT_XML_CONFIG},
         {"time",                       required_argument, 0, 't'},
         /// Frontend options
         {"circuit-dbg",                required_argument, 0,  0 },
#if HAVE_EXPERIMENTAL
         {"pdg-reduction",              optional_argument, 0,  0 },
         ///--- Flow options ---
         {"dump-constraints",           optional_argument, 0, OPT_DUMP_CONSTRAINTS},
         /// --- Design Space Exploration options ---
#if HAVE_BEAGLE
         {"dse",                        required_argument, 0, OPT_DSE},
         {"max_evaluations",            required_argument, 0, OPT_MAX_EVALUATIONS},
         //ACO
         {"aco-flow",                   optional_argument, 0, OPT_ACO_FLOW},
         {"generations",                optional_argument, 0, OPT_ACO_GENERATIONS},
         //Genetic algorithm
         {"mixed_synthesis",            optional_argument, 0, 'H'},
         {"seed",                       required_argument, 0, 'A'},
         {"run",                        required_argument, 0, 'B'},
         {"population",                 required_argument, 0, 'E'},
         {"generation",                 required_argument, 0, OPT_GENERATION},
         {"fitness_function",           required_argument, 0, 'P'},
         {"inheritance_rate",           required_argument, 0, 'F'},
         {"inheritance_mode",           required_argument, 0, 'G'},
         {"max_inheritance",            required_argument, 0, OPT_MAX_INHERITANCE},
         {"min_inheritance",            required_argument, 0, OPT_MIN_INHERITANCE},
         {"distance_rate",              required_argument, 0, 'M'},
         {"weighting_function",         required_argument, 0, 'N'},
         {"normalize",                  optional_argument, 0, 'Q'},
         {"time-weight",                required_argument, 0, OPT_TIME_WEIGHT}, //no short option
         {"area-weight",                required_argument, 0, OPT_AREA_WEIGHT}, //no short option
#endif
         /// --- Algorithms options ---
         {"explore-mux",                optional_argument,       0,  0 },
         {"explore-fu-reg",             required_argument,       0,  0 },
#endif
         /// Scheduling options
         {FIXED_SCHEDULING_OPT,         required_argument, 0, OPT_FIXED_SCHED},
         {PAR_LIST_BASED_OPT,           optional_argument, 0, OPT_LIST_BASED}, //no short option
#if HAVE_ILP_BUILT
         {"ilp",                        optional_argument, 0, 0},
         {"ilp-newform",                optional_argument, 0, 0},
         {"silp",                       optional_argument, 0, 0},
         {"ilp-loop",                   optional_argument, 0, 0},
#endif
         {"speculative",                no_argument,       0, 'S'},
         {"no-chaining",                no_argument,       0,  0 },
         /// Finite state machine options
         {"stg",                        optional_argument, 0, OPT_STG},
         /// module binding
         {"module-binding",             optional_argument, 0, 0 },
         /// register allocation
         {"register-allocation",        required_argument, 0, OPT_REGISTER_ALLOCATION },
         {"storage-value-insertion",    optional_argument, 0, 0 },
         /// Memory allocation
         {"memory-allocation",          optional_argument, 0,  0 },
         //interconnections
         {"interconnection",            optional_argument, 0, 'C' },
         /// Datapath options
         {"no-datapath",                no_argument,       0, OPT_DATAPATH},
         /// Controller options
         {"no-controller",              no_argument,       0, OPT_CONTROLLER},
#if HAVE_EXPERIMENTAL
         {"parallel-controller",        no_argument,       0, OPT_PARALLEL_CONTROLLER},
#if HAVE_CUDD
         {"logical-optimization",       required_argument, 0, OPT_LOGICAL_OPTIMIZATION},
#endif
         /// evaluation options
         {"evaluation",                 optional_argument, 0, OPT_EVALUTION},
         {"evaluation_model",           required_argument, 0, OPT_EVALUTION_MODEL},
         {"area_result",                required_argument, 0, OPT_AREA_RESULT},
         {"objective",                  required_argument, 0,  0 },
         {"timing-simulation",          no_argument,       0,  0 },
#if (HAVE_XILINX || HAVE_ALTERA) && HAVE_MODELSIM
         {"vcd-debug",                  no_argument,       0,  0 },
#endif
#endif
         {"assert-debug",               no_argument,       0,  0 },
         {"perform-synthesis",          no_argument,       0,  0 },
         {"device-name",                required_argument, 0, OPT_DEVICE_NAME},
         {"clock-period",               required_argument, 0, OPT_PERIOD_CLOCK},
         {"power-optimization",         no_argument,       0, OPT_POWER_OPTIMIZATION},
         {"no-connect-iob",             no_argument,       0, OPT_ENABLE_IOB},
         ///target options
         {"target-file",                required_argument, 0, 'b'},
         {"target-device-type",         required_argument, 0,  0 },
#if HAVE_EXPERIMENTAL
         {"edk-config",                 required_argument, 0,  0 },
#endif
         ///Output options
         {"writer",                     required_argument, 0, 'w'},
         {"dynamic-generators-dir",     required_argument, 0, OPT_DYNAMIC_GENERATORS_DIR},
         {"pretty-print",               required_argument, 0, OPT_PRETTY_PRINT},
         {"insert-memory-profile",      no_argument,       0, OPT_INSERT_MEMORY_PROFILE},
#if HAVE_EXPERIMENTAL
         {"generate-interface",         required_argument, 0,  0 },
#endif
         {"generate-tb",                required_argument, 0, OPT_TESTBENCH},
         {"generate-vcd",               no_argument,       0, OPT_GENERATE_VCD},
         {"simulate",                   optional_argument, 0,  0 },
         {"simulator",                  required_argument, 0,  0 },
         GCC_LONG_OPTIONS,
         {0, 0, 0, 0}
      };

   if (argc == 1)   // Bambu called without arguments, it simple prints help message
   {
      PrintUsage(std::cout);
      return EXIT_SUCCESS;
   }

   while (1)
   {
      next_option = getopt_long(argc, argv, short_options, long_options, &option_index);

      // no more options are available
      if (next_option == -1)
         break;

      switch (next_option)
      {
         ///general options
         case 'c':
            // set name of the function to be take into account as top function
            setOption(OPT_top_function_name, optarg);
            setOption(OPT_top_file, optarg);
            break;
         case 't':
            setOption(OPT_ilp_max_time, optarg);
            break;
#if HAVE_EXPERIMENTAL
         case OPT_XML_CONFIG:
            setOption(OPT_hls_flow, hls_flow::XML_CONFIG);
            setOption(OPT_xml_input_configuration, optarg);
            break;
         case OPT_DUMP_CONSTRAINTS:
            setOption(OPT_hls_flow, hls_flow::DUMP);
            setOption("dumpConstraints", true);
            if (optarg)
               setOption("dumpConstraints_file", optarg);
            break;
         ///frontend options
         case 'U': // compute the unrolling degree of a loop
            setOption("compute_unrolling_degree",true);
            if (optarg)
               setOption("IImax", optarg);
            else
               throw "BadParameters: compute unrolling degree needs a additional parameter";
            break;
         case OPT_INSERT_VERIFICATION_OPERATION:
            setOption("insert_verification_operation", true);
            break;
         ///design space exploration options
#if HAVE_BEAGLE
         case OPT_DSE:
         {
            setOption(OPT_hls_flow, hls_flow::DSE);
            if (std::string(optarg) == ("MOSA"))
               setOption("dse_algorithm", dse_hls::MOSA);
            else if (std::string(optarg) == ("MOTS"))
               setOption("dse_algorithm", dse_hls::MOTS);
            else if (std::string(optarg) == ("ACO"))
               setOption("dse_algorithm", dse_hls::ACO);
            else if (std::string(optarg) == ("GRASP"))
               setOption("dse_algorithm", dse_hls::GRASP);
            else if (std::string(optarg) == ("MOGA"))
               setOption("dse_algorithm", dse_hls::MOGA);
            else
               THROW_ERROR("Not supported exploration algorithm: " + std::string(optarg));
            break;
         }
         case OPT_MAX_EVALUATIONS:
            setOption("max_evaluations", optarg);
            break;
         //ACO
         case OPT_ACO_FLOW:
            setOption(OPT_hls_flow, hls_flow::DSE);
            setOption("dse_algorithm", dse_hls::ACO);
            if (optarg)
               setOption("aco_ant_num", optarg);
            break;
         case OPT_ACO_GENERATIONS:
            if (optarg)
               setOption("aco_generations", optarg);
            break;
         case OPT_MOSA_FLOW:
            setOption(OPT_hls_flow, hls_flow::DSE);
            setOption("dse_algorithm", dse_hls::MOSA);
            break;
         //multi-objective genetic algorithm
         case 'H':  // enable mixed high level synthesis
            setOption(OPT_hls_flow, hls_flow::DSE);
            setOption("dse_algorithm", dse_hls::MOGA);
            if (optarg)
               setOption("exploration_technique", optarg);
            else
               setOption("exploration_technique", dse_hls::BINDING);
            break;
         case 'Q':
            if (optarg)
               setOption("to_normalize", optarg);
            else
               setOption("to_normalize", 1);
            break;
         case 'P':  // mixed high level synthesis fitness function
            setOption("fitness_function", optarg);
            break;
         case 'A':
            setOption("seed", optarg);
            break;
         case 'B':
            setOption("run", optarg);
            break;
         case 'E':
            setOption("population", optarg);
            break;
         case OPT_GENERATION:
            setOption("GA_generation", optarg);
            break;
         case 'F':
            setOption("fitness_inheritance_rate", optarg);
            break;
         case 'G':
            setOption("inheritance_mode", optarg);
            break;
         case OPT_MAX_INHERITANCE:
            setOption("max_for_inheritance", optarg);
            break;
         case OPT_MIN_INHERITANCE:
            setOption("min_for_inheritance", optarg);
            break;
         case 'M':
            setOption("distance_rate", optarg);
            break;
         case 'N':
            setOption("weighting_function", optarg);
            break;
         case OPT_TIME_WEIGHT:
            setOption("time_weight", optarg);
            break;
         case OPT_AREA_WEIGHT:
            setOption("area_weight", optarg);
            break;
#endif
#endif
         ///scheduling options
         case OPT_FIXED_SCHED:
            if (scheduling_set_p)
               throw "BadParameters: only one scheduler can be specified";
            scheduling_set_p = true;
            setOption(OPT_scheduling_algorithm, scheduling::FIXED);
            if (optarg)
               setOption("fixed_scheduling_file", optarg);
            break;
         case OPT_LIST_BASED: // enable list based scheduling
            if (scheduling_set_p)
               throw "BadParameters: only one scheduler can be specified";
            scheduling_set_p = true;
            setOption(OPT_scheduling_algorithm, scheduling::LIST_BASED);
            if (optarg)
               setOption(OPT_scheduling_priority, optarg);
            break;
#if HAVE_EXPERIMENTAL
#if HAVE_ILP_BUILT
         case 'i': // enable ILP scheduling
            if (scheduling_set_p)
               throw "BadParameters: only one scheduler can be specified";
            scheduling_set_p = true;
            setOption(OPT_scheduling_algorithm, scheduling::ILP);
            setOption(OPT_chaining, true);/// ILP formulation tries to chain the operations
            if (optarg)
            {
               if (optarg[0] == 'G')
                  setOption(OPT_ilp_solver, meilp_solver::GLPK);
#if HAVE_COIN_OR
               else if (optarg[0] == 'C')
                  setOption(OPT_ilp_solver, meilp_solver::COIN_OR);
#endif
#if HAVE_LP_SOLVE
               else if (optarg[0] == 'L')
                  setOption(OPT_ilp_solver, meilp_solver::LP_SOLVE);
#endif
               else
                  throw "BadParameters: not recognized ilp solver";
            }
            break;
         case 'I': // enable ILP scheduling, new-formulation
            if (scheduling_set_p)
               throw "BadParameters: only one scheduler can be specified";
            scheduling_set_p = true;
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Set new ilp formulation");
            setOption(OPT_scheduling_algorithm, scheduling::ILP_NEW_FORM);
            if (optarg)
            {
               if (optarg[0] == 'G')
                  setOption(OPT_ilp_solver, meilp_solver::GLPK);
#if HAVE_COIN_OR
               else if (optarg[0] == 'C')
                  setOption(OPT_ilp_solver, meilp_solver::COIN_OR);
#endif
#if HAVE_LP_SOLVE
               else if (optarg[0] == 'L')
                  setOption(OPT_ilp_solver, meilp_solver::LP_SOLVE);
#endif
               else
                  throw "BadParameters: not recognized ilp solver";
            }
            break;
         case 'J': // enable ILP scheduling, loop pipelining
            if (scheduling_set_p)
               throw "BadParameters: only one scheduler can be specified";
            scheduling_set_p = true;
            setOption(OPT_scheduling_algorithm, scheduling::ILP_LOOP);
            if (optarg)
            {
               if (optarg[0] == 'G')
                  setOption(OPT_ilp_solver, meilp_solver::GLPK);
#if HAVE_COIN_OR
               else if (optarg[0] == 'C')
                  setOption(OPT_ilp_solver, meilp_solver::COIN_OR);
#endif
#if HAVE_LP_SOLVE
               else if (optarg[0] == 'L')
                  setOption(OPT_ilp_solver, meilp_solver::LP_SOLVE);
#endif
               else
                  throw "BadParameters: not recognized ilp solver";
            }
            break;
         case 'a': // enable SILP scheduling
            if (scheduling_set_p)
               throw "BadParameters: only one scheduler can be specified";
            scheduling_set_p = true;
            setOption(OPT_scheduling_algorithm, scheduling::SILP);
            if (optarg)
            {
               if (optarg[0] == 'G')
                  setOption(OPT_ilp_solver, meilp_solver::GLPK);
#if HAVE_COIN_OR
               else if (optarg[0] == 'C')
                  setOption(OPT_ilp_solver, meilp_solver::COIN_OR);
#endif
#if HAVE_LP_SOLVE
               else if (optarg[0] == 'L')
                  setOption(OPT_ilp_solver, meilp_solver::LP_SOLVE);
#endif
               else
                  throw "BadParameters: not recognized silp solver";
            }
            break;
#endif
         case 'S': // enable scheduling with speculative computation
            setOption(OPT_speculative, true);
            break;
#endif
         case OPT_STG:
            setOption(OPT_stg, true);
            if (optarg)
               setOption(OPT_stg_algorithm, optarg);
            break;
         ///binding options
         //register
         case OPT_REGISTER_ALLOCATION: // enable register allocation
         {
            if (std::string(optarg) == "COLORING")
               setOption(OPT_register_allocation_algorithm, reg_binding_creator::COLORING);
            else if(std::string(optarg) == "CHORDAL_COLORING")
               setOption(OPT_register_allocation_algorithm, reg_binding_creator::CHORDAL_COLORING);
#if HAVE_EXPERIMENTAL
            else if(std::string(optarg) == "LEFT_EDGE")
               setOption(OPT_register_allocation_algorithm, reg_binding_creator::LEFT_EDGE);
            else if(std::string(optarg) == "CYCLIC_ALLOCATION")
               setOption(OPT_register_allocation_algorithm, reg_binding_creator::CYCLIC_ALLOCATION);
            else if(std::string(optarg) == "K_COFAMILY")
               setOption(OPT_register_allocation_algorithm, reg_binding_creator::K_COFAMILY);
#endif
            else if(std::string(optarg) == "WEIGHTED_COLORING")
               setOption(OPT_register_allocation_algorithm, reg_binding_creator::WEIGHTED_COLORING);
            else if(std::string(optarg) == "BIPARTITE_MATCHING")
               setOption(OPT_register_allocation_algorithm, reg_binding_creator::BIPARTITE_MATCHING);
            else if(std::string(optarg) == "TTT_CLIQUE_COVERING")
               setOption(OPT_register_allocation_algorithm, reg_binding_creator::TTT_CLIQUE_COVERING);
            else if(std::string(optarg) == "UNIQUE_BINDING")
               setOption(OPT_register_allocation_algorithm, reg_binding_creator::UNIQUE_BINDING);
            else
               throw "BadParameters: register allocation not correctly specified";

            break;
         }
         //interconnection
         case 'C':
            setOption(OPT_datapath_interconnection, true);
            if (std::string(optarg) == "M")
               setOption(OPT_datapath_interconnection_style, conn_binding_creator::MUX);
            else
               throw "BadParameters: interconnection binding not correctly specified";
            break;
         ///datapath options
         case OPT_DATAPATH:
            setOption(OPT_datapath, false);
            break;
         ///controller options
         case OPT_CONTROLLER:
            setOption(OPT_controller, false);
            break;
#if HAVE_EXPERIMENTAL
         case OPT_PARALLEL_CONTROLLER:
            setOption(OPT_controller_architecture, controller_creator::PARALLEL_CONTROLLER);
            break;
#endif
         ///target options
         case 'b':
            setOption(OPT_target_device_file, optarg);
            break;
#if HAVE_EXPERIMENTAL
         ///evaluation
         case OPT_EVALUTION:
            setOption(OPT_evaluation, true);
            if (!optarg || std::string(optarg) == "LINEAR")
               setOption(OPT_evaluation_method, objective_evaluator::LINEAR);
            if (std::string(optarg) == "EXACT")
               setOption(OPT_evaluation_method, objective_evaluator::EXACT);
            if (std::string(optarg) == "STATISTICAL")
               setOption(OPT_evaluation_method, objective_evaluator::STATISTICAL);
            break;
         case OPT_EVALUTION_MODEL:
            setOption("evaluation_xml_model", optarg);
            break;
         case OPT_AREA_RESULT:
            setOption("area_result", true);
            if (optarg)
               setOption("area_final_variable", optarg);
            break;
#endif
         case OPT_DEVICE_NAME:
            {
               std::string tmp_string = optarg;
               std::vector<std::string> values = convert_string_to_vector<std::string>(tmp_string, std::string(","));
               if(values.size() == 2)
               {
                  setOption(OPT_device_string, values[1]);
               }
               else if (values.size() == 3)
               {
                  setOption("device_name", values[0]);
                  setOption("device_speed", values[1]);
                  setOption("device_package", values[2]);
               }
               else
                   THROW_ERROR("Malformed device: " + tmp_string);
               break;
            }
         case OPT_PERIOD_CLOCK:
            setOption(OPT_clock_period, optarg);
            break;
         case OPT_POWER_OPTIMIZATION:
            setOption("power_optimization", true);
            break;
         case OPT_ENABLE_IOB:
            setOption("connect_iob", false);
            break;
         ///output options
         case 'w':
            if (std::string(optarg) == "V")
               setOption(OPT_writer_language, language_writer::VERILOG);
#if HAVE_EXPERIMENTAL
            else if (std::string(optarg) == "S")
               setOption(OPT_writer_language, language_writer::SYSTEMC);
            else if (std::string(optarg) == "H")
               setOption(OPT_writer_language, language_writer::VHDL);
#endif
            else
               throw "BadParameters: backend language not correctly specified";
            break;
         case OPT_DYNAMIC_GENERATORS_DIR:
            setOption("dynamic_generators_dir", optarg);
            break;
         case OPT_PRETTY_PRINT:
            setOption(OPT_pretty_print, optarg);
            break;
         case OPT_INSERT_MEMORY_PROFILE:
            setOption(OPT_insert_memory_profile, true);
            break;          
         case OPT_TESTBENCH:
            setOption(OPT_generate_testbench, true);
            setOption("tb_input_xml", optarg);
            break;
         case OPT_GENERATE_VCD:
            setOption(OPT_generate_vcd, true);
            break;
#if HAVE_CUDD
         case OPT_LOGICAL_OPTIMIZATION:
            if (optarg)
               setOption("logical_optimization", optarg);
            break;
#endif
         case 0:
         {
            if (strcmp(long_options[option_index].name, "module-binding") == 0)
            {
               setOption(OPT_fu_binding, true);
               // set the module binding approach to be executed
               if (!optarg)
                  setOption(OPT_fu_binding_algorithm, fu_binding_creator::CDFC_WEIGHTED_COLORING);
               else
               {
                  if (std::string(optarg) == "TTT_FAST")
                     setOption(OPT_fu_binding_algorithm, fu_binding_creator::CDFC_TTT_FAST);
                  else if (std::string(optarg) == "TTT_FULL")
                     setOption(OPT_fu_binding_algorithm, fu_binding_creator::CDFC_TTT_FULL);
                  else if (std::string(optarg) == "COLORING")
                     setOption(OPT_fu_binding_algorithm, fu_binding_creator::CDFC_COLORING);
                  else if (std::string(optarg) == "WEIGHTED_COLORING")
                     setOption(OPT_fu_binding_algorithm, fu_binding_creator::CDFC_WEIGHTED_COLORING);
                  else if (std::string(optarg) == "BIPARTITE_MATCHING")
                     setOption(OPT_fu_binding_algorithm, fu_binding_creator::CDFC_BIPARTITE_MATCHING);
                  else if (std::string(optarg) == "UNIQUE")
                     setOption(OPT_fu_binding_algorithm, fu_binding_creator::UNIQUE);
                  else
                     throw "BadParameters: module binding option not correctly specified";
               }
               break;
            }
            if (strcmp(long_options[option_index].name, "perform-synthesis") == 0)
            {
               setOption(OPT_perform_synthesis, true);
               break;
            }
            if (strcmp(long_options[option_index].name, "memory-allocation") == 0)
            {
               // set the module binding approach to be executed
               if (!optarg)
                  setOption(OPT_memory_allocation_algorithm, memory_allocation::GSS);
               else
               {
                  if (std::string(optarg) == "GSS")
                     setOption(OPT_memory_allocation_algorithm, memory_allocation::GSS);
                  else if (std::string(optarg) == "ALL_BRAM")
                     setOption(OPT_memory_allocation_algorithm, memory_allocation::ALL_BRAM);
                  else if (std::string(optarg) == "NO_BRAM")
                     setOption(OPT_memory_allocation_algorithm, memory_allocation::NO_BRAM);
                  else
                     throw "BadParameters: memory allocation option not correctly specified";
               }
               break;
            }
#if HAVE_EXPERIMENTAL
            if (strcmp(long_options[option_index].name, "timing-simulation") == 0)
            {
               setOption(OPT_timing_simulation, true);
            }
#if (HAVE_XILINX || HAVE_ALTERA) && HAVE_MODELSIM
            if (strcmp(long_options[option_index].name, "vcd-debug") == 0)
            {
               setOption("vcd-debug", true);
            }
#endif
            //front-end analysis option
            if (strcmp(long_options[option_index].name, "pdg-reduction") == 0)
            {
               //Set the default
               if(optarg)
                  setOption("pdg-reduction", optarg);
            }
            if (strcmp(long_options[option_index].name, "explore-mux") == 0)
            {
               setOption("explore-mux", true);
               if (optarg) setOption("worst_case_delay", optarg);
            }
            if (strcmp(long_options[option_index].name, "explore-fu-reg") == 0)
            {
               setOption("explore-fu-reg", true);
               setOption("explore-fu-reg-param", optarg);
            }
#endif
            if (strcmp(long_options[option_index].name, "assert-debug") == 0)
            {
               setOption(OPT_assert_debug, true);
            }
            if (strcmp(long_options[option_index].name, "objective") == 0)
            {
               setOption(OPT_evaluation, true);
               setOption(OPT_evaluation_objectives, optarg);
            }
            if (strcmp(long_options[option_index].name, "circuit-dbg") == 0)
            {
               setOption(OPT_circuit_debug_level, optarg);
            }
            ///scheduling options
            if (strcmp(long_options[option_index].name, "no-chaining") == 0)
            {
               setOption(OPT_chaining, false);
            }
            ///target options
            if (strcmp(long_options[option_index].name, "target-device-type") == 0)
            {
               if (std::string(optarg) == "FPGA")
                  setOption(OPT_target_device_type, target_device::FPGA);
               else if (std::string(optarg) == "ASIC")
                  setOption(OPT_target_device_type, target_device::IC);
               else
                  THROW_ERROR("Not support target device type: \"" + std::string(optarg) + "\"");
            }
#if HAVE_EXPERIMENTAL
            if (strcmp(long_options[option_index].name, "generate-interface") == 0)
            {
               setOption(OPT_interface, true);
               if (std::string(optarg) == "FSL")
               {
                  setOption(OPT_interface_type, module_interface::FSL);
               }
               else
               {
                  THROW_ERROR("Not supported interface: " + std::string(optarg));
               }
            }
            if (strcmp(long_options[option_index].name, "edk-config") == 0)
            {
               setOption("edk_wrapper", true);
               setOption("edk_config_file", std::string(optarg));
            }
#endif
            if (strcmp(long_options[option_index].name, "simulate") == 0)
            {
               setOption(OPT_generate_testbench, true);
               if (optarg) setOption("tb_input_xml", optarg);
               else setOption("tb_input_xml", "test.xml");
               setOption(OPT_simulate, true);
            }
            if (strcmp(long_options[option_index].name, "simulator") == 0)
            {
               setOption(OPT_simulator, std::string(optarg));
            }
            break;
         }
         ///other options
         default:
         {
            bool exit_success=false;
            bool res = ManageGccOptions(next_option, optarg);
            if (res)
               res = ManageDefaultOptions(next_option, optarg, exit_success);
            if(exit_success)
               return EXIT_SUCCESS;
            if(res)
            {
               return PARAMETER_NOTPARSED;
            }
         }
      }
   }

#if HAVE_EXPERIMENTAL
   if (isOption(OPT_gcc_write_xml))
   {
      const std::string filenameXML = getOption<std::string>(OPT_gcc_write_xml);
      write_xml_configuration_file(filenameXML);
      PRINT_MSG("Configuration saved into file \"" + filenameXML + "\"");
      return EXIT_SUCCESS;
   }
#endif

   std::string cat_args;

   for(int i = 0; i < argc; i++)
   {
      cat_args += std::string(argv[i]) + " ";
   }
   setOption(OPT_cat_args, cat_args);

   if (getOption<int>(OPT_output_level) >= OUTPUT_LEVEL_MINIMUM)
   {
      std::cerr << " ==  Bambu executed with: ";
      std::cerr << cat_args;
      std::cerr << std::endl << std::endl;
   }

   bool is_constraints_file = true;
   bool is_tech_file = true;
   if (not getOption<bool>(OPT_use_raw))
   {
      while (optind < argc)
      {
         std::string extension_str = boost::filesystem::extension(boost::filesystem::path(argv[optind]));
         if(extension_str == ".xml" || extension_str == ".XML")
         {
            if(is_constraints_file)
            {
               is_constraints_file = false;
               setOption("constraintsfile", argv[optind]);
            }
            else if(is_tech_file)
            {
               is_tech_file = false;
               setOption("techfile", argv[optind]);
            }
            else
               throw "BadParameters: only two configuration files are considered";
         }
         else
         {
            std::string input_file;
            if(isOption(OPT_input_file))
            {
               input_file = getOption<std::string>(OPT_input_file) + STR_CST_string_separator;
            }
            setOption(OPT_input_file, input_file + argv[optind]);
         }
         optind++;
      }
   }
   else
   {
      if (optind < argc)                    // the first non-option param is input file
      {
         std::string input_file;
         if(isOption(OPT_input_file))
         {
            input_file = getOption<std::string>(OPT_input_file) + STR_CST_string_separator;
         }
         setOption(OPT_input_file, input_file + argv[optind]);
      }
      optind++;
      if (optind < argc)                    // the second non-option param is constraint file
      {
         setOption("constraintsfile", argv[optind]);
      }
      optind++;
      if (optind < argc)                    // the third non-option param is tech file
      {
         setOption("techfile", argv[optind]);
      }

   }

   CheckParameters();

   return PARAMETER_PARSED;

}

void BambuParameter::CheckParameters()
{
   if (not isOption(OPT_device_string))
   {
      std::string device_string = getOption<std::string>("device_name") + getOption<std::string>("device_speed")  + getOption<std::string>("device_package");
      setOption(OPT_device_string, device_string);
   }
#if HAVE_EXPERIMENTAL
#if HAVE_BEAGLE
   if (getOption<int>("exploration_technique") == dse_hls::PRIORITY and getOption<int>("to_normalize"))
      setOption("to_normalize", false);
#endif
#endif
   if (getOption<bool>(OPT_datapath) || getOption<bool>(OPT_controller))
   {
      setOption(OPT_register_allocation, true);
      setOption(OPT_fu_binding, true);
      setOption(OPT_datapath_interconnection, true);
   }

#if HAVE_CUDD
   if(!isOption("logical_optimization"))
      setOption("logical_optimization", 0);
#endif

   ///the finite state machine has to be created if the register allocation has to be performed. If the datapath or controller are
   ///requested, the register allocation has to be performed for sure, so this condition is sufficient to allow the finite state
   ///machine to be computed
   if (getOption<bool>(OPT_register_allocation))
   {
      setOption(OPT_stg, true);
   }

#if HAVE_EXPERIMENTAL
#if HAVE_BEAGLE
   if (isOption("dse_algorithm") and getOption<unsigned int>("dse_algorithm") == dse_hls::ACO)
   {
      if (!isOption("aco_ant_num"))
         setOption("aco_ant_num", 5);
      if (!isOption("aco_generations"))
         setOption("aco_generations", 10);
   }
#endif
#endif
   if (isOption(OPT_generate_vcd) and getOption<bool>(OPT_generate_vcd))
   {
      if (!isOption(OPT_generate_testbench))
         THROW_ERROR("Testbench generation required. (--generate-tb or --simulate undeclared).");
      setOption(OPT_assert_debug, true);
   }
   setOption(OPT_chaining_algorithm, chaining::STG_CHAINING);
#if HAVE_EXPERIMENTAL
   if (getOption<unsigned int>(OPT_controller_architecture) == controller_creator::PARALLEL_CONTROLLER)
   {
      //setOption(OPT_liveness_algorithm, liveness_computer::EPDG_DATAFLOW);
      setOption(OPT_liveness_algorithm, liveness_computer::DUMMY_LIVENESS);

      setOption(OPT_storage_value_insertion_algorithm, storage_value_insertion::VALUES);
      setOption(OPT_chaining_algorithm, chaining::CDFG_CHAINING);
   }
   if (getOption<unsigned int>(OPT_evaluation_method) == objective_evaluator::EXACT)
   {
      std::string objective_string = getOption<std::string>(OPT_evaluation_objectives);
      std::vector<std::string> objective_vector = convert_string_to_vector<std::string>(objective_string, ",");
      if (std::find(objective_vector.begin(), objective_vector.end(), "TIME") != objective_vector.end() && !getOption<bool>(OPT_generate_testbench))
      {
         setOption(OPT_simulate, true);
         setOption(OPT_generate_testbench, true);
         setOption("tb_input_xml", "test.xml");
         if (boost::filesystem::exists("test.xml")) boost::filesystem::remove("test.xml");
      }
   }
#endif
   if (getOption<int>(OPT_debug_level)>=DEBUG_LEVEL_VERY_PEDANTIC)
   {
      setOption(OPT_assert_debug, true);
   }

}

void BambuParameter::SetDefaults()
{
   // ---------- general options ----------- //
   /// Revision
   if(std::string(REVISION) == std::string(""))
      setOption(OPT_revision, "unknown");
   else
      setOption(OPT_revision, "r" + std::string(REVISION));
   setOption(OPT_output_directory, "HLS_output");
   setOption(OPT_simulation_output, "results.txt");
   /// Debugging level
   setOption(OPT_output_level, OUTPUT_LEVEL_MINIMUM);
   setOption(OPT_debug_level, DEBUG_LEVEL_NONE);
   setOption(OPT_circuit_debug_level, DEBUG_LEVEL_NONE);
   setOption(OPT_print_dot, false);
   /// maximum execution time: 0 means no time limit
   setOption(OPT_ilp_max_time, 0);
   setOption(OPT_use_raw, false);

#if HAVE_MAPPING_BUILT
   setOption(OPT_driving_component_type, "ARM");
#endif
   /// ---------- frontend analysis ----------//
   setOption(OPT_parse_pragma, false);
#if HAVE_FROM_RTL_BUILT
   setOption(OPT_use_rtl, false);
#endif

   //setOption("pdg-reduction", false);

#if HAVE_EXPERIMENTAL
   setOption("compute_unrolling_degree", false);
   setOption("insert_verification_operation",false);
#endif
   setOption(STR_OPT_frontend_statistics, false);

   /// ---------- HLS process options ----------- //
   setOption(OPT_hls_flow, hls_flow::STANDARD);

   /// ---------- HLS specification reference ----------- //
   setOption(OPT_generate_testbench, false);
   setOption(OPT_simulate, false);
   setOption(OPT_chaining, true);

   // High-level synthesis contraints dump -- //
   setOption("dumpConstraints", false);
   setOption("dumpConstraints_file", "Constraints.XML");

   // -- Scheduling -- //
   /// Scheduling algorithm (default is list based one)
   setOption(OPT_scheduling, true);
   setOption(OPT_scheduling_algorithm, scheduling::LIST_BASED);
   setOption(OPT_scheduling_priority, parametric_list_based::dynamic_mobility);
   /// ilp solver
#if HAVE_ILP_BUILT
   setOption(OPT_ilp_solver, meilp_solver::GLPK);
#endif
   /// speculative execution flag
   setOption(OPT_speculative, false);

   // -- Module binding -- //
   /// flag to check if module binding has to be performed
   setOption(OPT_fu_binding, false);
   /// module binding algorithm
   setOption(OPT_fu_binding_algorithm, fu_binding_creator::CDFC_WEIGHTED_COLORING);

   // -- Finite state machine -- //
   /// flag to check if finite state machine has to be created
   setOption(OPT_stg, false);
   /// state transition graph creation algorithm
   setOption(OPT_stg_algorithm, STG_creator::BB_BASED);

   // -- Dataflow analysis -- //
   setOption(OPT_liveness_algorithm, liveness_computer::FSM_NI_SSA_LIVENESS);

   // -- Register allocation -- //
   /// flag to check if register allocation has to be performed
   setOption(OPT_register_allocation, false);
   /// register allocation algorithm
   setOption(OPT_register_allocation_algorithm, reg_binding_creator::WEIGHTED_COLORING);
   /// storage value insertion algorithm
   setOption(OPT_storage_value_insertion_algorithm, storage_value_insertion::VALUES);
   
   // -- Memory allocation //
   setOption(OPT_memory_allocation_algorithm, memory_allocation::GSS);

   // -- Datapath -- //
   setOption(OPT_datapath, true);
   /// flag to check if interconnection binding has to be performed
   setOption(OPT_datapath_interconnection, false);
   /// interconnection architecture
   setOption(OPT_datapath_interconnection_style, conn_binding_creator::MUX);
   /// Datapath interconnection architecture
   setOption(OPT_datapath_interconnection_algorithm, conn_binding_creator::MUX_INTERCONNECTION_BINDING);
   /// flag to check if datapath structural representation has to be created
   /// Output file name for datapath
   setOption(OPT_datapath_file, "datapath");
   /// Datapath architecture
   setOption(OPT_datapath_architecture, datapath_creator::CLASSIC);

   // -- Controller -- //
   /// flag to check if controller structural representation has to be created
   setOption(OPT_controller, true);
   /// target architecture for the controller
   setOption(OPT_controller_architecture, controller_creator::FSM_CONTROLLER);
   /// Output file name for controller
   setOption(OPT_controller_file, "controller");

   // -- top entity -- //
   /// Output file name for top entity
   setOption(OPT_top_file, "top");

   /// backend HDL
   setOption(OPT_writer_language, language_writer::VERILOG);
   setOption("dynamic_generators_dir", BAMBU_DYNAMIC_GENERATOR_DIR);

#if HAVE_EXPERIMENTAL
   // -- Module Interfaces -- //
   setOption(OPT_interface, false);
   // -- Module Characterization -- //
   setOption(OPT_evaluation, false);
   setOption(OPT_evaluation_method, objective_evaluator::LINEAR);
   setOption(OPT_evaluation_objectives, "AREA,TIME");
   setOption("evaluation_statistic", false);
   setOption("evaluation_reduced", false);
   setOption("evaluation_output", "evaluation.txt");
   setOption("area_result", true);
   setOption("area_final_variable", "total_area");
#endif

   // -- Module Synthesis -- //
   setOption(OPT_perform_synthesis, false);
   setOption(OPT_rtl, true); ///the resulting specification will be a RTL description
   setOption(OPT_simulator, "MODELSIM");
   setOption("device_name", "xc5vlx50");
   setOption("device_speed", "-3");
   setOption("device_package", "ff1153");
   setOption(OPT_timing_simulation, false);
   setOption(OPT_target_device_type, target_device::FPGA);
#if HAVE_EXPERIMENTAL
   setOption("vcd-debug", false);
   setOption("edk_wrapper", false);
#endif
   setOption("connect_iob", true);

#if (HAVE_EXPERIMENTAL && HAVE_BEAGLE)
   // -- Parameters for the design space exploration -- //
   setOption("exploration_technique", dse_hls::BINDING);
   setOption("to_normalize", 0);
   setOption("seed", 0);
   setOption("run", 1);
   setOption("deme_size", 1);
   setOption("population", 50);
   setOption("GA_generation", 10);
   setOption("max_evaluations", 0);
   setOption("fitness_function", objective_evaluator::LINEAR);
   setOption("fitness_inheritance_rate", 0.0);
   /// if the parameter is even, the cache is analysed. Otherwise only the latest population
   setOption("inheritance_mode", 0);
   setOption("max_for_inheritance", 50);
   setOption("min_for_inheritance", 1);
   setOption("distance_rate", 0.20);
   setOption("weighting_function", FitnessFunction::QUADRATIC);
   setOption("time_weight", 0.5);
   setOption("area_weight", 0.5);
   setOption("remove_duplicated", true);
   setOption("lower", 0.05);
   setOption("upper", 0.05);
   setOption("prob_dupl", 0.50);
#endif

   // -- GCC options -- //
   setOption(OPT_default_compiler, CT_I386_GCC);
   setOption(OPT_compatible_compilers, CT_I386_GCC | CT_ARM_GCC | CT_SPARC_GCC);
   setOption(OPT_without_transformation, true);
   setOption(OPT_compute_size_of, true);
   setOption(OPT_precision, 3);
   setOption(OPT_gcc_opt_level, 0);
   setOption(OPT_gcc_config, false);
   setOption(OPT_gcc_costs, false);
   setOption(OPT_gcc_standard, "gnu89");
   setOption(OPT_gcc_compiler_target, CT_I386_GCC);
   setOption(OPT_gcc_optimization_set, OS_OBAMBU);
   setOption(OPT_gcc_include_sysdir, false);
   setOption(OPT_model_costs, false);

   std::string defines;
   if(isOption(OPT_gcc_defines))
      defines = getOption<std::string>(OPT_gcc_defines) + STR_CST_string_separator;
   defines += "__BAMBU__";
   setOption(OPT_gcc_defines, defines);
}
