/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file fsm_controller.cpp
 * @brief Starting from the FSM graph, it builds a state transition table
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 * $Locker:  $
 * $State: Exp $
 *
*/

#include "fsm_controller.hpp"

#include "hls.hpp"
#include "allocation.hpp"
#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "behavioral_helper.hpp"

#include "BambuParameter.hpp"

#include "conn_binding.hpp"
#include "schedule.hpp"
#include "fu_binding.hpp"
#include "commandport_obj.hpp"

#include "StateTransitionGraph.hpp"

#include <iostream>
#include <deque>
#include "exceptions.hpp"

#include "tree_manager.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"

#include "structural_objects.hpp"
#include "structural_manager.hpp"

#include "dbgPrintHelper.hpp"
#include "technology_builtin.hpp"
#include "technology_manager.hpp"
#include "time_model.hpp"

#include "op_node_info.hpp"
#include "op_edge_info.hpp"

fsm_controller::fsm_controller(const ParameterRef _Param, const hlsRef _HLS) :
   controller_creator(FSM_CONTROLLER, _Param, _HLS)
{

}

fsm_controller::~fsm_controller()
{

}

void fsm_controller::do_exec()
{
   THROW_ASSERT(HLS->STG, "State transition graph not created");

   /// top circuit creation
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "FSM based controller creation");
   
   const std::string function_name = HLS->FB->get_helper()->get_function_name();
   /// main circuit type
   structural_type_descriptorRef module_type = structural_type_descriptorRef(new structural_type_descriptor("controller_"+function_name));

   SM->set_top_info("Controller_i", module_type);
   structural_objectRef circuit = this->SM->get_circ();
   // Now the top circuit is created, just as an empty box. <circuit> is a reference to the structural object that
   // will contain all the circuit components

   ///Set some descriptions and legal stuff
   GetPointer<module>(circuit)->set_description("FSM based controller descrition for " + function_name);
   GetPointer<module>(circuit)->set_copyright("Copyright (C) 2004-2011 Politecnico di Milano");
   GetPointer<module>(circuit)->set_authors("module automatically generated by bambu");
   GetPointer<module>(circuit)->set_license("PANDA_GPLv3");

   
   //Add clock, reset, done and command ports
   this->add_common_ports(circuit);

   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Creating state machine representations...");
   std::string state_representation;
   this->create_state_machine(state_representation);

   SM->add_NP_functionality(circuit, NP_functionality::FSM, state_representation);
   
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Machine encoding");
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, state_representation);
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "****");

   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Circuit created without errors!");
   return;
}

std::string fsm_controller::get_kind_text() const
{
   return "FSM-creation";
}

static
std::string input_vector_to_string(const std::vector<long long int>& to_be_printed, bool with_comma)
{
   std::string output;
   for(unsigned int i = 0; i < to_be_printed.size(); i++)
   {
      output += (to_be_printed[i] == default_COND ? "-" : boost::lexical_cast<std::string>(to_be_printed[i]));
      if(i!=(to_be_printed.size()-1) && with_comma)
         output += ",";
   }
   return output;
}

void fsm_controller::create_state_machine(std::string &parse)
{
   const graph *stg = HLS->STG->get_stg();

   vertex entry = HLS->STG->get_entry_state();
   THROW_ASSERT(boost::out_degree(entry, *stg) == 1, "Non deterministic initial state");
   OutEdgeIterator oe,oend;
   tie(oe,oend) = boost::out_edges(entry, *stg);
   /// Getting first state (initial one). It will be also first state for resetting
   vertex first_state = boost::target(*oe, *stg);
   /// adding reset state to machine encoding
   parse += GET_STATE_NAME(stg, first_state) + " " + RESET_PORT_NAME + " " + START_PORT_NAME + " " + CLOCK_PORT_NAME + "; ";

   const std::map<unsigned int, std::map<std::pair<generic_objRef, unsigned int>, generic_objRef> >& selectors = HLS->Rconn->get_selectors();

   std::map<vertex, std::vector<long long int> > present_state;
   std::set<unsigned int> unbounded_ports;
   
   /// analysis for each state to compute the default output
   std::list<vertex> working_list;
   boost::topological_sort(*(HLS->STG->get_astg()), std::front_inserter(working_list));
   for(std::list<vertex>::iterator v = working_list.begin(); v != working_list.end(); v++)
   {
      present_state[*v] = std::vector<long long int>(out_num, 0);
      for(std::map<std::pair<generic_objRef, unsigned int>, generic_objRef>::const_iterator s = selectors.find(conn_binding::IN)->second.begin(); s != selectors.find(conn_binding::IN)->second.end(); s++)
      {
#ifndef NDEBUG
         std::map<vertex, std::set<vertex> > activations_check;
#endif
         const std::set<std::pair<vertex, vertex> >& activations = GetPointer<commandport_obj>(s->second)->get_activations();
         for(std::set<std::pair<vertex, vertex> >::const_iterator a = activations.begin(); a != activations.end(); a++)
         {
#ifndef NDEBUG
            if(activations_check.find(a->first) != activations_check.end())
            {
               THROW_ASSERT(!activations_check.find(a->first)->second.empty(), "empty set not expected here");
               if(activations_check.find(a->first)->second.find(a->second) == activations_check.find(a->first)->second.end())
               {
                  if(a->second == NULL_VERTEX)
                     THROW_ERROR("non compatible transitions added");
                  else if(activations_check.find(a->first)->second.find(NULL_VERTEX) != activations_check.find(a->first)->second.end())
                     THROW_ERROR("non compatible transitions added");
                  else
                     activations_check[a->first].insert(a->second);
               }
               else
               {
                  THROW_ERROR("activation already added");
               }
            }
            else
               activations_check[a->first].insert(a->second);
#endif
            if (a->first == *v)
            {
               present_state[*v][out_ports[s->second]] = 1;
               //std::cerr << "port " << s->second->get_string() << " " << out_ports[s->second] << " enabled from state " << HLS->STG->get_state_name(a->first) << " to " << HLS->STG->get_state_name(a->second) << std::endl;
            }
         }
      }

      const graph *data = HLS->FB->get_graph(function_behavior::CFG);

      const std::list<vertex>& operations = GET_OPERATION_LIST(HLS->STG->get_astg(), *v);
      for (std::list<vertex>::const_iterator op = operations.begin(); op != operations.end(); op++)
      {
         technology_nodeRef tn = HLS->ALL->get_fu(HLS->Rfu->get_assign(*op));
         technology_nodeRef op_tn = GetPointer<functional_unit>(tn)->get_operation(GET_OP(data, *op));
         THROW_ASSERT(GetPointer<operation>(op_tn)->time_m, "Time model not available for operation: " + GET_NAME(data, *op));
         if (((GET_TYPE(data, *op) & TYPE_EXTERNAL) || !GetPointer<operation>(op_tn)->is_bounded()) && !GET_DUMMY_STATE(stg, *v))
         {
            unsigned int unbounded_port = out_ports[HLS->Rconn->bind_selector_port(conn_binding::IN, commandport_obj::UNBOUNDED, *op, data)];
            unbounded_ports.insert(unbounded_port);
            present_state[*v][unbounded_port] = 1;
         }
      }

   }

   parse += "\n";

   const tree_managerRef TreeM = HLS->AppM->get_tree_manager();
   for(std::list<vertex>::iterator v = working_list.begin(); v != working_list.end(); v++)
   {
      if (HLS->STG->get_entry_state() == *v or HLS->STG->get_exit_state() == *v) continue;

      parse += GET_STATE_NAME(stg, *v) + " 0" + input_vector_to_string(present_state[*v],0);

      std::list<EdgeDescriptor> sorted;
      EdgeDescriptor default_edge;
      bool found_default=false;

      for(tie(oe,oend) = boost::out_edges(*v, *stg); oe != oend; oe++)
      {
         if(!found_default)
         {
            const std::set<std::pair<vertex,unsigned int> >& cond = GET_TRANSITION_CONDITION(stg, *oe);
            for(std::set<std::pair<vertex,unsigned int> >::const_iterator cond_it = cond.begin(); cond_it != cond.end(); cond_it++)
            {
               if(cond_it->second == default_COND)
               {
                  found_default = true;
                  default_edge = *oe;
                  break;
               }
            }
            if(!found_default)
               sorted.push_back(*oe);
         }
         else
            sorted.push_back(*oe);
      }
      if(found_default)
         sorted.push_back(default_edge);

      unsigned int j = 0;
      std::list<EdgeDescriptor>::const_iterator e_it_end = sorted.end();
      for(std::list<EdgeDescriptor>::const_iterator e_it = sorted.begin(); e_it != e_it_end; e_it++, j++)
      {
         std::vector<std::string> in(in_num, "-");

         const std::set<std::pair<vertex,unsigned int> >& cond = GET_TRANSITION_CONDITION(stg, *e_it);
         for(std::set<std::pair<vertex,unsigned int> >::const_iterator cond_it = cond.begin(); cond_it != cond.end(); cond_it++)
         {
            //std::cerr << cond_it->second << std::endl;
            std::string value;
            if(cond_it->second == T_COND)
            {
               THROW_ASSERT(in[cond_ports[cond_it->first]]=="-", "two different values for the same condition port");
               value = "1";
            }
            else if (cond_it->second == F_COND)
            {
               THROW_ASSERT(in[cond_ports[cond_it->first]]=="-", "two different values for the same condition port");
               value = "0";
            }
            else if(cond_it->second == default_COND)
            {
               value = in[cond_ports[cond_it->first]];
               if(value=="-")
                  value = STR(default_COND);
               else
                  value += "|" + STR(default_COND);
            }
            else
            {
               value = in[cond_ports[cond_it->first]];
               if(value=="-")
                  value = get_guard_value(TreeM, cond_it->second);
               else
                  value += "|" + get_guard_value(TreeM, cond_it->second);
            }
            in[cond_ports[cond_it->first]] = value;
         }

         parse += " : ";
         for(std::vector<std::string>::const_iterator in_it = in.begin(); in_it != in.end(); ++in_it)
            if(in_it == in.begin())
               parse += *in_it;
            else
               parse += ","+*in_it;

         vertex tgt = boost::target(*e_it, *stg);
         bool last_transition = tgt == HLS->STG->get_exit_state();
         vertex next_state = last_transition ? first_state : tgt;

         std::vector<long long int> transition_outputs(out_num, default_COND);
         for(unsigned int k = 0; k < out_num; k++)
         {
            if (present_state[*v][k] == 1 && unbounded_ports.find(k) == unbounded_ports.end()) transition_outputs[k] = 0;
         }
         for(std::map<std::pair<generic_objRef, unsigned int>, generic_objRef>::const_iterator s = selectors.find(conn_binding::IN)->second.begin(); s != selectors.find(conn_binding::IN)->second.end(); s++)
         {
            //std::cerr << jt->second->get_string() << std::endl;
            const std::set<std::pair<vertex, vertex> >& activations = GetPointer<commandport_obj>(s->second)->get_activations();
            for(std::set<std::pair<vertex, vertex> >::const_iterator it = activations.begin(); it != activations.end(); it++)
            {
               THROW_ASSERT(*v != NULL_VERTEX && it->first != NULL_VERTEX, "error on source vertex");
               if (it->first == *v && (it->second == tgt || it->second == NULL_VERTEX))
               {
                  THROW_ASSERT(present_state[*v][out_ports[s->second]] == 1, "unexpected condition");
                  transition_outputs[out_ports[s->second]] = 1;
                  //std::cerr << "Port " << s->second->get_string() << " " << out_ports[s->second] << " enabled from state " << HLS->STG->get_state_name(it->first) << " to " << HLS->STG->get_state_name(it->second) << std::endl;
               }
            }
         }
         for(unsigned int k = 0; k < out_num; k++)
         {
            if(present_state[*v][k] == transition_outputs[k])
               transition_outputs[k] = default_COND;
            else if (present_state[*v][k] != transition_outputs[k] && present_state[*v][k] == 1 && transition_outputs[k] == 0)
            {
               //std::cerr << "k " << k << " to " << HLS->STG->get_state_name(tgt)<< std::endl;
               //abort();
            }
         }

         parse += " " + GET_STATE_NAME(stg, next_state) + " " + (last_transition?"1":"-") + input_vector_to_string(transition_outputs,0);
     }

      parse += "; ";

      parse += "\n";

   }
   //std::cerr << "Finite_state_machine representation: " << std::endl;
   //std::cerr << parse << std::endl << std::endl;
 
}

std::string fsm_controller::get_guard_value(const tree_managerRef & TM, const unsigned int index)
{
   tree_nodeRef node = TM->get_tree_node_const(index);
   THROW_ASSERT(node->get_kind() == case_label_expr_K, "case_label_expr expected");
   case_label_expr * cle = GetPointer<case_label_expr>(node);
   THROW_ASSERT(cle->op0, "guard expected in a case_label_expr");
   integer_cst * ic = GetPointer<integer_cst>(GET_NODE(cle->op0));
   THROW_ASSERT(ic, "expected integer_cst object as guard in a case_label_expr");
   long long int low_result = tree_helper::get_integer_cst_value(ic);
   long long int high_result = 0;
   if(cle->op1)
   {
      integer_cst * ic_high = GetPointer<integer_cst>(GET_NODE(cle->op1));
      THROW_ASSERT(ic_high, "expected integer_cst object as guard in a case_label_expr");
      high_result = tree_helper::get_integer_cst_value(ic_high);
   }
   if(high_result == 0)
      return STR(low_result);
   else
   {
      std::string res_string;
      for(long long int current_value=low_result; current_value<=high_result; ++current_value)
         if(current_value==low_result)
            res_string = STR(current_value);
         else
            res_string += "|"+STR(current_value);
      return res_string;
   }
}
