/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file classic_datapath.cpp
 * @brief Base class for usual datapath creation.
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

#include "classic_datapath.hpp"

#include "hls.hpp"
#include "hls_target.hpp"
#include "hls_constraints.hpp"
#include "function_behavior.hpp"
#include "behavioral_helper.hpp"

#include "generic_obj.hpp"
#include "mux_conn.hpp"
#include "mux_obj.hpp"
#include "commandport_obj.hpp"

#include "technology_manager.hpp"
#include "technology_builtin.hpp"

#include "StateTransitionGraph.hpp"

#include "BambuParameter.hpp"
#include "schedule.hpp"
#include "fu_binding.hpp"
#include "reg_binding.hpp"
#include "conn_binding.hpp"
#include "memory.hpp"

#include "exceptions.hpp"

#include "structural_objects.hpp"
#include "structural_manager.hpp"

#include "tree_reindex.hpp"
#include "tree_manager.hpp"
#include "tree_node.hpp"

#include "dbgPrintHelper.hpp"
#include "Parameter.hpp"

#include <iostream>
#include <boost/lexical_cast.hpp>

classic_datapath::classic_datapath(const ParameterRef _Param, const hlsRef _HLS) :
   datapath_creator(_Param, _HLS)
{
   type = CLASSIC;
}

classic_datapath::~classic_datapath()
{

}

std::string classic_datapath::get_kind_text() const
{
   return "Classical-datapath";
}

void classic_datapath::exec()
{
   /// Test on previuos steps. They checks if schedule and connection binding have been performed. If they didn't,
   /// circuit cannot be created.
   THROW_ASSERT(HLS->Rfu, "Functional units not allocated");
   THROW_ASSERT(HLS->Rreg, "Register allocation not performed");
   THROW_ASSERT(HLS->Rconn, "Connection allocation not performed");
   /// Test on memory allocation
   THROW_ASSERT(HLS->HLS_C->Rmem, "Memory allocation not performed");

   /// main circuit type
   structural_type_descriptorRef module_type = structural_type_descriptorRef(new structural_type_descriptor("datapath_"+HLS->FB->get_helper()->get_function_name()));

   /// top circuit creation
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Main datapath creation");
   HLS->datapath = structural_managerRef(new structural_manager(HLS->Param));

   HLS->datapath->set_top_info("Datapath_i", module_type);
   const structural_objectRef& datapath_cir = HLS->datapath->get_circ();

   // Now the top circuit is created, just as an empty box. <circuit> is a reference to the structural object that
   // will contain all the circuit components

   ///Set some descriptions and legal stuff
   GetPointer<module>(datapath_cir)->set_description("Datapath RTL descrition for " + HLS->FB->get_helper()->get_function_name());
   GetPointer<module>(datapath_cir)->set_copyright("Copyright (C) 2004-2011 Politecnico di Milano");
   GetPointer<module>(datapath_cir)->set_authors("module automatically generated by bambu");
   GetPointer<module>(datapath_cir)->set_license("PANDA_GPLv3");

   /// add clock and reset to circuit. It increments in_port number and update in_port_map
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Adding clock and reset ports");
   structural_objectRef clock, reset;
   add_clock_reset(clock, reset);

   /// add all input ports
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Adding ports for primary inputs and outputs");
   add_ports();

   /// add registers, connecting them to clock and reset ports
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Adding registers");
   HLS->Rreg->add_to_SM(clock, reset);

   /// allocate functional units
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Adding functional units");
   HLS->Rfu->add_to_SM(HLS, clock, reset);

   /// allocate interconnections
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Adding interconnections");
   HLS->Rconn->add_to_SM(HLS, HLS->datapath);

   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Circuit created without errors!");

   /// circuit is now complete. circuit manager can be initialized and dot representation can be created
   HLS->datapath->INIT(true);
   if (debug_level >= DEBUG_LEVEL_VERBOSE) HLS->datapath->write_dot((HLS->FB->get_helper()->get_function_name()+"_Datapath.dot").c_str(), structural_manager::COMPLETE_G);

   return;
}

void classic_datapath::add_clock_reset(structural_objectRef& clock_obj, structural_objectRef& reset_obj)
{

   const structural_managerRef& SM = this->HLS->datapath;
   const structural_objectRef& circuit = SM->get_circ();

   /// define boolean type for clock and reset signal
   structural_type_descriptorRef port_type = structural_type_descriptorRef(new structural_type_descriptor("bool", 0));

   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "   * Start adding clock signal...");
   /// add clock port
   clock_obj = SM->add_port(CLOCK_PORT_NAME, port_o::IN, circuit, port_type);
   GetPointer<port_o>(clock_obj)->set_is_clock(true);
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "    Clock signal added!");

   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "   * Start adding reset signal...");
   /// add reset port
   reset_obj = SM->add_port(RESET_PORT_NAME, port_o::IN, circuit, port_type);
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "    Reset signal added!");

   return;
}

void classic_datapath::add_ports()
{
   bool need_start_done = false;
   const structural_managerRef SM = this->HLS->datapath;
   const structural_objectRef circuit = SM->get_circ();

   behavioral_helperRef BH = HLS->FB->get_helper();

   const std::list<unsigned int>& parameters = BH->get_parameters();
   for(std::list<unsigned int>::const_iterator p = parameters.begin(); p != parameters.end(); p++)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Parameter: " + BH->PrintVariable(*p) + " IN");

      conn_binding::direction_type direction = conn_binding::IN;
      generic_objRef port_obj = HLS->Rconn->get_port(*p, direction);
      structural_type_descriptorRef port_type;
      if(HLS->HLS_C->Rmem->has_base_address(*p) && !HLS->HLS_C->Rmem->is_parm_decl_stored(*p))
      {
         port_type = structural_type_descriptorRef(new structural_type_descriptor("bool", 32));
      }
      else
         port_type = HLS->get_data_type(*p);
      if(HLS->HLS_C->Rmem->has_base_address(*p) && (HLS->HLS_C->Rmem->is_parm_decl_stored(*p) || HLS->HLS_C->Rmem->is_parm_decl_copied(*p)))
         need_start_done = true;
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "  type is: " + port_type->get_name());
      std::string prefix = "in_port_";
      port_o::port_direction port_direction = port_o::IN;
      structural_objectRef p_obj = SM->add_port(prefix + BH->PrintVariable(*p), port_direction, circuit, port_type);
      port_obj->set_structural_obj(p_obj);
      port_obj->set_out_sign(p_obj);
   }

   std::map<std::string, generic_objRef> const_objs = HLS->Rconn->get_constant_objs();
   unsigned int num = 0;
   for(std::map<std::string, generic_objRef>::iterator c = const_objs.begin(); c != const_objs.end(); c++)
   {
      generic_objRef constant_obj = c->second;
      structural_objectRef const_obj = SM->add_module_from_technology_library("const_" + STR(num), CONSTANT_STD, LIBRARY_STD, circuit, HLS->HLS_T->get_technology_manager());

      std::string trimmed_value = "\"" + c->first + "\"";
      const_obj->set_parameter("value", trimmed_value);
      constant_obj->set_structural_obj(const_obj);
      std::string name = "out_const_" + boost::lexical_cast<std::string>(num);
      structural_type_descriptorRef sign_type = structural_type_descriptorRef(new structural_type_descriptor("bool", static_cast<unsigned int>(c->first.size())));
      structural_objectRef sign = SM->add_sign(name, circuit, sign_type);
      structural_objectRef out_port =const_obj->find_member("out1", port_o_K, const_obj);
      //customize output port size
      out_port->type_resize(static_cast<unsigned int>(c->first.size()));
      SM->add_connection(sign, out_port);
      constant_obj->set_out_sign(sign);
      num++;
   }

   if (BH->has_return())
   {
      unsigned int var = BH->get_function_return(BH->get_function_index());
      PRINT_DBG_STRING(DEBUG_LEVEL_PEDANTIC, debug_level, "Return type: " + BH->print_type(var));

      generic_objRef port_obj = HLS->Rconn->get_port(var, conn_binding::OUT);
      structural_type_descriptorRef port_type = HLS->get_data_type(var);
      structural_objectRef p_obj = SM->add_port(RETURN_PORT_NAME, port_o::OUT, circuit, port_type);
      port_obj->set_structural_obj(p_obj);
   }
   /// add start and done when needed
   if(need_start_done)
   {
      structural_type_descriptorRef bool_type = structural_type_descriptorRef(new structural_type_descriptor("bool", 0));
      SM->add_port(START_PORT_NAME, port_o::IN, circuit, bool_type);
      SM->add_port(DONE_PORT_NAME, port_o::OUT, circuit, bool_type);
   }
}

