/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file top_entity.cpp
 * @brief Implementation of the class creating the top entity.
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 * $Locker:  $
 * $State: Exp $
 *
*/
#include "top_entity.hpp"

#include "function_behavior.hpp"
#include "behavioral_helper.hpp"

#include "hls.hpp"
#include "schedule.hpp"
#include "fu_binding.hpp"
#include "conn_binding.hpp"

#include "structural_manager.hpp"
#include "structural_objects.hpp"
#include "generic_obj.hpp"
#include "commandport_obj.hpp"

#include "technology_manager.hpp"
#include "technology_builtin.hpp"

#include "BambuParameter.hpp"
#include "dbgPrintHelper.hpp"
#include "exceptions.hpp"

top_entity::top_entity(const ParameterRef _Param, const hlsRef _HLS) :
      HLS_step(_Param, _HLS)
{

}

top_entity::~top_entity()
{

}

void top_entity::exec()
{
   /// function name to be synthesized
   const std::string function_name = HLS->FB->get_helper()->get_function_name();

   /// Test on previuos steps. They checks if datapath and controller have been created. If they didn't,
   /// top circuit cannot be created.
   THROW_ASSERT(HLS->datapath, "Datapath not created");
   THROW_ASSERT(HLS->controller, "Controller not created");

   // reference to hls top circuit
   HLS->top = structural_managerRef(new structural_manager(Param));
   SM = HLS->top;
   structural_managerRef Datapath = HLS->datapath;
   structural_managerRef Controller = HLS->controller;

   /// top circuit creation
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Top circuit creation");

   /// main circuit type
   structural_type_descriptorRef module_type = structural_type_descriptorRef(new structural_type_descriptor(function_name));
   /// setting top circuit component
   SM->set_top_info(function_name, module_type);
   structural_objectRef circuit = SM->get_circ();
   THROW_ASSERT(circuit, "Top circuit is missing");
   // Now the top circuit is created, just as an empty box. <circuit> is a reference to the structural object that
   // will contain all the circuit components

   ///Set some descriptions and legal stuff
   GetPointer<module>(circuit)->set_description("Top circuit for " + function_name);
   GetPointer<module>(circuit)->set_copyright("Copyright (C) 2004-2011 Politecnico di Milano");
   GetPointer<module>(circuit)->set_authors("module automatically generated by bambu");
   GetPointer<module>(circuit)->set_license("PANDA_GPLv3");

   structural_objectRef datapath_circuit = Datapath->get_circ();
   THROW_ASSERT(datapath_circuit, "Missing datapath circuit");
   structural_objectRef controller_circuit = Controller->get_circ();
   THROW_ASSERT(controller_circuit, "Missing controller circuit");

   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Creating controller object");
   /// creating structural_manager
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Adding controller");
   datapath_circuit->set_owner(circuit);
   GetPointer<module>(circuit)->add_internal_object(datapath_circuit);

   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Creating datapath object");
   /// creating structural_manager
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Adding datapath");
   controller_circuit->set_owner(circuit);
   GetPointer<module>(circuit)->add_internal_object(controller_circuit);

   /// command signal type descriptor
   structural_type_descriptorRef bool_type = structural_type_descriptorRef(new structural_type_descriptor("bool", 0));

   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "\tStart adding clock signal...");
   /// add clock port
   structural_objectRef clock_obj = SM->add_port(CLOCK_PORT_NAME, port_o::IN, circuit, bool_type);
   GetPointer<port_o>(clock_obj)->set_is_clock(true);
   /// connect to datapath and controller clock
   structural_objectRef datapath_clock = datapath_circuit->find_member(CLOCK_PORT_NAME, port_o_K, datapath_circuit);
   SM->add_connection(datapath_clock, clock_obj);
   structural_objectRef controller_clock = controller_circuit->find_member(CLOCK_PORT_NAME, port_o_K, controller_circuit);
   SM->add_connection(controller_clock, clock_obj);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "\tClock signal added!");

   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "\tAdding reset signal...");
   /// add reset port
   structural_objectRef reset_obj = SM->add_port(RESET_PORT_NAME, port_o::IN, circuit, bool_type);
   /// connecting global reset port to the datapath one
   structural_objectRef datapath_reset = datapath_circuit->find_member(RESET_PORT_NAME, port_o_K, datapath_circuit);
   SM->add_connection(datapath_reset, reset_obj);
   /// connecting global reset port to the controller one
   structural_objectRef controller_reset = controller_circuit->find_member(RESET_PORT_NAME, port_o_K, controller_circuit);
   SM->add_connection(controller_reset, reset_obj);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "\tReset signal added!");

   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "\tAdding start signal...");
   /// start port
   structural_objectRef start_obj = SM->add_port(START_PORT_NAME, port_o::IN, circuit, bool_type);
   structural_objectRef controller_start = controller_circuit->find_member(START_PORT_NAME, port_o_K, controller_circuit);
   /// check if datapath has a start signal
   structural_objectRef datapath_start = datapath_circuit->find_member(START_PORT_NAME, port_o_K, datapath_circuit);
   structural_objectRef sync_datapath_controller;
   if(datapath_start)
   {
      SM->add_connection(start_obj, datapath_start);
      sync_datapath_controller = SM->add_sign("done2start", circuit, bool_type);
      structural_objectRef datapath_done = datapath_circuit->find_member(DONE_PORT_NAME, port_o_K, datapath_circuit);
      SM->add_connection(datapath_done, sync_datapath_controller);
   }
   else
      SM->add_connection(start_obj, controller_start);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "\tStart signal added!");

   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "\tStart adding Done signal...");
   /// add done port
   structural_objectRef done_obj = SM->add_port(DONE_PORT_NAME, port_o::OUT, circuit, bool_type);
   THROW_ASSERT(done_obj, "Done port not added in the top component");
   structural_objectRef controller_done = controller_circuit->find_member(DONE_PORT_NAME, port_o_K, controller_circuit);
   THROW_ASSERT(controller_done, "Done signal not found in the controller");
   if(datapath_start)
      SM->add_connection(sync_datapath_controller, controller_start);
   SM->add_connection(controller_done, done_obj);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "\tDone signal added!");

   /// add entry in in_port_map between port id and port index

   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "\tAdding input/output ports...");
   this->add_ports(circuit);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "\tInput/output ports added!");

   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "\tAdding command ports...");
   this->add_command_signals(circuit);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "\tCommand ports added!");

   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Circuit created without errors!");
   return;
}

void top_entity::add_ports(structural_objectRef circuit)
{
   const behavioral_helperRef& BH = HLS->FB->get_helper();

   const structural_objectRef& Datapath = HLS->datapath->get_circ();
   const std::list<unsigned int>& parameters = BH->get_parameters();
   conn_binding& conn = *(HLS->Rconn);
   for (std::list<unsigned int>::const_iterator i = parameters.begin(); i != parameters.end(); i++)
   {
      generic_objRef input = conn.get_port(*i, conn_binding::IN);
      structural_objectRef in_obj = input->get_structural_obj();
      structural_type_descriptorRef port_type = in_obj->get_typeRef();
      structural_objectRef top_obj = SM->add_port(HLS->FB->get_helper()->PrintVariable(*i), port_o::IN, circuit, port_type);
      SM->add_connection(in_obj, top_obj);
   }
   if (BH->has_return())
   {
      unsigned int var = BH->get_function_return(BH->get_function_index());

      generic_objRef return_port = conn.get_port(var, conn_binding::OUT);
      structural_objectRef ret_obj = return_port->get_structural_obj();
      structural_type_descriptorRef port_type = ret_obj->get_typeRef();
      structural_objectRef top_obj = SM->add_port(RETURN_PORT_NAME, port_o::OUT, circuit, port_type);
      SM->add_connection(ret_obj, top_obj);
   }

   bool is_top = (BH->get_function_name() == Param->getOption<std::string>(OPT_top_function_name));
   structural_objectRef master_port = Datapath->find_member("M_DataRdy", port_o_K, Datapath);
   ////////////////////////////////////////////////////////////////////////////////////////
   ////////////////////////////////////////////////////////////////////////////////////////
   ////////////////////////////////////////////////////////////////////////////////////////
   std::map<unsigned int, structural_objectRef> null_values;
   /// creating extern IN port on top starting from extern ports on datapath and add connection
   for(unsigned int j = 0; j < GetPointer<module>(Datapath)->get_in_port_size(); j++)
   {
      structural_objectRef port_in = GetPointer<module>(Datapath)->get_in_port(j);
      if(GetPointer<port_o>(port_in)->get_is_extern())
      {
         if (!GetPointer<port_o>(port_in)->get_is_memory() || !is_top || master_port)
         {
            structural_objectRef ext_port = this->SM->add_port(GetPointer<port_o>(port_in)->get_id(), port_o::IN, circuit, port_in->get_typeRef());
            GetPointer<port_o>(ext_port)->set_is_extern(true);
            if(GetPointer<port_o>(port_in)->get_is_memory())
               GetPointer<port_o>(ext_port)->set_is_memory(true);
            if(GetPointer<port_o>(port_in)->get_is_global())
               GetPointer<port_o>(ext_port)->set_is_global(true);
            //adding connection between datapath extern port and top extern port
            this->SM->add_connection(port_in, ext_port);
         }
         else
         {
            if (null_values.find(GET_TYPE_SIZE(port_in)) == null_values.end())
            {
               structural_objectRef const_obj = SM->add_constant("null_value_" + STR(GET_TYPE_SIZE(port_in)), circuit, port_in->get_typeRef(), STR(0));
               null_values[GET_TYPE_SIZE(port_in)] = const_obj;
            }
            SM->add_connection(port_in, null_values[GET_TYPE_SIZE(port_in)]);
         }
      }
      else if(GetPointer<port_o>(port_in)->get_is_memory())
      {
         if (!is_top || master_port)
         {
            structural_objectRef ext_port = this->SM->add_port(GetPointer<port_o>(port_in)->get_id(), port_o::IN, circuit, port_in->get_typeRef());
            GetPointer<port_o>(ext_port)->set_is_memory(true);
            if(GetPointer<port_o>(port_in)->get_is_slave())
               GetPointer<port_o>(ext_port)->set_is_slave(true);
            if(GetPointer<port_o>(port_in)->get_is_master())
               GetPointer<port_o>(ext_port)->set_is_master(true);
            if(GetPointer<port_o>(port_in)->get_is_global())
               GetPointer<port_o>(ext_port)->set_is_global(true);
            //adding connection between datapath extern port and top extern port
            this->SM->add_connection(port_in,ext_port);
         }
         else
         {
            if (null_values.find(GET_TYPE_SIZE(port_in)) == null_values.end())
            {
               structural_objectRef const_obj = SM->add_constant("null_value_" + STR(GET_TYPE_SIZE(port_in)), circuit, port_in->get_typeRef(), STR(0));
               null_values[GET_TYPE_SIZE(port_in)] = const_obj;
            }
            SM->add_connection(port_in, null_values[GET_TYPE_SIZE(port_in)]);
         }
      }
   }
   /// creating extern OUT port on top starting from extern ports on datapath and add connection
   for(unsigned int j = 0; j < GetPointer<module>(Datapath)->get_out_port_size(); j++)
   {
      structural_objectRef port_out = GetPointer<module>(Datapath)->get_out_port(j);
      if (GetPointer<port_o>(port_out)->get_is_memory() && is_top && !master_port) continue;
      if (GetPointer<port_o>(port_out)->get_is_extern())
      {
         structural_objectRef ext_port = this->SM->add_port(GetPointer<port_o>(port_out)->get_id(), port_o::OUT, circuit, port_out->get_typeRef());
         GetPointer<port_o>(ext_port)->set_is_extern(true);
         if(GetPointer<port_o>(port_out)->get_is_memory())
            GetPointer<port_o>(ext_port)->set_is_memory(true);
         if(GetPointer<port_o>(port_out)->get_is_global())
            GetPointer<port_o>(ext_port)->set_is_global(true);
         //adding connection between datapath extern port and top extern port
         this->SM->add_connection(port_out,ext_port);
      }
      else if(GetPointer<port_o>(port_out)->get_is_memory())
      {
         structural_objectRef ext_port = this->SM->add_port(GetPointer<port_o>(port_out)->get_id(), port_o::OUT, circuit, port_out->get_typeRef());
         GetPointer<port_o>(ext_port)->set_is_memory(true);
         if(GetPointer<port_o>(port_out)->get_is_slave())
            GetPointer<port_o>(ext_port)->set_is_slave(true);
         if(GetPointer<port_o>(port_out)->get_is_master())
               GetPointer<port_o>(ext_port)->set_is_master(true);
         if(GetPointer<port_o>(port_out)->get_is_global())
            GetPointer<port_o>(ext_port)->set_is_global(true);
         //adding connection between datapath extern port and top extern port
         this->SM->add_connection(port_out,ext_port);

      }
   }

   /// creating extern IO port on top starting from extern ports on datapath and add connection
   for(unsigned int j = 0; j < GetPointer<module>(Datapath)->get_in_out_port_size(); j++)
   {
      structural_objectRef port_in_out = GetPointer<module>(Datapath)->get_in_out_port(j);
      if(GetPointer<port_o>(port_in_out)->get_is_extern())
      {
         structural_objectRef ext_port = this->SM->add_port(GetPointer<port_o>(port_in_out)->get_id(), port_o::IO, circuit, port_in_out->get_typeRef());
         GetPointer<port_o>(ext_port)->set_is_extern(true);
         //adding connection between datapath extern port and top extern port
         this->SM->add_connection(port_in_out,ext_port);
      }
   }
   ////////////////////////////////////////////////////////////////////////////////////////
   ////////////////////////////////////////////////////////////////////////////////////////
   ////////////////////////////////////////////////////////////////////////////////////////

}

#if 0
void top_entity::add_output_ports(structural_objectRef circuit)
{
   /// define boolean type for command signals
   structural_type_descriptorRef int_type =
      structural_type_descriptorRef(new structural_type_descriptor("int", 0));
   structural_type_descriptorRef bool_type =
      structural_type_descriptorRef(new structural_type_descriptor("bool", 0));

   structural_objectRef Datapath = HLS->datapath->get_circ();
   structural_objectRef Controller = HLS->controller->get_circ();
   std::list<unsigned int>::const_iterator i = HLS->BM->get_write_port_list().begin();
   conn_binding& conn = *(HLS->Rconn);
   unsigned int num = 0;
   for (; i != HLS->BM->get_write_port_list().end(); i++, num++)
   {
      generic_objRef output = conn.get_port(*i, conn_binding::OUT);
      structural_objectRef out_obj = output->get_structural_obj();
      structural_objectRef top_obj = SM->add_port("output_" + boost::lexical_cast<std::string>(*i), port_o::OUT, circuit, int_type);
      SM->add_connection(out_obj, top_obj);

      std::string name = "en_output_" + boost::lexical_cast<std::string>(num);
      structural_objectRef en_out_obj = Controller->find_member(name, port_o_K, Controller);
      structural_objectRef en_top_obj = SM->add_port(name, port_o::OUT, circuit, bool_type);
      SM->add_connection(en_out_obj, en_top_obj);
   }
}
#endif

void top_entity::add_command_signals(structural_objectRef circuit)
{
   //std::tr1::unordered_set<vertex> levels = HLS->operations;
   const graphRef data = HLS->get_graph(function_behavior::DFG);

   structural_objectRef Datapath = HLS->datapath->get_circ();
   structural_objectRef Controller = HLS->controller->get_circ();

   const std::map<unsigned int, std::map<std::pair<generic_objRef, unsigned int>, generic_objRef> >& selectors = HLS->Rconn->get_selectors();
   for(std::map<unsigned int, std::map<std::pair<generic_objRef, unsigned int>, generic_objRef> >::const_iterator j = selectors.begin(); j != selectors.end(); j++)
   {
      for(std::map<std::pair<generic_objRef, unsigned int>, generic_objRef>::const_iterator l = j->second.begin(); l != j->second.end(); l++)
      {
         structural_objectRef datapath_obj = l->second->get_structural_obj();
         THROW_ASSERT(datapath_obj, "missing structural object associated with the selector " + l->second->get_string());
         std::string datapath_name = datapath_obj->get_id();
         structural_objectRef controller_obj = GetPointer<commandport_obj>(l->second)->get_controller_obj();
         ///it means that the operation has not to be executed
         if (!controller_obj) continue;
         std::string controller_name = controller_obj->get_id();
         structural_objectRef src = Controller->find_member(controller_name, port_o_K, Controller);
         THROW_ASSERT(src, "Missing select port in the controller");
         structural_objectRef tgt = Datapath->find_member(datapath_name, port_o_K, Datapath);
         THROW_ASSERT(tgt, "Missing select port in the datapath");
         structural_objectRef sign = SM->add_sign(datapath_name, circuit, src->get_typeRef());
         SM->add_connection(src, sign);
         SM->add_connection(sign, tgt);
      }
   }
}

std::string top_entity::get_kind_text() const
{
   return "top-creation";
}
