/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file mux_connection_binding.hpp
 * @brief Class to manage mux-based interconnections based on the FSM controller
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

#ifndef MUX_CONNECTION_BINDING_HPP
#define MUX_CONNECTION_BINDING_HPP

#include "conn_binding.hpp"
#include "conn_binding_creator.hpp"

#include "application_manager.hpp"

#include "ext_tree_node.hpp"

#include <tr1/unordered_map>

/**
 * Definition of hash function for EdgeDescriptor
 */
namespace std
{
   namespace tr1
   {
      template <>
      struct hash<std::tr1::tuple<generic_objRef, unsigned int> > : public unary_function<std::tr1::tuple<generic_objRef, unsigned int>, size_t>
      {
         size_t operator()(std::tr1::tuple<generic_objRef, unsigned int> p) const
         {
            size_t hash_value = 0;
            boost::hash_combine(hash_value, std::tr1::get<0>(p));
            boost::hash_combine(hash_value, std::tr1::get<1>(p));

            return hash_value;
         }
      };
   }
}



REF_FORWARD_DECL(graph);
/**
 * @class mux_connection_binding
 * @ingroup interconnect
 * @brief Class managing the connection binding of datapath. It extends the standard mux_connection_binding class.
 *
 * This class provides methods to compute interconnections among datapath elements (functional units, registers,
 * ports). Different architectural style have been implemented: mux-based or bus-based. They are implemented as
 * specialization of generic interconnections.
 *
 */
class mux_connection_binding : public conn_binding_creator
{
   private:

      /// variable used to assign a unique id to sparse logic
      unsigned int id;

      /// store the current phi use. Used during the analysis of phi nodes
      unsigned int cur_phi_tree_var;

      /**
       * Performs specialization of interconnections using mux architecture.
       */
      unsigned int mux_interconnection();

      /**
       * Computes logic for inputs. The connection can become directed or by multiplexer. If multiplexers are
       * allocated, then decoding logic is created.
       * @param src is the set of references to generic_obj, sources of connections
       * @param tgt is the reference to connection target
       * @param op is i-th operand of target element, where source is attached
       * @return number of multiplexer allocated
       */
      unsigned int input_logic(const std::map<generic_objRef, std::set<conn_binding::temporary> >& src, const generic_objRef tgt, unsigned int op, unsigned int iteration);

      /**
       * Determine the actual interconnection
       */
      void determine_connection(const vertex& op, const application_manager::io_binding_type& var, generic_objRef fu_obj, unsigned int port_num, const graphRef data, unsigned int precision);

      /**
       * Compute the bitsize given a io_binding type
       */
      unsigned int object_bitsize(const tree_managerRef TreeM, const application_manager::io_binding_type & obj) const;

      /**
      * Recursive function which returns the offset of a dynamic multidimensional array call
      */
      generic_objRef dynamic_multidimensional_array_handler(array_ref * ar, const vertex& op, const graphRef data, unsigned int &base_address_index_pointer, std::vector<unsigned int> &recursive_indexes_values, std::vector<unsigned int> &dims, generic_objRef &global_adder, const bool is_not_a_phi);


   private:

      /**
       *  create the connection object and update the unique table
       */
      void create_single_conn(const vertex& op, generic_objRef fu_obj_src,
                                            generic_objRef fu_obj, unsigned int port_num, unsigned int tree_var, unsigned int precision, const bool is_not_a_phi);

      /**
       * connect the fu_obj with the associated registers.
       */
      void connect_to_registers(vertex op, const graphRef data, generic_objRef fu_obj, unsigned int port_num, unsigned int tree_var, unsigned int precision, const bool is_not_a_phi);

   public:

      /**
       * Main constructor
       * @param hls_f is the hls data structure.
       */
      mux_connection_binding(const ParameterRef Param, const hlsRef HLS);

      /**
       * Destructor
       */
      virtual ~mux_connection_binding();

      /**
       * Function that starts connection binding computation.
       * @param hls_flow is the hls data structure.
       */
      void exec();

      /**
       * Creates the connections inside the architecture
       */
      void create_connections();

      /**
       * Returns the name of the algorithm
       */
      std::string get_kind_text() const;

};

#endif
