/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file conn_binding.cpp
 * @brief Class implementation of the interconnection binding data structure.
 *
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Autoheader include
#include "config_HAVE_EXPERIMENTAL.hpp"
#include "config_HAVE_FLOPOCO.hpp"

#include "conn_binding.hpp"

#include "application_manager.hpp"
#include "hls_target.hpp"

#include "connection_obj.hpp"
#include "mux_conn.hpp"

#include "dbgPrintHelper.hpp"
#include "dataport_obj.hpp"
#include "commandport_obj.hpp"
#include "conn_binding_creator.hpp"
#include "mux_conn.hpp"
#include "mux_obj.hpp"
#include "adder_conn_obj.hpp"
#include "multiplier_conn_obj.hpp"
#include "bitfield_obj.hpp"
#include "funit_obj.hpp"
#include "register_obj.hpp"
#include "fu_binding.hpp"

#include "hls.hpp"
#include "allocation.hpp"

#include "structural_manager.hpp"
#include "technology_manager.hpp"
#include "technology_builtin.hpp"
#include "tree_helper.hpp"
#include "tree_manager.hpp"
#include "time_model.hpp"

#include "Parameter.hpp"
#include "behavioral_helper.hpp"

#include <boost/lexical_cast.hpp>

#if HAVE_FLOPOCO
#include "flopoco_wrapper.hpp"
#include <mpfr.h>
namespace flopoco
{
   extern std::string fp2bin(mpfr_t x, int wE, int wF);
}
#endif

#define CONN_COLUMN_SIZE 40

conn_binding::conn_binding(const behavioral_helperRef _BH, int _debug_level) :
   debug_level(_debug_level), BH(_BH)
{
}

conn_binding::~conn_binding()
{

}

generic_objRef conn_binding::get_port(unsigned int var, conn_binding::direction_type dir)
{
   switch (dir)
   {
      case IN:
         THROW_ASSERT(input_ports.count(var), "Data port not stored into conn_binding class");
         return input_ports[var];
      case OUT:
         THROW_ASSERT(output_ports.count(var), "Data port not stored into conn_binding class");
         return output_ports[var];
      default:
         THROW_ERROR("Port kind not allowed");
   }
   return generic_objRef();
}

void conn_binding::add_data_transfer(const generic_objRef op1, const generic_objRef op2, unsigned int operand, temporary data)
{
   conn_variables[std::make_pair(op2, operand)][op1].insert(data);
}

void conn_binding::remove_data_transfer(const generic_objRef op1, const generic_objRef op2, unsigned int operand, temporary data)
{
   if (conn_variables.find(std::make_pair(op2, operand)) == conn_variables.end())
      return;
   if (conn_variables[std::make_pair(op2, operand)].find(op1) == conn_variables[std::make_pair(op2, operand)].end())
      return;
   if (conn_variables[std::make_pair(op2, operand)][op1].find(data) != conn_variables[std::make_pair(op2, operand)][op1].end())
   {
      conn_variables[std::make_pair(op2, operand)][op1].erase(data);
      if (conn_variables[std::make_pair(op2, operand)][op1].size() == 0)
         conn_variables[std::make_pair(op2, operand)].erase(op1);
   }
}

const std::map<std::pair<generic_objRef, unsigned int>, std::map<generic_objRef, std::set<temporary> > >& conn_binding::get_data_transfers() const
{
   return conn_variables;
}

const conn_binding::conn_implementation_map& conn_binding::get_connection_implementations() const
{
   return conn_implementation;
}

std::map<generic_objRef, std::set<temporary> > conn_binding::get_data_transfer(const generic_objRef op2, unsigned int operand) const
{
   return conn_variables.find(std::make_pair(op2, operand))->second;
}

std::set<temporary> conn_binding::get_data_transfer(const generic_objRef op1, const generic_objRef op2, unsigned int operand) const
{
   if (conn_variables.find(std::make_pair(op2, operand)) == conn_variables.end())
      return std::set<temporary>();
   if (conn_variables.find(std::make_pair(op2, operand))->second.find(op1) == conn_variables.find(std::make_pair(op2, operand))->second.end())
      return std::set<temporary>();
   return conn_variables.find(std::make_pair(op2, operand))->second.find(op1)->second;
}

void conn_binding::add_connection(const generic_objRef op1, const generic_objRef op2, unsigned int operand, connection_objRef conn)
{
   conn_implementation[connection(op1, op2, operand)] = conn;
}

connection_objRef conn_binding::get_connection(const generic_objRef op1, const generic_objRef op2, unsigned int operand) const
{
   THROW_ASSERT(is_connection(op1, op2, operand), "missing connection between source and target");
   return conn_implementation.find(connection(op1, op2, operand))->second;
}

bool conn_binding::is_connection(const generic_objRef op1, const generic_objRef op2, unsigned int operand) const
{
   return conn_implementation.find(connection(op1, op2, operand)) != conn_implementation.end();
}

generic_objRef conn_binding::get_command_port(const vertex& ver, conn_binding::direction_type dir, unsigned int mode)
{
   switch (dir)
   {
      case IN:
         THROW_ASSERT(command_input_ports.count(std::make_pair(ver,mode)), "Command port not stored into conn_binding class");
         return command_input_ports[std::make_pair(ver,mode)];
      case OUT:
         THROW_ASSERT(command_output_ports.count(ver), "Command port not stored into conn_binding class");
         return command_output_ports[ver];
      default:
         THROW_ERROR("Port kind not allowed");
   }
   return generic_objRef();
}

generic_objRef conn_binding::bind_port(unsigned int var, conn_binding::direction_type dir)
{
   switch (dir)
   {
      case IN:
      {
         if (input_ports.count(var) == 0)
            input_ports[var] = generic_objRef(new dataport_obj("IN_PORT_" + BH->PrintVariable(var)));
         return input_ports[var];
      }
      case OUT:
      {
         if (output_ports.count(var) == 0)
            output_ports[var] = generic_objRef(new dataport_obj("OUT_PORT_" + BH->PrintVariable(var)));
         return output_ports[var];
      }
      default:
         THROW_ERROR("Port kind not allowed");
   }
   return generic_objRef();
}

generic_objRef conn_binding::bind_command_port(const vertex& ver, conn_binding::direction_type dir, unsigned int mode, const graph* g)
{
   switch (dir)
   {
      case IN:
      {
         if (command_input_ports.count(std::make_pair(ver,mode)) == 0)
            command_input_ports[std::make_pair(ver,mode)] = generic_objRef(new commandport_obj(ver, mode, "IN_" + commandport_obj::get_mode_string(mode) + "_" + GET_NAME(g,ver)));
         return command_input_ports[std::make_pair(ver,mode)];
      }
      case OUT:
      {
         if (command_output_ports.count(ver) == 0)
            command_output_ports[ver] = generic_objRef(new commandport_obj(ver, mode, "OUT_" + commandport_obj::get_mode_string(mode) + "_" + GET_NAME(g,ver)));
         return command_output_ports[ver];
      }
      default:
         THROW_ERROR("Port kind not allowed");
   }
   return generic_objRef();
}

generic_objRef conn_binding::bind_selector_port(conn_binding::direction_type dir, unsigned int mode, const generic_objRef elem, unsigned int op)
{
   if (selectors.find(dir) == selectors.end() or selectors[dir].find(std::make_pair(elem, op)) == selectors[dir].end())
      selectors[dir][std::make_pair(elem, op)] = generic_objRef(new commandport_obj(elem, mode, ( dir == IN ? "IN_" : "OUT_") + elem->get_string() + "_" + commandport_obj::get_mode_string(mode) + "_" + STR(op)));
   return selectors[dir][std::make_pair(elem, op)];
}

generic_objRef conn_binding::bind_selector_port(conn_binding::direction_type dir, unsigned int mode, const vertex& cond, const graph* data)
{
   if (activation_ports.find(cond) != activation_ports.end() and activation_ports[cond].find(dir) != activation_ports[cond].end())
      return activation_ports[cond][dir];
   generic_objRef port = generic_objRef(new commandport_obj(cond, mode, ( dir == IN ? "IN_" : "OUT_") + commandport_obj::get_mode_string(mode) + "_" + GET_NAME(data,cond)));
   activation_ports[cond][dir] = port;
   return selectors[dir][std::make_pair(port, 0)] = port;
}

unsigned int conn_binding::get_component_num(const std::string& type) const
{
   if (component_num.find(type) == component_num.end())
      return 0;
   return component_num.find(type)->second;
}

void conn_binding::add_component(const std::string& type, unsigned int num)
{
   if (component_num.find(type) == component_num.end())
   {
      component_num[type] = num;
      return;
   }
   component_num[type] += num;
}

const std::map<std::string, unsigned int>& conn_binding::get_components() const
{
   return component_num;
}

unsigned int conn_binding::get_to_controller_ports() const
{
   return static_cast<unsigned int>(command_output_ports.size());
}

std::map<std::pair<vertex, unsigned int>, generic_objRef> conn_binding::get_command_input_ports() const
{
   return command_input_ports;
}

unsigned int conn_binding::get_from_controller_ports() const
{
   return static_cast<unsigned int>(command_input_ports.size());
}

void conn_binding::add_to_SM(const hlsRef HLS, const structural_managerRef SM)
{
    unsigned int connection_type = HLS->Param->getOption<unsigned int>(OPT_datapath_interconnection_style);

    /// add command ports
    PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, HLS->debug_level, "Adding command ports coming from controller");
    add_command_ports(HLS, SM);

    /// add sparse logic
    PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, HLS->debug_level, "Adding sparse logic to the datapath");
    add_sparse_logic(HLS, SM);

    /// up to now, circuit is general about interconnections. Now, proper interconnection architecture will be executed
    switch (connection_type)
    {
       case conn_binding_creator::MUX:
          this->mux_connection(HLS, SM);
          break;
       default:
          THROW_ERROR("Interconnection mode not allowed in datapath creation");
    }

    const structural_objectRef circuit = SM->get_circ();
    std::map<unsigned int, structural_objectRef> null_values;
    for (unsigned int i = 0; i < GetPointer<module>(circuit)->get_internal_objects_size(); i++)
    {
       structural_objectRef curr_gate = GetPointer<module>(circuit)->get_internal_object(i);
       if (!GetPointer<module>(curr_gate)) continue;
       for(unsigned int j = 0; j < GetPointer<module>(curr_gate)->get_in_port_size(); j++)
       {
          structural_objectRef port_i = GetPointer<module>(curr_gate)->get_in_port(j);
          //std::cerr << "port_i " << port_i->get_path() << " of size " << GET_TYPE_SIZE(port_i) << std::endl;
          if (GetPointer<port_o>(port_i) && GetPointer<port_o>(port_i)->find_bounded_object()) continue;
          if (GetPointer<port_vector_o>(port_i))
          {
             bool pempty = true;
             for(unsigned int p = 0; p < GetPointer<port_vector_o>(port_i)->get_ports_size() && pempty; ++p)
             {
                structural_objectRef port_d = GetPointer<port_vector_o>(port_i)->get_port(p);
                if (GetPointer<port_o>(port_d)->find_bounded_object()) pempty = false;
             }
             if (!pempty) continue;
          }
          //std::cerr << "  empty" << std::endl;
          if (null_values.find(GET_TYPE_SIZE(port_i)) == null_values.end())
          {
             structural_type_descriptorRef bool_type = structural_type_descriptorRef(new structural_type_descriptor("bool", GET_TYPE_SIZE(port_i)));
             structural_objectRef const_obj = SM->add_constant("null_value_" + STR(GET_TYPE_SIZE(port_i)), circuit, bool_type, STR(0));
             null_values[GET_TYPE_SIZE(port_i)] = const_obj;
          }
          SM->add_connection(port_i, null_values[GET_TYPE_SIZE(port_i)]);
       }
    }
}

void conn_binding::mux_connection(const hlsRef HLS, const structural_managerRef SM)
{
   structural_objectRef circuit = SM->get_circ();
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Datapath interconnection using mux architecture");

   //std::set<std::pair<std::string, std::string> > already_considered;
   for (std::map<std::tr1::tuple<generic_objRef, generic_objRef, unsigned int>, connection_objRef>::const_iterator i = conn_implementation.begin(); i != conn_implementation.end(); i++)
   {
      generic_objRef src = std::tr1::get<0>(i->first);
      generic_objRef tgt = std::tr1::get<1>(i->first);

      THROW_ASSERT(src, "a NULL src may come from uninitialized variables. Target: "+ tgt->get_string());
      unsigned int operand = std::tr1::get<2>(i->first);
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Creating CONNECTION between " + src->get_string() + " and " + tgt->get_string() + "(" + STR(operand) + ")");

      structural_objectRef src_module = src->get_structural_obj();
      structural_objectRef tgt_module = tgt->get_structural_obj();
      THROW_ASSERT(src, "No source signal to " + tgt->get_string());
      structural_objectRef port_src, port_tgt;
      THROW_ASSERT(src_module, "No object associated to " + src->get_string());
      if (src_module->get_kind() == component_o_K)
      {
         module* src_obj = GetPointer<module>(src_module);
         for(unsigned int ind = 0; ind < src_obj->get_out_port_size(); ind++)
         {
            if (src_obj->get_out_port(ind)->get_id() == DONE_PORT_NAME) continue;
            port_src = src_obj->get_out_port(ind);
            break;
         }
      }
      else if (src_module->get_kind() == port_o_K)
         port_src = src_module;

      THROW_ASSERT(tgt_module, "No object associated to " + tgt->get_string());
      if (tgt_module->get_kind() == component_o_K)
      {
         module* tgt_obj = GetPointer<module>(tgt_module);
         unsigned int num = 0;

         for(unsigned int ind = 0; ind < tgt_obj->get_in_port_size(); ind++)
         {
            if (tgt_obj->get_in_port(ind)->get_id() == CLOCK_PORT_NAME || tgt_obj->get_in_port(ind)->get_id() == RESET_PORT_NAME || tgt_obj->get_in_port(ind)->get_id() == START_PORT_NAME) continue;
            if (num == operand)
            {
               port_tgt = tgt_obj->get_in_port(ind);
               break;
            }
            num++;
         }
      }
      else if (tgt_module->get_kind() == port_o_K)
         port_tgt = tgt_module;

      structural_type_descriptorRef port_src_type = port_src->get_typeRef();
      structural_objectRef sign = src->get_out_sign();
      if (!sign)
      {
         std::string name = "out_" + src->get_structural_obj()->get_id();
         sign = SM->add_sign(name, circuit, port_src_type);
         THROW_ASSERT(port_src_type->size,"size greater than one expected");
         src->set_out_sign(sign);
         SM->add_connection(port_src, sign);
      }

      switch (i->second->get_type())
      {
         case connection_obj::DIRECT_CONN:
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Creating DIRECTED CONNECTION");
            add_datapath_connection(HLS->HLS_T->get_technology_manager(), SM, sign, port_tgt);
            break;
         }
         case connection_obj::BY_MUX:
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Creating MUX TREE");
            mux_allocation(HLS, SM, sign, port_tgt, i->second);
            break;
         }
         default:
            THROW_ERROR("Connection type not allowed: " + STR(i->second->get_type()));
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "");
   }

   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Datapath interconnections completed!");
}


void conn_binding::specialise_mux(const generic_objRef mux) const
{
   unsigned int data_size = GetPointer<mux_obj>(mux)->get_bitsize();

   structural_objectRef mux_obj = mux->get_structural_obj();
   const module* mux_module = GetPointer<module>(mux_obj);

   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Specialising " +  mux_obj->get_path() + ": " + STR(data_size));
   ///specializing multiplexer ports
   mux_module->get_in_port(1)->type_resize(data_size);
   mux_module->get_in_port(2)->type_resize(data_size);
   mux_module->get_out_port(0)->type_resize(data_size);
}

void conn_binding::mux_allocation(const hlsRef HLS, const structural_managerRef SM, structural_objectRef src, structural_objectRef tgt, connection_objRef conn)
{
   THROW_ASSERT(src, "mux_allocation - No source object");
   THROW_ASSERT(tgt, "mux_allocation - No target object");
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "conn_binding::mux_allocation between " + src->get_path() + " and " + tgt->get_path());

   THROW_ASSERT(GetPointer<mux_conn>(conn), "The connection is not implemented through multiplexers");
   mux_conn *mux_alloc = GetPointer<mux_conn>(conn);
   const std::vector<std::pair<generic_objRef, unsigned int> >& mux_tree = mux_alloc->get_mux_tree();
   THROW_ASSERT(mux_tree.size() > 0, "no mux into a mux connection");

   const structural_objectRef circuit = SM->get_circ();
   for (unsigned int i = 0; i < mux_tree.size(); i++)
   {
      structural_objectRef mux = mux_tree[i].first->get_structural_obj();
      unsigned int in_mux = mux_tree[i].second == T_COND ? 1 : 2;

      if (mux)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "MUX exists");
         module* mux_object = GetPointer<module>(mux);

         ///adding input connection
         structural_objectRef mux_input = mux_object->get_in_port(in_mux);
         THROW_ASSERT(mux_input, "classic_datapath::mux_allocation - In port does not exist");
         add_datapath_connection(HLS->HLS_T->get_technology_manager(), SM, src, mux_input);

         return;
      }
      else
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "MUX must be allocated");
         std::string name = mux_tree[i].first->get_string();
         mux = SM->add_module_from_technology_library(name, MUX_GATE_STD, HLS->HLS_T->get_technology_manager()->get_library(MUX_GATE_STD), circuit, HLS->HLS_T->get_technology_manager());
         mux_tree[i].first->set_structural_obj(mux);

         ///mux selector in datapath interface
         generic_objRef selector = GetPointer<mux_obj>(mux_tree[i].first)->get_selector();
         structural_objectRef sel_obj = selector->get_structural_obj();
         THROW_ASSERT(sel_obj, "Selector obj not created");
         ///selector in mux object
         structural_objectRef mux_sel = mux->find_member("sel", port_o_K, mux);
         THROW_ASSERT(mux_sel, "Mux selector not yet created");
         SM->add_connection(sel_obj, mux_sel);

         ///specializing allocated mux
         specialise_mux(mux_tree[i].first);

         module* mux_object = GetPointer<module>(mux);

         ///adding input connection
         structural_objectRef mux_input = mux_object->get_in_port(in_mux);
         THROW_ASSERT(mux_input, "classic_datapath::mux_allocation - In port does not exist");
         add_datapath_connection(HLS->HLS_T->get_technology_manager(), SM, src, mux_input);

         ///adding output signal to the multiplexer
         structural_objectRef port_out_mux = mux_object->get_out_port(0);
         THROW_ASSERT(port_out_mux, "classic_datapath::mux_allocation - Out port does not exist");
         structural_type_descriptorRef out_type = port_out_mux->get_typeRef();
         THROW_ASSERT(out_type->size,"size greater than one expected");
         std::string sig_name = "out_" + name;
         structural_objectRef sign = SM->add_sign(sig_name, circuit, out_type);
         mux_tree[i].first->set_out_sign(sign);
         SM->add_connection(sign, port_out_mux);
         src = sign;
      }
   }
   add_datapath_connection(HLS->HLS_T->get_technology_manager(), SM, src, tgt);
}

void conn_binding::add_datapath_connection(const technology_managerRef TM, const structural_managerRef SM, const structural_objectRef src, const structural_objectRef tgt)
{
   unsigned int bits_src = GET_TYPE_SIZE(src);
   unsigned int bits_tgt = GET_TYPE_SIZE(tgt);
   //std::cerr << "adding connection between " << src->get_path() << " and " << tgt->get_path() << std::endl;
   if (bits_src == bits_tgt)
   {
      THROW_ASSERT(src->get_owner() && tgt->get_owner(), "expected an owner for src and tgt");
      if(src->get_owner() == tgt->get_owner() && GetPointer<port_o>(src) && GetPointer<port_o>(tgt))
      {
         std::string name = "io_signal_" + src->get_id() + "_" + tgt->get_id();
         structural_type_descriptorRef sign_type = tgt->get_typeRef();
         structural_objectRef sign = SM->add_sign(name, src->get_owner(), sign_type);
         SM->add_connection(src, sign);
         SM->add_connection(sign, tgt);
      }
      else
         SM->add_connection(src, tgt);
      return;
   }

   structural_objectRef circuit = SM->get_circ();
   bool is_src_int = src->get_typeRef()->type == structural_type_descriptor::INT;
   bool is_tgt_int = tgt->get_typeRef()->type == structural_type_descriptor::INT;
   std::string name = "conv_" + src->get_id() + (is_src_int?"_I":"") + "_" + STR(bits_src) + (is_tgt_int?"_I":"") + "_" + STR(bits_tgt);
   if (converters.find(name) == converters.end())
   {
      //std::cerr << "name: " << name << std::endl;
      //std::cerr << src->get_typeRef()->get_name() << "--" << tgt->get_typeRef()->get_name() << std::endl;
      structural_objectRef c_obj;
      if(!is_src_int && !is_tgt_int)
      {
         std::string library_name = TM->get_library(UUDATA_CONVERTER_STD);
         c_obj = SM->add_module_from_technology_library(name, UUDATA_CONVERTER_STD, library_name, circuit, TM);
      }
      else if(!is_src_int && is_tgt_int)
      {
         std::string library_name = TM->get_library(UIDATA_CONVERTER_STD);
         c_obj = SM->add_module_from_technology_library(name, UIDATA_CONVERTER_STD, library_name, circuit, TM);
      }
      else if(is_src_int && !is_tgt_int)
      {
         std::string library_name = TM->get_library(IUDATA_CONVERTER_STD);
         c_obj = SM->add_module_from_technology_library(name, IUDATA_CONVERTER_STD, library_name, circuit, TM);
      }
      else
      {
         std::string library_name = TM->get_library(IIDATA_CONVERTER_STD);
         c_obj = SM->add_module_from_technology_library(name, IIDATA_CONVERTER_STD, library_name, circuit, TM);
      }
      ///fixing input stuff
      structural_objectRef in1 = GetPointer<module>(c_obj)->get_in_port(0);
      in1->type_resize(bits_src);
      SM->add_connection(src, in1);

      ///fixing output stuff
      structural_objectRef out0 = GetPointer<module>(c_obj)->get_out_port(0);
      out0->type_resize(bits_tgt);
      THROW_ASSERT(out0->get_typeRef()->size,"size greater than one expected");

      structural_objectRef sign = SM->add_sign("out_" + c_obj->get_id(), circuit, out0->get_typeRef());
      SM->add_connection(out0, sign);
      converters[name] = sign;
   }
   SM->add_connection(converters[name], tgt);

}

generic_objRef conn_binding::get_constant_obj(const std::string& value)
{
   THROW_ASSERT(value != "", "value expected");
   if (constant_values.find(value) == constant_values.end())
      constant_values[value] = generic_objRef(new dataport_obj("CONSTANT_"+value));
   return constant_values[value];
}

std::map<std::string, generic_objRef> conn_binding::get_constant_objs() const
{
   return constant_values;
}

std::string conn_binding::get_constant_string(const hlsRef HLS, unsigned int node, unsigned int precision)
{
   std::string trimmed_value;
   if(BH->is_real(node))
#if HAVE_FLOPOCO
   {
      unsigned int wF, wE;
      char *endptr;
      THROW_ASSERT(tree_helper::size(HLS->AppM->get_tree_manager(), tree_helper::get_type_index(HLS->AppM->get_tree_manager(), node)) == precision, "real precision mistmatch");
      flopoco_wrapper::DECODE_BITS(precision, wF, wE);
      std::string C_value = BH->print_constant(node);
      mpfr_t mpx;
      mpfr_init2 (mpx, static_cast<mpfr_prec_t>(wF+1));
      mpfr_strtofr (mpx, C_value.c_str(), &endptr, 10, GMP_RNDN);
      std::string flopoco_string = flopoco::fp2bin(mpx, static_cast<int>(wE), static_cast<int>(wF));
      trimmed_value = flopoco_string.substr(2);
   }
#else
   THROW_ERROR("Real constants not yet supported");
#endif
   else
   {
      tree_nodeRef ic_node = HLS->AppM->get_tree_manager()->get_tree_node_const(node);
      integer_cst *ic = GetPointer<integer_cst>(ic_node);
      unsigned long long int ull_value = static_cast<unsigned long long int>(tree_helper::get_integer_cst_value(ic));
      trimmed_value = "";
      for(unsigned int ind = 0; ind < precision; ind++)
         trimmed_value = trimmed_value + (((1LLU << (precision-ind-1)) & ull_value) ? '1' : '0');
   }
   return trimmed_value;
}


void conn_binding::add_sparse_logic(const hlsRef HLS, const structural_managerRef SM)
{
   const structural_objectRef circuit = SM->get_circ();
   std::set<generic_objRef>::const_iterator it_end = sparse_logic.end();
   structural_objectRef sparse_component;
   std::string resource_name, resource_instance_name;
   unsigned int resource_index = 0;
   unsigned int bitsize=0;
   const module* sparse_module;
   for(std::set<generic_objRef>::const_iterator it = sparse_logic.begin(); it != it_end; ++it)
   {
      switch((*it)->get_type())
      {
         case generic_obj::ADDER_CONN_OBJ:
         {
            resource_name = ADDER_STD;
            resource_instance_name = resource_name + "_adder_" + STR(resource_index);
            bitsize = GetPointer<adder_conn_obj>((*it))->get_bitsize();
            break;
         }
         case generic_obj::MULTIPLIER_CONN_OBJ:
         {
            resource_name = MULTIPLIER_STD;
            resource_instance_name = resource_name + "_multiplier_" + STR(resource_index);
            bitsize = GetPointer<multiplier_conn_obj>((*it))->get_bitsize();
            break;
         }
         case generic_obj::BITFIELD_OBJ:
         {
            resource_name = SIGNED_BITFIELD_FU_STD;
            resource_instance_name = resource_name + "_" + SIGNED_BITFIELD_FU_STD + "_" + STR(resource_index);
            bitsize = GetPointer<bitfield_obj>((*it))->get_bitsize();
            break;
         }
         default:
         {
            THROW_ERROR("sparse component not yet considered " + (*it)->get_string());
         }
      }
      ++resource_index;
      sparse_component = SM->add_module_from_technology_library(resource_instance_name, resource_name, HLS->HLS_T->get_technology_manager()->get_library(resource_name), circuit, HLS->HLS_T->get_technology_manager());
      (*it)->set_structural_obj(sparse_component);
      sparse_module = GetPointer<module>(sparse_component);

     PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, HLS->debug_level, "Specialising " +  sparse_component->get_path() + ": " + STR(bitsize));
     ///specializing sparse module ports
     sparse_module->get_in_port(0)->type_resize(bitsize);
     sparse_module->get_in_port(1)->type_resize(bitsize);
     sparse_module->get_out_port(0)->type_resize(bitsize);
   }
}

void conn_binding::print(std::ostream& os) const
{
   for (std::map<std::tr1::tuple<generic_objRef, generic_objRef, unsigned int>, connection_objRef>::const_iterator l = conn_implementation.begin(); l != conn_implementation.end(); l++)
   {
      generic_objRef src = std::tr1::get<0>(l->first);
      generic_objRef tgt = std::tr1::get<1>(l->first);
      unsigned int op = std::tr1::get<2>(l->first);

      os.width(CONN_COLUMN_SIZE);
      os.setf(std::ios_base::left, std::ios_base::adjustfield);
      os << "Source: " + src->get_string() << " ";

      os.width(CONN_COLUMN_SIZE);
      os << "Target: " + tgt->get_string() << "(" << op <<") ";

      os.width(20);
      os.setf(std::ios_base::left, std::ios_base::adjustfield);
      std::string str = l->second->get_string();
      if (l->second->get_type() == connection_obj::BY_MUX)
      {
         str = "MUX_TREE [" + STR(GetPointer<mux_conn>(l->second)->get_mux_tree_size()) + "] => " + str;
      }
      os << " connected by: " << str << std::endl;
      os.width(0);
   }
}

unsigned int conn_binding::determine_bit_level_mux() const
{
   std::set<generic_objRef> mux;
   for(conn_implementation_map::const_iterator it = conn_implementation.begin(); it != conn_implementation.end(); it++)
   {
      if (!GetPointer<mux_conn>(it->second)) continue;
      const std::vector<std::pair<generic_objRef,unsigned int> >& tree = GetPointer<mux_conn>(it->second)->get_mux_tree();
      for(unsigned int v = 0; v < tree.size(); v++)
         mux.insert(tree[v].first);
   }
   unsigned int bit_mux = 0;
   for(std::set<generic_objRef>::iterator m = mux.begin(); m != mux.end(); m++)
      bit_mux += GetPointer<mux_obj>(*m)->get_bitsize();
   return bit_mux;
}

void conn_binding::add_command_ports(const hlsRef HLS, const structural_managerRef SM)
{
   structural_objectRef circuit = SM->get_circ();

   const graphRef data = HLS->get_graph(function_behavior::DFG);

   /// define the type for boolean command signals
   structural_type_descriptorRef boolean_port_type = structural_type_descriptorRef(new structural_type_descriptor("bool", 0));

   const std::tr1::unordered_set<vertex>& operations = HLS->operations;
   for (std::tr1::unordered_set<vertex>::const_iterator j = operations.begin(); j != operations.end(); j++)
   {
      technology_nodeRef tn = HLS->ALL->get_fu(HLS->Rfu->get_assign(*j));
      technology_nodeRef op_tn = GetPointer<functional_unit>(tn)->get_operation(GET_OP(data, *j));
      THROW_ASSERT(GetPointer<operation>(op_tn)->time_m, "Time model not available for operation: " + GET_NAME(data, *j));
      if (GET_TYPE(data, *j) & TYPE_EXTERNAL || !GetPointer<operation>(op_tn)->is_bounded())
      {
         bind_selector_port(conn_binding::IN, commandport_obj::UNBOUNDED, *j, data.get());
         bind_selector_port(conn_binding::OUT, commandport_obj::UNBOUNDED, *j, data.get());
      }
   }

   std::map<structural_objectRef, std::list<structural_objectRef> > calls;
   if (selectors.find(conn_binding::IN) != selectors.end())
   {
      for(std::map<std::pair<generic_objRef, unsigned int>, generic_objRef>::const_iterator j = selectors.find(conn_binding::IN)->second.begin(); j != selectors.find(conn_binding::IN)->second.end(); j++)
      {
         //unit associate with selector
         const generic_objRef elem = j->first.first;
         //operation's index
         unsigned int oper = j->first.second;
         ///if elem is a functional unit and has more than one operation, add a selector for each operation (e.g. selector_LOAD_1...)
         switch(elem->get_type())
         {
            case generic_obj::FUNCTIONAL_UNIT:
            {
               unsigned int type_fu = GetPointer<funit_obj>(elem)->get_fu();
               std::vector<technology_nodeRef> tmp_ops_node = GetPointer<functional_unit>(HLS->ALL->get_fu(type_fu))->get_operations();
               THROW_ASSERT(GetPointer<commandport_obj>(j->second), "Not valid command port");
               structural_objectRef sel_obj = SM->add_port("fuselector_" + elem->get_string() + "_" + tmp_ops_node[oper]->get_name(), port_o::IN, circuit, boolean_port_type);
               (j->second)->set_structural_obj(sel_obj);
               structural_objectRef fu_mod = elem->get_structural_obj();
               THROW_ASSERT(fu_mod, "not correct module");
               structural_objectRef port_selector = fu_mod->find_member("sel_"+tmp_ops_node[oper]->get_name(), port_o_K, fu_mod);
               SM->add_connection(port_selector, sel_obj);
               break;
            }
            case generic_obj::REGISTER:
            {
               structural_objectRef reg_mod = elem->get_structural_obj();

               THROW_ASSERT(GetPointer<commandport_obj>(j->second), "Not valid command port");
               structural_objectRef sel_obj = SM->add_port("wrenable_" + reg_mod->get_id(), port_o::IN, circuit, boolean_port_type);
               (j->second)->set_structural_obj(sel_obj);

               structural_objectRef port_wenable = reg_mod->find_member(WENABLE_PORT_NAME, port_o_K, reg_mod);
               SM->add_connection(port_wenable, sel_obj);
               break;
            }
            case generic_obj::COMMAND_PORT:
            case generic_obj::CONNECTION_ELEMENT:
            {
               THROW_ASSERT(GetPointer<commandport_obj>(j->second), "Not valid command port");
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Adding selector_" + elem->get_string() + " " + STR(elem->get_type()));
               structural_objectRef sel_obj = SM->add_port("selector_" + elem->get_string(), port_o::IN, circuit, boolean_port_type);
               (j->second)->set_structural_obj(sel_obj);

               if (GetPointer<commandport_obj>(j->second)->get_command_type() == commandport_obj::UNBOUNDED)
               {
                  vertex op = GetPointer<commandport_obj>(j->second)->get_vertex();
                  generic_objRef fu_unit = HLS->Rfu->get(op);
                  structural_objectRef fu_obj = fu_unit->get_structural_obj();
                  structural_objectRef start = fu_obj->find_member(START_PORT_NAME, port_o_K, fu_obj);
                  calls[start].push_back(sel_obj);
               }
               break;
            }
            default:
            {
               THROW_ERROR("Not supported generic_obj " + STR(elem->get_type()));
            }
         }
      }
   }
   for (std::map<structural_objectRef, std::list<structural_objectRef> >::iterator c = calls.begin(); c != calls.end(); c++)
   {
      if (c->second.size() == 1)
      {
         SM->add_connection(c->first, c->second.front());
      }
      else
      {
         const technology_managerRef TM = HLS->HLS_T->get_technology_manager();
         std::string library = TM->get_library(OR_GATE_STD);
         structural_objectRef or_gate = SM->add_module_from_technology_library("or_" + c->first->get_owner()->get_id(), OR_GATE_STD, library, SM->get_circ(), TM);
         structural_objectRef sig = SM->add_sign("s_" + c->first->get_owner()->get_id(), SM->get_circ(), boolean_port_type);
         SM->add_connection(sig, or_gate->find_member("out1", port_o_K, or_gate));
         SM->add_connection(sig, c->first);
         structural_objectRef in = or_gate->find_member("in", port_vector_o_K, or_gate);
         port_vector_o *port = GetPointer<port_vector_o>(in);
         port->add_n_ports(static_cast<unsigned int>(c->second.size()), in);
         unsigned int num = 0;
         for(std::list<structural_objectRef>::iterator a = c->second.begin(); a != c->second.end(); a++, num++)
         {
            SM->add_connection(*a, port->get_port(num));
         }
      }
   }

   std::map<structural_objectRef, structural_objectRef> sig;
   ///output signals to the controller for condition evaluation
   if (selectors.find(conn_binding::OUT) != selectors.end())
   {
      for(std::map<std::pair<generic_objRef, unsigned int>, generic_objRef>::const_iterator j = selectors.find(conn_binding::OUT)->second.begin(); j != selectors.find(conn_binding::OUT)->second.end(); j++)
      {
         THROW_ASSERT(GetPointer<commandport_obj>(j->second), "Not valid command port");
         if(GetPointer<commandport_obj>(j->second)->get_command_type() == commandport_obj::SWITCH)
         {
            vertex op = GetPointer<commandport_obj>(j->second)->get_vertex();
            unsigned int var_written = HLS->AppM->get_produced_value(HLS->functionId, op);
            structural_type_descriptorRef switch_port_type = structural_type_descriptorRef(new structural_type_descriptor(var_written, BH));
            structural_objectRef sel_obj = SM->add_port(GetPointer<commandport_obj>(j->second)->get_string(), port_o::OUT, circuit, switch_port_type);
            (j->second)->set_structural_obj(sel_obj);
         }
         else
         {
            structural_objectRef sel_obj = SM->add_port(GetPointer<commandport_obj>(j->second)->get_string(), port_o::OUT, circuit, boolean_port_type);
            (j->second)->set_structural_obj(sel_obj);

            if (GetPointer<commandport_obj>(j->second)->get_command_type() == commandport_obj::UNBOUNDED)
            {
               vertex op = GetPointer<commandport_obj>(j->second)->get_vertex();
               generic_objRef fu_unit = HLS->Rfu->get(op);
               structural_objectRef fu_obj = fu_unit->get_structural_obj();
               structural_objectRef done = fu_obj->find_member(DONE_PORT_NAME, port_o_K, fu_obj);
               THROW_ASSERT(done, "Missing done_port from function call " + fu_unit->get_string());
               if (sig.find(done) == sig.end())
               {
                  sig[done] = SM->add_sign("s_done_" + done->get_owner()->get_id(), SM->get_circ(), boolean_port_type);
                  SM->add_connection(done, sig[done]);
               }
               SM->add_connection(sig[done], sel_obj);
            }
         }
      }
   }
}
