   /*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file cdfc_module_binding.cpp
 * @brief Class implementation of the module allocation structure.
 *
 * This file implements some of the module allocation member functions.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#include "cdfc_module_binding.hpp"
#include "config_HAVE_EXPERIMENTAL.hpp"
#if HAVE_EXPERIMENTAL
#include "controller_creator.hpp"
#endif
#include "clique_covering.hpp"

#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "behavioral_helper.hpp"
#include "behavioral_writer_helper.hpp"

#include "allocation.hpp"
#include "hls.hpp"
#include "schedule.hpp"
#include "fu_binding.hpp"
#include "reg_binding.hpp"
#include "liveness.hpp"
#include "storage_value_insertion.hpp"
#include "chaining.hpp"
#include "graph.hpp"
#include "op_graph.hpp"

#include "Parameter.hpp"
#include "utility.hpp"
#include "dbgPrintHelper.hpp"

#include <iostream>
#include <map>
#include <string>

cdfc_module_binding::cdfc_module_binding(const ParameterRef _Param, const hlsRef _HLS, type_t _clique_covering_method) :
   fu_binding_creator(_Param, _HLS), clique_covering_method(_clique_covering_method)
{

}

cdfc_module_binding::~cdfc_module_binding()
{

}
struct weight_edgeInfo : public edge_info
{
   int edge_weight;
};

class op_cdfc_writer
{
   public:
      /**
       * Constructor
       */
      op_cdfc_writer(const std::map<vertex,vertex> &_c2s, const graph *_g,  const behavioral_helperConstRef _helper) : c2s(_c2s), cdfg_writer(_g,_helper)
      {
      }

      /**
       * Operator used to print the label of a vertex
       * @param out is the stream where label has to printed
       * @param v is the vertex to be printed
       */
      void operator()(std::ostream& out, const vertex& v) const
      {
         cdfg_writer(out,c2s.find(v)->second);
      }
   private:
      const std::map<vertex,vertex> &c2s;
      cdfg_label_writer cdfg_writer;
};

class edge_cdfc_writer
{
   public:
      /**
       * Constructor
       */
      edge_cdfc_writer(const graph* _g) : g(_g) {}
      /**
       * Operator which print label of an EdgeDescriptor
       * @param out is the stream
       * @param e is the edge
       */
      void operator()(std::ostream& out, const EdgeDescriptor& e) const
      {
         const weight_edgeInfo * edge_info = Cget_edge_info<weight_edgeInfo>(e, *g);
         if(edge_info)
            out << "[label=\"" << edge_info->edge_weight << "\"]";
         else
            out << "[color=red3]";
      }
   private:
      const graph* g;
};


void cdfc_module_binding::exec()
{
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Starting module allocation based on ''control data flow chained graph''");

   // resource binding and allocation  info
   fu_binding& fu = *(HLS->Rfu);
   allocation& all = *(HLS->ALL);
   // pointer to a Control, Data dependence and antidependence graph graph
   const graph * sdg = HLS->FB->get_graph(function_behavior::SDG);
   size_t total_modules_allocated = 0;

   VertexIterator vIt, vItEnd;
   unsigned int fu_unit;
   /// compute unshared resources
   std::map<unsigned int, unsigned int> n_shared_fu;
   for(tie(vIt, vItEnd) = boost::vertices(*sdg); vIt != vItEnd; vIt++)
   {
      fu_unit = fu.get_assign(*vIt);
      if(all.is_vertex_bounded(fu_unit) || all.is_memory_unit(fu_unit)) continue;
      if(n_shared_fu.find(fu_unit) == n_shared_fu.end())
         n_shared_fu[fu_unit] = 1;
      else
         n_shared_fu[fu_unit] = 1 + n_shared_fu[fu_unit];
   }
   /// check easy binding and compute the list of vertices for which a sharing is possible
   std::list<vertex> candidate_vertices;
   for(tie(vIt, vItEnd) = boost::vertices(*sdg); vIt != vItEnd; vIt++)
   {
      if(fu.get_index(*vIt) != INFINITE_UINT) continue;
      fu_unit = fu.get_assign(*vIt);
      if(all.is_vertex_bounded(fu_unit) || all.is_memory_unit(fu_unit) || n_shared_fu.find(fu_unit)->second == 1)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Easy binding for -> " + GET_NAME(sdg, *vIt));
         ++total_modules_allocated;
         fu.bind(*vIt, fu_unit, 0);
      }
      else
         candidate_vertices.push_back(*vIt);
   }
   /// in case no vertices was left we have done
   if(!candidate_vertices.empty())
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Creating the cdfc for the module binding...");

      refcount<graphs_collection> cdfc_bulk = refcount<graphs_collection>(new graphs_collection(graph_infoRef(new op_graph_info)));

      std::map<vertex,vertex> s2c;
      std::map<vertex,vertex> c2s;

      /// add the vertices to the cdfc graph
      VertexIterator vi, vi_end;
      for (boost::tie(vi, vi_end) = boost::vertices(*sdg); vi != vi_end; ++vi)
      {
         vertex s = *vi;
         vertex c = cdfc_bulk->AddVertex();
         s2c[s] = c;
         c2s[c] = s;
      }
      /// add the control dependences edges to the cdfc graph
      EdgeIterator ei,ei_end;
      for(tie(ei,ei_end) = boost::edges(*sdg); ei != ei_end; ++ei)
      {
         vertex src = boost::source(*ei, *sdg);
         vertex tgt = boost::target(*ei, *sdg);
         if(HLS->Rchain->may_be_chained_ops(tgt, src)) /// only the chained operations are relavant
         {
            bool exists, inserted;
            vertex cdfc_src = s2c[src];
            vertex cdfc_tgt = s2c[tgt];
            EdgeDescriptor e;
            boost::tie(e, exists) = boost::edge(cdfc_src, cdfc_tgt, *cdfc_bulk);
            THROW_ASSERT(!exists, "unexpected case: edge already added");
            boost::tie(e, inserted) = boost::add_edge(cdfc_src, cdfc_tgt, EdgeProperty(CD_EDGE), *cdfc_bulk);
            THROW_ASSERT(inserted, "unexpected case: edge not added");
         }
      }

      int _w = 1;
      unsigned int fu_s1;
      unsigned int fu_s2;
      bool weighted = (HLS->Rreg != NULL);
#if HAVE_EXPERIMENTAL
      weighted = weighted and HLS->Param->getOption<unsigned int>(OPT_controller_architecture) != controller_creator::PARALLEL_CONTROLLER;
#endif

      std::list<vertex>::const_iterator cv_it_end = candidate_vertices.end();
      for(std::list<vertex>::const_iterator cv_it = candidate_vertices.begin(); cv_it != cv_it_end;)
      {
         fu_s1 = fu.get_assign(*cv_it);
         std::list<vertex>::const_iterator cv1_it = cv_it;
         ++cv_it;

         for(std::list<vertex>::const_iterator cv2_it = cv_it; cv2_it != cv_it_end; ++cv2_it)
         {
            fu_s2 = fu.get_assign(*cv2_it);
            if(fu_s1 != fu_s2) continue;
            const std::set<vertex>& op1_end = HLS->Rliv->get_state_where_end(*cv1_it);
            const std::set<vertex>& op2_end = HLS->Rliv->get_state_where_end(*cv2_it);
            const std::set<vertex>::const_iterator op1_end_it_end = op1_end.end();
            bool is_in_conflict = false;
            for(std::set<vertex>::const_iterator op1_end_it = op1_end.begin(); op1_end_it != op1_end_it_end && !is_in_conflict; ++op1_end_it)
               if(op2_end.find(*op1_end_it) != op2_end.end())
                  is_in_conflict = true;
            if(is_in_conflict) continue;

            const std::set<vertex>& op1_run = HLS->Rliv->get_state_where_run(*cv1_it);
            const std::set<vertex>& op2_run = HLS->Rliv->get_state_where_run(*cv2_it);
            const std::set<vertex>::const_iterator op1_run_it_end = op1_run.end();
            for(std::set<vertex>::const_iterator op1_run_it = op1_run.begin(); op1_run_it != op1_run_it_end && !is_in_conflict; ++op1_run_it)
               if(op2_run.find(*op1_run_it) != op2_run.end())
                  is_in_conflict = true;
            if(is_in_conflict) continue;
            // default weight for edge
            _w = 1;
            // check if you have to add weight (and calculate it)
            if (weighted)
            {
               /// an operation can end in different states
               const std::set<vertex>& end1 = HLS->Rliv->get_state_where_end(*cv1_it);
               const std::set<vertex>& end2 = HLS->Rliv->get_state_where_end(*cv2_it);
               unsigned int var_written1 = HLS->AppM->get_produced_value(HLS->functionId, *cv1_it);
               unsigned int var_written2 = HLS->AppM->get_produced_value(HLS->functionId, *cv2_it);
               std::vector<application_manager::io_binding_type> vars_read1 = HLS->AppM->get_required_values(HLS->functionId, *cv1_it);
               std::vector<application_manager::io_binding_type> vars_read2 = HLS->AppM->get_required_values(HLS->functionId, *cv2_it);

               int current_w;
               const std::set<vertex>::const_iterator e1_it_end = end1.end();
               const std::set<vertex>::const_iterator e2_it_end = end2.end();
               for(std::set<vertex>::const_iterator e1_it = end1.begin(); e1_it != e1_it_end; ++e1_it)
                  for(std::set<vertex>::const_iterator e2_it = end2.begin(); e2_it != e2_it_end; ++e2_it)
                  {
                     current_w = 1;
                     unsigned int o1, o2;

                     // save registers where operation results is stored
                     if(var_written1 != 0 && var_written2 != 0)
                     {
                        get_reg_out(var_written1, *e1_it, o1);
                        get_reg_out(var_written2, *e2_it, o2);

                        // each common register output increase edge weight
                        if (o1 == o2) current_w++;
                     }

                     // vector to contain input registers (i1 and i2)
                     std::list<unsigned int> i1, i2;
                     // save operation inputs
                     get_reg_in(vars_read1, *e1_it, *cv1_it, i1);
                     get_reg_in(vars_read2, *e2_it, *cv1_it, i2);

                     // compute common inputs
                     std::list<unsigned int>::iterator i_it1 = i1.begin();
                     std::list<unsigned int>::iterator i_it2 = i2.begin();
                     for (; i_it1 != i1.end() && i_it2 != i2.end(); i_it1++, i_it2++)
                     {
                        // each common register output increase edge weight
                        if (*i_it1 == *i_it2) current_w++;
                        //std::cerr << "in1: " << *i_it1 << " in2: " << *i_it2 << " _w: " << _w << std::endl;
                     }
                     /// chosing the max is one possible option...
                     _w = std::max(_w, current_w);
                  }
            }
            /// add compatibility edge with computed weight on both the directions
            EdgeDescriptor e;
            bool inserted;
            boost::tie(e, inserted) = boost::add_edge(s2c[*cv1_it], s2c[*cv2_it], EdgeProperty(COMPATIBILITY_EDGE), *cdfc_bulk);
            THROW_ASSERT(inserted, "unexpected case: edge selector already added");
            GET_EDGE_INFO(cdfc_bulk, weight_edgeInfo, e)->edge_weight = _w;
            boost::tie(e, inserted) = boost::add_edge(s2c[*cv2_it], s2c[*cv1_it], EdgeProperty(COMPATIBILITY_EDGE), *cdfc_bulk);
            THROW_ASSERT(inserted, "unexpected case: edge selector already added");
            GET_EDGE_INFO(cdfc_bulk, weight_edgeInfo, e)->edge_weight = _w;
         }
      }

      /// topologically sort vertex of CD_EDGE based graph
      graphRef CG = graphRef(new graph(cdfc_bulk.get(), COMPATIBILITY_EDGE));
      graphRef CD_chained_graph = graphRef(new graph(cdfc_bulk.get(), CD_EDGE));
      std::deque<vertex> sorted_vertices;
      boost::topological_sort(*CD_chained_graph, std::back_inserter(sorted_vertices));
      /// compute levels
      std::map<vertex, int> cd_levels;
      std::deque<vertex>::const_iterator sv_it_end = sorted_vertices.end();
      for(std::deque<vertex>::const_iterator sv_it = sorted_vertices.begin(); sv_it != sv_it_end; ++sv_it)
      {
         cd_levels[c2s[*sv_it]] = 0;
         InEdgeIterator ie, ie_end;
         for(tie(ie, ie_end) = boost::in_edges(*sv_it, *CD_chained_graph); ie != ie_end; ++ie)
            if(boost::in_degree(boost::source(*ie,*CD_chained_graph), *CG) != 0)
               cd_levels[c2s[*sv_it]] = std::max(cd_levels[c2s[*sv_it]], 1+cd_levels[c2s[boost::source(*ie, *CD_chained_graph)]]);
            else
               cd_levels[c2s[*sv_it]] = std::max(cd_levels[c2s[*sv_it]], cd_levels[c2s[boost::source(*ie, *CD_chained_graph)]]);
      }

      /// remove all cycles from the cdfc graph
      graphRef cdfc = graphRef(new graph(cdfc_bulk.get(), CD_EDGE|COMPATIBILITY_EDGE));

      unsigned int k=1;
      std::deque<vertex> candidate_cycle;
      bool found_a_loop;

      std::set<vertex> no_cycles;

      do
      {
         found_a_loop = false;
         std::list<vertex>::const_iterator cit_end = candidate_vertices.end();
         for(std::list<vertex>::const_iterator cit = candidate_vertices.begin(); cit != cit_end; ++cit)
         {
            if(no_cycles.find(*cit) != no_cycles.end()) continue;
            vertex start = s2c[*cit];
            if(cd_levels.find(*cit)->second != 0 && boost::in_degree(start, *CG) != 0)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Search loops starting from -> " + GET_NAME(sdg, *cit) + " iteration " + STR(k));
               bool vertex_has_a_loop = false;
               /// look for a cycle
               /// the seed is a CD_EDGE
               THROW_ASSERT(boost::in_degree(start, *CD_chained_graph) != 0, "something of unexpected happen");
               InEdgeIterator ie_cd, i_end_cd;
               for(tie(ie_cd, i_end_cd) = boost::in_edges(start, *CD_chained_graph); ie_cd != i_end_cd && !vertex_has_a_loop; ++ie_cd)
               {
                  std::map<vertex,vertex> successor_map;
                  std::deque<vertex> vertex_queue;
                  vertex seed = boost::source(*ie_cd, *CD_chained_graph);
                  vertex_queue.push_back(seed);
                  successor_map[seed] = start;
                  do
                  {
                     InEdgeIterator ie, iend;
                     vertex cur = vertex_queue.front();
                     vertex_queue.pop_front();
                     PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "vertex extracted from the queue -> " + GET_NAME(sdg, c2s[cur]));
                     for(tie(ie, iend) = boost::in_edges(cur, *cdfc); ie != iend; ++ie)
                     {
                        vertex src = boost::source(*ie,*cdfc);
                        if(successor_map.find(src) == successor_map.end())
                        {
                           if(src == start)
                           {
                              /// cycle found
                              /// fill the candidate_cycle
                              candidate_cycle.push_back(src);
                              PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Added to the cycle -> " + GET_NAME(sdg, c2s[src]));
                              while(cur != start)
                              {
                                 THROW_ASSERT(std::find(candidate_cycle.begin(), candidate_cycle.end(), cur) == candidate_cycle.end(), "something of unexpected happen " + GET_NAME(sdg, c2s[cur]));
                                 PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Added to the cycle -> " + GET_NAME(sdg, c2s[cur]));
                                 candidate_cycle.push_back(cur);
                                 THROW_ASSERT(successor_map.find(cur) != successor_map.end(), "something of unexpected happen");
                                 cur = successor_map[cur];
                              }
                              vertex_has_a_loop=true;
                              /// end the search
                              vertex_queue.clear();
                              break;
                           }
                           else
                           {
                              successor_map[src] = cur;
                              vertex_queue.push_back(src);
                           }
                        }
                     }
                  } while(!vertex_queue.empty());
               }

               if(!candidate_cycle.empty())
               {
                  found_a_loop = true;
                  /// remove the loop
                  std::deque<vertex>::const_iterator cl_it_end = candidate_cycle.end();
                  vertex cand_src = candidate_cycle.back();
                  vertex cand_tgt = candidate_cycle.front();
                  int cand_level_difference = std::abs(cd_levels[c2s[cand_src]]-cd_levels[c2s[cand_tgt]]);
                  size_t cand_out_degree = boost::out_degree(cand_src, *CG) + boost::out_degree(cand_tgt, *CG);
                  int cand_edge_weight = 0;
                  EdgeDescriptor cand_e;
                  bool cand_exists;
                  boost::tie(cand_e, cand_exists) = boost::edge(cand_src, cand_tgt, *CG);
                  if(cand_exists)
                  {
                     cand_edge_weight = CGET_EDGE_INFO(CG, weight_edgeInfo, cand_e)->edge_weight;
                     PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "analyzing compatibility between operations " + GET_NAME(sdg, c2s[cand_src]) + " and " + GET_NAME(sdg, c2s[cand_tgt]) + " - ld = " + STR(cand_level_difference) + " - d= " + STR(cand_out_degree) + " - w = " + STR(cand_edge_weight));
                  }
                  else
                     PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "analyzing compatibility between operations " + GET_NAME(sdg, c2s[cand_src]) + " and " + GET_NAME(sdg, c2s[cand_tgt]) + " - CD_EDGE ");
                  for(std::deque<vertex>::const_iterator cl_it = candidate_cycle.begin(); cl_it != cl_it_end; )
                  {
                     vertex src = *cl_it;
                     ++cl_it;
                     if(cl_it != cl_it_end)
                     {
                        vertex tgt = *cl_it;
                        EdgeDescriptor e;
                        bool exists;
                        boost::tie(e, exists) = boost::edge(src, tgt, *CG);
                        if(exists)
                        {
                           int level_difference = std::abs(cd_levels[c2s[src]]-cd_levels[c2s[tgt]]);
                           size_t out_degree = boost::out_degree(src, *CG) + boost::out_degree(tgt, *CG);
                           int edge_weight = CGET_EDGE_INFO(CG, weight_edgeInfo, e)->edge_weight;
                           PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "analyzing compatibility between operations " + GET_NAME(sdg, c2s[src]) + " and " + GET_NAME(sdg, c2s[tgt]) + " - ld = " + STR(level_difference) + " - d= " + STR(out_degree) + " - w = " + STR(edge_weight));
                           if(!cand_exists ||
                                 level_difference> cand_level_difference ||
                                 (level_difference == cand_level_difference && out_degree > cand_out_degree) ||
                                 (level_difference == cand_level_difference && out_degree == cand_out_degree && edge_weight < cand_edge_weight)
                                 )
                           {
                              cand_src = src;
                              cand_tgt = tgt;
                              cand_e = e;
                              cand_level_difference = level_difference;
                              cand_out_degree = out_degree;
                              cand_edge_weight = edge_weight;
                              cand_exists = exists;
                           }
                        }
                        else
                           PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "analyzing compatibility between operations " + GET_NAME(sdg, c2s[src]) + " and " + GET_NAME(sdg, c2s[tgt]) + " - CD_EDGE ");
                     }
                  }
                  THROW_ASSERT(cand_exists, "edge not identified");
                  /// remove both the compatibility edges
                  CG->set_selector(cand_e, 0);
                  bool exists;
                  boost::tie(cand_e, exists) = boost::edge(cand_tgt, cand_src, *CG);
                  THROW_ASSERT(exists, "edge already removed");
                  CG->set_selector(cand_e, 0);
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Removed compatibility between operations " + GET_NAME(sdg, c2s[cand_src]) + " and " + GET_NAME(sdg, c2s[cand_tgt]));
                  candidate_cycle.clear();
               }
               THROW_ASSERT(candidate_cycle.empty(), "candidate_cycle has to be empty");
               if(!vertex_has_a_loop)
                  no_cycles.insert(*cit);
            }
            else
               no_cycles.insert(*cit);
         }
         ++k;
      } while(found_a_loop);

      std::string functionName = HLS->FB->get_helper()->get_function_name();
      if (debug_level >= DEBUG_LEVEL_VERY_PEDANTIC)
      {
         cdfc->PrintDot(functionName+"_CD_COMP.dot", op_cdfc_writer(c2s,sdg,HLS->FB->get_helper()), edge_cdfc_writer(sdg));
      }

      /// partition vertices for clique covering or bind the easy functional units
      std::map<unsigned int, unsigned int> numModule;
      std::map<unsigned int, std::set<vertex> > partitions;
      std::list<vertex>::const_iterator cit_end = candidate_vertices.end();
      for(std::list<vertex>::const_iterator cit = candidate_vertices.begin(); cit != cit_end; ++cit)
      {
         fu_unit = fu.get_assign(*cit);
         /// check easy binding
         if(boost::out_degree(s2c[*cit], *CG) == 0)
         {
            unsigned int num=0;
            if(numModule.find(fu_unit) == numModule.end())
               numModule[fu_unit] = 1;
            else
               num = numModule[fu_unit]++;
            ++total_modules_allocated;
            fu.bind(*cit, fu_unit, num);
         }
         else
         {
            partitions[fu_unit].insert(s2c[*cit]);
         }
      }

      /// solve the binding problem for all the partitions
      std::map<unsigned int, std::set<vertex> >::const_iterator p_it_end = partitions.end();
      for(std::map<unsigned int, std::set<vertex> >::const_iterator p_it = partitions.begin(); p_it_end != p_it; ++p_it)
      {
         THROW_ASSERT(p_it->second.size()>1, "bad projection");
         /// build the clique covering solver
         refcount< clique_covering<vertex> > module_clique;
         switch(clique_covering_method)
         {
            case TTT_FAST:
            {
               module_clique = clique_covering<vertex>::create_solver(clique_covering<vertex>::TTT_CLIQUE_COVERING_FAST);
               break;
            }
            case TTT_FULL:
            {
               module_clique = clique_covering<vertex>::create_solver(clique_covering<vertex>::TTT_CLIQUE_COVERING);
               break;
            }
            case COLORING:
            {
               module_clique = clique_covering<vertex>::create_solver(clique_covering<vertex>::COLORING);
               break;
            }
            case WEIGHTED_COLORING:
            {
               module_clique = clique_covering<vertex>::create_solver(clique_covering<vertex>::WEIGHTED_COLORING);
               break;
            }
            case BIPARTITE_MATCHING:
            {
               module_clique = clique_covering<vertex>::create_solver(clique_covering<vertex>::BIPARTITE_MATCHING);
               break;
            }
            default:
               THROW_ERROR("This clique covering algorithm has not been implemented");
         }
         /// add vertex to the clique covering solver
         std::set<vertex>::const_iterator vert_it_end = p_it->second.end();
         for(std::set<vertex>::const_iterator vert_it = p_it->second.begin(); vert_it != vert_it_end; ++vert_it)
         {
            std::string el1_name = GET_NAME(sdg, c2s[*vert_it]) + "(" + GET_OP(sdg, c2s[*vert_it]) + ")";
            module_clique->add_vertex(c2s[*vert_it],el1_name);
         }
         /// add the edges
         EdgeIterator cg_ei,cg_ei_end;
         graphRef CG_subgraph = graphRef(new graph(cdfc_bulk.get(), COMPATIBILITY_EDGE, p_it->second));
         for(tie(cg_ei,cg_ei_end) = boost::edges(*CG_subgraph); cg_ei != cg_ei_end; ++cg_ei)
         {
            vertex src = c2s[boost::source(*cg_ei, *CG_subgraph)];
            vertex tgt = c2s[boost::target(*cg_ei, *CG_subgraph)];
            if(src > tgt) continue;/// only one edge is needed to build the undirected compatibility graph
            _w = CGET_EDGE_INFO(CG, weight_edgeInfo, *cg_ei)->edge_weight;
            module_clique->add_edge(src, tgt, _w);
         }

         if (debug_level >= DEBUG_LEVEL_VERY_PEDANTIC)
            module_clique->writeDot(functionName+"_"+all.get_string_name(p_it->first) +"_cdfc_module_binding_graph.dot");

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Compatibility graph created!");

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Starting clique covering on a graph with "+ STR(p_it->second.size()) + " vertices");

         /// performing clique covering
         module_clique->exec();

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Number of cliques covering the graph: "+ STR(module_clique->num_vertices()));
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Clique covering algorithm completed!");
         total_modules_allocated +=module_clique->num_vertices();

         /// retrieve the solution
         for (unsigned int i = 0; i < module_clique->num_vertices(); ++i)
         {
            std::set<vertex> clique = module_clique->get_clique(i);
            fu_unit = fu.get_assign(*(clique.begin()));
            THROW_ASSERT(fu_unit == p_it->first, "unexpected case");
            unsigned int num=0;
            if(numModule.find(fu_unit) == numModule.end())
               numModule[fu_unit] = 1;
            else
               num = numModule[fu_unit]++;
            std::set<vertex>::iterator v = clique.begin();
            for (; v != clique.end(); v++)
            {
               /// storing new binding results
               fu.bind(*v, fu_unit, num);
            }
         }
      }
   }
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Weighted module allocation performed!");
   PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Number of modules instantiated: " + STR(total_modules_allocated));
}

void cdfc_module_binding::get_reg_in(std::vector<application_manager::io_binding_type> &vars_read, vertex state, vertex op, std::list<unsigned int> &r_eg)
{
   r_eg.clear();
   for(unsigned int num = 0; num < vars_read.size(); num++)
   {
      unsigned int tree_var = std::tr1::get<0>(vars_read[num]);
      if(tree_var != 0)
      {
         bool found_state=false;
         vertex state_in;
         if(HLS->Rliv->has_state_in(state, op, tree_var))
         {
            const std::set<vertex>& states_in = HLS->Rliv->get_state_in(state, op, tree_var);
            const std::set<vertex>::const_iterator s_in_it_end = states_in.end();
            for(std::set<vertex>::const_iterator s_in_it = states_in.begin(); s_in_it != s_in_it_end; ++s_in_it)
            {
               state_in =  *s_in_it;
               if(state == state_in) continue;
               found_state = true;
               break;
            }
         }

         if(found_state && HLS->svi_data->is_a_storage_value(state_in, tree_var))
         {
            unsigned int storage_value = HLS->svi_data->get_storage_value_index(state_in, tree_var);
            r_eg.push_back(HLS->Rreg->get_register(storage_value));
         }
         else
            r_eg.push_back(static_cast<unsigned int>(-tree_var));
      }
      else
         r_eg.push_back(0);
   }
}

void cdfc_module_binding::get_reg_out(unsigned int var_written, vertex state, unsigned int &reg_out)
{
   if(HLS->svi_data->is_a_storage_value(state, var_written))
      reg_out = HLS->Rreg->get_register(HLS->svi_data->get_storage_value_index(state, var_written));
   else
      reg_out = static_cast<unsigned int>(-var_written);
}

std::string cdfc_module_binding::get_kind_text() const
{
   return "cdfc-module-binding";
}
