/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file fu_binding.cpp
 * @brief Class implementation of the functional-unit binding data structure.
 *
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Autoheader include
#include "config_HAVE_FLOPOCO.hpp"

#include "fu_binding.hpp"
#include "funit_obj.hpp"

#include "hls.hpp"
#include "hls_target.hpp"
#include "hls_constraints.hpp"
#include "allocation.hpp"
#include "memory.hpp"

#include "structural_manager.hpp"
#include "structural_objects.hpp"
#include "technology_builtin.hpp"

#include "Parameter.hpp"
#include "constant_strings.hpp"

#include "utility.hpp"

#include "tree_manager.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"
#include "behavioral_helper.hpp"
#include "application_manager.hpp"
#include "op_node_info.hpp"

#include "conn_binding.hpp"

#include "technology_node.hpp"
#include "technology_manager.hpp"

#include <boost/algorithm/string/replace.hpp>

#include <fstream>
#if HAVE_FLOPOCO
#include "flopoco_wrapper.hpp"
#include <mpfr.h>
namespace flopoco
{
   extern std::string fp2bin(mpfr_t x, int wE, int wF);
};
#endif

void fu_binding::bind(const vertex& v, unsigned int unit, unsigned int index)
{
   if (unique_table.count(std::make_pair(unit,index)) == 0)
      unique_table[std::make_pair(unit,index)] = generic_objRef(new funit_obj(ALL->get_string_name(unit)+"_i"+STR(index),unit,index));
   op_binding[v] = unique_table[std::make_pair(unit, index)];
   operations[std::make_pair(unit, index)].insert(v);
   if (index != INFINITE_UINT)
      update_allocation(unit, index + 1);
}

std::set<vertex> fu_binding::get_operations(unsigned int unit, unsigned int index) const
{
   if (operations.find(std::make_pair(unit, index)) == operations.end()) return std::set<vertex>();
   return operations.find(std::make_pair(unit, index))->second;
}


const funit_obj& fu_binding::operator[](const vertex& v)
{
   THROW_ASSERT(op_binding.find(v) != op_binding.end(), "vertex not preset");
   return *(GetPointer<funit_obj>(op_binding.find(v)->second));
}

std::set<unsigned int> fu_binding::get_allocation_list() const 
{
   std::set<unsigned int> allocation_list;
   std::map<unsigned int, unsigned int>::const_iterator i = allocation.begin();
   for (; i != allocation.end(); i++)
   {
      if (i->second > 0)
         allocation_list.insert(i->first);
   }
   return allocation_list;
}


void fu_binding::update_allocation(unsigned int unit, unsigned int number)
{
   if (number > allocation[unit])
      allocation[unit] = number;
}

unsigned int fu_binding::get_assign(vertex const &v) const
{
   return GetPointer<funit_obj>(op_binding.find(v)->second)->get_fu();
}

std::string fu_binding::get_fu_name(vertex const &v) const
{
   return ALL->get_string_name(get_assign(v));
}

unsigned int fu_binding::get_index(vertex const &v) const
{
   return GetPointer<funit_obj>(op_binding.find(v)->second)->get_index();
}

structural_objectRef fu_binding::add_gate(const hlsRef HLS, const technology_nodeRef fu, const std::string& _name, const std::set<vertex>& ops, structural_objectRef clock_port, structural_objectRef reset_port)
{
   int debug = HLS->debug_level;

   const graph* data = HLS->FB->get_graph(function_behavior::CFG);

   const structural_managerRef SM = HLS->datapath;
   const structural_objectRef circuit = SM->get_circ();

   std::string name = _name;
   structural_objectRef curr_gate = structural_objectRef(new component_o(debug, circuit));

   /// creating structural_manager starting from technology_node
   structural_objectRef curr_lib_instance;
   if (GetPointer<functional_unit>(fu)->fu_template_name == "")
      curr_lib_instance = GetPointer<functional_unit>(fu)->CM->get_circ();
   else
   {
      std::string template_name = GetPointer<functional_unit>(fu)->fu_template_name;
      std::string library_name = HLS->HLS_T->get_technology_manager()->get_library(template_name);
      curr_lib_instance = GetPointer<functional_unit>(GetPointer<functional_unit_template>(HLS->HLS_T->get_technology_manager()->get_fu(template_name, library_name))->FU)->CM->get_circ();
   }
   curr_lib_instance->copy(curr_gate);
   if (ops.size() == 1)
   {
      vertex operation = *(ops.begin());
      name = "fu_" + GET_NAME(data, operation);
   }
   curr_gate->set_id(name);

   ///connecting clock and reset ports, if any
   structural_objectRef port_ck = curr_gate->find_member(CLOCK_PORT_NAME, port_o_K, curr_gate);
   if(port_ck) SM->add_connection(clock_port, port_ck);
   structural_objectRef port_rst = curr_gate->find_member(RESET_PORT_NAME, port_o_K, curr_gate);
   if(port_rst) SM->add_connection(reset_port, port_rst);
   GetPointer<module>(circuit)->add_internal_object(curr_gate);

   return curr_gate;
}

void fu_binding::add_to_SM(const hlsRef HLS, structural_objectRef clock_port, structural_objectRef reset_port)
{
#ifndef NDEBUG
   int debug = HLS->debug_level;
#endif

   const structural_managerRef SM = HLS->datapath;

   const structural_objectRef circuit = SM->get_circ();

   std::set<structural_objectRef> memory_modules;

   /// add the MEMCPY_STD component when parameters has to be copied into the local storare
   const std::list<unsigned int>& parameters = HLS->FB->get_helper()->get_parameters();
   unsigned int sign_id = 0;
   structural_objectRef start_port = GetPointer<module>(circuit)->find_member(START_PORT_NAME, port_o_K, circuit);
   structural_objectRef done_port = GetPointer<module>(circuit)->find_member(DONE_PORT_NAME, port_o_K, circuit);
   structural_objectRef in_chain = start_port;
   for(std::list<unsigned int>::const_iterator p = parameters.begin(); p != parameters.end(); p++)
   {
      if(HLS->HLS_C->Rmem->is_parm_decl_copied(*p) && !HLS->HLS_C->Rmem->is_parm_decl_stored(*p))
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug, "Managing parameter copy: " + STR(*p));
         structural_objectRef curr_gate = add_gate(HLS, HLS->HLS_T->get_technology_manager()->get_fu(MEMCPY_STD, LIBRARY_STD_FU), "parameter_manager_" + STR(*p), std::set<vertex>(), clock_port, reset_port);
         manage_module_ports(SM, curr_gate, 0);
         memory_modules.insert(curr_gate);
         conn_binding::direction_type direction = conn_binding::IN;
         generic_objRef port_obj = HLS->Rconn->get_port(*p, direction);
         structural_objectRef in_par = port_obj->get_out_sign();
         structural_objectRef src = GetPointer<module>(curr_gate)->find_member("src", port_o_K, curr_gate);
         src->type_resize(STD_GET_SIZE(in_par->get_typeRef()));
         SM->add_connection(in_par, src);
         structural_objectRef dest = GetPointer<module>(curr_gate)->find_member("dest", port_o_K, curr_gate);
         dest->type_resize(STD_GET_SIZE(in_par->get_typeRef()));
         structural_objectRef dest_obj = SM->add_constant("constant_dest_" + STR(*p), circuit, dest->get_typeRef(), STR(HLS->HLS_C->Rmem->get_base_address(*p)));
         SM->add_connection(dest, dest_obj);
         structural_objectRef n = GetPointer<module>(curr_gate)->find_member("n", port_o_K, curr_gate);
         n->type_resize(STD_GET_SIZE(in_par->get_typeRef()));
         structural_objectRef n_obj = SM->add_constant("constant_n_" + STR(*p), circuit, dest->get_typeRef(), STR(tree_helper::size(HLS->AppM->get_tree_manager(), tree_helper::get_type_index(HLS->AppM->get_tree_manager(),*p))/8));
         SM->add_connection(n, n_obj);
         /// we do not use the return port but we correctly specialize to avoid synthesis warnings
         structural_objectRef return_port = GetPointer<module>(curr_gate)->find_member(RETURN_PORT_NAME, port_o_K, curr_gate);
         return_port->type_resize(STD_GET_SIZE(in_par->get_typeRef()));
         THROW_ASSERT(in_chain, "missing in chain element");
         structural_objectRef start_obj = GetPointer<module>(curr_gate)->find_member(START_PORT_NAME, port_o_K, curr_gate);
         if(in_chain == start_port)
            SM->add_connection(in_chain, start_obj);
         else
         {
            structural_objectRef sign = SM->add_sign(START_PORT_NAME+STR("_")+STR(sign_id), circuit, in_chain->get_typeRef());
            ++sign_id;
            SM->add_connection(sign,in_chain);
            SM->add_connection(sign,start_obj);
         }
         in_chain = GetPointer<module>(curr_gate)->find_member(DONE_PORT_NAME, port_o_K, curr_gate);
      }
      else if(HLS->HLS_C->Rmem->is_parm_decl_stored(*p))
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug, "Managing parameter initialization: " + STR(*p));
         structural_objectRef curr_gate = add_gate(HLS, HLS->HLS_T->get_technology_manager()->get_fu(MEMSTORE_STD, LIBRARY_STD_FU), "parameter_manager_" + STR(*p), std::set<vertex>(), clock_port, reset_port);
         manage_module_ports(SM, curr_gate, 0);
         memory_modules.insert(curr_gate);
         conn_binding::direction_type direction = conn_binding::IN;
         generic_objRef port_obj = HLS->Rconn->get_port(*p, direction);
         structural_objectRef in_par = port_obj->get_out_sign();
         structural_objectRef data = GetPointer<module>(curr_gate)->find_member("data", port_o_K, curr_gate);
         data->type_resize(STD_GET_SIZE(in_par->get_typeRef()));
         SM->add_connection(in_par, data);
         GetPointer<module>(curr_gate)->set_parameter("base_address", STR(HLS->HLS_C->Rmem->get_base_address(*p)));
         THROW_ASSERT(in_chain, "missing in chain element");
         structural_objectRef start_obj = GetPointer<module>(curr_gate)->find_member(START_PORT_NAME, port_o_K, curr_gate);
         if(in_chain == start_port)
            SM->add_connection(in_chain, start_obj);
         else
         {
            structural_objectRef sign = SM->add_sign(START_PORT_NAME+STR("_")+STR(sign_id), circuit, in_chain->get_typeRef());
            ++sign_id;
            SM->add_connection(sign,in_chain);
            SM->add_connection(sign,start_obj);
         }
         in_chain = GetPointer<module>(curr_gate)->find_member(DONE_PORT_NAME, port_o_K, curr_gate);
      }
   }
   if(in_chain)
      SM->add_connection(in_chain, done_port);
   else
      THROW_ASSERT(!done_port, "done port not connected");

   std::map<unsigned int, unsigned int> memory_units = HLS->ALL->get_memory_units();
   std::map<unsigned int, structural_objectRef> mem_obj;
   for(std::map<unsigned int, unsigned int>::iterator m = memory_units.begin(); m != memory_units.end(); ++m)
   {
      std::string name;
      if (HLS->ALL->get_fu_name(m->first).first == ARRAY_1D_STD_BRAM || HLS->ALL->get_fu_name(m->first).first == ARRAY_1D_STD_DISTRAM)
      {
         name = "array_" + STR(m->second);
      }
      else
         THROW_ERROR("Unit not currenty supported: " + HLS->ALL->get_fu_name(m->first).first);
      structural_objectRef curr_gate = add_gate(HLS, HLS->ALL->get_fu(m->first), name, operations[std::make_pair(m->first, 0)], clock_port, reset_port);

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug, "Memory Unit: " + HLS->ALL->get_string_name(m->first) + " for variable: " + HLS->FB->get_helper()->PrintVariable(m->second));
      unsigned int var = m->second;
      unsigned int base_address = HLS->HLS_C->Rmem->get_base_address(var);
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug, "  - base address: " + STR(base_address));
      specialise_fu(HLS, curr_gate, m->first, operations[std::make_pair(m->first, 0)], var);
      specialize_memory_unit(HLS, curr_gate, var, base_address);
      mem_obj[m->first] = curr_gate;
      manage_module_ports(SM, curr_gate, 0);
      memory_modules.insert(curr_gate);
   }

   std::set<unsigned int> fu_list = this->get_allocation_list();
   for (std::set<unsigned int>::iterator i = fu_list.begin(); i != fu_list.end(); i++)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug, "Functional Unit: " + HLS->ALL->get_string_name(*i));
      if (HLS->ALL->is_return(*i))
      {
         structural_objectRef obj = circuit->find_member(RETURN_PORT_NAME, port_o_K, circuit);
         generic_objRef module_obj = get(*i, 0);
         module_obj->set_structural_obj(obj);
      }

      if (!HLS->ALL->has_to_be_synthetized(*i)) continue;

      for (unsigned int num = 0; num < get_number(*i); num++)
      {
         generic_objRef module_obj = get(*i, num);
         THROW_ASSERT(module_obj, "module missing: " + HLS->ALL->get_fu_name(*i).first + " instance " + STR(num));
         std::string name = module_obj->get_string();

         structural_objectRef curr_gate;
         if(HLS->ALL->is_memory_unit(*i))
         {
            curr_gate = mem_obj[*i];
         }
         else
         {
            curr_gate = add_gate(HLS, HLS->ALL->get_fu(*i), name, operations[std::make_pair(*i,num)], clock_port, reset_port);
            const std::set<vertex>& mapped_operations = get_operations(*i, num);
            specialise_fu(HLS, curr_gate, *i, mapped_operations, 0);
            bool added_memory_element = manage_module_ports(SM, curr_gate, num);
            if (added_memory_element) memory_modules.insert(curr_gate);
         }
         module_obj->set_structural_obj(curr_gate);
      }
   }
   //std::cerr << "number of memory modules: " << memory_modules.size() << std::endl;
   manage_memory_ports(SM, memory_modules, circuit);
}

bool fu_binding::manage_module_ports(const structural_managerRef SM, const structural_objectRef curr_gate, unsigned int num)
{
   const structural_objectRef circuit = SM->get_circ();
   /// creating extern IN port on datapath starting from extern ports on module
   bool added_memory_element = false;
   for(unsigned int j = 0; j < GetPointer<module>(curr_gate)->get_in_port_size(); j++)
   {
      structural_objectRef port_in = GetPointer<module>(curr_gate)->get_in_port(j);
      manage_extern_global_port(SM, port_in, port_o::IN, circuit, num);
      if(GetPointer<port_o>(port_in)->get_is_memory())
      {
         added_memory_element = true;
      }
   }
   /// creating extern OUT port on datapath starting from extern ports on module
   for(unsigned int j = 0; j < GetPointer<module>(curr_gate)->get_out_port_size(); j++)
   {
      structural_objectRef port_out = GetPointer<module>(curr_gate)->get_out_port(j);
      manage_extern_global_port(SM, port_out, port_o::OUT, circuit, num);
   }
   /// creating extern IO port on datapath starting from extern ports on module
   for(unsigned int j = 0; j < GetPointer<module>(curr_gate)->get_in_out_port_size(); j++)
   {
      structural_objectRef port_in_out = GetPointer<module>(curr_gate)->get_in_out_port(j);
      manage_extern_global_port(SM, port_in_out, port_o::IO, circuit, num);
   }
   return added_memory_element;
}

void fu_binding::manage_memory_ports(const structural_managerRef SM, const std::set<structural_objectRef> &memory_modules, const structural_objectRef circuit)
{
   std::map<std::string, structural_objectRef> from_ports;
   std::map<std::string, structural_objectRef> primary_outs;
   structural_objectRef cir_port;
   unsigned int sign_id = 0;
   for (std::set<structural_objectRef>::iterator i = memory_modules.begin(); i != memory_modules.end(); i++)
   {
      //std::cerr << (*i)->get_path() << std::endl;
      for(unsigned int j = 0; j < GetPointer<module>(*i)->get_in_port_size(); j++)
      {
         structural_objectRef port_i = GetPointer<module>(*i)->get_in_port(j);
         if(GetPointer<port_o>(port_i)->get_is_memory() && (!GetPointer<port_o>(port_i)->get_is_global()) && (!GetPointer<port_o>(port_i)->get_is_extern()))
         {
            std::string port_name = GetPointer<port_o>(port_i)->get_id();
            THROW_ASSERT(port_name.substr(1,3)=="in_", "Expected the \"?in_\" prefix:"+port_name);
            std::string key = port_name.substr(0,1)+port_name.substr(4);///trimmed away the "?in_"prefix
            cir_port = circuit->find_member(port_name, port_i->get_kind(), circuit);
            THROW_ASSERT(!cir_port || GetPointer<port_o>(cir_port), "should be a port or null");
            if(!cir_port)
            {
               cir_port = SM->add_port(port_name, port_o::IN, circuit, port_i->get_typeRef());
               GetPointer<port_o>(cir_port)->set_is_memory(true);
               if(GetPointer<port_o>(port_i)->get_is_slave())
                  GetPointer<port_o>(cir_port)->set_is_slave(true);
               if(GetPointer<port_o>(port_i)->get_is_master())
                     GetPointer<port_o>(cir_port)->set_is_master(true);
               SM->add_connection(cir_port,port_i);
            }
            else
            {
               THROW_ASSERT(from_ports.find(key) != from_ports.end(), "somewhere the signal " + key + " should be produced");
               structural_objectRef sign = SM->add_sign(key+"_"+boost::lexical_cast<std::string>(sign_id), circuit, port_i->get_typeRef());
               ++sign_id;
               SM->add_connection(sign,port_i);
               SM->add_connection(sign,from_ports.find(key)->second);
               from_ports[key] = structural_objectRef();
            }
         }
      }
      for(unsigned int j = 0; j < GetPointer<module>(*i)->get_out_port_size(); j++)
      {
         structural_objectRef port_i = GetPointer<module>(*i)->get_out_port(j);
         if(GetPointer<port_o>(port_i)->get_is_memory() && (!GetPointer<port_o>(port_i)->get_is_global()) && (!GetPointer<port_o>(port_i)->get_is_extern()))
         {
            std::string port_name = GetPointer<port_o>(port_i)->get_id();
            THROW_ASSERT(port_name.substr(1,4)=="out_", "Expected the \"?out_\" prefix: "+port_name);
            std::string key = port_name.substr(0,1)+port_name.substr(5);///trimmed away the "?out_"prefix
            cir_port = circuit->find_member(port_name, port_i->get_kind(), circuit);
            THROW_ASSERT(!cir_port || GetPointer<port_o>(cir_port), "should be a port or null");
            if(!cir_port)
            {
               cir_port = SM->add_port(port_name, port_o::OUT, circuit, port_i->get_typeRef());
               GetPointer<port_o>(cir_port)->set_is_memory(true);
               if(GetPointer<port_o>(port_i)->get_is_slave())
                  GetPointer<port_o>(cir_port)->set_is_slave(true);
               if(GetPointer<port_o>(port_i)->get_is_master())
                     GetPointer<port_o>(cir_port)->set_is_master(true);
               primary_outs[port_name] = cir_port;
            }
            from_ports[key]=port_i;
         }
      }
   }
   std::map<std::string, structural_objectRef>::const_iterator po_end = primary_outs.end();
   for(std::map<std::string, structural_objectRef>::const_iterator po = primary_outs.begin(); po != po_end; ++po)
   {
      THROW_ASSERT(from_ports.find(po->first.substr(0,1)+po->first.substr(5)) != from_ports.end(), "Port source not present");
      SM->add_connection(po->second,from_ports.find(po->first.substr(0,1)+po->first.substr(5))->second);
   }
}

void fu_binding::manage_extern_global_port(const structural_managerRef SM, structural_objectRef port_in, port_o::port_direction dir, structural_objectRef circuit, unsigned int num)
{
   //checking flag is_extern of port_oOPT_debug_level
   //std::cerr << "fu_binding::manage_extern_global_port::port_in = " << port_in->get_path() << std::endl;
   //std::cerr << "is_extern() = " << GetPointer<port_o>(port_in)->get_is_extern() << std::endl;
   //std::cerr << "is_extern() = " << GetPointer<port_o>(port_in)->get_is_global() << std::endl;
   if(GetPointer<port_o>(port_in)->get_is_extern())
   {
      //std::cout<<"type: "<<GET_TYPE_NAME(port_in)<<"  size: "<<GET_TYPE_SIZE(port_in)<< std::endl;
      //const structural_type_descriptorRef ext_type = port_in->get_typeRef();
      structural_objectRef ext_port;
      if(GetPointer<port_o>(port_in)->get_is_global())
      {
         std::string port_name = GetPointer<port_o>(port_in)->get_id();
         ext_port = circuit->find_member(port_name, port_in->get_kind(), circuit);
         THROW_ASSERT(!ext_port || GetPointer<port_o>(ext_port), "should be a port or null");
         if(ext_port && GetPointer<port_o>(ext_port)->get_port_direction() != dir)
         {
            SM->change_port_direction(ext_port, dir, circuit);
            if(STD_GET_SIZE(ext_port->get_typeRef()) != STD_GET_SIZE(port_in->get_typeRef()))
               ext_port->type_resize(std::max(STD_GET_SIZE(ext_port->get_typeRef()), STD_GET_SIZE(port_in->get_typeRef())));
         }
         else if(!ext_port)
         {
            ext_port = SM->add_port(port_name, dir, circuit, port_in->get_typeRef());
            if(GetPointer<port_o>(port_in)->get_is_memory())
               GetPointer<port_o>(ext_port)->set_is_memory(true);
         }
         GetPointer<port_o>(ext_port)->set_is_global(true);
         //std::cerr << " . is global" << std::endl;
      }
      else
         ext_port = SM->add_port("ext_"+GetPointer<port_o>(port_in)->get_id()+"_"+boost::lexical_cast<std::string>(num), dir, circuit, port_in->get_typeRef());
      //GetPointer<port_o>(ext_port)->set_type(port_in->get_typeRef());
      //GetPointer<port_o>(ext_port)->type_resize(size_port);
      GetPointer<port_o>(ext_port)->set_is_extern(true);
      //std::cerr << " connecting with " << ext_port->get_path() << std::endl;
      SM->add_connection(port_in,ext_port);
   }
   else if(GetPointer<port_o>(port_in)->get_is_global())
   {
      structural_objectRef ext_port;
      std::string port_name = GetPointer<port_o>(port_in)->get_id();
      ext_port = circuit->find_member(port_name, port_in->get_kind(), circuit);
      THROW_ASSERT(!ext_port || GetPointer<port_o>(ext_port), "should be a port or null");
      if(!ext_port)
      {
         ext_port = SM->add_port(port_name, dir, circuit, port_in->get_typeRef());
         if(GetPointer<port_o>(port_in)->get_is_memory())
            GetPointer<port_o>(ext_port)->set_is_memory(true);
      }
      GetPointer<port_o>(ext_port)->set_is_global(true);
      SM->add_connection(port_in,ext_port);
   }
}

void fu_binding::specialise_fu(const hlsRef HLS, structural_objectRef fu_obj, unsigned int fu, const std::set<vertex>& mapped_operations, unsigned int ar)
{
   const module* fu_module = GetPointer<module>(fu_obj);
#ifndef NDEBUG
   int debug_level = HLS->debug_level;
#endif
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " Specializing " + fu_obj->get_path());
   std::map<unsigned int, unsigned int> required_variables;
   unsigned int produced_variables = 1;
   if (ar)
   {
      unsigned int elmt_bitsize=1;
      unsigned int type_index = tree_helper::get_type_index(HLS->AppM->get_tree_manager(), ar);
      tree_nodeRef type_node = HLS->AppM->get_tree_manager()->get_tree_node_const(type_index);
      greatest_bitsize(HLS->AppM->get_tree_manager(), type_node, type_index, elmt_bitsize);

      if(elmt_bitsize > 64)
         THROW_ERROR("Unexpected data size: " + STR(elmt_bitsize));
      if (HLS->ALL->get_fu_name(fu).first == ARRAY_1D_STD_BRAM || HLS->ALL->get_fu_name(fu).first == ARRAY_1D_STD_DISTRAM)
      {
         required_variables[0] = elmt_bitsize;
         required_variables[1] = 32;
         required_variables[2] = 8;
         produced_variables = elmt_bitsize;
      }
      else
         THROW_ERROR("Unit currently not supported: " + HLS->ALL->get_fu_name(fu).first);
   }
   else
   {
      const graph* data = HLS->FB->get_graph(function_behavior::CFG);
      for(std::set<vertex>::iterator op = mapped_operations.begin(); op != mapped_operations.end(); op++)
      {
         const std::vector<application_manager::io_binding_type>& vars = HLS->AppM->get_required_values(HLS->functionId, *op);
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "  vars = " + GET_OP(data, *op) + " " + STR(vars.size()) + " " + GET_NAME(data, *op));
         unsigned int out_var = HLS->AppM->get_produced_value(HLS->functionId, *op);
         if (HLS->ALL->get_fu_name(fu).first == MEMORY_STD)
         {
            unsigned int mem_var_size_in = 1;
            unsigned int mem_var_size_out = 1;

            if (GET_OP(data, *op) == STORE)
            {
               THROW_ASSERT(std::tr1::get<0>(vars[0]), "Expected a tree node in case of a value to store");
               mem_var_size_in = std::max(mem_var_size_in, tree_helper::size(HLS->AppM->get_tree_manager(), tree_helper::get_type_index(HLS->AppM->get_tree_manager(), std::tr1::get<0>(vars[0]))));
            }
            else if (GET_OP(data, *op) == LOAD)
            {
               THROW_ASSERT(out_var, "Expected a tree node in case of a value to load");
               mem_var_size_out = std::max(mem_var_size_out, tree_helper::size(HLS->AppM->get_tree_manager(), tree_helper::get_type_index(HLS->AppM->get_tree_manager(), out_var)));
            }
            ///specializing MEMORY_STD ports
            required_variables[0] = std::max(required_variables[0], mem_var_size_in);
            required_variables[1] = 32;
            required_variables[2] = 8;
            produced_variables = std::max(produced_variables, mem_var_size_out);
         }
         else if(MEMCPY_STD == GET_TYPE_NAME(fu_obj))
         {
            required_variables[0] = 32;
            required_variables[1] = 32;
            required_variables[2] = 32;
            produced_variables = 32;
         }
         else if(HLS->HLS_T->get_technology_manager()->get_library(HLS->ALL->get_fu_name(fu).first) != WORK_LIBRARY) //functions just synthesized shouldn't be customized
         {
            for(unsigned int i = 0; i < vars.size(); i++)
            {
               if (!std::tr1::get<0>(vars[i])/* || std::tr1::get<0>(vars[i]) == application_manager::CONDITION_RESULT*/) continue;
               required_variables[i] = std::max(required_variables[i], tree_helper::size(HLS->AppM->get_tree_manager(), tree_helper::get_type_index(HLS->AppM->get_tree_manager(), std::tr1::get<0>(vars[i]))));
            }
            if (out_var && out_var != application_manager::CONDITION_RESULT )
               produced_variables = std::max(produced_variables, tree_helper::size(HLS->AppM->get_tree_manager(), tree_helper::get_type_index(HLS->AppM->get_tree_manager(), out_var)));
         }
      }
   }

   unsigned int offset = 0;
   for(unsigned int i = 0; i < fu_module->get_in_port_size(); i++)
   {
      structural_objectRef port = fu_module->get_in_port(i);
      if (port->get_id() == CLOCK_PORT_NAME || port->get_id() == RESET_PORT_NAME || port->get_id() == START_PORT_NAME ) ++offset;
   }
   for(std::map<unsigned int, unsigned int>::iterator l = required_variables.begin(); l != required_variables.end(); l++)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " Specializing port in " + fu_module->get_in_port(l->first+offset)->get_path() + " " + STR(l->second));
      fu_module->get_in_port(l->first+offset)->type_resize(l->second);
   }
   offset = 0;
   for(unsigned int i = 0; i < fu_module->get_out_port_size(); i++)
   {
      structural_objectRef port = fu_module->get_out_port(i);
      if (port->get_id() == DONE_PORT_NAME) offset++;
   }
   if (offset < fu_module->get_out_port_size())
   {
      structural_objectRef port = fu_module->get_out_port(offset);
      if(port->get_typeRef()->type != structural_type_descriptor::BOOL)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, " Specializing port out " + fu_module->get_out_port(offset)->get_path() + " " + STR(produced_variables));
         port->type_resize(produced_variables);
      }
   }
   if (MEMORY_STD == GET_TYPE_NAME(fu_obj) || ARRAY_1D_STD_BRAM == GET_TYPE_NAME(fu_obj) || ARRAY_1D_STD_DISTRAM == GET_TYPE_NAME(fu_obj))
   {
      /// memory needs further specialization as a component
      structural_objectRef in1 = fu_obj->find_member("in1", port_o_K, fu_obj);
      THROW_ASSERT(in1, "Unexpected structure of MEMORY_STD/ARRAY_REF: in1");
      structural_objectRef out1 = fu_obj->find_member("out1", port_o_K, fu_obj);
      THROW_ASSERT(out1, "Unexpected structure of MEMORY_STD/ARRAY_REF: out1");
      structural_objectRef data_ram;
      if (MEMORY_STD == GET_TYPE_NAME(fu_obj))
      {
         unsigned int in1_size = STD_GET_SIZE(in1->get_typeRef());
         unsigned int out1_size = STD_GET_SIZE(out1->get_typeRef());
         data_ram = fu_obj->find_member("M_Rdata_ram", port_o_K, fu_obj);
         THROW_ASSERT(data_ram, "Unexpected structure of MEMORY_STD: M_Rdata_ram");
         data_ram->type_resize(std::max(in1_size, out1_size));
         data_ram = fu_obj->find_member("Min_Wdata_ram", port_o_K, fu_obj);
         THROW_ASSERT(data_ram, "Unexpected structure of MEMORY_STD Min_Wdata_ram");
         data_ram->type_resize(std::max(in1_size, out1_size));
         data_ram = fu_obj->find_member("Mout_Wdata_ram", port_o_K, fu_obj);
         THROW_ASSERT(data_ram, "Unexpected structure of MEMORY_STD Mout_Wdata_ram");
         data_ram->type_resize(std::max(in1_size, out1_size));
      }
      else if(ARRAY_1D_STD_BRAM == GET_TYPE_NAME(fu_obj) || ARRAY_1D_STD_DISTRAM == GET_TYPE_NAME(fu_obj))
      {
         unsigned int in1_size = STD_GET_SIZE(in1->get_typeRef());
         unsigned int out1_size = STD_GET_SIZE(out1->get_typeRef());
         data_ram = fu_obj->find_member("S_Wdata_ram", port_o_K, fu_obj);
         THROW_ASSERT(data_ram, "Unexpected structure of ARRAY_REF: S_Wdata_ram");
         data_ram->type_resize(std::max(in1_size, out1_size));
         data_ram = fu_obj->find_member("Sin_Rdata_ram", port_o_K, fu_obj);
         THROW_ASSERT(data_ram, "Unexpected structure of ARRAY_REF Sin_Rdata_ram");
         data_ram->type_resize(std::max(in1_size, out1_size));
         data_ram = fu_obj->find_member("Sout_Rdata_ram", port_o_K, fu_obj);
         THROW_ASSERT(data_ram, "Unexpected structure of ARRAY_REF Sout_Rdata_ram");
         data_ram->type_resize(std::max(in1_size, out1_size));
      }
      else
         THROW_ERROR("Unexpected MEMORY pattern");
   }
}

void fu_binding::specialize_memory_unit(const hlsRef HLS, structural_objectRef fu_obj, unsigned int ar, unsigned int base_address)
{
   module* fu_module = GetPointer<module>(fu_obj);
   /// base address specialization
   fu_module->set_parameter("base_address", boost::lexical_cast<std::string>(base_address));

   long long int vec_size=0;
   /// array ref initialization
   std::string init_filename="array_ref_"+boost::lexical_cast<std::string>(base_address)+".data";
   std::ofstream init_file(init_filename.c_str());
   unsigned int elts_size;
   fill_array_ref_memory(HLS->AppM->get_tree_manager(), init_file, ar, vec_size, elts_size, HLS->HLS_C->Rmem, GET_TYPE_NAME(fu_obj) == ARRAY_1D_STD_DISTRAM ? 8 : 4);
   THROW_ASSERT(vec_size, "at least one element is expected");
   fu_module->set_parameter("INIT_file", "\"\""+init_filename+"\"\"");

   /// specialize the number of elements in the array
   fu_module->set_parameter("n_elements", boost::lexical_cast<std::string>(vec_size));
   fu_module->set_parameter("data_size", boost::lexical_cast<std::string>(elts_size));
   //CM->set_parameter("BITSIZE_addr_ram", boost::lexical_cast<std::string>(32));

}

void fu_binding::fill_array_ref_memory(const tree_managerRef TreeM, std::ostream &init_file, unsigned int ar, long long int &vec_size, unsigned int &elts_size, const memoryRef mem, unsigned int nbyte_on_memory)
{
   unsigned int type_index;
   tree_nodeRef ar_node = TreeM->get_tree_node_const(ar);
   tree_nodeRef init_node;
   var_decl * vd = GetPointer<var_decl>(ar_node);
   if(vd && vd->init)
      init_node = GET_NODE(vd->init);
   else if(GetPointer<string_cst>(ar_node))
      init_node = ar_node;
   tree_nodeRef array_type_node = tree_helper::get_type_node(ar_node, type_index);
   //std::cerr << "fill_array_ref_memory for variable " << ar << std::endl;
   //std::cerr << " - " << array_type_node->get_kind_text() << std::endl;
   if (GetPointer<array_type>(array_type_node))
   {
      std::vector<unsigned int> dims;
      tree_helper::get_array_dimensions(TreeM, type_index, dims, elts_size);
      THROW_ASSERT(dims.size(), "something of wrong happen");
      vec_size = 1;
      for(std::vector<unsigned int>::const_iterator it = dims.begin(); it != dims.end(); ++it)
         vec_size *= *it;
   }
   else if (GetPointer<integer_type>(array_type_node) ||
            GetPointer<real_type>(array_type_node) ||
            GetPointer<enumeral_type>(array_type_node) ||
            GetPointer<pointer_type>(array_type_node) ||
            GetPointer<record_type>(array_type_node) ||
            GetPointer<union_type>(array_type_node) ||
            GetPointer<complex_type>(array_type_node))
   {
      elts_size=tree_helper::size(TreeM, type_index);
      vec_size = 1;
   }
   else
      THROW_ERROR("Type not supported: " + array_type_node->get_kind_text());

   if(init_node &&
         (
            (GetPointer<constructor>(init_node) && GetPointer<constructor>(init_node)->list_of_idx_valu.size()) ||
            (GetPointer<string_cst>(init_node) && GetPointer<string_cst>(init_node)->strg != "") ||
            (!GetPointer<constructor>(init_node) && !GetPointer<string_cst>(init_node))
         )
     )
   {
      std::vector<std::string> init_string;
      write_init(TreeM, ar_node, init_node, init_string, mem);
      std::vector<std::string> eightbit_string;
      for(unsigned int l = 0; l < init_string.size(); ++l)
         for(unsigned int base_index = 0; base_index < init_string[l].size(); base_index = base_index + 8)
            eightbit_string.push_back(init_string[l].substr(init_string[l].size()-8-base_index, 8));

      std::string str_bit;
      for(unsigned int l = 0; l < eightbit_string.size();)
      {
         str_bit ="";
         for(unsigned int counter=0; counter < nbyte_on_memory && l < eightbit_string.size(); counter++, l++)
            str_bit = eightbit_string[l] + str_bit;
         init_file << str_bit << std::endl;
      }
   }
   else
   {
      unsigned int counter=0;
      for(unsigned int i = 0; i < vec_size; ++i)
      {
         for(unsigned int j = 0; j < elts_size; ++j)
         {
            init_file << "0";
            counter++;
            if(counter%(nbyte_on_memory*8)==0)
               init_file << std::endl;
         }
      }
   }
}

void fu_binding::write_init(const tree_managerRef TreeM, tree_nodeRef var_node, tree_nodeRef init_node, std::vector<std::string>& init_file, const memoryRef mem)
{
   //std::cerr << "writing init " << init_node->get_kind_text() << std::endl;
   std::string trimmed_value;
   switch (init_node->get_kind())
   {
      case real_cst_K:
      {
#if HAVE_FLOPOCO
         unsigned int wF, wE;
         char *endptr;
         unsigned int type_index;
         tree_helper::get_type_node(init_node,type_index);
         unsigned int precision = tree_helper::size(TreeM, type_index);
         flopoco_wrapper::DECODE_BITS(precision, wF, wE);
         real_cst *rc = GetPointer<real_cst>(init_node);
         std::string C_value = rc->valr;
         mpfr_t mpx;
         mpfr_init2 (mpx, static_cast<mpfr_prec_t>(wF+1));
         mpfr_strtofr (mpx, C_value.c_str(), &endptr, 10, GMP_RNDN);
         std::string flopoco_string = flopoco::fp2bin(mpx, static_cast<int>(wE), static_cast<int>(wF));
         trimmed_value = flopoco_string.substr(2);
         init_file.push_back(trimmed_value);
#else
      THROW_ERROR("Real constants not yet supported");
#endif
      break;
      }
      case integer_cst_K:
      {
         integer_cst *ic = GetPointer<integer_cst>(init_node);
         unsigned long long int ull_value = static_cast<unsigned long long int>(tree_helper::get_integer_cst_value(ic));
         trimmed_value = "";
         unsigned int type_index;
         tree_helper::get_type_node(init_node,type_index);
         unsigned int precision = tree_helper::size(TreeM, type_index);
         for(unsigned int ind = 0; ind < precision; ind++)
            trimmed_value = trimmed_value + (((1LLU << (precision-ind-1)) & ull_value) ? '1' : '0');
         THROW_ASSERT(precision, "expected a size greater than 0");
         init_file.push_back(trimmed_value);
         break;
      }
      case constructor_K:
      {
         constructor* co = GetPointer<constructor>(init_node);
         std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::const_iterator i = co->list_of_idx_valu.begin();
         std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::const_iterator vend = co->list_of_idx_valu.end();
         bool designated_initializers_used = false;
         bool is_struct = false;
         bool is_union = false;
         unsigned int union_size=0;
         unsigned int struct_or_union_align = 0;
         std::vector<tree_nodeRef>* field_list = 0;
         ///check if designated initializers are really used
         tree_nodeRef firstnode = (i != vend) ? co->list_of_idx_valu.begin()->first : tree_nodeRef();
         if (firstnode && GET_NODE(firstnode)->get_kind() == field_decl_K)
         {
            field_decl * fd = GetPointer<field_decl>(GET_NODE(firstnode));
            tree_nodeRef scpe = GET_NODE(fd->scpe);

            if (scpe->get_kind() == record_type_K)
            {
               field_list = &(GetPointer<record_type>(scpe)->list_of_flds);
               struct_or_union_align = GetPointer<record_type>(scpe)->algn;
               is_struct =  true;
            }
            else if (scpe->get_kind() == union_type_K)
            {
               field_list = &(GetPointer<union_type>(scpe)->list_of_flds);
               is_union = true;
               union_size = tree_helper::size(TreeM,GET_INDEX_NODE(fd->scpe));
            }
            else
               THROW_ERROR("expected a record_type or a union_type");
            std::vector<tree_nodeRef>::const_iterator flend = field_list->end();
            std::vector<tree_nodeRef>::const_iterator fli = field_list->begin();
            for (; fli != flend && i != vend; i++, fli++)
            {
               if (i->first && GET_INDEX_NODE(i->first) != GET_INDEX_NODE(*fli))
                  break;
            }
            if (fli != flend && is_struct)
               designated_initializers_used = true;
         }

         if(designated_initializers_used)
         {
            THROW_ASSERT(field_list, "something of wrong happen");
            std::vector<tree_nodeRef>::const_iterator flend = field_list->end();
            std::vector<tree_nodeRef>::const_iterator fli = field_list->begin();
            i = co->list_of_idx_valu.begin();
            for (; fli != flend; fli++)
            {
               if (i != vend && GET_INDEX_NODE(i->first) == GET_INDEX_NODE(*fli))
               {
                  write_init(TreeM, GET_NODE(i->first), GET_NODE(i->second), init_file, mem);
                  i++;
               }
               else
               {
                  write_init(TreeM, GET_NODE(*fli), GET_NODE(*fli), init_file, mem);
               }
               if(!GetPointer<field_decl>(GET_NODE(*fli))->is_bitfield()
                     && (fli+1 != flend) /// alignment is not performed on the last field
                     )
               {
                  /// check if padding is needed
                  unsigned int type_index;
                  tree_helper::get_type_node(GET_NODE(*fli),type_index);
                  unsigned int field_decl_size = tree_helper::size(TreeM, type_index);
                  if(field_decl_size%struct_or_union_align != 0)
                  {
                     /// add padding
                     unsigned int nbits = struct_or_union_align - (field_decl_size%struct_or_union_align);
                     std::string init_string;
                     for(unsigned int j = 0; j < nbits; ++j)
                        init_string += "0";
                     init_file.push_back(init_string);
                  }
               }
            }
         }
         else
         {
            for(i = co->list_of_idx_valu.begin(); i != vend; i++)
            {
               write_init(TreeM, GET_NODE(i->first), GET_NODE(i->second), init_file, mem);
               if(is_struct  && !GetPointer<field_decl>(GET_NODE(i->first))->is_bitfield()
                   && (i+1 != vend) /// alignment is not performed on the last field
                     )
               {
                  /// check if padding is needed
                  unsigned int type_index;
                  tree_helper::get_type_node(GET_NODE(i->first),type_index);
                  unsigned int field_decl_size = tree_helper::size(TreeM, type_index);
                  if(field_decl_size%struct_or_union_align != 0)
                  {
                     /// add padding
                     unsigned int nbits = struct_or_union_align - (field_decl_size%struct_or_union_align);
                     std::string init_string;
                     for(unsigned int j = 0; j < nbits; ++j)
                        init_string += "0";
                     init_file.push_back(init_string);
                  }
               }
               else if(is_union)
               {
                  /// check if padding is needed
                  THROW_ASSERT(co->list_of_idx_valu.size() == 1, "just one itializer is possible");
                  unsigned int type_index;
                  tree_helper::get_type_node(GET_NODE(i->first),type_index);
                  unsigned int field_decl_size = tree_helper::size(TreeM, type_index);
                  if(field_decl_size != union_size)
                  {
                     /// add padding
                     unsigned int nbits = union_size - field_decl_size;
                     std::string init_string;
                     for(unsigned int j = 0; j < nbits; ++j)
                        init_string += "0";
                     init_file.push_back(init_string);
                  }
               }
            }
         }
         unsigned int type_index;
         tree_nodeRef type_n = tree_helper::get_type_node(var_node, type_index);
         if (GetPointer<array_type>(type_n))
         {
            unsigned int size_of_data;
            std::vector<unsigned int> dims;
            tree_helper::get_array_dimensions(TreeM, type_index, dims, size_of_data);
            unsigned int num_elements=dims[0];
            std::string value;
            THROW_ASSERT(num_elements>=co->list_of_idx_valu.size(), "not correctly initialized array" + STR(co->list_of_idx_valu.size()) + "-" + STR(num_elements));
            num_elements = num_elements - static_cast<unsigned int>(co->list_of_idx_valu.size());
            for (unsigned int l = 0; l < num_elements; l++)
            {
               value = "";
               for (unsigned int ii = 0; ii < size_of_data; ii++)
                  value += "0";
               init_file.push_back(value);
            }
         }
         break;
      }
      case string_cst_K:
      {
         string_cst *sc=GetPointer<string_cst>(init_node);
         std::string string_value=sc->strg;
         std::string tmp;
         for(unsigned int index=0; index < string_value.size(); ++index)
         {
            if(string_value[index] == '\\' && string_value[index+1] == '0')
            {
               tmp += '\0';
               ++index;
            }
            else
               tmp += string_value[index];
         }
         string_value = tmp;
         boost::replace_all(string_value, "\\a", "\a");
         boost::replace_all(string_value, "\\b", "\b");
         boost::replace_all(string_value, "\\t", "\t");
         boost::replace_all(string_value, "\\n", "\n");
         boost::replace_all(string_value, "\\v", "\v");
         boost::replace_all(string_value, "\\f", "\f");
         boost::replace_all(string_value, "\\r", "\r");
         boost::replace_all(string_value, "\\\"", "\"");
         unsigned int elmt_bitsize;
         std::vector<unsigned int> dims;

         tree_helper::get_array_dimensions(TreeM, GET_INDEX_NODE(sc->type), dims, elmt_bitsize);
         if(elmt_bitsize == 32) //wide char used
         {
            THROW_ERROR("wide char conversion not supported");
         }
         else
         {
            for(unsigned int j = 0; j < string_value.size(); j++)
            {
               unsigned long int ull_value = static_cast<unsigned long int>(string_value[j]);
               trimmed_value = "";
               for(unsigned int ind = 0; ind < elmt_bitsize; ind++)
                  trimmed_value = trimmed_value + (((1LLU << (elmt_bitsize-ind-1)) & ull_value) ? '1' : '0');
               init_file.push_back(trimmed_value);
            }
         }
         trimmed_value = "";
         for(unsigned int ind = 0; ind < elmt_bitsize; ind++)
            trimmed_value = trimmed_value + '0';
         init_file.push_back(trimmed_value);
         unsigned int type_index;
         tree_nodeRef type_n = tree_helper::get_type_node(var_node, type_index);
         THROW_ASSERT(GetPointer<array_type>(type_n), "expected an array_type");
         unsigned int size_of_data;
         dims.clear();
         tree_helper::get_array_dimensions(TreeM, type_index, dims, size_of_data);
         THROW_ASSERT(size_of_data == elmt_bitsize, "something of wrong happen");
         unsigned int num_elements=1;
         std::string value;
         for(std::vector<unsigned int>::const_iterator it = dims.begin(); it != dims.end(); ++it)
            num_elements *= *it;
         THROW_ASSERT(num_elements>=(string_value.size()+1), "not correctly initialized array" + STR(string_value.size()+1) + "-" + STR(num_elements));
         num_elements = num_elements - static_cast<unsigned int>(string_value.size()+1);
         for (unsigned int l = 0; l < num_elements; l++)
         {
            value = "";
            for (unsigned int ii = 0; ii < size_of_data; ii++)
               value += "0";
            init_file.push_back(value);
         }
         break;
      }
      case nop_expr_K:
      {
         nop_expr * ne = GetPointer<nop_expr>(init_node);
         if(GetPointer<addr_expr>(GET_NODE(ne->op)))
         {
            write_init(TreeM, GET_NODE(ne->op), GET_NODE(ne->op), init_file, mem);
         }
         else
            THROW_ERROR("Something of unexpected happened: " + GET_NODE(ne->op)->get_kind_text());
         break;
      }
      case addr_expr_K:
      {
         addr_expr* ae = GetPointer<addr_expr>(init_node);
         tree_nodeRef addr_expr_op = GET_NODE(ae->op);
         unsigned int addr_expr_op_idx = GET_INDEX_NODE(ae->op);
         unsigned long long int ull_value=0;
         unsigned int type_index;
         tree_helper::get_type_node(init_node,type_index);
         unsigned int precision = tree_helper::size(TreeM, type_index);
         switch (addr_expr_op->get_kind())
         {
            case ssa_name_K:
            case var_decl_K:
            case parm_decl_K:
            case string_cst_K:
            {
               THROW_ASSERT(mem->has_base_address(addr_expr_op_idx), "missing base address for: @" + STR(addr_expr_op_idx));
               ull_value = mem->get_base_address(addr_expr_op_idx);
               break;
            }
            case array_ref_K:
            {
               array_ref* ar = GetPointer<array_ref>(addr_expr_op);
               tree_nodeRef aridx = GET_NODE(ar->op1);
               if(aridx->get_kind() == integer_cst_K && GetPointer<integer_cst>(aridx))
               {
                  switch (GET_NODE(ar->op0)->get_kind())
                  {
                     case ssa_name_K:
                     case var_decl_K:
                     case parm_decl_K:
                     case string_cst_K:
                     {
                        unsigned int step = tree_helper::size(TreeM, tree_helper::get_type_index(TreeM,addr_expr_op_idx)) /8;
                        ull_value = mem->get_base_address(GET_INDEX_NODE(ar->op0)) + step*static_cast<unsigned int>(tree_helper::get_integer_cst_value(GetPointer<integer_cst>(aridx)));
                        break;
                     }
                     default:
                        THROW_ERROR("addr_expr-array_ref[0] pattern not supported: " + std::string(addr_expr_op->get_kind_text())+ " @" + STR(addr_expr_op_idx));
                  }
               }
               else
                  THROW_ERROR("addr_expr-array_ref[0] pattern not supported: " + std::string(addr_expr_op->get_kind_text())+ " @" + STR(addr_expr_op_idx));
               break;
            }
            default:
               THROW_ERROR("addr_expr pattern not supported: " + std::string(addr_expr_op->get_kind_text())+ " @" + STR(addr_expr_op_idx));
         }
         for(unsigned int ind = 0; ind < precision; ind++)
            trimmed_value = trimmed_value + (((1LLU << (precision-ind-1)) & ull_value) ? '1' : '0');
         init_file.push_back(trimmed_value);

         break;
      }
      case field_decl_K:
      {
         unsigned int type_index;
         tree_helper::get_type_node(init_node,type_index);
         unsigned int field_decl_size = tree_helper::size(TreeM, type_index);
         std::string init_string;
         for(unsigned int j = 0; j < field_decl_size; ++j)
            init_string += "0";
         if(field_decl_size)
            init_file.push_back(init_string);
         break;
      }
      default:
         THROW_ERROR("elements not yet supported: "+init_node->get_kind_text());
   }
}


void fu_binding::greatest_bitsize(const tree_managerRef TreeM, tree_nodeRef type_node, unsigned int type_index, unsigned int &bitsize)
{
   switch (type_node->get_kind())
   {
      case array_type_K:
      {
         array_type * atype =GetPointer<array_type>(type_node);
         greatest_bitsize(TreeM, GET_NODE(atype->elts), GET_INDEX_NODE(atype->elts), bitsize);
         break;
      }
      case record_type_K:
      {
         record_type * rt =GetPointer<record_type>(type_node);
         std::vector<tree_nodeRef>field_list = rt->list_of_flds;
         std::vector<tree_nodeRef>::const_iterator flend = field_list.end();
         for (std::vector<tree_nodeRef>::const_iterator fli = field_list.begin(); fli != flend; fli++)
         {
            greatest_bitsize(TreeM, GET_NODE(*fli), GET_INDEX_NODE(*fli), bitsize);
         }
         break;
      }
      case union_type_K:
      {
         union_type * ut =GetPointer<union_type>(type_node);
         std::vector<tree_nodeRef>field_list = ut->list_of_flds;
         std::vector<tree_nodeRef>::const_iterator flend = field_list.end();
         for (std::vector<tree_nodeRef>::const_iterator fli = field_list.begin(); fli != flend; fli++)
         {
            greatest_bitsize(TreeM, GET_NODE(*fli), GET_INDEX_NODE(*fli), bitsize);
         }
         break;
      }
      case field_decl_K:
      {
         unsigned int fd_type_index;
         tree_nodeRef fd_type_node = tree_helper::get_type_node(type_node, fd_type_index);
         greatest_bitsize(TreeM, fd_type_node, fd_type_index, bitsize);
         break;
      }
      case complex_type_K:
      {
         bitsize= std::max(bitsize, tree_helper::size(TreeM, type_index)/2); /// it is composed by two identical parts
         break;
      }
      case real_type_K:
      case integer_type_K:
      case enumeral_type_K:
      case pointer_type_K:
      {
         bitsize= std::max(bitsize, tree_helper::size(TreeM, type_index));
         break;
      }
      default:
         THROW_ERROR("elements not yet supported: "+type_node->get_kind_text());
   }

}
