/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file fu_binding.hpp
 * @brief Data structure used to store the functional-unit binding of the vertexes.
 *
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef FU_BINDING_HPP
#define FU_BINDING_HPP

#include <iostream>
#include <map>

#include "Vertex.hpp"

#include "funit_obj.hpp"


#include "refcount.hpp"
#include "utility.hpp"


/**
* @name forward declarations
*/
//@{
REF_FORWARD_DECL(hls);
REF_FORWARD_DECL(allocation);
REF_FORWARD_DECL(tree_manager);
REF_FORWARD_DECL(tree_node);
REF_FORWARD_DECL(memory);
//@}

/**
 * Class managing the functional-unit binding.
 * It stores the functional-unit binding, that is, the mapping of operations in the behavioral description onto the set of
 * allocated functional units.
 */
class fu_binding
{

   private:

      /// map between functional unit id and number of units allocated
      std::map< unsigned int, unsigned int >allocation;

      /// map between unit and allocated objects
      std::map< std::pair<unsigned int, unsigned int>, generic_objRef > unique_table;

      ///reverse map that associated each functional unit with the set of operations that are executed
      std::map< std::pair<unsigned int, unsigned int>, std::set<vertex> > operations;

      ///operation binding
      std::map<vertex, generic_objRef> op_binding;

      /**
       * Update number of allocated units. It takes effect only if new number is greather than stored one
       * @param unit is the identifier associated to functional unit
       * @param number is the new number of allocated units
       */
       void update_allocation(unsigned int unit, unsigned int number);

       ///allocation manager. Used to retrieve the string name of the functional units.
       const allocationRef ALL;

       /**
         * fill the memory of the array ref
         * @param TreeM is the tree_manager
         * @param filestream is the file where the data is written
         * @param ar is the array ref variable declaration
         * @param vec_size is the number of the element of the array
         * @param elts_size is the element size in bits
        */
       void fill_array_ref_memory(const tree_managerRef TreeM, std::ostream &init_file, unsigned int ar, long long int &vec_size, unsigned int &elts_size, const memoryRef mem, unsigned int nbyte_on_memory);

       /**
         * Manage the connections between memory ports
         */
       void manage_memory_ports(const structural_managerRef SM, const std::set<structural_objectRef> &memory_modules, const structural_objectRef circuit);

       /**
        * Add an instance of the current port
        */
       structural_objectRef add_gate(const hlsRef HLS, const technology_nodeRef fu, const std::string& name, const std::set<vertex>& operations,
                                     structural_objectRef clock_port, structural_objectRef reset_port);

       void greatest_bitsize(const tree_managerRef TreeM, tree_nodeRef type_node, unsigned int type_index, unsigned int &bitsize);

   public:

      /**
       * Constructor.
       */
       fu_binding(const allocationRef &ALL_) : ALL(ALL_){}

      /**
       * Destructor.
       */
      ~fu_binding() {}

      /**
       * Binds an operation vertex to a functional unit. The functional unit is identified by an id and
       * its index
       * @param v is operation vertex
       * @param id is the identifier of the functional unit
       * @param index is the functional unit index
       */
      void bind(const vertex &v, unsigned int id, unsigned int index = INFINITE_UINT);

      /**
       * Returns the functional unit assigned to the vertex.
       * @param v is the considered vertex.
       * @return the functional unit assigned with the vertex.
       */
      unsigned int get_assign(const vertex &v) const;

      /**
       * Returns the index of functional unit assigned to the vertex.
       * @param v is the considered vertex.
       * @return the index of the functional unit assigned to the vertex.
       */
      unsigned int get_index(const vertex &v) const;

      /**
       * Returns the name of the functional unit
       * @param v is the considered vertex.
       * @return the name of the functional unit assigned to the vertex.
       */
      std::string get_fu_name(vertex const &v) const;

      /**
       * Returns number of functional unit allocated
       * @param unit is the identifier associated to functional unit
       * @return the number of functional units allocated.
       */
      unsigned int get_number(unsigned int unit) const
      {
         std::map<unsigned int, unsigned int>::const_iterator it = allocation.find(unit);
         if (it != allocation.end())
            return it->second;
         else
            return 0;
      }

      /**
       * Redefinition of the [] operator. It is necessary because segfaults happen when the vertex is not into the map
       * and so the object has not been created yet. This operator can be used only to read informations since it returns
       * a constant object. It's necessary because a direct manipulation of object can create data inconsistence with
       * objects.
       * @param v is the vertex you want to get the object
       * @return the constant object related to vertex
       */
      const funit_obj& operator[](const vertex& v);

      /**
       * Returns reference to funit object associated with this vertex
       * @param v is the given vertex
       * @return the associated reference
       */
      generic_objRef get(vertex const& v) const
      {
         return op_binding.find(v)->second;
      }

      generic_objRef get(unsigned int name, unsigned int index)
      {
         return unique_table.count(std::make_pair(name,index)) ?
                   unique_table[std::make_pair(name,index)] :
                   generic_objRef();
      }

      /**
       * Returns the set of allocated unit
       * @return the set of allocated unit
       */
      std::set<unsigned int> get_allocation_list() const;

      /**
       * Instance the functional unit inside the structural representation of the datapath
       */
      void add_to_SM(const hlsRef HLS, structural_objectRef clock_port, structural_objectRef reset_port);

      void manage_extern_global_port(const structural_managerRef SM, structural_objectRef port_in, port_o::port_direction dir, structural_objectRef circuit, unsigned int num);

      /**
       * Return the operations that are executed by the given functional unit
       */
      std::set<vertex> get_operations(unsigned int unit, unsigned int index) const;

      /**
       * Specialise the functional unit based on variables associated with the corresponding operations
       */
      void specialise_fu(const hlsRef HLS, structural_objectRef fu_obj, unsigned int fu, const std::set<vertex>& operations, unsigned int ar);

      /**
       * Specialize a memory unit
       */
      void specialize_memory_unit(const hlsRef HLS, structural_objectRef fu_obj, unsigned int ar, unsigned int base_address);

      static
      void write_init(const tree_managerRef TreeM, tree_nodeRef var_node, tree_nodeRef init_node, std::vector<std::string>& init_file, const memoryRef mem);

      bool manage_module_ports(const structural_managerRef SM, const structural_objectRef curr_gate, unsigned int num);
};

/**
 * RefCount type definition of the fu_binding class structure
 */
typedef refcount<fu_binding> fu_bindingRef;

#endif
