/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file compatibility_based_register.hpp
 * @brief Base class specification for register allocation algorithm based on a compatibility graph
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef COMPATIBILITY_BASED_REGISTER_HPP
#define COMPATIBILITY_BASED_REGISTER_HPP

#include "reg_binding_creator.hpp"

#include "refcount.hpp"
/**
 * @name Forward declarations
 */
//@{
REF_FORWARD_DECL(storage_value_insertion);
REF_FORWARD_DECL(Parameter);
REF_FORWARD_DECL(hls);
//@}

#include <boost/graph/adjacency_list.hpp>
#include <vector>

class compatibility_based_register : public reg_binding_creator
{
   protected:

      /// edge property
      struct edge_compatibility_property
      {
         /// edge weight
         int weight;

         /**
             * Constructor with selector
             * @param _weight is the weight to be associated with the edge
             */
         edge_compatibility_property(int _weight) :
            weight(_weight) {}

      };

      typedef boost::adjacency_list<boost::vecS, boost::vecS, boost::undirectedS, boost::no_property, edge_compatibility_property > compatibility_graph;
      typedef boost::graph_traits<compatibility_graph>::vertex_descriptor CG_vertex_descriptor;
      typedef boost::graph_traits<compatibility_graph>::vertices_size_type CG_vertices_size_type;
      
      ///compatibility graph
      compatibility_graph CG;

      ///ordered vector containing the vertices of the compatibility graph
      std::vector<CG_vertex_descriptor> verts;

   public:

      /**
       * Constructor
       */
      compatibility_based_register(const ParameterRef _Param, const hlsRef HLS, const storage_value_insertionRef _svi_algorithm);

      /**
       * Destructor
       */
      virtual ~compatibility_based_register();

      /**
       * Creates the compatibility graph
       */
      void create_compatibility_graph();

      /**
       * Checks if two storage values are compatible
       */
      bool is_compatible(unsigned int sv1, unsigned int sv2) const;

      /**
       * This class only computes the compatibility graph. Thus, the exec method has nothing to do
       */
      virtual void exec();

      /**
       * Returns the name of the algorithm
       */
      virtual std::string get_kind_text() const;

};
///refcount definition of the class
typedef refcount<compatibility_based_register> compatibility_based_registerRef;

#endif
