/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file storage_value_insertion.hpp
 * @brief This package is used to define the storage value scheme adopted by the register allocation algorithm.
 *
 * @defgroup Storage Value Insertion
 * @ingroup HLS
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef STORAGE_VALUE_INSERTION_HPP
#define STORAGE_VALUE_INSERTION_HPP

#include "HLS_step.hpp"

#include "Vertex.hpp"

#include "refcount.hpp"
#include "config_HAVE_EXPERIMENTAL.hpp"

/**
 * @name Forward declarations.
*/
//@{
REF_FORWARD_DECL(storage_value_insertion);
REF_FORWARD_DECL(hls);
REF_FORWARD_DECL(Parameter);
REF_FORWARD_DECL(dataflow_analysis);
class xml_element;
//@}


/**
 * Base class used to define the priority associated 
 * with each vertex of a list base scheduling problem.
*/
class storage_value_insertion : public HLS_step
{
   protected:

      /// current number of storage values
      unsigned int number_of_storage_values;

      /// verbosity level
      int output_level;

   public:

      /// storage value insertion algorithms
      enum storage_value_insertion_type
      {
         VALUES = 0      //! Default solution, it just uses the results of the dataflow analysis
                          //! since we use single statement assignments (SSA) it is equivalent to the scheme 3(VALUES)-2(VARIABLES) described in the Stok's paper.
#if HAVE_EXPERIMENTAL
         ,
         VALUES_INSTANCES,
         LIMITED_VALUES_INSTANCES,
         PAULIN_SCHEME
#endif
      };

      /**
       * This is the constructor of the storage_value_insertion class.
       */
      storage_value_insertion(const ParameterRef Param, const hlsRef HLS);

      /**
       * Destructor.
       */
      virtual ~storage_value_insertion();

      /**
       * Creates solver based on selected option
       * @param type is the index of desidered storage value insertion solver
       */
      static
      storage_value_insertionRef create_solver(storage_value_insertion_type type, const ParameterRef Param, const hlsRef HLS);
      
      /**
       * Creates solver based on the configuration stored in the XML node
       * @param node is the XML node
       */
      static
      storage_value_insertionRef xload(const xml_element *node, const ParameterRef Param, const hlsRef HLS);

      /**
       * Returns the number of storage values inserted
       */
      unsigned int get_number_of_storage_values() const;

      /**
       * return true in case a storage value exist for the pair vertex variable
       * @param curr_vertex is the vertex
       * @param var_index is the variable
      */
      virtual bool is_a_storage_value(vertex curr_vertex, unsigned int var_index) const = 0;

      /**
       * Function that computes the storage value insertion data str
       */
      virtual void exec() = 0;

      /**
       * Returns the index of the storage value associated with the variable in a given vertex
       * @param curr_vertex is the vertex
       * @param var_index is the variable
       */
      virtual unsigned int get_storage_value_index(vertex curr_vertex, unsigned int var_index) = 0;
      
      /**
       * Returns the index of the variable associated with the storage value in a given vertex
       */
      virtual unsigned int get_variable_index(unsigned int storage_value_index) = 0;

      /**
       * return a weight that estimate how much two storage values are compatible.
       * An high value returned means an high compatibility between the two storage values.
       */
      virtual int get_compatibility_weight(unsigned int storage_value_index1, unsigned int storage_value_index2) = 0;

      /**
       * return the bitsize of a storage value
       * @param storage_value_index is the storage value
       */
      virtual unsigned int get_storage_value_bitsize(unsigned int storage_value_index) = 0;
};

#endif
