/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file chaining.hpp
 * @brief class supporting the chaining optimization in high level synthesis
 * @author Vito Giovanni Castellana <vcastellana@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
 */

#ifndef CHAINING_HPP
#define CHAINING_HPP

#include <map>

#include "refcount.hpp"
#include "HLS_step.hpp"
#include "graph.hpp"

/**
 * @name forward declarations
 */
REF_FORWARD_DECL(hls);
REF_FORWARD_DECL(Parameter);
REF_FORWARD_DECL(chaining);


class chaining : public HLS_step
{
   public:

      typedef enum
      {
         STG_CHAINING = 0,
         CDFG_CHAINING
      } chaining_algorithm;

   protected:

      /// vector storing all the chained vertices
      /// the vector contains only vertices that do not occur in chaining_with map as keys
      std::vector<vertex> chained_vertices;

      /// the key is a chained vertex (SAOD,DFG ops), values are vectors of vertices that the key is chained with
      /// only vertices scheduled in the same clock cycle (of the key) are stored
      std::map<vertex, std::vector<vertex> > chaining_with;

      /// store all pairs of vertices that may be in chaining
      std::map<vertex, std::set<vertex> > may_be_chained;

      /// key of the first map: a state
      /// key of the second map: an operation
      /// set: operations in chaining with the key
      std::map<vertex, std::map<vertex, std::set<vertex> > > chained_in_state;


   public:

      /**
       * Constructor
       */
      chaining(const ParameterRef Param, const hlsRef HLS);

      /**
       * Destructor
       */
      ~chaining();

      /**
       * return true in case the vertex is in chaining with something
       * @param v is an SADG/DFG node (operation)
       */
      bool is_chained_vertex(vertex v);


      /**
        * returns the chaining predecessors of an operation (DFG/SAODG node)
        */
      std::vector<vertex> get_chaining_with_vertices(vertex op);

      /**
        * returns true iff op is in chaining with source
      */
      bool is_chained_with_vertex(vertex tgt, vertex src);

      /**
       * check if two operations are chained in at least one state
       * @param op1 is the first vertex
       * @param op2 is the second vertex
       */
      bool may_be_chained_ops(vertex op1, vertex op2);

      /**
       * check if two operations are chained in the given state
       * @param state is the state in which the chaining condition is tested
       * @param op1 is the first vertex
       * @param op2 is the second vertex
       */
      bool are_chained_in_state(vertex state, vertex op1, vertex op2);

      /**
       * Factory method
       */
      static
      chainingRef create_solver(chaining_algorithm algorithm, const ParameterRef Param, const hlsRef HLS);

};

typedef refcount<chaining> chainingRef;

#endif
