/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file hls.hpp
 * @brief Data structure definition for high-level synthesis flow.
 *
 * This struct contains all informations useful to high-level synthesis flow.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef HLS_HPP
#define HLS_HPP

#include <iostream>
#include <map>
#include <set>
#include <string>

#include "refcount.hpp"
#include "dbgPrintHelper.hpp"

#include "graph.hpp"
#include "function_behavior.hpp"

/**
 * @name forward declarations
 */
//@{
REF_FORWARD_DECL(graph);
// problem specifics
REF_FORWARD_DECL(application_manager);
REF_FORWARD_DECL(function_behavior);
REF_FORWARD_DECL(HLS_constraints);
REF_FORWARD_DECL(HLS_target);
// intermediate results
REF_FORWARD_DECL(allocation);
REF_FORWARD_DECL(chaining);
REF_FORWARD_DECL(schedule);
REF_FORWARD_DECL(fu_binding);
REF_FORWARD_DECL(StateTransitionGraph);
REF_FORWARD_DECL(liveness);
REF_FORWARD_DECL(storage_value_insertion);
REF_FORWARD_DECL(reg_binding);
REF_FORWARD_DECL(conn_binding);
// structural representation
REF_FORWARD_DECL(structural_manager);
REF_FORWARD_DECL(structural_type_descriptor);
// self forward declaration
REF_FORWARD_DECL(hls);
// bambu parameters
REF_FORWARD_DECL(Parameter);
/// forward decl of xml Element
class xml_element;
//@}

#include "utility.hpp"


/**
 * @class hls
 * @ingroup HLS
 * Data structure that contains all informations about high level synthesis process
 */
struct hls
{
   ///reference to the application manager
   const application_managerRef AppM;

   ///this is the identifier of the function to be implemented
   unsigned int functionId;

   ///reference to the function to be implemented
   const function_behaviorRef FB;

   /// Set representing the subset of operations in the specification to be implemented
   const std::tr1::unordered_set<vertex> operations;

   // -------------- High Level Synthesis Specific -------------- //

   ///reference to the information representing the target for the synthesis
   const HLS_targetRef HLS_T;

   ///store the HLS constraints
   const HLS_constraintsRef HLS_C;

   // -------------- High Level Synthesis Results -------------- //

   /// Store the technology information
   allocationRef ALL;

   /// Store the refcounted scheduling of the operations.
   scheduleRef Rsch;

   /// Store the refcounted functional unit binding of the operations.
   fu_bindingRef Rfu;

   /// Store the refcounted state transition graph
   StateTransitionGraphRef STG;

   /// datastructure containing the variable liveness
   livenessRef Rliv;

   /// datastructure for storage values
   storage_value_insertionRef svi_data;

   /// Store the refcounted register binding of the variables.
   reg_bindingRef Rreg;

   /// Store the refcounted register group binding of the variables.
   reg_bindingRef RregGroup;

   /// Store the refcounted interconnection of datapath elements.
   conn_bindingRef Rconn;

   /// Store the refcounted chaining info
   chainingRef Rchain;


   // -------------- High Level Synthesis structural representation -------------- //

   /// Store the datapath description.
   structural_managerRef datapath;

   /// Store the controller description.
   structural_managerRef controller;

   /// Store the top description.
   structural_managerRef top;

   // -------------- Parameters -------------- //

   ///class containing all the parameters
   const ParameterRef Param;

   ///debugging level of the class
   int debug_level;

   ///verbosity level of the class
   int output_level;

   // -------------- Constructor & Destructor -------------- //

   /**
    *
    */
   hls(const ParameterRef Param,
       const application_managerRef AppM, unsigned int function_id, const std::tr1::unordered_set<vertex>& operations,
       const HLS_targetRef HLS_T, const HLS_constraintsRef HLS_C);

   /**
    * Destructor
    */
   ~hls();

   /**
    * Returns the subgraph associated with the operations to be synthesized
    * @param type is the type of the graph to be returned;
    */
   const graphRef get_graph(function_behavior::graph_type type) const;

   // -------------- Printing aids -------------- //

   /**
    * Prints the hls solution available up to now. It prints informations about:
    *   - scheduling
    *   - functional units binding
    *   - total number of control steps
    *   - register binding (if register allocation has been performed)
    *   - connection binding (if it has been performed)
    * @param os is the stream where the informations have to be printed
    */
   void print(std::ostream& os) const;

    /**
    * Prints on stream the scheduling (if it has been performed). It prints informations about
    *   - Operation name
    *   - Control step where it will start its execution
    *   - Functional unit where it has been mapped
    * @param os is the stream where the informations have to be printed
    */
   void print_scheduling(std::ostream& os) const;

  /**
    * Prints on stream the register binding (if it has been performed). It prints informations about
    *   - Variable name
    *   - Operations which variable is live beetween
    *   - Register where the variable has been stored
    * @param os is the stream where the informations have to be printed
    */
   void print_register_binding(std::ostream& os) const;

   /**
    * Prints on stream the register grouping (if it has been performed). It prints informations about
    *   - Variable name
    *   - Operations which variable is live beetween
    *   - Register where the variable has been stored
    * @param os is the stream where the informations have to be printed
    */
   void print_register_grouping(std::ostream& os) const;

   /**
    * Prints on stream the connection binding (if it has been performed). It prints informations about
    *   - Connection source and target
    *   - Element used to perform connection
    *   - Variables that could cross the connection
    * @param os is the stream where the informations have to be printed
    */
   void print_connection_binding(std::ostream& os) const;

   /**
    * Prints the summmary of allocated resources
    * @param os is the stream where the information has to be printed out
    */
   void print_resources(std::ostream& os) const;

   /**
    * Friend definition of the << operator.
    */
   friend std::ostream& operator<<(std::ostream& os, const hls& s)
   {
      s.print(os);
      return os;
   }

   void xload(const xml_element* rootnode);

   void xwrite(xml_element* rootnode);

   structural_type_descriptorRef get_data_type(unsigned int variable) const;

};

/// RefCount type definition of the hls class structure
typedef refcount<hls> hlsRef;

#endif
