/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file hls_constraints.hpp
 * @brief Data structure definition for high-level synthesis flow.
 *
 * This struct contains all informations useful to high-level synthesis flow.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef HLS_CONSTRAINTS_HPP
#define HLS_CONSTRAINTS_HPP

#include "constraint_manager.hpp"

#include <iostream>
#include <map>
#include <set>
#include <string>

#include "refcount.hpp"
/**
 * @name forward declarations
*/
//@{
REF_FORWARD_DECL(HLS_constraints);
REF_FORWARD_DECL(memory);
/// forward decl of xml Element
class xml_element;
//@}

#include "utility.hpp"
#include "graph.hpp"
#include <tr1/unordered_map>

/// macro used to convert the functional unit name and the library in an unique string.
#define ENCODE_FU_LIB(fu_name, library) fu_name+":"+library

/**
 * @class HLS_constraints
 * @ingroup HLS
 * Data structure used to store all the HLS contraints
 */
class HLS_constraints : public constraint_manager
{
      /// For each functional unit tech_constraints stores the number of resources.
      std::tr1::unordered_map<std::string, unsigned int> tech_constraints;

      ///put into relation the vertex name with the functional unit, its library and the functional unit index
      ///this map can be used to define a binding for a vertex.
      std::tr1::unordered_map<std::string, std::pair<std::string, std::pair<std::string, unsigned int> > > binding_constraints;

      ///map between an operation vertex and its given scheduling priority value
      std::tr1::unordered_map<std::string, int> scheduling_constraints;

   public:

      friend class allocation;

      /// Clock period in ns.
      double clock_period;

      /// Max power consumption per cycle
      double max_power_per_cycle;

      /// Variable storing the number of registers.
      unsigned int regis;

      /// Variable storing the number of busses.
      unsigned int bus;

      /// Variable storing the number of multiplexers.
      unsigned int mux;

      /// base address for memory space addressing
      unsigned int base_address;

      ///information about memory allocation
      memoryRef Rmem;

      /**
       * Default constructor. It sets by default clock period to 10.0 and maximum power for cycle, number of register,
       * buses and muxes to infinite value.
       */
      HLS_constraints(const ParameterConstRef Param);

      /**
       * Sets the maximum number of resources available in the design
       * @param name is the name of the functional unit.
       * @param library is the library of name.
       * @param n_resources is the maximum number of fu available (default is infinite).
      */
      void set_number_fu(const std::string &name, const std::string & library, unsigned int n_resources = std::numeric_limits<unsigned int>::max());

      /**
       * This method returns the maximum number of functional units available for the design.
       * @param name is the name of the functional unit.
       * @param library is the library of name.
       * @return the maximum number of functional units available.
      */
      unsigned int get_number_fu(const std::string &name, const std::string & library) const;

      /**
       * This method returns the maximum number of functional units available for the design.
       * @param combined is the name of the functional unit in the form <unit_name>:<library_name>
       * @return the maximum number of functional units available.
      */
      unsigned int get_number_fu(const std::string &combined) const;

      /**
       * Binds a vertex to a functional unit.
       * @param vertex_name is the name of the vertex.
       * @param fu_name is the name of the functional unit.
       * @param fu_library is the name of the library of the functional unit fu_name.
       * @param fu_index is the ith functional unit.
      */
      void bind_vertex_to_fu(const std::string &vertex_name, const std::string & fu_name, const std::string & fu_library, const unsigned int fu_index);

      /**
       * Binds a vertex to a functional unit.
       *
       */
      bool has_binding_to_fu(const std::string &vertex_name) const;

      /**
       * Store the priority value for an operation vertex
       * @param Ver is the operation vertex_name
       * @param Priority is the priority value related to the operation vertex
      */
      void set_scheduling_priority(const std::string &vertex_name, int Priority);

      /**
       * Return the datastructure containing scheduling priority constraints
       * @return a map between the operation vertex and its priority value
      */
      std::tr1::unordered_map<std::string,int> get_scheduling_priority() const;

      /**
       * Set the clock period.
       * @param period is the clock period
      */
      void set_clock_period(double period);

      /**
       * Get the clock period.
       * @return the clock period
      */
      double get_clock_period() const { return clock_period;}

      /**
       * Set the maximum power consumption.
       * @param power is the maximum power available
      */
      void set_max_power(double power);

      /**
       * Return the maximum power consumption per cycle
       * @return a double containing the maximum power consumption per cycle
      */
      double get_max_power() const { return max_power_per_cycle;}

      /**
       * Set the number of registers present in the design.
       * @param n_resources is the maximum number of registers available
      */
      void set_max_regis(unsigned int n_resources);

      /**
       * Returns the number of registers available.
       * @return the maximum number of register
      */
      unsigned int get_max_regis() const{ return regis;}

      /**
       * Set the number of busses present in the design.
       * @param n_resources is the maximum number of buses available
      */
      void set_max_bus(unsigned int n_resources);

      /**
       * Returns the number of buses available.
       * @return the maximum number of buses
      */
      unsigned int get_max_bus() const { return bus;}

      /**
       * Set the number of muxes present in the design.
       * @param n_resources is the maximum number of muxes available
      */
      void set_max_mux(unsigned int n_resources);

      /**
       * Returns the number of muxes available.
       * @return the maximum number of muxes
      */
      unsigned int get_max_mux() const { return mux;}

      /**
       * Function that prints the class HLS_constraints.
       * @param os is the output stream.
      */
      void print(std::ostream& os) const;

      /**
       * Load a high level synthesis constraints starting from an xml file.
       * @param Enode is a node of the xml tree.
       * @param HLS_C is the HLS_constraints data structure.
      */
      void xload(const xml_element* Enode, HLS_constraintsRef const & HLS_C);

      /**
       * write the HLS_constraints data structure on an xml tree.
       * @param rootnode is the root node at which the xml representation of the HLS_constraints is attached.
      */
      void xwrite(xml_element* rootnode);

      void xwriteFUConstraints(xml_element* rootnode, const std::string& fu_name, 
                               unsigned int number_fu, const std::string& fu_library, bool forDump = false);

      void xwriteHLSConstraints(xml_element* rootnode, bool forDump = false);

      void xwriteBindingConstraints(xml_element* rootnode, const std::string& vertex_name, const std::string& fu_name, 
                                    unsigned int fu_index, const std::string& fu_library);

      void xwriteSchedulingConstraints(xml_element* Enode, const std::string& vertex_name, int priority);

      ///definition of the get_kind_text method
      GET_KIND_TEXT(HLS_constraints);

  private:
    static const double clock_period_DEFAULT /*= 10.0*/;
};

/**
 * RefCount type definition of the HLS_constraints class structure
 */
typedef refcount<HLS_constraints> HLS_constraintsRef;

/**
 * Define the default constraints for the builtin resources
 */
void add_builtin_constraints(HLS_constraintsRef const & HLS_C);

#endif
