/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file hls_flow.hpp
 * @brief Base class for the high-level synthesis flow.
 *
 * This struct contains all the information useful to create a high-level synthesis flow.
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef HLS_FLOW_HPP
#define HLS_FLOW_HPP

///Autoheader include
#include "config_HAVE_BEAGLE.hpp"
#include "config_HAVE_EXPERIMENTAL.hpp"

#include <iostream>
#include <map>
#include <set>
#include <string>

#include "refcount.hpp"

/**
 * @defgroup HLS High-Level Synthesis
 * This group contains files that are used to implement high-level synthesis process.
 *
 * This package contains different subproject:
 * - \ref allocation this is a package to manage information about allocation and binding constraints
 *
 * To get more details into this process see \ref src_HLS_page
 */


/**
 * @name forward declarations
 */
//@{
REF_FORWARD_DECL(hls);
REF_FORWARD_DECL(hls_flow);
REF_FORWARD_DECL(Parameter);
REF_FORWARD_DECL(application_manager);
REF_FORWARD_DECL(HLS_target);
REF_FORWARD_DECL(HLS_constraints);
//@}

#include "graph.hpp"
#include <tr1/unordered_set>

/**
 * @class hls_flow
 * @ingroup HLS
 * Data structure that contains all the information about the HLS process
 */
struct hls_flow
{
   /// reference to the register where all options have been stored
   const ParameterRef Param;

   /// debug level
   const int debug_level;

   /// verbosity level
   const unsigned int output_level;

   /// HLS datastructures
   const hlsRef HLS;

   ///implementations resulting from the synthesis
   std::set<hlsRef> resulting_implementations;

   // -------------- High Level Synthesis flow -------------- //
   /// hls flow supported
   enum flow_type
   {
      STANDARD,
      VIRTUAL,
#if HAVE_EXPERIMENTAL
      XML_CONFIG,
      EXPLORE_MUX,
      FU_REG_BINDING,
      DUMP,
#endif
#if HAVE_BEAGLE
      DSE,
#endif
   };

   /**
    * Constructor
    * @param Param is the set of parameters
    * @param AppM is the manager of the application to be implemented
    */
   hls_flow(const ParameterRef Param, const hlsRef HLS);

   /**
    * Destructor
    */
   virtual ~hls_flow();


   // -------------- Execution methods -------------- //

   /**
    * Creates desidered high level synthesis flow specialization. You have to give the desidered design flow identifier.
    * @param type is the desidered high level synthesis flow descriptor
    * @param Param is the reference to the parameter register
    * @param HLS is the class containing the specification of the HLS synthesis
    */
   static
   hls_flowRef create_flow(flow_type type, const ParameterRef Param, const hlsRef HLS);

   /**
    * Creates the HLS flow starting from the given specification
    */
   static
   hlsRef create_base_HLS(const ParameterRef Param,
                          const application_managerRef AppM, unsigned int functionId,
                          const HLS_targetRef HLS_T, const HLS_constraintsRef HLS_C);

   /**
    * This method starts your desidered high-level synthesis flow. You have to specialize it to provide your desidered
    * flow.
    * @param hls_flow is the reference to hls datastructure to allow specialization to pass this reference as parameter
    */
   virtual void exec() = 0;

   /**
    * Returns all the implementations resulting from the synthesis
    */
   std::set<hlsRef> get_all_implementations() const;

};
/// RefCount type definition of the hls class structure
typedef refcount<hls_flow> hls_flowRef;

#endif
