/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file standard_hls.cpp
 * @brief Implementation of the methods to create the structural description of the component
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10622 $
 * $Date: 2012-03-14 09:23:58 +0100 (Wed, 14 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#include "standard_hls.hpp"

///Autoheader include
#include "config_HAVE_EXPERIMENTAL.hpp"

#include "datapath_creator.hpp"
#include "controller_creator.hpp"
#include "top_entity.hpp"
#if HAVE_EXPERIMENTAL
#include "evaluation.hpp"
#endif
#include "Parameter.hpp"
#include "dbgPrintHelper.hpp"
#include "cpu_time.hpp"

standard_hls::standard_hls(const ParameterRef _Param, const hlsRef _HLS) :
   hls_flow (_Param, _HLS)
{

}

standard_hls::~standard_hls()
{

}

void standard_hls::exec()
{
   if (output_level >= OUTPUT_LEVEL_VERBOSE)
   {
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "************************************************");
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "*                                              *");
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "*      HIGH LEVEL SYNTHESIS STANDARD FLOW      *");
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "*                                              *");
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "************************************************");
   }

   hls_flow::flow_type synthesis_flow = hls_flow::VIRTUAL;
#if HAVE_EXPERIMENTAL
   if (Param->isOption("explore-mux") && Param->getOption<bool>("explore-mux"))
      synthesis_flow = hls_flow::EXPLORE_MUX;
   if (Param->isOption("explore-fu-reg") && Param->getOption<bool>("explore-fu-reg"))
      synthesis_flow = hls_flow::FU_REG_BINDING;
#endif
   hls_flowRef virt_flow = hls_flow::create_flow(synthesis_flow, Param, HLS);
   virt_flow->exec();

   if (Param->getOption<bool>(OPT_datapath))
   {
      datapath_creator::datapath_type dt = static_cast<datapath_creator::datapath_type>(Param->getOption<unsigned int>(OPT_datapath_architecture));
      datapath_creatorRef DC = datapath_creator::create_creator(dt, Param, HLS);
      long datapath_time;
      START_TIME(datapath_time);
      DC->exec();
      STOP_TIME(datapath_time);
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Datapath creation (" + DC->get_kind_text() + ") completed in " + print_cpu_time(datapath_time) + " seconds");
   }

   if (Param->getOption<int>(OPT_controller))
   {
      controller_creator::controller_type ct = static_cast<controller_creator::controller_type>(Param->getOption<unsigned int>(OPT_controller_architecture));
      controller_creatorRef CC = controller_creator::create_creator(ct, Param, HLS);
      long controller_time;
      START_TIME(controller_time);
      CC->exec();
      STOP_TIME(controller_time);
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Controller creation (" + CC->get_kind_text() + ") completed in " + print_cpu_time(controller_time) + " seconds");
   }

   if (Param->getOption<bool>(OPT_datapath) && Param->getOption<bool>(OPT_datapath))
   {
      long top_time;
      START_TIME(top_time);
      top_entityRef TOP = top_entityRef(new top_entity(Param, HLS));
      TOP->exec();
      STOP_TIME(top_time);
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Top entity creation (" + TOP->get_kind_text() + ") completed in " + print_cpu_time(top_time) + " seconds");
   }

#if HAVE_EXPERIMENTAL
   if (Param->getOption<int>(OPT_evaluation))
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "\nHigh-level synthesis statistics:");
      evaluationRef estimate = evaluationRef(new evaluation(Param, HLS));
      estimate->exec();
   }
#endif

   PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "");
   resulting_implementations.insert(HLS);
   return;
}
