/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file memory.cpp
 * @brief Class for representing memory information
 *
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#include "memory.hpp"

#include "tree_helper.hpp"
#include "funit_obj.hpp"

#include "exceptions.hpp"
#include "utility.hpp"

/// we start to allocate from 8th byte to allign address to 64bit
/// we can't use address 0 since C assumes that no object is allocated at that address
#define internal_base_address_DEFAULT 8

memory::memory(unsigned int off_base_address) :
      next_base_address(internal_base_address_DEFAULT),
      next_off_base_address(off_base_address)
{

}

memory::~memory()
{

}

void memory::add_internal_variable(unsigned int funID_scope, unsigned int var)
{
   if (external.find(var) != external.end())
      THROW_WARNING("The variable " + STR(var) + " has been already set as off-chip");
   internal[funID_scope].insert(var);
}

void memory::add_external_variable(const tree_managerRef & TreeM, unsigned int var)
{
   external.insert(var);
   set_external_base_address(var, next_off_base_address);
   /// compute the next base address
   unsigned int size = tree_helper::size(TreeM, tree_helper::get_type_index(TreeM, var)) / 8;
   next_off_base_address += size;
}

unsigned int memory::get_memory_address() const
{
   return next_off_base_address;
}

bool memory::is_internal_variable(unsigned int funID_scope, unsigned int var) const
{
   return internal.find(funID_scope) != internal.end() && internal.find(funID_scope)->second.find(var) != internal.find(funID_scope)->second.end();
}

void memory::set_internal_base_address(unsigned int var, unsigned int base_address)
{
   internal_base_address_binding[var] = base_address;
}

void memory::set_external_base_address(unsigned int var, unsigned int base_address)
{
   external_base_address_binding[var] = base_address;
}

void memory::add_internal_base_address(const tree_managerRef & TreeM, unsigned int var)
{
   THROW_ASSERT(internal_base_address_binding.find(var) == internal_base_address_binding.end(), "internal_base_address for variable already allocated");
   /// set the base address
   set_internal_base_address(var, next_base_address);
   /// compute the next base address
   unsigned int size = tree_helper::size(TreeM, tree_helper::get_type_index(TreeM, var)) / 8;
   next_base_address += size;
}

unsigned int memory::get_internal_base_address(unsigned int var) const
{
   THROW_ASSERT(internal_base_address_binding.find(var) != internal_base_address_binding.end(), "Variable not yet allocated");
   return internal_base_address_binding.find(var)->second;
}

unsigned int memory::get_external_base_address(unsigned int var) const
{
   THROW_ASSERT(external_base_address_binding.find(var) != external_base_address_binding.end(), "Variable not yet allocated");
   return external_base_address_binding.find(var)->second;
}

std::set<unsigned int> memory::get_function_vars(unsigned int funID_scope) const
{
   if (internal.find(funID_scope) == internal.end())
      return std::set<unsigned int>();
   return internal.find(funID_scope)->second;
}

bool memory::has_internal_base_address(unsigned int var) const
{
   return internal_base_address_binding.find(var) != internal_base_address_binding.end();
}

bool memory::has_external_base_address(unsigned int var) const
{
   return external_base_address_binding.find(var) != external_base_address_binding.end();
}

bool memory::has_base_address(unsigned int var) const
{
   return external_base_address_binding.find(var) != external_base_address_binding.end() || internal_base_address_binding.find(var) != internal_base_address_binding.end();
}

unsigned int memory::get_base_address(unsigned int var) const
{
   THROW_ASSERT(has_base_address(var), "Variable not yet allocated: @" + STR(var));
   if (has_internal_base_address(var)) return get_internal_base_address(var);
   return get_external_base_address(var);
}

std::set<unsigned int> memory::get_memory_variables() const
{
   return external;
}

void memory::reserve_space(unsigned int space)
{
   //std::cerr << next_off_base_address << std::endl;
   next_off_base_address += space;
   //std::cerr << next_off_base_address << std::endl;
}

unsigned int memory::get_allocated_space() const
{
   return (next_base_address - internal_base_address_DEFAULT);
}

bool memory::is_parm_decl_copied(unsigned int var) const
{
   return parm_decl_copied.find(var) != parm_decl_copied.end();
}

void memory::add_parm_decl_copied(unsigned int var)
{
   parm_decl_copied.insert(var);
}

bool memory::is_parm_decl_stored(unsigned int var) const
{
   return parm_decl_stored.find(var) != parm_decl_stored.end();
}

void memory::add_parm_decl_stored(unsigned int var)
{
   parm_decl_stored.insert(var);
}

bool memory::is_actual_parm_loaded(unsigned int var) const
{
   return actual_parm_loaded.find(var) != actual_parm_loaded.end();
}

void memory::add_actual_parm_loaded(unsigned int var)
{
   actual_parm_loaded.insert(var);
}
