/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file memory.hpp
 * @brief Datastructure to represent memory information in high-level synthesis
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

#ifndef _MEMORY_HPP_
#define _MEMORY_HPP_

#include "refcount.hpp"

/**
 * @name forward declarations
*/
//@{
REF_FORWARD_DECL(tree_manager);
REF_FORWARD_DECL(generic_obj);
//@}

#include <set>
#include <map>

class memory
{
      ///set of variables allocated in on-chip memories
      std::map<unsigned int, std::set<unsigned int> > internal;

      ///set of variables allocated in off-chip memories
      std::set<unsigned int> external;

      /// Store the base address of the associated functional unit
      std::map<unsigned int, unsigned int> internal_base_address_binding;

      /// Store the base address of the associated functional unit
      std::map<unsigned int, unsigned int> external_base_address_binding;

      /// parm_decl that has to be copied from the caller
      std::set<unsigned int> parm_decl_copied;

      /// parm_decl storage has to be initialized from the formal parameter
      std::set<unsigned int> parm_decl_stored;

      /// actual parameter that has to be loaded from a stored value
      std::set<unsigned int> actual_parm_loaded;

      /// represent the next memory address
      unsigned int next_base_address;

      /// represent the next external base address
      unsigned int next_off_base_address;

   public:

      /**
       * Constructor
       */
      memory(unsigned int off_base_address);

      /**
       * Destructor
       */
      ~memory();

      /**
       * Add a variable to the set of on-chip variables
       */
      void add_internal_variable(unsigned int funID_scope, unsigned int var);

      /**
       * Add a variable to the set of on-chip variables
       */
      void add_external_variable(const tree_managerRef & TreeM, unsigned int var);

      /**
       * Add a variable to the set of on-chip variables
       */
      bool is_internal_variable(unsigned int funID_scope, unsigned int var) const;

      /**
       * Set the base address for an on-chip variable
       */
      void set_internal_base_address(unsigned int var, unsigned int base_address);

      /**
       * Set the memory address for an off-chip variable
       */
      void set_external_base_address(unsigned int var, unsigned int base_address);

      /**
       * Add an internal_base_address for the given variable
       */
      void add_internal_base_address(const tree_managerRef & TreeM, unsigned int var);

      /**
       * Get the current base address of the given variable
       */
      unsigned int get_internal_base_address(unsigned int var) const;

      /**
       * Get the current base address of the given variable
       */
      unsigned int get_external_base_address(unsigned int var) const;

      /**
       * Get the current base address of the given variable
       */
      unsigned int get_base_address(unsigned int var) const;

      /**
       * Check if there is a base address for the given variable
       */
      bool has_internal_base_address(unsigned int var) const;

      /**
       * Check if there is a base address for the given variable
       */
      bool has_external_base_address(unsigned int var) const;

      /**
       * Check if there is a base address for the given variable
       */
      bool has_base_address(unsigned int var) const;

      /**
       * return true in case the parm_decl parameter has to be copied from the caller
       */
      bool is_parm_decl_copied(unsigned int var) const;

      /**
       * add a parm_decl to the set of parm_decl written
       */
      void add_parm_decl_copied(unsigned int var);

      /**
       * return true in case the parm_decl parameter has to be initialized from the formal value
       */
      bool is_parm_decl_stored(unsigned int var) const;

      /**
       * add a parm_decl to the set of parm_decl that has to be initialized
       */
      void add_parm_decl_stored(unsigned int var);

      /**
       * return true in case the actual parameter has to be initialized from a stored value
       */
      bool is_actual_parm_loaded(unsigned int var) const;

      /**
       * add an actual parameter to the set of parameter that has to be initialized from a stored value
       */
      void add_actual_parm_loaded(unsigned int var);


      /**
       * Return the variables allocated within the space of a given function
       */
      std::set<unsigned int> get_function_vars(unsigned int funID_scope) const;

      /**
       * Return the first memory address
       */
      unsigned int get_memory_address() const;

      /**
       * Return external variables
       */
      std::set<unsigned int> get_memory_variables() const;

      void reserve_space(unsigned int space);

      unsigned int get_allocated_space() const;

};

typedef refcount<memory> memoryRef;

#endif
