/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file memory_allocation.cpp
 * @brief Memory allocation based on the dominator tree of the call graph.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#include "memory_allocation.hpp"

#include "memory.hpp"
#include "hls.hpp"
#include "hls_constraints.hpp"

#include "call_graph.hpp"
#include "call_graph_helper.hpp"
#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "behavioral_helper.hpp"

#include "tree_manager.hpp"
#include "tree_reindex.hpp"
#include "tree_node.hpp"

#include "Parameter.hpp"
#include "constant_strings.hpp"
#include "cpu_time.hpp"
#include "Dominance.hpp"

memory_allocation::memory_allocation(const ParameterRef _Param,
                                     const application_managerRef _AppM,
                                     const HLS_constraintsRef _HLS_C) :
   Param(_Param),
   debug_level(_Param->getOption<int>(OPT_debug_level)),
   output_level(_Param->getOption<int>(OPT_output_level)),
   AppM(_AppM),
   HLS_C(_HLS_C),
   allocation_algorithm(static_cast<type_t>(_Param->getOption<unsigned int>(OPT_memory_allocation_algorithm)))
{

}

memory_allocation::~memory_allocation()
{

}

void memory_allocation::exec()
{
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "   - Performing memory allocation...");
   const tree_managerRef TreeM = AppM->get_tree_manager();
   const call_graphRef CG = AppM->get_call_graph();

   long memory_time;
   START_TIME(memory_time);

   /// the analysis has to be performed only on the reachable functions
   const graph *cg = CG->get_graph();
   unsigned int top_function = TreeM->function_index(Param->getOption<std::string>(OPT_top_function_name));
   vertex top_vertex = CG->get_vertex(top_function);
   std::list<unsigned int> sort_list;
   //call the depth-first-search
   called_functions_visitor vis(sort_list, AppM);
   std::vector<boost::default_color_type> color_vec(boost::num_vertices(*cg));
   boost::depth_first_visit(*cg, top_vertex, vis, boost::make_iterator_property_map(color_vec.begin(), boost::get(boost::vertex_index_t(), *cg), boost::white_color));
   std::set<vertex> vertex_subset;
   std::list<unsigned int>::iterator It, End = sort_list.end();
   for(It = sort_list.begin(); It != End; ++It)
      vertex_subset.insert(CG->get_vertex(*It));

   refcount<dominance<graph> > cg_dominators;
   graphRef subgraph = CG->get_subgraph(vertex_subset);
   /// we do not need the exit vertex since the post-dominator graph is not used
   cg_dominators = refcount<dominance<graph> >(new dominance<graph>(*subgraph, top_vertex, NULL_VERTEX, Param));
   cg_dominators->calculate_dominance_info(dominance<graph>::CDI_DOMINATORS);
   const std::tr1::unordered_map<vertex, vertex>& cg_dominator_map = cg_dominators->get_dominator_map();

   std::map<unsigned int, std::set<vertex> > var_map;
   for(It = sort_list.begin(); It != End; ++It)
   {
      const function_behaviorRef FB = AppM->get_function_behavior(*It);
#ifndef NDEBUG
      const behavioral_helperRef BH = FB->get_helper();
#endif
      PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Analyzing function: " + BH->get_function_name());
      /// add parm_decls that have to be copied
      const std::set<unsigned int>& parm_decl_copied = FB->get_parm_decl_copied();
      for(std::set<unsigned int>::const_iterator p = parm_decl_copied.begin(); p != parm_decl_copied.end(); ++p)
      {
         HLS_C->Rmem->add_parm_decl_copied(*p);
      }
      /// add parm_decls that have to be stored
      const std::set<unsigned int>& parm_decl_stored = FB->get_parm_decl_stored();
      for(std::set<unsigned int>::const_iterator p = parm_decl_stored.begin(); p != parm_decl_stored.end(); ++p)
      {
         HLS_C->Rmem->add_parm_decl_stored(*p);
      }
      /// add actual parameters that have to be loaded
      const std::set<unsigned int>& parm_decl_loaded = FB->get_parm_decl_loaded();
      for(std::set<unsigned int>::const_iterator p = parm_decl_loaded.begin(); p != parm_decl_loaded.end(); ++p)
      {
         HLS_C->Rmem->add_actual_parm_loaded(*p);
      }
      vertex vert_dominator, current_vertex;
      current_vertex = CG->get_vertex(*It);
      THROW_ASSERT(cg_dominator_map.find(current_vertex) != cg_dominator_map.end(), "Dominator vertex not in the dominator tree: " + AppM->get_function_behavior(CG->get_function(current_vertex))->get_helper()->get_function_name());
      if(boost::in_degree(current_vertex, *cg) != 1)
         vert_dominator = cg_dominator_map.find(current_vertex)->second;
      else
         vert_dominator = current_vertex;
      const std::set<unsigned int>& function_mem = FB->get_function_mem();
      for(std::set<unsigned int>::const_iterator v = function_mem.begin(); v != function_mem.end(); v++)
      {
         var_map[*v].insert(vert_dominator);
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Dominator Vertex: " + AppM->get_function_behavior(CG->get_function(vert_dominator))->get_helper()->get_function_name() + " - Variable to be stored: " + BH->PrintVariable(*v));
      }
   }

   /// compute the number of istances for each function
   std::map<vertex, unsigned int> num_instances;
   num_instances[top_vertex] = 1;
   std::list<unsigned int>::reverse_iterator rIt, rEnd = sort_list.rend();
   for(rIt = sort_list.rbegin(); rIt != rEnd; ++rIt)
   {
      vertex cur = CG->get_vertex(*rIt);
      THROW_ASSERT(num_instances.find(cur) != num_instances.end(), "missing number of instances of function " + AppM->get_function_behavior(CG->get_function(cur))->get_helper()->get_function_name());
      unsigned int cur_instances = num_instances.find(cur)->second;
      OutEdgeIterator eo, eo_end;
      for(boost::tie(eo, eo_end) = boost::out_edges(cur, *cg); eo != eo_end; ++eo)
      {
         vertex tgt = boost::target(*eo, *cg);
         unsigned int n_call_points = static_cast<unsigned int>(GET_CALL_POINTS(cg, *eo).size());
         if(num_instances.find(tgt) == num_instances.end())
            num_instances[tgt] = cur_instances * n_call_points;
         else
            num_instances[tgt] += cur_instances * n_call_points;
      }
   }
   THROW_ASSERT(num_instances.find(top_vertex)->second == 1, "something of wrong happened");

   /// find the common dominator
   std::map<unsigned int, std::set<vertex> >::const_iterator it_end = var_map.end();
   for(std::map<unsigned int, std::set<vertex> >::const_iterator it = var_map.begin(); it != it_end; ++it)
   {
      unsigned int funID=0;
      unsigned int var_index = it->first;
      THROW_ASSERT(var_index, "null var index unexpected");
      if(it->second.size() == 1)
      {
         vertex cur = *it->second.begin();
         /// look for a single instance function
         while(num_instances.find(cur)->second != 1)
            cur = cg_dominator_map.find(cur)->second;
         funID = CG->get_function(cur);
      }
      else
      {
         std::set<vertex>::const_iterator vert_it_end = it->second.end();
         if(it->second.find(top_vertex) != vert_it_end)
            funID = CG->get_function(top_vertex);
         else
         {
            std::set<vertex>::const_iterator vert_it = it->second.begin();
            std::list<vertex> dominator_list1;
            vertex cur = *vert_it;
            dominator_list1.push_front(cur);
            do
            {
               cur = cg_dominator_map.find(cur)->second;
               dominator_list1.push_front(cur);
            }
            while(cur != top_vertex);
            ++vert_it;
            std::list<vertex>::const_iterator last=dominator_list1.end();
            for(; vert_it != vert_it_end; ++vert_it)
            {
               std::list<vertex> dominator_list2;
               cur = *vert_it;
               dominator_list2.push_front(cur);
               do
               {
                  cur = cg_dominator_map.find(cur)->second;
                  dominator_list2.push_front(cur);
               }
               while(cur != top_vertex);
               /// find the common dominator between two candidates
               std::list<vertex>::const_iterator dl1_it=dominator_list1.begin(), dl2_it=dominator_list2.begin(), dl2_it_end=dominator_list2.end(), cur_last=dominator_list1.begin();
               while(dl1_it!=last && dl2_it!=dl2_it_end && *dl1_it == *dl2_it && num_instances.find(*dl1_it)->second == 1)
               {
                  cur = *dl1_it;
                  ++dl1_it;
                  cur_last = dl1_it;
                  ++dl2_it;
               }
               last=cur_last;
               funID = CG->get_function(cur);
               if(cur == top_vertex)
                  break;
            }
         }
      }
      THROW_ASSERT(funID, "null function id index unexpected");
      const function_behaviorRef FB = AppM->get_function_behavior(funID);
      const behavioral_helperRef BH = FB->get_helper();
      bool is_on_chip = false;

      const tree_nodeRef& tn = TreeM->get_tree_node_const(var_index);
      if(allocation_algorithm== GSS)
      {
      if (GetPointer<var_decl>(tn) && (GetPointer<var_decl>(tn)->static_flag || !GetPointer<var_decl>(tn)->scpe))
         is_on_chip = true;
      if(GetPointer<string_cst>(tn))
         is_on_chip = true;
      }
      else if(allocation_algorithm == ALL_BRAM)
         is_on_chip = true;
      else if(allocation_algorithm == NO_BRAM)
         is_on_chip = false;
      else
         THROW_ERROR("not supported memory allocation algorithm");

      if (is_on_chip)
      {
         HLS_C->Rmem->add_internal_base_address(TreeM, var_index);
         PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Adding on-chip variable: " + STR(var_index) + " - " + BH->PrintVariable(var_index) + " in function " + BH->get_function_name());
         HLS_C->Rmem->add_internal_variable(funID, var_index);
      }
      else
      {
         PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Adding off-chip variable: " + STR(var_index) + " - " + BH->PrintVariable(var_index));
         HLS_C->Rmem->add_external_variable(TreeM, var_index);
      }
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Variable: " + BH->PrintVariable(var_index));
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "  - Id: " + STR(var_index));
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "  - Base Address: " + STR(HLS_C->Rmem->get_base_address(var_index)));
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "  - Size: " + STR(tree_helper::size(TreeM, tree_helper::get_type_index(TreeM, var_index)) / 8));
   }

   STOP_TIME(memory_time);
}

