/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file allocation.cpp
 * @brief Wrapper for technology used by the high-level synthesis flow.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10655 $
 * $Date: 2012-03-20 08:39:27 +0100 (Tue, 20 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#include "allocation.hpp"

#include "technology_manager.hpp"
#include "library_manager.hpp"
#include "technology_node.hpp"
#include "area_model.hpp"
#include "time_model.hpp"
#if HAVE_EXPERIMENTAL
#include "power_model.hpp"
#endif
#include "technology_builtin.hpp"
#include "memory.hpp"
#include "hls_target.hpp"
#include "hls_constraints.hpp"
#include "hls.hpp"
#include "utility.hpp"
#include "exceptions.hpp"
#include "utility.hpp"
#include "graph.hpp"
#include "tree_manager.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"

#include "structural_objects.hpp"
#include "structural_manager.hpp"
#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "behavioral_helper.hpp"
#include "op_node_info.hpp"

#include "memory.hpp"
#include "target_device.hpp"

#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>

#include <map>
#include <vector>

#include <iostream>

#include "moduleGenerator.hpp"

#include "Parameter.hpp"
#include "dbgPrintHelper.hpp"

#define MAP_ON_DISTRAM 0


static inline std::string encode_op_type(const std::string & op_name, const std::string & fu_supported_types)
{
   return op_name + ":" + fu_supported_types;
}

static inline unsigned int resize_to_8_16_32_64(unsigned int value)
{
   if(value  <= 8)
      return 8;
   else if (value <= 16)
      return 16;
   else if(value <= 32)
      return 32;
   else if(value <= 64)
      return 64;
   else
      THROW_ERROR("not expected size " + boost::lexical_cast<std::string>(value));
   return 0;
}

static inline std::string encode_op_type_prec(const std::string & op_name, const std::string & fu_supported_types, node_kind_prec_infoRef node_info)
{
   std::string op_type = encode_op_type(op_name, fu_supported_types);
   for (std::vector<unsigned int>::const_iterator itr = node_info->input_prec.begin(), end = node_info->input_prec.end(); itr != end; ++itr)
      op_type += ":" + STR(*itr);
   op_type += ":" + STR(node_info->output_prec);
   return op_type;
}

allocation::allocation(const ParameterRef _Param, const hlsRef _HLS):
   HLS_step(_Param, _HLS),
   AppM(_HLS->AppM),
   functionId(_HLS->functionId),
   HLS_T(_HLS->HLS_T),
   output_level(_Param->getOption<int>(OPT_output_level))
{
}

allocation::~allocation()
{

}


void allocation::exec()
{
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Starting allocation");
   const technology_managerRef TM = HLS_T->get_technology_manager();
   const tree_managerRef TreeM = AppM->get_tree_manager();
   const function_behaviorRef FB = AppM->get_function_behavior(functionId);
   const HLS_constraintsRef HLS_C = HLS->HLS_C;

   const std::set<unsigned int>& function_vars = HLS_C->Rmem->get_function_vars(functionId);
   for(std::set<unsigned int>::const_iterator l = function_vars.begin(); l != function_vars.end(); ++l)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, " - analyzing variable " + STR(*l));
#if MAP_ON_DISTRAM
      technology_nodeRef current_fu = get_fu(ARRAY_1D_STD_DISTRAM);
#else
      technology_nodeRef current_fu = get_fu(ARRAY_1D_STD_BRAM);
#endif
      unsigned int var = *l;
      unsigned int current_size = get_number_fu_types();
      PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, " - allocating unit " + current_fu->get_name() + " for variable " + FB->get_helper()->PrintVariable(*l) + " in position " + STR(current_size));
      list_of_FU.push_back(current_fu);
      tech_constraints.push_back(1);
      id_to_fu_names[current_size] = std::make_pair(current_fu->get_name(), TM->get_library(current_fu->get_name()));
      THROW_ASSERT(tech_constraints.size() == list_of_FU.size(), "Something of wrong happened");
      vars_to_memory_units[var] = current_size;
      memory_units[current_size] = var;
   }

   std::tr1::unordered_map<std::string, unsigned int> tech_vec = HLS_C->tech_constraints;
   const std::tr1::unordered_map<std::string, std::pair<std::string, std::pair<std::string, unsigned int> > > &binding_constraints = HLS_C->binding_constraints;
   std::tr1::unordered_map<std::string, std::map<unsigned int, unsigned int> > fu_name_to_id;
   std::set<vertex> vertex_to_analyse;
   std::set<vertex> vertex_analysed;
   std::set<vertex> support_ops;
   support_ops.insert(HLS->operations.begin(), HLS->operations.end());
   const graphRef g = FB->get_subgraph(function_behavior::CFG, support_ops);
   graph::vertex_iterator v, v_end;
   for (tie(v, v_end) = boost::vertices(*g); v != v_end; ++v)
   {
      std::string current_op = GET_OP(g.get(), *v);
      PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "- Processing operation: " + current_op + " - " + GET_NAME(g.get(), *v));
      if (LOAD == current_op || STORE == current_op)
      {
         const tree_nodeRef curr_tn = TreeM->get_tree_node_const(GET_NODEID(g.get(), *v).back());
         gimple_assign* me = GetPointer<gimple_assign>(curr_tn);
         THROW_ASSERT(me, "only gimple_assign's are allowed as memory operations");
         unsigned int var = 0;
         if (GET_OP(g, *v) == STORE)
            var = tree_helper::get_base_index(TreeM, GET_INDEX_NODE(me->op0));
         else
            var = tree_helper::get_base_index(TreeM, GET_INDEX_NODE(me->op1));
         if (var == 0 || function_vars.find(var) == function_vars.end())
         {
            vertex_to_analyse.insert(*v);
            PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "  * Operation " + current_op + " queued for allocation");
            continue;
         }
         THROW_ASSERT(vars_to_memory_units.find(var) != vars_to_memory_units.end(), "Not existing memory unit associated with the variable");
         binding[*v] = vars_to_memory_units[var];
         op_name_to_fus[current_op].insert(vars_to_memory_units[var]);
         technology_nodeRef current_fu = list_of_FU[vars_to_memory_units[var]];
         vertex_to_op[*v] = current_op;
         PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "  . Operation " + current_op + " named " + GET_NAME(g.get(), *v) + " mapped onto " + current_fu->get_name() + ", found in library " + TM->get_library(current_op) + " in position " + STR(vars_to_memory_units[var]));
      }
      // Artificial FUs
      else if (ASSIGN == current_op || ADDR_EXPR == current_op || READ_COND == current_op || NOP_EXPR == current_op || CONVERT_EXPR == current_op || SWITCH_COND == current_op || GIMPLE_LABEL == current_op || GIMPLE_GOTO == current_op || ENTRY == current_op || EXIT == current_op || NOP == current_op || GIMPLE_RETURN == current_op || GIMPLE_PHI == current_op)
      {
         unsigned int current_size = get_number_fu_types();
         technology_nodeRef current_fu;
         if (current_op == ASSIGN)
            current_fu = get_fu(ASSIGN_STD);
         else if (current_op == ADDR_EXPR)
            current_fu = get_fu(ADDR_EXPR_STD);
         else if (current_op == NOP_EXPR)
         {
            unsigned int modify_tree_index = *GET_NODEID(g.get(), *v).begin();

            //std::cout << NOP_EXPR << "->" << modify_tree_index << std::endl;
            tree_nodeRef modify_node = TreeM->get_tree_node_const(modify_tree_index);
            gimple_assign *gms = GetPointer<gimple_assign>(modify_node);

            unsigned int left_type_index;
            tree_nodeRef left_type_node = tree_helper::get_type_node(GET_NODE(gms->op0), left_type_index);
            nop_expr * ne = GetPointer<nop_expr>(GET_NODE(gms->op1));
            unsigned int right_type_index;
            tree_nodeRef right_type_node = tree_helper::get_type_node(GET_NODE(ne->op), right_type_index);

            bool unsignedR = tree_helper::is_unsigned(TreeM, right_type_index);
            bool unsignedL = tree_helper::is_unsigned(TreeM, left_type_index);
            bool intR = tree_helper::is_int(TreeM, right_type_index);
            bool intL = tree_helper::is_int(TreeM, left_type_index);
            bool enumR = tree_helper::is_an_enum(TreeM, right_type_index);
            bool enumL = tree_helper::is_an_enum(TreeM, left_type_index);
            bool boolR = tree_helper::is_bool(TreeM, right_type_index);
            bool boolL = tree_helper::is_bool(TreeM, left_type_index);
            bool is_a_pointerR = tree_helper::is_a_pointer(TreeM, right_type_index);
            bool is_a_pointerL = tree_helper::is_a_pointer(TreeM, left_type_index);
            bool is_realR = tree_helper::is_real(TreeM, right_type_index);
            bool is_realL = tree_helper::is_real(TreeM, left_type_index);

            if((unsignedR || is_a_pointerR || boolR) && (unsignedL || is_a_pointerL || boolL))
               current_fu = get_fu(UUDATA_CONVERTER_STD);
            else if((intR || enumR) && (unsignedL || is_a_pointerL || boolL))
               current_fu = get_fu(IUDATA_CONVERTER_STD);
            else if((unsignedR || is_a_pointerR || boolR) && (intL|| enumL))
               current_fu = get_fu(UIDATA_CONVERTER_STD);
            else if((intR || enumR) && (intL|| enumL))
               current_fu = get_fu(IIDATA_CONVERTER_STD);
            else if(is_realR && is_realL)
               current_fu = get_fu(FFDATA_CONVERTER_STD);
            else
               THROW_ERROR(std::string("Nop_Expr pattern not supported ") + boost::lexical_cast<std::string>(modify_tree_index));
         }
         else if (current_op == CONVERT_EXPR)
         {
            unsigned int modify_tree_index = *GET_NODEID(g.get(), *v).begin();

            //std::cout << CONVERT_EXPR << "->" << modify_tree_index << std::endl;
            tree_nodeRef modify_node = TreeM->get_tree_node_const(modify_tree_index);
            gimple_assign *gms = GetPointer<gimple_assign>(modify_node);
            unsigned int left_type_index;
            tree_nodeRef left_type_node = tree_helper::get_type_node(GET_NODE(gms->op0), left_type_index);
            convert_expr * ce = GetPointer<convert_expr>(GET_NODE(gms->op1));
            unsigned int right_type_index;
            tree_nodeRef right_type_node = tree_helper::get_type_node(GET_NODE(ce->op), right_type_index);

            bool unsignedR = tree_helper::is_unsigned(TreeM, right_type_index);
            bool unsignedL = tree_helper::is_unsigned(TreeM, left_type_index);
            bool intR = tree_helper::is_int(TreeM, right_type_index);
            bool intL = tree_helper::is_int(TreeM, left_type_index);
            bool boolR = tree_helper::is_bool(TreeM, right_type_index);
            bool boolL = tree_helper::is_bool(TreeM, left_type_index);
            bool is_a_pointerR = tree_helper::is_a_pointer(TreeM, right_type_index);
            bool is_a_pointerL = tree_helper::is_a_pointer(TreeM, left_type_index);

            if((unsignedR || is_a_pointerR || boolR) && (unsignedL || is_a_pointerL || boolL))
               current_fu = get_fu(UUCONVERTER_EXPR_STD);
            else if(intR && (unsignedL || is_a_pointerL || boolL))
               current_fu = get_fu(IUCONVERTER_EXPR_STD);
            else if((unsignedR || is_a_pointerR || boolR) && intL)
               current_fu = get_fu(UICONVERTER_EXPR_STD);
            else if(intR && intL)
               current_fu = get_fu(IICONVERTER_EXPR_STD);
            else
               THROW_ERROR(std::string("CONVERT_EXPR pattern not supported ") + boost::lexical_cast<std::string>(modify_tree_index) + ":" + STR(left_type_index) + ":" + STR(right_type_index));

         }
         else if (current_op == READ_COND)
            current_fu = get_fu(READ_COND_STD);
         else if (current_op == SWITCH_COND)
            current_fu = get_fu(SWITCH_COND_STD);
         else if (current_op == GIMPLE_LABEL)
            current_fu = get_fu(GIMPLE_LABEL_STD);
         else if (current_op == GIMPLE_GOTO)
            current_fu = get_fu(GIMPLE_GOTO_STD);
         else if (current_op == ENTRY)
            current_fu = get_fu(ENTRY_STD);
         else if (current_op == EXIT)
            current_fu = get_fu(EXIT_STD);
         else if (current_op == NOP)
            current_fu = get_fu(NOP_STD);
         else if (current_op == GIMPLE_RETURN)
            current_fu = get_fu(GIMPLE_RETURN_STD);
         else if (current_op == GIMPLE_PHI)
            current_fu = get_fu(GIMPLE_PHI_STD);
         else
            THROW_ERROR("Unexpected operation");
         // FU must exist
         THROW_ASSERT(current_fu, std::string("Not found ") + current_op + " in library " + TM->get_library(current_op));
         list_of_FU.push_back(current_fu);
         tech_constraints.push_back(1);
         THROW_ASSERT(tech_constraints.size() == list_of_FU.size(), "Something of wrong happen");
         is_vertex_bounded_rel.insert(current_size);
         binding[*v] = current_size;
         op_name_to_fus[current_op].insert(current_size);
         id_to_fu_names[current_size] = std::make_pair(current_fu->get_name(), TM->get_library(current_op));
         vertex_to_op[*v] = current_op;
         PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "  . Operation " + current_op + " mapped onto " + current_fu->get_name() + ", found in library " + TM->get_library(current_op));
      }
      // Constrained FUs
      else if(binding_constraints.find(GET_NAME(g.get(), *v)) != binding_constraints.end())
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "  . Current node is under constraints");
         const std::pair<std::string, std::pair<std::string, unsigned int> >& defined_binding = binding_constraints.find(GET_NAME(g.get(), *v))->second;
         std::string fu_name = defined_binding.first;
         std::string fu_library = defined_binding.second.first;
         unsigned int fu_index = defined_binding.second.second;
         std::string key = ENCODE_FU_LIB(fu_name, fu_library);
         if(fu_name_to_id.find(key) != fu_name_to_id.end())
         {
            if(fu_name_to_id[key].find(fu_index) != fu_name_to_id[key].end())
               binding[*v] = fu_name_to_id[key][fu_index];
            else
            {
               unsigned int current_size = get_number_fu_types();
               fu_name_to_id[key][fu_index] = current_size;
               technology_nodeRef current_fu = TM->get_fu(fu_name, fu_library);
               THROW_ASSERT(current_fu, std::string("Not found") + fu_name + " in library " + fu_library);
               list_of_FU.push_back(current_fu);
               tech_constraints.push_back(1);
               binding[*v] = current_size;
               op_name_to_fus[current_op].insert(current_size);
               id_to_fu_names[current_size] = std::make_pair(fu_name, fu_library);
               if(tech_vec.find(key) != tech_vec.end())
                  tech_vec[key]--;
            }
         }
         else
         {
            unsigned int current_size = get_number_fu_types();
            fu_name_to_id[key][fu_index] = current_size;
            technology_nodeRef current_fu = TM->get_fu(fu_name, fu_library);
            THROW_ASSERT(current_fu, std::string("Not found") + fu_name + " in library " + fu_library);
            list_of_FU.push_back(current_fu);
            tech_constraints.push_back(1);
            binding[*v] = current_size;
            op_name_to_fus[current_op].insert(current_size);
            id_to_fu_names[current_size] = std::make_pair(fu_name, fu_library);
            if(tech_vec.find(key) != tech_vec.end())
               tech_vec[key]--;
         }
      }
      else
      {
         vertex_to_analyse.insert(*v);
         PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "  * Operation " + current_op + " queued for allocation");
      }
   }

   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Starting allocation of operations in queued vertices");
   const std::vector<std::string>& libraries = TM->get_library_list();
   std::map<std::string,technology_nodeRef> new_fu;

   for(unsigned int l = 0; l < libraries.size(); l++)
   {
      const library_managerRef library = TM->get_library_manager(libraries[l]);
      const library_manager::fu_map_type& list_fu = library->get_library_fu();
      //library_manager::fu_map_type::const_iterator fu_end = list_fu.end();
      std::map<technology_nodeRef,std::map<unsigned int, unsigned int> >fu_list;
      for (library_manager::fu_map_type::const_iterator fu = list_fu.begin(); fu !=  list_fu.end(); fu++)
      {
         //std::cerr << fu->first << std::endl;

         ///LOAD/STORE operations allocated to such units have been already allocated
         if (fu->first == ARRAY_1D_STD_BRAM || fu->first == ARRAY_1D_STD_DISTRAM || fu->first == ARRAY_1D_STD_BRAM_ALIGNED) continue;

         technology_nodeRef current_fu = fu->second;
         if (GetPointer<functional_unit_template>(current_fu)) continue;

         std::tr1::unordered_map<std::string, unsigned int>::const_iterator tech_constrain_it = tech_vec.find(ENCODE_FU_LIB(current_fu->get_name(), libraries[l]));
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "- Considering functional unit: " + current_fu->get_name());
         if (tech_constrain_it != tech_vec.end() && tech_constrain_it->second == 0) continue; // forced to use 0 FUs of current ones
         structural_managerRef structManager_obj=GetPointer<functional_unit>(current_fu)->CM;
         if(structManager_obj ) ///generated cannot be directly considered for allocation
         {
            structural_objectRef modobj=structManager_obj->get_circ();
            if(GetPointer<module>(modobj)->get_generated())
               continue;
         }
         const functional_unit::operation_vec & Ops = GetPointer<functional_unit>(current_fu)->get_operations();
         functional_unit::operation_vec::const_iterator ops_end = Ops.end();
         bool found;
         unsigned int current_fu_id = get_number_fu_types();

         unsigned int current_id=current_fu_id;

         bool predicate_1 = library->get_library_name() != WORK_LIBRARY;
         bool predicate_2 = library->get_library_name() == WORK_LIBRARY;

         found = false;

         for (functional_unit::operation_vec::const_iterator ops = Ops.begin(); ops != ops_end; ops++)
         {
            operation* curr_op = GetPointer<operation>(*ops);
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  . Considering operation: " + (*ops)->get_name());
            bool varargs_fu;
            for (std::set<vertex>::const_iterator vert = vertex_to_analyse.begin(); vert != vertex_to_analyse.end(); vert++)
            {
               if (GET_OP(g.get(), *vert) != curr_op->get_name())
                  continue;
               else if((GET_OP(g.get(), *vert) == LOAD || GET_OP(g.get(), *vert) == STORE) && fu->first != MEMORY_STD)
                  continue;
               else if(predicate_1 && TM->get_fu(GET_OP(g.get(), *vert), WORK_LIBRARY) && GET_TYPE(g.get(), *vert) != TYPE_MEMCPY) continue;
               else if(predicate_2 && GET_TYPE(g.get(), *vert) == TYPE_MEMCPY) continue;

               varargs_fu = false;

               std::string specialized_fuName="";

               std::string current_op;
               node_kind_prec_infoRef node_info(new node_kind_prec_info());
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  . Considering vertex " + GET_NAME(g, *vert));
               bool isMemory = false;
               if (fu->first == MEMORY_STD) isMemory = true;

               if (!isMemory)
                  GET_NODE_TYPE_PREC(*vert, node_info);
               else
                  node_info->node_kind = "VECTOR_BOOL";

               // Check for correct type
               if (node_info->node_kind != "" && !curr_op->is_type_supported(node_info->node_kind))
                  continue; // FU does not support the operation type
               // Check for correct precision
               if (!curr_op->is_type_supported(node_info->node_kind, node_info->input_prec))
                  continue; // FU does support the operation type, but does not support correct precision

               if(structManager_obj && GetPointer<module>(structManager_obj->get_circ())->is_var_args())
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level,"Found a var args function");
                  varargs_fu = true;
                  moduleGeneratorRef modGen = moduleGeneratorRef(new moduleGenerator(Param, AppM));
                  std::vector<application_manager::io_binding_type> required_variables = AppM->get_required_values(functionId, *vert);
                  current_op=current_fu->get_name()+modGen->get_specialized_name(required_variables, FB);
                  specialized_fuName=current_op;

                  std::string fu_name=current_fu->get_name();
                  std::string check_lib = TM->get_library(specialized_fuName);
                  if(check_lib == libraries[l])
                  {
                     new_fu[specialized_fuName]=get_fu(specialized_fuName);
                  }
                  else if(new_fu.find(specialized_fuName)==new_fu.end())
                  {
                     modGen->specialize_fu(fu_name, *vert,libraries[l], TM, FB, specialized_fuName, new_fu);
                  }
               }
               else if (node_info->node_kind != "" && !isMemory)
                  current_op = encode_op_type_prec((*ops)->get_name(), curr_op->get_type_supported_string(), node_info);
               else if (node_info->node_kind != "")
                  current_op = encode_op_type((*ops)->get_name(), curr_op->get_type_supported_string());
               else
                  current_op = (*ops)->get_name();

               if (GetPointer<functional_unit>(current_fu)->fu_template_name != "")
               {
                  // If the current_fu is a specialization of a template compute required parameters for the specialization.
                  std::string required_prec = "";
                  for (std::vector<unsigned int>::const_iterator itr = node_info->input_prec.begin(), end = node_info->input_prec.end(); itr != end; ++itr)
                     required_prec += STR(*itr) + " ";
                  required_prec += STR(node_info->output_prec);

                  // if the computed parameters is different form what was used to build this specialization skip it.
                  if (required_prec != GetPointer<functional_unit>(current_fu)->fu_template_parameters) continue;
               }
               found = true;

               const library_managerRef libraryManager = TM->get_library_manager(libraries[l]);

               std::map<technology_nodeRef,std::map<unsigned int,unsigned int> >::iterator techMap;

               std::string functionalUnitName="";
               unsigned int specializedId=current_id;
               unsigned int max_prec = node_info->input_prec.begin() == node_info->input_prec.end() ? 0 : *std::max_element(node_info->input_prec.begin(), node_info->input_prec.end());
               if(varargs_fu)
               {
                  functionalUnitName=specialized_fuName;
                  techMap=fu_list.find(new_fu.find(functionalUnitName)->second);
                  PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Specialized unit: "+new_fu.find(functionalUnitName)->first);
                  if(techMap!=fu_list.end() && techMap->second.find(max_prec) != techMap->second.end())
                  {
                     specializedId=techMap->second.find(max_prec)->second;
                  }
                  else
                  {
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Insert into list of unit to add: "+new_fu.find(functionalUnitName)->first);
                     fu_list[new_fu.find(functionalUnitName)->second][max_prec]=current_id;
                     current_id++;
                  }
               }
               else
               {
                  functionalUnitName=current_fu->get_name();

                  techMap=fu_list.find(libraryManager->get_fu(functionalUnitName));
                  if(techMap!=fu_list.end() && techMap->second.find(max_prec) != techMap->second.end())
                  {
                     specializedId=techMap->second.find(max_prec)->second;
                  }
                  else
                  {
                     fu_list[libraryManager->get_fu(functionalUnitName)][max_prec]=current_id;
                     current_id++;
                  }
               }

               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "    . Match found for vertex: " + GET_NAME(g, *vert));
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "      . Adding candidate FU: " + functionalUnitName + " for operation: " + current_op + " in position " + STR(specializedId));
               vertex_analysed.insert(*vert);
               op_name_to_fus[current_op].insert(specializedId);
               id_to_fu_names[specializedId] = std::make_pair(functionalUnitName, libraries[l]);
               vertex_to_op[*vert] = current_op;
            }
         }
         if (found)
         {
            for(unsigned int pos=current_fu_id; pos<current_id; pos++)
            {
               for(std::map<technology_nodeRef,std::map<unsigned int,unsigned int> >::iterator iter=fu_list.begin(); iter!=fu_list.end(); iter++)
               {
                  for(std::map<unsigned int,unsigned int>::iterator iter2=iter->second.begin(); iter2!=iter->second.end(); iter2++)
                  {
                     if(iter2->second==pos)
                     {
                        PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Specialized unit: "+(iter->first->get_name())+" in position: "+STR(iter2->second));
                        list_of_FU.push_back(iter->first);
                        if (tech_constrain_it == tech_vec.end())
                           tech_constraints.push_back(INFINITE_UINT);
                        else
                           tech_constraints.push_back(tech_constrain_it->second);
                     }
                  }
               }
            }
            fu_list.clear();
            THROW_ASSERT(tech_constraints.size() == list_of_FU.size(), "something of wrong happen");
         }
      }
      for(std::map<std::string,technology_nodeRef>::iterator iter_new_fu=new_fu.begin(); iter_new_fu!=new_fu.end(); iter_new_fu++) {
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Adding functional unit: "+iter_new_fu->first+" in "+libraries[l]);
         TM->add(iter_new_fu->second,libraries[l]);
      }
      new_fu.clear();
   }

   // For debug purpose
   print_allocated_resources(g);

   // Check if each operation has been analysed
   for (std::set<vertex>::const_iterator ve = vertex_to_analyse.begin(); ve != vertex_to_analyse.end(); ve++)
   {
      if (vertex_analysed.find(*ve) != vertex_analysed.end()) continue;

      node_kind_prec_infoRef node_info(new node_kind_prec_info());
      GET_NODE_TYPE_PREC(*ve, node_info);
      std::string precisions;
      for (std::vector<unsigned int>::iterator itr = node_info->input_prec.begin(), end = node_info->input_prec.end(); itr < end; ++itr)
         precisions += " " + STR(*itr);
      precisions += " " + STR(node_info->output_prec);
      PRINT_OUT_MEX(OUTPUT_LEVEL_NONE, output_level, "Operation for which does not exist a functional unit in the resource library: " + GET_OP(g, *ve) + " in vertex: " + GET_NAME(g, *ve) + " with vertex type: " + node_info->node_kind + " and vertex prec:" + precisions);
   }
   if(vertex_to_analyse.size() > vertex_analysed.size())
      THROW_ERROR("Vertices not completely allocated");

   PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Function " + FB->get_helper()->get_function_name() + " uses " + STR(tech_constraints.size()) + " functional unit types");

}

void allocation::print(std::ostream& os) const
{
   std::vector<technology_nodeRef>::const_iterator fu_end = list_of_FU.end();
   unsigned int index = 0;
   for (std::vector<technology_nodeRef>::const_iterator fu =  list_of_FU.begin(); fu != fu_end; fu++)
   {
      os << index << " ";
      index++;
      (*fu)->print(os);
   }
   if (!op_name_to_fus.empty())
   {
      os << "Op_name relation with functional unit name and operations.\n";
      std::tr1::unordered_map<std::string, std::set<unsigned int> >::const_iterator op_end = op_name_to_fus.end();
      for (std::tr1::unordered_map<std::string, std::set<unsigned int> >::const_iterator it_op = op_name_to_fus.begin(); it_op != op_end; it_op++)
      {
         std::set<unsigned int>::const_iterator f_end = it_op->second.end();
         for (std::set<unsigned int>::const_iterator f_op = it_op->second.begin(); f_op != f_end; f_op++)
            os << "  [" << it_op->first << ", <" << list_of_FU[*f_op]->get_name() << ">]" << std::endl;
      }
   }
}

bool allocation::can_implement(const unsigned int fu_id, const vertex v, const graphRef g) const
{
   // Preliminar check
   THROW_ASSERT(fu_id < get_number_fu_types(), "functional unit id not meaningful: " + STR(fu_id));
   // Check if the vertex is staticly bound to a FU
   std::tr1::unordered_map<vertex, unsigned int>::const_iterator it = binding.find(v);
   std::tr1::unordered_map<vertex, unsigned int>::const_iterator it_end = binding.end();
   if (it != it_end)
      return (it->second == fu_id);

   std::string current_op = vertex_to_op.find(v) != vertex_to_op.end() ? vertex_to_op.find(v)->second : GET_OP(g, v);

   std::tr1::unordered_map<std::string, std::set<unsigned int> >::const_iterator fu_found = op_name_to_fus.find(current_op);
   if (fu_found == op_name_to_fus.end())
      return false; // Operation (name:type:prec) is supported
   if (fu_found->second.find(fu_id) == fu_found->second.end())
      return false; // Vertex can't be implemented by FU
   // If the execution gets here, then the previous tests was successful either in the is_real() case, and in the standard one.
   return true;
}

const std::set<unsigned int> & allocation::can_implement_set(const vertex v, const graphRef g) const
{
   std::string current_op = vertex_to_op.find(v) != vertex_to_op.end() ? vertex_to_op.find(v)->second : GET_OP(g, v);
   return op_name_to_fus.find(current_op)->second;
}

bool allocation::is_read_cond(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return list_of_FU[fu_name]->get_name() == READ_COND_STD;
}

bool allocation::is_assign(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return list_of_FU[fu_name]->get_name() == ASSIGN_STD;
}

bool allocation::is_return(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return list_of_FU[fu_name]->get_name() == GIMPLE_RETURN_STD;
}

bool allocation::is_memory_unit(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return memory_units.find(fu_name) != memory_units.end();
}

bool allocation::is_vertex_bounded(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return is_vertex_bounded_rel.find(fu_name) != is_vertex_bounded_rel.end();
}

bool allocation::is_vertex_bounded_with(const vertex v, unsigned int & fu_name) const
{
   if(binding.find(v) == binding.end())
      return false;
   else
      fu_name = binding.find(v)->second;
   return true;
}

bool allocation::is_artificial_fu(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   std::string fu_string_name = list_of_FU[fu_name]->get_name();
   if(fu_string_name == ASSIGN_STD || !has_to_be_synthetized(fu_name))
      return true;
   else
      return false;
}

bool allocation::has_to_be_synthetized(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   std::string fu_string_name = list_of_FU[fu_name]->get_name();
   if(fu_string_name == GIMPLE_RETURN_STD || fu_string_name == ENTRY_STD || fu_string_name == EXIT_STD || fu_string_name == NOP_STD || fu_string_name == GIMPLE_PHI_STD || fu_string_name == GIMPLE_LABEL_STD || fu_string_name == GIMPLE_GOTO_STD)
      return false;
   else
      return true;
}

unsigned int allocation::get_number_fu(unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return tech_constraints[fu_name];
}

technology_nodeRef allocation::get_fu(unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return list_of_FU[fu_name];
}

///ToBeCompleted
std::string allocation::get_string_name(unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   return list_of_FU[fu_name]->get_name() + "_" + STR(fu_name);
}

unsigned int allocation::get_initiation_time(const unsigned int fu_name, const vertex v, const graphRef g) const
{
   THROW_ASSERT(can_implement(fu_name, v, g), "This function (" + get_string_name(fu_name) + ") cannot implement the operation " + GET_OP(g, v));
   if(!has_to_be_synthetized(fu_name))
      return 1;
   technology_nodeRef node_op = GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(GET_OP(g, v));
   THROW_ASSERT(GetPointer<operation>(node_op)->time_m, "Timing information not specified for unit " + id_to_fu_names.find(fu_name)->second.first);
   return GetPointer<operation>(node_op)->time_m->get_initiation_time();
}


double allocation::get_execution_time(const unsigned int fu_name, const vertex v, const graphRef g) const
{
   THROW_ASSERT(can_implement(fu_name, v, g), "This function (" + get_string_name(fu_name) + ") cannot implement the operation " + GET_OP(g, v));
   if(!has_to_be_synthetized(fu_name))
      return 0.0;
   technology_nodeRef node_op = GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(GET_OP(g, v));
   THROW_ASSERT(GetPointer<operation>(node_op)->time_m, "Timing information not specified for unit " + id_to_fu_names.find(fu_name)->second.first);
   return GetPointer<operation>(node_op)->time_m->get_execution_time();
}

#if HAVE_EXPERIMENTAL
double allocation::get_power_consumption(const unsigned int fu_name, const vertex v, const graphRef g) const
{
   THROW_ASSERT(can_implement(fu_name, v, g), "This function (" + get_string_name(fu_name) + ") cannot implement the operation " + GET_OP(g, v));
   if(!has_to_be_synthetized(fu_name))
      return 0.0;
   technology_nodeRef node_op = GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(GET_OP(g, v));
   THROW_ASSERT(GetPointer<operation>(node_op)->power_m, "Power information not specified for unit " + id_to_fu_names.find(fu_name)->second.first);
   return GetPointer<operation>(node_op)->power_m->power_consumption;
}
#endif

double allocation::get_area(const unsigned int fu_name) const
{
   THROW_ASSERT(fu_name < get_number_fu_types(), "functional unit id not meaningful");
   if(!has_to_be_synthetized(fu_name))
      return 0.0;
   THROW_ASSERT(GetPointer<functional_unit>(list_of_FU[fu_name])->area_m, "Area information not specified for unit " + id_to_fu_names.find(fu_name)->second.first);
   return GetPointer<functional_unit>(list_of_FU[fu_name])->area_m->get_area_value();
}

unsigned int allocation::get_number_fu_types() const
{
   return static_cast<unsigned int>(list_of_FU.size());
}


double allocation::get_attribute_of_fu_per_op(const vertex v, const graphRef g, allocation::op_perfomed op, allocation::op_target target) const
{
   unsigned int fu_name;
   bool flag;
   double res =  get_attribute_of_fu_per_op(v, g, op, target, fu_name, flag);
   THROW_ASSERT(flag, "something of wrong happen");
   return res;
}


double allocation::get_attribute_of_fu_per_op(const vertex v, const graphRef g, allocation::op_perfomed op, allocation::op_target target, unsigned int &fu_name, bool & flag, const updatecopy_HLS_constraints_functor *CF ) const
{
   std::string current_op = vertex_to_op.find(v) != vertex_to_op.end() ? vertex_to_op.find(v)->second : GET_OP(g, v);
   const std::set<unsigned int> & fu_set = op_name_to_fus.find(current_op)->second;

   std::string op_name = GET_OP(g, v);
   const std::set<unsigned int>::const_iterator f_end = fu_set.end();
   std::set<unsigned int>::const_iterator f_i = fu_set.begin();
   unsigned int int_value;
   double double_value;
   flag = false;
   while (CF && f_i != f_end && ((*CF)(*f_i) <= 0 || (binding.find(v) != binding.end()  && binding.find(v)->second != *f_i))) f_i++;
   if (f_i == f_end) return -1.0;
   flag = true;

   switch (target)
   {
   case initiation_time:
   {
      unsigned int temp;
      fu_name = *f_i;
      if(!has_to_be_synthetized(fu_name))
         return 1.0;

      THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(op_name))->time_m, "Timing information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(fu_name)->second.first);
      int_value = GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(op_name))->time_m->get_initiation_time();

      if(binding.find(v) != binding.end() && binding.find(v)->second == fu_name)
         return static_cast<double>(int_value);
      f_i++;

      for (; f_i != f_end; f_i++)
      {
         if (CF && (*CF)(*f_i) <= 0) continue;
         switch (op)
         {
         case max:
            THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m, "Timing information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(*f_i)->second.first);
            temp = MAX(int_value, GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m->get_initiation_time());
            break;
         case min:
            THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m, "Timing information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(*f_i)->second.first);
            temp = MIN(int_value, GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m->get_initiation_time());
            break;
         default:
            temp = 0;
            THROW_ERROR(std::string("Not supported allocation::op_perfomed"));
            break;
         }
         if (temp != int_value)
         {
            fu_name = *f_i;
            int_value = temp;
         }
      }
      return static_cast<double>(int_value);
      break;
   }
#if HAVE_EXPERIMENTAL
   case power_consumption:
   {
      double temp;
      fu_name = *f_i;
      if(!has_to_be_synthetized(fu_name))
         return 0.0;
      THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(op_name))->time_m, "Power information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(fu_name)->second.first);
      double_value = GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(op_name))->power_m->power_consumption;
      if(binding.find(v) != binding.end() && binding.find(v)->second == fu_name)
         return double_value;
      f_i++;
      for (; f_i != f_end; f_i++)
      {
         if (CF && (*CF)(*f_i) <= 0) continue;
         switch (op)
         {
         case max:
            THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m, "Power information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(*f_i)->second.first);
            temp = MAX(double_value, GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->power_m->power_consumption);
            break;
         case min:
            THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m, "Power information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(*f_i)->second.first);
            temp = MIN(double_value, GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->power_m->power_consumption);
            break;
         default:
            temp = 0;
            THROW_ERROR(std::string("Not supported allocation::op_perfomed"));
            break;
         }
         if (temp != double_value)
         {
            fu_name = *f_i;
            double_value = temp;
         }
      }
      return double_value;
      break;
   }
#endif
   case execution_time:
   {
      double temp;
      fu_name = *f_i;
      if(!has_to_be_synthetized(fu_name))
         return 0.0;
      THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(op_name))->time_m, "Timing information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(fu_name)->second.first);
      double_value = GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[fu_name])->get_operation(op_name))->time_m->get_execution_time();      
      if(binding.find(v) != binding.end() && binding.find(v)->second == fu_name)
         return double_value;
      f_i++;
      for (; f_i != f_end; f_i++)
      {
         if (CF && (*CF)(*f_i) <= 0) continue;
         switch (op)
         {
         case max:
            THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m, "Timing information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(*f_i)->second.first);
            temp = MAX(double_value, GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m->get_execution_time());
            break;
         case min:
            THROW_ASSERT(GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m, "Timing information not specified for operation " + op_name + " on unit " + id_to_fu_names.find(*f_i)->second.first);
            temp = MIN(double_value, GetPointer<operation>(GetPointer<functional_unit>(list_of_FU[*f_i])->get_operation(op_name))->time_m->get_execution_time());
            break;
         default:
            temp = 0;
            THROW_ERROR(std::string("Not supported allocation::op_perfomed"));
            break;
         }
         if (temp != double_value)
         {
            fu_name = *f_i;
            double_value = temp;
         }
      }
      return double_value;
      break;
   }
   default:
      THROW_ERROR(std::string("Not supported allocation::op_target"));
      break;
   }
   return -1.0;
}


unsigned int allocation::max_number_of_resources(const vertex v, const graphRef g) const
{
   std::string current_op = vertex_to_op.find(v) != vertex_to_op.end() ? vertex_to_op.find(v)->second : GET_OP(g, v);
   const std::set<unsigned int> & fu_set = op_name_to_fus.find(current_op)->second;

   unsigned int tot_num_res = 0, num_res;
   const std::set<unsigned int>::const_iterator f_end = fu_set.end();

   for (std::set<unsigned int>::const_iterator f_i = fu_set.begin(); f_i != f_end; f_i++)
   {
      num_res = tech_constraints[*f_i];
      THROW_ASSERT(num_res != 0, "something of wrong happen");
      if (num_res == INFINITE_UINT)
         return num_res;
      else
         tot_num_res += num_res;
   }
   return tot_num_res;
}


unsigned int allocation::min_number_of_resources(const vertex v, const graphRef g) const
{
   std::string current_op = vertex_to_op.find(v) != vertex_to_op.end() ? vertex_to_op.find(v)->second : GET_OP(g, v);
   const std::set<unsigned int> & fu_set = op_name_to_fus.find(current_op)->second;

   unsigned int min_num_res = INFINITE_UINT, num_res;
   const std::set<unsigned int>::const_iterator f_end = fu_set.end();

   for (std::set<unsigned int>::const_iterator f_i = fu_set.begin(); f_i != f_end; f_i++)
   {
      num_res = tech_constraints[*f_i];
      THROW_ASSERT(num_res != 0, "something of wrong happen");
      min_num_res = min_num_res > num_res ? num_res : min_num_res;
   }
   return min_num_res;
}

unsigned int allocation::get_memory_var(const unsigned int fu_name) const
{
   THROW_ASSERT(is_memory_unit(fu_name), "functional unit id not meaningful");
   return memory_units.find(fu_name)->second;
}

unsigned int updatecopy_HLS_constraints_functor::operator() (const unsigned int name) const
{
   return tech[name];
}

void updatecopy_HLS_constraints_functor::update(const unsigned int name, int delta)
{
   if (tech[name] == INFINITE_UINT)
      return;
   tech[name] = static_cast<unsigned int>(static_cast<int>(tech[name]) + delta);
}


std::pair<std::string, std::string> allocation::get_fu_name(unsigned int id)
{
   //std::cerr << id << std::endl;
   THROW_ASSERT(id_to_fu_names.find(id) != id_to_fu_names.end(), "Functional unit name not stored!");
   return id_to_fu_names[id];
}

void allocation::DECODE_OP_TYPE_PREC(std::string & op_name, std::string & op_type, unsigned int & op_prec, const std::string & to_split) const
{
   std::vector<std::string> split;
   boost::algorithm::split(split, to_split, boost::algorithm::is_any_of(":"));
   op_name = split[0];
   if (split.size() >= 2)
      op_type = split[1];
   else
      op_type = "";
   if (split.size() >= 3)
   {
      std::vector<unsigned int> precisions;
      for (std::vector<unsigned int>::size_type i = 2; i < split.size(); ++i)
         precisions.push_back(boost::lexical_cast<unsigned int>(split[i]));
      op_prec = *std::max_element(precisions.begin(), precisions.end());
   }
   else
      op_prec = 0;
}

void allocation::GET_NODE_TYPE_PREC(const vertex node, node_kind_prec_infoRef info) const
{
   const tree_managerRef TreeM = AppM->get_tree_manager();
   std::vector<application_manager::io_binding_type> vars_read = AppM->get_required_values(functionId, node);
   unsigned int first_valid_id = 0;
   if (vars_read.size() == 0) return;
   std::vector<unsigned int> node_ids;

   for (std::vector<application_manager::io_binding_type>::iterator itr = vars_read.begin(), end = vars_read.end(); itr < end; ++itr)
   {
      unsigned int id = std::tr1::get<0>(*itr);
      if (id && !first_valid_id) first_valid_id = id;
      node_ids.push_back(id);
   }

   if(!first_valid_id)
   {
      THROW_ASSERT(first_valid_id, "Unexpected pattern");
   }

   long long int vec_size = 0;
   bool is_a_pointer = false;
   bool is_a_function = false;
   unsigned int type_index = 0;
   for (std::vector<unsigned int>::iterator itr = node_ids.begin(), end = node_ids.end(); itr < end; ++itr)
   {
      if (*itr && !tree_helper::is_constant(TreeM, *itr))
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "node_ids " + STR(*itr));
         type_index = tree_helper::get_type_index(TreeM, *itr);
         if(tree_helper::is_an_array(TreeM, *itr) || tree_helper::is_a_struct(TreeM, type_index) || tree_helper::is_an_union(TreeM, type_index) || tree_helper::is_a_complex(TreeM, type_index))
            info->input_prec.push_back(32);
         else
         {
            unsigned int size_value = tree_helper::size(TreeM, type_index);
            info->input_prec.push_back(size_value);
         }
      }
      else
         info->input_prec.push_back(0);
   }

   type_index = tree_helper::get_type_index(TreeM, first_valid_id, vec_size, is_a_pointer, is_a_function);

   if(is_a_pointer || tree_helper::is_an_array(TreeM, type_index) || tree_helper::is_a_struct(TreeM, type_index) || tree_helper::is_an_union(TreeM, type_index) || tree_helper::is_a_complex(TreeM, type_index))
   {
      info->node_kind = "VECTOR_BOOL";
   }
   else if (tree_helper::is_int(TreeM, type_index))
   {
      info->node_kind = "INT";
   }
   else if (tree_helper::is_real(TreeM, type_index))
   {
      info->node_kind = "REAL";
   }
   else if (tree_helper::is_unsigned(TreeM, type_index))
   {
      info->node_kind = "UINT";
   }
   else if (tree_helper::is_bool(TreeM, type_index))
   {
      info->node_kind = "BOOL";
   }
   else
   {
      THROW_ERROR("not supported type: " + STR(type_index));
   }

   unsigned int nodeOutput_id = AppM->get_produced_value(functionId, node);
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "nodeOutput_id " + STR(nodeOutput_id));
   if (nodeOutput_id)
   {
      if(nodeOutput_id != application_manager::CONDITION_RESULT)
      {
         type_index = tree_helper::get_type_index(TreeM, nodeOutput_id);
         if(tree_helper::is_an_array(TreeM, type_index) || tree_helper::is_a_struct(TreeM, type_index) || tree_helper::is_an_union(TreeM, type_index) || tree_helper::is_a_complex(TreeM, type_index))
            info->output_prec = 32;
         else
            info->output_prec = tree_helper::size(TreeM, type_index);
      }
      else
         info->output_prec = 1;
   }
   /// Now we need to normalize the size to be compliant with the technology library assumptions
   unsigned int max_size_in =  resize_to_8_16_32_64(*std::max_element(info->input_prec.begin(), info->input_prec.end()));
   unsigned int max_size_out =  resize_to_8_16_32_64(info->output_prec);
   unsigned int max_io = std::max(max_size_in, max_size_out);
   size_t n_inputs = info->input_prec.size();
   for (unsigned int i = 0; i < n_inputs; ++i)
      if(info->input_prec[i] != 0)
         info->input_prec[i] = max_io;
   info->output_prec =max_io;

}


inline void allocation::print_allocated_resources(const graphRef & DEBUG_PARAMETER(g)) const
{
#ifndef NDEBUG
   if (debug_level >= DEBUG_LEVEL_VERBOSE)
   {
      unsigned int i;

      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "\nDumping the list of all the fixed bindings FU <-> node");
      i = 0;
      for (std::tr1::unordered_map<vertex, unsigned int>::const_iterator bind = binding.begin(); bind != binding.end(); bind++)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "  . Entry number: " + STR(++i) + " (vertex " + GET_NAME(g, bind->first) + ")");
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "    . Corresponding operation: " + GET_OP(g, bind->first));
         functional_unit* fu = dynamic_cast<functional_unit*>(GetPointer<functional_unit>(list_of_FU[bind->second]));
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "    . Vertex bound to: " + fu->get_name());
      }

      PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Dumping the list of all the possible bindings FU <-> node");
      i = 0;
      for (std::tr1::unordered_map<std::string, std::set<unsigned int> >::const_iterator bind = op_name_to_fus.begin(); bind != op_name_to_fus.end(); bind++)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "  . Entry number: " + STR(++i) + " (encoded operation " + bind->first + ")");
         std::string op_name;
         std::string op_type;
         unsigned int op_prec;
         DECODE_OP_TYPE_PREC(op_name, op_type, op_prec, bind->first);
         PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "    . Decoded string");
         if (op_name != "")
            PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "      . Operation name: " + op_name);
         if (op_type != "")
            PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "      . Operation type: " + op_type);
         if (op_prec != 0)
            PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "      . Operation precision: " + STR(op_prec));
         PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "    . Operation can be implemented by the following FUs:");
         for (std::set<unsigned int>::const_iterator fu_id = bind->second.begin(); fu_id != bind->second.end(); fu_id++)
         {
            functional_unit* fu = dynamic_cast<functional_unit*>(GetPointer<functional_unit>(list_of_FU[*fu_id]));
            PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "      . FU name: " + fu->get_name());
         }
      }
   }
#endif
}

technology_nodeRef allocation::get_fu(const std::string& fu_name) const
{
   std::string library_name = HLS_T->get_technology_manager()->get_library(fu_name);
   return HLS_T->get_technology_manager()->get_fu(fu_name, library_name);
}

std::map<unsigned int, unsigned int> allocation::get_memory_units() const
{
   return memory_units;
}

bool allocation::is_operation_bounded(const graphRef g, const vertex& op, unsigned int fu_type) const
{
   const technology_nodeRef node = get_fu(fu_type);
   std::string op_string = GET_OP(g, op);
   const functional_unit* fu = GetPointer<functional_unit>(node);
   const technology_nodeRef op_node = fu->get_operation(op_string);
   return GetPointer<operation>(op_node)->is_bounded();
}

allocationRef allocation::create_solver(const ParameterRef Param, const hlsRef HLS)
{
   return allocationRef(new allocation(Param, HLS));
}
