/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file parametric_list_based.cpp
 * @brief Class implementation of the parametric_list_based structure.
 *
 * This file implements some of the parametric_list_based member functions.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

#include "parametric_list_based.hpp"
// #include "call_graph.hpp"
#include "function_behavior.hpp"
#include "op_node_info.hpp"
#include "exceptions.hpp"
#include "utility.hpp"

#include "schedule.hpp"
#include "fu_binding.hpp"
#include "hls.hpp"
#include "hls_constraints.hpp"
#include "allocation.hpp"
#include "ASLAP.hpp"

#include "tree_basic_block.hpp"
#include "basic_block.hpp"

#include "BambuParameter.hpp"
#include "cpu_time.hpp"


/**
 * Functor used to compare two vertices with respect to an order based on the control steps associated with the vertices.
 */
struct cs_ordering_functor
{
   private:
      ///copy of the order: control step associated with the vertices.
      const vertex2obj<double> & order;

   public:
      /**
       * functor function used to compare two vertices with respect to the control step associated with the vertices.
       * @param a is the first vertex
       * @param bis the second vertex
       * @return true when order(a) > order(b)
       */
      bool operator() (const vertex & a, const vertex & b) const 
      {
         return order(a) < order(b) || (order(a) == order(b) && a < b);
      }

      /**
       * Constructor
       * @param o is the order.
       */
      cs_ordering_functor(const vertex2obj<double> & o) : order(o) {}

      /**
       * Destructor
       */
      ~cs_ordering_functor() {}

};

/**
 * The key comparison function for edge-integer set based on levels
 */
class edge_integer_order_by_map : std::binary_function<vertex, vertex, bool>
{
   private:
      /// Topological sorted vertices
      const std::map<vertex, unsigned int> & ref;

      /// Graph
      const graph * g;

   public:
      /**
       * Constructor
       * @param ref is the map with the topological sort of vertices
       * @param g is a graph used only for debugging purpose to print name of vertex
       */
      edge_integer_order_by_map(const std::map<vertex, unsigned int> & _ref, const graph * _g) :
         ref(_ref),
         g(_g)
      {}

      /**
       * Compare position of two vertices in topological sorted
       * @param x is the first pair of vertex
       * @param y is the second pair of vertex
       * @return true if x precedes y in topological sort, false otherwise
       */
         bool operator()(const std::pair<std::pair<vertex, vertex>, unsigned int> & x, const std::pair<std::pair<vertex, vertex>, unsigned int> & y) const
      {
         THROW_ASSERT(ref.find(x.first.first) != ref.end(), "Vertex " + GET_NAME(g,x.first.first) + " is not in topological_sort");
         THROW_ASSERT(ref.find(y.first.first) != ref.end(), "Vertex " + GET_NAME(g,y.first.first) + " is not in topological_sort");
         THROW_ASSERT(ref.find(x.first.second) != ref.end(), "Vertex " + GET_NAME(g,x.first.second) + " is not in topological_sort");
         THROW_ASSERT(ref.find(y.first.second) != ref.end(), "Vertex " + GET_NAME(g,y.first.second) + " is not in topological_sort");
         if(ref.find(x.first.first)->second != ref.find(y.first.first)->second)
            return ref.find(x.first.first)->second < ref.find(y.first.first)->second;
         else if(ref.find(x.first.second)->second != ref.find(y.first.second)->second)
            return ref.find(x.first.second)->second < ref.find(y.first.second)->second;
         else
            return x.second < x.second;
      }
};

static
bool check_if_is_live_in_next_cycle(const std::set<vertex, cs_ordering_functor> &live_vertices, unsigned int current_cycle, const vertex2obj<double> &ending_time, double clock_cycle)
{
   std::set<vertex, cs_ordering_functor>::const_iterator live_vertex_it = live_vertices.begin();
   while(live_vertex_it != live_vertices.end())
   {
      if(ending_time(*live_vertex_it) > (current_cycle+1) * clock_cycle)
      {
         return true;
      }
      ++live_vertex_it;
   }
   return false;
}

parametric_list_based::parametric_list_based(const ParameterRef _Param, const hlsRef _HLS, bool _speculation, LB_method met) :
   scheduling(_Param, _HLS, _speculation),
   lb_meth(met)
{

}

parametric_list_based::~parametric_list_based()
{

}

void parametric_list_based::exec(const std::set<vertex>& operations, unsigned int current_cycle)
{
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Executing parametric_list_based::exec...");

   ///The scheduling
   schedule & LB = *(HLS->Rsch);

   ///Current ASAP
   vertex2obj<unsigned int> current_ASAP;

   ///The binding
   fu_binding & res_binding = *(HLS->Rfu);

   ///Number of predecessors of a vertex already scheduled
   vertex2obj<double> scheduled_predecessors;

   ///The control dependence graph
   cdg = HLS->get_graph(function_behavior::CDG);

   ///Vertices sorted according topological sort of cdg
   unsigned int counter = 0;
   const std::deque<vertex>& levels = HLS->FB->get_levels();
   for(std::deque<vertex>::const_iterator it = levels.begin(); it != levels.end(); it++)
      cdg_sorted[*it] = ++counter;

   ///The clock cycle
   clock_cycle = HLS->HLS_C->get_clock_period();

   ///The set of resources for which the list of ready operations is not empty
   std::set<unsigned int> ready_resources;

   ///Set of ready operations (needed for update priority)
   std::set<vertex> ready_vertices;

   ///vertices still live at the beginning of the current control step
   cs_ordering_functor et_order(ending_time);
   std::set<vertex, cs_ordering_functor> live_vertices(et_order);

   ///select the type of graph
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "   . Selecting the type of the graph...");
   switch(lb_meth)
   {
      case static_mobility:
      case static_fixed:
      case dynamic_mobility:
         if(speculation)
            flow_graph = HLS->FB->get_subgraph(function_behavior::SG, operations);
         else
            flow_graph = HLS->FB->get_subgraph(function_behavior::FLSAODG, operations);
         break;
      default:
         THROW_ERROR("not supported list based scheduling method");
   }

   ///Number of operation to be scheduled
   size_t operations_number = operations.size();

   long int cpu_time;
   /// compute asap and alap
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "   . Computing asap and alap...");
   START_TIME(cpu_time);
   ASLAPRef aslap;
   switch(lb_meth)
   {
      case static_mobility:
      case dynamic_mobility:
      {
         aslap = ASLAPRef(new ASLAP(HLS, speculation, debug_level, operations));
         if(HLS->Rsch->num_scheduled() == 0)
         {
            aslap->compute_ASAP(HLS->HLS_C, HLS->ALL);
            aslap->compute_ALAP(HLS, ASLAP::ALAP_fast);
         }
         else
         {
            aslap->compute_ASAP(HLS->HLS_C, HLS->ALL, HLS->Rsch);
            unsigned int est_upper_bound = static_cast<unsigned int>(operations_number);
            aslap->compute_ALAP(HLS, ASLAP::ALAP_with_partial_scheduling, HLS->Rsch, 0, est_upper_bound);
         }
         break;
      }
      case static_fixed:
         break;
      default:
         THROW_ERROR("not supported list based scheduling method");
   }
   STOP_TIME(cpu_time);

   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "   . Computing free input vertices...");
   ///compute the set of vertices without input edges.
   /// At least one vertex is expected
   for(std::set<vertex>::const_iterator vi = operations.begin(); vi != operations.end(); vi++)
   {
      ///Skip vertex if it is not in the current subgraph
      if(!flow_graph->is_in_subset(*vi))
         continue;
      ///Updating structure for already scheduled operations
      if(LB.is_scheduled(*vi))
      {
         ///check for chaining
         compute_starting_ending_time(*vi, res_binding.get_assign(*vi), LB.get_cstep(*vi), starting_time[*vi], ending_time[*vi]);
         live_vertices.insert(*vi);

      }
      else if(boost::in_degree(*vi, *flow_graph) <= 0)
      {
         ready_vertices.insert(*vi);
         if(GET_TYPE(flow_graph, *vi) == TYPE_ENTRY)
            entry_vertex = *vi;
      }
      else
      {
         ///Check if all its predecessors have been scheduled. In this case the vertex is ready
         InEdgeIterator ei, ei_end;
         for(boost::tie(ei, ei_end) = boost::in_edges(*vi, *flow_graph); ei != ei_end; ei++)
         {
            vertex source = boost::source(*ei, *flow_graph);
            if(!LB.is_scheduled(source))
               break;
         }
         if(ei == ei_end)
         {
            ready_vertices.insert(*vi);
         }
      }
   }
   THROW_ASSERT(ready_vertices.size(), "At least one vertex is expected");
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "   . Number of vertices: " + STR(ready_vertices.size()));

   ///Setting priority
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "   . Setting priority...");
   refcount<priority_data<int> > Priority;
   switch(lb_meth)
   {
      case static_mobility:
         Priority = refcount<priority_data<int> >(new priority_static_mobility(aslap));
         break;
      case dynamic_mobility:
      {
         Priority = refcount<priority_data<int> >(new priority_dynamic_mobility(aslap, ready_vertices));
         break;
      }
      case static_fixed:
      {
         std::tr1::unordered_map<vertex, int> priority_value; 
         std::tr1::unordered_map<std::string, int> string_priority_value = HLS->HLS_C->get_scheduling_priority();
         VertexIterator ki, ki_end;
         for (boost::tie(ki, ki_end) = boost::vertices(*flow_graph); ki != ki_end; ki++)
         {
            priority_value[*ki] = string_priority_value[GET_NAME(flow_graph, *ki)];
         }
         Priority = refcount<priority_data<int> >(new priority_fixed(priority_value));
         break;
      }
      default:
         THROW_ERROR("not supported list based scheduling method");
   }

   ///priory queues set up
   unsigned int n_resources = HLS->ALL->get_number_fu_types();
   priority_compare_functor<int>  priority_functors(*Priority, flow_graph);
   std::vector<rehashed_heap<int> > priority_queues(n_resources, priority_functors);

   std::set<vertex>::const_iterator rv, rv_end = ready_vertices.end();

   for(rv = ready_vertices.begin(); rv != rv_end; rv++)
      add_to_priority_queues(priority_queues, ready_resources, *rv);

   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "   . Starting scheduling...");
   unsigned int already_sch = LB.num_scheduled();
   bool unbounded = false;
   while((LB.num_scheduled() - already_sch) != operations_number)
   {
      unbounded = false;
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "      . LB.num_scheduled() " + boost::lexical_cast<std::string>(LB.num_scheduled()));
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "      . already_sch " + boost::lexical_cast<std::string>(already_sch));
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "      . operations_number " + boost::lexical_cast<std::string>(operations_number));
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "      . Scheduling in control step " + boost::lexical_cast<std::string>(current_cycle) + " (Time: " + boost::lexical_cast<std::string>(current_cycle * clock_cycle)+ ")");
      /// definition of the data structure used to check if a resource is available given a vertex
      ///in case a vertex is not included in a map this mean that the used resources are zero.
      ///First index is the functional unit type, second index is the controller node, third index is the condition
      std::tr1::unordered_map<unsigned int, unsigned int> used_resources;

      ///Operations which can be scheduled in this control step because precedence has satisfied, but can't be scheduled in this control step
      ///Index is the functional unit type
      std::tr1::unordered_map<unsigned int, std::tr1::unordered_set<vertex> > black_list;

      ///Adding information about operation still live
      std::set<vertex, cs_ordering_functor>::const_iterator live_vertex_it = live_vertices.begin();
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "         . Considering live vertices...");
      while(live_vertex_it != live_vertices.end())
      {
         if(ending_time(*live_vertex_it) <= current_cycle * clock_cycle)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "            . Vertex " + GET_NAME(flow_graph, *live_vertex_it) + " dies");
            live_vertices.erase(*live_vertex_it);
            live_vertex_it = live_vertices.begin();
         }
         else
         {
            bool schedulable;
            if(used_resources.find(res_binding.get_assign(*live_vertex_it)) == used_resources.end())
                  used_resources[res_binding.get_assign(*live_vertex_it)] = 0;
            schedulable = BB_update_resources_use(used_resources[res_binding.get_assign(*live_vertex_it)], *live_vertex_it, starting_time[*live_vertex_it], res_binding.get_assign(*live_vertex_it));
            if(!schedulable)
               THROW_ERROR("Unfeasible scheduling");
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "            . Vertex " + GET_NAME(flow_graph, *live_vertex_it) + " lives until " + boost::lexical_cast<std::string>(ending_time(*live_vertex_it)));
            live_vertex_it++;
         }
      }

      for(unsigned int fu_type = 0; fu_type < n_resources; fu_type++)
      {
         if(priority_queues[fu_type].size())
            ready_resources.insert(fu_type);
      }


      std::tr1::unordered_map<unsigned int, std::tr1::unordered_set<vertex> > postponed_resources;
      bool do_again;

      do
      {
         do_again = false;
         while(ready_resources.size())
         {
            unsigned int fu_type = *(ready_resources.begin());
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "         . Considering functional unit type " + boost::lexical_cast<std::string>(fu_type));
            ready_resources.erase(fu_type);
            rehashed_heap<int> & queue = priority_queues[fu_type];
            ///Ignore empty list
            while(queue.size())
            {
               vertex current_vertex = queue.top();
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "               . First operation ready for this unit is " + GET_NAME(flow_graph, current_vertex));
               queue.pop();
               ///Ignore already scheduled operation
               if(LB.is_scheduled(current_vertex))
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "               . Already scheduled");
                  ///remove current_vertex from the queue
                  continue;
               }
               bool is_live = check_if_is_live_in_next_cycle(live_vertices, current_cycle, ending_time, clock_cycle);
               THROW_ASSERT(!(GET_TYPE(flow_graph,current_vertex) & (TYPE_WHILE | TYPE_FOR)), "not expected operation type");
               ///put these type of operations as last operation scheduled for the basic block
               if((GET_TYPE(flow_graph,current_vertex) & (TYPE_IF | TYPE_RET | TYPE_SWITCH | TYPE_GOTO)) &&
                  ((LB.num_scheduled() - already_sch) != operations_number-1))
               {
                  postponed_resources[fu_type].insert(current_vertex);
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "            . Scheduling of Control Vertex " + GET_NAME(flow_graph, current_vertex) + " postponed ");
                  continue;
               }
               if((GET_TYPE(flow_graph,current_vertex) & (TYPE_IF | TYPE_RET | TYPE_SWITCH | TYPE_GOTO)) &&
                  (unbounded || is_live))
               {
                  black_list[fu_type].insert(current_vertex);
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "            . Scheduling of Control Vertex " + GET_NAME(flow_graph, current_vertex) + " postponed to the next cycle");
                  continue;
               }

               ///true if operation is schedulable
               bool schedulable = false;

               ///starting time of the operation
               double current_starting_time;

               ///ending time of the operation
               double current_ending_time;

               ///checking if predecessors have finished
               if(current_ASAP[current_vertex] > current_cycle)
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "                  . Depends on a live operation");
                  black_list[fu_type].insert(current_vertex);
                  continue;
               }

               ///checking if chaining is feasible
               compute_starting_ending_time(current_vertex, fu_type, current_cycle, current_starting_time, current_ending_time);
               if((current_starting_time == ((current_cycle + 1) * clock_cycle)) || ((current_starting_time > (current_cycle * clock_cycle)) && current_ending_time > (current_cycle + 1) * clock_cycle))
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "                  . Can not be chained: starting time is " + boost::lexical_cast<std::string>(current_starting_time) + " ending time is " + boost::lexical_cast<std::string>(current_ending_time)) ;
                  black_list[fu_type].insert(current_vertex);
                  continue;
               }
               else if ((current_starting_time > (current_cycle * clock_cycle)) && (GET_TYPE(flow_graph,current_vertex) & TYPE_RET))
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "                  . Chaining of return expression operation is not allowed by construction");
                  black_list[fu_type].insert(current_vertex);
                  continue;
               }
               else if ((current_starting_time > (current_cycle * clock_cycle)) && !HLS->Param->getOption<bool>(OPT_chaining))
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "                  . Chaining is possible but not allowed -> starting time " + boost::lexical_cast<std::string>(current_starting_time) + " ending time: " + boost::lexical_cast<std::string>(current_ending_time));
                  black_list[fu_type].insert(current_vertex);
                  continue;
               }
               else if(current_starting_time == current_cycle * clock_cycle && current_ending_time > (current_cycle + 1) * clock_cycle && unbounded)
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "                  . Multi-cycles operations cannot be scheduled together with unbounded operations");
                  black_list[fu_type].insert(current_vertex);
                  continue;
               }

               /// check if there exist enough resources available
               schedulable = BB_update_resources_use(used_resources[fu_type], current_vertex, current_starting_time, fu_type);

               if (!HLS->ALL->is_operation_bounded(flow_graph, current_vertex, fu_type))
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "                  . " + GET_NAME(flow_graph, current_vertex) + " is unbounded");
                  if (unbounded || live_vertices.size() > 0) schedulable = false;
                  else unbounded = true;
               }
               else
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "                  . " + GET_NAME(flow_graph, current_vertex) + " is bounded");
               }
               if(schedulable)
               {
                  ///set the schedule information
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "                  . " + GET_NAME(flow_graph, current_vertex) + " scheduled at " + boost::lexical_cast<std::string>(current_cycle*clock_cycle));
                  LB.set_execution(current_vertex, current_cycle);
                  starting_time[current_vertex] = current_starting_time;
                  if(unbounded)
                     ending_time[current_vertex] = clock_cycle*ceil(current_ending_time/clock_cycle) + (current_starting_time == clock_cycle*ceil(current_ending_time/clock_cycle) ? clock_cycle : 0);
                  else
                     ending_time[current_vertex] = current_ending_time;
                  ready_vertices.erase(current_vertex);

                  ///set the binding information
                  if (HLS->HLS_C->has_binding_to_fu(GET_NAME(flow_graph, current_vertex)))
                     res_binding.bind(current_vertex, fu_type, 0);
                  else
                     res_binding.bind(current_vertex, fu_type);

                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "                  . Current cycles ends at " + boost::lexical_cast<std::string>((current_cycle + 1) * clock_cycle) + "  - operation ends at " + boost::lexical_cast<std::string>(ending_time[current_vertex]));
                  if(ending_time[current_vertex] > (current_cycle + 1) * clock_cycle)
                     live_vertices.insert(current_vertex);

                  ///Check if some successors have become ready
                  OutEdgeIterator eo, eo_end;

                  std::list<std::pair<std::string, vertex> > successors;
                  for(boost::tie(eo, eo_end) = boost::out_edges(current_vertex, *flow_graph); eo != eo_end; eo++)
                  {
                     vertex target = boost::target(*eo, *flow_graph);
                     successors.push_back(std::make_pair(GET_NAME(flow_graph, target), target));
                  }
                  //successors.sort();

                  for(std::list<std::pair<std::string, vertex> >::iterator s = successors.begin(); s != successors.end(); s++)
                  {
                     PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "                  . Considering successor " + s->first);
                     scheduled_predecessors[s->second]++;
                     current_ASAP[s->second] = MAX(static_cast<unsigned int>(floor(ending_time(current_vertex) / clock_cycle)), current_ASAP[s->second]);
                     ///check if to_v should be considered as ready
                     if(boost::in_degree(s->second, *flow_graph) == scheduled_predecessors(s->second))
                     {
                        PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "                     . Become ready");
                        ready_vertices.insert(s->second);
                        add_to_priority_queues(priority_queues, ready_resources, s->second);
                     }
                  }
                  continue;
               }
               else
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "               . No free resource");
                  black_list[fu_type].insert(current_vertex);
               }
            }
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "         . Finished with unit type " + boost::lexical_cast<std::string>(fu_type));
         }
         if(!postponed_resources.empty())
         {
            std::tr1::unordered_map<unsigned int, std::tr1::unordered_set<vertex> >::const_iterator bl_end = postponed_resources.end();
            for(std::tr1::unordered_map<unsigned int, std::tr1::unordered_set<vertex> >::const_iterator bl_it = postponed_resources.begin(); bl_it != bl_end; ++bl_it)
            {
               std::tr1::unordered_set<vertex>::const_iterator v_end = bl_it->second.end();
               for(std::tr1::unordered_set<vertex>::const_iterator v = bl_it->second.begin(); v != v_end; ++v)
               {
                  priority_queues[bl_it->first].push(*v);
               }
               ready_resources.insert(bl_it->first);
            }
            postponed_resources.clear();
            if(black_list.empty())
            {
               do_again = true;
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "         . Restarted the schedulign loop to accomodate postponed vertices");
            }
         }
      } while(do_again);

      std::tr1::unordered_map<unsigned int, std::tr1::unordered_set<vertex> >::const_iterator bl_end = black_list.end();
      for(std::tr1::unordered_map<unsigned int, std::tr1::unordered_set<vertex> >::const_iterator bl_it = black_list.begin(); bl_it != bl_end; ++bl_it)
      {
         std::tr1::unordered_set<vertex>::const_iterator v_end = bl_it->second.end();
         for(std::tr1::unordered_set<vertex>::const_iterator v = bl_it->second.begin(); v != v_end; ++v)
         {
            priority_queues[bl_it->first].push(*v);
         }
      }
      /// update priorities and possibly rehash the queues
      if(Priority->update())
         for(unsigned int i = 0; i < n_resources; i++)
            priority_queues[i].rehash();
      ///move to the next cycle
      current_cycle++;
   }

   unsigned int steps = current_cycle;
   LB.set_csteps(steps);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "   . List Based finished");
}

std::string parametric_list_based::get_kind_text() const
{
   std::string name;
   switch(lb_meth)
   {
      case static_mobility:
         name = "Static ";
         break;
      case static_fixed:
         name = "Priority-Fixed ";
         break;
      case dynamic_mobility:
         name = "Dynamic mobility based ";
         break;
      default:
         THROW_ERROR("not supported list based scheduling method");
   }
   return name + "List-Based Scheduling";
}

bool parametric_list_based::check_for_chaining(unsigned int cs, const vertex& op, unsigned int , schedule & LB) const
{
   //double current_ending_time = cs * clock_cycle;
   InEdgeIterator ei, ei_end;
   for(boost::tie(ei, ei_end) = boost::in_edges(op, *flow_graph); ei != ei_end; ei++)
   {
      vertex from_vertex = boost::source(*ei, *flow_graph);
      if (cs == LB.get_cstep(from_vertex) && GET_NAME(flow_graph, from_vertex) != ENTRY)
         return true;
   }
   return false;
}

void parametric_list_based::compute_starting_ending_time(vertex v, unsigned int fu_type, unsigned int cs, double & current_starting_time, double & current_ending_time) const
{
   current_ending_time = cs * clock_cycle;
   InEdgeIterator ei, ei_end;
   for(boost::tie(ei, ei_end) = boost::in_edges(v, *flow_graph); ei != ei_end; ei++)
   {
      vertex from_vertex = boost::source(*ei, *flow_graph);
      current_ending_time = MAX(current_ending_time, ending_time(from_vertex));
   }
   current_starting_time = current_ending_time;
   current_ending_time += HLS->ALL->get_execution_time(fu_type, v, flow_graph);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "                  . Operation " + GET_NAME(flow_graph, v) + " starts at " + boost::lexical_cast<std::string>(current_starting_time) + " and ends at " + boost::lexical_cast<std::string>(current_ending_time));
}

bool parametric_list_based::BB_update_resources_use(unsigned int & used_resources, const vertex & DEBUG_PARAMETER(current_vertex), const double & DEBUG_PARAMETER(current_starting_time), const unsigned int fu_type) const
{
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "               . Checking if operation " + GET_NAME(flow_graph, current_vertex) + " is schedulable in this step on " + boost::lexical_cast<std::string>(fu_type) + " - starting_time is " + boost::lexical_cast<std::string>(current_starting_time));
   if(used_resources == HLS->ALL->get_number_fu(fu_type))
      return false;
   else
   {
      used_resources++;
      return true;
   }
}

void parametric_list_based::add_to_priority_queues(std::vector<rehashed_heap<int> > & priority_queue, std::set<unsigned int> & ready_resources, const vertex v) const
{
   unsigned int fu_name;
   if(HLS->ALL->is_vertex_bounded_with(v, fu_name))
   {
      priority_queue[fu_name].push(v);
      ready_resources.insert(fu_name);
   }
   else
   {
      const std::set<unsigned int> & fu_set = HLS->ALL->can_implement_set(v, flow_graph);
      const std::set<unsigned int>::const_iterator fu_set_it_end = fu_set.end();
      for(std::set<unsigned int>::const_iterator fu_set_it = fu_set.begin(); fu_set_it != fu_set_it_end; fu_set_it++)
      {
         priority_queue[*fu_set_it].push(v);
         ready_resources.insert(*fu_set_it);
      }
   }
}

void parametric_list_based::exec()
{
#if 1
   const bb_graph* bbg = HLS->FB->get_bb_graph();
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Number of basic blocks: " + STR(boost::num_vertices(*bbg)));
   std::deque<vertex> vertices;
   boost::topological_sort(*bbg, std::front_inserter(vertices));
   std::deque<vertex>::const_iterator viend = vertices.end();
   unsigned int ctrl_steps = 0;
   for(std::deque<vertex>::const_iterator vi = vertices.begin(); vi != viend; vi++)
   {
      std::set<vertex> operations;
      std::list<vertex> bb_operations = CGET_BB_NODE_INFO(bbg, *vi)->statements_list;
      for(std::list<vertex>::iterator l = bb_operations.begin(); l != bb_operations.end(); l++)
      {
         for(std::tr1::unordered_set<vertex>::const_iterator o = HLS->operations.begin(); o != HLS->operations.end(); o++)
            if (*l == *o) operations.insert(*l);
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "performing scheduling of basic block " + STR(CGET_BB_NODE_INFO(bbg, *vi)->block->number));
      PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "  .operations: " + STR(operations.size()));
      exec(operations, ctrl_steps);
      ctrl_steps = HLS->Rsch->get_csteps();
   }
   HLS->Rsch->set_csteps(ctrl_steps);
   PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Number of control steps: "+boost::lexical_cast<std::string>(ctrl_steps));
#else
   std::set<vertex> operations;
   for(std::tr1::unordered_set<vertex>::const_iterator l = HLS->operations.begin(); l != HLS->operations.end(); l++)
   {
      operations.insert(*l);
   }
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "  .operations: " + STR(operations.size()));
   exec(operations);
#endif
}

