/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file parametric_list_based.hpp
 * @brief Class definition of the list_based structure.
 *
 * This file defines the class performing the list_based scheduling.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

#ifndef PARAMETRIC_LIST_BASED_HPP
#define PARAMETRIC_LIST_BASED_HPP

#include <iostream>
#include <set>
#include <map>
#include <vector>

#include "rehashed_heap.hpp"
#include "scheduling.hpp"
#include "Vertex.hpp"
#include "refcount.hpp"
#include <tr1/unordered_map>
/**
 * @name forward declarations
*/
//@{
REF_FORWARD_DECL(hls);
REF_FORWARD_DECL(graph);
class schedule;
class graph;
//@}

/**
 * Class managing list based scheduling algorithms.
*/
class parametric_list_based : public scheduling
{
   public:
      enum LB_method
      {
         dynamic_mobility = 0,
         static_fixed,
         static_mobility
      };

      /**
       * This is the constructor of the list_based.
       * @param met is the method used to perform the list based scheduling.
       * @param speculation is the flag to check speculation
      */
      parametric_list_based(const ParameterRef Param, const hlsRef HLS, bool speculation, LB_method met);

      /**
       * Destructor.
      */
      ~parametric_list_based();

      /**
       * Function that computes the List-Based scheduling of the graph.
       * @param HLS is the high level synthesis global data structure.
       */
      virtual void exec();

      /**
      * Function printing the name of the scheduling performed.
      */
      virtual std::string get_kind_text() const;

      /**
       * Function that computes the List-Based scheduling of the graph.
       */
      void exec(const std::set<vertex>& operations, unsigned int current_cycle);

   private:

      /// method used to perform the list based scheduling
      LB_method lb_meth;

      ///The dependence graph
      graphRef flow_graph;

      ///The control dependence graph
      graphRef cdg;

      ///Entry vertex
      vertex entry_vertex;

      ///Topological sort of the control dependence graphs
      std::map<vertex, unsigned int> cdg_sorted;

      ///The starting time given the scheduling (used for chaining)
      vertex2obj<double> starting_time;

      ///The ending time given the scheduling (used for chaining)
      vertex2obj<double> ending_time;

      ///The clock cycle
      double clock_cycle;

      /**
       * Given the control step in which an operation is scheduled, compute the exact starting and ending time of an operation
       * @param v is the vertex of the operation
       * @param fu_type is the functional unit type
       * @param control_step is the control_step
       * @param starting_time is where starting_time will be stored
       * @param ending_time is where ending_time will be stored
      */
      void compute_starting_ending_time(const vertex v, const unsigned int fu_type, const unsigned int control_step, double & starting_time, double & ending_time) const;

      /**
       * Update the resource map
       * @param used_resources_fu is the resource map.
       * @param current_vertex is the vertex for which an update is required.
       * @param current_starting_time is the starting time of the operation of current_vertex
       * @param fu_type is the functional unit type on which operation is scheduled
       * @return true if the assignment is feasible
      */
      bool BB_update_resources_use(unsigned int & used_resources, const vertex & current_vertex, const double & current_starting_time, const unsigned int fu_type) const;

      /**
       * Adds the vertex v to the priority queues
       * @param priority_trees are the priority_queues
       * @param ready_resources is the set of resources which have at least one ready operation
       * @param v is the vertex.
      */
      void add_to_priority_queues(std::vector<rehashed_heap<int> > & priority_queue, std::set<unsigned int> & ready_resources, const vertex v) const;

      /**
       * Check if it is possible to perform chaining
       */
      bool check_for_chaining(unsigned int cs, const vertex& op, unsigned int fu_type, schedule & LB) const;

};
#endif
