/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file schedule.cpp
 * @brief Class implementation of the schedule data structure.
 *
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#include <ostream>
#include <boost/graph/graphviz.hpp>

#include "schedule.hpp"
#include "behavioral_writer_helper.hpp"
#include "graph.hpp"
#include "behavioral_helper.hpp"
#include "function_behavior.hpp"
#include "op_edge_info.hpp"

schedule::schedule() :
   tot_csteps(0)
{

}

schedule::~schedule()
{

}

void schedule::print(std::ostream& os, const graphRef data) const
{
   std::map<vertex, unsigned int>::const_iterator it, it_end = op_starting_cycle.end();
   for(it = op_starting_cycle.begin(); it != it_end; it++)
   {
      if(!data->is_in_subset(it->first)) continue;
      os << "Operation: ";
      os.width(COLUMN_SIZE);
      os.setf(std::ios_base::left, std::ios_base::adjustfield);
      os << GET_NAME(data, it->first) + "(" + GET_OP(data, it->first) + ")";
      os.width(0);

      os << " scheduled at control step (";
      os.width(0);
      os.setf(std::ios_base::left, std::ios_base::adjustfield);
      os << it->second;
      os << ")";
      os.width(3);

      os << std::endl;
   }
   os << "Total number of control steps: " << tot_csteps << std::endl;
}

class schedule_graph_property
{
   public:
      /**
       * Constructor
       * @param g is the graph used for the scheduling
       * @param sch is the schedule
       */
      schedule_graph_property(const graph &_g, const schedule &_sch) :
         g(_g),
         sch(_sch)
      {}

      /**
       * Redifinition of operator()
       */
      void operator()(std::ostream& os) const
      {
         os << "//Scheduling solution\n";
         os << "splines=true;\n";
         std::multimap<unsigned int, vertex> inverse_relation;
         VertexIterator v, v_end;
         for(boost::tie(v, v_end) = boost::vertices(g); v != v_end; v++)
            inverse_relation.insert(std::multimap<unsigned int, vertex>::value_type(sch.get_cstep(*v), *v));
         for (unsigned int level = 0; level < sch.get_csteps(); level++)
         {
            std::pair<std::multimap<unsigned int, vertex>::const_iterator, std::multimap<unsigned int, vertex>::const_iterator > p = inverse_relation.equal_range(level);
            if (p.first == p.second) continue;
            os << "//Control Step: " << level << std::endl;
            os << "CS" << level << " [style=plaintext]\n{rank=same; CS" << level << " ";
            for (std::multimap<unsigned int, vertex>::const_iterator i = p.first; i != p.second; ++i)
            {
               os << boost::get(boost::vertex_index_t(), g)[i->second] << " ";
            }
            os << ";}\n";
         }
         for (unsigned int level = 1; level < sch.get_csteps(); level++)
            os << "CS" << level - 1 << "-> CS" << level << ";\n";
      }
   private:
      const graph &g;
      const schedule & sch;

};

void schedule::write_dot(const std::string& file_name, const function_behaviorRef FB) const
{
   std::ofstream f(file_name.c_str());
   behavioral_helperRef helper = FB->get_helper();
   const graph * fsadg = FB->get_graph(function_behavior::FSADG);
   boost::write_graphviz(f, *fsadg, cdfg_label_writer(fsadg, helper), OpEdgePropertyWriter(fsadg, FSADG_SELECTOR, helper), schedule_graph_property(*fsadg, *this));
}

void schedule::set_execution(const vertex& op, unsigned int c_step)
{
   op_starting_cycle[op] = c_step;
}

bool schedule::is_scheduled(const vertex& op) const
{
   return op_starting_cycle.find(op) != op_starting_cycle.end();
}

unsigned int schedule::get_cstep(const vertex& op) const
{
   THROW_ASSERT(is_scheduled(op), "Operation has not been scheduled");
   return op_starting_cycle.find(op)->second;
}

unsigned int schedule::num_scheduled() const
{
   return static_cast<unsigned int>(op_starting_cycle.size());
}

void schedule::clear()
{
   tot_csteps = 0;
   spec.clear();
   op_starting_cycle.clear();
}

