/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file schedule.hpp
 * @brief Data structure used to store the schedule of the operations.
 *
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef SCHEDULE_HPP
#define SCHEDULE_HPP

#include "refcount.hpp"
/**
 * @name Forward declarations.
 */
//@{
REF_FORWARD_DECL(function_behavior);
REF_FORWARD_DECL(behavioral_manager);
REF_FORWARD_DECL(graph);
//@}

#include "graph.hpp"
#include <tr1/unordered_map>
#include <iostream>

/**
 * Class managing the schedule of the operations.
 */
class schedule
{
   private:

      ///total number of control steps
      unsigned int tot_csteps;

      ///map between the operation vertex and the corresponding initial clock cycle
      std::map<vertex, unsigned int> op_starting_cycle;

      ///Map for speculation property of each operation vertex. If true, it means that vertex is speculative executed,
      ///false otherwise
      std::tr1::unordered_map<vertex, bool> spec;

   public:

      /**
       * Constructor.
       */
      schedule();

      /**
       * Destructor.
       */
      ~schedule();

      /**
       * This method returns the number of control steps.
       * @return the number of control steps of the stored scheduling.
       */
      unsigned int get_csteps() const { return tot_csteps; }

      /**
       * This method sets the number of control steps.
       * @param cs is the number of control steps.
       */
      void set_csteps(unsigned int cs) { tot_csteps = cs; }

      /**
       * Function that prints the class schedule.
       */
      void print(std::ostream& os, const graphRef data) const;

      /**
       * Function that writes the dot file of the scheduling by using the AT&T direct graph representation.
       * @param file_name is the file name
       * @param BM is the behavioral_manager in which current must has been setted to the 
       * function to which this schedule refers to
       */
      void write_dot(const std::string& file_name, const function_behaviorRef FB) const;

      /**
       * Sets the speculation map
       */
      void set_spec(const std::tr1::unordered_map<vertex, bool>& spec_map) { spec = spec_map; }

      /**
       * Returns the speculation map
       */
      std::tr1::unordered_map<vertex,bool> get_spec() const { return spec; }

      /**
       * Sets the starting clock cycle for the given operation
       * @param op the the operation vertex
       * @param c_step is an integer representing the clock cycle where the operation starts the computation
       */
      void set_execution(const vertex& op, unsigned int c_step);

      /**
       * Returns true if the given operation has been already scheduled, false otherwise
       */
      bool is_scheduled(const vertex& op) const;

      /**
       * Returns the clock cycle where the given operation has been scheduled
       * @param op is the vertex of the operation
       * @return an integer representing the clock cycle where it starts the execution
       */
      unsigned int get_cstep(const vertex& op) const;

      /**
       * Returns the number of scheduled operations
       */
      unsigned int num_scheduled() const;

      /**
       * Erases the current results
       */
      void clear();

};
///Refcount definition of the class
typedef refcount<schedule> scheduleRef;

#endif
