/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file scheduling.hpp
 * @brief Base class for all scheduling algorithms.
 *
 * This class is a pure virtual one, that has to be specilized in order to implement a particular scheduling algorithm.
 * It has a internal attribute to check if a vertex is speculate. To specialize this class into an algorithm
 *
 * @author Matteo Barbati <mbarbati@gmail.com>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

#ifndef SCHEDULING_HPP
#define SCHEDULING_HPP

#include "HLS_step.hpp"

///Autoheader include
#include "config_HAVE_ILP_BUILT.hpp"

#include "graph.hpp"
#include <map>

#include "refcount.hpp"
/**
 * @name forward declarations
*/
//@{
REF_FORWARD_DECL(hls);
REF_FORWARD_DECL(scheduling);
REF_FORWARD_DECL(graph);
class xml_element;
//@}

#include <tr1/unordered_map>

/**
 * Generic class managing scheduling algorithms.
 */
class scheduling : public HLS_step
{
   private:

      /**
       * @name switch statements data structures and functions.
       */
      //@{
      /// store for each switch the number of outgoing branches.
      std::tr1::unordered_map<vertex, unsigned int > switch_map_size;

      /// for each controlling vertex defines a relation between switch tags and branch tags
      std::tr1::unordered_map<vertex, std::tr1::unordered_map< unsigned int, unsigned int> > switch_normalizing_map;
      //@}

   protected:

      /// Map for speculation property of each operation vertex. If true, it means that vertex is speculative executed,
      /// false otherwise
      std::tr1::unordered_map<vertex, bool> spec;

      /// flag to check speculation
      const bool speculation;

      /// verbosity level
      int output_level;

      /**
       * compute the branch tag given the edge and the control data flow graph.
       * @param e is the edge.
       * @param cdg is the controlling dependency graph.
       * @param switch_set is filled with the switch tags in case the source of e is a switch vertex.
       */
      unsigned int compute_b_tag(const EdgeDescriptor& e, const graphRef cdg, std::set<unsigned int>::const_iterator &switch_it, std::set<unsigned int>::const_iterator &switch_it_end) const;

      /**
       * Return the number of branches associated with the controlling vertex.
       * @param cdg is the control dependency graph.
       * @param controlling_vertex is the controlling vertex.
       */
      unsigned int compute_b_tag_size(const graphRef cdg, vertex controlling_vertex) const;

      /**
       * return an unsigned int from 0 to compute_b_tag_size(controlling_vertex)-1 given a branch tag not normalized.
       * @param controlling_vertex is the controlling vertex (used only for switch vertex).
       * @param b_tag_not_normalized is the non normalized tag.
       */
      unsigned int b_tag_normalize(vertex controlling_vertex, unsigned int b_tag_not_normalized) const;

      /**
       * initialize the maps associated with the switch vertices.
       * @param controlling_vertex is a switch vertex.
       * @param cdg is the control dependency graph.
       */
      void init_switch_maps(vertex controlling_vertex, const graphRef cdg);

      /**
       * Move up zero time operations in the previous step when possible
       * @param HLS is the structure storing all the High level synthesis information
       * @param dependence_graph is the graph to be scheduled
       * @return the new last step in which an operation is scheduled
       */
      unsigned int anticipate_operations(const graphRef dependence_graph);

   public:

      /// Scheduling algorithms implemented
      enum scheduling_type
      {
         LIST_BASED = 0,
#if HAVE_ILP_BUILT
         ILP,
         ILP_NEW_FORM,
         SILP,
         ILP_LOOP,
#endif
         FIXED
      };

      /**
       * This is the constructor of the scheduling class.
       * @param Param is the data structure containing all the parameters
       * @param HLS is the data structure storing all the HLS results
       * @param speculation is the flag to check speculation
       */
      scheduling(const ParameterRef Param, const hlsRef HLS, bool speculation);

      /**
       * Destructor.
       */
      virtual ~scheduling();

      /**
       * It returns speculation property map
       * @return the map associated
       */
      const std::tr1::unordered_map<vertex, bool>& get_spec() const { return spec; }

      /**
       * Creates solver based on selected option
       * @param type is the desidered scheduling solver
       * @param Param is the data structure containing all the parameters
       * @param HLS is the data structure storing all the HLS results
       * @return the reference to scheduling specialization
       */
      static
      schedulingRef create_solver(scheduling_type type, const ParameterRef Param, const hlsRef HLS);

      /**
       * Creates solver based on configuration stored in the XML node
       * @param node is the XML node where the configuration is stored
       * @param Param is the data structure containing all the parameters
       * @param HLS is the data structure storing all the HLS results
       * @return the reference to scheduling specialization
       */
      static
      schedulingRef xload(const xml_element* node, const ParameterRef Param, const hlsRef HLS);

};
///refcount definition of the class
typedef refcount<scheduling> schedulingRef;

#endif
