/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file HLStestbench.cpp
 * @brief Implementation about hls testbenches
 *
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/
///Autoheader include
#include "config_HAVE_FLOPOCO.hpp"

#include "config_PACKAGE_BUGREPORT.hpp"
#include "config_PACKAGE_NAME.hpp"
#include "config_PACKAGE_VERSION.hpp"

#include "HLStestbench.hpp"

#include "application_manager.hpp"
#include "tree_manager.hpp"
#include "tree_reindex.hpp"
#include "function_behavior.hpp"
#include "behavioral_helper.hpp"
#include "var_pp_functor.hpp"
#include "hls_constraints.hpp"
#include "memory.hpp"
#include "fu_binding.hpp"

#include "Parameter.hpp"
#include "dbgPrintHelper.hpp"
#include "simple_indent.hpp"
#include "utility.hpp"
#include "fileIO.hpp"
#include "polixml.hpp"
#include "xml_dom_parser.hpp"

#include "prettyPrintVertex.hpp"
#include "boost/filesystem.hpp"
#include "xml_helper.hpp"

#include "HDL_manager.hpp"
#include "language_writer.hpp"

#include "technology_builtin.hpp"

#include "gcc_wrapper.hpp"

/// ----------- C backend ----------- ///
#include "c_backend.hpp"
#include "c_writer.hpp"
#include "hls_instruction_writer.hpp"

#if HAVE_FLOPOCO
#include "flopoco_wrapper.hpp"
#include <mpfr.h>
namespace flopoco
{
   extern std::string fp2bin(mpfr_t x, int wE, int wF);
};
#endif

#include <fstream>
#include <iostream>

#define TESTBENCH_COPYING3_SHORT_NROW 15
const char* TESTBENCH_COPYING3_SHORT[TESTBENCH_COPYING3_SHORT_NROW] = {
   "************************************************************************\n",
   "This hardware description is free; you can redistribute it and/or modify\n",
   "it under the terms of the GNU General Public License as published by\n",
   "the Free Software Foundation; either version 3, or (at your option)\n",
   "any later version.\n",
   "\n",
   "This hardware description is distributed in the hope that it will be useful,\n",
   "but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY\n",
   "or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License\n",
   "for more details.\n",
   "\n",
   "You should have received a copy of the GNU General Public License\n",
   "along with the PandA framework; see the files COPYING\n",
   "If not, see <http://www.gnu.org/licenses/>.\n",
   "************************************************************************\n\n"
};

HLStestbench::HLStestbench(const ParameterRef& _Param) :
      Param(_Param),
      debug_level(_Param->getOption<int>(OPT_debug_level))
{
   output_directory = Param->getOption<std::string>(OPT_output_directory) + "/simulation/";
   if (!boost::filesystem::exists(output_directory)) boost::filesystem::create_directories(output_directory);
}

HLStestbench::~HLStestbench()
{

}

unsigned int HLStestbench::parse_test_vectors(const application_managerRef AppM, const std::string& tb_input_xml, std::vector<std::map<std::string, std::string> >& test_vectors)
{
   const tree_managerRef TreeM = AppM->get_tree_manager();
   unsigned int function_id = TreeM->function_index(Param->getOption<std::string>(OPT_top_function_name));
   const behavioral_helperRef BH = AppM->get_function_behavior(function_id)->get_helper();

   if (!boost::filesystem::exists(tb_input_xml))
   {
      THROW_WARNING("XML file \"" + tb_input_xml + "\" cannot be opened, creating a stub with random values");
      xml_document document;
      xml_element* nodeRoot = document.create_root_node("function");

      xml_element* node = nodeRoot->add_child_element("testbench");

      std::list<unsigned int> parameters = BH->get_parameters();
      for(std::list<unsigned int>::iterator p = parameters.begin(); p != parameters.end(); p++)
      {
         if (BH->is_a_pointer(*p)) continue;
         std::string param = BH->PrintVariable(*p);

         long long int value = (rand() % 20);
         node->set_attribute(param, boost::lexical_cast<std::string>(value));
      }

      document.write_to_file_formatted(tb_input_xml);
   }
   try
   {
      fileIO_istreamRef sname = fileIO_istream_open(tb_input_xml);
      xml_dom_parser parser;
      //parser.set_validate();
      parser.parse_stream(*sname);
      if (parser)
      {
         //Walk the tree:
         const xml_element* node = parser.get_document()->get_root_node(); //deleted by DomParser.
         const xml_node::node_list list = node->get_children();
         for (xml_node::node_list::const_iterator iter = list.begin(); iter != list.end(); ++iter)
         {
            const xml_element* Enode = GetPointer<const xml_element>(*iter);

            if(!Enode || Enode->get_name() != "testbench") continue;

            std::map<std::string, std::string> test_vector;

            std::list<unsigned int> parameters = BH->get_parameters();
            for(std::list<unsigned int>::iterator p = parameters.begin(); p != parameters.end(); p++)
            {
               std::string param = BH->PrintVariable(*p);
               PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Parameter: " + param + (BH->is_a_pointer(*p) ? " (memory access)" : " (input value)"));
               if ((Enode)->get_attribute(param))
               {
                  test_vector[param] = boost::lexical_cast<std::string>((Enode)->get_attribute(param)->get_value());
               }
               else if (!BH->is_a_pointer(*p))
               {
                  THROW_ERROR("Missing input value for parameter: " + param);
               }
            }
            test_vectors.push_back(test_vector);
         }

         return static_cast<unsigned int>(test_vectors.size());
      }
   }
   catch (const char * msg)
   {
      std::cerr << msg << std::endl;
   }
   catch (const std::string & msg)
   {
      std::cerr << msg << std::endl;
   }
   catch (const std::exception& ex)
   {
      std::cout << "Exception caught: " << ex.what() << std::endl;
   }
   catch ( ... )
   {
      std::cerr << "unknown exception" << std::endl;
   }
   THROW_ERROR("Error during parsing of testbench file " + tb_input_xml);
   return 0;
}

std::string HLStestbench::print_var_init(const behavioral_helperRef& BH, const tree_managerRef& TreeM, unsigned int var, const memoryRef& mem)
{
   std::vector<std::string> init_els;
   const tree_nodeRef& tn = TreeM->get_tree_node_const(var);
   tree_nodeRef init_node;
   var_decl * vd = GetPointer<var_decl>(tn);
   if(vd && vd->init)
      init_node = GET_NODE(vd->init);

   if (init_node && (!GetPointer<constructor>(init_node) || GetPointer<constructor>(init_node)->list_of_idx_valu.size()))
   {
      fu_binding::write_init(TreeM, tn, init_node, init_els, mem);
   }
   else if(GetPointer<string_cst>(tn) || GetPointer<integer_cst>(tn) || GetPointer<real_cst>(tn))
   {
      fu_binding::write_init(TreeM, tn, tn, init_els, mem);
   }
   else
   {
      if (BH->is_an_array(var))
      {
         unsigned int size_of_data;
         unsigned int type_index;
         std::vector<unsigned int> dims;
         tree_helper::get_type_node(tn, type_index);
         tree_helper::get_array_dimensions(TreeM, type_index, dims, size_of_data);
         unsigned int num_elements=1;
         std::string value;
         for(std::vector<unsigned int>::const_iterator it = dims.begin(); it != dims.end(); ++it)
            num_elements *= *it;
         for (unsigned int l = 0; l < num_elements; l++)
         {
            value = "";
            for (unsigned int i = 0; i < size_of_data; i++)
               value += "0";
            init_els.push_back(value);
         }
      }
      else
      {
         unsigned int type_index;
         tree_helper::get_type_node(tn, type_index);
          unsigned int size_of_data = tree_helper::size(TreeM, type_index);
         std::string value;
         for (unsigned int i = 0; i < size_of_data; i++)
            value += "0";
         init_els.push_back(value);
      }
   }
   std::string init;
   for (unsigned int l = 0; l < init_els.size(); l++)
   {
      if (l) init += ",";
      init += init_els[l];
   }
   //std::cerr << "variable: " << BH->PrintVariable(var) << std::endl;
   //std::cerr << "   init: " << init << std::endl;

   return init;
}

bool HLStestbench::is_all_8zeros(std::string & str)
{
   size_t size = str.size();
   if(size%8 != 0 || size == 8) return false;
   for(size_t i = 0; i < size; ++i)
      if(str.at(i) != '0')
         return false;
   return true;
}

std::string HLStestbench::creating_c_testbench(const application_managerRef AppM, const HLS_constraintsRef& HLS_C, std::map<std::string, std::string>& test_vector, const memoryRef& Rmem, const std::string& input_file)
{
   std::ostringstream os;
   simple_indent PP('{', '}', 3);
   CBackendRef backend(new CBackend(AppM, Param));
   CWriterRef cWriter = CWriterRef(new CWriter(AppM, Param, debug_level >= DEBUG_LEVEL_VERBOSE));
   InstructionWriterRef instrWriter = InstructionWriterRef(new HLSInstructionWriter(Param));
   cWriter->add_instruction_writer(instrWriter);
   backend->registerWriter(cWriter);
   backend->writeHeader();
   backend->writeIncludes();
   backend->WriteGlobalDeclarations();
   backend->writeImplementations();

   PP(os, "//Original pretty printed C description\n\n");
   PP(os, cWriter->getoutStream()->str() + "\n");
   PP(os, "//Testbench part for High-Level Synthesis\n\n");
   PP(os, "#include <stdio.h>\n\n");

   PP(os, "//global variable used to store the output file\n");
   PP(os, "FILE * global_testvector_file;\n");
   PP(os, "//variable used to detect a standard end of the main (exit has not been called)\n");
   PP(os, "unsigned int __standard_exit;\n");
   PP(os, "//definition of __exit function\n\n");
   PP(os, "void __exit(void)\n");
   PP(os, "{\n");
   PP(os, "if(!__standard_exit)\n");
   PP(os, "{\n");
   PP(os, "fprintf(global_testvector_file, \"//expected value for return value\\n\");\n");
   PP(os, "fprintf(global_testvector_file, \"o00000000000000000000000000000000\\n\");\n");
   PP(os, "fprintf(global_testvector_file, \"e\\n\");\n");
   PP(os, "}\n");
   PP(os, "}\n\n");

   PP(os, "void _Dec2Bin_(FILE * testvector_file, long long int num, unsigned int precision)\n");
   PP(os, "{\n");
   PP(os, "int i;\n");
   PP(os, "unsigned long long int ull_value = (unsigned long long int) num;\n");
   PP(os, "for(i=0;i<precision;++i)\n");
   PP(os, "fprintf(testvector_file, \"%c\", (((1LLU << (precision-i-1)) & ull_value) ? '1' : '0'));\n");
   PP(os, "}\n\n");

#if HAVE_FLOPOCO
   PP(os, "void _Ptd2Bin_(FILE * testvector_file, unsigned char * num, unsigned int precision)\n");
   PP(os, "{\n");
   PP(os, "int i,j;\n");
   PP(os, "char value;\n");
   PP(os, "for(i=0; i<precision; i=i+8)\n");
   PP(os, "{\n");
   PP(os, "value = *(num + precision/8-i/8-1);\n");
   PP(os, "for(j=0;j<8;++j)\n");
   PP(os, "fprintf(testvector_file, \"%c\", (((1LLU << (8-j-1)) & value) ? '1' : '0'));\n");
   PP(os, "}\n\n");
   PP(os, "}\n\n");
#endif

   PP(os, "int main()\n{\n\n");
   PP(os, "FILE * testvector_file;\n");
   PP(os, "unsigned int __testbench_index,__testbench_index0;\n");
   PP(os, "atexit(__exit);\n");
   PP(os, "__standard_exit=0;\n");
   PP(os, "global_testvector_file=testvector_file=fopen(\"" +  input_file + "\", \"a\");\n");
   PP(os, "fprintf(testvector_file, \"//base address " + STR(HLS_C->base_address) + "\\n\");\n");
   std::string trimmed_value;
   for(unsigned int ind = 0; ind < 32; ind++)
      trimmed_value = trimmed_value + (((1LLU << (31-ind)) & HLS_C->base_address) ? '1' : '0');
   PP(os, "fprintf(testvector_file, \"b" + trimmed_value + "\\n\");\n");
   const tree_managerRef TreeM = AppM->get_tree_manager();
   unsigned int function_id = TreeM->function_index(Param->getOption<std::string>(OPT_top_function_name));
   const behavioral_helperRef& BH = AppM->get_function_behavior(function_id)->get_helper();
   const std::set<unsigned int>& mem_vars = Rmem->get_memory_variables();
   std::map<unsigned int, unsigned int> address;
   for(std::set<unsigned int>::const_iterator m = mem_vars.begin(); m != mem_vars.end(); m++)
   {
      address[Rmem->get_external_base_address(*m)] = *m;
   }

   std::list<unsigned int> mem;
   for(std::map<unsigned int, unsigned int>::const_iterator ma = address.begin(); ma != address.end(); ma++)
   {
      //std::cerr << "address = " << ma->first << std::endl;
      //std::cerr << "  variable = " << ma->second << std::endl;
      //std::cerr << "  variable = " << BH->PrintVariable(ma->second) << std::endl;
      //std::cerr << "  size = " << tree_helper::size(TreeM, ma->second) << std::endl;
      //std::cerr << "  elements = " << tree_helper::size(TreeM, ma->second)/8 << std::endl;
      //std::cerr << "  end_address = " << ma->first+(tree_helper::size(TreeM, ma->second)/8) << std::endl;
      mem.push_back(ma->second);
   }
   const std::list<unsigned int>& parameters = BH->get_parameters();
   for(std::list<unsigned int>::const_iterator p = parameters.begin(); p != parameters.end(); p++)
   {
      if (BH->is_a_pointer(*p) && mem_vars.find(*p) == mem_vars.end())
      {
         mem.push_back(*p);
      }
   }
   size_t byte_allocated, actual_byte;
   for (std::list<unsigned int>::iterator l = mem.begin(); l != mem.end(); l++)
   {
      std::string param = BH->PrintVariable(*l);
      if(param[0] == '"')
         param = "@"+STR(*l);

      std::vector<std::string> splitted;
      std::string test_v;
      bool is_memory = false;
      unsigned int base_type = tree_helper::get_type_index(TreeM, *l);
      byte_allocated = 0;
      actual_byte = tree_helper::size(TreeM, base_type)/8;
      if (mem_vars.find(*l) != mem_vars.end() && std::find(parameters.begin(), parameters.end(), *l) == parameters.end())
      {
         is_memory = true;
         test_v = print_var_init(BH, TreeM, *l, Rmem);
         PP(os, "fprintf(testvector_file, \"//memory initialization for variable " + param + "\\n\");\n");
      }
      else if(test_vector.find(param) != test_vector.end())
         test_v = test_vector[param];
      else
         test_v ="0";
      unsigned int base_type_size=0;
      boost::algorithm::split(splitted, test_v , boost::algorithm::is_any_of(","));
      if (tree_helper::is_a_pointer(TreeM, *l) && !is_memory)
      {
         tree_nodeRef pt_node = TreeM->get_tree_node_const(base_type);
         THROW_ASSERT(pt_node->get_kind() == pointer_type_K, "A pointer type is expected");
         unsigned int ptd_base_type = GET_INDEX_NODE(GetPointer<pointer_type>(pt_node)->ptd);
         base_type_size = tree_helper::size(TreeM, ptd_base_type)/8;
         unsigned int reserve_space = (static_cast<unsigned int>(splitted.size()))*base_type_size;
         actual_byte = reserve_space;
         /// FIXME in case the following tests require more space nothing work
         if(param_address.find(*l) == param_address.end())
         {
            param_address[*l] = Rmem->get_memory_address();
            Rmem->reserve_space(reserve_space);
         }
      }
      else if(!is_memory)
            THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "not yet supported case");
      else
         base_type_size = static_cast<unsigned int>(splitted[0].size()/8);
      /// check for regularity
      bool all_equal = splitted.size()>1;
      for(unsigned int i = 1; i < splitted.size(); i++)
         if(splitted[i] != splitted[0])
            all_equal = false;
      if(all_equal)
      {
         PP(os, "for(__testbench_index0=0; __testbench_index0 < " + STR(splitted.size()) + "; ++__testbench_index0)\n{\n");
         byte_allocated += (base_type_size)*(splitted.size()-1);
      }
      for(unsigned int i = 0; i < splitted.size() && (!all_equal || i==0); i++)
      {
         THROW_ASSERT(splitted[i] != "", "Not well formed test vector: " + test_v);

         std::string initial_string = splitted[i];
         std::string binary_string;
         unsigned int size_of_data = static_cast<unsigned int>(initial_string.size());
         if (!is_memory)
         {
            std::string init_value_copy = initial_string;
            boost::replace_all(init_value_copy, "\\", "\\\\");
            PP(os, "fprintf(testvector_file, \"//memory initialization for variable " + param + " value: "  + init_value_copy + "\\n\");\n");
            tree_nodeRef pt_node = TreeM->get_tree_node_const(base_type);
            THROW_ASSERT(pt_node->get_kind() == pointer_type_K, "A pointer type is expected");
            unsigned int ptd_base_type = GET_INDEX_NODE(GetPointer<pointer_type>(pt_node)->ptd);
            size_of_data = tree_helper::size(TreeM, ptd_base_type);
            if(BH->is_a_struct(ptd_base_type) || BH->is_an_union(ptd_base_type))
            {
               std::vector<std::string> splitted_fields;
               boost::algorithm::split(splitted_fields, initial_string, boost::algorithm::is_any_of("|"));
               std::list<unsigned int> fields = tree_helper::get_fields_type(TreeM, ptd_base_type);
               size_t n_values = splitted_fields.size();
               unsigned int index=0;
               unsigned int field_size;
               for(std::list<unsigned int>::const_iterator it=fields.begin(); it != fields.end(); ++it, ++index)
               {
                  unsigned int field_type = *it;
                  field_size = tree_helper::size(TreeM, field_type);
                  if(index < n_values)
                     binary_string = convert_in_binary(BH, field_type, splitted_fields[index], field_size);
                  else
                     binary_string = convert_in_binary(BH, field_type, "0", field_size);
                  if(is_all_8zeros(binary_string))
                  {
                     PP(os, "for(__testbench_index=0; __testbench_index < " + STR(binary_string.size()/8) + "; ++__testbench_index)\n   fprintf(testvector_file, \"m00000000\\n\");\n");
                     byte_allocated += binary_string.size()/8;
                  }
                  else
                  {
                     for(unsigned int base_index = 0; base_index < field_size; base_index = base_index + 8)
                     {
                        PP(os, "fprintf(testvector_file, \"m" + binary_string.substr(field_size-8-base_index, 8) + "\\n\");\n");
                        ++byte_allocated;
                     }
                  }
               }
            }
            else
            {
               binary_string =  convert_in_binary(BH, ptd_base_type, initial_string, size_of_data);
               if(is_all_8zeros(binary_string))
               {
                  PP(os, "for(__testbench_index=0; __testbench_index < " + STR(binary_string.size()/8) + "; ++__testbench_index)\n   fprintf(testvector_file, \"m00000000\\n\");\n");
                  byte_allocated += binary_string.size()/8;
               }
               else
               {
                  for(unsigned int base_index = 0; base_index < size_of_data; base_index = base_index + 8)
                  {
                     PP(os, "fprintf(testvector_file, \"m" + binary_string.substr(size_of_data-8-base_index, 8) + "\\n\");\n");
                     ++byte_allocated;
                  }
               }
            }
         }
         else
         {
            if(is_all_8zeros(initial_string))
            {
               PP(os, "for(__testbench_index=0; __testbench_index < " + STR(initial_string.size()/8) + "; ++__testbench_index)\n   fprintf(testvector_file, \"m00000000\\n\");\n");
               byte_allocated += initial_string.size()/8;
            }
            else
            {
               for(unsigned int base_index = 0; base_index < size_of_data; base_index = base_index + 8)
               {
                  PP(os, "fprintf(testvector_file, \"m" + initial_string.substr(size_of_data-8-base_index, 8) + "\\n\");\n");
                  ++byte_allocated;
               }
            }
         }
      }
      if(all_equal)
         PP(os, "}\n");

      if(byte_allocated != actual_byte)
         THROW_ERROR("mismatch between allocated bytes and object size: All(" + STR(byte_allocated) +"), ObjSize(" + STR(actual_byte) + ") for " + STR(*l));
   }
   //std::cerr << "starting declaration of variables" << std::endl;
   PP(os, "//base address: " + STR(HLS_C->base_address) + "\n");
   PP(os, "//declaration of variables\n");
   for(std::list<unsigned int>::const_iterator p = parameters.begin(); p != parameters.end(); p++)
   {
      unsigned int type_id = BH->get_type(*p);
      std::string type = BH->print_type(type_id);
      std::string param = BH->PrintVariable(*p);
      if (BH->is_a_pointer(*p))
      {
         unsigned int base_type = tree_helper::get_type_index(TreeM, *p);
         tree_nodeRef pt_node = TreeM->get_tree_node_const(base_type);
         THROW_ASSERT(pt_node->get_kind() == pointer_type_K, "A pointer type is expected");
         base_type = GET_INDEX_NODE(GetPointer<pointer_type>(pt_node)->ptd);
         unsigned int base_type_size= tree_helper::size(TreeM, base_type)/8;

         std::vector<std::string> splitted;
         std::string test_v;
         if(test_vector.find(param) != test_vector.end())
            test_v = test_vector[param];
         else
            test_v ="0";
         boost::algorithm::split(splitted, test_v , boost::algorithm::is_any_of(","));

         PP(os, type + " " + param + "=malloc(" + boost::lexical_cast<std::string>(base_type_size*splitted.size()) + ");\n");

         for(unsigned int i = 0; i < splitted.size(); i++)
         {
            if(BH->is_a_struct(base_type) || BH->is_an_union(base_type))
            {
               std::vector<std::string> splitted_fields;
               std::string field_value = splitted[i];
               boost::algorithm::split(splitted_fields, field_value, boost::algorithm::is_any_of("|"));
               std::list<unsigned int> fields = tree_helper::get_fields_type(TreeM, base_type);
               size_t n_values = splitted_fields.size();
               unsigned int index=0;
               for(std::list<unsigned int>::const_iterator it=fields.begin(); it != fields.end(); ++it, ++index)
               {
                  if(index < n_values)
                     PP(os, param + "["+boost::lexical_cast<std::string>(i)+ "]." + BH->PrintVariable(tree_helper::get_field_idx(TreeM, base_type, index))+" = "+splitted_fields[index]+";\n");
                  else
                     PP(os, param + "["+boost::lexical_cast<std::string>(i)+ "]." + BH->PrintVariable(tree_helper::get_field_idx(TreeM, base_type, index))+" = 0;\n");
               }
            }
            else
               PP(os, param + "["+boost::lexical_cast<std::string>(i)+ "] = " + splitted[i] + ";\n");
         }
         std::string memory_addr;
         THROW_ASSERT(param_address.find(*p) != param_address.end(), "parameter does not have an address");
         memory_addr = boost::lexical_cast<std::string>(param_address[*p]);
         PP(os, "fprintf(testvector_file, \"//parameter: " + param + " value: "  + memory_addr + "\\n\");\n");
         PP(os, "fprintf(testvector_file, \"" + convert_in_binary(BH, 0, memory_addr, 32) + "\\n\");\n");
      }
      else
      {
         PP(os, type + " " + param + " = " + test_vector[param] + ";\n");
         PP(os, "fprintf(testvector_file, \"//parameter: " + param + " value: "  + test_vector[param] + "\\n\");\n");
         PP(os, "fprintf(testvector_file, \"" + convert_in_binary(BH, type_id, test_vector[param], tree_helper::size(TreeM, type_id)) + "\\n\");\n");
      }
   }

   PP(os, "//function call\n");
   if (BH->has_return())
   {
      unsigned int var = BH->get_function_return(BH->get_function_index());
      std::string ret_type = BH->print_type(var);
      PP(os, ret_type + " " + std::string(RETURN_PORT_NAME) + " = ");
   }
   std::string function_name = BH->get_function_name();
   if(function_name == "main")
      function_name = "_" + function_name;
   PP(os, function_name + "(");
   bool is_first = true;
   for(std::list<unsigned int>::const_iterator p = parameters.begin(); p != parameters.end(); p++)
   {
      if (!is_first) PP(os, ", ");
      else is_first = false;
      std::string param = BH->PrintVariable(*p);
      PP(os, param);
   }
   PP(os, ");\n");
   for(std::list<unsigned int>::const_iterator p = parameters.begin(); p != parameters.end(); p++)
   {
      std::string param = BH->PrintVariable(*p);
      if (BH->is_a_pointer(*p))
      {
         std::vector<std::string> splitted;
         std::string test_v;
         if(test_vector.find(param) != test_vector.end())
            test_v = test_vector[param];
         else
            test_v ="0";
         boost::algorithm::split(splitted, test_v , boost::algorithm::is_any_of(","));
         unsigned int base_type = tree_helper::get_type_index(TreeM, *p);
         tree_nodeRef pt_node = TreeM->get_tree_node_const(base_type);
         THROW_ASSERT(pt_node->get_kind() == pointer_type_K, "A pointer type is expected");
         base_type = GET_INDEX_NODE(GetPointer<pointer_type>(pt_node)->ptd);
         unsigned int base_type_bitsize= tree_helper::size(TreeM, base_type);
         for(unsigned int i = 0; i < splitted.size(); i++)
         {
            PP(os, "fprintf(testvector_file, \"//expected value for output " + param + "[" + boost::lexical_cast<std::string>(i) + "]: %" + (BH->is_real(base_type) ? std::string("g") : std::string("d")) + "\\n\", " + param + "[" + boost::lexical_cast<std::string>(i) + "]);\n");
            PP(os, "fprintf(testvector_file, \"o\");\n");
            if(BH->is_real(base_type))
            {
#if HAVE_FLOPOCO
               PP(os, "_Ptd2Bin_(testvector_file, (unsigned char*)&" + param + "[" + boost::lexical_cast<std::string>(i) + "], "+ boost::lexical_cast<std::string>(base_type_bitsize) + ");\n");
#else
               THROW_ERROR("Real constants not yet supported");
#endif
            }
            else if(BH->is_a_struct(base_type) || BH->is_an_union(base_type))
            {
               PP(os, "_Ptd2Bin_(testvector_file, (unsigned char*)&" + param + "[" + boost::lexical_cast<std::string>(i) + "], "+ boost::lexical_cast<std::string>(base_type_bitsize) + ");\n");
            }
            else
               PP(os, "_Dec2Bin_(testvector_file, " + param + "[" + boost::lexical_cast<std::string>(i) + "], "+ boost::lexical_cast<std::string>(base_type_bitsize) + ");\n");
            PP(os, "fprintf(testvector_file, \"\\n\");\n");
         }
         PP(os, "fprintf(testvector_file, \"e\\n\");\n");
      }
   }
   if (BH->has_return())
   {
      unsigned int var = BH->get_function_return(BH->get_function_index());
      PP(os, "fprintf(testvector_file, \"//expected value for return value\\n\");\n");
      PP(os, "fprintf(testvector_file, \"o\");\n");
      unsigned int base_type = BH->get_type(var);
      unsigned int base_type_bitsize= tree_helper::size(TreeM, base_type);
      if(BH->is_real(base_type))
      {
#if HAVE_FLOPOCO
         PP(os, "_Ptd2Bin_(testvector_file, (unsigned char*)&" + std::string(RETURN_PORT_NAME) + ", "+ boost::lexical_cast<std::string>(base_type_bitsize) + ");\n");
#else
         THROW_ERROR("Real constants not yet supported");
#endif
      }
      else if(BH->is_a_struct(base_type) || BH->is_an_union(base_type))
      {
         PP(os, "_Ptd2Bin_(testvector_file, (unsigned char*)&" + std::string(RETURN_PORT_NAME) + ", "+ boost::lexical_cast<std::string>(base_type_bitsize) + ");\n");
      }
      else
         PP(os, "_Dec2Bin_(testvector_file, " + std::string(RETURN_PORT_NAME) + "," + boost::lexical_cast<std::string>(base_type_bitsize) + ");\n");
      PP(os, "fprintf(testvector_file, \"\\n\");\n");
      PP(os, "fprintf(testvector_file, \"e\\n\");\n");
   }
   PP(os, "__standard_exit=1;\n");
   PP(os, "exit(0);\n");
   PP(os, "}\n");

   // Creating output file
   std::string fileName = output_directory + "values.c";
   std::ofstream fileOut(fileName.c_str(), std::ios::out);
   fileOut << os.str() << std::endl;
   fileOut.close();

   return fileName;
}

void HLStestbench::generate_input_file(const std::string& fileName)
{
   std::tr1::unordered_set<std::string> file_sources;
   file_sources.insert(fileName);
   const GccWrapperConstRef gcc_wrapper(new GccWrapper(Param, CT_I386_GCC, OS_O0));
   gcc_wrapper->CreateExecutable(file_sources, output_directory + "test", std::string(" -D") + SIMULATION_DEFINE);

   ///executing the test to generate inputs and exected outputs values
   const std::string command = output_directory + "test";
   int ret = PandaSystem(Param, command);
   if(IsError(ret))
   {
      THROW_ERROR("Error in generating the tests");
   }

}

void HLStestbench::generate_testbench(const application_managerRef AppM, const technology_managerRef&, const HLS_constraintsRef& HLS_C, const memoryRef& mem)
{
   PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "  . Generation of the testbenches");

   std::vector<std::map<std::string, std::string> > test_vectors;

   std::string tb_input_xml = Param->getOption<std::string>("tb_input_xml");
   unsigned int vectors = parse_test_vectors(AppM, tb_input_xml, test_vectors);
   if (vectors == 0)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "  No input vectors to be tested");
      return;
   }
   else
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "  Number of input vectors: " + boost::lexical_cast<std::string>(vectors));
   }

   std::string input_file = output_directory + "values.txt";
   if (boost::filesystem::exists(input_file)) boost::filesystem::remove_all(input_file);

   for (unsigned int l = 0; l < test_vectors.size(); l++)
   {
      ///creating C file for generating the values
      std::string fileName = creating_c_testbench(AppM, HLS_C, test_vectors[l], mem, input_file);

      ///executing the test file and getting the testbench values
      generate_input_file(fileName);
   }

   Param->setOption("tb_inputs", input_file);
}

std::string HLStestbench::convert_in_binary(const behavioral_helperRef& BH, unsigned int base_type, const std::string &C_value, unsigned int precision)
{
   std::string trimmed_value;
   if(base_type && BH->is_real(base_type))
#if HAVE_FLOPOCO
   {
      unsigned int wF, wE;
      char *endptr;
      flopoco_wrapper::DECODE_BITS(precision, wF, wE);
      mpfr_t mpx;
      mpfr_init2 (mpx, static_cast<mpfr_prec_t>(wF+1));
      mpfr_strtofr (mpx, C_value.c_str(), &endptr, 10, GMP_RNDN);
      std::string flopoco_string = flopoco::fp2bin(mpx, static_cast<int>(wE), static_cast<int>(wF));
      trimmed_value = flopoco_string.substr(2);
   }
#else
   THROW_ERROR("Real constants not yet supported");
#endif
   else
   {
      long long int ll_value;
      if(C_value[0] == '\'')
      {
         trimmed_value = C_value.substr(1);
         trimmed_value = trimmed_value.substr(0, trimmed_value.find('\''));
         //std::cout << "-" << trimmed_value << "-" << std::endl;
         if(trimmed_value[0] == '\\')
            ll_value = boost::lexical_cast<long long int>(trimmed_value.substr(1));
         else
            ll_value = boost::lexical_cast<char>(trimmed_value);
      }
      else
         ll_value = boost::lexical_cast<long long int>(C_value);
      unsigned long long int ull_value = static_cast<unsigned long long int>(ll_value);
      trimmed_value = "";
      for(unsigned int ind = 0; ind < precision; ind++)
         trimmed_value = trimmed_value + (((1LLU << (precision-ind-1)) & ull_value) ? '1' : '0');
   }
   return trimmed_value;
}

std::string HLStestbench::create_testbench(const std::string& filename_bench, const structural_objectRef &cir, const double target_period, bool xilinx_isim, const tree_managerConstRef TreeM) const
{
   if (!Param->getOption<bool>(OPT_generate_testbench) or !Param->isOption("tb_inputs"))
      return "";

   language_writerRef writer = language_writer::create_writer(language_writer::VERILOG, debug_level);
   std::string simulation_values_path = Param->getOption<std::string>("tb_inputs");
   bool generate_vcd_output = Param->isOption(OPT_generate_vcd) && Param->getOption<bool>(OPT_generate_vcd);

   std::string file_name = output_directory + filename_bench + writer->get_extension();
   std::ofstream hdl_file(file_name.c_str());

   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  writing testbench");
   writer->write_comment(hdl_file, std::string("File automatically generated by: ") + PACKAGE_NAME + " framework version=" + PACKAGE_VERSION + "\n");
   writer->write_comment(hdl_file, std::string("Send any bug to: ") + PACKAGE_BUGREPORT + "\n");
   for(unsigned int row=0; row < TESTBENCH_COPYING3_SHORT_NROW; ++row)
   {
      writer->write_comment(hdl_file, std::string(TESTBENCH_COPYING3_SHORT[row]));
   }
   //write testbech for simulation
   write_testbench(hdl_file, writer, cir, target_period, simulation_values_path, generate_vcd_output, xilinx_isim, TreeM);

   hdl_file.close();

   return file_name;
}

void HLStestbench::write_testbench(std::ostream& hdl_file, language_writerRef writer, const structural_objectRef &cir, double target_period, std::string simulation_values_path, bool generate_vcd_output, bool xilinx_isim, const tree_managerConstRef TreeM) const
{
   std::string input_values_filename = simulation_values_path;

   /// generation of VCD output
   std::string vcd_output_filename = output_directory + "test.vcd";

   module * mod = GetPointer<module>(cir);

   simple_indent PP(STD_OPENING_CHAR, STD_CLOSING_CHAR, 2);
   PP(hdl_file, "`timescale 1ns / 1ps\n");
   writer->write_comment(hdl_file, "CONSTANTS DECLARATION\n");
   PP(hdl_file, "`define EOF 32'hFFFF_FFFF\n`define NULL 0\n`define MAX_COMMENT_LENGTH 1000\n`define SIMULATION_LENGTH 20000000\n\n");
   PP(hdl_file, "`define HALF_CLOCK_PERIOD "+boost::lexical_cast<std::string>(target_period/2)+"\n\n");
   PP(hdl_file, "`define CLOCK_PERIOD (2*`HALF_CLOCK_PERIOD)\n\n");
   PP(hdl_file, "`define MEM_DELAY_READ 2\n\n");
   PP(hdl_file, "`define MEM_DELAY_WRITE 1\n\n");

   writer->write_comment(hdl_file, "MODULE DECLARATION\n");
   PP(hdl_file, "module " + cir->get_id() + "_tb;\n");
   PP.indent();

   PP(hdl_file, "parameter MEMSIZE = 524288;\n");

   writer->write_comment(hdl_file, "AUXILIARY VARIABLES DECLARATION\n");
   PP(hdl_file, "integer res_file, file, _ch_, _r_, _n_, _i_, startTime, endTime, sim_time; // Temporary variables\n");
   PP(hdl_file, "reg compare_outputs, success; // Flag: True if input vector specifies expected outputs\n");
   PP(hdl_file, "reg [8*`MAX_COMMENT_LENGTH:0] line; // Comment line read from file\n\n");

   PP(hdl_file, "reg [31:0] addr, base_addr;\n");
   PP(hdl_file, "reg signed [7:0] mem [0:MEMSIZE-1];\n\n");

   bool with_memory = false;

   /// write input signals declaration
   if (mod->get_in_port_size())
   {
      writer->write_comment(hdl_file, "INPUT SIGNALS\n");
      for (unsigned int i = 0; i < mod->get_in_port_size(); i++)
      {
         const structural_objectRef& port_obj = mod->get_in_port(i);
         if(GetPointer<port_o>(port_obj)->get_is_memory())
         {
            if (GetPointer<port_o>(port_obj)->get_id().find("M") == 0)
               with_memory=true;
            PP(hdl_file, "wire ");
         }
         else
            PP(hdl_file, "reg ");
         PP(hdl_file, writer->type_converter(port_obj->get_typeRef()) + writer->type_converter_size(port_obj));
         if (port_obj->get_kind() != port_o_K && port_obj->get_typeRef()->type != structural_type_descriptor::VECTOR_BOOL)
         {
            unsigned int lsb = GetPointer<port_vector_o>(mod->get_in_port(i))->get_lsb();
            PP(hdl_file, "[" + boost::lexical_cast<std::string>(GetPointer<port_vector_o>(mod->get_in_port(i))->get_ports_size() - 1 + lsb) + ":" + boost::lexical_cast<std::string>(lsb) + "] ");
         }
         PP(hdl_file, HDL_manager::convert_to_identifier(writer.get(), mod->get_in_port(i)->get_id()) + ";\n");
         if (port_obj->get_typeRef()->treenode > 0 && tree_helper::is_a_pointer(TreeM, port_obj->get_typeRef()->treenode))
         {
            PP(hdl_file, "reg ["+boost::lexical_cast<std::string>(tree_helper::size(TreeM, tree_helper::get_pointed_type(TreeM, tree_helper::get_type_index(TreeM,port_obj->get_typeRef()->treenode)))-1)+":0] ex_" + port_obj->get_id() + ";\n");
         }
      }
      PP(hdl_file, "\n");
   }

   /// write output signals declaration
   if (mod->get_out_port_size())
   {
      writer->write_comment(hdl_file, "OUTPUT SIGNALS\n");
      for (unsigned int i = 0; i < mod->get_out_port_size(); i++)
      {
         PP(hdl_file, "wire " + writer->type_converter(mod->get_out_port(i)->get_typeRef()) + writer->type_converter_size(mod->get_out_port(i)));
         if (mod->get_out_port(i)->get_kind() != port_o_K && mod->get_out_port(i)->get_typeRef()->type != structural_type_descriptor::VECTOR_BOOL)
         {
            unsigned int lsb = GetPointer<port_vector_o>(mod->get_out_port(i))->get_lsb();
            PP(hdl_file, "[" + boost::lexical_cast<std::string>(GetPointer<port_vector_o>(mod->get_out_port(i))->get_ports_size() - 1 + lsb) + ":" + boost::lexical_cast<std::string>(lsb) + "] ");
         }
         PP(hdl_file, HDL_manager::convert_to_identifier(writer.get(), mod->get_out_port(i)->get_id()) + ";\n");
         if (mod->get_out_port(i)->get_id() == RETURN_PORT_NAME)
         {
            PP(hdl_file, "reg " + writer->type_converter(mod->get_out_port(i)->get_typeRef()) + writer->type_converter_size(mod->get_out_port(i)));
            PP(hdl_file, "ex_" + HDL_manager::convert_to_identifier(writer.get(), mod->get_out_port(i)->get_id()) + ";\n");
            PP(hdl_file, "reg " + writer->type_converter(mod->get_out_port(i)->get_typeRef()) + writer->type_converter_size(mod->get_out_port(i)));
            PP(hdl_file, "registered_" + HDL_manager::convert_to_identifier(writer.get(), mod->get_out_port(i)->get_id()) + ";\n");
         }
      }
      PP(hdl_file, "\n");
   }

   /// write inout signals declaration
   if (mod->get_in_out_port_size())
   {
      THROW_ERROR("Module with IO ports cannot synthesized");
   }
   if(with_memory)
   {
      PP(hdl_file, "integer reg_DataReady;\n");
      PP(hdl_file, "wire [63:0] mask;\n\n");
   }
   /// write module instantiation and ports binding
   writer->write_comment(hdl_file, "MODULE INSTANTIATION AND PORTS BINDING\n");
   writer->write_module_instance_begin(hdl_file, cir, !xilinx_isim);
   std::string prefix = "";
   if (mod->get_in_port_size())
   {
      for (unsigned int i = 0; i < mod->get_in_port_size(); i++)
      {
         if(i == 0)
            prefix = ".";
         else
            prefix = ", .";

         PP(hdl_file, prefix + HDL_manager::convert_to_identifier(writer.get(), mod->get_in_port(i)->get_id()) + "(" + HDL_manager::convert_to_identifier(writer.get(), mod->get_in_port(i)->get_id()) + ")");
      }
   }
   if (mod->get_out_port_size())
   {
      for (unsigned int i = 0; i < mod->get_out_port_size(); i++)
      {
         PP(hdl_file, prefix + HDL_manager::convert_to_identifier(writer.get(), mod->get_out_port(i)->get_id()) + "(" + HDL_manager::convert_to_identifier(writer.get(), mod->get_out_port(i)->get_id()) + ")");
      }
   }

   writer->write_module_instance_end(hdl_file, cir);

   if(xilinx_isim)
       PP(hdl_file, "glbl #("+ boost::lexical_cast<std::string>(target_period/2*1000) + ", 0) glbl();");
   /// initial operation
   PP(hdl_file, "\n");
   writer->write_comment(hdl_file, "Operation to be executed just one time\n");
   PP(hdl_file, "initial\nbegin\n");
   PP.indent();

   /// VCD output generation (optional)
   if(generate_vcd_output)
   {
      writer->write_comment(hdl_file, "VCD file generation\n");
      PP(hdl_file, "$dumpfile(\"" + vcd_output_filename + "\");\n");
      PP(hdl_file, "$dumpvars;\n");
   }

   /// open file with values
   writer->write_comment(hdl_file, "OPEN FILE WITH VALUES FOR SIMULATION\n");
   PP(hdl_file, "file = $fopen(\"" + input_values_filename + "\",\"r\");\n");
   PP(hdl_file, "if (file == `NULL) ");
   writer->write_comment(hdl_file, "Error in file open\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "$display(\"ERROR - Error opening the file\");\n");
   PP(hdl_file, "$finish; ");
   writer->write_comment(hdl_file, "Terminate\n");
   PP.deindent();
   PP(hdl_file, "end\n");

   /// open file with results
   writer->write_comment(hdl_file, "OPEN FILE WHERE results will be written\n");
   std::string result_file = Param->getOption<std::string>(OPT_simulation_output);
   PP(hdl_file, "res_file = $fopen(\"" + result_file + "\",\"w\");\n\n");
   PP(hdl_file, "if (res_file == `NULL) ");
   writer->write_comment(hdl_file, "Error in file open\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "$display(\"ERROR - Error opening the res_file\");\n");
   PP(hdl_file, "$fclose(file);\n");
   PP(hdl_file, "$finish; ");
   writer->write_comment(hdl_file, "Terminate\n");
   PP.deindent();
   PP(hdl_file, "end\n");

   /// auxiliary variables initialization
   writer->write_comment(hdl_file, "Variables initialization\n");
   PP(hdl_file, "sim_time = 0;\n");
   PP(hdl_file, "startTime = 0;\n");
   PP(hdl_file, "endTime = 0;\n");
   PP(hdl_file, "_ch_ = 0;\n");
   PP(hdl_file, "_n_ = 0;\n");
   PP(hdl_file, "_r_ = 0;\n");
   PP(hdl_file, "line = 0;\n");
   PP(hdl_file, "_i_ = 0;\n");
   PP(hdl_file, "compare_outputs = 0;\n");
   PP(hdl_file, "success = 1;\n");
   /// write input signals initialization
   if (mod->get_in_port_size())
   {
      for (unsigned int i = 0; i < mod->get_in_port_size(); i++)
      {
         const structural_objectRef& port_obj = mod->get_in_port(i);
         if(GetPointer<port_o>(port_obj)->get_is_memory()) continue;
         PP(hdl_file, HDL_manager::convert_to_identifier(writer.get(), port_obj->get_id()) + " = 0;\n");
         if (port_obj->get_typeRef()->treenode > 0 && tree_helper::is_a_pointer(TreeM, port_obj->get_typeRef()->treenode))
         {
            PP(hdl_file, "ex_" + port_obj->get_id() + " = 0;\n");
         }
      }
      PP(hdl_file, "\n");
   }
   if (mod->find_member(RETURN_PORT_NAME, port_o_K, cir))
   {
      PP(hdl_file, "ex_" + std::string(RETURN_PORT_NAME) + " = 0;\n");
      PP(hdl_file, "registered_" + std::string(RETURN_PORT_NAME) + " = 0;\n");
      PP(hdl_file, "\n");
   }
   if(with_memory)
   {
      PP(hdl_file, "reg_DataReady <= 0;\n\n");

   }

   PP(hdl_file, "for (addr = 0; addr < MEMSIZE; addr = addr + 1)\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "mem[addr] = 8'b0;\n");
   PP.deindent();
   PP(hdl_file, "end\n");

   PP(hdl_file, "#0  reset <= 1'b1; " + std::string(START_PORT_NAME) + " <= 1'b0;\n");
   PP(hdl_file, "#`HALF_CLOCK_PERIOD reset <= 1'b0;\n");
   PP(hdl_file, "#`CLOCK_PERIOD " + std::string(START_PORT_NAME) + " <= 1'b1;\n");

   PP.deindent();
   PP(hdl_file, "end\n"); /// end of initial block

   /// file reading operations
   PP(hdl_file, "\n");

   writer->write_comment(hdl_file, "Assigning values\n");
   PP(hdl_file, "always @ (posedge " + std::string(START_PORT_NAME) + ")\nbegin\n");
   PP.indent();
   PP(hdl_file, "if (" + std::string(START_PORT_NAME) + " == 1'b1)\n");
   PP(hdl_file, "begin\n");
   PP.indent();

   writer->write_comment(hdl_file, "reading base address memory --------------------------------------------------------------\n");
   PP(hdl_file, "_ch_ = $fgetc(file);\n");
   PP(hdl_file, "_i_ = 0;\n");
   PP(hdl_file, "while (_ch_ == \"/\" || _ch_ == \"\\n\" || _ch_ == \"b\") // skip comments and empty lines\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "if (_ch_ == \"b\")\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "_r_ = $fscanf(file,\"%b\\n\", base_addr); ");
   PP(hdl_file, "_i_ = _i_ + 1;\n");
   PP.deindent();
   PP(hdl_file, "end\n");
   PP(hdl_file, "else\n");
   PP.indent();
   PP(hdl_file, "_r_ = $fgets(line, file);\n");
   PP.deindent();
   PP(hdl_file, "_ch_ = $fgetc(file);\n");
   PP.deindent();
   PP(hdl_file, "end\n");
   PP(hdl_file, "_r_ = $ungetc(_ch_, file); // Undo the first $fgetc\n");

   writer->write_comment(hdl_file, "initializing memory --------------------------------------------------------------\n");
   PP(hdl_file, "_ch_ = $fgetc(file);\n");
   PP(hdl_file, "_i_ = 0;\n");
   PP(hdl_file, "while (_ch_ == \"/\" || _ch_ == \"\\n\" || _ch_ == \"m\") // skip comments and empty lines\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "if (_ch_ == \"m\")\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "_r_ = $fscanf(file,\"%b\\n\", mem[_i_]); ");
   PP(hdl_file, "_i_ = _i_ + 1;\n");
   PP.deindent();
   PP(hdl_file, "end\n");
   PP(hdl_file, "else\n");
   PP.indent();
   PP(hdl_file, "_r_ = $fgets(line, file);\n");
   PP.deindent();
   PP(hdl_file, "_ch_ = $fgetc(file);\n");
   PP.deindent();
   PP(hdl_file, "end\n");
   PP(hdl_file, "_r_ = $ungetc(_ch_, file); // Undo the first $fgetc\n");
   bool first_valid_input = true;

   /// iterate over all input ports
   if (mod->get_in_port_size())
   {
      std::string input_name = "";

      /// write code for input values extraction
      for (unsigned int i = 0; i < mod->get_in_port_size(); i++)
      {
         if(GetPointer<port_o>(mod->get_in_port(i))->get_is_memory()) continue;
         input_name = HDL_manager::convert_to_identifier(writer.get(), mod->get_in_port(i)->get_id());

         if(input_name != CLOCK_PORT_NAME && input_name != RESET_PORT_NAME && input_name != START_PORT_NAME)
         {
            PP(hdl_file, "\n");
            writer->write_comment(hdl_file, "Read a value for " + input_name + " --------------------------------------------------------------\n");
            PP(hdl_file, "_ch_ = $fgetc(file);\n");
            PP(hdl_file, "while (_ch_ == \"/\" || _ch_ == \"\\n\") ");
            writer->write_comment(hdl_file, "skip comments and empty lines\n");
            PP(hdl_file, "begin\n");
            PP.indent();
            PP(hdl_file, "_r_ = $fgets(line, file);\n");
            PP(hdl_file, "_ch_ = $fgetc(file);\n");
            PP.deindent();
            PP(hdl_file, "end\n");

            /// if it's the first input value
            if(first_valid_input)
            {
               /// write statement for new vectors' check
               writer->write_comment(hdl_file, "If no character found\n");
               PP(hdl_file, "if (_ch_ == -1)\n");
               PP(hdl_file, "begin\n");
               PP.indent();
               PP(hdl_file, "$display(\"No more values found. Simulation(s) executed: %d.\\n\", _n_);\n");
               PP(hdl_file, "$fclose(res_file);\n");
               PP(hdl_file, "$fclose(file);\n");
               PP(hdl_file, "$finish;\n");
               PP.deindent();
               PP(hdl_file, "end\n");
               PP(hdl_file, "else\n");
               PP(hdl_file, "begin\n");
               PP.indent();
               writer->write_comment(hdl_file, "Vectors count\n");
               PP(hdl_file, "_n_ = _n_ + 1;\n");
               PP(hdl_file, "$display(\"Start reading vector %d's values from input file.\\n\", _n_);\n");
               PP.deindent();
               PP(hdl_file, "end\n");
               first_valid_input = false;
            }

            PP(hdl_file, "_r_ = $ungetc(_ch_, file); ");
            writer->write_comment(hdl_file, "Undo the first $fgetc\n");
            PP(hdl_file, "_r_ = $fscanf(file,\"%b\\n\", " + input_name + "); ");
            writer->write_comment(hdl_file, "expected format: bbb...b (example: 00101110)\n");

            PP(hdl_file, "if (_r_ != 1) ");
            writer->write_comment(hdl_file, "error\n");
            PP(hdl_file, "begin\n");
            PP.indent();
            PP(hdl_file, "_ch_ = $fgetc(file);\n");
            PP(hdl_file, "if (_ch_ == `EOF) ");
            writer->write_comment(hdl_file, "end-of-file reached\n");
            PP(hdl_file, "begin\n");
            PP.indent();
            PP(hdl_file, "$display(\"ERROR - End of file reached before getting all the values for the parameters\");\n");
            PP(hdl_file, "$fclose(res_file);\n");
            PP(hdl_file, "$fclose(file);\n");
            PP(hdl_file, "$finish;\n");
            PP.deindent();
            PP(hdl_file, "end\n");
            PP(hdl_file, "else ");
            writer->write_comment(hdl_file, "generic error\n");
            PP(hdl_file, "begin\n");
            PP.indent();
            PP(hdl_file, "$display(\"ERROR - Unknown error while reading the file. Character found: %c\", _ch_);\n");
            PP(hdl_file, "$fclose(res_file);\n");
            PP(hdl_file, "$fclose(file);\n");
            PP(hdl_file, "$finish;\n");
            PP.deindent();
            PP(hdl_file, "end\n");
            PP.deindent();
            PP(hdl_file, "end\n");
            PP(hdl_file, "else\n");
            PP.indent();
            PP(hdl_file, "$display(\"Value found for input " + input_name + ": %b\", " + input_name + ");");
            PP.deindent();
            writer->write_comment(hdl_file, "Value for " + input_name + " found ---------------------------------------------------------------\n");
         }
      }
   }

   PP(hdl_file, "\n#`CLOCK_PERIOD; ");
   writer->write_comment(hdl_file, "Wait some cycles (needed for a correct simulation)\n\n");

   writer->write_comment(hdl_file, "Simulation start\n");
   PP(hdl_file, std::string(START_PORT_NAME) + " <= 1'b0;\nstartTime = $time;\n$display(\"Reading of vector values from input file completed. Simulation started.\");\n");
   PP.deindent();
   PP(hdl_file, "end\n");
   PP.deindent();
   PP(hdl_file, "end\n\n");

   /// write clock switching operation
   writer->write_comment(hdl_file, "Clock switching: 1 cycle every CLOCK_PERIOD seconds\n");
   PP(hdl_file, "always # `HALF_CLOCK_PERIOD clock = !clock;\n\n");

   PP(hdl_file, "always @(posedge clock)\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "if (($time - startTime)/`CLOCK_PERIOD > `SIMULATION_LENGTH)\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "$display(\"Simulation not completed into %d cycles\", `SIMULATION_LENGTH);\n");
   PP(hdl_file, "$fwrite(res_file, \"X\\t\");\n");
   PP(hdl_file, "$fwrite(res_file, \"%d\\n\", `SIMULATION_LENGTH);\n");
   PP(hdl_file, "$fclose(res_file);\n");
   PP(hdl_file, "$fclose(file);\n");
   PP(hdl_file, "$finish;\n");
   PP.deindent();
   PP(hdl_file, "end\n");
   PP.deindent();
   PP(hdl_file, "end\n\n");
   if(with_memory)
   {
      PP(hdl_file, "assign mask = (1 << Mout_data_ram_size) -1;\n");
      std::string mem_aggregate = "{";
      structural_objectRef Mout_Wdata_ram_port = mod->find_member("Mout_Wdata_ram", port_o_K, cir);
      long long int bitsize = Mout_Wdata_ram_port->get_typeRef()->size * Mout_Wdata_ram_port->get_typeRef()->vector_size;
      THROW_ASSERT(Mout_Wdata_ram_port, "Mout_Wdata_ram port is missing");
      for(unsigned int bitsize_index=0; bitsize_index < bitsize; bitsize_index = bitsize_index + 8)
      {
         if(bitsize_index) mem_aggregate += ", ";
         mem_aggregate += "mem[Mout_addr_ram + " + boost::lexical_cast<std::string>((bitsize - bitsize_index)/8-1) + " - base_addr]";
      }
      mem_aggregate += "}";
      writer->write_comment(hdl_file, "OffChip Memory write\n");
      PP(hdl_file, "always @(posedge clock)\n");
      PP(hdl_file, "begin\n");
      PP.indent();
      PP(hdl_file, "if (Mout_we_ram === 1'b1 && base_addr <= Mout_addr_ram && Mout_addr_ram < (base_addr + MEMSIZE))\n");
      PP(hdl_file, "begin\n");
      PP.indent();
      PP(hdl_file, mem_aggregate + " <= (Mout_Wdata_ram & mask) | (" + mem_aggregate + " & ~mask);\n");
      PP.deindent();
      PP(hdl_file, "end\n");
      PP.deindent();
      PP(hdl_file, "end\n\n");

      structural_objectRef Min_Wdata_ram_port = mod->find_member("Min_Wdata_ram", port_o_K, cir);
      THROW_ASSERT(Min_Wdata_ram_port, "Min_Wdata_ram port is missing");
      PP(hdl_file, "assign Min_Wdata_ram = 0;\n");

      structural_objectRef M_Rdata_ram_port = mod->find_member("M_Rdata_ram", port_o_K, cir);
      THROW_ASSERT(M_Rdata_ram_port, "M_Rdata_ram port is missing");
      bitsize = M_Rdata_ram_port->get_typeRef()->size * M_Rdata_ram_port->get_typeRef()->vector_size;

      structural_objectRef Sout_Rdata_ram_port = mod->find_member("Sout_Rdata_ram", port_o_K, cir);
      if(Sout_Rdata_ram_port)
      {
         PP(hdl_file, "assign M_Rdata_ram = Sout_Rdata_ram;\n");
         PP(hdl_file, "assign Sin_Rdata_ram = (Mout_oe_ram===1'b1 && base_addr <= Mout_addr_ram && Mout_addr_ram < (base_addr + MEMSIZE)) ? " + mem_aggregate + " : " + boost::lexical_cast<std::string>(bitsize) + "'b0;\n");
      }
      else
      {
         PP(hdl_file, "assign M_Rdata_ram = (Mout_oe_ram===1'b1 && base_addr <= Mout_addr_ram && Mout_addr_ram < (base_addr + MEMSIZE)) ? " + mem_aggregate + " : " + boost::lexical_cast<std::string>(bitsize) + "'b0;\n");
      }
      structural_objectRef Sin_Wdata_ram_port = mod->find_member("S_Wdata_ram", port_o_K, cir);
      if(Sin_Wdata_ram_port)
         PP(hdl_file, "assign S_Wdata_ram = Mout_Wdata_ram;\n");

      structural_objectRef Sout_DataRdy_port = mod->find_member("Sout_DataRdy", port_o_K, cir);
      if(Sout_DataRdy_port)
         PP(hdl_file,"assign M_DataRdy = Sout_DataRdy;\n");
      structural_objectRef Sin_DataRdy_port = mod->find_member("Sin_DataRdy", port_o_K, cir);

      PP(hdl_file, "always @(posedge clock)\n");
      PP(hdl_file, "begin\n");

      PP.indent();
      PP(hdl_file, "if ((Mout_oe_ram===1'b1 && base_addr <= Mout_addr_ram && Mout_addr_ram < (base_addr + MEMSIZE)))\n");
      PP(hdl_file, "begin\n");
      PP.indent();
      PP(hdl_file, "if (reg_DataReady >= 0 && reg_DataReady < `MEM_DELAY_READ-1)\n");
      PP.indent();
      PP(hdl_file, "reg_DataReady <= 1 + reg_DataReady;\n");
      PP.deindent();
      PP(hdl_file, "else\n");
      PP.indent();
      PP(hdl_file, "reg_DataReady <= 0;\n");
      PP.deindent();
      PP.deindent();
      PP(hdl_file, "end\n\n");

      PP(hdl_file, "else if ((Mout_we_ram===1'b1 && base_addr <= Mout_addr_ram && Mout_addr_ram < (base_addr + MEMSIZE)))\n");
      PP(hdl_file, "begin\n");
      PP.indent();
      PP(hdl_file, "if (reg_DataReady >= 0 && reg_DataReady < `MEM_DELAY_WRITE-1)\n");
      PP.indent();
      PP(hdl_file, "reg_DataReady <= 1 + reg_DataReady;\n");
      PP.deindent();
      PP(hdl_file, "else\n");
      PP.indent();
      PP(hdl_file, "reg_DataReady <= 0;\n");
      PP.deindent();
      PP.deindent();
      PP(hdl_file, "end\n\n");

      PP(hdl_file, "else\n");
      PP.indent();
      PP(hdl_file, "reg_DataReady <= 0;\n");
      PP.deindent();
      PP.deindent();
      PP(hdl_file, "end\n\n");

      if(Sin_DataRdy_port)
         PP(hdl_file,"assign Sin_DataRdy = (Mout_oe_ram===1'b1 && (reg_DataReady == `MEM_DELAY_READ-1)) || (Mout_we_ram===1'b1 && (reg_DataReady == `MEM_DELAY_WRITE-1));\n");
      else
         PP(hdl_file,"assign M_DataRdy = (Mout_oe_ram===1'b1 && (reg_DataReady == `MEM_DELAY_READ-1)) || (Mout_we_ram===1'b1 && (reg_DataReady == `MEM_DELAY_WRITE-1));\n");

      /// Master inputs set up
      structural_objectRef Min_oe_ram_port = mod->find_member("Min_oe_ram", port_o_K, cir);
      THROW_ASSERT(Min_oe_ram_port, "Min_oe_ram port is missing");
      PP(hdl_file, "assign Min_oe_ram = 1'b0;\n");
      structural_objectRef Min_we_ram_port = mod->find_member("Min_we_ram", port_o_K, cir);
      THROW_ASSERT(Min_we_ram_port, "Min_we_ram port is missing");
      PP(hdl_file, "assign Min_we_ram = 1'b0;\n");
      structural_objectRef Min_addr_ram_port = mod->find_member("Min_addr_ram", port_o_K, cir);
      THROW_ASSERT(Min_addr_ram_port, "Min_addr_ram port is missing");
      PP(hdl_file, "assign Min_addr_ram = 0;\n");
      structural_objectRef Min_data_ram_size_port = mod->find_member("Min_data_ram_size", port_o_K, cir);
      THROW_ASSERT(Min_data_ram_size_port, "Min_data_ram_size port is missing");
      PP(hdl_file, "assign Min_data_ram_size = 0;\n");
      /// Slave inputs connections
      structural_objectRef Mout_oe_ram_port = mod->find_member("Mout_oe_ram", port_o_K, cir);
      THROW_ASSERT(Mout_oe_ram_port, "Mout_oe_ram port is missing");
      structural_objectRef S_oe_ram_port = mod->find_member("S_oe_ram", port_o_K, cir);
      if(S_oe_ram_port)
      {
         PP(hdl_file, "assign S_oe_ram = Mout_oe_ram;\n");
      }
      structural_objectRef Mout_we_ram_port = mod->find_member("Mout_we_ram", port_o_K, cir);
      THROW_ASSERT(Mout_we_ram_port, "Mout_we_ram port is missing");
      structural_objectRef S_we_ram_port = mod->find_member("S_we_ram", port_o_K, cir);
      if(S_we_ram_port)
      {
         PP(hdl_file, "assign S_we_ram = Mout_we_ram;\n");
      }
      structural_objectRef Mout_addr_ram_port = mod->find_member("Mout_addr_ram", port_o_K, cir);
      THROW_ASSERT(Mout_addr_ram_port, "Mout_addr_ram port is missing");
      structural_objectRef S_addr_ram_port = mod->find_member("S_addr_ram", port_o_K, cir);
      if(S_addr_ram_port)
      {
         PP(hdl_file, "assign S_addr_ram = Mout_addr_ram;\n");
      }

      structural_objectRef Mout_data_ram_size_port = mod->find_member("Mout_data_ram_size", port_o_K, cir);
      THROW_ASSERT(Mout_data_ram_size_port, "Mout_data_ram_size port is missing");
      structural_objectRef S_data_ram_size_port = mod->find_member("S_data_ram_size", port_o_K, cir);
      if(S_data_ram_size_port)
      {
         PP(hdl_file, "assign S_data_ram_size = Mout_data_ram_size;\n");
      }

   }

   /// write code for simulation time computation
   writer->write_comment(hdl_file, "Check done signal\n");
   PP(hdl_file, "always @(negedge clock)\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "if (done_port == 1)\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "endTime = $time;\n\n");
   writer->write_comment(hdl_file, "Simulation time (clock cycles) = 1+(time elapsed (seconds) / clock cycle (seconds per cycle)) (until done_port is 1)\n");
   PP(hdl_file, "sim_time = 1 + (endTime + `HALF_CLOCK_PERIOD - startTime)/`CLOCK_PERIOD;\n\n");

   PP(hdl_file, "success = 1;\n");
   PP(hdl_file, "compare_outputs = 0;\n");

   if (mod->find_member(RETURN_PORT_NAME, port_o_K, cir))
   {
      PP(hdl_file, "registered_" + std::string(RETURN_PORT_NAME) + " = " + std::string(RETURN_PORT_NAME) + ";\n");
   }

   PP(hdl_file, "\n#`CLOCK_PERIOD; ");
   writer->write_comment(hdl_file, "Wait some cycles (needed for a correct simulation)\n\n");

   /// write code for expected output values extraction
   for (unsigned int i = 0; i < mod->get_in_port_size(); i++)
   {
      if (GetPointer<port_o>(mod->get_in_port(i))->get_is_memory() || !mod->get_in_port(i)->get_typeRef()->treenode || !tree_helper::is_a_pointer(TreeM, mod->get_in_port(i)->get_typeRef()->treenode)) continue;

      std::string unmangled_name = mod->get_in_port(i)->get_id();
      std::string port_name = HDL_manager::convert_to_identifier(writer.get(), unmangled_name);
      std::string output_name = "ex_" + unmangled_name;
      long long int bitsize = tree_helper::size(TreeM, tree_helper::get_pointed_type(TreeM, tree_helper::get_type_index(TreeM,mod->get_in_port(i)->get_typeRef()->treenode)));

      PP(hdl_file, "\n");
      writer->write_comment(hdl_file, "OPTIONAL - Read a value for " + output_name + " --------------------------------------------------------------\n");

      PP(hdl_file, "_i_ = 0;\n");
      PP(hdl_file, "_ch_ = $fgetc(file);\n");
      PP(hdl_file, "while (_ch_ == \"/\" || _ch_ == \"\\n\" || _ch_ == \"o\") ");
      PP(hdl_file, "begin\n");
      PP.indent();
      PP(hdl_file, "if (_ch_ == \"o\")\n");
      PP(hdl_file, "begin\n");
      PP.indent();
      PP(hdl_file, "compare_outputs = 1;\n");
      PP(hdl_file, "_r_ = $fscanf(file,\"%b\\n\", " + output_name + "); ");
      writer->write_comment(hdl_file, "expected format: bbb...b (example: 00101110)\n");
      PP(hdl_file, "if (_r_ != 1)\n");
      PP(hdl_file, "begin\n");
      PP.indent();
      writer->write_comment(hdl_file, "error\n");
      PP(hdl_file, "$display(\"ERROR - Unknown error while reading the file. Character found: %c\", _ch_);\n");
      PP(hdl_file, "$fclose(res_file);\n");
      PP(hdl_file, "$fclose(file);\n");
      PP(hdl_file, "$finish;\n");
      PP.deindent();
      PP(hdl_file, "end\n");
      PP(hdl_file, "else\n");
      PP.indent();
      PP(hdl_file, "$display(\"Value found for output " + output_name + ": %b\", " + output_name + ");\n");
      PP.deindent();
      PP(hdl_file, "$display(\" res = %b " + port_name + " = %d " " mem[" + port_name  + " + %d - base_addr] = %d  expected = %d \\n\", ");
      PP(hdl_file, "{");
      for(unsigned int bitsize_index=0; bitsize_index < bitsize; bitsize_index = bitsize_index + 8)
      {
         if(bitsize_index) PP(hdl_file, ", ");
         PP(hdl_file, "mem[" + port_name + " + _i_*" + boost::lexical_cast<std::string>(bitsize/8) + " + " + boost::lexical_cast<std::string>((bitsize - bitsize_index)/8-1) + " - base_addr]");
      }
      PP(hdl_file,  "} == " + output_name + ", ");
      PP(hdl_file,  port_name + ", _i_*" + boost::lexical_cast<std::string>(bitsize/8) + ", {");
      for(unsigned int bitsize_index=0; bitsize_index < bitsize; bitsize_index = bitsize_index + 8)
      {
         if(bitsize_index) PP(hdl_file, ", ");
         PP(hdl_file, "mem[" + port_name + " + _i_*" + boost::lexical_cast<std::string>(bitsize/8) + " + " + boost::lexical_cast<std::string>((bitsize - bitsize_index)/8-1) + " - base_addr]");
      }
      PP(hdl_file, "}, " + output_name+ ");\n");

      PP(hdl_file, "if ({");
      for(unsigned int bitsize_index=0; bitsize_index < bitsize; bitsize_index = bitsize_index + 8)
      {
         if(bitsize_index) PP(hdl_file, ", ");
         PP(hdl_file, "mem[" + port_name + " + _i_*" + boost::lexical_cast<std::string>(bitsize/8) + " + " + boost::lexical_cast<std::string>((bitsize - bitsize_index)/8-1) + " - base_addr]");
      }
      PP(hdl_file,  "} !== " + output_name);
      PP(hdl_file, ")\n");
      PP.indent();
      PP(hdl_file, "success = 0;\n");
      PP.deindent();

      PP(hdl_file, "_i_ = _i_ + 1;\n");
      PP(hdl_file, "_ch_ = $fgetc(file);\n");
      PP.deindent();
      PP(hdl_file, "end\n");
      PP(hdl_file, "else\n");
      PP(hdl_file, "begin\n");
      PP.indent();
      writer->write_comment(hdl_file, "skip comments and empty lines\n");
      PP(hdl_file, "_r_ = $fgets(line, file);\n");
      PP(hdl_file, "_ch_ = $fgetc(file);\n");
      PP.deindent();
      PP(hdl_file, "end\n");

      PP.deindent();
      PP(hdl_file, "end\n");

      PP(hdl_file, "if (_ch_ == \"e\")\n");
      PP.indent();
      PP(hdl_file, "_r_ = $fgets(line, file);\n");
      PP.deindent();
      PP(hdl_file, "else\n");
      PP(hdl_file, "begin\n");
      PP.indent();
      writer->write_comment(hdl_file, "error\n");
      PP(hdl_file, "$display(\"ERROR - Unknown error while reading the file. Character found: %c\", _ch_);\n");
      PP(hdl_file, "$fclose(res_file);\n");
      PP(hdl_file, "$fclose(file);\n");
      PP(hdl_file, "$finish;\n");
      PP.deindent();
      PP(hdl_file, "end\n");

   }

   if (mod->find_member(RETURN_PORT_NAME, port_o_K, cir))
   {
      std::string output_name = "ex_" + std::string(RETURN_PORT_NAME);

      PP(hdl_file, "_i_ = 0;\n");
      PP(hdl_file, "_ch_ = $fgetc(file);\n");
      PP(hdl_file, "while (_ch_ == \"/\" || _ch_ == \"\\n\" || _ch_ == \"o\") ");
      PP(hdl_file, "begin\n");
      PP.indent();
      PP(hdl_file, "if (_ch_ == \"o\")\n");
      PP(hdl_file, "begin\n");
      PP.indent();
      PP(hdl_file, "compare_outputs = 1;\n");
      PP(hdl_file, "_r_ = $fscanf(file,\"%b\\n\", " + output_name + "); ");
      writer->write_comment(hdl_file, "expected format: bbb...b (example: 00101110)\n");
      PP(hdl_file, "if (_r_ != 1)\n");
      PP(hdl_file, "begin\n");
      PP.indent();
      writer->write_comment(hdl_file, "error\n");
      PP(hdl_file, "$display(\"ERROR - Unknown error while reading the file. Character found: %c\", _ch_);\n");
      PP(hdl_file, "$fclose(res_file);\n");
      PP(hdl_file, "$fclose(file);\n");
      PP(hdl_file, "$finish;\n");
      PP.deindent();
      PP(hdl_file, "end\n");
      PP(hdl_file, "else\n");
      PP.indent();
      PP(hdl_file, "$display(\"Value found for output " + output_name + ": %b\", " + output_name + ");\n");
      PP.deindent();
      PP(hdl_file, "$display(\" " + std::string(RETURN_PORT_NAME) + " = %d   expected = %d \\n\", registered_" + std::string(RETURN_PORT_NAME) + ", ex_" + std::string(RETURN_PORT_NAME) + ");\n");
      PP(hdl_file, "if (registered_" + std::string(RETURN_PORT_NAME) + " !== " + output_name + ")\n");
      PP.indent();
      PP(hdl_file, "success = 0;\n");
      PP.deindent();

      PP(hdl_file, "_i_ = _i_ + 1;\n");
      PP(hdl_file, "_ch_ = $fgetc(file);\n");
      PP.deindent();
      PP(hdl_file, "end\n");
      PP(hdl_file, "else\n");
      PP(hdl_file, "begin\n");
      PP.indent();
      writer->write_comment(hdl_file, "skip comments and empty lines\n");
      PP(hdl_file, "_r_ = $fgets(line, file);\n");
      PP(hdl_file, "_ch_ = $fgetc(file);\n");
      PP.deindent();
      PP(hdl_file, "end\n");

      PP.deindent();
      PP(hdl_file, "end\n");

      PP(hdl_file, "if (_ch_ == \"e\")\n");
      PP.indent();
      PP(hdl_file, "_r_ = $fgets(line, file);\n");
      PP.deindent();
      PP(hdl_file, "else\n");
      PP(hdl_file, "begin\n");
      PP.indent();
      writer->write_comment(hdl_file, "error\n");
      PP(hdl_file, "$display(\"ERROR - Unknown error while reading the file. Character found: %c\", _ch_);\n");
      PP(hdl_file, "$fclose(res_file);\n");
      PP(hdl_file, "$fclose(file);\n");
      PP(hdl_file, "$finish;\n");
      PP.deindent();
      PP(hdl_file, "end\n");
   }
   writer->write_comment(hdl_file, "Compare output with expected output (if given)\n");
   PP(hdl_file, "if (compare_outputs == 1)\n");

   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "$display(\"Simulation ended after %d cycles.\\n\", sim_time);\n");
   PP(hdl_file, "if (success == 1)\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "$display(\"Simulation completed with SUCCESS\\n\");\n");
   PP(hdl_file, "$fwrite(res_file, \"1\\t\");\n");
   PP.deindent();
   PP(hdl_file, "end\n");
   PP(hdl_file, "else\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "$display(\"Simulation FAILED\\n\");\n");
   PP(hdl_file, "$fwrite(res_file, \"0\\t\");\n");
   PP.deindent();
   PP(hdl_file, "end\n");
   PP.deindent();
   PP(hdl_file, "end\n");


   PP(hdl_file, "else\n");
   PP(hdl_file, "begin\n");
   PP.indent();
   PP(hdl_file, "$display(\"Simulation ended after %d cycles (no expected outputs specified).\\n\", sim_time);\n");
   PP(hdl_file, "$fwrite(res_file, \"-\\t\");\n");
   PP.deindent();
   PP(hdl_file, "end\n");
   PP(hdl_file, "$fwrite(res_file, \"%d\\n\", sim_time);\n");
   if(first_valid_input)
   {
      PP(hdl_file, "# `HALF_CLOCK_PERIOD; ");
      writer->write_comment(hdl_file, "wait a cycle (needed for a correct simulation)\n");
      PP(hdl_file, "$fclose(res_file);\n");
      PP(hdl_file, "$fclose(file);\n");
      PP(hdl_file, "$finish;\n");
   }
   else
   {
      PP(hdl_file, "#`HALF_CLOCK_PERIOD ");
      writer->write_comment(hdl_file, "Restart a new computation if possible\n");
      PP(hdl_file, std::string(START_PORT_NAME) + " <= 1'b1;\n");
   }
   PP.deindent();
   PP(hdl_file, "end\n");
   PP.deindent();
   PP(hdl_file, "end\n\n");
   PP.deindent();
   PP(hdl_file, "endmodule\n\n");
}
