/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file StateTransitionGraph.cpp
 * @brief File contanining the structures necessary to manage a graph that will represent a state transition graph
 *
 * This file contains the necessary data structures used to represent a state transition graph
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
 */

#include "StateTransitionGraph.hpp"
#include "StateTransitionGraph_constructor.hpp"

#include <math.h>

#include "hls.hpp"
#include "behavioral_helper.hpp"
#include "function_behavior.hpp"
#include "var_pp_functor.hpp"

#include "schedule.hpp"
#include "fu_binding.hpp"
#include "allocation.hpp"

#include "Parameter.hpp"
#include "boost/lexical_cast.hpp"
#include "constant_strings.hpp"

#include "dbgPrintHelper.hpp"
#include "basic_block.hpp"

//Boost includes
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/breadth_first_search.hpp>
#include <boost/graph/reverse_graph.hpp>
#include "cyclic_topological_sort.hpp"
#include <boost/graph/graphviz.hpp>
#include <boost/lexical_cast.hpp>

#include <vector>

StateTransitionGraph::StateTransitionGraph(const ParameterRef _Param, const behavioral_helperRef BH, const graphRef _op_function_graph) :
   Param(_Param),
   debug_level(_Param->getOption<int>(OPT_debug_level)),
   output_level(_Param->getOption<int>(OPT_output_level)),
   op_function_graph(_op_function_graph)
{
   //base graph
   STG_bulk = new graphs_collection(graph_infoRef(new StateTransitionGraphInfo));

   //filtered graphs
   ACYCLIC_STG_graph = new graph(STG_bulk, TransitionInfo::EDGE_NORMAL_T);
   STG_graph = new graph(STG_bulk, TransitionInfo::EDGE_NORMAL_T | TransitionInfo::EDGE_FEEDBACK_T);

   GET_GRAPH_INFO(STG_bulk, StateTransitionGraphInfo)->BH = BH;
   GET_GRAPH_INFO(STG_bulk, StateTransitionGraphInfo)->op_function_graph = op_function_graph;

   STG_builder = StateTransitionGraph_constructorRef(new StateTransitionGraph_constructor(STG_bulk, BH, op_function_graph));
   STG_builder->create_entry_state();
   STG_builder->create_exit_state();

}

StateTransitionGraph::~StateTransitionGraph()
{
   delete STG_bulk;
   delete STG_graph;
   delete ACYCLIC_STG_graph;
}

void StateTransitionGraph::writeDot(const std::string& fileName, graph_type type) const
{
   graph *g;
   switch (type)
   {
      case ACYCLIC:
         g = ACYCLIC_STG_graph;
         break;
      case COMPLETE:
         g = STG_graph;
         break;
      default:
         g=0;
         THROW_ERROR("Graph type not supported)");
   }
   std::string output_directory = Param->getOption<std::string>(OPT_output_directory) + "/HLS_dot/";
   std::string functionName = GET_GRAPH_INFO(STG_bulk, StateTransitionGraphInfo)->BH->get_function_name();
   output_directory += functionName +"/";
   if (!boost::filesystem::exists(output_directory)) boost::filesystem::create_directories(output_directory);
   std::ofstream f((output_directory + fileName).c_str());
   boost::write_graphviz(f, *g, NodeWriter(GET_GRAPH_INFO(STG_bulk, StateTransitionGraphInfo)->BH, g,
                                           op_function_graph,
                                           GET_GRAPH_INFO(STG_bulk, StateTransitionGraphInfo)->entry_node,
                                           GET_GRAPH_INFO(STG_bulk, StateTransitionGraphInfo)->exit_node),
                         EdgeWriter(GET_GRAPH_INFO(STG_bulk, StateTransitionGraphInfo)->BH, g, op_function_graph));
}

/// compute and return states where the operation could be executed
std::set<vertex> StateTransitionGraph::get_execution_states(const vertex& op) const
{
   std::set<vertex> vertex_set;
   VertexIterator v, vend;
   for (tie(v, vend) = boost::vertices(*STG_graph); v != vend; v++)
   {
      const std::list<vertex>& operations = GET_OPERATION_LIST(STG_graph, *v);
      std::list<vertex>::const_iterator op_it_end = operations.end();
      bool stop_flag;
      stop_flag = false;
      for(std::list<vertex>::const_iterator op_it = operations.begin(); op_it !=  op_it_end && !stop_flag; ++op_it)
      {
         if(*op_it == op)
         {
            stop_flag = true;
            vertex_set.insert(*v);
         }
      }

   }
   THROW_ASSERT(vertex_set.size() > 0, "Something wrong! Operation " + GET_NAME(op_function_graph, op) + " is executed "
                + "into no states");
   return vertex_set;
}


struct compute_csteps : public boost::bfs_visitor<>
{
   compute_csteps(graph *stg_, const vertex& entry_, std::tr1::unordered_map<vertex, int>& computed_) :
         entry(entry_), computed(computed_), stg(stg_) { computed[entry] = -1; }

   template <typename Vertex, typename Graph>
   void finish_vertex(const Vertex& u, Graph&)
   {
      if (u == entry)
         return;
      InEdgeIterator ie, iend;
      int max = INT_MIN;
      for (tie(ie, iend) = boost::in_edges(u, *stg); ie != iend; ie++)
      {
         if (computed[boost::source(*ie, *stg)] > max)
            max = computed[boost::source(*ie, *stg)];
      }
      computed[u] = max + 1;
   }

protected:
   const vertex& entry;
   std::tr1::unordered_map<vertex, int>& computed;
   graph *stg;
};

int StateTransitionGraph::get_control_step(const vertex& state) const
{
   std::tr1::unordered_map<vertex, int> CSteps;
   compute_csteps vis(ACYCLIC_STG_graph, GET_GRAPH_INFO(STG_bulk, StateTransitionGraphInfo)->entry_node, CSteps);
   //call the breadth-first-search on the finite state machine graph, using its visitor
   boost::breadth_first_search(*ACYCLIC_STG_graph,
                               GET_GRAPH_INFO(STG_bulk, StateTransitionGraphInfo)->entry_node,
                               visitor(vis));
   THROW_ASSERT(CSteps.find(state) != CSteps.end(), "Not computed control step");
   return CSteps.find(state)->second;

}

vertex StateTransitionGraph::get_entry_state() const
{
   return GET_GRAPH_INFO(STG_bulk, StateTransitionGraphInfo)->entry_node;
}

vertex StateTransitionGraph::get_exit_state() const
{
   return GET_GRAPH_INFO(STG_bulk, StateTransitionGraphInfo)->exit_node;
}

void StateTransitionGraph::print_statistics() const
{
   PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Number of states: " + STR(boost::num_vertices(*STG_bulk) - 2));
}

std::set<std::pair<vertex, unsigned int> > StateTransitionGraph::get_conditions(const vertex& v) const
{
   std::set<std::pair<vertex, unsigned int> > Conditions;

   InEdgeIterator ie, iend;
   for (tie(ie, iend) = boost::in_edges(v, *STG_graph); ie != iend; ie++)
   {
      std::set<std::pair<vertex, unsigned int> > EdgeConditions = GET_EDGE_INFO(STG_bulk, TransitionInfo, *ie)->Cond;
      for (std::set<std::pair<vertex, unsigned int> >::const_iterator c = EdgeConditions.begin(); c != EdgeConditions.end(); c++)
      {
         if (Conditions.find(*c) != Conditions.end()) continue;
         if (c->second == TransitionInfo::DONTCARE)
         {
            if (Conditions.find(std::make_pair(c->first, T_COND)) != Conditions.end())
               Conditions.erase(std::make_pair(c->first, T_COND));
            if (Conditions.find(std::make_pair(c->first, F_COND)) != Conditions.end())
               Conditions.erase(std::make_pair(c->first, F_COND));
            Conditions.insert(*c);
         }
         if (c->second == T_COND)
         {
            if (Conditions.find(std::make_pair(c->first, F_COND)) != Conditions.end())
            {
               Conditions.erase(std::make_pair(c->first, F_COND));
               Conditions.insert(std::make_pair(c->first, TransitionInfo::DONTCARE));
            }
            if (Conditions.find(std::make_pair(c->first, TransitionInfo::DONTCARE)) == Conditions.end())
               Conditions.insert(*c);
         }
         if (c->second == F_COND)
         {
            if (Conditions.find(std::make_pair(c->first, T_COND)) != Conditions.end())
            {
               Conditions.erase(std::make_pair(c->first, T_COND));
               Conditions.insert(std::make_pair(c->first, TransitionInfo::DONTCARE));
            }
            if (Conditions.find(std::make_pair(c->first, TransitionInfo::DONTCARE)) == Conditions.end())
               Conditions.insert(*c);
         }
         else
         {
            Conditions.insert(*c);
         }
      }
   }

   return Conditions;
}


void StateInfo::print(std::ostream& os) const
{
   const var_pp_functorConstRef vpp = var_pp_functorConstRef(new std_var_pp_functor(BH));
   simple_indent indenter('{', '}', 3);

   std::list<vertex>::const_iterator i, i_end;
   i_end = executing_operations.end();
   i = executing_operations.begin();

   if (i != i_end && GET_TYPE(op_function_graph, *i) == TYPE_ENTRY)
   {
      os << "START";
      return;
   }

   if (i != i_end && GET_TYPE(op_function_graph, *i) == TYPE_EXIT)
   {
      os << "END";
      return;
   }

   os << name << " | ";
   os << " { ";
   if (i != i_end)
   {
      os << GET_NAME(op_function_graph, *i) << " --\\> ";
      std::ostringstream sstream;
      BH->print_vertex(sstream, op_function_graph.get(), *i, vpp, indenter, true);
      std::string vertex_print = sstream.str();
      boost::replace_all(vertex_print, ">", "\\>");
      boost::replace_all(vertex_print, "<", "\\<");
      boost::replace_all(vertex_print, "||", "or");
      boost::replace_all(vertex_print, "|", "or");
      os << vertex_print;
      i++;
      unsigned int num = 1;
      for (; i != i_end; i++, num++)
      {
         os << GET_NAME(op_function_graph, *i) << " --\\> ";
         std::ostringstream sstream1;
         BH->print_vertex(sstream1, op_function_graph.get(), *i, vpp, indenter, true);
         vertex_print = sstream1.str();
         boost::replace_all(vertex_print, ">", "\\>");
         boost::replace_all(vertex_print, "<", "\\<");
         boost::replace_all(vertex_print, "||", "or");
         boost::replace_all(vertex_print, "|", "or");
         os << vertex_print;
      }
   }
   else
      os << " [none] ";

   os << " | ";
   i_end = ending_operations.end();
   i = ending_operations.begin();
   if (i != i_end)
   {
      os << GET_NAME(op_function_graph, *i) << " --\\> ";
      std::ostringstream sstream;
      BH->print_vertex(sstream, op_function_graph.get(), *i, vpp, indenter, true);
      std::string vertex_print = sstream.str();
      boost::replace_all(vertex_print, ">", "\\>");
      boost::replace_all(vertex_print, "<", "\\<");
      boost::replace_all(vertex_print, "||", "or");
      boost::replace_all(vertex_print, "|", "or");
      os << vertex_print;
      i++;
      unsigned int num = 1;
      for (; i != i_end; i++, num++)
      {
            os << GET_NAME(op_function_graph, *i) << " --\\> ";
            std::ostringstream sstream1;
            BH->print_vertex(sstream1, op_function_graph.get(), *i, vpp, indenter, true);
            vertex_print = sstream1.str();
            boost::replace_all(vertex_print, ">", "\\>");
            boost::replace_all(vertex_print, "<", "\\<");
            boost::replace_all(vertex_print, "||", "or");
            boost::replace_all(vertex_print, "|", "or");
            os << vertex_print;
      }
   }
   else
      os << " [none] ";
   os << "}";
}

void TransitionInfo::print(std::ostream& os) const
{
   for (std::set<std::pair<vertex, unsigned int> >::const_iterator it = Cond.begin(); it != Cond.end(); it++)
   {
      if (it->second == T_COND)
         os << GET_NAME(op_function_graph, it->first) << "(T)\\n";
      else if (it->second == F_COND)
         os << GET_NAME(op_function_graph, it->first) << "(F)\\n";
      else if (it->second == DONTCARE)
         os << GET_NAME(op_function_graph, it->first) << "(-)\\n";
      else if (it->second == default_COND)
         os << GET_NAME(op_function_graph, it->first) << "(default)\\n";
      else
      {
         const var_pp_functorConstRef std(new std_var_pp_functor(BH));
         os << GET_NAME(op_function_graph, it->first) << " (" << BH->print_constant(it->second, std) << ")\\n";
      }
   }
}

void NodeWriter::operator()(std::ostream& out, const vertex& v) const
{
   const StateInfo *temp = Cget_node_info< StateInfo > (v, *stg);
   out << "[";
   if (v == entry_node or v == exit_node)
      out << "color=blue,shape=Msquare,";
   else
      out << "shape=record,";
   out << "label=\"";
   temp->print(out);
   out << "\"]";
}

void EdgeWriter::operator()(std::ostream& out, const EdgeDescriptor& e) const
{
   const TransitionInfo *temp = Cget_edge_info<TransitionInfo>(e, *stg);
   if (TransitionInfo::EDGE_NORMAL_T & stg->get_selector(e))
      out << "[color=red3";
   else
      if (TransitionInfo::EDGE_FEEDBACK_T & stg->get_selector(e))
         out << "[color=green2";
      else
         THROW_ERROR("InconsistentDataStructure");
   out << ",label=\"";
   temp->print(out);
   out << "\"]";
}

TransitionInfo * StateTransitionGraph::getTransitionInfo(EdgeDescriptor e) const
{
   return GET_EDGE_INFO(STG_bulk, TransitionInfo, e);
}

StateInfo * StateTransitionGraph::getStateInfo(vertex v) const
{
   return GET_NODE_INFO(STG_bulk, StateInfo, v);
}

