/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file StateTransitionGraph.hpp
 * @brief This file contains the structures needed to manage a graph that will represent the state transition graph
 *
 * This file contains the needed data structures used to represent a graph used to represent a state 
 * transition graph associated to the specification. The State Transition Graph (STG) is defined as follows:
 *  - the nodes represent the states to be executed. Each node contains the set of operations to be executed when
 *    the control reaches the related state.
 *  - the edges represent the transition, subjected to the related conditions. In particular, a condition is the
 *    result of the evalution of a control operation that modifies the control flow.
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
 */
#ifndef STATETRANSITIONGRAPH_HPP
#define STATETRANSITIONGRAPH_HPP

//Utility includes
#include "refcount.hpp"

//STL includes
#include <tr1/unordered_map>

//Basic structure describing a property associated with each vertex
#include <node_info.hpp>
//Basic structure describing a property associated with each edge
#include <edge_info.hpp>
//Basic structure describing a property associated with the graph
#include <graph_info.hpp>
//Basic class representing a graph: we will simply used it in conjunction
//to the properties defined below: this is all it is necessary to define a
//new graph
#include <graph.hpp>

/**
 * @name Forward declarations
 */
//@{
REF_FORWARD_DECL(Parameter);
REF_FORWARD_DECL(StateTransitionGraph_constructor);
REF_FORWARD_DECL(graph);
REF_FORWARD_DECL(behavioral_helper);
//@}

//-----------------defining the graph property structures-----------//
/**
 * Structure holding information about the whole graph.
 */
struct StateTransitionGraphInfo : public graph_info
{
   /// this is the helper associated to the actual module
   behavioral_helperRef BH;

   /// reference to operation graph
   graphRef op_function_graph;

   /// this vertex represents the entry state of the state transition graph
   vertex entry_node;

   /// this vertex represents the exit state of the state transition graph
   vertex exit_node;

};
///definition of the refcount
typedef refcount<StateTransitionGraphInfo> StateTransitionGraphInfoRef;

//-----------------dec  fining the node property structures-----------//
/**
 * Structure holding informations about a node into graph. A node represent an execution state of the machine. It
 * contains operations to be started when execution will reach this state.
 */
struct StateInfo : public NodeInfo
{
   /// this is the helper associated to the actual module
   behavioral_helperRef BH;

   /// pointer to graph storing informations about operations
   graphRef op_function_graph;

   ///string used to give a label to the state
   std::string name;

   ///set of operation vertices which require inputs in this state
   std::list<vertex> executing_operations;

   ///set of operation vertices completing their execution in this state
   std::list<vertex> ending_operations;

   ///flag to check if the state is dummy or not
   bool is_dummy;

   /**
    * Implementation of print method for this kind of node. It simply prints the list of operations contained into this
    * state
    * @param os is the stream reference where you want to print
    */
   void print(std::ostream& os) const;

   /**
    * Constructor
    */
   StateInfo(): is_dummy(false) {}

};

#define GET_STATE_NAME(g,vertex_index) \
   Cget_node_info<StateInfo>(vertex_index,*(g))->name

#define GET_OPERATION_LIST(data,vertex_index) \
   Cget_node_info<StateInfo>(vertex_index,*(data))->executing_operations

#define GET_ENDING_OPERATION_LIST(data,vertex_index) \
   Cget_node_info<StateInfo>(vertex_index,*(data))->ending_operations

#define GET_DUMMY_STATE(data,vertex_index) \
   Cget_node_info<StateInfo>(vertex_index,*(data))->is_dummy

/// refcount definition
typedef refcount<StateInfo> StateInfoRef;


//-----------------defining the edge property structures-----------//

/**
 * Structure holding the information about an edge into the graph. It specialize generic edge_info
 * The property associated with edge is the control condition
 */
struct TransitionInfo : public edge_info
{
   /// this is the helper associated to the actual module
   behavioral_helperRef BH;

   /// pointer to graph storing informations about operations
   graphRef op_function_graph;

   ///Types associated with the edges of the graph.
   ///Normal edge
   static const int EDGE_NORMAL_T      = 1 << 1;
   ///Feedback edge
   static const int EDGE_FEEDBACK_T    = 1 << 2;

   ///Don't care information
   static const unsigned int DONTCARE  = 1 << 0;

   ///set of operation vertices which require inputs in this transition
   std::list<vertex> executing_operations;

   ///set of operation vertices completing their execution in this transition
   std::list<vertex> ending_operations;

   /**
    * This set stores a set of pair where the first represents the controlling operations while the second encodes the controlling condition value/label.
    * It can be 0 (without any control dependence), T_COND (it depends on
    * a true value of a condition) or F_COND (it depends on a false value of a condition) or a label in case of a switch statement
    */
   std::set<std::pair<vertex, unsigned int> > Cond;

   /**
    * Print the information associated with the edge of the graph.
    * @param os is the output stream.
    */
   void print(std::ostream& os) const;

};

#define GET_TRANSITION_CONDITION(data,edge_index) \
   Cget_edge_info<TransitionInfo>(edge_index,*(data))->Cond

#define GET_TRANSITION_OPERATION_LIST(data,edge_index) \
   Cget_edge_info<TransitionInfo>(edge_index,*(data))->executing_operations

#define GET_TRANSITION_ENDING_OPERATION_LIST(data,edge_index) \
   Cget_edge_info<TransitionInfo>(edge_index,*(data))->ending_operations


/// refcount about edge info
typedef refcount<TransitionInfo> TransitionInfoRef;


//-----------------defining the writer functors -----------//

/**
 * Functor template used to write the content of the nodes to a dotty file.
 */
class NodeWriter
{

   private:

      /// this is the helper associated to the actual module
      const behavioral_helperRef BH;

      /// reference to the STG graph
      const graph* stg;

      /// reference to operation graph
      const graphRef op_function_graph;

      /// vertex that contains entry node
      vertex entry_node;

      /// vertex that contains exit node
      vertex exit_node;

   public:

      /**
       * Constructor. It initialize reference to the graph provided as parameter
       */
      NodeWriter(const behavioral_helperRef _BH, const graph* _stg, const graphRef _op_function_graph, const vertex& _entry_node, const vertex& _exit_node) :
            BH(_BH), stg(_stg), op_function_graph(_op_function_graph), entry_node(_entry_node), exit_node(_exit_node)
      { }

      /**
       * Functor actually called by the boost library to perform the writing
       */
      void operator()(std::ostream& out, const vertex& v) const;
};

/**
 * Functor used to write the content of the edges to a dotty file and it is used to write specific edge properties
 * such as condition.
 */
class EdgeWriter
{

   private:

      /// this is the helper associated to the actual module
      behavioral_helperRef BH;

      /// reference to the STG graph
      const graph* stg;

      /// reference to operation graph
      const graphRef op_function_graph;

   public:

      /**
       * Constructor. It initialize reference to the graph provided as parameter
       */
      EdgeWriter(behavioral_helperRef _BH, const graph* _stg, const graphRef _op_function_graph) : BH(_BH), stg(_stg), op_function_graph(_op_function_graph) {}

      /**
       * Functor actually called by the boost library to perform the writing
       */
      void operator()(std::ostream& out, const EdgeDescriptor& e) const;

};


//------------- defining the manager for graph that will be used to represent finite state machine -----------//

/**
 * Class used to manage a graph into finite state machine representation; it contains methods to build the graph,
 * to add nodes, edges... and also to write the dotty representation of the graph. Two different graphs have been stored:
 * the finite state machine and the related acyclic version (without any feedback edges)
 */
class StateTransitionGraph
{
   public:

      /// reference to the class for building the graph
      StateTransitionGraph_constructorRef STG_builder;

   private:

      /// class containing all the parameters
      const ParameterRef Param;

      /// debugging level
      int debug_level;

      /// verbosity level
      int output_level;

      /// reference to operation graph
      const graphRef op_function_graph;

      /// pointer to the base graph
      graphs_collection *STG_bulk;

      /// Pointer to acyclic version of stg (it's a filtering of the base one)
      graph *ACYCLIC_STG_graph;

      /// Pointer to complete version of stg (it's a filtering of the base one)
      graph *STG_graph;

   public:

      /// These are graph type defined into this class:
      ///  - COMPLETE is the complete finite state machine graph (also with feedback edges)
      ///  - ACYCLIC is directed acyclic version of finite state machine graph (without feedback edges)
      enum graph_type
      {
         COMPLETE = 0,
         ACYCLIC
      };

      /**
       * Constructor of the class. It creates a new empty graph and it sets reference to hls class
       */
      StateTransitionGraph(const ParameterRef Param, const behavioral_helperRef BH, const graphRef op_function_graph);

      /**
       * Destructor
       */
      ~StateTransitionGraph();

      /**
       * Writes the graph to a dotty file
       * @param fileName the name of the output dotty file
       * @param type is the type of graph to be written (complete stg is the default option)
       */
      void writeDot(const std::string& fileName, graph_type type = COMPLETE) const;

      std::set<vertex> get_execution_states(const vertex& op) const;

      int get_control_step(const vertex& state) const;

      /**
       * Returns pointer to state transition graph created.
       * @return reference to a graph that contains informations about operations to be executed and control
       *         edges
       */
      const graph* get_stg() const { return STG_graph; }

      const graph* get_astg() const { return ACYCLIC_STG_graph; }

      void print_statistics() const;

      /**
       * Returns the name associated to given vertex
       * @param v is graph vertex
       * @return a string that contains the name associated to vertex
       */
      std::string get_state_name(const vertex& v) const
      {
         return GET_STATE_NAME(STG_bulk, v);
      }

      /**
       * Gets vertex that represents state that contains entry node
       * @return the vertex of state associated to entry node
       */
      vertex get_entry_state() const;

      /**
       * Gets vertex that represents state that contains exit node
       * @return the vertex of state associated to exit node
       */
      vertex get_exit_state() const;

      TransitionInfo * getTransitionInfo(EdgeDescriptor e) const;

      StateInfo * getStateInfo(vertex e) const;

      std::set<std::pair<vertex, unsigned int> > get_conditions(const vertex& v) const;

};

/// refcount definition to allocate the class
typedef refcount<StateTransitionGraph> StateTransitionGraphRef;

/// state name prefix
#define STATE_NAME_PREFIX "S_"

#endif
