/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file StateTransitionGraph_constructor.cpp
 * @brief File contanining the structures necessary to manage a graph that will represent a state transition graph
 *
 * This file contains the necessary data structures used to represent a state transition graph
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
 */

#include "StateTransitionGraph_constructor.hpp"

#include "StateTransitionGraph.hpp"

#include "op_graph.hpp"
#include "op_edge_info.hpp"

StateTransitionGraph_constructor::StateTransitionGraph_constructor(graphs_collection *& og, const behavioral_helperRef _BH, const graphRef _op_function_graph) :
   state_index(0),
   STG_bulk(og),
   BH(_BH),
   op_function_graph(_op_function_graph)
{

}

void StateTransitionGraph_constructor::create_entry_state()
{
   vertex newVertex = STG_bulk->AddVertex();
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->BH = BH;
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->op_function_graph = op_function_graph;
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->name = "ENTRY";
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->executing_operations.push_back(GET_ENTRY_VERTEX(op_function_graph.get()));
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->ending_operations.push_back(GET_ENTRY_VERTEX(op_function_graph.get()));

   GET_GRAPH_INFO(STG_bulk, StateTransitionGraphInfo)->entry_node = newVertex;
}

void StateTransitionGraph_constructor::create_exit_state()
{
   vertex newVertex = STG_bulk->AddVertex();
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->BH = BH;
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->op_function_graph = op_function_graph;
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->name = "EXIT";
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->executing_operations.push_back(GET_EXIT_VERTEX(op_function_graph.get()));
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->ending_operations.push_back(GET_EXIT_VERTEX(op_function_graph.get()));

   GET_GRAPH_INFO(STG_bulk, StateTransitionGraphInfo)->exit_node = newVertex;
}

vertex StateTransitionGraph_constructor::create_state(std::list<vertex> & exec_op, std::list<vertex> & end_op)
{
   vertex newVertex = STG_bulk->AddVertex();
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->BH = BH;
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->op_function_graph = op_function_graph;
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->name = std::string(STATE_NAME_PREFIX + boost::lexical_cast<std::string>(state_index));
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->executing_operations = exec_op;
   GET_NODE_INFO(STG_bulk, StateInfo, newVertex)->ending_operations = end_op;

   state_index++;

   return newVertex;
}


EdgeDescriptor StateTransitionGraph_constructor::connect_state(const vertex& src, const vertex& tgt, int type, std::list<vertex> & tr_exec_op, std::list<vertex> & tr_end_op)
{
   //get the vertex iterator
   VertexIterator vIterBeg, vIterEnd;
   tie(vIterBeg, vIterEnd) = boost::vertices(*STG_bulk);
   //check that source and target have been already added
   THROW_ASSERT(std::find(vIterBeg, vIterEnd, src) != vIterEnd and std::find(vIterBeg, vIterEnd, tgt) != vIterEnd, "Source vertex or target one is not present into graph");
   EdgeDescriptor e;
   bool exists;
   boost::tie(e, exists) = boost::edge(src, tgt, *STG_bulk);
   THROW_ASSERT(!exists, "transition already added");
   //edge creation
   bool inserted;
   boost::tie(e, inserted) = boost::add_edge(src, tgt, EdgeProperty(type), *STG_bulk);
   //edge property setting
   TransitionInfoRef eInfo = TransitionInfoRef(new TransitionInfo);
   eInfo->BH = BH;
   eInfo->op_function_graph = op_function_graph;
   eInfo->executing_operations = tr_exec_op;
   eInfo->ending_operations = tr_end_op;
   SET_EDGE_INFO_REF(STG_bulk, TransitionInfo, e, eInfo);

   return e;
}

void StateTransitionGraph_constructor::set_condition(const EdgeDescriptor& e, const std::set<std::pair<vertex, unsigned int> >& cond)
{
   THROW_ASSERT(GET_EDGE_INFO(STG_bulk, TransitionInfo, e), "Not valid edge info");
   // getting informations from graph
   GET_EDGE_INFO(STG_bulk, TransitionInfo, e)->Cond.clear();
   add_conditions(cond, GET_EDGE_INFO(STG_bulk, TransitionInfo, e)->Cond);
}

void StateTransitionGraph_constructor::add_conditions(const std::set<std::pair<vertex, unsigned int> > & EdgeConditions, std::set<std::pair<vertex, unsigned int> > & Conditions)
{
   for (std::set<std::pair<vertex, unsigned int> >::iterator Cond = EdgeConditions.begin(); Cond != EdgeConditions.end(); Cond++)
   {
      if (Conditions.find(std::make_pair(Cond->first, TransitionInfo::DONTCARE)) != Conditions.end())
      {
         continue;
      }
      if (Cond->second == TransitionInfo::DONTCARE)
      {
         Conditions.erase(std::make_pair(Cond->first, T_COND));
         Conditions.erase(std::make_pair(Cond->first, F_COND));
         Conditions.insert(std::make_pair(Cond->first, TransitionInfo::DONTCARE));
      }
      else if (Cond->second == T_COND)
      {
         if (Conditions.find(std::make_pair(Cond->first, F_COND)) != Conditions.end())
         {
            Conditions.erase(std::make_pair(Cond->first, F_COND));
            Conditions.insert(std::make_pair(Cond->first, TransitionInfo::DONTCARE));
         }
         else
            Conditions.insert(std::make_pair(Cond->first, T_COND));
      }
      else if (Cond->second == F_COND)
      {
         if (Conditions.find(std::make_pair(Cond->first, T_COND)) != Conditions.end())
         {
            Conditions.erase(std::make_pair(Cond->first, T_COND));
            Conditions.insert(std::make_pair(Cond->first, TransitionInfo::DONTCARE));
         }
         else
            Conditions.insert(std::make_pair(Cond->first, F_COND));
      }
      else
      {
         Conditions.insert(*Cond);
      }
   }
}
