/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file StateTransitionGraph_constructor.hpp
 * @brief File contanining the structures necessary to manage a graph that will represent a state transition graph
 *
 * This file contains the necessary data structures used to represent a graph used to represent a state transition graph
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
 */


#ifndef STATETRANSITIONGRAPHCONSTR_HPP
#define STATETRANSITIONGRAPHCONSTR_HPP

#include "graph.hpp"

#include "refcount.hpp"

REF_FORWARD_DECL(graph);
REF_FORWARD_DECL(behavioral_helper);

class StateTransitionGraph_constructor
{

   private:
      ///Index of the next state to be created
      unsigned int state_index;

      /// pointer to the base graph
      graphs_collection *& STG_bulk;

      /// this is the helper associated to the actual module
      behavioral_helperRef BH;

      /// pointer to graph storing informations about operations
      graphRef op_function_graph;

   public:

      /**
       * Constructor of the class. It creates a new empty graph and it sets reference to hls class
       * @param data is the reference the graph related to operations
       */
      StateTransitionGraph_constructor(graphs_collection *& og, const behavioral_helperRef BH, const graphRef op_function_graph);

      /**
       * Adds a new state managing the operations given as parameters
       * @param exec_op is a list of operation vertices that will be executed by the created state
       * @param end_op is a list of operation vertices that will end in the created state
       * @return the created vertex
       */
      vertex create_state(std::list<vertex> &exec_op, std::list<vertex> &end_op);

      /**
       *  create the STG entry vertex
       */
      void create_entry_state();
      /**
       *  create the STG exit vertex
       */
      void create_exit_state();

      /**
       * Creates a connection between two vertices into the graph. You can also specify the edge kind. The method checks
       * if the two vertices are stored into the graph, if one of them (or both) isn't, an exception is thrown.
       * @param src is the source vertex (as returned by create_state method)
       * @param tgt is the target vertex (as returned by create_state method)
       * @param type is the edge type that it's going to be created (normal, direct edge is default)
       * @param exec_op is a list of operation vertices that will be executed by the transition
       * @param end_op is a list of operation vertices that will end in the transition
       * @return the descriptor of edge here created
       */
      EdgeDescriptor connect_state(const vertex& src, const vertex& tgt, int type, std::list<vertex> & exec_op, std::list<vertex> & end_op);

      /**
       * Changes the control condition associated to an edge. If no condition is given, previous one is erased and
       * edge becomes without a true or false control dependence.
       * @param e is the edge you want to change the control condition
       * @param cond is the new control condition to be associated to edge (no control dependence condition as default)
       */
      void set_condition(const EdgeDescriptor& e, const std::set<std::pair<vertex, unsigned int> >& cond);

      static
      void add_conditions(const std::set<std::pair<vertex, unsigned int> > & EdgeConditions, std::set<std::pair<vertex, unsigned int> > & Conditions);

};

/// refcount definition to StateTransitionGraph_constructor class
typedef refcount<StateTransitionGraph_constructor> StateTransitionGraph_constructorRef;

#endif
