/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file clique_covering.hpp
 * @brief This header file include four different algorithms solving heuristically the clique covering problem.
 *
 * This header has 4 classes solving clique covering problem. Three of them consider the weighted of the edges during the clique selection.
 *
 * The first two heuristics are based on this two papers:
 *  - E. Tomita, A. Tanaka, H. Takahashi
 *    "The worst-case time complexity for generating all maximal cliques and computational experiments",
 *     Theoretical Computer Science, Volume 363, Issue 1, 2006.
 *  - K. C. Dukka Bahadur, Tatsuya Akutsu, Etsuji Tomita, and Tomokazu Seki.
 *    "Protein side-chain packing problem: a maximum edge-weight clique algorithmic approach",
 *    In Proceedings of the second conference on Asia-Pacific bioinformatics - Volume 29 (APBC '04),
 *    Yi-Ping Phoebe Chen (Ed.), Vol. 29. Australian Computer Society, Inc., Darlinghurst, Australia, Australia, 191-200.
 * The third one is a simple clique covering obtained by coloring the complement graph.
 * The last one performs iteratively the clique covering on a filtered graph induced by different edge weights.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
 */
#ifndef CLIQUE_COVERING_HPP
#define CLIQUE_COVERING_HPP

#include <set>
#include <tr1/unordered_set>
#include <vector>
#include <algorithm>

/// Graph includes
#include <boost/config.hpp>
#include <boost/pending/disjoint_sets.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/filtered_graph.hpp>
#include <boost/graph/graph_utility.hpp>
#include <boost/graph/graphviz.hpp>
#include <boost/graph/topological_sort.hpp>
#include <boost/version.hpp>
#include <boost/graph/properties.hpp>

/// matrix includes
#include <boost/numeric/ublas/matrix.hpp>

#include "exceptions.hpp"
#include "refcount.hpp"
#include "dsatur2_coloring.hpp"
#include "bipartite_matching.hpp"

/// unordered_set_difference
template <typename IIter1, typename Container, typename OIter>
OIter unordered_set_difference(IIter1 first1, IIter1 last1, Container second, OIter result)
{
   while (first1 != last1)
   {
      if (second.find(*first1) == second.end())
      {
         *result = *first1;
         ++result;
      }
      ++first1;
   }
   return result;
}

/// unordered_set_intersection
template <typename IIter1, typename Container, typename OIter>
OIter unordered_set_intersection(IIter1 first1, IIter1 last1, Container second, OIter result)
{
   while (first1 != last1)
   {
      if (second.find(*first1) != second.end())
      {
         *result = *first1;
         ++result;
      }
      ++first1;
   }
   return result;
}

struct edge_compatibility_selector
{
   public:

      /// The selector associated with the edge
      int selector;

      /// edge weight
      int weight;

      /**
       * Constructor with selector
       * @param _selector is the selector to be associated with the edge
       * @param _weight is the weight to be associated with the edge
       */
      edge_compatibility_selector(int _selector, int _weight) :
         selector(_selector), weight(_weight) {}

};

/// bulk compatibility graph
typedef boost::adjacency_list< boost::vecS, boost::vecS, boost::undirectedS, boost::property<boost::vertex_index_t, std::size_t>, edge_compatibility_selector>  boost_cc_compatibility_graph;

/// Predicate functor object used to select the proper set of vertices
template <typename Graph>
struct cc_compatibility_graph_vertex_selector
{
   public:
      typedef typename boost::graph_traits<Graph>::vertex_descriptor vertex_descriptor;
      typedef std::tr1::unordered_set<vertex_descriptor> SET_container;
      /// constructor
      cc_compatibility_graph_vertex_selector() : all(true), support(0) {}
      /// constructor
      cc_compatibility_graph_vertex_selector(SET_container * _support) : all(false), support(_support) {}
      /// selector operator
      bool operator()(const vertex_descriptor &v) const
      {
         if(all)
            return true;
         else
            return support->find(v) != support->end();
      }
   private:
      bool all;
      SET_container *support;
};

/// Predicate functor object used to select the proper set of edges
template <typename Graph>
struct cc_compatibility_graph_edge_selector
{
   private:
      /// The selector associated with the filtered graph
      int selector;

      /// bulk graph
      Graph * g;

      bool all;

   public:

      /**
       * Constructor for filtering only on selector
       * @param _selector is the selector of the filtered graph
       * @param _g is the graph
       */
      cc_compatibility_graph_edge_selector(const int _selector, Graph * _g) :
         selector(_selector),
         g(_g), all(false)
      {}

      /// all edges selector
      cc_compatibility_graph_edge_selector() :
         selector(0),
         g(0), all(true)
      {}

      /// edge selector operator
      template <typename Edge>
      bool operator()(const Edge& e) const
      {
         if(all)
            return true;
         else
            return selector & (*g)[e].selector;
      }
};

/// compatibility graph
typedef boost::filtered_graph<boost_cc_compatibility_graph, cc_compatibility_graph_edge_selector<boost_cc_compatibility_graph>, cc_compatibility_graph_vertex_selector<boost_cc_compatibility_graph> > cc_compatibility_graph;

/// refcount version of cc_compatibility_graph
typedef refcount<cc_compatibility_graph> cc_compatibility_graphRef;

/// cc_compatibility_graph vertex
typedef boost::graph_traits<cc_compatibility_graph>::vertex_descriptor C_vertex;
typedef boost::graph_traits<cc_compatibility_graph>::out_edge_iterator C_outEdgeIterator;

template<typename VertexType>
class clique_covering
{

   public:

      /// Defines all clique covering algorithm
      typedef enum { COLORING = 0, WEIGHTED_COLORING, TTT_CLIQUE_COVERING, TTT_CLIQUE_COVERING_FAST, BIPARTITE_MATCHING } algorithm_type;

   protected:


   public:

      /**
       * Default constructor.
       */
      clique_covering() {}

      /**
       * Destructor
       */
      virtual ~clique_covering() {}

      /**
       * Creates a reference to desidered solver
       * @param solver is the solver which you want to perform clique covering with
       * @param graph_type is the kind of graph you are going to create
       * @return a reference to the desidered solver
       */
      static
      typename refcount< clique_covering< VertexType> > create_solver(algorithm_type solver);
      /**
       * Adds a vertex to graph. It checks if element is already into graph. If it is, an assertion fails, otherwise
       * the vertex is added and the new index is saved for future checks
       * @param VertexType& is the reference to element that deals with compatibility
       * @return the new vertex index
       */
      virtual C_vertex add_vertex(const VertexType& element, const std::string& name ) = 0;

      /**
       * Adds an edge to graph. It check if source and vertex are stored into graph. If one of them isn't into graph,
       * the related assertion fails. If both vertices are stored, a (weighted edge is added)
       * @param src is the index of first vertex
       * @param dest is the index of second vertex
       * @param weight is the weight associated to edge (optional, set to unitary value as default)
       */
      virtual void add_edge(const VertexType& src, const VertexType& dest, int _weight) = 0;

      /**
       * Returns number of vertices into graph
       */
      virtual size_t num_vertices() = 0;

      /**
       * Abstract method that will execute clique covering algorithm. If you want to specialize the implementation
       * with you favourite algorithm, you have to implement this method.
       */
      virtual void exec() = 0;

      /**
       * Returns a clique
       * @param i is the i-th clique into graph
       * @return set of elements into clique
       */
      virtual std::set<VertexType> get_clique(unsigned int i) = 0;

      /**
       * Returns number of cliques into graph after performing clique covering
       */
      virtual size_t num_cliques()
      {
         return num_vertices();
      }

      /**
       * Writes a dotty reprentation of the actual graph
       * @param filename is the output filename
       */
      virtual void writeDot(const std::string &filename) const  = 0;


};



/**
 * Functor used by boost::write_graphviz to write the edge info
 */
class compatibility_edge_writer
{

   private:

      /// reference to graph where the node is stored
      const boost_cc_compatibility_graph& g;

   public:

      /**
       * Constructor. It initialize reference to the graph provided as parameter
       */
      compatibility_edge_writer(const boost_cc_compatibility_graph& _g) : g(_g) {}

      /**
       * Functor actually called by the boost library to perform the writing
       */
      void operator()(std::ostream& out, const boost::graph_traits<cc_compatibility_graph>::edge_descriptor& e) const
      {
         out << "[label=\"" << g[e].weight << "\"]";
      }
};


/**
 * Functor used by boost::write_graphviz to write the node info
 */
class compatibility_node_info_writer
{
   private:

      /// names of the vertices
      const std::map<C_vertex, std::string> &names;

   public:

      /// Constructor
      compatibility_node_info_writer(const std::map<C_vertex, std::string> &_names) : names(_names) {}

      /// Functor used to print a node
      void operator()(std::ostream& out, const C_vertex& v) const
      {
         out << "[label=\"" <<  names.find(v)->second << "\"]";
      }

};

/**
 * Class computing the maximal weighted clique from a generic graph
 */
template <typename Graph>
class TTT_maximal_weighted_clique
{
      /// vertex iterator
      typedef typename boost::graph_traits<Graph>::vertex_iterator vertex_iterator;
      /// vertex object
      typedef typename boost::graph_traits<Graph>::vertex_descriptor vertex;
      /// adjacency iterator
      typedef typename boost::graph_traits<Graph>::adjacency_iterator adjacency_iterator;
      /// out edge iterator
      typedef typename boost::graph_traits<Graph>::out_edge_iterator edge_iterator;

      /// set of vertices of the current clique
      std::tr1::unordered_set<vertex> Q;
      /// set of vertices of the maximum clique found so far
      std::set<vertex> Q_max;
      /// weight of Q
      int W_Q;
      /// weight of Q_max
      int W_Q_max;

      std::map<C_vertex, std::string> &names;

      /// return the vertex of subg with the maximum intersection with cand
      vertex get_max_weighted_adiacent_intersection(const std::tr1::unordered_set<vertex> &subg, const std::tr1::unordered_set<vertex>&cand, const Graph &g)
      {
         vertex result=boost::graph_traits<Graph>::null_vertex();
         THROW_ASSERT(!subg.empty(), "at least one element should belong to subg");
         int max_weighted_intersection = -1;
         const typename std::tr1::unordered_set<vertex>::const_iterator it_end = subg.end();
         for(typename std::tr1::unordered_set<vertex>::const_iterator it = subg.begin(); it != it_end; ++it)
         {
            int weight_intersection = 0;
            edge_iterator ei,ei_end;
            boost::tie(ei,ei_end) = boost::out_edges(*it, g);
            for(;ei != ei_end; ++ei)
            {
               if(cand.find(boost::target(*ei, g)) != cand.end())
                  weight_intersection+=g[*ei].weight;
            }
            if(weight_intersection>max_weighted_intersection)
            {
               max_weighted_intersection = weight_intersection;
               result = *it;
            }
         }
         THROW_ASSERT(max_weighted_intersection>=0, "something of wrong happen");
         return result;
      }

      /// return the vertex of ext having the maximum degree with respect to the graph g
      vertex get_max_weight_vertex(std::tr1::unordered_set<vertex> & ext, const Graph &g)
      {
         vertex result=boost::graph_traits<Graph>::null_vertex();
         int max_weight = -1;
         THROW_ASSERT(!ext.empty(), "at least one element should belong to ext");
         const typename std::tr1::unordered_set<vertex>::const_iterator it_end = ext.end();
         for(typename std::tr1::unordered_set<vertex>::const_iterator it = ext.begin(); it != it_end; ++it)
         {
            int cur_weight=0;
            edge_iterator ei,ei_end;
            boost::tie(ei,ei_end) = boost::out_edges(*it, g);
            for(;ei != ei_end; ++ei)
            {
               cur_weight+=g[*ei].weight;
            }
            if(cur_weight>max_weight)
            {
               result = *it;
               max_weight = cur_weight;
            }
         }
         THROW_ASSERT(max_weight>=0, "something of wrong happen");
         return result;
      }

      /// compute the delta of the weight by adding q_vertex to the clique
      int compute_delta_weight(vertex q_vertex, std::tr1::unordered_set<vertex> Q_set, const Graph &g)
      {
         int result=0;
         edge_iterator ei,ei_end;
         boost::tie(ei,ei_end) = boost::out_edges(q_vertex, g);
         for(;ei != ei_end; ++ei)
            if(Q_set.find(boost::target(*ei, g)) != Q_set.end())
               result+=g[*ei].weight;
         return result;
      }

      /// recorsive procedure expand defined in first cited paper
      void expand(std::tr1::unordered_set<vertex> &subg, std::tr1::unordered_set<vertex>&cand, const Graph &g, int upper_bound)
      {
         if(subg.empty() && Q.size() >= Q_max.size())
         {
            if(Q.size() > Q_max.size() || W_Q > W_Q_max)
            {
               Q_max.clear();
               Q_max.insert(Q.begin(), Q.end());
               W_Q_max = W_Q;
               //std::cerr << "clique W=" << W_Q_max << " size=" << Q_max.size() << std::endl;
            }
            return;
         }
         else if(cand.empty())
            return;

         /// get the vertex in subg with the maximum of adjacent vertices in cand
         vertex u =get_max_weighted_adiacent_intersection(subg, cand, g);
         //vertex u =get_max_adiacent_intersection(subg, cand, g);
         /// get adjacent vertices of u
         adjacency_iterator vi, vi_end;
         std::pair<adjacency_iterator, adjacency_iterator> p;
         p = boost::adjacent_vertices(u, g);
         vi = p.first;
         vi_end = p.second;
         /// set of vertices adjacent to u
         std::tr1::unordered_set<vertex> gamma_u(vi, vi_end);
         /// compute EXT_u = CAND - gamma_u
         std::tr1::unordered_set<vertex> EXT_u;
         unordered_set_difference(cand.begin(), cand.end(), gamma_u, std::inserter(EXT_u, EXT_u.end()));
         while(!EXT_u.empty())
         {
            vertex q = get_max_weight_vertex(EXT_u, g);
            //vertex q = get_max_degree_vertex(EXT_u, g);
            //std::cerr << names[q] << "," << std::endl;
            Q.insert(q);
            int W_Q_pre = W_Q;
            /// compute delta_weight
            int delta = compute_delta_weight(q, Q, g);
            W_Q += delta;
            boost::tie(vi, vi_end) = boost::adjacent_vertices(q, g);
            std::tr1::unordered_set<vertex> gamma_q(vi, vi_end);
            std::tr1::unordered_set<vertex> subg_q;
            unordered_set_intersection(subg.begin(), subg.end(), gamma_q, std::inserter(subg_q, subg_q.end()));
            std::tr1::unordered_set<vertex> cand_q;
            unordered_set_intersection(cand.begin(), cand.end(), gamma_q, std::inserter(cand_q, cand_q.end()));
            expand(subg_q, cand_q, g, upper_bound);
            if(upper_bound <= W_Q_max) return;
            cand.erase(q);
            //std::cerr << "back," << std::endl;
            Q.erase(q);
            W_Q = W_Q_pre;
            EXT_u.clear();
            unordered_set_difference(cand.begin(), cand.end(), gamma_u, std::inserter(EXT_u, EXT_u.end()));
         }
      }

   public:

      /// return the weighted maximal clique of a graph g
      const std::set<vertex> get_weighted_maximal_cliques(const Graph &g, int upper_bound)
      {
         Q.clear();
         Q_max.clear();
         W_Q=0;
         W_Q_max = std::numeric_limits<int>::min();
         std::tr1::unordered_set<vertex> subg;
         std::tr1::unordered_set<vertex> cand;
         vertex_iterator  vi,vi_end;
         boost::tie(vi,vi_end) = boost::vertices(g);
         subg.insert(vi,vi_end);
         cand.insert(vi,vi_end);
         expand(subg,cand, g, upper_bound);
         return Q_max;
      }

      /// return the last weight of the maximum clique
      int get_last_W_Q_max() {return W_Q_max;}
      TTT_maximal_weighted_clique(std::map<C_vertex, std::string> &_names) : names(_names) {}
};



template<typename vertex_type>
class TTT_based_clique_covering : public clique_covering<vertex_type>
{
      /// bulk undirected graph
      boost_cc_compatibility_graph clique_covering_graph_bulk;
      /// set of maximal clique computed
      std::vector<std::set<C_vertex> > cliques;
      /// map between vertex_type and C_vertex
      std::map<vertex_type, C_vertex> v2uv;
      /// map between C_vertex and vertex_type
      std::map<C_vertex, vertex_type> uv2v;
      /// name map for the C_vertex vertices
      std::map<C_vertex, std::string> names;
      /// edge selector
      static const int COMPATIBILITY_EDGE = 1;
   public:

      /// constructor
      TTT_based_clique_covering() {}

      /// add a vertex
      /// vertex weight not considered
      C_vertex add_vertex(const vertex_type& element, const std::string& name)
      {
         C_vertex result;
         THROW_ASSERT(v2uv.find(element) == v2uv.end(), "vertex already added");
         v2uv[element] = result = boost::add_vertex(clique_covering_graph_bulk);
         uv2v[result] = element;
         names[result] = name;
         return result;
      }

      /// add an edge
      void add_edge(const vertex_type& src, const vertex_type& dest, int _weight)
      {
         THROW_ASSERT(src != dest, "autoloops are not allowed in a compatibility graph");
         THROW_ASSERT(v2uv.find(src) != v2uv.end(), "src not added");
         THROW_ASSERT(v2uv.find(dest) != v2uv.end(), "dest not added");
         C_vertex SRC = v2uv.find(src)->second;
         C_vertex DEST = v2uv.find(dest)->second;
         boost::add_edge(SRC, DEST, edge_compatibility_selector(COMPATIBILITY_EDGE, _weight), clique_covering_graph_bulk);
      }

      /// return the number of vertices of the clique
      size_t num_vertices()
      {
         return cliques.size();
      }

      std::set<vertex_type> get_clique(unsigned int i)
      {
         std::set<vertex_type> result;
         std::set<C_vertex> & cur_clique = cliques[i];
         std::set<C_vertex>::const_iterator it_end = cur_clique.end();
         for(std::set<C_vertex>::const_iterator it = cur_clique.begin(); it != it_end; ++it)
         {
            THROW_ASSERT(uv2v.find(*it) != uv2v.end(), "vertex not added");
            result.insert(uv2v.find(*it)->second);
         }
         return result;
      }

      void exec()
      {
         TTT_maximal_weighted_clique<cc_compatibility_graph> MWC(names);
         int upper_bound = std::numeric_limits<int>::max();
         std::tr1::unordered_set<C_vertex> support;
         //std::cerr << "TTT based clique covering" << std::endl;
         typename std::map<vertex_type, C_vertex>::const_iterator it_end = v2uv.end();
         for(typename std::map<vertex_type, C_vertex>::const_iterator it = v2uv.begin(); it != it_end; ++it)
         {
            if(boost::out_degree(it->second, clique_covering_graph_bulk)==0)
            {
               //std::cerr << "trivial case: no outgoing edges for " << names[it->second] << std::endl;
               std::set<C_vertex> singularity;
               singularity.insert(it->second);
               cliques.push_back(singularity);
            }
            else
               support.insert(it->second);
         }
         cc_compatibility_graphRef CG = cc_compatibility_graphRef(new cc_compatibility_graph(clique_covering_graph_bulk,
                                                                                    cc_compatibility_graph_edge_selector<boost_cc_compatibility_graph>(COMPATIBILITY_EDGE, &clique_covering_graph_bulk),
                                                                                    cc_compatibility_graph_vertex_selector<boost_cc_compatibility_graph>(&support)));
         while(!support.empty())
         {
            //std::cerr << "Looking for a maximum weighted clique" << std::endl;
            cliques.push_back(MWC.get_weighted_maximal_cliques(*CG, upper_bound));
            /// update the upper bound with the last weight
            upper_bound = MWC.get_last_W_Q_max();
            //std::cerr << "Found one of size " << cliques.back().size() << std::endl;
            //std::set_difference(support.begin(), support.end(), cliques.back().begin(), cliques.back().end(), std::inserter(difference, difference.begin()));
            std::tr1::unordered_set<C_vertex>::iterator s_it_end = support.end();
            for(std::tr1::unordered_set<C_vertex>::iterator s_it=support.begin(); s_it != s_it_end;)
            {
               std::tr1::unordered_set<C_vertex>::iterator current = s_it++;
               if(cliques.back().find(*current) != cliques.back().end())
                  support.erase(current);
            }

            /// check trivial cases
            s_it_end = support.end();
            for(std::tr1::unordered_set<C_vertex>::iterator s_it=support.begin(); s_it != s_it_end;)
            {
               std::tr1::unordered_set<C_vertex>::iterator current = s_it++;
               if(boost::out_degree(*current, *CG)==0)
               {
                  {
                     //std::cerr << "trivial case: no outgoing edges for " << names[*s_it] << std::endl;
                     std::set<C_vertex> singularity;
                     singularity.insert(*current);
                     cliques.push_back(singularity);
                     support.erase(current);
                  }
               }
            }
         }
      }

      void writeDot(const std::string &filename) const
      {
         std::ofstream f(filename.c_str());
         boost::write_graphviz(f, clique_covering_graph_bulk, compatibility_node_info_writer(names), compatibility_edge_writer(clique_covering_graph_bulk));
      }

};

/**
 * fast version that just return the first maximal clique found
 */
template <typename Graph>
class TTT_maximal_weighted_clique_fast
{
      /// vertex iterator
      typedef typename boost::graph_traits<Graph>::vertex_iterator vertex_iterator;
      /// vertex object
      typedef typename boost::graph_traits<Graph>::vertex_descriptor vertex;
      /// adjacency iterator
      typedef typename boost::graph_traits<Graph>::adjacency_iterator adjacency_iterator;
      /// out edge iterator
      typedef typename boost::graph_traits<Graph>::out_edge_iterator edge_iterator;

      /// set of vertices of the current clique
      std::tr1::unordered_set<vertex> Q;
      /// set of vertices of the maximum clique found so far
      std::set<vertex> Q_max;
      /// weight of Q
      int W_Q;
      /// weight of Q_max
      int W_Q_max;

      std::map<C_vertex, std::string> &names;

      /// return the vertex of subg with the maximum intersection with cand
      vertex get_max_weighted_adiacent_intersection(const std::tr1::unordered_set<vertex> &subg, const std::tr1::unordered_set<vertex>&cand, const Graph &g)
      {
         vertex result=boost::graph_traits<Graph>::null_vertex();
         THROW_ASSERT(!subg.empty(), "at least one element should belong to subg");
         int max_weighted_intersection = -1;
         const typename std::tr1::unordered_set<vertex>::const_iterator it_end = subg.end();
         for(typename std::tr1::unordered_set<vertex>::const_iterator it = subg.begin(); it != it_end; ++it)
         {
            int weight_intersection = 0;
            edge_iterator ei,ei_end;
            boost::tie(ei,ei_end) = boost::out_edges(*it, g);
            for(;ei != ei_end; ++ei)
            {
               if(cand.find(boost::target(*ei, g)) != cand.end())
                  weight_intersection+=g[*ei].weight;
            }
            if(weight_intersection>max_weighted_intersection)
            {
               max_weighted_intersection = weight_intersection;
               result = *it;
            }
         }
         THROW_ASSERT(max_weighted_intersection>=0, "something of wrong happen");
         return result;
      }

      /// return the vertex of ext having the maximum degree with respect to the graph g
      vertex get_max_weight_vertex(std::tr1::unordered_set<vertex> & ext, const Graph &g)
      {
         vertex result=boost::graph_traits<Graph>::null_vertex();
         int max_weight = -1;
         THROW_ASSERT(!ext.empty(), "at least one element should belong to ext");
         const typename std::tr1::unordered_set<vertex>::const_iterator it_end = ext.end();
         for(typename std::tr1::unordered_set<vertex>::const_iterator it = ext.begin(); it != it_end; ++it)
         {
            int cur_weight=0;
            edge_iterator ei,ei_end;
            boost::tie(ei,ei_end) = boost::out_edges(*it, g);
            for(;ei != ei_end; ++ei)
            {
               cur_weight+=g[*ei].weight;
            }
            if(cur_weight>max_weight)
            {
               result = *it;
               max_weight = cur_weight;
            }
         }
         THROW_ASSERT(max_weight>=0, "something of wrong happen");
         return result;
      }

      /// compute the delta of the weight by adding q_vertex to the clique
      int compute_delta_weight(vertex q_vertex, std::tr1::unordered_set<vertex> Q_set, const Graph &g)
      {
         int result=0;
         edge_iterator ei,ei_end;
         boost::tie(ei,ei_end) = boost::out_edges(q_vertex, g);
         for(;ei != ei_end; ++ei)
            if(Q_set.find(boost::target(*ei, g)) != Q_set.end())
               result+=g[*ei].weight;
         return result;
      }

      /// recorsive procedure expand defined in first cited paper
      void expand(std::tr1::unordered_set<vertex> &subg, std::tr1::unordered_set<vertex>&cand, const Graph &g)
      {
         if(subg.empty() && Q.size() >= Q_max.size())
         {
            if(Q.size() > Q_max.size() || W_Q > W_Q_max)
            {
               Q_max.clear();
               Q_max.insert(Q.begin(), Q.end());
               W_Q_max = W_Q;
               //std::cerr << "clique W=" << W_Q_max << " size=" << Q_max.size() << std::endl;
            }
            return;
         }
         else if(cand.empty())
            return;

         /// get the vertex in subg with the maximum of adjacent vertices in cand
         vertex u =get_max_weighted_adiacent_intersection(subg, cand, g);
         //vertex u =get_max_adiacent_intersection(subg, cand, g);
         /// get adjacent vertices of u
         adjacency_iterator vi, vi_end;
         std::pair<adjacency_iterator, adjacency_iterator> p;
         p = boost::adjacent_vertices(u, g);
         vi = p.first;
         vi_end = p.second;
         /// set of vertices adjacent to u
         std::tr1::unordered_set<vertex> gamma_u(vi, vi_end);
         /// compute EXT_u = CAND - gamma_u
         std::tr1::unordered_set<vertex> EXT_u;
         unordered_set_difference(cand.begin(), cand.end(), gamma_u, std::inserter(EXT_u, EXT_u.end()));
         while(!EXT_u.empty())
         {
            vertex q = get_max_weight_vertex(EXT_u, g);
            //vertex q = get_max_degree_vertex(EXT_u, g);
            //std::cerr << names[q] << "," << std::endl;
            Q.insert(q);
            int W_Q_pre = W_Q;
            /// compute delta_weight
            int delta = compute_delta_weight(q, Q, g);
            W_Q += delta;
            boost::tie(vi, vi_end) = boost::adjacent_vertices(q, g);
            std::tr1::unordered_set<vertex> gamma_q(vi, vi_end);
            std::tr1::unordered_set<vertex> subg_q;
            unordered_set_intersection(subg.begin(), subg.end(), gamma_q, std::inserter(subg_q, subg_q.end()));
            std::tr1::unordered_set<vertex> cand_q;
            unordered_set_intersection(cand.begin(), cand.end(), gamma_q, std::inserter(cand_q, cand_q.end()));
            expand(subg_q, cand_q, g);
            if(W_Q_max>=0) return;
            cand.erase(q);
            //std::cerr << "back," << std::endl;
            Q.erase(q);
            W_Q = W_Q_pre;
            EXT_u.clear();
            unordered_set_difference(cand.begin(), cand.end(), gamma_u, std::inserter(EXT_u, EXT_u.end()));
         }
      }

   public:

      /// return the weighted maximal clique of a graph g
      const std::set<vertex> get_weighted_maximal_cliques(const Graph &g)
      {
         Q.clear();
         Q_max.clear();
         W_Q=0;
         W_Q_max = std::numeric_limits<int>::min();
         std::tr1::unordered_set<vertex> subg;
         std::tr1::unordered_set<vertex> cand;
         vertex_iterator  vi,vi_end;
         boost::tie(vi,vi_end) = boost::vertices(g);
         subg.insert(vi,vi_end);
         cand.insert(vi,vi_end);
         expand(subg,cand, g);
         return Q_max;
      }

      TTT_maximal_weighted_clique_fast(std::map<C_vertex, std::string> &_names) : names(_names) {}
};


/// fast version of the TTT_based_clique_covering. The first maximal clique found is used by the greedy clique covering.
template<typename vertex_type>
class TTT_based_clique_covering_fast : public clique_covering<vertex_type>
{
      /// bulk undirected graph
      boost_cc_compatibility_graph clique_covering_graph_bulk;
      /// set of maximal clique computed
      std::vector<std::set<C_vertex> > cliques;
      /// map between vertex_type and C_vertex
      std::map<vertex_type, C_vertex> v2uv;
      /// map between C_vertex and vertex_type
      std::map<C_vertex, vertex_type> uv2v;
      /// name map for the C_vertex vertices
      std::map<C_vertex, std::string> names;
      /// edge selector
      static const int COMPATIBILITY_EDGE = 1;
   public:

      /// constructor
      TTT_based_clique_covering_fast() {}

      /// add a vertex
      C_vertex add_vertex(const vertex_type& element, const std::string& name)
      {
         C_vertex result;
         THROW_ASSERT(v2uv.find(element) == v2uv.end(), "vertex already added");
         /// vertex weight not considered
         v2uv[element] = result = boost::add_vertex(clique_covering_graph_bulk);
         uv2v[result] = element;
         names[result] = name;
         return result;
      }

      /// add an edge
      void add_edge(const vertex_type& src, const vertex_type& dest, int _weight)
      {
         THROW_ASSERT(src != dest, "autoloops are not allowed in a compatibility graph");
         THROW_ASSERT(v2uv.find(src) != v2uv.end(), "src not added");
         THROW_ASSERT(v2uv.find(dest) != v2uv.end(), "dest not added");
         C_vertex SRC = v2uv.find(src)->second;
         C_vertex DEST = v2uv.find(dest)->second;
         boost::add_edge(SRC, DEST, edge_compatibility_selector(COMPATIBILITY_EDGE, _weight), clique_covering_graph_bulk);
      }

      /// return the number of vertices of the clique
      size_t num_vertices()
      {
         return cliques.size();
      }

      std::set<vertex_type> get_clique(unsigned int i)
      {
         std::set<vertex_type> result;
         std::set<C_vertex> & cur_clique = cliques[i];
         std::set<C_vertex>::const_iterator it_end = cur_clique.end();
         for(std::set<C_vertex>::const_iterator it = cur_clique.begin(); it != it_end; ++it)
         {
            THROW_ASSERT(uv2v.find(*it) != uv2v.end(), "vertex not added");
            result.insert(uv2v.find(*it)->second);
         }
         return result;
      }

      void exec()
      {
         TTT_maximal_weighted_clique_fast<cc_compatibility_graph> MWC(names);
         std::tr1::unordered_set<C_vertex> support;
         //std::cerr << "TTT based clique covering" << std::endl;
         typename std::map<vertex_type, C_vertex>::const_iterator it_end = v2uv.end();
         for(typename std::map<vertex_type, C_vertex>::const_iterator it = v2uv.begin(); it != it_end; ++it)
         {
            if(boost::out_degree(it->second, clique_covering_graph_bulk)==0)
            {
               //std::cerr << "trivial case: no outgoing edges for " << names[it->second] << std::endl;
               std::set<C_vertex> singularity;
               singularity.insert(it->second);
               cliques.push_back(singularity);
            }
            else
               support.insert(it->second);
         }
         cc_compatibility_graphRef CG = cc_compatibility_graphRef(new cc_compatibility_graph(clique_covering_graph_bulk,
                                                                                    cc_compatibility_graph_edge_selector<boost_cc_compatibility_graph>(COMPATIBILITY_EDGE, &clique_covering_graph_bulk),
                                                                                    cc_compatibility_graph_vertex_selector<boost_cc_compatibility_graph>(&support)));
         while(!support.empty())
         {
            //std::cerr << "Looking for a maximum weighted clique" << std::endl;
            cliques.push_back(MWC.get_weighted_maximal_cliques(*CG));
            //std::cerr << "Found one of size " << cliques.back().size() << std::endl;
            std::tr1::unordered_set<C_vertex>::iterator s_it_end = support.end();
            for(std::tr1::unordered_set<C_vertex>::iterator s_it=support.begin(); s_it != s_it_end;)
            {
               std::tr1::unordered_set<C_vertex>::iterator current = s_it++;
               if(cliques.back().find(*current) != cliques.back().end())
                  support.erase(current);
            }

            /// check trivial cases
            s_it_end = support.end();
            for(std::tr1::unordered_set<C_vertex>::iterator s_it=support.begin(); s_it != s_it_end;)
            {
               std::tr1::unordered_set<C_vertex>::iterator current = s_it++;
               if(boost::out_degree(*current, *CG)==0)
               {
                  {
                     //std::cerr << "trivial case: no outgoing edges for " << names[*s_it] << std::endl;
                     std::set<C_vertex> singularity;
                     singularity.insert(*current);
                     cliques.push_back(singularity);
                     support.erase(current);
                  }
               }
            }
         }
      }

      void writeDot(const std::string &filename) const
      {
         std::ofstream f(filename.c_str());
         boost::write_graphviz(f, clique_covering_graph_bulk, compatibility_node_info_writer(names), compatibility_edge_writer(clique_covering_graph_bulk));
      }

};

template<typename vertex_type>
class coloring_based_clique_covering : public clique_covering<vertex_type>
{
   /// bulk undirected graph
   boost_cc_compatibility_graph clique_covering_graph_bulk;
   /// set of maximal clique computed
   std::vector<std::set<C_vertex> > cliques;
   /// map between vertex_type and C_vertex
   std::map<vertex_type, C_vertex> v2uv;
   /// map between C_vertex and vertex_type
   std::map<C_vertex, vertex_type> uv2v;
   /// name map for the C_vertex vertices
   std::map<C_vertex, std::string> names;
   /// edge selector to select all edges of the compatibility graph
   static const int COMPATIBILITY_ALL_EDGES = ~0;
   int max_level;
   bool all_edges;

   typedef boost::property_map<boost_cc_compatibility_graph, boost::vertex_index_t>::type vertex_index_pmap_t;
   /// rank property map definition
   typedef boost::iterator_property_map<std::vector<boost::graph_traits<boost_cc_compatibility_graph>::vertices_size_type>::iterator, vertex_index_pmap_t> rank_pmap_type;
   /// parent property map definition
   typedef boost::iterator_property_map<std::vector<C_vertex>::iterator, vertex_index_pmap_t> pred_pmap_type;

public:
   /// constructor
   coloring_based_clique_covering(bool _all_edges) : max_level(0), all_edges(_all_edges) {}

   /// add a vertex
   C_vertex add_vertex(const vertex_type& element, const std::string& name)
   {
      C_vertex result;
      THROW_ASSERT(v2uv.find(element) == v2uv.end(), "vertex already added");
      /// vertex weight not considered
      v2uv[element] = result = boost::add_vertex(clique_covering_graph_bulk);
      uv2v[result] = element;
      names[result] = name;
      return result;
   }

   /// add an edge
   void add_edge(const vertex_type& src, const vertex_type& dest, int _weight)
   {
      THROW_ASSERT(src != dest, "autoloops are not allowed in a compatibility graph");
      THROW_ASSERT(v2uv.find(src) != v2uv.end(), "src not added");
      THROW_ASSERT(v2uv.find(dest) != v2uv.end(), "dest not added");
      THROW_ASSERT(_weight>0&& _weight<32, "weights from 1 to 31 are allowed");
      max_level = std::max(max_level, _weight);
      C_vertex SRC = v2uv.find(src)->second;
      C_vertex DEST = v2uv.find(dest)->second;
      boost::add_edge(SRC, DEST, edge_compatibility_selector(1 << _weight, _weight), clique_covering_graph_bulk);
   }

   /// return the number of vertices of the clique
   size_t num_vertices()
   {
      return cliques.size();
   }

   std::set<vertex_type> get_clique(unsigned int i)
   {
      //std::cerr << "solution: " << i << std::endl;
      std::set<vertex_type> result;
      std::set<C_vertex> & cur_clique = cliques[i];
      std::set<C_vertex>::const_iterator it_end = cur_clique.end();
      for(std::set<C_vertex>::const_iterator it = cur_clique.begin(); it != it_end; ++it)
      {
         THROW_ASSERT(uv2v.find(*it) != uv2v.end(), "vertex not added");
         //std::cerr << names[*it] << std::endl;
         result.insert(uv2v.find(*it)->second);
      }
      return result;
   }

   void color_the_conflict_graph(const cc_compatibility_graphRef filteredCG, boost::disjoint_sets<rank_pmap_type,pred_pmap_type> &ds)
   {
      typedef boost::adjacency_list<boost::vecS, boost::vecS, boost::undirectedS> conflict_graph;
      typedef boost::graph_traits<conflict_graph>::vertices_size_type cg_vertices_size_type;
      typedef boost::property_map<conflict_graph, boost::vertex_index_t>::const_type cg_vertex_index_map;
      boost::iterator_property_map < cg_vertices_size_type*, cg_vertex_index_map, cg_vertices_size_type, cg_vertices_size_type& > color;

      std::vector<cg_vertices_size_type> color_vec;
      /// conflict graph
      conflict_graph cg;

      typedef boost::graph_traits<cc_compatibility_graph>::vertex_descriptor vertex_descriptor_ug;
      typedef boost::graph_traits<conflict_graph>::vertex_descriptor vertex_descriptor_cg;
      std::map<vertex_descriptor_ug, vertex_descriptor_cg> vmap;
      std::vector<vertex_descriptor_ug> reverse_map;

      unsigned int vertex_index=0;
      BGL_FORALL_VERTICES(v, *filteredCG, cc_compatibility_graph)
      {
        vmap[v] = boost::add_vertex(cg);
        reverse_map.push_back(v);
        ++vertex_index;
      }
      color_vec.resize(vertex_index);
      color = boost::iterator_property_map < cg_vertices_size_type*, cg_vertex_index_map, cg_vertices_size_type, cg_vertices_size_type& >(&color_vec.front(), boost::get(boost::vertex_index, cg));

      BGL_FORALL_VERTICES(u, *filteredCG, cc_compatibility_graph)
      {
        std::vector<vertex_descriptor_ug> neighbors(boost::adjacent_vertices(u, *filteredCG).first, adjacent_vertices(u, *filteredCG).second);
        std::sort(neighbors.begin(), neighbors.end());
        BGL_FORALL_VERTICES(v, *filteredCG, cc_compatibility_graph) {
           if(u>v) continue; /// the graph is an undirected graph...
          // Might want to check for self-loops
          if (!std::binary_search(neighbors.begin(), neighbors.end(), v))
            boost::add_edge(vmap[u], vmap[v], cg);
        }
      }

      /// coloring based on DSATUR 2 heuristic
      cg_vertices_size_type num_colors = dsatur2_coloring(cg, color);
      std::vector<unsigned int>colors(num_colors);

      for(unsigned int i=0; i < num_colors; ++i)
      {
         colors[i] = std::numeric_limits<unsigned int>::max();
      }
      for(unsigned int i=0; i < vertex_index; ++i)
      {
         cg_vertices_size_type c= color_vec[i];
         if(colors[c] == std::numeric_limits<unsigned int>::max())
            colors[c] = i;
         else
         {
            vertex_descriptor_ug ug_vertex_i = reverse_map[i];
            vertex_descriptor_ug ug_vertex_c = reverse_map[colors[c]];
            ds.union_set(ug_vertex_i,ug_vertex_c);
         }
      }
   }

   /// build partitions
   void build_partitions(std::tr1::unordered_set<C_vertex> & support, boost::disjoint_sets<rank_pmap_type,pred_pmap_type> & ds,   std::map<C_vertex, std::set<C_vertex> > & current_partitions)
   {
      std::tr1::unordered_set<C_vertex>::iterator it_end = support.end();
      for(std::tr1::unordered_set<C_vertex>::iterator it=support.begin(); it != it_end;)
      {
         std::tr1::unordered_set<C_vertex>::iterator current = it++;
         C_vertex rep = ds.find_set(*current);
         C_vertex cur = *current;
         if(rep!= cur)
         {
            if(current_partitions.find(rep) == current_partitions.end())
            {
               std::set<C_vertex> singularity;
               singularity.insert(cur);
               current_partitions[rep] = singularity;
            }
            else
               current_partitions.find(rep)->second.insert(cur);
         }
         else
         {
            if(current_partitions.find(rep) == current_partitions.end())
            {
               std::set<C_vertex> null_set;
               current_partitions[rep] = null_set;
            }
         }
      }
   }

   void exec()
   {
      boost::graph_traits<boost_cc_compatibility_graph>::vertices_size_type n = boost::num_vertices(clique_covering_graph_bulk);
      std::vector<boost::graph_traits<boost_cc_compatibility_graph>::vertices_size_type> rank_map(n);
      std::vector<C_vertex> pred_map(n);
      vertex_index_pmap_t cindex_pmap = boost::get(boost::vertex_index_t(), clique_covering_graph_bulk);
      rank_pmap_type rank_pmap = boost::make_iterator_property_map(rank_map.begin(), cindex_pmap, rank_map[0]);
      pred_pmap_type pred_pmap = boost::make_iterator_property_map(pred_map.begin(), cindex_pmap, pred_map[0]);
      boost::disjoint_sets<rank_pmap_type,pred_pmap_type> ds(rank_pmap, pred_pmap);
      typedef boost::graph_traits<boost_cc_compatibility_graph>::vertex_iterator u_vertex_iterator;
      std::tr1::unordered_set<C_vertex> support;

      u_vertex_iterator ui, uiend;
      for (boost::tie(ui, uiend) = boost::vertices(clique_covering_graph_bulk); ui != uiend; ui++)
      {
         ds.make_set(*ui);
         support.insert(*ui);
      }
      cc_compatibility_graphRef completeCG = cc_compatibility_graphRef(new cc_compatibility_graph(clique_covering_graph_bulk,
                                                                                 cc_compatibility_graph_edge_selector<boost_cc_compatibility_graph>(COMPATIBILITY_ALL_EDGES, &clique_covering_graph_bulk),
                                                                                 cc_compatibility_graph_vertex_selector<boost_cc_compatibility_graph>()));

      std::map<C_vertex, std::set<C_vertex> > current_partitions;
      if(all_edges)
      {
         color_the_conflict_graph(completeCG, ds);
      }
      else
      {
         /// color the conflict graph projection
         int cur_level = max_level;
         int selector = 1 << cur_level;
         cc_compatibility_graphRef filteredCG = cc_compatibility_graphRef(new cc_compatibility_graph(clique_covering_graph_bulk,
                                                                                    cc_compatibility_graph_edge_selector<boost_cc_compatibility_graph>(selector, &clique_covering_graph_bulk),
                                                                                    cc_compatibility_graph_vertex_selector<boost_cc_compatibility_graph>(&support)));
         while(cur_level>0)
         {
            color_the_conflict_graph(filteredCG, ds);
            /// build the current partitions
            build_partitions(support, ds, current_partitions);
            /// remove non conformant edges
            bool restart;
            do
            {
               restart = false;
               std::map<C_vertex, std::set<C_vertex> >::iterator cp_it_end = current_partitions.end();
               for(std::map<C_vertex, std::set<C_vertex> >::iterator cp_it = current_partitions.begin(); cp_it != cp_it_end; ++cp_it)
               {
                  C_vertex rep = cp_it->first;
                  std::set<C_vertex> & current_cliques = cp_it->second;
                  std::set<C_vertex>::const_iterator c_it_end = current_cliques.end();
                  for(std::set<C_vertex>::iterator c_it=current_cliques.begin(); c_it != c_it_end;)
                  {
                     std::set<C_vertex>::iterator current = c_it++;
                     C_vertex cur = *current;
                     C_outEdgeIterator ei,ei_end;
                     /// remove edges given the current set of clique
                     for(boost::tie(ei,ei_end) = boost::out_edges(rep, *completeCG); ei != ei_end; ++ei)
                     {
                        C_vertex rep_target = boost::target(*ei, *completeCG);
                        if(rep_target != cur)
                        {
                           std::vector<C_vertex> neighbors(boost::adjacent_vertices(cur, *completeCG).first, adjacent_vertices(cur, *completeCG).second);
                           std::sort(neighbors.begin(), neighbors.end());
                           if (!std::binary_search(neighbors.begin(), neighbors.end(), rep_target))
                           {
                              (*completeCG)[*ei].selector=0;
                              //std::cerr << names[cur] << "|"<< names[rep] << " -0- " << names[rep_target] << std::endl;
                              c_it = current_cliques.begin();/// restart the pruning
                              restart = true; /// and then restart the whole partition analysis
                           }
                        }
                     }
                     for(boost::tie(ei,ei_end) = boost::out_edges(cur, *completeCG); ei != ei_end; ++ei)
                     {
                        C_vertex cur_target = boost::target(*ei, *completeCG);
                        if(cur_target != rep)
                        {
                           std::vector<C_vertex> neighbors(boost::adjacent_vertices(rep, *completeCG).first, adjacent_vertices(rep, *completeCG).second);
                           std::sort(neighbors.begin(), neighbors.end());
                           if (!std::binary_search(neighbors.begin(), neighbors.end(), cur_target))
                           {
                              (*completeCG)[*ei].selector=0;
                              //std::cerr << names[rep] << "|"<< names[cur] << " -1- " << names[cur_target] << std::endl;
                           }
                        }
                     }
                  }
               }
            } while(restart);
            current_partitions.clear();

            /// recompute the support
            /// by keeping only the representative of each partition
            std::tr1::unordered_set<C_vertex>::iterator it_end = support.end();
            for(std::tr1::unordered_set<C_vertex>::iterator it=support.begin(); it != it_end;)
            {
               std::tr1::unordered_set<C_vertex>::iterator current = it++;
               C_vertex rep = ds.find_set(*current);
               C_vertex cur = *current;
               if(rep != cur || boost::out_degree(rep,*completeCG) == 0)
               {
                  //std::cerr << "remove vertex " << names[cur] << std::endl;
                  support.erase(current);
               }
            }
            /// rebuild the graph starting from which the coloring will be performed
            --cur_level;
            selector = selector | 1 << cur_level;
            //std::cerr << "new selector " << selector << std::endl;
            filteredCG = cc_compatibility_graphRef(new cc_compatibility_graph(clique_covering_graph_bulk,
                                                                                       cc_compatibility_graph_edge_selector<boost_cc_compatibility_graph>(selector, &clique_covering_graph_bulk),
                                                                                       cc_compatibility_graph_vertex_selector<boost_cc_compatibility_graph>(&support)));
         }
         /// rebuild the partitions
         boost::tie(ui, uiend) = boost::vertices(clique_covering_graph_bulk);
         support.clear();
         support.insert(ui,uiend);
      }

      build_partitions(support, ds, current_partitions);
      /// and then fill the cliques
      std::map<C_vertex, std::set<C_vertex> >::const_iterator cp_it_end = current_partitions.end();
      for(std::map<C_vertex, std::set<C_vertex> >::const_iterator cp_it = current_partitions.begin(); cp_it != cp_it_end; ++cp_it)
      {
         cliques.push_back(cp_it->second);
         cliques.back().insert(cp_it->first);
      }
   }

   void writeDot(const std::string &filename) const
   {
      std::ofstream f(filename.c_str());
      boost::write_graphviz(f, clique_covering_graph_bulk, compatibility_node_info_writer(names), compatibility_edge_writer(clique_covering_graph_bulk));
   }
};

/// functor cost where the cost just depend on x
struct y_independent_cost : BM_cost_functor
{
   public:
      y_independent_cost(const std::vector<int> & _cost) : cost(_cost) {}

      int operator() (size_t x, size_t /*notused*/) const
      {
         return cost[x];
      }
      int max_row(size_t x) const
      {
         return cost[x];
      }

   private:
      const std::vector<int> & cost;
};

struct full_cost_matrix : BM_cost_functor
{

   public:
      full_cost_matrix(const boost::numeric::ublas::matrix<int> & _cost) : cost(_cost) {}

      int operator() (size_t x, size_t y) const
      {
         return cost(x,y);
      }
      int max_row(size_t x) const
      {
         int res = cost(x,0);
         for (unsigned j = 1; j < cost.size2(); ++j)
            res = std::max(res, cost(x,j));
         return res;
      }

   private:
      const boost::numeric::ublas::matrix<int> &cost;

};

/// clique covering exploiting based on bipartite matching procedure
template<typename vertex_type>
class bipartite_matching_clique_covering : public clique_covering<vertex_type>
{
   private:
      /// bulk undirected graph
      boost_cc_compatibility_graph clique_covering_graph_bulk;
      /// set of maximal clique computed
      std::vector<std::tr1::unordered_set<C_vertex> > cliques;
      /// map between vertex_type and C_vertex
      std::map<vertex_type, C_vertex> v2uv;
      /// map between C_vertex and vertex_type
      std::map<C_vertex, vertex_type> uv2v;
      /// name map for the C_vertex vertices
      std::map<C_vertex, std::string> names;
      /// maximum weight
      int max_weight;
      /// edge selector
      static const int COMPATIBILITY_EDGE = 1;
      static const int BIG_NUMBER = 10000000;

      /// partition the set of vertex in set where each element is in conflict with the others
      /// the maximum size of these sets is the number of the cliques used to cover the graph
      size_t color_the_cc_compatibility_graph(const boost_cc_compatibility_graph & CG, std::map<size_t, std::set<boost::graph_traits<boost_cc_compatibility_graph>::vertex_descriptor> > &partitions)
      {
         typedef boost::graph_traits<boost_cc_compatibility_graph>::vertices_size_type cg_vertices_size_type;
         typedef boost::property_map<boost_cc_compatibility_graph, boost::vertex_index_t>::const_type cg_vertex_index_map;
         boost::iterator_property_map < cg_vertices_size_type*, cg_vertex_index_map, cg_vertices_size_type, cg_vertices_size_type& > color;

         std::vector<cg_vertices_size_type> color_vec;

         size_t n_vertices=boost::num_vertices(CG);
         color_vec.resize(n_vertices);
         color = boost::iterator_property_map < cg_vertices_size_type*, cg_vertex_index_map, cg_vertices_size_type, cg_vertices_size_type& >(&color_vec.front(), boost::get(boost::vertex_index, CG));

         /// coloring based on DSATUR 2 heuristic
         dsatur2_coloring(CG, color);
         size_t max_size=1;
         boost::graph_traits<boost_cc_compatibility_graph>::vertex_iterator ui, uiend;
         for (boost::tie(ui, uiend) = boost::vertices(clique_covering_graph_bulk); ui != uiend; ++ui)
         {
            cg_vertices_size_type c= color[*ui];
            std::map<size_t, std::set<boost::graph_traits<boost_cc_compatibility_graph>::vertex_descriptor> >::iterator p_it = partitions.find(c);
            if(p_it == partitions.end())
            {
               std::set<boost::graph_traits<boost_cc_compatibility_graph>::vertex_descriptor> singularity;
               singularity.insert(*ui);
               partitions[c]=singularity;
            }
            else
            {
               p_it->second.insert(*ui);
               max_size=std::max(max_size, p_it->second.size());
            }
         }
         return max_size;
      }

   public:

      /// constructor
      bipartite_matching_clique_covering() : max_weight(std::numeric_limits<int>::min()) {}

      /// add a vertex
      C_vertex add_vertex(const vertex_type& element, const std::string& name)
      {
         C_vertex result;
         THROW_ASSERT(v2uv.find(element) == v2uv.end(), "vertex already added");
         /// vertex weight not considered
         v2uv[element] = result = boost::add_vertex(clique_covering_graph_bulk);
         uv2v[result] = element;
         names[result] = name;
         return result;
      }

      /// add an edge
      void add_edge(const vertex_type& src, const vertex_type& dest, int _weight)
      {
         THROW_ASSERT(src != dest, "autoloops are not allowed in a compatibility graph");
         THROW_ASSERT(v2uv.find(src) != v2uv.end(), "src not added");
         THROW_ASSERT(v2uv.find(dest) != v2uv.end(), "dest not added");
         THROW_ASSERT(_weight>0, "weights greater than 0 are allowed");
         C_vertex SRC = v2uv.find(src)->second;
         C_vertex DEST = v2uv.find(dest)->second;
         max_weight = std::max(max_weight, _weight);
         boost::add_edge(SRC, DEST, edge_compatibility_selector(COMPATIBILITY_EDGE, _weight), clique_covering_graph_bulk);
      }

      /// return the number of vertices of the clique
      size_t num_vertices()
      {
         return cliques.size();
      }

      std::set<vertex_type> get_clique(unsigned int i)
      {
         std::set<vertex_type> result;
         std::tr1::unordered_set<C_vertex> & cur_clique = cliques[i];
         std::tr1::unordered_set<C_vertex>::const_iterator it_end = cur_clique.end();
         for(std::tr1::unordered_set<C_vertex>::const_iterator it = cur_clique.begin(); it != it_end; ++it)
         {
            THROW_ASSERT(uv2v.find(*it) != uv2v.end(), "vertex not added");
            result.insert(uv2v.find(*it)->second);
         }
         return result;
      }

      void extend_and_assign(boost::numeric::ublas::matrix<int> &cost_matrix, size_t &final_num_cols, size_t num_rows)
      {
         cost_matrix.resize(final_num_cols+1, final_num_cols+1, true);
         for(unsigned int i=0; i <= final_num_cols; ++i)
            cost_matrix(i,final_num_cols) = 0;
         for(unsigned int i=0; i < final_num_cols; ++i)
            cost_matrix(final_num_cols,i) = 0;
         cost_matrix(num_rows,final_num_cols) = BIG_NUMBER;
         std::tr1::unordered_set<C_vertex> empty;
         cliques.push_back(empty);
         ++final_num_cols;
      }

      void exec()
      {
         std::map<size_t, std::set<boost::graph_traits<boost_cc_compatibility_graph>::vertex_descriptor> > partitions;
         /// now color the graph and then do the bipartite matching on the vertex having the same color
         size_t num_cols =color_the_cc_compatibility_graph(clique_covering_graph_bulk, partitions);
         //std::cerr << "initial num_cols " << num_cols << std::endl;

         size_t final_num_cols = num_cols;

         /// initializes the cliques with empty sets
         for(unsigned int i = 0; i < num_cols; ++i)
         {
            std::tr1::unordered_set<C_vertex> empty;
            cliques.push_back(empty);
         }
         size_t offset = partitions.size();
         //std::cerr << "offset " << offset <<std::endl;
         boost::numeric::ublas::matrix<int> cost_matrix(num_cols,num_cols);
         /// compute the assignment for each element of a partition
         std::map<size_t, std::set<boost::graph_traits<boost_cc_compatibility_graph>::vertex_descriptor> >::reverse_iterator p_it_end = partitions.rend();
         for(std::map<size_t, std::set<boost::graph_traits<boost_cc_compatibility_graph>::vertex_descriptor> >::reverse_iterator p_it = partitions.rbegin(); p_it != p_it_end; ++p_it)
         {
            boost::numeric::ublas::noalias(cost_matrix)=boost::numeric::ublas::zero_matrix<int>(num_cols,num_cols);
            std::tr1::unordered_set<unsigned int> assigned;
            unsigned int num_rows=0;
            std::set<boost::graph_traits<boost_cc_compatibility_graph>::vertex_descriptor>::const_iterator v_it_end = p_it->second.end();
            for(std::set<boost::graph_traits<boost_cc_compatibility_graph>::vertex_descriptor>::const_iterator v_it = p_it->second.begin(); v_it != v_it_end; ++v_it, ++num_rows)
            {
               boost::graph_traits<boost_cc_compatibility_graph>::out_edge_iterator ei,eibegin,ei_end;
               boost::tie(ei,ei_end) = boost::out_edges(*v_it, clique_covering_graph_bulk);
               eibegin =ei;
               bool compatible_exist = false;
               for(unsigned int y = 0; y < num_cols; ++y)
               {
                  size_t compatibles = 0;
                  int acc_cost_y = 0;
                  for(ei=eibegin; ei != ei_end; ++ei)
                     if(cliques[y].find(boost::target(*ei, clique_covering_graph_bulk)) != cliques[y].end())
                     {
                        ++compatibles;
                        acc_cost_y += (clique_covering_graph_bulk)[*ei].weight;
                     }
                  if(compatibles == cliques[y].size() && assigned.find(y) == assigned.end())
                  {
                     if(compatibles == 0)
                     {
                        cost_matrix(num_rows,y) = BIG_NUMBER;
                        if(!compatible_exist)
                        {
                           /// in case the vertex is compatible with no one we reserve a slot for it
                           assigned.insert(y);
                           compatible_exist = true;
                           //std::cerr << "cost_matrix(num_rows,y) " << cost_matrix(num_rows,y) << " nr " << names[*v_it] << " y " << y << " size " << cliques[y].size() << std::endl;
                           break;
                        }
                     }
                     else
                     {
                        int curr_cost = (max_weight*acc_cost_y)/static_cast<int>(compatibles)+max_weight*static_cast<int>(offset-compatibles);
                        THROW_ASSERT(curr_cost>0, "cost has to be positive");
                        cost_matrix(num_rows,y) = curr_cost;
                        compatible_exist = true;
                     }
                     //std::cerr << "cost_matrix(num_rows,y) " << cost_matrix(num_rows,y) << " nr " << names[*v_it] << " y " << y << " size " << cliques[y].size() << std::endl;
                  }
                  else
                     cost_matrix(num_rows,y) = 0;
               }
               if(!compatible_exist)
               {
                  extend_and_assign(cost_matrix, final_num_cols, num_rows);
                  //std::cerr << "cost_matrix(num_rows,y) " << cost_matrix(num_rows,final_num_cols-1) << " nr " << names[*v_it] << " y " << final_num_cols-1<< std::endl;
               }
            }
            num_cols = final_num_cols;
            //std::cerr << "num_rows " << num_rows << std::endl;
            //std::cerr << "num_cols " << num_cols << std::endl;
            THROW_ASSERT(num_rows<=num_cols, "something of unexpected happen");
            bool restart_bipartite;
            do
            {
               restart_bipartite = false;
               full_cost_matrix fcm(cost_matrix);
               bipartite_matching<full_cost_matrix> WBM(num_rows, num_cols, fcm);

               WBM.solve_bipartite_matching();
               std::vector<size_t> &sol = WBM.get_solution();
               size_t analyzed_rows = 0;
               std::set<boost::graph_traits<boost_cc_compatibility_graph>::vertex_descriptor>::const_iterator v_it = p_it->second.begin();
               for(unsigned int i = 0; i < num_rows; ++i, ++v_it)
               {
                  size_t s = sol[i];
                  //std::cerr << names[*v_it] << " assigned to " << s << std::endl;
                  /// consistency check
                  boost::graph_traits<boost_cc_compatibility_graph>::out_edge_iterator ei,ei_end;
                  boost::tie(ei,ei_end) = boost::out_edges(*v_it, clique_covering_graph_bulk);
                  size_t compatibles = 0;
                  for(;ei != ei_end; ++ei)
                     if(cliques[s].find(boost::target(*ei, clique_covering_graph_bulk)) != cliques[s].end())
                        ++compatibles;
                  if(compatibles != cliques[s].size())
                  {
                     //std::cerr << "Wrong result from bipartite matching for " + names[*v_it] + " assigned to " + STR(s) << std::endl;
                     restart_bipartite = true;
                     extend_and_assign(cost_matrix, final_num_cols, i);
                     num_cols = final_num_cols;
                     analyzed_rows = i;
                     //std::cerr << "cost_matrix(num_rows,y) " << cost_matrix(i,final_num_cols-1) << " nr " << names[*v_it] << " y " << final_num_cols-1<< std::endl;
                     break;
                  }
                  cliques[s].insert(*v_it);
               }
               if(restart_bipartite)
               {
                  /// undo the last step
                  v_it = p_it->second.begin();
                  for(unsigned int i = 0; i < analyzed_rows; ++i, ++v_it)
                  {
                     size_t s = sol[i];
                     cliques[s].erase(*v_it);
                  }
               }
            } while(restart_bipartite);
         }
      }

      void writeDot(const std::string &filename) const
      {
         std::ofstream f(filename.c_str());
         boost::write_graphviz(f, clique_covering_graph_bulk, compatibility_node_info_writer(names), compatibility_edge_writer(clique_covering_graph_bulk));
      }

};


template<typename VertexType>
refcount< clique_covering< VertexType> > clique_covering<VertexType>::create_solver(algorithm_type solver)
{
   switch (solver)
   {
      case COLORING:
         return refcount< clique_covering<VertexType> > (new coloring_based_clique_covering<VertexType>(true));
      case WEIGHTED_COLORING:
         return refcount< clique_covering<VertexType> > (new coloring_based_clique_covering<VertexType>(false));
      case TTT_CLIQUE_COVERING:
         return refcount< clique_covering<VertexType> > (new TTT_based_clique_covering<VertexType>());
      case TTT_CLIQUE_COVERING_FAST:
         return refcount< clique_covering<VertexType> > (new TTT_based_clique_covering_fast<VertexType>());
      case BIPARTITE_MATCHING:
         return refcount< clique_covering<VertexType> > (new bipartite_matching_clique_covering<VertexType>());
      default:
         THROW_ERROR("This clique covering algorithm has not been implemented");
   }
   return refcount< clique_covering< VertexType> >();
}

#endif // TTT_BASED_CLIQUE_COVERING_HPP
