/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file weak_dominance.hpp
 * @brief Class specifying weak dominance calculus.
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
 */

///Header include
#include "weak_dominance.hpp"

///Dominance include
#include "Dominance.hpp"

///Utility include
#include "utility.hpp"

void weak_dominance::calculate_weak_dominance_info(graphs_collection * output, std::tr1::unordered_map<vertex, vertex> & i2o, std::tr1::unordered_map<vertex, vertex> & o2i)
{
   if(boost::num_vertices(*input) == 0)
      return;
   //creating maps
   THROW_ASSERT(i2o.size() == o2i.size(), "Different map sizes");
   if(i2o.size() == 0)
   {
      graph::vertex_iterator v, v_end;
      for(boost::tie(v, v_end) = boost::vertices(*input); v != v_end; v++)
      {
         vertex new_vertex = output->AddVertex();
         i2o[*v] = new_vertex;
         o2i[new_vertex] = *v;
      }
   }

   std::list<vertex> levels;
   boost::topological_sort(*input, std::front_inserter(levels));
   std::tr1::unordered_map<vertex, unsigned int> sorted_nodes;
   unsigned int counter = 0;
   for(std::list<vertex>::iterator it = levels.begin(); it != levels.end(); it++)
      sorted_nodes[*it] = ++counter;

   dominance<graph> dm = dominance<graph>(*input, start, end, param);
   dm.calculate_dominance_info(dominance<graph>::CDI_POST_DOMINATORS);
   std::tr1::unordered_map<vertex, vertex> post_dominators = dm.get_dominator_map();

   //iterate over outgoing edges of the input graph
   EdgeIterator ei, ei_end;
   for (boost::tie(ei, ei_end) = boost::edges(*input); ei != ei_end ; ++ei )
   {
      vertex A = boost::source(*ei, *input);
      vertex B = boost::target(*ei, *input);
      InEdgeIterator pd_ei, pd_ei_end;
      {
         vertex current_node = B;
         while (current_node && current_node != A && current_node != post_dominators[A])
         {
            if(sorted_nodes[current_node]>sorted_nodes[A])
               add_edge(i2o[A], i2o[current_node], output);
            else
               break;
            current_node = post_dominators[current_node];
         }
      }
   }
}

weak_dominance::weak_dominance(const graph * _input, vertex _start, vertex _end, const ParameterConstRef _param, int _selector) :
      input(_input),
      start(_start),
      end(_end),
      selector(_selector),
      param(_param),
      debug_level(_param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE))
{}

void weak_dominance::add_edge(vertex source, vertex target, graphs_collection * output)
{
   EdgeDescriptor e;
   bool inserted;
   boost::tie(e, inserted) = boost::edge(source, target, *output);
   if(inserted)
      output->set_selector(e, output->get_selector(e) | selector);
   else
      boost::add_edge(source, target, EdgeProperty(selector), *output);
}
