/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file loop.hpp
 * @brief interface of a loop
 *
 * A loop is a strongly connected component in a CFG. We use Tarjan algorithm
 * to build a loop forest (see loops.h).
 * Loop represents a single loop. A loop is composed by a set of nodes and arcs
 * which belong to the CFG. Some of the nodes in a loop are special and have
 * been given a name. Some nodes outside the loop are also somehow related to
 * loops and have a special name.
 * Here we introduce the naming conventions we will be using.
 *
 *               -----------------------
 *               |      PreHeader      |
 *               -----------------------
 *                          |
 *          |-------------->|
 *          |               V
 *          |    -----------------------
 *          |    |       Header        |
 *          |    -----------------------
 *          |               |
 *          |               |
 *          |               V
 *          |    -----------------------
 *          |    |    Any CFG allowed  |
 *          |    -----------------------
 *          |               |
 *          |               |
 *          |               V
 *          |    -----------------------
 *          |    |        Exit         |
 *          |    -----------------------
 *          |               |
 *          ----------------|
 *                          V
 *               -----------------------
 *               |     Landing pad     |
 *               -----------------------
 *
 * This example shows an example of a loop, but it is not the most general case.
 * The example has the following properties:
 *
 * 1) it is bottom tested (the exit condition is evaulated at the end of the loop,
 *    which means the loop is do-while)
 * 2) it is reducible (single entry point)
 * 3) has a single pre header (for reducible loops it is always possible to change the
 *    CFG to meet this condition)
 * 4) has a single exit
 * 5) has a single landing pad
 * 6) has a single back edge (for a definition of back edge see for example Tarjan)
 *
 * In general a loop can be much more complicated, having multiple entries (in C this can
 * be achieved using gotos), mutiple exits, landing pads and back edges.
 *
 * In our Loop representation a loop has a header (and only one!) if and only if the loop is
 * reducible. If the loop is not reducible then there's no header info. Exits are kept in a
 * list and landing pads as well. An interesting case is when we have N landing pads and N-1 landing
 * pads have a single successor being the Nth landing pad. This case is generate by break statements
 * in C/C++. For this reason a Loop has a special property to signal this case.
 *
 * @author Marco Garatti <m.garatti@gmail.com>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

#ifndef LOOP_HPP
#define LOOP_HPP

///Boost include
#include "graph.hpp"

///STL include
#include <list>
#include <set>
#include "tr1/unordered_set"

/**
 * @name Constants identifying the type of the loops
 */
//@{
///loop with single exit
#define SINGLE_EXIT_LOOP 1

///unknown loop
#define UNKNOWN_LOOP 2

///while or for loop
#define WHILE_LOOP 4

///for loop
#define FOR_LOOP 8

///parallelizable for loop
#define OMP_FOR_LOOP 16

///do while loop
#define DO_WHILE_LOOP 32

///countable loop
#define COUNTABLE_LOOP 64

//@}

class bb_graph;
CONSTREF_FORWARD_DECL(Loop);
REF_FORWARD_DECL(Loop);
REF_FORWARD_DECL(Loops);

class Loop 
{
   private:
      ///Friend definition of loop
      friend class Loops;

      /**
       * ???
       */
      void ComputeLandingPadExits();

      ///The basic block control flow graph
      const bb_graph* g;

      /// tells if there aren't loop nested in this
      bool is_innermost_loop;

      ///Parent loop
      Wrefcount<Loop> parent_loop;

      ///Child loops
      std::set<LoopConstRef> children;

      ///Blocks which belong to this loop
      std::tr1::unordered_set<vertex> blocks;

      /// exit blocks for this loop
      std::list<vertex> exits;

      /// landing_pads for this loop
      std::tr1::unordered_set<vertex> landing_pads;

      ///???
      vertex primary_landing_pad_block;

      /// the header of the loop
      vertex header_block;

      /// the id of the loop
      unsigned int loop_id;

      ///Map storing the association between an exit basic block and the corresponding landing pads
      std::map<vertex, std::tr1::unordered_set<vertex> > exit_landing_association;

   public:
      ///Nesting depth of this loop
      unsigned int depth;

      /// loop type
      int loop_type;

      ///loop memory footprint
      long long footprint_size;

      ///loop instruction footprint
      long long instruction_size;

      /// body start basic block (at the moment defined only for SIMPLE_LOOP & WHILE_LOOP)
      vertex body_start;

      ///the main induction variable of countable loop;
      unsigned int main_iv;

      ///Increment of induction variable
      long long int increment;

      ///Initial value of induction variable
      long long int lower_bound;

      ///Final value of induction variable
      long long int upper_bound;

      ///flag for induction variable close interval
      bool close_interval;

      /**
       * Constructor for empty loop (used for irreducible)
       * @param g is the basic block control flow grah
       */
      Loop(const bb_graph* g);

      /**
       * Constructor for reducible loop
       * @param g is the basic block control flow graph
       * @param header is the header basic block
       */
      Loop(const bb_graph* g, vertex header);

      /**
       * returns the loop id
       * @return the loop id
      */
      unsigned int GetId() const;

      /**
       * tells if the loop is innermost
       * @return true if the loop is innermost
       */
      bool is_innermost() const;

      /**
       * tells if the loop is reducible
       * @return true if the loop is reducible
       */
      bool IsReducible() const;

      /**
       * returns loop header
       * @return returns loop header
       */
      vertex GetHeader() const;

      /**
       * returns the parent loop
       * @return the parent loop if this loop is nested, NULL otherwise
       */
      const LoopRef Parent() const;

      /**
       * adds a block to this loop
       * @param block is the basic block to be added
       */
      void add_block(vertex block);

      /**
       * returns the number of basic blocks belonging to this loop
       * @return the number of basic blocks
       */
      size_t num_blocks() const;

      /**
       * returns the blocks   
       * @return the blocks
       */
      const std::tr1::unordered_set<vertex> & get_blocks() const;

      // Info on the exit blocks
      size_t num_exits() const;
      std::list<vertex>::const_iterator exit_block_iter_begin() const;
      std::list<vertex>::const_iterator exit_block_iter_end() const;

      // Info on the landing pads
      size_t num_landing_pads() const;

      /**
       * Return the landing pads of the loops
       * @param the basic block landing pads of the loop
      */
      const std::tr1::unordered_set<vertex> GetLandingPadBlocks() const;

      // Return the primary landing pad if it exists, NULL otherwise
      // The primary landing pad is the one that is the unique successor
      // of all the other LPs
      vertex primary_landing_pad() const;

      /**
       * Sets parent for this loop
       * @param parent is the parent loop
       */
      void SetParent(LoopRef parent);

      /**
       * Adds a child loop
       * @param child is the child loop
       */
      void AddChild(LoopRef child);

      /**
       * Returns the children of this loop in the loop forest
       */
      const std::set<LoopConstRef> & GetChildren() const;

      /**
       * Returns the basic blocks which belong to this loop and to loop nested in this loop
       * @param ret is the returned set of basic block of this loop and of its children
       */
      void get_recursively_bb(std::tr1::unordered_set<vertex> & ret) const;

      void get_recursively_op(std::set<vertex> & ret) const;

      /**
       * Returns the map exit_landing_association
       * @return the map exit_landing_association
       */
      const std::map<vertex, std::tr1::unordered_set<vertex> > & get_exit_landing_association() const;

      ///Definition of friend class add_loop_nop
      friend class add_loop_nop;
};
///refcount definition of the class
typedef refcount<Loop> LoopRef;

#endif // LOOP_HPP

