/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file loops.hpp
 * @brief interface of loops finding algorithm
 *
 * Loops are detected via Shreedar, Gao and Lee algorithm, described
 * in ?? ref to paper ??
 *
 * @author Marco Garatti <m.garatti@gmail.com>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/

#ifndef LOOPS_HPP
#define LOOPS_HPP

///Autoheader include
#include "config_HAVE_HOST_PROFILING_BUILT.hpp"

#include "refcount.hpp"
#include "graph.hpp"

#include <set>
#include <list>
#include <tr1/unordered_map>

/**
 * @name forward declarations
 */
//@{
REF_FORWARD_DECL(function_behavior);
CONSTREF_FORWARD_DECL(Loop);
REF_FORWARD_DECL(Loop);
CONSTREF_FORWARD_DECL(Parameter);
CONSTREF_FORWARD_DECL(ProfilingInformation);
class graph;
template <typename Graph> class dominance;
class bb_graph;
//@}

// This class represents a loop forest for a given method.
// The loop forest is built by Tarjan algorithm and each loop
// is represented by a loop instance (see loop.h).
//
// The class allows a client to iterate on the loops in a method
// and provides a map from basic block to loop to know if a block
// belongs to a loop.
//
// If necessary it is possible to extend the class to navigate the
// hierarchy in an ordered way.
//
class Loops
{
   private:

      ///The function behavior
      const Wrefcount<function_behavior> FB;

      ///class containing all the parameters
      const ParameterConstRef Param;

      ///Debug level
      int debug_level;

      ///flag for dot writing
      int write_dot;

      ///Maps between basic block and loop to which it belongs
      std::tr1::unordered_map<vertex, LoopRef> block_to_loop;

      ///List of found loops
      std::list<LoopRef> modifiable_loops_list;
      std::list<LoopConstRef> const_loops_list;

      typedef std::pair<vertex, vertex> vertex_pair;

      Loops();

      /**
       * Actually computes the loops of the control flow graph
       */
      void DetectLoops();

      /**
       * ???
       * @param djg is ???
       * @param visited is ???
       * @param loop is ???
       * @param node is ???
       * @param header is ???
       */
      void DetectReducibleLoop(const bb_graph* djg, std::set<vertex> &visited, LoopRef loop, vertex node, vertex header);

      bool is_edge_in_list(std::list<vertex_pair>& l, vertex source, vertex target);

      void DetectIrreducibleLoop(const bb_graph* djg, int min_level, std::tr1::unordered_map<vertex, int>& vertex_level);

      void tarjan_scc(const bb_graph* djg, vertex v,
                      std::tr1::unordered_map<vertex, unsigned int>& dfs_order,
                      std::tr1::unordered_map<vertex, unsigned int>& lowlink,
                      std::list<vertex>& s,
                      std::set<vertex> u,
                      unsigned int& max_dfs);


      bool stack_contains(std::list<vertex> stack, vertex v);

      /**
       * Creates Loop zero data structure
       */
      void BuildZeroLoop();

      /**
       * Sets depth for each loop in the forest starting from loop
       * @param loop is the root of the loop forest
       */
      void computeDepth(const LoopConstRef loop);

   public:

      /**
       * The constructor builds the loop forest for the given method
       * @param FB is the function behavior of the control flow graph to be analyzed
       * @param parameter is the set of input parameters
       */
      Loops(const function_behaviorRef FB, const ParameterConstRef parameter);

      /**
       * Returns the number of loops
       */
      size_t NumLoops() const;

      /**
       * Returns the list of loops (const)
       * @param return the const list of loops
       */
      const std::list<LoopConstRef>& GetList() const;

      /**
       * Return the list of loops 
       * @param return the const list of loops
       */
      const std::list<LoopRef>& GetModifiableList() const;

      /**
       * Returns a loop given the id 
       * @param id is the id of the loop
       */
      const LoopConstRef GetLoop(unsigned int id) const;

      /**
       * Returns a loop given the id 
       * @param id is the id of the loop
       */
      const LoopRef GetModifiableLoop(unsigned int id) const;

      /**
       * Write dot files representing the loop forest
       * @param file_name is the file name to be produced
       * @param profiling_information is the profiling information used to print data about loop iterations
       */
      void WriteDot(const std::string & file_name
#if HAVE_HOST_PROFILING_BUILT
         ,const ProfilingInformationConstRef profiling_information = ProfilingInformationConstRef()
#endif
      ) const;
};

#endif // LOOPS_HPP
