/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file c_backend.cpp
 * @brief Simple class used to drive the backend in order to be able to print c source code
 *
 * @author Luca Fossati <fossati@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
///Header include
#include  "c_backend.hpp"

///Backend include
#include "c_writer.hpp"
#include "var_pp_functor.hpp"

///Behavior include
#include "application_manager.hpp"
#include "behavioral_helper.hpp"
#include "call_graph.hpp"
#include "function_behavior.hpp"
#include "prettyPrintVertex.hpp"

///Graph include
#include "graph.hpp"

///Paramter include
#include "Parameter.hpp"

///STD include
#include <fstream>
#include <iostream>
#include <ostream>
#include <sstream>
#include <string>

///STL include
#include <deque>
#include <list>
#include <map>
#include <set>
#include <vector>

///Tree include
#include "tree_helper.hpp"

///Utility include
#include <boost/config.hpp>
#include <boost/lexical_cast.hpp>
#include "boost/filesystem/operations.hpp"
#include "boost/filesystem/path.hpp"
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/topological_sort.hpp>
#include <boost/graph/filtered_graph.hpp>
#include <boost/graph/graph_utility.hpp>
#include "refcount.hpp"

CBackend::CBackend(const application_managerConstRef _AppM, const ParameterConstRef _Param) :
   Param(_Param),
   debug_level(_Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE)),
   AppM(_AppM),
   TM(_AppM->get_tree_manager())
{
}

CBackend::~CBackend()
{
}

void CBackend::registerWriter(const CWriterRef writer)
{
   writers.push_back(writer);
}

void CBackend::write(std::string fileName)
{
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->CBackend: writing " + fileName);
   writeHeader();

   writeIncludes();

   WriteGlobalDeclarations();

   writeImplementations();
   for(unsigned int i = 0; i < writers.size(); i++)
   {
      writers[i]->writeFile(fileName);
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--Cbackend: writing " + fileName);
}

void CBackend::writeHeader()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->C backend: writing header");
   for(unsigned int i = 0; i < writers.size(); i++)
   {
      writers[i]->writeHeader();
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--C backend: written header");
}

void CBackend::WriteGlobalDeclarations()
{
   for(unsigned int i = 0; i < writers.size(); i++)
   {
      writers[i]->WriteGlobalDeclarations();
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Writing function prototypes");
   for(unsigned int i = 0; i < writers.size(); i++)
   {
      std::list<unsigned int> extFunctions;
      this->AppM->get_functions_without_body(extFunctions);
      std::list<unsigned int>::iterator extBeg, extEnd;

      for (extBeg = extFunctions.begin(), extEnd = extFunctions.end(); extBeg != extEnd; extBeg++)
      {
         const behavioral_helperConstRef BH = this->AppM->get_function_behavior(*extBeg)->get_helper();
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Writing external function prototype: " + BH->get_function_name());
         bool is_system;
         std::string decl = std::tr1::get<0>(BH->get_definition(*extBeg, is_system));
         if (not is_system and decl != "<built-in>")
         {
            writers[i]->DeclareFunctionTypes(*extBeg);
            writers[i]->WriteFunctionDeclaration(*extBeg);
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Written external function prototype: " + BH->get_function_name());
      }
   }

   for(unsigned int i = 0; i < writers.size(); i++)
   {
      std::list<unsigned int> functions;
      AppM->get_functions_with_body(functions);
      THROW_ASSERT(functions.size() > 0, "at least one function is expected");
      std::list<unsigned int>::const_iterator it, it_end = functions.end();
      for(it = functions.begin(); it != it_end; it++)
      {
         const behavioral_helperConstRef BH = this->AppM->get_function_behavior(*it)->get_helper();
         bool is_system;
         std::string decl = std::tr1::get<0>(BH->get_definition(*it, is_system));
         if (!is_system and decl != "<built-in>")
         {
            writers[i]->DeclareFunctionTypes(*it);
            writers[i]->WriteFunctionDeclaration(*it);
         }
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Written function prototypes");
}

void CBackend::writeImplementations()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Writing implementation");
   for(unsigned int i = 0; i < writers.size(); i++)
   {
      //First of all I declare the functions and then the tasks
      std::list<unsigned int> functions;
      AppM->get_functions_with_body(functions);
      THROW_ASSERT(functions.size() > 0, "at least one function is expected");
      std::list<unsigned int>::const_iterator it, it_end = functions.end();
      for(it = functions.begin(); it != it_end; it++)
      {
         const behavioral_helperConstRef BH = this->AppM->get_function_behavior(*it)->get_helper();
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Writing implementation of function " + BH->get_function_name());
         bool is_system;
         std::string decl = std::tr1::get<0>(BH->get_definition(*it, is_system));
         if (!is_system and decl != "<built-in>")
            writeFunctionImplementation(writers[i], *it);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Written implementation of function " + BH->get_function_name());
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Written implementation");
}

void CBackend::writeFunctionImplementation(const CWriterRef writer, unsigned int funId)
{
   const behavioral_helperConstRef BH = AppM->get_function_behavior(funId)->get_helper();
   var_pp_functorRef variableFunctor(new std_var_pp_functor(BH));
   writer->StartFunctionBody(funId);
   declareLocalVariables(writer, funId);
   writeBodyFunction(writer, funId);
   writer->endFunctionBody(funId, variableFunctor);
}

void CBackend::writeBodyFunction(const CWriterRef writer, unsigned int funId)
{
   const behavioral_helperConstRef BH = AppM->get_function_behavior(funId)->get_helper();
   var_pp_functorRef variableFunctor(new std_var_pp_functor(BH));
   const std::tr1::unordered_set<vertex>& vertices = getVertices(AppM->get_function_behavior(funId)->get_graph(function_behavior::DFG));
   THROW_ASSERT(vertices.size() > 0, "Graph for function " + BH->get_function_name() + " is empty");
   routineFunctorRef codeFunc(new routineFunctor(AppM, AppM->get_function_behavior(funId)));
   writer->writeRoutineInstructions(vertices, variableFunctor, codeFunc);
}

std::tr1::unordered_set<vertex> CBackend::getVertices(const graph *inGraph) const
{
   std::tr1::unordered_set<vertex> vertices;
   VertexIterator v, vEnd;
   for (boost::tie(v, vEnd) = boost::vertices(*inGraph); v != vEnd; v++)
      vertices.insert(*v);

   return vertices;
}

void CBackend::declareLocalVariables(const CWriterRef writer, unsigned int funId)
{
   const behavioral_helperConstRef BH = AppM->get_function_behavior(funId)->get_helper();
   std::tr1::unordered_set<unsigned int> vars = getVariables(funId);

   const std::list<unsigned int>& funParams = BH->get_parameters();
   for (std::list<unsigned int>::const_iterator i = funParams.begin(); i != funParams.end(); i++)
   {
      if (vars.find(*i) != vars.end())
         vars.erase(*i);
   }

   const std::tr1::unordered_set<unsigned int> &gblVariables = AppM->get_global_variables();
   for (std::tr1::unordered_set<unsigned int>::const_iterator i = gblVariables.begin(); i != gblVariables.end(); i++)
   {
      if (vars.find(*i) != vars.end())
         vars.erase(*i);
   }

   var_pp_functorRef variableFunctor(new std_var_pp_functor(BH));
   std::tr1::unordered_set<unsigned int> already_declared_variables;
   std::tr1::unordered_set<std::string> locally_declared_types;
   writer->DeclareLocalVariables(vars, already_declared_variables, locally_declared_types, BH, variableFunctor);
}

std::tr1::unordered_set<unsigned int> CBackend::getVariables(unsigned int funId) const
{
   const graph* inGraph = this->AppM->get_function_behavior(funId)->get_graph(function_behavior::DFG);
   const behavioral_helperRef & BH = AppM->get_function_behavior(funId)->get_helper();
   std::tr1::unordered_set<unsigned int> vars;
   //I simply have to go over all the vertices and get the used variables;
   //the variables which have to be declared are all those variables but
   //the globals ones
   VertexIterator v, vEnd;
   for (boost::tie(v, vEnd) = boost::vertices(*inGraph); v != vEnd; v++)
   {
      std::tr1::unordered_set<unsigned int> varsTemp;
      prettyPrintVertex::get_referenced_vars(*v, inGraph, varsTemp, BH);
      vars.insert(varsTemp.begin(), varsTemp.end());
   }
   return vars;
}

void CBackend::writeIncludes()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->CBackend: writing includes");
   ///Set of decl_nodes to be examinated
   std::tr1::unordered_set<unsigned int> decl_nodes;

   std::set<std::string> writtenIncludes;

   std::list<unsigned int> extFunctions;
   this->AppM->get_functions_without_body(extFunctions);
   std::list<unsigned int>::iterator extBeg, extEnd;
   for (extBeg = extFunctions.begin(), extEnd = extFunctions.end(); extBeg != extEnd; extBeg++)
   {
      behavioral_helperRef BH = this->AppM->get_function_behavior(*extBeg)->get_helper();
      AnalyzeInclude(*extBeg, BH, writtenIncludes);
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Written includes for external functions");

   std::list<unsigned int> Functions;
   AppM->get_functions_with_body(Functions);
   for (extBeg = Functions.begin(), extEnd = Functions.end(); extBeg != extEnd; extBeg++)
   {
      const behavioral_helperConstRef BH = AppM->get_function_behavior(*extBeg)->get_helper();
      std::tr1::unordered_set<unsigned int> tmp_vars = CBackend::getVariables(*extBeg);
      decl_nodes.insert(tmp_vars.begin(), tmp_vars.end());
      const std::list<unsigned int>& funParams = BH->get_parameters();
      decl_nodes.insert(funParams.begin(), funParams.end());
   }

   const behavioral_helperRef & BH = AppM->get_function_behavior(*(Functions.begin()))->get_helper();
   std::tr1::unordered_set<unsigned int> vars = AppM->get_global_variables();
   decl_nodes.insert(vars.begin(), vars.end());
   for (std::tr1::unordered_set<unsigned int>::iterator v = decl_nodes.begin(); v != decl_nodes.end(); v++)
   {
      const unsigned int variable_type = BH->get_type(*v);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Analyzing " + BH->PrintVariable(*v) + " of type " + boost::lexical_cast<std::string>(variable_type));
      AnalyzeInclude(variable_type, BH, writtenIncludes);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
   }
   for(unsigned int i = 0; i < writers.size(); i++)
   {
      for(std::set<std::string>::iterator s = writtenIncludes.begin(); s != writtenIncludes.end(); s++)
      {
         writers[i]->writeInclude(*s);
      }
      std::vector<std::string> addHeaders = writers[i]->getIncludes();
      for(std::vector<std::string>::iterator s = addHeaders.begin(); s != addHeaders.end(); s++)
      {
         if (writtenIncludes.find(*s) == writtenIncludes.end())
            writers[i]->writeInclude(*s);
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Written includes for variables");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--CBackend: written includes");
}

void CBackend::AnalyzeInclude(unsigned int index, const behavioral_helperConstRef BH, std::set<std::string>& writtenIncludes)
{
   if(already_visited.find(index) != already_visited.end())
   {
      return;
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing include for " + boost::lexical_cast<std::string>(index));
   already_visited.insert(index);
   bool is_system;
   const std::string decl = std::tr1::get<0>(BH->get_definition(index, is_system));
   if(not decl.empty() and decl != "<built-in>" and is_system)
   {
      writtenIncludes.insert(decl);
   }
   else
   {
      const unsigned int type = tree_helper::get_type_index(TM, index);
      const std::tr1::unordered_set<unsigned int> types_to_be_declared = tree_helper::GetTypesToBeDeclared(TM, type);
      std::tr1::unordered_set<unsigned int>::const_iterator type_to_be_declared, type_to_be_declared_end = types_to_be_declared.end();
      for(type_to_be_declared = types_to_be_declared.begin(); type_to_be_declared != type_to_be_declared_end; type_to_be_declared++)
      {
         AnalyzeInclude(*type_to_be_declared, BH, writtenIncludes);
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed");
}

void CBackend::compute_variables(const graph *inGraph, const behavioral_helperConstRef BH, const std::tr1::unordered_set<unsigned int> &gblVariables, std::list<unsigned int> &funParams, std::tr1::unordered_set<unsigned int> & vars)
{
   //I simply have to go over all the vertices and get the used variables;
   //the variables which have to be declared are all those variables but
   //the globals ones
   VertexIterator v, vEnd;
   for(boost::tie(v, vEnd) = boost::vertices(*inGraph); v != vEnd; v++)
   {
      std::tr1::unordered_set<unsigned int> varsTemp;
      prettyPrintVertex::get_referenced_vars(*v, inGraph, varsTemp, BH);
      vars.insert(varsTemp.begin(), varsTemp.end());
   }

   //I have to take out the variables global to the whole program and the function parameters
   std::set<unsigned int> varsTemp;
   std::tr1::unordered_set<unsigned int>::const_iterator it, it_end = gblVariables.end();
   for(it = gblVariables.begin(); it != it_end; it++)
   {
      vars.erase(*it);
   }
   std::list<unsigned int>::const_iterator it2, it2_end = funParams.end();
   for(it2 = funParams.begin(); it2 != it2_end; it2++)
   {
      vars.erase(*it2);
   }
}

