/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file c_backend.hpp
 * @brief Simple class used to drive the backend in order to be able to print c source code
 *
 * @author Luca Fossati <fossati@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

#ifndef C_BACKEND_HPP
#define C_BACKEND_HPP

///graph include
#include "graph.hpp"

///Std include
#include <ostream>
#include <fstream>
#include <iostream>
#include <sstream>

///Stl include
#include <list>
#include <map>
#include <set>
#include <string>
#include <tr1/unordered_set>
#include <vector>

///utility include
#include "refcount.hpp"

CONSTREF_FORWARD_DECL(behavioral_helper);
REF_FORWARD_DECL(CWriter);
REF_FORWARD_DECL(InstructionWriter);
CONSTREF_FORWARD_DECL(Parameter);
CONSTREF_FORWARD_DECL(application_manager);
CONSTREF_FORWARD_DECL(tree_manager);

/**
 * Class simply used to drive the backend in order to print C code
 */
class CBackend
{
   protected:

      ///The set of input parameters
      const ParameterConstRef Param;

      ///The debug level
      int debug_level;


      ///The set of already analyzed nodes during search of header to include; it is used to avoid infinite recursion
      std::tr1::unordered_set<unsigned int> already_visited;

      ///the general writers
      std::vector<CWriterRef> writers;

      ///the manager of the application
      const application_managerConstRef AppM;

      ///The tree_manager
      const tree_managerConstRef TM;

      /**
       * Computes the variables which have to be declared for a function
       * @param inGraph is the data flow graph of the function
       * @param BH is the behavioral helper
       * @param gblVariables is the set containing the node ids of the global variables
       * @param funParams is the list of function parameters
       * @param computed_variables is the set where the computed variables will be stored
       */
      void compute_variables(const graph *inGraph, const behavioral_helperConstRef BH, const std::tr1::unordered_set<unsigned int> &gblVariables, std::list<unsigned int> &funParams, std::tr1::unordered_set<unsigned int> & computed_variables);

      /**
       * Return the set of vertices of a function
       * @param inGraph is the graph
       * @return the set of vertices
       */
      std::tr1::unordered_set<vertex> getVertices(const graph *inGraph) const;

      virtual void writeFunctionImplementation(const CWriterRef writer, unsigned int funId);

      virtual void writeBodyFunction(const CWriterRef writer, unsigned int funId);

      virtual void declareLocalVariables(const CWriterRef writer, unsigned int funId);

      virtual std::tr1::unordered_set<unsigned int> getVariables(unsigned int funId) const;

      /**
       * Analyze a variable or a type to identify the includes to be added
       * @param index is the index of the variable or of the type 
       * @param BH is the behavioral helper
       * @param includes is where include has to be inseted
       */
      virtual void AnalyzeInclude(unsigned int index, const behavioral_helperConstRef BH, std::set<std::string>& includes);

   public:
      /**
       * Constructor
       * @param AppM is the manager of the application
       * @param Param is the set of input parameters
       */
      CBackend(const application_managerConstRef _AppM, const ParameterConstRef Param);

      /**
       * Destructor
       */
      virtual ~CBackend();

      /**
       * Write the source code
       * @param fileName is name of the file
       */
      virtual void write(std::string fileName);

      /**
       * Register a writer
       * @param writer is the writer
       */
      virtual void registerWriter(const CWriterRef writer);

      /**
       * Writes the file header, i.e the comments at the beginning of the file
       * @param library is the stream where to write the header
       */
      virtual void writeHeader();

      /**
       * Writes the file header, i.e the comments at the beginning of the file
       */
      virtual void writeIncludes();

      /**
       * Write the global declarations
       */
      virtual void WriteGlobalDeclarations();

      virtual void writeImplementations();
};

/// Refcount definition for the class CBackend
typedef refcount<CBackend> CBackendRef;


#endif
