/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file c_writer.cpp
 * @brief This file contains the routines necessary to create a C executable program starting from an abstract decription of the threads composing the application.
 *
 * This file contains the routines necessary to create a C executable program
 * starting from an abstract decription of the threads composing the application.
 *
 * @author Luca Fossati <fossati@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Autoheader include
#include "config_HAVE_ARM_COMPILER.hpp"
#include "config_HAVE_SPARC_COMPILER.hpp"
#include "config_HAVE_TARGET_PROFILING.hpp"
#include "config_HAVE_TASKGRAPH_BUILT.hpp"
#include "config_PACKAGE_NAME.hpp"
#include "config_PACKAGE_VERSION.hpp"

///Header include
#include "c_writer.hpp"

///backend include
#include "instruction_writer.hpp"
#include "prettyPrintVertex.hpp"
#include "var_pp_functor.hpp"

///boost include
#include <boost/config.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/topological_sort.hpp>
#include <boost/graph/filtered_graph.hpp>
#include <boost/graph/graph_utility.hpp>
#include "boost/filesystem/operations.hpp"
#include "boost/filesystem/path.hpp"
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>

#include "graph.hpp"
#if HAVE_TASKGRAPH_BUILT
#include "thread_c_writer.hpp"
#include "TaskGraph.hpp"
#endif

///frontend include
#include "basic_block.hpp"
#include "behavioral_helper.hpp"
#include "application_manager.hpp"
#include "cdfg_edge_info.hpp"
#include "function_behavior.hpp"
#include "loop.hpp"
#include "loops.hpp"
#include "op_node_info.hpp"
#include "Parameter.hpp"

///Std include
#include <fstream>
#include <iostream>
#include <ostream>
#include <string>
#include <set>
#include <vector>

///Tree include
#include "tree_basic_block.hpp"
#include "tree_helper.hpp"
#include "tree_manager.hpp"
#include "tree_reindex.hpp"

///utility include
#include "refcount.hpp"
#include "exceptions.hpp"
#include "simple_indent.hpp"
#include "Dominance.hpp"
#include <boost/date_time/posix_time/posix_time.hpp>

CWriter::CWriter(const application_managerConstRef _AppM, const ParameterConstRef _Param, bool _verbose) :
      AppM(_AppM),
      TM(_AppM->get_tree_manager()),
      outStream(new std::ostringstream),
      //created outstream for file.h
      outStream_h(new std::ostringstream),
      indenter('{', '}', 3),
      added_goto(0),
      verbose(_verbose),
      Param(_Param),
      debug_level(_Param->get_class_debug_level("CWriter", DEBUG_LEVEL_NONE)),
      output_level(_Param->getOption<int>(OPT_output_level)),
      fake_max_tree_node_id(TM->get_next_available_tree_node_id())
{

}

CWriter::~CWriter()
{
   if (this->outStream != NULL)
      delete this->outStream;
   if (this->outStream_h != NULL)
      delete this->outStream_h;
}

void CWriter::writeFile(const std::string& fileName)
{
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->CWriter writing file: " + fileName);
   // Opening output file
   std::ofstream fileOut(fileName.c_str(), std::ios::out);

   // writing code stream
   if (outStream)
   {
      fileOut << outStream->str() << std::endl;
      ///Cleaning the stream
      delete this->outStream;
      this->outStream = new std::ostringstream;
   }

   fileOut.close();
   globally_declared_types.clear();
   globallyDeclVars.clear();
   const unsigned int tree_added_goto = TM->get_added_goto();
   if (Param->getOption<bool>(STR_OPT_frontend_statistics))
   {
      PRINT_OUT_MEX(OUTPUT_LEVEL_NONE, output_level, "************************************");
      PRINT_OUT_MEX(OUTPUT_LEVEL_NONE, output_level, "*   Starting Architecture Reading  *");
      PRINT_OUT_MEX(OUTPUT_LEVEL_NONE, output_level, "************************************");
      PRINT_OUT_MEX(OUTPUT_LEVEL_NONE, output_level, "goto added directly in tree: " + boost::lexical_cast<std::string>(tree_added_goto));
      PRINT_OUT_MEX(OUTPUT_LEVEL_NONE, output_level, "goto Added during c dumping: " + boost::lexical_cast<std::string>(added_goto));
      PRINT_OUT_MEX(OUTPUT_LEVEL_NONE, output_level, "Pointer plus removed from the tree: " + boost::lexical_cast<std::string>(TM->get_removed_pointer_plus()));
      PRINT_OUT_MEX(OUTPUT_LEVEL_NONE, output_level, "Pointer plus removed during the backend printing: " + boost::lexical_cast<std::string>(TM->get_removable_pointer_plus()));
      PRINT_OUT_MEX(OUTPUT_LEVEL_NONE, output_level, "Pointer plus not removed: " + boost::lexical_cast<std::string>(TM->get_unremoved_pointer_plus()));
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--CWriter written file: " + fileName);
}

void CWriter::add_instruction_writer(const InstructionWriterRef new_instrWriter)
{
   instrWriter = new_instrWriter;
}

void CWriter::writeHeader()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->CWriter: writing header");

   indenter(*outStream, "/*\n");
   indenter(*outStream, " * Politecnico di Milano\n");
   indenter(*outStream, std::string(" * Code created using the ") + PACKAGE_NAME + " framework version=" + PACKAGE_VERSION);
#if !RELEASE
   boost::posix_time::ptime current_time = boost::posix_time::second_clock::local_time();
   indenter(*outStream,  " - Revision " + boost::lexical_cast<std::string>(Param->getOption<std::string>(OPT_revision)) + " - Date " + boost::posix_time::to_simple_string(current_time));
#endif
   indenter(*outStream, "\n");
#if !RELEASE
   if (Param->isOption(OPT_cat_args))
      indenter(*outStream, " * " + Param->getOption<std::string>(OPT_program_name) + " executed with: " + Param->getOption<std::string>(OPT_cat_args) + "\n");
#endif
   indenter(*outStream, " */\n");
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--CWriter: written header");
}

void CWriter::WriteGlobalDeclarations()
{
   ///Writing declarations of global variables
   std::list<unsigned int> functions;
   AppM->get_functions_with_body(functions);
   THROW_ASSERT(functions.size() > 0, "at least one function is expected");
   unsigned int first_fun = *functions.begin();
   const behavioral_helperRef& BH = AppM->get_function_behavior(first_fun)->get_helper();

   const std::tr1::unordered_set<unsigned int>& gblVariables = AppM->get_global_variables();
   //Write the declarations for the global variables
   var_pp_functorRef variableFunctor(new std_var_pp_functor(BH));
   std::tr1::unordered_set<unsigned int>::const_iterator gblVars, gblVarsEnd;
   for (gblVars = gblVariables.begin(), gblVarsEnd = gblVariables.end(); gblVars != gblVarsEnd; gblVars++)
   {
      declare_variable(*gblVars, globallyDeclVars, globally_declared_types, BH, variableFunctor);
   }
   indenter(*outStream, "\n");

}

void CWriter::DeclareFunctionTypes(const unsigned int funId)
{
   const function_behaviorRef FB = AppM->get_function_behavior(funId);

   const behavioral_helperRef BH = FB->get_helper();
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Declaring function types for " + BH->get_function_name());

   //In case the function parameters are of a non built_in type I have
   //to declare their type

   std::tr1::unordered_set<unsigned int> parameter_types;
   BH->get_parameters_type(parameter_types);
   std::tr1::unordered_set<unsigned int>::const_iterator parameter_type, parameter_type_end = parameter_types.end();
   for(parameter_type = parameter_types.begin(); parameter_type != parameter_type_end; parameter_type++)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Parameter type " + boost::lexical_cast<std::string>(*parameter_type));
      declare_type(*parameter_type, BH, globally_declared_types);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Parameter type " + boost::lexical_cast<std::string>(*parameter_type));
   }

   if (BH->has_return())
   {
      const unsigned int return_type = BH->get_function_return(funId);
      declare_type(return_type, BH, globally_declared_types);
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Declared function types for " + BH->get_function_name());
}

void CWriter::WriteFunctionDeclaration(unsigned int funId)
{
   const function_behaviorConstRef FB = AppM->get_function_behavior(funId);
   const behavioral_helperConstRef BH = FB->get_helper();
   const std::string& funName = BH->get_function_name();
#if HAVE_ARM_COMPILER
   if(Param->getOption<CompilerTarget>(OPT_default_compiler) == CT_ARM_GCC and BH->is_var_args())
   {
      THROW_ERROR_CODE(VARARGS_EC, "Source code containing vargs function (" + BH->get_function_name() + ")produced using arm compiler can not be compiled by x86 compilers");
   }
#endif
#if HAVE_SPARC_COMPILER
   if(Param->getOption<CompilerTarget>(OPT_default_compiler) == CT_SPARC_GCC and BH->is_var_args())
   {
      THROW_ERROR_CODE(VARARGS_EC, "Source code containing vargs function (" + BH->get_function_name() + ") produced using sparc compiler can not be compiled by x86 compilers");
   }
#endif
   if (funName != "main")
   {
      simple_indent PP(STD_OPENING_CHAR, STD_CLOSING_CHAR, 3);
      var_pp_functorRef protFunctor(new std_var_pp_functor(BH));
      routineFunctorRef codeFunc(new routineFunctor(AppM, FB));
      this->instrWriter->declareFunction(*outStream, PP, protFunctor, codeFunc);
      *outStream << ";" << std::endl << std::endl;
   }
}

void CWriter::StartFunctionBody(const unsigned int funId)
{
   //First of all I have to check whether the function represents
   //main: in case I have to print separately its declaration
   //and the initialization. Otherwise I print the prototype
   //of the function; then the body of the function can be printed
   //in the same way for both the type of functions
   //Actually also in case the function is not called main but its id
   //is equal to this->mainFunId (not that no function id can be equal
   //to 0), then it is treated as if it were main.

   const function_behaviorRef& FB = AppM->get_function_behavior(funId);
   const behavioral_helperConstRef BH = FB->get_helper();

   var_pp_functorRef prototypeFunc(new std_var_pp_functor(BH));
   routineFunctorRef codeFunc(new routineFunctor(AppM, FB));
   instrWriter->declareFunction(*outStream, indenter, prototypeFunc, codeFunc);
   indenter(*outStream, "\n{\n");
}

void CWriter::endFunctionBody(unsigned int funId, var_pp_functorConstRef taskFunctor)
{
   //In case the function has a return value I have to print
   //the return instruction
   const function_behaviorRef& FB = AppM->get_function_behavior(funId);
   const behavioral_helperConstRef BH = FB->get_helper();
   if (BH->has_return())
   {
      routineFunctorRef codeFunc(new routineFunctor(AppM, FB));
      this->instrWriter->printReturn(*outStream, indenter, taskFunctor, codeFunc);
   }

   indenter(*outStream, "}\n");
   if (this->verbose)
      indenter(*outStream, "//end of function; id: " +
               boost::lexical_cast<std::string>(funId) + "\n");
   indenter(*outStream, "\n");
}

void CWriter::writeRoutineInstructions_rec
(
   const bb_graph & bb_fcfgGraph,
   const graph * cfgGraph,
   vertex current_vertex,
   std::set<vertex> &bb_frontier,
   std::set<vertex> &bb_analyzed,
   const std::tr1::unordered_set<vertex> &instructions,
   const std::map<unsigned int, std::string> &basic_blocks_labels,
   const dominance<bb_graph> * dominators,
   const dominance<bb_graph> * post_dominators,
   std::set<vertex> & goto_list,
   var_pp_functorConstRef variableFunctor,
   const routineFunctorConstRef codeFunc,
   bool bracket,
   loop_writerRef loops_writer
)
{
   const behavioral_helperConstRef BH = codeFunc->FB->get_helper();
#if HAVE_TARGET_PROFILING || !defined(NDEBUG)
   unsigned int funId = BH->get_function_index();
#endif
   const bb_node_info* bi = CGET_BB_NODE_INFO(&bb_fcfgGraph, current_vertex);
   const unsigned int bb_number = bi->block->number;
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Starting writing BB" + boost::lexical_cast<std::string>(bb_number));
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->");
   THROW_ASSERT(bb_frontier.find(current_vertex) == bb_frontier.end(), "current_vertex cannot be part of the basic block frontier");

   if (bb_analyzed.find(current_vertex) != bb_analyzed.end())
      return;

   bool add_phi_nodes_assignment = basic_block_tail.find(bb_number) != basic_block_tail.end();
   bool add_phi_nodes_assignment_prefix = basic_block_prefix.find(bb_number) != basic_block_prefix.end();
   bb_analyzed.insert(current_vertex);
   if (this->verbose)
      indenter(*outStream, "//Basic block " + boost::lexical_cast<std::string>(bb_number) + " - loop " + boost::lexical_cast<std::string>(bi->loop_id) + "\n");

   ///get immediate post-dominator
   vertex bb_PD = post_dominators->get_immediate_dominator(current_vertex);
#ifndef NDEBUG
   const bb_node_info* bi_PD = CGET_BB_NODE_INFO(&bb_fcfgGraph, bb_PD);
   const unsigned int bb_number_PD = bi_PD->block->number;
   std::string frontier_string;
   std::set<vertex>::iterator bb_frontier_begin, bb_frontier_end = bb_frontier.end();
   for (bb_frontier_begin = bb_frontier.begin(); bb_frontier_begin != bb_frontier_end; bb_frontier_begin++)
   {
      frontier_string += "BB" + boost::lexical_cast<std::string>(CGET_BB_NODE_INFO(&bb_fcfgGraph, *bb_frontier_begin)->block->number) + " ";
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Frontier at the moment is: " + frontier_string);
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Its post-dominator is BB" + boost::lexical_cast<std::string>(bb_number_PD));
#endif
   bool analyze_bb_PD = bb_frontier.find(bb_PD) == bb_frontier.end() && bb_analyzed.find(bb_PD) == bb_analyzed.end();
   if (analyze_bb_PD)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Post dominator will be examinated");
      bb_frontier.insert(bb_PD);
   }
   ///compute the last statement
   vertex last_stmt = 0;
   std::list<vertex>::const_reverse_iterator vRIter, vRIterEnd;
   bool is_there = false;
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Operations of task ");
   if (debug_level >= DEBUG_LEVEL_VERY_PEDANTIC)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->");
      std::tr1::unordered_set<vertex>::const_iterator it, it_end;
      it_end = instructions.end();
      for (it = instructions.begin(); it != it_end; it++)
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, GET_NAME(cfgGraph, *it));
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Looking for last statement");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->");
   for (vRIter = bi->statements_list.rbegin(), vRIterEnd = bi->statements_list.rend(); vRIter != vRIterEnd ; vRIter++)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Considering operation " + GET_NAME(cfgGraph, *vRIter));
      if (instructions.find(*vRIter) == instructions.end()) continue;
      if (GET_TYPE(cfgGraph, *vRIter) & TYPE_VPHI) continue;
      if ((GET_TYPE(cfgGraph, *vRIter) & TYPE_INIT) != 0) continue;
      is_there = true;
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "This is basic block is not empty in this task. Last operation to be printed id " + GET_NAME(cfgGraph, *vRIter));
      last_stmt = *vRIter;
      break;
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
   ///check the feasibility
   bool last_statement_is_a_cond_or_goto = is_there && BH->end_with_a_cond_or_goto(bi->block) != 0 && last_stmt == bi->statements_list.back();
   THROW_ASSERT(!last_statement_is_a_cond_or_goto || !is_there || (last_statement_is_a_cond_or_goto && last_stmt == bi->statements_list.back()), "inconsistent recursion");
   ///check for basic block label
   bool start_with_a_label = BH->start_with_a_label(bi->block);

   if (start_with_a_label)
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Basic block starts with a label");

   bool add_bb_label = goto_list.find(current_vertex) != goto_list.end();

   if (add_bb_label)
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Basic block should start with a label");

   if (!add_bb_label && !start_with_a_label && boost::in_degree(current_vertex, bb_fcfgGraph) > 1)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Basic block has an indegree > 1 and not associated label");
      InEdgeIterator inE, inEEnd;
      for (boost::tie(inE, inEEnd) = boost::in_edges(current_vertex, bb_fcfgGraph); inE != inEEnd; inE++)
      {
         vertex source = boost::source(*inE, bb_fcfgGraph);
         const bb_node_info* pred_bi = CGET_BB_NODE_INFO(&bb_fcfgGraph, source);

         //This condition match first basic block of case preceded by a case without break
         if (pred_bi->statements_list.size() && (GET_TYPE(cfgGraph, *(pred_bi->statements_list.rbegin())) & TYPE_SWITCH) && post_dominators->get_immediate_dominator(source) != current_vertex)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Basic block is the first case of a case preceded by a case without break");
            add_bb_label = true;
            break;
         }
         //Basic block start the body of a short circuit
         else if (bb_analyzed.find(source) == bb_analyzed.end() and !((FB_CFG_SELECTOR & bb_fcfgGraph.get_selector(*inE))))
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Basic block should start with a label since is the body of a short-circuit");
            add_bb_label = true;
            break;
         }
         //Basic block is a header loop, but it does not end with while or for
         else if ((bb_analyzed.find(source) == bb_analyzed.end() or current_vertex == source) and (bi->statements_list.empty() or ((GET_TYPE(cfgGraph, *(bi->statements_list.rbegin())) & (TYPE_WHILE | TYPE_FOR)) == 0)))
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Basic block is the header of a loop and it does not end with while or for");
            add_bb_label = true;
            break;
         }
      }
   }
   add_bb_label = add_bb_label && !start_with_a_label;
   bool add_semicolon = false;
   ///print each instruction
   if (bracket)
   {
      if (analyze_bb_PD || is_there || add_bb_label || add_phi_nodes_assignment || add_phi_nodes_assignment_prefix)
      {
         indenter(*outStream, "{\n");
      }
      else
         add_semicolon = true;
   }
   if (add_bb_label)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "A label should be added at the beginning");
      THROW_ASSERT(basic_blocks_labels.find(bb_number) != basic_blocks_labels.end(), "I do not know the destination");
      indenter(*outStream, basic_blocks_labels.find(bb_number)->second + ":\n");
      add_semicolon = true;
   }
   else if (start_with_a_label)
   {
      add_semicolon = true;
   }
   std::list<vertex>::const_iterator vIter, vIterEnd, vIterBegin;
   vIter = bi->statements_list.begin();
   vIterBegin = vIter;
   vIterEnd = bi->statements_list.end();
   if (is_there)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "There are instructions to be printed for this pair task - basic block");
      unsigned int analyzed_statement = 0;
      /// fill the renamin table in case is needed
      if(renaming_table.find(current_vertex) != renaming_table.end())
      {
         const std::map<unsigned int, std::string> & rvt = renaming_table.find(current_vertex)->second;
         std::map<unsigned int, std::string>::const_iterator rvt_it_end = rvt.end();
         for(std::map<unsigned int, std::string>::const_iterator rvt_it = rvt.begin(); rvt_it != rvt_it_end; ++rvt_it)
            behavioral_helper::rename_a_variable(rvt_it->first, rvt_it->second);
      }
      bool printed_prefix = basic_block_prefix.find(bb_number) == basic_block_prefix.end(); /// nothing to print imply that we already have printed...
      do
      {
         if(!printed_prefix && ((analyzed_statement > 0 && start_with_a_label) || (analyzed_statement==0 && !start_with_a_label)))
         {
            printed_prefix = true;
            indenter(*outStream, basic_block_prefix.find(bb_number)->second);
         }
         if (instructions.find(*vIter) == instructions.end()) continue;
         analyzed_statement++;
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Preparing printing of operation " + GET_NAME(cfgGraph, *vIter));
         //Now I print the instruction
         if (this->verbose)
            indenter(*outStream, "//Instruction: " + GET_NAME(cfgGraph, *vIter) + "\n");

         if (start_with_a_label && vIter == vIterBegin)
         {
            InEdgeIterator inE, inEEnd;
            for (tie(inE, inEEnd) = boost::in_edges(current_vertex, bb_fcfgGraph); inE != inEEnd ; inE++)
            {
               if (FB_BB_SELECTOR & bb_fcfgGraph.get_selector(*inE))
                  indenter(*outStream, "//start of a loop\n");
            }
         }
         bool isLastIntruction = last_stmt == *vIter;
         /// in case we have phi nodes we check if some assignments should be printed
         bool print_phi_now = ((GET_TYPE(cfgGraph, *vIter) & (TYPE_IF | TYPE_WHILE | TYPE_FOR | TYPE_SWITCH))) || BH->end_with_a_cond_or_goto(bi->block);
         if(add_phi_nodes_assignment && isLastIntruction && print_phi_now)
            indenter(*outStream, basic_block_tail.find(bb_number)->second);
         if (loops_writer && (GET_TYPE(cfgGraph, *vIter) & (TYPE_WHILE | TYPE_FOR)))
         {
            analyzed_statement--;
         }
         else if ((GET_TYPE(cfgGraph, *vIter) & (TYPE_INIT | TYPE_VPHI)) == 0)
         {
            instrWriter->write(*outStream, *vIter, indenter, variableFunctor, codeFunc, isLastIntruction);
            if ((GET_TYPE(cfgGraph, *vIter) &  TYPE_LABEL) == 0)
               add_semicolon = false;
         }
         else if (this->verbose)
         {
            if ((GET_TYPE(cfgGraph, *vIter) & TYPE_INIT) != 0)
               indenter(*outStream, "//(removed init instruction)\n");
            else
               indenter(*outStream, "//(removed virtual phi instruction)\n");
         }
         if (!isLastIntruction)
            continue;
         behavioral_helper::clear_renaming_table();
         if(add_phi_nodes_assignment && !print_phi_now)
            indenter(*outStream, basic_block_tail.find(bb_number)->second);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---This is not the last statement");
         //Now I check if this is a control statement and I consequently print
         //the instructions contained in its branches
         if (GET_TYPE(cfgGraph, *vIter) & TYPE_IF)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Operation is an if");
            const unsigned int bb_true_number = bi->block->true_edge;
            const vertex true_vertex = bb_fcfgGraph.GetConstBBGraphInfo()->bb_index_map.find(bb_true_number)->second;
            const unsigned int bb_false_number = bi->block->false_edge;
            vertex false_vertex = bb_fcfgGraph.GetConstBBGraphInfo()->bb_index_map.find(bb_false_number)->second;
            bool add_false_to_goto = false;
            if (bb_frontier.find(true_vertex) == bb_frontier.end())
            {
               if (bb_frontier.find(false_vertex) == bb_frontier.end() && goto_list.find(false_vertex) == goto_list.end())
               {
                  goto_list.insert(false_vertex);
                  add_false_to_goto = true;
               }
               if (bb_analyzed.find(true_vertex) == bb_analyzed.end())
                  writeRoutineInstructions_rec
                  (
                     bb_fcfgGraph,
                     cfgGraph,
                     true_vertex,
                     bb_frontier, bb_analyzed, instructions, basic_blocks_labels, dominators, post_dominators, goto_list, variableFunctor, codeFunc, true, loops_writer);
               else
               {
                  THROW_ASSERT(basic_blocks_labels.find(bb_true_number) != basic_blocks_labels.end(), "I do not know the destination");
                  added_goto++;
                  indenter(*outStream, "   goto " + basic_blocks_labels.find(bb_true_number)->second + ";\n");
               }
            }
            else
               indenter(*outStream, "{}\n");
            if (add_false_to_goto)
               goto_list.erase(false_vertex);
            if (bb_frontier.find(false_vertex) == bb_frontier.end())
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "False BB is not on the frontier");
               EdgeDescriptor e;
               bool ins;
               false_vertex = bb_fcfgGraph.GetConstBBGraphInfo()->bb_index_map.find(bb_false_number)->second;
               boost::tie(e, ins) = boost::edge(current_vertex, false_vertex, bb_fcfgGraph);
               if (not(FB_CFG_SELECTOR & bb_fcfgGraph.get_selector(e)))
               {
                  indenter(*outStream, "else\n");
                  if (bb_analyzed.find(false_vertex) == bb_analyzed.end())
                     writeRoutineInstructions_rec
                     (
                        bb_fcfgGraph,
                        cfgGraph,
                        false_vertex,
                        bb_frontier,
                        bb_analyzed,
                        instructions,
                        basic_blocks_labels,
                        dominators,
                        post_dominators,
                        goto_list,
                        variableFunctor,
                        codeFunc,
                        true,
                        loops_writer
                     );
                  else
                  {
                     THROW_ASSERT(basic_blocks_labels.find(bb_false_number) != basic_blocks_labels.end(), "I do not know the destination");
                     added_goto++;
                     indenter(*outStream, "   goto " + basic_blocks_labels.find(bb_false_number)->second + ";\n");
                  }
               }
               ///Feedback edge on the false path of an if
               else
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Target is a header");
                  const vertex target = boost::target(e, bb_fcfgGraph);
                  const bb_node_info * target_bi = CGET_BB_NODE_INFO(&bb_fcfgGraph, target);
                  ///Target is not a while or a for
                  if(target_bi->statements_list.empty() or ((GET_TYPE(cfgGraph, *(target_bi->statements_list.rbegin())) & (TYPE_WHILE | TYPE_FOR)) == 0))
                  {
                     indenter(*outStream, "else\n");
                     THROW_ASSERT(basic_blocks_labels.find(bb_false_number) != basic_blocks_labels.end(), "I do not know the destination");
                     added_goto++;
                     indenter(*outStream, "   goto " + basic_blocks_labels.find(bb_false_number)->second + ";\n");
                  }
               }
            }
         }
         else if (GET_TYPE(cfgGraph, *vIter) & (TYPE_WHILE | TYPE_FOR))
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Operation is a while or a for");
            const unsigned int bb_true_number = bi->block->true_edge;
            const vertex true_vertex = bb_fcfgGraph.GetConstBBGraphInfo()->bb_index_map.find(bb_true_number)->second;
            if (bb_frontier.find(true_vertex) == bb_frontier.end())
            {
               if (bb_analyzed.find(true_vertex) == bb_analyzed.end())
               {
                  if (analyzed_statement == instructions.size())
                     return;
                  if (!loops_writer or !(*loops_writer)(*outStream, codeFunc, bb_number))
                     writeRoutineInstructions_rec
                     (
                        bb_fcfgGraph,
                        cfgGraph,
                        true_vertex,
                        bb_frontier,
                        bb_analyzed,
                        instructions,
                        basic_blocks_labels,
                        dominators,
                        post_dominators,
                        goto_list,
                        variableFunctor,
                        codeFunc,
                        true,
                        loops_writer
                     );
               }
               else
               {
                  THROW_ERROR("Body of a loop has yet been printed before the while statement");
               }
            }
            else
            {
               return;
            }

            const unsigned int bb_false_number = bi->block->false_edge;
            const vertex false_vertex = bb_fcfgGraph.GetConstBBGraphInfo()->bb_index_map.find(bb_false_number)->second;
            if (bb_frontier.find(false_vertex) == bb_frontier.end())
            {
               if (bb_analyzed.find(false_vertex) == bb_analyzed.end())
               {
                  writeRoutineInstructions_rec
                  (
                     bb_fcfgGraph,
                     cfgGraph,
                     false_vertex,
                     bb_frontier,
                     bb_analyzed,
                     instructions,
                     basic_blocks_labels,
                     dominators,
                     post_dominators,
                     goto_list,
                     variableFunctor,
                     codeFunc,
                     false,
                     loops_writer
                  );
               }
               else
               {
                  added_goto++;
                  indenter(*outStream, "goto " + basic_blocks_labels.find(bb_false_number)->second + ";\n");
                  goto_list.insert(false_vertex);
               }
            }
         }
         else if (GET_TYPE(cfgGraph, *vIter) & TYPE_SWITCH)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Operation is a switch");
            ///now we can analyze the following basic blocks
            indenter(*outStream,  "{\n");
            OutEdgeIterator oE, oEEnd;
            for (boost::tie(oE, oEEnd) = boost::out_edges(current_vertex, bb_fcfgGraph); oE != oEEnd; oE++)
            {
               bool empty_block = false;
               vertex next_bb = boost::target(*oE, bb_fcfgGraph);
#ifndef NDEBUG
               const bb_node_info* bi_next_bb = CGET_BB_NODE_INFO(&bb_fcfgGraph, next_bb);
               const unsigned int bb_number_next_bb = bi_next_bb->block->number;
#endif
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Examining successor " + boost::lexical_cast<std::string>(bb_number_next_bb));
               std::set<unsigned int>::const_iterator eIdBeg, eIdEnd;
               std::set<unsigned int> Set = bb_fcfgGraph.get_bb_edge_info(*oE)->get_labels(BB_SELECTOR);
               for (eIdBeg = Set.begin(), eIdEnd = Set.end(); eIdBeg != eIdEnd; eIdBeg++)
               {
                  if (*eIdBeg == default_COND)
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Successor is default case");
                     if (next_bb == post_dominators->get_immediate_dominator(current_vertex))
                     {
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Default is empty");
                        empty_block = true;
                        indenter(*outStream, "default:\n");
                        if (current_vertex == dominators->get_immediate_dominator(next_bb))
                           analyze_bb_PD = true;
                        break;
                     }
                     indenter(*outStream,  "default");
                  }
                  else
                  {
                     if (next_bb == bb_PD) ///then adjust post dominator
                     {
                        empty_block = true;
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Removed current basic block");
                     }

                     indenter(*outStream,  "case " + BH->print_constant(*eIdBeg));
                  }
                  indenter(*outStream,  ":\n");
               }
               if (empty_block)
               {
                  indenter(*outStream,  "break;\n");
                  continue;
               }
               if (bb_analyzed.find(next_bb) == bb_analyzed.end())
               {
                  writeRoutineInstructions_rec(bb_fcfgGraph, cfgGraph, next_bb, bb_frontier, bb_analyzed, instructions, basic_blocks_labels, dominators, post_dominators, goto_list, variableFunctor, codeFunc, true, loops_writer);
                  indenter(*outStream,  "break;\n");
               }
               else
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Successor has already been examined");
                  const bb_node_info* next_bi = CGET_BB_NODE_INFO(&bb_fcfgGraph, next_bb);
                  const unsigned int next_bb_number = next_bi->block->number;

                  THROW_ASSERT(basic_blocks_labels.find(next_bb_number) != basic_blocks_labels.end(), "I do not know the destination " + boost::lexical_cast<std::string>(next_bb_number));
                  added_goto++;
                  indenter(*outStream, "   goto " + basic_blocks_labels.find(next_bb_number)->second + ";\n");
               }
            }
            indenter(*outStream,  "}\n");
         }
         else if (BH->end_with_a_cond_or_goto(bi->block))
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Basic block ends with a cond or a goto");
            if (last_statement_is_a_cond_or_goto)
            {
               ///now we can analyze the following basic blocks
               OutEdgeIterator oE, oEEnd;
               for (boost::tie(oE, oEEnd) = boost::out_edges(current_vertex, bb_fcfgGraph); oE != oEEnd; oE++)
               {
                  vertex next_bb = boost::target(*oE, bb_fcfgGraph);
                  if (bb_frontier.find(next_bb) != bb_frontier.end()) continue;
                  goto_list.insert(next_bb);
               }
            }
         }
         else
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---It is not a special operation");
            const vertex bbentry = bb_fcfgGraph.GetConstBBGraphInfo()->entry_vertex;
            if (current_vertex == bbentry)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Ended writing basic block " + boost::lexical_cast<std::string>(bb_number));
               return;
            }
            THROW_ASSERT(boost::out_degree(current_vertex, bb_fcfgGraph) <= 1, "only one edge expected");
            OutEdgeIterator oE, oEEnd;
            for (boost::tie(oE, oEEnd) = boost::out_edges(current_vertex, bb_fcfgGraph); oE != oEEnd; oE++)
            {
               vertex next_bb = boost::target(*oE, bb_fcfgGraph);
               if (bb_frontier.find(next_bb) != bb_frontier.end())
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Not adding goto since target is in the frontier");
                  continue;
               }
               if (FB_CFG_SELECTOR & bb_fcfgGraph.get_selector(*oE))
               {
                  const vertex target = boost::target(*oE, bb_fcfgGraph);
                  const bb_node_info * target_bi = CGET_BB_NODE_INFO(&bb_fcfgGraph, target);
                  if(target_bi->statements_list.size() and (GET_TYPE(cfgGraph, *(target_bi->statements_list.rbegin())) & (TYPE_WHILE | TYPE_FOR)))
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Not adding a goto since target is a while/for");
                     continue;
                  }
               }
               if (boost::in_degree(next_bb, bb_fcfgGraph) == 1)
               {
                  writeRoutineInstructions_rec
                  (
                     bb_fcfgGraph,
                     cfgGraph,
                     next_bb,
                     bb_frontier,
                     bb_analyzed,
                     instructions,
                     basic_blocks_labels,
                     dominators,
                     post_dominators,
                     goto_list,
                     variableFunctor,
                     codeFunc,
                     false,
                     loops_writer
                  );
               }
               else
               {
                  const bb_node_info* next_bi = CGET_BB_NODE_INFO(&bb_fcfgGraph, next_bb);
                  const unsigned int next_bb_number = next_bi->block->number;
                  THROW_ASSERT(basic_blocks_labels.find(next_bb_number) != basic_blocks_labels.end(), "I do not know the destination");
                  added_goto++;
                  indenter(*outStream, "   goto " + basic_blocks_labels.find(next_bb_number)->second + ";\n");
                  goto_list.insert(next_bb);
               }
            }
         }
      }
      while (*vIter++ != last_stmt);
      if (add_semicolon)
      {
         indenter(*outStream, ";\n");///added a fake indent
      }
   }
   else
   {

      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "is_there is false");
      /// in case we have phi nodes we check if some assignments should be printed
      if(add_phi_nodes_assignment)
      {
         indenter(*outStream, basic_block_tail.find(bb_number)->second);
         add_semicolon = false;
      }
      if (!BH->end_with_a_cond_or_goto(bi->block) && ((bi->statements_list.empty()) || ((GET_TYPE(cfgGraph, *bi->statements_list.rbegin()) & (TYPE_SWITCH | TYPE_WHILE | TYPE_FOR)) == 0)))
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Not end with a cond or goto nor switch");
         const vertex bbentry = bb_fcfgGraph.GetConstBBGraphInfo()->entry_vertex;
         if (current_vertex == bbentry) 
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Ended writing basic block " + boost::lexical_cast<std::string>(bb_number));
            return;
         }
         THROW_ASSERT(boost::out_degree(current_vertex, bb_fcfgGraph) <= 1, "only one edge expected BB(" + boost::lexical_cast<std::string>(bb_number) + ") Fun(" + boost::lexical_cast<std::string>(funId) + ")");
         OutEdgeIterator oE, oEEnd;
         for (boost::tie(oE, oEEnd) = boost::out_edges(current_vertex, bb_fcfgGraph); oE != oEEnd; oE++)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Examining the only? successor");
            vertex next_bb = boost::target(*oE, bb_fcfgGraph);
            if (bb_frontier.find(next_bb) != bb_frontier.end() or boost::in_degree(next_bb, bb_fcfgGraph) == 1)
               continue;
            ///Last basic block of a while/for loop
            if(FB_CFG_SELECTOR & bb_fcfgGraph.get_selector(*oE))
            {
               const vertex target = boost::target(*oE, bb_fcfgGraph);
               const bb_node_info * target_bi = CGET_BB_NODE_INFO(&bb_fcfgGraph, target);
               if(target_bi->statements_list.size() and (GET_TYPE(cfgGraph, *(target_bi->statements_list.rbegin())) & (TYPE_WHILE | TYPE_FOR)))
                  continue;
            }

            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Successor does not belong to frontier");
            const bb_node_info* next_bi = CGET_BB_NODE_INFO(&bb_fcfgGraph, next_bb);
            const unsigned int next_bb_number = next_bi->block->number;
            THROW_ASSERT(basic_blocks_labels.find(next_bb_number) != basic_blocks_labels.end(), "I do not know the destination");
            added_goto++;
            indenter(*outStream, "   goto " + basic_blocks_labels.find(next_bb_number)->second + ";\n");
            goto_list.insert(next_bb);
            add_semicolon = false;
         }
         if (add_semicolon)
         {
            indenter(*outStream, ";\n");///added a fake indent
         }
      }
      else if (add_semicolon)
         indenter(*outStream, "   ;\n");///added a fake indent
   }

   if (analyze_bb_PD)
   {
      //recurse on the post dominator
      bb_frontier.erase(bb_PD);
      THROW_ASSERT(bb_analyzed.find(bb_PD) == bb_analyzed.end(), "something of wrong happen " + boost::lexical_cast<std::string>(CGET_BB_NODE_INFO(&bb_fcfgGraph, bb_PD)->block->number) + " Fun(" + boost::lexical_cast<std::string>(funId) + ")");
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Printing the post dominator");
      writeRoutineInstructions_rec(bb_fcfgGraph, cfgGraph, bb_PD, bb_frontier, bb_analyzed, instructions, basic_blocks_labels, dominators, post_dominators, goto_list, variableFunctor, codeFunc, false, loops_writer);
   }
   if ((analyze_bb_PD || is_there || add_bb_label|| add_phi_nodes_assignment || add_phi_nodes_assignment_prefix) && bracket)
      indenter(*outStream, "}\n");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Ended writing basic block " + boost::lexical_cast<std::string>(bb_number));
}

void CWriter::writeRoutineInstructions
(
   const std::tr1::unordered_set<vertex>& instructions,
   var_pp_functorConstRef variableFunctor,
   const routineFunctorConstRef codefunc,
   vertex bb_start,
   std::set<vertex> bb_end,
   loop_writerRef loops_writer
)
{
   const behavioral_helperConstRef BH = codefunc->FB->get_helper();
   const graph *cfgGraph = codefunc->FB->get_graph(function_behavior::FCFG);
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->CWriter::writeRoutineInstructions - Start");
   if (instructions.empty())
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--CWriter::writeRoutineInstructions - instructions is an empty set");
      return;
   }
   else if (instructions.size() == 1)
   {
      if (GET_TYPE(cfgGraph, (*instructions.begin())) & TYPE_ENTRY)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--CWriter::writeRoutineInstructions - instructions is a set with only entry");
         return;
      }
   }
   /// compute the assignment introduced by the phi nodes destruction
   const bb_graph *bb_domGraph = codefunc->FB->get_bb_graph(function_behavior::DOM_TREE);
   const bb_graph *bb_fcfgGraph = codefunc->FB->get_bb_graph(function_behavior::FBB);
   std::tr1::unordered_set<unsigned int> phi_instructions;
   std::tr1::unordered_set<vertex>::const_iterator i_it_end = instructions.end();
   for(std::tr1::unordered_set<vertex>::const_iterator i_it = instructions.begin(); i_it != i_it_end; ++i_it)
   {
      if (GET_TYPE(cfgGraph, *i_it) & TYPE_PHI)
      {
         phi_instructions.insert(*(GET_NODEID(cfgGraph, *i_it).begin()));
      }
   }
   if(!phi_instructions.empty())
   {
      std::map<unsigned int, unsigned int> created_variables;
      std::map<unsigned int, std::string> symbol_table;
      std::map<unsigned int, std::deque<std::string> > array_of_stacks;
      insert_copies(bb_domGraph->GetConstBBGraphInfo()->entry_vertex, bb_domGraph, bb_fcfgGraph, variableFunctor, phi_instructions, created_variables, symbol_table, array_of_stacks);
      /// in case we declare the variableS introduced during the phi nodes destruction
      if(!created_variables.empty())
      {
         std::map<unsigned int, unsigned int>::const_iterator cv_it_end = created_variables.end();
         for(std::map<unsigned int, unsigned int>::const_iterator cv_it= created_variables.begin(); cv_it != cv_it_end; ++cv_it)
         {
            unsigned real_var = cv_it->second;
            indenter(*outStream, tree_helper::print_type(TM, tree_helper::get_type_index(TM, real_var)));
            indenter(*outStream, " " + symbol_table[cv_it->first] + ";\n");
         }
      }
   }
   ///Then I compute all the labels associated with a basic block with more than one entering edge.
   std::map<unsigned int, std::string> basic_blocks_labels;
   const dominance<bb_graph> * dominators = codefunc->FB->dominators;
   const dominance<bb_graph> * post_dominators = codefunc->FB->post_dominators;
   VertexIterator vi, vi_end;
   vertex bbentry;
   std::set<vertex> bb_exit;
   if (!bb_start)
      bbentry = bb_fcfgGraph->GetConstBBGraphInfo()->entry_vertex;
   else
      bbentry = bb_start;
   if (bb_end.empty())
   {
      bb_exit.insert(bb_fcfgGraph->GetConstBBGraphInfo()->exit_vertex);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "There are " + boost::lexical_cast<std::string>(bb_exit.size()) + " exit basic blocks");
   }
   else
   {
      bb_exit = bb_end;
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing labels");
   for (boost::tie(vi, vi_end) = boost::vertices(*bb_fcfgGraph); vi != vi_end; vi++)
   {
      size_t delta = bb_exit.find(*vi) != bb_exit.end() ? 1u : 0u;
      if (boost::in_degree(*vi, *bb_fcfgGraph) <= (1 + delta))
         continue;
      const bb_node_info* bi = CGET_BB_NODE_INFO(bb_fcfgGraph, *vi);
      unsigned int le = BH->start_with_a_label(bi->block);
      if (le)
         basic_blocks_labels[bi->block->number] = BH->get_label_name(le);
#if HAVE_TASKGRAPH_BUILT
      else if (GetPointer<const taskFunctor>(codefunc))
      {
         std::string task_label = GetPointer<const taskFunctor>(codefunc)->TG->getTaskProp(GetPointer<const taskFunctor>(codefunc)->task)->name;
         basic_blocks_labels[bi->block->number] = "BB_" + task_label + "_LABEL_" + boost::lexical_cast<std::string>(bi->block->number);
      }
#endif
      else
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Adding label to BB" + boost::lexical_cast<std::string>(bi->block->number));
         basic_blocks_labels[bi->block->number] = "BB_LABEL_" + boost::lexical_cast<std::string>(bi->block->number);
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed labels");
   ///set of basic block already analyzed
   std::set<vertex> bb_analyzed;
   bb_analyzed.insert(bb_exit.begin(), bb_exit.end());
   ///store for which basic block the goto has been used
   std::set<vertex> goto_list;
   ///basic block frontier over which writeRoutineInstructions_rec cannot go.
   std::set<vertex> bb_frontier;
   bb_frontier.insert(bb_exit.begin(), bb_exit.end());
   OutEdgeIterator oE, oEEnd;
   ///some statements can be in entry
   writeRoutineInstructions_rec(*bb_fcfgGraph, cfgGraph, bbentry,
                                bb_frontier, bb_analyzed, instructions, basic_blocks_labels, dominators,
                                post_dominators, goto_list, variableFunctor, codefunc, false, loops_writer);
   if (!bb_start && bb_end.size() == 0)
   {
      for (boost::tie(oE, oEEnd) = boost::out_edges(bbentry, *bb_fcfgGraph); oE != oEEnd; oE++)
      {
         if (bb_exit.find(boost::target(*oE, *bb_fcfgGraph)) != bb_exit.end())
            continue;
         else
         {
            writeRoutineInstructions_rec
                  (
                     *bb_fcfgGraph,
                     cfgGraph,
                     boost::target(*oE, *bb_fcfgGraph),
                     bb_frontier,
                     bb_analyzed,
                     instructions,
                     basic_blocks_labels,
                     dominators,
                     post_dominators,
                     goto_list,
                     variableFunctor,
                     codefunc,
                     false,
                     loops_writer
                     );
         }
      }
   }
   std::set<vertex> not_yet_considered;
   std::set_difference(goto_list.begin(), goto_list.end(), /*first set*/
                       bb_analyzed.begin(), bb_analyzed.end(), /*second set*/
                       std::inserter(not_yet_considered, not_yet_considered.begin()) /*result*/);
   while (!not_yet_considered.empty())
   {
      vertex next_bb = *not_yet_considered.begin();
      not_yet_considered.erase(next_bb);
      writeRoutineInstructions_rec(*bb_fcfgGraph, cfgGraph, next_bb, bb_frontier, bb_analyzed, instructions, basic_blocks_labels, dominators, post_dominators, goto_list, variableFunctor, codefunc, false, loops_writer);
      not_yet_considered.clear();
      std::set_difference(goto_list.begin(), goto_list.end(), /*first set*/
                          bb_analyzed.begin(), bb_analyzed.end(), /*second set*/
                          std::inserter(not_yet_considered, not_yet_considered.begin()) /*result*/);
   }
   const vertex exit = bb_fcfgGraph->GetConstBBGraphInfo()->exit_vertex;
   if (goto_list.find(exit) != goto_list.end() && basic_blocks_labels.find(bloc::EXIT_BLOCK_ID) != basic_blocks_labels.end())
   {
      indenter(*outStream, basic_blocks_labels.find(bloc::EXIT_BLOCK_ID)->second + ":\n");
   }
   basic_block_prefix.clear();
   basic_block_tail.clear();
   renaming_table.clear();
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--CWriter::writeRoutineInstructions - End");
}

void CWriter::declare_type(unsigned int varType, const behavioral_helperConstRef BH, std::tr1::unordered_set<std::string> & locally_declared_types)
{
#ifndef NDEBUG
   const std::string& routine_name = BH->get_function_name();
#endif

   const unsigned int real_var_type = tree_helper::GetRealType(TM, varType);
   const std::string type_name = tree_helper::name_type(TM, real_var_type);

   //Check that the variable really needs the declaration of a new type
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Declaration of type " + type_name + " " + boost::lexical_cast<std::string>(varType) + "(" + boost::lexical_cast<std::string>(real_var_type) + ") in function " + routine_name);

   if (globally_declared_types.find(type_name) == globally_declared_types.end() and locally_declared_types.find(type_name) == locally_declared_types.end())
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---This type has not been declared in this function");
      if(tree_helper::is_system(TM, varType))
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--System type");
         globally_declared_types.insert(type_name);
         return;
      }
      locally_declared_types.insert(type_name);
      const std::tr1::unordered_set<unsigned int> types_to_be_declared = tree_helper::GetTypesToBeDeclared(TM, real_var_type);
      std::tr1::unordered_set<unsigned int>::const_iterator type_to_be_declared, type_to_be_declared_end = types_to_be_declared.end();
      ///True if the declaration of this type has to be written before declaration of the types from which it depends
      bool anticipated_declaration = false;
      if (BH->is_named_pointer(real_var_type) and ((BH->is_a_struct(BH->get_pointed_type(real_var_type)) and not tree_helper::GetUnqualified(TM, BH->get_pointed_type(real_var_type))) or (BH->is_an_union(BH->get_pointed_type(real_var_type)) and not tree_helper::GetUnqualified(TM, BH->get_pointed_type(real_var_type)))))
      {
         anticipated_declaration = true;
      }
      ///typedecl struct ... without alignment
      if (tree_helper::is_a_struct(TM, real_var_type) and tree_helper::GetUnqualified(TM, real_var_type) and not tree_helper::IsAligned (TM, real_var_type) and tree_helper::GetTypesToBeDeclared(TM, real_var_type).size() == 1 and *(tree_helper::GetTypesToBeDeclared(TM, real_var_type).begin()) ==  tree_helper::GetUnqualified(TM, real_var_type))
      {
         anticipated_declaration = true;
      }

      if(not anticipated_declaration)
      {
        for(type_to_be_declared = types_to_be_declared.begin(); type_to_be_declared != type_to_be_declared_end; type_to_be_declared++)
        {
            this->declare_type(*type_to_be_declared, BH, locally_declared_types);
        }
      }
      if (tree_helper::HasToBeDeclared(TM, real_var_type))
      {
         if (this->verbose)
         {
            indenter(*outStream,  "//declaration of type " + boost::lexical_cast<std::string>(varType) + "(" + boost::lexical_cast<std::string>(real_var_type) + ")\n");
         }
         prettyPrintVertex::print_type_declaration(*outStream, real_var_type, indenter, BH);
      }
      if (anticipated_declaration)
      {
        for(type_to_be_declared = types_to_be_declared.begin(); type_to_be_declared != type_to_be_declared_end; type_to_be_declared++)
        {
            this->declare_type(*type_to_be_declared, BH, locally_declared_types);
        }
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Declared type " + boost::lexical_cast<std::string>(varType));
}

void CWriter::declare_variable
(
   unsigned int curVar,
   std::tr1::unordered_set<unsigned int> & already_declared_variables,
   std::tr1::unordered_set<std::string> & locally_declared_types,
   const behavioral_helperConstRef BH,
   const var_pp_functorConstRef varFunc
)
{
   if (already_declared_variables.find(curVar) != already_declared_variables.end())
      return;
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Declaring variable " + BH->PrintVariable(curVar));
   already_declared_variables.insert(curVar);

   std::tr1::unordered_set<unsigned int> initVars;
   std::tr1::unordered_set<unsigned int>::const_iterator initVarsIter, initVarsIterEnd;
   if(BH->GetInit(curVar, initVars))
   {
      for (initVarsIter = initVars.begin(), initVarsIterEnd = initVars.end(); initVarsIter != initVarsIterEnd; initVarsIter++)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "For variable " + boost::lexical_cast<std::string>(curVar) + " recursing on " + boost::lexical_cast<std::string>(*initVarsIter));
         if (already_declared_variables.find(*initVarsIter) == already_declared_variables.end() && globallyDeclVars.find(*initVarsIter) == globallyDeclVars.end())
         {
            declare_variable(*initVarsIter, already_declared_variables, locally_declared_types, BH, varFunc);
         }
      }
   }
   const unsigned int variable_type = tree_helper::get_type_index(TM, curVar);
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Type is " + boost::lexical_cast<std::string>(variable_type));
   declare_type(variable_type, BH, locally_declared_types);
   if (verbose)
      indenter(*outStream, "//declaring variable " + boost::lexical_cast<std::string>(curVar) + " - type: " + boost::lexical_cast<std::string>(BH->get_type(curVar)) + "\n");

   if (BH->get_attributes(curVar))
   {
      std::string tmp = BH->print_attributes(BH->get_attributes(curVar), varFunc);
      if (tmp.find("section") != std::string::npos)
         indenter(*outStream, "#pragma " + tmp + "\n");
   }
   BH->print_var_declaration(*outStream, curVar, varFunc, indenter, true);
   indenter(*outStream, ";\n");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Declared variable " + BH->PrintVariable(curVar));
}

void CWriter::writeCustomCode(std::ostringstream &code)
{
   indenter(*outStream, code.str());
}

loop_writer::loop_writer(const CWriterRef _writer, const InstructionWriterRef _instr_writer, simple_indent & _indenter, const var_pp_functorConstRef _variableFunctor, const ParameterConstRef _parameters) :
      writer(_writer),
      instr_writer(_instr_writer),
      variableFunctor(_variableFunctor),
      indenter(_indenter),
      parameters(_parameters)
{
   debug_level = parameters->get_class_debug_level(GET_CLASS(*this));
}

loop_writer::loop_writer(const CWriterRef _writer, const var_pp_functorConstRef _variableFunctor, const ParameterConstRef _parameters) :
      writer(_writer),
      instr_writer(_writer->instrWriter),
      variableFunctor(_variableFunctor),
      indenter(_writer->indenter),
      parameters(_parameters)
{
   debug_level = parameters->get_class_debug_level(GET_CLASS(*this));
}


loop_writer::~loop_writer()
{}

bool loop_writer::operator() (std::ostringstream &stream, const routineFunctorConstRef codeFunc, const unsigned int loop_id)
{
   const function_behaviorConstRef FB = codeFunc->FB;
   const bb_graph * fbb = FB->get_bb_graph(function_behavior::FBB);
   const LoopConstRef loop = FB->GetLoops()->GetLoop(loop_id);
   const vertex operation = CGET_BB_NODE_INFO(fbb, loop->GetHeader())->statements_list.back();
   instr_writer->write(stream, operation, indenter, variableFunctor, codeFunc);
   return false;
}

const InstructionWriterRef CWriter::getInstructionWriter() const
{
   return instrWriter;
}

void CWriter::WriteComment(const std::string& text)
{
   indenter(*outStream, "//" + text + "\n");
}

void CWriter::writeInclude(const std::string& file_name)
{
   if (file_name.find(".h") == std::string::npos || writtenIncludes.find(file_name) != writtenIncludes.end())
      return;
   writtenIncludes.insert(file_name);
   *outStream << "#include \"" << file_name << "\"" << std::endl;
}

void CWriter::addIncludes(const std::string& include)
{
   additionalIncludes.push_back(include);
}

std::vector<std::string> CWriter::getIncludes() const
{
   return additionalIncludes;
}

void CWriter::DeclareLocalVariables(std::tr1::unordered_set<unsigned int> & to_be_declared, std::tr1::unordered_set<unsigned int> & already_declared_variables, std::tr1::unordered_set<std::string> & already_declared_types, const behavioral_helperConstRef BH, const var_pp_functorConstRef varFunc)
{
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->Declaring local variables");
   std::tr1::unordered_set<unsigned int>::const_iterator var, var_end = to_be_declared.end();
   for (var = to_be_declared.begin(); var != var_end; var++)
   {
      if (BH->is_to_declare(*var))
      {
         declare_variable(*var, already_declared_variables, already_declared_types, BH, varFunc);
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--Declaring local variables");
}

void CWriter::insert_copies(vertex b, const bb_graph *bb_domGraph, const bb_graph *bb_fcfgGraph, var_pp_functorConstRef variableFunctor, const std::tr1::unordered_set<unsigned int>& phi_instructions, std::map<unsigned int, unsigned int> &created_variables, std::map<unsigned int, std::string> &symbol_table, std::map<unsigned int, std::deque<std::string> > &array_of_stacks)
{
   std::list<unsigned int> pushed;
   /// fill the renaming table for basic block b
   std::map<unsigned int, std::deque<std::string> >::const_iterator aos_it_end =array_of_stacks.end();
   for(std::map<unsigned int, std::deque<std::string> >::const_iterator aos_it = array_of_stacks.begin(); aos_it != aos_it_end; ++aos_it)
   {
      if(!aos_it->second.empty())
      {
         renaming_table[b][aos_it->first] = aos_it->second.back();
         symbol_table[aos_it->first] = aos_it->second.back();
      }
   }
   schedule_copies(b, bb_domGraph, bb_fcfgGraph, variableFunctor, phi_instructions, created_variables, symbol_table, pushed, array_of_stacks);
   OutEdgeIterator oi, oend;
   for (tie(oi, oend) = boost::out_edges(b, *bb_domGraph); oi != oend; ++oi)
   {
      vertex c = boost::target(*oi, *bb_domGraph);
      insert_copies(c, bb_domGraph, bb_fcfgGraph, variableFunctor, phi_instructions, created_variables, symbol_table, array_of_stacks);
   }
   pop_stack(pushed, array_of_stacks);
}

void CWriter::schedule_copies(vertex b, const bb_graph *bb_domGraph, const bb_graph *bb_fcfgGraph, var_pp_functorConstRef variableFunctor, const std::tr1::unordered_set<unsigned int>& phi_instructions, std::map<unsigned int, unsigned int> &created_variables, std::map<unsigned int, std::string> &symbol_table, std::list<unsigned int> &pushed, std::map<unsigned int, std::deque<std::string> > &array_of_stacks)
{
   /// Pass One: initialize the data structures
   const bb_node_info* bi = CGET_BB_NODE_INFO(bb_fcfgGraph, b);
   unsigned int bi_id = bi->block->number;

   std::set<std::pair<tree_nodeRef, tree_nodeRef> > copy_set;
   std::set<std::pair<tree_nodeRef, tree_nodeRef> > worklist;
   std::map<unsigned int, unsigned int> map;
   std::set<unsigned int> used_by_another;
   std::map<unsigned int, unsigned int> bb_dest_definition;
   OutEdgeIterator oi, oend;
   for (tie(oi, oend) = boost::out_edges(b, *bb_fcfgGraph); oi != oend; ++oi)
   {
      vertex s = boost::target(*oi, *bb_fcfgGraph);
      const bb_node_info* si = CGET_BB_NODE_INFO(bb_fcfgGraph, s);
      std::vector<tree_nodeRef>::const_iterator phi_op, phi_op_end = si->block->list_of_phi.end();
      for(phi_op = si->block->list_of_phi.begin(); phi_op != phi_op_end; ++phi_op)
      {
         if(phi_instructions.find(GET_INDEX_NODE(*phi_op)) == phi_instructions.end()) continue;
         gimple_phi * pn = GetPointer<gimple_phi>(GET_NODE(*phi_op));
         tree_nodeRef dest = pn->res;
         unsigned int dest_i = GET_INDEX_NODE(pn->res);
         bool is_virtual = pn->virtual_flag;
         if(!is_virtual)
         {
            bb_dest_definition[dest_i] = si->block->number;
            std::vector<std::pair< tree_nodeRef, unsigned int> >::iterator it_end = pn->list_of_def_edge.end();
            for (std::vector<std::pair< tree_nodeRef, unsigned int> >::iterator it = pn->list_of_def_edge.begin(); it != it_end; it++)
            {
               if(it->second == bi_id)
               {
                  tree_nodeRef src = it->first;
                  unsigned int src_i = GET_INDEX_NODE(it->first);
                  copy_set.insert(std::pair<tree_nodeRef, tree_nodeRef>(src, dest));
                  map[src_i] = src_i;
                  map[dest_i] = dest_i;
                  used_by_another.insert(src_i);
                  break;
               }
            }
         }
      }
   }

   /// Pass two: Set up the worklist of initial copies
   std::set<std::pair<tree_nodeRef, tree_nodeRef> >::iterator cs_it_end = copy_set.end();
   for(std::set<std::pair<tree_nodeRef, tree_nodeRef> >::iterator cs_it = copy_set.begin(); cs_it != cs_it_end; )
   {
      std::set<std::pair<tree_nodeRef, tree_nodeRef> >::iterator current_it = cs_it;
      ++cs_it;
      if(used_by_another.find(GET_INDEX_NODE(current_it->second)) == used_by_another.end())
      {
         worklist.insert(*current_it);
         copy_set.erase(current_it);
      }
   }

   /// Pass Three: Iterate over the worklist, inserting copies
   while(!worklist.empty() || !copy_set.empty())
   {
      std::set<std::pair<tree_nodeRef, tree_nodeRef> > worklist_restart;
      do
      {
         std::set<std::pair<tree_nodeRef, tree_nodeRef> >::iterator wl_it_end = worklist.end();
         for(std::set<std::pair<tree_nodeRef, tree_nodeRef> >::iterator wl_it = worklist.begin(); wl_it != wl_it_end; )
         {
            std::set<std::pair<tree_nodeRef, tree_nodeRef> >::iterator current_it = wl_it;
            ++wl_it;
            tree_nodeRef src = GET_NODE(current_it->first);
            unsigned int src_i = GET_INDEX_NODE(current_it->first);
            unsigned int dest_i = GET_INDEX_NODE(current_it->second);
            /// if dest \belongs live\_out(b)
            /// wrt the original algorithm an optimization has been added: in case b does not dominate any other node we can skip the creation of t
            if(boost::out_degree(b, *bb_domGraph) > 0 && bi->block->live_out.find(dest_i) != bi->block->live_out.end())
//            bool add_copy=false;
//            OutEdgeIterator o_it, o_it_end;
//            for (tie(o_it, o_it_end) = boost::out_edges(b, *bb_domGraph); o_it != o_it_end; ++o_it)
//            {
//               vertex tgt_bb = boost::target(*o_it, *bb_domGraph);
//               if(tgt_bb == b) continue;
//               const bb_node_info* tgt_bi = CGET_BB_NODE_INFO(bb_domGraph, tgt_bb);
//               if(tgt_bi && tgt_bi->block->live_in.find(dest_i) != tgt_bi->block->live_in.end())
//               {
//                  add_copy = true;
//               }
//            }
//            if(add_copy)
            {
               //THROW_ERROR("check the source code @" + STR(dest_i));
               ///   insert a copy from dest to a new temp t at phi-node defining dest
               unsigned int t_i = create_new_identifier(symbol_table);
               std::string copy_statement = symbol_table[t_i] + " = " + (*variableFunctor)(dest_i) + ";\n";
               //std::cerr << bi_id << " " << copy_statement;
               basic_block_prefix[bb_dest_definition[dest_i]] += copy_statement;
               created_variables[t_i] = dest_i;
               map[t_i] = t_i;
               ///   push(t, Stack[dest])
               push_stack(symbol_table[t_i], dest_i, pushed, array_of_stacks);
            }
            /// insert a copy operation from map[src] to dest at the end of b
            std::string copy_statement;
            if(symbol_table.find(map[src_i]) != symbol_table.end())
               copy_statement = (*variableFunctor)(dest_i)  + " = " + symbol_table[map[src_i]] + ";\n";
            else
               copy_statement += (*variableFunctor)(dest_i)  + " = " + (*variableFunctor)(map[src_i]) + ";\n";

            //std::cerr << bi_id << "-" << copy_statement;
            basic_block_tail[bi_id] += copy_statement;
            map[src_i] = dest_i;

            std::set<std::pair<tree_nodeRef, tree_nodeRef> >::iterator cs1_it_end = copy_set.end();
            for(std::set<std::pair<tree_nodeRef, tree_nodeRef> >::iterator cs1_it = copy_set.begin(); cs1_it != cs1_it_end; )
            {
               std::set<std::pair<tree_nodeRef, tree_nodeRef> >::iterator current1_it = cs1_it;
               ++cs1_it;
               if(src == current1_it->second)
               {
                  worklist_restart.insert(*current1_it);
                  copy_set.erase(current1_it);
               }
            }
            worklist.erase(current_it);
         }
         std::swap(worklist_restart, worklist);
      } while(!worklist.empty());

      std::set<std::pair<tree_nodeRef, tree_nodeRef> >::iterator cs2_it_end = copy_set.end();
      std::set<std::pair<tree_nodeRef, tree_nodeRef> >::iterator cs2_it = copy_set.begin();
      if(cs2_it != cs2_it_end)
      {
         std::set<std::pair<tree_nodeRef, tree_nodeRef> >::iterator current2_it = cs2_it;
         unsigned int dest_i = GET_INDEX_NODE(current2_it->second);
         /// check if dest_i is source of any other pair in copy_set
         /// this optimization is not described in the original algorithm
         ++cs2_it;
         bool add_temporary = false;
         for(; cs2_it != cs2_it_end; ++cs2_it)
         {
            if(GET_INDEX_NODE(cs2_it->first) == dest_i)
            {
               add_temporary = true;
               break;
            }
         }
         if(add_temporary)
         {
            /// create a new symbol
            unsigned int t_i = create_new_identifier(symbol_table);
            /// insert a copy from dest to a new temp t at the end of b
            std::string copy_statement = symbol_table[t_i] + " = " + (*variableFunctor)(dest_i) + ";\n";
            //std::cerr << bi_id << " " << copy_statement;
            basic_block_tail[bi_id] += copy_statement;
            created_variables[t_i] = dest_i;
            map[t_i] = t_i;
            map[dest_i] = t_i;
         }
         worklist.insert(*current2_it);
         copy_set.erase(current2_it);
      }
   }
}

unsigned int CWriter::create_new_identifier(std::map<unsigned int, std::string> &symbol_table)
{
   unsigned int counter = 0;
   unsigned int node_id_this = 0;
   std::string new_name;
   do
   {
      new_name = "__t__"+boost::lexical_cast<std::string>(fake_max_tree_node_id)+"_"+boost::lexical_cast<std::string>(counter);
      node_id_this = TM->find_identifier_nodeID(new_name);
      counter++;
   } while(node_id_this>0);
   symbol_table[fake_max_tree_node_id] = new_name;
   node_id_this = fake_max_tree_node_id;
   ++fake_max_tree_node_id;
   return node_id_this;
}

void CWriter::push_stack(std::string symbol_name, unsigned int dest_i, std::list<unsigned int> &pushed, std::map<unsigned int, std::deque<std::string> > &array_of_stacks)
{
   array_of_stacks[dest_i].push_back(symbol_name);
   pushed.push_back(dest_i);
}

void CWriter::pop_stack(std::list<unsigned int> &pushed, std::map<unsigned int, std::deque<std::string> > &array_of_stacks)
{
   std::list<unsigned int>::iterator p_it_end = pushed.end();
   for(std::list<unsigned int>::iterator p_it = pushed.begin(); p_it != p_it_end;)
   {
      std::list<unsigned int>::iterator current = p_it;
      ++p_it;
      THROW_ASSERT(array_of_stacks.find(*current) != array_of_stacks.end(), "the array of stacks is inconsistent");
      array_of_stacks.find(*current)->second.pop_back();
      pushed.erase(current);
   }
}
