/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file c_writer.hpp
 * @brief This file contains the routines necessary to create a C executable program
 *
 * This file contains the routines necessary to create a C executable program starting from a behavioral specification
 *
 * @author Luca Fossati <fossati@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

#ifndef CWRITER_HPP
#define CWRITER_HPP

///Graph include
#include "graph.hpp"

///STL include
#include <deque>
#include <list>
#include <map>
#include <set>
#include <tr1/unordered_map>
#include <tr1/unordered_set>
#include <vector>

///Utility include
#include <fstream>
#include "refcount.hpp"
#include <iostream>
#include <ostream>
#include <string>

CONSTREF_FORWARD_DECL(call_graph);
CONSTREF_FORWARD_DECL(application_manager);
CONSTREF_FORWARD_DECL(behavioral_helper);
REF_FORWARD_DECL(CWriter);
CONSTREF_FORWARD_DECL(function_behavior);
CONSTREF_FORWARD_DECL(InstructionWriter);
REF_FORWARD_DECL(InstructionWriter);
REF_FORWARD_DECL(Loop);
REF_FORWARD_DECL(machine_node);
REF_FORWARD_DECL(ProgModelBackend);
CONSTREF_FORWARD_DECL(Parameter);
CONSTREF_FORWARD_DECL(tree_manager);
CONSTREF_FORWARD_DECL(var_pp_functor);

class bb_graph;
class graph;
class instrumented_call_instr_writer;

template <typename Graph> class dominance;

#include "simple_indent.hpp"

struct routineFunctor
{
   const application_managerConstRef AppM;
   const function_behaviorConstRef FB;
   routineFunctor(const application_managerConstRef _AppM, const function_behaviorConstRef _FB) : AppM(_AppM), FB(_FB) {}
   virtual ~routineFunctor() {}
};
typedef refcount<routineFunctor> routineFunctorRef;
typedef refcount<const routineFunctor> routineFunctorConstRef;

/**
 * Functor used to print loop
 */
struct loop_writer
{
   protected:

      ///Contains the class used to write
      const CWriterRef writer;

      ///Contains the class used to write instructions
      const InstructionWriterRef instr_writer;

      ///the variable functor
      const var_pp_functorConstRef variableFunctor;

      ///the indenter
      simple_indent & indenter;

      ///The set of input parameters
      const ParameterConstRef parameters;

      ///the debug level
      int debug_level;

   public:

      /**
       * Constructor
       * @param instr_writer is the instruction writer
       * @param identer is the indenter
       * @param var_pp_functor is the variable functor
       */
      loop_writer(const CWriterRef writer, const InstructionWriterRef instr_writer, simple_indent & indenter, const var_pp_functorConstRef variableFunctor, const ParameterConstRef parameters);

      loop_writer(const CWriterRef writer, const var_pp_functorConstRef variableFunctor, const ParameterConstRef parameters);

      /**
      * Destructor
      */
      virtual ~loop_writer();

      /**
       * Operator used to print loop header and body
       * @param stream is the stream where loop is printed
       * @param codeFunct is the function functor
       * @param loop_is is the index of the loop
       * @return true if body is printed
      */
      virtual bool operator() (std::ostringstream &stream, const routineFunctorConstRef codeFunc, const unsigned int loop_id);

};

typedef refcount<loop_writer> loop_writerRef;


/**
 * Class use to write the C code representing a parallelized program,
 * i.e. a program composed of many tasks; this class can't be directly
 * instantiated since the backend is not specified yet. So a subclass must
 * be created which redefines all the abstract methods of this class.
 */
class CWriter
{
   private:
      friend class loop_writer;

   protected:

      ///the manager of the application
      const application_managerConstRef AppM;

      ///The tree manager
      const tree_managerConstRef TM;

      ///Represents the stream we are currently writing to
      std::ostringstream * outStream;
      std::ostringstream * outStream_h;

      ///Function used to indent the printed code
      simple_indent indenter;

      ///goto added during printing
      unsigned int added_goto;

      ///This set contains the list of the non built_in types already declared in the global scope
      std::tr1::unordered_set<std::string> globally_declared_types;

      std::tr1::unordered_set<unsigned int> globallyDeclVars;

      ///Contains the class used to write instructions
      InstructionWriterRef instrWriter;

      ///Verbosity means that a comment is printed for each line in the output file
      bool verbose;

      ///set of parameters
      const ParameterConstRef Param;

      ///the debug level
      int debug_level;

      ///the output level
      int output_level;

      unsigned int fake_max_tree_node_id;

      /// string to be printed at the beginning of a given basic block
      std::map<unsigned int, std::string> basic_block_prefix;

      /// string to be printed at the end of a given basic block
      std::map<unsigned int, std::string> basic_block_tail;

      /// renaming table used by phi node destruction procedure
      std::map<vertex, std::map<unsigned int, std::string> > renaming_table;

      /**
       * Write recursively instructions belonging to a basic block of task or of a function
       * @param bb_graph is the basic block control flow graph with feedback
       * @param cfgGraph is the operation control flow graph
       * @param current_vertex is the basic block which is being printed
       * @param bb_frontier is ???
       * @param bb_analyzed are the basic block already printed for this task/function
       * @param instructions are the instructions which belongs to this task/function
       * @param basic_blocks_labels is ???
       * @param dominators is the dominators tree
       * @param post_dominators is the post dominators tree
       * @param goto_list is ???
       * @param variableFunctor is the pretty functor
       * @param bracket tells if bracket should be added before and after this basic block
       * @param loops_writer is the functor used to print the loop
       */
      virtual void writeRoutineInstructions_rec
      (
         const bb_graph & bb_fcfgGraph,
         const graph * cfgGraph,
         vertex current_vertex,
         std::set<vertex> &bb_frontier,
         std::set<vertex> &bb_analyzed,
         const std::tr1::unordered_set<vertex>& instructions,
         const std::map<unsigned int, std::string> &basic_blocks_labels,
         const dominance<bb_graph> * dominators,
         const dominance<bb_graph> * post_dominators,
         std::set<vertex> & goto_list,
         var_pp_functorConstRef variableFunctor,
         const routineFunctorConstRef codefunc,
         bool bracket,
         loop_writerRef loops_writer
      );

      ///Mapping between measure region name and measure region identifier
      ///if analysis level is all, there is only a key ALL
      ///if analysis level is source, there are only two keys SOURCE and SYSTEM
      ///If analysis level is component, key is the componenent
      ///if analysis level is function, key is the function
      std::tr1::unordered_map<std::string, unsigned long long> measure_region_identifiers;

      std::vector<std::string> additionalIncludes;
      std::set<std::string> writtenIncludes;

      /**
       * Declares the local variable; in case the variable used in the intialization of
       * curVar hasn't been declared yet it get declared
       * @param curVar is the variable to be declared
       * @param already_decl_variables is the set of already declared variables
       * @param locally_declared_type is the set of already declared types
       * @param routine_name is the name of the routine (function or thread)
       * @param funId is the index of the function
       * @param varFunc is the printer functor
       */
      virtual void declare_variable
      (
         unsigned int curVar,
         std::tr1::unordered_set<unsigned int> & already_declared_variables,
         std::tr1::unordered_set<std::string> & locally_declared_type,
         const behavioral_helperConstRef BH,
         const var_pp_functorConstRef varFunc
      );

      /**
       * Compute the copy assignments needed by the phi nodes destruction
       * Further details can be found in:
       * - Preston Briggs, Keith D. Cooper, Timothy J. Harvey and L. Taylor Simpson,
       *   "Practical Improvements to the Construction and Destruction of Static Single Assignment Form",
       *   Software -- Practice and Experience 1998
       */
      void insert_copies(vertex b, const bb_graph *bb_domGraph, const bb_graph *bb_fcfgGraph, var_pp_functorConstRef variableFunctor, const std::tr1::unordered_set<unsigned int>& phi_instructions, std::map<unsigned int, unsigned int> &created_variables, std::map<unsigned int, std::string> & symbol_table, std::map<unsigned int, std::deque<std::string> > &array_of_stacks);

      /**
       * insert copies according the algorithm described in Briggs et. al.
       */
      void schedule_copies(vertex b, const bb_graph *bb_domGraph, const bb_graph *bb_fcfgGraph, var_pp_functorConstRef variableFunctor, const std::tr1::unordered_set<unsigned int>& phi_instructions, std::map<unsigned int, unsigned int> &created_variables, std::map<unsigned int, std::string> & symbol_table, std::list<unsigned int> &pushed, std::map<unsigned int, std::deque<std::string> > &array_of_stacks);

      /**
       * create an identifier for the temporaries created by phi node destruction
       * @param symbol_table is the symbol table where the new id is inserted
       * return an id for the symbol created
       */
      unsigned int create_new_identifier(std::map<unsigned int, std::string> &symbol_table);

      /**
       * push on the stack of temporary variables that has to replaced
       * @param symbol_name is the new identifier for dest_i
       * @param dest_i is the id of the variable to be renamed
       * @param pushed is the list of variables to be renamed
       * @param array_of_stacks is the array of stacks used by the phi node destruction procedure
       */
      void push_stack(std::string symbol_name, unsigned int dest_i, std::list<unsigned int> &pushed, std::map<unsigned int, std::deque<std::string> > &array_of_stacks);

      /**
       * remove from the stack all the temporaries
       * @param pushed is the list of variables to be renamed
       * @param array_of_stacks is the array of stacks used by the phi node destruction procedure
       */
      void pop_stack(std::list<unsigned int> &pushed, std::map<unsigned int, std::deque<std::string> > &array_of_stacks);

   public:
      /**
       * Constructor of the class
       * @param AppM is the manager of the application
       * @param Param is the set of parameters
       * @param verbose tells if annotations
       */
      CWriter(const application_managerConstRef _AppM, const ParameterConstRef Param, bool verbose = true);

      /**
       * Destructor
       */
      virtual ~CWriter();

      /**
       * Writes the final C file
       * @param fileName is the name of the file to be generated
       */
      virtual void writeFile(const std::string& fileName);

      /**
       * Adds the instruction_writer
       */
      virtual void add_instruction_writer(const InstructionWriterRef instrWriter);

      /**
       * Returns the instruction writer
       * @return the instruction writer
       */
      const InstructionWriterRef getInstructionWriter() const;

      /**
       * Writes a string into the output file; this is used to add custom code
       * to the program; note that the string is automatically indented, there is
       * no need to manyally insert tabulation spaces
       */
      virtual void writeCustomCode(std::ostringstream &code);

      /**
       * Writes the body of the function to the specified stream
       * @param funId is the index of the function
       */
      virtual void StartFunctionBody(const unsigned int funId);

      /**
       * Writes the code necessary to close a function (this function was a function
       * also present in the original specification)
       */
      virtual void endFunctionBody(unsigned int funId, var_pp_functorConstRef taskFunctor);

      /**
       * Writes the instructions of the current routine, being it a task or a function of the original program.
       * @param funId is the index of the function
       * @param instructions is the instructions which have to be printed
       * @param var_pp_functor is the variable functor
       * @param routineName is the name of the routine
       * @param task_label is the label which has to be added to bb label
       * @param bb_start is the first basic block to be printed
       * @param bb_end is the set of first basic block not to be printed
       * @param loops_writer is the functor used to print the loop
       * @param backend is the backend used to print task graph manager
       */
      virtual void writeRoutineInstructions
      (
         const std::tr1::unordered_set<vertex>& instructions,
         var_pp_functorConstRef variableFunctor,
         const routineFunctorConstRef codefunc,
         vertex bb_start = 0,
         std::set<vertex> bb_end = std::set<vertex>(),
         loop_writerRef loops_writer = loop_writerRef()
      );

      /**
       * Writes the declaration of the function whose id in the tree is funId
       */
      virtual void WriteFunctionDeclaration(const unsigned int funId);

      /**
       * Writes the header of the file
       */
      virtual void writeHeader();

      /**
       * Writes the global declarations
       */
      virtual void WriteGlobalDeclarations();

      /**
       * Writes an include directive
       * @param file_name the name of the header file to be included
       */
      virtual void writeInclude(const std::string& file_name);

      /**
       * Writes a comment
       * @param comment is the string to be printed
       */
      virtual void WriteComment(const std::string& comment);

      /**
       * This method should be called only if the type associated with the variable
       * is a non built_in type and, in case the non built_in type hasn't been
       * declared yet, it declares it; this method is used to declared new types
       * with scope limited to a routine (be it a function of a task)
       * @param varType the index of the type to be declared
       * @param BH the behavioral helper associated to the function which declare type
       * @param locally_declared_type is the set of type already declared in this function
       * @param routine_name is the name of the routina (function or thread)
       */
      virtual void declare_type(unsigned int varType, const behavioral_helperConstRef BH, std::tr1::unordered_set<std::string> & locally_declared_type);

      /**
       * Declares the local variable; in case the variable used in the intialization of
       * curVar hasn't been declared yet it get declared
       * @param to_be_declared is the set of variables which have to be declared 
       * @param already_decl_variables is the set of already declared variables
       * @param locally_declared_type is the set of already declared types
       * @param helper is the behavioral helper associated with the function
       * @param varFunc is the printer functor
       */
      virtual void DeclareLocalVariables
      (
         std::tr1::unordered_set<unsigned int> & to_be_declared,
         std::tr1::unordered_set<unsigned int> & already_declared_variables,
         std::tr1::unordered_set<std::string> & locally_declared_type,
         const behavioral_helperConstRef BH,
         const var_pp_functorConstRef varFunc
      );

      virtual void addIncludes(const std::string& include);

      virtual std::vector<std::string> getIncludes() const;

      /// return the stream
      std::ostringstream* getoutStream() {return outStream;}

      /**
       * Declares the types of the parameters of a function
       * @param fun_id is the index of the function
       */
      void DeclareFunctionTypes(const unsigned int);
};

typedef refcount<CWriter> CWriterRef;

#endif
