/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file HDL_manager.hpp
 * @brief This class writes different HDL based descriptions (VHDL, Verilog, SystemC) starting from a structural representation.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef HDL_MANAGER_HPP
#define HDL_MANAGER_HPP

///Autoheader include
#include "config_HAVE_FLOPOCO.hpp"

#include <ostream>
#include <string>
#include <vector>
#include <list>

#include "refcount.hpp"
#include "language_writer.hpp"

/**
 * @name Forward declarations.
 */
//@{
REF_FORWARD_DECL(HDL_manager);
REF_FORWARD_DECL(structural_object);
REF_FORWARD_DECL(language_writer);
REF_FORWARD_DECL(target_device);
#if HAVE_FLOPOCO
REF_FORWARD_DECL(flopoco_wrapper);
#endif
CONSTREF_FORWARD_DECL(technology_manager);
CONSTREF_FORWARD_DECL(Parameter);
//@}

class HDL_manager
{
   private:

      ///reference to the target device
      const target_deviceRef device;

      ///reference to the class containing all the technology information
      const technology_managerConstRef TM;

#if HAVE_FLOPOCO
      ///wrapper to the FloPoCo library
      flopoco_wrapperRef flopo_wrap;
#endif

      ///reference to the class containing all the parameters
      const ParameterConstRef Param;

      ///debugging level of the class
      int debug_level;

      ///list of files generated by the default writer
      std::list<std::string> HDL_files;

      ///list of files generated by the auxiliary writers
      std::list<std::string> HDL_aux_files;

      /**
       * Returns the list of components that have a structural-based description.
       * This list of components is relative sorted such that if a component C_i uses a component C_j then C_j is before C_i.
       * To obtain this list the hierarchy is visited in a post-order fashion.
       * @param lan is the chosen language writer object.
       * @param cir is the structural object under analysis.
       * @param list_of_com is the list of components.
       */
      void get_post_order_structural_components(const structural_objectRef &cir, std::list<structural_objectRef> &list_of_com) const;

      /**
       * Generates the HDL description for the given components in the specified language
       */
      std::string write_components(std::string filename, language_writer::language lan, const std::string& top, const std::list<structural_objectRef>& components, bool equation) const;

      /**
       * Determines the proper language for each component and generates the corresponding HDL descriptions
       */
      void write_components(std::string filename, const std::string& top, const std::list<structural_objectRef>& components, bool equation);

      /**
       * Writes the module description.
       * @param lan is the chosen language writer object.
       * @param os is the output stream.
       * @param cir is the module to be written. The analysis does not consider the inner objects but just one level of the hierarchy.
       */
      void write_module(std::ostream& os, const language_writerRef writer, const structural_objectRef cir, bool equation) const;

      /**
       * Writes the FloPoCo module description to a VHDL file.
       * @param cir is the module to be fixed.
       */
      void write_flopoco_module(const structural_objectRef &cir) const;

      /**
       * Writes signal port connection post fix.
       * @param lan is the chosen language writer object.
       * @param os is the output stream.
       * @param po is the primary port.
       * @param lspf is true when the first post is written
       */
      void io_signal_fix_ith(std::ostream& os, const language_writerRef writer, const structural_objectRef po, bool& lspf) const;

      /**
       * Returns true if the module has a FSM description associated with, false otherwise.
       * @param cir is the module.
       */
      bool is_fsm(const structural_objectRef &cir) const;

      /**
       * Writes a mealy/moore finite state machine behavioral description.
       * @param os is the output stream.
       * @param writer is the chosen language writer object.
       * @param cir is the module.
       * @param fsm_desc is the string-based FSM description.
       */
      void write_fsm(std::ostream& os, const language_writerRef writer, const structural_objectRef &cir, const std::string &fsm_desc) const;

      /**
       * Writes the behavioral description associated with the component
       * @param os is the output stream.
       * @param writer is the chosen language writer object.
       * @param cir is the module.
       * @param behav is the string-based behavioral description.
       */
      void write_behavioral(std::ostream& os, const language_writerRef writer, const structural_objectRef &cir, const std::string &behav) const;

   public:

      /**
       * Constructor
       * @param device is the data structure containing information about the target device
       * @param Param is the data structure containing all the parameters
       */
      HDL_manager(const target_deviceRef device, const ParameterConstRef Param);

      /**
       * Destructor
       */
      ~HDL_manager();

      /**
       * Generates HDL code.
       * @param filename is the output file name.
       * @param cir is the structural object representing the component to be generated
       * @param equation specifies if the equation-based version of the component has to be generated
       */
      void hdl_gen(std::string filename, const structural_objectRef cir, bool equation = false);

      /**
       * Generates HDL code for a specific set of components
       * @param filename is the output file name
       * @param top is the name of the top component
       * @param cirs is the list of components to be written
       * @param hierarchical is a flag to specify if the analysis has to be hierarchically performed
       * @param equation specifies if the equation-based version of the component has to be generated
       */
      void hdl_gen(std::string filename, const std::string& top, const std::vector<structural_objectRef> &cirs, bool hierarchical, bool equation = false);

      /**
       * Returns the list of generated files
       */
      std::string get_HDL_files() const;

      /**
       * Returns the list of auxiliary files separated by ";"
       */
      std::string get_aux_files() const;

      /**
       * Converts a generic string to a language compliant identifier
       */
      static
      std::string convert_to_identifier(const language_writer* lan, const std::string &id);

      /**
       * Converts a generic string to a language compliant identifier
       */
      static
      std::string convert_to_identifier(const std::string &id);

      /**
       * Returns the module typename taking into account even the flopoco customizations
       * @param lan is the chosen language writer object.
       * @param cir is the module.
       */
      static
      std::string get_mod_typename(const language_writer* lan, const structural_objectRef& cir);

};
///refcount definition of the class
typedef refcount<HDL_manager> HDL_managerRef;
#endif
