/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file bambu.cpp
 * @brief High level Synthesis tool.
 *
 * Main file used to perform high-level synthesis starting from a C-based specification.
 * See \ref src_bambu for further informations
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/

///Autoheader include
#include "config_PACKAGE_VERSION.hpp"
#include "config_HAVE_ALTERA.hpp"
#include "config_HAVE_EXPERIMENTAL.hpp"
#include "config_HAVE_MODELSIM.hpp"
#include "config_HAVE_VCD_BUILT.hpp"
#include "config_HAVE_XILINX.hpp"

#include <iostream>
#include <cstdlib>

/// ----------- Frontend ----------- ///
/// Frontend Analysis
#include "BaseAnalysis.hpp"
#include "BambuFrontendFlow.hpp"

/// ----------- Intermediate Representation ----------- ///
/// Intermediate Representation Parsing
#include "parse_tree.hpp"
/// Intermediate Representation Datastructure
#include "tree_manager.hpp"
/// Intermediate Representation Management
#include "application_manager.hpp"
#include "call_graph.hpp"
#include "call_graph_helper.hpp"
#include "function_behavior.hpp"
#include "behavioral_helper.hpp"
#include "tree_node.hpp"
#include "hls.hpp"
#include "hls_target.hpp"

/// ----------- C backend ----------- ///
#include "c_backend.hpp"
#include "c_writer.hpp"
#include "instruction_writer.hpp"
#if HAVE_EXPERIMENTAL
#include "memory_profile_c_writer.hpp"
#include "memory_profile_instruction_writer.hpp"
#endif
/// ----------- HDL backend ----------- ///
#include "HDL_manager.hpp"
#include "language_writer.hpp"
#include "structural_manager.hpp"

#include "HLStestbench.hpp"

/// ----------- Resource Library ----------- ///
/// Resource Library Parsing
#include "parse_technology.hpp"
/// Resource Library Datastructure
#include "technology_manager.hpp"
#include "technology_builtin.hpp"
/// technology information
#include "target_technology.hpp"
#include "target_device.hpp"
#include "time_model.hpp"
/// Memories
#include "memory_allocation.hpp"
#include "memory.hpp"

/// ----------- High-Level Synthesis Constraints ----------- ///
/// HLS Constraints Parsing
#include "HLS_constraintsIO.hpp"
/// HLS Constraints Datastructure
#include "hls_constraints.hpp"

/// ----------- vcd debug ----------- ///
#if HAVE_XILINX && HAVE_MODELSIM && HAVE_VCD_BUILT
#include "vcd_data.hpp"
#include "vcd_parser.hpp"
#include "vcd_utility.hpp"
#include "vcd_debug_c_writer.hpp"
#include "vcd_debug_instruction_writer.hpp"
#endif

/// ----------- High-Level Synthesis Flow ----------- ///
#include "hls_flow.hpp"

/// ----------- Controller data Structures ----------- ///
#include "controller_creator.hpp"

/// ----------- Generation of interfaces ----------- ///
#if HAVE_EXPERIMENTAL
#include "module_interface.hpp"
#endif
#include "top_entity.hpp"

/// ----------- Scripts generation ----------- ///
#include "BackendFlow.hpp"

/// ----------- General PandA includes ----------- ///
#include "dbgPrintHelper.hpp"
#include "utility.hpp"
#include "exceptions.hpp"
#include "cpu_time.hpp"

/// ----------- Parameters Management ----------- ///
#include "BambuParameter.hpp"
#include "constant_strings.hpp"

///GCC wrapper
#include "gcc_wrapper.hpp"

#include "global_variables.hpp"

void createCallGraph(const ParameterRef parameters, const application_managerRef AppM, const HLS_targetRef HLS_T, const HLS_constraintsRef HLS_C)
{
#ifndef NDEBUG
   int debug_level = parameters->getOption<int>(OPT_debug_level);
#endif
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, " ==== Starting Initialization of Behavioral Specification ====");
   std::tr1::unordered_map<FrontendAnalysisType, unsigned int> Transformations;
   Transformations[SSA_DATA_FLOW_ANALYSIS] = true;
   Transformations[CONTROL_DEPENDENCE_COMPUTATION] = true;
   Transformations[ANTI_DEP_COMPUTATION] = true;
   Transformations[LOOP_REGIONS_COMPUTATION] = false;
   Transformations[SPLIT_PHINODES] = false;
   Transformations[LOOPS_ANALYSIS] = false;
   Transformations[ADD_OP_FLOW_EDGES] = true;
   Transformations[CHECK_SYSTEM_TYPE] = true;

   if (parameters->getOption<bool>(OPT_speculative))
      Transformations[SPECULATION_EDGES_COMPUTATION] = true;
#if 0
   if (parameters->getOption<bool>("compute_unrolling_degree"))
      Transformations[UNROLLING_DEGREE] = true;
   if (parameters->getOption<bool>("insert_verification_operation"))
   {
      Transformations[VERIFICATION_OPERATION] = true;
      Transformations[SPLIT_PHINODES] = true;
   }
#endif
#if HAVE_EXPERIMENTAL
   if (parameters->getOption<int>(OPT_controller_architecture) == controller_creator::PARALLEL_CONTROLLER)
   {
      Transformations[EXTENDED_PDG_COMPUTATION] = true;
      if(parameters->isOption(OPT_chaining) and parameters->getOption<bool>(OPT_chaining))
         Transformations[PARALLEL_REGIONS_GRAPH_COMPUTATION] = true;
   }
   if(parameters->isOption("pdg-reduction") and parameters->getOption<bool>("pdg-reduction"))
      Transformations[REDUCED_PDG_COMPUTATION] = true;
      
#endif
   Transformations[DETERMINE_MEMORY_ACCESSES] = true;
   Transformations[NI_SSA_LIVENESS] = true;

   FrontendFlowRef fendFlow = FrontendFlowRef(new BambuFrontendFlow(parameters, AppM, HLS_T, HLS_C, Transformations));
   ///execution of the flow
   fendFlow->exec();
}

void writeStructuralDescription(const ParameterRef& Param, const HDL_managerRef& HM, const structural_managerRef& SM)
{
   std::string file_name;
   if (Param->getOption<bool>(OPT_datapath) and !Param->getOption<bool>(OPT_controller))
   {
      file_name = Param->getOption<std::string>(OPT_datapath_file);
   }

   if (!Param->getOption<bool>(OPT_datapath) and Param->getOption<bool>(OPT_controller))
   {
      file_name = Param->getOption<std::string>(OPT_controller_file);
   }

   if (Param->getOption<bool>(OPT_datapath) and Param->getOption<bool>(OPT_controller))
   {
      file_name = Param->getOption<std::string>(OPT_top_file);
   }
   HM->hdl_gen(file_name, SM->get_circ());
}

void generateBackendScripts(const ParameterRef Param, const hlsRef HLS)
{
   int output_level = Param->getOption<int>(OPT_output_level);

   structural_managerRef SM;
   if (Param->getOption<bool>(OPT_datapath) and Param->getOption<bool>(OPT_controller))
   {
      SM = HLS->top;
   }
   else if (Param->getOption<bool>(OPT_datapath))
   {
      SM = HLS->datapath;
   }
   else if (Param->getOption<bool>(OPT_controller))
   {
      SM = HLS->controller;
   }
   if (!SM) return;

   const target_deviceRef device = HLS->HLS_T->get_target_device();

   long cpu_time;
   START_TIME(cpu_time);
   //Finally, the whole structural description is dumped
   HDL_managerRef HM = HDL_managerRef(new HDL_manager(device, Param));
   writeStructuralDescription(Param, HM, SM);
   STOP_TIME(cpu_time);

   PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Structural description written (" + print_cpu_time(cpu_time) + " seconds) in the following files:");
   std::string aux_file_string = HM->get_aux_files();
   if (aux_file_string.size() > 0)
   {
      std::vector<std::string> aux_files = convert_string_to_vector<std::string>(aux_file_string, ";");
      for(unsigned int a = 0; a < aux_files.size(); a++)
      {
         PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, " - Auxiliary file: \"" + aux_files[a] + "\"");
      }
   }
   std::string HDL_file_string = HM->get_HDL_files();
   std::vector<std::string> HDL_files = convert_string_to_vector<std::string>(HDL_file_string, ";");
   for(unsigned int f = 0; f < HDL_files.size(); f++)
   {
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, " - HDL file: \"" + HDL_files[f] + "\"");
   }
   PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "");

   ///generate the synthesis scripts
   BackendFlowRef flow = BackendFlow::CreateFlow(Param, "Synthesis", device);
   std::string script_file = flow->GenerateSynthesisScripts(SM, HM);
   PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Generated the synthesis script: \"" + script_file + "\"\n");

   ///determine if the synthesis has to be executed or not
   if (Param->getOption<bool>(OPT_perform_synthesis))
   {
      flow->ExecuteSynthesis();
   }

   if (Param->getOption<bool>(OPT_generate_testbench) && Param->isOption("tb_inputs"))
   {
      const tree_managerRef TreeM = HLS->AppM->get_tree_manager();
      double clock_period_value = HLS->HLS_C->get_clock_period();
      HLStestbenchRef hls_tb(new HLStestbench(Param));

      ///creating testbench
      std::string bench_name = "testbench_" + SM->get_circ()->get_id();
      std::string filename_bench = hls_tb->create_testbench(bench_name, SM->get_circ(), clock_period_value, false, TreeM);
      if (filename_bench.size() == 0)
         THROW_ERROR("Testbench cannot be created!");

      ///generate the simulation files
      std::string simulation_script = flow->GenerateSimulationScripts(SM->get_circ()->get_id(), filename_bench, HM, "_beh");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Generated the simulation script: \"" + simulation_script + "\"\n");

      if (Param->getOption<bool>(OPT_simulate))
      {
         ///perform the simulation
         unsigned long long int average_execution = flow->ExecuteSimulation();
         PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Simulation(s) completed! Average execution: " + boost::lexical_cast<std::string>(average_execution) + " cycles");
         PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "");
      }
   }
}

/**
 * Main file used to perform high-level synthesis starting from a C specification.
 * @anchor MainBambu
 * @param argc is the number of arguments
 * @param argv is the array of arguments passed to the program.
 */
int main(int argc, char *argv[])
{
   srand(static_cast<unsigned int>(time(NULL)));

   // General options register
   ParameterRef parameters;

   try
   {
      // ---------- Initialization ------------ //

      // Synthesis cpu time
      long total_time;
      START_TIME(total_time);

      // ---------- Parameter parsing ------------ //
      long cpu_time;
      START_TIME(cpu_time);
      parameters = ParameterRef(new BambuParameter(argv[0]));

      switch(parameters->exec(argc, argv))
      {
         case PARAMETER_NOTPARSED:
         {
            exit_code = PARAMETER_NOTPARSED;
            THROW_ERROR("Bad Parameters format");
         }
         case EXIT_SUCCESS:
         {
            if(not (parameters->getOption<bool>(OPT_no_clean)))
            {
               boost::filesystem::remove_all(parameters->getOption<std::string>(OPT_output_temporary_directory));
            }
            return EXIT_SUCCESS;
         }
         case PARAMETER_PARSED:
         {
            exit_code = EXIT_FAILURE;
            break;
         }
         default:
         {
            THROW_ERROR("Bad Parameters parsing");
         }
      }

      int debug_level = parameters->getOption<int>(OPT_debug_level);
      int output_level = parameters->getOption<int>(OPT_output_level);
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "          ***************************************************");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "                   ____                  _");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "                  | __ )  __ _ _ __ ___ | |_   _   _");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "                  |  _ \\ / _` | \'_ ` _ \\| \'_ \\| | | |");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "                  | |_) | (_| | | | | | | |_) | |_| |");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "                  |____/ \\__,_|_| |_| |_|_.__/ \\__,_|");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "          ***************************************************");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "                       High-Level Synthesis Tool");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "                      Politecnico di Milano - DEI");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "                       System Architectures Group");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "          ***************************************************");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "             Copyright (c) 2004-2012 Politecnico di Milano");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "          ***************************************************");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "          Version: " + STR(PACKAGE_VERSION) + " - Build: " + boost::lexical_cast<std::string>(parameters->getOption<std::string>(OPT_revision)));
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "");

      //Include sysdir
      if(parameters->getOption<bool>(OPT_gcc_include_sysdir))
      {
         const GccWrapperRef gcc_wrapper(new GccWrapper(parameters, CT_NO_GCC, OS_O0));
         std::vector<std::string> system_includes;
         gcc_wrapper->GetSystemIncludes(system_includes);
         std::vector<std::string>::const_iterator system_include, system_include_end = system_includes.end();
         for(system_include = system_includes.begin(); system_include != system_include_end; system_include++)
         {
            INDENT_OUT_MEX(0, 0, *system_include);
         }
         if(not (parameters->getOption<bool>(OPT_no_clean)))
         {
            boost::filesystem::remove_all(parameters->getOption<std::string>(OPT_output_temporary_directory));
         }
         return EXIT_SUCCESS;
      }

      if(parameters->getOption<bool>(OPT_gcc_config))
      {
         const GccWrapperRef gcc_wrapper(new GccWrapper(parameters, CT_NO_GCC, OS_O0));
         gcc_wrapper->GetGccConfig();
         if(not (parameters->getOption<bool>(OPT_no_clean)))
         {
            boost::filesystem::remove_all(parameters->getOption<std::string>(OPT_output_temporary_directory));
         }
         return EXIT_SUCCESS;
      }
      STOP_TIME(cpu_time);
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Parameters parsed in " + print_cpu_time(cpu_time) + " seconds\n");

      // up to now all parameters have been parsed and datastructures created, so synthesis can start
      std::string output_directory = parameters->getOption<std::string>(OPT_output_directory);
      if (boost::filesystem::exists(output_directory))
      {
         boost::filesystem::remove_all(output_directory);
      }
      ///creating the output directory
      boost::filesystem::create_directories(output_directory);

      /// ==== Creating behavioral specification ==== ///
      application_managerRef AppM = application_managerRef(new application_manager(parameters));

      /// ==== Creating target for the synthesis ==== ///
      HLS_targetRef HLS_T = HLS_targetRef(new HLS_target);

      /// ==== Creating High-level synthesis constraints reference ==== ///
      HLS_constraintsRef HLS_C = HLS_constraintsRef(new HLS_constraints(parameters));

      /// ==== Loading high-level synthesis constraints ==== ///
      START_TIME(cpu_time);
      /// add builtin constraints for the builtin resources
      add_builtin_constraints(HLS_C);
      /// add user defined constraints
      if (parameters->isOption("constraintsfile"))
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "parsing the constraint file...");
         HLS_C = read_HLS_constraints_File(parameters->getOption<std::string>("constraintsfile").c_str(), HLS_C);
         PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "Constraint file parsed");
      }
      THROW_ASSERT(HLS_C, "HLS Constraints datastructure not defined");
      /// define the clock period
      if(parameters->isOption(OPT_clock_period))
         HLS_C->set_clock_period(parameters->getOption<double>(OPT_clock_period));
      double clock_period_value = HLS_C->get_clock_period();
      parameters->setOption(OPT_clock_period, clock_period_value);
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Clock period: " << clock_period_value << "ns (" << 1000.0 / clock_period_value << "MHz)");
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "");
      // ---------- Save XML file ------------ //
      if (debug_level >= DEBUG_LEVEL_PEDANTIC)
         write_HLS_constraints_File("HLS_constraints.out.XML", HLS_C);
      STOP_TIME(cpu_time);
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "High-level synthesis constraints loaded in " + print_cpu_time(cpu_time) + " seconds\n");

      /// ==== Loading resource library ==== ///
      START_TIME(cpu_time);
      // Technology library manager
      technology_managerRef TM = technology_managerRef(new technology_manager(parameters));
      /// this adds predefined resources
      add_builtin_resources(TM, parameters);
      PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Built-in resources added");
      // if technology file is given, it is parsed
      if (parameters->isOption("techfile"))
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "parsing the technology file...");
         read_technology_File(parameters->getOption<std::string>("techfile").c_str(), TM, parameters);
         PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "==== Technology file parsed ====");
      }
      THROW_ASSERT(TM, "Technology datastructure not defined");
      HLS_T->set_technology_manager(TM);
      // ---------- Save XML file ------------ //
      if (debug_level >= DEBUG_LEVEL_PEDANTIC)
         write_xml_technology_File("technology.out.XML", TM);
      STOP_TIME(cpu_time);
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Resource library loaded in " + print_cpu_time(cpu_time) + " seconds\n");

      ///creating the datastructure representing the target device
      unsigned int target_device = parameters->getOption<unsigned int>(OPT_target_device_type);
      target_deviceRef device = target_device::create_device(static_cast<target_device::type_t>(target_device), parameters, TM);
      device->set_parameter("clock_period", clock_period_value);
      HLS_T->set_target_device(device);
      target_technologyRef technology = device->get_target_technology();
      HLS_T->set_target_technology(technology);

      /// ==== Creating intermediate representation ==== ///
      START_TIME(cpu_time);
      // Call_graph reference, where problem specification is contained
      createCallGraph(parameters, AppM, HLS_T, HLS_C);
      const tree_managerRef TreeM = AppM->get_tree_manager();
      if(!parameters->isOption(OPT_top_function_name))
      {
         unsigned int function_id_main = AppM->get_main_function();
         if(function_id_main)
         {
            parameters->setOption(OPT_top_function_name, "main");
            PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Top function used is : main");
         }
         else
         {
            std::tr1::unordered_set<unsigned int> functions;
            AppM->get_call_graph()->get_root_functions(functions);
            if(functions.empty())
               THROW_ERROR("No synthesizable functions in the source code");
            unsigned int first_of_root = *functions.begin();
            std::string name_of_first_root = tree_helper::name_function(TreeM, first_of_root);
            parameters->setOption(OPT_top_function_name, name_of_first_root);
            PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Top function used is : " + name_of_first_root);
         }
         PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "");
      }
      if (output_level >= OUTPUT_LEVEL_VERBOSE)
      {
         unsigned int function_id = TreeM->function_index(parameters->getOption<std::string>(OPT_top_function_name));
         THROW_ASSERT(function_id, "BadParameters: RAW file does not include function <" + parameters->getOption<std::string>(OPT_top_function_name) + ">");
         const function_behaviorRef FB = AppM->get_function_behavior(function_id);
         std::string file_name = FB->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(function_id) + "_OP_FFLSAODG.dot";
         FB->get_graph(function_behavior::FFLSAODG)->print_to_dot(parameters, file_name, FB->get_helper());
         file_name = FB->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(function_id) + "_OP_FLSAODG.dot";
         FB->get_graph(function_behavior::FLSAODG)->print_to_dot(parameters, file_name, FB->get_helper());
      }

      STOP_TIME(cpu_time);
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Intermediate representation created in " + print_cpu_time(cpu_time) + " seconds\n");

      if (parameters->isOption(OPT_pretty_print))
      {
         START_TIME(cpu_time);
         CBackendRef backend(new CBackend(AppM, parameters));
#if HAVE_EXPERIMENTAL
         if (parameters->isOption(OPT_insert_memory_profile))
         {
            MemoryProfileCWriterRef cWriter;
            MemoryProfileInstructionWriterRef instrWriter;
            cWriter = MemoryProfileCWriterRef(new MemoryProfileCWriter(AppM, parameters, debug_level >= DEBUG_LEVEL_VERBOSE));
            instrWriter = MemoryProfileInstructionWriterRef(new MemoryProfileInstructionWriter(parameters));
            cWriter->add_instruction_writer(instrWriter);
            backend->registerWriter(cWriter);
            std::string outFileName = parameters->getOption<std::string>(OPT_pretty_print);
            backend->write(outFileName);
         }
         else
#endif
         {
            CWriterRef cWriter;
            InstructionWriterRef instrWriter;
            cWriter = CWriterRef(new CWriter(AppM, parameters, debug_level >= DEBUG_LEVEL_VERBOSE));
            instrWriter = InstructionWriterRef (new InstructionWriter(parameters));
            cWriter->add_instruction_writer(instrWriter);
            backend->registerWriter(cWriter);
            std::string outFileName = parameters->getOption<std::string>(OPT_pretty_print);
            backend->write(outFileName);
         }

         STOP_TIME(cpu_time);
         PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Pretty print completed in " + print_cpu_time(cpu_time) + " seconds\n");
      }

      /// ==== High-Level Synthesis ==== ///
      /* Let's start to synthetize */
      long hls_time;
      START_TIME(hls_time);
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, " ==== Starting High-Level Synthesis ====\n");

      ///creating the memory datastructure
      HLS_C->Rmem = memoryRef(new memory(HLS_C->base_address));

      ///performing the memory allocation
      long memory_time;
      START_TIME(memory_time);
      memory_allocationRef m_alloc = memory_allocationRef(new memory_allocation(parameters, AppM, HLS_C));
      m_alloc->exec();
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Total amount of allocated memory: " + STR(HLS_C->Rmem->get_allocated_space()) + " bytes");
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "External memory allocated space: " + STR(HLS_C->Rmem->get_memory_address()-HLS_C->base_address));
      STOP_TIME(memory_time);
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Memory allocation completed in " + print_cpu_time(memory_time) + " seconds\n");

      if (parameters->getOption<bool>(OPT_generate_testbench))
      {
         HLStestbenchRef tb = HLStestbenchRef(new HLStestbench(parameters));
         tb->generate_testbench(AppM, TM, HLS_C, HLS_C->Rmem);
      }

      std::list<unsigned int> functions;
      const call_graphRef CG = AppM->get_call_graph();
      const graph* call_graph = CG->get_graph();

      unsigned int top_function = TreeM->function_index(parameters->getOption<std::string>(OPT_top_function_name));
      vertex starting_function = CG->get_vertex(top_function);
      //call the depth-first-search
      called_functions_visitor vis(functions, AppM);
      std::vector<boost::default_color_type> color_vec(boost::num_vertices(*call_graph));
      boost::depth_first_visit(*call_graph, starting_function, vis, boost::make_iterator_property_map(color_vec.begin(), boost::get(boost::vertex_index_t(), *call_graph), boost::white_color));

      for(std::list<unsigned int>::iterator f = functions.begin(); f != functions.end(); f++)
      {
         PRINT_OUT_MEX(OUTPUT_LEVEL_PEDANTIC, output_level, "Function to be synthetized: " + tree_helper::name_function(TreeM, *f));
      }
      PRINT_OUT_MEX(OUTPUT_LEVEL_PEDANTIC, output_level, "");

      std::map<unsigned int, hlsRef> implementations;
      for(std::list<unsigned int>::iterator f = functions.begin(); f != functions.end(); f++)
      {
         std::string function_name = tree_helper::normalized_ID(tree_helper::name_function(TreeM, *f));
         if(TM->get_library(function_name) != "")
            continue;
         PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Synthesizing function: " + function_name);
         structural_managerRef SM;
         TM->add_resource(WORK_LIBRARY, function_name, SM);
         TM->add_operation(WORK_LIBRARY, function_name, function_name);

         hlsRef HLS = hls_flow::create_base_HLS(parameters, AppM, *f, HLS_T, HLS_C);
         // Creates the specificated high-level synthesis flow
         hls_flow::flow_type flow = static_cast<hls_flow::flow_type>(parameters->getOption<unsigned int>(OPT_hls_flow));
         hls_flowRef HlsFlow = hls_flow::create_flow(flow, parameters, HLS);
         // Executes the high-level synthesis on the specified module
         HlsFlow->exec();
         ///save the resulting implementation
         implementations[*f] = HLS;

         ///updating the structural description into the library
         if (parameters->getOption<bool>(OPT_datapath) and parameters->getOption<bool>(OPT_controller))
         {
            SM = HLS->top;
         }
         else if (parameters->getOption<bool>(OPT_datapath))
         {
            SM = HLS->datapath;
         }
         else if (parameters->getOption<bool>(OPT_controller))
         {
            SM = HLS->controller;
         }
         TM->add_resource(WORK_LIBRARY, function_name, SM);
         const functional_unit* fu = GetPointer<functional_unit>(TM->get_fu(function_name, WORK_LIBRARY));
         operation* op = GetPointer<operation>(fu->get_operation(function_name));
         op->bounded = false;  /// all functions are unbounded by default
         op->time_m = time_model::create_model(device->get_target_technology(), parameters);
         op->time_m->set_execution_time(clock_period_value, 1);
         PRINT_OUT_MEX(OUTPUT_LEVEL_PEDANTIC, output_level, "\nAdding " << function_name << " to WORK_LIBRARY");

#if HAVE_EXPERIMENTAL
         HDL_managerRef HDL = HDL_managerRef(new HDL_manager(device, parameters));
         HDL->hdl_gen(function_name, SM->get_circ());
#endif

///vcd-debug
#if (HAVE_XILINX || HAVE_ALTERA) && HAVE_MODELSIM && HAVE_VCD_BUILT
         if (parameters->getOption<bool>("vcd-debug"))
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "Starting vcd-debug\n");
            vcd_utilityRef vcd_util = vcd_utilityRef(new vcd_utility(debug_level));

            unsigned int function_id = TreeM->function_index(parameters->getOption<std::string>(OPT_top_function_name));
            const function_behaviorRef FB = AppM->get_function_behavior(function_id);
            vcd_util->compute_graph(FB, HLS);
            std::vector<std::string> fl = vcd_util->map_reg_to_signal();
            vcd_util->parse_vcd_file("test.vcd", fl);
            std::map<unsigned int, std::vector<sig_variationRef> > var;

            const CWriterRef vcd_writer = CWriterRef(new CWriterVcdDebug(AppM, parameters, false));
            const InstructionWriterRef instrWriter = InstructionWriterRef(new VcdDebugInstructionWriter(AppM, parameters));

            CBackendRef backend_vcd(new CBackend(AppM, parameters));

            vcd_writer->add_instruction_writer(instrWriter);
            backend_vcd->registerWriter(vcd_writer);
            backend_vcd->write(function_name+"_vcd_debug.c");
            PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "Terminating vcd-debug\n");
         }
#endif
         PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "");
      }
      THROW_ASSERT(implementations.find(top_function) != implementations.end(), "Implementation of top function does not exist at the end of the analysis");
      hlsRef HLS = implementations[top_function];
      if (output_level >= OUTPUT_LEVEL_VERBOSE)
      {
         HLS->print_resources(std::cerr);
         PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "");
      }
      STOP_TIME(hls_time);
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, " ==== High-Level Synthesis completed in " + print_cpu_time(hls_time) + " seconds ==== \n");

#if HAVE_EXPERIMENTAL
      ///Creates the module interface, if required.
      if (parameters->getOption<bool>(OPT_interface))
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Creating the module interface");
         module_interface::interface_t int_type = static_cast<module_interface::interface_t>(parameters->getOption<unsigned int>(OPT_interface_type));
         module_interfaceRef miRef = module_interface::create_module(int_type, parameters, TM, TreeM);
         if (!HLS->top) THROW_ERROR("Interface can be created only for the top entity");
         HLS->top = miRef->exec(HLS->top);
      }
#endif

      ///Generates all the backend scripts
      generateBackendScripts(parameters, HLS);

      STOP_TIME(total_time);
      PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "Total Execution Time: " + print_cpu_time(total_time) + " seconds\n");

      if(not (parameters->getOption<bool>(OPT_no_clean)))
      {
         boost::filesystem::remove_all(parameters->getOption<std::string>(OPT_output_temporary_directory));
      }
      return EXIT_SUCCESS; // Bambu tool has completed execution without errors
   }

   // exception catching
   catch (const char * str)
   {
      if(EXIT_SUCCESS == exit_code)
         exit_code = EXIT_FAILURE;
      std::cerr << str << std::endl;
   }
   catch (const std::string& str)
   {
      if(EXIT_SUCCESS == exit_code)
         exit_code = EXIT_FAILURE;
      std::cerr << str << std::endl;
   }
   catch (std::exception &e)
   {
      std::cerr << e.what() << std::endl;
   }
   catch (...)
   {
      if(EXIT_SUCCESS == exit_code)
         exit_code = EXIT_FAILURE;
      std::cerr << "Unknown error type" << std::endl;
   }

   switch(exit_code)
   {
      case PARAMETER_NOTPARSED:
      {
         parameters->PrintUsage(std::cout);
         break;
      }
      case EXIT_FAILURE:
      {
         if(parameters)
            parameters->PrintBugReport(std::cout);
         break;
      }
      default:
      {

      }
   }
   if(parameters && not (parameters->getOption<bool>(OPT_no_clean)))
   {
      boost::filesystem::remove_all(parameters->getOption<std::string>(OPT_output_temporary_directory));
   }
   return exit_code;

}

