/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file application_manager.cpp
 * @brief Implementation of some methods to manage a generic C application
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Autoheader include
#include "config_HAVE_POLIXML_BUILT.hpp"
#include "config_HAVE_PRAGMA_BUILT.hpp"

///Header include
#include "application_manager.hpp"

///Application includes
#include "call_graph.hpp"
#include "behavioral_helper.hpp"
#include "tree_manager.hpp"
#include "tree_reindex.hpp"
#include "ext_tree_node.hpp"
#include "loops.hpp"
#include "loop.hpp"
#include "op_node_info.hpp"
#include "function_behavior.hpp"

///POLIXML includes
#include "polixml.hpp"
#include "xml_helper.hpp"

///GCC wrapper include
#include "gcc_wrapper.hpp"

///Utility includes
#include "Parameter.hpp"
#include "constant_strings.hpp"
#include "exceptions.hpp"
#include "utility.hpp"


application_manager::application_manager(const ParameterConstRef _Param) :
   Param(_Param),
   main_function(0),  ///it means that the main function has not been identified yet
   debug_level(_Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE))
{

}

application_manager::~application_manager()
{

}

void application_manager::set_tree_manager(const tree_managerRef _TM)
{
   TM = _TM;
   main_function = TM->function_index("main");
   CG = call_graphRef();
   functions.clear();
   global_variables.clear();
}

const tree_managerRef application_manager::get_tree_manager() const
{
   return TM;
}

void application_manager::set_call_graph(const call_graphRef _CG)
{
   CG = _CG;
   functions.clear();
   global_variables.clear();
}

const call_graphRef application_manager::get_call_graph() const
{
   return CG;
}

int application_manager::write_call_graph_dot(const std::string& file_name) const
{
   THROW_ASSERT(CG, "Call graph not yet created");
   return CG->write_dot(file_name, functions);
}

unsigned int application_manager::get_main_function() const
{
   return main_function;
}

bool application_manager::is_function(unsigned int index) const
{
   return functions.find(index) != functions.end();
}

function_behaviorRef application_manager::get_function_behavior(unsigned int index) const
{
   THROW_ASSERT(functions.find(index) != functions.end(), "There is no function with index " + STR(index));
   return functions.find(index)->second;
}

const std::tr1::unordered_set<unsigned int>& application_manager::get_global_variables() const
{
   return global_variables;
}

void application_manager::add_global_variable(unsigned int var)
{
   global_variables.insert(var);
}

void application_manager::add_function(unsigned int current, unsigned int index, const behavioral_helperRef helper, unsigned int node_stmt)
{
   add_function(index, helper);
   CG->add_function(current, index, helper, node_stmt);
}

void application_manager::add_function(unsigned int index, const behavioral_helperRef helper)
{
   if(functions.find(index) == functions.end())
   {
      THROW_ASSERT(helper, "expected a valid helper "+ boost::lexical_cast<std::string>(index));
      functions[index] = function_behaviorRef(new function_behavior(helper));
   }
   CG->add_function(index, helper);
}

void application_manager::get_functions_with_body(std::list<unsigned int> &withBody) const
{
   std::map<unsigned int, function_behaviorRef>::const_iterator beg, end;
   for(beg = functions.begin(), end = functions.end(); beg != end; beg++)
   {
      if(beg->second->get_helper()->has_implementation())
         withBody.push_back(beg->first);
   }
}

unsigned int application_manager::get_first_function_id() const
{
   std::map<unsigned int, function_behaviorRef>::const_iterator beg, end;
   for(beg = functions.begin(), end = functions.end(); beg != end; beg++)
   {
      if(beg->second->get_helper()->has_implementation())
         return beg->first;
   }
   PANDA_UNREACHABLE();
   return 0;
}

void application_manager::get_functions_without_body(std::list<unsigned int> &withoutBody) const
{
   std::map<unsigned int, function_behaviorRef>::const_iterator beg, end;
   for(beg = functions.begin(), end = functions.end(); beg != end; beg++)
   {
      if(!(beg->second->get_helper()->has_implementation()))
         withoutBody.push_back(beg->first);
   }
}

#if HAVE_PRAGMA_BUILT
void application_manager::set_pragma_manager(const pragma_managerRef _PM)
{
   PM = _PM;
}

const pragma_managerRef application_manager::get_pragma_manager() const
{
   return PM;
}
#endif

#if HAVE_POLIXML_BUILT
void application_manager::dump_source_code_statistics(xml_element * local_root)
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Dumping source code statistics");
   xml_element * source_code_lines = local_root->add_child_element(STR_XML_source_code_lines);
   size_t value = GccWrapper::GetSourceCodeLines(Param);
   WRITE_XNVM(value, boost::lexical_cast<std::string>(value), source_code_lines);

   std::list<unsigned int> functions_with_body;
   get_functions_with_body(functions_with_body);
   value = functions_with_body.size();
   xml_element * functions_with_body_node = local_root->add_child_element(STR_XML_source_code_function_with_body);
   WRITE_XNVM(value, boost::lexical_cast<std::string>(value), functions_with_body_node);


   size_t loop_number = 0;
   unsigned int maximum_depth = 0;
   unsigned int conditional_construct = 0;
   std::list<unsigned int>::const_iterator f, f_end = functions_with_body.end();
   for(f = functions_with_body.begin(); f != f_end; f++)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Dumping function " + boost::lexical_cast<std::string>(*f));
      const std::list<LoopConstRef> loops = get_function_behavior(*f)->GetLoops()->GetList();
      ///Minus one because of loop zero
      loop_number += loops.size() - 1;
      std::list<LoopConstRef>::const_iterator l, l_end = loops.end();
      for(l = loops.begin(); l != l_end; l++)
      {
         if((*l)->depth > maximum_depth)
            maximum_depth = (*l)->depth;
      }
      const graph * cfg = get_function_behavior(*f)->get_graph(function_behavior::CFG);
      VertexIterator v, v_end;
      for(boost::tie(v, v_end) = boost::vertices(*cfg); v != v_end; v++)
      {
         if(GET_TYPE(cfg, *v) & (TYPE_SWITCH | TYPE_IF))
         {
            conditional_construct++;
         }
      }
   }
   value = loop_number;
   xml_element * loop_number_node = local_root->add_child_element(STR_XML_source_code_loop_number);
   WRITE_XNVM(value, boost::lexical_cast<std::string>(value), loop_number_node);

   value = maximum_depth;
   xml_element * maximum_depth_node = local_root->add_child_element(STR_XML_source_code_maximum_depth);
   WRITE_XNVM(value, boost::lexical_cast<std::string>(value), maximum_depth_node);

   value = conditional_construct;
   xml_element * conditional_construct_node = local_root->add_child_element(STR_XML_source_code_conditional_construct);
   WRITE_XNVM(value, boost::lexical_cast<std::string>(value), conditional_construct_node);
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Dumped source code statistics");

}
#endif

unsigned int application_manager::get_produced_value(unsigned int fun_id, const vertex& v)
{
   const graph* cfg = get_function_behavior(fun_id)->get_graph(function_behavior::CFG);
   const std::list<unsigned int>& NodeID = GET_NODEID(cfg, v);
   return (NodeID.size() ? get_produced_value(TM->get_tree_node_const(NodeID.back())) : 0);
}

std::vector<application_manager::io_binding_type> application_manager::get_required_values(unsigned int fun_id, const vertex& v)
{
   const graph* cfg = get_function_behavior(fun_id)->get_graph(function_behavior::CFG);
   const std::list<unsigned int>& NodeID = GET_NODEID(cfg, v);
   std::vector<io_binding_type> required;
   if (NodeID.size())
      get_required_values(required, TM->get_tree_node_const(NodeID.back()));
   return required;
}

unsigned int application_manager::get_produced_value(const tree_nodeRef& tn)
{
   switch (tn->get_kind())
   {
      case while_expr_K:
      case gimple_cond_K:
         return CONDITION_RESULT;
      case gimple_label_K:
      case gimple_return_K:
      case gimple_call_K:
      case gimple_goto_K:
      {
         break;
      }
      case gimple_phi_K:
      {
         gimple_phi* gp = GetPointer<gimple_phi>(tn);
         return GET_INDEX_NODE(gp->res);
      }
      case gimple_switch_K:
      {
         gimple_switch * se = GetPointer<gimple_switch>(tn);
         return GET_INDEX_NODE(se->op0);
      }
      case gimple_assign_K:
      {
         gimple_assign* gm = GetPointer<gimple_assign>(tn);
         tree_nodeRef op0 = GET_NODE(gm->op0);
         if(gm->init_assignment)
            break;
         else if(op0->get_kind() == array_ref_K )
            break;
         else if (op0->get_kind() == indirect_ref_K)
            break;
         else if (op0->get_kind() == mem_ref_K)
            break;
         else if (op0->get_kind() == target_mem_ref_K)
            break;
         else if (op0->get_kind() == target_mem_ref461_K)
            break;
         else
         {
            return GET_INDEX_NODE(gm->op0);
         }
      }
      default:
         THROW_ERROR("Operation not yet supported: " + std::string(tn->get_kind_text()));
   }
   return 0;
}

void application_manager::get_required_values(std::vector<io_binding_type>& required, const tree_nodeRef& tn, unsigned int index)
{
   switch (tn->get_kind())
   {
      case ssa_name_K:
      case string_cst_K:
      case real_cst_K:
      case integer_cst_K:
      case vector_cst_K:
      case var_decl_K:
      case parm_decl_K:
      {
         required.push_back(io_binding_type(index,0));
         break;
      }
      case while_expr_K:
      {
         while_expr* we = GetPointer<while_expr>(tn);
         get_required_values(required, GET_NODE(we->op0), GET_INDEX_NODE(we->op0));
         break;
      }
      case CASE_BINARY_EXPRESSION:
      {
         binary_expr* be = GetPointer<binary_expr>(tn);
         get_required_values(required, GET_NODE(be->op0), GET_INDEX_NODE(be->op0));
         get_required_values(required, GET_NODE(be->op1), GET_INDEX_NODE(be->op1));
         break;
      }
      case CASE_UNARY_EXPRESSION:
      {
         unary_expr* ue = GetPointer<unary_expr>(tn);
         if(tn->get_kind() == addr_expr_K || tn->get_kind() == imagpart_expr_K || tn->get_kind() == realpart_expr_K)
         {
            required.push_back(io_binding_type(index,0));
         }
         else
            get_required_values(required, GET_NODE(ue->op), GET_INDEX_NODE(ue->op));
         break;
      }
      case cond_expr_K:
      {
         cond_expr* ce = GetPointer<cond_expr>(tn);
         get_required_values(required, GET_NODE(ce->op0), GET_INDEX_NODE(ce->op0));
         get_required_values(required, GET_NODE(ce->op1), GET_INDEX_NODE(ce->op1));
         get_required_values(required, GET_NODE(ce->op2), GET_INDEX_NODE(ce->op2));
         break;
      }
      case gimple_cond_K:
      {
         gimple_cond* gc = GetPointer<gimple_cond>(tn);
         get_required_values(required, GET_NODE(gc->op0), GET_INDEX_NODE(gc->op0));
         break;
      }
      case gimple_switch_K:
      {
         gimple_switch* se = GetPointer<gimple_switch>(tn);
         get_required_values(required, GET_NODE(se->op0), GET_INDEX_NODE(se->op0));
         break;
      }
      case array_ref_K:
      {
         array_ref* ar = GetPointer<array_ref>(tn);
         required.push_back(io_binding_type(GET_INDEX_NODE(ar->op0),0));
         get_required_values(required, GET_NODE(ar->op1), GET_INDEX_NODE(ar->op1));
         break;
      }
      case target_mem_ref_K:
      {
         target_mem_ref* tmr = GetPointer<target_mem_ref>(tn);
         if(tmr->symbol)
            required.push_back(io_binding_type(GET_INDEX_NODE(tmr->symbol),0));
         else
            required.push_back(io_binding_type(0,0));
         if(tmr->base)
            required.push_back(io_binding_type(GET_INDEX_NODE(tmr->base),0));
         else
            required.push_back(io_binding_type(0,0));
         if(tmr->idx)
            required.push_back(io_binding_type(GET_INDEX_NODE(tmr->idx),0));
         else
            required.push_back(io_binding_type(0,0));
         if(tmr->step)
            required.push_back(io_binding_type(GET_INDEX_NODE(tmr->step),0));
         else
            required.push_back(io_binding_type(0,0));
         if(tmr->offset)
            required.push_back(io_binding_type(GET_INDEX_NODE(tmr->offset),0));
         else
            required.push_back(io_binding_type(0,0));
         break;
      }
      case target_mem_ref461_K:
      {
         target_mem_ref461* tmr = GetPointer<target_mem_ref461>(tn);
         if(tmr->base)
            required.push_back(io_binding_type(GET_INDEX_NODE(tmr->base),0));
         else
            required.push_back(io_binding_type(0,0));
         if(tmr->idx)
            required.push_back(io_binding_type(GET_INDEX_NODE(tmr->idx),0));
         else
            required.push_back(io_binding_type(0,0));
         if(tmr->step)
            required.push_back(io_binding_type(GET_INDEX_NODE(tmr->step),0));
         else
            required.push_back(io_binding_type(0,0));
         if(tmr->idx2)
            required.push_back(io_binding_type(GET_INDEX_NODE(tmr->idx2),0));
         else
            required.push_back(io_binding_type(0,0));
         if(tmr->offset)
            required.push_back(io_binding_type(GET_INDEX_NODE(tmr->offset),0));
         else
            required.push_back(io_binding_type(0,0));
         break;
      }
      case gimple_assign_K:
      {
         gimple_assign* gm = GetPointer<gimple_assign>(tn);
         if(!gm->init_assignment)
         {
            tree_nodeRef op0 = GET_NODE(gm->op0);
            tree_nodeRef op1 = GET_NODE(gm->op1);

            if (op0->get_kind() == component_ref_K || op0->get_kind() == indirect_ref_K  || op0->get_kind() == mem_ref_K || op0->get_kind() == array_ref_K || op0->get_kind() == target_mem_ref_K || op0->get_kind() == target_mem_ref461_K || op0->get_kind() ==  bit_field_ref_K)
            {
               get_required_values(required, op1, GET_INDEX_NODE(gm->op1));
               get_required_values(required, op0, GET_INDEX_NODE(gm->op0));
            }
            else
            {
               if (op1->get_kind() == component_ref_K || op1->get_kind() == indirect_ref_K || op1->get_kind() == mem_ref_K || op1->get_kind() == array_ref_K || op1->get_kind() == target_mem_ref_K || op1->get_kind() == target_mem_ref461_K || op1->get_kind() == bit_field_ref_K)
                  required.push_back(io_binding_type(0,0));
               get_required_values(required, op1, GET_INDEX_NODE(gm->op1));
            }
         }
         break;
      }
      case gimple_return_K:
      {
         gimple_return* rt = GetPointer<gimple_return>(tn);
         if (rt->op) get_required_values(required, GET_NODE(rt->op), GET_INDEX_NODE(rt->op));
         break;
      }
      case gimple_phi_K:
      {
         gimple_phi* gp = GetPointer<gimple_phi>(tn);
         std::vector<std::pair< tree_nodeRef, unsigned int> >::const_iterator lde_it_end = gp->list_of_def_edge.end();
         for(std::vector<std::pair< tree_nodeRef, unsigned int> >::const_iterator lde_it = gp->list_of_def_edge.begin(); lde_it != lde_it_end; ++lde_it)
            required.push_back(io_binding_type(GET_INDEX_NODE(lde_it->first),0));
         break;
      }
      case label_decl_K:
      case gimple_label_K:
      case gimple_goto_K:
      {
         ///this has not to be synthesized
         break;
      }
      case call_expr_K:
      {
         call_expr* ce = GetPointer<call_expr>(tn);
         const std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            required.push_back(io_binding_type(GET_INDEX_NODE(*arg), 0));
         }
         break;
      }
      case gimple_call_K:
      {
         gimple_call* ce = GetPointer<gimple_call>(tn);
         const std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            required.push_back(io_binding_type(GET_INDEX_NODE(*arg), 0));
         }
         break;
      }
      case tree_list_K:
      {
         tree_list* tl = GetPointer<tree_list>(tn);
         required.push_back(io_binding_type(GET_INDEX_NODE(tl->valu),0));
         if (tl->chan) get_required_values(required, GET_NODE(tl->chan), GET_INDEX_NODE(tl->chan));
         break;
      }
      case component_ref_K:
      {
         component_ref* cr = GetPointer<component_ref>(tn);
         required.push_back(io_binding_type(GET_INDEX_NODE(cr->op0),0));
         get_required_values(required, GET_NODE(cr->op1), GET_INDEX_NODE(cr->op1));
         break;
      }
      case bit_field_ref_K:
      {
         bit_field_ref* bfr = GetPointer<bit_field_ref>(tn);
         required.push_back(io_binding_type(GET_INDEX_NODE(bfr->op0),0));
         get_required_values(required, GET_NODE(bfr->op1), GET_INDEX_NODE(bfr->op1));
         get_required_values(required, GET_NODE(bfr->op2), GET_INDEX_NODE(bfr->op2));
         break;
      }
      case field_decl_K:
      {
         field_decl* fd = GetPointer<field_decl>(tn);
         integer_cst *ic = GetPointer<integer_cst>(GET_NODE(fd->bpos));
         THROW_ASSERT(ic, "non-constant field offset (variable lenght object) currently not supported: " + STR(GET_INDEX_NODE(fd->bpos)));
         unsigned long long int ull_value = static_cast<unsigned long long int>(tree_helper::get_integer_cst_value(ic));
         required.push_back(io_binding_type(0, static_cast<unsigned int>(ull_value/8))); /// bpos has an offset in bits
         if(ull_value%8 != 0)
            THROW_ERROR("bitfields are not yet supported");
         break;
      }
      default:
         THROW_ERROR("Operation not yet supported: " + std::string(tn->get_kind_text()));
   }
}
