/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file application_manager.hpp
 * @brief Definition of the class representing a generic C application
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef _APPLICATION_MANAGER_HPP_
#define _APPLICATION_MANAGER_HPP_

///Autoheader
#include "config_HAVE_FROM_PRAGMA_BUILT.hpp"

#include "graph.hpp"
///Utility include
#include "refcount.hpp"
CONSTREF_FORWARD_DECL(Parameter);
REF_FORWARD_DECL(tree_manager);
REF_FORWARD_DECL(tree_node);
REF_FORWARD_DECL(call_graph);
REF_FORWARD_DECL(graph);
REF_FORWARD_DECL(function_behavior);
REF_FORWARD_DECL(behavioral_helper);
REF_FORWARD_DECL(pragma_manager);
class xml_element;

#include <map>
#include <list>
#include <vector>
#include <tr1/unordered_set>

class application_manager
{
   public:

      ///tuple set used to represent the required values or the constant default value associated with the inputs of a node
      typedef std::tr1::tuple<unsigned int, unsigned int> io_binding_type;

      ///value returned for if, while and switch
      static const unsigned int CONDITION_RESULT = static_cast<unsigned int>(-1);

   protected:

      ///class containing all the parameters
      const ParameterConstRef Param;

      ///class representing the application information at low level
      tree_managerRef TM;

      ///class representing the call graph of the application
      call_graphRef CG;

      ///id of the main function
      unsigned int main_function;

      ///put into relation the index of the function and the corresponding data structure
      std::map<unsigned int, function_behaviorRef> functions;

      ///set of global variables
      std::tr1::unordered_set<unsigned int> global_variables;

#if HAVE_FROM_PRAGMA_BUILT
      ///class representing the source code pragmas
      pragma_managerRef PM;
#endif

      ///debugging level of the class
      const int debug_level;

      /**
       * Returns the values produced by a vertex (recursive version)
       */
      unsigned int get_produced_value(const tree_nodeRef& tn);

      /**
       * Returns the values required by a vertex (recursive version)
       */
      void get_required_values(std::vector<io_binding_type>& required, const tree_nodeRef& tn, unsigned int index = 0);

   public:

      /**
       * Constructor
       * @param _Param is the reference to the class containing all the parameters
       */
      application_manager(const ParameterConstRef _Param);

      /**
       * Destructor
       */
      virtual ~application_manager();

      /**
       * Sets the new tree manager for the application. This method invalidates the data depending from this datastructure.
       * @param _TM is the new call graph for the application
       */
      void set_tree_manager(const tree_managerRef _TM);

      /**
       * Returns the tree manager associated with the application
       */
      const tree_managerRef get_tree_manager() const;

      /**
       * Sets the new call graph for the application. This method invalidates the data depending from this datastructure.
       * @param _CG is the new call graph for the application
       */
      void set_call_graph(const call_graphRef _CG);

      /**
       * Returns the call graph associated with the application
       */
      const call_graphRef get_call_graph() const;

      /**
       * Writes the call graph in a dot format
       */
      int write_call_graph_dot(const std::string& file_name) const;

      /**
       * Returns the id of the main function, if there exist. Otherwise it returns 0.
       */
      unsigned int get_main_function() const;

      /**
       * Creates a new function_behavior for function index when required and update the called_by_graph.
       * @param current is the current function
       * @param index is the index of the called function
       * @param helper is the behavioral_helper associated with the called function
       */
      void add_function(unsigned int current, unsigned int index, const behavioral_helperRef helper, unsigned int node_stmt = 0);

      /**
       * Creates a new function_behavior for function index when required and update the called_by_graph. Note that index could not be called by any other function.
       * @param index is the index of the function
       * @param helper is the behavioral_helper associated with the called function
       */
      void add_function(unsigned int index, const behavioral_helperRef helper);

      /**
       * Returns true if the function (represented by the identified) has been already analyzed, i.e., there is a data structure associated with.
       * @param index is the identifier of the function
       */
      bool is_function(unsigned int index) const;

      /**
       * Returns the datastructure associated with the given identifier. This method returns an error if the function does not exist.
       * @param index is the identified of the function to be returned
       * @return the function_behavior associated with the given function
       */
      function_behaviorRef get_function_behavior(unsigned int index) const;

      /**
       * Returns the lists of functions whose implementation is present in the parsed
       * input specification (i.e. which has a non empty graph)
       * @param withBody is the list of functions to be returned
       */
      void get_functions_with_body(std::list<unsigned int>& withBody) const;

      /**
       * Returns the list of functions whose implementation is not present in the parsed
       * input specification (i.e. the ones with an empty Control Flow Graph)
       * @param withoutBody is the list of functions to be returned
       */
      void get_functions_without_body(std::list<unsigned int>& withoutBody) const;

      /**
       * Returns the first id of a function with the body
       */
      unsigned int get_first_function_id() const;

      /**
       * Adds a global variable
       * @param var is the global variable to be added
       */
      void add_global_variable(unsigned int var);

      /**
       * Returns the set of original global variables
       * @return a set containing the identified of the global variables
       */
      const std::tr1::unordered_set<unsigned int>& get_global_variables() const;

#if HAVE_FROM_PRAGMA_BUILT
      /**
       * Sets the reference to the manager of the source code pragmas
       * @param _PM is the new reference to the pragma manager
       */
      void set_pragma_manager(const pragma_managerRef _PM);

      /**
       * Returns the reference to the manager for the source code pragmas
       */
      const pragma_managerRef get_pragma_manager() const;
#endif

      /**
       * Prints statistics about input source code
       * @param node is the xml node which the results have to be nested to
       */
      void dump_source_code_statistics(xml_element* local_root);

      /**
       * Returns the value produced by a vertex
       */
      unsigned int get_produced_value(unsigned int fun_id, const vertex& v);

      /**
       * Returns the values required by a vertex
       */
      std::vector<io_binding_type> get_required_values(unsigned int fun_id, const vertex& v);

};
///refcount definition of the class
typedef refcount<application_manager> application_managerRef;
///constant refcount definition of the class
typedef refcount<const application_manager> application_managerConstRef;

#endif
