/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file basic_block.cpp
 * @brief Class implementation of the basic_block structure.
 *
 * This file implements some of the basic_block member functions.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/

///Autoheader include
#include "config_HAVE_HOST_PROFILING_BUILT.hpp"

///Header include
#include "basic_block.hpp"
#include "tree_basic_block.hpp"
///Behavior include
#include "behavioral_writer_helper.hpp"
#include "behavioral_helper.hpp"

///Graph include
#include "graph.hpp"

///Utility include
#include <boost/graph/graphviz.hpp>
#include <fstream>
#include "Parameter.hpp"

bb_node_info::bb_node_info() :
   loop_id(0),
   cer(0)
{
}


void bb_node_info::add_operation_node(const vertex op)
{
   statements_list.push_back(op);
}

size_t bb_node_info::size() const
{
   return statements_list.size();
}

vertex bb_node_info::get_first_operation() const
{
   return statements_list.front();
}

vertex bb_node_info::get_last_operation() const
{
   return statements_list.back();
}

bool bb_node_info::empty() const
{
   return statements_list.empty();
}

unsigned int bb_node_info::get_bb_index() const
{
   return block->number;
}

const std::set<unsigned int> & bb_node_info::get_live_in() const
{
   return block->live_in;
}

const std::set<unsigned int> & bb_node_info::get_live_out() const
{
   return block->live_out;
}

/*const std::map<unsigned int, unsigned int> & bb_node_info::get_live_out_phi_defs() const
{
   return block->live_out_phi_defs;
}*/

bb_edge_info::bb_edge_info() :
   epp_value(0)
{
}

bb_edge_info::~bb_edge_info()
{
}

void bb_edge_info::set_epp_value(unsigned long long _epp_value)
{
   epp_value = _epp_value;
}

unsigned long long bb_edge_info::get_epp_value() const
{
   return epp_value;
}

#if HAVE_HOST_PROFILING_BUILT
BBGraphInfo::BBGraphInfo(const ProfilingInformationConstRef _profiling_information) :
   profiling_information(_profiling_information)
{}
#endif

void bb_graphs_collection::print(std::ostream& os)
{
   VertexIterator v, v_end;
   for(boost::tie(v, v_end); v!= v_end; v++)
      GET_BB_NODE_INFO(this, *v)->print(os);
}

bb_graphs_collection::bb_graphs_collection()  :
   graphs_collection(graph_infoRef(new BBGraphInfo))
{

}

#if HAVE_HOST_PROFILING_BUILT
bb_graphs_collection::bb_graphs_collection(const ProfilingInformationConstRef profiling_information) :
   graphs_collection(graph_infoRef(new BBGraphInfo(profiling_information)))
{

}
#endif

bb_graphs_collection::~bb_graphs_collection()
{

}

bb_edge_info * bb_graphs_collection::create_bb_edge_info(const EdgeDescriptor e)
{
   edge_infoRef & info = (*this)[e].info;
   THROW_ASSERT(not info or dynamic_cast<bb_edge_info *>(info.get()), "Edge with associated a node info different from bb_edge_info");
   if (not info)
   {
      info = refcount<edge_info>(new bb_edge_info());
   }
   return dynamic_cast<bb_edge_info *>(info.get());
}

bb_edge_info * bb_graphs_collection::get_bb_edge_info(const EdgeDescriptor e) const
{
   const edge_infoRef & info = (*this)[e].info;
   THROW_ASSERT(info and dynamic_cast<bb_edge_info *>(info.get()), "Edge without bb_edge_info associated");
   return dynamic_cast<bb_edge_info *>(info.get());
}

bb_graph::bb_graph(bb_graphs_collection * _g, int _selector) :
   graph(_g, _selector),
   g(_g)
{
}

bb_graph::bb_graph(bb_graphs_collection * _g, int _selector, std::set<boost::graph_traits<bb_graphs_collection>::vertex_descriptor > sub) :
   graph(_g,_selector,sub),
   g(_g)
{
}

int bb_graph::PrintToDot(const ParameterConstRef Param, std::string file_name, const behavioral_helperConstRef BH, std::tr1::unordered_set<vertex> annotated) const
{
   std::string output_directory = Param->getOption<std::string>(OPT_output_directory) + "/dot/";
   if (BH) output_directory += BH->get_function_name() + "/";
   if (!boost::filesystem::exists(output_directory)) boost::filesystem::create_directories(output_directory);
   std::ofstream f((output_directory + file_name).c_str());
   boost::write_graphviz(f, *this, bb_label_writer(this, BH, annotated), bb_edge_property_writer(this, selector, BH));
   return 1;
}

size_t bb_graph::num_bblocks() const
{
   return boost::num_vertices(m_g) - 2;
}

bb_edge_info * bb_graph::get_bb_edge_info(const EdgeDescriptor e) const
{
   return g->get_bb_edge_info(e);
}

bool bb_edge_info::cdg_edge_T() const
{
   return labels.find(CDG_SELECTOR) != labels.end() and labels.find(CDG_SELECTOR)->second.find(T_COND) != labels.find(CDG_SELECTOR)->second.end();
}

bool bb_edge_info::cdg_edge_F() const
{
   return labels.find(CDG_SELECTOR) != labels.end() and labels.find(CDG_SELECTOR)->second.find(F_COND) != labels.find(CDG_SELECTOR)->second.end();
}

bool bb_edge_info::cfg_edge_T() const
{
   return labels.find(BB_SELECTOR) != labels.end() and labels.find(BB_SELECTOR)->second.find(T_COND) != labels.find(BB_SELECTOR)->second.end();
}

bool bb_edge_info::cfg_edge_F() const
{
   return labels.find(BB_SELECTOR) != labels.end() and labels.find(BB_SELECTOR)->second.find(F_COND) != labels.find(BB_SELECTOR)->second.end();
}

bool bb_edge_info::switch_p() const
{
   return !cdg_edge_T() and !cdg_edge_F() and !cfg_edge_T() and !cfg_edge_F() and (labels.find(CDG_SELECTOR) != labels.end() or labels.find(BB_SELECTOR) != labels.end());
}

const std::set<unsigned int> bb_edge_info::get_labels(const int selector) const
{
   if(labels.find(selector) != labels.end())
      return labels.find(selector)->second;
   else
      return std::set<unsigned int>();
}

void bb_edge_info::print(std::ostream& ) const
{
   THROW_ASSERT(false, "Not implemented function");
}

