/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file basic_block.hpp
 * @brief Class specification of the basic_block structure.
 *
 * This structure is used to represent graphs where nodes are basic_block.
 * A basic block is a sequence of instructions with only one entry and
 * only one exit. If any of the instructions are executed, they
 * will be all executed, and in sequence from first to last.
 * Jumps, if any, start a block, and jumps end a block.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef BASIC_BLOCK_HPP
#define BASIC_BLOCK_HPP

///Autoheader include
#include "config_HAVE_HOST_PROFILING_BUILT.hpp"

///Superclass include
#include "cdfg_edge_info.hpp"

///Behavior include
#include "graph.hpp"
#include "graph_info.hpp"
#include "op_edge_info.hpp"

///STL include
#include <list>
#include <tr1/unordered_map>
#include <tr1/unordered_set>

///Utility include
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/topological_sort.hpp>
#include <boost/graph/filtered_graph.hpp>
#include <boost/graph/graph_utility.hpp>
#include "refcount.hpp"
/**
 * @name forward declarations
 */
//@{
class BasicBlocksGraphConstructor;
REF_FORWARD_DECL(bloc);
CONSTREF_FORWARD_DECL(Parameter);
CONSTREF_FORWARD_DECL(behavioral_helper);
#if HAVE_HOST_PROFILING_BUILT
CONSTREF_FORWARD_DECL(ProfilingInformation);
#endif
//@}

/// Basic block control flow graph edge selector
#define BB_SELECTOR CFG_SELECTOR 
/// Basic block feedback control flow graph edge selector
#define FB_BB_SELECTOR FB_CFG_SELECTOR

/// dominator graph edge selector
#define D_BB_SELECTOR 16
/// post-dominator graph edge selector
#define PD_BB_SELECTOR 32

/// j graph edge selector for dj graph (used during loop computation)
#define J_BB_SELECTOR 64
//@}

/**
 * Definition of the node_info object for the basic_block graph.
 * This object info defines the list of vertexes associated with the basic block node.
*/
struct bb_node_info : public NodeInfo
{

   public:

      ///id of the loop to which basic block belongs to (0 if it doesn't belong to any loop)
      unsigned int loop_id;

      ///id of the control equivalent region
      unsigned int cer;

      ///Structure associated with this basic block in the GCC tree
      blocRef block;

      ///List of operation vertices associated with basic block node
      std::list<vertex> statements_list;

      /**
       * Empty constructor
       */
      bb_node_info();

      /**
       * Constructor which uses gcc information
       * @param _block is the block in the gcc dump
       */
      bb_node_info(blocRef _block);

      /**
       * Adds an operation to the list of the statements
       * @param op is the operation to be added
       */
      void add_operation_node(const vertex op);

      /**
       * Returns the first operation vertex associated with the basic block
       * @return the first operation vertex of the basic block.
       */
      vertex get_first_operation() const;

      /**
       * Returns the last operation vertex associated with the basic block.
       * @return the last operation statement of the basic block.
      */
      vertex get_last_operation() const;

      /**
       * Returns true if there is no node associated with the basic block.
       * @return true if there is no node associated with the basic block
      */
      bool empty() const;

      /**
       * Returns the number of vertices of the graph representing
       * the original program, associated with the current vertex of
       * the basic block graph
       * @return size of basic block
       */
      size_t size() const;

      /**
       * Returns the index of the basic block
       */
      unsigned int get_bb_index() const;

      /**
       * Returns the live in of the basic block
       */
      const std::set<unsigned int> & get_live_in() const;

      /**
       * Returns the live in of the basic block
       */
      const std::set<unsigned int> & get_live_out() const;

};

/**
 * Information associated with a basic block edge
 */
struct bb_edge_info : public CdfgEdgeInfo
{
   private:

      friend class BasicBlocksGraphConstructor;
      friend class function_behavior;

      /// edge instrumented weight
      unsigned long long epp_value;

   public:

      /**
       * Constructor
       */
      bb_edge_info();

      /**
       * Destructor
       */
      ~bb_edge_info();

      /**
       * Function returning true when the edge is a then control dependence edge, false otherwise
       */
      bool cdg_edge_T() const;

      /**
       * Function returning true when the edge is an else control dependence edge, false otherwise
       */
      bool cdg_edge_F() const;

      /**
       * Function returning true when the edge is a then control flow edge, false otherwise
       */
      bool cfg_edge_T() const;

      /**
       * Function returning true when the edge is an else control flow edge, false otherwise
       */
      bool cfg_edge_F() const;

      /**
       * Function returning true when the edge is a control flow edge exiting from a switch, false otherwise
       */
      bool switch_p() const;

      /**
       * Return the labels associated with a selector
       */
      const std::set<unsigned int> get_labels(const int selector) const;

      /**
       * Function that sets the epp_edge associated with the edge
       */
      void set_epp_value(unsigned long long _epp_value);

      /**
       * Function returning the epp_value associated with the edge
       */
      unsigned long long get_epp_value() const;

      /**
       * Prints the information associated with the edge of the graph.
       * @param os is the output stream.
       */
      void print(std::ostream& os) const;
};

/**
 * Information associated with the whole basic-block graph
 */
struct BBGraphInfo : public graph_info
{
   std::tr1::unordered_map<unsigned int, vertex> bb_index_map;

   /// Index identifying the entry basic block.
   vertex entry_vertex;

   /// Index identifying the exit basic block.
   vertex exit_vertex;

#if HAVE_HOST_PROFILING_BUILT
   ///Profiling information
   const ProfilingInformationConstRef profiling_information;

   /**
    * Constructor with profiling information
    * @param _profiling_information is the profiling information
    */
   BBGraphInfo(const ProfilingInformationConstRef _profiling_information);
#endif

   /**
    * Empty constructor
    */
   BBGraphInfo() {}
};

/**
 * This structure defines graphs where nodes are basic_blocks.
 * Graphs defined are: control flow graph, control dependence graph, dominator tree and post-dominator tree
 * The basic_block structure and the control flow graph can be extracted directly from the raw structure when cfg
 * pass is done or built up starting from a CFG where nodes are standard vertexes.
*/
class bb_graphs_collection : public graphs_collection
{
   private:
      friend class BasicBlocksGraphConstructor;

      /**
       * Creates and returns the info associated with an edge
       */
      bb_edge_info * create_bb_edge_info(const EdgeDescriptor e);

   public:

      /**
       * Empty constructory
       */
      bb_graphs_collection();

#if HAVE_HOST_PROFILING_BUILT
      /**
       * Constructor with profiling information
       * @param _profiling_information is the profiling information
       */
      bb_graphs_collection(const ProfilingInformationConstRef _profiling_information);
#endif

      /**
       * Destructor
       */
      ~bb_graphs_collection();

      /**
       * Function that prints the class basic_block.
       * @param os is the output stream.
      */
      void print(std::ostream& os);

      /**
       * Returns the info associated with an edge; if it does not exist, this method throws assert;
       * info should be always present since BasicBlocksGraphConstructor::add_edge creates it
       */
      bb_edge_info * get_bb_edge_info(const EdgeDescriptor e) const;

      /**
       * Returns the property associated with the graph
       * @return the proprty associated with the graph
       */
      inline
      const BBGraphInfo * GetConstBBGraphInfo() const
      {
         const graph_infoRef info = boost_get_graph_property(*this);
         return GetPointer<BBGraphInfo>(info);
      }

      /**
       * Returns the property associated with the graph
       * @return the property associated with the graph
       */
      inline
      BBGraphInfo * GetBBGraphInfo()
      {
         graph_infoRef info = boost_get_graph_property(*this);
         return GetPointer<BBGraphInfo>(info);
      }
};

/**
 * Class used to describe a particular graph with basic blocks as nodes
 */
struct bb_graph : public graph
{
   private:

      ///The bb_graph collection
      bb_graphs_collection * g;

   public:

      /// @name graph constructors
      //@{
      /**
       * Standard constructor.
       * @param g is the bulk graph.
       * @param g is the selector used to filter the bulk graph.
       */
      bb_graph(bb_graphs_collection * g, int selector);

      /**
       * Sub-graph constructor.
       * @param g is the bulk graph.
       * @param g is the selector used to filter the bulk graph.
       * @param sub is the set of vertexes on which the graph is filtered.
       */
      bb_graph(bb_graphs_collection * g, int selector, std::set<boost::graph_traits<bb_graphs_collection>::vertex_descriptor > sub);
      //@}

      /**
       * Destructor
       */
      ~bb_graph(){}

      /**
       * Writes this graph in dot format
       * @param file_name is the file where the graph has to be printed
       * @param BH is the helper to be used
       * @param annotated is the set of the vertices to be annotated
       * @return 1 if no error occurs
       */
      int PrintToDot(const ParameterConstRef Param, std::string file_name, const behavioral_helperConstRef BH, std::tr1::unordered_set<vertex> annotated = std::tr1::unordered_set<vertex>()) const;

      /**
       * Returns the number of basic blocks contained into the graph
       */
      size_t num_bblocks() const;

      /**
       * Returns the info associated with an edge; if it does not exist it, this method throws assert
       * info should be always present since BasicBlocksGraphConstructor::add_edge creates it
       */
      bb_edge_info * get_bb_edge_info(const EdgeDescriptor e) const;

      /**
       * Returns the property associated with the graph
       * @return the proprty associated with the graph
       */
      inline
      const BBGraphInfo * GetConstBBGraphInfo() const
      {
         return g->GetConstBBGraphInfo();
      }

      /**
       * Returns the property associated with the graph
       * @return the property associated with the graph
       */
      inline
      BBGraphInfo * GetBBGraphInfo()
      {
         return g->GetBBGraphInfo();
      }
};
///refcount definition of the class
typedef refcount<bb_graph> bb_graphRef;

/**
 * Helper Macro returning const info associated with a basic block node
 * @param data is the basic block graph
 * @param vertex_index is the vertex node
*/
#define CGET_BB_NODE_INFO(data, vertex_index) \
   CGET_NODE_INFO(data, bb_node_info, vertex_index)

/**
 * Helper Macro returning info associated with a basic block nodes
 * @param data is the basic block graphs
 * @param vertex_index is the node
 */
#define GET_BB_NODE_INFO(data, vertex_index) \
   GET_NODE_INFO(data, bb_node_info, vertex_index)

/**
 * The key comparison function for vertices set based on levels
 */
class bb_vertex_order_by_map : std::binary_function<vertex, vertex, bool>
{
   private:

      ///Topological sorted vertices
      const std::map<vertex, unsigned int> & ref;

   public:

      /**
       * Constructor
       * @param ref_ is the map with the topological sort of vertices
       */
      bb_vertex_order_by_map(const std::map<vertex, unsigned int> & _ref) :
         ref(_ref)
         {}

      /**
       * Compares position of two vertices sorted in topological order
       * @param x is the first vertex
       * @param y is the second vertex
       * @return true if x precedes y in the topological order, false otherwise
       */
      bool operator()(const vertex x, const vertex y) const
      {
         return ref.find(x)->second < ref.find(y)->second;
      }
};

#endif
