/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file behavioral_writer_helper.cpp
 * @brief Implement all structs used to write a graph in the graphviz format
 *
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
///Autoheader include
#include "config_HAVE_HOST_PROFILING_BUILT.hpp"

///Header include
#include "behavioral_writer_helper.hpp"

///Behavior include
#include "basic_block.hpp"
#include "behavioral_helper.hpp"
#include "op_edge_info.hpp"
#include "op_node_info.hpp"
#if HAVE_HOST_PROFILING_BUILT
#include "profiling_information.hpp"
#endif

///STD include
#include <algorithm>

///Tree include
#include "tree_reindex.hpp"
#include "tree_basic_block.hpp"

///Utility include
#include "boost/algorithm/string/replace.hpp"
#include "exceptions.hpp"
#include "simple_indent.hpp"
#include "var_pp_functor.hpp"

bb_label_writer::bb_label_writer(const bb_graph * _g, const behavioral_helperConstRef _helper, std::tr1::unordered_set<vertex> _annotated) :
   g(_g),
   helper(_helper),
   annotated(_annotated)
{}

void bb_label_writer::operator()(std::ostream& out, const vertex& v) const
{
   const BBGraphInfo * info = g->GetConstBBGraphInfo();
   const bb_node_info * operations = CGET_BB_NODE_INFO(g, v);
   if(v == info->entry_vertex)
   {
      out << "[color=blue,shape=Msquare, ";
      if(annotated.find(v) != annotated.end())
         out << " style=filled, fillcolor=black, fontcolor=white,";
      out << "label=\"ENTRY";
   }
   else if(v == info->exit_vertex)
   {
      out << "[color=blue,shape=Msquare, ";
      if(annotated.find(v) != annotated.end())
      {
         out << " style=filled, fillcolor=black, fontcolor=white,";
      }
      out << "label=\"EXIT";
   }
   else
   {
      out << "[shape=box";
      if(annotated.find(v) != annotated.end())
         out << ", style=filled, fillcolor=black, fontcolor=white";
      if(operations && operations->block)
      {
         out << ", label=\"BB" << operations->block->number << " - Cer: " << operations->cer;
         out << " - Loop " << operations->loop_id;
#if HAVE_HOST_PROFILING_BUILT
         out << " - Executions: " << info->profiling_information->GetBBExecutions(v);
#endif
      }
      if(operations && !operations->block->list_of_phi.empty())
      {
         std::vector<tree_nodeRef>::const_iterator phi_op, phi_op_end = operations->block->list_of_phi.end();
         if(helper)
         {
            out << "\\n";
            for(phi_op = operations->block->list_of_phi.begin(); phi_op != phi_op_end; phi_op++)
            {
               const var_pp_functorConstRef svpf(new std_var_pp_functor(helper));
               std::string res = STR(GET_INDEX_NODE(*phi_op)) + " -> " + helper->print_node(GET_INDEX_NODE(*phi_op), 0, svpf);
               std::string temp;
               for (unsigned int i = 0; i < res.size(); i++)
               {
                  if (res[i] == '\"')
                     temp += "\\\"";
                  else if (res[i] != '\n')
                     temp += res[i];
               }
               out << temp << "\\n";
            }
         }
      }
      if(operations && !operations->block->list_of_stmt.empty())
      {
         std::vector<tree_nodeRef>::const_iterator op, op_end = operations->block->list_of_stmt.end();
         if(helper)
         {
            if(operations->block->list_of_phi.empty())
               out << "\\n";
            for(op = operations->block->list_of_stmt.begin();op != op_end; op++)
            {
               const var_pp_functorConstRef svpf(new std_var_pp_functor(helper));
               std::string res = STR(GET_INDEX_NODE(*op)) + " -> " + helper->print_node(GET_INDEX_NODE(*op), 0, svpf);
               const tree_nodeRef node = GET_NODE(*op);
               switch (node->get_kind())
               {
                  case gimple_assign_K:
                  case gimple_call_K:
                  case gimple_asm_K:
                  case gimple_goto_K:
                     res += ";";
                     break;
                  default:
                     break;
               }
               std::string temp;
               for (unsigned int i = 0; i < res.size(); i++)
               {
                  if (res[i] == '\"')
                     temp += "\\\"";
                  else if (res[i] != '\n')
                     temp += res[i];
               }
               out << temp << "\\n";
            }
         }
      }
   }
   out << "\"]";
}

OpEdgePropertyWriter::OpEdgePropertyWriter(const graph *_g, int selector, const behavioral_helperConstRef _BH) :
   g(_g),
   s(selector),
   BH(_BH)
{
}

void OpEdgePropertyWriter::operator()(std::ostream& out, const EdgeDescriptor& e) const
{
   if((FB_CFG_SELECTOR) & s & g->get_selector(e))
      out << "[color=gold";
   else if((CFG_SELECTOR) & s & g->get_selector(e))
      out << "[fontcolor=red3";
   else if((FB_CDG_SELECTOR) & s & g->get_selector(e) && FB_DFG_SELECTOR & s & g->get_selector(e))
      out << "[color=gold,style=dotted";
   else if((FB_CDG_SELECTOR) & s & g->get_selector(e))
      out << "[color=gold";
   else if(((CDG_SELECTOR) & s & g->get_selector(e)) && ((DFG_SELECTOR) & s & g->get_selector(e)))
      out << "[color=red3,style=dotted";
   else if((CDG_SELECTOR) & s & g->get_selector(e))
      out << "[color=red3";
   else if((FB_DFG_SELECTOR) & s & g->get_selector(e))
      out << "[color=lightblue";
   else if((DFG_SSA_SELECTOR) & s & g->get_selector(e))
      out << "[color=blue, style=dotted";
   else if((DFG_SER_SELECTOR) & s & g->get_selector(e))
      out << "[color=blue";
   else if((FB_ADG_SELECTOR) & s & g->get_selector(e))
      out << "[color=lawngreen";
   else if((ADG_SELECTOR) & s & g->get_selector(e))
      out << "[color=green4";
   else if((ODG_SELECTOR) & s & g->get_selector(e))
      out << "[color=green4";
   else if((CDG_SELECTOR) & s & g->get_selector(e))
      out << "[color=red3";
   else if((DFG_SELECTOR) & s & g->get_selector(e))
      out << "[color=blue";
   else if((FB_CDG_SELECTOR) & s & g->get_selector(e))
      out << "[color=gold";
   else if((FB_DFG_SELECTOR) & s & g->get_selector(e))
      out << "[color=lightblue";
   else if((FLG_SELECTOR) & s & g->get_selector(e))
      out << "[color=red3";
   else if((CSG_SELECTOR) & s & g->get_selector(e) && DFG_SELECTOR & s & g->get_selector(e))
      out << "[color=pink,style=dotted";
   else if((CSG_SELECTOR) & s & g->get_selector(e))
      out << "[color=pink";
   else if((FB_FG_SELECTOR) & s & g->get_selector(e))
      out << "[color=pink";

   const OpEdgeInfo * edge_info = Cget_edge_info<OpEdgeInfo>(e, *g);

   if(edge_info)
   {
      out << ",label=\"";
      if(g->get_selector(e) & s & FCDG_SELECTOR)
      {
         if(edge_info->CdgEdgeT() || edge_info->CdgEdgeF() || edge_info->Switch())
         {
            out << edge_info->PrintLabels(CDG_SELECTOR, BH);
         }
      }
      if(g->get_selector(e) & s & FCFG_SELECTOR)
      {
         if(edge_info->CfgEdgeT() || edge_info->CfgEdgeF() || edge_info->Switch())
         {
            out << edge_info->PrintLabels(CFG_SELECTOR, BH);
         }
      }
      if(g->get_selector(e) & s & FLG_SELECTOR)
      {
         if(edge_info->FlgEdgeT() || edge_info->FlgEdgeF())
         {
            out << edge_info->PrintLabels(FLG_SELECTOR, BH);
         }
      }
      if(g->get_selector(e) & s & FDFG_SELECTOR)
      {
         out << edge_info->PrintLabels(DFG_SELECTOR, BH);
      }
      if(g->get_selector(e) & s & FADG_SELECTOR)
      {
         out << edge_info->PrintLabels(ADG_SELECTOR, BH);
      }
      out << "\"";
   }
   out << "]";
}

bb_edge_property_writer::bb_edge_property_writer(const bb_graph * _g, int _s, const behavioral_helperConstRef _BH) :
   g(_g),
   s(_s),
   BH(_BH)
{
}

void bb_edge_property_writer::operator()(std::ostream & out, const EdgeDescriptor & e) const
{
   if(FB_BB_SELECTOR & g->get_selector(e))
      out << "[fontcolor=blue, color=gold";
   else if(BB_SELECTOR & g->get_selector(e))
      out << "[fontcolor=blue, color=red3";
   else if(CDG_SELECTOR & g->get_selector(e))
      out << "[fontcolor=blue, color=red3";
   else if(D_BB_SELECTOR & g->get_selector(e))
      out << "[fontcolor=blue";
   else if(PD_BB_SELECTOR & g->get_selector(e))
      out << "[fontcolor=blue";
   else if(FLG_SELECTOR & g->get_selector(e))
      out << "[fontcolor=blue, color=gold";
   else if(PP_SELECTOR & g->get_selector(e))
      out << "[fontcolor=blue, color=gold";
   else
      THROW_ERROR("Not supported graph type in printing");
   const bb_edge_info * edge_info = g->get_bb_edge_info(e);
   if(s & FLG_SELECTOR)
   {
      out << ",label=\"";
      out <<  boost::lexical_cast<std::string>(edge_info->get_epp_value());
      out << "\"";
   }
   else if(s & FCFG_SELECTOR)
   {
      if(edge_info->CfgEdgeT() || edge_info->CfgEdgeF() || edge_info->Switch())
      {
         out << ",label=\"";
         out << edge_info->PrintLabels(BB_SELECTOR, BH);
         out << "\"";
      }
   }
   else if(s & FCDG_SELECTOR)
   {
      if(edge_info->CdgEdgeT() || edge_info->CdgEdgeF() || edge_info->Switch())
      {
         out << ",label=\"";
         out << edge_info->PrintLabels(CDG_SELECTOR, BH);
         out << "\"";
      }
   }
   out << "]";
}

cdfg_label_writer::cdfg_label_writer(const graph *_g, const behavioral_helperConstRef _helper) : g(_g), helper(_helper){}

void cdfg_label_writer::operator()(std::ostream& out, const vertex& v) const
{
   THROW_ASSERT((!(GET_TYPE(g, v) & TYPE_EXTERNAL) && GET_CALLED(g, v).empty()) || ((GET_TYPE(g, v) & TYPE_EXTERNAL) && GET_CALLED(g, v).size()), GET_NAME(g,v) + " is an external node without function id associated");
   if(GET_TYPE(g, v) & (TYPE_IF | TYPE_SWITCH))
      out << "[color=red,shape=diamond,";
   else if(GET_TYPE(g, v) & TYPE_LOAD)
      out << "[color=green,shape=box,";
   else if(GET_TYPE(g, v) & TYPE_STORE)
      out << "[color=red,shape=box,";
   else if(GET_TYPE(g, v) & TYPE_EXTERNAL)
      out << "[color=green,shape=Msquare,";
   else if(GET_TYPE(g, v) & TYPE_MEMCPY)
      out << "[color=burlywood,shape=diamond,";
   else if(GET_TYPE(g, v) & TYPE_GOTO)
      out << "[color=yellow,shape=box,";
   else if(GET_TYPE(g, v) & TYPE_ASSIGN)
      out << "[color=burlywood,shape=box,";
   else if(GET_TYPE(g, v) & (TYPE_ENTRY | TYPE_EXIT))
      out << "[color=blue,shape=Msquare,";
   else
      out << "[";
   out << "label=\"" << GET_NAME(g, v);
   const var_pp_functorConstRef svpf(new std_var_pp_functor(helper));
   if(!GET_NODEID(g, v).empty())
   {
      out << "\\n";
      simple_indent PP(STD_OPENING_CHAR, STD_CLOSING_CHAR, 0);
      helper->print_vertex(out, g, v, svpf, PP, true);
   }
   else
   {
      if(GET_TYPE(g, v) & TYPE_ASSIGN)
      {
         const std::tr1::unordered_set<unsigned int>& ssa_written = GET_SSA_WRITTEN(g, v);
         for(std::tr1::unordered_set<unsigned int>::const_iterator it = ssa_written.begin(); it !=ssa_written.end(); it++)
            out << helper->PrintVariable(*it);
         const std::list<unsigned int>& ssa_read = GET_SSA_READ(g, v);
         if(!ssa_read.empty())
         {
            out << " = ";
            for(std::list<unsigned int>::const_iterator it = ssa_read.begin(); it != ssa_read.end(); it++)
               out << helper->PrintVariable(*it);
            out << ";";
         }
      }
   }
   out << "\"]";
}
