/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file behavioral_writer_helper.hpp
 * @brief Collect all structs used to write a graph in the graphviz format
 *
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef BEHAVIORAL_WRITER_HELPER_HPP
#define BEHAVIORAL_WRITER_HELPER_HPP

///Autoheader include
#include "config_HAVE_HOST_PROFILING_BUILT.hpp"

///STD include
#include <ostream>

///STL include
#include <tr1/unordered_set>

///Utility include
#include "refcount.hpp"

/**
 * @name forward declarations
*/
//@{
CONSTREF_FORWARD_DECL(behavioral_helper);
#if HAVE_HOST_PROFILING_BUILT
CONSTREF_FORWARD_DECL(ProfilingInformation);
#endif
REF_FORWARD_DECL(basic_block_graph);
class bb_graph;
//@}

#include "graph.hpp"

/**
 * Edge writer for operation graph
*/
class OpEdgePropertyWriter
{
   private:
      /// operation graph
      const graph *g;

      ///Selector of the graph
      int s;

      ///The helper used to print the labels
      const behavioral_helperConstRef BH;

   public:
      /**
       * Constructor
       * @param _g is the operation graph
       * @param selector is the selector of the graph
       * @param BH is the helper used to print the labels
       */
      OpEdgePropertyWriter(const graph *_g, int selector, const behavioral_helperConstRef BH);

      /**
       * Operator which print label of an EdgeDescriptor
       * @param out is the stream
       * @param e is the edge
       */
      void operator()(std::ostream& out, const EdgeDescriptor& e) const;
};

/**
 * Class which prints the edge of a basic block graph in dot format
 */
class bb_edge_property_writer
{
   private:
      ///graph to be printed
      const bb_graph * g;

      ///Selector of the graph
      int s;

      ///The helper used to print the labels
      const behavioral_helperConstRef BH;

   public:
      /**
       * Constructor
       * @param g is the bb_graph to be printed
       * @param s is the selector of the bb_graph
       * @param BH is the behavioral helper
       */
      bb_edge_property_writer(const bb_graph * g, int s, const behavioral_helperConstRef BH);

      /**
       * Operator used to print an edge
       * @param out is the stream on which edge should be printed
       * @param e is the edge to be printed
       */
      void operator()(std::ostream & out, const EdgeDescriptor & e) const;
};

class cdfg_label_writer
{
   public:
      cdfg_label_writer(const graph *_g,  const behavioral_helperConstRef _helper);
      void operator()(std::ostream& out, const vertex& v) const;
   private:
      const graph *g;
      const behavioral_helperConstRef helper;
};

class bb_label_writer
{
   private:
      ///The graph to be printed
      const bb_graph * g;

      ///The helper
      const behavioral_helperConstRef helper;

      ///The set of vertices to be annotated
      std::tr1::unordered_set<vertex> annotated;

   public:
      /**
       * The constructor
       * @param g is the graph to be printed
       * @param helper is the helper used to pint labels
       * @param profiling_information is the profiling information
       * @param annotated is the set of the vertices to be annotated
       */
      bb_label_writer(const bb_graph * g, const behavioral_helperConstRef helper, std::tr1::unordered_set<vertex> annotated = std::tr1::unordered_set<vertex>());

      /**
       * Operator used to print the label of a vertex
       * @param out is the stream where label has to printed
       * @param v is the vertex to be printed
       */
      void operator()(std::ostream& out, const vertex& v) const;
};

#endif
