/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file call_graph.cpp
 * @brief Call graph hierarchy.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Autoheader include
#include "config_HAVE_HOST_PROFILING_BUILT.hpp"
#include "config_HAVE_POLIXML_BUILT.hpp"

///Header include
#include "call_graph.hpp"

///Behavior include
#include "behavioral_helper.hpp"
#include "function_behavior.hpp"
#include "loop.hpp"
#include "loops.hpp"
#include "op_node_info.hpp"
#if HAVE_HOST_PROFILING_BUILT
#include "profiling_information.hpp"
#endif
///Boost include
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>

///Graph include
#include "graph.hpp"
#include "basic_block.hpp"
#include "tree_basic_block.hpp"
#include <boost/graph/graphviz.hpp>

///Paramter include
#include "Parameter.hpp"

///Utility include
#include <boost/date_time/posix_time/posix_time.hpp>

///Wrapper include
#include "gcc_wrapper.hpp"

///XML include
#if HAVE_POLIXML_BUILT
#include "xml_helper.hpp"
#include "polixml.hpp"
#include "xml_dom_parser.hpp"
#include "xml_document.hpp"
#endif

/**
 * Helper macro adding a call point to an edge of the call graph
 * @param g is the graph
 * @param e is the edge
 * @param newstmt is the call point to be added
 */
#define ADD_CALL_POINT(g, e, newstmt) \
   get_edge_info<function_graph_edge_info>(e, *(g))->call_points.insert(newstmt)


/**
 * @name function graph selector
 */
//@{
/// Data line selector
#define STD_SELECTOR 1 << 0
/// Clock line selector
#define FEEDBACK_SELECTOR 1 << 1
//@}

/**
 * data structure used to store the graph describing the relation called_by.
 */
struct function_graph : public graphs_collection
{
   ///Constructor of graph.
   function_graph() : graphs_collection(graph_infoRef(new graph_info)) {}

   ///Destructor
   ~function_graph(){}

};

template<class info_object>
const info_object* Cget_graph_info(const function_graph &g)
{
   graph_info * info = g.get_const_graph_property().get();
   THROW_ASSERT(!info || dynamic_cast<info_object*>(info) != 0, "Function Cget_graph_info: the graph property associated with the graph is not derived from info_object\n\tCheck the actual type of info_object and the type of the graph property");
   return info ? dynamic_cast<info_object*>(info) : 0;
}

/**
 * Functor used by write_graphviz to write the label of the vertices of a function graph
*/
class function_graph_label_writer
{
   public:

      /**
       * constructor
       * @param g is the graph
       */
      function_graph_label_writer(const graph *_g, const std::map<unsigned int, function_behaviorRef> &_b) : g(_g), behaviors(_b) {}

      /**
       * operator function returning the label of the vertex
       * @param out is the output stream
       * @param v is the vertex
      */
      void operator()(std::ostream& out, const vertex& v) const
      {
         const function_behaviorRef FB = behaviors.find(GET_FUNCTION_ID(g, v))->second;
         out << "[shape=box, label=\"" << FB->get_helper()->get_function_name();
         const std::set<unsigned int>& mem_nodeID = FB->get_function_mem();
         if (mem_nodeID.size())
         {
            out << "\\n(";
            bool first = true;
            for(std::set<unsigned int>::const_iterator l = mem_nodeID.begin(); l != mem_nodeID.end(); l++)
            {
               if (!first) out << " ";
               out << FB->get_helper()->PrintVariable(*l);
               first = false;
            }
            out << ")";
         }
         out << "\"]" ;
      }

   private:

      ///reference to the graph
      const graph *g;
      ///reference to the behaviors
      const std::map<unsigned int, function_behaviorRef> &behaviors;
};

/**
 * Functor used by write_graphviz to write the edges of a function graph
*/
class function_graph_edge_writer
{
   public:
      /**
       * constructor
       * @param g is the graph
      */
      function_graph_edge_writer(const graph * _g, const std::map<unsigned int, function_behaviorRef> &_b) : g(_g), behaviors(_b) {}

      /**
       * operator function returning the edge description
       * @param out is the output stream
       * @param e is the edge
       */
      void operator()(std::ostream& out, const EdgeDescriptor& e) const
      {
         if (STD_SELECTOR & g->get_selector(e))
            out << "[color=blue, label=\"";
         else if (FEEDBACK_SELECTOR & g->get_selector(e))
               out << "[color=red, label=\"";
         else
            THROW_ERROR(std::string("InconsistentDataStructure"));
         const std::set<unsigned int>& call_points = GET_CALL_POINTS(g, e);
         std::set<unsigned int>::const_iterator s, s_end = call_points.end();
         for(s = call_points.begin(); s != s_end; s++)
         {
            out << *s << " ";
         }
         out << "\"]";
      }

   private:

      ///reference to the graph
      const graph * g;

      ///reference to the behaviors
      const std::map<unsigned int, function_behaviorRef> &behaviors;
};

/**
 * Class used to build the graph describing the relation called_by
 */
struct function_graph_constructor
{
   public:
      /**
       * Constructor.
      */
      function_graph_constructor(std::map<unsigned int, vertex>& functionID_vertex_map_) : functionID_vertex_map(functionID_vertex_map_){}
      /**
       * Add vertex to the call graph
       * @param fg is the call graph 
       * @param functionID is the function identifier
       */
      vertex add_vertex(function_graph* fg, unsigned int functionID)
      {
         THROW_ASSERT(functionID_vertex_map.find(functionID) == functionID_vertex_map.end(), "this vertex has been already added " + boost::lexical_cast<std::string>(functionID));
         vertex v = fg->AddVertex();
         GET_NODE_INFO(fg, function_graph_node_info, v)->nodeID = functionID;
         functionID_vertex_map[functionID] = v;
         return v;

      }
      /**
       * return true in case the vertex has been already created
       * @param functionID is the function identifier
       */
      bool check_vertex(unsigned int functionID) const
      {
         return functionID_vertex_map.find(functionID) != functionID_vertex_map.end();
      }
      /**
       * return a vertex of the graph given the functionID.
       * @param functionID is the function identifier
       */
      vertex get_vertex(unsigned int functionID) const
      {
         THROW_ASSERT(functionID_vertex_map.find(functionID) != functionID_vertex_map.end(), "this vertex does not exist " + boost::lexical_cast<std::string>(functionID));
         return functionID_vertex_map.find(functionID)->second;
      }
      /**
       * add an edge to the call graph
       * @param fg is the call graph
       * @param v1 is the source vertex
       * @param v2 is the target vertex
       * @param selector is the type of the edge
       */
      void add_edge(function_graph* fg, vertex v1, vertex v2, int selector) const 
      {
         ADD_EDGE(v1, v2, selector, *fg);
      }
   private:
      /// put into relation functionID with vertex
      std::map<unsigned int, vertex>& functionID_vertex_map;
};

void function_graph_node_info::print(std::ostream& os) const
{
   os << nodeID << std::endl;
}

void function_graph_edge_info::print(std::ostream& os) const
{
   std::set<unsigned int>::const_iterator s, s_end = call_points.end();
   for(s = call_points.begin(); s != s_end; s++)
   {
      os << *s << " ";
   }
   os << std::endl;
}

call_graph::call_graph(const ParameterConstRef _Param) :
   Param(_Param),
   debug_level(_Param->get_class_debug_level(GET_CLASS(*this)))
{
   bulk_called_by_graph = new function_graph;
   called_by_graph = new graph(bulk_called_by_graph, STD_SELECTOR | FEEDBACK_SELECTOR);
   FCG = function_graph_constructorRef(new function_graph_constructor(index_to_vertex));
}

call_graph::~call_graph()
{
   delete bulk_called_by_graph;
   delete called_by_graph;
}

void call_graph::add_function(unsigned int current, unsigned int index, const behavioral_helperRef helper, unsigned int node_stmt)
{
   THROW_ASSERT(FCG, "this function cannot be called after the call graph has been consolidated");
   add_function(index, helper);
   THROW_ASSERT(FCG->check_vertex(index), "this function should be already added to the call_graph");
   vertex source = FCG->get_vertex(current);
   vertex target = FCG->get_vertex(index);
   if(called_by[current].find(index) == called_by[current].end())
   {
      called_by[current].insert(index);
      FCG->add_edge(bulk_called_by_graph, source, target, STD_SELECTOR);
   }
   EdgeDescriptor e;
   bool found;
   tie(e,found) = boost::edge(source, target, *called_by_graph);
   ADD_CALL_POINT(bulk_called_by_graph, e, node_stmt);
}

void call_graph::add_function(unsigned int index, const behavioral_helperRef helper)
{
   if(!FCG->check_vertex(index))
      FCG->add_vertex(bulk_called_by_graph, index);
   function_helper[index] = helper;
}

void call_graph::print(std::ostream& os) const
{
   os << "Call graph " << std::endl;
   std::map<unsigned int, behavioral_helperRef>::const_iterator it_end = function_helper.end();
   for(std::map<unsigned int, behavioral_helperRef>::const_iterator it = function_helper.begin(); it != it_end; it++)
   {
      //it->second->print(os);
      os << it->second->get_function_name();
      if(called_by.find(it->first) == called_by.end())
      {
         os << std::endl;
         continue;
      }
      const std::set<unsigned int> & called = called_by.find(it->first)->second;
      std::set<unsigned int>::const_iterator cit_end = called.end();
      std::set<unsigned int>::const_iterator cit = called.begin();
      if(cit != cit_end)
         os << " Use/call";
      for(; cit != cit_end; cit++)
         os << " " << function_helper.find(*cit)->second->get_function_name();
      os << std::endl;
   }
}

int call_graph::write_dot(const std::string &file_name, const std::map<unsigned int, function_behaviorRef>& behaviors) const
{
   std::string output_directory = Param->getOption<std::string>(OPT_output_directory) + "/dot/";
   if (!boost::filesystem::exists(output_directory)) boost::filesystem::create_directories(output_directory);
   std::ofstream f((output_directory + file_name).c_str());
   boost::write_graphviz(f, *called_by_graph, function_graph_label_writer(called_by_graph, behaviors), function_graph_edge_writer(called_by_graph, behaviors));
   return 1;
}

void call_graph::consolidate()
{
   FCG = function_graph_constructorRef();
}

vertex call_graph::get_vertex(unsigned int index)
{
   THROW_ASSERT(index_to_vertex.find(index) != index_to_vertex.end(), "Function not stored");
   return index_to_vertex.find(index)->second;
}

unsigned int call_graph::get_function(vertex node) const
{
   for(std::map<unsigned int, vertex>::const_iterator i = index_to_vertex.begin(); i != index_to_vertex.end(); i++)
      if (i->second == node)
         return i->first;
   return 0;
}

const graph* call_graph::get_graph() const
{
    return called_by_graph;
}

const graphRef call_graph::get_subgraph(const std::set<vertex>& subset) const
{
   return graphRef(new graph(bulk_called_by_graph,STD_SELECTOR | FEEDBACK_SELECTOR,subset));
}

void call_graph::replace_call_point(EdgeDescriptor e, const unsigned int orig, const unsigned int repl)
{
   std::set<unsigned int> & call_points = get_edge_info<function_graph_edge_info>(e, *called_by_graph)->call_points;
   call_points.erase(call_points.find(orig));
   call_points.insert(repl);
}

const std::set<unsigned int> call_graph::get_called_by(unsigned int index) const
{
   if(called_by.find(index) != called_by.end())
      return called_by.find(index)->second;
   else
      return std::set<unsigned int>();
}

const std::tr1::unordered_set<unsigned int> call_graph::get_called_by(const graph* cfg, const vertex & caller) const
{
   return GET_CALLED(cfg, caller);
}

void call_graph::get_root_functions(std::tr1::unordered_set<unsigned int> & functions)
{
   VertexIterator function, function_end;
   for (boost::tie(function, function_end) = boost::vertices(*called_by_graph); function != function_end; function++)
   {
      unsigned int fun_id = get_function(*function);
      THROW_ASSERT(fun_id>0, "expected a meaningful function id");
      if (boost::in_degree(*function, *called_by_graph) == 0 && function_helper.find(fun_id) != function_helper.end() && function_helper.find(fun_id)->second->has_implementation())
         functions.insert(fun_id);
   }
}

