/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file call_graph.hpp
 * @brief Call graph hierarchy.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef CALL_GRAPH_HPP
#define CALL_GRAPH_HPP

///Graph include
#include "graph.hpp"
#include "graph_info.hpp"
#include "node_info.hpp"

///STL include
#include <list>
#include <map>
#include <set>
#include <tr1/unordered_set>
#include <tr1/unordered_map>

///Utility include
#include "refcount.hpp"

/**
 * @name forward declarations
 */
//@{
REF_FORWARD_DECL(behavioral_helper);
REF_FORWARD_DECL(bloc);
REF_FORWARD_DECL(call_graph);
REF_FORWARD_DECL(function_behavior);
REF_FORWARD_DECL(function_graph_constructor);
CONSTREF_FORWARD_DECL(Parameter);

class function_graph;
class xml_dom_parser;
class xml_element;
//@}


/**
 * This class collects information concerning the set of functions that will be analyzed by the PandA framework
*/
class call_graph
{
      ///definition of the class that can access the private methods/attributes
      friend class application_manager;

   private:

      ///describe the relation called_by in term of graph.
      function_graph * bulk_called_by_graph;

      ///external view of the call graph
      graph * called_by_graph;

      ///put into relation function F_i and the list of functions called by F_i
      std::map<unsigned int, std::set<unsigned int> > called_by;

      ///put into relation function F_i and the list of functions called by F_i
      std::map<unsigned int, vertex> index_to_vertex;

      ///datastructure used to build the call_graph
      function_graph_constructorRef FCG;

      ///set of input parameters
      const ParameterConstRef Param;

      ///the debug level
      const int debug_level;

      std::map<unsigned int, behavioral_helperRef> function_helper;

      /**
       * Creates a new function_behavior for function index when required and update the called_by_graph.
       * @param current is the current function
       * @param index is the index of the called function
       * @param helper is the behavioral_helper associated with the called function
       * @param node_stmt is the id of the instruction where the call is performed
       */
      void add_function(unsigned int current, unsigned int index, const behavioral_helperRef helper, unsigned int node_stmt = 0);


      /**
       * Creates a new function_behavior for function index when required and update the called_by_graph. Note that index could not be called by any other function.
       * @param index is the index of the function
       * @param helper is the behavioral_helper associated with the called function
       */
      void add_function(unsigned int index, const behavioral_helperRef helper);

      /**
       * Writes the call graph in a dot format.
       * @param file_name is the name of the file to be written.
       * @return 1 if no error occurs
       */
      int write_dot(const std::string &file_name, const std::map<unsigned int, function_behaviorRef>& behaviors) const;

  public:

      /**
       * Constructor. The data structure is initialized.
       * @param main is the index of the main function
       * @param Param is the set of input parameters
       */
      call_graph(const ParameterConstRef Param);

      /**
       * Destructor
       */
      ~call_graph();

      /**
       * Returns the set of functions called by a function
       * @param index is the index of the caller function
       */
      const std::set<unsigned int> get_called_by(unsigned int index) const;

      /**
       * Returns the set of functions called by an operation vertex
       * @param cfg is the pointer to the graph which the operation belongs to
       * @param caller is the caller vertex
       */
      const std::tr1::unordered_set<unsigned int> get_called_by(const graph* cfg, const vertex & caller) const;

      /**
       * Deallocates the function graph constructor. After the call of this function is not possibile to change the call graph data structure anymore.
      */
      void consolidate();

      /**
       * Function that prints the class call_graph.
       * @param os is the output stream.
       */
      void print(std::ostream& os) const;

      /**
       * Friend definition of the << operator.
       * @param os is the output stream
       */
      friend std::ostream& operator<<(std::ostream& os, call_graph & s)
      {
         s.print(os);
         return os ;
      }

      /**
       * Friend definition of the << operator. Pointer version.
       * @param os is the output stream
       */
      friend std::ostream& operator<<(std::ostream& os, call_graphRef & s)
      {
         if (s)
            s->print(os);
         return os;
      }

      /**
       * Returns the vertex associated with an index of a function
       * @param the index of the function
       * @return the vertex of the function in the call graph
       */
      vertex get_vertex(unsigned int index);

      /**
       * Given a vertex of the call graph, this returns the index of the corresponding function
       * @param node is the vertex of the function
       * @return the index of the function
       */
      unsigned int get_function(vertex node) const;

      /**
       * Returns the call graph
       * @param return the call graph;
       */
      const graph* get_graph() const;

      /**
       * Returns the a subset of the call graph
       * @param subset is the set of vertex belonging to the new sub graph
       */
      const graphRef get_subgraph(const std::set<vertex>& subset) const;
      /**
       * Replaces a call point
       * @param e is the edge in the call graph
       * @param orig is the old call graph point
       * @param repl is the new call graph point
       */
      void replace_call_point(EdgeDescriptor e, const unsigned int orig, const unsigned int repl);

      /**
       * Returns the root functions (i.e., the functions that are not called by any other ones)
       * @param functions is the set where the results will be stored
       */
      void get_root_functions(std::tr1::unordered_set<unsigned int> & functions);

};

/**
 * Information associated with a call_graph node.
 */
struct function_graph_node_info : public NodeInfo 
{
   ///this is the nodeID of the function associated with the vertex
   unsigned int nodeID;

   /**
    * Prints the information associated with the current node of the graph.
    * @param os is the output stream.
    */
   void print(std::ostream& os) const;

   /**
    * Constructor
    */
   function_graph_node_info() : nodeID(0) {}
};

/**
 * Information associated with a call_graph edge.
 */
struct function_graph_edge_info : public edge_info
{
   ///the index of the statements of the caller function where the target is called;
   std::set<unsigned int> call_points;

   /**
    * Prints the information associated with the current node of the graph.
    * @param os is the output stream.
    */
   void print(std::ostream& os) const;

   /**
    * Constructor
    */
   function_graph_edge_info() {}
};

/**
 * Helper macro returning the nodeID of the function associated with the vertex.
 * @param g is the graph.
 * @param v is the vertex.
 */
#define GET_FUNCTION_ID(g, v) \
   Cget_node_info<function_graph_node_info>(v, *(g))->nodeID

#define ADD_GLOBAL(g, index) \
   get_graph_info<function_graph_info>(*(g))->global_variables.insert(index)

#define GET_GLOBALS(g) \
   Cget_graph_info<function_graph_info>(*(g))->global_variables


/**
 * Helper macro returning the nodes where the source function calls the target
 * @param g is the graph
 * @param e is the edge caller-callee
 */
#define GET_CALL_POINTS(g, e) \
   Cget_edge_info<function_graph_edge_info>(e, *(g))->call_points


#endif
