/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file function_behavior.cpp
 * @brief A brief description of the C++ Source File
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Autoheader include
#include "config_HAVE_EXPERIMENTAL.hpp"
#include "config_HAVE_HOST_PROFILING_BUILT.hpp"
#include "config_HAVE_POLIXML_BUILT.hpp"

///Header include
#include "function_behavior.hpp"

///Behavior include
#include "application_manager.hpp"
#include "behavioral_helper.hpp"
#include "loop.hpp"
#include "loops.hpp"
#include "op_graph.hpp"
#include "op_node_info.hpp"
#if HAVE_HOST_PROFILING_BUILT
#include "profiling_information.hpp"
#endif
///Boost include
#include <boost/filesystem.hpp>

///Graph include
#include "basic_block.hpp"
#include "basic_blocks_graph_constructor.hpp"
#include "level_constructor.hpp"
#include "operations_graph_constructor.hpp"
#if HAVE_EXPERIMENTAL
#include "extended_pdg.hpp"
#include "extended_pdg_constructor.hpp"
#include "parallel_regions_graph.hpp"
#include "parallel_regions_graph_constructor.hpp"
#endif
#include "graph.hpp"
#include "Dominance.hpp"

///Tree include
#include "tree_basic_block.hpp"
#include "tree_common.hpp"
#include "tree_helper.hpp"
#include "tree_manager.hpp"
#include "tree_reindex.hpp"
#include "ext_tree_node.hpp"

///Utility include
#include "constant_strings.hpp"
#include "utility.hpp"

///XML include
#if HAVE_POLIXML_BUILT
#include "xml_helper.hpp"
#include "polixml.hpp"
#include "xml_dom_parser.hpp"
#include "xml_element.hpp"
#endif

memory_access::memory_access(unsigned int _node_id, unsigned int _base_address, unsigned int _offset) :
   node_id(_node_id),
   base_address(_base_address),
   offset(_offset)
{

}

function_behavior::function_behavior(const behavioral_helperRef _helper) :
   helper(_helper),
#if HAVE_EXPERIMENTAL
   is_cuda_compliant(false),
#endif
#if HAVE_HOST_PROFILING_BUILT
   profiling_information(ProfilingInformationRef(new ProfilingInformation())),
#endif
   dominators(0),
   post_dominators(0)
{
   og = new graphs_collection(graph_infoRef(new op_graph_info));
   bg = new bb_graphs_collection(
#if HAVE_HOST_PROFILING_BUILT
      profiling_information
#endif
   );
#if HAVE_EXPERIMENTAL
   epdg_bulk = new epdg_graphs_collection;
   rpdg = new graph(og, RPDG_SELECTOR);
   prg_bulk = new parallel_regions_graphs_collection;
#endif
   ogc = operations_graph_constructorRef(new operations_graph_constructor(og));
   bbgc = BasicBlocksGraphConstructorRef(new BasicBlocksGraphConstructor(bg));

   fb_cfg = new graph(og, FB_CFG_SELECTOR);
   cfg = new graph(og, CFG_SELECTOR);
   fcfg = new graph(og, FCFG_SELECTOR);
   cdg = new graph(og, CDG_SELECTOR);
   fcdg = new graph(og, FCDG_SELECTOR);
   dfg = new graph(og, DFG_SELECTOR);
   fdfg = new graph(og, FDFG_SELECTOR);
   adg = new graph(og, ADG_SELECTOR);
   fadfg = new graph(og, FDFG_SELECTOR | ADG_SELECTOR | FB_ADG_SELECTOR);
   fadg = new graph(og, ADG_SELECTOR | FB_ADG_SELECTOR);
   odg = new graph(og, ODG_SELECTOR);
   sdg = new graph(og, SDG_SELECTOR);
   fsdg = new graph(og, FSDG_SELECTOR);
   sadg = new graph(og, SADG_SELECTOR);
   fsadg = new graph(og, FSADG_SELECTOR);
   saodg = new graph(og, SAODG_SELECTOR);
   fsaodg = new graph(og, FSAODG_SELECTOR);
   flsadg = new graph(og, SADG_SELECTOR | FLG_SELECTOR);
   fflsadg = new graph(og, FSADG_SELECTOR | FLG_SELECTOR);
   flsaodg = new graph(og, SAODG_SELECTOR | FLG_SELECTOR);
   fflsaodg = new graph(og, FSAODG_SELECTOR | FLG_SELECTOR | FB_FG_SELECTOR);
   flaoddg = new graph(og, DFG_SELECTOR | ADG_SELECTOR | ODG_SELECTOR | FLG_SELECTOR);
   flg = new graph(og, FLG_SELECTOR);
   sg = new  graph(og, SG_SELECTOR | FLG_SELECTOR);
   hg = new graph(og, HG_SELECTOR);

   fbb = new bb_graph(bg, BB_SELECTOR | FB_BB_SELECTOR);
   bb = new  bb_graph(bg, BB_SELECTOR);
   bb_extended_cfg = new bb_graph(bg, BB_SELECTOR | FLG_SELECTOR);
   cdg_bb = new  bb_graph(bg, CDG_SELECTOR);
   dt = new  bb_graph(bg, D_BB_SELECTOR);
   pdt = new  bb_graph(bg, PD_BB_SELECTOR);
   ppg = new bb_graph(bg, PP_SELECTOR | BB_SELECTOR);
   dj = new bb_graph(bg, D_BB_SELECTOR | J_BB_SELECTOR);

#if HAVE_EXPERIMENTAL
   cepdg = new extended_pdg_graph(epdg_bulk, extended_pdg_edgeInfo::EDGE_CONTROL /*| extended_pdg_edgeInfo::EDGE_CONTROL_FLOW*/);
   depdg = new extended_pdg_graph(epdg_bulk, extended_pdg_edgeInfo::EDGE_DATA_RAW | extended_pdg_edgeInfo::EDGE_DATA_WAW_WAR);
   cdepdg = new extended_pdg_graph(epdg_bulk, extended_pdg_edgeInfo::EDGE_DATA_RAW | extended_pdg_edgeInfo::EDGE_DATA_WAW_WAR |extended_pdg_edgeInfo::EDGE_CONTROL);
   cdcfepdg = new extended_pdg_graph(epdg_bulk, extended_pdg_edgeInfo::EDGE_DATA_RAW | extended_pdg_edgeInfo::EDGE_DATA_WAW_WAR |extended_pdg_edgeInfo::EDGE_CONTROL | extended_pdg_edgeInfo::EDGE_CONTROL_FLOW);
   epdg = new extended_pdg_graph(epdg_bulk, extended_pdg_edgeInfo::EDGE_DATA_RAW | extended_pdg_edgeInfo::EDGE_DATA_WAW_WAR | extended_pdg_edgeInfo::EDGE_CONTROL | extended_pdg_edgeInfo::EDGE_OUTPUT | extended_pdg_edgeInfo::EDGE_CONTROL_FLOW);
   fepdg = new extended_pdg_graph(epdg_bulk, extended_pdg_edgeInfo::EDGE_DATA_RAW | extended_pdg_edgeInfo::EDGE_DATA_WAW_WAR | extended_pdg_edgeInfo::EDGE_CONTROL | extended_pdg_edgeInfo::EDGE_FEEDBACK | extended_pdg_edgeInfo::EDGE_OUTPUT | extended_pdg_edgeInfo::EDGE_CONTROL_FLOW);
   afg = new extended_pdg_graph(epdg_bulk, extended_pdg_edgeInfo::EDGE_AF_STANDARD);
   fafg = new extended_pdg_graph(epdg_bulk, extended_pdg_edgeInfo::EDGE_AF_STANDARD | extended_pdg_edgeInfo::EDGE_AF_FEEDBACK);
   epdgc = extended_pdg_constructorRef(new extended_pdg_constructor(epdg_bulk, saodg));
   is_cuda_compliant = true;
   prg = new parallel_regions_graph(prg_bulk, parallel_regions_graph_edgeInfo::EDGE_CONTROL | parallel_regions_graph_edgeInfo::EDGE_DATA);
   prgc = parallel_regions_graph_constructorRef(new parallel_regions_graph_constructor(prg_bulk));
#endif
   lm = level_constructorRef(new level_constructor(map_levels, deque_levels));
   bb_lm = level_constructorRef(new level_constructor(bb_map_levels, bb_deque_levels));

}

function_behavior::~function_behavior()
{
   delete cfg;
   delete fb_cfg;
   delete fcfg;
   delete cdg;
   delete fcdg;
   delete dfg;
   delete fdfg;
   delete adg;
   delete fadfg;
   delete fadg;
   delete odg;
   delete sdg;
   delete fsdg;
   delete sadg;
   delete fsadg;
   delete saodg;
   delete fsaodg;
   delete flsadg;
   delete fflsadg;
   delete flsaodg;
   delete fflsaodg;
   delete flaoddg;
   delete flg;
   delete sg;
   delete hg;

   delete fbb;
   delete bb;
   delete bb_extended_cfg;
   delete cdg_bb;
   delete dt;
   delete pdt;
   delete ppg;
   delete og;
   if (dominators)
      delete dominators;
   if (post_dominators)
      delete post_dominators;

#if HAVE_EXPERIMENTAL
   delete epdg_bulk;
   delete cepdg;
   delete depdg;
   delete cdepdg;
   delete cdcfepdg;
   delete epdg;
   delete fepdg;
   delete afg;
   delete fafg;
   delete rpdg;
   delete prg;
#endif
}


const graph* function_behavior::get_graph(function_behavior::graph_type gt) const
{
   switch ( gt )
   {
      case CFG:
         return cfg;
      case FB_CFG:
         return fb_cfg;
      case FCFG:
         return fcfg;
      case CDG:
         return cdg;
      case FCDG:
         return fcdg;
      case DFG:
         return dfg;
      case FDFG:
         return fdfg;
      case ADG:
         return adg;
      case FADFG:
         return fadfg;
      case FADG:
         return fadg;
      case ODG:
         return odg;
      case SDG:
         return sdg;
      case FSDG:
         return fsdg;
      case SADG:
         return sadg;
      case FSADG:
         return fsadg;
      case SAODG:
         return saodg;
      case FSAODG:
         return fsaodg;
      case FLSADG:
         return flsadg;
      case FFLSADG:
         return fflsadg;
      case FLSAODG:
         return flsaodg;
      case FFLSAODG:
         return fflsaodg;
      case FLAODDG:
         return flaoddg;
      case FLG:
         return flg;
      case SG:
         return sg;
      case HG:
         return hg;
#if HAVE_EXPERIMENTAL
      case RPDG:
         return rpdg;
#endif
      default:
         THROW_ERROR(std::string("Not supported graph type"));
   }
   return 0;
}

///optimization in case the subset is equal to the whole set of vertices is possible
const graphRef function_behavior::get_subgraph(function_behavior::graph_type gt, const std::set<vertex>& subset) const
{
   switch ( gt )
   {
      case FB_CFG:
         return graphRef(new graph(og,FB_CFG_SELECTOR,subset));

      case FLADG:
         return graphRef(new graph(og,DFG_SELECTOR | ADG_SELECTOR | FLG_SELECTOR,subset));

      case CFG:
         return graphRef(new graph(og,CFG_SELECTOR,subset));

      case FCFG:
         return graphRef(new graph(og,FCFG_SELECTOR,subset));

      case CDG:
         return graphRef(new graph(og,CDG_SELECTOR,subset));

      case FCDG:
         return graphRef(new graph(og,FCDG_SELECTOR,subset));

      case DFG:
         return graphRef(new graph(og,DFG_SELECTOR,subset));

      case FDFG:
         return graphRef(new graph(og,FDFG_SELECTOR,subset));

      case FADFG:
         return graphRef(new graph(og,FDFG_SELECTOR | ADG_SELECTOR | FB_ADG_SELECTOR,subset));

      case ADG:
         return graphRef(new graph(og,ADG_SELECTOR,subset));

      case FADG:
         return graphRef(new graph(og,ADG_SELECTOR | FB_ADG_SELECTOR,subset));

      case ODG:
         return graphRef(new graph(og,ODG_SELECTOR,subset));

      case SDG:
         return graphRef(new graph(og,SDG_SELECTOR,subset));

      case FSDG:
         return graphRef(new graph(og,FSDG_SELECTOR,subset));

      case SADG:
         return graphRef(new graph(og,SADG_SELECTOR,subset));

      case FSADG:
         return graphRef(new graph(og,FSADG_SELECTOR,subset));

      case SAODG:
         return graphRef(new graph(og,SAODG_SELECTOR,subset));

      case FSAODG:
         return graphRef(new graph(og,FSAODG_SELECTOR,subset));

      case FLSADG:
         return graphRef(new graph(og,SADG_SELECTOR | FLG_SELECTOR,subset));

      case FFLSADG:
         return graphRef(new graph(og,FSADG_SELECTOR | FLG_SELECTOR,subset));

      case FLSAODG:
         return graphRef(new graph(og,SAODG_SELECTOR | FLG_SELECTOR,subset));

      case FFLSAODG:
         return graphRef(new graph(og,FSAODG_SELECTOR | FLG_SELECTOR,subset));

      case FLAODDG:
         return graphRef(new graph(og, DFG_SELECTOR | ADG_SELECTOR | ODG_SELECTOR | FLG_SELECTOR, subset));

      case FLG:
         return graphRef(new graph(og, FLG_SELECTOR, subset));

      case SG:
         return graphRef(new graph(og,SG_SELECTOR,subset));

      case HG:
         return graphRef(new graph(og,HG_SELECTOR,subset));

#if HAVE_EXPERIMENTAL
      case RPDG:
         return graphRef(new graph(og, SAODG_SELECTOR & RPDG_SELECTOR, subset));
#endif

      default:
         THROW_ERROR(std::string("Not supported graph type"));
   }
   return graphRef();
}

#if HAVE_EXPERIMENTAL
const extended_pdg_graph* function_behavior::get_epdg_graph(function_behavior::graph_type gt) const
{
   switch ( gt )
   {
      case EPDG:
         return epdg;
      case FEPDG:
         return fepdg;
      case CEPDG:
         return cepdg;
      case DEPDG:
         return depdg;
      case CDEPDG:
         return cdepdg;
      case CDCFEPDG:
         return cdcfepdg;
      case AFG:
         return afg;
      case FAFG:
         return fafg;
      default:
         THROW_ERROR(std::string("Not supported graph type"));
   }
   return 0;
}

const parallel_regions_graph* function_behavior::get_prg_graph(function_behavior::graph_type gt) const
{
   switch ( gt )
   {
      case PRG:
         return this->prg;
      default:
         THROW_ERROR(std::string("Not supported graph type"));
   }
   return 0;
}
#endif

const bb_graph* function_behavior::get_bb_graph(function_behavior::bb_graph_type gt) const
{
   switch ( gt )
   {
      case BB:
         return bb;
      case FBB:
         return fbb;
      case FLCFG:
         return bb_extended_cfg;
      case CDG_BB:
         return cdg_bb;
      case DOM_TREE:
         return dt;
      case POST_DOM_TREE:
         return pdt;
      case PPG:
         return ppg;
      case DJ:
         return dj;
   }
   return 0;
}

void function_behavior::print(std::ostream& os) const
{
   os << "Function " << helper->get_function_name() << std::endl;
   //os << "Bulk operation graph of " << og;
   //os << ", , ";
}

const std::deque<vertex> & function_behavior::get_levels() const
{
   return deque_levels;
}

const std::deque<vertex> & function_behavior::get_bb_levels() const
{
   return bb_deque_levels;
}

const std::map<vertex, unsigned int> & function_behavior::get_map_levels() const
{
   return map_levels;
}


const std::map<vertex, unsigned int> & function_behavior::get_bb_map_levels() const
{
   return bb_map_levels;
}

void function_behavior::set_epp(EdgeDescriptor e, unsigned long long value)
{
   ppg->get_bb_edge_info(e)->set_epp_value(value);
}

#if HAVE_EXPERIMENTAL
void function_behavior::setCUDAFlag(bool _is_cuda_compliant)
{
   this->is_cuda_compliant = _is_cuda_compliant;
}

bool function_behavior::isCUDACompliant() const
{
   return this->is_cuda_compliant;
}
#endif

const LoopsConstRef function_behavior::GetLoops() const
{
   return loops;
}

#if HAVE_HOST_PROFILING_BUILT
const ProfilingInformationRef function_behavior::GetProfilingInformation() const
{
   return profiling_information;
}
#endif

void function_behavior::add_function_mem(unsigned int node_id)
{
   mem_nodeID.insert(node_id);
}

bool function_behavior::is_variable_mem(unsigned int node_id)
{
   return mem_nodeID.find(node_id) != mem_nodeID.end();
}

void function_behavior::add_parm_decl_copied(unsigned int node_id)
{
   parm_decl_copied.insert(node_id);
}

void function_behavior::add_parm_decl_loaded(unsigned int node_id)
{
   parm_decl_loaded.insert(node_id);
}

void function_behavior::add_parm_decl_stored(unsigned int node_id)
{
   parm_decl_stored.insert(node_id);
}

std::set<unsigned int> function_behavior::get_function_mem() const
{
   return mem_nodeID;
}

const std::set<unsigned int>& function_behavior::get_parm_decl_copied() const
{
   return parm_decl_copied;
}

const std::set<unsigned int>& function_behavior::get_parm_decl_loaded() const
{
   return parm_decl_loaded;
}

const std::set<unsigned int>& function_behavior::get_parm_decl_stored() const
{
   return parm_decl_stored;
}

std::set<unsigned int> function_behavior::get_local_variables(const application_managerConstRef AppM) const
{
   std::set<unsigned int> vars;
   //I simply have to go over all the vertices and get the used variables;
   //the variables which have to be declared are all those variables but
   //the globals ones
   VertexIterator v, vEnd;
   for (boost::tie(v, vEnd) = boost::vertices(*cfg); v != vEnd; v++)
   {
      std::tr1::unordered_set<unsigned int> varsTemp = GET_REF_VAR(cfg, *v);
      vars.insert(varsTemp.begin(), varsTemp.end());
   }
   const std::list<unsigned int>& funParams = helper->get_parameters();
   for (std::list<unsigned int>::const_iterator i = funParams.begin(); i != funParams.end(); i++)
   {
      if (vars.find(*i) != vars.end())
         vars.erase(*i);
   }

   const std::tr1::unordered_set<unsigned int>& gblVariables = AppM->get_global_variables();
   for (std::tr1::unordered_set<unsigned int>::const_iterator i = gblVariables.begin(); i != gblVariables.end(); i++)
   {
      if (vars.find(*i) != vars.end())
         vars.erase(*i);
   }
   return vars;
}

