/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file function_behavior.hpp
 * @brief A brief description of the C++ Header File
 *
 * Here goes a detailed description of the file
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef FUNCTION_BEHAVIOR_HPP
#define FUNCTION_BEHAVIOR_HPP

///Autoheader include
#include "config_HAVE_EXPERIMENTAL.hpp"
#include "config_HAVE_HOST_PROFILING_BUILT.hpp"

/// Graph include
//TODO: this should be removed when graph and graphs_collection will not be used anymore
#include "graph.hpp"
#include "Vertex.hpp"

///STL include
#include <deque>
#include <map>
#include <tr1/unordered_map>
#include <tr1/unordered_set>
#include <tr1/tuple>

///Utility include
#include "refcount.hpp"

/**
 * @name forward declarations
 */
//@{
CONSTREF_FORWARD_DECL(application_manager);
REF_FORWARD_DECL(function_behavior);
CONSTREF_FORWARD_DECL(tree_manager);
REF_FORWARD_DECL(BasicBlocksGraphConstructor);
REF_FORWARD_DECL(behavioral_helper);
REF_FORWARD_DECL(extended_pdg_constructor);
REF_FORWARD_DECL(parallel_regions_graph_constructor);
REF_FORWARD_DECL(graph);
REF_FORWARD_DECL(level_constructor);
REF_FORWARD_DECL(operations_graph_constructor);
#if HAVE_HOST_PROFILING_BUILT
REF_FORWARD_DECL(ProfilingInformation);
#endif
REF_FORWARD_DECL(tree_node);
CONSTREF_FORWARD_DECL(Loops);

class bb_graph;
class bb_graphs_collection;
template <typename Graph> class dominance;
class epdg_graphs_collection;
class extended_pdg_graph;
class parallel_regions_graph;
class parallel_regions_graphs_collection;
class sequence_info;
class xml_element;
typedef unsigned int tree_class;
//@}

///Struct representing memory information
struct memory_access
{
   unsigned int node_id;

   unsigned int base_address;

   unsigned int offset;

   memory_access(unsigned int _node_id, unsigned int _base_address, unsigned int _offset = 0);
};
typedef refcount<memory_access> memory_accessRef;

/**
 * 
*/
class function_behavior
{
   private:
      /**
       * @name some friend classes
      */
      //@{
      friend class add_loop_nop;
      friend class cdg_computation;
      friend class hpp_profiling;
      friend class instr_sequences_detection;
      friend struct loop_regions_computation;
      friend class loops_computation;
      friend class LoopsProfiling;
      friend class probability_path;
      friend class profiling;
      friend class read_profiling_data;
      friend class tp_profiling;
      friend class loops_analysis;
      //@}

      /// Behavioral helper associated with this behavioral_graph_manager
      behavioral_helperRef helper;

      /// Global graph storing CFG, DFG, FCFG, FDFG, SDG, FSDG, CDG. The nodes of this graph are operations.
      graphs_collection * og;

      /// Global graph storing CFG, dominators and post-dominators. The nodes of this graph are basic blocks.
      bb_graphs_collection * bg;

      /// The control flow graph.
      graph * cfg;

      /// The feedback edges for the control flow graph.
      graph * fb_cfg;

      /// The control flow graph with feedback.
      graph * fcfg;

      /// The control dependence graph.
      graph * cdg;

      /// The control dependence graph with feedback
      graph * fcdg;

      /// The data flow graph.
      graph * dfg;

      /// The data flow graph with feedback.
      graph * fdfg;

      /// The anti-dependencies graph.
      graph * adg;

      /// The anti-dependencies graph with feedback
      graph * fadg;

      /// The data-dependencies graph, anti-dependencies graph with feedback
      graph * fadfg;

      /// The output-dependencies flow graph.
      graph * odg;

      /// The system dependence graph.
      graph * sdg;

      /// The system dependence graph with feedback.
      graph * fsdg;

      /// The system dependence and antidependence graph.
      graph * sadg;

      /// The system dependence and antidependence graph with feedback.
      graph * fsadg;

      /// The system dependence, antidependence and output dependence graph.
      graph * saodg;

      /// The system dependence, antidependence and output dependence graph with feedback.
      graph * fsaodg;

      /// The system dependence and antidependence and flow graph with flow edges.
      graph * flsadg;

      /// The system dependence and antidependence and flow graph with feedback;
      graph * fflsadg;

      /// The system dependence, antidependence, output dependence and flow graph with flow edges.
      graph * flsaodg;

      /// The system dependence, antidependence, output dependence and flow graph with feedback;
      graph * fflsaodg;

      /// The anti-dependence + data dependence + output dependence + flow graph
      graph * flaoddg;

      /// The flow edge operation graph
      graph * flg;

      /// The speculation graph
      graph * sg;

      /// The helping graph for reg allocation in ilp
      graph * hg;

#if HAVE_EXPERIMENTAL
      ///bulk graph for extended PDG
      epdg_graphs_collection * epdg_bulk;

      ///extended PDG with only control edges
      extended_pdg_graph *cepdg;

      ///extended PDG with only data edges
      extended_pdg_graph *depdg;

      ///extended PDG with only control and data edges
      extended_pdg_graph *cdepdg;

      ///extended PDG with only control, data edges and control flow
      extended_pdg_graph *cdcfepdg;

      ///extended PDG (acyclic version)
      extended_pdg_graph *epdg;

      ///extended PDG with feedback edges
      extended_pdg_graph *fepdg;

      ///represents activation functions (acyclic version)
      extended_pdg_graph *afg;

      ///represents activation functions
      extended_pdg_graph *fafg;

      ///bulk graph for the PRG
      parallel_regions_graphs_collection * prg_bulk;

      ///parallel regions graph (Basic Block version)
      parallel_regions_graph * prg;

      /// The reduced program dependence graph
      graph * rpdg;
#endif

      /// The basic block CFG.
      bb_graph * bb;

      /// The basic block CFG with feedback
      bb_graph * fbb;

      /// The basic block Control Flow Graph + Flow Edges; flow edges impose that basic block inside a loop are executed 
      /// before what follows the loop
      bb_graph * bb_extended_cfg;

      /// The control dependence graph among basic blocks
      bb_graph * cdg_bb;

      /// The dominator tree of the CFG on basic blocks.
      bb_graph * dt;

      /// The post-dominator tree of the CFG on basic blocks.
      bb_graph * pdt;

      /// The support basic block graph for path profiling
      bb_graph * ppg;

      ///The dj graph (used for loop computation)
      bb_graph * dj;

      /// Map operation vertex to position in topological order in control flow graph; in the sorting then part vertices come before else part ones
      std::map<vertex, unsigned int> map_levels;

      /// Map basic block vertex to position in topological order in control flow graph; in the sorting then part vertices come before else part ones
      std::map<vertex, unsigned int> bb_map_levels;

      /// list of operations vertices sorted by topological order in control flow graph; in the sorting then part vertices come before else part ones
      std::deque<vertex> deque_levels;

      /// list of operations vertices sorted by topological order in control flow graph; in the sorting then part vertices come before else part ones
      std::deque<vertex> bb_deque_levels;

      std::set<unsigned int> constants;

#if HAVE_EXPERIMENTAL
      ///defines whether the function is CUDA-compliant or not
      bool is_cuda_compliant;
#endif
      /// Loops of the function
      LoopsConstRef loops;

#if HAVE_HOST_PROFILING_BUILT
      ///Profiling information about this function
      const ProfilingInformationRef profiling_information;
#endif

      ///this set represents the memory variables accessed by the function
      std::set<unsigned int> mem_nodeID;

      ///this set represents the parameters that have to be copied from the caller
      std::set<unsigned int> parm_decl_copied;

      /// this set represents the actual parameters that has to be loaded into the formal parameter from the actual parameter
      std::set<unsigned int> parm_decl_loaded;

      /// this set represents the formal parameters that has to be stored into the formal parameter from the actual parameter
      std::set<unsigned int> parm_decl_stored;

   public:

      /**
       * Constructor
       * @param _helper is the helper associated with the function
       */
      function_behavior(const behavioral_helperRef _helper);

      /**
       * Destructor
       */
      ~function_behavior();

      /**
       * Declaration of enum representing the type of graph
      */
      enum graph_type {
         CFG,           /**< Control flow graph */
         FB_CFG,        /**< Feedback edges in the Control flow graph */ 
         FCFG,          /**< Control flow graph with feedback */ 
         CDG,           /**< Control dependence graph */
         FCDG,          /**< Control dependence graph with feedback */
         DFG,           /**< Data flow graph */
         FDFG,          /**< Data flow graph with feedback */
         ADG,           /**< Anti-dependence graph */
         FADG,          /**< Anti-dependence graph with feedback */
         FADFG,         /**< Data flow graph + Anti-dependence graph + with feedback */
         ODG,           /**< Output dependence graph */
         SDG,           /**< System dependence graph */
         FSDG,          /**< System dependence graph with feedback */
         SADG,          /**< System dependence + anti-dependence graph */
         FSADG,         /**< System dependence + anti-dependence graph with feedback */
         SAODG,         /**< System dependence + anti-dependence graph + output dependence graph */
         FSAODG,        /**< System dependence + anti-dependence graph + output dependence graph with feedback */
         FLSADG,        /**< System dependence + anti-dependence + flow graph */
         FFLSADG,       /**< System dependence + anti-dependence + flow graph with feedback */
         FLSAODG,       /**< System dependence + anti-dependence + output dependence graph + flow graph */
         FFLSAODG,      /**< System dependence + anti-dependence + output dependence graph + flow graph with feedback */
         FLAODDG,       /**< Anti-dependence + data dependence + output dependence + flow graph */
         FLADG,         /**< Data dependence + anti-dependence + flow graph */
         FLG,           /**< Flow edges graph */
         SG,            /**< Speculation graph */
         HG,            /**< Helper graph */

#if HAVE_EXPERIMENTAL
         EPDG,          /**< Extended PDG */
         FEPDG,         /**< Extended PDG with feedback edges */
         CEPDG,         /**< Extended PDG with control edges */
         DEPDG,         /**< Extended PDG with data edges */
         CDEPDG,        /**< Extended PDG with control and data edges */
         CDCFEPDG,      /**< Extended PDG with control, data and control flow edges */
         AFG,           /**< Activation Function Graph */
         FAFG,          /**< Activation Function Graph with feedback */
         RPDG,           /**< Reduced Program Dependence graph */
         PRG            /**< Parallel Regions Graph */
#endif
      };

      /**
       * Declaration of enum representing the type of bb_graph
       */
      enum bb_graph_type {
         BB,            /**< Basic block control flow graph */
         FBB,           /**< Basic block control flow graph with feedback*/
         FLCFG,         /**< Basic block control flow graph + + Flow Edges; flow edges impose that basic block inside a loop are executed before what follows the loop*/
         CDG_BB,        /**< Basic block control dependence graph */
         DOM_TREE,      /**< Basic block dominator tree */
         POST_DOM_TREE, /**< Basic block post-dominator tree */
         PPG,           /**< Support basic block for path profiling */
         DJ             /**< DJ basic block graph (used for loop computation) */
      };

      /// Mutual exclusion between basic blocks (based on control flow graph with flow edges)
      std::map<vertex, std::set<vertex> > bb_reachability;

      ///reference to the operations graph constructor
      operations_graph_constructorRef ogc;

      ///reference to the basic block graph constructor
      BasicBlocksGraphConstructorRef bbgc;

#if HAVE_EXPERIMENTAL
      ///reference to the extended PDG constructor
      extended_pdg_constructorRef epdgc;

      ///reference to the basic block PRG constructor
      parallel_regions_graph_constructorRef prgc;

#endif
      ///reference to the level constructor
      level_constructorRef lm;

      ///reference to the level constructor
      level_constructorRef bb_lm;

      /// This class stores dominator information.
      dominance<bb_graph> * dominators;

      /// This class stores post-dominator information.
      dominance<bb_graph> * post_dominators;

      ///map between node id and the corresponding memory allocation
      std::map<unsigned int, memory_accessRef> memory_info;

      /**
       * Returns the helper associated with the function
      */
      const behavioral_helperRef & get_helper() const {return helper;}

      /**
       * Returns the set of local variables
       */
      std::set<unsigned int> get_local_variables(const application_managerConstRef AppM) const;

      /**
       * Return the vector of vertex index sorted in topological order.
       */
      const std::deque<vertex> & get_levels() const;

      /**
       * Return the vector of bb vertex index sorted in topological order.
       */
      const std::deque<vertex> & get_bb_levels() const;

      /**
       * Return the map of vertex index sorted in topological order.
       */
      const std::map<vertex, unsigned int> & get_map_levels() const;

      /**
       * Return the map of bb vertex index sorted in topological order.
       */
      const std::map<vertex, unsigned int> & get_bb_map_levels() const;

      /**
       * This method returns the operation graphs.
       * @param gt is the type of the graph to be returned
       * @return the pointer to the graph.
       */
      const graph* get_graph(function_behavior::graph_type gt ) const;

      /**
       * This method returns the operation graph having as vertices the vertices of subset
       * @param gt is the type of the graph to be returned
       * @param subset is the set of subgraph vertices
       * @return the refcount to the subgraph
       */
      const graphRef get_subgraph(function_behavior::graph_type gt, const std::set<vertex> &subset) const;

      /**
       * This method returns the basic block graphs.
       * @param gt is the type of the bb_graph to be returned
       * @return the pointer to the bb_graph.
       */
      const bb_graph* get_bb_graph(function_behavior::bb_graph_type gt = function_behavior::BB) const;

#if HAVE_EXPERIMENTAL
      /**
       * This method returns the extended PDG graph.
       * @param gt is the type of the graph to be returned
       * @return the pointer to the corresponding extended_pdg_graph.
       */
      const extended_pdg_graph* get_epdg_graph(function_behavior::graph_type gt) const;

      /**
       * This method returns the PRG.
       * @param gt is the type of the graph to be returned
       * @return the pointer to the corresponding parallel_regions_graph.
       */
      const parallel_regions_graph* get_prg_graph(function_behavior::graph_type gt) const;
#endif

      /**
       * Function that prints the class behavioral_manager.
       */
      void print(std::ostream& os) const;

      /**
       * Add a nodeid to the set of constant values used by this function
       * @param nodeid is the nodeid associated to the constant value
       */
      void add_constant(unsigned int nodeid) { constants.insert(nodeid); }

      /**
       * Get the constant values
       * @return the set of variables representing constant values
       */
      std::set<unsigned int> get_constants() const { return constants; }

      /**
       * Set epp associated with an edges
       * @param e is the edges
       * @param value is the value
       */
      void set_epp(EdgeDescriptor e, unsigned long long value);

      /**
       * Friend definition of the << operator.
       * @param os is the output stream
       */
      friend std::ostream& operator<<(std::ostream& os, const function_behavior & s)
      {
         s.print(os);
         return os ;
      }

      /**
       * Friend definition of the << operator. Pointer version.
       * @param os is the output stream
       */
      friend std::ostream& operator<<(std::ostream& os, const function_behaviorRef & s)
      {
         if (s)
            s->print(os);
         return os;
      }

#if HAVE_EXPERIMENTAL
      /**
       * This method is used to set the CUDA flag for the current function.
       * @param is_cuda_compliant defines the compliancy to the CUDA specifications
       */
      void setCUDAFlag(bool is_cuda_compliant);

      /**
       * This method is used to check whether the current function is CUDA-compliant or not.
       * @return <b>true</b> if and only if the function adheres to CUDA function-specifications 
       */
      bool isCUDACompliant() const;
#endif

      /**
       * Return the loops
       */
      const LoopsConstRef GetLoops() const;

#if HAVE_HOST_PROFILING_BUILT
      /**
       * Return the profiling information
       * @return the profiling information
       */
      const ProfilingInformationRef GetProfilingInformation() const;
#endif

      /**
       * Adds an identifier to the set of memory variables
       */
      void add_parm_decl_copied(unsigned int node_id);

      /**
       * add an actual parameter that has to be loaded into the formal parameter
       */
      void add_parm_decl_loaded(unsigned int node_id);

      /**
       * add a formal parameter that has to be initialized from the actual value
       */
      void add_parm_decl_stored(unsigned int node_id);

      /**
       * Adds a memory variable to the function
       */
      void add_function_mem(unsigned int node_id);

      /**
       * Checks if a variable has been allocated in memory
       */
      bool is_variable_mem(unsigned int node_id);

      /**
       * Returns the set of memory variables
       */
      std::set<unsigned int> get_function_mem() const;

      /**
       * Returns the set of parameters to be copied
       */
      const std::set<unsigned int>& get_parm_decl_copied() const;

      /**
       * Returns the set of the actual parameters that has to be loaded into the formal parameter
       */
      const std::set<unsigned int>& get_parm_decl_loaded() const;

      /**
       * Returns the set of the formal parameters that has to be stored into the formal parameter
       */
      const std::set<unsigned int>& get_parm_decl_stored() const;

};

typedef refcount<function_behavior> function_behaviorRef;
typedef refcount<const function_behavior> function_behaviorConstRef;

/**
 * The key comparison function for vertices set based on levels
 */
class op_vertex_order_by_map : std::binary_function<vertex, vertex, bool>
{
   private:
      /// Topological sorted vertices
      const std::map<vertex, unsigned int> & ref;

      /// Graph
      const graph * g;

   public:
      /**
       * Constructor
       * @param ref_ is the map with the topological sort of vertices
       * @param g_ is a graph used only for debugging purpose to print name of vertex
       */
      op_vertex_order_by_map(const std::map<vertex, unsigned int> & ref_, const graph * g_) :
         ref(ref_),
             g(g_)
             {}

      /**
       * Compare position of two vertices in topological sorted
       * @param x is the first vertex
       * @param y is the second vertex
       * @return true if x precedes y in topological sort, false otherwise
       */
      bool operator()(const vertex x, const vertex y) const
      {
         THROW_ASSERT(ref.find(x) != ref.end(), "Vertex " + GET_NAME(g,x) + " is not in topological_sort");
         THROW_ASSERT(ref.find(y) != ref.end(), "Second " + GET_NAME(g,y) + " vertex is not in topological_sort");
         return ref.find(x)->second < ref.find(y)->second;
      }
};

#endif
