/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file op_node_info.hpp
 * @brief Specialization of class node_info for operation graphs
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef OP_NODE_INFO_HPP
#define OP_NODE_INFO_HPP

///Superclass include
#include "typed_node_info.hpp"

///STD include
#include <ostream>

///STL include
#include <tr1/unordered_set>

///Utility include
#include "refcount.hpp"

/**
 * @name Node name and type definition for the cdfg node.
*/
//@{
/**
 * constant identifying a node of opaque type
 */
#define TYPE_OPAQUE 1 << 0

/**
 * constant identifying the node type of a EXTERNAL operation (a function call)
*/
#define TYPE_EXTERNAL 1 << 4

/**
 * constant identifying the node type of an IF operation. 
*/
#define TYPE_IF 1 << 5

/**
 * constant identifying the node type of a SWITCH operation. 
*/
#define TYPE_SWITCH 1 << 6

/**
 * constant string identifying the node type of an WHILE operation. 
*/
#define TYPE_WHILE 1 << 7

/**
 * constant string identifying the node type of an WHILE operation. 
 */
#define TYPE_FOR 1 << 8

/**
 * constant string identifying the node type of an DO operation. 
*/
#define TYPE_DO 1 << 9

/**
 * constant string identifying the node type of an ASSIGN operation. 
*/
#define TYPE_ASSIGN 1 << 10

/**
 * constant string identifying a type for a no operation. Only used for operations associated with empty basic blocks.
*/
#define TYPE_NOP 1 << 11

/**
 * constant string identifying an operation node of type PHI
 */
#define TYPE_PHI 1 << 12

/**
 * constant string identifying an operation node of type return expr
 */
#define TYPE_RET 1 << 13

/**
 * Constant string identifying an operation that has to be removed
*/
#define TYPE_TO_BE_REMOVED 1 << 14

/**
 * A vertex is of type TYPE_LABEL when it is a target of a goto expression.
 * Used to define the first vertex of a basic block header of a loop.
*/
#define TYPE_LABEL 1 << 15

/**
 * A vertex is of type TYPE_GOTO when it is associated with a goto expression.
 * Mainly used in loops and in computed goto expressions.
*/
#define TYPE_GOTO 1 << 16

/**
 * constant string identifying an operation node of type virtual phi-nodes
 */
#define TYPE_VPHI 1 << 17

/**
 * Constant string identifying an operation that is a variable initialization
*/
#define TYPE_INIT 1 << 18

/**
 * Constant string identifying a memory load operation
*/
#define TYPE_LOAD 1 << 19

/**
 * Constant string identifying a memory store operation
*/
#define TYPE_STORE 1 << 20

/**
 * A vertex is of type TYPE_MEMCPY when it is associated with a assignment between struct/union.
*/
#define TYPE_MEMCPY 1 << 21

/**
 * A vertex is of type TYPE_WAS_GIMPLE_PHI when it is comes from a split of phi nodes
*/
#define TYPE_WAS_GIMPLE_PHI 1 << 22

/**
 * A vertex of type FIRST_OP if it is the first operation of the application
 */
#define TYPE_FIRST_OP 1 << 23

/**
 * A vertex of type LAST_OP if it is the last operation of the application
 */
#define TYPE_LAST_OP 1 << 24

//@}

/**
 * constant string identifying the operation performed by an assignment. 
 */
#define ASSIGN "ASSIGN"
/**
 * constant string identifying the operation performed by a store in memory.
 */
#define STORE  "STORE"
/**
 * constant string identifying the operation performed by a load from memory.
 */
#define LOAD   "LOAD"


/**
 * constant string identifying the operation performed by a READ_COND. 
 */
#define READ_COND "READ_COND"

/**
 * constant string identifying the operation performed by a SWITCH_COND. 
 */
#define SWITCH_COND "SWITCH_COND"

/**
 * constant string identifying the operation performed by a GIMPLE_LABEL.
 */
#define GIMPLE_LABEL "gimple_label"

/**
 * constant string identifying the operation performed by a GIMPLE_GOTO.
 */
#define GIMPLE_GOTO "gimple_goto"

/**
 * constant string identifying a no operation. Only used for operations associated with empty basic blocks.
 */
#define NOP "NOP"

/**
 * constant string identifying the operation performed by a gimple_return.
 */
#define GIMPLE_RETURN "gimple_return"

/**
 * constant string identifying the operation performed by a gimple_phi.
 */
#define GIMPLE_PHI "gimple_phi"

/**
 * constant string identifying the operation performed when two objects are memcopied.
 */
#define MEMCPY "memcpy"

/**
 * constant string identifying the addressing operation.
 */
#define ADDR_EXPR "addr_expr"

/**
 * constant string identifying some conversion expressions
 */
#define NOP_EXPR                  "nop_expr"

/**
 * constant string identifying some conversion expressions
 */
#define CONVERT_EXPR              "convert_expr"

/**
 * Information associated with a generic operation node.
 */
struct OpNodeInfo : public TypedNodeInfo
{
   /// set of parameters
   std::list<unsigned int> parameters;

   /// set of return variables
   std::list<unsigned int> returns;

   /// set of referenced vars_written
   std::tr1::unordered_set<unsigned int> referenced_var;

   /// set of ssa variables written. This set store only the nodeID of the variables.
   std::tr1::unordered_set<unsigned int> ssa_written;

   /// set of variables that are redefined by this operation. They have to be considered for anti-dependences
   std::tr1::unordered_set<unsigned int> vdef_ops;

   /// set of ssa variables read. This set stores only the nodeID of the variables.
   std::list<unsigned int> ssa_read;

   /// It contains the variables where the operation stores some values
   std::tr1::unordered_set<unsigned int> stores;

   /// It contains the variables where the operation loads some values from
   std::tr1::unordered_set<unsigned int> loads;

   /// list of the nodeID of tree manager associated with the cdfg node. 
   /// To a cdfg node can be associated one operation and any number of assignments.
   std::list<unsigned int> NodeID;

   /// Store the index of called functions
   std::tr1::unordered_set<unsigned int> called;

   /// Store the index of the basic block which this operation vertex belongs to
   unsigned int bb_index;

   /// Store the index of the control equivalent region
   unsigned int cer;

   /**
    * Constructor
    */
   OpNodeInfo();

   /**
    * add a var to the set of ssa written variables
    * @param NodeID is the NodeID associated with the variable
    */
   void add_ssa_written(unsigned int NodeID);

   /**
    * Adding a variable to the set of redefined variables
    * @param NodeID is the NodeID associated with the variable
    */
   void add_redefined_var(unsigned int NodeID);

   /**
    * add a var to the set of ssa read variables
    * @param NodeID is the NodeID associated with the variable
    */
   void add_ssa_read(unsigned int NodeID);

   /**
    * add a var to the set of referenced var
    * @param NodeID is the NodeId associated with the variable
    */
   void add_ref_var(unsigned int NodeID);

   /**
    * add a var to the set of variables stored by the operation
    * @param NodeID is the NodeId associated with the variable
    */
   void add_store_var(unsigned int NodeID);

   /**
    * add a var to the set of variables loaded by the operation
    * @param NodeID is the NodeId associated with the variable
    */
   void add_load_var(unsigned int NodeID);

   /**
    * add a NodeID to the list of the nodeID associated with the cdfg node when not previously added.
    * @param nodeid is the nodeID added.
    */
   void add_NodeID(unsigned int nodeid);

   /**
    * add a var to the set of parameters
    * @param NodeID is the NodeId associated with the variable
    */
   void add_parameter(unsigned int NodeID);

   /**
    * add a var to the set of return variables
    * @param NodeID is the NodeId associated with the variable
    */
   void add_return(unsigned int NodeID);
};

/**
 * Helper macro returning the NodeID associated with a node.
 * @param data is the graph.
 * @param vertex_index is the index of the node.
*/
#define GET_NODEID(data, vertex_index) \
   Cget_node_info<OpNodeInfo>(vertex_index, *(data))->NodeID

/**
 * Add a ssa variable written by the operation performed by vertex_index node.
 * @param data is the graph.
 * @param vertex_index is the index of the cdfg node.
 * @param var_index is the NodeID of the variable.
 */
#define ADD_SSA_WRITTEN(data, vertex_index, var_index) \
   get_node_info<OpNodeInfo>(vertex_index, *(data))->add_ssa_written(var_index)

/**
 * Macro returning the set of ssa variables written by the operation performed by vertex_index node.
 * @param data is the graph.
 * @param vertex_index is the index of the cdfg node.
 */
#define GET_SSA_WRITTEN(data, vertex_index) \
   Cget_node_info<OpNodeInfo>(vertex_index, *(data))->ssa_written

/**
 * Add a ssa variable read by the operation performed by vertex_index node.
 * @param data is the graph.
 * @param vertex_index is the index of the cdfg node.
 * @param var_index is the NodeID of the variable.
 */
#define ADD_SSA_READ(data, vertex_index, var_index) \
   get_node_info<OpNodeInfo>(vertex_index, *(data))->add_ssa_read(var_index)

/**
 * Macro returning the set of ssa variables read by the operation performed by vertex_index node.
 * @param data is the graph.
 * @param vertex_index is the index of the cdfg node.
 */
#define GET_SSA_READ(data, vertex_index) \
   Cget_node_info<OpNodeInfo>(vertex_index, *(data))->ssa_read

#define ADD_REDEFINED_VAR(data, vertex_index, var_index) \
   get_node_info<OpNodeInfo>(vertex_index, *(data))->add_redefined_var(var_index)

#define GET_REDEFINED_VAR(data, vertex_index) \
   Cget_node_info<OpNodeInfo>(vertex_index, *(data))->vdef_ops

#define ADD_STORE_VAR(data, vertex_index, var_index) \
   get_node_info<OpNodeInfo>(vertex_index, *(data))->add_store_var(var_index)

#define GET_STORE_VAR(data, vertex_index) \
   Cget_node_info<OpNodeInfo>(vertex_index, *(data))->stores

#define ADD_LOAD_VAR(data, vertex_index, var_index) \
   get_node_info<OpNodeInfo>(vertex_index, *(data))->add_load_var(var_index)

#define GET_LOAD_VAR(data, vertex_index) \
   Cget_node_info<OpNodeInfo>(vertex_index, *(data))->loads

/**
 * Add a variable to set of referenced var
 * @param data is the graph.
 * @param vertex_index is the index of the cdfg node.
 * @param var_index is the NodeID of the variable.
 */
#define ADD_REF_VAR(data, vertex_index, var_index) \
   get_node_info<OpNodeInfo>(vertex_index, *(data))->add_ref_var(var_index)

/**
 * Get variables referenced by this node
 * @param data is the graph.
 * @param vertex_index is the index of the cdfg node.
 */
#define GET_REF_VAR(data, vertex_index) \
   Cget_node_info<OpNodeInfo>(vertex_index, *(data))->referenced_var

/**
 * Add a parameter to the function call performed by vertex_index node.
 * @param data is the graph.
 * @param vertex_index is the index of the cdfg node.
 * @param var_index is the NodeID of the variable.
 */
#define ADD_PARAMETER(data, vertex_index, var_index) \
   get_node_info<OpNodeInfo>(vertex_index, *(data))->add_parameter(var_index)

/**
 * Macro returning only the index of the original parameters of the call (not derived ones)
 * @param data is the graph.
 * @param vertex_index is the index of the cdfg node.
 */
#define GET_PARAMETERS(data, vertex_index) \
   Cget_node_info<OpNodeInfo>(vertex_index, *(data))->parameters

/**
 * Add the id of the call node to vertex_index; NODEID associated with it is about the function and
 * not about the call
 * @param data is the graph.
 * @param vertex_index is the index of the cdfg node.
 * @param nodeID is the nodeID of the call
 */
#define ADD_CALLED(data, vertex_index, function_index) \
   get_node_info<OpNodeInfo>(vertex_index, *(data))->called.insert(function_index)

/**
 * Get the id of function called by this vertex
 * @param data is the graph.
 * @param vertex_index is the index of the cdfg node.
 * @param nodeID is the nodeID of the call
 */
#define GET_CALLED(data, vertex_index) \
   Cget_node_info<OpNodeInfo>(vertex_index, *(data))->called

/**
 * Add a nodeID to the vertex_index node.
 * @param data is the graph.
 * @param vertex_index is the index of the cdfg node.
 * @param nodeID is the nodeID associated with the cdfg node. 
*/
#define ADD_NODEID(data, vertex_index, nodeID) \
   get_node_info<OpNodeInfo>(vertex_index, *(data))->add_NodeID(nodeID)

/**
 * Macro returning the list of nodeID associated with the vertex_index node.
 * @param data is the graph.
 * @param vertex_index is the index of the cdfg node.
*/
#define GET_NODEID(data, vertex_index) \
   Cget_node_info<OpNodeInfo>(vertex_index, *(data))->NodeID

/**
 * Macro returning the index of the basic block which the node belongs to
 * @param vertex_index is the index of the cdfg node.
*/
#define GET_BB_INDEX(data, vertex_index) \
   Cget_node_info<OpNodeInfo>(vertex_index, *(data))->bb_index

/**
 * Macro returning the control equivalent region of the node
 * @param var_index is the NodeID of the variable
*/
#define GET_CER(data, vertex_index) \
   Cget_node_info<OpNodeInfo>(vertex_index, *(data))->cer

#endif
