/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file operations_graph_constructor.cpp
 * @brief This class provides methods to build a operations graph.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Header include
#include "operations_graph_constructor.hpp"

///Behavior include
#include "op_edge_info.hpp"
#include "op_graph.hpp"
#include "op_node_info.hpp"

operations_graph_constructor::operations_graph_constructor(graphs_collection * _og) : og(_og)
{
}


operations_graph_constructor::~operations_graph_constructor()
{
}


vertex operations_graph_constructor::getIndex(const std::string &source)
{
   if (index_map.find(source) != index_map.end())
      return index_map.find(source)->second;
   const vertex v_og = og->AddVertex();
   index_map[source] = v_og;
   GET_NODE_INFO(og, OpNodeInfo, v_og)->vertex_name = source;
   return index_map[source];
}


vertex operations_graph_constructor::CgetIndex(const std::string &source) const
{
   THROW_ASSERT(index_map.find(source) != index_map.end(), "Index with name " + source + " doesn't exist");
   return index_map.find(source)->second;
}

EdgeDescriptor operations_graph_constructor::add_edge(const vertex source, const vertex dest, int selector)
{
   ADD_EDGE(source, dest, selector, *og);
   EdgeDescriptor o_edge;
   bool inserted;
   boost::tie(o_edge, inserted) = boost::edge(source, dest, *og);
   THROW_ASSERT(inserted, "Edge not added");
   return o_edge;
}

void operations_graph_constructor::add_ssa_written(const vertex& Ver, unsigned int Var)
{
   ADD_SSA_WRITTEN(og, Ver, Var);
}

void operations_graph_constructor::add_ssa_read(const vertex& Ver, unsigned int Var)
{
   ADD_SSA_READ(og, Ver, Var);
}

void operations_graph_constructor::add_load_var(const vertex& Ver, unsigned int Var)
{
   ADD_LOAD_VAR(og, Ver, Var);
}

void operations_graph_constructor::add_store_var(const vertex& Ver, unsigned int Var)
{
   ADD_STORE_VAR(og, Ver, Var);
}

void operations_graph_constructor::add_redefined_var(const vertex& Ver, unsigned int Var)
{
   ADD_REDEFINED_VAR(og, Ver, Var);
}

void operations_graph_constructor::add_ref_var(const vertex& Ver, unsigned int Var)
{
   ADD_REF_VAR(og, Ver, Var);
}

void operations_graph_constructor::add_edge_info(const vertex src, const vertex tgt, const int selector, unsigned int NodeID)
{
   EdgeDescriptor e;
   bool inserted;
   boost::tie(e, inserted) = boost::edge(src, tgt, *og);
   THROW_ASSERT(inserted, "Edge from " + GET_NAME(og, src) + " to " + GET_NAME(og, tgt) + " doesn't exists");
   get_edge_info<OpEdgeInfo>(e, *(og))->add_nodeID(NodeID, selector);
}

void operations_graph_constructor::set_edge_type(const vertex src, const vertex tgt, int selector)
{
   EdgeDescriptor e;
   bool inserted;
   boost::tie(e, inserted) = boost::edge(src, tgt, *og);
   THROW_ASSERT(inserted, "Edge from " + GET_NAME(og, src) + " to " + GET_NAME(og, tgt) + " doesn't exists");
   og->set_selector(e, selector);
}

void operations_graph_constructor::add_operation(const std::string &src, const std::string &operation_t, unsigned int bb_index)
{
   THROW_ASSERT(src != "", "Vertex name empty");
   THROW_ASSERT(operation_t != "", "Operation empty");
   vertex current = getIndex(src);
   GET_NODE_INFO(og, OpNodeInfo, current)->bb_index = bb_index;
   GET_NODE_INFO(og, OpNodeInfo, current)->node_operation = operation_t;
   if (src == ENTRY)
      GET_GRAPH_INFO(og, op_graph_info)->entry_vertex = current;
   if (src == EXIT)
      GET_GRAPH_INFO(og, op_graph_info)->exit_vertex = current;
}

void operations_graph_constructor::add_type(const std::string &src, unsigned int type_t)
{
   THROW_ASSERT(src != "", "Vertex name empty");
   THROW_ASSERT(type_t != 0, "Type of vertex " + src + " is zero");
   vertex src_index = getIndex(src);
   if(GET_NODE_INFO(og, OpNodeInfo, src_index)->node_type != TYPE_GENERIC)
      GET_NODE_INFO(og, OpNodeInfo, src_index)->node_type |= type_t;
   else
      GET_NODE_INFO(og, OpNodeInfo, src_index)->node_type = type_t;
}

void operations_graph_constructor::add_nodeID(const std::string &source, unsigned int nodeID)
{
   THROW_ASSERT(source != "", "Vertex name empty");
   THROW_ASSERT(nodeID > 0, "NodeID not positive");
   ADD_NODEID(og, getIndex(source), nodeID);
}

void operations_graph_constructor::add_parameter(const vertex& Ver, unsigned int Var)
{
   ADD_PARAMETER(og, Ver, Var);
}

void operations_graph_constructor::add_called_function(const std::string &source, unsigned int called_function)
{
   ADD_CALLED(og, getIndex(source), called_function);
}
