/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file operations_graph_constructor.hpp
 * @brief This class provides methods to build an operations graph.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef OPERATIONS_GRAPH_CONSTRUCTOR_HPP
#define OPERATIONS_GRAPH_CONSTRUCTOR_HPP

///Graph include
#include "graph.hpp"

///Utility include
#include "refcount.hpp"

/**
 * @name Forward declarations
 */
//@{
REF_FORWARD_DECL(operations_graph_constructor);
//@}

/**
 * class providing methods to manage an operations graph.
*/
class operations_graph_constructor
{
   private:

      ///reference to the bulk operations graph
      graphs_collection * og;

      /// Mapping beetwen id to index
      std::map<std::string, vertex> index_map;

   public:
      /**
       * Return the vertex index given the id of the vertex node.
       * @param source is the name of the vertex.
       * @return the index associatet with the source.
       */
      vertex getIndex(const std::string &source);

      /**
       * Return the vertex index given the id of the vertex node.
       * @param source is the name of the vertex.
       * @return the index associatet with the source.
       */
      vertex CgetIndex(const std::string & source) const;

      /**
       * add an edge between vertex source and vertex dest
       * @param source is the source vertex
       * @param dest is the dest vertexes
       * @param selector is the type of the edge
       */
      EdgeDescriptor add_edge(const vertex source, const vertex dest, int selector);

      /**
       * set the selector of an edge between vertex source and vertex dest
       * @param source is the source vertex
       * @param dest is the dest vertexes
       * @param selector is the type of the edge
       */
      void set_edge_type(const vertex src, const vertex tgt, int selector);

      /**
       * Add edge info to the graph.
       * @param src is an unique id representing the source node.
       * @param tgt is an unique id representing the target node.
       * @param ef is the family of the edge. See cdfg_edge_info class for details.
       * @param NodeID is the NodeID of the variable carrying the data through the edge.
       */
      void add_edge_info(const vertex src, const vertex tgt, const int selector, unsigned int NodeID);

      /**
       * Add the operation associated with a vertex.
       * @param source is the vertex name at which the operation is associated.
       * @param oper is a string representing the operation associated with source.
       * @param bb_index is the basic block index associated with the operation.
      */
      void add_operation(const std::string &source, const std::string &oper, unsigned int bb_index);

      /**
       * Add the type associated with a vertex.
       * @param source is the vertex name at which the type is associated.
       * @param type is an unsigned int representing the type associated with source.
       */
      void add_type(const std::string &source, unsigned int type);

      /**
       * Add a nodeID to a vertex.
       * @param source is the vertex name.
       * @param nodeID is the nodeID to be associated to the vertex.
       */
      void add_nodeID(const std::string &source, unsigned int nodeID);

      /**
       * Constructor.
       * @param _og is the reference to the bulk operations graph.
       */
      operations_graph_constructor(graphs_collection * _og);

      /**
       * Destructor.
       */
      ~operations_graph_constructor();

      /**
       * Adds a (ssa-)variable to the set of variables written by the operation vertex
       * @param Ver is the operation vertex
       * @param Var is the node id associated to the variable written by the operation vertex
       */
      void add_ssa_written(const vertex& Ver, unsigned int Vargc);

      /**
       * Adds a (ssa-)variable to the set of variables stored by the operation vertex
       * @param Ver is the operation vertex
       * @param Var is the node id associated to the variable stored by the operation vertex
       */
      void add_store_var(const vertex& Ver, unsigned int Vargc);

      /**
       * Adds a (ssa-)variable to the set of variables redefined by the operation vertex
       * @param Ver is the operation vertex
       * @param Var is the node id associated to the variable redefined by the operation vertex
       */
      void add_redefined_var(const vertex& Ver, unsigned int Vargc);

      /**
       * Adds a (ssa-)variable to the set of variables loaded by the operation vertex
       * @param Ver is the operation vertex
       * @param Var is the node id associated to the variable loaded by the operation vertex
       */
      void add_load_var(const vertex& Ver, unsigned int Vargc);

      /**
       * Adds a (ssa-)variable to the set of variables read by the operation vertex
       * @param Ver is the operation vertex
       * @param Var is the node id associated to the variable read by the operation vertex
       */
      void add_ssa_read(const vertex& Ver, unsigned int Vargc);

      /**
       * Adds a (ssa-)variable to the set of variables referred by the operation vertex
       * @param Ver is the operation vertex
       * @param Var is the node id associated to the variable referred by the operation vertex
       */
      void add_ref_var(const vertex& Ver, unsigned int Vargc);

      /**
       * Adds a parameter to the vertex
       * @param Ver is the operation vertex
       * @param Var is the node id associated to the variable referred by the operation vertex
       */
      void add_parameter(const vertex& Ver, unsigned int Vargc);

      /**
       * Adds a call to the vertex
       * @param source is the vertex name at which the type is associated.
       * @param called is the called function
       */
      void add_called_function(const std::string &source, unsigned int called_function);
};

#endif
