/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file BaseAnalysis.hpp
 * @brief This class contains the base representation for a generic analysis step
 *
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

#ifndef BASEANALYSIS_HPP
#define BASEANALYSIS_HPP

///Autoheader include
#include "config_HAVE_BAMBU_BUILT.hpp"
#include "config_HAVE_EXPERIMENTAL.hpp"
#include "config_HAVE_FROM_PRAGMA_BUILT.hpp"
#include "config_HAVE_RTL_BUILT.hpp"
#include "config_HAVE_ZEBU_BUILT.hpp"

// Generic includes
#include "refcount.hpp"

#include <set>
#include <ostream>

/**
 * @name Forward declarations
 */
//@{
REF_FORWARD_DECL(BaseAnalysis);
REF_FORWARD_DECL(application_manager);
REF_FORWARD_DECL(Parameter);
CONSTREF_FORWARD_DECL(Parameter);
//@}

/// Identifier to represent which functions the pre/post conditions have to operate on.
typedef enum
{
   GEN=0,             //! Generic analysis
   ID,              //! The same function
   ID_deep,         //! All the functions directly called by this one
   ID_others,       //! All the other functions
   ID_n             //! All the functions
} FrontendAnalysisRelationship;

/// Identifier of the transformations that could be executed
typedef enum
{
   ADD_BB_FLOW_EDGES=0,
   ADD_LOOP_NOP,
#if HAVE_ZEBU_BUILT
   ADD_MAPPING,
#endif
   ADD_OP_CFG_EDGES,
   ADD_OP_FLOW_EDGES,
   ANTI_DEP_COMPUTATION,
#if HAVE_ZEBU_BUILT
   ARRAY_REF_FIX,
#endif
   BASIC_BLOCKS_CFG_COMPUTATION,
   BB_FEEDBACK_EDGES_IDENTIFICATION,
   BLOCK_FIX,
   CFG_CREATION,
   CHECK_SYSTEM_TYPE,                  //! Set the system flag to variables and types
   CONTROL_DEPENDENCE_COMPUTATION,
   CREATE_TREE_MANAGER,
   CUDA_ANALYSIS,
   CUDA_DEPENDENCY_ANALYSIS,
   DEAD_CODE_ELIMINATION,
#if HAVE_BAMBU_BUILT
   DETERMINE_MEMORY_ACCESSES,
#endif
   DFG_CLEANING,
   DOM_POST_DOM_COMPUTATION,
#if HAVE_EXPERIMENTAL
   EXTENDED_PDG_COMPUTATION,
   REDUCED_PDG_COMPUTATION,
   PARALLEL_REGIONS_GRAPH_COMPUTATION,
#endif
   FUNCTION_ANALYSIS,                  //! Creation of the call graph
   FUNCTION_CALL_SERIALIZATION,
#if HAVE_ZEBU_BUILT
   HEADER_STRUCTURING,
#endif
   HPP_PROFILING,
#if HAVE_ZEBU_BUILT
   INSTRUCTION_SEQUENCES_COMPUTATION,
#endif
#if HAVE_ZEBU_BUILT
   LOOP_COMPUTATION,
#endif
   LOOP_REGIONS_COMPUTATION,
   LOOP_REGIONS_FLOW_COMPUTATION,
   LOOPS_ANALYSIS,
   LOOPS_IDENTIFICATION,
   LOOPS_PROFILING,
#if HAVE_BAMBU_BUILT
   NI_SSA_LIVENESS,
#endif
   OP_FEEDBACK_EDGES_IDENTIFICATION,
   ORDER_COMPUTATION,
#if HAVE_ZEBU_BUILT
   PARALLEL_LOOP_SWAP,
#endif
#if HAVE_FROM_PRAGMA_BUILT
   PRAGMA_ANALYSIS,
#endif
   PREDICTABILITY_ANALYSIS,
   PROBABILITY_PATH,
   REACHABILITY_COMPUTATION,
   READ_PROFILING_DATA,
#if HAVE_ZEBU_BUILT
   SHORT_CIRCUIT_STRUCTURING,
#endif
#if HAVE_BAMBU_BUILT
   SPECULATION_EDGES_COMPUTATION,
#endif
   SPLIT_PHINODES,
   SSA_DATA_FLOW_ANALYSIS,
   TP_PROFILING,
   TREE_MANAGER_FIX,
#if HAVE_BAMBU_BUILT
   UNROLLING_DEGREE,
#endif
#if HAVE_RTL_BUILT
   UPDATE_RTL_WEIGHT,
#endif
#if HAVE_ZEBU_BUILT
   UPDATE_TREE_WEIGHT,
#endif
   USE_COUNTING,
   VAR_ANALYSIS,
   VAR_DECL_FIX,
#if HAVE_BAMBU_BUILT
   VERIFICATION_OPERATION
#endif
} FrontendAnalysisType;

/**
 * Definition of hash function for FrontendAnalysisType
 */
namespace std
{
   namespace tr1
   {
      template <>
      struct hash<FrontendAnalysisType> : public unary_function<FrontendAnalysisType, size_t>
      {
         size_t operator()(FrontendAnalysisType algorithm) const
         {
            hash<int> hasher;
            return hasher(static_cast<int>(algorithm));
         }
      };
   }
}

class BaseAnalysis
{
   protected:

      ///reference to the C/C++ application manager
      application_managerRef AppM;

      /// Identifier of the analysis step that the class will execute
      const FrontendAnalysisType analysis_type;

      /// Identifier of the function where the analysis step will work on (if the value is 0, it means that this
      /// information is not useful, e.g. general analysis step)
      unsigned int funID;

      /// Set of the preconditions of this analysis step. It represents the set of transformations to be executed before the execution of this one.
      std::set< std::pair<FrontendAnalysisRelationship, FrontendAnalysisType> > PreConds;

      /// Set of the postconditions of this analysis step
      std::set< std::pair<FrontendAnalysisRelationship, FrontendAnalysisType> > PostConds;

      ///when true the frontend breaks the execution when an irriducible loop is identified.
      bool loop_irriducible_breaks;

      ///the set of parameters
      const ParameterConstRef Param;

      ///The output message level
      const int output_level;

      ///debug level
      int debug_level;

   public:
      /**
       * Constructor
       * @param _Param is the set of the parameters
       * @param _AppM is the reference to the application manager
       * @param _analysis is the type of this analysis step
       * @param _funID is the identifier of the function where the analysis step will be executed on (if the value is 0, it means that this
       *              information is not useful, e.g. general analysis step)
       */
      BaseAnalysis(const ParameterConstRef _Param, const application_managerRef _AppM, const FrontendAnalysisType _analysis, const unsigned int _funID);

      /**
       * Destructor
       */
      virtual ~BaseAnalysis();

      /**
       * Pure virtual method. It is the method to be called by each class that extends this base one to execute the analysis step.
       */
      virtual void exec() = 0;

      /**
       * Adds a new precondition to the set of preconditions of this analysis step
       * @param relationship represents the characteristics of the functions to be considered 
       * @param analysis represents the analysis to be executed (@see analysis_t)
       */
      void SetPreCond(const FrontendAnalysisRelationship relationship, const FrontendAnalysisType analysis);

      /**
       * Returns the set of preconditions associated to this analysis step
       * @return the preconditions
       */
      const std::set< std::pair<FrontendAnalysisRelationship, FrontendAnalysisType> > & GetPreConds() const;

      /**
       * Adds a new postcondition to the set of preconditions of this analysis step
       * @param relationship represents the characteristics of the functions to be considered 
       * @param analysis represents the analysis to be executed (@see analysis_t)
       */
      void SetPostCond(const FrontendAnalysisRelationship relationship, const FrontendAnalysisType analysis);

      /**
       * Returns the postconditions associated to this analysis step
       * @return the set of pairs where the first element is a pair between the identifier for the functions and the identifier of the analysis
       *         step and the second element is a condition that could be associated to the postcondition
       */
      const std::set< std::pair<FrontendAnalysisRelationship, FrontendAnalysisType> > & GetPostConds() const;

      /**
       * @name Printing methods
       */
      //@{

      /**
       * Function that prints the class BaseAnalysis.
       * @param os is the output stream.
       */
      void print(std::ostream& os) const;

      /**
       * Virtual function returning the name of the actual class.
       */
      virtual std::string get_kind_text() const = 0;

      /**
       * Returns the type of this analysis
       */
      FrontendAnalysisType get_type() const;

      /**
       * Returns the function ID of the function associated to this analysis
       */
      unsigned int get_funID() const;

      /**
       * Friend definition of the << operator.
       * @param os is the output stream
       */
      friend std::ostream& operator<<(std::ostream& os, BaseAnalysis & s)
      {
         s.print(os);
         return os ;
      }

      /**
       * Friend definition of the << operator. Pointer version.
       * @param os is the output stream
       */
      friend std::ostream& operator<<(std::ostream& os, BaseAnalysisRef & s)
      {
         if (s)
            s->print(os);
         return os;
      }
      //@}

};

/// Refcount definition for the class BaseAnalysis
typedef refcount<BaseAnalysis> BaseAnalysisRef;

#endif
