/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file FrontendFlow.cpp
 * @brief This file contains the implementation of some methods of the flow from the raw to
 *        the intermediate representation used by the PandA framework.
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Autoheader include
#include "config_HAVE_EXPERIMENTAL.hpp"
#include "config_HAVE_FROM_PRAGMA_BUILT.hpp"
#include "config_HAVE_HOST_PROFILING_BUILT.hpp"
#include "config_HAVE_TASKGRAPH_BUILT.hpp"
#include "config_HAVE_ZEBU_BUILT.hpp"

///Header include
#include "FrontendFlow.hpp"

#include "application_manager.hpp"

///Algorithm includes
#include "dom_post_dom_computation.hpp"
#include "loops_computation.hpp"

///Behavior includes
#include "basic_blocks_cfg_computation.hpp"
#include "behavioral_helper.hpp"
#include "call_graph.hpp"
#include "call_graph_computation.hpp"
#include "function_behavior.hpp"
#include "loops.hpp"
#include "operations_cfg_computation.hpp"
#include "tree_helper.hpp"

///Frontend includes
#include "Parameter.hpp"

///Graph includes
#include "graph.hpp"
#include <boost/graph/graphviz.hpp>

///Passes includes
#include "BaseAnalysis.hpp"
#include "add_bb_flow_edges.hpp"
#include "add_op_flow_edges.hpp"
#include "add_op_cfg_edges.hpp"
#include "anti_dependence_computation.hpp"
#include "bb_feedback_edges_computation.hpp"
#include "block_fix.hpp"
#include "cdg_computation.hpp"
#include "check_system_type.hpp"
#include "create_tree_manager.hpp"
#include "data_flow_cleaning.hpp"
#if HAVE_EXPERIMENTAL
#include "extended_pdg_computation.hpp"
#include "reduced_pdg_computation.hpp"
#include "parallel_regions_graph_computation.hpp"
#endif
#include "loop_regions_computation.hpp"
#include "loop_regions_flow_computation.hpp"
#include "op_feedback_edges_computation.hpp"
#include "order_computation.hpp"
#if HAVE_HOST_PROFILING_BUILT
#include "profiling.hpp"
#endif
#include "reachability_computation.hpp"
#include "ssa_data_dependence_computation.hpp"
#include "split_phi_nodes.hpp"
#include "tree_manager_fix.hpp"
#include "use_counting.hpp"
#include "var_computation.hpp"
#include "var_decl_fix.hpp"

///TaskGraphEstimator
#if HAVE_TASKGRAPH_BUILT
#include "TaskGraphEstimator.hpp"
#endif

///Tree include
#include "tree_manager.hpp"

///Utility include
#include "cpu_time.hpp"
#include "cpu_stats.hpp"
#include <iostream>

/**
 * @name function graph selector
 */
//@{
/// Precedence selector
#define STD_SELECTOR 1 << 0

#define COND_SELECTOR 1 << 1
//@}

/**
 * data structure used to store the graph describing the relation among the analysis steps
*/
struct analysis_graph : public graphs_collection
{
   ///Constructor of graph.
   analysis_graph() : graphs_collection(graph_infoRef(new graph_info)){}

   ///Destructor
   ~analysis_graph(){}

};

/**
 * Functor used by write_graphviz to write the label of the vertices of a analysis graph
*/
class analysis_graph_label_writer
{
   public:
      /**
       * constructor
       * @param g is the graph
      */
      analysis_graph_label_writer(const graph *_g, vertex _entry, vertex _exit, const application_managerRef _AppM) : g(_g), entry(_entry), exit(_exit), AppM(_AppM) {}

      /**
       * operator function returning the label of the vertex
       * @param out is the output stream
       * @param v is the vertex
      */
      void operator()(std::ostream& out, const vertex& v) const
      {
         if (v == exit)
            out << "[shape=box, label=\"EXIT\"]" ;
         else
         {
            unsigned int funID = GET_ANALYSIS(g,v)->get_funID();
            out << "[shape=box, label=\"" << GET_ANALYSIS(g,v)->get_kind_text() << ";<";
            if(funID)
               out << AppM->get_function_behavior(funID)->get_helper()->get_function_name() << "," << funID ;
            else
               out << "NO-FUNCTION";
            out << ">\"]" ;
         }
      }

   private:

      ///reference to the graph
      const graph *g;

      /// Vertices that represent entry and exit points
      vertex entry, exit;

      /// datastructure containing information about the specification (used to rescue the function name)
      const application_managerRef AppM;
};

/**
 * Functor used by write_graphviz to write the edges of a function graph
*/
class analysis_graph_edge_writer
{
   public:
      /**
       * constructor
       * @param g is the graph
      */
      analysis_graph_edge_writer(const graph * _g) : g(_g) {}

      /**
       * operator function returning the edge description
       * @param out is the output stream
       * @param e is the edge
       */
      void operator()(std::ostream& out, const EdgeDescriptor& e) const
      {
         if (STD_SELECTOR & g->get_selector(e))
            out << "[color=blue]";
         else if (COND_SELECTOR & g->get_selector(e))
               out << "[color=red]";
         else
            THROW_ERROR(std::string("InconsistentDataStructure"));
      }

   private:

      ///reference to the graph
      const graph * g;
};

/**
 * Class used to build the graph describing the relation among the analysis steps
 */
struct analysis_graph_constructor
{
   public:
      /**
       * Constructor.
      */
      analysis_graph_constructor(){}
      /**
       * Add vertex to the call graph
       * @param fg is the call graph
       * @param functionID is the function identifier
       */
      vertex add_vertex(analysis_graph* ag, BaseAnalysisRef BA)
      {
         if (BA)
            THROW_ASSERT(BA_vertex_map.find(BA) == BA_vertex_map.end(), "this vertex has been already added ");
         vertex v = ag->AddVertex();
         GET_NODE_INFO(ag, analysis_node_info, v)->BA = BA;
         BA_vertex_map[BA] = v;
         return v;

      }

      /**
       * return true in case the vertex has been already created
       * @param functionID is the function identifier
       */
      bool check_vertex(const BaseAnalysisRef& BA) const
      {
         return BA_vertex_map.find(BA) != BA_vertex_map.end();
      }

      /**
       * return a vertex of the graph given the functionID.
       * @param functionID is the function identifier
       */
      vertex get_vertex(const BaseAnalysisRef& BA) const
      {
         THROW_ASSERT(BA_vertex_map.find(BA) != BA_vertex_map.end(), "this vertex does not exist ");
         return BA_vertex_map.find(BA)->second;
      }

      /**
       * Check if the analysis of a function has been already created as vertex
       * @param analysis_type is the identifier of the analysis step
       * @param function_index the function identifier
       * @return true is the vertex (analysis of this function) has been already inserted, false otherwise
       */
      bool is_vertex(const FrontendAnalysisType analysis_type, const unsigned int function_index) const
      {
         for(std::map<BaseAnalysisRef, vertex>::const_iterator p = BA_vertex_map.begin(); p != BA_vertex_map.end(); p++)
         {
            if (!p->first) continue;
            if (p->first->get_type() == analysis_type and p->first->get_funID() == function_index)
               return true;
         }
         return false;
      }

      /**
       * Return the analysis of a function already created as vertex
       * @param analysis_type is the identifier of the analysis step
       * @param function_index is the function identifier
       * @return the vertex associated to that analysis
       */
      vertex get_vertex(const FrontendAnalysisType analysis_type, const unsigned int function_index) const
      {
         for(std::map<BaseAnalysisRef, vertex>::const_iterator p = BA_vertex_map.begin(); p != BA_vertex_map.end(); p++)
         {
            if (!p->first) continue;
            if (p->first->get_type() == analysis_type and p->first->get_funID() == function_index)
               return p->second;
         }
         THROW_ERROR("this vertex does not exist ");
         return 0;
      }

      /**
       * add an edge to the analysis graph
       * @param ag is the analysis graph
       * @param v1 is the source vertex
       * @param v2 is the target vertex
       * @param selector is the type of the edge
       */
      void add_edge(analysis_graph* ag, vertex v1, vertex v2, int selector = STD_SELECTOR) const
      {
         ADD_EDGE(v1, v2, selector, *ag);
      }

      /**
       * check if an edge is stored in the analysis graph
       * @param ag is the analysis graph
       * @param v1 is the source vertex
       * @param v2 is the target vertex
       * @return true if the edge has been found, false otherwise
       */
      bool is_edge(analysis_graph* ag, vertex v1, vertex v2) const
      {
         EdgeDescriptor e;
         bool found;
         tie(e,found) = boost::edge(v1, v2, *ag);
         return found;
      }

   private:

      /// put into relation the analysis reference with vertex
      std::map<BaseAnalysisRef, vertex> BA_vertex_map;
};


FrontendFlow::FrontendFlow(const std::tr1::unordered_map<FrontendAnalysisType, unsigned int>& transformations, const application_managerRef _AppM, const ParameterConstRef _Param) :
   Transformations(transformations),
   AppM(_AppM),
   Param(_Param),
   debug_level(_Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE))
{
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Constructor of the frontend flow");
}

FrontendFlow::~FrontendFlow ()
{
}

int FrontendFlow::write_dot(const std::string &file_name) const
{
   std::string output_directory = Param->getOption<std::string>(OPT_output_directory) + "/dot/";
   if (!boost::filesystem::exists(output_directory)) boost::filesystem::create_directories(output_directory);
   std::ofstream f((output_directory + file_name).c_str());
   boost::write_graphviz(f, *AGraph, analysis_graph_label_writer(AGraph, entry, exit, AppM), analysis_graph_edge_writer(AGraph));
   return 1;
}

void FrontendFlow::consolidate()
{
   VertexIterator v,vend;
   for(tie(v,vend) = boost::vertices(*AGraph); v != vend; v++)
   {
      if (*v == exit or *v == entry) continue;
      if (boost::out_degree(*v, *AGraph) == 0)
         AGC->add_edge(bulk_analysis_graph, *v, exit);
      if (boost::in_degree(*v, *AGraph) == 0)
         AGC->add_edge(bulk_analysis_graph, entry, *v);
   }

   if (boost::num_edges(*AGraph) == 0)
      AGC->add_edge(bulk_analysis_graph, entry, exit);

   AGC = analysis_graph_constructorRef();

#ifndef NDEBUG
   try
   {
      std::deque<vertex> vertices;
      boost::topological_sort(*AGraph, std::front_inserter(vertices));
   }
   catch (const char * msg)
   {
      write_dot("cyclic_analysis_graph.dot");
      THROW_ERROR("Analysis graph cyclic");
   }
   catch (const std::string & msg)
   {
      write_dot("cyclic_analysis_graph.dot");
      THROW_ERROR("Analysis graph cyclic");
   }
   catch (const std::exception& ex)
   {
      write_dot("cyclic_analysis_graph.dot");
      THROW_ERROR("Analysis graph cyclic");
   }
   catch ( ... )
   {
      write_dot("cyclic_analysis_graph.dot");
      THROW_ERROR("Analysis graph cyclic");
   }
#endif
}

void FrontendFlow::exec()
{
   int output_level = Param->getOption<int>(OPT_output_level);

   bulk_analysis_graph = new analysis_graph;
   AGraph = new graph(bulk_analysis_graph, STD_SELECTOR | COND_SELECTOR);
   AGC = analysis_graph_constructorRef(new analysis_graph_constructor);
   exit = AGC->add_vertex(bulk_analysis_graph, BaseAnalysisRef());
   expansion_functor = expandRef(new expand());
   if (output_level >= OUTPUT_LEVEL_VERY_PEDANTIC)
      util_print_cpu_stats(std::cerr);

   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "-->Starting execution of frontend analsysis flow");

   ///parsing of the behavioral specification
   BaseAnalysisRef base = create_analysis(CREATE_TREE_MANAGER, 0);
   INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "-->ANALYSIS-FLOW -- Starting "+ std::string(base->get_kind_text()));
   long analysis_time = 0;
   START_TIME(analysis_time);
   base->exec();
   INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "<--ANALYSIS-FLOW -- Ended: "+ print_cpu_time(STOP_TIME(analysis_time)));
   tree_managerRef TM = AppM->get_tree_manager();

   if (debug_level >= DEBUG_LEVEL_VERY_PEDANTIC)
   {
      const std::string gimple_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "frontend_flow_begin.gimple";
      std::ofstream gimple_file(gimple_file_name.c_str());
      TM->PrintGimple(gimple_file, true);
      gimple_file.close();
   }

#if HAVE_FROM_PRAGMA_BUILT
   if (Param->getOption<bool>(OPT_parse_pragma))
   {
#ifndef NDEBUG
      size_t old_indentation = indentation;
#endif
      base = create_analysis(PRAGMA_ANALYSIS, 0);
      INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "-->ANALYSIS-FLOW -- Starting "+ std::string(base->get_kind_text()));
      START_TIME(analysis_time);
      base->exec();
      INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "<--ANALYSIS-FLOW -- Ended: "+ print_cpu_time(STOP_TIME(analysis_time)));
      THROW_ASSERT(indentation == old_indentation, "Pragma analysis introduces indentation");
   }
#endif

   base = create_analysis(TREE_MANAGER_FIX, 0);
   INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "-->ANALYSIS-FLOW -- Starting "+ std::string(base->get_kind_text()));
   START_TIME(analysis_time);
   base->exec();
   INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "<--ANALYSIS-FLOW -- Ended: "+ print_cpu_time(STOP_TIME(analysis_time)));
   if (output_level >= OUTPUT_LEVEL_VERY_PEDANTIC) util_print_cpu_stats(std::cerr);

   ///call_graph computation
   base = create_analysis(FUNCTION_ANALYSIS, 0);
   INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "-->ANALYSIS-FLOW -- Starting "+ std::string(base->get_kind_text()));
   START_TIME(analysis_time);
   base->exec();
   INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "<--ANALYSIS-FLOW -- Ended: "+ print_cpu_time(STOP_TIME(analysis_time)));
   if (output_level >= OUTPUT_LEVEL_VERY_PEDANTIC) util_print_cpu_stats(std::cerr);

#if HAVE_ZEBU_BUILT
   if (Param->isOption(STR_OPT_hand_mapping))
   {
      base = create_analysis(ADD_MAPPING, 0);
      INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "-->ANALYSIS-FLOW -- Starting "+ std::string(base->get_kind_text()));
      START_TIME(analysis_time);
      base->exec();
      INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "<--ANALYSIS-FLOW -- Ended: "+ print_cpu_time(STOP_TIME(analysis_time)));
   }
#endif
   ///just after the call-graph computation, parse the entire functions list and execute the appropriate frontend analysis
   std::list<unsigned int> funIDs;
   AppM->get_functions_with_body(funIDs);

   const std::tr1::unordered_map<FrontendAnalysisType, unsigned int>::const_iterator Tr_end = Transformations.end();
   for(std::tr1::unordered_map<FrontendAnalysisType, unsigned int>::const_iterator Tr = Transformations.begin(); Tr != Tr_end; Tr++)
   {
      const std::list<unsigned int>::const_iterator fid_end = funIDs.end();
      for(std::list<unsigned int>::const_iterator fid = funIDs.begin(); fid != fid_end; fid++)
      {
         if (Tr->second)
            create_node(Tr->first, *fid);
      }
   }

   entry = AGC->get_vertex(FUNCTION_ANALYSIS, 0);

   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "FrontendFlow::exec - Before consolidate");

   consolidate();

   if (output_level >= OUTPUT_LEVEL_VERBOSE || Param->getOption<bool>(OPT_print_dot))
   {
      write_dot("analysis_graph.dot");
   }

   StartAnalysis();

   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "FrontendFlow::exec - Exit");

   if (bulk_analysis_graph) delete bulk_analysis_graph;
   if (AGraph) delete AGraph;
   expansion_functor = expandRef();

   std::tr1::unordered_set<BaseAnalysisRef> profiling_analyses;

   if (Param->isOption(OPT_profiling_method) and Param->getOption<ProfilingMethod>(OPT_profiling_method) != PM_NONE)
   {
      ProfilingMethod profiling_method = Param->getOption<ProfilingMethod>(OPT_profiling_method);
      if(profiling_method & PM_PATH_PROBABILITY)
      {
         profiling_analyses.insert(create_analysis(PROBABILITY_PATH, 0));
      }
      if(profiling_method & PM_HPP)
      {
         profiling_analyses.insert(create_analysis(HPP_PROFILING, 0));
      }
      if(profiling_method & PM_TP)
      {
         profiling_analyses.insert(create_analysis(TP_PROFILING, 0));
      }
      if(profiling_method & PM_MAX_LOOP_ITERATIONS)
      {
         profiling_analyses.insert(create_analysis(LOOPS_PROFILING, 0));
      }
      if(profiling_method & PM_XML_FILE)
      {
         profiling_analyses.insert(create_analysis(READ_PROFILING_DATA, 0));
      }
   }
   if (profiling_analyses.size())
   {
      std::tr1::unordered_set<BaseAnalysisRef>::const_iterator analysis, analysis_end = profiling_analyses.end();
      for(analysis = profiling_analyses.begin(); analysis != analysis_end; analysis++)
      {
         INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "-->POST-ANALYSIS-FLOW -- Starting "+ std::string((*analysis)->get_kind_text()));
         START_TIME(analysis_time);
         (*analysis)->exec();
         INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "<--POST-ANALYSIS-FLOW -- Ended: "+ print_cpu_time(STOP_TIME(analysis_time)));
      }
   }
#if HAVE_ZEBU_BUILT
   if(Param->isOption(OPT_sequence_length) and Param->getOption<size_t>(OPT_sequence_length) > 1)
   {
      std::list<unsigned int>::const_iterator it, it_end = funIDs.end();
      for(it = funIDs.begin(); it != it_end; it++)
      {
         const BaseAnalysisRef post_flow_analysis = create_analysis(INSTRUCTION_SEQUENCES_COMPUTATION, *it);
         INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "-->POST-ANALYSIS-FLOW -- Starting "+ std::string(post_flow_analysis->get_kind_text()));
         START_TIME(analysis_time);
         post_flow_analysis->exec();
         INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "<--POST-ANALYSIS-FLOW -- Ended: "+ print_cpu_time(STOP_TIME(analysis_time)));
      }
   }
#endif
   if (debug_level >= DEBUG_LEVEL_VERY_PEDANTIC)
   {
      const std::string gimple_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "frontend_flow_end.gimple";
      std::ofstream gimple_file(gimple_file_name.c_str());
      TM->PrintGimple(gimple_file, true);
      gimple_file.close();
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "<--Ended execution of frontend analsysis flow");
}

vertex FrontendFlow::create_node(const FrontendAnalysisType analysis_type, const unsigned int function_index)
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Creating node " + boost::lexical_cast<std::string>(analysis_type) + " for function " + boost::lexical_cast<std::string>(function_index));
   vertex v;
   BaseAnalysisRef Step;
   if (!AGC->is_vertex(analysis_type, function_index))
   {
      Step = create_analysis(analysis_type, function_index);
      v = AGC->add_vertex(bulk_analysis_graph, Step);
   }
   else
   {
      v = AGC->get_vertex(analysis_type, function_index);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Already exists");
      return v;
   }

   const std::set< std::pair<FrontendAnalysisRelationship, FrontendAnalysisType> > & PreConds = Step->GetPreConds();
   std::set< std::pair<FrontendAnalysisRelationship, FrontendAnalysisType> >::const_iterator pre_it, pre_it_end = PreConds.end();
   for (pre_it = PreConds.begin(); pre_it != pre_it_end; pre_it++)
   {
      std::set<vertex> PreVertex;
      switch(pre_it->first)
      {
         case GEN:
         {
            PreVertex.insert(create_node(pre_it->second, 0));
            break;
         }
         case ID:
         {
            PreVertex.insert(create_node(pre_it->second, Step->get_funID()));
            break;
         }
         case ID_deep:
         {
            std::set<unsigned int> Funs = AppM->get_call_graph()->get_called_by(Step->get_funID());
            std::list<unsigned int> EmptyFuns;
            AppM->get_functions_without_body(EmptyFuns);
            for (std::set<unsigned int>::iterator p = Funs.begin(); p != Funs.end(); p++)
            {
               if (Step->get_funID() == *p || std::find(EmptyFuns.begin(), EmptyFuns.end(), *p) != EmptyFuns.end()) continue;
               PreVertex.insert(create_node(pre_it->second, *p));
            }
            break;
         }
         case ID_others:
         {
            std::list<unsigned int> Funs;
            AppM->get_functions_with_body(Funs);
            for (std::list<unsigned int>::iterator p = Funs.begin(); p != Funs.end(); p++)
            {
               if (Step->get_funID() == *p) continue;
               PreVertex.insert(create_node(pre_it->second, *p));
            }
            break;
         }
         case ID_n:
         {
            std::list<unsigned int> Funs;
            AppM->get_functions_with_body(Funs);
            for (std::list<unsigned int>::iterator p = Funs.begin(); p != Funs.end(); p++)
            {
               PreVertex.insert(create_node(pre_it->second, *p));
            }
            break;
         }
         default:
            THROW_ERROR("Not supported function identifier");
      }
      for(std::set<vertex>::iterator v_it = PreVertex.begin(); v_it != PreVertex.end(); v_it++)
      {
         if (!AGC->is_edge(bulk_analysis_graph, *v_it, v))
         {
            AGC->add_edge(bulk_analysis_graph, *v_it, v);
         }
      }
   }
   const std::set<std::pair<FrontendAnalysisRelationship, FrontendAnalysisType> > & PostConds = Step->GetPostConds();
   std::set<std::pair<FrontendAnalysisRelationship, FrontendAnalysisType> >::const_iterator post_it, post_it_end = PostConds.end();
   for (post_it = PostConds.begin(); post_it != post_it_end; post_it++)
   {
      std::set<vertex> PostVertex;
      switch(post_it->first)
      {
         case GEN:
         {
            PostVertex.insert(create_node(post_it->second, 0));
            break;
         }
         case ID:
         {
            PostVertex.insert(create_node(post_it->second, Step->get_funID()));
            break;
         }
         case ID_deep:
         {
            std::set<unsigned int> Funs = AppM->get_call_graph()->get_called_by(Step->get_funID());
            for (std::set<unsigned int>::iterator p = Funs.begin(); p != Funs.end(); p++)
            {
               if (Step->get_funID() == *p)
                  continue;
               PostVertex.insert(create_node(post_it->second, *p));
            }
            break;
         }
         case ID_others:
         {
            std::list<unsigned int> Funs;
            AppM->get_functions_with_body(Funs);
            for (std::list<unsigned int>::iterator p = Funs.begin(); p != Funs.end(); p++)
            {
               if (Step->get_funID() == *p)
                  continue;
               PostVertex.insert(create_node(post_it->second, *p));
            }
            break;
         }
         case ID_n:
         {
            std::list<unsigned int> Funs;
            AppM->get_functions_with_body(Funs);
            for (std::list<unsigned int>::iterator p = Funs.begin(); p != Funs.end(); p++)
            {
               PostVertex.insert(create_node(post_it->second, *p));
            }
            break;
         }
         default:
            THROW_ERROR("Not supported function identifier");
      }
      for(std::set<vertex>::iterator successor = PostVertex.begin(); successor != PostVertex.end(); successor++)
      {
         if (!AGC->is_edge(bulk_analysis_graph, v, *successor))
            AGC->add_edge(bulk_analysis_graph, v, *successor);
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Created node " + boost::lexical_cast<std::string>(analysis_type) + " for function " + boost::lexical_cast<std::string>(function_index));
   return v;
}

void FrontendFlow::StartAnalysis()
{
   unsigned int output_level = Param->getOption<unsigned int>(OPT_output_level);

   std::map<vertex, bool> Mark;

   VertexIterator v,vend;
   for(tie(v,vend) = boost::vertices(*AGraph); v != vend; v++)
   {
      Mark[*v] = false;
   }

   std::list<vertex> WorkList;
   WorkList.push_back(entry);
   do
   {
      while(WorkList.size())
      {
         vertex Point = WorkList.front();
         WorkList.pop_front();
         if (Point != entry)
         {
            BaseAnalysisRef BA = GET_ANALYSIS(AGraph, Point);
            std::string function_name;
            if(output_level >= OUTPUT_LEVEL_VERBOSE)
            {
               if(BA->get_funID())
               {
                  function_name = " on function " + AppM->get_function_behavior(BA->get_funID())->get_helper()->get_function_name() + " - funId " + boost::lexical_cast<std::string>(BA->get_funID());
               }
               INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "-->ANALYSIS-FLOW -- Starting "+ std::string(BA->get_kind_text()) + function_name);
            }
            long analysis_time;
            if(output_level >= OUTPUT_LEVEL_VERBOSE)
            {
               START_TIME(analysis_time);
               analysis_timer.push_back(analysis_time);
            }
            BA->exec();
            if(output_level >= OUTPUT_LEVEL_VERBOSE)
            {
               analysis_time = analysis_timer.back();
               analysis_timer.pop_back();
            }
            INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "<--ANALYSIS-FLOW -- Ended: "+ print_cpu_time(STOP_TIME(analysis_time)));
            if (output_level >= OUTPUT_LEVEL_VERY_PEDANTIC) util_print_cpu_stats(std::cerr);
            /// to support loops in the analysis graph at this point I should invalidate all the following analysis
         }
         Mark[Point] = true;

         OutEdgeIterator oi,oend;
         for (tie(oi,oend) = boost::out_edges(Point, *AGraph); oi != oend; oi++)
         {
            vertex Tgt = boost::target(*oi, *AGraph);
            bool Ready = true;
            InEdgeIterator ie, iend;
            for(tie(ie, iend) = boost::in_edges(Tgt, *AGraph); ie != iend; ie++)
            {
               Ready = Ready and Mark[boost::source(*ie, *AGraph)];
            }
            if (Ready and Tgt != exit)
               WorkList.push_back(Tgt);
         }
      }
   }while(WorkList.size());

   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "************************************");
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "*      Ended Frontend Analysis     *");
   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "************************************");
}

BaseAnalysisRef FrontendFlow::create_analysis(const FrontendAnalysisType analysis, const unsigned int fun_id)
{
   switch(analysis)
   {
      case ADD_BB_FLOW_EDGES:
      {
         return BaseAnalysisRef(new add_bb_flow_edges(Param, AppM, fun_id));
      }
      case ADD_OP_CFG_EDGES:
      {
         return BaseAnalysisRef(new add_op_cfg_edges(Param, AppM, fun_id));
      }
      case ADD_OP_FLOW_EDGES:
      {
         return BaseAnalysisRef(new add_op_flow_edges(Param, AppM, fun_id));
      }
      case ANTI_DEP_COMPUTATION:
      {
         return BaseAnalysisRef(new anti_dependence_computation(Param, AppM, fun_id));
      }
      case BASIC_BLOCKS_CFG_COMPUTATION:
      {
         return BaseAnalysisRef(new BasicBlocksCfgComputation(Param, AppM, fun_id));
      }
      case BB_FEEDBACK_EDGES_IDENTIFICATION:
      {
         return BaseAnalysisRef(new bb_feedback_edges_computation(Param, AppM, fun_id));
      }
      case BLOCK_FIX:
      {
         return BaseAnalysisRef(new block_fix(Param, AppM, fun_id));
      }
      case CFG_CREATION:
      {
         return BaseAnalysisRef(new operations_cfg_computation(Param, AppM, fun_id));
      }
      case CHECK_SYSTEM_TYPE:
      {
         return BaseAnalysisRef(new check_system_type(Param, AppM, fun_id));
      }
      case CONTROL_DEPENDENCE_COMPUTATION:
      {
         return BaseAnalysisRef(new cdg_computation(Param, AppM, fun_id));
      }
      case CREATE_TREE_MANAGER:
      {
         return BaseAnalysisRef(new create_tree_manager(Param, AppM));
      }
      case DFG_CLEANING:
      {
         return BaseAnalysisRef(new data_flow_cleaning(Param, AppM, fun_id));
      }
      case DOM_POST_DOM_COMPUTATION:
      {
         return BaseAnalysisRef(new dom_post_dom_computation(Param, AppM, fun_id));
      }
#if HAVE_EXPERIMENTAL
      case EXTENDED_PDG_COMPUTATION:
      {
         return BaseAnalysisRef(new extended_pdg_computation(Param, AppM, fun_id));
      }
#endif
      case FUNCTION_ANALYSIS:
      {
         return BaseAnalysisRef(new call_graph_computation(Param, AppM, expansion_functor));
      }
      case LOOP_REGIONS_COMPUTATION:
      {
         return BaseAnalysisRef(new loop_regions_computation(Param, AppM, fun_id));
      }
      case LOOP_REGIONS_FLOW_COMPUTATION:
      {
         return BaseAnalysisRef(new loop_regions_flow_computation(Param, AppM, fun_id));
      }
      case LOOPS_IDENTIFICATION:
      {
         return BaseAnalysisRef(new loops_computation(Param, AppM, fun_id));
      }
      case OP_FEEDBACK_EDGES_IDENTIFICATION:
      {
         return BaseAnalysisRef(new op_feedback_edges_computation(Param, AppM, fun_id));
      }
      case ORDER_COMPUTATION:
      {
         return BaseAnalysisRef(new order_computation(Param, AppM, fun_id));
      }
#if HAVE_EXPERIMENTAL
      case PARALLEL_REGIONS_GRAPH_COMPUTATION:
      {
         return BaseAnalysisRef(new parallel_regions_graph_computation(Param, AppM, fun_id));
      }
#endif
      case REACHABILITY_COMPUTATION:
      {
         return BaseAnalysisRef(new reachability_computation(Param, AppM, fun_id));
      }
#if HAVE_EXPERIMENTAL
      case REDUCED_PDG_COMPUTATION:
      {
         return BaseAnalysisRef(new reduced_pdg_computation(Param, AppM, fun_id));
      }
#endif
      case SPLIT_PHINODES:
      {
         return BaseAnalysisRef(new split_phi_nodes(Param, AppM, fun_id));
      }
      case SSA_DATA_FLOW_ANALYSIS:
      {
         return BaseAnalysisRef(new ssa_data_dependence_computation(Param, AppM, fun_id));
      }
      case TREE_MANAGER_FIX:
      {
         return BaseAnalysisRef(new tree_manager_fix(Param, AppM));
      }
      case USE_COUNTING:
      {
         return BaseAnalysisRef(new use_counting(Param, AppM, fun_id));
      }
      case VAR_ANALYSIS:
      {
         return BaseAnalysisRef(new var_computation(Param, AppM, fun_id));
      }
      case VAR_DECL_FIX:
      {
         return BaseAnalysisRef(new var_decl_fix(Param, AppM, fun_id));
      }
      default:
         return create_tool_analysis(analysis, fun_id);
   }
   return BaseAnalysisRef();
}

BaseAnalysisRef FrontendFlow::create_tool_analysis(const FrontendAnalysisType, const unsigned int)
{
   THROW_ERROR("Analysis step not yet supported");
   return BaseAnalysisRef();
}
