/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file FrontendFlow.hpp
 * @brief This class defines the flow from the raw to the intermediate representation used by the PandA framework.
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

#ifndef FRONTENDFLOW_HPP
#define FRONTENDFLOW_HPP

// Generic includes
#include "refcount.hpp"
#include "dbgPrintHelper.hpp"

#include "BaseAnalysis.hpp"
#include "graph.hpp"
#include "node_info.hpp"

/**
 * @name Forward declarations
 */
//@{
/// Forward declaration
REF_FORWARD_DECL(analysis_graph_constructor);
REF_FORWARD_DECL(expand);
CONSTREF_FORWARD_DECL(Parameter);
class analysis_graph;
class graph;
//@}

#include <tr1/unordered_map>

class FrontendFlow
{
   protected:

      /// This map is used to represent the optional transformations to be executed by the analysis. The key of the map is the identifier of the
      /// transformation (@see analysis_t) and the value is an integer representing if the analysis has to be performed or not (value 0 means
      /// it has not). A integer is used to allow to specify also some parameters to the transformation.
      std::tr1::unordered_map<FrontendAnalysisType, unsigned int> Transformations;

      ///describe the relation among the analysis steps in form of graph
      analysis_graph * bulk_analysis_graph;

      /// External view of the analysis graph
      graph * AGraph;

      /// Graph aid constructor
      analysis_graph_constructorRef AGC;

      /// If represent the entry point of the analysis
      vertex entry;

      /// If represent the exit point of the analysis (when it has been reached, it means that the analysis has been completed)
      vertex exit;

      expandRef expansion_functor;

      ///timing of analysis step
      std::vector<long> analysis_timer;

      ///datastructure representing the input application
      const application_managerRef AppM;

      ///class containing all parameter information
      const ParameterConstRef Param;

      /// Debug level
      const int debug_level;

      /**
       * Creates a vertex in the analysis graph in according to analysis step and function identifier given as parameters
       * @param analysis_type is the identifier of the required analysis step
       * @param function_index the identifier of the function where the analysis has to work on (if the value is 0, it means that it
       *               is a global analysis
       * @return the related vertex into the graph. Note that if the vertex has been already created (e.g. the analysis step of a function has
       *         been already inserted), the vertex returned is the old one (a new one is not created).
       */
      vertex create_node(const FrontendAnalysisType analysis_type, const unsigned int function_index);

      /**
       * Starts the analysis in according to the analysis graph created
       * @param CG is the reference to the call graph datastructure
       * @param TM is the reference to the tree manager datastructure
       */
      void StartAnalysis();

      /**
       * Creates a tool specific analysis step in according to requirements
       * @param analysis_type is the identifier of the required analysis step
       * @param function_index the identifier of the function where the analysis has to work on (if the value is 0, it means that it
       *               is a global analysis
       * @return a reference to the implementation of the proper analysis step
      */
      virtual BaseAnalysisRef create_tool_analysis(const FrontendAnalysisType analysis_type, const unsigned int function_index);

   public:
      /**
       * Constructor
       * @param transformations is the map the represent the optional transformations to be executed or not
       * @param CG is the call graph where infromation will be stored
       * @param Param is the set of the input parameters
       */
      FrontendFlow(const std::tr1::unordered_map<FrontendAnalysisType, unsigned int>& transformations, const application_managerRef AppM, const ParameterConstRef Param);

      /**
       * Destructor
       */
      virtual ~FrontendFlow ();

      /**
       * write the analysis graph in a dot format.
       * @param file_name is the name of the file where the file is written.
       * @return 1 if no error occurs
       */
      int write_dot(const std::string &file_name) const;

      /**
       * deallocate the function graph constructor. After the call of this function is not possibile to change the analysis graph data structure.
       */
      void consolidate();

      /**
       * This method executes the frontend analysis in according with requests
       * @param TM is the reference to the manager of tree nodes
       */
      void exec();

      /**
       * allow the customization of the expansion_functor
       */
      void set_expansion_functor(expandRef & exp) {expansion_functor=exp;}

      /**
       * Creates the analysis step in according to requirements
       * @param analysis_type is the identifier of the required analysis step
       * @param function_index the identifier of the function where the analysis has to work on (if the value is 0, it means that it
       *               is a global analysis
       * @return a reference to the implementation of the proper analysis step
       */
      BaseAnalysisRef create_analysis(const FrontendAnalysisType analysis_type, const unsigned int function_index);

};

/// Refcount definition for the FrontendFlow class
typedef refcount<FrontendFlow> FrontendFlowRef;

/**
 * Information associated with a circuit graph node.
 */
struct analysis_node_info : public NodeInfo
{
   /// is the nodeID of the function associated with the vertex
   BaseAnalysisRef BA;

   /**
    * Print the information associated with the current node of the graph.
    * @param os is the output stream.
    */
   void print(std::ostream& os) const
   {
      os << BA->get_kind_text() << ";" << BA->get_funID() << std::endl;
   }

   /// Constructor
   analysis_node_info() {}
};


/**
 * Helper macro returning the nodeID of the function associated with the vertex.
 * @param g is the graph.
 * @param v is the vertex.
 */
#define GET_ANALYSIS(g, v) \
   Cget_node_info<analysis_node_info>(v, *(g))->BA


#endif
