/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file block_fix.cpp
 * @brief Analysis step which modifies the control flow graph of the tree to make it more compliant
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
 */

///header include
#include "block_fix.hpp"

#include "application_manager.hpp"

///Parameter include
#include "Parameter.hpp"

///STD include
#include <fstream>

///Tree include
#include "token_interface.hpp"
#include "tree_basic_block.hpp"
#include "tree_manager.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"

///Utility include
#include "dbgPrintHelper.hpp"

bool block_fix::tree_dumped = false;

block_fix::block_fix(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
   BaseAnalysis(_Param, _AppM, BLOCK_FIX, function_id)
{
   debug_level = _Param->get_class_debug_level("block_fix", DEBUG_LEVEL_NONE);
}

block_fix::~block_fix()
{
}

void block_fix::exec()
{
   const tree_managerRef TM = AppM->get_tree_manager();
   if(debug_level >= DEBUG_LEVEL_PEDANTIC && !tree_dumped)
   {
      tree_dumped = true;
      std::string raw_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "before_block_fix.raw";
      std::ofstream raw_file(raw_file_name.c_str());
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Tree-Manager dumped for debug purpose");
      raw_file << TM;
      raw_file.close();
   }
   tree_nodeRef temp = TM->get_tree_node_const(funID);
   function_decl * fd = GetPointer<function_decl>(temp);
   statement_list * sl = GetPointer<statement_list>(GET_NODE(fd->body));

   std::map<unsigned int, blocRef> & list_of_bloc = sl->list_of_bloc;
   std::map<unsigned int, blocRef>::iterator it3, it3_end = list_of_bloc.end();
   for(it3 = list_of_bloc.begin(); it3 != it3_end; it3++)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Examining basic block " + boost::lexical_cast<std::string>(it3->first));
      std::vector<tree_nodeRef> & list_of_stmt = it3->second->list_of_stmt;
      //Checking for switch
      if(list_of_stmt.size() && GET_NODE(*(list_of_stmt.rbegin()))->get_kind() == gimple_switch_K)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Found basic block which ends with a switch " + boost::lexical_cast<std::string>(it3->first));
         std::tr1::unordered_set<unsigned int> to_be_fixed;
         std::vector<unsigned int> & list_of_succ = it3->second->list_of_succ;
         std::vector<unsigned int>::const_iterator s, s_end = list_of_succ.end();
         for(s = list_of_succ.begin(); s != s_end; s++)
         {
            //if this check is true, *s follows a case without break or there is an empty case and *s is the basic block which closes the switch
            if(list_of_bloc[*s]->list_of_pred.size() > 1)
               to_be_fixed.insert(*s);
         }
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Found " + boost::lexical_cast<std::string>(to_be_fixed.size()) + " basic blocks to be fixed");
         std::tr1::unordered_set<unsigned int>::const_iterator t, t_end = to_be_fixed.end();
         for(t = to_be_fixed.begin(); t != t_end; t++)
         {
            //Creating new basic block
            blocRef new_bb = blocRef(new bloc());
            new_bb->number = list_of_bloc.rbegin()->first + 1;
            list_of_bloc[new_bb->number] = new_bb;
            new_bb->list_of_pred.push_back(it3->first);
            new_bb->list_of_succ.push_back(*t);
            size_t i, i_end;
            i_end = it3->second->list_of_succ.size();
            for(i = 0; i != i_end; i++)
            {
               if(it3->second->list_of_succ[i] == *t)
               {
                  it3->second->list_of_succ[i] = new_bb->number;
                  break;
               }
            }
            //if this flag true, there are more than one predecessor of *t basic block which end with a switch
            bool multiple_pred_switch = false;
            i_end = list_of_bloc[*t]->list_of_pred.size();
            for(i = 0; i != i_end; i++)
            {
               if(list_of_bloc[*t]->list_of_pred[i] == it3->first)
               {
                  list_of_bloc[*t]->list_of_pred[i] = new_bb->number;
                  std::vector<tree_nodeRef> & list_of_phi = list_of_bloc[*t]->list_of_phi;
                  std::vector<tree_nodeRef>::iterator phi, phi_end = list_of_phi.end();
                  for(phi = list_of_phi.begin(); phi != phi_end; phi++)
                  {
                     gimple_phi * current_phi = GetPointer<gimple_phi>(GET_NODE(*phi));
                     std::vector<std::pair< tree_nodeRef, unsigned int> > & list_of_def_edge = current_phi->list_of_def_edge;
                     std::vector<std::pair< tree_nodeRef, unsigned int> >::iterator definition, definition_end = list_of_def_edge.end();
                     for(definition = list_of_def_edge.begin(); definition != definition_end; definition++)
                     {
                        if(definition->second == it3->first)
                        {
                           definition->second = new_bb->number;
                        }
                     }

                  }
               }
               else
               {
                  if(list_of_bloc[list_of_bloc[*t]->list_of_pred[i]]->list_of_stmt.size())
                  {
                     if(GET_NODE(*(list_of_bloc[list_of_bloc[*t]->list_of_pred[i]]->list_of_stmt.rbegin()))->get_kind() == gimple_switch_K)
                     {
                        multiple_pred_switch = true;
                     }
                  }
               }
            }
            PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Created basic block " + boost::lexical_cast<std::string>(new_bb->number));
            //moving label expr
            THROW_ASSERT(GET_NODE(*(list_of_bloc[*t]->list_of_stmt.begin()))->get_kind() == gimple_label_K, "Block " + boost::lexical_cast<std::string>(*t) + " which follows a switch " + boost::lexical_cast<std::string>(it3->first) + " but it does not start with a label");
            if(multiple_pred_switch)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Duplicating label_decl + gimple_label");
               tree_nodeRef label_expr_node = *(list_of_bloc[*t]->list_of_stmt.begin());
               gimple_label * le = GetPointer<gimple_label>(GET_NODE(label_expr_node));
               tree_nodeRef label_decl_node = le->op;
               label_decl * ld = GetPointer<label_decl>(GET_NODE(label_decl_node));
               unsigned int new_label_decl_id = TM->new_tree_node_id();
               std::map<treeVocabularyTokenTypes::token_enum, std::string> IR_schema;
               IR_schema[TOK(TOK_TYPE)] = boost::lexical_cast<std::string>(GET_INDEX_NODE(ld->type));
               IR_schema[TOK(TOK_SCPE)] = boost::lexical_cast<std::string>(GET_INDEX_NODE(ld->scpe));
               IR_schema[TOK(TOK_SRCP)] = ld->include_name + ":" + boost::lexical_cast<std::string>(ld->line_number)+":"+boost::lexical_cast<std::string>(ld->column_number);
               IR_schema[TOK(TOK_ARTIFICIAL)] = boost::lexical_cast<std::string>(ld->artificial_flag);
               TM->create_tree_node(new_label_decl_id, TOK(TOK_LABEL_DECL), IR_schema);
               unsigned int new_label_expr_id = TM->new_tree_node_id();
               IR_schema.clear();
               IR_schema[TOK(TOK_SCPE)] = boost::lexical_cast<std::string>(GET_INDEX_NODE(ld->scpe));
               IR_schema[TOK(TOK_OP)] = boost::lexical_cast<std::string>(new_label_decl_id);
               IR_schema[TOK(TOK_SRCP)] = le->include_name + ":" + boost::lexical_cast<std::string>(le->line_number)+":"+boost::lexical_cast<std::string>(ld->column_number);
               TM->create_tree_node(new_label_expr_id, TOK(TOK_GIMPLE_LABEL), IR_schema);
               gimple_switch * se = GetPointer<gimple_switch>(GET_NODE(*(list_of_stmt.rbegin())));
               tree_vec * te = GetPointer<tree_vec>(GET_NODE(se->op1));
               std::vector<tree_nodeRef> & list_of_op = te->list_of_op;
               for(size_t ind = 0; ind < list_of_op.size(); ind++)
               {
                  case_label_expr * cl = GetPointer<case_label_expr>(GET_NODE(list_of_op[ind]));
                  if(GET_INDEX_NODE(cl->got) == GET_INDEX_NODE(le->op))
                  {
                     tree_nodeRef new_label_decl_reindex = tree_nodeRef(new tree_reindex(new_label_decl_id, TM->add(new_label_decl_id)));
                     cl->got = new_label_decl_reindex;
                     break;
                  }
               }
               new_bb->list_of_stmt.resize(1);
               new_bb->list_of_stmt[0] = tree_nodeRef(new tree_reindex(new_label_expr_id, TM->add(new_label_expr_id)));
            }
            else
            {
               tree_nodeRef label = *(list_of_bloc[*t]->list_of_stmt.begin());
               list_of_bloc[*t]->list_of_stmt.erase(list_of_bloc[*t]->list_of_stmt.begin());
               new_bb->list_of_stmt.resize(1);
               new_bb->list_of_stmt[0] = label;
            }
         }
      }
   }
   if(debug_level >= DEBUG_LEVEL_PEDANTIC)
   {
      tree_dumped = true;
      std::string raw_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "after_block_fix.raw";
      std::ofstream raw_file(raw_file_name.c_str());
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Tree-Manager dumped for debug purpose");
      raw_file << TM;
      raw_file.close();
   }
}

std::string block_fix::get_kind_text() const
{
   return "block_fix";
}

