/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file check_system_type.cpp
 * @brief analyse srcp of variables and types to detect system ones; the identified one are flagged
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Autoheader include
#include "config_HAVE_LEON3.hpp"

///Header include
#include "check_system_type.hpp"

///Behavior include
#include "behavioral_helper.hpp"
#include "application_manager.hpp"
#include "function_behavior.hpp"

///Parameter include
#include "Parameter.hpp"

///STD include
#include <fstream>

///Tree include
#include "ext_tree_node.hpp"
#include "token_interface.hpp"
#include "tree_manager.hpp"
#include "tree_basic_block.hpp"
#include "tree_reindex.hpp"

///Utility include
#include "boost/filesystem/operations.hpp"
#include "boost/filesystem/path.hpp"
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>

///Wrapper include
#include "gcc_wrapper.hpp"

#define FILENAME_NORM(name) ((boost::filesystem::path(name)).normalize().string())


bool check_system_type::tree_dumped = false;

std::vector<std::string> check_system_type::systemIncPath;

std::tr1::unordered_map<std::string, std::string> check_system_type::inclNameToPath;

std::tr1::unordered_map<std::string, std::string> check_system_type::rename_function;

std::tr1::unordered_set<std::string> check_system_type::library_system_functions;

std::tr1::unordered_set<std::string> check_system_type::library_system_includes;

#if HAVE_LEON3
std::tr1::unordered_set<std::string> check_system_type::not_supported_leon3_functions;
#endif

std::tr1::unordered_map<std::string, std::string> check_system_type::undefined_library_function_include;

check_system_type::check_system_type(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
   BaseAnalysis(_Param, _AppM, CHECK_SYSTEM_TYPE, function_id)
{
   debug_level = Param->get_class_debug_level("check_system_type", DEBUG_LEVEL_NONE);
   if(inclNameToPath.size() == 0)
      build_include_structures();
}

check_system_type::~check_system_type()
{
}

void check_system_type::exec()
{
   const tree_managerRef TM = AppM->get_tree_manager();

   if(debug_level >= DEBUG_LEVEL_PEDANTIC && !tree_dumped)
   {
      tree_dumped = true;
      std::string raw_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "before_check_system_type.raw";
      std::ofstream raw_file(raw_file_name.c_str());
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Tree-Manager dumped for debug purpose");
      raw_file << TM;
      raw_file.close();
   }

   BH = AppM->get_function_behavior(funID)->get_helper();
   tree_nodeRef & curr_tn = TM->add(funID);
   function_decl * fd = GetPointer<function_decl>(curr_tn);
   statement_list * sl = GetPointer<statement_list>(GET_NODE(fd->body));

   std::list<unsigned int> without_body_function;
   AppM->get_functions_without_body(without_body_function);
   std::list<unsigned int>::const_iterator f, f_end = without_body_function.end();
   for(f = without_body_function.begin(); f != f_end; f++)
   {
      recursive_examinate(TM->get_tree_node_const(*f), *f);
   }

   std::map<unsigned int, blocRef> &blocks = sl->list_of_bloc;
   std::map<unsigned int, blocRef>::iterator it, it_end;
   it_end = blocks.end();
   for(it = blocks.begin(); it != it_end; it++)
   {
      std::vector<tree_nodeRef> & list_of_stmt = it->second->list_of_stmt;
      std::vector<tree_nodeRef>::iterator it2, it2_end = list_of_stmt.end();
      for(it2 = list_of_stmt.begin(); it2 != it2_end; it2++)
         recursive_examinate(*it2);
   }

   if(debug_level >= DEBUG_LEVEL_PEDANTIC)
   {
      std::string raw_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "after_check_system_type.raw";
      std::ofstream raw_file(raw_file_name.c_str());
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Tree-Manager dumped for debug purpose");
      raw_file << TM;
      raw_file.close();
   }
}

void check_system_type::recursive_examinate(const tree_nodeRef &tn)
{
   THROW_ASSERT(tn->get_kind() == tree_reindex_K, "Not Passed a tree_reindex");
   recursive_examinate(GET_NODE(tn), GET_INDEX_NODE(tn));
}

void check_system_type::recursive_examinate(const tree_nodeRef & curr_tn, const unsigned int index)
{
   THROW_ASSERT(curr_tn->get_kind() != tree_reindex_K, "Passed tree_reindex instead of real node");
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Checking" + boost::lexical_cast<std::string>(index));
   switch(curr_tn->get_kind())
   {
      case call_expr_K:
      {
         const call_expr * ce = GetPointer<call_expr>(curr_tn);
         recursive_examinate(ce->fn);
         const std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            recursive_examinate(*arg);
         }
         break;
      }
      case gimple_call_K:
      {
         const gimple_call * ce = GetPointer<gimple_call>(curr_tn);
         recursive_examinate(ce->fn);
         const std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            recursive_examinate(*arg);
         }
         break;
      }
      case gimple_assign_K:
      {
         const gimple_assign * gm = GetPointer<gimple_assign>(curr_tn);
         recursive_examinate(gm->op0);
         recursive_examinate(gm->op1);
         break;
      }
      case gimple_nop_K:
      {
         break;
      }
      case var_decl_K:
      case parm_decl_K:
      case field_decl_K:
      case function_decl_K:
      case label_decl_K:
      case result_decl_K:
      case type_decl_K:
      {
         decl_node * dn = GetPointer<decl_node>(curr_tn);
         function_decl * fd = GetPointer<function_decl>(curr_tn);
         bool is_system;
         std::string include = std::tr1::get<0>(BH->get_definition(index, is_system));
         if(!dn->operating_system_flag and !dn->library_system_flag and (is_system || is_system_include(include)))
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "System variable");
            srcp * sr = GetPointer<srcp>(curr_tn);
            if(!is_system && sr && sr->include_name != "")
            {
               std::string new_include;
               getRealInclName(include, new_include);
               sr->include_name = new_include;
            }
            std::string include_name = sr->include_name;
            fd = GetPointer<function_decl>(curr_tn);
            if(fd)
            {
               std::string function_name = fd->get_function_name();
               if(library_system_functions.find(function_name) != library_system_functions.end() or
                  library_system_includes.find(include_name) != library_system_includes.end())
               {
                  dn->library_system_flag = true;
               }
               else
               {
                  dn->operating_system_flag = true;
               }
            }
            else
            {
               dn->operating_system_flag = true;
            }
         }
         else if(fd and undefined_library_function_include.count(fd->get_function_name()))
         {
            dn->library_system_flag = true;
            srcp * sr = GetPointer<srcp>(curr_tn);
            sr->include_name = undefined_library_function_include[fd->get_function_name()];
         }
         //Checking for implicit declaration
         if(curr_tn->get_kind() == function_decl_K)
         {
            fd = GetPointer<function_decl>(curr_tn);
            srcp * sr = GetPointer<srcp>(curr_tn);
            if(fd and fd->type and fd->undefined_flag and !fd->operating_system_flag and !fd->library_system_flag and sr->include_name != "<built-in>")
            {
               function_type * ft = GetPointer<function_type>(GET_NODE(fd->type));
               if(!ft->prms and ft->retn and GetPointer<integer_type>(GET_NODE(ft->retn)))
               {
                  integer_type * it = GetPointer<integer_type>(GET_NODE(ft->retn));
                  if(it->name and GetPointer<type_decl>(GET_NODE(it->name)))
                  {
                     type_decl * td = GetPointer<type_decl>(GET_NODE(it->name));
                     if(td->name and GetPointer<identifier_node>(GET_NODE(td->name)))
                     {
                        identifier_node * in = GetPointer<identifier_node>(GET_NODE(td->name));
                        if(in->strg == "int")
                           fd->include_name = "<built-in>";
                     }
                  }
               }
            }
            if(fd and fd->name and GET_NODE(fd->name)->get_kind() == identifier_node_K)
            {
               identifier_node * in = GetPointer<identifier_node>(GET_NODE(fd->name));
#if HAVE_LEON3
               if(not_supported_leon3_functions.find(in->strg) != not_supported_leon3_functions.end())
               {
                  if(Param->getOption<bool>(OPT_without_operating_system))
                  {
                     THROW_ERROR("Leon3 without operating system does not support function " + in->strg);
                  }
               }
#endif
               if(rename_function.find(in->strg) != rename_function.end())
                  in->strg = rename_function.find(in->strg)->second;
            }
         }
         recursive_examinate(dn->type);
         break;
      }
      case ssa_name_K:
      {
         const ssa_name * sn = GetPointer<ssa_name>(curr_tn);
         recursive_examinate(sn->var);
         break;
      }
      case tree_list_K:
      {
         tree_nodeRef current = curr_tn;
         while(current)
         {
            if(GetPointer<tree_list>(current)->purp) recursive_examinate(GetPointer<tree_list>(current)->purp);
            recursive_examinate(GetPointer<tree_list>(current)->valu);
            current = GetPointer<tree_list>(current)->chan;
            if(current)
               current = GET_NODE(current);
         }
         break;
      }
      case CASE_UNARY_EXPRESSION:
      {
         if(curr_tn->get_kind() == addr_expr_K)
         {
            if(already_visited_ae.find(index) != already_visited_ae.end())
            {
               break;
            }
            already_visited_ae.insert(index);
         }
         const unary_expr * ue = GetPointer<unary_expr>(curr_tn);
         recursive_examinate(ue->op);
         break;
      }
      case CASE_BINARY_EXPRESSION:
      {
         const binary_expr * be = GetPointer<binary_expr>(curr_tn);
         recursive_examinate(be->op0);
         recursive_examinate(be->op1);
         break;
      }
      case CASE_TERNARY_EXPRESSION :
      {
         const ternary_expr * te = GetPointer<ternary_expr>(curr_tn);
         recursive_examinate(te->op0);
         if(te->op1)
            recursive_examinate(te->op1);
         if(te->op2)
            recursive_examinate(te->op2);
         break;
      }
      case CASE_QUATERNARY_EXPRESSION :
      {
         const quaternary_expr * qe = GetPointer<quaternary_expr>(curr_tn);
         recursive_examinate(qe->op0);
         if(qe->op1)
            recursive_examinate(qe->op1);
         if(qe->op2)
            recursive_examinate(qe->op2);
         if(qe->op3)
            recursive_examinate(qe->op3);
         break;
      }
      case constructor_K:
      {
         const constructor * co = GetPointer<constructor>(curr_tn);
         const std::vector<std::pair< tree_nodeRef, tree_nodeRef> > & list_of_idx_valu = co->list_of_idx_valu;
         std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::const_iterator it, it_end = list_of_idx_valu.end();
         for(it = list_of_idx_valu.begin(); it != it_end; it++)
         {
            recursive_examinate(it->second);
         }
         break;
      }
      case gimple_cond_K:
      {
         const gimple_cond * gc = GetPointer<gimple_cond>(curr_tn);
         recursive_examinate(gc->op0);
         break;
      }
      case gimple_switch_K:
      {
         const gimple_switch * se = GetPointer<gimple_switch>(curr_tn);
         recursive_examinate(se->op0);
         break;
      }
      case gimple_return_K:
      {
         const gimple_return * re = GetPointer<gimple_return>(curr_tn);
         if(re->op)
            recursive_examinate(re->op);
         break;
      }
      case for_expr_K:
      {
         const for_expr * fe = GetPointer<for_expr>(curr_tn);
         recursive_examinate(fe->op0);
         recursive_examinate(fe->op1);
         recursive_examinate(fe->op2);
         break;
      }
      case while_expr_K:
      {
         const while_expr * we = GetPointer<while_expr>(curr_tn);
         recursive_examinate(we->op0);
         break;
      }
      case gimple_goto_K:
      {
         const gimple_goto * ge = GetPointer<gimple_goto>(curr_tn);
         recursive_examinate(ge->op);
         break;
      }

      case boolean_type_K:
      {
         //_Bool is C99: we replace it for MAGIC compatibility
         boolean_type * bt = GetPointer<boolean_type>(curr_tn);
         if(bt->name and GetPointer<type_decl>(GET_NODE(bt->name)))
         {
            type_decl * td = GetPointer<type_decl>(GET_NODE(bt->name));
            if(td->name and GetPointer<identifier_node>(GET_NODE(td->name)))
            {
               identifier_node * in = GetPointer<identifier_node>(GET_NODE(td->name));
               if(Param->getOption<std::string>(OPT_gcc_standard) == "c99" and in->strg == "_Bool")
               {
                  const std::string INT = "int";
                  in->strg = INT;
               }
            }
         }
         break;
      }
      case vector_type_K:
      case array_type_K:
      case complex_type_K:
      case real_type_K:
      case record_type_K:
      case union_type_K:
      case pointer_type_K:
      case enumeral_type_K:
      case integer_type_K:
      case function_type_K:
      case void_type_K:
      case reference_type_K:
      {
         if(already_visited_ae.find(index) != already_visited_ae.end())
         {
            break;
         }
         already_visited_ae.insert(index);
         type_node * ty = GetPointer<type_node>(curr_tn);
         if(ty->name)
            recursive_examinate(ty->name);
         switch(curr_tn->get_kind())
         {
            case vector_type_K:
            {
               const vector_type * vt = GetPointer<vector_type>(curr_tn);
               recursive_examinate(vt->elts);
               break;
            }
            case array_type_K:
            {
               const array_type * at = GetPointer<array_type>(curr_tn);
               recursive_examinate(at->elts);
               break;
            }
            case record_type_K:
            {
               const record_type * rt = GetPointer<record_type>(curr_tn);
               const std::vector<tree_nodeRef> & list_of_flds = rt->list_of_flds;
               std::vector<tree_nodeRef>::const_iterator it, it_end = list_of_flds.end();
               for(it = list_of_flds.begin(); it != it_end; it++)
                  recursive_examinate(*it);
               break;
            }
            case union_type_K:
            {
               const union_type * ut = GetPointer<union_type>(curr_tn);
               const std::vector<tree_nodeRef> & list_of_flds = ut->list_of_flds;
               std::vector<tree_nodeRef>::const_iterator it, it_end = list_of_flds.end();
               for(it = list_of_flds.begin(); it != it_end; it++)
                  recursive_examinate(*it);
               break;
            }
            case pointer_type_K:
            {
               const pointer_type * pt = GetPointer<pointer_type>(curr_tn);
               recursive_examinate(pt->ptd);
               break;
            }
            case function_type_K:
            {
               const function_type * ft = GetPointer<function_type>(curr_tn);
               if(ft->retn)
                  recursive_examinate(ft->retn);
               if(ft->prms)
                  recursive_examinate(ft->prms);
               break;
            }
            case reference_type_K:
            {
               const reference_type * rt = GetPointer<reference_type>(curr_tn);
               if(rt->refd)
                  recursive_examinate(rt->refd);
               break;
            }
            case boolean_type_K:
            case enumeral_type_K:
            case integer_type_K:
            case complex_type_K:
            case real_type_K:
            case void_type_K:
               break;
            default:
               THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "Not supported node: " + std::string(curr_tn->get_kind_text()));
         }
         bool is_system;
         std::string include = std::tr1::get<0>(BH->get_definition(index, is_system));
         if(!ty->system_flag && (is_system || is_system_include(include)))
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "System type");
            ty->system_flag = true;
            srcp * sr = GetPointer<srcp>(curr_tn);
            if(!is_system && sr && sr->include_name   != "")
            {
               std::string new_include;
               getRealInclName(include, new_include);
               sr->include_name = new_include;
            }
         }
         break;
      }
      case target_mem_ref_K:
      {
         const target_mem_ref * tmr = GetPointer<target_mem_ref>(curr_tn);
         if(tmr->symbol)
            recursive_examinate(tmr->symbol);
         if(tmr->base)
            recursive_examinate(tmr->base);
         if(tmr->idx)
            recursive_examinate(tmr->idx);
         break;
      }
      case target_mem_ref461_K:
      {
         const target_mem_ref461 * tmr = GetPointer<target_mem_ref461>(curr_tn);
         if(tmr->base)
            recursive_examinate(tmr->base);
         if(tmr->idx)
            recursive_examinate(tmr->idx);
         if(tmr->idx2)
            recursive_examinate(tmr->idx2);
         break;
      }
      case real_cst_K:
      case complex_cst_K:
      case string_cst_K:
      case integer_cst_K:
      case vector_cst_K:
      case tree_vec_K:
      case case_label_expr_K:
      case gimple_label_K:
      case gimple_asm_K:
      case pragma_node_K:
      case identifier_node_K:
         break;
      default:
         THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "Not supported node: " + std::string(curr_tn->get_kind_text()));
   }
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Checked" + boost::lexical_cast<std::string>(index));
   return;
}

bool check_system_type::is_system_include(std::string include) const
{
   if(include == "<built-in>")
      return true;
   std::string include_p(FILENAME_NORM(include));
   for (unsigned int i = 0; i < systemIncPath.size(); i++)
   {
      if (include_p.compare(0, systemIncPath[i].size() + 1, systemIncPath[i] + "/") == 0)
         return true;
   }
   return false;
}


void check_system_type::build_include_structures()
{
   //Ok, now I put here the includes which must be modified
   inclNameToPath[FILENAME_NORM("bits/ipc.h")] = FILENAME_NORM("sys/ipc.h");
   inclNameToPath[FILENAME_NORM("bits/sem.h")] = FILENAME_NORM("sys/sem.h");
   inclNameToPath[FILENAME_NORM("bits/mathcalls.h")] = FILENAME_NORM("math.h");
   inclNameToPath[FILENAME_NORM("bits/types.h")] = FILENAME_NORM("sys/types.h");
   inclNameToPath[FILENAME_NORM("bits/stat.h")] = FILENAME_NORM("sys/stat.h");
   inclNameToPath[FILENAME_NORM("bits/mman.h")] = FILENAME_NORM("sys/mman.h");
   inclNameToPath[FILENAME_NORM("bits/in.h")] = FILENAME_NORM("netinet/in.h");
   inclNameToPath[FILENAME_NORM("bits/errno.h")] = FILENAME_NORM("errno.h");
   inclNameToPath[FILENAME_NORM("bits/fcntl.h")] = FILENAME_NORM("fcntl.h");
   inclNameToPath[FILENAME_NORM("bits/link.h")] = FILENAME_NORM("link.h");
   inclNameToPath[FILENAME_NORM("bits/shm.h")] = FILENAME_NORM("sys/shm.h");
   inclNameToPath[FILENAME_NORM("bits/stdio.h")] = FILENAME_NORM("stdio.h");
   inclNameToPath[FILENAME_NORM("bits/resource.h")] = FILENAME_NORM("sys/resource.h");
   inclNameToPath[FILENAME_NORM("bits/sigthread.h")] = FILENAME_NORM("pthread.h");
   inclNameToPath[FILENAME_NORM("bits/string2.h")] = FILENAME_NORM("string.h");
   inclNameToPath[FILENAME_NORM("bits/time.h")] = FILENAME_NORM("sys/time.h");
   inclNameToPath[FILENAME_NORM("bits/pthreadtypes.h")] = FILENAME_NORM("pthread.h");
   inclNameToPath[FILENAME_NORM("bits/sched.h")] = FILENAME_NORM("sched.h");
   inclNameToPath[FILENAME_NORM("bits/stdio2.h")] = FILENAME_NORM("stdio.h");
   inclNameToPath[FILENAME_NORM("libio.h")] = FILENAME_NORM("stdio.h");

   std::vector<std::string> Splitted;
   const GccWrapperConstRef gcc_wrapper(new GccWrapper(Param, CT_NO_GCC, OS_O0));
   gcc_wrapper->GetSystemIncludes(Splitted);

   for (std::vector<std::string>::iterator tok_iter = Splitted.begin(); tok_iter != Splitted.end(); ++tok_iter)
   {
      if (*tok_iter != "")
      {
         std::string temp;
         if (getenv("MINGW_INST_DIR"))
         {
            std::string mingw_prefix = getenv("MINGW_INST_DIR");
            temp = *tok_iter;
            if (temp.find("z:/mingw") == 0)
               temp = temp.replace(0, 8, FILENAME_NORM(mingw_prefix)); ///replace z:/mingw at the beginning of the string
            temp = FILENAME_NORM(temp);
            systemIncPath.push_back(temp);
         }
         else
         {
            temp = FILENAME_NORM(*tok_iter);
            systemIncPath.push_back(temp);
         }
      }
   }
   systemIncPath.push_back("/usr/local/share/hframework/include");

   ///Building the rename function map
   rename_function["_IO_putc"] = "putc";
   rename_function["_IO_getc"] = "getc";

   ///Building library system function
   library_system_includes.insert("math.h");

#if HAVE_LEON3
   ///Building not supported function
   not_supported_leon3_functions.insert("fopen");
#endif

   undefined_library_function_include["atof"] = "stdlib.h";
   undefined_library_function_include["atoi"] = "stdlib.h";
   undefined_library_function_include["srand48"] = "stdlib.h";
   undefined_library_function_include["__builtin_va_start"] = "stdarg.h";
   undefined_library_function_include["__builtin_va_end"] = "stdarg.h";
}

void check_system_type::getRealInclName(const std::string include, std::string & real_name) const
{
   //Now I have to see if one of the elements in systemIncPath is the start of the include:
   //in case I eliminate it and look the remaining part of the string in the map
   std::string include_p(FILENAME_NORM(include));
   for (unsigned int i = 0; i < systemIncPath.size(); i++)
   {
      if (include_p.compare(0, systemIncPath[i].size() + 1, systemIncPath[i] + "/") == 0)
      {
         std::string trimmed = include_p.substr(systemIncPath[i].size() + 1);
         if (inclNameToPath.find(trimmed) != inclNameToPath.end())
            real_name = inclNameToPath.find(trimmed)->second;
         else
            real_name = trimmed;
         return;
      }
   }

   //If, finally, the include is not a system one I simply print it back as it is
   real_name = include;
}

std::string check_system_type::get_kind_text() const 
{
   return "check_system_type";
}

