/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file check_system_type.hpp
 * @brief analyse srcp of variables and types to detect system ones; the identified one are flagged
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef CHECK_SYSTEM_TYPE_HPP
#define CHECK_SYSTEM_TYPE_HPP

///Autoheader include
#include "config_HAVE_LEON3.hpp"

///Superclass include
#include "BaseAnalysis.hpp"

///STD include
#include <tr1/unordered_map>
#include <tr1/unordered_set>
#include <vector>

///Utility include
#include "refcount.hpp"
#include "utility.hpp"

REF_FORWARD_DECL(behavioral_helper);
REF_FORWARD_DECL(tree_node);

/**
 * Class which system_flag to tree_node of variables and types when necessary
 */
class check_system_type : public BaseAnalysis
{
   private:
      /// flag to check if initial tree has been dumped
      static bool tree_dumped;

      ///The helper associated with the current function
      behavioral_helperRef BH;

      ///Already visited address expression (used to avoid infite recursion)
      std::tr1::unordered_set<unsigned int> already_visited_ae;

      ///Contains the list of the folders containining the system header files
      static
      std::vector<std::string> systemIncPath;

      ///Associates to each system header file its full path
      static
      std::tr1::unordered_map<std::string, std::string> inclNameToPath;

      ///Associates a function to its original name
      static
      std::tr1::unordered_map<std::string, std::string> rename_function;

      ///The set of functions which have to be considered library_system
      static
      std::tr1::unordered_set<std::string> library_system_functions;

       ///The set of headers which contains function which have to be considered library_system
      static
      std::tr1::unordered_set<std::string> library_system_includes;

#if HAVE_LEON3
      ///The set of system function not supported by bcc
      static
      std::tr1::unordered_set<std::string> not_supported_leon3_functions;
#endif

      ///Map undefined library function to corresponding header
      static
      std::tr1::unordered_map<std::string, std::string> undefined_library_function_include;

      /**
       * Examinate recursively the tree to detect system types and system variables
       * @param tn is the root of the subtree to be examinated; it must be a tree_reindex
       */
      void recursive_examinate(const tree_nodeRef &tn);

      /**
       * Examinate recursively the tree to detect system types and system variables
       * @param curr_tn is the root of the subtree to be examinated; it must not be a tree_reindex
       * @param index is the index of the tree_node
       */
      void recursive_examinate(const tree_nodeRef & curr_tn, const unsigned int index);

      /**
       * Check if an header is a system header
       * @param include is the header file
       */
      bool is_system_include(std::string include) const;

      /**
       * Build the include map, the function rename map and library system sets
       */
      void build_include_structures();

      /**
       * Given the string stored in a srcp of the raw return the correct include name 
       * @param include is the string of srcp to be checked
       * @param real_name is the string fixed
       */
      void getRealInclName(const std::string include, std::string & real_name) const;

   public:

      /**
       * Constructor
       * @param Param is the set of the parameters
       * @param function_id is the index of the function
       */
      check_system_type(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id);

      /**
       * Destructor
       */
      virtual ~check_system_type();

      /**
       * Adds the system_flag to the tree_node's of variables and types when necessary
       */
      void exec();

      /**
       * Returns the name of the analysis step
       */
      std::string get_kind_text() const;
};

#endif
