/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file data_flow_cleaninig.cpp
 * @brief Analysis step removing fake data dependences
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/

///Header include
#include "data_flow_cleaning.hpp"

///Behavior include
#include "behavioral_helper.hpp"
#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "op_node_info.hpp"

///Graph include
#include "graph.hpp"
#include "basic_block.hpp"
#include "operations_graph_constructor.hpp"

///Tree include
#include "tree_node.hpp"
#include "tree_basic_block.hpp"
#include "tree_reindex.hpp"
#include "tree_manager.hpp"

///Frontend include
#include "Parameter.hpp"


data_flow_cleaning::data_flow_cleaning(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
   BaseAnalysis(_Param, _AppM, DFG_CLEANING, function_id)
{
   this->SetPreCond(ID, SSA_DATA_FLOW_ANALYSIS);
   this->SetPreCond(ID, REACHABILITY_COMPUTATION);
}


data_flow_cleaning::~data_flow_cleaning()
{

}

void data_flow_cleaning::exec()
{
   TM = AppM->get_tree_manager();
   function_behaviorRef FB = AppM->get_function_behavior(funID);
   const graph * dfg = FB->get_graph(function_behavior::DFG);
   const graph * adg = FB->get_graph(function_behavior::ADG);
   ///The control flow graph of basic block
   const bb_graph * bb_cfg = FB->get_bb_graph(function_behavior::BB);
   helper = FB->get_helper();
   operations_graph_constructorRef & ogc = FB->ogc;

   //set containing pointers which are deferenced directly or by constant index
   std::set<unsigned int> candidate_pointers;

   //set containing pointers which are deferenced by not constant index
   std::set<unsigned int> not_candidate_pointers;

   //set containing pointers which are only deferenced directly or by constant index
   std::set<unsigned int> pointers;

   VertexIterator vi, vi_end;
   for (boost::tie(vi, vi_end) = boost::vertices(*dfg); vi != vi_end; vi++)
   {
      std::list<unsigned int> nodeId = GET_NODEID(dfg, *vi);
      if(nodeId.size())
         get_candidate_pointers(nodeId.back(), candidate_pointers, not_candidate_pointers);
   }

   std::set_difference(
      candidate_pointers.begin(),
      candidate_pointers.end(),
      not_candidate_pointers.begin(),
      not_candidate_pointers.end(),
      std::inserter(pointers, pointers.begin())
   );
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Candidate pointers are " + boost::lexical_cast<std::string>(pointers.size()));
   std::set<unsigned int>::const_iterator it2_end = pointers.end();
   for(std::set<unsigned int>::const_iterator it2 = pointers.begin(); it2 != it2_end; it2++)
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  " + boost::lexical_cast<std::string>(*it2));

   std::map<unsigned int, std::pair<unsigned int, long long int> >  pointer_to_base_and_offset;
   std::map<unsigned int, std::set<std::pair<vertex, unsigned int> > > ptr_info_to_vertex_and_pointer;
   std::map<unsigned int, unsigned int> ptr_info_to_base;

   for (boost::tie(vi, vi_end) = boost::vertices(*dfg); vi != vi_end; vi++)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Examining vertex " + GET_NAME(dfg, *vi));
      std::list<unsigned int> var_read = GET_SSA_READ(dfg, *vi);
      std::list<unsigned int>::const_iterator it_end = var_read.end();
      for(std::list<unsigned int>::const_iterator it = var_read.begin(); it != it_end; it++)
      {
         if(helper->is_a_pointer(*it))
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Examining pointer read " + boost::lexical_cast<std::string>(*it));
            tree_nodeRef tn = TM->get_tree_node_const(*it);
            if(tn->get_kind() != ssa_name_K)
               continue;
            ssa_name * sn = GetPointer<ssa_name>(tn);
            unsigned int ptr_info;
            if(sn->ptr_info)
               ptr_info = GET_INDEX_NODE(sn->ptr_info);
            else
               continue;
            tree_nodeRef pointer = GET_NODE(sn->var);
            tree_nodeRef base = GET_NODE(sn->var);
            unsigned int base_id = GET_INDEX_NODE(sn->var);
            unsigned int pointer_id = *it;
            long long int offset = 0;
            if(GetPointer<decl_node>(pointer))
            {
               decl_node * dn = GetPointer<decl_node>(pointer);
               //This is case a[something]
               if(dn->artificial_flag)
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "It's artificial");
                  //This is an artificial variable which is computed in another vertex. Searching for it
                  InEdgeIterator ei, ei_end;
                  for(boost::tie(ei, ei_end) = boost::in_edges(*vi, *dfg); ei != ei_end; ei++)
                  {
                     std::set<unsigned int> nodeId = EDGE_GET_NODEID(dfg, *ei, DFG_SELECTOR);
                     if(nodeId.find(*it) != nodeId.end())
                     {
                        //In new current we could have the pointer_plus_expr
                        vertex new_current = boost::source(*ei, *dfg);
                        PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "It comes from " + GET_NAME(dfg, new_current));
                        if(GET_NODEID(dfg, new_current).size() > 0)
                        {
                           //Checking if it's a plus_pointer_expr
                           unsigned int op_id = GET_NODEID(dfg, new_current).back();
                           tree_nodeRef op = TM->get_tree_node_const(op_id);
                           PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Examining its operation " + boost::lexical_cast<std::string>(op_id));
                           if(op->get_kind() != gimple_assign_K)
                              break;
                           gimple_assign * gm = GetPointer<gimple_assign>(op);
                           op_id = GET_INDEX_NODE(gm->op1);
                           op = GET_NODE(gm->op1);
                           if(GetPointer<pointer_plus_expr>(op))
                           {
                              pointer_plus_expr * pe = GetPointer<pointer_plus_expr>(op);
                              if(helper->is_a_constant(GET_INDEX_NODE(pe->op0)))
                              {
                                 PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "First element is constant");
                                 ssa_name * ssan = GetPointer<ssa_name>(GET_NODE(pe->op1));
                                 base = GET_NODE(ssan->var);
                                 base_id = GET_INDEX_NODE(ssan->var);
                                 integer_cst * ic = GetPointer<integer_cst>(GET_NODE(pe->op0));
                                 offset = tree_helper::get_integer_cst_value(ic);
                              }
                              else if(helper->is_a_constant(GET_INDEX_NODE(pe->op1)))
                              {
                                 PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Second element is constant");
                                 if(GetPointer<ssa_name>(GET_NODE(pe->op0)))
                                 {
                                    ssa_name * ssan = GetPointer<ssa_name>(GET_NODE(pe->op0));
                                    base = GET_NODE(ssan->var);
                                    base_id = GET_INDEX_NODE(ssan->var);
                                    PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Base is " + boost::lexical_cast<std::string>(base_id));
                                    PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Offset is " + boost::lexical_cast<std::string>(GET_INDEX_NODE(pe->op1)));
                                    integer_cst * ic = GetPointer<integer_cst>(GET_NODE(pe->op1));
                                    offset = tree_helper::get_integer_cst_value(ic);
                                 }
                              }
                           }
                        }
                     }
                  }
               }
            }
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Ptr_info is " + boost::lexical_cast<std::string>(ptr_info) + " base is " + boost::lexical_cast<std::string>(ptr_info_to_base[ptr_info]) + " - this base is " + boost::lexical_cast<std::string>(base_id));
            //In this way we say that ptr_info is not a candidate
            if(ptr_info_to_base[ptr_info] != base_id)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "ptr_info does not match");
               if(ptr_info_to_base[ptr_info] || candidate_pointers.find(base_id) == candidate_pointers.end())
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "ptr_info " + boost::lexical_cast<std::string>(ptr_info) + "should be ignored");
                  ptr_info_to_base[ptr_info] = ptr_info;
                  continue;
               }
               else
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Associating to " + boost::lexical_cast<std::string>(ptr_info) + " base " + boost::lexical_cast<std::string>(base_id));
                  ptr_info_to_base[ptr_info] = base_id;
               }
            }
            pointer_to_base_and_offset[pointer_id] = std::pair<unsigned int, long long int>(base_id, offset);
            std::pair<vertex, unsigned int> temp(*vi, pointer_id);
            ptr_info_to_vertex_and_pointer[ptr_info].insert(temp);
         }
      }
   }

   std::map<vertex, std::set<vertex> > & bb_reachability = FB->bb_reachability;

   std::map<vertex, unsigned int> op_levels = FB->get_map_levels();

   std::map<unsigned int, std::set<std::pair<vertex, unsigned int> > >::const_iterator it_end = ptr_info_to_vertex_and_pointer.end();
   for(std::map<unsigned int, std::set<std::pair<vertex, unsigned int> > >::const_iterator it = ptr_info_to_vertex_and_pointer.begin(); it != it_end; it++)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Examining memory tag " + boost::lexical_cast<std::string>(it->first));
      tree_nodeRef tn = TM->get_tree_node_const(it->first);
      THROW_ASSERT(GetPointer<memory_tag>(tn), "ptr_info does not points to a memory tag");
      memory_tag * mt = GetPointer<memory_tag>(tn);
      if(mt->list_of_aliases.size() != 1 || ptr_info_to_base[it->first] == it->first)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "It's alias of more then one element");
         continue;
      }
      unsigned int alias = GET_INDEX_NODE(mt->list_of_aliases.front());
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Alias is " + boost::lexical_cast<std::string>(alias));
      EdgeIterator e, e_end;
      for(boost::tie(e, e_end) = boost::edges(*dfg); e != e_end; e++)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Examining edge from " + GET_NAME(dfg, boost::source(*e, *dfg)) + " to " + GET_NAME(dfg, boost::target(*e, *dfg)));
         if(EDGE_GET_NODEID(dfg, *e, DFG_SELECTOR).size() == 1)
         {
            unsigned int label_id = *(EDGE_GET_NODEID(dfg, *e, (DFG_SELECTOR)).begin());
            tree_nodeRef label = TM->get_tree_node_const(label_id);
            if(GetPointer<ssa_name>(label))
            {
               ssa_name * sn = GetPointer<ssa_name>(label);
               label_id = GET_INDEX_NODE(sn->var);
            }
            if(label_id == alias)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Removing edge");
               int old_selector =dfg->get_selector(*e);
               ogc->set_edge_type(boost::source(*e, *dfg), boost::target(*e, *dfg), old_selector & ~DFG_SELECTOR);
            }
         }
      }
      for(boost::tie(e, e_end) = boost::edges(*adg); e != e_end; e++)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Examining edge from " + GET_NAME(dfg, boost::source(*e, *adg)) + " to " + GET_NAME(dfg, boost::target(*e, *adg)));
         if(EDGE_GET_NODEID(adg, *e, ADG_SELECTOR).size() == 1)
         {
            unsigned int label_id = *(EDGE_GET_NODEID(dfg, *e, (ADG_SELECTOR)).begin());
            tree_nodeRef label = TM->get_tree_node_const(label_id);
            if(GetPointer<ssa_name>(label))
            {
               ssa_name * sn = GetPointer<ssa_name>(label);
               label_id = GET_INDEX_NODE(sn->var);
            }
            if(label_id == alias)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Removing edge");
               int old_selector = adg->get_selector(*e);
               ogc->set_edge_type(boost::source(*e, *adg), boost::target(*e, *adg), old_selector & ~ADG_SELECTOR);
            }
         }
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Finished with removing of edges for this memory tag");
      std::set<std::pair<vertex, unsigned int> > temp_map = it->second;
      std::set<std::pair<vertex, unsigned int> >::const_iterator iit2_end = temp_map.end();
      for(std::set<std::pair<vertex, unsigned int> >::const_iterator it2 = temp_map.begin(); it2 != iit2_end;  it2++)
      {
         vertex ext_vertex = it2->first;
         unsigned int ext_pointer = it2->second;
         long long int ext_offset = pointer_to_base_and_offset[ext_pointer].second;
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "First vertex is " + GET_NAME(dfg, ext_vertex) + " - pointer " + boost::lexical_cast<std::string>(ext_pointer) + " - offset " + boost::lexical_cast<std::string>(ext_offset));
         std::set<std::pair<vertex, unsigned int> >::const_iterator it3_end = temp_map.end();
         for(std::set<std::pair<vertex, unsigned int> >::const_iterator it3 = temp_map.begin(); it3 != it3_end;  it3++)
         {
            vertex int_vertex = it3->first;
            unsigned int int_pointer = it3->second;
            long long int int_offset = pointer_to_base_and_offset[int_pointer].second;
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Second vertex is " + GET_NAME(dfg, int_vertex) + " - pointer " + boost::lexical_cast<std::string>(int_pointer) + " - offset " + boost::lexical_cast<std::string>(int_offset));
            if(int_offset == ext_offset && ext_vertex != int_vertex)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Vertices are different and reference same memory location");
               vertex source;
               vertex dest;
               bool r__ = false;
               bool w__ = false;
               bool __r = false;
               bool __w = false;
               unsigned int ext_bb = CGET_NODE_INFO(dfg, OpNodeInfo, ext_vertex)->bb_index;
               unsigned int int_bb = CGET_NODE_INFO(dfg, OpNodeInfo, int_vertex)->bb_index;
               const std::tr1::unordered_map<unsigned int, vertex> & bb_index_map = bb_cfg->GetConstBBGraphInfo()->bb_index_map;
               const vertex ext_bb_vertex = bb_index_map.find(ext_bb)->second;
               const vertex int_bb_vertex = bb_index_map.find(int_bb)->second;
               if(ext_bb == int_bb)
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Vertices belong to the same basic block");
                  if(op_levels[ext_vertex] < op_levels[int_vertex])
                  {
                     source = ext_vertex;
                     dest = int_vertex;
                  }
                  else
                  {
                     source = int_vertex;
                     dest = ext_vertex;
                  }
               }
               else
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Vertices don't belong to the same basic block");
                  if(bb_reachability[ext_bb_vertex].find(int_bb_vertex) == bb_reachability[ext_bb_vertex].end())
                  {
                     if(bb_reachability[int_bb_vertex].find(ext_bb_vertex) == bb_reachability[int_bb_vertex].end())
                     {
                        PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "There is no path between vertices");
                        continue;
                     }
                     source = int_vertex;
                     dest = ext_vertex;
                  }
                  else
                  {
                     source = ext_vertex;
                     dest = int_vertex;
                  }
               }
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Source node is " + GET_NAME(dfg, source));
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Dest node is " + GET_NAME(dfg, dest));
               std::list<unsigned int> nodeId = GET_NODEID(dfg, source);
               unsigned int op_id = nodeId.back();
               tree_nodeRef op = TM->get_tree_node_const(op_id);
               THROW_ASSERT(GetPointer<gimple_assign>(op), "Not a gimple_assign");
               gimple_assign * gm = GetPointer<gimple_assign>(op);
               if(!(gm->virtual_operands))
                  continue;
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Souce has virtual_operands");
               vopsRef virtual_operands = gm->virtual_operands;
               std::vector<vdefRef> list_of_vdefs = virtual_operands->list_of_vdefs;
               for(unsigned int it4 = 0; it4 < list_of_vdefs.size() && !w__; it4++)
               {
                  std::vector<tree_nodeRef> list_of_stores = list_of_vdefs[it4]->list_of_stores;
                  for(unsigned int it5 = 0; it5 < list_of_stores.size() && !w__; it5++)
                     if(GET_INDEX_NODE(list_of_stores[it5]) == alias)
                        w__ = true;
               }
               if(!w__)
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Source does not store memory tag");
                  std::vector<vuseRef> list_of_vuses = virtual_operands->list_of_vuses;
                  for(unsigned int it4 = 0; it4 < list_of_vuses.size() && !r__; it4++)
                  {
                     std::vector<tree_nodeRef> list_of_loads = list_of_vuses[it4]->list_of_loads;
                     for(unsigned int it5 = 0; it5 < list_of_loads.size() && r__; it5++)
                        if(GET_INDEX_NODE(list_of_loads[it5]) == alias)
                           r__ = true;
                  }
                  if(!r__)
                     continue;
               }

               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Finished analyis of source");
               nodeId = GET_NODEID(dfg, dest);
               op_id = nodeId.back();
               op = TM->get_tree_node_const(op_id);
               THROW_ASSERT(GetPointer<gimple_assign>(op), "Not a gimple_assign");
               gm = GetPointer<gimple_assign>(op);
               if(!(gm->virtual_operands))
                  continue;
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Dest has virtual_operands");
               virtual_operands = gm->virtual_operands;
               list_of_vdefs = virtual_operands->list_of_vdefs;
               for(unsigned int it4 = 0; it4 < list_of_vdefs.size() && !__w; it4++)
               {
                  std::vector<tree_nodeRef> list_of_stores = list_of_vdefs[it4]->list_of_stores;
                  for(unsigned int it5 = 0; it5 < list_of_stores.size() && !__w; it5++)
                     if(GET_INDEX_NODE(list_of_stores[it5]) == alias)
                        __w = true;
               }
               if(!__w)
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Dest does not store memory tag");
                  std::vector<vuseRef> list_of_vuses = virtual_operands->list_of_vuses;
                  for(unsigned int it4 = 0; it4 < list_of_vuses.size() && !__r; it4++)
                  {
                     std::vector<tree_nodeRef> list_of_loads = list_of_vuses[it4]->list_of_loads;
                     for(unsigned int it5 = 0; it5 < list_of_loads.size() && !__r; it5++)
                        if(GET_INDEX_NODE(list_of_loads[it5]) == alias)
                           __r = true;
                  }
                  if(!__r)
                     continue;   
               }
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "An edge has to be added");
               if(w__ && __r)
               {
                  ogc->add_edge(source, dest, DFG_SELECTOR);
                  ogc->add_edge_info(source, dest, DFG_SELECTOR, alias);
               }
               if(r__ && __w)
               {
                  ogc->add_edge(source, dest, ADG_SELECTOR);
                  ogc->add_edge_info(source, dest, ADG_SELECTOR, alias);
               }
            }
         }
      }
   }



   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, helper->get_function_name() + " Finished data flow cleaning");
   if(output_level >= OUTPUT_LEVEL_VERY_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
   {
      std::string file_name = AppM->get_function_behavior(funID)->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_FDFG.dot";
      dfg->print_to_dot(Param, file_name, helper);

      file_name = AppM->get_function_behavior(funID)->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_FADG.dot";
      adg->print_to_dot(Param, file_name, helper);
   }
}

void data_flow_cleaning::get_candidate_pointers(unsigned int NodeID, std::set<unsigned int> & candidate_pointers, std::set<unsigned int> & not_candidate_pointers)
{
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "data_flow_cleaning::get_candidate_pointers - Start - NodeID: " + boost::lexical_cast<std::string>(NodeID));
   const tree_nodeRef tn = TM->get_tree_node_const(NodeID);

   switch(tn->get_kind())
   {
      case indirect_ref_K:
      {
         indirect_ref * ir = GetPointer<indirect_ref>(tn);
         if(GET_NODE(ir->op)->get_kind() != pointer_plus_expr_K)
            break;
         get_candidate_pointers(GET_INDEX_NODE(ir->op), candidate_pointers, not_candidate_pointers);
         break;
      }
      case pointer_plus_expr_K:
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Node is of type pointer_plus_expr");
         pointer_plus_expr * pe = GetPointer<pointer_plus_expr>(tn);
         if(helper->is_a_constant(GET_INDEX_NODE(pe->op0)))
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "First element is constant");
            tree_nodeRef pointer = GET_NODE(pe->op1);
            if(GetPointer<decl_node>(pointer))
               candidate_pointers.insert(GET_INDEX_NODE(pe->op1));
            else if (GetPointer<ssa_name>(pointer))
            {
               ssa_name * sn = GetPointer<ssa_name>(pointer);
               candidate_pointers.insert(GET_INDEX_NODE(sn->var));
            }
         }
         else if(helper->is_a_constant(GET_INDEX_NODE(pe->op1)))
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Second element is constant");
            tree_nodeRef pointer = GET_NODE(pe->op0);
            if(GetPointer<decl_node>(pointer))
               candidate_pointers.insert(GET_INDEX_NODE(pe->op0));
            else if (GetPointer<ssa_name>(pointer))
            {
               ssa_name * sn = GetPointer<ssa_name>(pointer);
               candidate_pointers.insert(GET_INDEX_NODE(sn->var));
            }
         }
         else
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "None of element is constant");
            tree_nodeRef pointer = GET_NODE(pe->op0);
            if(GetPointer<ssa_name>(pointer) || GetPointer<decl_node>(pointer))
               not_candidate_pointers.insert(GET_INDEX_NODE(pe->op0));
            pointer = GET_NODE(pe->op1);
            if(GetPointer<ssa_name>(pointer) || GetPointer<decl_node>(pointer))
               not_candidate_pointers.insert(GET_INDEX_NODE(pe->op1));
         }
         break;
      }
      case gimple_assign_K:
      {
         gimple_assign * gm = GetPointer<gimple_assign>(tn);
         get_candidate_pointers(GET_INDEX_NODE(gm->op0), candidate_pointers, not_candidate_pointers);
         get_candidate_pointers(GET_INDEX_NODE(gm->op1), candidate_pointers, not_candidate_pointers);
         break;
      }
      case gimple_return_K:
      {
         gimple_return * re = GetPointer<gimple_return>(tn);
         if(re->op)
            get_candidate_pointers(GET_INDEX_NODE(re->op), candidate_pointers, not_candidate_pointers);
         break;
      }
      case call_expr_K:
      {
         call_expr * ce = GetPointer<call_expr>(tn);
         const std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            unsigned int index = GET_INDEX_NODE(*arg);
            get_candidate_pointers(index, candidate_pointers, not_candidate_pointers);
         }
         break;
      }
      case gimple_call_K:
      {
         gimple_call * ce = GetPointer<gimple_call>(tn);
         const std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            unsigned int index = GET_INDEX_NODE(*arg);
            get_candidate_pointers(index, candidate_pointers, not_candidate_pointers);
         }
         break;
      }
      case cond_expr_K:
      {
         cond_expr * ce = GetPointer<cond_expr>(tn);
         get_candidate_pointers(GET_INDEX_NODE(ce->op0), candidate_pointers, not_candidate_pointers);
         get_candidate_pointers(GET_INDEX_NODE(ce->op1), candidate_pointers, not_candidate_pointers);
         get_candidate_pointers(GET_INDEX_NODE(ce->op2), candidate_pointers, not_candidate_pointers);
         break;
      }
      case gimple_cond_K:
      {
         gimple_cond * gc = GetPointer<gimple_cond>(tn);
         get_candidate_pointers(GET_INDEX_NODE(gc->op0), candidate_pointers, not_candidate_pointers);
         break;
      }
      case gimple_switch_K:
      {
         gimple_switch * se = GetPointer<gimple_switch>(tn);
         get_candidate_pointers(GET_INDEX_NODE(se->op0), candidate_pointers, not_candidate_pointers);
         if(se->op1)
            get_candidate_pointers(GET_INDEX_NODE(se->op1), candidate_pointers, not_candidate_pointers);
         break;
      }
      //Unary expression
      case nop_expr_K:
      case abs_expr_K:
      case addr_expr_K:
      case arrow_expr_K:
      case bit_not_expr_K:
      case buffer_ref_K:
      case card_expr_K:
      case cast_expr_K:
      case cleanup_point_expr_K:
      case conj_expr_K:
      case convert_expr_K:
      case exit_expr_K:
      case fix_ceil_expr_K:
      case fix_floor_expr_K:
      case fix_round_expr_K:
      case fix_trunc_expr_K:
      case float_expr_K:
      case gimple_goto_K:
      case imagpart_expr_K:
      case loop_expr_K:
      case negate_expr_K:
      case non_lvalue_expr_K:
      case realpart_expr_K:
      case reference_expr_K:
      case reinterpret_cast_expr_K:
      case gimple_resx_K:
      case sizeof_expr_K:
      case static_cast_expr_K:
      case throw_expr_K:
      case truth_not_expr_K:
      case unsave_expr_K:
      case va_arg_expr_K:
      case view_convert_expr_K:
      {
         unary_expr * ue = GetPointer<unary_expr>(tn);
         get_candidate_pointers(GET_INDEX_NODE(ue->op), candidate_pointers, not_candidate_pointers);
         break;
      }
      case vector_cst_K:
      {
         vector_cst* vc = GetPointer<vector_cst>(tn);
         std::vector<tree_nodeRef> list_of_valu = vc->list_of_valu;
         for(unsigned int i = 0; i < list_of_valu.size(); i++)
            get_candidate_pointers(GET_INDEX_NODE(list_of_valu[i]), candidate_pointers, not_candidate_pointers);
         break;
      }
      case tree_list_K:
      {
         if (GetPointer<tree_list>(tn)->purp)
         {
            unsigned int index = GET_INDEX_NODE(GetPointer<tree_list>(tn)->purp);
            get_candidate_pointers(index, candidate_pointers, not_candidate_pointers);
         }
         tree_nodeRef actual_args = tn;
         while (actual_args)
         {
            unsigned int index = GET_INDEX_NODE(GetPointer<tree_list>(actual_args)->valu);
            get_candidate_pointers(index, candidate_pointers, not_candidate_pointers);
            if (GetPointer<tree_list>(actual_args)->chan)
               actual_args = GET_NODE(GetPointer<tree_list>(actual_args)->chan);
            else 
               break;
         }
         break;
      }
      case tree_vec_K:
      {
         tree_vec* tv = GetPointer<tree_vec>(tn);
         std::vector<tree_nodeRef> list_of_op = tv->list_of_op;
         for(unsigned int i = 0; i < list_of_op.size(); i++)
            get_candidate_pointers(GET_INDEX_NODE(list_of_op[i]), candidate_pointers, not_candidate_pointers);
         break;
      }
      //Ternary expr
      case bit_field_ref_K:
      case component_ref_K:
      {
         ternary_expr * expr = GetPointer<ternary_expr>(tn);

         get_candidate_pointers(GET_INDEX_NODE(expr->op0), candidate_pointers, not_candidate_pointers);
         break;
      }
      case bit_and_expr_K:
      case bit_ior_expr_K:
      case bit_xor_expr_K:
      case catch_expr_K:
      case ceil_div_expr_K:
      case ceil_mod_expr_K:
      case complex_expr_K:
      case compound_expr_K:
      case eh_filter_expr_K:
      case eq_expr_K:
      case exact_div_expr_K:
      case fdesc_expr_K:
      case floor_div_expr_K:
      case floor_mod_expr_K:
      case ge_expr_K:
      case gt_expr_K:
      case goto_subroutine_K:
      case in_expr_K:
      case init_expr_K:
      case le_expr_K:
      case lrotate_expr_K:
      case lshift_expr_K:
      case lt_expr_K:
      case max_expr_K:
      case min_expr_K:
      case minus_expr_K:
      case mult_expr_K:
      case ne_expr_K:
      case ordered_expr_K:
      case plus_expr_K:
      case postdecrement_expr_K:
      case postincrement_expr_K:
      case predecrement_expr_K:
      case preincrement_expr_K:
      case range_expr_K:
      case rdiv_expr_K:
      case round_div_expr_K:
      case round_mod_expr_K:
      case rrotate_expr_K:
      case rshift_expr_K:
      case set_le_expr_K:
      case trunc_div_expr_K:
      case trunc_mod_expr_K:
      case truth_and_expr_K:
      case truth_andif_expr_K:
      case truth_or_expr_K:
      case truth_orif_expr_K:
      case truth_xor_expr_K:
      case try_catch_expr_K:
      case try_finally_K:
      case uneq_expr_K:
      case ltgt_expr_K:
      case unge_expr_K:
      case ungt_expr_K:
      case unle_expr_K:
      case unlt_expr_K:
      case unordered_expr_K:
      case with_size_expr_K:
      {
         binary_expr * be = GetPointer<binary_expr>(tn);
         get_candidate_pointers(GET_INDEX_NODE(be->op0), candidate_pointers, not_candidate_pointers);
         get_candidate_pointers(GET_INDEX_NODE(be->op1), candidate_pointers, not_candidate_pointers);

         break;
      }
      case array_ref_K:
      {
         quaternary_expr * expr = GetPointer<quaternary_expr>(tn);
         get_candidate_pointers(GET_INDEX_NODE(expr->op0), candidate_pointers, not_candidate_pointers);
         get_candidate_pointers(GET_INDEX_NODE(expr->op1), candidate_pointers, not_candidate_pointers);
         if(expr->op2)
            get_candidate_pointers(GET_INDEX_NODE(expr->op2), candidate_pointers, not_candidate_pointers);
         if (expr->op3)
            get_candidate_pointers(GET_INDEX_NODE(expr->op3), candidate_pointers, not_candidate_pointers);

         break;
      }
      case constructor_K:
      {
         constructor* con = GetPointer<constructor>(tn);
         std::vector<std::pair< tree_nodeRef, tree_nodeRef> > list_of_idx_valu = con->list_of_idx_valu;
         for(std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::iterator el = list_of_idx_valu.begin(); el != list_of_idx_valu.end(); el++)
         {
            if (el->first)
               get_candidate_pointers(GET_INDEX_NODE(el->first), candidate_pointers, not_candidate_pointers);
            if (el->second)
               get_candidate_pointers(GET_INDEX_NODE(el->second), candidate_pointers, not_candidate_pointers);
         }
         break;
      }
      //Special node
      case case_label_expr_K:
      case gimple_label_K:
      case gimple_asm_K:
      //Decl node
      case field_decl_K:
      case function_decl_K:
      case label_decl_K:
      case result_decl_K:
      //Cast
      case real_cst_K:
      case complex_cst_K:
      case string_cst_K:
      case integer_cst_K:
      //Var
      case parm_decl_K:
      case var_decl_K:
      case ssa_name_K:
      {
         break;
      }
      default:
         THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "Not supported rvalue NodeID: " + std::string(tn->get_kind_text()));
   }
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "data_flow_cleaning::get_candidate_pointers - End - NodeID: " + boost::lexical_cast<std::string>(NodeID));
   return;
}

std::string data_flow_cleaning::get_kind_text() const
{
   return "data_flow_cleaning";
}

