/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file determine_memory_accesses.cpp
 * @brief Class to determine the variable to be stored in memory
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#include "determine_memory_accesses.hpp"

#include "application_manager.hpp"
#include "function_behavior.hpp"

#include "tree_basic_block.hpp"
#include "tree_manager.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"
#include "behavioral_helper.hpp"

#include "Parameter.hpp"


determine_memory_accesses::determine_memory_accesses(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int _function_id):
      BaseAnalysis(_Param, _AppM, DETERMINE_MEMORY_ACCESSES, _function_id)
{
   debug_level = Param->get_class_debug_level("determine_memory_accesses", DEBUG_LEVEL_NONE);
   this->SetPreCond(ID_deep, DETERMINE_MEMORY_ACCESSES);
   this->SetPreCond(ID, BASIC_BLOCKS_CFG_COMPUTATION);
   this->SetPostCond(ID, CFG_CREATION);
}

determine_memory_accesses::~determine_memory_accesses()
{
}

void determine_memory_accesses::exec()
{
   const tree_managerRef TM = AppM->get_tree_manager();
   tree_nodeRef tn = TM->get_tree_node_const(funID);
   function_decl * fd = GetPointer<function_decl>(tn);
   if (!fd || !fd->body)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Node is not a function or it hasn't a body");
      return;
   }

   const behavioral_helperRef BH = AppM->get_function_behavior(funID)->get_helper();
   std::string function_name = BH->get_function_name();

   if(debug_level >= DEBUG_LEVEL_PEDANTIC)
   {
      std::string raw_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "before_determine_memory_accesses." + function_name + ".raw";
      std::ofstream raw_file(raw_file_name.c_str());
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Tree-Manager dumped for debug purpose");
      raw_file << TM;
      raw_file.close();
   }

   /// analyze formal parameters
   std::vector <tree_nodeRef>::const_iterator formal_it_end = fd->list_of_args.end();
   for(std::vector <tree_nodeRef>::const_iterator formal_it = fd->list_of_args.begin(); formal_it != formal_it_end; ++formal_it)
      analyze_node(GET_INDEX_NODE(*formal_it), TM, BH, false);

   statement_list * sl = GetPointer<statement_list>(GET_NODE(fd->body));
   THROW_ASSERT(sl, "Body is not a statement_list");
   std::map<unsigned int, blocRef>::iterator it_bb, it_bb_end = sl->list_of_bloc.end();
   for(it_bb = sl->list_of_bloc.begin(); it_bb != it_bb_end ; it_bb++)
   {
      if (it_bb->second->number == BB_ENTRY || it_bb->second->number == BB_EXIT) continue;
      const std::vector<tree_nodeRef>& list_of_phi = it_bb->second->list_of_phi;
      for(unsigned int s = 0; s < list_of_phi.size(); s++)
      {
         analyze_node(GET_INDEX_NODE(list_of_phi[s]), TM, BH, false);
      }
      const std::vector<tree_nodeRef>& list_of_stmt = it_bb->second->list_of_stmt;
      for(unsigned int s = 0; s < list_of_stmt.size(); s++)
      {
         analyze_node(GET_INDEX_NODE(list_of_stmt[s]), TM, BH, false);
      }
   }

   if(output_level >= OUTPUT_LEVEL_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
      AppM->write_call_graph_dot("CG_memory_analysis.dot");
}  

std::string determine_memory_accesses::get_kind_text() const
{
   return "determine_memory_accesses";
}

void determine_memory_accesses::analyze_node(unsigned int node_id, const tree_managerRef& TM, const behavioral_helperRef& BH, bool left_p)
{
   const tree_nodeRef tn = TM->get_tree_node_const(node_id);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node " + boost::lexical_cast<std::string>(node_id) + " of type " + tn->get_kind_text());
   const function_behaviorRef FB = AppM->get_function_behavior(funID);

   switch(tn->get_kind())
   {
      case gimple_assign_K:
      {
         gimple_assign* gm = GetPointer<gimple_assign>(tn);
         analyze_node(GET_INDEX_NODE(gm->op0), TM, BH, true);
         analyze_node(GET_INDEX_NODE(gm->op1), TM, BH, false);
         break;
      }
      case CASE_UNARY_EXPRESSION:
      {
         unary_expr* ue = GetPointer<unary_expr>(tn);
         if (GetPointer<addr_expr>(tn))
         {
            if (GetPointer<var_decl>(GET_NODE(ue->op)))
            {
               FB->add_function_mem(GET_INDEX_NODE(ue->op));
            }
            else if (GetPointer<parm_decl>(GET_NODE(ue->op)))
            {
               FB->add_function_mem(GET_INDEX_NODE(ue->op));
               /// an address of a parm decl may be used in writing so it has to be copied
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node: formal parameter copied " + STR(GET_INDEX_NODE(ue->op)));
               FB->add_parm_decl_copied(GET_INDEX_NODE(ue->op));
            }
            else if (GetPointer<string_cst>(GET_NODE(ue->op)))
            {
               FB->add_function_mem(GET_INDEX_NODE(ue->op));
            }
            else if(GetPointer<component_ref>(GET_NODE(ue->op)) ||
                    GetPointer<realpart_expr>(GET_NODE(ue->op)) ||
                    GetPointer<imagpart_expr>(GET_NODE(ue->op)) ||
                    GetPointer<array_ref>(GET_NODE(ue->op)))
               analyze_node(GET_INDEX_NODE(ue->op), TM, BH, true);
            else if(GetPointer<function_decl>(GET_NODE(ue->op)))
               THROW_ERROR("pointers to functions not yet supported");
            else if(GetPointer<mem_ref>(GET_NODE(ue->op)))
               ; /// nothing to be done
            else
               THROW_ERROR("determine_memory_accesses addressing currently not supported: " + GET_NODE(ue->op)->get_kind_text()+" @"+STR(node_id));
         }
         else
            analyze_node(GET_INDEX_NODE(ue->op), TM, BH, left_p);
         break;
      }
      case CASE_BINARY_EXPRESSION:
      {
         binary_expr* be = GetPointer<binary_expr>(tn);
         analyze_node(GET_INDEX_NODE(be->op0), TM, BH, left_p);
         analyze_node(GET_INDEX_NODE(be->op1), TM, BH, left_p);
         break;
      }
      case gimple_cond_K:
      {
         gimple_cond* gc = GetPointer<gimple_cond>(tn);
         analyze_node(GET_INDEX_NODE(gc->op0), TM, BH, left_p);
         break;
      }
      case gimple_switch_K:
      {
         gimple_switch* se = GetPointer<gimple_switch>(tn);
         if (se->op0) analyze_node(GET_INDEX_NODE(se->op0), TM, BH, left_p);
         break;
      }
      case gimple_phi_K:
      {
         gimple_phi * gp = GetPointer<gimple_phi>(tn);
         size_t n_sources = gp->list_of_def_edge.size();
         for(unsigned int i = 0; i < n_sources; ++i)
            analyze_node(GET_INDEX_NODE(gp->list_of_def_edge[i].first), TM, BH, left_p);
         break;
      }
      case CASE_TERNARY_EXPRESSION:
      {
         ternary_expr* te = GetPointer<ternary_expr>(tn);
         if (te->op0) analyze_node(GET_INDEX_NODE(te->op0), TM, BH, left_p);
         if (te->op1) analyze_node(GET_INDEX_NODE(te->op1), TM, BH, left_p);
         if (te->op2) analyze_node(GET_INDEX_NODE(te->op2), TM, BH, left_p);
         break;
      }
      case CASE_QUATERNARY_EXPRESSION:
      {
         quaternary_expr* qe = GetPointer<quaternary_expr>(tn);
         if (qe->op0) analyze_node(GET_INDEX_NODE(qe->op0), TM, BH, left_p);
         if (qe->op1) analyze_node(GET_INDEX_NODE(qe->op1), TM, BH, left_p);
         if (qe->op2) analyze_node(GET_INDEX_NODE(qe->op2), TM, BH, left_p);
         if (qe->op3) analyze_node(GET_INDEX_NODE(qe->op3), TM, BH, left_p);
         break;
      }
      case gimple_return_K:
      {
         gimple_return* re = GetPointer<gimple_return>(tn);
         if (re->op) analyze_node(GET_INDEX_NODE(re->op), TM, BH, left_p);
         break;
      }
      case call_expr_K:
      {
         call_expr* ce = GetPointer<call_expr>(tn);
         std::vector<tree_nodeRef> & args = ce->args;
         addr_expr* ae = GetPointer<addr_expr>(GET_NODE(ce->fn));
         THROW_ASSERT(ae, "pointers to functions not yet supported");
         function_decl* fd = GetPointer<function_decl>(GET_NODE(ae->op));
         bool is_var_args_p = GetPointer<function_type>(GET_NODE(fd->type))->varargs_flag;
         THROW_ASSERT(fd, "expected a function_decl");
         std::vector<tree_nodeRef>::iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; ++arg)
         {
            analyze_node(GET_INDEX_NODE(*arg), TM, BH, left_p);
         }
         if(!fd->undefined_flag)
         {
            std::vector <tree_nodeRef>::const_iterator formal_it = fd->list_of_args.begin();
            std::vector <tree_nodeRef>::const_iterator formal_it_end = fd->list_of_args.end();
            std::string function_name = BH->get_function_name();
            if(!(is_var_args_p || fd->list_of_args.size() == args.size()))
                  THROW_ERROR("In function " + function_name + " a different number of formal and actual parameters is found when function " + fd->get_function_name() + " is called: " + STR(fd->list_of_args.size()) + " - " +STR(args.size()) + "\n Check the C source code since an actual parameter is passed to a function that does have the associated formal parameter");
            for(arg = args.begin(); arg != arg_end && formal_it != formal_it_end; ++arg, ++formal_it)
            {
               unsigned int actual_par_index = GET_INDEX_NODE(*arg);
               unsigned int formal_par_index = GET_INDEX_NODE(*formal_it);
               unsigned int calledFundID = GET_INDEX_NODE(ae->op);
               function_behaviorRef FBcalled = AppM->get_function_behavior(calledFundID);
               /// check if the actual parameter has been allocated in memory
               if(FB->is_variable_mem(actual_par_index))
               {
                  /// if the formal parameter has not been allocated in memory then it has to be initialized with a MEMLOAD_STD
                  if(!FBcalled->is_variable_mem(formal_par_index))
                  {
                     PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node: actual parameter loaded " + STR(actual_par_index));
                     FB->add_parm_decl_loaded(actual_par_index);
                  }
               }
               /// check if the formal parameter has been allocated in memory.
               if(FBcalled->is_variable_mem(formal_par_index))
               {
                  /// If the actual has not been allocated in memory then the formal parameter storage has to be initialized with the actual value with a MEMSTORE_STD
                  tree_nodeRef actual_par = GET_NODE(*arg);
                  switch(actual_par->get_kind())
                  {
                     case ssa_name_K:
                     {
                        if(!FB->is_variable_mem(actual_par_index))
                        {
                           PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node: formal parameter stored " + STR(formal_par_index));
                           FBcalled->add_parm_decl_stored(formal_par_index);
                        }
                        break;
                     }
                     case real_cst_K:
                     case string_cst_K:
                     case integer_cst_K:
                     case addr_expr_K:
                     {
                        PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node: formal parameter stored " + STR(formal_par_index));
                        FBcalled->add_parm_decl_stored(formal_par_index);
                        break;
                     }
                     case indirect_ref_K:
                     case array_ref_K:
                     case component_ref_K:
                     {
                        PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node: formal parameter copied " + STR(formal_par_index));
                        FBcalled->add_parm_decl_copied(formal_par_index);
                        break;
                     }
                     default:
                     {
                        THROW_ASSERT(FB->is_variable_mem(actual_par_index), "actual parameter non allocated in memory: calling @" + STR(calledFundID) + " actual @" + STR(actual_par_index));
                        break;
                     }
                  }
               }
            }
         }
         break;
      }
      case gimple_call_K:
      {
         gimple_call* ce = GetPointer<gimple_call>(tn);
         std::vector<tree_nodeRef> & args = ce->args;
         addr_expr* ae = GetPointer<addr_expr>(GET_NODE(ce->fn));
         THROW_ASSERT(ae, "pointers to functions not yet supported");
         function_decl* fd = GetPointer<function_decl>(GET_NODE(ae->op));
         bool is_var_args_p = GetPointer<function_type>(GET_NODE(fd->type))->varargs_flag;
         THROW_ASSERT(fd, "expected a function_decl");
         std::vector<tree_nodeRef>::iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; ++arg)
         {
            analyze_node(GET_INDEX_NODE(*arg), TM, BH, left_p);
         }
         if(!fd->undefined_flag)
         {
            std::vector <tree_nodeRef>::const_iterator formal_it = fd->list_of_args.begin();
            std::vector <tree_nodeRef>::const_iterator formal_it_end = fd->list_of_args.end();
            std::string function_name = BH->get_function_name();
            if(!(is_var_args_p || fd->list_of_args.size() == args.size()))
                  THROW_ERROR("In function " + function_name + " a different number of formal and actual parameters is found when function " + fd->get_function_name() + " is called: " + STR(fd->list_of_args.size()) + " - " +STR(args.size()) + "\n Check the C source code since an actual parameter is passed to a function that does have the associated formal parameter");
            for(arg = args.begin(); arg != arg_end && formal_it != formal_it_end; ++arg, ++formal_it)
            {
               unsigned int actual_par_index = GET_INDEX_NODE(*arg);
               unsigned int formal_par_index = GET_INDEX_NODE(*formal_it);
               unsigned int calledFundID = GET_INDEX_NODE(ae->op);
               function_behaviorRef FBcalled = AppM->get_function_behavior(calledFundID);
               /// check if the actual parameter has been allocated in memory
               if(FB->is_variable_mem(actual_par_index))
               {
                  /// if the formal parameter has not been allocated in memory then it has to be initialized with a MEMLOAD_STD
                  if(!FBcalled->is_variable_mem(formal_par_index))
                  {
                     PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node: actual parameter loaded " + STR(actual_par_index));
                     FB->add_parm_decl_loaded(actual_par_index);
                  }
               }
               /// check if the formal parameter has been allocated in memory.
               if(FBcalled->is_variable_mem(formal_par_index))
               {
                  /// If the actual has not been allocated in memory then the formal parameter storage has to be initialized with the actual value with a MEMSTORE_STD
                  tree_nodeRef actual_par = GET_NODE(*arg);
                  switch(actual_par->get_kind())
                  {
                     case ssa_name_K:
                     {
                        if(!FB->is_variable_mem(actual_par_index))
                        {
                           PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node: formal parameter stored " + STR(formal_par_index));
                           FBcalled->add_parm_decl_stored(formal_par_index);
                        }
                        break;
                     }
                     case real_cst_K:
                     case string_cst_K:
                     case integer_cst_K:
                     case addr_expr_K:
                     {
                        PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node: formal parameter stored " + STR(formal_par_index));
                        FBcalled->add_parm_decl_stored(formal_par_index);
                        break;
                     }
                     case indirect_ref_K:
                     case array_ref_K:
                     case component_ref_K:
                     {
                        PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node: formal parameter copied " + STR(formal_par_index));
                        FBcalled->add_parm_decl_copied(formal_par_index);
                        break;
                     }
                     default:
                     {
                        THROW_ASSERT(FB->is_variable_mem(actual_par_index), "actual parameter non allocated in memory: calling @" + STR(calledFundID) + " actual @" + STR(actual_par_index));
                        break;
                     }
                  }
               }
            }
         }
         break;
      }
      case ssa_name_K:
      {
         ssa_name* sn = GetPointer<ssa_name>(tn);
         var_decl *vd = GetPointer<var_decl>(GET_NODE(sn->var));
         if(vd && (GET_NODE(vd->type)->get_kind() == complex_type_K))
            FB->add_function_mem(node_id);
         break;
      }
      case real_cst_K:
      case integer_cst_K:
      case gimple_label_K:
      case label_decl_K:
      {
         break;
      }
      case complex_cst_K:
      case vector_cst_K:
      case string_cst_K:
      {
         FB->add_function_mem(node_id);
         break;
      }
      case parm_decl_K:
      {
         parm_decl *pd = GetPointer<parm_decl>(tn);
         if(GET_NODE(pd->type)->get_kind() == record_type_K || //records have to be allocated
            GET_NODE(pd->type)->get_kind() == union_type_K || // unions have to be allocated
            GET_NODE(pd->type)->get_kind() == complex_type_K //complexes are like structs and so they are allocated
            )
         {
            FB->add_function_mem(node_id);
            if(left_p)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node: formal parameter copied " + STR(node_id));
               FB->add_parm_decl_copied(node_id);
            }
         }
         break;
      }
      case result_decl_K:
      {
         result_decl *rd = GetPointer<result_decl>(tn);
         if(GET_NODE(rd->type)->get_kind() == record_type_K || //records have to be allocated
            GET_NODE(rd->type)->get_kind() == union_type_K || // unions have to be allocated
            GET_NODE(rd->type)->get_kind() == complex_type_K //complexes are like structs and so they are allocated
            )
         {
            THROW_ERROR("structs or unions returned by copy are not yet supported: @" + STR(node_id) + " in function @" + STR(funID));
            FB->add_parm_decl_copied(node_id);
            FB->add_function_mem(node_id);
         }
         break;
      }
      case tree_list_K:
      {
         tree_list* tl = GetPointer<tree_list>(tn);
         analyze_node(GET_INDEX_NODE(tl->valu), TM, BH, left_p);
         if (tl->chan) analyze_node(GET_INDEX_NODE(tl->chan), TM, BH, left_p);
         break;
      }
      case var_decl_K:
      {
         var_decl *vd = GetPointer<var_decl>(tn);
         if (!vd->scpe || GET_NODE(vd->scpe)->get_kind() == translation_unit_decl_K) //memory has to be allocated in case of global variables
         {
            FB->add_function_mem(node_id);
            if (vd->init) analyze_node(GET_INDEX_NODE(vd->init), TM, BH, false);
         }
         else
         {
            THROW_ASSERT(GET_NODE(vd->scpe)->get_kind() != translation_unit_decl_K, "translation_unit_decl not expected a translation unit in this point @" + STR(node_id));
            if(vd->static_flag || //memory has to be allocated in case of local static variables
                  tree_helper::is_volatile(TM,node_id) || // volatile vars have to be allocated
                  GET_NODE(vd->type)->get_kind() == array_type_K || //arrays have to be allocated
                  GET_NODE(vd->type)->get_kind() == complex_type_K || //complexes are like structs and so they are allocated
                  GET_NODE(vd->type)->get_kind() == record_type_K || //records have to be allocated
                  GET_NODE(vd->type)->get_kind() == union_type_K
                  )
            {
               FB->add_function_mem(node_id);
            }
            else
            {
               //nothing have to be allocated for the variable
               // maybe something has to be allocated for its initialization
               if (vd->init) analyze_node(GET_INDEX_NODE(vd->init), TM, BH, left_p);
            }
         }
         break;
      }
      case constructor_K:
      {
         constructor* con = GetPointer<constructor>(tn);
         std::vector<std::pair< tree_nodeRef, tree_nodeRef> > list_of_idx_valu = con->list_of_idx_valu;
         for(std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::iterator el = list_of_idx_valu.begin(); el != list_of_idx_valu.end(); el++)
         {
            if (el->first) analyze_node(GET_INDEX_NODE(el->first), TM, BH, left_p);
            if (el->second) analyze_node(GET_INDEX_NODE(el->second), TM, BH, left_p);
         }
         break;
      }
      case gimple_goto_K:
      {
         gimple_goto* ge = GetPointer<gimple_goto>(tn);
         analyze_node(GET_INDEX_NODE(ge->op), TM, BH, left_p);
         break;
      }
      case field_decl_K:
      {
         break;
      }
      case target_mem_ref_K:
      {
         target_mem_ref* tmr = GetPointer<target_mem_ref>(tn);
         if(tmr->symbol) analyze_node(GET_INDEX_NODE(tmr->symbol), TM, BH, left_p);
         if(tmr->base) analyze_node(GET_INDEX_NODE(tmr->base), TM, BH, left_p);
         if(tmr->idx) analyze_node(GET_INDEX_NODE(tmr->idx), TM, BH, left_p);
         break;
      }
      case target_mem_ref461_K:
      {
         target_mem_ref461* tmr = GetPointer<target_mem_ref461>(tn);
         if(tmr->base) analyze_node(GET_INDEX_NODE(tmr->base), TM, BH, left_p);
         if(tmr->idx) analyze_node(GET_INDEX_NODE(tmr->idx), TM, BH, left_p);
         if(tmr->idx2) analyze_node(GET_INDEX_NODE(tmr->idx2), TM, BH, left_p);
         break;
      }
      case function_decl_K:
      {
         THROW_ERROR("pointers to functions not yet supported");
         break;
      }
      default:
      {
         THROW_ERROR("Not supported node (@" + STR(node_id) + ") of type " + std::string(tn->get_kind_text()));
      }
   }
}
