/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file loop_regions_computation.cpp
 * @brief Analysis step creating the regions for loops.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/

///Header include
#include "loop_regions_computation.hpp"

///Behavior include
#include "behavioral_helper.hpp"
#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "loop.hpp"
#include "loops.hpp"

///Frontend include
#include "Parameter.hpp"

///Graph include
#include "graph.hpp"
#include "basic_block.hpp"
#include "basic_blocks_graph_constructor.hpp"

///STD include
#include <fstream>

///Tree include
#include "tree_basic_block.hpp"
#include "tree_manager.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"

///Utility include
#include "Dominance.hpp"

bool loop_regions_computation::tree_dumped = false;

loop_regions_computation::loop_regions_computation(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
      BaseAnalysis(_Param, _AppM, LOOP_REGIONS_COMPUTATION, function_id)
{
   this->SetPreCond(ID, LOOPS_IDENTIFICATION);
   if(!(Param->getOption<bool>(OPT_without_transformation)))
      SetPreCond(ID, LOOPS_ANALYSIS);
   this->SetPostCond(ID, OP_FEEDBACK_EDGES_IDENTIFICATION);
   this->SetPostCond(ID, CFG_CREATION);
}


loop_regions_computation::~loop_regions_computation()
{

}

void loop_regions_computation::exec()
{
   const tree_managerRef TM = AppM->get_tree_manager();
   if(debug_level >= DEBUG_LEVEL_PEDANTIC && !tree_dumped)
   {
      tree_dumped = true;
      std::string raw_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "before_loop_regions_computation.raw";
      std::ofstream raw_file(raw_file_name.c_str());
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Tree-Manager dumped for debug purpose");
      raw_file << TM;
      raw_file.close();
   }

   bool added_block = false;
   const function_behaviorRef FB = AppM->get_function_behavior(funID);
   bb_graph * fbb = FB->fbb;
   behavioral_helperRef & helper = FB->helper;

   ///first label expressions are inserted
   const std::list<LoopConstRef> & loops = FB->GetLoops()->GetList();
   std::list<LoopConstRef>::const_iterator loop, loop_end = loops.end();
   for(loop = loops.begin(); loop != loop_end; loop++)
   {
      if ((*loop)->IsReducible())
      {
         if((*loop)->GetId() != 0 and (((*loop)->loop_type & WHILE_LOOP) == 0 or ((*loop)->loop_type & SINGLE_EXIT_LOOP) == 0))
         {
            bb_node_info* bi = GET_BB_NODE_INFO(fbb, (*loop)->GetHeader());
            if(!helper->start_with_a_label(bi->block))
            {
               ///create the tree node for a label expression
               helper->create_label(bi->block, helper->get_function_index());
            }
         }
      }
      else
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level,  "Found an irreducible loop");
      }
   }
   ///then gotos are added
   loop_end = loops.end();
   for(loop = loops.begin(); loop != loop_end; loop++)
   {
      if ((*loop)->IsReducible())
      {
         if((*loop)->GetId() != 0 and (((*loop)->loop_type & WHILE_LOOP) == 0 or ((*loop)->loop_type & SINGLE_EXIT_LOOP) == 0))
         {
            vertex bb_header = (*loop)->GetHeader();
            const bb_node_info* bi = CGET_BB_NODE_INFO(fbb, bb_header);
            unsigned int label_expr_nid = helper->start_with_a_label(bi->block);
            THROW_ASSERT(label_expr_nid, "gimple_label has to be inserted into the intermediate representation");
            InEdgeIterator ei, ei_end;
            const std::tr1::unordered_set<vertex> loop_blocks = (*loop)->get_blocks();
            std::tr1::unordered_set<vertex>::const_iterator bb_it = loop_blocks.begin(), bb_it_end = loop_blocks.end();
            std::tr1::unordered_set<vertex> incoming;
            for(boost::tie(ei, ei_end) = boost::in_edges(bb_header, *fbb); ei != ei_end; ei++)
            {
               incoming.insert(boost::source(*ei, *fbb));
            }
            std::tr1::unordered_set<vertex>::const_iterator it, it_end = incoming.end();
            for(it = incoming.begin(); it != it_end; it++)
            {
               vertex from_bb = *it;
               bb_node_info* bi_from = GET_BB_NODE_INFO(fbb, from_bb);
               if(std::find(bb_it, bb_it_end, from_bb) != bb_it_end)
               {
                  if(!helper->end_with_a_cond_or_goto(bi_from->block))
                  {
                     helper->create_goto(bi_from->block, helper->get_function_index(), label_expr_nid);
                  }
                  else
                  {
                     //Checking if block finishes with an if
                     const std::vector<tree_nodeRef> & list_of_stmt = bi_from->block->list_of_stmt;
                     if(list_of_stmt.size() && GET_NODE(list_of_stmt.back())->get_kind() == gimple_cond_K)
                     {
                        bool true_case;
                        unsigned int source_index = bi_from->block->number;
                        unsigned int target_index = bi->block->number;
                        if(bi_from->block->true_edge == target_index)
                           true_case = true;
                        else
                           true_case = false;
                        //We have to insert a basic block on the feedback edge
                        function_decl * fd = GetPointer<function_decl>(TM->add(funID));
                        statement_list * sl = GetPointer<statement_list>(GET_NODE(fd->body));
                        std::map<unsigned int, blocRef> & list_of_bloc = sl->list_of_bloc;
                        //Creating new block
                        blocRef new_bb = blocRef(new bloc());
                        new_bb->number = (list_of_bloc.rbegin())->second->number + 1;
                        new_bb->list_of_pred.push_back(source_index);
                        new_bb->list_of_succ.push_back(target_index);
                        list_of_bloc[new_bb->number] = new_bb;
                        //Updated target
                        for(unsigned int index = 0; index < list_of_bloc[target_index]->list_of_pred.size(); index++)
                        {
                           if(list_of_bloc[target_index]->list_of_pred[index] == source_index)
                           {
                              list_of_bloc[target_index]->list_of_pred[index] = new_bb->number;
                              break;
                           }
                        }

                        //Updated source
                        if(true_case)
                           bi_from->block->true_edge = new_bb->number;
                        else
                           bi_from->block->false_edge = new_bb->number;
                        for(unsigned int index = 0; index < list_of_bloc[source_index]->list_of_succ.size(); index++)
                        {
                           if(list_of_bloc[source_index]->list_of_succ[index] == target_index)
                           {
                              list_of_bloc[source_index]->list_of_succ[index] = new_bb->number;
                              break;
                           }
                        }

                        //Creating new vertex
                        const BasicBlocksGraphConstructorRef bbgc = FB->bbgc;
                        bbgc->add_vertex(new_bb);

                        vertex source = bbgc->Cget_vertex(source_index);
                        vertex target = bbgc->Cget_vertex(target_index);
                        vertex new_vertex = bbgc->Cget_vertex(new_bb->number);

                        //Updating basic block graph
                        bbgc->AddEdge(source, new_vertex, BB_SELECTOR);
                        bbgc->AddEdge(new_vertex, target, FB_BB_SELECTOR);
                        if(true_case)
                           bbgc->add_bb_edge_info(source, new_vertex, BB_SELECTOR, T_COND);
                        else
                           bbgc->add_bb_edge_info(source, new_vertex, BB_SELECTOR, F_COND);
                        bbgc->SetEdge(source, target, 0);

                        //Updating dominators
                        FB->dominators->dom[new_vertex] = source;
                        FB->post_dominators->dom[new_vertex] = target;
                        bbgc->AddEdge(source, new_vertex, D_BB_SELECTOR);
                        bbgc->AddEdge(target, new_vertex, PD_BB_SELECTOR);

                        //Updating loop information
                        unsigned int loop_id = GET_BB_NODE_INFO(fbb, source)->loop_id;
                        GET_BB_NODE_INFO(fbb, new_vertex)->loop_id = loop_id;
                        FB->GetLoops()->GetModifiableLoop(loop_id)->add_block(new_vertex);
                        added_block = true;
                        helper->create_goto(new_bb, helper->get_function_index(), label_expr_nid);
                     }
                  }
               }
            }
         }
      }
      else
      {
         if(loop_irriducible_breaks)
         {
            THROW_ERROR_CODE(IRREDUCIBLE_LOOPS_EC,"Irreducible loops not yet supported");
            THROW_ERROR("Irreducible loops not yet supported");
         }
         else
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level,  "Found an irreducible loop");
      }
   }
   if(added_block && (output_level >= OUTPUT_LEVEL_PEDANTIC || Param->getOption<bool>(OPT_print_dot)))
   {
      behavioral_helperRef null_helper;
      FB->get_bb_graph(function_behavior::BB)->PrintToDot(Param, (helper->get_function_name() + "_" + boost::lexical_cast<std::string>(helper->get_function_index()) + "_BB_CFG.dot").c_str(),helper);
      FB->get_bb_graph(function_behavior::FBB)->PrintToDot(Param, (helper->get_function_name() + "_" + boost::lexical_cast<std::string>(helper->get_function_index()) + "_BB_FCFG.dot").c_str(),helper);
      FB->get_bb_graph(function_behavior::DOM_TREE)->PrintToDot(Param, (helper->get_function_name() + "_" + boost::lexical_cast<std::string>(helper->get_function_index()) + "_BB_dom_tree.dot").c_str(),null_helper);
      FB->get_bb_graph(function_behavior::POST_DOM_TREE)->PrintToDot(Param, (helper->get_function_name() + "_" + boost::lexical_cast<std::string>(helper->get_function_index()) + "_BB_post_dom_tree.dot").c_str(),null_helper);

   }
   if(debug_level >= DEBUG_LEVEL_PEDANTIC)
   {
      std::string raw_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "after_loop_regions_computation.raw";
      std::ofstream raw_file(raw_file_name.c_str());
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Tree-Manager dumped for debug purpose");
      raw_file << TM;
      raw_file.close();
   }

}
