/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file operations_cfg_computation.cpp
 * @brief Analysis step creating the control flow graph of the operations
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/
///Header include
#include "operations_cfg_computation.hpp"

///Behavior include
#include "basic_block.hpp"
#include "basic_blocks_graph_constructor.hpp"
#include "behavioral_helper.hpp"
#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "op_node_info.hpp"
#include "operations_graph_constructor.hpp"

///Graph include
#include "graph.hpp"

///Parameter include
#include "Parameter.hpp"

///Tree include
#include "ext_tree_node.hpp"
#include "tree_basic_block.hpp"
#include "tree_helper.hpp"
#include "tree_manager.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"

operations_cfg_computation::operations_cfg_computation(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
   BaseAnalysis(_Param, _AppM, CFG_CREATION, function_id)
{
   debug_level = Param->get_class_debug_level("operations_cfg_computation", DEBUG_LEVEL_NONE);
   SetPreCond(ID, BB_FEEDBACK_EDGES_IDENTIFICATION);
}


operations_cfg_computation::~operations_cfg_computation()
{

}

void operations_cfg_computation::exec()
{
   const tree_managerRef TM = AppM->get_tree_manager();
   const function_behaviorRef FB = AppM->get_function_behavior(funID);
   const bb_graph * fbb = FB->get_bb_graph(function_behavior::FBB);
   const behavioral_helperRef & helper = FB->get_helper();
   operations_graph_constructorRef & ogc = FB->ogc;
   const BasicBlocksGraphConstructorRef bbgc = FB->bbgc;

   VertexIterator v_iter, v_iter_end;
   std::string res;
   std::string f_name = helper->get_function_name() + "_" + boost::lexical_cast<std::string>(funID);

   /// entry and exit computation
   clean_start_nodes();
   ogc->add_operation(ENTRY, ENTRY, BB_ENTRY);
   bbgc->add_operation_to_bb(ogc->getIndex(ENTRY), BB_ENTRY);
   ogc->add_type(ENTRY, TYPE_ENTRY);

   ogc->add_operation(EXIT, EXIT, BB_EXIT);
   bbgc->add_operation_to_bb(ogc->getIndex(EXIT), BB_EXIT);
   ogc->add_type(EXIT, TYPE_EXIT);

   /// first cycle to compute the label_decl_map and the first_statement maps.
   for(boost::tie(v_iter, v_iter_end) = boost::vertices(*fbb); v_iter != v_iter_end; ++v_iter)
   {
      if(/**v_iter == fbb->GetConstBBGraphInfo()->entry_vertex || */*v_iter == fbb->GetConstBBGraphInfo()->exit_vertex)
         continue;
      const bb_node_info* bi = CGET_BB_NODE_INFO(fbb, *v_iter);
      const blocRef & b = bi->block;
      if(!b->list_of_stmt.empty())
      {
         tree_nodeRef front = b->list_of_stmt.front();
         if(GET_NODE(front)->get_kind() == gimple_label_K)
         {
            gimple_label* le = GetPointer<gimple_label>(GET_NODE(front));
            THROW_ASSERT(le->op, "Label expr without object");
            res = get_first_node(front, f_name);
            label_decl_map[GET_INDEX_NODE(le->op)] = res;
         }
         else if(b->list_of_phi.empty())
            res = get_first_node(front, f_name);
         else
            res = get_first_node(*b->list_of_phi.begin(), f_name);
      }
      else if(!b->list_of_phi.empty())
         res = get_first_node(*b->list_of_phi.begin(), f_name);
      else if(*v_iter == fbb->GetConstBBGraphInfo()->entry_vertex)
      {
         res = ENTRY;
         first_statement[b->number] = res;
         continue;
      }
      else/*empty basic block*/
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Empty basic block");
         res = f_name + "_empty_" + boost::lexical_cast<std::string>(b->number);
         ogc->add_operation(res, NOP, b->number);
         bbgc->add_operation_to_bb(ogc->getIndex(res), b->number);
         ogc->add_type(res, TYPE_NOP);
      }
      first_statement[b->number] = res;
   }
   /// second cycle doing the real job
   tree_nodeRef last_instruction;
   for(boost::tie(v_iter, v_iter_end) = boost::vertices(*fbb); v_iter != v_iter_end; ++v_iter)
   {
      if(/* *v_iter == fbb->GetConstBBGraphInfo()->entry_vertex || */*v_iter == fbb->GetConstBBGraphInfo()->exit_vertex)
         continue;
      const bb_node_info* bi = CGET_BB_NODE_INFO(fbb, *v_iter);
      const blocRef & b = bi->block;
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Building operation of basic block BB" + boost::lexical_cast<std::string>(b->number));
      bool skip_first_stmt = false;
      if(!b->list_of_stmt.empty())
      {
         tree_nodeRef front = b->list_of_stmt.front();
         if(GET_NODE(front)->get_kind() == gimple_label_K)
         {
            actual_name = get_first_node(front, f_name);
            THROW_ASSERT(actual_name == first_statement[b->number], "the name of the first vertice has to be the same of the label expression vertex");
            if(!empty_start_nodes())
               connect_start_nodes(ogc, actual_name);
            init_start_nodes(actual_name);
            build_operation_recursive(TM, ogc, front, f_name, b->number);
            ogc->add_nodeID(actual_name, GET_INDEX_NODE(front));
            bbgc->add_operation_to_bb(ogc->getIndex(actual_name), b->number);
            skip_first_stmt = true;
         }
      }
      std::vector<tree_nodeRef>::iterator p_end = b->list_of_phi.end();
      for(std::vector<tree_nodeRef>::iterator p = b->list_of_phi.begin(); p != p_end; p++)
      {
         actual_name = get_first_node(*p, f_name);
         if(!empty_start_nodes())
            connect_start_nodes(ogc, actual_name);
         init_start_nodes(actual_name);
         build_operation_recursive(TM, ogc, *p, f_name, b->number);
         ogc->add_nodeID(actual_name, GET_INDEX_NODE(*p));
         bbgc->add_operation_to_bb(ogc->getIndex(actual_name), b->number);
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "List of operations size " + boost::lexical_cast<std::string>(b->list_of_stmt.size()));
      std::vector<tree_nodeRef>::iterator s_end = b->list_of_stmt.end();
      std::vector<tree_nodeRef>::iterator s = b->list_of_stmt.begin();
      if(skip_first_stmt)
         s++;
      for(; s != s_end; s++)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Analyzing operation " + boost::lexical_cast<std::string>(GET_INDEX_NODE(*s)));
         actual_name = get_first_node(*s, f_name);
         if(!empty_start_nodes())
            connect_start_nodes(ogc, actual_name);
         init_start_nodes(actual_name);
         build_operation_recursive(TM, ogc, *s, f_name, b->number);
         ogc->add_nodeID(actual_name, GET_INDEX_NODE(*s));
         bbgc->add_operation_to_bb(ogc->getIndex(actual_name), b->number);
      }
      clean_start_nodes();
      if(b->list_of_stmt.empty() )
      {
         if(!b->list_of_phi.empty())
            last_instruction = b->list_of_phi.back();
         else
            last_instruction = tree_nodeRef();
      }
      else if(b->list_of_stmt.size() == 1 && skip_first_stmt)
      {
         if(!b->list_of_phi.empty())
            last_instruction = b->list_of_phi.back();
         else
            last_instruction = b->list_of_stmt.back();
      }
      else
      {
         last_instruction = b->list_of_stmt.back();
      }
      if(last_instruction && GET_NODE(last_instruction)->get_kind() != gimple_switch_K)
      {
         init_start_nodes(get_first_node(last_instruction, f_name));
      }
      else if(!last_instruction)
      {
         init_start_nodes(first_statement[b->number]);
      }
      if(b->list_of_succ.size() == 0 and funID == AppM->get_main_function())
      {
         std::list<std::string>::iterator operation, operation_end = start_nodes.end();
         for(operation = start_nodes.begin(); operation != operation_end; operation++)
         {
            ogc->add_type(*operation, TYPE_LAST_OP);
         }
      }
      else
      {
         std::vector<unsigned int>::iterator su_end = b->list_of_succ.end();
         for(std::vector<unsigned int>::iterator su = b->list_of_succ.begin(); su != su_end; su++)
         {
            if((*su)== bloc::EXIT_BLOCK_ID)
            {
               connect_start_nodes(ogc, EXIT);
               if(funID == AppM->get_main_function())
               {
                  std::list<std::string>::iterator operation, operation_end = start_nodes.end();
                  for(operation = start_nodes.begin(); operation != operation_end; operation++)
                  {
                     ogc->add_type(*operation, TYPE_LAST_OP);
                  }
               }
            }
            else
            {
               if(*su == b->true_edge)
                  connect_start_nodes(ogc, first_statement[*su], true, false);
               else if(*su == b->false_edge)
                  connect_start_nodes(ogc, first_statement[*su], false, true);
               else if(last_instruction && GetPointer<gimple_goto>(GET_NODE(last_instruction)) && b->list_of_succ.size() > 1)
                  connect_start_nodes(ogc, first_statement[*su], true, true, *su);
               else
                  connect_start_nodes(ogc, first_statement[*su]);
            }
         }
      }
      clean_start_nodes();
   }
   label_decl_map.clear();
   first_statement.clear();

   if(output_level >= OUTPUT_LEVEL_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
   {
      std::string file_name = FB->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_CFG_Cleaned.dot";
      FB->get_graph(function_behavior::CFG)->print_to_dot(Param, file_name, helper);
   }
}


std::string operations_cfg_computation::get_first_node(const tree_nodeRef &tn, const std::string & f_name) const
{
   tree_nodeRef curr_tn;
   if (tn->get_kind() == tree_reindex_K)
   {
      curr_tn = GET_NODE(tn);
   }
   else
   {
      curr_tn = tn;
   }
   unsigned int ind = GET_INDEX_NODE(tn);
   std::string src, res = "";
   src = f_name + "_" + boost::lexical_cast<std::string>(ind);

   switch (curr_tn->get_kind())
   {
      case gimple_return_K:
      case gimple_assign_K:
      case gimple_nop_K:
      case gimple_call_K:
      case gimple_switch_K:
      case gimple_cond_K:
      case while_expr_K:
      case for_expr_K:
      case gimple_resx_K:
      case gimple_phi_K:
      case gimple_asm_K:
      case gimple_goto_K:
      case gimple_label_K:
      case pragma_node_K:
      case gimple_predict_K:
         return src;
      case case_label_expr_K:
      {
         case_label_expr* cle = GetPointer<case_label_expr>(curr_tn);
         ind = GET_INDEX_NODE(cle->got);
         THROW_ASSERT(label_decl_map.find(ind) != label_decl_map.end(), "Label " + boost::lexical_cast<std::string>(ind) + " doesn't exist");
         return label_decl_map.find(ind)->second;
      }
      default:
         THROW_ERROR(std::string("Node not supported (") + boost::lexical_cast<std::string>(ind) + std::string("): ") + curr_tn->get_kind_text());
   };
   return "";
}


void operations_cfg_computation::insert_start_node(const std::string &start_node)
{
   start_nodes.push_back(start_node);
}

void operations_cfg_computation::clean_start_nodes()
{
   start_nodes.clear();
}

bool operations_cfg_computation::empty_start_nodes() const
{
   return start_nodes.empty();
}

void operations_cfg_computation::init_start_nodes(const std::string &start_node)
{
   start_nodes.clear();
   start_nodes.push_back(start_node);
}

void operations_cfg_computation::connect_start_nodes(operations_graph_constructorRef & ogc, const std::string & next, bool true_edge, bool false_edge, unsigned int nodeid)
{
   std::string Start_node;
   std::list<std::string>::iterator s_end = start_nodes.end();
   for(std::list<std::string>::iterator s = start_nodes.begin(); s != s_end; s++)
   {
      Start_node = *s;
      ///Mark first operation of the application
      if(funID == AppM->get_main_function() and Start_node == ENTRY)
         ogc->add_type(*s, TYPE_FIRST_OP);
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Adding edge from " + Start_node + " to " + next);
      ogc->add_edge(ogc->getIndex(Start_node), ogc->getIndex(next), CFG_SELECTOR);
      if(true_edge && false_edge)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Adding label " + boost::lexical_cast<std::string>(nodeid));
         ogc->add_edge_info(ogc->getIndex(Start_node), ogc->getIndex(next), CFG_SELECTOR, nodeid);
      }
      if(true_edge && !false_edge)
         ogc->add_edge_info(ogc->getIndex(Start_node), ogc->getIndex(next), CFG_SELECTOR, T_COND);
      if(false_edge && !true_edge)
         ogc->add_edge_info(ogc->getIndex(Start_node), ogc->getIndex(next), CFG_SELECTOR, F_COND);
   }
}

void operations_cfg_computation::build_operation_recursive(const tree_managerRef TM, operations_graph_constructorRef & ogc, const tree_nodeRef &tn, const std::string &f_name, unsigned int bb_index)
{
   const tree_nodeRef &curr_tn = GET_NODE(tn);
   unsigned int ind = GET_INDEX_NODE(tn);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, f_name + " Building CFG of node " + boost::lexical_cast<std::string>(ind) + " of type " + curr_tn->get_kind_text());
   switch (curr_tn->get_kind())
   {
      case gimple_return_K:
      {
         ogc->add_operation(actual_name, curr_tn->get_kind_text(), bb_index);
         ogc->add_type(actual_name, TYPE_RET);
         break;
      }
      case gimple_assign_K:
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Building CFG of node " + boost::lexical_cast<std::string>(ind) + " - gimple_assign");
         gimple_assign* me = GetPointer<gimple_assign>(curr_tn);

         tree_nodeRef op0 = GET_NODE(me->op0);
         tree_nodeRef op1 = GET_NODE(me->op1);
         tree_nodeRef op0_type = tree_helper::get_type_node(op0);
         tree_nodeRef op1_type = tree_helper::get_type_node(op1);
         const std::set<unsigned int>& fun_mem_data = AppM->get_function_behavior(funID)->get_function_mem();
         bool load_candidate = op1->get_kind() == bit_field_ref_K ||op1->get_kind() == component_ref_K || op1->get_kind() == indirect_ref_K || op1->get_kind() == mem_ref_K || op1->get_kind() == array_ref_K || op1->get_kind() == target_mem_ref_K || op1->get_kind() == target_mem_ref461_K || op1->get_kind() == realpart_expr_K || op1->get_kind() == imagpart_expr_K;
         bool store_candidate = op0->get_kind() == bit_field_ref_K || op0->get_kind() == component_ref_K || op0->get_kind() == indirect_ref_K || op0->get_kind() == mem_ref_K || op0->get_kind() == array_ref_K || op0->get_kind() == target_mem_ref_K || op0->get_kind() == target_mem_ref461_K || op0->get_kind() == realpart_expr_K || op0->get_kind() == imagpart_expr_K;

         if(me->init_assignment)
         {
            ogc->add_operation(actual_name, NOP, bb_index);
            ogc->add_type(actual_name, TYPE_NOP);
         }
         else if(op1->get_kind() == complex_expr_K)
         {
            ogc->add_operation(actual_name, op1->get_kind_text(), bb_index);
            ogc->add_type(actual_name, TYPE_GENERIC);
         }
         else if(op0_type && op1_type &&
                ((op0_type->get_kind()== record_type_K && op1_type->get_kind()== record_type_K) ||
                (op0_type->get_kind()== union_type_K && op1_type->get_kind()== union_type_K) ||
                (op0_type->get_kind() == array_type_K) ||
                (fun_mem_data.find(GET_INDEX_NODE(me->op0)) != fun_mem_data.end() && fun_mem_data.find(GET_INDEX_NODE(me->op1)) != fun_mem_data.end()) ||
                (fun_mem_data.find(GET_INDEX_NODE(me->op0)) != fun_mem_data.end() && load_candidate) ||
                (store_candidate && fun_mem_data.find(GET_INDEX_NODE(me->op1)) != fun_mem_data.end())
                )
               )
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, " - set as MEMCPY operation");
            ogc->add_operation(actual_name, MEMCPY, bb_index);
            ogc->add_type(actual_name, TYPE_MEMCPY);
         }
         else if (store_candidate || fun_mem_data.find(GET_INDEX_NODE(me->op0)) != fun_mem_data.end())
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, " - set as STORE operation");
            ogc->add_operation(actual_name, STORE, bb_index);
            ogc->add_type(actual_name, TYPE_STORE);
         }
         else if (load_candidate || fun_mem_data.find(GET_INDEX_NODE(me->op1)) != fun_mem_data.end())
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, " - set as LOAD operation");
            ogc->add_operation(actual_name, LOAD, bb_index);
            ogc->add_type(actual_name, TYPE_LOAD);
         }
         else
         {
            ogc->add_operation(actual_name, ASSIGN, bb_index);
            ogc->add_type(actual_name, TYPE_ASSIGN);
            if(me->orig)
               ogc->add_type(actual_name, TYPE_WAS_GIMPLE_PHI);
            build_operation_recursive(TM, ogc, me->op1, f_name, bb_index);
         }
         break;
      }
      case gimple_nop_K:
      {
         ogc->add_operation(actual_name, curr_tn->get_kind_text(), bb_index);
         ogc->add_type(actual_name, TYPE_GENERIC);
         break;
      }
      case call_expr_K:
      {
         call_expr* ce = GetPointer<call_expr>(curr_tn);
         tree_nodeRef temp_node = GET_NODE(ce->fn);
         function_decl * fd=0;

         if(temp_node->get_kind() == addr_expr_K)
         {
            unary_expr* ue = GetPointer<unary_expr>(temp_node);
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Building CFG of node " + boost::lexical_cast<std::string>(ind) + " - call_expr + addr_expr");
            temp_node = ue->op;
            fd = GetPointer<function_decl>(GET_NODE(temp_node));
         }
         else if(temp_node->get_kind() == obj_type_ref_K)
         {
            temp_node = tree_helper::find_obj_type_ref_function(ce->fn);
            fd = GetPointer<function_decl>(GET_NODE(temp_node));
         }
         if(fd)
         {
            std::string fun_name = fd->get_function_name();
            fun_name = tree_helper::normalized_ID(fun_name);
            const std::string builtin_prefix("__builtin_");
            if(fun_name.find(builtin_prefix) == 0)
               fun_name = fun_name.substr(builtin_prefix.size());
            const unsigned int call_id = GET_INDEX_NODE(temp_node);
            //Creating node of call
            ogc->add_operation(actual_name, fun_name, bb_index);
            ogc->add_type(actual_name, TYPE_EXTERNAL);
            if(fun_name == "exit" || fun_name == "abort")
               ogc->add_type(actual_name, TYPE_LAST_OP);
            ogc->add_called_function(actual_name, call_id);
         }
         else
         {
            //Call of not an usual function decl (for example array_ref of an array of function pointer)
            //This call is needed to set the basic block index of this operation. Otherwise
            //vertex is created but basic block index is not setted
            ogc->add_operation(actual_name, curr_tn->get_kind_text(), bb_index);
         }
         break;
      }
      case gimple_call_K:
      {
         gimple_call* ce = GetPointer<gimple_call>(curr_tn);
         tree_nodeRef temp_node = GET_NODE(ce->fn);
         function_decl * fd=0;

         if(temp_node->get_kind() == addr_expr_K)
         {
            unary_expr* ue = GetPointer<unary_expr>(temp_node);
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Building CFG of node " + boost::lexical_cast<std::string>(ind) + " - gimple_call + addr_expr");
            temp_node = ue->op;
            fd = GetPointer<function_decl>(GET_NODE(temp_node));
         }
         else if(temp_node->get_kind() == obj_type_ref_K)
         {
            temp_node = tree_helper::find_obj_type_ref_function(ce->fn);
            fd = GetPointer<function_decl>(GET_NODE(temp_node));
         }
         if(fd)
         {
            std::string fun_name = fd->get_function_name();
            fun_name = tree_helper::normalized_ID(fun_name);
            const std::string builtin_prefix("__builtin_");
            if(fun_name.find(builtin_prefix) == 0)
               fun_name = fun_name.substr(builtin_prefix.size());
            const unsigned int call_id = GET_INDEX_NODE(temp_node);
            //Creating node of call
            ogc->add_operation(actual_name, fun_name, bb_index);
            ogc->add_type(actual_name, TYPE_EXTERNAL);
            if(fun_name == "exit" || fun_name == "abort")
               ogc->add_type(actual_name, TYPE_LAST_OP);
            ogc->add_called_function(actual_name, call_id);
         }
         else
         {
            //Call of not an usual function decl (for example array_ref of an array of function pointer)
            //This call is needed to set the basic block index of this operation. Otherwise
            //vertex is created but basic block index is not setted
            ogc->add_operation(actual_name, curr_tn->get_kind_text(), bb_index);
         }
         break;
      }
      case gimple_goto_K:
      {
         ogc->add_operation(actual_name, curr_tn->get_kind_text(), bb_index);
         ogc->add_type(actual_name, TYPE_GOTO);
         break;
      }
      case gimple_label_K:
      {
         ogc->add_operation(actual_name, curr_tn->get_kind_text(), bb_index);
         gimple_label *le = GetPointer<gimple_label>(curr_tn);
         THROW_ASSERT(le->op, "expected a gimple_label operand");
         label_decl * ld = GetPointer<label_decl>(GET_NODE(le->op));
         if(ld->artificial_flag)
            ogc->add_type(actual_name, TYPE_NOP);
         else
            ogc->add_type(actual_name, TYPE_LABEL);
         break;
      }
      case gimple_cond_K:
      {
         gimple_cond* gc = GetPointer<gimple_cond>(curr_tn);
         ogc->add_operation(actual_name, READ_COND, bb_index);
         build_operation_recursive(TM, ogc, gc->op0, f_name, bb_index);
         ogc->add_type(actual_name, TYPE_IF);
         break;
      }
      case while_expr_K:
      {
         while_expr* we = GetPointer<while_expr>(curr_tn);
         ogc->add_operation(actual_name, READ_COND, bb_index);
         build_operation_recursive(TM, ogc, we->op0, f_name, bb_index);
         ogc->add_type(actual_name, TYPE_WHILE);
         break;
      }
      case for_expr_K:
      {
         ogc->add_operation(actual_name, READ_COND, bb_index);
         ogc->add_type(actual_name, TYPE_FOR);
         break;
      }
      case gimple_switch_K:
      {
         gimple_switch* se = GetPointer<gimple_switch>(curr_tn);

         ogc->add_operation(actual_name, SWITCH_COND, bb_index);
         build_operation_recursive(TM, ogc, se->op0, actual_name, bb_index);
         ogc->add_type(actual_name, TYPE_SWITCH);

         tree_nodeRef case_label_exprs;
         if(se->op1)
            case_label_exprs = GET_NODE(se->op1);
         else
            THROW_ERROR(std::string("op2 in gimple_switch not yet supported (") + boost::lexical_cast<std::string>(ind) + std::string(")"));
         if(case_label_exprs->get_kind() == tree_vec_K)
         {
            tree_vec* tv = GetPointer<tree_vec>(case_label_exprs);
            std::vector<tree_nodeRef>::iterator end = tv->list_of_op.end();
            for(std::vector<tree_nodeRef>::iterator i = tv->list_of_op.begin(); i != end; i++)
            {
               std::string res = get_first_node(*i, f_name);
               THROW_ASSERT(res != "", "Impossible to find first operation of case " + boost::lexical_cast<std::string>(GET_INDEX_NODE(*i)));
               if(GET_NODE(*i)->get_kind() == case_label_expr_K)
               {
                  case_label_expr * cl = GetPointer<case_label_expr>(GET_NODE(*i));
                  if(cl->default_flag)
                     connect_start_nodes(ogc, res, true, true, default_COND);
                  else
                     connect_start_nodes(ogc, res, true, true, GET_INDEX_NODE(*i));
               }
               else
                  connect_start_nodes(ogc, res, true, true, GET_INDEX_NODE(*i));
            }
         }
         else
            THROW_ERROR(std::string("expected tree_vec in op1 in gimple_switch (") + boost::lexical_cast<std::string>(ind) + std::string(")"));
         break;
      }
      case abs_expr_K:
      case addr_expr_K:
      case arrow_expr_K:
      case bit_not_expr_K:
      case buffer_ref_K:
      case card_expr_K:
      case cast_expr_K:
      case cleanup_point_expr_K:
      case conj_expr_K:
      case convert_expr_K:
      case exit_expr_K:
      case fix_ceil_expr_K:
      case fix_floor_expr_K:
      case fix_round_expr_K:
      case fix_trunc_expr_K:
      case float_expr_K:
      case imagpart_expr_K:
      case indirect_ref_K:
      case loop_expr_K:
      case negate_expr_K:
      case non_lvalue_expr_K:
      case nop_expr_K:
      case realpart_expr_K:
      case reference_expr_K:
      case reinterpret_cast_expr_K:
      case gimple_resx_K:
      case sizeof_expr_K:
      case static_cast_expr_K:
      case throw_expr_K:
      case truth_not_expr_K:
      case unsave_expr_K:
      case va_arg_expr_K:
      case view_convert_expr_K:
      case gimple_predict_K:
      {
         const unary_expr * ue = GetPointer<unary_expr>(curr_tn);
         build_operation_recursive(TM, ogc, ue->op, f_name, bb_index);
         ogc->add_operation(actual_name, curr_tn->get_kind_text(), bb_index);
         ogc->add_type(actual_name, TYPE_GENERIC);
         break;
      }
      case CASE_BINARY_EXPRESSION:
      {
         const binary_expr * be = GetPointer<binary_expr>(curr_tn);
         build_operation_recursive(TM, ogc, be->op0, f_name, bb_index);
         build_operation_recursive(TM, ogc, be->op1, f_name, bb_index);
         ogc->add_operation(actual_name, curr_tn->get_kind_text(), bb_index);
         ogc->add_type(actual_name, TYPE_GENERIC);
         break;
      }
      case CASE_TERNARY_EXPRESSION:
      {
         const ternary_expr * te = GetPointer<ternary_expr>(curr_tn);
         build_operation_recursive(TM, ogc, te->op0, f_name, bb_index);
         if(te->op1)
            build_operation_recursive(TM, ogc, te->op1, f_name, bb_index);
         if(te->op2)
            build_operation_recursive(TM, ogc, te->op2, f_name, bb_index);
         ogc->add_operation(actual_name, curr_tn->get_kind_text(), bb_index);
         ogc->add_type(actual_name, TYPE_GENERIC);
         break;
      }
      case CASE_QUATERNARY_EXPRESSION:
      {
         const quaternary_expr * qe = GetPointer<quaternary_expr>(curr_tn);
         build_operation_recursive(TM, ogc, qe->op0, f_name, bb_index);
         if(qe->op1)
            build_operation_recursive(TM, ogc, qe->op1, f_name, bb_index);
         if(qe->op2)
            build_operation_recursive(TM, ogc, qe->op2, f_name, bb_index);
         if(qe->op3)
            build_operation_recursive(TM, ogc, qe->op3, f_name, bb_index);
         ogc->add_operation(actual_name, curr_tn->get_kind_text(), bb_index);
         ogc->add_type(actual_name, TYPE_GENERIC);
         break;
      }
      case result_decl_K:
      case parm_decl_K:
      case var_decl_K:
      case ssa_name_K:
      case integer_cst_K:
      case real_cst_K:
      case string_cst_K:
      case vector_cst_K:
      case complex_cst_K:
      case field_decl_K:
      case exc_ptr_expr_K:
      case filter_expr_K:
      case label_decl_K:
      case constructor_K:
      case target_mem_ref_K:
      case target_mem_ref461_K:
      case pragma_node_K:
      {
         break;
      }
      case gimple_asm_K:
      {
         gimple_asm* ae = GetPointer<gimple_asm>(curr_tn);
         if(!ae->volatile_flag && (ae->in && ae->out))
            ogc->add_type(actual_name, TYPE_GENERIC);
         else
            ogc->add_type(actual_name, TYPE_GENERIC|TYPE_OPAQUE);
         ogc->add_operation(actual_name, curr_tn->get_kind_text(), bb_index);
         break;
      }
      case gimple_phi_K:
      {
         gimple_phi * phi = GetPointer<gimple_phi>(curr_tn);
         ogc->add_operation(actual_name, curr_tn->get_kind_text(), bb_index);
         ogc->add_type(actual_name, phi->virtual_flag ? TYPE_VPHI : TYPE_PHI);
         break;
      }
      default:
      {
         ogc->add_operation(actual_name, curr_tn->get_kind_text(), bb_index);
         ogc->add_type(actual_name, TYPE_GENERIC);
         break;
      }
   };
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, f_name + " Finished building CFG of node " + boost::lexical_cast<std::string>(ind));
}

