/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file var_computation.cpp
 * @brief Analyzes operations and creates the sets of read and written variables.
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Header include
#include "var_computation.hpp"

///Behavior include
#include "behavioral_helper.hpp"
#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "graph.hpp"
#include "op_node_info.hpp"
#include "operations_graph_constructor.hpp"

///Parameter include
#include "Parameter.hpp"

///STD include
#include <fstream>

///Tree include
#include "ext_tree_node.hpp"
#include "tree_helper.hpp"
#include "tree_manager.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"

///Utility include
#include "boost/lexical_cast.hpp"
#include "dbgPrintHelper.hpp"
#include "exceptions.hpp"



#define TOSTRING(id) boost::lexical_cast<std::string>(id)

var_computation::var_computation(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
   BaseAnalysis(_Param, _AppM, VAR_ANALYSIS, function_id), executed(false), isRef(true)
{
   this->SetPreCond(ID, CFG_CREATION);
}


var_computation::~var_computation()
{
}

void var_computation::exec()
{
   if(executed)
      THROW_ERROR("Variable computation executed twice");
   executed = true;
   const graph* cfg = AppM->get_function_behavior(funID)->get_graph(function_behavior::CFG);
   VertexIterator VerIt, VerItEnd;
   std::list<vertex> Vertices;
   for(tie(VerIt, VerItEnd) = boost::vertices(*cfg); VerIt != VerItEnd; VerIt++)
   {
      Vertices.push_back(*VerIt);
   }
   std::list<vertex> PhiNodes;
   for(std::list<vertex>::iterator Ver = Vertices.begin(); Ver != Vertices.end(); )
   {
      std::list<vertex>::iterator curr_Ver = Ver;
      ++Ver;
      if (GET_TYPE(cfg, *curr_Ver) == TYPE_VPHI)
      {
         PhiNodes.push_back(*curr_Ver);
         Vertices.erase(curr_Ver);
      }
   }
   for(std::list<vertex>::iterator Ver = Vertices.begin(); Ver != Vertices.end(); Ver++)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Analyzing operation vertex: " + GET_NAME(cfg, *Ver));
      std::list<unsigned int> NodeID = GET_NODEID(cfg, *Ver);
      for(std::list<unsigned int>::iterator It = NodeID.begin(); It != NodeID.end(); It++)
      {
         AnalyzeNodeID(*Ver, *It);
      }
   }
   for(std::list<vertex>::iterator Ver = PhiNodes.begin(); Ver != PhiNodes.end(); Ver++)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Analyzing phi-node Vertex: " + GET_NAME(cfg, *Ver));
      std::list<unsigned int> NodeID = GET_NODEID(cfg, *Ver);
      for(std::list<unsigned int>::iterator It = NodeID.begin(); It != NodeID.end(); It++)
      {
         AnalyzeNodeID(*Ver, *It);
      }
   }
   if(debug_level>=DEBUG_LEVEL_VERY_PEDANTIC)
      report();
   ref_vars.clear();
   vops.clear();
}

void var_computation::AnalyzeNodeID(const vertex &Ver, unsigned int NodeID)
{
   const tree_managerRef TM = AppM->get_tree_manager();
   const tree_nodeRef tn = TM->get_tree_node_const(NodeID);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node " + boost::lexical_cast<std::string>(NodeID) + " of type: " + tn->get_kind_text());

   switch(tn->get_kind())
   {
      case pragma_node_K:
      case gimple_predict_K:
         break;
      case gimple_assign_K:
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing left side of statement: " + TOSTRING(NodeID));
         tree_nodeRef lop = GetPointer<gimple_assign>(tn)->op0;
         Analyze_lvalue(Ver, GET_INDEX_NODE(lop));
         tree_nodeRef rside = GET_NODE(GetPointer<gimple_assign>(tn)->op1);
         if(
               !tree_helper::is_a_vector(TM, GET_INDEX_NODE(GetPointer<gimple_assign>(tn)->op0)) &&
               (
                  (
                     tree_helper::is_an_array(TM, GET_INDEX_NODE(GetPointer<gimple_assign>(tn)->op0)) &&
                     !tree_helper::is_a_pointer(TM, GET_INDEX_NODE(GetPointer<gimple_assign>(tn)->op0))
                  ) ||
                  (rside->get_kind() == constructor_K)
               )
            )
         {
            if(
                  !tree_helper::is_an_array(TM, GET_INDEX_NODE(GetPointer<gimple_assign>(tn)->op0)) ||
                  (
                     (rside->get_kind() == constructor_K || (rside->get_kind() == var_decl_K && GetPointer<var_decl>(rside)->init) || rside->get_kind() == string_cst_K) &&
                     (GetPointer<decl_node>(GET_NODE(GetPointer<gimple_assign>(tn)->op0)) || GetPointer<ssa_name>(GET_NODE(GetPointer<gimple_assign>(tn)->op0)))
                  )
              )
            {
               AppM->get_function_behavior(funID)->get_helper()->add_initialization(GET_INDEX_NODE(GetPointer<gimple_assign>(tn)->op0), GET_INDEX_NODE(GetPointer<gimple_assign>(tn)->op1));
               operations_graph_constructorRef ogc = AppM->get_function_behavior(funID)->ogc;
               std::string name = GET_NAME(AppM->get_function_behavior(funID)->get_graph(function_behavior::CFG), Ver);
               ogc->add_type(name, TYPE_INIT);
            }
         }

         if (GetPointer<target_mem_ref>(GET_NODE(GetPointer<gimple_assign>(tn)->op0)))
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing left side of statement: " + TOSTRING(NodeID));
            lop = GetPointer<gimple_assign>(tn)->op0;
            Analyze_rvalue(Ver, GET_INDEX_NODE(lop));
         }
         if (GetPointer<target_mem_ref461>(GET_NODE(GetPointer<gimple_assign>(tn)->op0)))
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing left side of statement: " + TOSTRING(NodeID));
            lop = GetPointer<gimple_assign>(tn)->op0;
            Analyze_rvalue(Ver, GET_INDEX_NODE(lop));
         }

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing right side of statement: " + TOSTRING(NodeID));
         tree_nodeRef rop = GetPointer<gimple_assign>(tn)->op1;
         Analyze_rvalue(Ver, GET_INDEX_NODE(rop));

         if (GetPointer<gimple_assign>(tn)->virtual_operands)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing vops of statement: " + TOSTRING(NodeID));
            add_vops(Ver, GetPointer<gimple_assign>(tn)->virtual_operands);
         }
         break;
      }
      case gimple_nop_K:
      {
         break;
      }
      case gimple_phi_K:
      {
         gimple_phi * pn = GetPointer<gimple_phi>(tn);
         bool is_virtual = pn->virtual_flag;

         std::vector<std::pair< tree_nodeRef, unsigned int> >::iterator it_end = pn->list_of_def_edge.end();
         for (std::vector<std::pair< tree_nodeRef, unsigned int> >::iterator it = pn->list_of_def_edge.begin(); it != it_end; it++)
         {
            if (is_virtual && GetPointer<ssa_name>(GET_NODE(it->first)))
            {
               ssa_name* sa = GetPointer<ssa_name>(GET_NODE(it->first));
               vops.insert(GET_INDEX_NODE(sa->var));
            }
            if(is_virtual) vops.insert(GET_INDEX_NODE(it->first));
            Analyze_rvalue(Ver, GET_INDEX_NODE(it->first));
         }
         if (is_virtual && GetPointer<ssa_name>(GET_NODE(pn->res)))
         {
            ssa_name* sa = GetPointer<ssa_name>(GET_NODE(pn->res));
            vops.insert(GET_INDEX_NODE(sa->var));
         }
         if(is_virtual) vops.insert(GET_INDEX_NODE(pn->res));
         Analyze_lvalue(Ver, GET_INDEX_NODE(pn->res));

         break;
      }
      case gimple_return_K:
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing right side of statement: " + TOSTRING(NodeID));
         tree_nodeRef op = GetPointer<gimple_return>(tn)->op;
         if (op)
         {
            Analyze_rvalue(Ver, GET_INDEX_NODE(op));
         }

         if (GetPointer<gimple_return>(tn)->virtual_operands)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing vops of statement: " + TOSTRING(NodeID));
            add_vops(Ver, GetPointer<gimple_return>(tn)->virtual_operands);
         }
#if !RELEASE
         const graph *cfg = AppM->get_function_behavior(funID)->get_graph(function_behavior::CFG);
         THROW_ASSERT(GET_TYPE(cfg,Ver) & TYPE_RET, "Wrong type for return statement");
#endif
         break;
      }

      case gimple_call_K:
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing gimple_call statement: " + TOSTRING(NodeID));
         ///Needed to correctly support function pointers
         if(GET_NODE(GetPointer<gimple_call>(tn)->fn)->get_kind() == ssa_name_K)
         {
            unsigned int Index = GET_INDEX_NODE(GetPointer<gimple_call>(tn)->fn);
            Analyze_lvalue(Ver, Index);
         }
         const std::vector<tree_nodeRef> & args = GetPointer<gimple_call>(tn)->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            unsigned int Index = GET_INDEX_NODE(*arg);
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing actual args of statement: " + (*arg)->ToString());
            ///add parameter to the vertex
            AppM->get_function_behavior(funID)->ogc->add_parameter(Ver, Index);
            if (tree_helper::is_a_variable(TM, Index))
               Analyze_rvalue(Ver, Index);
            else
               Analyze_lvalue(Ver, Index);
         }

         if (GetPointer<gimple_call>(tn)->virtual_operands)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing vops of statement: " + TOSTRING(NodeID));
            add_vops(Ver, GetPointer<gimple_call>(tn)->virtual_operands);
         }
         break;
      }
      case gimple_cond_K:
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing condition statement: "  + TOSTRING(NodeID));
         Analyze_rvalue(Ver, GET_INDEX_NODE(GetPointer<gimple_cond>(tn)->op0));
         break;
      }
      case while_expr_K:
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing while statement: "  + TOSTRING(NodeID));
         Analyze_rvalue(Ver, GET_INDEX_NODE(GetPointer<while_expr>(tn)->op0));
         break;
      }
      case for_expr_K:
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing for statement: "  + TOSTRING(NodeID));
         Analyze_rvalue(Ver, GET_INDEX_NODE(GetPointer<for_expr>(tn)->op0));
         AnalyzeNodeID(Ver, GET_INDEX_NODE(GetPointer<for_expr>(tn)->op1));
         Analyze_lvalue(Ver, GET_INDEX_NODE(GetPointer<for_expr>(tn)->op2));
         break;
      }
      case gimple_switch_K:
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing switch statement: "  + TOSTRING(NodeID));
         Analyze_rvalue(Ver, GET_INDEX_NODE(GetPointer<gimple_switch>(tn)->op0));

         if (GetPointer<gimple_switch>(tn)->op1)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing first operand of switch: "  + TOSTRING(NodeID));
            tree_nodeRef rop = GetPointer<gimple_switch>(tn)->op1;
            Analyze_rvalue(Ver, GET_INDEX_NODE(rop));
         }

         break;
      }
      case gimple_label_K:
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing label associated to statement: " + TOSTRING(NodeID));
         tree_nodeRef op = GetPointer<gimple_label>(tn)->op;
         Analyze_rvalue(Ver, GET_INDEX_NODE(op));
         break;
      }
      case gimple_goto_K:
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing goto associated to statement: " + TOSTRING(NodeID));
         tree_nodeRef op = GetPointer<gimple_goto>(tn)->op;
         Analyze_rvalue(Ver, GET_INDEX_NODE(op));
         break;
      }
      case gimple_asm_K:
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing asm associated to statement: " + TOSTRING(NodeID));
         gimple_asm* asme = GetPointer<gimple_asm>(tn);

         /// it could be wrong!

         if (asme->out)
            Analyze_lvalue(Ver, GET_INDEX_NODE(asme->out));
         if (asme->in)
            Analyze_rvalue(Ver, GET_INDEX_NODE(asme->in));
         if (asme->clob)
            Analyze_rvalue(Ver, GET_INDEX_NODE(asme->clob));

         if (GetPointer<gimple_asm>(tn)->virtual_operands)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing vops of statement: " + TOSTRING(NodeID));
            add_vops(Ver, GetPointer<gimple_asm>(tn)->virtual_operands);
         }
         break;
      }
      case gimple_resx_K:
      {
         break;
      }
      default:
         THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "Not supported statement: " + std::string(tn->get_kind_text()));
   }
   return;
}


void var_computation::Analyze_lvalue(const vertex &Ver, unsigned int NodeID, bool onlyReference)
{
   const tree_managerRef TM = AppM->get_tree_manager();
   const tree_nodeRef tn = TM->get_tree_node_const(NodeID);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node " + boost::lexical_cast<std::string>(NodeID) + " of type " + tn->get_kind_text() + " - left value");

   operations_graph_constructorRef ogc = AppM->get_function_behavior(funID)->ogc;

   switch(tn->get_kind())
   {
      case real_cst_K:
      case complex_cst_K:
      case string_cst_K:
      case integer_cst_K:
      {
         //This is possible if this nodes have been collapsed as parameter of a call
         break;
      }
      case function_decl_K:
      {
         THROW_WARNING("Function pointer");
         break;
      }
      case var_decl_K:
      case parm_decl_K:
      case result_decl_K:
      case ssa_name_K:
      {
         if (!onlyReference)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Adding written variable " + TOSTRING(NodeID));
            ogc->add_ssa_written(Ver, NodeID);
         }
         if (GetPointer<ssa_name>(tn))
         {
            tree_nodeRef tnVar = GET_NODE(GetPointer<ssa_name>(tn)->var);
            if (GetPointer<var_decl>(tnVar) && !GetPointer<var_decl>(tnVar)->artificial_flag)
               isRef = true;
            if (GetPointer<var_decl>(tnVar) && GetPointer<var_decl>(tnVar)->static_flag)
               isRef = false;
            if ((GetPointer<ssa_name>(tn)->volatile_flag || GET_NODE(GetPointer<ssa_name>(tn)->def_stmts.front())->get_kind() == nop_expr_K) and isRef && vops.find(NodeID) == vops.end())
            {
               ogc->add_ref_var(ogc->CgetIndex(ENTRY), NodeID);
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Adding referenced variable " + TOSTRING(NodeID) + " from ENTRY vertex");
            }
         }
         if (isRef && vops.find(NodeID) == vops.end())
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Adding referenced variable " + TOSTRING(NodeID));
            ogc->add_ref_var(Ver, NodeID);
         }
         isRef = true;
         var_decl *vd = GetPointer<var_decl>(tn);
         if (vd and (!vd->scpe || GET_NODE(vd->scpe)->get_kind() == translation_unit_decl_K))
            AppM->add_global_variable(NodeID);

         break;
      }
      case tree_list_K:
      {
         if (GetPointer<tree_list>(tn)->purp)
         {
            unsigned int Index = GET_INDEX_NODE(GetPointer<tree_list>(tn)->purp);
            Analyze_rvalue(Ver, Index);
         }

         tree_nodeRef actual_args = tn;
         while (actual_args)
         {
            unsigned int Index = GET_INDEX_NODE(GetPointer<tree_list>(actual_args)->valu);
            Analyze_lvalue(Ver, Index);
            if (GetPointer<tree_list>(actual_args)->chan)
               actual_args = GET_NODE(GetPointer<tree_list>(actual_args)->chan);
            else
               break;
         }
         break;
      }
      case CASE_UNARY_EXPRESSION:
      {
         Analyze_rvalue(Ver, GET_INDEX_NODE(GetPointer<unary_expr>(tn)->op));
         if(tn->get_kind() == indirect_ref_K)
         {
            indirect_ref * ir = GetPointer<indirect_ref>(tn);
            type_node * treen = GetPointer<type_node>(GET_NODE(ir->type));
            THROW_ASSERT(treen, "Expected a type_node");
            AppM->get_function_behavior(funID)->add_constant(GET_INDEX_NODE(treen->size));
         }
         break;
      }
      case CASE_BINARY_EXPRESSION:
      {
         binary_expr * be = GetPointer<binary_expr>(tn);
         if(tn->get_kind() == postincrement_expr_K || tn->get_kind() == with_size_expr_K)
            Analyze_lvalue(Ver, GET_INDEX_NODE(be->op0), true);
         else
            Analyze_rvalue(Ver, GET_INDEX_NODE(be->op0));
         Analyze_rvalue(Ver, GET_INDEX_NODE(be->op1));
         break;
      }
      case CASE_TERNARY_EXPRESSION:
      {
         ternary_expr * expr = GetPointer<ternary_expr>(tn);

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing ternary first operator: " + TOSTRING(GET_INDEX_NODE(expr->op0)));
         Analyze_lvalue(Ver, GET_INDEX_NODE(expr->op0), true);

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing ternary second operator: " + TOSTRING(GET_INDEX_NODE(expr->op1)));
         Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op1));

         if (expr->op2)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing ternary third operator: " + TOSTRING(GET_INDEX_NODE(expr->op2)));
            Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op2));
         }

         break;
      }
      case CASE_QUATERNARY_EXPRESSION:
      {
         quaternary_expr * expr = GetPointer<quaternary_expr>(tn);

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing quaternary first operator: " + TOSTRING(GET_INDEX_NODE(expr->op0)));
         Analyze_lvalue(Ver, GET_INDEX_NODE(expr->op0), true);

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing quaternary second operator: " + TOSTRING(GET_INDEX_NODE(expr->op1)));
         Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op1));

         if (expr->op2)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing quaternary third operator: " + TOSTRING(GET_INDEX_NODE(expr->op2)));
            Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op2));
         }

         if (expr->op3)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing quaternary forth operator: " + TOSTRING(GET_INDEX_NODE(expr->op3)));
            Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op3));
         }
         break;
      }
      case exc_ptr_expr_K:
      case filter_expr_K:
      {
         break;
      }
      case target_mem_ref_K:
      {
         target_mem_ref * tm = GetPointer<target_mem_ref>(tn);
         if (tm->symbol) Analyze_lvalue(Ver,GET_INDEX_NODE(tm->symbol),true);
         if (tm->base) Analyze_lvalue(Ver,GET_INDEX_NODE(tm->base),true);
         if (tm->idx) Analyze_rvalue(Ver,GET_INDEX_NODE(tm->idx));
         break;
      }
      case target_mem_ref461_K:
      {
         target_mem_ref461 * tm = GetPointer<target_mem_ref461>(tn);
         if (tm->base) Analyze_lvalue(Ver,GET_INDEX_NODE(tm->base),true);
         if (tm->idx) Analyze_rvalue(Ver,GET_INDEX_NODE(tm->idx));
         if (tm->idx2) Analyze_rvalue(Ver,GET_INDEX_NODE(tm->idx2));
         break;
      }
      default:
         THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "Not supported lvalue NodeID: " + std::string(tn->get_kind_text()) + " " + STR(NodeID));
   }
   return;
}

void var_computation::Analyze_rvalue(const vertex &Ver, unsigned int NodeID)
{
   const tree_managerRef TM = AppM->get_tree_manager();
   const tree_nodeRef tn = TM->get_tree_node_const(NodeID);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing node " + boost::lexical_cast<std::string>(NodeID) + " of type " + std::string(tn->get_kind_text()) + " - right value");

   operations_graph_constructorRef ogc = AppM->get_function_behavior(funID)->ogc;

   switch(tn->get_kind())
   {
      case call_expr_K:
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing call_expr statement: " + TOSTRING(NodeID));
         ///Needed to correctly support function pointers
         if(GET_NODE(GetPointer<call_expr>(tn)->fn)->get_kind() == ssa_name_K)
         {
            unsigned int Index = GET_INDEX_NODE(GetPointer<call_expr>(tn)->fn);
            Analyze_lvalue(Ver, Index);
         }
         const std::vector<tree_nodeRef> & args = GetPointer<call_expr>(tn)->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            unsigned int Index = GET_INDEX_NODE(*arg);
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing actual args of statement: " + (*arg)->ToString());
            ///add parameter to the vertex
            AppM->get_function_behavior(funID)->ogc->add_parameter(Ver, Index);
            if (tree_helper::is_a_variable(TM, Index))
               Analyze_rvalue(Ver, Index);
            else
               Analyze_lvalue(Ver, Index);
         }
      }
      case real_cst_K:
      case complex_cst_K:
      case string_cst_K:
      case integer_cst_K:
      {
         AppM->get_function_behavior(funID)->add_constant(NodeID);
         break;
      }
      case field_decl_K:
      case function_decl_K:
      case label_decl_K:
         break;
      case result_decl_K:
      {
         ogc->add_ref_var(Ver, NodeID);
         break;
      }
      case parm_decl_K:
      case var_decl_K:
      case ssa_name_K:
      {
         bool entryWrite = false;

         ssa_name *sa = GetPointer<ssa_name>(tn);
         if (sa && (sa->volatile_flag || GET_NODE(sa->def_stmts.front())->get_kind() == gimple_nop_K))
            entryWrite = true;
         parm_decl *pd = GetPointer<parm_decl>(tn);
         if(pd)
            entryWrite = true;

         if (entryWrite)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Adding written variable " + TOSTRING(NodeID) + " to ENTRY");
            ogc->add_ssa_written(ogc->CgetIndex(ENTRY), NodeID);
            if (vops.find(NodeID) == vops.end())
            {
               ogc->add_ref_var(ogc->CgetIndex(ENTRY), NodeID);
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Adding referenced variable " + TOSTRING(NodeID) + " from ENTRY");
            }
         }

         var_decl *vd = GetPointer<var_decl>(tn);
         if(vd && vd->init)
            Analyze_rvalue(Ver, GET_INDEX_NODE(vd->init));
         if (vd and (!vd->scpe || GET_NODE(vd->scpe)->get_kind() == translation_unit_decl_K))
            AppM->add_global_variable(NodeID);
         if (vd or pd)
         {
            if(vd && vd->name && GetPointer<identifier_node>(GET_NODE(vd->name)))
            {
               identifier_node * id = GetPointer<identifier_node>(GET_NODE(vd->name));
               if(id->strg == "__FUNCTION__" || id->strg == "__PRETTY_FUNCTION__")
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Found special variable __FUNCTION__");
               }
               else
               {
                  ogc->add_ref_var(Ver, NodeID);
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Adding referenced variable " + TOSTRING(NodeID));
               }
            }
            else
            {
               ogc->add_ref_var(Ver, NodeID);
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Adding referenced variable " + TOSTRING(NodeID));
            }
         }

         ogc->add_ssa_read(Ver, NodeID);
         if (vops.find(NodeID) == vops.end())
         {
            if (sa and vops.find(GET_INDEX_NODE(sa->var)) == vops.end() and tree_helper::is_a_variable(TM, GET_INDEX_NODE(sa->var)))
            {
               ogc->add_ref_var(Ver, NodeID);
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Adding referenced variable " + TOSTRING(NodeID) + " as ssa-variable");
            }
         }

         break;
      }
      case vector_cst_K:
      {
         vector_cst* vc = GetPointer<vector_cst>(tn);
         std::vector<tree_nodeRef> list_of_valu = vc->list_of_valu;
         for(unsigned int i = 0; i < list_of_valu.size(); i++)
            Analyze_rvalue(Ver, GET_INDEX_NODE(list_of_valu[i]));
         break;
      }
      case tree_list_K:
      {
         if (GetPointer<tree_list>(tn)->purp)
         {
            unsigned int Index = GET_INDEX_NODE(GetPointer<tree_list>(tn)->purp);
            Analyze_rvalue(Ver, Index);
         }
         tree_nodeRef actual_args = tn;
         while (actual_args)
         {
            unsigned int Index = GET_INDEX_NODE(GetPointer<tree_list>(actual_args)->valu);
            Analyze_rvalue(Ver, Index);
            if (GetPointer<tree_list>(actual_args)->chan)
               actual_args = GET_NODE(GetPointer<tree_list>(actual_args)->chan);
            else 
               break;
         }
         break;
      }
      case tree_vec_K:
      {
         tree_vec* tv = GetPointer<tree_vec>(tn);
         std::vector<tree_nodeRef> list_of_op = tv->list_of_op;
         for(unsigned int i = 0; i < list_of_op.size(); i++)
            Analyze_rvalue(Ver, GET_INDEX_NODE(list_of_op[i]));
         break;
      }
      case case_label_expr_K:
      {
         case_label_expr* expr = GetPointer<case_label_expr>(tn);
         if (expr->op0)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing case_label first operator: " + TOSTRING(GET_INDEX_NODE(expr->op0)));
            Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op0));
         }

         if (expr->op1)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing case_label second operator: " + TOSTRING(GET_INDEX_NODE(expr->op1)));
            Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op1));
         }

         if (expr->got)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing goto operator: " + TOSTRING(GET_INDEX_NODE(expr->got)));
            Analyze_rvalue(Ver, GET_INDEX_NODE(expr->got));
         }

         break;
      }
      case CASE_UNARY_EXPRESSION:
      {
         unary_expr * expr = GetPointer<unary_expr>(tn);
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing unary operator: " + TOSTRING(GET_INDEX_NODE(expr->op)));
         Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op));
         if(tn->get_kind() == indirect_ref_K)
         {
            indirect_ref * ir = GetPointer<indirect_ref>(tn);
            type_node * treen = GetPointer<type_node>(GET_NODE(ir->type));
            THROW_ASSERT(treen, "Expected a type_node");
            AppM->get_function_behavior(funID)->add_constant(GET_INDEX_NODE(treen->size));
         }
         break;
      }
      case CASE_TERNARY_EXPRESSION:
      {
         ternary_expr * expr = GetPointer<ternary_expr>(tn);

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing ternary first operator: " + TOSTRING(GET_INDEX_NODE(expr->op0)));
         Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op0));

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing ternary second operator: " + TOSTRING(GET_INDEX_NODE(expr->op1)));
         Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op1));

         if (expr->op2)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing ternary third operator: " + TOSTRING(GET_INDEX_NODE(expr->op2)));
            Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op2));
         }

         break;
      }
      case CASE_BINARY_EXPRESSION:
      {
         binary_expr * expr = GetPointer<binary_expr>(tn);

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing binary first operator: " + TOSTRING(GET_INDEX_NODE(expr->op0)));
         Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op0));

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing binary second operator: " + TOSTRING(GET_INDEX_NODE(expr->op1)));
         Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op1));
         break;
      }
      case CASE_QUATERNARY_EXPRESSION:
      {
         quaternary_expr * expr = GetPointer<quaternary_expr>(tn);

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing quaternary first operator: " + TOSTRING(GET_INDEX_NODE(expr->op0)));
         Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op0));

         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing quaternary second operator: " + TOSTRING(GET_INDEX_NODE(expr->op1)));
         Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op1));

         if (expr->op2)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing quaternary third operator: " + TOSTRING(GET_INDEX_NODE(expr->op2)));
            Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op2));
         }

         if (expr->op3)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzing quaternary forth operator: " + TOSTRING(GET_INDEX_NODE(expr->op3)));
            Analyze_rvalue(Ver, GET_INDEX_NODE(expr->op3));
         }

         break;
      }
      case constructor_K:
      {
         constructor* con = GetPointer<constructor>(tn);
         std::vector<std::pair< tree_nodeRef, tree_nodeRef> > list_of_idx_valu = con->list_of_idx_valu;
         for(std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::iterator el = list_of_idx_valu.begin(); el != list_of_idx_valu.end(); el++)
         {
            if (el->first)
               Analyze_rvalue(Ver, GET_INDEX_NODE(el->first));
            if (el->second)
               Analyze_rvalue(Ver, GET_INDEX_NODE(el->second));
         }
         break;
      }
      case target_mem_ref_K:
      {
         target_mem_ref * tm = GetPointer<target_mem_ref>(tn);
         if (tm->symbol) Analyze_rvalue(Ver,GET_INDEX_NODE(tm->symbol));
         if (tm->base) Analyze_rvalue(Ver,GET_INDEX_NODE(tm->base));
         if (tm->idx) Analyze_rvalue(Ver,GET_INDEX_NODE(tm->idx));
         break;
      }
      case target_mem_ref461_K:
      {
         target_mem_ref461 * tm = GetPointer<target_mem_ref461>(tn);
         if (tm->base) Analyze_rvalue(Ver,GET_INDEX_NODE(tm->base));
         if (tm->idx) Analyze_rvalue(Ver,GET_INDEX_NODE(tm->idx));
         if (tm->idx2) Analyze_rvalue(Ver,GET_INDEX_NODE(tm->idx2));
         break;
      }
      case exc_ptr_expr_K:
      case filter_expr_K:
      {
         break;
      }
      case gimple_cond_K:
      {
         const gimple_cond * gc = GetPointer<gimple_cond>(tn);
         Analyze_rvalue(Ver, GET_INDEX_NODE(gc->op0));
         break;
      }
      default:
         THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "Not supported rvalue NodeID: " + std::string(tn->get_kind_text()));
   }
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Analyzed node "+ TOSTRING(NodeID));
   return;
}

void var_computation::add_vops(const vertex &Ver, const vopsRef& vop)
{
   operations_graph_constructorRef ogc = AppM->get_function_behavior(funID)->ogc;
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Adding virtual operands"); 
   /// VIRTUAL USES
   if (vop->list_of_vuses.size())
   {
      std::vector<vuseRef> list_of_vuses = vop->list_of_vuses;
      for (std::vector<vuseRef>::iterator it = list_of_vuses.begin(); it != list_of_vuses.end(); it++)
      {
         /// Adding virtual variables used by the operation
         std::vector<tree_nodeRef> list_of_vuse_ops = (*it)->list_of_vuse_ops;
         for (std::vector<tree_nodeRef>::iterator it2 = list_of_vuse_ops.begin(); it2 != list_of_vuse_ops.end(); it2++)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "    Adding read variable " + TOSTRING(GET_INDEX_NODE(*it2)));
            ogc->add_ssa_read(Ver, GET_INDEX_NODE(*it2));
            /// if the variable is a volatile ssa one, it means that the entry node writes it; it is function parameter or something coming
            /// from outside the function (e.g. a global parameter)
            /*if (GetPointer<ssa_name>(tm->get_tree_node_const(info.second)) and
            (GetPointer<ssa_name>(tm->get_tree_node_const(info.second))->volatile_flag ||  || GET_NODE(GetPointer<ssa_name>(tm->get_tree_node_const(info.second))->def_stmts.front())->get_kind() == nop_expr_K))
            {
               //bm->add_ssa_written(ENTRY, info.second, info.first);
            }*/
         }
      }
   }

   /// VIRTUAL DEFINITIONS
   if (vop->list_of_vdefs.size())
   {
      std::vector<vdefRef> list_of_vdefs = vop->list_of_vdefs;
      for (std::vector<vdefRef>::iterator it = list_of_vdefs.begin(); it != list_of_vdefs.end(); it++)
      {
         /// Adding the (virtual) variable defined by the operation
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "    Adding written variable " + TOSTRING(GET_INDEX_NODE((*it)->vdefinition)));
         ogc->add_ssa_written(Ver, GET_INDEX_NODE((*it)->vdefinition));
         vops.insert(GET_INDEX_NODE((*it)->vdefinition));
         /// Adding virtual definition variables (variables redefined by the operation)
         std::vector<tree_nodeRef> list_of_vdef_ops = (*it)->list_of_vdef_ops;
         for (std::vector<tree_nodeRef>::iterator it2 = list_of_vdef_ops.begin(); it2 != list_of_vdef_ops.end(); it2++)
         {
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "    Adding redefined variable " + TOSTRING(GET_INDEX_NODE(*it2)));
            ogc->add_redefined_var(Ver, GET_INDEX_NODE(*it2));
            vops.insert(GET_INDEX_NODE((*it)->vdefinition));
         }
      }
   }

}

struct printvar : public std::unary_function<unsigned int, bool>
{
   std::ostream& os;
   printvar (std::ostream& os_) : os(os_) {}
   void operator() (unsigned int x) { os << x << " "; }
};


void var_computation::report()
{
   const tree_managerRef TM = AppM->get_tree_manager();
   std::ofstream os((Param->getOption<std::string>(OPT_output_temporary_directory) + "dump_"+ TOSTRING(funID) + ".txt").c_str());
   const graph *cfg = AppM->get_function_behavior(funID)->get_graph(function_behavior::CFG);
   VertexIterator v, vend;
   os << "Function name: " << tree_helper::name_function(TM, funID) << std::endl;
   for(tie(v,vend) = boost::vertices(*cfg); v != vend; v++)
   {
      os << "Name: " << GET_NAME(cfg, *v) << std::endl;
      os << "OP: " << GET_OP(cfg, *v) << std::endl;
      os << "TYPE: " << GET_TYPE(cfg, *v) << std::endl;
      std::tr1::unordered_set<unsigned int >SET = GET_SSA_WRITTEN(cfg, *v);
      if (SET.size())
      {
         os << "  - Var Written: ";
         std::for_each(SET.begin(), SET.end(), printvar(os));
         os << std::endl;
      }
      SET = GET_REDEFINED_VAR(cfg, *v);
      if (SET.size())
      {
         os << "  - Var Redefined: ";
         std::for_each(SET.begin(), SET.end(), printvar(os));
         os << std::endl;
      }
      std::list<unsigned int >LIST = GET_SSA_READ(cfg, *v);
      if (LIST.size())
      {
         os << "  - Var Read: ";
         std::for_each(LIST.begin(), LIST.end(), printvar(os));
         os << std::endl;
      }
      SET = GET_REF_VAR(cfg, *v);
      if (SET.size())
      {
         os << "  - Var Referenced: ";
         std::for_each(SET.begin(), SET.end(), printvar(os));
         os << std::endl;
      }
      SET = GET_STORE_VAR(cfg, *v);
      if (SET.size())
      {
         os << "  - Var Stored: ";
         std::for_each(SET.begin(), SET.end(), printvar(os));
         os << std::endl;
      }
      SET = GET_LOAD_VAR(cfg, *v);
      if (SET.size())
      {
         os << "  - Var Loaded: ";
         std::for_each(SET.begin(), SET.end(), printvar(os));
         os << std::endl;
      }
      LIST = GET_NODEID(cfg, *v);
      if (LIST.size())
      {
         os << "  - NodeID: ";
         std::for_each(LIST.begin(), LIST.end(), printvar(os));
         unsigned int Node = LIST.front();
         os << std::endl;
         os << "  - Type: " << TM->get_tree_node_const(Node)->get_kind_text();
         os << std::endl;
      }
      os << "***" << std::endl;
   }
}
