/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file var_decl_fix.cpp
 * @brief Pre-analysis step fixing var_decl duplications.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Header include
#include "var_decl_fix.hpp"

///Behavior include
#include "application_manager.hpp"
#include "function_behavior.hpp"

///Graph include
#include "basic_block.hpp"
#include "basic_blocks_graph_constructor.hpp"

///Parameter include
#include "Parameter.hpp"

///STD include
#include <fstream>

///STL include
#include <map>
#include <string>

///Tree include
#include "ext_tree_node.hpp"
#include "tree_basic_block.hpp"
#include "tree_manager.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"
#include "token_interface.hpp"

///Utility include
#include "dbgPrintHelper.hpp"
#include "exceptions.hpp"

bool var_decl_fix::tree_dumped = false;

var_decl_fix::var_decl_fix(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int fun_id) :
   BaseAnalysis(_Param, _AppM, VAR_DECL_FIX, fun_id)
{
   debug_level = Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE);
   this->SetPreCond(ID, CHECK_SYSTEM_TYPE);
}


var_decl_fix::~var_decl_fix()
{

}

void var_decl_fix::exec()
{
   const tree_managerRef TM = AppM->get_tree_manager();
   if(debug_level >= DEBUG_LEVEL_PEDANTIC && !tree_dumped)
   {
      tree_dumped = true;
      std::string raw_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "before_var_decl_fix.raw";
      std::ofstream raw_file(raw_file_name.c_str());
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Tree-Manager dumped for debug purpose");
      raw_file << TM;
      raw_file.close();
   }

   tree_nodeRef & curr_tn = TM->add(funID);
   function_decl * fd = GetPointer<function_decl>(curr_tn);
   statement_list * sl = GetPointer<statement_list>(GET_NODE(fd->body));

   //TODO: this step not only fixes variable declaration but also add entry and exit block
   //Adding entry block
   blocRef entry_bloc = blocRef(new bloc());
   entry_bloc->number = BB_ENTRY;
   //Set of successor of entry
   std::vector<unsigned int>& succ_entry = entry_bloc->list_of_succ;
   std::map<unsigned int, blocRef> &blocks = sl->list_of_bloc;
   std::map<unsigned int, blocRef>::iterator it, it_end;
   it_end = blocks.end();
   for(it = blocks.begin(); it != it_end; it++)
   {
      std::vector<unsigned int>::iterator it2, it2_end;
      it2_end = it->second->list_of_pred.end();
      for(it2 = it->second->list_of_pred.begin(); it2 != it2_end; it2++)
      {
         if(*it2 == BB_ENTRY)
            succ_entry.push_back(it->second->number);
      }
   }
   ///set of predecessor of exit is missing! TO BE CHECK
   //Adding exit
   blocRef exit_bloc = blocRef(new bloc());
   exit_bloc->number = BB_EXIT;
   sl->list_of_bloc[BB_ENTRY] = entry_bloc;
   sl->list_of_bloc[BB_EXIT] = exit_bloc;
   it_end = blocks.end();
   for(it = blocks.begin(); it != it_end; it++)
   {
      std::vector<tree_nodeRef> & list_of_stmt = it->second->list_of_stmt;
      std::vector<tree_nodeRef>::iterator it1, it2, it2_end = list_of_stmt.end();
      it1 = list_of_stmt.begin();
      for(it2 = list_of_stmt.begin(); it2 != it2_end; it2 = it1)
      {
         it1++;
         recursive_examinate(*it2);
      }
   }
   if(debug_level >= DEBUG_LEVEL_PEDANTIC)
   {
      std::string raw_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "after_var_decl_fix.raw";
      std::ofstream raw_file(raw_file_name.c_str());
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Tree-Manager dumped for debug purpose");
      raw_file << TM;
      raw_file.close();
   }
}

void var_decl_fix::recursive_examinate(const tree_nodeRef & tn)
{
   THROW_ASSERT(tn->get_kind() == tree_reindex_K, "Node is not a tree reindex");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Update recursively " + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)));
   const tree_managerRef TM = AppM->get_tree_manager();
   const tree_nodeRef curr_tn = GET_NODE(tn);
   switch(curr_tn->get_kind())
   {
      case call_expr_K:
      {
         const call_expr * ce = GetPointer<call_expr>(curr_tn);
         const std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            recursive_examinate(*arg);
         }
         break;
      }
      case gimple_call_K:
      {
         const gimple_call * ce = GetPointer<gimple_call>(curr_tn);
         const std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            recursive_examinate(*arg);
         }
         break;
      }
      case gimple_assign_K:
      {
         gimple_assign * gm = GetPointer<gimple_assign>(curr_tn);
         recursive_examinate(gm->op0);
         recursive_examinate(gm->op1);
         if (GET_NODE(gm->op0)->get_kind() == var_decl_K && (GET_NODE(gm->op1)->get_kind() == string_cst_K
                                                             || GET_NODE(gm->op1)->get_kind() == constructor_K))
         {
            var_decl* vd = GetPointer<var_decl>(GET_NODE(gm->op0));
            THROW_ASSERT(vd, "not valid variable");
            vd->init = gm->op1;
            gm->init_assignment = true; /// Removing statements create more problems that what it may solve.
                                        /// So the solution for bambu is to not synthesize them.
         }
         break;
      }
      case gimple_nop_K:
      {
         break;
      }
      case var_decl_K:
      case parm_decl_K:
      {
         if(already_examinated_decls.find(GET_INDEX_NODE(tn)) == already_examinated_decls.end())
         {
            already_examinated_decls.insert(GET_INDEX_NODE(tn));
            decl_node * dn = GetPointer<decl_node>(TM->add(GET_INDEX_NODE(tn)));
            recursive_examinate(dn->type);
            if(dn && dn->name)
            {
               //check if the var_decl
               if(curr_tn->get_kind() == var_decl_K)
               {  /* this is a variable declaration */
                  var_decl *cast_res = GetPointer<var_decl>(curr_tn);
                  if(cast_res->static_flag == true)
                  {  
                     PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "Found a static variable with identifier <" +
                                                                      GetPointer<identifier_node>(GET_NODE(dn->name))->strg +
                                                                      "> within function #" + boost::lexical_cast<std::string>(funID));

#if HAVE_EXPERIMENTAL
                     //update function feature
                     function_behaviorRef fbr = AppM->get_function_behavior(funID);
                     fbr->setCUDAFlag(false);
#endif
                  }
               }

               std::string name_id = GetPointer<identifier_node>(GET_NODE(dn->name))->strg;
               if(already_examinated_names.find(name_id) == already_examinated_names.end())
                  already_examinated_names.insert(name_id);
               else
               {
                  PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, name_id + " is a duplicated var_decl");
                  ///create a new identifier_node tree node
                  std::map<treeVocabularyTokenTypes::token_enum, std::string> IR_schema;
                  unsigned int var_decl_name_nid_test;
                  unsigned var_decl_unique_id=0;
                  do
                  {
                     IR_schema[TOK(TOK_STRG)] = name_id+boost::lexical_cast<std::string>(var_decl_unique_id++);
                     var_decl_name_nid_test = TM->find(TOK(TOK_IDENTIFIER_NODE), IR_schema);
                  } while(var_decl_name_nid_test);
                  unsigned int var_decl_name_nid = TM->new_tree_node_id();
                  TM->create_tree_node(var_decl_name_nid, TOK(TOK_IDENTIFIER_NODE), IR_schema);
                  IR_schema.clear();
                  tree_nodeRef tr_new_id = tree_nodeRef(new tree_reindex(var_decl_name_nid, TM->add(var_decl_name_nid)));
                  dn->name = tr_new_id;
               }
            }
         }
         break;
      }
      case ssa_name_K:
      {
         const ssa_name * sn = GetPointer<ssa_name>(curr_tn);
         recursive_examinate(sn->var);
         break;
      }
      case tree_list_K:
      {
         tree_nodeRef current = tn;
         while(current)
         {
            recursive_examinate(GetPointer<tree_list>(GET_NODE(current))->valu);
            current = GetPointer<tree_list>(GET_NODE(current))->chan;
         }
         break;
      }
      case CASE_UNARY_EXPRESSION :
      {
         if(curr_tn->get_kind() == addr_expr_K)
         {
            if(already_visited_ae.find(GET_INDEX_NODE(tn)) != already_visited_ae.end())
            {
               break;
            }
            already_visited_ae.insert(GET_INDEX_NODE(tn));
         }
         const unary_expr * ue = GetPointer<unary_expr>(curr_tn);
         recursive_examinate(ue->op);
         break;
      }
      case CASE_BINARY_EXPRESSION :
      {
         const binary_expr * be = GetPointer<binary_expr>(curr_tn);
         recursive_examinate(be->op0);
         recursive_examinate(be->op1);
         break;
      }
      case CASE_TERNARY_EXPRESSION :
      {
         const ternary_expr * te = GetPointer<ternary_expr>(curr_tn);
         recursive_examinate(te->op0);
         if(te->op1)
            recursive_examinate(te->op1);
         if(te->op2)
            recursive_examinate(te->op2);
         break;
      }
      case CASE_QUATERNARY_EXPRESSION :
      {
         const quaternary_expr * qe = GetPointer<quaternary_expr>(curr_tn);
         recursive_examinate(qe->op0);
         if(qe->op1)
            recursive_examinate(qe->op1);
         if(qe->op2)
            recursive_examinate(qe->op2);
         if(qe->op3)
            recursive_examinate(qe->op3);
         break;
      }
      case constructor_K:
      {
         const constructor * co = GetPointer<constructor>(curr_tn);
         const std::vector<std::pair< tree_nodeRef, tree_nodeRef> > & list_of_idx_valu = co->list_of_idx_valu;
         std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::const_iterator it, it_end = list_of_idx_valu.end();
         for(it = list_of_idx_valu.begin(); it != it_end; it++)
         {
            recursive_examinate(it->second);
         }
         break;
      }
      case gimple_cond_K:
      {
         const gimple_cond * gc = GetPointer<gimple_cond>(curr_tn);
         recursive_examinate(gc->op0);
         break;
      }
      case gimple_switch_K:
      {
         const gimple_switch * se = GetPointer<gimple_switch>(curr_tn);
         recursive_examinate(se->op0);
         break;
      }
      case gimple_return_K:
      {
         const gimple_return * re = GetPointer<gimple_return>(curr_tn);
         if(re->op)
            recursive_examinate(re->op);
         break;
      }
      case for_expr_K:
      {
         const for_expr * fe = GetPointer<for_expr>(curr_tn);
         recursive_examinate(fe->op0);
         recursive_examinate(fe->op1);
         recursive_examinate(fe->op2);
         break;
      }
      case while_expr_K:
      {
         const while_expr * we = GetPointer<while_expr>(curr_tn);
         recursive_examinate(we->op0);
         break;
      }
      case vector_type_K:
      case array_type_K:
      case complex_type_K:
      case real_type_K:
      case record_type_K:
      case union_type_K:
      case pointer_type_K:
      case boolean_type_K:
      case enumeral_type_K:
      case integer_type_K:
      {
         if(already_examinated_decls.find(GET_INDEX_NODE(tn)) == already_examinated_decls.end())
         {
            already_examinated_decls.insert(GET_INDEX_NODE(tn));
            type_node * ty = GetPointer<type_node>(TM->add(GET_INDEX_NODE(tn)));
            if(ty && ty->name && GET_NODE(ty->name)->get_kind() == type_decl_K)
            {
               recursive_examinate(ty->name);
            }
            else if(ty and (not ty->system_flag) and ty->name and GET_NODE(ty->name)->get_kind() == identifier_node_K)
            {
               std::string name_id = GetPointer<identifier_node>(GET_NODE(ty->name))->strg;
               if(already_examinated_type_names.find(name_id) == already_examinated_type_names.end())
               {
                  already_examinated_type_names.insert(name_id);
               }
               else
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---" + name_id + " is a duplicated type");
                  ///create a new identifier_node tree node
                  std::map<treeVocabularyTokenTypes::token_enum, std::string> IR_schema;
                  unsigned int var_decl_name_nid_test;
                  unsigned var_decl_unique_id=0;
                  do
                  {
                     IR_schema[TOK(TOK_STRG)] = name_id+boost::lexical_cast<std::string>(var_decl_unique_id++);
                     var_decl_name_nid_test = TM->find(TOK(TOK_IDENTIFIER_NODE), IR_schema);
                  } while(var_decl_name_nid_test);
                  unsigned int var_decl_name_nid = TM->new_tree_node_id();
                  TM->create_tree_node(var_decl_name_nid, TOK(TOK_IDENTIFIER_NODE), IR_schema);
                  IR_schema.clear();
                  tree_nodeRef tr_new_id = tree_nodeRef(new tree_reindex(var_decl_name_nid, TM->add(var_decl_name_nid)));
                  ty->name = tr_new_id;
               }
            }
         }
         break;
      }
      case type_decl_K:
      {
         type_decl * td = GetPointer<type_decl>(GET_NODE(tn));
         if(td && td->name && td->include_name != "<built-in>" and (not td->operating_system_flag) and (not td->library_system_flag))
         {
            std::string name_id = GetPointer<identifier_node>(GET_NODE(td->name))->strg;
            if(already_examinated_type_names.find(name_id) == already_examinated_type_names.end())
               already_examinated_type_names.insert(name_id);
            else
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---" + name_id + " is a duplicated type");
               ///create a new identifier_node tree node
               std::map<treeVocabularyTokenTypes::token_enum, std::string> IR_schema;
               unsigned int var_decl_name_nid_test;
               unsigned var_decl_unique_id=0;
               do
               {
                  IR_schema[TOK(TOK_STRG)] = name_id+boost::lexical_cast<std::string>(var_decl_unique_id++);
                  var_decl_name_nid_test = TM->find(TOK(TOK_IDENTIFIER_NODE), IR_schema);
               } while(var_decl_name_nid_test);
               unsigned int var_decl_name_nid = TM->new_tree_node_id();
               TM->create_tree_node(var_decl_name_nid, TOK(TOK_IDENTIFIER_NODE), IR_schema);
               IR_schema.clear();
               tree_nodeRef tr_new_id = tree_nodeRef(new tree_reindex(var_decl_name_nid, TM->add(var_decl_name_nid)));
               td->name = tr_new_id;
            }
         }
         break;
      }
      case target_mem_ref_K:
      {
         const target_mem_ref * tmr = GetPointer<target_mem_ref>(curr_tn);
         if(tmr->symbol)
            recursive_examinate(tmr->symbol);
         if(tmr->base)
            recursive_examinate(tmr->base);
         if(tmr->idx)
            recursive_examinate(tmr->idx);
         break;
      }
      case target_mem_ref461_K:
      {
         const target_mem_ref461 * tmr = GetPointer<target_mem_ref461>(curr_tn);
         if(tmr->base)
            recursive_examinate(tmr->base);
         if(tmr->idx)
            recursive_examinate(tmr->idx);
         if(tmr->idx2)
            recursive_examinate(tmr->idx2);
         break;
      }
      case real_cst_K:
      case complex_cst_K:
      case string_cst_K:
      case integer_cst_K:
      case field_decl_K:
      case function_decl_K:
      case label_decl_K:
      case result_decl_K:
      case vector_cst_K:
      case tree_vec_K:
      case case_label_expr_K:
      case gimple_label_K:
      case gimple_asm_K:
      case gimple_goto_K:
      case pragma_node_K:
         break;
      default:
      {
         THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "Not supported node: " + std::string(curr_tn->get_kind_text()));
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Updated recursively " + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)));
   return;
}

std::string var_decl_fix::get_kind_text() const
{
    return "var_decl_fix";
}

