/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file create_tree_manager.cpp
 * @brief Implementation of the class for creating the tree_manager starting from the source code files
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Autoheader include
#include "config_HAVE_FROM_PRAGMA_BUILT.hpp"
#include "config_NPROFILE.hpp"

///Header include
#include "create_tree_manager.hpp"

#include "application_manager.hpp"
/// Intermediate Representation at raw level
#include "parse_tree.hpp"
#include "tree_manager.hpp"
#if HAVE_FROM_PRAGMA_BUILT
#include "pragma_manager.hpp"
#endif

///Wrapper include
#include "gcc_wrapper.hpp"

///Utility include
#include "cpu_time.hpp"
#include "global_enums.hpp"
#include "Parameter.hpp"

create_tree_manager::create_tree_manager(const ParameterConstRef _Param, const application_managerRef _AppM) :
   BaseAnalysis(_Param, _AppM, CREATE_TREE_MANAGER, 0)
{
   debug_level = Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE);
}

create_tree_manager::~create_tree_manager()
{

}

void create_tree_manager::exec()
{
   long cpu_time;
   START_TIME(cpu_time);
   const CompilerTarget compiler_target = Param->getOption<CompilerTarget>(OPT_gcc_compiler_target);
   const OptimizationSet optimization_set = Param->getOption<OptimizationSet>(OPT_gcc_optimization_set);

   const GccWrapperRef gcc_wrapper = GccWrapperRef(new GccWrapper(Param, compiler_target, optimization_set));

#if !RELEASE
   // if a XML configuration file has been specified for the GCC parameters
   if (Param->isOption(OPT_gcc_read_xml))
      gcc_wrapper->ReadXml(Param->getOption<std::string>(OPT_gcc_read_xml));
#endif
   if(!Param->isOption(OPT_input_file))
      THROW_ERROR("At least one source file has to be passed to the tool");
   std::tr1::unordered_set<std::string> input_files = Param->getOption<const std::tr1::unordered_set<std::string> >(OPT_input_file);

   PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Created list of files: " + boost::lexical_cast<std::string>(input_files.size()) + " input source code files");
#ifndef NDEBUG
   if (debug_level >= DEBUG_LEVEL_PEDANTIC)
   {
      std::tr1::unordered_set<std::string>::const_iterator file, file_end = input_files.end();
      for(file = input_files.begin(); file != file_end; file++)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "  - " + *file);
      }
   }
#endif
   if(Param->isOption(OPT_no_parse_files))
   {
      const std::tr1::unordered_set<std::string> no_parse_files = Param->getOption<const std::tr1::unordered_set<std::string> >(OPT_no_parse_files);
      PRINT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Created list of no-parse files: " + boost::lexical_cast<std::string>(no_parse_files.size()) + " input source code files not to be parsed");
      std::tr1::unordered_set<std::string>::const_iterator no_parse_file, no_parse_file_end = no_parse_files.end();
      for(no_parse_file = no_parse_files.begin(); no_parse_file != no_parse_file_end; no_parse_file++)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "  - " + *no_parse_file);
      }
   }

   ///datastructure to contain the tree representation
   tree_managerRef TreeM;

   if (Param->getOption<bool>(OPT_use_raw))
   {
#if !NPROFILE
      long int tree_time = 0;
      START_TIME(tree_time);
#endif
      if(output_level >= OUTPUT_LEVEL_MINIMUM)
      {
         if(output_level >= OUTPUT_LEVEL_VERBOSE)
         {
            INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "");
            INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "");
            INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "");
            INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "*********************************************************************************");
            INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "*               Building internal representation from raw files                 *");
            INDENT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "*********************************************************************************");
         }
         else
         {
            INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, "");
            INDENT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, " =============== Building internal representation from raw files ===============");
         }
      }
      TreeM  = tree_managerRef(new tree_manager(Param));
      const std::tr1::unordered_set<std::string> raw_files = Param->getOption<const std::tr1::unordered_set<std::string> >(OPT_input_file);
      std::tr1::unordered_set<std::string>::const_iterator raw_file, raw_file_end = raw_files.end();
      for(raw_file = raw_files.begin(); raw_file != raw_file_end; raw_file++)
      {
         if(!boost::filesystem::exists(boost::filesystem::path(*raw_file)))
         {
            THROW_ERROR("File " + *raw_file + " does not exist");
         }
         tree_managerRef TreeM_tmp = ParseTreeFile(Param, *raw_file);
         TreeM->merge_tree_managers(TreeM_tmp);
      }
#if !NPROFILE
      STOP_TIME(tree_time);
      if (output_level >= OUTPUT_LEVEL_MINIMUM)
      {
         dump_exec_time("Tree analysis time", tree_time);
      }
#endif
   }
   else
   {
#if !NPROFILE
      long int wrapping_time = 0;
      START_TIME(wrapping_time);
#endif
#if HAVE_FROM_PRAGMA_BUILT
      const pragma_managerRef PM = AppM->get_pragma_manager();
      /// Pragma init
      if (Param->getOption<bool>(OPT_parse_pragma))
      {
         PM->substitutePragma(input_files);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "Modified list of files: " + boost::lexical_cast<std::string>(input_files.size()) + " input source code files after pragma substitution");
         if (debug_level >= DEBUG_LEVEL_PEDANTIC)
         {
            std::tr1::unordered_set<std::string>::const_iterator input_file, input_file_end = input_files.end();
            for(input_file = input_files.begin(); input_file != input_file_end; input_file++)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "---" + *input_file);
            }
         }
      }
#endif
      gcc_wrapper->CreateTreeManager(TreeM, input_files);
#if !NPROFILE
      STOP_TIME(wrapping_time);
      if (debug_level >= DEBUG_LEVEL_MINIMUM)
      {
         dump_exec_time("Gcc wrapping time", wrapping_time);
      }
#endif
   }

   THROW_ASSERT(TreeM, "Error with tree manager creation");
#if !RELEASE
   if (Param->isOption(OPT_gcc_write_xml))
      gcc_wrapper->WriteXml(Param->getOption<std::string>(OPT_gcc_write_xml));
#endif

   AppM->set_tree_manager(TreeM);

   STOP_TIME(cpu_time);
   PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Behavioral specification created in " + print_cpu_time(cpu_time) + " seconds\n");
}

std::string create_tree_manager::get_kind_text() const
{
   return "create_tree_manager";
}

