/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file add_bb_flow_edges.hpp
 * @brief Analysis step which adds flow edges to basic block graphs
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/
///Header include
#include "add_bb_flow_edges.hpp"

///Behavior include
#include "basic_block.hpp"
#include "basic_blocks_graph_constructor.hpp"
#include "behavioral_helper.hpp"
#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "loop.hpp"
#include "loops.hpp"
#include "tree_basic_block.hpp"

///Graph include
#include "graph.hpp"

///Parameter include
#include "Parameter.hpp"

///STL include
#include <list>
#include <tr1/unordered_set>

///Utility include
#include "boost/lexical_cast.hpp"
#include "dbgPrintHelper.hpp"
#include "exceptions.hpp"
#include "utility.hpp"

add_bb_flow_edges::add_bb_flow_edges(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
   BaseAnalysis(_Param, _AppM, ADD_BB_FLOW_EDGES, function_id)
{
   debug_level = Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE);
   SetPreCond(ID, BB_FEEDBACK_EDGES_IDENTIFICATION);
   SetPreCond(ID, ORDER_COMPUTATION);
}

add_bb_flow_edges::~add_bb_flow_edges()
{
}

void add_bb_flow_edges::exec()
{
   ///The function behavior
   const function_behaviorRef FB = AppM->get_function_behavior(funID);

   ///The ordering among basic block in control flow graph
   const std::map<vertex, unsigned int> & bb_map_levels = FB->get_bb_map_levels();

   ///The control flow graph with feedback of basic blocks
   const bb_graph * fbb = FB->get_bb_graph(function_behavior::FBB);

   ///The loop structure
   const std::list<LoopConstRef> & loops = FB->GetLoops()->GetList();

   ///Adding edges in basic block graphs from sources of feedback edges to landing pads
   std::list<LoopConstRef>::const_iterator loop, loop_end = loops.end();
   for(loop = loops.begin(); loop != loop_end; loop++)
   {
      if((*loop)->GetId() == 0)
         continue;
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Considering loop " + boost::lexical_cast<std::string>((*loop)->GetId()));

      ///Sources of feedback loop
      std::tr1::unordered_set<vertex> sources;

      ///Landing pads
      std::tr1::unordered_set<vertex> landing_pads = (*loop)->GetLandingPadBlocks();

      ///The targets of the flow edges they can be different from landing_pads of this loop if the edge which connects a block
      ///of the loop to a landing_pads is the feedback edge of an external loop. In this case the block must be connected to the
      ///landing pads of the external loop
      std::tr1::unordered_set<vertex> targets;

      ///Header of the loop
      const vertex header = (*loop)->GetHeader();

      InEdgeIterator ei, ei_end;
      for(boost::tie(ei, ei_end) = boost::in_edges(header, *fbb); ei != ei_end; ei++)
      {
         if((FB_BB_SELECTOR) & fbb->get_selector(*ei))
         {
            sources.insert(boost::source(*ei, *fbb));
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Found source BB" + boost::lexical_cast<std::string>(CGET_BB_NODE_INFO(fbb, boost::source(*ei, *fbb))->block->number));
         }
      }
      std::tr1::unordered_set<vertex>::const_iterator l, l_end = landing_pads.end();
      for(l = landing_pads.begin(); l != l_end; l++)
      {
         vertex target = *l;
         THROW_ASSERT(bb_map_levels.find(target) != bb_map_levels.end(), "Target not found in bb_map_levels");
         THROW_ASSERT(bb_map_levels.find(header) != bb_map_levels.end(), "Header not found in bb_map_levels");
         if(bb_map_levels.find(target)->second < bb_map_levels.find(header)->second)
         {
            unsigned int external_loop_index = CGET_BB_NODE_INFO(fbb, target)->loop_id;
            THROW_ASSERT(CGET_BB_NODE_INFO(fbb, target)->block->number == external_loop_index, "Target BB" + boost::lexical_cast<std::string>(CGET_BB_NODE_INFO(fbb, boost::source(*ei, *fbb))->block->number) + " is not an header");
            const LoopConstRef external_loop = FB->GetLoops()->GetLoop(external_loop_index);
            ///Landing pads of external loop
            std::tr1::unordered_set<vertex> external_landing_pads = external_loop->GetLandingPadBlocks();
            targets.insert(external_landing_pads.begin(), external_landing_pads.end());
         }
         else
         {
            targets.insert(target);
         }
      }

      std::tr1::unordered_set<vertex>::const_iterator s, s_end = sources.end(), t, t_end = targets.end();
      for(s = sources.begin(); s != s_end; s++)
      {
         for(t = targets.begin(); t != t_end; t++)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Adding edge from BB" + boost::lexical_cast<std::string>(CGET_BB_NODE_INFO(fbb, *s)->block->number) + " to BB" + boost::lexical_cast<std::string>(CGET_BB_NODE_INFO(fbb, *t)->block->number));
            FB->bbgc->AddEdge(*s, *t, FLG_SELECTOR);
         }
      }
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Considered loop " + boost::lexical_cast<std::string>((*loop)->GetId()));
   }

   if(output_level>= OUTPUT_LEVEL_VERY_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Printing FLCFG graph");
      std::string file_name = FB->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_BB_FLCFG.dot";
      FB->get_bb_graph(function_behavior::FLCFG)->PrintToDot(Param, file_name,FB->get_helper());
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Printed FLCFG graph");
   }

#ifndef NDEBUG
   try
   {
      const bb_graph * flcfg = FB->get_bb_graph(function_behavior::FLCFG);
      std::deque<vertex> vertices;
      boost::topological_sort(*flcfg, std::front_inserter(vertices));
   }
   catch (const char* msg)
   {
      THROW_ASSERT(false, "flcfg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
   catch (const std::string & msg)
   {
      THROW_ASSERT(false, "flcfg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
   catch (const std::exception& ex)
   {
      THROW_ASSERT(false, "flcfg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
   catch ( ... )
   {
      THROW_ASSERT(false, "flcfg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
#endif
}

std::string add_bb_flow_edges::get_kind_text() const
{
   return "add_bb_flow_edges";
}
