/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file add_op_cfg_edges.hpp
 * @brief Analysis step which adds CFG edges for the operation graphs
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
///Header include
#include "add_op_cfg_edges.hpp"

///Behavior include
#include "basic_block.hpp"
#include "behavioral_helper.hpp"
#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "op_node_info.hpp"
#include "operations_graph_constructor.hpp"

///Tree include
#include "tree_basic_block.hpp"

///Parameter include
#include "Parameter.hpp"

add_op_cfg_edges::add_op_cfg_edges(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
   BaseAnalysis(_Param, _AppM, ADD_OP_CFG_EDGES, function_id)
{
   SetPreCond(ID, OP_FEEDBACK_EDGES_IDENTIFICATION);
   SetPreCond(ID, CONTROL_DEPENDENCE_COMPUTATION);
   SetPreCond(ID, ANTI_DEP_COMPUTATION);
   SetPreCond(ID, SSA_DATA_FLOW_ANALYSIS);
   SetPreCond(ID, REACHABILITY_COMPUTATION);
}

add_op_cfg_edges::~add_op_cfg_edges()
{
}

void add_op_cfg_edges::exec()
{
   const function_behaviorRef FB = AppM->get_function_behavior(funID);

   ///The control flow graph of basic blocks
   const bb_graph * fbb = FB->get_bb_graph(function_behavior::FBB);
   ///The control flow graph of operations
   const graph * fcfg = FB->get_graph(function_behavior::FCFG);

   bool is_feedback;
   EdgeIterator e, e_end;
   for (boost::tie(e, e_end) = boost::edges(*fbb); e != e_end ; e++)
   {
      vertex bb_src = boost::source(*e, *fbb);
      vertex bb_tgt = boost::target(*e, *fbb);

      const bb_node_info * target_info = CGET_BB_NODE_INFO(fbb, bb_tgt);
      const std::list<vertex>  & target_statements_list = target_info->statements_list;
      std::list<vertex>::const_iterator target_stmt_it_end = target_statements_list.end();

      const bb_node_info * source_info = CGET_BB_NODE_INFO(fbb, bb_src);
      const std::list<vertex>  & source_statements_list = source_info->statements_list;
      std::list<vertex>::const_iterator source_stmt_it_end = source_statements_list.end();

      is_feedback = FB_BB_SELECTOR & fbb->get_selector(*e);

      for(std::list<vertex>::const_iterator source_stmt_it = source_statements_list.begin(); source_stmt_it_end != source_stmt_it; ++source_stmt_it)
      {
         for(std::list<vertex>::const_iterator target_stmt_it = target_statements_list.begin(); target_stmt_it_end != target_stmt_it; ++target_stmt_it)
         {
            if(is_feedback)
               FB->ogc->add_edge(*source_stmt_it, *target_stmt_it, FB_FG_SELECTOR);
            else
               FB->ogc->add_edge(*source_stmt_it, *target_stmt_it, FLG_SELECTOR);
         }
      }
      /// check for return
      vertex last_target = *(target_statements_list.rbegin());
      if((GET_TYPE(fcfg, last_target) & TYPE_RET) != 0)
      {
         for(std::list<vertex>::const_iterator target_stmt_it = target_statements_list.begin(); target_stmt_it_end != target_stmt_it; ++target_stmt_it)
         {
            if(last_target != *target_stmt_it)
               FB->ogc->add_edge(*target_stmt_it, last_target, FLG_SELECTOR);
         }
      }
   }

   if(output_level> OUTPUT_LEVEL_PEDANTIC || Param->getOption<int>(OPT_print_dot))
   {
      std::string file_name = FB->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_FL.dot";
      FB->get_graph(function_behavior::FLG)->print_to_dot(Param, file_name, FB->get_helper());
   }
   if(output_level> OUTPUT_LEVEL_PEDANTIC || Param->getOption<int>(OPT_print_dot))
   {
      std::string file_name = FB->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_FFLSAODG.dot";
      FB->get_graph(function_behavior::FFLSAODG)->print_to_dot(Param, file_name, FB->get_helper());
   }
#ifndef NDEBUG
   try
   {
      const graph * flsaodg = FB->get_graph(function_behavior::FLSAODG);
      std::deque<vertex> vertices;
      boost::topological_sort(*flsaodg, std::front_inserter(vertices));
   }
   catch (const char* msg)
   {
      THROW_ASSERT(false, "flsaodg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
   catch (const std::string & msg)
   {
      THROW_ASSERT(false, "flsaodg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
   catch (const std::exception& ex)
   {
      THROW_ASSERT(false, "flsaodg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
   catch ( ... )
   {
      THROW_ASSERT(false, "flsaodg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
#endif
}

std::string add_op_cfg_edges::get_kind_text() const
{
   return "add_op_cfg_edges";
}
