/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file add_op_flow_edges.hpp
 * @brief Analysis step which adds flow edges for scheduling to operation graphs
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/
///Header include
#include "add_op_flow_edges.hpp"

///Algotithm include
#include "loop.hpp"
#include "loops.hpp"

///Behavior include
#include "basic_block.hpp"
#include "behavioral_helper.hpp"
#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "operations_graph_constructor.hpp"
#include "op_node_info.hpp"

///Tree include
#include "tree_basic_block.hpp"

///Parameter include
#include "Parameter.hpp"

add_op_flow_edges::add_op_flow_edges(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
   BaseAnalysis(_Param, _AppM, ADD_OP_FLOW_EDGES, function_id)
{
   SetPreCond(ID, OP_FEEDBACK_EDGES_IDENTIFICATION);
   SetPreCond(ID, CONTROL_DEPENDENCE_COMPUTATION);
   SetPreCond(ID, ANTI_DEP_COMPUTATION);
   SetPreCond(ID, SSA_DATA_FLOW_ANALYSIS);
   SetPreCond(ID, REACHABILITY_COMPUTATION);
}

add_op_flow_edges::~add_op_flow_edges()
{
}

void add_op_flow_edges::exec()
{
   const function_behaviorRef FB = AppM->get_function_behavior(funID);

   ///The control flow graph of operation
   const graph * fcfg = FB->get_graph(function_behavior::FCFG);

   ///The graph used for the scheduling
   const graph * fflsaodg = FB->get_graph(function_behavior::FFLSAODG);

   ///The control flow graph of basic blocks
   const bb_graph * fbb = FB->get_bb_graph(function_behavior::FBB);

   ///Ordering of operations
   const std::map<vertex, unsigned int> &map_levels = FB->get_map_levels();

   ///Map between bb index and bb vertex
   const std::tr1::unordered_map<unsigned int, vertex> & bb_index_map = fbb->GetConstBBGraphInfo()->bb_index_map;

   ///Operations of each loop
   std::tr1::unordered_map<unsigned int, std::tr1::unordered_set<vertex> > loop_operations;

   ///The loop structure
   const std::list<LoopConstRef> & loops = FB->GetLoops()->GetList();

   ///Adding edges from last operation of header to all operations of the loop
   std::list<LoopConstRef>::const_iterator loop, loop_end = loops.end();
   for(loop = loops.begin(); loop != loop_end; loop++)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Considering loop " + boost::lexical_cast<std::string>((*loop)->GetId()));
      ///Operations which belong to the loop
      const std::tr1::unordered_set<vertex> & loop_bb = (*loop)->get_blocks();
      std::tr1::unordered_set<vertex>::const_iterator it, it_end = loop_bb.end();
      for(it = loop_bb.begin(); it != it_end; it++)
      {
         const bb_node_info * info = CGET_BB_NODE_INFO(fbb, *it);
         ///Skip operation belonging to the header itself
         const std::list<vertex>  & statements_list = info->statements_list;
         std::list<vertex>::const_iterator it2, it2_end = statements_list.end();
         for(it2 = statements_list.begin(); it2 != it2_end; it2++)
         {
            if(info->block->number != info->loop_id)
            {
               PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "   " + GET_NAME(fcfg, *it2));
               loop_operations[(*loop)->GetId()].insert(*it2);
            }
         }
      }

      ///Searching last operation of the header
      const vertex header = (*loop)->GetHeader();
      const bb_node_info * info = CGET_BB_NODE_INFO(fbb, header);
      const std::list<vertex>  & statements_list = info->statements_list;
      THROW_ASSERT(statements_list.size(), "Header of a loop " + boost::lexical_cast<std::string>((*loop)->GetId()) + " is empty");
      const vertex last_statement = *(statements_list.rbegin());

      /// add a flow edge from the last operation of the header and the operations of the loop
      /// useful only for the speculation graph
      std::tr1::unordered_set<vertex>::const_iterator it3, it3_end = loop_operations[(*loop)->GetId()].end();
      for(it3 = loop_operations[(*loop)->GetId()].begin(); it3 != it3_end; it3++)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Adding flow edge from " + GET_NAME(fcfg, last_statement) + " to " + GET_NAME(fcfg, *it3));
         FB->ogc->add_edge(last_statement, *it3, FLG_SELECTOR);
      }
      /// add a feedback flow edge from the operations of the loop to the first statement of the header
      const vertex first_statement = *(statements_list.begin());
      for(it3 = loop_operations[(*loop)->GetId()].begin(); it3 != it3_end; it3++)
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Adding a feedback flow edge from " + GET_NAME(fcfg, *it3) + " to " + GET_NAME(fcfg, first_statement));
         FB->ogc->add_edge(*it3, first_statement, FB_FG_SELECTOR);
      }
   }

   ///Adding operation to empty return
   VertexIterator v, v_end;
   for (tie(v, v_end) = boost::vertices(*fcfg); v != v_end; ++v)
   {
      unsigned int v_bb_index = CGET_NODE_INFO(fcfg, OpNodeInfo, *v)->bb_index;
      const vertex v_bb_vertex = bb_index_map.find(v_bb_index)->second;
      if((GET_TYPE(fcfg, *v) & TYPE_RET) != 0)
      {
         VertexIterator u, u_end;
         for (boost::tie(u, u_end) = boost::vertices(*fcfg); u != u_end; u++)
         {
            unsigned int u_bb_index = CGET_NODE_INFO(fcfg, OpNodeInfo, *u)->bb_index;
            const vertex u_bb_vertex = bb_index_map.find(u_bb_index)->second;

            bool reachability = false;

            if(FB->bb_reachability[u_bb_vertex].find(v_bb_vertex) != FB->bb_reachability[u_bb_vertex].end() && boost::out_degree(*u, *fflsaodg) == 0)
            {
               reachability = true;
            }

            if(v_bb_vertex == u_bb_vertex && map_levels.find(*u)->second < map_levels.find(*v)->second)
            {
               reachability = true;
            }

            if(reachability)
               FB->ogc->add_edge(*u, *v, FLG_SELECTOR);
         }
      }
   }

   if(output_level>= OUTPUT_LEVEL_VERY_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
   {
      std::string file_name = FB->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_FL.dot";
      FB->get_graph(function_behavior::FLG)->print_to_dot(Param, file_name, FB->get_helper());
   }
   if(output_level>= OUTPUT_LEVEL_VERY_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
   {
      std::string file_name = FB->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_FFLSAODG.dot";
      FB->get_graph(function_behavior::FFLSAODG)->print_to_dot(Param, file_name, FB->get_helper());
   }
#ifndef NDEBUG
   try
   {
      const graph * flsaodg = FB->get_graph(function_behavior::FLSAODG);
      std::deque<vertex> vertices;
      boost::topological_sort(*flsaodg, std::front_inserter(vertices));
   }
   catch (const char* msg)
   {
      THROW_ASSERT(false, "flsaodg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
   catch (const std::string & msg)
   {
      THROW_ASSERT(false, "flsaodg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
   catch (const std::exception& ex)
   {
      THROW_ASSERT(false, "flsaodg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
   catch ( ... )
   {
      THROW_ASSERT(false, "flsaodg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
#endif
}

std::string add_op_flow_edges::get_kind_text() const
{
   return "add_op_flow_edges";
}
