/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file anti_dependence_computation.cpp
 * @brief Analysis step performing anti dependence computation.
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/
///Header include
#include "anti_dependence_computation.hpp"

///Behavior include
#include "application_manager.hpp"
#include "basic_block.hpp"
#include "behavioral_helper.hpp"
#include "function_behavior.hpp"
#include "operations_graph_constructor.hpp"
#include "op_node_info.hpp"

///Parameter include
#include "Parameter.hpp"

///STD include
#include <iostream>

anti_dependence_computation::anti_dependence_computation(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
      BaseAnalysis(_Param, _AppM, ANTI_DEP_COMPUTATION, function_id)
{
   this->SetPreCond(ID, REACHABILITY_COMPUTATION);
   this->SetPreCond(ID, VAR_ANALYSIS);
   this->SetPreCond(ID, SSA_DATA_FLOW_ANALYSIS);
   debug_level = Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE);
}


anti_dependence_computation::~anti_dependence_computation()
{
}

void anti_dependence_computation::exec()
{
   function_behaviorRef FB = AppM->get_function_behavior(funID);
   const graph * cfg = FB->get_graph(function_behavior::CFG);
   const behavioral_helperConstRef BH = FB->get_helper();
   const std::map<vertex, unsigned int>& map_levels = FB->get_map_levels();
   const std::map<vertex, std::set<vertex> > & bb_reachability = FB->bb_reachability;

   ///The map between index of basic block and corresponding vertex
   const std::tr1::unordered_map<unsigned int, vertex> & bb_index_map = FB->get_bb_graph(function_behavior::BB)->GetConstBBGraphInfo()->bb_index_map;

   if(output_level >= OUTPUT_LEVEL_VERY_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
   {
      std::string file_name = BH->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_Initial_ADG.dot";
      FB->get_graph(function_behavior::ADG)->print_to_dot(Param, file_name, BH);
   }

   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing uses and defines");
   //Maps between ssa_name and vertex where they are used or defined
   std::tr1::unordered_map<unsigned int, std::tr1::unordered_set<vertex> > uses;
   VertexIterator vi, vi_end;
   for (boost::tie(vi, vi_end) = boost::vertices(*cfg); vi != vi_end; vi++)
   {
      const std::list<unsigned int>& ssa_read = GET_SSA_READ(cfg, *vi);
      const std::list<unsigned int>::const_iterator vr_end = ssa_read.end();
      for (std::list<unsigned int>::const_iterator vr = ssa_read.begin(); vr != vr_end; vr++)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---" + BH->PrintVariable(*vr) + " is used in " + GET_NAME(cfg, *vi));
         uses[*vr].insert(*vi);
      }
      const std::tr1::unordered_set<unsigned int>& ssa_written = GET_SSA_WRITTEN(cfg, *vi);
      const std::tr1::unordered_set<unsigned int>::const_iterator vw_end = ssa_written.end();
      for (std::tr1::unordered_set<unsigned int>::const_iterator vw = ssa_written.begin(); vw != vw_end; vw++)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---" + BH->PrintVariable(*vw) + " is defined in " + GET_NAME(cfg, *vi));
         uses[*vw].insert(*vi);
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed uses and defines");

   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing anti dependencies");
   for (boost::tie(vi, vi_end) = boost::vertices(*cfg); vi != vi_end; vi++)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Examining variables written in " + GET_NAME(cfg, *vi));
      ///Index of the basic block of operation *vi
      unsigned int vi_bb_index = CGET_NODE_INFO(cfg, OpNodeInfo, *vi)->bb_index;

      ///Vertex of the basic block of operation *vi
      const vertex vi_bb_vertex = bb_index_map.find(vi_bb_index)->second;

      std::tr1::unordered_set<unsigned int> all_writings;

      const std::tr1::unordered_set<unsigned int>& redef = GET_REDEFINED_VAR(cfg, *vi);

      const std::tr1::unordered_set<unsigned int>& writes = GET_SSA_WRITTEN(cfg, *vi);

      all_writings.insert(redef.begin(), redef.end());
      all_writings.insert(writes.begin(), writes.end());

      const std::tr1::unordered_set<unsigned int>::const_iterator vr_end = all_writings.end();
      for (std::tr1::unordered_set<unsigned int>::const_iterator vr = all_writings.begin(); vr != vr_end; vr++)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Examining written variable " + BH->PrintVariable(*vr));
         if(uses.find(*vr) != uses.end())
         {
            const std::tr1::unordered_set<vertex> &this_uses = uses[*vr];
            const std::tr1::unordered_set<vertex>::const_iterator it_end = this_uses.end();
            for(std::tr1::unordered_set<vertex>::const_iterator it = this_uses.begin(); it != it_end; it++)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Examining use in " + GET_NAME(cfg, *it));
               bool reachability = false;
               unsigned int it_bb_index = CGET_NODE_INFO(cfg, OpNodeInfo, *it)->bb_index;
               const vertex it_bb_vertex = bb_index_map.find(it_bb_index)->second;
               ///Operations in different basic blocks
               if(bb_reachability.find(it_bb_vertex) != bb_reachability.end() && bb_reachability.find(it_bb_vertex)->second.find(vi_bb_vertex) != bb_reachability.find(it_bb_vertex)->second.end())
               {
                  reachability = true;
               }
               ///If operations are in same basic block check the ordering
               else if(it_bb_vertex == vi_bb_vertex && map_levels.find(*it)->second < map_levels.find(*vi)->second)
               {
                  reachability = true;
               }
               if(reachability)
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Antidependence between " + GET_NAME(cfg, *it) + " and " + GET_NAME(cfg, *vi));
                  FB->ogc->add_edge(*it, *vi, ADG_SELECTOR);
                  FB->ogc->add_edge_info(*it, *vi, ADG_SELECTOR, *vr);
               }
               else
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Not Antidependence between " + GET_NAME(cfg, *it) + " and " + GET_NAME(cfg, *vi));
            }
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
      }
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed anti dependencies");

   if(output_level>= OUTPUT_LEVEL_VERY_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
   {
      std::string file_name = FB->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_ADG.dot";
      FB->get_graph(function_behavior::ADG)->print_to_dot(Param, file_name, FB->get_helper());
   }

   if(output_level>= OUTPUT_LEVEL_VERY_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
   {
      std::string file_name = FB->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_SAODG.dot";
      FB->get_graph(function_behavior::SAODG)->print_to_dot(Param, file_name, FB->get_helper());
   }
#ifndef NDEBUG
   try
   {
      const graph * saodg = FB->get_graph(function_behavior::SAODG);
      std::deque<vertex> vertices;
      boost::topological_sort(*saodg, std::front_inserter(vertices));
   }
   catch (const char* msg)
   {
      THROW_ASSERT(false, "saodg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
   catch (const std::string & msg)
   {
      THROW_ASSERT(false, "saodg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
   catch (const std::exception& ex)
   {
      THROW_ASSERT(false, "saodg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
   catch ( ... )
   {
      THROW_ASSERT(false, "saodg graph of function " + FB->get_helper()->get_function_name() + " is not acyclic");
   }
#endif
}

std::string anti_dependence_computation::get_kind_text() const
{
   return "anti_dependence_computation";
}

