/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file bb_feedback_edges_computation.cpp
 * @brief Analysis step computing feedback edges of basic block control flow graph
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#include "bb_feedback_edges_computation.hpp"

#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "behavioral_helper.hpp"
#include "loops.hpp"
#include "loop.hpp"
#include "basic_block.hpp"
#include "operations_graph_constructor.hpp"
#include "basic_blocks_graph_constructor.hpp"

#include "Parameter.hpp"

bb_feedback_edges_computation::bb_feedback_edges_computation(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
      BaseAnalysis(_Param, _AppM, BB_FEEDBACK_EDGES_IDENTIFICATION, function_id)
{
   this->SetPreCond(ID, LOOPS_IDENTIFICATION);
   debug_level = Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE);
}


bb_feedback_edges_computation::~bb_feedback_edges_computation()
{
}

void bb_feedback_edges_computation::exec()
{
   function_behaviorRef FB = AppM->get_function_behavior(funID);
   const bb_graph * fbb = FB->get_bb_graph(function_behavior::FBB);
   const behavioral_helperRef & helper = FB->get_helper();
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, helper->get_function_name() + " Starting feedback edge computation");
    ///then consider loops
   std::list<LoopConstRef> loops = FB->GetLoops()->GetList();
   std::list<LoopConstRef>::const_iterator loop_end = loops.end();
   for(std::list<LoopConstRef>::const_iterator loop = loops.begin(); loop != loop_end; loop++)
   {
      if((*loop)->GetId() == 0)
         continue;

      if ((*loop)->IsReducible())
      {
         vertex bb_header = (*loop)->GetHeader();
         InEdgeIterator ei, ei_end;
         std::tr1::unordered_set<vertex> blocks;
         (*loop)->get_recursively_bb(blocks);
         for (boost::tie(ei, ei_end) = boost::in_edges(bb_header, *fbb); ei != ei_end; ei++)
         {
            vertex from_bb = boost::source(*ei, *fbb);
            if(std::find(blocks.begin(), blocks.end(), from_bb) != blocks.end())
            {
               FB->bbgc->SetEdge(from_bb, bb_header, FB_BB_SELECTOR);
            }
         }
      }
      else
      {
         if(loop_irriducible_breaks)
         {
            THROW_ERROR_CODE(IRREDUCIBLE_LOOPS_EC,"Irreducible loops not yet supported");
            THROW_ERROR("Irreducible loops not yet supported");
         }
         else
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level,  "Found an irreducible loop");
      }
   }
   if(output_level> OUTPUT_LEVEL_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
   {
      FB->get_bb_graph(function_behavior::FBB)->PrintToDot(Param, (helper->get_function_name() + "_" + boost::lexical_cast<std::string>(helper->get_function_index()) + "_BB_FCFG.dot").c_str(), helper);
   }
}

std::string bb_feedback_edges_computation::get_kind_text() const
{
   return "bb_feedback_edges_computation";
}
