/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file cdg_computation.cpp
 * @brief Analysis step performing control dependence computation.
 *
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/
///Header include
#include "cdg_computation.hpp"

///Behavior include
#include "basic_block.hpp"
#include "basic_blocks_graph_constructor.hpp"
#include "behavioral_helper.hpp"
#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "loops.hpp"
#include "op_graph.hpp"
#include "op_node_info.hpp"
#include "operations_graph_constructor.hpp"

///Graph include
#include "graph.hpp"

///Parameter include
#include "Parameter.hpp"

///STD include
#include <iostream>

///Utility include
#include "Dominance.hpp"

cdg_computation::cdg_computation(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
   BaseAnalysis(_Param, _AppM, CONTROL_DEPENDENCE_COMPUTATION, function_id)
{
   this->SetPreCond(ID, BB_FEEDBACK_EDGES_IDENTIFICATION);
   this->SetPreCond(ID, CFG_CREATION);
   this->SetPreCond(ID, DOM_POST_DOM_COMPUTATION);
   this->SetPreCond(ID, ORDER_COMPUTATION);
}


cdg_computation::~cdg_computation()
{
}

void cdg_computation::exec()
{
   function_behaviorRef FB = AppM->get_function_behavior(funID);
   const bb_graph * bb = FB->get_bb_graph(function_behavior::BB);
   graph * fcfg = FB->fcfg;
   const dominance<bb_graph> * post_dominators = FB->post_dominators;
   const behavioral_helperRef & helper = FB->get_helper();
   if(boost::num_vertices(*fcfg) == 0)
      return;

   EdgeIterator ei, ei_end;
   std::list<vertex> bb_levels;
   boost::topological_sort(*bb, std::front_inserter(bb_levels));
   std::map<vertex, unsigned int> bb_sorted;
   unsigned int counter = 0;
   for(std::list<vertex>::iterator it = bb_levels.begin(); it != bb_levels.end(); it++)
      bb_sorted[*it] = ++counter;
   //iterate over outgoing edges of the basic block CFG.
   for (boost::tie(ei, ei_end) = boost::edges(*bb); ei != ei_end ; ++ei )
   {
      vertex A = boost::source(*ei, *bb);
      vertex B = boost::target(*ei, *bb);
      InEdgeIterator pd_ei, pd_ei_end;
      vertex current_node = B;
      while (current_node && current_node != A && current_node != post_dominators->get_immediate_dominator(A))
      {
         const std::list<vertex> operations_A = CGET_BB_NODE_INFO(bb, A)->statements_list;
         const std::list<vertex> operations_B = CGET_BB_NODE_INFO(bb, current_node)->statements_list;
         if (operations_A.size() && operations_B.size())
         {
            std::list<vertex>::const_iterator op_B, op_B_end = operations_B.end();
            if(bb_sorted[current_node]>bb_sorted[A])
            {
               FB->bbgc->AddEdge(A, current_node, CDG_SELECTOR);
               std::set<unsigned int> labels = bb->get_bb_edge_info(*ei)->get_labels(BB_SELECTOR);
               if(labels.size())
               {
                  std::set<unsigned int>::iterator it_end = labels.end();
                  for(std::set<unsigned int>::iterator it = labels.begin(); it != it_end; it++)
                  {
                     FB->bbgc->add_bb_edge_info(A, current_node, CDG_SELECTOR, *it);
                  }
               }
               for (op_B = operations_B.begin();op_B != op_B_end; op_B++)
               {
                  FB->ogc->add_edge(operations_A.back(), *op_B, CDG_SELECTOR);
                  if (labels.size())
                  {
                     std::set<unsigned int>::iterator it_end = labels.end();
                     for(std::set<unsigned int>::iterator it = labels.begin(); it != it_end; it++)
                        FB->ogc->add_edge_info(operations_A.back(), *op_B, CDG_SELECTOR, *it);
                  }
               }
            }
            else
            {
               break;
            }
         }
         current_node = post_dominators->get_immediate_dominator(current_node);
      }
   }

   bb_graph * cdg_bb = FB->cdg_bb;

   //Counter used to enumerate different
   unsigned int cer_counter = 0;
   //Map control equivalent region codification to control equivalent index;
   //The codification is the set of pair predecessor-edge label in the cdg_computation
   std::map<std::set<std::pair<vertex, std::set<unsigned int> > >, unsigned int> cdg_to_index;

   const std::deque<vertex> & topological_sorted_nodes = FB->get_bb_levels();
   std::deque<vertex>::const_iterator it, it_end;
   it_end = topological_sorted_nodes.end();
   for(it = topological_sorted_nodes.begin(); it != it_end; it++)
   {
      unsigned int cer_index = cer_counter;
      bb_node_info * binfo = GET_BB_NODE_INFO(cdg_bb, *it);
      if (boost::in_degree(*it, *cdg_bb) > 0)
      {
         //codification of this basic block
         std::set<std::pair<vertex, std::set<unsigned int> > > this_cod;
         InEdgeIterator eii, eii_end;
         for(boost::tie(eii, eii_end) = boost::in_edges(*it, *cdg_bb); eii != eii_end; eii++)
         {
            this_cod.insert(std::pair<vertex, std::set<unsigned int> >(boost::source(*eii, *cdg_bb), cdg_bb->get_bb_edge_info(*eii)->get_labels(CDG_SELECTOR)));
         }
         if(cdg_to_index.find(this_cod) == cdg_to_index.end())
         {
            cdg_to_index[this_cod] = cer_counter;
            cer_counter++;
         }
         else
         {
            cer_index = cdg_to_index[this_cod];
         }
         binfo->cer = cer_index;
         std::list<vertex> statements_list = binfo->statements_list;
         std::list<vertex>::iterator it2, it2_end = statements_list.end();
         for(it2 = statements_list.begin(); it2 != it2_end; it2++)
         {
           OpNodeInfo * cinfo = GET_NODE_INFO(fcfg, OpNodeInfo, *it2);
           cinfo->cer = cer_index;
         }
      }
   }

   if(output_level>= OUTPUT_LEVEL_VERY_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
   {
      const graph * cdg = FB->get_graph(function_behavior::CDG);
      vertex entry_vertex = GET_ENTRY_VERTEX(cdg);
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level,  "* Number of operations controlled by entry: " + boost::lexical_cast<std::string>(boost::out_degree(entry_vertex, *cdg)));
      OutEdgeIterator eo, eo_end;
      unsigned int controlled_areas=0;
      for (boost::tie(eo, eo_end) = boost::out_edges(entry_vertex, *cdg); eo != eo_end; eo++)
      {
         vertex to_v = boost::target(*eo, *cdg);
         unsigned int vtype = GET_TYPE(cdg, to_v);
         if(vtype & (TYPE_IF | TYPE_SWITCH | TYPE_WHILE | TYPE_FOR))
            controlled_areas++;
      }
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level,  "* Number of controlled areas: " + boost::lexical_cast<std::string>(controlled_areas));

      FB->get_bb_graph(function_behavior::CDG_BB)->PrintToDot(Param, helper->get_function_name() + "_" + boost::lexical_cast<std::string>(helper->get_function_index()) + "_BB_CDG.dot",helper);

      std::string file_name = FB->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_CDG.dot";
      FB->get_graph(function_behavior::CDG)->print_to_dot(Param, file_name, helper);
      FB->get_graph(function_behavior::FCDG)->print_to_dot(Param, file_name, helper);
   }
}

std::string cdg_computation::get_kind_text() const
{
   return "cdg_computation";
}
