/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file dom_post_dom_computation.cpp
 * @brief Analysis step performing dominators and post dominators computation.
 *
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/
///Header include
#include "dom_post_dom_computation.hpp"

#include "application_manager.hpp"
#include "graph.hpp"
#include "function_behavior.hpp"
#include "behavioral_helper.hpp"
#include "basic_blocks_graph_constructor.hpp"
#include "Dominance.hpp"
#include "basic_block.hpp"
#include "Parameter.hpp"
#include <tr1/unordered_map>

dom_post_dom_computation::dom_post_dom_computation(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
      BaseAnalysis(_Param, _AppM, DOM_POST_DOM_COMPUTATION, function_id)
{
   this->SetPreCond(ID, BASIC_BLOCKS_CFG_COMPUTATION);
   debug_level = Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE);
}


dom_post_dom_computation::~dom_post_dom_computation()
{

}

void dom_post_dom_computation::exec()
{
   function_behaviorRef FB = AppM->get_function_behavior(funID);
   const bb_graph * fbb = FB->get_bb_graph(function_behavior::FBB);

   const behavioral_helperRef & helper = FB->get_helper();
   ///dominators computation
   THROW_ASSERT(!FB->dominators, "Dominators already built");
   const vertex bbentry = fbb->GetConstBBGraphInfo()->entry_vertex;
   const vertex bbexit = fbb->GetConstBBGraphInfo()->exit_vertex;
   FB->dominators = new dominance<bb_graph>(*fbb, bbentry, bbexit, Param);
   FB->dominators->calculate_dominance_info(dominance<bb_graph>::CDI_DOMINATORS);
   std::tr1::unordered_map<vertex, vertex> dominator_map = FB->dominators->get_dominator_map();
   for(std::tr1::unordered_map<vertex, vertex>::iterator it = dominator_map.begin(); it != dominator_map.end(); it++)
   {
      if(it->first != bbentry)
      {
         FB->bbgc->AddEdge(it->second, it->first, D_BB_SELECTOR);
      }
   }
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Built dominators tree of " + helper->get_function_name());
   if(output_level>= OUTPUT_LEVEL_VERY_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
   {
      behavioral_helperRef null_helper;
      FB->get_bb_graph(function_behavior::DOM_TREE)->PrintToDot(Param, (helper->get_function_name() + "_" + boost::lexical_cast<std::string>(helper->get_function_index()) + "_BB_dom_tree.dot").c_str(),null_helper);
   }
   ///post-dominators computation
   THROW_ASSERT(!FB->post_dominators, "Post dominators yet built");
   FB->post_dominators = new dominance<bb_graph>(*fbb, bbentry, bbexit, Param);
   FB->post_dominators->calculate_dominance_info(dominance<bb_graph>::CDI_POST_DOMINATORS);
   std::tr1::unordered_map<vertex, vertex> post_dominator_map = FB->post_dominators->get_dominator_map();
   for(std::tr1::unordered_map<vertex, vertex>::iterator it = post_dominator_map.begin(); it != post_dominator_map.end(); it++)
   {
      if(it->first != bbexit)
      {
         FB->bbgc->AddEdge(it->second, it->first, PD_BB_SELECTOR);
      }
   }
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Built post-dominators tree of " + helper->get_function_name());

   if(output_level >= OUTPUT_LEVEL_VERY_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
   {
      behavioral_helperRef null_helper;
      FB->get_bb_graph(function_behavior::POST_DOM_TREE)->PrintToDot(Param, (helper->get_function_name() + "_" + boost::lexical_cast<std::string>(helper->get_function_index()) + "_BB_post_dom_tree.dot").c_str(),null_helper);
   }
}
