/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file loop_regions_flow_computation.cpp
 * @brief Analysis step performing edge flow computation for loop regions.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/

#include "loop_regions_flow_computation.hpp"
#include "application_manager.hpp"
#include "graph.hpp"
#include "function_behavior.hpp"
#include "behavioral_helper.hpp"
#include "basic_block.hpp"
#include "loops.hpp"
#include "loop.hpp"
#include "operations_graph_constructor.hpp"
#include "tree_basic_block.hpp"
#include "Parameter.hpp"

loop_regions_flow_computation::loop_regions_flow_computation(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
      BaseAnalysis(_Param, _AppM, LOOP_REGIONS_FLOW_COMPUTATION, function_id)
{
   this->SetPreCond(ID, OP_FEEDBACK_EDGES_IDENTIFICATION);
   this->SetPreCond(ID, LOOP_REGIONS_COMPUTATION);
   this->SetPreCond(ID, CONTROL_DEPENDENCE_COMPUTATION);
   debug_level = Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE);
}


loop_regions_flow_computation::~loop_regions_flow_computation()
{

}

void loop_regions_flow_computation::exec()
{
   const function_behaviorRef FB = AppM->get_function_behavior(funID);
   const bb_graph * fbb = FB->get_bb_graph(function_behavior::FBB);
   const bb_graph * dom = FB->get_bb_graph(function_behavior::CDG_BB);
   const behavioral_helperRef & helper = FB->get_helper();

   ///first consider the label
   std::list<LoopConstRef> loops = FB->GetLoops()->GetList();
   std::list<LoopConstRef>::const_iterator loop_end = loops.end();
   for(std::list<LoopConstRef>::const_iterator loop = loops.begin(); loop != loop_end; loop++)
   {
      if ((*loop)->IsReducible())
      {
         if((*loop)->GetId() != 0 and (((*loop)->loop_type & WHILE_LOOP) == 0 or ((*loop)->loop_type & SINGLE_EXIT_LOOP) == 0))
         {
            vertex header = (*loop)->GetHeader();
            const bb_node_info* bi = CGET_BB_NODE_INFO(fbb, header);
            //THROW_ASSERT((*loop)->num_blocks() != 1 || helper->end_with_a_cond_or_goto(bi->block), "malformed loop region "+boost::lexical_cast<std::string>(bi->block->number));

            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Examing loop " + boost::lexical_cast<std::string>(bi->block->number));
            InEdgeIterator ei, ei_end;
            boost::tie(ei, ei_end) = boost::in_edges(header, *dom);
            THROW_ASSERT(boost::in_degree(header, *dom) >= 1, "Header of loop " + boost::lexical_cast<std::string>(bi->block->number) + " hasn't controller");

            vertex header_dom = boost::source(*ei, *dom);
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Controller of header is BB" + boost::lexical_cast<std::string>(CGET_BB_NODE_INFO(dom, header_dom)->block->number));
            if(helper->start_with_a_label(bi->block))
            {
               std::list<vertex>::const_iterator vcur_it_end = bi->statements_list.end();
               std::list<vertex>::const_iterator vcur_it = bi->statements_list.begin();
               if(vcur_it_end != vcur_it)
               {
                  vertex label_vertex = *vcur_it;
                  vcur_it++;
                  for(; vcur_it != vcur_it_end; vcur_it++)
                  {
                     FB->ogc->add_edge(label_vertex, *vcur_it, FLG_SELECTOR);
                  }
                  OutEdgeIterator eo, eo_end;
                  for(boost::tie(eo, eo_end) = boost::out_edges(header_dom, *dom); eo != eo_end; eo++)
                  {
                     PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Controller controlls BB" + boost::lexical_cast<std::string>(CGET_BB_NODE_INFO(dom, boost::target(*eo, *dom))->block->number));
                     if(CGET_BB_NODE_INFO(dom, boost::target(*eo, *dom))->loop_id == bi->block->number && CGET_BB_NODE_INFO(dom, boost::target(*eo, *dom))->block->number != bi->block->number)
                     {
                        const bb_node_info* bii = CGET_BB_NODE_INFO(fbb, boost::target(*eo, *dom));
                        std::list<vertex>::const_iterator vcur_iit_end = bii->statements_list.end();
                        std::list<vertex>::const_iterator vcur_iit;
                        for(vcur_iit = bii->statements_list.begin(); vcur_iit != vcur_iit_end; vcur_iit++)
                        {
                           FB->ogc->add_edge(label_vertex, *vcur_iit, FLG_SELECTOR);
                        }
                     }
                  }
               }
               else
                  THROW_ERROR("something of wrong happened");
            }
            else
               THROW_ERROR("something of wrong happened");
         }
      }
      else
      {
         if(loop_irriducible_breaks)
         {
            THROW_ERROR_CODE(IRREDUCIBLE_LOOPS_EC,"Irreducible loops not yet supported");
            THROW_ERROR("Irreducible loops not yet supported");
         }
         else
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level,  "Found an irreducible loop");
      }
   }
   ///and then the gotos
   for(std::list<LoopConstRef>::const_iterator loop = loops.begin(); loop != loop_end; loop++)
   {
      if ((*loop)->IsReducible())
      {
         if((*loop)->GetId() != 0 and (((*loop)->loop_type & WHILE_LOOP) == 0 or ((*loop)->loop_type & SINGLE_EXIT_LOOP) == 0))
         {
            vertex bb_header = (*loop)->GetHeader();
#ifndef NDEBUG
            const bb_node_info* bi = CGET_BB_NODE_INFO(fbb, bb_header);
            unsigned int label_expr_nid = helper->start_with_a_label(bi->block);
            THROW_ASSERT(label_expr_nid, "gimple_label has to be inserted into the intermediate representation");
#endif
            InEdgeIterator ei, ei_end;
            const std::tr1::unordered_set<vertex> blocks = (*loop)->get_blocks();
            std::tr1::unordered_set<vertex>::const_iterator bb_it = blocks.begin(), bb_it_end = blocks.end();
            for (boost::tie(ei, ei_end) = boost::in_edges(bb_header, *fbb); ei != ei_end; ei++)
            {
               vertex from_bb = boost::source(*ei, *fbb);
               if(std::find(bb_it, bb_it_end, from_bb) != bb_it_end)
               {
                  const bb_node_info* bi_from = CGET_BB_NODE_INFO(fbb, from_bb);
                  if(helper->end_with_a_cond_or_goto(bi_from->block))
                  {
                     vertex goto_vertex = bi_from->statements_list.back();
   #if 1
                     std::tr1::unordered_set<vertex>::const_iterator bb_it_local = bb_it;
                     while(bb_it_local != bb_it_end)
                     {
                        const std::list<vertex> &sl = CGET_BB_NODE_INFO(fbb, *bb_it_local)->statements_list;

                        std::list<vertex>::const_iterator vcur_it_end = sl.end();
                        for(std::list<vertex>::const_iterator vcur_it = sl.begin(); vcur_it != vcur_it_end; vcur_it++)
                           if(*vcur_it != goto_vertex)
                              FB->ogc->add_edge(*vcur_it, goto_vertex, FLG_SELECTOR);
                        bb_it_local++;
                     }
   #else
                     const std::list<vertex> &sl = bi_from->statements_list;
                     std::list<vertex>::const_iterator vcur_it_end = sl.end();
                     std::list<vertex>::const_iterator vcur_it = sl.begin();
                     for(std::list<vertex>::const_iterator vcur_it = sl.begin(); vcur_it != vcur_it_end; vcur_it++)
                        if(*vcur_it != goto_vertex)
                           FB->ogc->add_edge(*vcur_it, goto_vertex, FLG_SELECTOR);
   #endif
                  }
                  else
                     THROW_ERROR("something of wrong happened");
               }
            }
         }
      }
      else
      {
         if(loop_irriducible_breaks)
         {
            THROW_ERROR_CODE(IRREDUCIBLE_LOOPS_EC,"Irreducible loops not yet supported");
            THROW_ERROR("Irreducible loops not yet supported");
         }
         else
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level,  "Found an irreducible loop");
      }
   }
   if(output_level>= OUTPUT_LEVEL_VERY_PEDANTIC || Param->getOption<bool>(OPT_print_dot))
   {
      std::string file_name = AppM->get_function_behavior(funID)->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_FG.dot";
      AppM->get_function_behavior(funID)->get_graph(function_behavior::FFLSADG)->print_to_dot(Param, file_name, helper);
   }
}
