/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file op_feedback_edges_computation.cpp
 * @brief Analysis step computing Analysis step computing feedback edges for operation control flow graph
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/
///Header include
#include "op_feedback_edges_computation.hpp"

///Behavior include
#include "behavioral_helper.hpp"
#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "loop.hpp"
#include "loops.hpp"

///Graph include
#include "basic_block.hpp"
#include "operations_graph_constructor.hpp"
#include "tree_basic_block.hpp"

///Frontend include
#include "Parameter.hpp"

op_feedback_edges_computation::op_feedback_edges_computation(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
   BaseAnalysis(_Param, _AppM, OP_FEEDBACK_EDGES_IDENTIFICATION, function_id)
{
   this->SetPreCond(ID, CFG_CREATION);
}


op_feedback_edges_computation::~op_feedback_edges_computation()
{

}

void op_feedback_edges_computation::exec()
{
   const function_behaviorRef FB = AppM->get_function_behavior(funID);
   const bb_graph * fbb = FB->get_bb_graph(function_behavior::FBB);
    ///then consider loops
   std::list<LoopConstRef> loops = FB->GetLoops()->GetList();
   std::list<LoopConstRef>::const_iterator loop_end = loops.end();
   for(std::list<LoopConstRef>::const_iterator loop = loops.begin(); loop != loop_end; loop++)
   {
      if((*loop)->GetId() == 0)
         continue;
      if ((*loop)->IsReducible())
      {
         vertex bb_header = (*loop)->GetHeader();
         const bb_node_info* bi = CGET_BB_NODE_INFO(fbb, bb_header);
         vertex label_vertex = bi->statements_list.front();
         InEdgeIterator ei, ei_end;
         std::tr1::unordered_set<vertex> blocks;
         (*loop)->get_recursively_bb(blocks);
         for (boost::tie(ei, ei_end) = boost::in_edges(bb_header, *fbb); ei != ei_end; ei++)
         {
            vertex from_bb = boost::source(*ei, *fbb);
            if(std::find(blocks.begin(), blocks.end(), from_bb) != blocks.end())
            {
               const bb_node_info* bi_from = CGET_BB_NODE_INFO(fbb, from_bb);
               THROW_ASSERT(bi_from->statements_list.size(), "Empty block " + boost::lexical_cast<std::string>(bi_from->block->number));
               vertex goto_vertex = bi_from->statements_list.back();
               ///add the feedback control dependence and the feedback control flow graph edges
               FB->ogc->set_edge_type(goto_vertex, label_vertex, FB_CDG_SELECTOR|FB_CFG_SELECTOR);
            }
         }
      }
      else
      {
         if(loop_irriducible_breaks)
         {
            THROW_ERROR_CODE(IRREDUCIBLE_LOOPS_EC,"Irreducible loops not yet supported");
            THROW_ERROR("Irreducible loops not yet supported");
         }
         else
            PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level,  "Found an irreducible loop");
      }
   }

   if(output_level >= OUTPUT_LEVEL_VERBOSE || Param->getOption<bool>(OPT_print_dot))
   {
      std::string file_name = FB->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_FCFG.dot";
      FB->get_graph(function_behavior::FCFG)->print_to_dot(Param, file_name, FB->get_helper());
      file_name = FB->get_helper()->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_CFG.dot";
      FB->get_graph(function_behavior::CFG)->print_to_dot(Param, file_name, FB->get_helper());
   }
}

std::string op_feedback_edges_computation::get_kind_text() const
{
   return "op_feedback_edges_computation";
}
