/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file order_computation.cpp
 * @brief Analysis step computing a topological order of the operations.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Header include
#include "order_computation.hpp"

///Behavior include
#include "application_manager.hpp"
#include "basic_block.hpp"
#include "basic_blocks_graph_constructor.hpp"
#include "behavioral_helper.hpp"
#include "function_behavior.hpp"
#include "level_constructor.hpp"
#include "operations_graph_constructor.hpp"

///Tree include
#include "tree_manager.hpp"

///Graph include
#include "graph.hpp"

///Parameter include
#include "Parameter.hpp"

///STD include
#include <list>

order_computation::order_computation(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
   BaseAnalysis(_Param, _AppM, ORDER_COMPUTATION, function_id)
{
   debug_level = Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE);
   this->SetPreCond(ID, OP_FEEDBACK_EDGES_IDENTIFICATION);
}


order_computation::~order_computation()
{
}

void order_computation::exec()
{
   function_behaviorRef FB = AppM->get_function_behavior(funID);
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Starting order computation on function " + FB->get_helper()->get_function_name());
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Starting order computation on Operation CFG");
   const graph * cfg = FB->get_graph(function_behavior::CFG);
   graph::vertex_iterator v, v_end;
   std::list<vertex> to_visit;
   unsigned int index;
   /// Mark for visit in different functions.
   std::map<graph::vertex_descriptor, bool> MARK;

   /// MARK initialization
   for (tie(v, v_end) = boost::vertices(*cfg); v != v_end; ++v)
   {
      MARK[*v] = false;
   }
   //Vertex list to be visited
   to_visit.push_front(FB->ogc->CgetIndex(ENTRY));
   index = 0;
   while(!to_visit.empty())
   {
      vertex actual = to_visit.front();
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Checking vertex " + GET_NAME(cfg, actual) + " : " + boost::lexical_cast<std::string>(index));
      to_visit.pop_front();
      MARK[actual] = true;
      FB->lm->add(actual, index++);
      graph::out_edge_iterator o, o_end;
      graph::in_edge_iterator i, i_end;
      vertex then = 0;
      ///Examining successors
      for(tie(o, o_end) = boost::out_edges(actual, *cfg);o != o_end;o++)
      {
         bool toadd = true;
         vertex next = boost::target(*o, *cfg);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Considering successor " + GET_NAME(cfg, next));
         ///Checking if all successor's predecessors have been examinated
         for(tie(i, i_end) = boost::in_edges(next, *cfg); i!= i_end; i++)
         {
            if(!MARK[boost::source(*i, *cfg)])
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Not adding because of predecessor " + GET_NAME(cfg, boost::source(*i, *cfg)));
               toadd = false;
               break;
            }
         }
         if(toadd)
         {
            if(Cget_edge_info<OpEdgeInfo>(*o, *cfg) && CFG_TRUE_CHECK(cfg, *o))
               then = next;
            ///Vertex can be added to list
            if(next != then)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Adding next " + GET_NAME(cfg, next));
               to_visit.push_front(next);
            }
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
      }
      if(then)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Adding then " + GET_NAME(cfg, then));
         to_visit.push_front(then);
      }
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
   }
   ///Checking if EXIT has been added
   const std::map<vertex, unsigned int> &map_levels = FB->get_map_levels();
   if(map_levels.find(FB->ogc->CgetIndex(EXIT)) == map_levels.end())
   {
      FB->lm->add(FB->ogc->CgetIndex(EXIT), index++);
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");


   const bb_graph * bb_flcfg = FB->get_bb_graph(function_behavior::FLCFG);
   bb_graph::vertex_iterator bb_v, bb_v_end;

   /// MARK initialization
   for (tie(v, v_end) = boost::vertices(*bb_flcfg); v != v_end; ++v)
   {
      MARK[*v] = false;
   }
   ///Vertex list to be visited
   to_visit.push_front(bb_flcfg->GetConstBBGraphInfo()->entry_vertex);
   index = 0;
   while(!to_visit.empty())
   {
      vertex actual = to_visit.front();
      to_visit.pop_front();
      MARK[actual] = true;
      FB->bb_lm->add(actual, index++);
      graph::out_edge_iterator o, o_end;
      graph::in_edge_iterator i, i_end;
      vertex then = 0;
      ///Examining successors
      for(tie(o, o_end) = boost::out_edges(actual, *bb_flcfg);o != o_end;o++)
      {
         bool toadd = true;
         vertex next = boost::target(*o, *bb_flcfg);
         ///Checking if all successor's predecessors have been examinated
         for(tie(i, i_end) = boost::in_edges(next, *bb_flcfg); i!= i_end; i++)
         {
            if(!MARK[boost::source(*i, *bb_flcfg)])
            {
               toadd = false;
               break;
            }
         }
         if(toadd)
         {
            if(bb_flcfg->get_bb_edge_info(*o)->cfg_edge_T())
               then = next;
            ///Vertex can be added to list
            if(next != then)
            {
               to_visit.push_front(next);
            }
         }
      }
      if(then)
      {
         to_visit.push_front(then);
      }
   }

   ///Checking if EXIT has been added
   const std::map<vertex, unsigned int> & bb_map_levels = FB->get_bb_map_levels();
   if(bb_map_levels.find(bb_flcfg->GetConstBBGraphInfo()->exit_vertex) == bb_map_levels.end())
   {
      FB->bb_lm->add(bb_flcfg->GetConstBBGraphInfo()->exit_vertex, index++);
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
}
