/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file reachability_computation.cpp
 * @brief Analysis step computing reachability between operations
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#include "reachability_computation.hpp"

#include "application_manager.hpp"
#include "function_behavior.hpp"
#include "graph.hpp"
#include "basic_block.hpp"
#include "tree_basic_block.hpp"
#include <boost/graph/topological_sort.hpp>
#include <tr1/unordered_map>
#include <tr1/unordered_set>

#include "Parameter.hpp"

reachability_computation::reachability_computation(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
      BaseAnalysis(_Param, _AppM, REACHABILITY_COMPUTATION, function_id)
{
   debug_level = _Param->get_class_debug_level("reachability_computation", DEBUG_LEVEL_NONE);
   this->SetPreCond(ID, BB_FEEDBACK_EDGES_IDENTIFICATION);
   this->SetPreCond(ID, ADD_BB_FLOW_EDGES);
}


reachability_computation::~reachability_computation()
{
}

void reachability_computation::exec()
{
   ///The function behavior
   const function_behaviorRef FB = AppM->get_function_behavior(funID);

   ///The extended controlo flow graph of basic block (CFG + flow edges
   const bb_graph * flcfg = FB->get_bb_graph(function_behavior::FLCFG);

   ///The reachability among basic blocks
   std::map<vertex, std::set<vertex> > & bb_reachability = FB->bb_reachability;

   std::deque<vertex> container;
   boost::topological_sort(*flcfg, std::back_inserter(container));
   std::deque<vertex>::const_iterator it, it_end;
   it_end = container.end();
   for(it = container.begin(); it != it_end; it++)
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "  Examining basic block " + boost::lexical_cast<std::string>(CGET_BB_NODE_INFO(flcfg, *it)->block->number));
      OutEdgeIterator eo, eo_end;
      for(boost::tie(eo, eo_end) = boost::out_edges(*it, *flcfg); eo != eo_end; eo++)
      {
         vertex previous = boost::target(*eo, *flcfg);
         bb_reachability[*it].insert(previous);
         bb_reachability[*it].insert(bb_reachability[previous].begin(), bb_reachability[previous].end());
      }
   }
}

std::string reachability_computation::get_kind_text() const
{
   return "reachability_computation";
}
