/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file ssa_data_dependence_computation.cpp
 * @brief Analysis step performing data flow analysis based on ssa variables
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10609 $
 * $Date: 2012-03-12 11:40:21 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: pilato $
 *
 */

///Header include 
#include "ssa_data_dependence_computation.hpp"

///Behavior include
#include "application_manager.hpp"
#include "basic_block.hpp"
#include "behavioral_helper.hpp"
#include "function_behavior.hpp"
#include "op_node_info.hpp"
#include "operations_graph_constructor.hpp"
#include "loop.hpp"
#include "loops.hpp"

///Parameter include
#include "Parameter.hpp"

///STD include
#include <iostream>

ssa_data_dependence_computation::ssa_data_dependence_computation(const ParameterConstRef _Param, const application_managerRef _AppM, unsigned int function_id) :
   BaseAnalysis(_Param, _AppM, SSA_DATA_FLOW_ANALYSIS, function_id)
{
   debug_level = Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE);
   this->SetPreCond(ID, REACHABILITY_COMPUTATION);
   this->SetPreCond(ID, VAR_ANALYSIS);
}


ssa_data_dependence_computation::~ssa_data_dependence_computation()
{
}

void ssa_data_dependence_computation::exec()
{
   const function_behaviorRef FB = AppM->get_function_behavior(funID);
   const behavioral_helperConstRef BH = FB->get_helper();
   const graph * cfg = FB->get_graph(function_behavior::CFG);

   const std::map<vertex, std::set<vertex> > & bb_reachability = FB->bb_reachability;

   const std::map<vertex, unsigned int> &map_levels = FB->get_map_levels();

   const std::tr1::unordered_map<unsigned int, vertex> & bb_index_map = FB->get_bb_graph(function_behavior::BB)->GetConstBBGraphInfo()->bb_index_map;

   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing definitions");
   //Maps between ssa_name and vertex where they are defined
   std::map<unsigned int, std::set<vertex> > defs;
   VertexIterator vi, vi_end;
   for (boost::tie(vi, vi_end) = boost::vertices(*cfg); vi != vi_end; vi++)
   {
      std::tr1::unordered_set<unsigned int> ssa_written = GET_SSA_WRITTEN(cfg, *vi);
      std::tr1::unordered_set<unsigned int>::const_iterator vw_end = ssa_written.end();
      for (std::tr1::unordered_set<unsigned int>::iterator vw = ssa_written.begin(); vw != vw_end; vw++)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---" + BH->PrintVariable(*vw) + " is defined in " + GET_NAME(cfg, *vi));
         defs[*vw].insert(*vi);
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed definitions");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Computing dependencies");
   for (boost::tie(vi, vi_end) = boost::vertices(*cfg); vi != vi_end; vi++)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Analyzing uses in " + GET_NAME(cfg, *vi));
      unsigned int vi_bb_index = CGET_NODE_INFO(cfg, OpNodeInfo, *vi)->bb_index;
      const vertex vi_bb_vertex = bb_index_map.find(vi_bb_index)->second;

      std::list<unsigned int> uses = GET_SSA_READ(cfg, *vi);
      std::list<unsigned int>::const_iterator vr_end = uses.end();
      for (std::list<unsigned int>::const_iterator vr = uses.begin(); vr != vr_end; vr++)
      {
         //TODO: this should become check on default flag of ssa_name
         if(defs.find(*vr) != defs.end())
         {
            std::set<vertex> this_defs = defs[*vr];
            std::set<vertex>::iterator it_end = this_defs.end();
            for(std::set<vertex>::iterator it = this_defs.begin(); it != it_end; it++)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Check if definition of " + BH->PrintVariable(*vr) + " in " + GET_NAME(cfg, *it) + " reaches " + GET_NAME(cfg, *vi));
               unsigned int it_bb_index = CGET_NODE_INFO(cfg, OpNodeInfo, *it)->bb_index;
               const vertex it_bb_vertex = bb_index_map.find(it_bb_index)->second;

               bool forward_dependence = false;
               bool feedback_dependence = false;

               if((bb_reachability.find(it_bb_vertex) != bb_reachability.end() and bb_reachability.find(it_bb_vertex)->second.find(vi_bb_vertex) != bb_reachability.find(it_bb_vertex)->second.end()))
               {
                  forward_dependence = true;
               }

               if ((bb_reachability.find(vi_bb_vertex) != bb_reachability.end() and bb_reachability.find(vi_bb_vertex)->second.find(it_bb_vertex) != bb_reachability.find(vi_bb_vertex)->second.end()))
               {
                  feedback_dependence = true;
               }

               if(vi_bb_vertex == it_bb_vertex)
               {
                  THROW_ASSERT(map_levels.find(*it) != map_levels.end() and map_levels.find(*vi) != map_levels.end(), "Something wrong in map levels data structure");
                  if(*it == *vi)
                     feedback_dependence = true;
                  else if(map_levels.find(*it)->second < map_levels.find(*vi)->second)
                  {
                     forward_dependence = true;
                  }
                  else
                     feedback_dependence = true;
               }
               THROW_ASSERT(!(forward_dependence and feedback_dependence), "Dependence between operation " + GET_NAME(cfg, *it) + " and " + GET_NAME(cfg, *vi) + " is in both the direction");
               if(forward_dependence)
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Data dependence between " + GET_NAME(cfg, *it) + " and " + GET_NAME(cfg, *vi));
                  FB->ogc->add_edge(*it, *vi, DFG_SELECTOR);
                  FB->ogc->add_edge_info(*it, *vi, DFG_SELECTOR, *vr);
               }

               if(feedback_dependence)
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Antidependence(added in ssa_data_dependence) between " + GET_NAME(cfg, *vi) + " and " + GET_NAME(cfg, *it));
                  FB->ogc->add_edge(*it, *vi, FB_DFG_SELECTOR);
                  FB->ogc->add_edge_info(*it, *vi, DFG_SELECTOR, *vr);
                  if(*vi != *it)
                  {
                     FB->ogc->add_edge(*vi, *it, ADG_SELECTOR);
                     FB->ogc->add_edge_info(*vi, *it, ADG_SELECTOR, *vr);
                  }
               }
            }
         }
      }
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Computed dependencies");

   if(output_level >= OUTPUT_LEVEL_VERBOSE || Param->getOption<bool>(OPT_print_dot))
   {
      std::string file_name = BH->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_DFG.dot";
      FB->get_graph(function_behavior::DFG)->print_to_dot(Param, file_name, BH);
   }

   if(output_level >= OUTPUT_LEVEL_VERBOSE || Param->getOption<bool>(OPT_print_dot))
   {
      std::string file_name = BH->get_function_name() + "_" + boost::lexical_cast<std::string>(funID) + "_OP_FDFG.dot";
      FB->get_graph(function_behavior::FDFG)->print_to_dot(Param, file_name, BH);
   }
#ifndef NDEBUG
   try
   {
      const graph * dfg = FB->get_graph(function_behavior::DFG);
      std::deque<vertex> vertices;
      boost::topological_sort(*dfg, std::front_inserter(vertices));
   }
   catch (const char* msg)
   {
      THROW_ASSERT(false, "dfg graph of function " + BH->get_function_name() + " is not acyclic");
   }
   catch (const std::string & msg)
   {
      THROW_ASSERT(false, "dfg graph of function " + BH->get_function_name() + " is not acyclic");
   }
   catch (const std::exception& ex)
   {
      THROW_ASSERT(false, "dfg graph of function " + BH->get_function_name() + " is not acyclic");
   }
   catch ( ... )
   {
      THROW_ASSERT(false, "dfg graph of function " + BH->get_function_name() + " is not acyclic");
   }
#endif

}

std::string ssa_data_dependence_computation::get_kind_text() const
{
   return "ssa_data_dependence_computation";
}
