/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file graph.hpp
 * @brief Class specification of the graph structures.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef GRAPH_HPP
#define GRAPH_HPP

///Graph include
#include "node_info.hpp"
#include "edge_info.hpp"
#include "graph_info.hpp"

///STD include
#include <fstream>
#include <ostream>

///Utility include
#include <boost/config.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/filtered_graph.hpp>
#include <boost/graph/graph_utility.hpp>
#include <boost/graph/graphviz.hpp>
#include <boost/graph/topological_sort.hpp>
#include <boost/version.hpp>
#include "exceptions.hpp"
#include "refcount.hpp"

/**
 * @name forward declarations
 */
//@{
CONSTREF_FORWARD_DECL(Parameter);
CONSTREF_FORWARD_DECL(behavioral_helper);
REF_FORWARD_DECL(NodeInfo);
//@}

/**
 * The property associated with edge
 */
struct EdgeProperty
{
   public:
      ///The info associated with an edge
      edge_infoRef info;

      ///The selector associated with the edge
      int selector;

      /**
       * Default constructor
       */
      EdgeProperty() :
         selector(0)
      {}

      /**
       * Constructor with selector
       * @param _selector is the selector to be associated with the edge
       */
      EdgeProperty(int _selector) :
         selector(_selector)
      {}

      /**
       * Constructor with property
       * @param _info is the property to be associated with the edge
       */
      EdgeProperty(edge_infoRef _info) :
         info(_info)
      {}
};
#if BOOST_VERSION >= 104600
typedef boost::adjacency_list< boost::listS, boost::listS, boost::bidirectionalS, boost::property<boost::vertex_index_t, std::size_t, boost::property<boost::vertex_color_t, boost::default_color_type, NodeInfoRef> >, EdgeProperty, graph_infoRef>  boost_graphs_collection;

typedef boost::adjacency_list< boost::listS, boost::listS, boost::undirectedS, boost::property<boost::vertex_index_t, std::size_t, boost::property<boost::vertex_color_t, boost::default_color_type, NodeInfoRef> >, EdgeProperty, graph_infoRef>  undirected_boost_graphs_collection;

#define boost_get_graph_property(graph_arg) (graph_arg)[boost::graph_bundle]

#else

/**
 * Custom graph property: graph_info.
 * This property provide a refcount pointer to the graph_info object. This object store all the information associatet with the whole graph.
 */
struct graph_info_t
{
   ///typedef defining graph_info_t as graph object
   typedef boost::graph_property_tag kind;
};

typedef boost::property<graph_info_t, graph_infoRef > GraphProperty;

typedef boost::adjacency_list< boost::listS, boost::listS, boost::bidirectionalS, boost::property<boost::vertex_index_t, std::size_t, boost::property<boost::vertex_color_t, boost::default_color_type, NodeInfoRef> >, EdgeProperty, GraphProperty>  boost_graphs_collection;

typedef boost::adjacency_list< boost::listS, boost::listS, boost::undirectedS, boost::property<boost::vertex_index_t, std::size_t, boost::property<boost::vertex_color_t, boost::default_color_type, NodeInfoRef> >, EdgeProperty, GraphProperty>  undirected_boost_graphs_collection;

#define boost_get_graph_property(graph_arg) boost::get_property(graph_arg, graph_info_t())

#endif

/**
 * bulk graph. All the edge of a graph are store in this object
*/
struct graphs_collection : public boost_graphs_collection
{
   protected:
      /**
       * Add an edge; if it exists, does nothing
       * @param source is the source of the edge
       * @param target is the target of the edge
       * @param selector is the selector to be added
       * @return the created edge
       */
      template <class EdgeInfoType>
      boost::graph_traits<boost_graphs_collection>::edge_descriptor GetEdge(const boost::graph_traits<boost_graphs_collection>::vertex_descriptor source, const boost::graph_traits<boost_graphs_collection>::vertex_descriptor target, const int selector)
      {
         boost::graph_traits<boost_graphs_collection>::edge_descriptor edge;
         bool already_inserted;
         boost::tie(edge, already_inserted) = boost::edge(source, target, *this);
         if(not already_inserted)
         {
            boost::add_edge(source, target, EdgeProperty(edge_infoRef(new EdgeInfoType())), *this);
            boost::tie(edge, already_inserted) = boost::edge(source, target, *this);
         }
         (*this)[edge].selector |= selector;

         return edge;
      }

      /**
       * Create a new vertex
       * @return the created vertex
       */
      template <class NodeInfoType>
      inline boost::graph_traits<boost_graphs_collection>::vertex_descriptor GetVertex()
      {
         size_t index = boost::num_vertices(*this);
         boost::graph_traits<boost_graphs_collection>::vertex_descriptor v = boost::add_vertex(*this);
         boost::property_map<boost_graphs_collection, boost::vertex_index_t>::type index_map = boost::get(boost::vertex_index_t(), *this);
         index_map[v] = index;
         NodeInfoRef & info = (*this)[v];
         info = NodeInfoRef(new NodeInfoType());
         return v;
      }

      /**
       * Get the node property; if it does not exists create it
       * @param node is the vertex whose property is asked
       * @return the associated property
       */
      template <class NodeInfoType>
      inline NodeInfoType * GetNodeInfo(typename boost::graph_traits<graphs_collection>::vertex_descriptor node)
      {
         NodeInfoRef & info = (*this)[node];
         THROW_ASSERT(info and GetPointer<NodeInfoType>(info), "Node with associated a node info different from expected");
         return GetPointer<NodeInfoType>(info);
      }

      /**
       * Get the node property
       * @param node is the node whose property is asked
       * @return the associated property
       */
      template <class NodeInfoType>
      inline const NodeInfoType * CGetNodeInfo(typename boost::graph_traits<graphs_collection>::vertex_descriptor node) const
      {
         const NodeInfoRef & info = (*this)[node];
         THROW_ASSERT(info and GetPointer<NodeInfoType>(info), "Node without node_info associated");
         return GetPointer<NodeInfoType>(info);
      }

      /**
       * Get the edge property; if it does not exists create it
       * @param source is the source vertex of the edge
       * @param target is the target vertex of the edge
       * @return the associated property
       */
      template <class EdgeInfoType>
      inline EdgeInfoType * GetEdgeInfo(typename boost::graph_traits<graphs_collection>::vertex_descriptor source, typename boost::graph_traits<graphs_collection>::vertex_descriptor target)
      {
         bool found;
         typename boost::graph_traits<graphs_collection>::edge_descriptor edge;
         boost::tie(edge, found) = boost::edge(source, target, *this);
         THROW_ASSERT(found, "Edge not present in the graph");
         return GetEdgeInfo<EdgeInfoType>(edge);
      }

      /**
       * Get the edge property; if it does not exists create it
       * @param edge is the edge whose property is asked
       * @return the associated property
       */
      template <class EdgeInfoType>
      inline EdgeInfoType * GetEdgeInfo(typename boost::graph_traits<graphs_collection>::edge_descriptor edge)
      {
         edge_infoRef & info = (*this)[edge].info;
         THROW_ASSERT(info and GetPointer<EdgeInfoType>(info), "Edge with associated an edge info different from expected");
         return GetPointer<EdgeInfoType>(info);
      }

      /**
       * Get the graph info; if it does not exists create it
       * @return the associated property
       */
      template <class GraphInfoType>
      inline GraphInfoType * GetGraphInfo()
      {
         graph_infoRef info = boost_get_graph_property(*this);
         return GetPointer<GraphInfoType>(info);
      }

   public:
      /**
       * Constructor of graph.
       * @param info is the property associated with the graph
       */
      graphs_collection(graph_infoRef info)
      {
         boost_get_graph_property(*this) = info;
      }

      ///Destructor
      virtual ~graphs_collection(){}

      /**
       * Return the selectors associated with an edge
       * @param e is the edge
       * @return the associated selector
       */
      inline int get_selector(const edge_descriptor e) const
      {
         return (*this)[e].selector;
      }

      /**
       * Set the selector associated with an edge
       * @param e is the edge
       * #param selector is the new selector
       */
      inline void set_selector(edge_descriptor e, const int selector)
      {
         (*this)[e].selector = selector;
      }

      /**
       * Add a vertex to this graph
       * @return the added vertex
       */
      inline boost::graph_traits<boost_graphs_collection>::vertex_descriptor AddVertex()
      {
         size_t index = boost::num_vertices(*this);
         boost::graph_traits<boost_graphs_collection>::vertex_descriptor v = boost::add_vertex(*this);
         boost::property_map<boost_graphs_collection, boost::vertex_index_t>::type index_map = boost::get(boost::vertex_index_t(), *this);
         index_map[v] = index;
         return v;
      }

      /**
       * Remove a vertex from this graph
       * @param v is the vertex to be removed
       */
      inline void RemoveVertex(boost::graph_traits<boost_graphs_collection>::vertex_descriptor v)
      {
         boost::remove_vertex(v, *this);
         boost::property_map<boost_graphs_collection, boost::vertex_index_t>::type index_map = boost::get(boost::vertex_index_t(), *this);
         boost::graph_traits<boost_graphs_collection>::vertex_iterator v_it, v_it_end;
         size_t index = 0;
         for (boost::tie(v_it, v_it_end) = boost::vertices(*this); v_it != v_it_end; v_it++, index++)
         {
            index_map[*v_it] = index;
         }
      }

      /**
       * Set the graph property
       * @param info is the new property
       */
      inline void set_graph_property(const graph_infoRef node)
      {
         boost_get_graph_property(*this) = node;
      }

      /**
       * Get the graph property
       * @return the property associated with the graph
       */
      inline const graph_infoRef & get_const_graph_property() const
      {
         return boost_get_graph_property(*this);
      }

      /**
       * Get the graph property
       * @return the property associated with the graph
       */
      inline graph_infoRef & get_graph_property()
      {
         return boost_get_graph_property(*this);
      }

};

typedef refcount<graphs_collection> graphs_collectionRef;

/**
 * bulk graph. All the edge of a graph are store in this object
*/
struct undirected_graphs_collection : public undirected_boost_graphs_collection
{
      ///Constructor of graph.
      undirected_graphs_collection(){}

      ///Destructor
      ~undirected_graphs_collection(){}

      /**
       * Add a vertex to this graph
       * @return the added vertex
       */
      inline boost::graph_traits<undirected_boost_graphs_collection>::vertex_descriptor AddVertex()
      {
         size_t index = boost::num_vertices(*this);
         boost::graph_traits<undirected_boost_graphs_collection>::vertex_descriptor v = boost::add_vertex(*this);
         boost::property_map<undirected_boost_graphs_collection, boost::vertex_index_t>::type index_map = boost::get(boost::vertex_index_t(), *this);
         index_map[v] = index;
         return v;
      }

      /**
       * Remove a vertex from this graph
       * @param v is the vertex to be removed
       */
      inline void RemoveVertex(boost::graph_traits<undirected_boost_graphs_collection>::vertex_descriptor v)
      {
         boost::remove_vertex(v, *this);
         boost::property_map<undirected_boost_graphs_collection, boost::vertex_index_t>::type index_map = boost::get(boost::vertex_index_t(), *this);
         boost::graph_traits<undirected_boost_graphs_collection>::vertex_iterator v_it, v_it_end;
         size_t index = 0;
         for (boost::tie(v_it, v_it_end) = boost::vertices(*this); v_it != v_it_end; v_it++, index++)
         {
            index_map[*v_it] = index;
         }
      }

      /**
       * Set a property to a vertex
       * @param v is the vertex
       * @param info is the info to be associated with a vertex
       */
      inline void set_vertex_property(boost::graph_traits<undirected_boost_graphs_collection>::vertex_descriptor v, NodeInfoRef info)
      {
         (*this)[v] = info;
      }

      /**
       * Get the property of a vertex
       * @param v is the vertex
       * @return the property
       */
      inline NodeInfoRef get_vertex_property(boost::graph_traits<undirected_boost_graphs_collection>::vertex_descriptor v)
      {
         return (*this)[v];
      }

      /**
       * Return the selectors associated with an edge
       * @param e is the edge
       * @return the associated selector
       */
      inline int get_selector(const edge_descriptor e) const
      {
         return (*this)[e].selector;
      }

      /**
       * Set the selector associated with an edge
       * @param e is the edge
       * #param selector is the new selector
       */
      inline void set_selector(edge_descriptor e, const int selector)
      {
         (*this)[e].selector = selector;
      }
};

typedef refcount<undirected_graphs_collection> undirected_graphs_collectionRef;

/**
 * Function returning the reference to the edge information associated with the specified edge.
 * This function modifies the graph if edge_info is not yet associated with the edge.
 * @param e is the considered edge.
 * @param g is the graph.
 * @return the reference to the edge_info object associated with e.
*/
template<class info_object, class Graph>
NodeInfoRef& Cget_node_infoRef(typename boost::graph_traits<Graph>::vertex_descriptor v, Graph &g)
{
   const NodeInfoRef& info = g[v];
   if (!info)
      info = NodeInfoRef(new info_object);
   return info;
}

/**
 * Function returning the node information associated with the specified node.
 * This function modifies the graph if NodeInfo is not yet associated with the node.
 * @param v is the considered node.
 * @param g is the graph.
 * @return the NodeInfo object associated with v.
*/
template<class info_object, class Graph>
void set_node_info(typename boost::graph_traits<Graph>::vertex_descriptor v, const refcount<info_object>& obj, Graph &g)
{
   NodeInfoRef& info = g[v];
   info = obj;
}

#define SET_NODE_INFO_REF(data, NodeInfo, obj, vertex_index) \
   set_node_info<NodeInfo>(vertex_index, obj, *(data))

#define GET_NODE_INFO_REF(data, NodeInfo, vertex_index) \
   Cget_node_infoRef<NodeInfo>(vertex_index, *(data))

/**
 * Function returning the node information associated with the specified node.
 * This function modifies the graph if NodeInfo is not yet associated with the node.
 * @param v is the considered node.
 * @param g is the graph.
 * @return the NodeInfo object associated with v.
*/
template<class info_object, class Graph>
info_object* get_node_info(typename boost::graph_traits<Graph>::vertex_descriptor v, Graph &g)
{
   NodeInfoRef& info = g[v];
   if (!info)
      info = NodeInfoRef(new info_object);
   THROW_ASSERT(GetPointer<info_object>(info) != 0, "Function get_node_info: the vertices associated with the graph used are not derived from info_object\n\tCheck the actual type of info_object and the type of the node of Graph");
   return GetPointer<info_object>(info);
}

/**
 * Function returning the node information associated with the specified node.
 * This function does not modify the graph.
 * @param v is the considered node.
 * @param g is the graph.
 * @return the node_info object associated with v.
*/
template<class info_object, class Graph>
const info_object* Cget_node_info(typename boost::graph_traits<Graph>::vertex_descriptor v, const Graph &g)
{
   const NodeInfo * info = g[v].get();
   THROW_ASSERT(!info || dynamic_cast<const info_object*>(info) != 0, "Function Cget_node_info: the nodes associated with the graph used are not derived from info_object\n\tCheck the actual type of info_object and the type of the node of Graph");
   return info ? dynamic_cast<const info_object*>(info) : 0;
}

#define GET_NODE_INFO(data, NodeInfo, vertex_index) \
   get_node_info<NodeInfo>(vertex_index, *(data))

#define CGET_NODE_INFO(data, NodeInfo, vertex_index) \
   Cget_node_info<NodeInfo>(vertex_index, *(data))


/**
 * Function returning the edge information associated with the specified edge.
 * This function modifies the graph if edge_info is not yet associated with the edge.
 * @param e is the considered edge.
 * @param g is the graph.
 * @return the edge_info object associated with e.
*/
template<class info_object, class Graph>
info_object* get_edge_info(typename boost::graph_traits<Graph>::edge_descriptor e, Graph &g)
{
   refcount<edge_info> & info = g[e].info;
   if (!info)
      info = refcount<edge_info>(new info_object);
   THROW_ASSERT(GetPointer<info_object>(info) != 0, "Function get_edge_info: the edges associated with the graph used are not derived from info_object\n\tCheck the actual type of info_object and the type of the edge of Graph");
   return GetPointer<info_object>(info);
}
/**
 * Function returning the reference to the edge information associated with the specified edge.
 * This function modifies the graph if edge_info is not yet associated with the edge.
 * @param e is the considered edge.
 * @param g is the graph.
 * @return the reference to the edge_info object associated with e.
*/
template<class info_object, class Graph>
refcount<edge_info> & get_edge_infoRef(typename boost::graph_traits<Graph>::edge_descriptor e, Graph &g)
{
   refcount<edge_info> & info = g[e].info;
   if (!info)
      info = refcount<edge_info>(new info_object);
   return info;
}
/**
 * Function that set the edge information associated with the specified edge.
 * @param e is the considered edge.
 * @param g is the graph.
 * @param val is the edge information.
*/
template <class Graph>
void set_edge_infoRef(typename boost::graph_traits<Graph>::edge_descriptor e, const refcount<edge_info> & val, Graph &g)
{
   g[e].info = val;
}

/**
 * Function returning the edge information associated with the specified edge.
 * This function does not modify the graph.
 * @param v is the considered node.
 * @param g is the graph.
 * @return the edge_info object associated with v.
*/
template<class info_object, class Graph>
const info_object* Cget_edge_info(typename boost::graph_traits<Graph>::edge_descriptor e, const Graph &g)
{
   const edge_info * info = g[e].info.get();
   THROW_ASSERT(!info || dynamic_cast<const info_object*>(info) != 0, "Function Cget_edge_info: the edges associated with the graph used are not derived from info_object\n\tCheck the actual type of info_object and the type of the edge of Graph");
   return info ? dynamic_cast<const info_object*>(info) : 0;
}

#define GET_EDGE_INFO(data, edge_info, edge_index) \
   get_edge_info<edge_info>(edge_index, *(data))

#define CGET_EDGE_INFO(data, edge_info, edge_index) \
   Cget_edge_info<edge_info>(edge_index, *(data))

#define SET_EDGE_INFO_REF(data, edge_info, edge_index, val) \
   set_edge_infoRef(edge_index, val, *(data))

#define GET_EDGE_INFO_REF(data, edge_info, edge_index) \
   get_edge_infoRef<edge_info>(edge_index, *(data))

/**
 * Function returning the graph information associated with the specified graph.
 * This function modifies the graph if graph_info is not yet associated with the graph.
 * @param g is the graph.
 * @return the graph_info object associated with g.
*/
template<class info_object, class Graph>
info_object* get_graph_info(Graph &g)
{
   graph_infoRef info = g.get_graph_property();
   if (!info)
   {
      graph_infoRef & prop = g.get_graph_property();
      prop = refcount<graph_info>(new info_object);
      return GetPointer<info_object>(prop);
   }
   THROW_ASSERT(GetPointer<info_object>(info) != 0, "Function get_graph_info: the graph property associated with the graph is not derived from info_object\n\tCheck the actual type of info_object and the type of the graph property");
   return GetPointer<info_object>(info);
}
/**
 * Function returning the reference to the graph information associated with the specified graph.
 * This function modifies the graph if graph_info is not yet associated with the graph.
 * @param e is the considered graph.
 * @return the reference to the graph_info object associated with e.
*/
template<class info_object, class Graph>
refcount<graph_info> & get_graph_infoRef( Graph &g)
{
   graph_infoRef info = g.get_graph_property();
   if (!info)
      info = refcount<graph_info>(new info_object);

   return info;
}

#define GET_GRAPH_INFO(data, graph_info) \
   get_graph_info<graph_info>(*(data))

#define CGET_GRAPH_INFO(data, graph_info) \
   Cget_graph_info<graph_info>(*(data))

#define SET_GRAPH_INFO_REF(data, val) \
   set_graph_infoRef(val, *(data))

#define GET_GRAPH_INFO_REF(data, graph_info) \
   get_graph_infoRef<graph_info>(*(data))

/**
 * Predicate functor object used to select the proper set of vertexes
*/
template<typename Graph>
struct select_vertex
{
   select_vertex() : all(true) {}
   select_vertex(const std::set<typename boost::graph_traits<Graph>::vertex_descriptor > &subsetentry) : all(false), subset(subsetentry)
   {
   }
   bool operator()(const typename boost::graph_traits<Graph>::vertex_descriptor &v) const
   {
      if (all)
         return true;
      else
      {
         return (subset.find(v) != subset.end());
      }
   }
   bool all;
   std::set<typename boost::graph_traits<Graph>::vertex_descriptor > subset;

};

/**
 * Predicate functor object used to select the proper set of edges
*/
template <typename Graph>
struct SelectEdge 
{
   private:
      ///The selector associated with the filtered graph
      int selector;

      ///The bulk graph
      Graph * g;

      ///The vertices of subgraph
      std::set<typename boost::graph_traits<Graph>::vertex_descriptor > subgraph_vertices;

   public:
      /**
       * Default constructor
       */
      SelectEdge():
         selector(0),
         g(0)
      {}

      /**
       * Constructor for filtering only on selector
       * @param _selector is the selector of the filtered graph
       * @param _g is the graph
       */
      SelectEdge(const int _selector, Graph * _g) :
         selector(_selector),
         g(_g)
      {}

      /**
       * Constructor for filtering also on vertices
       * @param _selector is the selector of the filtered graph
       * @param _g is the graph
       * @param _subgraph_vertices is the set of vertices of the filtered graph
       */
      SelectEdge(const int _selector, Graph * _g, const std::set<typename boost::graph_traits<Graph>::vertex_descriptor > & _subgraph_vertices) :
         selector(_selector),
         g(_g),
         subgraph_vertices(_subgraph_vertices)
      {}

      template <typename Edge>
      bool operator()(const Edge& e) const
      {
         if (subgraph_vertices.empty())
            return selector & (*g)[e].selector;
         else
         {
            typename boost::graph_traits<Graph>::vertex_descriptor u, v;
            u = boost::source(e, *g);
            v = boost::target(e, *g);
            if (subgraph_vertices.find(v) != subgraph_vertices.end() && subgraph_vertices.find(u) != subgraph_vertices.end())
               return selector & (*g)[e].selector;
            else
               return false;
      }
   }
};


/**
 * General class used to describe a graph in PandA.
*/
struct graph : public boost::filtered_graph<boost_graphs_collection, SelectEdge<boost_graphs_collection>, select_vertex<boost_graphs_collection> >
{
   protected:
      ///The graph collection
      graphs_collection * collection;

      /// selector
      const int selector;

   public:
      /// @name graph constructors
      //@{
      /**
      * Standard constructor.
      * @param g is the bulk graph.
      * @param _selector is the selector used to filter the bulk graph.
      */
      graph(graphs_collection * g, const int _selector) :
         boost::filtered_graph<boost_graphs_collection, SelectEdge<boost_graphs_collection>, select_vertex<boost_graphs_collection> > (*g, SelectEdge<boost_graphs_collection>(_selector, g), select_vertex<boost_graphs_collection>()),
         collection(g),
         selector(_selector)
      {}

      /**
      * Sub-graph constructor.
      * @param g is the bulk graph.
      * @param _selector is the selector used to filter the bulk graph.
      * @param vertices is the set of vertexes on which the graph is filtered.
      */
      graph(graphs_collection * g, const int _selector, const std::set<boost::graph_traits<graphs_collection>::vertex_descriptor > & vertices) :
         boost::filtered_graph<boost_graphs_collection, SelectEdge<boost_graphs_collection>, select_vertex<boost_graphs_collection> > (*g, SelectEdge<boost_graphs_collection>(_selector, g, vertices), select_vertex<boost_graphs_collection>(vertices)),
         collection(g),
         selector(_selector)
      {}
      //@}

      /// this function can access the bulk graph
      friend boost::graph_traits<graph>::vertex_descriptor VERTEX(const boost::graph_traits<graph>::vertices_size_type, const graph&);

      /// Destructor
      ~graph(){}

      /**
       * return true in case the vertex is a vertex of the subgraph.
      */
      bool is_in_subset(const boost::graph_traits<graph>::vertex_descriptor v) const
      {
         return m_vertex_pred(v);
      }

      /**
       * Write this graph in dot format
       * @param file_name is the file where print the graph
       * @param helper is the helper used to print label 
       * @return 1 if no error occurs
       */
      ///FIXME: helper should become a reference inside graph
      int print_to_dot(const ParameterConstRef Param, std::string & file_name, const behavioral_helperConstRef BH) const;

      /**
       * Return the selectors associated with an edge
       * @param e is the edge
       * @return the associated selector
       */
      inline int get_selector(const edge_descriptor e) const
      {
         return collection->get_selector(e);
      }

      /**
       * Set the selector associated with an edge
       * @param e is the edge
       * #param edge_selector is the new selector
       */
      inline void set_selector(edge_descriptor e, const int edge_selector)
      {
         collection->set_selector(e, edge_selector);
      }

      /**
       * Set the graph property
       * @param info is the new property
       */
      inline void set_graph_property(const graph_infoRef node)
      {
         collection->set_graph_property(node);
      }

      /**
       * Get the graph property
       * @return the property associated with the graph
       */
      inline const graph_infoRef & get_const_graph_property() const
      {
         return collection->get_const_graph_property();
      }

      /**
       * Get the graph property
       * @return the property associated with the graph
       */
      inline graph_infoRef & get_graph_property()
      {
         return collection->get_graph_property();
      }

      /**
       * Print the graph in dot format
       * @param file_name is the name of the file to be created
       * @param node_writer is the functor used to print the node labels
       * @param edge_writer is the functor used to print the edge labels
       */
      template <class NodeWriter, class EdgeWriter>
      inline int PrintDot(const std::string & file_name, const NodeWriter node_writer, const EdgeWriter edge_writer) const
      {
         std::ofstream file_stream(file_name.c_str());
         boost::write_graphviz(file_stream, *this, node_writer, edge_writer);
         return 1;
      }

      /**
       * Get the edge property
       * @param edge is the edge whose property is asked
       * @return the associated property
       */
      template <class EdgeInfoType>
      inline const EdgeInfoType * CGetEdgeInfo(typename boost::graph_traits<graphs_collection>::edge_descriptor edge) const
      {
         const edge_infoRef & info = (*this)[edge].info;
         THROW_ASSERT(info and GetPointer<EdgeInfoType>(info), "Edge without edge_info associated");
         return GetPointer<EdgeInfoType>(info);
      }

      /**
       * Get the vertex property
       * @param node is the node whose property is asked
       * @return the associated property
       */
      template <class NodeInfoType>
      inline const NodeInfoType * CGetNodeInfo(typename boost::graph_traits<graphs_collection>::vertex_descriptor node) const
      {
         const NodeInfoRef & info = (*this)[node];
         THROW_ASSERT(info and GetPointer<NodeInfoType>(info), "Edge without node_info associated");
         return GetPointer<NodeInfoType>(info);
      }

#if BOOST_VERSION >= 104200
      typedef boost::edge_property_type<graphs_collection>::type   edge_property_type;
      typedef boost::vertex_property_type<graphs_collection>::type vertex_property_type;
      typedef boost::graph_property_type<graphs_collection>::type graph_property_type;
#endif

};

typedef refcount<graph> graphRef;


#if BOOST_VERSION >= 104600
namespace boost {
  namespace detail {
    struct filtered_graph_property_selector {
      template <class FilteredGraph, class Property, class Tag>
      struct bind_ {
        typedef typename FilteredGraph::graph_type Graph;
        typedef property_map<Graph, Tag> Map;
        typedef typename Map::type type;
        typedef typename Map::const_type const_type;
      };
    };    
  } // namespace detail
  template <>  
  struct vertex_property_selector<filtered_graph_tag> {
    typedef detail::filtered_graph_property_selector type;
  };
  template <>  
  struct edge_property_selector<filtered_graph_tag> {
    typedef detail::filtered_graph_property_selector type;
  };

} // namespace boost
#endif


/**
 * General class used to describe a graph in PandA.
*/
struct ugraph : public  boost::filtered_graph<undirected_boost_graphs_collection, SelectEdge<undirected_boost_graphs_collection>, select_vertex<undirected_boost_graphs_collection> >
{
   protected:
      ///The graph collection
      undirected_graphs_collection * collection;
      /// selector
      const int selector;

   public:
      /// @name graph constructors
      //@{
      /**
      * Standard constructor.
      * @param g is the bulk graph.
      * @param _selector is the selector used to filter the bulk graph.
      */
      ugraph(undirected_graphs_collection * g, const int _selector) :
         boost::filtered_graph<undirected_boost_graphs_collection, SelectEdge<undirected_boost_graphs_collection>, select_vertex<undirected_boost_graphs_collection> > (*g, SelectEdge<undirected_boost_graphs_collection>(_selector, g), select_vertex<undirected_boost_graphs_collection>()),
         collection(g),
         selector(_selector)
      {}

      /**
      * Sub-graph constructor.
      * @param g is the bulk graph.
      * @param _selector is the selector used to filter the bulk graph.
      * @param vertices is the set of vertexes on which the graph is filtered.
      */
      ugraph(undirected_graphs_collection * g, const int _selector, const std::set<boost::graph_traits<undirected_boost_graphs_collection>::vertex_descriptor > & vertices) :
         boost::filtered_graph<undirected_boost_graphs_collection, SelectEdge<undirected_boost_graphs_collection>, select_vertex<undirected_boost_graphs_collection> > (*g, SelectEdge<undirected_boost_graphs_collection>(_selector, g, vertices), select_vertex<undirected_boost_graphs_collection>(vertices)),
         collection(g),
         selector(_selector)

      {}
      //@}

      /// this function can access the bulk graph
      friend boost::graph_traits<ugraph>::vertex_descriptor VERTEX(const boost::graph_traits<ugraph>::vertices_size_type, const ugraph&);

      /// Destructor
      ~ugraph(){}

      /**
       * return true in case the vertex is a vertex of the subgraph.
       */
      bool is_in_subset(const boost::graph_traits<ugraph>::vertex_descriptor v) const
      {
         return m_vertex_pred(v);
      }

      /**
      * Write this graph in dot format
      * @param file_name is the file where print the graph
      */
      void print_to_dot(std::string file_name);

      /**
       * Return the selectors associated with an edge
       * @param e is the edge
       * @return the associated selector
       */
      inline int get_selector(const edge_descriptor e) const
      {
         return collection->get_selector(e);
      }

      /**
       * Set the selector associated with an edge
       * @param e is the edge
       * #param edge_selector is the new selector
       */
      inline void set_selector(edge_descriptor e, const int edge_selector)
      {
         collection->set_selector(e, edge_selector);
      }

#if BOOST_VERSION >= 104200
      typedef boost::edge_property_type<undirected_graphs_collection>::type edge_property_type;
      typedef boost::vertex_property_type<undirected_graphs_collection>::type vertex_property_type;
      typedef boost::graph_property_type<undirected_graphs_collection>::type graph_property_type;
#endif
};

typedef refcount<ugraph> ugraphRef;

/**
 * Functor used to sort edges
 */
template <typename Graph>
struct ltedge
{
   private:
      ///the graph
      const Graph * g;

   public:
      /** 
       * The constructors
       */
      ltedge(const Graph * _g):
         g(_g)
         {}

      /**
       * Redefinition of binary operator as less than
       */
      bool operator()(const typename boost::graph_traits<Graph>::edge_descriptor first, const typename boost::graph_traits<Graph>::edge_descriptor second) const
      {
         if(boost::source(first, *g) < boost::source(second, *g))
            return true;
         if(boost::source(first, *g) > boost::source(second, *g))
            return false;
         return boost::target(first, *g) < boost::target(second, *g);
      }
};


/**
 * Partial specialization of Cget_graph_info with respect to the graph view.
 * This function does not modify the graph.
 * @return the graph_info object associated with g.
 */
template<class info_object>
const info_object* Cget_graph_info(const graph &g)
{
   const graph_info * info = g.get_const_graph_property().get();
   THROW_ASSERT(!info || dynamic_cast<const info_object*>(info) != 0, "Function Cget_graph_info: the graph property associated with the graph is not derived from info_object\n\tCheck the actual type of info_object and the type of the graph property");
   return info ? dynamic_cast<const info_object*>(info) : 0;
}

///vertex definition.
typedef boost::graph_traits<graph>::vertex_descriptor vertex;
///null vertex definition
#define NULL_VERTEX boost::graph_traits<graph>::null_vertex()
///vertex_iterator definition.
typedef boost::graph_traits<graph>::vertex_iterator VertexIterator;

///in_edge_iterator definition.
typedef boost::graph_traits<graph>::in_edge_iterator InEdgeIterator;
///out_edge_iterator definition.
typedef boost::graph_traits<graph>::out_edge_iterator OutEdgeIterator;
///edge_iterator definition.
typedef boost::graph_traits<graph>::edge_iterator EdgeIterator;
///edge definition.
typedef boost::graph_traits<graph>::edge_descriptor EdgeDescriptor;

/**
 * Definition of hash function for EdgeDescriptor
 */
namespace std
{
   namespace tr1
   {
      template <>
      struct hash<EdgeDescriptor> : public unary_function<EdgeDescriptor, size_t>
      {
         size_t operator()(EdgeDescriptor edge) const
         {
            size_t hash_value = 0;
            boost::hash_combine(hash_value, edge.m_source);
            boost::hash_combine(hash_value, edge.m_target);
            return hash_value;
         }
      };
   }
}

///vertex definition.
typedef boost::graph_traits<ugraph>::vertex_descriptor uvertex;
///null vertex definition
#define NULL_UVERTEX boost::graph_traits<ugraph>::null_vertex()

///vertex definition for undirected_graphs_collection.
typedef boost::graph_traits<undirected_graphs_collection>::vertex_descriptor UGCvertex;
#define NULL_UGCVERTEX boost::graph_traits<undirected_graphs_collection>::null_vertex()

///vertex_iterator definition.
typedef boost::graph_traits<ugraph>::vertex_iterator UVertexIterator;

///in_edge_iterator definition.
typedef boost::graph_traits<ugraph>::in_edge_iterator UInEdgeIterator;
///out_edge_iterator definition.
typedef boost::graph_traits<ugraph>::out_edge_iterator UOutEdgeIterator;
///edge_iterator definition.
typedef boost::graph_traits<ugraph>::edge_iterator UEdgeIterator;
///edge definition.
typedef boost::graph_traits<ugraph>::edge_descriptor UEdgeDescriptor;

/**
 * Given a filtered graph and an index returns the vertex exploiting the boost::vertex applied on the original graph.
 * @param i is the index of the vertex.
 * @param g is the graph for which the vertex is asked.
 * @return the vertex with the index i of the graph g.
 */
inline
boost::graph_traits<graph>::vertex_descriptor
VERTEX(const boost::graph_traits<graph>::vertices_size_type i, const graph&g)
{
   return boost::vertex(i, (g.m_g));
}

/**
 * Given a filtered ugraph and an index returns the vertex exploiting the boost::vertex applied on the original graph.
 * @param i is the index of the vertex.
 * @param g is the ugraph for which the vertex is asked.
 * @return the vertex with the index i of the graph g.
 */
inline
boost::graph_traits<ugraph>::vertex_descriptor
VERTEX(const boost::graph_traits<ugraph>::vertices_size_type i, const ugraph&g)
{
   return boost::vertex(i, (g.m_g));
}

/**
 * this template function adds an edge to the bulk graph when it does not exist, otherwise it modifies the associated edge property
 */
template <class Graph>
void ADD_EDGE(typename boost::graph_traits<Graph>::vertex_descriptor A, typename boost::graph_traits<Graph>::vertex_descriptor B, int selector, Graph& g)
{
   EdgeDescriptor e;
   bool inserted;
   boost::tie(e, inserted) = boost::edge(A, B, g);
   if(inserted)
   {
      g.set_selector(e, g.get_selector(e) | selector);
   }
   else
   {
      boost::add_edge(A, B, EdgeProperty(selector), g);
   }

}

template <class Graph>
void ADD_UEDGE(typename boost::graph_traits<Graph>::vertex_descriptor A, typename boost::graph_traits<Graph>::vertex_descriptor B, int selector, Graph& g)
{
   UEdgeDescriptor e;
   bool inserted;
   boost::tie(e, inserted) = boost::edge(A, B, g);
   if(inserted)
      g[e].selector = g[e].selector | selector;
   else
      boost::add_edge(A, B, selector, g);
}
#endif
