/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file RTL_characterization.cpp
 * @brief Class implementing RTL characterization
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Marco Minutoli <mminutoli@gmail.com>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10655 $
 * $Date: 2012-03-20 08:39:27 +0100 (Tue, 20 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Autoheader includes
#include "config_HAVE_FLOPOCO.hpp"

#include "RTL_characterization.hpp"

#include "technology_manager.hpp"
#include "library_manager.hpp"
#include "technology_builtin.hpp"

#include "target_device.hpp"
#include "area_model.hpp"
#include "time_model.hpp"
#include "clb_model.hpp"
#include "LUT_model.hpp"
#if HAVE_EXPERIMENTAL
#include "power_model.hpp"
#endif
#include "structural_manager.hpp"
#include "structural_objects.hpp"
#include "NP_functionality.hpp"

#include "BackendFlow.hpp"
#include "XilinxBackendFlow.hpp"

#include "HDL_manager.hpp"
#include "language_writer.hpp"

#include "polixml.hpp"
#include "xml_helper.hpp"
#include "op_node_info.hpp"
#include "Parameter.hpp"

#define DUMMY_SYNTHESIS 0

RTL_characterization::RTL_characterization(const ParameterConstRef _Param) :
      Param(_Param),
      debug_level(_Param->getOption<int>(OPT_debug_level)),
      output_level(_Param->getOption<int>(OPT_output_level))
{

}

RTL_characterization::~RTL_characterization()
{

}

void RTL_characterization::exec(const technology_managerRef& TM, const target_deviceRef& device)
{
   const std::vector<std::string>& libraries = TM->get_library_list();

   for(unsigned int l = 0; l < libraries.size(); l++)
   {
      const library_managerRef& LM = TM->get_library_manager(libraries[l]);
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, " - Analyzing the library " + LM->get_library_name());
      characterize_library(TM, LM, device);
   }

   xwrite_device_file(TM, device);
}

void RTL_characterization::xwrite_device_file(const technology_managerRef& TM, const target_deviceRef& device)
{
   std::string file_name = Param->getOption<std::string>(OPT_estimation_output);
   try
   {
      xml_document document;
      xml_element* nodeRoot = document.create_root_node("target");

      device->xwrite(nodeRoot);

      xwrite_characterization(TM, device, nodeRoot);

      document.write_to_file_formatted(file_name);
   }
   catch (const char * msg)
   {
      std::cerr << msg << std::endl;
   }
   catch (const std::string & msg)
   {
      std::cerr << msg << std::endl;
   }
   catch (const std::exception& ex)
   {
      std::cout << "Exception caught: " << ex.what() << std::endl;
   }
   catch ( ... )
   {
      std::cerr << "unknown exception" << std::endl;
   }
}

void RTL_characterization::xwrite_characterization(const technology_managerRef& TM, const target_deviceRef& device, xml_element* nodeRoot)
{
   xml_element* tmRoot = nodeRoot->add_child_element("technology");

   const std::vector<std::string>& libraries = TM->get_library_list();
   for(unsigned int l = 0; l < libraries.size(); l++)
   {
      const library_managerRef& LM = TM->get_library_manager(libraries[l]);
      xml_element* lmRoot = tmRoot->add_child_element("library");
      xml_element* name_el = lmRoot->add_child_element("name");
      name_el->add_child_text(LM->get_library_name());

      const library_manager::fu_map_type& fus = LM->get_library_fu();
      for(library_manager::fu_map_type::const_iterator f = fus.begin(); f != fus.end(); f++)
      {
         technology_nodeRef tn = f->second;
         functional_unit * current_fu = GetPointer<functional_unit>(tn);
         if (!current_fu)
         {
            current_fu = GetPointer<functional_unit>(GetPointer<functional_unit_template>(tn)->FU);
         }
         if (completed.find(current_fu->functional_unit_name) == completed.end()) continue;

         if (!current_fu->area_m)
         {
            /// set to the default value
            current_fu->area_m = area_model::create_model(device->get_target_technology(), Param);
         }

         xml_element* cell_el = lmRoot->add_child_element("cell");
         xml_element* cell_name_el = cell_el->add_child_element("name");
         cell_name_el->add_child_text(f->first);

         xml_element* attribute_el = cell_el->add_child_element("attribute");
         WRITE_XNVM2("name", "area", attribute_el);
         WRITE_XNVM2("value_type", "float64", attribute_el);
         attribute_el->add_child_text(STR(current_fu->area_m->get_area_value()));

         if (current_fu->fu_template_name != "" && current_fu->fu_template_parameters != "")
         {
            xml_element* template_el = cell_el->add_child_element("template");
            WRITE_XNVM2("name", current_fu->fu_template_name, template_el);
            WRITE_XNVM2("parameter", current_fu->fu_template_parameters, template_el);

            if(current_fu->characterizing_constant_value != "")
            {
               xml_element* constant_el = cell_el->add_child_element("characterizing_constant_value");
               constant_el->add_child_text(STR(current_fu->characterizing_constant_value));
            }
         }

         functional_unit::operation_vec ops = current_fu->get_operations();
         for(unsigned int o = 0; o < ops.size(); o++)
         {
            operation * current_op = GetPointer<operation>(ops[o]);
            current_op->xwrite(cell_el, ops[o], Param);
         }
         if(current_fu->CM && current_fu->CM->get_circ() && GetPointer<module>(current_fu->CM->get_circ()) && GetPointer<module>(current_fu->CM->get_circ())->get_specialized() != "")
               current_fu->CM->xwrite(cell_el);
      }
   }
}

void RTL_characterization::specialize_fu(const structural_objectRef& obj, unsigned int prec)
{
   const module* mod = GetPointer<module>(obj);
   for(unsigned int i = 0; i < mod->get_in_port_size(); i++)
   {
      const structural_objectRef& port = mod->get_in_port(i);
      if (port->get_typeRef()->type == structural_type_descriptor::BOOL || (GetPointer<port_o>(port) && GetPointer<port_o>(port)->get_is_memory())) continue;
      port->type_resize(prec);
   }
   for(unsigned int i = 0; i < mod->get_out_port_size(); i++)
   {
      const structural_objectRef& port = mod->get_out_port(i);
      if (port->get_typeRef()->type == structural_type_descriptor::BOOL || (GetPointer<port_o>(port) && GetPointer<port_o>(port)->get_is_memory())) continue;
      port->type_resize(prec);
   }
}

void RTL_characterization::characterize_library(const technology_managerRef& TM, const library_managerRef& LM, const target_deviceRef& device)
{
   // A copy of the map is needed becouse we are going to add functional unit to the library.
   const library_manager::fu_map_type fus = LM->get_library_fu();
   size_t n_fus = fus.size();
   size_t index = 0;
   for(library_manager::fu_map_type::const_iterator f = fus.begin(), end = fus.end(); f != end; f++)
   {
      characterize_fu(f->second, TM, LM, device);
      if(index%10 == 0)
         PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "||||||| Functional units analyzed " + STR(index) + "/" + STR(n_fus));
      ++index;
   }
}

void RTL_characterization::characterize_fu(const technology_nodeRef& f_unit, const technology_managerRef& TM, const library_managerRef& LM, const target_deviceRef& device)
{
   std::set<unsigned int> precision;
   functional_unit * fu_curr = GetPointer<functional_unit>(f_unit);
   if(fu_curr && fu_curr->fu_template_name != "") return; ///previous characterization is not considered
   bool isTemplate = false;
   if (!fu_curr && GetPointer<functional_unit_template>(f_unit))
   {
      fu_curr = GetPointer<functional_unit>(GetPointer<functional_unit_template>(f_unit)->FU);
      isTemplate = true;
   }
   std::string fu_name = fu_curr->functional_unit_name;
   std::string fu_base_name = fu_name;
   const functional_unit::operation_vec & Ops = fu_curr->get_operations();
   functional_unit::operation_vec::const_iterator ops_end = Ops.end();
   for (functional_unit::operation_vec::const_iterator ops = Ops.begin(); ops != ops_end; ops++)
   {
      operation* curr_op = GetPointer<operation>(*ops);
      std::map< std::string, std::vector<unsigned int> >::const_iterator supported_type_it_end =  curr_op->supported_types.end();
      if(curr_op->supported_types.begin() == curr_op->supported_types.end())
      {
         if(isTemplate)
         {
            precision.insert(8);
            precision.insert(16);
            precision.insert(64);
         }
         precision.insert(32);
      }
      else
      {
         for(std::map< std::string, std::vector<unsigned int> >::const_iterator supported_type_it =  curr_op->supported_types.begin(); supported_type_it != supported_type_it_end; ++supported_type_it)
         {
            std::vector<unsigned int>::const_iterator prec_it_end = supported_type_it->second.end();
            std::vector<unsigned int>::const_iterator prec_it = supported_type_it->second.begin();
            if(prec_it == prec_it_end)
            {
               if(isTemplate)
               {
                  precision.insert(8);
                  precision.insert(16);
                  precision.insert(64);
               }
               precision.insert(32);
            }
            else
            {
               for(; prec_it != prec_it_end; ++prec_it)
                  precision.insert(*prec_it);
            }

         }
      }
   }

   PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, " - Characterizing " + fu_name);
   for (std::set<unsigned int>::iterator prec = precision.begin(), end = precision.end(); prec != end; ++prec)
   {
      if (!fu_curr->CM)
      {
         completed.insert(fu_base_name);
      }
      else
      {
         const structural_objectRef obj = fu_curr->CM->get_circ();
         NP_functionalityRef NPF = GetPointer<module>(obj)->get_NP_functionality();
         bool is_xilinx = device->has_parameter("vendor") && device->get_parameter<std::string>("vendor") == "Xilinx";
         bool is_altera = device->has_parameter("vendor") && device->get_parameter<std::string>("vendor") == "Altera";


         if (!(NPF->exist_NP_functionality(NP_functionality::VERILOG_PROVIDED)
#if HAVE_FLOPOCO
               || (NPF->exist_NP_functionality(NP_functionality::FLOPOCO_PROVIDED) && is_xilinx)
#endif
               || (NPF->exist_NP_functionality(NP_functionality::VHDL_PROVIDED))
               || (NPF->exist_NP_functionality(NP_functionality::SYSTEM_VERILOG_PROVIDED))
               )
             || (fu_base_name == ARRAY_1D_STD_BRAM && !is_xilinx && !is_altera)
             || (fu_base_name == ARRAY_1D_STD_DISTRAM && !is_xilinx)
             || fu_base_name == ARRAY_1D_STD_BRAM_ALIGNED
             || fu_base_name == LUT_GATE_STD
             || fu_base_name == AND_GATE_STD
             || fu_base_name == NAND_GATE_STD
             || fu_base_name == OR_GATE_STD
             || fu_base_name == NOR_GATE_STD
             || fu_base_name == XOR_GATE_STD
             || fu_base_name == XNOR_GATE_STD
             || fu_base_name == JOIN_STD
             || fu_base_name == PT_STD
             || fu_base_name == "split_signal"
             || fu_base_name == "FSL_handler"
             || fu_base_name == "trunc_mod_expr_FU"
             //|| fu_base_name == "fp_sqrt_FU"
             || fu_base_name.find(CONSTANT_STD) != std::string::npos)
         {
            // keep the existing one
            completed.insert(fu_base_name);
         }
         else
         {
            const module* mod = GetPointer<module>(obj);
            unsigned int n_ports = mod->get_in_port_size();
            ///check for a single port
            unsigned int n_port_to_be_specialized = 0;
            for (unsigned int i = 0; i < n_ports; ++i)
            {
               structural_objectRef port_c = mod->get_in_port(i);
               if (port_c && (port_c->get_id() == CLOCK_PORT_NAME || port_c->get_id() == RESET_PORT_NAME || port_c->get_id() == START_PORT_NAME || (GetPointer<port_o>(port_c) && GetPointer<port_o>(port_c)->get_is_memory()))) continue;
               ++n_port_to_be_specialized;
            }
            unsigned int constPort;
            if(n_port_to_be_specialized <= 1 || !isTemplate)
               constPort = n_ports; // Set constPort to in_port_size to immediately stop the loop after one iteration.
            else
               constPort = 0;
            for (; constPort < n_ports + 1; ++constPort)
            {
               structural_objectRef port_c = n_ports !=  constPort ? mod->get_in_port(constPort) : structural_objectRef();
               if (port_c && (port_c->get_id() == CLOCK_PORT_NAME || port_c->get_id() == RESET_PORT_NAME || port_c->get_id() == START_PORT_NAME || (GetPointer<port_o>(port_c) && GetPointer<port_o>(port_c)->get_is_memory()))) continue;
               std::string template_parameters;
               if (isTemplate)
               {
                  fu_name = fu_base_name;
                  template_parameters="";
                  for (unsigned int iport = 0; iport < n_ports; ++iport)
                  {
                     structural_objectRef port = mod->get_in_port(iport);
                     if (port->get_id() == CLOCK_PORT_NAME || port->get_id() == RESET_PORT_NAME || port->get_id() == START_PORT_NAME || (GetPointer<port_o>(port_c) && GetPointer<port_o>(port_c)->get_is_memory())) continue;
                     if (template_parameters != "")
                        template_parameters += " ";
                     if (iport != constPort)
                     {
                        fu_name += "_" + STR(*prec);
                        template_parameters += STR(*prec);
                     }
                     else
                     {
                        fu_name += "_" + STR(0);
                        template_parameters += STR(0);
                     }
                  }
                  // output port
                  for (unsigned int oport = 0; oport < mod->get_out_port_size(); ++oport)
                  {
                     fu_name += "_" + STR(*prec);
                     template_parameters += " " + STR(*prec);
                  }
               }

               const technology_nodeRef& tn = TM->get_fu(fu_name, LM->get_library_name());
               functional_unit* fu = GetPointer<functional_unit>(tn);
               if (!fu)
               {
                  // Analizing a template, specializations of that template won't be found in the library.
                  const technology_nodeRef& specialized_fu = create_template_instance(GetPointer<functional_unit_template>(f_unit)->FU, fu_name, device, *prec);
                  fu = GetPointer<functional_unit>(specialized_fu);
                  fu->fu_template_parameters = template_parameters;
                  LM->add(specialized_fu);
               }
               // The constant 6148914691236517205 has been choosen because from a binary point of view is made of alternating 1 and 0.
               if(constPort != n_ports)
                  fu->characterizing_constant_value = STR(6148914691236517205);
               else
                  fu->characterizing_constant_value = "";
               THROW_ASSERT(NPF->exist_NP_functionality(NP_functionality::VERILOG_PROVIDED)
#if HAVE_FLOPOCO
                            || NPF->exist_NP_functionality(NP_functionality::FLOPOCO_PROVIDED)
#endif
                            || NPF->exist_NP_functionality(NP_functionality::VHDL_PROVIDED)
                            || NPF->exist_NP_functionality(NP_functionality::SYSTEM_VERILOG_PROVIDED)
                            , "Verilog, VHDL, SystemVerilog or Flopoco description not provided for functional unit " + fu_base_name);
               structural_managerRef SM = structural_managerRef(new structural_manager(Param));
               /// main circuit type
               structural_type_descriptorRef module_type = structural_type_descriptorRef(new structural_type_descriptor(fu_name + "_wrapper"));
               /// setting top circuit component
               SM->set_top_info(fu_name + "_wrapper", module_type);
               structural_objectRef circuit = SM->get_circ();
               THROW_ASSERT(circuit, "Top circuit is missing");
               structural_objectRef template_circuit = SM->add_module_from_technology_library(fu_base_name, fu_base_name, LM->get_library_name(), circuit, TM);

               PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, " - Generating HDL of functional unit " + fu_name);
               specialize_fu(template_circuit, *prec);

               module* spec_module = GetPointer<module>(template_circuit);
               if(fu_base_name == "MC_FU") /// add further specializations for this module
               {
                  spec_module->set_parameter("EXECUTION_TIME", STR(2));
                  spec_module->set_parameter("BITSIZE", STR(8));
                  spec_module->set_parameter("INITIATION_TIME", STR(1));
               }
               else if(fu_base_name == ARRAY_1D_STD_BRAM || fu_base_name == ARRAY_1D_STD_DISTRAM)
               {
                  unsigned int base_address = 8;
                  std::string init_filename="array_ref_"+boost::lexical_cast<std::string>(base_address)+".data";
                  std::ofstream init_file(init_filename.c_str());
                  unsigned int counter=0;
                  unsigned int nbyte_on_memory = fu_base_name == ARRAY_1D_STD_DISTRAM ? 8 : 4;
                  unsigned int vec_size = 64;
                  unsigned int elts_size = 8;
                  for(unsigned int i = 0; i < vec_size; ++i)
                  {
                     for(unsigned int j = 0; j < elts_size; ++j)
                     {
                        init_file << "0";
                        counter++;
                        if(counter%(nbyte_on_memory*8)==0)
                           init_file << std::endl;
                     }
                  }
                  init_file.close();
                  spec_module->set_parameter("base_address", boost::lexical_cast<std::string>(base_address));
                  spec_module->set_parameter("INIT_file", "\"\""+init_filename+"\"\"");
                  spec_module->set_parameter("n_elements", boost::lexical_cast<std::string>(vec_size));
                  spec_module->set_parameter("data_size", boost::lexical_cast<std::string>(elts_size));
               }
               else if(fu_base_name == MEMLOAD_STD)
                  spec_module->set_parameter("base_address", boost::lexical_cast<std::string>(8));
               else if(fu_base_name == MEMSTORE_STD)
                  spec_module->set_parameter("base_address", boost::lexical_cast<std::string>(8));
               structural_objectRef e_port, one_port, r_signal;

               structural_type_descriptorRef bool_type = structural_type_descriptorRef(new structural_type_descriptor("bool", 1));
               one_port = SM->add_constant(fu_name+"_constant_" + STR(1), circuit, bool_type, STR(1));
               std::string register_library = TM->get_library(register_STD);
               structural_objectRef clock_port, reset_port;

               ///add clock and reset
               for(unsigned int i = 0; i < spec_module->get_in_port_size(); i++)
               {
                  const structural_objectRef port_in = spec_module->get_in_port(i);
                  if (port_in->get_id() == CLOCK_PORT_NAME)
                  {
                     clock_port = SM->add_port(GetPointer<port_o>(port_in)->get_id(), port_o::IN, circuit, port_in->get_typeRef());
                     SM->add_connection(port_in, clock_port);
                  }
                  if (port_in->get_id() == RESET_PORT_NAME)
                  {
                     reset_port = SM->add_port(GetPointer<port_o>(port_in)->get_id(), port_o::IN, circuit, port_in->get_typeRef());
                     SM->add_connection(port_in, reset_port);
                  }
                  else if(port_in->get_id() == START_PORT_NAME)
                  {
                     e_port = SM->add_port(GetPointer<port_o>(port_in)->get_id(), port_o::IN, circuit, port_in->get_typeRef());
                     SM->add_connection(port_in, e_port);
                  }
               }
               if(!clock_port)
               {
                  clock_port = SM->add_port(CLOCK_PORT_NAME, port_o::IN, circuit, bool_type);
               }
               if(!reset_port)
               {
                  reset_port = SM->add_port(RESET_PORT_NAME, port_o::IN, circuit, bool_type);
               }

               for(unsigned int i = 0; i < spec_module->get_in_port_size(); i++)
               {
                  const structural_objectRef port_in = spec_module->get_in_port(i);
                  if (port_in->get_id() == CLOCK_PORT_NAME || port_in->get_id() == RESET_PORT_NAME || port_in->get_id() == START_PORT_NAME) continue;
                  if (isTemplate && i == constPort)
                  {
                     e_port = SM->add_constant("constant_" + STR(constPort), circuit, port_in->get_typeRef(), fu->characterizing_constant_value);
                     SM->add_connection(port_in, e_port);
                  }
                  else
                  {
                     e_port = SM->add_port(GetPointer<port_o>(port_in)->get_id(), port_o::IN, circuit, port_in->get_typeRef());

                     /// add register on inputs
                     structural_objectRef reg_mod = SM->add_module_from_technology_library(GetPointer<port_o>(port_in)->get_id()+"_REG", register_STD, register_library, circuit, TM);
                     GetPointer<module>(reg_mod)->get_in_port(0)->type_resize(GET_TYPE_SIZE(port_in));
                     GetPointer<module>(reg_mod)->get_out_port(0)->type_resize(GET_TYPE_SIZE(port_in));

                     structural_objectRef port_ck = reg_mod->find_member(CLOCK_PORT_NAME, port_o_K, reg_mod);
                     SM->add_connection(clock_port, port_ck);
                     structural_objectRef port_rst = reg_mod->find_member(RESET_PORT_NAME, port_o_K, reg_mod);
                     SM->add_connection(reset_port, port_rst);
                     structural_objectRef port_wen = reg_mod->find_member(WENABLE_PORT_NAME, port_o_K, reg_mod);
                     SM->add_connection(port_wen, one_port);

                     r_signal = SM->add_sign(GetPointer<port_o>(port_in)->get_id()+"_SIGI1", circuit, port_in->get_typeRef());
                     SM->add_connection(e_port, r_signal);
                     SM->add_connection(GetPointer<module>(reg_mod)->get_in_port(0), r_signal);

                     r_signal = SM->add_sign(GetPointer<port_o>(port_in)->get_id()+"_SIGI2", circuit, port_in->get_typeRef());
                     SM->add_connection(GetPointer<module>(reg_mod)->get_out_port(0), r_signal);
                     SM->add_connection(port_in, r_signal);
                  }

               }
               for(unsigned int i = 0; i < spec_module->get_out_port_size(); i++)
               {
                  const structural_objectRef port_out = spec_module->get_out_port(i);
                  e_port = SM->add_port(GetPointer<port_o>(port_out)->get_id(), port_o::OUT, circuit, port_out->get_typeRef());
                  /// add register on outputs
                  structural_objectRef reg_mod = SM->add_module_from_technology_library(GetPointer<port_o>(port_out)->get_id()+"_REG", register_STD, register_library, circuit, TM);
                  GetPointer<module>(reg_mod)->get_in_port(0)->type_resize(GET_TYPE_SIZE(port_out));
                  GetPointer<module>(reg_mod)->get_out_port(0)->type_resize(GET_TYPE_SIZE(port_out));

                  structural_objectRef port_ck = reg_mod->find_member(CLOCK_PORT_NAME, port_o_K, reg_mod);
                  SM->add_connection(clock_port, port_ck);
                  structural_objectRef port_rst = reg_mod->find_member(RESET_PORT_NAME, port_o_K, reg_mod);
                  SM->add_connection(reset_port, port_rst);
                  structural_objectRef port_wen = reg_mod->find_member(WENABLE_PORT_NAME, port_o_K, reg_mod);
                  SM->add_connection(port_wen, one_port);

                  r_signal = SM->add_sign(GetPointer<port_o>(port_out)->get_id()+"_SIGO1", circuit, port_out->get_typeRef());
                  SM->add_connection(port_out, r_signal);
                  SM->add_connection(GetPointer<module>(reg_mod)->get_in_port(0), r_signal);
                  r_signal = SM->add_sign(GetPointer<port_o>(port_out)->get_id()+"_SIGO2", circuit, port_out->get_typeRef());
                  SM->add_connection(GetPointer<module>(reg_mod)->get_out_port(0), r_signal);
                  SM->add_connection(e_port, r_signal);
               }

               // get the wrapped circuit.
               HDL_managerRef HDL = HDL_managerRef(new HDL_manager(device, Param));
               HDL->hdl_gen(fu_name, circuit);

               try
               {
                  /// generate the synthesis scripts
                  BackendFlowRef flow = BackendFlow::CreateFlow(Param, "Characterization", device);
                  flow->GenerateSynthesisScripts(SM, HDL);
                  PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Performing characterization of functional unit " + fu_name);
#if !DUMMY_SYNTHESIS
                  flow->ExecuteSynthesis();
                  ///the synthesis has been successfully completed
                  ///setting the used resources
                  fu->area_m = flow->get_used_resources();
#else
                  fu->area_m = area_model::create_model(device->get_target_technology(), Param);

#endif
                  ///setting the timing values for each operation
                  functional_unit::operation_vec ops = fu->get_operations();
                  for(unsigned int o = 0; o < ops.size(); o++)
                  {
                     operation * new_op = GetPointer<operation>(ops[o]);
                     time_modelRef synthesis_results = flow->get_timing_results();
                     double exec_time = synthesis_results->get_execution_time();
                     if(!new_op->time_m)
                        new_op->time_m = time_model::create_model(device->get_target_technology(), Param);
                     if(fu_base_name == ARRAY_1D_STD_BRAM && new_op->operation_name == LOAD && GetPointer<module>(fu->CM->get_circ())->get_specialized() == "")
                     {
                        new_op->time_m->set_execution_time(exec_time, 2);
                        new_op->time_m->set_initiation_time(1);
                     }
                     else if (fu_base_name == ARRAY_1D_STD_BRAM && new_op->operation_name == STORE && GetPointer<module>(fu->CM->get_circ())->get_specialized() == "")
                        new_op->time_m->set_execution_time(exec_time, 1);
                     else if ((fu_base_name == BUILTIN_EXIT_STD || fu_base_name == BUILTIN_ABORT_STD) && GetPointer<module>(fu->CM->get_circ())->get_specialized() == "")
                        new_op->time_m->set_execution_time(exec_time, 1);
                     else if(synthesis_results && new_op->time_m->get_cycles() == 0)
                     {
                        /// possible setup/hold times are trimmed away (meaningful delay are above 1ns)
                        if(exec_time <= 1.0)
                           exec_time =0.0;
                        new_op->time_m->set_execution_time(exec_time, time_model::cycles_time_DEFAULT);
                     }
                  }

                  completed.insert(fu->functional_unit_name);
                  continue;
               }
               catch (const char * str)
               {
                  std::cerr << str << std::endl;
               }
               catch (const std::string& str)
               {
                  std::cerr << str << std::endl;
               }
               catch (std::exception &e)
               {
                  std::cerr << e.what() << std::endl;
               }
               catch (...)
               {
                  std::cerr << "Unknown error type" << std::endl;
               }
               THROW_ERROR("Error during the characterization of unit " + fu_name);
            }
         }
      }
   }
}


technology_nodeRef RTL_characterization::create_template_instance(const technology_nodeRef& fu_template, std::string & name, const target_deviceRef& device, unsigned int prec)
{
   functional_unit * curr_fu = GetPointer<functional_unit>(fu_template);
   THROW_ASSERT(curr_fu, "Null functional unit template");

   functional_unit * specialized_fu = new functional_unit;
   specialized_fu->functional_unit_name = name;
   specialized_fu->fu_template_name = curr_fu->functional_unit_name;
   specialized_fu->CM = curr_fu->CM;
   specialized_fu->XML_description = curr_fu->XML_description;

   for (functional_unit::operation_vec::const_iterator itr = curr_fu->get_operations().begin(), end = curr_fu->get_operations().end(); itr < end; ++itr)
   {
      operation * const op = GetPointer<operation>(*itr);
      operation * new_op = new operation;
      new_op->operation_name = op->operation_name;
      new_op->bounded = op->bounded;

      new_op->time_m = time_model::create_model(device->get_target_technology(), Param);
      if(op->time_m)
      {
         new_op->time_m->set_execution_time(op->time_m->get_execution_time(), op->time_m->get_cycles());
         new_op->time_m->set_initiation_time(op->time_m->get_initiation_time());
      }
#if HAVE_EXPERIMENTAL
      new_op->power_m = power_modelRef(new power_model());
#endif
      new_op->commutative = op->commutative;
      std::map< std::string, std::vector<unsigned int> >::const_iterator supported_type_it_end =  op->supported_types.end();
      if(op->supported_types.begin() != op->supported_types.end())
      {
         for(std::map< std::string, std::vector<unsigned int> >::const_iterator supported_type_it =  op->supported_types.begin(); supported_type_it != supported_type_it_end; ++supported_type_it)
         {
            new_op->supported_types[supported_type_it->first].push_back(prec);
         }
      }
      specialized_fu->add_operation(technology_nodeRef(new_op));
   }

   specialized_fu->area_m = area_model::create_model(device->get_target_technology(), Param);

   return technology_nodeRef(specialized_fu);
}

