/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file core_generation.cpp
 * @brief Class implementing the converter from HDL to XML-based representation
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#include "core_generation.hpp"

#include "target_technology.hpp"
#include "technology_manager.hpp"
#include "technology_node.hpp"
#include "structural_manager.hpp"
#include "target_device.hpp"

#include "language_writer.hpp"
#include "HDL_manager.hpp"

#include "polixml.hpp"

#include "fileIO.hpp"
#include "dbgPrintHelper.hpp"
#include "Parameter.hpp"

#include "boost/filesystem.hpp"

core_generation::core_generation(const ParameterConstRef _Param) :
      Param(_Param),
      debug_level(_Param->getOption<int>(OPT_debug_level))
{

}

core_generation::~core_generation()
{

}

void core_generation::convert_vhdl_to_XML(const std::string& vhdl_module, const std::string& output_file)
{
   PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "Converting \"" + vhdl_module + "\" into \"" + output_file + "\"");

   std::string entity_name;
   std::string ip_component;
   std::string body_string;

   std::ifstream istr(vhdl_module.c_str());
   if (istr.is_open())
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "File \"" + vhdl_module + "\" opened");

      bool is_port = false;
      bool is_body = false;

      std::vector<std::string> components;

      while (!istr.eof())
      {
         std::string line;
         getline(istr, line);
         std::string trimmed_line = line;
         boost::trim(trimmed_line);

         if (trimmed_line.find("entity") != std::string::npos)
         {
            std::vector<std::string> tokens = convert_string_to_vector<std::string>(trimmed_line, " ");
            if (tokens[0] == "entity" and tokens[2] == "is")
            {
               entity_name = tokens[1];
            }
            else if (tokens[0] == "end")
            {
               is_port = false;
            }
         }
         if (trimmed_line.find("component") != std::string::npos)
         {
            std::vector<std::string> tokens = convert_string_to_vector<std::string>(trimmed_line, " ");
            if (tokens[0] == "component" and tokens[2] == "is")
            {
               components.push_back(tokens[1]);
            }
         }

         if (is_port)
         {

         }
         if (is_body)
         {
            boost::replace_all(line, ">", "&gt;");
            boost::replace_all(line, "<", "&lt;");
            body_string += line + "\\n";
         }

         if (!is_body and trimmed_line.find("port") != std::string::npos)
         {
            is_port = true;
         }
         if (trimmed_line.find("architecture") != std::string::npos)
         {
            std::vector<std::string> tokens = convert_string_to_vector<std::string>(trimmed_line, " ");
            if (tokens[0] == "architecture" and tokens[4] == "is")
            {
               is_body = true;
            }
         }
      }

      for (unsigned int c = 0; c < components.size(); c++)
      {
         if (ip_component.size()) ip_component += ",";
         ip_component += components[c];
      }
      if (entity_name.size())
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "Entity name = \"" + entity_name + "\"");
      }
      if (ip_component.size())
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "IP_COMPONENT = \"" + ip_component + "\"");
      }
      if (body_string.size())
      {
         PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "VHDL_PROVIDED = \"" + body_string + "\"");
      }
   }
   if (!entity_name.size()) return;

   structural_managerRef SM = structural_managerRef(new structural_manager(Param));
   structural_type_descriptorRef module_type(new structural_type_descriptor(entity_name));
   SM->set_top_info(entity_name, module_type);
   if (ip_component.size()) SM->add_NP_functionality(SM->get_circ(), NP_functionality::IP_COMPONENT, ip_component);
   SM->add_NP_functionality(SM->get_circ(), NP_functionality::VHDL_PROVIDED, body_string);

   technology_nodeRef tn = technology_nodeRef(new functional_unit);
   GetPointer<functional_unit>(tn)->functional_unit_name = entity_name;
   GetPointer<functional_unit>(tn)->CM = SM;

   xml_document document;
   xml_element* nodeRoot = document.create_root_node("technology");
   xml_element* libraryRoot = nodeRoot->add_child_element("library");
   xml_element* cellRoot = libraryRoot->add_child_element("cell");
   tn->xwrite(cellRoot, tn, Param);
   document.write_to_file_formatted(output_file);
}

void core_generation::convert_to_XML(const std::string& hdl_module)
{
   PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "Converting module \"" + hdl_module + "\"");
   std::string output_file = Param->getOption<std::string>(OPT_output_file);
   std::string extension = boost::filesystem::extension(hdl_module);
   if (extension == ".vhd")
   {
      convert_vhdl_to_XML(hdl_module, output_file);
   }
}

void core_generation::export_core(const technology_managerRef TM, const std::string& module_name)
{
   PRINT_DBG_MEX(DEBUG_LEVEL_MINIMUM, debug_level, "Exporting module \"" + module_name + "\"");
   std::string library = TM->get_library(module_name);
   technology_nodeRef tn = TM->get_fu(module_name, library);

   std::string output_file = Param->getOption<std::string>(OPT_output_file);

   language_writerRef lan;
   std::string extension = boost::filesystem::extension(output_file);
   if (extension == ".v")
   {
      lan = language_writer::create_writer(language_writer::VERILOG, debug_level);
   }
#if HAVE_EXPERIMENTAL
   else if (extension == ".vhd")
   {
      lan = language_writer::create_writer(language_writer::VHDL, debug_level);
   }
#endif
   else
   {
      THROW_ERROR("Output language not supported for file \"" + output_file + "\"");
   }
   output_file = output_file.substr(0, output_file.find(extension));

   unsigned int target_device = Param->getOption<unsigned int>(OPT_target_device_type);
   target_deviceRef device = target_device::create_device(static_cast<target_device::type_t>(target_device), Param, TM);
   HDL_manager hdl_man(device, Param);
   hdl_man.hdl_gen(output_file, GetPointer<functional_unit>(tn)->CM->get_circ());
}
