/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file time_model.hpp
 * @brief Class specification for time_model
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef _TIME_MODEL_HPP_
#define _TIME_MODEL_HPP_

#include "refcount.hpp"
///refcount definition for the class
CONSTREF_FORWARD_DECL(Parameter);
CONSTREF_FORWARD_DECL(target_technology);
REF_FORWARD_DECL(time_model);
class xml_element;

#if BOOST_VERSION >= 104600
#else
#include <tr1/tuple>
#define BOOST_TR1_TUPLE_HPP_INCLUDED
#endif
#include <boost/math/distributions.hpp>

#include <string>
#include <map>
#include <vector>

class time_model
{
   public:

      /// statistical delay for this type of operation on a given functional unit.
      boost::math::normal statistical_delay;

      /// map representing the pin-to-pin delay
      std::map<std::string, std::map<std::string, double> > pin_to_pin_delay;

      ///type of the timing path
      typedef enum
      {
         POST_SYNTHESIS = 1,
         POST_LAYOUT    = 2
      } path_t;

   protected:

      ///class containing all the parameters
      const ParameterConstRef Param;

      ///reference to the class containing all the information about the target technology
      const target_technologyConstRef target;

      ///map containing all the information about critical paths
      std::map<unsigned int, std::vector<std::string> > critical_paths;

      std::map<unsigned int, float> max_delay;

      /// initiation time, in terms of cycle_units, for this type of operation on a given functional unit.
      unsigned int initiation_time;

      unsigned int cycles;

      /// execution time, in terms of ns, for this type of operation on a given functional unit.
      double execution_time;


   public:

      /**
       * @name Constructors and destructors.
       */
      //@{
      ///Constructor.
      time_model(const target_technologyConstRef target, const ParameterConstRef Param);

      ///Destructor.
      virtual ~time_model();
      //@}

      static
      time_modelRef create_model(const target_technologyConstRef target, const ParameterConstRef Param);

      virtual void xwrite(xml_element* pin_node, const std::string& output_pin) = 0;

      virtual unsigned int xload_timing_path(xml_element* node);

      std::vector<std::string> get_critical_path(unsigned int type) const;

      bool has_max_delay(unsigned int type) const;

      void set_max_delay(unsigned int type, float value);

      float get_max_delay(unsigned int type) const;

      void set_initiation_time(unsigned int initiation_time);

      unsigned int get_initiation_time() const;

      unsigned int get_cycles() const;

      void set_execution_time(double execution_time, unsigned int cycles);

      double get_execution_time() const;

   public:

     /**
      * @name default values associated with the members of the operation class.
     */
     //@{
     static const double execution_time_DEFAULT /*= 10.0*/;
     static const unsigned int initiation_time_DEFAULT /*= 1*/;
     static const unsigned int cycles_time_DEFAULT /*=0.0*/;
     //@}
};

typedef refcount<time_model> time_modelRef;

#endif
