/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file behavioral_helper.cpp
 * @brief Helper for reading data from tree_manager
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
*/

///Autoheader include
#include "config_HAVE_SPARC_COMPILER.hpp"
#include "config_SKIP_WARNING_SECTIONS.hpp"

///Header include
#include "behavioral_helper.hpp"

///Behavior include
#include "application_manager.hpp"
#include "cdfg_edge_info.hpp"
#include "function_behavior.hpp"
#include "op_node_info.hpp"

///Parameter include
#include "Parameter.hpp"

///Parser include
#include "token_interface.hpp"

///STD include
#include <string>

///Tree include
#include "ext_tree_node.hpp"
#include "tree_basic_block.hpp"
#include "tree_helper.hpp"
#include "tree_manager.hpp"
#include "tree_reindex.hpp"

///Utility include
#include "boost/lexical_cast.hpp"
#include "exceptions.hpp"
#include "global_enums.hpp"
#include "simple_indent.hpp"
#include "utility.hpp"
#include "var_pp_functor.hpp"


std::map<std::string, unsigned int> behavioral_helper::used_name;

std::map<unsigned int, std::string> behavioral_helper::vars_symbol_table;

std::map<unsigned int, std::string> behavioral_helper::vars_renaming_table;

///Max length of a row (at the moment checked only during constructor printing)
#define MAX_ROW_LENGTH 128

behavioral_helper::behavioral_helper(const ParameterConstRef _Param, const tree_managerRef _TM, unsigned int _index, bool _body) :
      Param(_Param),
      TM(_TM),
      debug_level(Param->get_class_debug_level("behavioral_helper", DEBUG_LEVEL_NONE)),
      function_index(_index),
      function_name(tree_helper::name_function(TM, function_index)),
      body(_body),
      opaque(!_body),
      goto_label_unique_id(0)
{

}

behavioral_helper::~behavioral_helper()
{

}

std::tr1::tuple<std::string, unsigned int, unsigned int> behavioral_helper::get_definition(unsigned int index, bool &is_system) const
{
   THROW_ASSERT(index, "expected a meaningful index");
   return tree_helper::get_definition(TM, index, is_system);
}

void behavioral_helper::print_vertex(std::ostream & os, const graph * g, const vertex v, const var_pp_functorConstRef vppf, simple_indent & PP, bool dot) const
{
   std::list<unsigned int> nodeID = GET_NODEID(g, v);
   std::string res;
   //const tree_nodeRef node = TM->get_tree_node_const(nodeID.back());
   if (nodeID.size() && TM->get_tree_node_const(nodeID.back())->get_kind() != gimple_nop_K)
   {
      res = print_node(nodeID.back(), v, vppf);
      const tree_nodeRef node = TM->get_tree_node_const(nodeID.back());

      switch (node->get_kind())
      {
         case (gimple_assign_K):
         case (gimple_call_K):
         case (gimple_asm_K):
         case (gimple_goto_K):
            res += ";";
            break;
         default:
            break;
      }
   }
   if (dot)
   {
      std::string temp;
      for (unsigned int i = 0; i < res.size(); i++)
      {
         if (res[i] == '\"')
            temp += "\\\"";
         else if (res[i] != '\n')
            temp += res[i];
         else
            temp += "\\n";
      }
      os << temp << "\\n";
   }
   else if (res != "")
      PP(os, res + "\n");
}

std::string behavioral_helper::print_init(unsigned int var, const var_pp_functorConstRef vppf) const
{
   const tree_nodeRef node = TM->get_tree_node_const(var);
   std::string res;
   switch (node->get_kind())
   {
      case constructor_K:
      {
         constructor *constr = GetPointer<constructor>(node);
         bool designated_initializers_needed = false;
         res += '{';
         std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::const_iterator i = constr->list_of_idx_valu.begin();
         std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::const_iterator vend = constr->list_of_idx_valu.end();
         ///check if designated initializers are really needed
         tree_nodeRef firstnode = i != vend ? constr->list_of_idx_valu.begin()->first : tree_nodeRef();
         if (firstnode && GET_NODE(firstnode)->get_kind() == field_decl_K)
         {
            field_decl * fd = GetPointer<field_decl>(GET_NODE(firstnode));
            tree_nodeRef scpe = GET_NODE(fd->scpe);
            std::vector<tree_nodeRef>* field_list = 0;
            if (scpe->get_kind() == record_type_K)
               field_list = &(GetPointer<record_type>(scpe)->list_of_flds);
            else if (scpe->get_kind() == union_type_K)
               field_list = &(GetPointer<union_type>(scpe)->list_of_flds);
            else
               THROW_ERROR("expected a record_type or a union_type");
            std::vector<tree_nodeRef>::const_iterator flend = field_list->end();
            std::vector<tree_nodeRef>::const_iterator fli = field_list->begin();
            for (; fli != flend && i != vend; i++, fli++)
            {
               if (i->first && GET_INDEX_NODE(i->first) != GET_INDEX_NODE(*fli))
                  break;
            }
            if (fli != flend && i != vend)
               designated_initializers_needed = true;
         }
         else
            designated_initializers_needed = true;

         size_t current_length = res.size();
         for (i = constr->list_of_idx_valu.begin(); i != vend;)
         {
            std::string current;
            if (designated_initializers_needed && i->first && GET_NODE(i->first)->get_kind() == field_decl_K)
            {
               current += ".";
               current += PrintVariable(GET_INDEX_NODE(i->first));
               current += "=";
            }
            tree_nodeRef val = GET_NODE(i->second);
            THROW_ASSERT(val, "Something of unexpected happen");
            if (val->get_kind() == addr_expr_K)
            {
               addr_expr *ae = GetPointer<addr_expr>(val);
               tree_nodeRef op = GET_NODE(ae->op);
               if (op->get_kind() == function_decl_K)
               {
                  val = GET_NODE(ae->op);
                  THROW_ASSERT(val, "Something of unexpected happen");
               }
            }
            if (val->get_kind() == function_decl_K)
            {
               current += tree_helper::print_function_name(TM, GetPointer<function_decl>(val));
            }
            else if (val->get_kind() == constructor_K)
            {
               current += print_init(GET_INDEX_NODE(i->second), vppf);
            }
            else if (val->get_kind() == var_decl_K)
            {
               current += PrintVariable(GET_INDEX_NODE(i->second));
            }
            else if (val->get_kind() == ssa_name_K)
            {
               current += PrintVariable(GET_INDEX_NODE(i->second));
            }
            else
            {
               current += print_node(GET_INDEX_NODE(i->second), vertex(), vppf);
            }
            i++;
            if (i != vend)
               current += ", ";
            if((current.size() + current_length) > MAX_ROW_LENGTH)
            {
               current_length = current.size();
               res += "\n" + current;
            }
            else
            {
               current_length += current.size();
               res += current;
            }
         }
         res += '}';
         break;
      }
      case addr_expr_K:
      case nop_expr_K:
      case integer_cst_K:
      case real_cst_K:
      case string_cst_K:
      case vector_cst_K:
      case complex_cst_K:
      {
         res = print_constant(var, vppf);
         break;
      }
      case pointer_plus_expr_K:
      case plus_expr_K:
      {
         vertex dummy_vertex = 0;
         res += print_node(var, dummy_vertex, vppf);
#if 0
         bool binary_op_cast = is_a_pointer(var);
         if (binary_op_cast)
            res += "((" + tree_helper::print_type(TM, get_type(var)) + ")(";
         const char *op = tree_helper::op_symbol (node);
         binary_expr *be = GetPointer<binary_expr>(node);
         unsigned int left_op = GET_INDEX_NODE(be->op0);
         bool left_op_cast = is_a_pointer(left_op);
         unsigned int right_op = GET_INDEX_NODE(be->op1);
         bool right_op_cast = is_a_pointer(right_op);
         if (left_op_cast)
            res += "((unsigned long int)";
         res += vppf(left_op);
         if (left_op_cast)
            res += ")";
         res += std::string(" ") + op + " ";
         if (right_op_cast)
            res += "((unsigned long int)";
         res += vppf(right_op);
         if (right_op_cast)
            res += ")";
         if (binary_op_cast)
            res += "))";
#endif
         break;
      }
      case var_decl_K:
      {
         var_decl * vd = GetPointer<var_decl>(node);
         THROW_ASSERT(vd->init, "expected a initialization value: "+STR(var));
         res += print_init(GET_INDEX_NODE(vd->init), vppf);
         break;
      }
      default:
         THROW_ERROR("Currently not supported nodeID " + boost::lexical_cast<std::string>(var));
   }
   return res;
}

std::string behavioral_helper::print_attributes(unsigned int var, const var_pp_functorConstRef vppf, bool first) const
{
   std::string res;
   if (first)
      res += "__attribute__ (";
   tree_list* tl = GetPointer<tree_list>(TM->get_tree_node_const(var));
   if (tl->purp)
      res += "(" + PrintVariable(GET_INDEX_NODE(tl->purp));
   if (tl->valu && GET_NODE(tl->valu)->get_kind() == tree_list_K)
   {
      res += print_attributes(GET_INDEX_NODE(tl->valu), vppf, false);
   }
   else if (tl->valu && (GET_NODE(tl->valu)->get_kind() == string_cst_K || GET_NODE(tl->valu)->get_kind() == integer_cst_K))
   {
      res += "(" + print_constant(GET_INDEX_NODE(tl->valu)) + ")";
   }
   else if (tl->valu)
      THROW_ERROR("Not yet supported: " + std::string(GET_NODE(tl->valu)->get_kind_text()));
   if (tl->purp)
      res += ")";
   if (tl->chan)
   {
      res += " " + print_attributes(GET_INDEX_NODE(tl->chan), vppf, false);
   }
   if (first)
      res += ")";
   return res;
}

std::string behavioral_helper::PrintVariable(unsigned int var) const
{
   THROW_ASSERT(var != application_manager::CONDITION_RESULT, "Condition result should not be printed or considered by this function");
   if(vars_renaming_table.find(var) != vars_renaming_table.end())
      return vars_renaming_table.find(var)->second;
   if (vars_symbol_table[var] != "")
   {
      return vars_symbol_table[var];
   }
   if (var == default_COND)
      return "default";
   const tree_nodeRef temp = TM->get_tree_node_const(var);
   if (temp->get_kind() == indirect_ref_K)
   {
      indirect_ref * ir = GetPointer<indirect_ref>(temp);
      unsigned int pointer = GET_INDEX_NODE(ir->op);
      std::string pointer_name = PrintVariable(pointer);
      vars_symbol_table[var] = "*" + pointer_name;
      return vars_symbol_table[var];
   }
   THROW_ASSERT(temp->get_kind() != mem_ref_K, "unxpected condition @" + STR(var));
   if (temp->get_kind() == identifier_node_K)
   {
      identifier_node * in = GetPointer<identifier_node>(temp);
      vars_symbol_table[var] = in->strg;
      return vars_symbol_table[var];
   }
   if (temp->get_kind() == field_decl_K)
   {
      field_decl * fd = GetPointer<field_decl>(temp);
      if (fd->name)
      {
         identifier_node * id = GetPointer<identifier_node>(GET_NODE(fd->name));
         return id->strg;
      }
      else
      {
         return INTERNAL + boost::lexical_cast<std::string>(var);
      }
   }
   if (temp->get_kind() == function_decl_K)
   {
      function_decl * fd = GetPointer<function_decl>(temp);
      return fd->get_function_name();
   }
   if (temp->get_kind() == ssa_name_K)
   {
      ssa_name * sa = GetPointer<ssa_name>(temp);
      unsigned int ssa_index = GET_INDEX_NODE(sa->var);
      std::string name = PrintVariable(ssa_index);
      if (!sa->volatile_flag && GET_NODE(sa->def_stmts.front())->get_kind() != gimple_nop_K)
      {
         name += ("_" + boost::lexical_cast<std::string>(sa->vers));
      }
      return name;
   }
   if (temp->get_kind() == var_decl_K || temp->get_kind() == parm_decl_K)
   {
      decl_node * dn = GetPointer<decl_node>(temp);
      if (dn->name)
      {
         identifier_node * id = GetPointer<identifier_node>(GET_NODE(dn->name));
         vars_symbol_table[var] = tree_helper::normalized_ID(id->strg);
         return vars_symbol_table[var];
      }
   }
   if (is_a_constant(var))
      vars_symbol_table[var] = print_constant(var);
   if (vars_symbol_table[var] == "")
      vars_symbol_table[var] = INTERNAL + boost::lexical_cast<std::string>(var);
   return vars_symbol_table[var];
}

std::string behavioral_helper::print_constant(unsigned int var, const var_pp_functorConstRef vppf) const
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Printing constant " + boost::lexical_cast<std::string>(var));
   THROW_ASSERT(is_a_constant(var), std::string("Object is not a constant ") + boost::lexical_cast<std::string>(var));
   if (var == default_COND)
      return "default";
   std::string res;
   const tree_nodeRef node = TM->get_tree_node_const(var);
   switch (node->get_kind())
   {
      case integer_cst_K:
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->integer_cst");
         integer_cst *ic = GetPointer<integer_cst>(node);
         tree_nodeRef type = GET_NODE(ic->type);
         integer_type* it = GetPointer<integer_type>(type);
         bool unsigned_flag = (it && it->unsigned_flag) || type->get_kind() == pointer_type_K || type->get_kind() == boolean_type_K || type->get_kind() == enumeral_type_K;
#if 0
         ///check if the IR type is consistent with the type name
         if (it)
         {
            std::string predicted_type;
            if (it->unsigned_flag)
               predicted_type += "unsigned ";
            if (it->algn == 8)
               predicted_type += "char";
            else if (it->algn == 16)
               predicted_type += "short";
            else if (it->algn == 32)
               predicted_type += "int";
            else if (it->algn == 64)
               predicted_type += "long long int";
            std::string actual_type = tree_helper::print_type(TM, GET_INDEX_NODE(ic->type));
            if (predicted_type != actual_type && actual_type != "bit_size_type")
               res = "(" + actual_type + ")";
         }
#endif
         long long value = tree_helper::get_integer_cst_value(ic);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Value is " + boost::lexical_cast<std::string>(value));
         if ((it && it->prec == 64) && (value == (static_cast<long long int>(-0x08000000000000000LL))))
            res = "(long long int)-0x08000000000000000";
         else if ((it && it->prec == 32) && (value == (static_cast<long int>(-0x080000000L))))
            res = "(long int)-0x080000000";
         else
         {
            res += boost::lexical_cast<std::string>(value);
         }
         if (it && it->prec > 32)
         {
            if (it && it->unsigned_flag)
               res += "LLU";
            else
               res += "LL";
         }
         else
         {
            if (unsigned_flag)
               res += "u";
            else if (type->get_kind() == pointer_type_K)
               res += "/*B*/";
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
         break;
      }
      case real_cst_K:
      {
         real_cst *rc = GetPointer<real_cst>(node);
         if (rc->overflow_flag)
            res += " overflow";
         tree_nodeRef type = GET_NODE(rc->type);
         if (type->get_kind() == real_type_K)
         {
            if (GetPointer<real_type>(type)->prec == 80) ///long double

            {
               if (rc->valr == "Inf")
               {
                  if (rc->valx[0] == '-')
                     res += "-";
                  res += "__builtin_infl()";
               }
               else if (rc->valr == "Nan")
               {
                  if (rc->valx[0] == '-')
                     res += "-";
                  res += "__builtin_nanl(\"\")";
               }
               else
               {
                  res += rc->valr;
                  res += "L";
               }
            }
            else if (GetPointer<real_type>(type)->prec == 64) ///double

            {
               if (rc->valr == "Inf")
               {
                  if (rc->valx[0] == '-')
                     res += "-";
                  res += "__builtin_inf()";
               }
               else if (rc->valr == "Nan")
               {
                  if (rc->valx[0] == '-')
                     res += "-";
                  res += "__builtin_nan(\"\")";
               }
               else
               {
                  res += rc->valr;
               }
            }
            else if (rc->valr == "Inf")///float

            {
               if (rc->valx[0] == '-')
                  res += "-";
               res += "__builtin_inff()";
            }
            else if (rc->valr == "Nan")
            {
               if (rc->valx[0] == '-')
                  res += "-";
               res += "__builtin_nanf(\"\")";
            }
            else
               ///FIXME: float can not be used for imaginary part of complex number
               res += /*"(float)" +*/ rc->valr;
         }
         else
            THROW_ERROR(std::string("Node not yet supported: ") + node->get_kind_text());
         break;
      }
      case complex_cst_K:
      {
         complex_cst *cc = GetPointer<complex_cst>(node);

         res += print_constant(GET_INDEX_NODE(cc->real), vppf);
         res += "+";
         res += print_constant(GET_INDEX_NODE(cc->imag), vppf);
         res += "i";
         break;
      }
      case string_cst_K:
      {
         string_cst * sc = GetPointer<string_cst>(node);
         if (sc->type)
         {
            array_type *at = GetPointer<array_type>(GET_NODE(sc->type));
            THROW_ASSERT(at, "Expected an array type");
            integer_type *elts = GetPointer<integer_type>(GET_NODE(at->elts));
            if (elts->prec == 32)//wide char string
            {
               if(elts->unsigned_flag and not tree_helper::is_system(TM, GET_INDEX_NODE(at->elts)))
               {
                  THROW_ERROR_CODE(C_EC, "Unsigned wide char not supported");
               }
               res = "L";
            }
         }
         res += "\"" + sc->strg + "\"";
         break;
      }
      case vector_cst_K:
      {
         vector_cst *vc = GetPointer<vector_cst>(node);
         //res += "((" + tree_helper::print_type(TM, get_type(GET_INDEX_NODE(vc->type))) + ") ";
         res += "{ ";
         for (unsigned int i = 0; i < (vc->list_of_valu).size(); i++ ) //vector elements

         {
            res += print_constant(GET_INDEX_NODE(vc->list_of_valu[i]), vppf);
            if (i != (vc->list_of_valu).size() - 1) //not the last element element
               res += ", ";
         }
         res += " }";
         //res += " )";
         break;
      }
      case case_label_expr_K:
      {
         case_label_expr * cl = GetPointer<case_label_expr>(node);
         if (cl->default_flag)
            res += "default";
         else
         {
            if (cl->op1)
            {
               THROW_ASSERT(GetPointer<integer_cst>(GET_NODE(cl->op0)), "Case label expression " + boost::lexical_cast<std::string>(var) + " does not use integer_cst");
               THROW_ASSERT(GetPointer<integer_cst>(GET_NODE(cl->op1)), "Case label expression " + boost::lexical_cast<std::string>(var) + " does not use integer_cst");
               long long low = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(cl->op0)));
               long long high = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(cl->op1)));
               while (low < high)
               {
                  res += boost::lexical_cast<std::string>(low) + "u : case ";
                  low++;
               }
               res += boost::lexical_cast<std::string>(low) + "u";
            }
            else
               res += print_constant(GET_INDEX_NODE(cl->op0), vppf);
         }
         break;
      }
      case label_decl_K:
      {
         label_decl * ld = GetPointer<label_decl>(node);
         THROW_ASSERT(ld->name, "name expected in a label_decl");
         identifier_node * id = GetPointer<identifier_node>(GET_NODE(ld->name));
         THROW_ASSERT(id, "expected an identifier_node");
         res += id->strg;
         break;
      }
      case nop_expr_K:
      {
         unary_expr *ue = GetPointer<unary_expr>(node);
         if (!(GET_NODE(ue->op)->get_kind() == addr_expr_K && GET_NODE(GetPointer<addr_expr>(GET_NODE(ue->op))->op)->get_kind() == label_decl_K))
            res += "(" + tree_helper::print_type(TM, GET_INDEX_NODE(ue->type)) + ") ";
         res += print_constant(GET_INDEX_NODE(ue->op), vppf);
         break;
      }
      case addr_expr_K:
      {
         unary_expr *ue = GetPointer<unary_expr>(node);
         if (is_a_constant(GET_INDEX_NODE(ue->op)))
         {
            if (GET_NODE(ue->op)->get_kind() == label_decl_K)
               res += "&";
            res += "&" + print_constant(GET_INDEX_NODE(ue->op));
         }
         else if (GET_NODE(ue->op)->get_kind() == function_decl_K)
            res += tree_helper::print_function_name(TM, GetPointer<function_decl>(GET_NODE(ue->op)));
         else if (is_an_array(GET_INDEX_NODE(ue->op)))
         {
            if (vppf)
               res += (*vppf)(GET_INDEX_NODE(ue->op));
            else
               res += PrintVariable(GET_INDEX_NODE(ue->op));
         }
         else
         {
            if (vppf)
               res += "&" + (*vppf)(GET_INDEX_NODE(ue->op));
            else
               res += "&" + PrintVariable(GET_INDEX_NODE(ue->op));
         }
         break;
      }
      default:
         THROW_ERROR("Var object is not a constant " + boost::lexical_cast<std::string>(var) + " " + node->get_kind_text());
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Printed constant " + res);
   return res;
}

unsigned int behavioral_helper::get_size(unsigned int var) const
{
   return tree_helper::size(TM, get_type(var));
}

unsigned int behavioral_helper::get_object_size(unsigned int var) const
{
   return tree_helper::get_object_size(TM->get_tree_node_const(var));
}

std::string behavioral_helper::get_function_name() const
{
    return function_name;
}

unsigned int behavioral_helper::get_function_index() const
{
    return function_index;
}

unsigned int behavioral_helper::get_function_return(unsigned int obj) const
{
   const tree_nodeRef function = TM->get_tree_node_const(obj);
   function_decl* fd = GetPointer<function_decl>(function);
   THROW_ASSERT(fd, "NodeId is not a function declaration");
   tree_nodeRef fun_type = GET_NODE(fd->type);
   function_type* ft = GetPointer<function_type>(fun_type);
   THROW_ASSERT(ft, "NodeId is not related to a valid function type");
   tree_nodeRef in_type = GET_NODE(ft->retn);
   return GET_INDEX_NODE(ft->retn);
}

const std::list<unsigned int> behavioral_helper::get_parameters() const
{
    return parameters;
}

void behavioral_helper::add_variable(unsigned int var, const std::string _name, bool written)
{
    std::string name;
    if (_name == "")
        name = INTERNAL + boost::lexical_cast<std::string>(var);
    else
        name = _name;
    name = tree_helper::normalized_ID(name);
    if (variables.find(var) == variables.end())
    {
        if (used_name.find(name) == used_name.end())
        {
            used_name[name] = var;
        }
        if (written)
            variables.insert(var);
        vars_symbol_table[var] = name;
    }
    else
    {
        THROW_ASSERT(vars_symbol_table[var] == name, "Impossible to add variable; index " + boost::lexical_cast<std::string>(var) + " Current name " + name + " Previous name "+ vars_symbol_table[var]);
        if (written)
            variables.insert(var);
    }
}

void behavioral_helper::add_parameter(unsigned int param)
{
    parameters.push_back(param);
}

bool behavioral_helper::has_implementation() const
{
    return body;
}

void behavioral_helper::add_initialization(unsigned int var, unsigned int init)
{
    initializations[var] = init;
}

void behavioral_helper::set_opaque()
{
    opaque = true;
}

void behavioral_helper::set_not_opaque()
{
    opaque = false;
}

bool behavioral_helper::get_opaque() const
{
    return opaque;
}


unsigned int behavioral_helper::get_type(const unsigned int var) const
{
   return tree_helper::get_type_index(TM, var);
}

unsigned int behavioral_helper::get_pointed_type(const unsigned int type) const
{
   return tree_helper::get_pointed_type(TM, type);
}

unsigned int behavioral_helper::GetElements(const unsigned int type) const
{
   return tree_helper::GetElements(TM, type);
}

std::list<unsigned int> behavioral_helper::get_fields_type(unsigned int obj) const
{
   return tree_helper::get_fields_type(TM, get_type(obj)) ;
}

bool behavioral_helper::isCallExpression(unsigned int nodeID) const
{
   //get the tree node associated with the provided ID
   const tree_nodeRef node = TM->get_tree_node_const(nodeID);
   if (node->get_kind() == call_expr_K)
      return true;
   else
      return false;
}

unsigned int behavioral_helper::getCallExpressionIndex(std::string launch_code) const
{
   std::string temp = launch_code;
   size_t pos = temp.find('(');
   temp.erase(pos);

   //now 'temp' variable contains the called-function name
   unsigned int index = TM->function_index(temp);
   return index;
}

void behavioral_helper::print_var_declaration(std::ostream &os, unsigned int var, var_pp_functorConstRef vppf, simple_indent &PP,  bool init_has_to_be_printed) const
{
   const tree_nodeRef & curr_tn = TM->get_tree_node_const(var);
   THROW_ASSERT(GetPointer<decl_node>(curr_tn) || GetPointer<ssa_name>(curr_tn) , "Call print_var_declaration on node "  + boost::lexical_cast<std::string>(var) + " which is of type " + curr_tn->get_kind_text());
   decl_node * dn = 0;
   if (GetPointer<decl_node>(curr_tn))
      dn = GetPointer<decl_node>(curr_tn);
   ///If it is not a decl node (then it is an ssa-name) or it's a not system decl_node
   if (!dn or !(dn->operating_system_flag or dn->library_system_flag))
   {
      PP(os, tree_helper::print_type(TM, get_type(var), false, false, init_has_to_be_printed, var, vppf));
      unsigned int attributes = get_attributes(var);
      std::tr1::unordered_set<unsigned int> list_of_variables;
      const unsigned int init = GetInit(var, list_of_variables);
      if (attributes)
         PP(os, " " + print_attributes(attributes, vppf));
      if (init && init_has_to_be_printed)
         PP(os, " = " + print_init(init, vppf));
   }
}


unsigned int behavioral_helper::is_coming_from_phi_node(unsigned int nodeID) const
{
   tree_nodeRef tn = TM->get_tree_node_const(nodeID);
   gimple_assign * gms = GetPointer<gimple_assign>(tn);
   if (gms && gms->orig)
      return GET_INDEX_NODE(gms->orig);
   else
      return 0;
}

bool behavioral_helper::is_var_args() const
{
   tree_nodeRef tn = TM->get_tree_node_const(function_index);
   THROW_ASSERT(tn->get_kind() == function_decl_K, "function_index is not a function decl");
   tn = GET_NODE(GetPointer<function_decl>(tn)->type);
   return GetPointer<function_type>(tn)->varargs_flag;
}

std::list<std::string> behavioral_helper::print_function_pragma(unsigned int funId, const var_pp_functorConstRef vppf) const
{
   tree_nodeRef tn = TM->get_tree_node_const(funId);
   THROW_ASSERT(tn->get_kind() == function_decl_K, "funId is not a function decl");
   std::list<std::string> pragmas;
   function_decl* fn = GetPointer<function_decl>(tn);
   for (std::vector<tree_nodeRef>::const_iterator i = fn->pragmas.begin(); i != fn->pragmas.end(); i++)
      pragmas.push_back(print_node(GET_INDEX_NODE(*i), 0, vppf));
   return pragmas;
}

std::string behavioral_helper::print_node(unsigned int index, vertex v, const var_pp_functorConstRef vppf) const
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Printing node " + boost::lexical_cast<std::string>(index));
   std::string res = "";
   const tree_nodeRef node = TM->get_tree_node_const(index);
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Type is " + std::string(node->get_kind_text()));
   switch (node->get_kind())
   {
         /* Binary arithmetic and logic expressions.  */
      case plus_expr_K:
      {
         const std::string op = tree_helper::op_symbol (node);
         binary_expr *be = GetPointer<binary_expr>(node);
         unsigned int left_op = GET_INDEX_NODE(be->op0);
         unsigned int right_op = GET_INDEX_NODE(be->op1);
         unsigned int prec = 0;
         unsigned int return_index;
         const tree_nodeRef type = tree_helper::get_type_node(node, return_index);
         if (type && (type->get_kind() == integer_type_K))
            prec = GetPointer<integer_type>(type)->prec;
         unsigned int algn = 0;
         if (type && (type->get_kind() == integer_type_K))
            algn = GetPointer<integer_type>(type)->algn;
         //bitfield type
         if (prec != algn && prec % algn)
         {
            res += "((";
         }
         res += "(" + tree_helper::print_type(TM, GET_INDEX_NODE(be->type)) + ")(";
         if (GetPointer<decl_node>(GET_NODE(be->op0)) || GetPointer<ssa_name>(GET_NODE(be->op0)))
            res += print_node(left_op, v, vppf);
         else
            res += ("(" + print_node(left_op, v, vppf) + ")");
         res += std::string(" ") + op + " ";
         if (GetPointer<decl_node>(GET_NODE(be->op1)) || GetPointer<ssa_name>(GET_NODE(be->op1)))
            res += print_node(right_op, v, vppf);
         else
            res += ("(" + print_node(right_op, v, vppf) + ")");
         res += ")";
         if (prec != algn && prec % algn)
         {
            res += ")%(1";
            if (prec > 32)
               res += "LL";
            if (GetPointer<integer_type>(type)->unsigned_flag)
               res += "U";
            res += " << " + boost::lexical_cast<std::string>(prec) + "))";
         }
         break;
      }
      case bit_ior_expr_K:
      case bit_xor_expr_K:
      case bit_and_expr_K:
      {
         const std::string op = tree_helper::op_symbol (node);
         binary_expr *be = GetPointer<binary_expr>(node);
         unsigned int left_op = GET_INDEX_NODE(be->op0);
         unsigned int right_op = GET_INDEX_NODE(be->op1);
         unsigned int prec = 0;
         unsigned int return_index;
         const tree_nodeRef type = tree_helper::get_type_node(node, return_index);
         bool bit_expression = type && type->get_kind() == pointer_type_K;

         if (type && (type->get_kind() == integer_type_K))
            prec = GetPointer<integer_type>(type)->prec;
         unsigned int algn = 0;
         if (type && (type->get_kind() == integer_type_K))
            algn = GetPointer<integer_type>(type)->algn;
         //bitfield type
         if (prec != algn && prec % algn)
         {
            res += "((";
         }
         if (bit_expression)
            res += "((" + tree_helper::print_type(TM, get_type(index)) + ")(((unsigned)(";
         if (GetPointer<decl_node>(GET_NODE(be->op0)) || GetPointer<ssa_name>(GET_NODE(be->op0)))
            res += print_node(left_op, v, vppf);
         else
            res += ("(" + print_node(left_op, v, vppf) + ")");
         if (bit_expression)
            res += "))";
         res += std::string(" ") + op + " ";
         if (bit_expression)
            res += "((unsigned)(";
         if (GetPointer<decl_node>(GET_NODE(be->op1)) || GetPointer<ssa_name>(GET_NODE(be->op1)))
            res += print_node(right_op, v, vppf);
         else
            res += ("(" + print_node(right_op, v, vppf) + ")");
         if (bit_expression)
            res += "))))";
         if (prec != algn && prec % algn)
         {
            res += ")%(1";
            if (prec > 32)
               res += "LL";
            if (GetPointer<integer_type>(type)->unsigned_flag)
               res += "U";
            res += " << " + boost::lexical_cast<std::string>(prec) + "))";
         }
         break;
      }
      case mult_expr_K:
      case minus_expr_K:
      case trunc_div_expr_K:
      case ceil_div_expr_K:
      case floor_div_expr_K:
      case round_div_expr_K:
      case trunc_mod_expr_K:
      case ceil_mod_expr_K:
      case floor_mod_expr_K:
      case round_mod_expr_K:
      case rdiv_expr_K:
      case exact_div_expr_K:
      case lshift_expr_K:
      case rshift_expr_K:
      case truth_andif_expr_K:
      case truth_orif_expr_K:
      case truth_and_expr_K:
      case truth_or_expr_K:
      case truth_xor_expr_K:
      {
         const std::string op = tree_helper::op_symbol (node);
         binary_expr *be = GetPointer<binary_expr>(node);
         unsigned int left_op = GET_INDEX_NODE(be->op0);
         unsigned int right_op = GET_INDEX_NODE(be->op1);
         unsigned int return_index;
         const tree_nodeRef type = tree_helper::get_type_node(node, return_index);

         if (GetPointer<decl_node>(GET_NODE(be->op0)) || GetPointer<ssa_name>(GET_NODE(be->op0)))
            res += print_node(left_op, v, vppf);
         else
            res += ("(" + print_node(left_op, v, vppf) + ")");
         res += std::string(" ") + op + " ";
         if (GetPointer<decl_node>(GET_NODE(be->op1)) || GetPointer<ssa_name>(GET_NODE(be->op1)))
            res += print_node(right_op, v, vppf);
         else
            res += ("(" + print_node(right_op, v, vppf) + ")");
         break;
      }
      case widen_sum_expr_K:
      case widen_mult_expr_K:
      {
         const std::string op = tree_helper::op_symbol (node);
         binary_expr *be = GetPointer<binary_expr>(node);
         unsigned int left_op = GET_INDEX_NODE(be->op0);
         unsigned int right_op = GET_INDEX_NODE(be->op1);
         unsigned int return_index;
         tree_helper::get_type_node(node, return_index);

         res += "(("+ tree_helper::print_type(TM, return_index) + ")(" + print_node(left_op, v, vppf) + "))";
         res += std::string(" ") + op + " ";
         res += "(("+ tree_helper::print_type(TM, return_index) + ")(" + print_node(right_op, v, vppf) + "))";
         break;
      }
      case pointer_plus_expr_K:
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->");
         bool binary_op_cast = is_a_pointer(index);
         const std::string op = tree_helper::op_symbol (node);
         pointer_plus_expr *ppe = GetPointer<pointer_plus_expr>(node);
         unsigned int temp = 0;
         tree_nodeRef type_node = tree_helper::get_type_node(GET_NODE(ppe->op0), temp);
         unsigned int left_op = GET_INDEX_NODE(ppe->op0);
         unsigned int right_op = GET_INDEX_NODE(ppe->op1);
         const bool left_op_cast = is_a_pointer(left_op);
#ifndef NDEBUG
         if(left_op_cast)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Left part is a pointer");
         }
         bool right_op_cast = is_a_pointer(right_op);
#endif
         bool do_reverse_pointer_arithmetic = false;
         tree_nodeRef right_op_node = GET_NODE(ppe->op1);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Starting right op node is " + boost::lexical_cast<std::string>(GET_INDEX_NODE(ppe->op1)) + " - " + right_op_node->get_kind_text());
         bool right_cost = right_op_node->get_kind() == integer_cst_K;
         THROW_ASSERT(!right_op_cast, "expected a right operand different from a pointer");
         THROW_ASSERT(GET_NODE(ppe->type)->get_kind() == pointer_type_K, "expected a pointer type");

         /// check possible pointer arithmetic reverse
         long long int deltabit;
         tree_nodeRef pointed_type = TM->get_tree_node_const(get_pointed_type(tree_helper::get_type_index(TM,GET_INDEX_NODE(ppe->type))));
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Pointed type (" +  boost::lexical_cast<std::string>(get_pointed_type(tree_helper::get_type_index(TM,left_op))) + ") is " + pointed_type->get_kind_text());
         if (pointed_type->get_kind() == void_type_K)
         {
            void_type * vt = GetPointer<void_type>(pointed_type);
            deltabit = vt->algn;
         }
         else
         {
            deltabit = tree_helper::get_object_size(pointed_type);
         }

         long long int pointer_offset = 0;
         std::string right_offset_var;
         if (right_cost)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Offset is constant");
            pointer_offset = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(right_op_node));
            if (pointer_offset % (deltabit / 8) == 0)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Arithmetic pointer pattern matched");
               integer_cst * ic = GetPointer<integer_cst>(right_op_node);
               tree_nodeRef ic_type = GET_NODE(ic->type);
               integer_type* it = GetPointer<integer_type>(ic_type);
               if (it && (it->prec == 32))
               {
                  pointer_offset = static_cast<int>(pointer_offset) / static_cast<int>(deltabit / 8);
                  pointer_offset = static_cast<unsigned int>(pointer_offset);
               }
               else
               {
                  pointer_offset = pointer_offset / (deltabit / 8);
               }
               do_reverse_pointer_arithmetic = true;
            }
            else
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Arithmetic pointer pattern not matched " + boost::lexical_cast<std::string>(pointer_offset) + " vs " + boost::lexical_cast<std::string>(deltabit / 8));
            }
         }
         else
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Offset is not constant");
            bool exit = false;
            while(!do_reverse_pointer_arithmetic and !exit)
            {
               switch(right_op_node->get_kind())
               {
                  case(ssa_name_K) :
                  {
                     if(GetPointer<ssa_name>(right_op_node)->def_stmts.size() == 0)
                     {
                        exit = true;
                        break;
                     }
                     ssa_name * rssa = GetPointer<ssa_name>(right_op_node);
                     tree_nodeRef defstmt = GET_NODE(rssa->def_stmts.front());
                     if(defstmt->get_kind() != gimple_assign_K)
                     {
                        exit = true;
                        break;
                     }
                     right_op_node = GET_NODE(GetPointer<gimple_assign>(defstmt)->op1);
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "New op node is " + boost::lexical_cast<std::string>(GET_INDEX_NODE(GetPointer<gimple_assign>(defstmt)->op1)) + " - " + right_op_node->get_kind_text());
                     break;
                  }
                  case(mult_expr_K) :
                  {
                     mult_expr * mult = GetPointer<mult_expr>(right_op_node);
                     if (GET_NODE(mult->op1)->get_kind() == integer_cst_K)
                     {
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Right part of multiply is an integer constant " + boost::lexical_cast<std::string>(GET_INDEX_NODE(mult->op1)));
                        long long size_of_pointer = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(mult->op1)));
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Size of pointer is " + boost::lexical_cast<std::string>(size_of_pointer));
                        if (size_of_pointer == (deltabit / 8))
                        {
                           INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Constant is the size of the pointed");
                           right_offset_var += print_node(GET_INDEX_NODE(mult->op0), v, vppf);
                           do_reverse_pointer_arithmetic = true;
                        }
                        else
                        {
                           INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Constant is not the size of the pointed: " + boost::lexical_cast<std::string>(size_of_pointer) + " vs " + boost::lexical_cast<std::string>(deltabit / 8));
                           unsigned int integer_type_index = 0;
                           tree_nodeRef temp1 = tree_helper::get_type_node(GET_NODE(mult->op1), integer_type_index);
                           THROW_ASSERT(GetPointer<integer_type>(temp1), "Type of integer cast " + boost::lexical_cast<std::string>(GET_INDEX_NODE(mult->op1)) + " is not an integer type");
                           const integer_type * it = GetPointer<integer_type>(temp1);
                           unsigned long long max_int = static_cast<unsigned long long>(tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(it->max))));
                           long long new_size_of_pointer = static_cast<long long>(max_int) + 1 - size_of_pointer;
                           if (new_size_of_pointer == (deltabit / 8))
                           {
                              INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Constant is minus the size of the pointed");
                              right_offset_var += "-(" + print_node(GET_INDEX_NODE(mult->op0), v, vppf) + ")";
                              do_reverse_pointer_arithmetic = true;
                           }
                           else if((size_of_pointer % ((deltabit / 8)) == 0))
                           {
                              INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Constant is a multiple of the  size of the pointed");
                              right_offset_var += print_node(GET_INDEX_NODE(mult->op0), v, vppf);
                              right_offset_var += " * ";
                              right_offset_var += boost::lexical_cast<std::string>(size_of_pointer / ((deltabit / 8)));
                              do_reverse_pointer_arithmetic = true;
                           }

                        }
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Right offset variable is " + right_offset_var);
                     }
                     else if (GET_NODE(mult->op0)->get_kind() == integer_cst_K)
                     {
                        long long size_of_pointer = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(mult->op0)));
                        if (size_of_pointer == (deltabit / 8))
                        {
                           right_offset_var += print_node(GET_INDEX_NODE(mult->op1), v, vppf);
                           do_reverse_pointer_arithmetic = true;
                        }
                        else
                        {
                           unsigned int integer_type_index = 0;
                           tree_nodeRef temp1 = tree_helper::get_type_node(GET_NODE(mult->op0), integer_type_index);
                           THROW_ASSERT(GetPointer<integer_type>(temp1), "Type of integer cast " + boost::lexical_cast<std::string>(GET_INDEX_NODE(mult->op0)) + " is not an integer type");
                           const integer_type * it = GetPointer<integer_type>(temp1);
                           unsigned long long max_int = static_cast<unsigned long long>(tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(it->max))));
                           long long new_size_of_pointer = static_cast<long long>(max_int) + 1 - size_of_pointer;
                           if (new_size_of_pointer == (deltabit / 8))
                           {
                              right_offset_var += "-" + print_node(GET_INDEX_NODE(mult->op1), v, vppf);
                              do_reverse_pointer_arithmetic = true;
                           }
                           else if((size_of_pointer % ((deltabit / 8)) == 0))
                           {
                              right_offset_var += print_node(GET_INDEX_NODE(mult->op1), v, vppf);
                              right_offset_var += " * ";
                              right_offset_var += boost::lexical_cast<std::string>(size_of_pointer / ((deltabit / 8)));
                              do_reverse_pointer_arithmetic = true;
                           }
                        }
                     }
                     exit = true;
                     break;
                  }
                  case(negate_expr_K) :
                  {
                     right_offset_var += "-";
                     negate_expr * ne = GetPointer<negate_expr>(right_op_node);
                     right_op_node = GET_NODE(ne->op);
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "New op node is " + boost::lexical_cast<std::string>(GET_INDEX_NODE(ne->op)) + " - " + right_op_node->get_kind_text());
                     break;
                  }
                  case(nop_expr_K) :
                  {
                     nop_expr * ne = GetPointer<nop_expr>(right_op_node);
                     right_op_node = GET_NODE(ne->op);
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "New op node is " + boost::lexical_cast<std::string>(GET_INDEX_NODE(ne->op)) + " - " + right_op_node->get_kind_text());
                     break;
                  }
                  default:
                  {
                     exit = true;
                     break;
                  }
               }
            }
            if (deltabit == 8)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Reversing pointer arithmetic sucessful");
               do_reverse_pointer_arithmetic = true;
               right_offset_var = print_node(right_op, v, vppf);
            }
         }
         bool char_pointer = false;
         if (!do_reverse_pointer_arithmetic and get_size(GET_INDEX_NODE(GetPointer<pointer_type>(type_node)->ptd)) == 8)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Reversing pointer arithmetic sucessful because of char pointer");
            do_reverse_pointer_arithmetic = true;
            char_pointer = true;
         }
         if (binary_op_cast && !do_reverse_pointer_arithmetic)
            res += "(" + tree_helper::print_type(TM, GET_INDEX_NODE(ppe->type)) + ")(";
         if ((left_op_cast and (pointed_type->get_kind() == void_type_K)) or not do_reverse_pointer_arithmetic)
         {
            res += "((unsigned char*)";
         }
         res += print_node(left_op, v, vppf);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "After printing of left part " + res);
         if ((left_op_cast and (pointed_type->get_kind() == void_type_K)) or not do_reverse_pointer_arithmetic)
            res += ")";
         res += std::string(" ") + op + " ";
         if (!do_reverse_pointer_arithmetic)
         {
            TM->increment_unremoved_pointer_plus();
         }
         else
         {
            TM->increment_removable_pointer_plus();
         }

         if (do_reverse_pointer_arithmetic and !char_pointer)
         {
            if (right_offset_var != "")
               res += right_offset_var;
            else
            {
               res += boost::lexical_cast<std::string>(pointer_offset);
               tree_nodeRef type = GET_NODE(GetPointer<integer_cst>(right_op_node)->type);
               integer_type* it = GetPointer<integer_type>(type);
               bool unsigned_flag = (it && it->unsigned_flag) || type->get_kind() == pointer_type_K || type->get_kind() == boolean_type_K || type->get_kind() == enumeral_type_K;
               if (unsigned_flag)
                  res += "u";
            }
         }
         else
         {
            if(right_op_node->get_kind() == integer_cst_K)
            {
               res += boost::lexical_cast<std::string>(tree_helper::get_integer_cst_value(GetPointer<integer_cst>(right_op_node)));
            }
            else
            {
               res += print_node(right_op, v, vppf);
            }
         }
         if (binary_op_cast and not do_reverse_pointer_arithmetic)
            res += ")";
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
         break;
      }
      /* Binary relational expressions.  */
      case lt_expr_K:
      case le_expr_K:
      case gt_expr_K:
      case ge_expr_K:
      case eq_expr_K:
      case ne_expr_K:
      {
         bool binary_op_cast = is_a_pointer(index);
         if (binary_op_cast)
            res += "((" + tree_helper::print_type(TM, get_type(index)) + ")(";
         const std::string op = tree_helper::op_symbol (node);
         binary_expr *be = GetPointer<binary_expr>(node);
         unsigned int left_op = GET_INDEX_NODE(be->op0);
         unsigned int right_op = GET_INDEX_NODE(be->op1);
         bool left_op_cast = is_a_pointer(left_op) and is_a_pointer(right_op) and (tree_helper::size(TM, left_op) != tree_helper::size(TM, right_op));
         bool right_op_cast = is_a_pointer(right_op) and is_a_pointer(left_op) and (tree_helper::size(TM, left_op) != tree_helper::size(TM, right_op));
         bool left_op_bracket = !(GetPointer<decl_node>(GET_NODE(be->op0)) || GetPointer<ssa_name>(GET_NODE(be->op0)));
         bool right_op_bracket = !(GetPointer<decl_node>(GET_NODE(be->op1)) || GetPointer<ssa_name>(GET_NODE(be->op1)));
         if (left_op_bracket)
            res += "(";
         if (left_op_cast)
            res += "((unsigned long int)";
         res += print_node(left_op, v, vppf);
         if (left_op_cast)
            res += ")";
         if (left_op_bracket)
            res += ")";
         res += std::string(" ") + op + " ";
         if (right_op_bracket)
            res += "(";
         if (right_op_cast)
            res += "((unsigned long int)";
         res += print_node(right_op, v, vppf);
         if (right_op_cast)
            res += ")";
         if (right_op_bracket)
            res += ")";
         if (binary_op_cast)
            res += "))";
         break;
      }
      case rrotate_expr_K:
      case lrotate_expr_K:
      {
         bool binary_op_cast = is_a_pointer(index);
         unsigned int type_index = get_type(index);
         const tree_nodeRef type = TM->get_tree_node_const(type_index);
         if (binary_op_cast)
            res += "((" + tree_helper::print_type(TM, type_index) + ")(";
         binary_expr *be = GetPointer<binary_expr>(node);
         unsigned int left_op = GET_INDEX_NODE(be->op0);
         bool left_op_cast = is_a_pointer(left_op);
         unsigned int right_op = GET_INDEX_NODE(be->op1);
         bool right_op_cast = is_a_pointer(right_op);
         if (left_op_cast)
            res += "((unsigned long int)";
         res += print_node(left_op, v, vppf);
         if (left_op_cast)
            res += ")";
         if (node->get_kind() == rrotate_expr_K)
            res += " >> ";
         else
            res += " << ";
         if (right_op_cast)
            res += "((unsigned long int)";
         res += print_node(right_op, v, vppf);
         if (right_op_cast)
            res += ")";
         res += " | ";
         if (left_op_cast)
            res += "((unsigned long int)";
         res += print_node(left_op, v, vppf);
         if (left_op_cast)
            res += ")";
         if (node->get_kind() == rrotate_expr_K)
            res += " << ";
         else
            res += " >> ";
         res += "(" + boost::lexical_cast<std::string>(GetPointer<integer_type>(type)->prec) + "-";
         if (right_op_cast)
            res += "((unsigned long int)";
         res += print_node(right_op, v, vppf);
         if (right_op_cast)
            res += ")";
         res += ")";
         if (binary_op_cast)
            res += "))";
         break;
      }
      /* Unary arithmetic and logic expressions.  */
      case negate_expr_K:
      case bit_not_expr_K:
      case truth_not_expr_K:
      case reference_expr_K:
      case predecrement_expr_K:
      case preincrement_expr_K:
      {
         const std::string op = tree_helper::op_symbol (node);
         unary_expr *ue = GetPointer<unary_expr>(node);
         res = res + " " + op + "(" + print_node(GET_INDEX_NODE(ue->op), v, vppf) + ")";
         break;
      }
      case realpart_expr_K:
      case imagpart_expr_K:
      {
         const std::string op = tree_helper::op_symbol (node);
         unary_expr *ue = GetPointer<unary_expr>(node);
         res = res + " " + op + print_node(GET_INDEX_NODE(ue->op), v, vppf);
         ssa_name* sa = GetPointer<ssa_name>(GET_NODE(ue->op));
         if (sa and (sa->volatile_flag || GET_NODE(sa->def_stmts.front())->get_kind() == gimple_nop_K) and GetPointer<var_decl>(GET_NODE(sa->var)))
         {
            res += " = 0";
         }
         break;
      }
      case addr_expr_K:
      {
         addr_expr *ue = GetPointer<addr_expr>(node);
         if(GetPointer<component_ref>(GET_NODE(ue->op)) and has_bit_field(GET_INDEX_NODE(GetPointer<component_ref>(GET_NODE(ue->op))->op1)))
         {
            THROW_ERROR_CODE(BITFIELD_EC, "Trying to get the address of a bitfield");
         }
         if(GET_NODE(ue->op)->get_kind() == array_ref_K)
         {
            const array_ref * ar = GetPointer<array_ref>(GET_NODE(ue->op));
            ///&string[0]
            if(GET_NODE(ar->op0)->get_kind() == string_cst_K and GET_NODE(ar->op1)->get_kind() == integer_cst_K and tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(ar->op1))) == 0)
            {
               res += print_node(GET_INDEX_NODE(ar->op0), v, vppf);
               break;
            }
         }
         bool prefix = true;
         if (GetPointer<var_decl>(GET_NODE(ue->op)))
         {
            var_decl* vd = GetPointer<var_decl>(GET_NODE(ue->op));
            switch (GET_NODE(vd->type)->get_kind())
            {
               case array_type_K:
                  prefix = false;
                  break;
               default:
                  prefix = true;
            }
         }
         res += " ";
         if (prefix)
         {
            const std::string op = tree_helper::op_symbol (node);
            res += op;
            if (GET_NODE(ue->op)->get_kind() == label_decl_K)
               res += op;
         }
         res += print_node(GET_INDEX_NODE(ue->op), v, vppf);
         break;
      }
      case function_decl_K:
      {
         function_decl * fd = GetPointer<function_decl>(node);
         res += tree_helper::print_function_name(TM, fd);
         break;
      }
      case fix_trunc_expr_K:
      case fix_ceil_expr_K:
      case fix_floor_expr_K:
      case fix_round_expr_K:
      case float_expr_K:
      case convert_expr_K:
      case nop_expr_K:
      {
         unary_expr *ue = GetPointer<unary_expr>(node);
         unsigned int return_index = 0;
         const tree_nodeRef type = tree_helper::get_type_node(node, return_index);
         unsigned int prec = 0;
         if (type && (type->get_kind() == integer_type_K))
            prec = GetPointer<integer_type>(type)->prec;
         unsigned int algn = 0;
         if (type && (type->get_kind() == integer_type_K))
            algn = GetPointer<integer_type>(type)->algn;
         //bitfield type
         if (prec != algn && prec % algn)
         {
            res += "(";
            res += print_node(GET_INDEX_NODE(ue->op), v, vppf);
            res += ")%(1";
            if (prec > 32)
               res += "LL";
            if (GetPointer<integer_type>(type)->unsigned_flag)
               res += "U";
            res += " << " + boost::lexical_cast<std::string>(prec) + ")";
         }
         else
         {
            res = res + "(" + tree_helper::print_type(TM, GET_INDEX_NODE(ue->type)) + ") (";
            res += print_node(GET_INDEX_NODE(ue->op), v, vppf);
            res += ")";
         }
         break;
      }
      case view_convert_expr_K:
      {
         const view_convert_expr * vce = GetPointer<view_convert_expr>(node);
         if(GetPointer<integer_cst>(GET_NODE(vce->op)))
         {
            THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "view_convert_expr of integer_cst not supported");
         }
         res = res + "*((" + tree_helper::print_type(TM, GET_INDEX_NODE(vce->type)) + " * ) &(" + print_node(GET_INDEX_NODE(vce->op), v, vppf)+ "))";
         break;
      }
      case component_ref_K:
      {
         component_ref * cr = GetPointer<component_ref>(node);
         res = "(" + print_node(GET_INDEX_NODE(cr->op0), v, vppf) + ")." + print_node(GET_INDEX_NODE(cr->op1), v, vppf);
         break;
      }
      case indirect_ref_K:
      {
         indirect_ref * ir = GetPointer<indirect_ref>(node);
         res = "*(" + print_node(GET_INDEX_NODE(ir->op), v, vppf) + ")";
         break;
      }
      case mem_ref_K:
      {
         mem_ref * mr = GetPointer<mem_ref>(node);
         long long int offset = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(mr->op1)));
         if(offset==0)
            res = "*(" + print_node(GET_INDEX_NODE(mr->op0), v, vppf) + ")";
         else
            res = "*((" + tree_helper::print_type(TM, GET_INDEX_NODE(mr->type)) + "*)(((unsigned char*)" + print_node(GET_INDEX_NODE(mr->op0), v, vppf) + ") + " + boost::lexical_cast<std::string>(offset) + "))";
         break;
      }
      case target_mem_ref_K:
      {
         target_mem_ref * tmr = GetPointer<target_mem_ref>(node);
         bool need_plus = false;
         res = "(*((" + tree_helper::print_type(TM, GET_INDEX_NODE(tmr->type)) + "*)(";
         if (tmr->symbol)
         {
            res += "((unsigned char*)" + (tree_helper::is_a_struct(TM, GET_INDEX_NODE(tmr->symbol)) || tree_helper::is_an_union(TM, GET_INDEX_NODE(tmr->symbol))?std::string("&"):std::string("")) + print_node(GET_INDEX_NODE(tmr->symbol), v, vppf) + ")";
            need_plus = true;
         }
         if (tmr->base)
         {
            if (need_plus)
               res += "+";
            else
               need_plus = true;
            if(tmr->symbol)
               res += "(" + print_node(GET_INDEX_NODE(tmr->base), v, vppf) + ")";
            else
               res += "((unsigned char*)" + print_node(GET_INDEX_NODE(tmr->base), v, vppf) + ")";
         }
         if (tmr->step)
         {
            if (need_plus)
               res += "+";
            need_plus = false;
            res += print_node(GET_INDEX_NODE(tmr->step), v, vppf) + "*";
            THROW_ASSERT(tmr->idx, "idx expected!");
         }
         if (tmr->idx)
         {
            if (need_plus)
               res += "+";
            else
               need_plus = true;
            res += print_node(GET_INDEX_NODE(tmr->idx), v, vppf);
         }
         if (tmr->offset)
         {
            if (need_plus)
               res += "+";
            res += print_node(GET_INDEX_NODE(tmr->offset), v, vppf);
         }
         res += ")))";
         break;
      }
      case target_mem_ref461_K:
      {
         target_mem_ref461 * tmr = GetPointer<target_mem_ref461>(node);
         bool need_plus = false;
         res = "(*((" + tree_helper::print_type(TM, GET_INDEX_NODE(tmr->type)) + "*)(";
         if (tmr->base)
         {
            need_plus = true;
            res += "((unsigned char*)" + print_node(GET_INDEX_NODE(tmr->base), v, vppf) + ")";
         }
         if (tmr->step)
         {
            if (need_plus)
               res += "+";
            need_plus = false;
            res += print_node(GET_INDEX_NODE(tmr->step), v, vppf) + "*";
            THROW_ASSERT(tmr->idx, "idx expected!");
         }
         if (tmr->idx)
         {
            if (need_plus)
               res += "+";
            else
               need_plus = true;
            res += print_node(GET_INDEX_NODE(tmr->idx), v, vppf);
         }
         if (tmr->idx2)
         {
            if (need_plus)
               res += "+";
            else
               need_plus = true;
            res += print_node(GET_INDEX_NODE(tmr->idx2), v, vppf);
         }
         if (tmr->offset)
         {
            if (need_plus)
               res += "+";
            res += print_node(GET_INDEX_NODE(tmr->offset), v, vppf);
         }
         res += ")))";
         break;
      }
      case array_ref_K:
      {
         array_ref * ar = GetPointer<array_ref>(node);
         tree_nodeRef base = GET_NODE(ar->op0);
         //tree_nodeRef offset = GET_NODE(ar->op1);
         if(base->get_kind() == mem_ref_K)
         {
            mem_ref * mr = GetPointer<mem_ref>(base);
            long long int offset = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(mr->op1)));
            if(offset==0)
               res = "(" + print_node(GET_INDEX_NODE(mr->op0), v, vppf) + ")";
            else
               res = "((" + tree_helper::print_type(TM, GET_INDEX_NODE(ar->type)) + "*)(((unsigned char*)" + print_node(GET_INDEX_NODE(mr->op0), v, vppf) + ") + " + boost::lexical_cast<std::string>(offset) + "))";
         }
         else
            res = "(" + print_node(GET_INDEX_NODE(ar->op0), v, vppf) + ")";
         res += "[" + print_node(GET_INDEX_NODE(ar->op1), v, vppf) + "]";
         break;
      }
      case bit_field_ref_K:
      {
#if HAVE_SPARC_COMPILER
         if(Param->getOption<CompilerTarget>(OPT_default_compiler) == CT_SPARC_GCC)
         {
            THROW_ERROR_CODE(BITFIELD_EC, "Bitfield not supported by sparc cross compiler");
         }
#endif
         bit_field_ref * bf = GetPointer<bit_field_ref>(node);
         long long int bpos = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(bf->op2)));
         res += "*((" + tree_helper::print_type(TM, GET_INDEX_NODE(bf->type)) + "* ) (((unsigned long int) &(" + print_node(GET_INDEX_NODE(bf->op0), v, vppf) + ")) + (unsigned long int)" + print_node(GET_INDEX_NODE(bf->op2), v, vppf) + "/8))";
         if(bpos % 8)
            res += " >> " + boost::lexical_cast<std::string>(bpos % 8);
         break;
      }
      /* post/pre increment and decrement operator should be treated as unary even if they are binary_expr*/
      case postdecrement_expr_K:
      case postincrement_expr_K:
      {
         const std::string op = tree_helper::op_symbol (node);
         binary_expr *be = GetPointer<binary_expr>(node);
         res += print_node(GET_INDEX_NODE(be->op0), v, vppf) + op + print_node(GET_INDEX_NODE(be->op1), v, vppf);
         break;
      }
      case min_expr_K:
      {
         min_expr *me = GetPointer<min_expr>(node);
         std::string op_0 = print_node(GET_INDEX_NODE(me->op0), v, vppf), op_1 = print_node(GET_INDEX_NODE(me->op1), v, vppf);
         res += op_0 + " < " + op_1 + " ? " + op_0 + " : " + op_1;
         break;
      }
      case max_expr_K:
      {
         max_expr *me = GetPointer<max_expr>(node);
         std::string op_0 = print_node(GET_INDEX_NODE(me->op0), v, vppf), op_1 = print_node(GET_INDEX_NODE(me->op1), v, vppf);
         res += op_0 + " > " + op_1 + " ? " + op_0 + " : " + op_1;
         break;
      }
      case unordered_expr_K:
      {
         binary_expr *be = GetPointer<binary_expr>(node);
         std::string op_0 = print_node(GET_INDEX_NODE(be->op0), v, vppf), op_1 = print_node(GET_INDEX_NODE(be->op1), v, vppf);
         res += "__builtin_isunordered(" + op_0 + "," + op_1 + ")";
         break;
      }
      case ordered_expr_K:
      {
         binary_expr *be = GetPointer<binary_expr>(node);
         std::string op_0 = print_node(GET_INDEX_NODE(be->op0), v, vppf), op_1 = print_node(GET_INDEX_NODE(be->op1), v, vppf);
         res += "!__builtin_isunordered(" + op_0 + "," + op_1 + ")";
         break;
      }
      case unlt_expr_K:
      {
         binary_expr *be = GetPointer<binary_expr>(node);
         std::string op_0 = print_node(GET_INDEX_NODE(be->op0), v, vppf), op_1 = print_node(GET_INDEX_NODE(be->op1), v, vppf);
         res += "!__builtin_isgreaterequal(" + op_0 + "," + op_1 + ")";
         break;
      }
      case unle_expr_K:
      {
         binary_expr *be = GetPointer<binary_expr>(node);
         std::string op_0 = print_node(GET_INDEX_NODE(be->op0), v, vppf), op_1 = print_node(GET_INDEX_NODE(be->op1), v, vppf);
         res += "!__builtin_isgreater(" + op_0 + "," + op_1 + ")";
         break;
      }
      case ungt_expr_K:
      {
         binary_expr *be = GetPointer<binary_expr>(node);
         std::string op_0 = print_node(GET_INDEX_NODE(be->op0), v, vppf), op_1 = print_node(GET_INDEX_NODE(be->op1), v, vppf);
         res += "!__builtin_islessequal(" + op_0 + "," + op_1 + ")";
         break;
      }
      case unge_expr_K:
      {
         binary_expr *be = GetPointer<binary_expr>(node);
         std::string op_0 = print_node(GET_INDEX_NODE(be->op0), v, vppf), op_1 = print_node(GET_INDEX_NODE(be->op1), v, vppf);
         res += "!__builtin_isless(" + op_0 + "," + op_1 + ")";
         break;
      }
      case uneq_expr_K:
      {
         binary_expr *be = GetPointer<binary_expr>(node);
         std::string op_0 = print_node(GET_INDEX_NODE(be->op0), v, vppf), op_1 = print_node(GET_INDEX_NODE(be->op1), v, vppf);
         res += "!__builtin_islessgreater(" + op_0 + "," + op_1 + ")";
         break;
      }
      case ltgt_expr_K:
      {
         binary_expr *be = GetPointer<binary_expr>(node);
         std::string op_0 = print_node(GET_INDEX_NODE(be->op0), v, vppf), op_1 = print_node(GET_INDEX_NODE(be->op1), v, vppf);
         res += "__builtin_islessgreater(" + op_0 + "," + op_1 + ")";
         break;
      }
      case abs_expr_K:
      {
         abs_expr *ae = GetPointer<abs_expr>(node);
         tree_nodeRef type = GET_NODE(ae->type);
         std::string op_0 = print_node(GET_INDEX_NODE(ae->op), v, vppf);
         if (GetPointer<real_type>(type))
         {
            real_type * rt = GetPointer<real_type>(type);
            if (rt->prec == 80)
               res += "__builtin_fabsl(" + op_0 + ")";
            else if (rt->prec == 64)
               res += "__builtin_fabs(" + op_0 + ")";
            else if (rt->prec == 32)
               res += "__builtin_fabsf(" + op_0 + ")";
            else
               THROW_ERROR("Abs on a real number with not supported precision");
         }
         else
            res += "(" + op_0 + ") >= 0 ? (" + op_0 + ") : -(" + op_0 + ")";
         //            res += "__builtin_llabs(" + op_0 + ")";
         break;
      }
      case complex_expr_K:
      {
         complex_expr *ce = GetPointer<complex_expr>(node);
         std::string op_0 = print_node(GET_INDEX_NODE(ce->op0), v, vppf), op_1 = print_node(GET_INDEX_NODE(ce->op1), v, vppf);
         res += op_0 + "+ 1i*" + op_1;
         break;
      }
      case cond_expr_K:
      {
         cond_expr *ce = GetPointer<cond_expr>(node);
         res = print_node(GET_INDEX_NODE(ce->op0), v, vppf) + " ? " + print_node(GET_INDEX_NODE(ce->op1), v, vppf) + " : " + print_node(GET_INDEX_NODE(ce->op2), v, vppf);
         break;
      }
      case gimple_cond_K:
      {
         gimple_cond *gc = GetPointer<gimple_cond>(node);
         std::vector<tree_nodeRef> pragmas = gc->pragmas;
         std::vector<tree_nodeRef>::const_iterator pragma, pragma_end = pragmas.end();
         for(pragma = pragmas.begin(); pragma != pragma_end; pragma++)
         {
            res += print_node(GET_INDEX_NODE(*pragma), v, vppf) + "\n";
         }
         res = "if (" + print_node(GET_INDEX_NODE(gc->op0), v, vppf) + ")";
         break;
      }
      case while_expr_K:
      {
         while_expr *we = GetPointer<while_expr>(node);
         res = "while (" + print_node(GET_INDEX_NODE(we->op0), v, vppf) + ")";
         break;
      }
      case for_expr_K:
      {
         for_expr * fe = GetPointer<for_expr>(node);
#if !RELEASE
         if(fe->omp_for)
         {
            res = "//#pragma omp " + print_node(GET_INDEX_NODE(fe->omp_for), v, vppf) + "\n";
         }
#endif
         res += "for (" + print_node(GET_INDEX_NODE(fe->op1), v, vppf) + "; ";
         res += print_node(GET_INDEX_NODE(fe->op0), v, vppf) + "; ";
         res += print_node(GET_INDEX_NODE(fe->op2), v, vppf);
         res += ")";
         break;
      }
      case gimple_switch_K:
      {
         gimple_switch *se = GetPointer<gimple_switch>(node);
         res += "switch(" + print_node(GET_INDEX_NODE(se->op0), v, vppf) + ")";
         break;
      }
      case gimple_assign_K:
      {
         gimple_assign * ms = GetPointer<gimple_assign>(node);
         if(!ms->init_assignment)
         {
            std::vector<tree_nodeRef> pragmas = ms->pragmas;
            std::vector<tree_nodeRef>::const_iterator pragma, pragma_end = pragmas.end();
            for(pragma = pragmas.begin(); pragma != pragma_end; pragma++)
            {
               res += print_node(GET_INDEX_NODE(*pragma), v, vppf) + "\n";
            }
            res = "";
            if (is_an_array(GET_INDEX_NODE(ms->op0)))
            {
               unsigned int type = tree_helper::get_type_index(TM, GET_INDEX_NODE(ms->op0));
               long long size = tree_helper::size(TM, type);
               res += "memcpy(";
               res += print_node(GET_INDEX_NODE(ms->op0), v, vppf) + ", ";
               if(GET_NODE(ms->op1)->get_kind() == view_convert_expr_K)
               {
                  const view_convert_expr * vce = GetPointer<view_convert_expr>(GET_NODE(ms->op1));
                  res += "&" + print_node(GET_INDEX_NODE(vce->op), v, vppf) + ", ";
               }
               else
               {
                  res += print_node(GET_INDEX_NODE(ms->op1), v, vppf) + ", ";
               }
               res += boost::lexical_cast<std::string>(size / 8) + ")";
               break;
            }
            if((is_a_struct(GET_INDEX_NODE(ms->op0)) or is_an_union(GET_INDEX_NODE(ms->op0))) and (is_a_struct(GET_INDEX_NODE(ms->op1)) or is_an_union(GET_INDEX_NODE(ms->op1))) and tree_helper::name_type(TM, tree_helper::GetRealType(TM, tree_helper::get_type_index(TM, GET_INDEX_NODE(ms->op0)))) != tree_helper::name_type(TM, tree_helper::GetRealType(TM, tree_helper::get_type_index(TM, GET_INDEX_NODE(ms->op1)))))
            {
               THROW_ERROR_CODE(C_EC, "Implicit struct type defintion not supported in gimple assignment " + boost::lexical_cast<std::string>(index));
            }
            res += print_node(GET_INDEX_NODE(ms->op0), v, vppf) + " = ";
            unsigned int return_index;
            const tree_nodeRef type = tree_helper::get_type_node(GET_NODE(ms->op0), return_index);
            const tree_nodeRef right = GET_NODE(ms->op1);
            ///check for type conversion
            switch (right->get_kind())
            {
               case constructor_K:
               case vector_cst_K:
               {
                  res += "(" + tree_helper::print_type(TM, get_type(GET_INDEX_NODE(ms->op0))) + ") ";
                  res += print_node(GET_INDEX_NODE(ms->op1), v, vppf);
                     break;
               }
               default:
               {
                  res += print_node(GET_INDEX_NODE(ms->op1), v, vppf);
                  break;
               }
            }
         }
         break;
      }
      case gimple_nop_K:
      {
         break;
      }
      case init_expr_K:
      {
         binary_expr *be = GetPointer<binary_expr>(node);
         res = print_node(GET_INDEX_NODE(be->op0), v, vppf) + " = ";
         const tree_nodeRef right = GET_NODE(be->op1);
         ///check for type conversion
         switch (right->get_kind())
         {
            case fix_trunc_expr_K:
            case fix_ceil_expr_K:
            case fix_floor_expr_K:
            case fix_round_expr_K:
            case float_expr_K:
            case convert_expr_K:
            case view_convert_expr_K:
            case nop_expr_K:
            {
               unary_expr *ue = GetPointer<unary_expr>(right);
               res = res + "(" + tree_helper::print_type(TM, get_type(GET_INDEX_NODE(be->op0))) + ") ";
               res += print_node(GET_INDEX_NODE(ue->op), v, vppf);
               break;
            }
            default:
               res += print_node(GET_INDEX_NODE(be->op1), v, vppf);
         }
         break;
      }
      case gimple_return_K:
      {
         gimple_return *re = GetPointer<gimple_return>(node);
         std::vector<tree_nodeRef> pragmas = re->pragmas;
         std::vector<tree_nodeRef>::const_iterator pragma, pragma_end = pragmas.end();
         for(pragma = pragmas.begin(); pragma != pragma_end; pragma++)
         {
            res += print_node(GET_INDEX_NODE(*pragma), v, vppf) + "\n";
         }
         res += "return ";
         if (re->op != NULL)
         {
            const tree_nodeRef return_node = GET_NODE(re->op);
            ///check for type conversion
            switch (return_node->get_kind())
            {
               case fix_trunc_expr_K:
               case fix_ceil_expr_K:
               case fix_floor_expr_K:
               case fix_round_expr_K:
               case float_expr_K:
               case convert_expr_K:
               case nop_expr_K:
                  {
                     unary_expr *ue = GetPointer<unary_expr>(return_node);
                     res += "(" + tree_helper::print_type(TM, get_type(GET_INDEX_NODE(re->op))) + ") (";
                     res += print_node(GET_INDEX_NODE(ue->op), v, vppf);
                     res += ")";
                     break;
                  }
               case constructor_K:
               case vector_cst_K:
                  {
                     res += "(" + tree_helper::print_type(TM, get_type(GET_INDEX_NODE(re->op))) + ") ";
                     res += print_node(GET_INDEX_NODE(re->op), v, vppf);
                     break;
                  }
               default:
                  {
                     res += print_node(GET_INDEX_NODE(re->op), v, vppf);
                     break;
                  }
            }
         }
         res += ";";
         break;
      }
      case call_expr_K:
      {
         call_expr *ce = GetPointer<call_expr>(node);
         function_decl * fd = 0;
         tree_nodeRef op0 = GET_NODE(ce->fn);
         bool is_va_start_end = false;

         //sizeof workaroung
         bool is_sizeof = false;

         switch (op0->get_kind())
         {
            case addr_expr_K:
            {
               unary_expr *ue = GetPointer<unary_expr>(op0);
               tree_nodeRef fn = GET_NODE(ue->op);
               switch (fn->get_kind())
               {
                  case function_decl_K:
                  {
                     fd = GetPointer<function_decl>(fn);
                     ///__builtin_va_start should be ad-hoc managed
                     std::string fname = tree_helper::print_function_name(TM, fd);
                     if (fname == "__builtin_va_start" || fname == "__builtin_va_end" || fname == "__builtin_va_copy")
                        is_va_start_end = true;
                     if (fname == "__builtin_constant_p")
                     {
                        THROW_ERROR_CODE(C_EC, "Not supported function " + fname); 
                     }
                     if (fname == STR_CST_string_sizeof)
                     {
                        is_sizeof = true;
                        res += "sizeof";
                     }
                     else
                        res += fname;
                     break;
                  }
                  default:
                     THROW_ERROR(std::string("tree node not currently supported ") + fn->get_kind_text());
               }
               break;
            }
            case obj_type_ref_K:
            {
               tree_nodeRef fn = tree_helper::find_obj_type_ref_function(ce->fn);
               switch (GET_NODE(fn)->get_kind())
               {
                  case function_decl_K:
                  {
                     function_decl * local_fd = GetPointer<function_decl>(GET_NODE(fn));
                     if (tree_helper::has_function_return(TM, GET_INDEX_NODE(fn)))
                        res += print_node(GET_INDEX_NODE(fn), v, vppf) + " = ";
                     res += tree_helper::print_function_name(TM, local_fd);
                     break;
                  }
                  default:
                     THROW_ERROR(std::string("tree node not currently supported ") + fn->get_kind_text());
               }
               break;
            }
            case ssa_name_K:
            {
               res += "(*" + print_node(GET_INDEX_NODE(ce->fn), v, vppf) + ")";
               break;
            }
            default:
               THROW_ERROR(std::string("tree node not currently supported ") + op0->get_kind_text());
         }
         /// Print parameters.
         res += "(";
         if (is_va_start_end)
         {
            THROW_ASSERT(ce->args.size(), "va_start or va_end have to have arguments");
            const tree_nodeRef par1 = GET_NODE(ce->args[0]);
            const unsigned int par1_index = GET_INDEX_NODE(ce->args[0]);
            //print the first removing the address
            if (GetPointer<addr_expr>(par1))
            {
               res += print_node(GET_INDEX_NODE(GetPointer<addr_expr>(par1)->op), v, vppf);
            }
            else if (GetPointer<var_decl>(par1) || GetPointer<ssa_name>(par1))
            {
               res += "*(" + print_node(par1_index, v, vppf) + ")";
            }
            else
               THROW_ERROR("expected an address or a variable " + boost::lexical_cast<std::string>(par1_index));
            size_t arg_index;
            for(arg_index = 1; arg_index < ce->args.size(); arg_index++)
            {
               res += ", ";
               res += print_node(GET_INDEX_NODE(ce->args[arg_index]), v, vppf);
            }
         }
         else if(is_sizeof)
         {
            THROW_ASSERT(ce->args.size() == 1, "Wrong number of arguments: " + boost::lexical_cast<std::string>(ce->args.size()));
            const tree_nodeRef arg1 = GET_NODE(ce->args[0]);
#ifndef NDEBUG
            THROW_ASSERT(GetPointer<addr_expr>(arg1), "Argument is not an addr_expr but a " + std::string(arg1->get_kind_text()));
            const addr_expr * ae = GetPointer<addr_expr>(arg1);
            THROW_ASSERT(GetPointer<array_ref>(GET_NODE(ae->op)), "Argument is not an array ref but a " + std::string(GET_NODE(ae->op)->get_kind_text()));
            const array_ref * ar = GetPointer<array_ref>(GET_NODE(ae->op));
            THROW_ASSERT(GetPointer<string_cst>(GET_NODE(ar->op0)), "Argument is not a string cast but a " + std::string(GET_NODE(ar->op0)->get_kind_text()));
#endif
            std::string argument = print_node(GET_INDEX_NODE(ce->args[0]), v, vppf);
            std::string unquoted_argument = argument.substr(1, argument.size() - 2);
            res += unquoted_argument;
         }
         else
         {
            if (ce->args.size())
            {
               const std::vector<tree_nodeRef> & actual_args = ce->args;
               std::vector<tree_nodeRef> formal_args;
               if(fd)
                  formal_args = fd->list_of_args;
               std::vector<tree_nodeRef>::const_iterator actual_arg, actual_arg_end = actual_args.end();
               std::vector<tree_nodeRef>::const_iterator formal_arg, formal_arg_end = formal_args.end();
               for(actual_arg = actual_args.begin(), formal_arg = formal_args.begin(); actual_arg != actual_arg_end; actual_arg++)
               {
                  if(formal_arg != formal_arg_end and (is_a_struct(GET_INDEX_NODE(*actual_arg)) or is_an_union(GET_INDEX_NODE(*actual_arg))) and (is_a_struct(GET_INDEX_NODE(*formal_arg)) or is_an_union(GET_INDEX_NODE(*formal_arg))) and (tree_helper::GetRealType(TM, tree_helper::get_type_index(TM, GET_INDEX_NODE(*actual_arg))) != tree_helper::GetRealType(TM, tree_helper::get_type_index(TM, GET_INDEX_NODE(*formal_arg)))))
                  {
                     THROW_ERROR_CODE(C_EC, "Implicit struct type defintion not supported in gimple assignment " + boost::lexical_cast<std::string>(index));
                  }
                  if(actual_arg != actual_args.begin())
                     res += ", ";
                  if(GET_NODE(*actual_arg)->get_kind() == vector_cst_K)
                     res += "(" + tree_helper::print_type(TM, get_type(GET_INDEX_NODE(*actual_arg))) + ") ";

                  res += print_node(GET_INDEX_NODE(*actual_arg), v, vppf);
                  if(formal_arg != formal_arg_end)
                  {
                     formal_arg++;
                  }
               }
            }
         }
         res += ")";
         break;
      }
      case gimple_call_K:
      {
         gimple_call *ce = GetPointer<gimple_call>(node);
         function_decl * fd = 0;
         std::vector<tree_nodeRef> pragmas = ce->pragmas;
         std::vector<tree_nodeRef>::const_iterator pragma, pragma_end = pragmas.end();
         for(pragma = pragmas.begin(); pragma != pragma_end; pragma++)
         {
            res += print_node(GET_INDEX_NODE(*pragma), v, vppf) + "\n";
         }
         tree_nodeRef op0 = GET_NODE(ce->fn);
         bool is_va_start_end = false;

         //sizeof workaroung
         bool is_sizeof = false;

         switch (op0->get_kind())
         {
            case addr_expr_K:
            {
               unary_expr *ue = GetPointer<unary_expr>(op0);
               tree_nodeRef fn = GET_NODE(ue->op);
               switch (fn->get_kind())
               {
                  case function_decl_K:
                  {
                     fd = GetPointer<function_decl>(fn);
                     ///__builtin_va_start should be ad-hoc managed
                     std::string fname = tree_helper::print_function_name(TM, fd);
                     if (fname == "__builtin_va_start" || fname == "__builtin_va_end" || fname == "__builtin_va_copy")
                        is_va_start_end = true;
                     if (fname == "__builtin_constant_p")
                     {
                        THROW_ERROR_CODE(C_EC, "Not supported function " + fname);
                     }
                     if (fname == STR_CST_string_sizeof)
                     {
                        is_sizeof = true;
                        res += "sizeof";
                     }
                     else
                        res += fname;
                     break;
                  }
                  default:
                     THROW_ERROR(std::string("tree node not currently supported ") + fn->get_kind_text());
               }
               break;
            }
            case obj_type_ref_K:
            {
               tree_nodeRef fn = tree_helper::find_obj_type_ref_function(ce->fn);
               switch (GET_NODE(fn)->get_kind())
               {
                  case function_decl_K:
                  {
                     function_decl * local_fd = GetPointer<function_decl>(GET_NODE(fn));
                     if (tree_helper::has_function_return(TM, GET_INDEX_NODE(fn)))
                        res += print_node(GET_INDEX_NODE(fn), v, vppf) + " = ";
                     res += tree_helper::print_function_name(TM, local_fd);
                     break;
                  }
                  default:
                     THROW_ERROR(std::string("tree node not currently supported ") + fn->get_kind_text());
               }
               break;
            }
            case ssa_name_K:
            {
               res += "(*" + print_node(GET_INDEX_NODE(ce->fn), v, vppf) + ")";
               break;
            }
            default:
               THROW_ERROR(std::string("tree node not currently supported ") + op0->get_kind_text());
         }
         /// Print parameters.
         res += "(";
         if (is_va_start_end)
         {
            THROW_ASSERT(ce->args.size(), "va_start or va_end have to have arguments");
            const tree_nodeRef par1 = GET_NODE(ce->args[0]);
            const unsigned int par1_index = GET_INDEX_NODE(ce->args[0]);
            //print the first removing the address
            if (GetPointer<addr_expr>(par1))
            {
               res += print_node(GET_INDEX_NODE(GetPointer<addr_expr>(par1)->op), v, vppf);
            }
            else if (GetPointer<var_decl>(par1) || GetPointer<ssa_name>(par1))
            {
               res += "*(" + print_node(par1_index, v, vppf) + ")";
            }
            else
               THROW_ERROR("expected an address or a variable " + boost::lexical_cast<std::string>(par1_index));
            size_t arg_index;
            for(arg_index = 1; arg_index < ce->args.size(); arg_index++)
            {
               res += ", ";
               res += print_node(GET_INDEX_NODE(ce->args[arg_index]), v, vppf);
            }
         }
         else if(is_sizeof)
         {
            THROW_ASSERT(ce->args.size() == 1, "Wrong number of arguments: " + boost::lexical_cast<std::string>(ce->args.size()));
            const tree_nodeRef arg1 = GET_NODE(ce->args[0]);
   #ifndef NDEBUG
            THROW_ASSERT(GetPointer<addr_expr>(arg1), "Argument is not an addr_expr but a " + std::string(arg1->get_kind_text()));
            const addr_expr * ae = GetPointer<addr_expr>(arg1);
            THROW_ASSERT(GetPointer<array_ref>(GET_NODE(ae->op)), "Argument is not an array ref but a " + std::string(GET_NODE(ae->op)->get_kind_text()));
            const array_ref * ar = GetPointer<array_ref>(GET_NODE(ae->op));
            THROW_ASSERT(GetPointer<string_cst>(GET_NODE(ar->op0)), "Argument is not a string cast but a " + std::string(GET_NODE(ar->op0)->get_kind_text()));
   #endif
            std::string argument = print_node(GET_INDEX_NODE(ce->args[0]), v, vppf);
            std::string unquoted_argument = argument.substr(1, argument.size() - 2);
            res += unquoted_argument;
         }
         else
         {
            if (ce->args.size())
            {
               const std::vector<tree_nodeRef> & actual_args = ce->args;
               std::vector<tree_nodeRef> formal_args;
               if(fd)
                  formal_args = fd->list_of_args;
               std::vector<tree_nodeRef>::const_iterator actual_arg, actual_arg_end = actual_args.end();
               std::vector<tree_nodeRef>::const_iterator formal_arg, formal_arg_end = formal_args.end();
               for(actual_arg = actual_args.begin(), formal_arg = formal_args.begin(); actual_arg != actual_arg_end; actual_arg++)
               {
                  if(formal_arg != formal_arg_end and (is_a_struct(GET_INDEX_NODE(*actual_arg)) or is_an_union(GET_INDEX_NODE(*actual_arg))) and (is_a_struct(GET_INDEX_NODE(*formal_arg)) or is_an_union(GET_INDEX_NODE(*formal_arg))) and (tree_helper::GetRealType(TM, tree_helper::get_type_index(TM, GET_INDEX_NODE(*actual_arg))) != tree_helper::GetRealType(TM, tree_helper::get_type_index(TM, GET_INDEX_NODE(*formal_arg)))))
                  {
                     THROW_ERROR_CODE(C_EC, "Implicit struct type defintion not supported in gimple assignment " + boost::lexical_cast<std::string>(index));
                  }
                  if(actual_arg != actual_args.begin())
                     res += ", ";
                  if(GET_NODE(*actual_arg)->get_kind() == vector_cst_K)
                     res += "(" + tree_helper::print_type(TM, get_type(GET_INDEX_NODE(*actual_arg))) + ") ";

                  res += print_node(GET_INDEX_NODE(*actual_arg), v, vppf);
                  if(formal_arg != formal_arg_end)
                  {
                     formal_arg++;
                  }
               }
            }
         }
         res += ")";
         break;
      }
      case gimple_resx_K:
      {
         gimple_resx *re = GetPointer<gimple_resx>(node);
         res += "__gimple_resx(" + print_node(GET_INDEX_NODE(re->op), v, vppf) + ")";
         break;
      }
      case gimple_asm_K:
      {
         gimple_asm *ae = GetPointer<gimple_asm>(node);
         std::vector<tree_nodeRef> pragmas = ae->pragmas;
         std::vector<tree_nodeRef>::const_iterator pragma, pragma_end = pragmas.end();
         for(pragma = pragmas.begin(); pragma != pragma_end; pragma++)
         {
            res += print_node(GET_INDEX_NODE(*pragma), v, vppf);
         }
         res += "asm";
         if (ae->volatile_flag)
            res += " __volatile__ ";
         res += "(\"" + ae->str +"\"";
         if (ae->out)
         {
            res += ":";
            tree_list * tl = GetPointer<tree_list>(GET_NODE(ae->out));
            std::string out_string;
            do
            {
               tree_list * tl_purp = GetPointer<tree_list>(GET_NODE(tl->purp));
               out_string = "";
               do
               {
                  out_string += print_constant(GET_INDEX_NODE(tl_purp->valu), vppf);
                  if (tl_purp->chan)
                     tl_purp = GetPointer<tree_list>(GET_NODE(tl_purp->chan));
                  else
                     tl_purp = 0;
               }
               while (tl_purp);
               if (out_string == "\"=\"")
                  out_string = "\"=r\"";
               res += out_string;
               res += "(" + print_node(GET_INDEX_NODE(tl->valu), v, vppf) + ")";
               if (tl->chan)
               {
                  res += ",";
                  tl = GetPointer<tree_list>(GET_NODE(tl->chan));
               }
               else
                  tl = 0;
            }
            while (tl);
         }
         else if (ae->in || ae->clob)
            res += ":";
         if (ae->in)
         {
            res += ":";
            tree_list * tl = GetPointer<tree_list>(GET_NODE(ae->in));
            std::string in_string;
            do
            {
               tree_list * tl_purp = GetPointer<tree_list>(GET_NODE(tl->purp));
               in_string = "";
               do
               {
                  in_string += print_constant(GET_INDEX_NODE(tl_purp->valu), vppf);
                  if (tl_purp->chan)
                     tl_purp = GetPointer<tree_list>(GET_NODE(tl_purp->chan));
                  else
                     tl_purp = 0;
               }
               while (tl_purp);
               if (in_string == "\"\"")
                  in_string = "\"r\"";
               res += in_string;
               res += "(" + print_node(GET_INDEX_NODE(tl->valu), v, vppf) + ")";
               if (tl->chan)
               {
                  res += ",";
                  tl = GetPointer<tree_list>(GET_NODE(tl->chan));
               }
               else
                  tl = 0;
            }
            while (tl);
         }
         else if (ae->clob)
            res += ":";
         if (ae->clob)
         {
            res += ":";
            tree_list * tl = GetPointer<tree_list>(GET_NODE(ae->clob));
            do
            {
               res += print_node(GET_INDEX_NODE(tl->valu), v, vppf);
               if (tl->chan)
               {
                  res += " ";
                  tl = GetPointer<tree_list>(GET_NODE(tl->chan));
               }
               else
                  tl = 0;
            }
            while (tl);
         }
         res += ")";
         break;
      }
      case gimple_phi_K:
      {
         gimple_phi * pn = GetPointer<gimple_phi>(node);
         res += "/* " + print_node(GET_INDEX_NODE(pn->res), v, vppf) +" = gimple_phi(";
         std::vector<std::pair< tree_nodeRef, unsigned int> >::const_iterator it_uses_end = pn->list_of_def_edge.end();
         for(std::vector<std::pair< tree_nodeRef, unsigned int> >::const_iterator it_uses = pn->list_of_def_edge.begin(); it_uses != it_uses_end; )
         {
            res += print_node(GET_INDEX_NODE(it_uses->first), v, vppf);
            ++it_uses;
            if(it_uses != it_uses_end)
               res+= ", ";
         }
         res += ") */";
         break;
      }
      case ssa_name_K:
      {
         res += (*vppf)(index);
         //res += print_node(GET_INDEX_NODE(sn->var), v, vppf);
         /*if (!sn->volatile_flag || GET_NODE(sn->def_stmts.front())->get_kind() == gimple_nop_K)
          res += "_" + boost::lexical_cast<std::string>(sn->vers);*/
         break;
      }
      case integer_cst_K:
      case real_cst_K:
      case string_cst_K:
      case vector_cst_K:
      case complex_cst_K:
      {
         res = print_constant(index, vppf);
         break;
      }
      case var_decl_K:
      case result_decl_K:
      case parm_decl_K:
      case struct_field_tag_K:
      case symbol_memory_tag_K:
      case name_memory_tag_K:
      case memory_partition_tag_K:
      {
         res = (*vppf)(index);
         break;
      }
      case field_decl_K:
      {
         res = PrintVariable(index);
         break;
      }
      case label_decl_K:
      {
         label_decl * ld = GetPointer<label_decl>(node);
         if(ld->name)
         {
            identifier_node * id = GetPointer<identifier_node>(GET_NODE(ld->name));
            res = id->strg;
         }
         else
            res = "_unnamed_label_" + boost::lexical_cast<std::string>(index);
         break;
      }
      case gimple_label_K:
      {
         gimple_label *le = GetPointer<gimple_label>(node);
         if (!GetPointer<label_decl>(GET_NODE(le->op))->artificial_flag)
         {
            res += print_node(GET_INDEX_NODE(le->op), v, vppf);
            res += ":";
         }
         break;
      }
      case gimple_goto_K:
      {
         gimple_goto *ge = GetPointer<gimple_goto>(node);
         bool is_a_label = GetPointer<label_decl>(GET_NODE(ge->op)) != 0;
         res += (is_a_label ? "goto " : "goto *") + print_node(GET_INDEX_NODE(ge->op), v, vppf);
         break;
      }
      case constructor_K:
      {
         res += print_init(index, vppf);
         break;
      }
      case with_size_expr_K:
      {
         with_size_expr * wse = GetPointer<with_size_expr>(node);
         res += print_node(GET_INDEX_NODE(wse->op0), v, vppf);
         break;
      }
      case filter_expr_K:
      {
         res = "<filter_expr>()";
         break;
      }
      case exc_ptr_expr_K:
      {
         res = "<exc_ptr_expr>()";
         break;
      }
      case gimple_predict_K:
      case null_node_K:
      {
         break;
      }
      case pragma_node_K:
      {
         pragma_node * pn = GetPointer<pragma_node>(node);
         if(pn->directive and GetPointer<omp_for_pragma>(GET_NODE(pn->directive)))
         {
            break;
         }
         if (pn->is_block && !pn->is_opening)
         {
            res += "\n}";
         }
         else
         {
            res += "#pragma ";
            if (!pn->scope)
            {
               res += pn->line;
            }
            else
            {
               res += print_node(GET_INDEX_NODE(pn->scope), v, vppf);
               res += " ";
               res += print_node(GET_INDEX_NODE(pn->directive), v, vppf);
               if (pn->is_block)
                  res += "\n{";
            }
         }
         break;
      }
      case omp_pragma_K:
      {
         res += "omp";
         break;
      }
      case omp_for_pragma_K:
      {
         omp_for_pragma * fp = GetPointer<omp_for_pragma>(node);
         res += "for ";
         /// now print clauses
         for (std::tr1::unordered_map<std::string, std::string>::iterator i = fp->clauses.begin(); i != fp->clauses.end(); i++)
            res += " " + i->first + "(" + i->second + ")";
         break;
      }
      case omp_parallel_pragma_K:
      {
         omp_parallel_pragma * pn = GetPointer<omp_parallel_pragma>(node);
         if (!pn->is_shortcut)
         {
            res += "parallel";
         }
         /// now print clauses
         for (std::tr1::unordered_map<std::string, std::string>::iterator i = pn->clauses.begin(); i != pn->clauses.end(); i++)
            res += " " + i->first + "(" + i->second + ")";
         break;
      }
      case omp_sections_pragma_K:
      {
         omp_sections_pragma * pn = GetPointer<omp_sections_pragma>(node);
         if (!pn->is_shortcut)
         {
            res += "sections";
         }
         break;
      }
      case omp_parallel_sections_pragma_K:
      {
         res += "parallel sections";
         omp_parallel_sections_pragma * pn = GetPointer<omp_parallel_sections_pragma>(node);
         res += print_node(GET_INDEX_NODE(pn->op0), v, vppf);
         res += " ";
         res += print_node(GET_INDEX_NODE(pn->op1), v, vppf);
         break;
      }
      case omp_section_pragma_K:
      {
         res += "section";
         break;
      }
      case map_pragma_K:
      {
         res += STR_CST_pragma_keyword_map;
         break;
      }
      case call_hw_pragma_K:
      {
         res += STR_CST_pragma_keyword_call_hw" ";
         call_hw_pragma * ch = GetPointer<call_hw_pragma>(node);
         res += ch->HW_component;
         if (ch->ID_implementation.size())
            res += " " + boost::lexical_cast<std::string>(ch->ID_implementation);
         break;
      }
      case call_point_hw_pragma_K:
      {
         res += STR_CST_pragma_keyword_call_point_hw" ";
         call_point_hw_pragma * ch = GetPointer<call_point_hw_pragma>(node);
         res += ch->HW_component;
         if (ch->ID_implementation.size())
            res += " " + boost::lexical_cast<std::string>(ch->ID_implementation);
         if (ch->recursive)
         {
            res += " " + std::string(STR_CST_pragma_keyword_recursive);
         }
         break;
      }
      case issue_pragma_K:
      {
         res = "issue";
         break;
      }
      case profiling_pragma_K:
      {
         res = "profiling";
         break;
      }
      case blackbox_pragma_K:
      {
         res = "blackbox";
         break;
      }
      case statistical_profiling_K:
      {
         res = "profiling";
         break;
      }
      default:
         THROW_ERROR(std::string("tree node not currently supported ") + node->get_kind_text());
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Printed node " + boost::lexical_cast<std::string>(index));
   return res;
}

std::string behavioral_helper::print_type_declaration(unsigned int type) const
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Printing type declaration " + boost::lexical_cast<std::string>(type));
   std::string res;
   const tree_nodeRef node_type = TM->get_tree_node_const(type);
   switch (node_type->get_kind())
   {
      case record_type_K:
      {
         record_type *rt = GetPointer<record_type>(node_type);
         THROW_ASSERT(tree_helper::GetRealType(TM, type) == type, "Printing declaration of fake type " + boost::lexical_cast<std::string>(type));
         if(rt->unql)
         {
            res += "typedef ";
         }
         unsigned int qualifiers = rt->qual;
         if(qualifiers > 0)
            res += tree_helper::return_C_qualifiers(qualifiers, false);
         res += "struct ";
         if (not rt->unql)
         {
            if(rt->name)
            {
               res += tree_helper::print_type(TM, GET_INDEX_NODE(rt->name)) + " ";
            }
            else
            {
               res += "Internal_" + boost::lexical_cast<std::string>(type) + " ";
            }
         }
         if (not rt->unql or not GetPointer<record_type>(GET_NODE(rt->unql))->name)
         {
            /// Print the contents of the structure
            res += "\n{\n";
            null_deleter null_del;
            long long current_accumulated_size = 0;
            const var_pp_functorConstRef std_vpf(new std_var_pp_functor(behavioral_helperConstRef(this, null_del)));
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Printing content of the structure");
            for (unsigned int i = 0; i < rt->list_of_flds.size();i++)
            {
               unsigned int field = GET_INDEX_NODE(rt->list_of_flds[i]);
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Printing field " + boost::lexical_cast<std::string>(field));
               const field_decl * fd = GetPointer<field_decl>(TM->get_tree_node_const(field));
               const unsigned int field_type_index = get_type(field);
               const tree_nodeRef field_type = TM->get_tree_node_const(field_type_index);
               if (has_bit_field(field))
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Bit field");
                  const long long field_size = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(fd->size)));
                  const long long bpos = tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(fd->bpos)));
                  ///Only boolean and integer bitfied are allowed
                  if(field_size > 32)
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Size of field is " + boost::lexical_cast<std::string>(field_size));
                     THROW_ERROR_CODE(BITFIELD_EC, "Bitfield larger than int bitsize " + boost::lexical_cast<std::string>(field));
                  }
                  if(not GetPointer<boolean_type>(field_type) and (current_accumulated_size + field_size) > 32 and bpos % 32)
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Current accumulated size is " + boost::lexical_cast<std::string>(current_accumulated_size + field_size) + " - bpos is " + boost::lexical_cast<std::string>(bpos));
                     THROW_ERROR_CODE(BITFIELD_EC, "Bitfield memory alignment not supported");
                  }
                  if(not GetPointer<boolean_type>(field_type) and bpos % 32 and bpos % 32 != current_accumulated_size)
                  {
                     INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Bpos is " + boost::lexical_cast<std::string>(bpos) + " - Current accumlated size " + boost::lexical_cast<std::string>(current_accumulated_size));
                     THROW_ERROR_CODE(BITFIELD_EC, "Bitfield memory alignment not supported");
                  }
                  if(bpos % 32 == 0)
                     current_accumulated_size = 0;
                  current_accumulated_size += field_size;
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---New accumulated size is " + boost::lexical_cast<std::string>(current_accumulated_size));
                  if(tree_helper::is_unsigned(TM, field))
                  {
                     res += "unsigned ";
                  }
                  res += "int ";
                  res += (*std_vpf)(field);
                  res += " : ";
                  res += print_constant(GET_INDEX_NODE(fd->size));
               }
               else
               {
                  current_accumulated_size = 0;
                  res += tree_helper::print_type(TM, field_type_index, false, false, false, field, std_vpf);
               }
               res += ";\n";
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
            }
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
            res += '}';
            res += " ";
         }
         if (rt->unql and rt->name)
         {
            const record_type * rt_unqal = GetPointer<record_type>(GET_NODE(rt->unql));
            if(rt_unqal->name)
            {
               res += tree_helper::print_type(TM, GET_INDEX_NODE(rt_unqal->name)) + " ";
            }
            res += tree_helper::print_type(TM, GET_INDEX_NODE(rt->name));
         }
         if(rt->unql and rt->algn != GetPointer<record_type>(GET_NODE(rt->unql))->algn)
            res += " __attribute__ ((aligned (" + boost::lexical_cast<std::string>(rt->algn/8) + ")))";
         break;
      }
      case union_type_K:
      {
         union_type * ut = GetPointer<union_type>(node_type);
         THROW_ASSERT(tree_helper::GetRealType(TM, type) == type, "Printing declaration of fake type " + boost::lexical_cast<std::string>(node_type));
         if(ut->unql)
         {
            res += "typedef ";
         }
         unsigned int qualifiers = ut->qual;
         if(qualifiers > 0)
            res += tree_helper::return_C_qualifiers(qualifiers, false);
         res += "union ";
         if (not ut->unql)
         {
            if(ut->name)
            {
               res += tree_helper::print_type(TM, GET_INDEX_NODE(ut->name)) + " ";
            }
            else
            {
               res += "Internal_" + boost::lexical_cast<std::string>(type) + " ";
            }
         }
         if (not ut->unql or not GetPointer<union_type>(GET_NODE(ut->unql))->name)
         {
            /// Print the contents of the structure
            res += "\n{\n";
            null_deleter null_del;
            const var_pp_functorConstRef std_vpf(new std_var_pp_functor(behavioral_helperConstRef(this, null_del)));
            for (unsigned int i = 0; i < ut->list_of_flds.size();i++)
            {
               unsigned int field = GET_INDEX_NODE(ut->list_of_flds[i]);
               res += tree_helper::print_type(TM, get_type(field), false, false, false, field, std_vpf);
               res += ";\n";
            }
            res += '}';
            res += " ";
         }
         if (ut->unql and ut->name)
         {
            const union_type * ut_unqal = GetPointer<union_type>(GET_NODE(ut->unql));
            if(ut_unqal->name)
            {
               res += tree_helper::print_type(TM, GET_INDEX_NODE(ut_unqal->name)) + " ";
            }
            res += tree_helper::print_type(TM, GET_INDEX_NODE(ut->name));
         }
         if(ut->unql and ut->algn != GetPointer<union_type>(GET_NODE(ut->unql))->algn)
            res += " __attribute__ ((aligned (" + boost::lexical_cast<std::string>(ut->algn/8) + "))) ";
         break;
      }
      case enumeral_type_K:
      {
         const enumeral_type *et = GetPointer<enumeral_type>(node_type);
         if (et->unql)
         {
            res += "typedef ";
         }
         unsigned int quals = et->qual;
         if (quals > 0)
         {
            res += tree_helper::return_C_qualifiers(quals, false);
         }
         res += "enum ";
         if(not et->unql)
         {
            if(et->name)
            {
               res+=tree_helper::print_type(TM, GET_INDEX_NODE(et->name)) + " ";
            }
            else
            {
               res += "Internal_" + boost::lexical_cast<std::string>(type) + " ";
            }
         }
         if (not et->unql or not GetPointer<enumeral_type>(GET_NODE(et->unql))->name)
         {
            res += "{";
            tree_list * tl = GetPointer<tree_list>(GET_NODE(et->csts));
            while (tl)
            {
               res += tree_helper::print_type(TM, GET_INDEX_NODE(tl->purp));
               if (tl->valu)
               {
                  res += " = " + print_constant(GET_INDEX_NODE(tl->valu));
               }
               if (tl->chan)
               {
                  res += ",";
                  tl = GetPointer<tree_list>(GET_NODE(tl->chan));
               }
               else
                  tl = 0;
            }
            res += "}";
         }
         if(et->unql and et->name)
         {
            const enumeral_type * et_unql = GetPointer<enumeral_type>(GET_NODE(et->unql));
            if(et_unql->name)
            {
               res += tree_helper::print_type(TM, GET_INDEX_NODE(et_unql->name)) + " ";
            }
            res += tree_helper::print_type(TM, GET_INDEX_NODE(et->name));
         }
         break;
      }
      case array_type_K:
      {
         const array_type * at = GetPointer<array_type>(node_type);
         ///Compute the dimensions
         if (not at->size)
         {
            THROW_ERROR_CODE(C_EC, "Declaration of array type without size");
         }
         const tree_nodeRef array_length = GET_NODE(at->size);
         const tree_nodeRef array_t = GET_NODE (at->elts);
         if (array_length->get_kind() != integer_cst_K)
         {
            THROW_ERROR_CODE(C_EC, "Declaration of array type without fixed size");
         }
         const integer_cst *arr_ic = GetPointer<integer_cst>(array_length);
         const type_node * tn = GetPointer<type_node>(array_t);
         const integer_cst *eln_ic = GetPointer<integer_cst>(GET_NODE(tn->size));

         res += "typedef ";
         res += tree_helper::print_type(TM, GET_INDEX_NODE(at->elts));
         res += " ";
         THROW_ASSERT(at->name, "Trying to declare array without name " + boost::lexical_cast<std::string>(type));
         res +=  tree_helper::print_type(TM, GET_INDEX_NODE(at->name));
         res += "[";
         res += boost::lexical_cast<std::string>(tree_helper::get_integer_cst_value(arr_ic) / tree_helper::get_integer_cst_value(eln_ic));
         res += "]";
         break;
      }
      ///NOTE: this case cannot be moved because of break absence
      case pointer_type_K:
      {
         const pointer_type * pt = GetPointer<pointer_type>(node_type);
         if (pt->unql && GET_NODE(pt->ptd)->get_kind() == function_type_K)
         {
            function_type * ft = GetPointer<function_type>(GET_NODE(pt->ptd));
            unsigned int quals = GetPointer<type_node>(node_type)->qual;
            res += "typedef ";
            res += tree_helper::print_type(TM, GET_INDEX_NODE(ft->retn));
            res += " (* ";
            if (quals > 0)
               res += tree_helper::return_C_qualifiers(quals, false);
            if (GetPointer<type_node>(node_type)->name)
               res += tree_helper::print_type(TM, GET_INDEX_NODE(GetPointer<type_node>(node_type)->name));
            else
               res += "Internal_" + boost::lexical_cast<std::string>(type);
            res += ") (";
            if (ft->prms)
            {
               res += tree_helper::print_type(TM, GET_INDEX_NODE(ft->prms));
            }
            res += ")";
            return res;
         }
         //in this point break has not been forgotten. Pointer to not function type could be treated normally
      }
      default:
      {
         if (GetPointer<type_node>(node_type)->unql)
         {
            unsigned int quals = GetPointer<type_node>(node_type)->qual;
            res += "typedef ";

            if (quals > 0)
               res += tree_helper::return_C_qualifiers(quals, false);

            res += tree_helper::print_type(TM, GET_INDEX_NODE(GetPointer<type_node>(node_type)->unql));
            res += " ";
            /*            if(GetPointer<type_node>(node_type)->algn != 8)
             res += "__attribute__ ((aligned (" + boost::lexical_cast<std::string>(GetPointer<type_node>(node_type)->algn/8) + "))) ";*/
            if (GetPointer<type_node>(node_type)->name)
            {
               res += tree_helper::print_type(TM, GET_INDEX_NODE(GetPointer<type_node>(node_type)->name));
            }
            else
               res += "Internal_" + boost::lexical_cast<std::string>(type);
         }
         else
         {
            THROW_ASSERT(GetPointer<type_node>(node_type)->name, "Expected a typedef declaration with a name " + boost::lexical_cast<std::string>(type));
            tree_nodeRef name = GetPointer<type_node>(node_type)->name;
/* #ifndef NDEBUG
            identifier_node * id = GetPointer<identifier_node>(GET_NODE(name));
            THROW_ASSERT(id && id->strg == "bit_size_type", "Expected bit_size_type " + boost::lexical_cast<std::string>(type));
#endif */
            res += "typedef long long int bit_size_type";
         }
         break;
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Printed type declaration " + boost::lexical_cast<std::string>(type));
   return res;
}

bool behavioral_helper::is_bool(unsigned int index) const
{
   return tree_helper::is_bool(TM, get_type(index));
}

bool behavioral_helper::is_int(unsigned int index) const
{
   return tree_helper::is_int(TM, get_type(index));
}

bool behavioral_helper::is_unsigned(unsigned int index) const
{
   return tree_helper::is_unsigned(TM, get_type(index));
}

bool behavioral_helper::is_real(unsigned int index) const
{
   return tree_helper::is_real(TM, get_type(index));
}

bool behavioral_helper::is_a_struct(unsigned int variable) const
{
   return tree_helper::is_a_struct(TM, get_type(variable));
}

bool behavioral_helper::is_an_union(unsigned int variable) const
{
   return tree_helper::is_an_union(TM, get_type(variable));
}

bool behavioral_helper::is_an_array(unsigned int variable) const
{
   return tree_helper::is_an_array(TM, get_type(variable));
}

bool behavioral_helper::is_a_pointer(unsigned int variable) const
{
   return tree_helper::is_a_pointer(TM, variable);
}

bool behavioral_helper::is_an_indirect_ref(unsigned int variable) const
{
   const tree_nodeRef temp = TM->get_tree_node_const(variable);
   if (temp->get_kind() == indirect_ref_K)
      return true;
   else
      return false;
}

bool behavioral_helper::is_an_array_ref(unsigned int variable) const
{
   const tree_nodeRef temp = TM->get_tree_node_const(variable);
   if (temp->get_kind() == array_ref_K)
      return true;
   else
      return false;
}

bool behavioral_helper::is_a_component_ref(unsigned int variable) const
{
   const tree_nodeRef temp = TM->get_tree_node_const(variable);
   if (temp->get_kind() == component_ref_K)
      return true;
   else
      return false;
}

bool behavioral_helper::is_an_addr_expr(unsigned int variable) const
{
   const tree_nodeRef temp = TM->get_tree_node_const(variable);
   if (temp->get_kind() == addr_expr_K)
      return true;
   else
      return false;
}

bool behavioral_helper::is_a_mem_ref(unsigned int variable) const
{
   const tree_nodeRef temp = TM->get_tree_node_const(variable);
   if (temp->get_kind() == mem_ref_K)
      return true;
   else
      return false;
}

bool behavioral_helper::is_a_constant(unsigned int obj) const
{
   const tree_nodeRef node = TM->get_tree_node_const(obj);
   switch (node->get_kind())
   {
      case addr_expr_K:
         return true;
      case nop_expr_K:
      {
         unary_expr *ue = GetPointer<unary_expr>(node);
         return is_a_constant(GET_INDEX_NODE(ue->op));
      }
      case integer_cst_K:
      case real_cst_K:
      case string_cst_K:
      case vector_cst_K:
      case complex_cst_K:
      case case_label_expr_K:
      case label_decl_K:
         return true;
      default:
         return false;
   }
}

bool behavioral_helper::is_a_result_decl(unsigned int obj) const
{
   const tree_nodeRef node = TM->get_tree_node_const(obj);
   if (node->get_kind() == result_decl_K)
      return true;
   else
      return false;
}

bool behavioral_helper::is_a_call_expr_with_return(unsigned int obj)
{
   const tree_nodeRef node = TM->get_tree_node_const(obj);
   call_expr *ce = GetPointer<call_expr>(node);
   if (!ce)
      return false;
   tree_nodeRef op0 = GET_NODE(ce->fn);
   if (op0->get_kind() == addr_expr_K)
   {
      unary_expr *ue = GetPointer<unary_expr>(op0);
      tree_nodeRef fn = GET_NODE(ue->op);
      if (fn->get_kind() == function_decl_K)
      {
         if (tree_helper::has_function_return(TM, GET_INDEX_NODE(ue->op)))
            return true;
         else
            return false;
      }
      else
         return false;
   }
   else
      return false;
}

bool behavioral_helper::is_a_realpart_expr(unsigned int obj) const
{
   const tree_nodeRef temp = TM->get_tree_node_const(obj);
   if (temp->get_kind() == realpart_expr_K)
      return true;
   else
      return false;
}

bool behavioral_helper::is_a_imagpart_expr(unsigned int obj) const
{
   const tree_nodeRef temp = TM->get_tree_node_const(obj);
   if (temp->get_kind() == imagpart_expr_K)
      return true;
   else
      return false;
}

bool behavioral_helper::is_operating_system_function(const unsigned int obj) const
{
   const tree_nodeRef curr_tn = TM->get_tree_node_const(obj);
   const function_decl * fd = GetPointer<function_decl>(curr_tn);
   if(!fd)
      return false;
   return fd->operating_system_flag;
}

unsigned int behavioral_helper::get_indirect_ref_var(unsigned int obj) const
{
   THROW_ASSERT(is_an_indirect_ref(obj), "obj assumed to be an inderect_ref object");
   const tree_nodeRef temp = TM->get_tree_node_const(obj);
   indirect_ref * ir = GetPointer<indirect_ref>(temp);
   return GET_INDEX_NODE(ir->op);
}

unsigned int behavioral_helper::get_array_ref_array(unsigned int obj) const
{
   THROW_ASSERT(is_an_array_ref(obj), "obj assumed to be an array_ref object");
   const tree_nodeRef temp = TM->get_tree_node_const(obj);
   array_ref * ar = GetPointer<array_ref>(temp);
   if(GET_NODE(ar->op0)->get_kind() == mem_ref_K)
   {
      mem_ref * mr = GetPointer<mem_ref>(GET_NODE(ar->op0));
      THROW_ASSERT(tree_helper::get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(mr->op1)))==0, "unexpected pattern");
      return GET_INDEX_NODE(mr->op0);
   }
   else
      return GET_INDEX_NODE(ar->op0);
}

unsigned int behavioral_helper::get_array_ref_index(unsigned int obj) const
{
   THROW_ASSERT(is_an_array_ref(obj), "obj assumed to be an array_ref object");
   const tree_nodeRef temp = TM->get_tree_node_const(obj);
   array_ref * ar = GetPointer<array_ref>(temp);
   return GET_INDEX_NODE(ar->op1);
}

unsigned int behavioral_helper::get_component_ref_record(unsigned int obj) const
{
   THROW_ASSERT(is_a_component_ref(obj), "obj assumed to be a component_ref object");
   const tree_nodeRef temp = TM->get_tree_node_const(obj);
   component_ref * cr = GetPointer<component_ref>(temp);
   return GET_INDEX_NODE(cr->op0);
}

unsigned int behavioral_helper::get_component_ref_field(unsigned int obj) const
{
   THROW_ASSERT(is_a_component_ref(obj), "obj assumed to be a component_ref object");
   const tree_nodeRef temp = TM->get_tree_node_const(obj);
   component_ref * cr = GetPointer<component_ref>(temp);
   return GET_INDEX_NODE(cr->op1);
}

unsigned int behavioral_helper::get_mem_ref_base(unsigned int obj) const
{
   THROW_ASSERT(is_a_mem_ref(obj), "obj assumed to be a mem_ref object");
   const tree_nodeRef temp = TM->get_tree_node_const(obj);
   mem_ref * mr = GetPointer<mem_ref>(temp);
   return GET_INDEX_NODE(mr->op0);
}

unsigned int behavioral_helper::get_mem_ref_offset(unsigned int obj) const
{
   THROW_ASSERT(is_a_mem_ref(obj), "obj assumed to be a mem_ref object");
   const tree_nodeRef temp = TM->get_tree_node_const(obj);
   mem_ref * mr = GetPointer<mem_ref>(temp);
   return GET_INDEX_NODE(mr->op1);
}



unsigned int behavioral_helper::get_operand_from_unary_expr(unsigned int obj) const
{
   THROW_ASSERT(is_an_addr_expr(obj) || is_a_realpart_expr(obj) || is_a_imagpart_expr(obj), "obj assumed to be an addr_expr, a realpart_expr or an imagpart_expr object. obj is "
                + boost::lexical_cast<std::string>(obj));
   const tree_nodeRef temp = TM->get_tree_node_const(obj);
   unary_expr * ue = GetPointer<unary_expr>(temp);
   return GET_INDEX_NODE(ue->op);
}

unsigned int behavioral_helper::get_var_from_ssa(unsigned int index) const
{
   //THROW_ASSERT(TM->get_tree_node_const(index)->get_kind() == ssa_name_K, "obj assumed to be an ssa name " + boost::lexical_cast<std::string>(index));
   const tree_nodeRef temp = TM->get_tree_node_const(index);
   ssa_name * sn = GetPointer<ssa_name>(temp);
   if (sn)
      return GET_INDEX_NODE(sn->var);
   else
      return index;
}

unsigned int behavioral_helper::get_intermediate_var(unsigned int obj) const
{
   const tree_nodeRef node = TM->get_tree_node_const(obj);
   switch (node->get_kind())
   {
      case modify_expr_K:
      case init_expr_K:
      {
         binary_expr *be = GetPointer<binary_expr>(node);
         const tree_nodeRef right = GET_NODE(be->op1);
         ///check for type conversion
         switch (right->get_kind())
         {
               //case fix_trunc_expr_K:
            case fix_ceil_expr_K:
            case fix_floor_expr_K:
            case fix_round_expr_K:
            case float_expr_K:
            case convert_expr_K:
            case view_convert_expr_K:
            case nop_expr_K:
            case realpart_expr_K:
            case imagpart_expr_K:
            {
               unary_expr *ue = GetPointer<unary_expr>(right);
               return GET_INDEX_NODE(ue->op);
            }
            default:
               return GET_INDEX_NODE(be->op1);
         }
      }
      default:
         return 0;
   }
}

void behavioral_helper::get_parameters_type(std::tr1::unordered_set< unsigned int> & ret) const
{
   tree_nodeRef node = TM->get_tree_node_const(function_index);
   function_decl * fd = GetPointer<function_decl>(node);
   if (!fd)
      return;
   const std::vector<tree_nodeRef> & list_of_args = fd->list_of_args;
   std::vector<tree_nodeRef>::const_iterator it, it_end = list_of_args.end();
   for(it = list_of_args.begin(); it != it_end; it++)
   {
      const parm_decl * pd = GetPointer<parm_decl>(GET_NODE(*it));
      ret.insert(GET_INDEX_NODE(pd->type));
   }
   function_type * ft = GetPointer<function_type>(GET_NODE(fd->type));
   if (!ft || !ft->prms)
      return;
   tree_list * tl = GetPointer<tree_list>(GET_NODE(ft->prms));
   while (tl)
   {
      ret.insert(GET_INDEX_NODE(tl->valu));
      if (tl->chan)
         tl = GetPointer<tree_list>(GET_NODE(tl->chan));
      else
         break;
   }
}

unsigned int behavioral_helper::is_named_pointer(const unsigned int index) const
{
   THROW_ASSERT(index, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   unsigned int type = get_type(index);
   THROW_ASSERT(type, "this index does not exist: " + boost::lexical_cast<std::string>(type));
   const tree_nodeRef Type_node = TM->get_tree_node_const(type);
   THROW_ASSERT(Type_node, "this index does not exist: " + boost::lexical_cast<std::string>(type));
   if (Type_node->get_kind() == pointer_type_K)
   {
      pointer_type * pt = GetPointer<pointer_type>(Type_node);
      if (pt->name)
         return type;
      else
         return is_named_pointer(GET_INDEX_NODE(pt->ptd));
   }
   else
      return 0;
}

bool behavioral_helper::has_return() const
{
   return tree_helper::has_function_return(TM, function_index);
}

bool behavioral_helper::is_va_start_call(unsigned int stm) const
{
   if (is_var_args())
   {
      const tree_nodeRef node = TM->get_tree_node_const(stm);
      if (node->get_kind() == gimple_call_K)
      {
         gimple_call* ce = GetPointer<gimple_call>(node);
         tree_nodeRef cefn = GET_NODE(ce->fn);
         if (cefn->get_kind() == addr_expr_K)
         {
            unary_expr* ue = GetPointer<unary_expr>(cefn);
            function_decl * fd = GetPointer<function_decl>(GET_NODE(ue->op));
            if (fd && fd->get_function_name() == "__builtin_va_start")
            {
               return true;
            }
            else
               return false;
         }
         else
            return false;
      }
      else
         return false;
   }
   else
      return false;
}

bool behavioral_helper::has_bit_field(unsigned int variable) const
{
   const tree_nodeRef node = TM->get_tree_node_const(variable);
   if (node->get_kind() == field_decl_K)
   {
      field_decl * fd = GetPointer<field_decl>(node);
      if ((fd->list_attr.find(treeVocabularyTokenTypes::TOK_BITFIELD)) != fd->list_attr.end())
         return true;
      else
         return false;
   }
   else
      return false;
}

unsigned int behavioral_helper::get_attributes(unsigned int var) const
{
   const tree_nodeRef node = TM->get_tree_node_const(var);
   switch (node->get_kind())
   {
      case parm_decl_K:
      case ssa_name_K:
      {
         /*ssa_name * sn = GetPointer<ssa_name>(node);
          return get_attributes(GET_INDEX_NODE(sn->var));*/
         return 0;
      }
      case function_decl_K:
      case result_decl_K:
      case var_decl_K:
      {
         THROW_ASSERT(GetPointer<decl_node>(node), "get_attributes is only for decl_node: " + boost::lexical_cast<std::string>(var));
         return GetPointer<decl_node>(node)->attributes ? GET_INDEX_NODE(GetPointer<decl_node>(node)->attributes) : 0;
      }
      default:
         THROW_ERROR("Not supported: " + std::string(node->get_kind_text()));
   }
   return 0;
}

unsigned int behavioral_helper::GetInit(unsigned int var, std::tr1::unordered_set<unsigned int> & list_of_variables) const
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Get init of " + PrintVariable(var));
   if (initializations.find(var) != initializations.end())
   {
      unsigned int init = var;
      while (initializations.find(init) != initializations.end())
         init = initializations.find(init)->second;
      if(TM->get_tree_node_const(init)->get_kind() == var_decl_K)
      {
         list_of_variables.insert(init);
         const unsigned int init_of_init = GetInit(init, list_of_variables);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Init is " + boost::lexical_cast<std::string>(init_of_init));
         return init_of_init;
      }
      else
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Init is " + boost::lexical_cast<std::string>(init));
         return init;
      }
   }
   expandRef exp = expandRef(new expand());
   const tree_nodeRef node = TM->get_tree_node_const(var);
   switch (node->get_kind())
   {
      case ssa_name_K:
      {
         ssa_name * sn = GetPointer<ssa_name>(node);
         const unsigned ssa_init = GetInit(GET_INDEX_NODE(sn->var), list_of_variables);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Init is " + boost::lexical_cast<std::string>(ssa_init));
         return ssa_init;
      }
      case var_decl_K:
      {
         var_decl *vd = GetPointer<var_decl>(node);
         if (vd->init)
         {
            tree_helper::get_used_variables(true, exp, vd->init, list_of_variables);
            const unsigned var_init = GET_INDEX_NODE(vd->init);
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Init is " + boost::lexical_cast<std::string>(var_init));
            return var_init;
         }
         else
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Without init");
            return 0;
         }
      }
      /* Binary arithmetic and logic expressions.  */
      case mult_expr_K:
      case plus_expr_K:
      case minus_expr_K:
      case trunc_div_expr_K:
      case ceil_div_expr_K:
      case floor_div_expr_K:
      case round_div_expr_K:
      case trunc_mod_expr_K:
      case ceil_mod_expr_K:
      case floor_mod_expr_K:
      case round_mod_expr_K:
      case rdiv_expr_K:
      case exact_div_expr_K:
      case lshift_expr_K:
      case rshift_expr_K:
      case bit_ior_expr_K:
      case bit_xor_expr_K:
      case bit_and_expr_K:
      case truth_andif_expr_K:
      case truth_orif_expr_K:
      case truth_and_expr_K:
      case truth_or_expr_K:
      case truth_xor_expr_K:
         /* Binary relational expressions.  */
      case lt_expr_K:
      case le_expr_K:
      case gt_expr_K:
      case ge_expr_K:
      case eq_expr_K:
      case ne_expr_K:
         /* other binary expressions */
      case postdecrement_expr_K:
      case postincrement_expr_K:
      case rrotate_expr_K:
      case lrotate_expr_K:
         /* Unary arithmetic and logic expressions.  */
      case negate_expr_K:
      case bit_not_expr_K:
      case truth_not_expr_K:
      case reference_expr_K:
      case predecrement_expr_K:
      case preincrement_expr_K:
      case realpart_expr_K:
      case imagpart_expr_K:
         /* other unary expressions */
      case fix_trunc_expr_K:
      case fix_ceil_expr_K:
      case fix_floor_expr_K:
      case fix_round_expr_K:
      case float_expr_K:
      case convert_expr_K:
      case view_convert_expr_K:
      case nop_expr_K:
         /*other expressions */
      case call_expr_K:
      case gimple_call_K:
      case bit_field_ref_K:
      case min_expr_K:
      case max_expr_K:
      case unordered_expr_K:
      case ordered_expr_K:
      case unlt_expr_K:
      case unle_expr_K:
      case ungt_expr_K:
      case unge_expr_K:
      case uneq_expr_K:
      case ltgt_expr_K:
      case abs_expr_K:
      case complex_expr_K:
      case function_decl_K:
      case parm_decl_K:
      case integer_cst_K:
      case real_cst_K:
      case string_cst_K:
      case vector_cst_K:
      case complex_cst_K:
      case result_decl_K:
      case addr_expr_K:
      case array_ref_K:
      case component_ref_K:
      case indirect_ref_K:
      case struct_field_tag_K:
      case symbol_memory_tag_K:
      case memory_partition_tag_K:
      case field_decl_K:
      case label_decl_K:
      case cond_expr_K:
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Without init");
         return 0;
      }
      case constructor_K:
      {
         constructor * co = GetPointer<constructor>(TM->get_tree_node_const(var));
         std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::const_iterator vend = co->list_of_idx_valu.end();
         for (std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::const_iterator i = co->list_of_idx_valu.begin(); i != vend; i++)
         {
            tree_helper::get_used_variables(true, exp, i->second, list_of_variables);
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Init is " + boost::lexical_cast<std::string>(var));
         return var;
      }
      default:
         THROW_ASSERT(false, "get_init: Tree node not yet supported " + boost::lexical_cast<std::string>(var) + " " + node->get_kind_text());
   }
   return 0;
}

void behavioral_helper::create_label(blocRef &block, const unsigned int function_decl_nid)
{
   std::map<treeVocabularyTokenTypes::token_enum, std::string> IR_schema;
   ///check if the first statement is a gimple_label but with a unnamed label_decl.
   if (!block->list_of_stmt.empty())
   {
      tree_nodeRef first_stmt = block->list_of_stmt.front();
      gimple_label * le = GetPointer<gimple_label>(GET_NODE(first_stmt));
      if (le)
      {
         THROW_ASSERT(le->op && GET_NODE(le->op)->get_kind() == label_decl_K, "label expression pattern not yet supported");
         label_decl * ld = GetPointer<label_decl>(GET_NODE(le->op));
         THROW_ASSERT(!ld->name, "gimple_label already added");
         /// addd the name to the label_decl
         //@436     identifier_node  strg: "GOTOLABEL0"   lngt: 10
         unsigned int label_decl_name_nid_test;
         do
         {
            IR_schema[TOK(TOK_STRG)] = "GOTOLABEL" + boost::lexical_cast<std::string>(goto_label_unique_id++);
            label_decl_name_nid_test = TM->find(TOK(TOK_IDENTIFIER_NODE), IR_schema);
         }
         while (label_decl_name_nid_test);
         unsigned int label_decl_name_nid = TM->new_tree_node_id();//436
         TM->create_tree_node(label_decl_name_nid, TOK(TOK_IDENTIFIER_NODE), IR_schema);
         IR_schema.clear();
         tree_nodeRef tr_new_label_name = tree_nodeRef(new tree_reindex(label_decl_name_nid, TM->add(label_decl_name_nid)));
         ld->name = tr_new_label_name;
         ld->artificial_flag = false;
         return;
      }
   }

   //@27     identifier_node  strg: "void"   lngt: 4
   IR_schema[TOK(TOK_STRG)] = "void";
   unsigned int type_decl_name_nid = TM->find(TOK(TOK_IDENTIFIER_NODE), IR_schema);
   if (!type_decl_name_nid)
   {
      type_decl_name_nid = TM->new_tree_node_id();//27
      TM->create_tree_node(type_decl_name_nid, TOK(TOK_IDENTIFIER_NODE), IR_schema);
   }
   IR_schema.clear();

   //@436     identifier_node  strg: "GOTOLABEL0"   lngt: 10
   unsigned int label_decl_name_nid_test;
   do
   {
      IR_schema[TOK(TOK_STRG)] = "GOTOLABEL" + boost::lexical_cast<std::string>(goto_label_unique_id++);
      label_decl_name_nid_test = TM->find(TOK(TOK_IDENTIFIER_NODE), IR_schema);
   }
   while (label_decl_name_nid_test);
   unsigned int label_decl_name_nid = TM->new_tree_node_id();//436
   TM->create_tree_node(label_decl_name_nid, TOK(TOK_IDENTIFIER_NODE), IR_schema);
   IR_schema.clear();

   //@23     type_decl        name: @27      type: @15      srcp: "<built-in>":0
   IR_schema[TOK(TOK_NAME)] = boost::lexical_cast<std::string>(type_decl_name_nid);
   IR_schema[TOK(TOK_SRCP)] = "<built-in>:0:0";
   unsigned int label_void_type_name_nid = TM->find(TOK(TOK_TYPE_DECL), IR_schema);
   unsigned int label_type_nid;
   if (!label_void_type_name_nid)
   {
      label_void_type_name_nid = TM->new_tree_node_id();//23
      label_type_nid = TM->new_tree_node_id();//15
      IR_schema[TOK(TOK_TYPE)] = boost::lexical_cast<std::string>(label_type_nid);
      TM->create_tree_node(label_void_type_name_nid, TOK(TOK_TYPE_DECL), IR_schema);
      IR_schema.clear();
      //@15     void_type        name: @23      algn: 8
      IR_schema[TOK(TOK_NAME)] = boost::lexical_cast<std::string>(label_void_type_name_nid);
      IR_schema[TOK(TOK_ALGN)] = boost::lexical_cast<std::string>(8);
      TM->create_tree_node(label_type_nid, TOK(TOK_VOID_TYPE), IR_schema);
   }
   else
   {
      IR_schema.clear();
      //@15     void_type        name: @23      algn: 8
      IR_schema[TOK(TOK_NAME)] = boost::lexical_cast<std::string>(label_void_type_name_nid);
      IR_schema[TOK(TOK_ALGN)] = boost::lexical_cast<std::string>(8);
      label_type_nid = TM->find(TOK(TOK_VOID_TYPE), IR_schema);
      if (!label_type_nid)
      {
         label_type_nid = TM->new_tree_node_id();//15
         label_void_type_name_nid = TM->new_tree_node_id();//23
         IR_schema[TOK(TOK_NAME)] = boost::lexical_cast<std::string>(label_void_type_name_nid);
         TM->create_tree_node(label_type_nid, TOK(TOK_VOID_TYPE), IR_schema);
         IR_schema.clear();
         IR_schema[TOK(TOK_NAME)] = boost::lexical_cast<std::string>(type_decl_name_nid);
         IR_schema[TOK(TOK_TYPE)] = boost::lexical_cast<std::string>(label_type_nid);
         IR_schema[TOK(TOK_SRCP)] = "<built-in>:0:0";
         TM->create_tree_node(label_void_type_name_nid, TOK(TOK_TYPE_DECL), IR_schema);
      }
      else
      {
         THROW_ASSERT(label_void_type_name_nid == GET_INDEX_NODE(GetPointer<void_type>(TM->get_tree_node_const(label_type_nid))->name), "somenthing of wrong happen");
      }
   }
   IR_schema.clear();

   //@77     gimple_label       type: @15      srcp: "<built-in>":0     op  : @242
   unsigned int label_expr_nid = TM->new_tree_node_id();//77

   unsigned int label_decl_nid = TM->new_tree_node_id();//242
   IR_schema[TOK(TOK_OP)] = boost::lexical_cast<std::string>(label_decl_nid);
   IR_schema[TOK(TOK_SRCP)] = "<built-in>:0:0";
   TM->create_tree_node(label_expr_nid, TOK(TOK_GIMPLE_LABEL), IR_schema);
   IR_schema.clear();

   //@242    label_decl       name: @436     type: @15      scpe: @32     srcp: "<built-in>":0                UID : 5
   IR_schema[TOK(TOK_NAME)] = boost::lexical_cast<std::string>(label_decl_name_nid);
   IR_schema[TOK(TOK_TYPE)] = boost::lexical_cast<std::string>(label_type_nid);
   IR_schema[TOK(TOK_SCPE)] = boost::lexical_cast<std::string>(function_decl_nid);
   IR_schema[TOK(TOK_SRCP)] = "<built-in>:0:0";
   TM->create_tree_node(label_decl_nid, TOK(TOK_LABEL_DECL), IR_schema);
   IR_schema.clear();

   tree_nodeRef tr_new_stmt = tree_nodeRef(new tree_reindex(label_expr_nid, TM->add(label_expr_nid)));
   block->add_stmt_in_front(tr_new_stmt);
#if 0
   static int nid = 0;
   std::string raw_file_name = boost::lexical_cast<std::string>(nid++) + "create_label.raw";
   std::ofstream raw_file(raw_file_name.c_str());
   raw_file << TM;
   raw_file.close();
#endif
}

void behavioral_helper::create_goto(blocRef &block, const unsigned int, const unsigned int label_expr_nid)
{
   TM->add_goto();
   ///compute the label_decl_nid
   gimple_label * le = GetPointer<gimple_label>(TM->get_tree_node_const(label_expr_nid));
   THROW_ASSERT(le, "expected a label expression");
   unsigned int label_decl_nid = GET_INDEX_NODE(le->op);
   ///create the gimple_goto tree node
   std::map<treeVocabularyTokenTypes::token_enum, std::string> IR_schema;

   //@27     identifier_node  strg: "void"   lngt: 4
   IR_schema[TOK(TOK_STRG)] = "void";
   unsigned int type_decl_name_nid = TM->find(TOK(TOK_IDENTIFIER_NODE), IR_schema);
   if (!type_decl_name_nid)
   {
      type_decl_name_nid = TM->new_tree_node_id();//27
      TM->create_tree_node(type_decl_name_nid, TOK(TOK_IDENTIFIER_NODE), IR_schema);
   }
   IR_schema.clear();

   //@23     type_decl        name: @27      type: @15      srcp: "<built-in>":0
   IR_schema[TOK(TOK_NAME)] = boost::lexical_cast<std::string>(type_decl_name_nid);
   IR_schema[TOK(TOK_SRCP)] = "<built-in>:0:0";
   unsigned int label_void_type_name_nid = TM->find(TOK(TOK_TYPE_DECL), IR_schema);
   unsigned int label_type_nid;
   if (!label_void_type_name_nid)
   {
      label_void_type_name_nid = TM->new_tree_node_id();//23
      label_type_nid = TM->new_tree_node_id();//15
      IR_schema[TOK(TOK_TYPE)] = boost::lexical_cast<std::string>(label_type_nid);
      TM->create_tree_node(label_void_type_name_nid, TOK(TOK_TYPE_DECL), IR_schema);
      IR_schema.clear();
      //@15     void_type        name: @23      algn: 8
      IR_schema[TOK(TOK_NAME)] = boost::lexical_cast<std::string>(label_void_type_name_nid);
      IR_schema[TOK(TOK_ALGN)] = boost::lexical_cast<std::string>(8);
      TM->create_tree_node(label_type_nid, TOK(TOK_VOID_TYPE), IR_schema);
   }
   else
   {
      IR_schema.clear();
      //@15     void_type        name: @23      algn: 8
      IR_schema[TOK(TOK_NAME)] = boost::lexical_cast<std::string>(label_void_type_name_nid);
      IR_schema[TOK(TOK_ALGN)] = boost::lexical_cast<std::string>(8);
      label_type_nid = TM->find(TOK(TOK_VOID_TYPE), IR_schema);
      if (!label_type_nid)
      {
         label_type_nid = TM->new_tree_node_id();//15
         label_void_type_name_nid = TM->new_tree_node_id();//23
         IR_schema[TOK(TOK_NAME)] = boost::lexical_cast<std::string>(label_void_type_name_nid);
         TM->create_tree_node(label_type_nid, TOK(TOK_VOID_TYPE), IR_schema);
         IR_schema.clear();
         IR_schema[TOK(TOK_NAME)] = boost::lexical_cast<std::string>(type_decl_name_nid);
         IR_schema[TOK(TOK_TYPE)] = boost::lexical_cast<std::string>(label_type_nid);
         IR_schema[TOK(TOK_SRCP)] = "<built-in>:0:0";
         TM->create_tree_node(label_void_type_name_nid, TOK(TOK_TYPE_DECL), IR_schema);
      }
      else
      {
         THROW_ASSERT(label_void_type_name_nid == GET_INDEX_NODE(GetPointer<void_type>(TM->get_tree_node_const(label_type_nid))->name), "somenthing of wrong happen");
      }
   }
   IR_schema.clear();

   //@60     gimple_goto        type: @15      srcp: "<built-in>":6      op  : @242
   unsigned int goto_expr_nid = TM->new_tree_node_id();//60
   IR_schema[TOK(TOK_TYPE)] = boost::lexical_cast<std::string>(label_type_nid);
   IR_schema[TOK(TOK_OP)] = boost::lexical_cast<std::string>(label_decl_nid);
   IR_schema[TOK(TOK_SRCP)] = "<built-in>:0:0";
   TM->create_tree_node(goto_expr_nid, TOK(TOK_GIMPLE_GOTO), IR_schema);
   IR_schema.clear();

   tree_nodeRef tr_new_stmt = tree_nodeRef(new tree_reindex(goto_expr_nid, TM->add(goto_expr_nid)));
   block->add_stmt(tr_new_stmt);
}

// void behavioral_helper::add_variable(unsigned int var, const std::string _name, bool written)
// {
//    behavioral_helper::add_variable(var, _name, written);
//    unsigned int type = get_type(var);
//    if(is_a_struct(var))
//    {
//       tree_nodeRef node_type = TM->get_tree_node_const(type);
//       THROW_ASSERT(node_type->get_kind() == record_type_K, "Type of struct variable " + boost::lexical_cast<std::string>(var) + " is not a record type");
//       record_type * rt = GetPointer<record_type>(node_type);
//       if(rt->unql)
//       {
//          definitions[GET_INDEX_NODE(rt->unql)] = get_definition(type);
//       }
//    }
//    if(is_an_union(var))
//    {
//       tree_nodeRef node_type = TM->get_tree_node_const(type);
//       THROW_ASSERT(node_type->get_kind() == union_type_K, "Type of union variable " + boost::lexical_cast<std::string>(var) + " is not a union type");
//       union_type * ut = GetPointer<union_type>(node_type);
//       if(ut->unql)
//       {
//          definitions[GET_INDEX_NODE(ut->unql)] = get_definition(type);
//       }
//    }
// }

std::string behavioral_helper::print_phinode_res(unsigned int phi_node_id, vertex v, const var_pp_functorConstRef vppf) const
{
   const tree_nodeRef node = TM->get_tree_node_const(phi_node_id);
   gimple_phi *phi = GetPointer<gimple_phi>(node);

   THROW_ASSERT(phi, "NodeId is not related to a gimple_phi");

   return print_node(GET_INDEX_NODE(phi->res), v, vppf);
}

unsigned int behavioral_helper::start_with_a_label(const blocRef &block) const
{
   if (block->list_of_stmt.empty())
      return 0;
   tree_nodeRef first_stmt = block->list_of_stmt.front();
   gimple_label * le = GetPointer<gimple_label>(GET_NODE(first_stmt));
   if (le && le->op && GET_NODE(le->op)->get_kind() == label_decl_K)
   {
      label_decl * ld = GetPointer<label_decl>(GET_NODE(le->op));
      if (ld->name)
         return GET_INDEX_NODE(first_stmt);
      else
         return 0;
   }
   else
      return 0;
}

const std::string behavioral_helper::get_label_name(unsigned int label_expr_nid) const
{
   tree_nodeRef tn = TM->get_tree_node_const(label_expr_nid);
   gimple_label * le = GetPointer<gimple_label>(tn);
   THROW_ASSERT(le->op && GET_NODE(le->op)->get_kind() == label_decl_K, "label decl expected");
   label_decl * ld = GetPointer<label_decl>(GET_NODE(le->op));
   THROW_ASSERT(ld->name && GET_NODE(ld->name)->get_kind() == identifier_node_K, "identifier_node expected");
   identifier_node * id = GetPointer<identifier_node>(GET_NODE(ld->name));
   return id->strg;
}

unsigned int behavioral_helper::end_with_a_cond_or_goto(const blocRef &block) const
{
   if (block->list_of_stmt.empty())
      return 0;
   tree_nodeRef last = block->list_of_stmt.back();
   gimple_cond * gc = GetPointer<gimple_cond>(GET_NODE(last));
   if (gc)
      return GET_INDEX_NODE(last);
   gimple_switch * se = GetPointer<gimple_switch>(GET_NODE(last));
   if (se)
      return GET_INDEX_NODE(last);
   gimple_goto * ge = GetPointer<gimple_goto>(GET_NODE(last));
   if (ge)
      return GET_INDEX_NODE(last);
   return 0;
}

bool behavioral_helper::is_to_declare(unsigned int obj) const
{
   if (std::find(parameters.begin(), parameters.end(), obj) != parameters.end())
      return false;
   const tree_nodeRef node = TM->get_tree_node_const(obj);
   if (node->get_kind() == parm_decl_K)
      return false;
   ssa_name* sa = GetPointer<ssa_name>(node);
   if (sa and (sa->volatile_flag || GET_NODE(sa->def_stmts.front())->get_kind() == gimple_nop_K)
         and      GET_NODE(sa->var)->get_kind() == parm_decl_K)
      return false;
   return true;
}

void behavioral_helper::create_gimple_modify_stmt(unsigned int, blocRef &, tree_nodeRef, tree_nodeRef )
{
   THROW_ERROR("Not implemented");
}

std::string behavioral_helper::print_forward_declaration(unsigned int type) const
{
   std::string res;
   const tree_nodeRef node_type = TM->get_tree_node_const(type);
   switch (node_type->get_kind())
   {
      case record_type_K:
      {
         record_type *rt = GetPointer<record_type>(node_type);
         if (rt->name)
            res += "struct " + tree_helper::print_type(TM, GET_INDEX_NODE(rt->name));
         else
            res += "struct Internal_" + boost::lexical_cast<std::string>(type);
      }
      break;
      case union_type_K:
      {
         union_type *rt = GetPointer<union_type>(node_type);
         if (rt->name)
            res += "union " + tree_helper::print_type(TM, GET_INDEX_NODE(rt->name));
         else
            res += "union Internal_" + boost::lexical_cast<std::string>(type);
      }
      break;
      case enumeral_type_K:
      {
         enumeral_type *rt = GetPointer<enumeral_type>(node_type);
         if (rt->name)
            res += "enum " + tree_helper::print_type(TM, GET_INDEX_NODE(rt->name));
         else
            res += "enum Internal_" + boost::lexical_cast<std::string>(type);
      }
      break;
      default:
      {
         THROW_ERROR("Not yet supported " + std::string(node_type->get_kind_text()));
      }
   }
   return res;
}

bool behavioral_helper::is_empty_return(unsigned int index) const
{
   const tree_nodeRef retNode = TM->get_tree_node_const(index);
   gimple_return *re = GetPointer<gimple_return>(retNode);
   THROW_ASSERT(re, "Expected a return statement");
   return !re->op;
}

void behavioral_helper::add_phi_init(
   std::map<unsigned int, std::list<std::pair<unsigned int, unsigned int> > >&phi_init, unsigned int phi_index) const
{
   tree_nodeRef tn = TM->get_tree_node_const(phi_index);
   THROW_ASSERT(tn->get_kind() == gimple_phi_K, "Node " + boost::lexical_cast<std::string>(phi_index) +
                " is not a phi-node");
   gimple_phi* phi = GetPointer<gimple_phi>(tn);
   unsigned int phi_res = GET_INDEX_NODE(phi->res);
   std::vector<std::pair< tree_nodeRef, unsigned int> >::const_iterator vend = phi->list_of_def_edge.end();
   for (std::vector<std::pair< tree_nodeRef, unsigned int> >::const_iterator i = phi->list_of_def_edge.begin(); i != vend; i++)
      if (phi_res != GET_INDEX_NODE(i->first))
         phi_init[i->second].push_back(std::pair<unsigned int, unsigned int>(phi_res, GET_INDEX_NODE(i->first)));
}

bool behavioral_helper::isSystemVariable(unsigned int varId, const std::vector<std::string>& systemIncPath) const
{
   tree_nodeRef tn = TM->get_tree_node_const(varId);
   std::string include = GetPointer<srcp>(tn)->include_name;
   for (unsigned int i = 0; i < systemIncPath.size(); i++)
   {
      if (include.find(systemIncPath[i]) == 0)
         return true;
   }
   return false;
}

const std::tr1::unordered_set<unsigned int> behavioral_helper::GetTypesToBeDeclared(const unsigned int index) const
{
   return tree_helper::GetTypesToBeDeclared(TM, index);
}

unsigned int behavioral_helper::GetUnqualified(const unsigned int index) const
{
   return tree_helper::GetUnqualified(TM, index);
}

std::string behavioral_helper::print_type(unsigned int type, bool global, bool print_qualifiers, bool print_storage, unsigned int var, const var_pp_functorConstRef vppf, const std::string prefix, const std::string tail) const
{
   return tree_helper::print_type(TM, type, global, print_qualifiers, print_storage, var, vppf, prefix, tail);
}

void behavioral_helper::rename_a_variable(unsigned int var, const std::string & new_name)
{
   vars_renaming_table[var] = new_name;
}

void behavioral_helper::clear_renaming_table()
{
   vars_renaming_table.clear();
}
