/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file prettyPrintVertex.cpp
 * @brief Helper class supporting the printing of vertexes of a graph.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * @version $Revision: 10608 $
 * @date $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *

*/
///Header include
#include "prettyPrintVertex.hpp"

///Behavior include
#include "behavioral_helper.hpp"
#include "op_node_info.hpp"

///STD include
#include <algorithm>

///utility include
#include "var_pp_functor.hpp"

void prettyPrintVertex::print_type_declaration(std::ostream &os, unsigned int type, simple_indent &PP, const behavioral_helperConstRef BH)
{
   PP(os, BH->print_type_declaration(type)+";\n");
}

void prettyPrintVertex::print_prototype_declaration(std::ostream &os, unsigned int function, simple_indent &PP, const var_pp_functorConstRef vppf, const behavioral_helperConstRef BH)
{
   PP(os, BH->print_type(function, false, true, false, 0, vppf));
}

void prettyPrintVertex::print_var_name(std::ostream &os, unsigned int var, const var_pp_functorConstRef vppf, simple_indent &PP)
{
   PP(os, (*vppf)(var));//the second parameter is null since there is no statement associated with a var.
}


void prettyPrintVertex::get_written_vars(const vertex &v, const graph * g, std::tr1::unordered_set<unsigned int> &list_of_variables, const behavioral_helperConstRef)
{
   const std::tr1::unordered_set<unsigned int> &tempWritten = GET_SSA_WRITTEN(g, v);

   const std::tr1::unordered_set<unsigned int>::const_iterator itw_end = tempWritten.end();
   for(std::tr1::unordered_set<unsigned int>::const_iterator itw = tempWritten.begin(); itw != itw_end; itw++)
   {
      //unsigned int var = BH->get_var_from_ssa(*itw);
      list_of_variables.insert(*itw);
   }
}


void prettyPrintVertex::get_read_vars(const vertex &v, const graph * g, std::tr1::unordered_set<unsigned int> &list_of_variables, const behavioral_helperConstRef )
{
   const std::list<unsigned int> &tempRead = GET_SSA_READ(g, v);

   const std::list<unsigned int>::const_iterator itr_end = tempRead.end();
   for(std::list<unsigned int>::const_iterator itr = tempRead.begin(); itr != itr_end; itr++)
   {
      //unsigned int var = BH->get_var_from_ssa(*itr);
      list_of_variables.insert(*itr);
   }
}

void prettyPrintVertex::get_internal_vars(const vertex &v, const graph * g, std::tr1::unordered_set<unsigned int> &list_of_variables, const behavioral_helperConstRef BH)
{
   std::list<unsigned int>::const_iterator it_end = GET_NODEID(g,v).end();
   for(std::list<unsigned int>::const_iterator it = GET_NODEID(g,v).begin(); it != it_end; it++)
   {
      unsigned intermediate_var = BH->get_intermediate_var(*it);
      if(intermediate_var && !BH->is_an_indirect_ref(intermediate_var) && !BH->is_a_constant(intermediate_var) &&
            !BH->is_a_component_ref(intermediate_var) && !BH->is_an_array_ref(intermediate_var) && !BH->is_an_addr_expr(intermediate_var))
      {
         list_of_variables.insert(intermediate_var);
      }
      else if(intermediate_var && BH->is_an_addr_expr(intermediate_var))
      {
         unsigned int var = BH->get_operand_from_unary_expr(intermediate_var);
         while(1)
         {
            if (BH->is_a_component_ref(var))
            {
               var = BH->get_component_ref_record(var);
               continue;
            }
            if (BH->is_an_array_ref(var))
            {
               var = BH->get_array_ref_array(var);
               continue;
            }
            if (BH->is_an_indirect_ref(var))
            {
               var = BH->get_indirect_ref_var(var);
               continue;
            }
            break;
         }
         if(!BH->is_a_constant(var))
            list_of_variables.insert(var);
      }
   }
}


void prettyPrintVertex::get_referenced_vars(const vertex &v, const graph * g, std::tr1::unordered_set<unsigned int> &list_of_variables, const behavioral_helperConstRef )
{
   list_of_variables = GET_REF_VAR(g, v);
   /*
   OutEdgeIterator oi,oend;
   for(tie(oi, oend) = boost::out_edges(v, *g); oi != oend; oi++)
   {
      vertex target = boost::target(*oi, *g);
      if (GET_TYPE(g, target) == TYPE_PHI)
      {
         std::set<unsigned int> written = GET_SSA_WRITTEN(g, target);
         list_of_variables.insert(written.begin(), written.end());
      }
   }*/
}

void prettyPrintVertex::print_forward_declaration(std::ostream &os, unsigned int type, simple_indent &PP, const behavioral_helperConstRef BH)
{
   PP(os, BH->print_forward_declaration(type)+";\n");
}
