/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file tree_basic_block.hpp
 * @brief Data structure describing a basic block at tree level.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#ifndef TREE_BASIC_BLOCK_HPP
#define TREE_BASIC_BLOCK_HPP

///STD include
#include <iostream>

///STL include
#include <list>
#include <map>
#include <vector>
#include <set>

///tree include
#include "tree_node.hpp"

///Utility include
#include "refcount.hpp"

/**
 * @name forward declarations
 */
//@{
REF_FORWARD_DECL(bloc);
//@}


/**
 * constant identifying the basic block node of type entry
 */
#define BB_ENTRY 0

/**
 * constant identifying the basic block node of type exit
*/
#define BB_EXIT 1

/**
 * This struct specifies the field bloc (basic block).
  * The tree walker structure of this node is:
  * #(TOK_BLOC TOK_NUMBER (pred)+ (succ)+ (phi)* (stmt)*)
*/
struct bloc
{
   /// list_of_pred is the list of predecessors.
   std::vector<unsigned int> list_of_pred;

   /// list_of_succ is the list of successors.
   std::vector<unsigned int> list_of_succ;

   /// number is the index of the basic block.
   unsigned int number;

   /// list_of_phi is a list of eventual phi node presents in the basic block.
   std::vector<tree_nodeRef> list_of_phi;

   /// Number of removed phi
   unsigned int removed_phi;

   /// list_of_stmt is the list of statements stored in the basic block.
   std::vector<tree_nodeRef> list_of_stmt;

   /// Live_In of the basic block
   std::set<unsigned int> live_in;

   /// Live_Out of the basic block
   std::set<unsigned int> live_out;

   /// store the relation between the phi defs inserted in the live_out set and phi uses
   //std::map<unsigned int, unsigned int> live_out_phi_defs;

   /// set of ssa uses
   std::set<tree_nodeRef> ssa_uses;

   /// flag for header of reducible parallelizable loop
   unsigned int hpl;

   /// in case the last statement is a gimple_cond associated with an if statement this member contains the basic block reference to the then statements.
   unsigned int true_edge;

   /// in case the last statement is a gimple_cond associated with an if statement this member contains the basic block reference to the else statements.
   unsigned int false_edge;

   /**
    * constructor
    */
   bloc();

   /**
    * Destructor
    */
   virtual ~bloc();

   /**
   * Add a value to list of pred.
   * @param a is a pred.
   */
   void add_pred(const unsigned int a)
   {
      list_of_pred.push_back(a);
   }

   /**
   * Add a value to list of succ.
   * @param a is a succ.
   */
   void add_succ(const unsigned int a)
   {
      list_of_succ.push_back(a);
   }
   /**
   * Add a value to list of phi node.
   * @param a is a NODE_ID.
   */
   void add_phi(const tree_nodeRef a)
   {
      list_of_phi.push_back(a);
   }

   /**
   * Add a value to list of stmt.
   * @param a is a NODE_ID.
   */
   void add_stmt(const tree_nodeRef a)
   {
      list_of_stmt.push_back(a);
   }

   /**
    * Add a statement in front
    * @param a is a NODE_ID.
   */
   void add_stmt_in_front(const tree_nodeRef a)
   {
      list_of_stmt.insert(list_of_stmt.begin(), a);
   }

   /**
    * Add a statement op after statement pos
    * @param op is the statement to be added
    * @param pos is the statement after which operation has to be added;
    */
   void add_stmt(const tree_nodeRef op, tree_nodeRef pos);

   /// constant indentify the entry basic block
   static const unsigned int ENTRY_BLOCK_ID;

   /// constant indentify the exit basic block
   static const unsigned int EXIT_BLOCK_ID;

   /**
    * virtual function used to traverse the tree_node data structure.
    * @param v is a reference to the tree_node visitor class
    */
   virtual void visit(tree_node_visitor * const v) const;

   ///visitor enum
   enum {GETID(list_of_phi)=0, GETID(list_of_stmt)};
};

#endif
