/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file tree_helper.cpp
 * @brief This file collects some utility functions.
 *
 *
 * @author Katia Turati <turati@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Header include
#include "tree_helper.hpp"

///Tree include
#include "ext_tree_node.hpp"
#include "token_interface.hpp"
#include "tree_basic_block.hpp"
#include "tree_manager.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"

#include "var_pp_functor.hpp"

///Utility include
#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/lexical_cast.hpp>
#include "exceptions.hpp"
#include <iostream>

/**
 * Used to test if a type name has a substring in a set of string
 */
struct find_SC_string
{
   //predicate function
   bool operator()(const std::string& el)
   {
      return target == el;
   }
   find_SC_string(const std::string& str) : target(str) {}
   private:
      const std::string& target;
};

///built in type without parameter
///"sc_in_resolved" is not a template and inherits from sc_in<sc_dt::sc_logic>
///"sc_inout_resolved" is not a template and inherits from sc_inout<sc_dt::sc_logic>
///"sc_out_resolved" is not a template and inherits from sc_inout_resolved
///"sc_in_clk" is not a template and is a typedef of sc_in<bool>
///"sc_out_clk" is not a template and is a typedef of sc_out<bool>
///"sc_inout_clk" is not a template and is a typedef of sc_inout<bool>

///builtin types with parameter
///"sc_lv" is a template, has a parameter (W)
///"sc_bv" is a template, has a parameter (W)
///"sc_in_rv" is a template, has a parameter (W) and inherits from sc_in<sc_dt::sc_lv<W> >
///"sc_out_rv" is a template, has a parameter (W) and inherits from sc_inout_rv<W>
///"sc_inout_rv" is a template, has a parameter (W) and inherits from sc_inout<sc_dt::sc_lv<W> >
///"sc_signal_rv" is a template, has a parameter (W) and inherits from sc_signal<sc_dt::sc_lv<W> >

///"sc_fifo_in" is a template, has a parameter (T) and inherits from sc_port<sc_fifo_in_if<T>,0>
///"sc_fifo_out" is a template, has a parameter (T) and inherits from sc_port<sc_fifo_out_if<T>,0>

static
      const std::string SC_tmpl_class_[] =
{"sc_signal", "sc_in",
    "sc_out",
    "sc_inout",
    "sc_port", "sc_export"
};

static
      const size_t size_SC_tmpl_class_ = sizeof SC_tmpl_class_ / sizeof SC_tmpl_class_[0];

static
      const std::string SC_builtin_scalar_type_[] =
{"sc_int", "sc_uint", "sc_biguint", "sc_bigint",
    "sc_lv", "sc_in_rv", "sc_out_rv", "sc_inout_rv",
    "sc_bv",
    "sc_signal_rv"
};/*check "sc_fix","sc_ufix"  "sc_fifo", "sc_buffer", "sc_fixed", "sc_ufixed" */

static
      const std::string TLM_SC_builtin_scalar_type_[] =
{"tlm_blocking_put_if", "tlm_blocking_get_if", "tlm_nonblocking_get_if", "tlm_nonblocking_put_if", "tlm_transport_if" //TLM objects
};

static
      const size_t size_SC_builtin_scalar_type_ = sizeof SC_builtin_scalar_type_ / sizeof SC_builtin_scalar_type_[0];

static
      const size_t TLM_size_SC_builtin_scalar_type_ = sizeof TLM_SC_builtin_scalar_type_ / sizeof TLM_SC_builtin_scalar_type_[0];

static
      std::string vec_tmpl_logic[] = {"sc_signal_rv", "sc_lv"};

static
      const size_t size_tmpl_logic = sizeof vec_tmpl_logic / sizeof vec_tmpl_logic[0];

int tree_helper::debug_level = 0;

std::set<std::string> tree_helper::SC_tmpl_class = std::set<std::string> (SC_tmpl_class_, SC_tmpl_class_ + size_SC_tmpl_class_);
std::set<std::string> tree_helper::SC_builtin_scalar_type = std::set<std::string> (SC_builtin_scalar_type_, SC_builtin_scalar_type_ + size_SC_builtin_scalar_type_);
std::set<std::string> tree_helper::TLM_SC_builtin_scalar_type = std::set<std::string> (TLM_SC_builtin_scalar_type_, TLM_SC_builtin_scalar_type_ + TLM_size_SC_builtin_scalar_type_);

tree_helper::tree_helper()
{}


tree_helper::~tree_helper()
{}

unsigned int tree_helper::size(const tree_managerConstRef tm, unsigned int index)
{
   const tree_nodeRef t = tm->get_tree_node_const(index);

   switch (t->get_kind())
   {
      case var_decl_K:
      {
         THROW_ASSERT(GetPointer<var_decl>(t)->type, "expected a var decl type");
         return size(tm, GET_INDEX_NODE(GetPointer<var_decl>(t)->type));
      }
      case parm_decl_K:
      {
         THROW_ASSERT(GetPointer<parm_decl>(t)->type, "expected a var decl type");
         return size(tm, GET_INDEX_NODE(GetPointer<parm_decl>(t)->type));
      }
      case function_decl_K:
      {
         return 0;
      }
      case result_decl_K:
      {
         THROW_ASSERT(GetPointer<result_decl>(t)->type, "expected a var decl type");
         return size(tm, GET_INDEX_NODE(GetPointer<result_decl>(t)->type));
      }
      case ssa_name_K:
      {
         THROW_ASSERT(GetPointer<ssa_name>(t), "Expected an ssa_name");
         return size(tm, GET_INDEX_NODE(GetPointer<ssa_name>(t)->var));
      }
      case field_decl_K:
      {
         THROW_ASSERT(GetPointer<field_decl>(t)->type, "expected a field decl type");
         return size(tm, GET_INDEX_NODE(GetPointer<field_decl>(t)->type));
      }
      case pointer_type_K:
      {
         return size(tm, GET_INDEX_NODE(GetPointer<pointer_type>(t)->size));
      }
      case reference_type_K:
      {
         THROW_ASSERT(GetPointer<reference_type>(t)->refd, "expected a reference declaration");
         return size(tm, GET_INDEX_NODE((GetPointer<reference_type>(t))->refd));
      }
      case record_type_K:
      {
         THROW_ASSERT(!is_module(tm, index) && !is_channel(tm, index), "The size of a module/channel is not supported by this function.");
         record_type* rect = GetPointer<record_type>(t);

         integer_cst * ic = GetPointer<integer_cst>(GET_NODE(rect->size));
         return static_cast<unsigned int>(get_integer_cst_value(ic));
#if 0
         std::string nt;
         if (rect->name && GET_NODE(rect->name)->get_kind() == type_decl_K)
         {
            type_decl* td = GetPointer<type_decl>(GET_NODE(rect->name));
            if (GET_NODE(td->name)->get_kind() == identifier_node_K)
            {
               identifier_node* idn = GetPointer<identifier_node>(GET_NODE(td->name));
               nt = idn->strg;
            }
            else
               THROW_ERROR("unexpected record type pattern of node " + boost::lexical_cast<std::string>(index));
         }
         else
         {
            integer_cst * ic = GetPointer<integer_cst>(GET_NODE(rect->size));
            return static_cast<unsigned int>(get_integer_cst_value(ic));
         }

         if(std::find_if(SC_tmpl_class.begin(), SC_tmpl_class.end(), find_SC_string(nt)) == SC_tmpl_class.end())
         {
            if(std::find_if(SC_builtin_scalar_type.begin(), SC_builtin_scalar_type.end(), find_SC_string(nt)) != SC_builtin_scalar_type.end())
            {
               tree_vec* tvec = GetPointer<tree_vec>(GET_NODE(rect->tmpl_args));
               return size(tm, GET_INDEX_NODE(tvec->list_of_op[0]));
            }
            else if(std::find_if(TLM_SC_builtin_scalar_type.begin(), TLM_SC_builtin_scalar_type.end(), find_SC_string(nt)) != TLM_SC_builtin_scalar_type.end())
            {
               tree_vec * rtv = GetPointer<tree_vec>(GET_NODE(rect->tmpl_args));
               unsigned int tot = 0;
               for (size_t i = 0; i < rtv->lngt; i++)
                  tot += size(tm, GET_INDEX_NODE(rtv->list_of_op[i]));
               return tot;
            }
            else
            {
               if (nt == "_Bool" ||
                   nt == "bool" ||
                   nt == "sc_logic" ||
                   nt == "sc_in_resolved" ||
                   nt == "sc_inout_resolved" ||
                   nt == "sc_out_resolved" ||
                   nt == "sc_in_clk" ||
                   nt == "sc_inout_clk" ||
                   nt == "sc_out_clk" ||
                   nt == "sc_bit" ||
                   nt == "sc_clock"
                  )
                  return 1;
               else if(nt == "sc_event") //sc_event specialization
                  return 0;
               else if ((nt == "sc_fix") || (nt == "sc_ufix"))
                  THROW_ERROR("not expected record_type" + nt);
               else /*the size is the sum of each member size (sum of flds size)*/
               {
                  unsigned int tot = 0;
                  for (size_t i = 0; i < rect->list_of_flds.size(); i++)
                  {
                     if (rect->list_of_flds[i])
                        if (GET_NODE(rect->list_of_flds[i])->get_kind() == field_decl_K)
                     {
                        field_decl* fd = GetPointer<field_decl>(GET_NODE(rect->list_of_flds[i]));
                        if (fd->type)
                        {
                           unsigned int w = size(tm, GET_INDEX_NODE(fd->type));
                           tot = tot + w;
                        }
                     }
                  }
                  return tot;
               }
            }
         }
         else if (rect->tmpl_args) /*the class has template parameters*/
         {
            tree_vec * rtv = GetPointer<tree_vec>(GET_NODE(rect->tmpl_args));
            THROW_ASSERT(rtv->lngt == 1 || nt == "sc_port", "Expected just one element");
            return size(tm, GET_INDEX_NODE(rtv->list_of_op[0]));
         }
         else
            THROW_ERROR("Unexpected template parameter pattern");
#endif
         break;
      }
      case union_type_K:
      {
         return size(tm, GET_INDEX_NODE(GetPointer<union_type>(t)->size));
      }
      case array_type_K:
      {
         array_type* at = GetPointer<array_type>(t);
         if(at->size)
         {
            tree_nodeRef domain = GET_NODE(at->domn);
            integer_type * it = GetPointer<integer_type>(domain);
            THROW_ASSERT(it, "expected an integer type as array domain");
            unsigned int vec_size;
            if(it->max)
               vec_size = size(tm, GET_INDEX_NODE(it->max)) + 1;
            else
               vec_size = 1;
            return size(tm, GET_INDEX_NODE(at->elts))*vec_size;
         }
         else
            return 0;
      }
      break;

      case integer_cst_K:
      {
         integer_cst *ic = GetPointer<integer_cst>(t);
         return static_cast<unsigned int>(get_integer_cst_value(ic));
      }
      case boolean_type_K:
      {
         return 1;
      }
      case void_type_K:
      {
         return 32;
      }
      case vector_type_K:
      case enumeral_type_K:
      case CharType_K:
      case real_type_K:
      case complex_type_K:
      case function_type_K:
      {
         THROW_ASSERT(GetPointer<type_node>(t)->size,"unexpected pattern " + STR(index));
         return size(tm, GET_INDEX_NODE(GetPointer<type_node>(t)->size));
      }
      case integer_type_K:
      {
         integer_type * it = GetPointer<integer_type>(t);
         THROW_ASSERT(it->size,"unexpected pattern " + STR(index));
         unsigned int prec = it->prec;
         unsigned int algn = it->algn;
         if (prec != algn && prec % algn)
            return prec;
         else
         {
            integer_cst *ic = GetPointer<integer_cst>(GET_NODE(it->size));
            THROW_ASSERT(ic,"unexpected pattern " + STR(index));
            return static_cast<unsigned int>(get_integer_cst_value(ic));
         }
      }
      case call_expr_K:
      {
         call_expr * ce = GetPointer<call_expr>(t);
         return size(tm, GET_INDEX_NODE(ce->type));
/*
         addr_expr * ae = GetPointer<addr_expr>(GET_NODE(ce->fn));
         if(ae)
         {
            function_type * ft;
            if(GET_NODE(ae->op)->get_kind() == function_decl_K)
               ft = GetPointer<function_type>(GET_NODE(GetPointer<function_decl>(GET_NODE(ae->op))->type));
            else
               ft = GetPointer<function_type>(GET_NODE(ae->op));
            if(ft && ft->retn)
            {
               return size (tm, GET_INDEX_NODE(ft->retn));
            }
         }
*/
      }
      case CASE_UNARY_EXPRESSION:
      {
         unary_expr * ue = GetPointer<unary_expr>(t);
         THROW_ASSERT(ue->type, "Expected an unary expr with type");
         return size(tm, GET_INDEX_NODE(ue->type));
      }
      case CASE_BINARY_EXPRESSION:
      {
         binary_expr * be = GetPointer<binary_expr>(t);
         return size(tm, GET_INDEX_NODE(be->type));
      }
      case array_ref_K:
      {
         array_ref * ar = GetPointer<array_ref>(t);
         return size(tm, GET_INDEX_NODE(ar->type));
      }
      case component_ref_K:
      {
         component_ref * cr = GetPointer<component_ref>(t);
         return size(tm, GET_INDEX_NODE(cr->type));
      }
      case string_cst_K:
      {
         string_cst * sc = GetPointer<string_cst>(t);
         return static_cast<unsigned int>(size(tm, GET_INDEX_NODE(sc->type)));
      }
      case vector_cst_K:
      {
         vector_cst * vc = GetPointer<vector_cst>(t);
         return static_cast<unsigned int>(size(tm, GET_INDEX_NODE(vc->type)));
      }
      default:
         THROW_ERROR(std::string("Unexpected type pattern ") + t->get_kind_text());
         return 0;
   }
   return 0;
}

std::string tree_helper::name_type(const tree_managerConstRef tm, int unsigned index)
{
   const tree_nodeRef t = tm->get_tree_node_const(index);
   THROW_ASSERT(GetPointer<type_node>(t) || t->get_kind() == tree_list_K, std::string("expected a type_decl got ") + t->get_kind_text());

   switch (t->get_kind())
   {
      case pointer_type_K:
      {
         return "*" + name_type(tm, GET_INDEX_NODE((GetPointer<pointer_type>(t))->ptd));
      }
      case reference_type_K:
      {
         return name_type(tm, GET_INDEX_NODE((GetPointer<reference_type>(t))->refd));
      }
      case record_type_K:
      {
         record_type* rect = GetPointer<record_type>(t);
         std::string nt;
         if(rect->name)
         {
            if (GET_NODE(rect->name)->get_kind() == type_decl_K)
            {
               type_decl* td = GetPointer<type_decl>(GET_NODE(rect->name));
               if (GET_NODE(td->name)->get_kind() == identifier_node_K)
               {
                  identifier_node* idn = GetPointer<identifier_node>(GET_NODE(td->name));
                  nt = idn->strg;
               }
               else
                  THROW_ERROR("unexpected record type pattern: " + STR(index));
            }
            else if (GET_NODE(rect->name)->get_kind() == identifier_node_K)
            {
               identifier_node* idn = GetPointer<identifier_node>(GET_NODE(rect->name));
               nt = "struct "+ idn->strg;
            }
            else
               THROW_ERROR("unexpected record type pattern: " + STR(index));
         }
         else
         {
            return "_unnamed_" + STR(index);
         }
         if(std::find_if(SC_tmpl_class.begin(), SC_tmpl_class.end(), find_SC_string(nt)) == SC_tmpl_class.end())
            return nt;
         else if (rect->tmpl_args) /*the class has template parameters*/
         {
            tree_vec * rtv = GetPointer<tree_vec>(GET_NODE(rect->tmpl_args));
            THROW_ASSERT(rtv->lngt == 1 || nt == "sc_port", "Expected just one element");
            return name_type(tm, GET_INDEX_NODE(rtv->list_of_op[0]));
         }
         else
            THROW_ERROR("Unexpected template parameter pattern");
         return "";//unreachable code

      }
      case union_type_K:
      {
         union_type* unt = GetPointer<union_type>(t);
         std::string nt;
         if(unt->name)
         {
            if (GET_NODE(unt->name)->get_kind() == type_decl_K)
            {
               type_decl* td = GetPointer<type_decl>(GET_NODE(unt->name));
               if (GET_NODE(td->name)->get_kind() == identifier_node_K)
               {
                  identifier_node* idn = GetPointer<identifier_node>(GET_NODE(td->name));
                  nt = idn->strg;
               }
               else
                  THROW_ERROR("unexpected record type pattern: " + STR(index));
            }
            else if (GET_NODE(unt->name)->get_kind() == identifier_node_K)
            {
               identifier_node* idn = GetPointer<identifier_node>(GET_NODE(unt->name));
               nt = "union "+ idn->strg;
            }
            else
               THROW_ERROR("unexpected record type pattern: " + STR(index));
         }
         else
         {
            return "_unnamed_" + STR(index);
         }
         if(std::find_if(SC_tmpl_class.begin(), SC_tmpl_class.end(), find_SC_string(nt)) == SC_tmpl_class.end())
            return nt;
         else
            THROW_ERROR("Unexpected template parameter pattern");
         return "";//unreachable code

      }
      case array_type_K:
      {
         array_type* at = GetPointer<array_type>(t);
         long long int vec_size = 0;
         if(at->domn)
         {
            tree_nodeRef domain = GET_NODE(at->domn);
            integer_type * it = GetPointer<integer_type>(domain);
            THROW_ASSERT(it, "expected an integer type as array domain");
            if(it->max)
               vec_size = size(tm, GET_INDEX_NODE(it->max)) + 1;
         }
         std::string vec_size_string;
         if(vec_size)
            vec_size_string = boost::lexical_cast<std::string>(vec_size);
         return name_type(tm, GET_INDEX_NODE(at->elts)) + "[" + vec_size_string + "]";
      }
      case enumeral_type_K:
      {
         const enumeral_type * et = GetPointer<enumeral_type>(t);
         if(et->name)
         {
            if(GET_NODE(et->name)->get_kind() == type_decl_K)
            {
               const type_decl * td = GetPointer<type_decl>(GET_NODE(et->name));
               const identifier_node * in = GetPointer<identifier_node>(GET_NODE(td->name));
               return in->strg;
            }
            else if(GET_NODE(et->name)->get_kind() == identifier_node_K)
            {
               const identifier_node * in = GetPointer<identifier_node>(GET_NODE(et->name));
               return "enum " + in->strg;
            }
         }
         return "enum Internal_" + boost::lexical_cast<std::string>(index);
      }
      case boolean_type_K:
      case integer_type_K:
      case CharType_K:
      case real_type_K:
      case complex_type_K:
      case void_type_K:
      case vector_type_K:
      {
         type_node* tnode = GetPointer<type_node>(t);
         if(not tnode->name)
         {
            if(t->get_kind() == integer_type_K)
            {
               return "int";
            }
            return boost::lexical_cast<std::string>(index);
         }
         if (GET_NODE(tnode->name)->get_kind() == type_decl_K)
         {
            type_decl* tdecl = GetPointer<type_decl>(GET_NODE(tnode->name));
            THROW_ASSERT(GET_NODE(tdecl->name)->get_kind() == identifier_node_K, "unexpected type name pattern");
            identifier_node* idn = GetPointer<identifier_node>(GET_NODE(tdecl->name));
            return idn->strg;
         }
         else if (GET_NODE(tnode->name)->get_kind() == identifier_node_K)
         {
            identifier_node* idn = GetPointer<identifier_node>(GET_NODE(tnode->name));
            return idn->strg;
         }
         else
            THROW_ASSERT(false, std::string("unexpected builtin type pattern ") + t->get_kind_text() + " " + boost::lexical_cast<std::string>(index));
         break;
      }
      case function_type_K:
      {
         std::string retn = name_type(tm, GET_INDEX_NODE(GetPointer<function_type>(t)->retn));
         retn += "(*)(";
         if(GetPointer<function_type>(t)->prms)
            retn += name_type(tm, GET_INDEX_NODE(GetPointer<function_type>(t)->prms));
         retn += ")";
         return retn;
      }
      case method_type_K:
      {
         std::string retn = name_type(tm, GET_INDEX_NODE(GetPointer<method_type>(t)->retn));
         retn += "(*)(";
         retn += name_type(tm, GET_INDEX_NODE(GetPointer<method_type>(t)->prms));
         retn += ")";
         return retn;
      }
      case tree_list_K:
      {
         tree_list* tl = GetPointer<tree_list>(t);
         std::string retn;
         if (tl->valu)
         {
            retn = name_type(tm, GET_INDEX_NODE(tl->valu));
         }
         if (tl->chan)
            retn += "," + name_type(tm, GET_INDEX_NODE(tl->chan));
         return retn;
      }
      default:
         THROW_ASSERT(false, std::string("unexpected type pattern ") + t->get_kind_text() + " " + boost::lexical_cast<std::string>(index));
         return "";
   }
   return "";
}

std::string tree_helper::name_tmpl(const tree_managerConstRef tm, const unsigned int index)
{
   const tree_nodeRef t = tm->get_tree_node_const(index);
   if (t->get_kind() == record_type_K)
   {
      record_type* rect = GetPointer<record_type>(t);
      if (rect->tmpl_args) /*the class is a template*/
      {
         for (unsigned int i = 0; i < rect->list_of_flds.size(); i++)
         {
            if (GET_NODE(rect->list_of_flds[i])->get_kind() == type_decl_K)
            {
               type_decl* td = GetPointer<type_decl>(GET_NODE(rect->list_of_flds[i]));
               if (GET_NODE(td->name)->get_kind() == identifier_node_K)
               {
                  identifier_node* idn = GetPointer<identifier_node>(GET_NODE(td->name));
                  return (idn->strg);
               }
            }
         }
      }
   }
   return "";
}

std::string tree_helper::record_name(const tree_managerConstRef tm, const unsigned int index)
{
   const tree_nodeRef t = tm->get_tree_node_const(index);
   if (t->get_kind() == record_type_K)
   {
      record_type* rect = GetPointer<record_type>(t);
      if(rect->name)
      {
         if (GET_NODE(rect->name)->get_kind() == type_decl_K)
         {
            type_decl* td = GetPointer<type_decl>(GET_NODE(rect->name));
            if (GET_NODE(td->name)->get_kind() == identifier_node_K)
            {
               identifier_node* idn = GetPointer<identifier_node>(GET_NODE(td->name));
               return (idn->strg);
            }
         }
      }
   }
   return "";
}

std::string tree_helper::name_function(const tree_managerConstRef tm, const unsigned int index)
{
   const tree_nodeRef t = tm->get_tree_node_const(index);
   if (t->get_kind() == function_decl_K)
   {
      function_decl* fd = GetPointer<function_decl>(t);
      return print_function_name(tm, fd);
   }
   else
      THROW_ERROR(std::string("Node not yet supported ") + t->get_kind_text());
   return "";
}

std::string tree_helper::print_function_name(const tree_managerConstRef TM, const function_decl * fd)
{
   tree_nodeRef name;
   if(fd->mngl)
   {
      name = GET_NODE(fd->mngl);
      if (GET_NODE(fd->name)->get_kind() == identifier_node_K)
      {
         identifier_node *in = GetPointer<identifier_node>(GET_NODE(fd->name));
         std::string function_name1 = in->strg;
         if(function_name1.find("builtin") == std::string::npos)
         {
            name = GET_NODE(fd->name);
         }
      }
      else
         THROW_ERROR(std::string("Node not yet supported ") + name->get_kind_text());
   }
   else
      name = GET_NODE(fd->name);
   std::string res;
   if (name->get_kind() == identifier_node_K)
   {
      identifier_node *in = GetPointer<identifier_node>(name);
      if (in->operator_flag)
      {
         res = "operator ";
         std::set<unsigned int>::const_iterator vend = fd->list_attr.end();
         for (std::set<unsigned int>::const_iterator i = fd->list_attr.begin(); i != vend; i++)
         {
            if (*i == treeVocabularyTokenTypes::TOK_PUBLIC || *i == treeVocabularyTokenTypes::TOK_PRIVATE || *i
                  == treeVocabularyTokenTypes::TOK_PROTECTED || *i == treeVocabularyTokenTypes::TOK_OPERATOR || *i
                  == treeVocabularyTokenTypes::TOK_MEMBER)
               continue;
            else if (*i == treeVocabularyTokenTypes::TOK_CONVERSION)
            {
               function_type *ft = GetPointer<function_type>(GET_NODE(fd->type));
               if (ft)
                  print_type(TM, GET_INDEX_NODE(ft->retn));
               else
               {
                  method_type *mt = GetPointer<method_type>(GET_NODE(fd->type));
                  print_type(TM, GET_INDEX_NODE(mt->retn));
               }
            }
            else if (*i == treeVocabularyTokenTypes::TOK_RSHIFT)
               res = res + ">>";
            else if (*i == treeVocabularyTokenTypes::TOK_LSHIFT)
               res = res + "<<";
            else if (*i == treeVocabularyTokenTypes::TOK_ASSIGN)
               res = res + "=";
            else
               res = res + TI_getTokenName(*i);
         }
         res = normalized_ID(res);
      }
      else
         res = normalized_ID(in->strg);
   }
   else
      THROW_ERROR(std::string("Node not yet supported ") + name->get_kind_text());
   return res;
}

std::tr1::tuple<std::string, unsigned int, unsigned int> tree_helper::get_definition(const tree_managerConstRef tm, const unsigned int index, bool & is_system)
{
   is_system = false;
   tree_nodeRef node = tm->get_tree_node_const(index);
   std::string include_name;
   unsigned int line_number = 0;
   unsigned int column_number = 0;
   if(GetPointer<type_node>(node))
   {
      type_node * tn = GetPointer<type_node>(node);
      if(tn->name && GetPointer<decl_node>(GET_NODE(tn->name)))
      {
         node = GET_NODE(tn->name);
      }
      else
      {
         if(GetPointer<union_type>(node))
         {
            std::vector<tree_nodeRef> list_of_flds = GetPointer<union_type>(node)->list_of_flds;
            if(list_of_flds.size())
            {
               tree_nodeRef field = GET_NODE(list_of_flds[0]);
               if(GetPointer<decl_node>(field))
               {
                  decl_node * dn = GetPointer<decl_node>(field);
                  include_name = dn->include_name;
                  line_number = dn->line_number;
                  column_number = dn->column_number;
                  is_system = dn->operating_system_flag or dn->library_system_flag;
               }
            }
         }
         if(GetPointer<record_type>(node))
         {
            std::vector<tree_nodeRef> list_of_flds = GetPointer<record_type>(node)->list_of_flds;
            if(list_of_flds.size())
            {
               tree_nodeRef field = GET_NODE(list_of_flds[0]);
               if(GetPointer<decl_node>(field))
               {
                  decl_node * dn = GetPointer<decl_node>(field);
                  include_name = dn->include_name;
                  line_number = dn->line_number;
                  column_number = dn->column_number;
                  is_system = dn->operating_system_flag or dn->library_system_flag;
               }
            }
         }
      }
   }
   if(GetPointer<decl_node>(node))
   {
      decl_node * dn = GetPointer<decl_node>(node);
      include_name = dn->include_name;
      line_number = dn->line_number;
      column_number = dn->column_number;
      is_system = dn->operating_system_flag or dn->library_system_flag;
   }
   return std::tr1::tuple<std::string, unsigned int, unsigned int>(include_name,line_number, column_number);
}

bool tree_helper::has_function_return(const tree_managerConstRef tm, const unsigned int index)
{
   const tree_nodeRef &curr_tn = tm->get_tree_node_const(index);
   function_decl* fd = GetPointer<function_decl>(curr_tn);
   tree_nodeRef fd_type = GET_NODE(fd->type);
   function_type * ft = GetPointer<function_type>(fd_type);
   if(GET_NODE(ft->retn)->get_kind() != void_type_K) //not a procedure but a function
      return true;
   return false;

}

std::set<unsigned int> tree_helper::calls_from_function(const tree_managerConstRef tm, const unsigned int index)
{
   std::set<unsigned int> temp_set;
   temp_set.clear();
   tree_helper::calls_from_function(tm, index, temp_set);
   return temp_set;
}

void tree_helper::calls_from_function(const tree_managerConstRef tm, const unsigned int index, std::set<unsigned int> & set_index)
{
   const tree_nodeRef t = tm->get_tree_node_const(index);

   if (GetPointer<unary_expr>(t))
   {
      unary_expr * un = GetPointer<unary_expr>(t);
      if (un->op)
         calls_from_function(tm, GET_INDEX_NODE(un->op), set_index);
   }
   else if (GetPointer<binary_expr>(t))
   {
      binary_expr * bin = GetPointer<binary_expr>(t);
      if (bin->op0)
         calls_from_function(tm, GET_INDEX_NODE(bin->op0), set_index);
      if (bin->op1)
         calls_from_function(tm, GET_INDEX_NODE(bin->op1), set_index);
   }
   else if (GetPointer<ternary_expr>(t))
   {
      ternary_expr * tern = GetPointer<ternary_expr>(t);
      if (tern->op0)
         calls_from_function(tm, GET_INDEX_NODE(tern->op0), set_index);
      if (tern->op1)
         calls_from_function(tm, GET_INDEX_NODE(tern->op1), set_index);
      if (tern->op2)
         calls_from_function(tm, GET_INDEX_NODE(tern->op2), set_index);
   }

   switch (t->get_kind())
   {
      case function_decl_K:
      {
         if ((GetPointer<function_decl>(t))->body)
            calls_from_function(tm, GET_INDEX_NODE((GetPointer<function_decl>(t))->body), set_index);
         else if (tm->get_implementation_node(index))
            calls_from_function(tm, tm->get_implementation_node(index), set_index);
         break;
      }
      case gimple_bind_K:
      {
         if ((GetPointer<gimple_bind>(t))->body)
            calls_from_function(tm, GET_INDEX_NODE((GetPointer<gimple_bind>(t))->body), set_index);
         break;
      }
      case statement_list_K:
      {
         statement_list * sl = GetPointer<statement_list>(t);
         if (sl->list_of_stmt.size())
         {
            for (unsigned int i = 0; i < sl->list_of_stmt.size(); i++)
            {
               calls_from_function(tm, GET_INDEX_NODE(sl->list_of_stmt[i]), set_index);
            }
         }
         else
         {
            std::map<unsigned int, blocRef>::iterator ib_end = sl->list_of_bloc.end();
            for(std::map<unsigned int, blocRef>::iterator ib = sl->list_of_bloc.begin(); ib != ib_end; ib++)
            {
               for (unsigned int i = 0; i < ib->second->list_of_stmt.size(); i++)
                  calls_from_function(tm, GET_INDEX_NODE(ib->second->list_of_stmt[i]), set_index);
            }
         }
         break;
      }
      case call_expr_K:
      {
         call_expr * ce = GetPointer<call_expr>(t);
         if (ce->fn)
         {
            if (GET_NODE(ce->fn)->get_kind() == addr_expr_K)
            {
               addr_expr * adr = GetPointer<addr_expr>(GET_NODE(ce->fn));
               if ((adr->op) && (GET_NODE(adr->op)->get_kind() == function_decl_K))
               {
                  set_index.insert(GET_INDEX_NODE(adr->op));
                  return;
               }
            }
         }
         break;
      }
      case gimple_call_K:
      {
         gimple_call * ce = GetPointer<gimple_call>(t);
         if (ce->fn)
         {
            if (GET_NODE(ce->fn)->get_kind() == addr_expr_K)
            {
               addr_expr * adr = GetPointer<addr_expr>(GET_NODE(ce->fn));
               if ((adr->op) && (GET_NODE(adr->op)->get_kind() == function_decl_K))
               {
                  set_index.insert(GET_INDEX_NODE(adr->op));
                  return;
               }
            }
         }
         break;
      }
      default:
      {
         THROW_ERROR("not expected case" + t->get_kind_text());
      }
   }
   return;
}

void tree_helper::get_used_variables(bool first_level_only, const expandRef dnt, const tree_nodeRef tRI, std::tr1::unordered_set<unsigned int> & list_of_variable)
{
   if (!tRI)
      return;
   tree_nodeRef t = tRI;
   THROW_ASSERT(tRI->get_kind() == tree_reindex_K, "Node is not a tree reindex");
   t = GET_NODE(tRI);
   switch (t->get_kind())
   {
      case result_decl_K://tree_to_graph considers this object as particular type of variable
      {
         result_decl * rd = GetPointer<result_decl>(t);
         list_of_variable.insert(GET_INDEX_NODE(tRI));
         if(rd->init)
            get_used_variables(first_level_only, dnt, rd->init, list_of_variable);
         break;
      }
      case var_decl_K:
      {
         var_decl * vd = GetPointer<var_decl>(t);
         list_of_variable.insert(GET_INDEX_NODE(tRI));
         if(vd->init)
            get_used_variables(first_level_only, dnt, vd->init, list_of_variable);
         break;
      }
      case ssa_name_K:
      {
         ssa_name * sn= GetPointer<ssa_name>(t);
         get_used_variables(first_level_only, dnt, sn->var, list_of_variable);
         var_decl * vd = GetPointer<var_decl>(GET_NODE(sn->var));
         list_of_variable.insert(GET_INDEX_NODE(tRI));
         if(vd && vd->init)
            get_used_variables(first_level_only, dnt, vd->init, list_of_variable);
         break;
      }
      case parm_decl_K:
         list_of_variable.insert(GET_INDEX_NODE(tRI));
         break;
      case function_decl_K:
      {
         function_decl* fd = GetPointer<function_decl>(t);
         bool expand_p = !first_level_only && (*dnt)(t);
         std::vector<tree_nodeRef>::const_iterator vend = fd->list_of_args.end();
         list_of_variable.insert(GET_INDEX_NODE(tRI));
         if (fd->body && expand_p)
         {
            for (std::vector<tree_nodeRef>::const_iterator i = fd->list_of_args.begin(); i != vend; i++)
            {
               get_used_variables(first_level_only, dnt, *i, list_of_variable);
            }
            // body analysis
            get_used_variables(first_level_only, dnt, fd->body, list_of_variable);
         }
      }
      break;
      case statement_list_K:
      {
         statement_list* sl = GetPointer<statement_list>(t);
         std::vector<tree_nodeRef>::iterator end = sl->list_of_stmt.end();
         std::vector<tree_nodeRef>::iterator i = sl->list_of_stmt.begin();
         if(i != end)
            for (; i != end; i++)
               get_used_variables(first_level_only, dnt, *i, list_of_variable);
         else
         {
            std::map<unsigned int, blocRef>::iterator ib_end = sl->list_of_bloc.end();
            for(std::map<unsigned int, blocRef>::iterator ib = sl->list_of_bloc.begin(); ib != ib_end; ib++)
            {
               end = ib->second->list_of_stmt.end();
               for(i = ib->second->list_of_stmt.begin(); i != end; i++)
                  get_used_variables(first_level_only, dnt, *i, list_of_variable);
            }
         }
      }
      break;
      case tree_vec_K:
      {
         tree_vec* tv = GetPointer<tree_vec>(t);
         std::vector<tree_nodeRef>::iterator end = tv->list_of_op.end();
         for (std::vector<tree_nodeRef>::iterator i = tv->list_of_op.begin(); i != end; i++)
         {
            get_used_variables(first_level_only, dnt, *i, list_of_variable);
         }
      }
      break;
      case gimple_cond_K:
      {
         gimple_cond* gc = GetPointer<gimple_cond>(t);
         get_used_variables(first_level_only, dnt, gc->op0, list_of_variable);
      }
      break;
      case gimple_assign_K:
      {
         gimple_assign * me = GetPointer<gimple_assign>(t);
         get_used_variables(first_level_only, dnt, me->op0, list_of_variable);
         get_used_variables(first_level_only, dnt, me->op1, list_of_variable);
      }
      break;
      case gimple_return_K:
      {
         gimple_return* re = GetPointer<gimple_return>(t);
         if (re->op)
            get_used_variables(first_level_only, dnt, re->op, list_of_variable);
      }
      break;
      case CASE_UNARY_EXPRESSION:
      {
         unary_expr* ue = GetPointer<unary_expr>(t);
         if(list_of_variable.find(GET_INDEX_NODE(ue->op)) != list_of_variable.end())
         {
            break;
         }
         get_used_variables(first_level_only, dnt, ue->op, list_of_variable);
         break;
      }
      case CASE_BINARY_EXPRESSION:
      {
         binary_expr* be = GetPointer<binary_expr>(t);
         get_used_variables(first_level_only, dnt, be->op0, list_of_variable);
         get_used_variables(first_level_only, dnt, be->op1, list_of_variable);
         break;
      }
      case CASE_TERNARY_EXPRESSION:
      {
         ternary_expr * tern = GetPointer<ternary_expr>(t);
         get_used_variables(first_level_only, dnt, tern->op0, list_of_variable);
         get_used_variables(first_level_only, dnt, tern->op1, list_of_variable);
         get_used_variables(first_level_only, dnt, tern->op2, list_of_variable);
         break;
      }
      /* quaternary expressions.*/
      case CASE_QUATERNARY_EXPRESSION:
      {
         quaternary_expr*qe = GetPointer<quaternary_expr>(t);
         get_used_variables(first_level_only, dnt, qe->op0, list_of_variable);
         get_used_variables(first_level_only, dnt, qe->op1, list_of_variable);
         get_used_variables(first_level_only, dnt, qe->op2, list_of_variable);
         get_used_variables(first_level_only, dnt, qe->op3, list_of_variable);
         break;
      }
      case gimple_switch_K:
      {
         gimple_switch * s = GetPointer<gimple_switch>(t);
         get_used_variables(first_level_only, dnt, s->op0, list_of_variable);
         break;
      }
      case label_decl_K:
      case integer_cst_K:
      case real_cst_K:
      case string_cst_K:
      case vector_cst_K:
      case complex_cst_K:
         break;
      case field_decl_K://used to specify the displacement
      case case_label_expr_K:
      case gimple_label_K:
      case filter_expr_K:
      case exc_ptr_expr_K:
         break;
      case constructor_K:
      {
         constructor* co = GetPointer<constructor>(t);
         for(std::vector<std::pair< tree_nodeRef, tree_nodeRef> >::const_iterator i = co->list_of_idx_valu.begin(); i != co->list_of_idx_valu.end(); i++)
            get_used_variables(first_level_only, dnt, i->second, list_of_variable);
         break;
      }
      case call_expr_K:
      {
         call_expr*ce = GetPointer<call_expr>(t);
         const std::vector<tree_nodeRef> args = ce->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            get_used_variables(first_level_only, dnt, *arg, list_of_variable);
         }
         break;
      }
      case gimple_call_K:
      {
         gimple_call*ce = GetPointer<gimple_call>(t);
         const std::vector<tree_nodeRef> args = ce->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            get_used_variables(first_level_only, dnt, *arg, list_of_variable);
         }
         break;
      }
      case gimple_asm_K:
      {
         gimple_asm * ae = GetPointer<gimple_asm>(t);
         ///str and clob are always string
         get_used_variables(first_level_only, dnt, ae->in, list_of_variable);
         if(ae->out)
            get_used_variables(first_level_only, dnt, ae->out, list_of_variable);
         break;
      }
      case tree_list_K:
      {
         tree_list* tl = GetPointer<tree_list>(t);
         if (tl->purp)
            get_used_variables(first_level_only, dnt, tl->purp, list_of_variable);
         if (tl->valu)
            get_used_variables(first_level_only, dnt, tl->valu, list_of_variable);
         if (tl->chan)
            get_used_variables(first_level_only, dnt, tl->chan, list_of_variable);
         break;
      }
      case for_expr_K:
      {
         const for_expr * fe = GetPointer<for_expr>(t);
         get_used_variables(first_level_only, dnt, fe->op0, list_of_variable);
         get_used_variables(first_level_only, dnt, fe->op1, list_of_variable);
         get_used_variables(first_level_only, dnt, fe->op2, list_of_variable);
         break;
      }
      case while_expr_K:
      {
         const while_expr * we = GetPointer<while_expr>(t);
         get_used_variables(first_level_only, dnt, we->op0, list_of_variable);
         break;
      }
      default:
         THROW_ERROR(std::string("Node not yet supported ") + t->get_kind_text());

   }
}

bool tree_helper::look_for_binfo_inheritance(binfo* b, std::string bcs)
{
   bool found = false;
   if (b)
   {
      if (b->type)
      {
         record_type * rt = GetPointer<record_type>(GET_NODE(b->type));
         if (rt && rt->get_maybe_name() == bcs)
            return true;
      }
      for (unsigned int i = 0; i < b->get_baseinfo_size(); i++)
      {
         tree_nodeRef binf = b->get_base(i);
         binfo* bnf = GetPointer<binfo>(GET_NODE(binf));
         found = look_for_binfo_inheritance(bnf, bcs);
         if (found)
            return true;
      }
   }
   return false;
}

tree_nodeRef tree_helper::find_obj_type_ref_function(const tree_nodeRef tn)
{
   const tree_nodeRef curr_tn = GET_NODE(tn);
   unsigned int ind = GET_INDEX_NODE(tn);
   obj_type_ref* otr = GetPointer<obj_type_ref>(curr_tn);
   THROW_ASSERT(otr, "tree node is not an obj_type_ref");
   THROW_ASSERT(otr->type && otr->op1 && otr->op2, "obj_type_ref has missing fields");
   tree_nodeRef type;
   unsigned int function_type;
   
   pointer_type* t_pt = GetPointer<pointer_type>(GET_NODE(otr->type));
   THROW_ASSERT(t_pt, "Expected a pointer_type");
   function_type = GET_INDEX_NODE(t_pt->ptd);
   THROW_ASSERT(GetPointer<method_type>(GET_NODE(t_pt->ptd)), "expected a method_type");
   type = GET_NODE(GetPointer<method_type>(GET_NODE(t_pt->ptd))->clas);
#if 0
   var_decl* vd = GetPointer<var_decl>(GET_NODE(otr->op1));
   if(vd)
      type = vd->type;
   else
   {
      parm_decl* pd = GetPointer<parm_decl>(GET_NODE(otr->op1));
      if(pd)
         type = pd->type;
      else
         THROW_ERROR(std::string("not supported case for obj_type_ref(") + boost::lexical_cast<std::string>(ind) + std::string(")"));
   }
#endif
   if(type)
   {
      record_type* rt = GetPointer<record_type>(type);
#if 0
      pointer_type* pt = GetPointer<pointer_type>(GET_NODE(type));
      if(pt)
         rt = GetPointer<record_type>(GET_NODE(pt->ptd));
      else
      {
         reference_type * Rt = GetPointer<reference_type>(GET_NODE(type));
         if(Rt)
            rt = GetPointer<record_type>(GET_NODE(Rt->refd));
         else
            THROW_ERROR(std::string("not supported case for obj_type_ref(") + boost::lexical_cast<std::string>(ind) + std::string(")"));
      }
#endif
      if(rt)
      {
         for(std::vector<tree_nodeRef>::const_iterator x = rt->list_of_fncs.begin(); x != rt->list_of_fncs.end(); x++)
         {
            THROW_ASSERT(GET_NODE(*x)->get_kind() == function_decl_K || GET_NODE(*x)->get_kind() == template_decl_K, "expected a function decl or a template_decl");
            function_decl* fd = GetPointer<function_decl>(GET_NODE(*x));
            if(GET_INDEX_NODE(fd->type) == function_type)
               return *x;
         }
      }
      else
         THROW_ERROR(std::string("not supported case for obj_type_ref(") + boost::lexical_cast<std::string>(ind) + std::string(")"));
   }
   else
      THROW_ERROR(std::string("not supported case for obj_type_ref(") + boost::lexical_cast<std::string>(ind) + std::string(")"));
   THROW_ERROR(std::string("obj_type_ref Function not found (") + boost::lexical_cast<std::string>(ind) + std::string(")"));
   return tree_nodeRef();
}

bool tree_helper::is_system(const tree_managerConstRef  TM, const unsigned int index)
{
   const tree_nodeRef curr_tn = TM->get_tree_node_const(index);
   if(GetPointer<decl_node>(curr_tn))
      return GetPointer<decl_node>(curr_tn)->operating_system_flag or GetPointer<decl_node>(curr_tn)->library_system_flag;
   if(GetPointer<type_node>(curr_tn))
      return GetPointer<type_node>(curr_tn)->system_flag;
   return false;
}

const std::tr1::unordered_set<unsigned int> tree_helper::GetTypesToBeDeclared(const tree_managerConstRef TM, const unsigned int index)
{
   return RecursiveGetTypesToBeDeclared(TM, index, false);
}

const std::tr1::unordered_set<unsigned int> tree_helper::RecursiveGetTypesToBeDeclared(const tree_managerConstRef TM, const unsigned int index, const bool recursion)
{
   std::tr1::unordered_set<unsigned int> returned_types;
   const tree_nodeRef type = TM->get_tree_node_const(index);
   switch(type->get_kind())
   {
      case pointer_type_K:
      {
         returned_types = RecursiveGetTypesToBeDeclared(TM, get_pointed_type(TM, index), true);
         break;
      }
      case reference_type_K:
      {
         reference_type * rt = GetPointer<reference_type>(type);
         returned_types = RecursiveGetTypesToBeDeclared(TM, GET_INDEX_NODE(rt->refd), true);
         break;
      }
      case array_type_K:
      case vector_type_K:
      {
         const type_node * tn = GetPointer<type_node>(type);
         if(recursion and tn->name and GET_NODE(tn->name)->get_kind() == type_decl_K)
         {
            returned_types.insert(index);
         }
         else
         {
            returned_types = RecursiveGetTypesToBeDeclared(TM, GetElements(TM, index), true);
         }
         break;
      }
      case record_type_K:
      {
         if(recursion)
         {
            returned_types.insert(index);
         }
         else
         {
            const record_type * rt = GetPointer<record_type>(TM->get_tree_node_const(index));
            if(rt->unql and GetPointer<record_type>(GET_NODE(rt->unql))->name)
            {
               returned_types.insert(GET_INDEX_NODE(rt->unql));
            }
            else
            {
               const std::list<unsigned int> field_types = get_fields_type(TM, index);
               std::list<unsigned int>::const_iterator field_type, field_type_end = field_types.end();
               for(field_type = field_types.begin(); field_type != field_type_end; field_type++)
               {
                  const std::tr1::unordered_set<unsigned int> local_types = RecursiveGetTypesToBeDeclared(TM, *field_type, true);
                  returned_types.insert(local_types.begin(), local_types.end());
               }
            }
         }
         break;
      }
      case union_type_K:
      {
         if(recursion)
         {
            returned_types.insert(index);
         }
         else
         {
            const union_type * ut = GetPointer<union_type>(TM->get_tree_node_const(index));
            if(ut->unql and GetPointer<union_type>(GET_NODE(ut->unql))->name)
            {
               returned_types.insert(GET_INDEX_NODE(ut->unql));
            }
            else
            {
               const std::list<unsigned int> field_types = get_fields_type(TM, index);
               std::list<unsigned int>::const_iterator field_type, field_type_end = field_types.end();
               for(field_type = field_types.begin(); field_type != field_type_end; field_type++)
               {
                  const std::tr1::unordered_set<unsigned int> local_types = RecursiveGetTypesToBeDeclared(TM, *field_type, true);
                  returned_types.insert(local_types.begin(), local_types.end());
               }
            }
         }
         break;
      }
      case enumeral_type_K:
      {
         if(recursion)
         {
            returned_types.insert(index);
         }
         else
         {
            const enumeral_type * et = GetPointer<enumeral_type>(TM->get_tree_node_const(index));
            if(et->unql and GetPointer<enumeral_type>(GET_NODE(et->unql))->name)
            {
               returned_types.insert(GET_INDEX_NODE(et->unql));
            }
         }
         break;
      }
      case boolean_type_K:
      case complex_type_K:
      case integer_type_K:
      case real_type_K:
      {
         const type_node * tn = GetPointer<type_node>(type);
         if(tn->name and GET_NODE(tn->name)->get_kind() == type_decl_K)
         {
            returned_types.insert(index);
         }
         break;
      }
      case void_type_K:
      {
         break;
      }
      case function_type_K:
      {
         const unsigned int return_type = get_return_type(TM, index);
         if(return_type)
         {
            const std::tr1::unordered_set<unsigned int> return_types = RecursiveGetTypesToBeDeclared(TM, return_type, true);
            returned_types.insert(return_types.begin(), return_types.end());
         }
         std::list<unsigned int> parameters;
         get_parameter_types(TM, index, parameters);
         std::list<unsigned int>::const_iterator parameter, parameter_end = parameters.end();
         for (parameter = parameters.begin(); parameter != parameter_end; parameter++)
         {
            const std::tr1::unordered_set<unsigned int> local_types = RecursiveGetTypesToBeDeclared(TM, *parameter, true);
            returned_types.insert(local_types.begin(), local_types.end());
         }
         break;
      }
      default:
      {
         THROW_ASSERT(false, "Unexpected node: " + boost::lexical_cast<std::string>(index));
      }
   }
   return returned_types;
}

unsigned int tree_helper::GetRealType(const tree_managerConstRef TM, unsigned int index)
{
   const record_type * rt = GetPointer<record_type>(TM->get_tree_node_const(index));
   if(rt and rt->unql and (not rt->name or GET_NODE(rt->name)->get_kind() == identifier_node_K))
   {
      return GET_INDEX_NODE(rt->unql);
   }
   const union_type * ut = GetPointer<union_type>(TM->get_tree_node_const(index));
   if(ut and ut->unql and (not ut->name or GET_NODE(ut->name)->get_kind() == identifier_node_K))
   {
      return GET_INDEX_NODE(ut->unql);
   }
   const enumeral_type * et = GetPointer<enumeral_type>(TM->get_tree_node_const(index));
   if(et and et->unql and (not et->name or GET_NODE(et->name)->get_kind() == identifier_node_K))
   {
      return GET_INDEX_NODE(et->unql);
   }
   return index;
}

unsigned int tree_helper::get_type_index(const tree_managerConstRef TM, const unsigned int index, long long int &vec_size, bool &is_a_pointer, bool &is_a_function)
{
   is_a_pointer =  false;
   is_a_function = false;
   vec_size = 0;
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(T, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      //decl_node * dn = GetPointer<decl_node>(T);
      //THROW_ASSERT(dn, "expected a declaration object");
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index " + boost::lexical_cast<std::string>(index));
      //type_index = GET_INDEX_NODE(dn->type);
   }
   if (Type->get_kind() == pointer_type_K)
      is_a_pointer = true;
   else if (Type->get_kind() == reference_type_K)
      is_a_pointer = true;//reference objects are assimilated to pointers
   else if (Type->get_kind() == array_type_K)
   {
      array_type * at = GetPointer<array_type>(Type);
      if(!at->domn)
         is_a_pointer = true;
   }
   else if (T->get_kind() == function_decl_K)
      is_a_function = true;
   return type_index;
}

unsigned int tree_helper::get_return_type(const tree_managerConstRef TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
   }
   THROW_ASSERT(GetPointer<function_type>(Type), "Index " + boost::lexical_cast<std::string>(index) + " does not correspond to a function type");
   if(GetPointer<function_type>(Type)->retn)
      return(GET_INDEX_NODE(GetPointer<function_type>(Type)->retn));
   else
      return 0;
}

unsigned int tree_helper::get_pointed_type(const tree_managerConstRef TM, const int unsigned index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(GetPointer<pointer_type>(T), "Index " + boost::lexical_cast<std::string>(index) + " does not correspond to a pointer type");
   return GET_INDEX_NODE(GetPointer<pointer_type>(T)->ptd);
}

unsigned int tree_helper::GetElements(const tree_managerConstRef TM, const unsigned int index)
{
   const array_type * at = GetPointer<array_type>(TM->get_tree_node_const(index));
   if(at)
   {
      return GET_INDEX_NODE(at->elts);
   }
   const vector_type * vt = GetPointer<vector_type>(TM->get_tree_node_const(index));
   if(vt)
   {
      return GET_INDEX_NODE(vt->elts);
   }
   THROW_ASSERT(false, "Tree node " + boost::lexical_cast<std::string>(index) + " is of type " + TM->get_tree_node_const(index)->get_kind_text());
   return 0;
}

std::string tree_helper::get_type_name(const tree_managerConstRef  TM, const unsigned int index)
{
   unsigned int type_index;
   tree_nodeRef type = get_type_node(TM->get_tree_node_const(index), type_index);
   THROW_ASSERT(GetPointer<type_node>(type), "Node type not type_node");
   type_node * tn = GetPointer<type_node>(type);
   if(tn->name)
   {
      tree_nodeRef name;
      if(GET_NODE(tn->name)->get_kind() == type_decl_K)
      {
         name = GetPointer<type_decl>(GET_NODE(tn->name))->name;
      }
      else
         name = tn->name;
      THROW_ASSERT(GET_NODE(name)->get_kind() == identifier_node_K, "Not an identifier node");
      identifier_node * id = GetPointer<identifier_node>(GET_NODE(name));
      return id->strg;
   }
   else
      return "Internal_" + boost::lexical_cast<std::string>(type_index);
}

void tree_helper::get_parameter_types(const tree_managerConstRef TM, const unsigned int index, std::list<unsigned int> & params)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
   }
   THROW_ASSERT(GetPointer<function_type>(Type), "Index " + boost::lexical_cast<std::string>(index) + " does not correspond to a function type");
   if(GetPointer<function_type>(Type)->prms)
   {
      tree_list * tl = GetPointer<tree_list>(GET_NODE(GetPointer<function_type>(Type)->prms));
      params.push_back(GET_INDEX_NODE(tl->valu));
      while(tl->chan)
      {
         tl = GetPointer<tree_list>(GET_NODE(tl->chan));
         params.push_back(GET_INDEX_NODE(tl->valu));
      }
   }
}

unsigned int tree_helper::get_type_index(const tree_managerConstRef TM, const unsigned int index)
{
   bool is_a_pointer;
   bool is_a_function;
   long long int vec_size;
   return get_type_index(TM, index, vec_size, is_a_pointer, is_a_function);
}

std::list<unsigned int> tree_helper::get_fields_type(const tree_managerConstRef TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(T, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   tree_nodeRef Type;
   if(GetPointer<type_node>(T))
   {
      Type = T;
   }
   else
   {
      unsigned int type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
   }
   std::list<unsigned int> ret;
   if (Type->get_kind() == record_type_K)
   {
      record_type * rt = GetPointer<record_type>(Type);
      for(std::vector<tree_nodeRef>::iterator it = rt->list_of_flds.begin(); it != rt->list_of_flds.end(); it++)
      {
         unsigned int field_type = tree_helper::get_type_index(TM, GET_INDEX_NODE(*it));
         ret.push_back(field_type);
      }
   }
   else if (Type->get_kind() == union_type_K)
   {
      union_type * ut = GetPointer<union_type>(Type);
      for(std::vector<tree_nodeRef>::iterator it = ut->list_of_flds.begin(); it != ut->list_of_flds.end(); it++)
      {
         unsigned int field_type = tree_helper::get_type_index(TM, GET_INDEX_NODE(*it));
         ret.push_back(field_type);
      }
   }
   else
      THROW_ERROR("Asking fields type of a variable not of struct type. Variable is " + boost::lexical_cast<std::string>(index));
   return ret;
}


unsigned int tree_helper::get_field_idx(const tree_managerConstRef  TM, const unsigned int index, unsigned int idx)
{
   tree_nodeRef node = TM->get_tree_node_const(index);
   THROW_ASSERT(GetPointer<record_type>(node) || GetPointer<union_type>(node), "expected record or union type");
   record_type * rt = GetPointer<record_type>(node);
   union_type * ut = GetPointer<union_type>(node);
   if(rt)
   {
      THROW_ASSERT(idx< rt->list_of_flds.size(), "unexpected index for list of fields");
      return GET_INDEX_NODE(rt->list_of_flds[idx]);
   }
   else
   {
      THROW_ASSERT(idx< ut->list_of_flds.size(), "unexpected index for list of fields");
      return GET_INDEX_NODE(ut->list_of_flds[idx]);
   }
}

unsigned int tree_helper::local_return_index = 0;

tree_nodeRef tree_helper::get_type_node (const tree_nodeRef &node, unsigned int &return_index)
{
   switch (node->get_kind())
   {
      case call_expr_K:
      {
         call_expr * ce = GetPointer<call_expr>(node);
         return_index = GET_INDEX_NODE(ce->type);
         return GET_NODE(ce->type);
         /*otherwise fall through*/
      }
      case gimple_asm_K:
      case gimple_return_K:
      case while_expr_K:
      case for_expr_K:
      case gimple_switch_K:
      case gimple_label_K:
      case gimple_goto_K:
      case gimple_nop_K:
      case gimple_call_K:
      case gimple_cond_K:
      {
            return_index = 0;
            return tree_nodeRef();
      }
      case CASE_UNARY_EXPRESSION:
      case CASE_BINARY_EXPRESSION:
      case CASE_TERNARY_EXPRESSION:
      case CASE_QUATERNARY_EXPRESSION:
      case exc_ptr_expr_K:
      case filter_expr_K:
      {
         expr_node * en = GetPointer<expr_node>(node);
         THROW_ASSERT(en && en->type, std::string("this NODE does not have a type: ") + node->get_kind_text());
         return_index = GET_INDEX_NODE(en->type);
         return GET_NODE(en->type);
      }
      case gimple_assign_K:
      {
         gimple_assign * gm = GetPointer<gimple_assign>(node);
         return get_type_node(GET_NODE(gm->op0), return_index);
      }
      case integer_cst_K:
      {
         integer_cst*ic = GetPointer<integer_cst>(node);
         return_index = GET_INDEX_NODE(ic->type);
         return GET_NODE(ic->type);
      }
      case real_cst_K:
      {
         real_cst*rc = GetPointer<real_cst>(node);
         return_index = GET_INDEX_NODE(rc->type);
         return GET_NODE(rc->type);
      }
      case string_cst_K:
      {
         string_cst*sc = GetPointer<string_cst>(node);
         return_index = sc->type ? GET_INDEX_NODE(sc->type) : 0;
         return sc->type ? GET_NODE(sc->type) : node;
      }
      case vector_cst_K:
      {
         vector_cst*vc = GetPointer<vector_cst>(node);
         return_index = vc->type ? GET_INDEX_NODE(vc->type) : 0;
         return vc->type ? GET_NODE(vc->type) : node;
      }
      case complex_cst_K:
      {
         complex_cst * cc = GetPointer<complex_cst>(node);
         return_index = cc->type ? GET_INDEX_NODE(cc->type) : 0;
         return cc->type ? GET_NODE(cc->type) : node;
      }
      case constructor_K:
      {
         constructor*c = GetPointer<constructor>(node);
         if(c->type)
         {
            return_index = GET_INDEX_NODE(c->type);
            return GET_NODE(c->type);
         }
         else
         {
            return_index = 0;
            return tree_nodeRef();
         }
      }
      case const_decl_K:
      case field_decl_K:
      case function_decl_K:
      case label_decl_K:
      case namespace_decl_K:
      case result_decl_K:
      case template_decl_K:
      case translation_unit_decl_K:
      case type_decl_K:
      case parm_decl_K:
      case struct_field_tag_K:
      case memory_partition_tag_K:
      case name_memory_tag_K:
      case symbol_memory_tag_K:
      case var_decl_K:
      {
         decl_node *dn = GetPointer<decl_node>(node);
         return_index = GET_INDEX_NODE(dn->type);
         return GET_NODE(dn->type);
      }
      case ssa_name_K:
      {
         ssa_name * sa = GetPointer<ssa_name>(node);
         return get_type_node(GET_NODE(sa->var), return_index);
      }
      case target_mem_ref_K:
      {
         target_mem_ref * tm = GetPointer<target_mem_ref>(node);
         return get_type_node(GET_NODE(tm->orig), return_index);
      }
      case target_mem_ref461_K:
      {
         target_mem_ref461 * tm = GetPointer<target_mem_ref461>(node);
         return_index = GET_INDEX_NODE(tm->type);
         return GET_NODE(tm->type);
      }

      default:
         THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC,std::string("Node not yet supported ") + node->get_kind_text());
   }
   return node;
}

bool tree_helper::is_an_enum(const tree_managerConstRef  TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(T, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
   }
   if(Type->get_kind() == enumeral_type_K)
      return true;
   return false;
}

bool tree_helper::is_a_struct(const tree_managerConstRef  TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(T, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      //decl_node * dn = GetPointer<decl_node>(T);
      //THROW_ASSERT(dn, "expected a declaration object");
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
      //type_index = GET_INDEX_NODE(dn->type);
   }
   if(Type->get_kind() == record_type_K)
      return true;
   return false;
}

bool tree_helper::is_an_union(const tree_managerConstRef  TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(T, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      //decl_node * dn = GetPointer<decl_node>(T);
      //THROW_ASSERT(dn, "expected a declaration object");
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
      //type_index = GET_INDEX_NODE(dn->type);
   }
   if(Type->get_kind() == union_type_K)
   {
      return true;
   }
   return false;
}

bool tree_helper::is_a_complex(const tree_managerConstRef  TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(T, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      //decl_node * dn = GetPointer<decl_node>(T);
      //THROW_ASSERT(dn, "expected a declaration object");
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
      //type_index = GET_INDEX_NODE(dn->type);
   }
   if(Type->get_kind() == complex_type_K)
   {
      return true;
   }
   return false;
}

bool tree_helper::is_an_array(const tree_managerConstRef  TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(T, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      //decl_node * dn = GetPointer<decl_node>(T);
      //THROW_ASSERT(dn, "expected a declaration object");
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
      //type_index = GET_INDEX_NODE(dn->type);
   }
   if(Type->get_kind() == array_type_K)
      return true;
   return false;
}

bool tree_helper::is_a_pointer(const tree_managerConstRef  TM, const unsigned int index)
{
   long long int vec_size;
   bool ret;
   bool is_a_function;
   tree_helper::get_type_index(TM, index, vec_size, ret, is_a_function);
   return ret;
}

bool tree_helper::is_a_function(const tree_managerConstRef  TM, const unsigned int index)
{
   long long int vec_size;
   bool is_a_pointer;
   bool ret;
   tree_helper::get_type_index(TM, index, vec_size, is_a_pointer, ret);
   return ret;
}

bool tree_helper::is_a_vector(const tree_managerConstRef TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(T, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
   }
   if(Type->get_kind() == vector_type_K)
      return true;
   return false;
}

bool tree_helper::is_an_addr_expr(const tree_managerConstRef TM, const unsigned int index)
{
   return (TM->get_tree_node_const(index)->get_kind() == addr_expr_K);
}


bool tree_helper::HasToBeDeclared(const tree_managerConstRef TM, const unsigned int index)
{
   const tree_nodeRef type = TM->get_tree_node_const(index);
   THROW_ASSERT(GetPointer<type_node>(type), "Tree node " + boost::lexical_cast<std::string>(index) + " is not a type_node but " + type->get_kind_text());
   if(GetPointer<type_node>(type)->name)
   {
      tree_nodeRef name = GET_NODE(GetPointer<type_node>(type)->name);
      if(name->get_kind() == type_decl_K)
      {
         type_decl * td = GetPointer<type_decl>(name);
         if(td->include_name == "<built-in>")
         {
            return false;
         }
         if(GetPointer<complex_type>(type))
         {
            std::string name1 = tree_helper::name_type(TM, index);
            std::vector<std::string> splitted;
            boost::algorithm::split(splitted, name1, boost::algorithm::is_any_of(" "));
            if(splitted.size() > 1 and (splitted[0] == "_Complex" or splitted[0] == "__complex__" or splitted[0] == "complex"))
            {
               return false;
            }
         }
         return true;
      }
   }
   if(type->get_kind() == record_type_K or type->get_kind() == union_type_K or type->get_kind() == enumeral_type_K)
   {
      return true;
   }
   return false;
}

bool tree_helper::is_function_type(const tree_managerConstRef TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
   }
   if(GetPointer<function_type>(Type))
   {
      return true;
   }
   return false;
}

bool tree_helper::is_function_pointer_type(const tree_managerConstRef TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
   }
   if(GetPointer<pointer_type>(Type))
   {
      Type = GET_NODE(GetPointer<pointer_type>(Type)->ptd);
      if(GetPointer<function_type>(Type))
         return true;
   }
   return false;
}

bool tree_helper::is_bool(const tree_managerConstRef  TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(T, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      //decl_node * dn = GetPointer<decl_node>(T);
      //THROW_ASSERT(dn, "expected a declaration object");
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
      //type_index = GET_INDEX_NODE(dn->type);
   }
   if(GetPointer<boolean_type>(Type))
      return true;
   std::string type_name = name_type(TM, type_index);
   if (type_name == "_Bool" ||
       type_name == "bool" ||
       type_name == "sc_logic" ||
       type_name == "sc_in_resolved" ||
       type_name == "sc_inout_resolved" ||
       type_name == "sc_out_resolved" ||
       type_name == "sc_in_clk" ||
       type_name == "sc_inout_clk" ||
       type_name == "sc_out_clk" ||
       type_name == "sc_bit" ||
       type_name == "sc_clock"
      )
      return true;
   else
      return false;
}

bool tree_helper::is_int(const tree_managerConstRef  TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(T, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      //decl_node * dn = GetPointer<decl_node>(T);
      //THROW_ASSERT(dn, "expected a declaration object");
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
      //type_index = GET_INDEX_NODE(dn->type);
   }
   if(GetPointer<enumeral_type>(Type) && !GetPointer<enumeral_type>(Type)->unsigned_flag)
      return true;
   if(GetPointer<integer_type>(Type) and not GetPointer<integer_type>(Type)->unsigned_flag)
      return true;
   std::string type_name = name_type(TM, type_index);
   if (type_name == std::string("sc_int"))
      return true;
   return false;
}

bool tree_helper::is_real(const tree_managerConstRef TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(T, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      //decl_node * dn = GetPointer<decl_node>(T);
      //THROW_ASSERT(dn, "expected a declaration object");
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
      //type_index = GET_INDEX_NODE(dn->type);
   }
   if(GetPointer<real_type>(Type))
      return true;
   else
      return false;
}

bool tree_helper::is_unsigned(const tree_managerConstRef  TM, const unsigned int index)
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->is_unsigned " + boost::lexical_cast<std::string>(index));
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(T, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      //decl_node * dn = GetPointer<decl_node>(T);
      //THROW_ASSERT(dn, "expected a declaration object");
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
      //type_index = GET_INDEX_NODE(dn->type);
   }
   if(GetPointer<enumeral_type>(Type) and GetPointer<enumeral_type>(Type)->unsigned_flag)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--yes");
      return true;
   }

   if(GetPointer<complex_type>(Type) and GetPointer<complex_type>(Type)->unsigned_flag)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--yes");
      return true;
   }

   if(GetPointer<integer_type>(Type) and GetPointer<integer_type>(Type)->unsigned_flag)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--yes");
      return true;
   }

   std::string type_name = name_type(TM, type_index);
   if (type_name == std::string("unsigned") ||
       type_name.find(" unsigned") != std::string::npos ||
       type_name.find("unsigned ") != std::string::npos ||
       type_name ==  std::string("sc_uint") ||
       type_name ==  std::string("sc_lv") ||
       type_name ==  std::string("sc_in_rv") ||
       type_name ==  std::string("sc_out_rv") ||
       type_name ==  std::string("sc_inout_rv") ||
       type_name ==  std::string("sc_bv") ||
       type_name ==  std::string("sc_signal_rv")
      )
      return true;
   if(GetPointer<array_type>(Type))
   {
      const bool return_value = is_unsigned(TM, GetElements(TM, index));
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--It depends on element");
      return return_value;
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--no");
   return false;
}

bool tree_helper::is_module(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   const std::string mod_st = "sc_module";
   const std::string mod_name_st = "sc_module_name";
   const std::string ifc_st = "sc_interface";
   const record_type* rt = GetPointer<record_type>(TM->get_tree_node_const(index));
   if (rt && rt->binf)
   {
      binfo* bi = GetPointer<binfo>(GET_NODE(rt->binf));
      if (bi && look_for_binfo_inheritance(bi, mod_st) && !look_for_binfo_inheritance(bi, mod_name_st) && !look_for_binfo_inheritance(bi, ifc_st))
         return true;
   }
   return false;
}

bool tree_helper::is_builtin_channel(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(GetPointer<record_type>(TM->get_tree_node_const(index)), "a record type is expected");
   std::string rec_name = tree_helper::record_name(TM, index);
   if(rec_name == "sc_fifo" ||
      rec_name == "tlm_fifo" ||
      rec_name == "sc_mutex" ||
      rec_name == "sc_semaphore"
     )
      return true;
   else
      return false;
}

bool tree_helper::is_channel(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   const std::string mod_st = "sc_module";
   const std::string ifc_st = "sc_interface";
   const record_type* rt = GetPointer<record_type>(TM->get_tree_node_const(index));
   if (rt && rt->binf)
   {
      binfo* bi = GetPointer<binfo>(GET_NODE(rt->binf));
      if (bi && look_for_binfo_inheritance(bi, mod_st) && look_for_binfo_inheritance(bi, ifc_st))
         return true;
      else
         return is_builtin_channel(TM, index);
   }
   else if(rt)
      return is_builtin_channel(TM, index);
   return false;
}

#if 0
rt->get_maybe_name().find("sc_signal<") == 0 ||
                                rt->get_maybe_name().find("sc_signal_resolved") == 0  ||  // inherit from sc_signal
                                rt->get_maybe_name().find("sc_signal_rv<") == 0  || // inherit from sc_signal
                                rt->get_maybe_name().find("sc_buffer") == 0)  // inherit from sc_signal
#endif
bool tree_helper::is_signal(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   const std::string ifc_st = "sc_interface";
   const std::string pch_st = "sc_prim_channel";
   const std::string sig_st = "sc_signal";
   const std::string clock_st = "sc_clock";
   const record_type* rt = GetPointer<record_type>(TM->get_tree_node_const(index));
   if (rt && rt->binf)
   {
      binfo* bi = GetPointer<binfo>(GET_NODE(rt->binf));
      if (bi &&
          look_for_binfo_inheritance(bi, pch_st) &&
          (
                (look_for_binfo_inheritance(bi, sig_st) && look_for_binfo_inheritance(bi, ifc_st) )
          ) &&
          rt->get_maybe_name() != clock_st
         )
         return true;
   }
   return false;
}

bool tree_helper::is_clock(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   const std::string ifc_st = "sc_interface";
   const std::string pch_st = "sc_prim_channel";
   const std::string clock_st = "sc_clock";
   const record_type* rt = GetPointer<record_type>(TM->get_tree_node_const(index));
   if (rt && rt->binf)
   {
      binfo* bi = GetPointer<binfo>(GET_NODE(rt->binf));
      if (bi && (tree_helper::look_for_binfo_inheritance(bi, ifc_st) &&
          tree_helper::look_for_binfo_inheritance(bi, pch_st) &&
          rt->get_maybe_name() == clock_st
                ))
         return true;
   }
   return false;
}

bool tree_helper::is_SC_BIND_PROXY_NIL(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   tree_nodeRef curr_tn = TM->get_tree_node_const(index);
   if(curr_tn->get_kind() == addr_expr_K)
   {
      curr_tn = GET_NODE(GetPointer<addr_expr>(curr_tn)->op);
   }
   if(curr_tn->get_kind() == var_decl_K)
   {
      var_decl * vd = GetPointer<var_decl>(curr_tn);
      if(vd->name && GET_NODE(vd->name)->get_kind() == identifier_node_K)
      {
         identifier_node * id = GetPointer<identifier_node>(GET_NODE(vd->name));
         std::string strg = id->strg;
         return strg.find("SC_BIND_PROXY_NIL") != std::string::npos;
      }
   }
   return false;
}


#if 0
rt->get_maybe_name().find("sc_in<") == 0 || //inherit from sc_port
                                rt->get_maybe_name().find("sc_in_rv<") == 0 || //inherit from sc_in
                                rt->get_maybe_name().find("sc_in_resolved") == 0 || //inherit from sc_in
                                rt->get_maybe_name().find("sc_in_clk") == 0 || // equal to sc_in<bool>
                                rt->get_maybe_name().find("sc_out<") == 0 || //inherit from sc_port
                                rt->get_maybe_name().find("sc_out_rv<") == 0 || //inherit from sc_inout_rv
                                rt->get_maybe_name().find("sc_out_resolved") == 0 || //inherit from sc_inout_resolved
                                rt->get_maybe_name().find("sc_out_clk") == 0 || //equal to sc_out<bool>
                                rt->get_maybe_name().find("sc_inout<") == 0 || //inherit from sc_port
                                rt->get_maybe_name().find("sc_inout_rv<") == 0 || //inherit from sc_inout
                                rt->get_maybe_name().find("sc_inout_resolved") == 0 || //inherit from sc_inout
                                rt->get_maybe_name().find("sc_inout_clk") == 0 || //equal to sc_inout<bool>
                                rt->get_maybe_name().find("sc_fifo_in<") == 0 || //inherit from sc_port
                                rt->get_maybe_name().find("sc_fifo_out<") == 0 //inherit from sc_port
#endif
bool tree_helper::is_port(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   const std::string port_st = "sc_port";
   const std::string sc_export_st = "sc_export";
   const record_type* rt = GetPointer<record_type>(TM->get_tree_node_const(index));
   if (rt && rt->binf)
   {
      binfo* bi = GetPointer<binfo>(GET_NODE(rt->binf));
      if (bi && (look_for_binfo_inheritance(bi, port_st) ||
          rt->get_maybe_name().find(sc_export_st) == 0
                ))
         return true;
   }
   return false;
}

bool tree_helper::is_in_port(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   const std::string sc_in_st = "sc_in";
   const record_type* rt = GetPointer<record_type>(TM->get_tree_node_const(index));
   if (rt && rt->binf)
   {
      binfo* bi = GetPointer<binfo>(GET_NODE(rt->binf));
      if (bi && look_for_binfo_inheritance(bi, sc_in_st))
         return true;
   }
   return false;
}

bool tree_helper::is_out_port(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   const std::string sc_out_st = "sc_out"; //several out port are actually inout port (e.g., sc_out_resolved and sc_out_rv)
   const record_type* rt = GetPointer<record_type>(TM->get_tree_node_const(index));
   if (rt && rt->binf)
   {
      binfo* bi = GetPointer<binfo>(GET_NODE(rt->binf));
      if (bi && look_for_binfo_inheritance(bi, sc_out_st))
         return true;
   }
   return false;
}

bool tree_helper::is_inout_port(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   const std::string sc_inout_st = "sc_inout";
   const record_type* rt = GetPointer<record_type>(TM->get_tree_node_const(index));
   if (rt && rt->binf)
   {
      binfo* bi = GetPointer<binfo>(GET_NODE(rt->binf));
      if (bi && look_for_binfo_inheritance(bi, sc_inout_st))
         return true;
   }
   return false;
}

bool tree_helper::is_event(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   const std::string event_st = "sc_event";
   const record_type* rt = GetPointer<record_type>(TM->get_tree_node_const(index));
   if (rt && rt->binf)
   {
      binfo* bi = GetPointer<binfo>(GET_NODE(rt->binf));
      if (bi && look_for_binfo_inheritance(bi, event_st))
         return true;
   }
   return false;
}

expand::serialization expand::get_serialization(std::string name) const
{
   if(no_serialize.find(name) != no_serialize.end())
      return none;
   if(internal_serialize.find(name) != internal_serialize.end())
      return internal;
   return total;
}

bool expand::is_transparent(std::string name) const
{
   return(transparent.find(name) != transparent.end());
}

void expand::check_lib_type(const tree_nodeRef var)
{
   tree_nodeRef curr_tn;
   if (var->get_kind() == tree_reindex_K)
   {
      curr_tn = GET_NODE(var);
   }
   else
   {
      curr_tn = var;
   }
   THROW_ASSERT(GetPointer<decl_node>(curr_tn), "Checking type of not a decl_node");
   decl_node * dn = GetPointer<decl_node>(curr_tn);
   std::string include_name = dn->include_name;
   std::set<std::string>::iterator it_end = headers.end();
   for(std::set<std::string>::iterator it = headers.begin(); it != it_end; it++)
   {
      if(include_name.find(*it) != std::string::npos && dn->type)
      {
         if(GetPointer<type_node>(GET_NODE(dn->type)))
         {
            lib_types.insert(GET_NODE(dn->type));
            type_node * tn = GetPointer<type_node>(GET_NODE(dn->type));
            if(tn && tn->unql)
               lib_types.insert(GET_NODE(tn->unql));
         }
      }
   }
}

bool expand::operator() (const tree_nodeRef &tn) const
{
   tree_nodeRef curr_tn;
   if (tn->get_kind() == tree_reindex_K)
   {
      curr_tn = GET_NODE(tn);
   }
   else
   {
      curr_tn = tn;
   }
   THROW_ASSERT(GetPointer<type_node>(curr_tn) || GetPointer<function_decl>(curr_tn), "tn is not a node of type type_node nor function_decl");
   if(lib_types.find(curr_tn) != lib_types.end())
      return false;
   if(curr_tn->get_kind() == record_type_K)
   {
      record_type * rt = GetPointer<record_type>(curr_tn);
      if(rt->ptrmem_flag)
         return false;
   }
   type_node * type = GetPointer<type_node>(curr_tn);
   if(type && type->name)
   {
      type_decl * td = GetPointer<type_decl>(GET_NODE(type->name));
      if(td)
      {
         std::string include_name = td->include_name;
         std::set<std::string>::iterator it_end = headers.end();
         for(std::set<std::string>::iterator it = headers.begin(); it != it_end; it++)
            if(include_name.find(*it) != std::string::npos)
            {
               return false;
            }
      }
   }
   return true;
}

expand::expand()
{
   internal_serialize.insert("printf");
   transparent.insert("__builtin_va_start");
   headers.insert("stdio.h");
}

bool SC_expand::namespace_check(const tree_nodeRef &tn) const
{
   tree_nodeRef curr_tn;
   if (tn->get_kind() == tree_reindex_K)
      curr_tn = GET_NODE(tn);
   else
      curr_tn = tn;
   record_type * rt;
   function_decl* fd = GetPointer<function_decl>(curr_tn);
   decl_node * dn = GetPointer<decl_node>(curr_tn);
   if (fd && fd->scpe)
   {
      rt = GetPointer<record_type>(GET_NODE(fd->scpe));
      namespace_decl* nd = GetPointer<namespace_decl>(GET_NODE(fd->scpe));
      if (nd)
      {
         identifier_node* in = GetPointer<identifier_node>(GET_NODE(nd->name));
         std::string namespace_string = in->strg;
         return expand_ns.find(namespace_string) == expand_ns.end();
      }
   }
   else if (dn && dn->type)
   {
      rt = GetPointer<record_type>(GET_NODE(dn->type));
   }
   else
   {
      rt = GetPointer<record_type>(curr_tn);
   }
   if (rt && rt->name)
   {
      type_decl* td = GetPointer<type_decl>(GET_NODE(rt->name));
      if (td && td->scpe)
      {
         namespace_decl* nd = GetPointer<namespace_decl>(GET_NODE(td->scpe));
         if (nd)
         {
            identifier_node* in = GetPointer<identifier_node>(GET_NODE(nd->name));
            std::string namespace_string = in->strg;
            return expand_ns.find(namespace_string) == expand_ns.end();
         }
      }
   }
   return true;
}

bool SC_expand::member_check(const tree_nodeRef &curr_tn) const
{
   bool expand_p = true;
   function_decl* fd = GetPointer<function_decl>(curr_tn);
   if (fd && fd->is_destructor())
      return false;
   if (fd && fd->scpe)
   {
      record_type* rt = GetPointer<record_type>(GET_NODE(fd->scpe));
      if (rt)
      {
         if (fd->is_constructor() && !tree_helper::look_for_binfo_inheritance(GetPointer<binfo>(GET_NODE(rt->binf)), "sc_module"))
            return false;
         type_decl* td = GetPointer<type_decl>(GET_NODE(rt->name));
         if (td)
         {
            identifier_node* in = GetPointer<identifier_node>(GET_NODE(td->name));
            std::string function_name_string = in->strg;

            for (std::set<std::string>::const_iterator i = expand_data.begin(); i != expand_data.end() && expand_p; ++i)
            {
               std::string::size_type j = function_name_string.find(*i);
               if (j != std::string::npos)
                  expand_p = false;
            }
         }

      }
   }
   return expand_p;
}

bool SC_expand::check_field(const tree_nodeRef &curr_tn) const
{
   bool expand_p = true;
   field_decl* fd = GetPointer<field_decl>(curr_tn);
   if (fd && fd->scpe)
   {
      record_type* rt = GetPointer<record_type>(GET_NODE(fd->scpe));
      if (rt)
      {
         type_decl* td = GetPointer<type_decl>(GET_NODE(rt->name));
         if (td)
         {
            identifier_node* in = GetPointer<identifier_node>(GET_NODE(td->name));
            std::string field_name_string = in->strg;

            for (std::set<std::string>::const_iterator i = expand_ns.begin(); i != expand_ns.end() && expand_p; ++i)
            {
               std::string::size_type j = field_name_string.find(*i);
               if (j == 0)//check for prefix
                  expand_p = false;
            }
         }

      }
   }
   return expand_p;
}


bool SC_expand::operator() (const tree_nodeRef &t) const
{
   if(t->get_kind() == function_decl_K)
   {
      function_decl * fd = GetPointer<function_decl>(t);
      if(fd->is_constructor())
         return namespace_check(t) && member_check(t) && expand::operator()(t);
      else
         return false;
   }
   return namespace_check(t) && member_check(t) && expand::operator()(t);
}

SC_expand::SC_expand()
{
   expand_ns.insert("sc_core");
   expand_ns.insert("sc_dt");
   expand_ns.insert("std");
   expand_ns.insert("tlm");
   expand_data.insert("sc_port_b"); //check if really needed
   expand_data.insert("sc_in"); //check if really needed
   expand_data.insert("sc_inout"); //check if really needed
   expand_data.insert("sc_port_base"); //check if really needed
   expand_data.insert("basic_ostream"); //check if really needed
   expand_data.insert("facet"); //check if really needed
   expand_data.insert("initialize_once"); //check if really needed
   expand_data.insert("router_port");
   expand_data.insert("tlm_blocking_put_if_probe");
   expand_data.insert("tlm_nonblocking_put_if_probe");
   expand_data.insert("tlm_transport_if_probe");
}

bool tree_helper::is_a_variable(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   tree_nodeRef node = TM->get_tree_node_const(index);
   switch(node->get_kind())
   {
      case integer_cst_K:
      case real_cst_K:
      case complex_cst_K:
      case vector_cst_K:
      case ssa_name_K:
      case function_decl_K:
      case var_decl_K:
      case parm_decl_K:
      case label_decl_K:
      case result_decl_K:
      case call_expr_K:
        return true;
      /// The following one are not considered as variables, but as operations on the variables
      case addr_expr_K:
      case string_cst_K:
      case indirect_ref_K:
      case component_ref_K:
      case array_ref_K:
      case struct_field_tag_K:
      case symbol_memory_tag_K:
      case memory_partition_tag_K:
      case name_memory_tag_K:
      case nop_expr_K:
      case exc_ptr_expr_K:
      case convert_expr_K:
      case float_expr_K:
      case CASE_BINARY_EXPRESSION:
         return false;
      default:
         THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "tree_helper::is_a_variable - variable is not supported: " + std::string(node->get_kind_text()));
   }
   return true;
}

unsigned int tree_helper::get_parameters_number(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   tree_nodeRef node = TM->get_tree_node_const(index);
   switch(node->get_kind())
   {
      case ssa_name_K:
      {
         ssa_name * sn = GetPointer<ssa_name>(node);
         return get_parameters_number(TM, GET_INDEX_NODE(sn->var));
      }
      case var_decl_K:
      {
         var_decl * vd = GetPointer<var_decl>(node);
         return get_parameters_number(TM, GET_INDEX_NODE(vd->type));
      }
      case parm_decl_K:
      {
         parm_decl * pd = GetPointer<parm_decl>(node);
         return get_parameters_number(TM, GET_INDEX_NODE(pd->type));
      }
      case pointer_type_K:
      {
         pointer_type * pt = GetPointer<pointer_type>(node);
         return get_parameters_number(TM, GET_INDEX_NODE(pt->ptd));
      }
      case function_decl_K:
      {
         if(TM->get_implementation_node(index))
            node = TM->get_tree_node_const(TM->get_implementation_node(index));
         function_decl * fd = GetPointer<function_decl>(node);
         return get_parameters_number(TM, GET_INDEX_NODE(fd->type));
      }
      case method_type_K:
      case function_type_K:
      {
         function_type * ft = GetPointer<function_type>(node);
         if(ft->prms)
            return get_parameters_number(TM, GET_INDEX_NODE(ft->prms));
         else
            return 0;
      }
      case tree_list_K:
      {
         tree_list * tl = GetPointer<tree_list>(node);
         if(tl->chan)
            return get_parameters_number(TM, GET_INDEX_NODE(tl->chan)) + 1;
         else
            return 1;
      }
      default:
         THROW_ERROR("Searching for number of parameters not in a function. Node id is " + boost::lexical_cast<std::string>(index));
   }
   return 0;
}

unsigned int tree_helper::get_base_index(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   tree_nodeRef node = TM->get_tree_node_const(index);
   switch(node->get_kind())
   {
      case ssa_name_K:
      {
         ssa_name * sn = GetPointer<ssa_name>(node);
         return get_base_index(TM, GET_INDEX_NODE(sn->var));
      }
      case result_decl_K:
      case parm_decl_K:
      case var_decl_K:
      case struct_field_tag_K:
      case memory_partition_tag_K:
      case symbol_memory_tag_K:
      case name_memory_tag_K:
      case string_cst_K:
      {
         return index;
      }
      case indirect_ref_K:
      {
         indirect_ref * ir = GetPointer<indirect_ref>(node);
         return get_base_index(TM, GET_INDEX_NODE(ir->op));
      }
      case mem_ref_K:
      {
         mem_ref * mr = GetPointer<mem_ref>(node);
         return get_base_index(TM, GET_INDEX_NODE(mr->op0));
      }
      case array_ref_K:
      {
         array_ref * ar = GetPointer<array_ref>(node);
         return get_base_index(TM, GET_INDEX_NODE(ar->op0));
      }
      case component_ref_K:
      {
         component_ref * cr = GetPointer<component_ref>(node);
         return get_base_index(TM, GET_INDEX_NODE(cr->op0));
      }
      case realpart_expr_K:
      {
         realpart_expr * rpe = GetPointer<realpart_expr>(node);
         return get_base_index(TM, GET_INDEX_NODE(rpe->op));
      }
      case imagpart_expr_K:
      {
         imagpart_expr * rpe = GetPointer<imagpart_expr>(node);
         return get_base_index(TM, GET_INDEX_NODE(rpe->op));
      }
      case bit_field_ref_K:
      {
         bit_field_ref * bfr = GetPointer<bit_field_ref>(node);
         return get_base_index(TM, GET_INDEX_NODE(bfr->op0));
      }
      case target_mem_ref_K:
      {
         target_mem_ref * tmr = GetPointer<target_mem_ref>(node);
         if(tmr->symbol)
            return get_base_index(TM, GET_INDEX_NODE(tmr->symbol));
         else if(tmr->base)
            return get_base_index(TM, GET_INDEX_NODE(tmr->base));
         else if(tmr->idx)
            return get_base_index(TM, GET_INDEX_NODE(tmr->idx));
         else
         {
            THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "tree_helper::get_base_index::target_mem_ref_K - variable type pattern not supported: " + STR(index));
         }
      }
      case target_mem_ref461_K:
      {
         target_mem_ref461 * tmr = GetPointer<target_mem_ref461>(node);
         if(tmr->base)
            return get_base_index(TM, GET_INDEX_NODE(tmr->base));
         else if(tmr->idx)
            return get_base_index(TM, GET_INDEX_NODE(tmr->idx));
         else if(tmr->idx2)
            return get_base_index(TM, GET_INDEX_NODE(tmr->idx2));
         else
         {
            THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "tree_helper::get_base_index::target_mem_ref461_K - variable type pattern not supported: " + STR(index));
         }
      }
      case addr_expr_K:
      {
         addr_expr * ae = GetPointer<addr_expr>(node);
         tree_nodeRef addr_expr_op = GET_NODE(ae->op);

         switch (addr_expr_op->get_kind())
         {
            case ssa_name_K:
            case var_decl_K:
            case parm_decl_K:
            {
               return GET_INDEX_NODE(ae->op);
               break;
            }
            case array_ref_K:
            {
               array_ref* ar = GetPointer<array_ref>(addr_expr_op);
               tree_nodeRef idx = GET_NODE(ar->op1);
               if(idx->get_kind() == integer_cst_K and get_integer_cst_value(GetPointer<integer_cst>(idx)) == 0)
               {
                  switch (GET_NODE(ar->op0)->get_kind())
                  {
                     case ssa_name_K:
                     case var_decl_K:
                     case parm_decl_K:
                     case string_cst_K:
                     {
                         return GET_INDEX_NODE(ar->op0);
                        break;
                     }
                     default:
                        THROW_ERROR("addr_expr-array_ref[0] pattern not supported: " + std::string(addr_expr_op->get_kind_text())+ " @" + STR(index));
                  }
               }
               else
                  return index;
               break;
            }
            default:
               THROW_ERROR("addr_expr pattern not supported: " + std::string(addr_expr_op->get_kind_text())+ " @" + STR(index));
         }
         break;
      }
      default:
         THROW_ERROR_CODE(NODE_NOT_YET_SUPPORTED_EC, "tree_helper::get_base_index - variable type is not supported: " + STR(index) + "-" +std::string(node->get_kind_text()));
   }
   return 0;
}

bool tree_helper::is_volatile(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   tree_nodeRef node = TM->get_tree_node_const(index);
   ssa_name *sa = GetPointer<ssa_name>(node);
   if (!sa)
   {
      //even a variable maybe volatile
      var_decl * vd = GetPointer<var_decl>(node);
      if(!vd)
         return false;
      tree_nodeRef n =  get_type_node(node);
      type_node * tn = GetPointer<type_node>(n);
      return tn->qual == treeVocabularyTokenTypes::TOK_QUAL_V || tn->qual == treeVocabularyTokenTypes::TOK_QUAL_VR || tn->qual == treeVocabularyTokenTypes::TOK_QUAL_CVR;
   }
   return sa->volatile_flag;
}

bool tree_helper::is_parameter(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   tree_nodeRef node = TM->get_tree_node_const(index);
   if(GetPointer<parm_decl>(node)) return true;
   ssa_name *sn = GetPointer<ssa_name>(node);
   if(!sn) return false;
   return GET_NODE(sn->def_stmts.front())->get_kind() == gimple_nop_K && GET_NODE(sn->var)->get_kind() == parm_decl_K;
}

bool tree_helper::is_ssa_name(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   tree_nodeRef node = TM->get_tree_node_const(index);
   ssa_name *sn = GetPointer<ssa_name>(node);
   return sn ? true : false;
}
bool tree_helper::is_virtual(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   tree_nodeRef node = TM->get_tree_node_const(index);
   ssa_name *sn = GetPointer<ssa_name>(node);
   if(!sn) return false;
   return sn->virtual_flag;
}

unsigned int tree_helper::get_bb_edge_def(const tree_managerConstRef TM, const unsigned int def, unsigned int i)
{
   THROW_ASSERT(def > 0, "expected positive non zero numbers");
   tree_nodeRef node = TM->get_tree_node_const(def);
   ssa_name *sn = GetPointer<ssa_name>(node);
   THROW_ASSERT(sn, "expected a ssa_name");
   THROW_ASSERT(sn->def_stmts.size() == 1, "expected a single statement defining the variable");
   tree_nodeRef def_stmt = GET_NODE(sn->def_stmts.front());
   gimple_phi* gp = GetPointer<gimple_phi>(def_stmt);
   THROW_ASSERT(gp, "expected a gimple_phi");
   THROW_ASSERT(i < gp->list_of_def_edge.size(), "incorrect index");
   return gp->list_of_def_edge[i].second;
}

unsigned int tree_helper::get_bb_edge(const tree_managerConstRef TM, const unsigned int phi, unsigned int i)
{
   THROW_ASSERT(phi > 0, "expected positive non zero numbers");
   tree_nodeRef node = TM->get_tree_node_const(phi);
   gimple_phi* gp = GetPointer<gimple_phi>(node);
   THROW_ASSERT(gp, "expected a gimple_phi @" + STR(phi));
   THROW_ASSERT(i < gp->list_of_def_edge.size(), "incorrect index");
   return gp->list_of_def_edge[i].second;
}

bool tree_helper::is_static(const tree_managerConstRef TM, const unsigned int index)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   tree_nodeRef node = TM->get_tree_node_const(index);
   var_decl *vd = GetPointer<var_decl>(node);
   if (!vd)
      return false;
   return vd->static_flag;
}

bool tree_helper::is_const_type(const tree_managerConstRef TM, const unsigned int index)
{
   const tree_nodeRef T = TM->get_tree_node_const(index);
   THROW_ASSERT(T, "this index does not exist: " + boost::lexical_cast<std::string>(index));
   tree_nodeRef Type;
   unsigned int type_index;
   if(GetPointer<type_node>(T))
   {
      type_index = index;
      Type = T;
   }
   else
   {
      //decl_node * dn = GetPointer<decl_node>(T);
      //THROW_ASSERT(dn, "expected a declaration object");
      type_index = 0;
      Type = get_type_node(T, type_index );
      THROW_ASSERT(type_index > 0, "expected a type index");
      //type_index = GET_INDEX_NODE(dn->type);
   }
   unsigned int quals = GetPointer<type_node>(Type)->qual;
   if(quals > 0 && (quals == treeVocabularyTokenTypes::TOK_QUAL_C || quals == treeVocabularyTokenTypes::TOK_QUAL_CV || quals == treeVocabularyTokenTypes::TOK_QUAL_CVR))
      return true;
   else
      return false;
}

std::string tree_helper::return_qualifier_prefix(unsigned int quals)
{
   switch(quals)
   {
      case treeVocabularyTokenTypes::TOK_QUAL_R:
         return "r_";
      case treeVocabularyTokenTypes::TOK_QUAL_V:
         return "v_";
      case treeVocabularyTokenTypes::TOK_QUAL_VR:
         return "vr_";
      case treeVocabularyTokenTypes::TOK_QUAL_C:
         return "c_";
      case treeVocabularyTokenTypes::TOK_QUAL_CR:
         return "cr_";
      case treeVocabularyTokenTypes::TOK_QUAL_CV:
         return "cv_";
      case treeVocabularyTokenTypes::TOK_QUAL_CVR:
         return "cvr_";
      default:
         THROW_ERROR("not supported qualifier "+boost::lexical_cast<std::string>(quals));
      return "";
   }
}

std::string tree_helper::return_C_qualifiers(unsigned int quals, bool real_const)
{
   switch(quals)
   {
      case treeVocabularyTokenTypes::TOK_QUAL_R:
         return "__restrict__ ";
      case treeVocabularyTokenTypes::TOK_QUAL_V:
         return "volatile ";
      case treeVocabularyTokenTypes::TOK_QUAL_VR:
         return "volatile __restrict__ ";
      case treeVocabularyTokenTypes::TOK_QUAL_C:
         return real_const ? "const " : "/*const*/ ";
      case treeVocabularyTokenTypes::TOK_QUAL_CR:
         return (real_const ? "const" : "/*const*/") + std::string(" __restrict__ ");
      case treeVocabularyTokenTypes::TOK_QUAL_CV:
         return (real_const ? "const" : "/*const*/") + std::string(" volatile ");
      case treeVocabularyTokenTypes::TOK_QUAL_CVR:
         return (real_const ? "const" : "/*const*/") + std::string(" volatile __restrict__ ");
      default:
         THROW_ERROR("not supported qualifier "+boost::lexical_cast<std::string>(quals));
         return "";
   }
}

long long tree_helper::get_integer_cst_value(const integer_cst * ic)
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Getting integer const value");
   tree_nodeRef type = GET_NODE(ic->type);
#ifndef NDEBUG
   integer_type* it = GetPointer<integer_type>(type);
   THROW_ASSERT(it || type->get_kind() == pointer_type_K || type->get_kind() == boolean_type_K || type->get_kind()
         == enumeral_type_K, "Expected a integer_type, pointer_type, boolean_type, enumeral_type "
         + boost::lexical_cast<std::string>(GET_INDEX_NODE(ic->type)));
#endif
   long long result = 0;
   ///If high is null or if high is a sign extension or low is enough to express an int (64 bit system)

   result = ic->value;
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Constant is " + boost::lexical_cast<std::string>(result));
   return result;
}


unsigned int tree_helper::get_object_size(const tree_nodeRef &t)
{
   switch (t->get_kind())
   {
      case record_type_K:
      case union_type_K:
      case integer_type_K:
      case real_type_K:
      case pointer_type_K:
      case function_type_K:
      case array_type_K:
      case complex_type_K:
      case vector_type_K:
      {
         type_node * tn = GetPointer<type_node>(t);
         integer_cst * ic = GetPointer<integer_cst>(GET_NODE(tn->size));
         THROW_ASSERT(ic, "expected integer_cst object");
         long long int result = get_integer_cst_value(ic);
         return static_cast<unsigned int>(result);
      }
      case void_type_K:
      {
         THROW_ERROR(std::string("tree_helper::get_object_size: void type does not have a size "));
         return 0;
      }
      default:
         THROW_ERROR(std::string("tree_helper::get_object_size: Unexpected type pattern ") + t->get_kind_text());
         return 0;
   }
   return 0;
}

unsigned int tree_helper::get_array_var(const tree_managerConstRef  TM, const unsigned int index, bool is_written, bool &two_dim_p)
{
   THROW_ASSERT(index > 0, "expected positive non zero numbers");
   tree_nodeRef node = TM->get_tree_node_const(index);
   gimple_assign *gms = GetPointer<gimple_assign>(node);
   array_ref * ar;
   if(is_written)
      ar = GetPointer<array_ref>(GET_NODE(gms->op0));
   else
      ar = GetPointer<array_ref>(GET_NODE(gms->op1));
   two_dim_p = false;
   if(ar)
   {
      if(GetPointer<array_ref>(GET_NODE(ar->op0)))
      {
         ar = GetPointer<array_ref>(GET_NODE(ar->op0));
         if(GetPointer<array_ref>(GET_NODE(ar->op0)))
            THROW_ERROR("n-dimension array not yet supported (n > 2)");
         two_dim_p = true;
         return GET_INDEX_NODE(ar->op0);
      }
      else
         return GET_INDEX_NODE(ar->op0);
   }
   target_mem_ref * tmr;
   if(is_written)
      tmr = GetPointer<target_mem_ref>(GET_NODE(gms->op0));
   else
      tmr = GetPointer<target_mem_ref>(GET_NODE(gms->op1));
   if(tmr)
   {
      if(tmr->symbol)
         return GET_INDEX_NODE(tmr->symbol);
      else if(tmr->base)
         return GET_INDEX_NODE(tmr->base);
      else
         THROW_ERROR("Unexpected pattern");
   }
   target_mem_ref461 * tmr461;
   if(is_written)
      tmr461 = GetPointer<target_mem_ref461>(GET_NODE(gms->op0));
   else
      tmr461 = GetPointer<target_mem_ref461>(GET_NODE(gms->op1));
   if(tmr461)
   {
      if(tmr461->base)
         return GET_INDEX_NODE(tmr461->base);
      else
         THROW_ERROR("Unexpected pattern");
   }
   addr_expr * ae = GetPointer<addr_expr>(GET_NODE(gms->op1));
   if(ae)
   {
      ar = GetPointer<array_ref>(GET_NODE(ae->op));
      if(ar)
      {
         if(GetPointer<array_ref>(GET_NODE(ar->op0)))
         {
            ar = GetPointer<array_ref>(GET_NODE(ar->op0));
            if(GetPointer<array_ref>(GET_NODE(ar->op0)))
               THROW_ERROR("n-dimension array not yet supported (n > 2)");
            two_dim_p = true;
            return GET_INDEX_NODE(ar->op0);
         }
         else
            return GET_INDEX_NODE(ar->op0);
      }
      else
      {
         THROW_ERROR("Unexpected pattern " + boost::lexical_cast<std::string>(index));
         return 0;
      }
   }
   nop_expr * ne = GetPointer<nop_expr>(GET_NODE(gms->op1));
   if(ne)
   {
      ae = GetPointer<addr_expr>(GET_NODE(ne->op));
      if(ae)
      {
         ar = GetPointer<array_ref>(GET_NODE(ae->op));
         if(ar)
         {
            if(GetPointer<array_ref>(GET_NODE(ar->op0)))
            {
               ar = GetPointer<array_ref>(GET_NODE(ar->op0));
               if(GetPointer<array_ref>(GET_NODE(ar->op0)))
                  THROW_ERROR("n-dimension array not yet supported (n > 2)");
               two_dim_p = true;
               return GET_INDEX_NODE(ar->op0);
            }
            else
               return GET_INDEX_NODE(ar->op0);
         }
         var_decl * vd = GetPointer<var_decl>(GET_NODE(ae->op));
         if(vd)
            return GET_INDEX_NODE(ae->op);
         else
         {
            THROW_ERROR("Unexpected pattern " + boost::lexical_cast<std::string>(index));
            return 0;
         }
      }
   }
   THROW_ERROR("Unexpected pattern " + boost::lexical_cast<std::string>(index));
   return 0;
}


bool tree_helper::is_constant(const tree_managerConstRef  TM, const unsigned int index)
{
   tree_nodeRef node = TM->get_tree_node_const(index);
   switch(node->get_kind())
   {
      case string_cst_K:
      case real_cst_K:
      case complex_cst_K:
      case vector_cst_K:
      case integer_cst_K:
         return true;
      default:
         return false;
   }
   return false;
}

std::string tree_helper::op_symbol(const tree_nodeRef &op)
{
   return op_symbol(op.get());
}

std::string tree_helper::op_symbol(const tree_node * op)
{
   THROW_ASSERT(op, "Null tree node");
   switch (op->get_kind())
   {
      case modify_expr_K:
         return "=";

      case truth_or_expr_K:
      case truth_orif_expr_K:
         return "||";

      case truth_and_expr_K:
      case truth_andif_expr_K:
         return "&&";

      case bit_ior_expr_K:
         return "|";

      case truth_xor_expr_K:
      case bit_xor_expr_K:
         return "^";

      case addr_expr_K:
      case bit_and_expr_K:
      {
         if (dynamic_cast<const addr_expr *>(op))
         {
            const addr_expr* ae = dynamic_cast<const addr_expr *>(op);
            tree_nodeRef tn = GET_NODE(ae->op);
            if (GetPointer<array_ref>(tn))
            {
               array_ref* ar = GetPointer<array_ref>(tn);
               tree_nodeRef base = GET_NODE(ar->op0);
               tree_nodeRef offset = GET_NODE(ar->op1);
               if (base->get_kind() == string_cst_K and offset->get_kind() == integer_cst_K and !tree_helper::get_integer_cst_value(GetPointer<integer_cst>(offset)))
                  return "";
            }
            if (GetPointer<string_cst>(tn))
               return "";
         }
         return "&";
      }

      case eq_expr_K:
         return "==";

      case ne_expr_K:
         return "!=";

      case lt_expr_K:
      case unlt_expr_K:
         return "<";

      case le_expr_K:
      case unle_expr_K:
         return "<=";

      case gt_expr_K:
      case ungt_expr_K:
         return ">";

      case ge_expr_K:
      case unge_expr_K:
         return ">=";

      case lshift_expr_K:
         return "<<";

      case rshift_expr_K:
         return ">>";

      case plus_expr_K:
      case pointer_plus_expr_K:
      case widen_sum_expr_K:
         return "+";

      case negate_expr_K:
      case minus_expr_K:
         return "-";

      case bit_not_expr_K:
         return "~";

      case truth_not_expr_K:
         return "!";

      case mult_expr_K:
      case indirect_ref_K:
      case widen_mult_expr_K:
         return "*";

      case trunc_div_expr_K:
      case ceil_div_expr_K:
      case floor_div_expr_K:
      case round_div_expr_K:
      case rdiv_expr_K:
      case exact_div_expr_K:
         return "/";

      case trunc_mod_expr_K:
      case ceil_mod_expr_K:
      case floor_mod_expr_K:
      case round_mod_expr_K:
         return "%";

      case predecrement_expr_K:
         return " --";

      case preincrement_expr_K:
         return " ++";

      case postdecrement_expr_K:
         return "-- ";

      case postincrement_expr_K:
         return "+=";

      case reference_expr_K:
         return "";
      case realpart_expr_K:
         return "__real__ ";
      case imagpart_expr_K:
         return "__imag__ ";
      case fix_trunc_expr_K:
      case min_expr_K:
         return "";
      default:
         THROW_ERROR(std::string("op_symbol not yet supported ") + op->get_kind_text());
         return "";
   }
}

void tree_helper::get_array_dimensions(const tree_managerConstRef TM, const unsigned int index, std::vector<unsigned int> &dims, unsigned int& elmt_bitsize)
{
   tree_nodeRef node = TM->get_tree_node_const(index);
   THROW_ASSERT(node->get_kind() == array_type_K, "array_type expected: @" + STR(index));
   array_type * at = GetPointer<array_type>(node);
   tree_nodeRef domn = GET_NODE(at->domn);
   THROW_ASSERT(domn->get_kind() == integer_type_K, "expected an integer type as domain");
   integer_type *it = GetPointer<integer_type>(domn);
   unsigned int min_value = 0;
   unsigned int max_value = 0;
   if(it->min)
      min_value = static_cast<unsigned int>(get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(it->min))));
   if(it->max)
      max_value = static_cast<unsigned int>(get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(it->max))));
   unsigned int range_domain = max_value - min_value + 1;
   dims.push_back(range_domain);
   THROW_ASSERT(at->elts, "elements type expected");
   tree_nodeRef elts = GET_NODE(at->elts);
   if(elts->get_kind() == array_type_K)
      get_array_dimensions(TM,GET_INDEX_NODE(at->elts),dims,elmt_bitsize);
   else
   {
      elmt_bitsize = size(TM, get_type_index(TM, GET_INDEX_NODE(at->elts)));
   }
}

void tree_helper::extract_array_indexes(const tree_managerConstRef TM, const unsigned int index, std::vector<unsigned int> &indexes, std::vector<unsigned int> &size_indexes, unsigned int &base_object)
{
   tree_nodeRef node = TM->get_tree_node_const(index);
   THROW_ASSERT(node->get_kind() == array_ref_K, "array_ref expected: @" + STR(index));
   array_ref * ar = GetPointer<array_ref>(node);
   base_object = GET_INDEX_NODE(ar->op0);
   if(GET_NODE(ar->op0)->get_kind() == array_ref_K)
   {
      array_ref * nested_ar = GetPointer<array_ref>(GET_NODE(ar->op0));
      array_type * at = GetPointer<array_type>(GET_NODE(nested_ar->type));
      tree_nodeRef domn = GET_NODE(at->domn);
      THROW_ASSERT(domn->get_kind() == integer_type_K, "expected an integer type as domain");
      integer_type *it = GetPointer<integer_type>(domn);
      unsigned int min_value = 0;
      unsigned int max_value = 0;
      if(it->min)
         min_value = static_cast<unsigned int>(get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(it->min))));
      if(it->max)
         max_value = static_cast<unsigned int>(get_integer_cst_value(GetPointer<integer_cst>(GET_NODE(it->max))));
      unsigned int range_domain = max_value - min_value + 1;
      size_indexes.push_back(range_domain);
      extract_array_indexes(TM, base_object, indexes, size_indexes, base_object);
      indexes.push_back(GET_INDEX_NODE(nested_ar->op1));
   }
}

unsigned int tree_helper::GetUnqualified(const tree_managerConstRef TM, unsigned int type)
{
   const tree_nodeRef typeNode = TM->get_tree_node_const(type);
   if (GetPointer<type_node>(typeNode) and GetPointer<type_node>(typeNode)->unql)
      return GET_INDEX_NODE(GetPointer<type_node>(typeNode)->unql);
   return 0;
}

bool tree_helper::IsAligned (const tree_managerConstRef TM, unsigned int type)
{
   const tree_nodeRef node = TM->get_tree_node_const(type);
   const type_node * tn = GetPointer<type_node>(node);
   THROW_ASSERT(tn, "Tree node " + boost::lexical_cast<std::string>(type) + " is of type " + node->get_kind_text()); 
   if(tn->unql and tn->algn != GetPointer<type_node>(GET_NODE(tn->unql))->algn)
      return true;
   else
      return false;
}

std::string tree_helper::normalized_ID(const std::string &id)
{
    std::string strg = id;
    for (unsigned int i = 0;i < strg.size();i++)
    {
        if (strg[i] == '$')
            strg[i] = '_';
        if (strg[i] == '.')
            strg[i] = '_';
    }
    return strg;
}

std::string tree_helper::print_type(const tree_managerConstRef TM, unsigned int original_type, bool global, bool print_qualifiers, bool print_storage, unsigned int var, const var_pp_functorConstRef vppf, const std::string prefix, const std::string tail)
{
   bool skip_var_printing = false;
   const unsigned int type = tree_helper::GetRealType(TM, original_type);
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Printing type " + boost::lexical_cast<std::string>(original_type) + " - Var " + boost::lexical_cast<std::string>(var));
   std::string res;
   const tree_nodeRef node_type = TM->get_tree_node_const(type);
   if (var)
   {
      const tree_nodeRef node_var = TM->get_tree_node_const(var);
      if (node_var->get_kind() == var_decl_K)
      {
         var_decl *vd = GetPointer<var_decl>(node_var);
         if ((vd->extern_flag || vd->extern_flag) and print_storage)
            res += "extern ";
         if ((vd->static_flag || vd->static_static_flag) and print_storage)
            res += "static ";
      }
   }
   switch (node_type->get_kind())
   {
      case function_decl_K:
      {
         function_decl *fd = GetPointer<function_decl>(node_type);
         if (fd->undefined_flag)
            res = "extern ";
         if (fd->static_flag)
            res = "static ";
         decl_node *dn = GetPointer<decl_node>(node_type);
         THROW_ASSERT(dn, "expected a declaration node");
         tree_nodeRef ftype = GET_NODE(dn->type);
         /* Print type declaration.  */
         if (ftype->get_kind() == function_type_K)
         {
            function_type *ft = GetPointer<function_type>(ftype);
            res += print_type(TM, GET_INDEX_NODE(ft->retn), global);
         }
         else
            THROW_ERROR(std::string("tree node not currently supported ") + node_type->get_kind_text());
         res += " ";

         /* Print function name.  */
         THROW_ASSERT(dn->name, "expected a name");
         res += tree_helper::print_function_name(TM, fd);
         res += "(";
         if (fd->list_of_args.size())
         {
            for (unsigned int i = 0; i < (fd->list_of_args.size());i++)
            {
               if (i > 0)
                  res += ", ";
               unsigned int argom = GET_INDEX_NODE(fd->list_of_args[i]);
               res = res + print_type(TM, get_type_index(TM,argom), global, print_qualifiers, print_storage, argom, vppf);
            }
         }
         else if (TM->get_implementation_node(type) and TM->get_implementation_node(type) != type)
         {
            skip_var_printing = true;
            res = print_type(TM, TM->get_implementation_node(type), global, print_qualifiers, print_storage, 0, vppf);
            break;
         }
         else if (GetPointer<function_type>(ftype)->prms)
         {
            if (ftype->get_kind() == function_type_K)
            {
               function_type *ft = GetPointer<function_type>(ftype);
               res += print_type(TM, GET_INDEX_NODE(ft->prms), global);
            }
            else
            {
               THROW_ERROR(std::string("tree node not currently supported ") + node_type->get_kind_text());
            }
         }
         if ((fd->list_of_args.size() || GetPointer<function_type>(ftype)->prms) && GetPointer<function_type>(ftype)->varargs_flag)
            res += ", ... ";
         res += ")";
         break;
      }
      case type_decl_K:
      {
         type_decl *td = GetPointer<type_decl>(node_type);
         if (td->name)
         {
            tree_nodeRef name = GET_NODE(td->name);
            if (name->get_kind() == identifier_node_K)
            {
               identifier_node *in = GetPointer<identifier_node>(name);
               ///patch for unsigned char
               std::string typename_value = tree_helper::normalized_ID(in->strg);
               if (typename_value == "char" && GetPointer<integer_type>(GET_NODE(td->type)) && GetPointer<integer_type>(GET_NODE(td->type))->unsigned_flag)
                  res += "unsigned " + typename_value;
               ///patch for va_list
               else if(typename_value == "__va_list_tag")
                  res += "va_list";
               else if(GET_NODE(td->type)->get_kind() == complex_type_K)
               {
                  std::vector<std::string> splitted;
                  boost::algorithm::split(splitted, typename_value, boost::algorithm::is_any_of(" "));
                  if((splitted[0] == "_Complex" || splitted[0] == "__complex__" || splitted[0] == "complex"))
                  {
                     res += "__complex__";
                     for(unsigned int ci=1; ci < splitted.size(); ci++)
                        res += " " + splitted[ci];
                  }
                  else
                     res += typename_value;

               }
               else
                  res += typename_value;
            }
            else
               THROW_ERROR(std::string("Node not yet supported: ") + node_type->get_kind_text());
         }
         else
            THROW_ERROR(std::string("Node not yet supported: ") + node_type->get_kind_text());
         skip_var_printing = true;
         break;
      }
      case identifier_node_K:
      {
         identifier_node *in = GetPointer<identifier_node>(node_type);
         res += tree_helper::normalized_ID(in->strg);
         skip_var_printing = true;
         break;
      }
      case void_type_K:
      case integer_type_K:
      case real_type_K:
      case complex_type_K:
      case vector_type_K:
      case boolean_type_K:
      case CharType_K:
      {
         type_node *tn = GetPointer<type_node>(node_type);
         unsigned int quals = tn->qual;
         /* const internally are not considered as constant...*/
         if (quals > 0)
         {
            tree_nodeRef node;
            if(var)
               node = TM->get_tree_node_const(var);
            if (var > 0 && !tn->name)
            {
               if ((GetPointer<var_decl>(node) && !(GetPointer<var_decl>(node)->scpe)) or GetPointer<parm_decl>(node))
               {
                  res += tree_helper::return_C_qualifiers(quals, true);
               }
               else
                  res += tree_helper::return_C_qualifiers(quals, false);
            }
            else if (global or print_qualifiers or (node and GetPointer<parm_decl>(node)))
            {
               res += tree_helper::return_C_qualifiers(quals, true);
            }
         }
         if (tn->name)
         {
            tree_nodeRef name = GET_NODE(tn->name);
            if (name->get_kind() == identifier_node_K)
            {
               identifier_node *in = GetPointer<identifier_node>(name);
               res += tree_helper::normalized_ID(in->strg);
            }
            else if (name->get_kind() == type_decl_K)
            {
               res += print_type(TM, GET_INDEX_NODE(tn->name), global);
            }
            else
               THROW_ERROR(std::string("Node not yet supported: ") + node_type->get_kind_text());//p_string(buffer, get_unnamed(node));
         }
         else if (node_type->get_kind() == complex_type_K)
         {
            res += "_Complex ";
            if (GetPointer<complex_type>(node_type)->unsigned_flag)
               res += "unsigned ";
            if (tn->algn == 8)
               res += "char";
            else if (tn->algn == 16)
               res += "short";
            else if (tn->algn == 32)
               res += "int";
            else if (tn->algn == 64)
               res += "long long";
            else
               THROW_ERROR(std::string("Node not yet supported: ") + node_type->get_kind_text() + boost::lexical_cast<std::string>(var));
         }
         else if (node_type->get_kind() == vector_type_K)
         {
            vector_type *vt = GetPointer<vector_type>(node_type);
            if (vt->unql)
               res += print_type(TM, GET_INDEX_NODE(vt->unql), global);
            else
            {
               //THROW_ERROR(std::string("Node not yet supported:<unnamed type> ") + node_type->get_kind_text()+boost::lexical_cast<std::string>(type));
               THROW_ASSERT(vt->elts, "expected the type of the elements of the vector");
               res += print_type(TM, GET_INDEX_NODE(vt->elts), global);
               res += " __attribute__((vector_size(" + boost::lexical_cast<std::string>(vt->algn / 8) + ")))";
            }
         }
         else
         {
            switch (node_type->get_kind())
            {
               case integer_type_K:
               {
                  integer_type* it = GetPointer<integer_type>(node_type);
                  if (it->unsigned_flag)
                     res += "unsigned ";
                  if (it->prec != tn->algn)
                  {
                     if (it->prec > 32)
                        res += "long long int";
                     else if (it->prec > 16)
                        res += "int";
                     else if (it->prec > 8)
                        res += "short";
                     else
                        res += "char";
                  }
                  else if (tn->algn == 8)
                     res += "char";
                  else if (tn->algn == 16)
                     res += "short";
                  else if (tn->algn == 32)
                     res += "int";
                  else if (tn->algn == 64)
                     res += "long long";
                  else
                     THROW_ERROR(std::string("Node not yet supported: ") + node_type->get_kind_text());
                  break;
               }
               case enumeral_type_K:
                  break;
               case void_type_K:
                  res += "void";
                  break;
               case real_type_K:
               {
                  real_type* rt = GetPointer<real_type>(node_type);
                  if (rt->prec == 32)
                     res += "_Decimal32";
                  else if (rt->prec == 64)
                     res += "_Decimal64";
                  else if (rt->prec == 128)
                     res += "_Decimal128";
                  else
                     THROW_ERROR(std::string("Real type not yet supported ") + boost::lexical_cast<std::string>(original_type));
                  break;
               }
               default:
                  THROW_ERROR(std::string("Node not yet supported ") + node_type->get_kind_text()+ " " + boost::lexical_cast<std::string>(original_type));
            }
         }
         break;
      }
      case pointer_type_K:
      case reference_type_K:
      {
         unsigned int ptd_type;
         if (node_type->get_kind() == pointer_type_K)
         {
            pointer_type *tree_type = GetPointer<pointer_type>(node_type);
            if (tree_type->name && GET_NODE(tree_type->name)->get_kind() == type_decl_K)
            {
               type_decl * td = GetPointer<type_decl>(GET_NODE(tree_type->name));
               if (td->name && GET_NODE(td->name)->get_kind() == identifier_node_K)
               {
                  identifier_node * id = GetPointer<identifier_node>(GET_NODE(td->name));
                  if (id->strg == "va_list")
                  {
                     res = "va_list";
                     break;
                  }
               }
            }
            if(tree_type->ptd && GET_NODE(tree_type->ptd)->get_kind() == record_type_K
               && GetPointer<record_type>(GET_NODE(tree_type->ptd))->name
               && GET_NODE(GetPointer<record_type>(GET_NODE(tree_type->ptd))->name)->get_kind() == type_decl_K
              )
            {
               type_decl * td = GetPointer<type_decl>(GET_NODE(GetPointer<record_type>(GET_NODE(tree_type->ptd))->name));
               if (td->name && GET_NODE(td->name)->get_kind() == identifier_node_K)
               {
                  identifier_node * id = GetPointer<identifier_node>(GET_NODE(td->name));
                  if (id->strg == "va_list" || id->strg == "__va_list_tag")
                  {
                     res = "va_list";
                     break;
                  }
               }
            }
            res = "*";
            unsigned int quals = tree_type->qual;
            /* const internally are not considered as constant...*/
            if (quals > 0)
            {
               tree_nodeRef node;
               if(var)
                  node = TM->get_tree_node_const(var);
               if((node and GetPointer<parm_decl>(node)) or print_qualifiers)
               {
                  res += tree_helper::return_C_qualifiers(quals, true);
               }
               else
                  res += tree_helper::return_C_qualifiers(quals, false);
            }
            ptd_type = GET_INDEX_NODE(tree_type->ptd);
            res = print_type(TM, ptd_type, global, print_qualifiers, print_storage, var, vppf, prefix + res, tail);
            skip_var_printing = true;
            break;
         }
         else
         {
            res = "/*&*/"; ///references in C are not supported
            reference_type *tree_type = GetPointer<reference_type>(node_type);
            ptd_type = GET_INDEX_NODE(tree_type->refd);
            res = print_type(TM, ptd_type, global, print_qualifiers, print_storage, var, vppf, prefix + res, tail);
            skip_var_printing = true;
            break;
         }
         //res +=print_type(TM, ptd_type) + " " + res;
         break;
      }
      case function_type_K:
      {
         function_type *ft = GetPointer<function_type>(node_type);
         res += print_type(TM, GET_INDEX_NODE(ft->retn), global, true);
         res += "(" + prefix;
         if (var)
         {
            THROW_ASSERT(vppf, "expected a functor");
            res += " " + (*vppf)(var);
         }
         res += tail + ")(";
         if (ft->prms)
         {
            res += print_type(TM, GET_INDEX_NODE(ft->prms), global, true);
         }
         if (ft->varargs_flag)
            res += ", ... ";
         res += ")";
         skip_var_printing = true;
         break;
      }
      case method_type_K:
      {
         method_type *mt = GetPointer<method_type>(node_type);
         res += print_type(TM, GET_INDEX_NODE(mt->clas), global);
         res += "::";
         res += print_type(TM, GET_INDEX_NODE(mt->retn), global);
         res += "(" + prefix;
         if (var)
         {
            THROW_ASSERT(vppf, "expected a functor");
            res += " " + (*vppf)(var);
         }
         res += tail + ")(";
         if (mt->prms)
            res += print_type(TM, GET_INDEX_NODE(mt->prms), global);
         res += ")";
         skip_var_printing = true;
         break;
      }
      case array_type_K:
      {
         const array_type * at = GetPointer<array_type>(node_type);
         if(at->name)
         {
            res += print_type(TM, GET_INDEX_NODE(at->name));
         }
         else
         {
            std::string local_tail;
            /* Print array's type */
            local_tail += "[";
            ///Compute the dimensions
            if (at->size)
            {
               tree_nodeRef array_length = GET_NODE(at->size);
               tree_nodeRef array_t = GET_NODE (at->elts);
               if (array_length->get_kind() == integer_cst_K)
               {
                  integer_cst *arr_ic = GetPointer<integer_cst>(array_length);
                  type_node *tn = GetPointer<type_node>(array_t);
                  integer_cst *eln_ic = GetPointer<integer_cst>(GET_NODE(tn->size));
                  local_tail += boost::lexical_cast<std::string>(tree_helper::get_integer_cst_value(arr_ic) / tree_helper::get_integer_cst_value(eln_ic));
               }
               else if(array_length->get_kind() == var_decl_K)
               {
                  THROW_ERROR_CODE(VLA_EC, "Variable lenght array not supported");
               }
               else
               {
                  THROW_ERROR("array print_type not supported " + STR(GET_INDEX_NODE(at->size)));
                  /*null_deleter nullDel;
                  const var_pp_functorConstRef std_vpf(new std_var_pp_functor(behavioral_helperConstRef(this, nullDel)));
                  local_tail += (*std_vpf)(GET_INDEX_NODE(at->size));*/
               }
            }
            local_tail += "]";
            std::string local_prefix;
            if (prefix != "")
               local_prefix = "(" + prefix;
            if (var)
            {
               THROW_ASSERT(vppf, "expected a functor");
               local_prefix += " " + (*vppf)(var);
            }
            if (prefix != "")
               local_tail = ")" + local_tail;

            res += print_type(TM, GET_INDEX_NODE(at->elts), global, print_qualifiers, print_storage, 0, var_pp_functorConstRef(), "", local_prefix + tail + local_tail);
            ///add alignment
            if (var && TM->get_tree_node_const(var)->get_kind() == field_decl_K)
            {
               unsigned int type_align = at->algn;
               const tree_nodeRef node_var = TM->get_tree_node_const(var);
               unsigned int var_align;
               bool is_a_pointerP = false;
               bool is_static = false;
               switch (node_var->get_kind())
               {
                  case field_decl_K:
                     var_align = GetPointer<field_decl>(node_var)->algn;
                     is_a_pointerP = GET_NODE(GetPointer<field_decl>(node_var)->type)->get_kind() == pointer_type_K;
                     break;
                  case parm_decl_K:
                     //var_align = GetPointer<parm_decl>(node_var)->algn;
                     var_align = type_align;
                     break;
                  case result_decl_K:
                     //var_align = GetPointer<result_decl>(node_var)->algn;
                     var_align = type_align;
                     break;
                  case var_decl_K:
                     var_align = GetPointer<var_decl>(node_var)->algn;
                     is_a_pointerP = GET_NODE(GetPointer<var_decl>(node_var)->type)->get_kind() == pointer_type_K;
                     is_static = GetPointer<var_decl>(node_var)->static_flag;
                     break;
                  default:
                     var_align = type_align;
                     break;
               }
               if (var_align > type_align && !is_a_pointerP && !is_static)
                  res += " __attribute__((aligned(" + boost::lexical_cast<std::string>(var_align / 8) + ")))";
            }
            skip_var_printing = true;
         }
         break;
      }
      case component_ref_K:
      {
         component_ref *cr = GetPointer<component_ref>(node_type);
         res += print_type(TM, GET_INDEX_NODE(cr->type), global);
         break;
      }
      case enumeral_type_K:
      {
         enumeral_type *et = GetPointer<enumeral_type>(node_type);
         if (et->name and (GET_NODE(et->name)->get_kind() == type_decl_K or et->unql))
            res += print_type(TM, GET_INDEX_NODE(et->name), global);
         else if (et->name)
            res += "enum " + print_type(TM, GET_INDEX_NODE(et->name), global);
         else if (et->unql)
            res += "Internal_" + boost::lexical_cast<std::string>(type);
         else
            res += "enum Internal_" + boost::lexical_cast<std::string>(type);
         break;
      }
      case record_type_K:
      {
         record_type *rt = GetPointer<record_type>(node_type);
         unsigned int quals = rt->qual;
         if (quals > 0 and print_qualifiers)
            res += tree_helper::return_C_qualifiers(quals, true);
         if (rt->name and (GET_NODE(rt->name)->get_kind() == type_decl_K or (rt->unql && !rt->system_flag)))
            res += print_type(TM, GET_INDEX_NODE(rt->name), global);
         else if (rt->name)
         {
            std::string struct_name = print_type(TM, GET_INDEX_NODE(rt->name), global);
            if(struct_name == "_IO_FILE")
               res += "FILE";
            else
               res += "struct " + struct_name;
         }
         else if (rt->unql)
            res += "Internal_" + boost::lexical_cast<std::string>(type);
         else
            res += "struct Internal_" + boost::lexical_cast<std::string>(type);
         break;
      }
      case union_type_K:
      {
         union_type *ut = GetPointer<union_type>(node_type);
         unsigned int quals = ut->qual;
         if (quals > 0 and print_qualifiers)
            res += tree_helper::return_C_qualifiers(quals, true);
         if (ut->name and (GET_NODE(ut->name)->get_kind() == type_decl_K or (ut->unql && !ut->system_flag)))
            res += print_type(TM, GET_INDEX_NODE(ut->name), global);
         else if (ut->name)
            res += "union " + print_type(TM, GET_INDEX_NODE(ut->name), global);
         else if (ut->unql)
            res += "Internal_" + boost::lexical_cast<std::string>(type);
         else
            res += "union Internal_" + boost::lexical_cast<std::string>(type);
         break;
      }
      case tree_list_K:
      {
         THROW_ASSERT(var == 0, "Received somenthing of unexpected");
         tree_list *lnode = GetPointer<tree_list>(node_type);
         res += print_type(TM, GET_INDEX_NODE(lnode->valu), global, print_qualifiers);
         ///tree_list are use for parameters declaration: in that case void_type has to be removed from the last type parameter
         if (lnode->chan)
         {
            tree_list *lnode_next = GetPointer<tree_list>(GET_NODE(lnode->chan));
            THROW_ASSERT(lnode_next, "tree_list expected");
            if (!GetPointer<void_type>(GET_NODE(lnode_next->valu)))
               res += "," + print_type(TM, GET_INDEX_NODE(lnode->chan), global, print_qualifiers);
         }
         break;
      }
      default:
         THROW_ASSERT(false, "Type not yet supported " + boost::lexical_cast<std::string>(original_type) + " " + node_type->get_kind_text() + " " + boost::lexical_cast<std::string>(var));
   }
   if(!skip_var_printing)
   {
      res += prefix;
      if (var)
      {
         THROW_ASSERT(vppf, "expected a functor");
         res += " " + (*vppf)(var);
      }
      res += tail;
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Printed type " + boost::lexical_cast<std::string>(original_type));
   return res;
}
