/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file tree_manager.cpp
 * @brief Class implementation of the manager of the tree structures extracted from the raw file.
 *
 * This file implements some of the tree_manager member functions.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 *
*/

///Autoheader include
#include "config_HAVE_CODE_ESTIMATION_BUILT.hpp"
#include "config_HAVE_MAPPING_BUILT.hpp"

///Skipping warnings due to operator() redefinition
#pragma GCC diagnostic ignored "-Woverloaded-virtual"

///Header include
#include "tree_manager.hpp"

///Machine include
#if HAVE_MAPPING_BUILT
#include "machine_node.hpp"
#endif
///Parameter include
#include "Parameter.hpp"

///Tree include
#include "ext_tree_node.hpp"
#include "gimple_writer.hpp"
#include "raw_writer.hpp"
#include "token_interface.hpp"
#include "tree_basic_block.hpp"
#include "tree_node.hpp"
#include "tree_node_finder.hpp"
#include "tree_node_factory.hpp"
#include "tree_nodes_merger.hpp"
#include "tree_reindex.hpp"
#if HAVE_CODE_ESTIMATION_BUILT
#include "weight_information.hpp"
#endif

///Wrapper include
#include "gcc_wrapper.hpp"

#include "dbgPrintHelper.hpp"

tree_manager::tree_manager(const ParameterConstRef _Param) :
   n_pl(0),
   added_goto(0),
   removed_pointer_plus(0),
   debug_level(_Param->get_class_debug_level(GET_CLASS(*this), DEBUG_LEVEL_NONE)),
   last_node_id(1),
   Param(_Param),
   next_vers(0),
   collapse_into_counter(0)
{
}

tree_manager::~tree_manager()
{}

unsigned int tree_manager::get_implementation_node(unsigned int decl_node) const
{
   THROW_ASSERT(GetPointer<function_decl>(get_tree_node_const(decl_node)), "Node "+ boost::lexical_cast<std::string>(decl_node) + " is not a function decl: "+get_tree_node_const(decl_node)->get_kind_text());
   if(GetPointer<function_decl>(get_tree_node_const(decl_node))->body)
      return decl_node;
   else
      return 0;
}

// *****************************************************************************************

unsigned int tree_manager::find_sc_main_node() const
{
   return function_index("sc_main");
}


unsigned int tree_manager::function_index(std::string function_name) const
{
   unsigned int function_id = 0;
   for(std::map<unsigned int, tree_nodeRef>::const_iterator it = function_decl_nodes.begin(); it != function_decl_nodes.end(); it++)
   {
      tree_nodeRef curr_tn = it->second;
      function_decl * fd = GetPointer<function_decl>(curr_tn);
      if(!fd->body)
         continue;
      std::string name = fd->get_function_name();
      if(name == function_name || function_name == std::string("-"))
         function_id = it->first;
   }
   return function_id;
}

void tree_manager::print(std::ostream& os) const
{
#if HAVE_MAPPING_BUILT
   std::string component_type_string = Param->getOption<std::string>(OPT_driving_component_type);
#endif
   raw_writer RW(
#if HAVE_MAPPING_BUILT
      processingElement::get_component_type(component_type_string),
#endif
      os);

   os << STOK(TOK_GCC_VERSION) << ": \"" << GccWrapper::current_gcc_version << "\"\n";
   os << STOK(TOK_PLUGIN_VERSION) << ": \"" << GccWrapper::current_plugin_version << "\"\n";

   std::tr1::unordered_map<unsigned int, tree_nodeRef>::const_iterator ti_end = tree_nodes.end();
   for (std::tr1::unordered_map<unsigned int, tree_nodeRef>::const_iterator ti = tree_nodes.begin(); ti != ti_end; ++ti)
   {
      if(ti->second)
      {
         //PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Dumping tree_node " + boost::lexical_cast<std::string>(ti->first));
         os << "@" << ti->first << " " ;
         ti->second->visit(&RW);
         os << std::endl;
      }
   }
}

void tree_manager::PrintGimple(std::ostream & os, const bool use_uid) const
{
   GimpleWriter gimple_writer(os, use_uid);

   std::map<unsigned int, tree_nodeRef>::const_iterator function, function_end = function_decl_nodes.end();
   for(function = function_decl_nodes.begin(); function != function_end; function++)
   {
      if(GetPointer<function_decl>(function->second)->body)
      {
         function->second->visit(&gimple_writer);
      }
   }
}


#define CREATE_TREE_NODE_CASE_BODY(tree_node_name, node_id, tree_node_schema)\
{\
   tree_node_name * tnn = new tree_node_name;\
   tree_nodeRef cur = tree_nodeRef(tnn);\
   add(node_id, cur);\
   tnn->tree_node_constructor(tree_node_schema, *this);\
   break;\
}

void tree_manager::create_tree_node(const unsigned int node_id, const unsigned int tree_node_type, std::map<treeVocabularyTokenTypes::token_enum, std::string> &tree_node_schema)
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Creating tree node " + boost::lexical_cast<std::string>(node_id));
   tree_node_factory TNF(tree_node_schema,*this);
   TNF.create_tree_node(node_id, tree_node_type);
}

unsigned int tree_manager::new_tree_node_id(const unsigned int ask)
{
   if(ask && tree_nodes.find(ask) == tree_nodes.end())
   {
      add(ask);
      return ask;
   }
   unsigned int temp = last_node_id;
   add(last_node_id); ///add change last_node_id
   return temp;
}

#define FIND_TREE_NODE_CASE_BODY(ti, tree_node_name, node_id)\
{\
   THROW_ASSERT(ti->second, "there is a null ref in the tree_nodes data structure"+boost::lexical_cast<std::string>(node_id));\
   if(GetPointer<tree_node_name>(ti->second) && TNF.check(ti->second))\
   {\
      find_cache[key]=node_id;\
      return node_id;\
   }\
   break;\
}

unsigned int tree_manager::find(unsigned int tree_node_type, const std::map<treeVocabularyTokenTypes::token_enum, std::string> &tree_node_schema)
{
   if (tree_node_type == TOK(TOK_IDENTIFIER_NODE))
   {
      std::string id;
      if (tree_node_schema.find(TOK(TOK_STRG)) != tree_node_schema.end())
         id = tree_node_schema.find(TOK(TOK_STRG))->second;
      else if (tree_node_schema.find(TOK(TOK_OPERATOR)) != tree_node_schema.end())
         id = STOK(TOK_OPERATOR);
      else
         THROW_ERROR("Incorrect schema for identifier_node: no TOK_STRG nor TOK_OPERATOR");
      return find_identifier_nodeID(id);
   }
   std::string key = STOK2(tree_node_type);
   const std::map<treeVocabularyTokenTypes::token_enum, std::string>::const_iterator tns_end =tree_node_schema.end();
   for(std::map<treeVocabularyTokenTypes::token_enum, std::string>::const_iterator tns =tree_node_schema.begin(); tns != tns_end; tns++)
   {
      key += ";" + STOK2(tns->first) + "=" + tns->second ;
   }
   //std::cout << "KEY: " + key + "[" << (find_cache.find(key) != find_cache.end() ? find_cache.find(key)->second : 0) << "]" << std::endl;
   if(find_cache.find(key) != find_cache.end())
      return find_cache.find(key)->second;
   tree_node_finder TNF(tree_node_schema);
   unsigned int node_id;
   std::tr1::unordered_map<unsigned int, tree_nodeRef>::const_iterator ti_end = tree_nodes.end();
   for (std::tr1::unordered_map<unsigned int, tree_nodeRef>::const_iterator ti = tree_nodes.begin(); ti != ti_end; ++ti)
   {
      node_id = ti->first;
      switch(tree_node_type)
      {
         case TOK(TOK_ABS_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, abs_expr, node_id)
         case TOK(TOK_ADDR_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, addr_expr, node_id)
         case TOK(TOK_ARRAY_RANGE_REF):
            FIND_TREE_NODE_CASE_BODY(ti, array_range_ref, node_id)
         case TOK(TOK_ARRAY_REF):
            FIND_TREE_NODE_CASE_BODY(ti, array_ref, node_id)
         case TOK(TOK_ARRAY_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, array_type, node_id)
         case TOK(TOK_ARROW_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, arrow_expr, node_id)
         case TOK(TOK_GIMPLE_ASM):
            FIND_TREE_NODE_CASE_BODY(ti, gimple_asm, node_id)
         case TOK(TOK_BASELINK):
            FIND_TREE_NODE_CASE_BODY(ti, baselink, node_id)
         case TOK(TOK_GIMPLE_BIND):
            FIND_TREE_NODE_CASE_BODY(ti, gimple_bind, node_id)
         case TOK(TOK_BINFO):
            FIND_TREE_NODE_CASE_BODY(ti, binfo, node_id)
         case TOK(TOK_BIT_AND_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, bit_and_expr, node_id)
         case TOK(TOK_BIT_FIELD_REF):
            FIND_TREE_NODE_CASE_BODY(ti, bit_field_ref, node_id)
         case TOK(TOK_BIT_IOR_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, bit_ior_expr, node_id)
         case TOK(TOK_BIT_NOT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, bit_not_expr, node_id)
         case TOK(TOK_BIT_XOR_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, bit_xor_expr, node_id)
         case TOK(TOK_BLOCK):
            FIND_TREE_NODE_CASE_BODY(ti, block, node_id)
         case TOK(TOK_BOOLEAN_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, boolean_type, node_id)
         case TOK(TOK_BUFFER_REF):
            FIND_TREE_NODE_CASE_BODY(ti, buffer_ref, node_id)
         case TOK(TOK_CALL_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, call_expr, node_id)
         case TOK(TOK_GIMPLE_CALL):
            FIND_TREE_NODE_CASE_BODY(ti, gimple_call, node_id)
         case TOK(TOK_CARD_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, card_expr, node_id)
         case TOK(TOK_CASE_LABEL_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, case_label_expr, node_id)
         case TOK(TOK_CAST_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, cast_expr, node_id)
         case TOK(TOK_CATCH_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, catch_expr, node_id)
         case TOK(TOK_CEIL_DIV_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, ceil_div_expr, node_id)
         case TOK(TOK_CEIL_MOD_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, ceil_mod_expr, node_id)
         case TOK(TOK_CHAR_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, CharType, node_id)
         case TOK(TOK_CLEANUP_POINT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, cleanup_point_expr, node_id)
         case TOK(TOK_COMPLEX_CST):
            FIND_TREE_NODE_CASE_BODY(ti, complex_cst, node_id)
         case TOK(TOK_COMPLEX_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, complex_expr, node_id)
         case TOK(TOK_COMPLEX_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, complex_type, node_id)
         case TOK(TOK_COMPONENT_REF):
            FIND_TREE_NODE_CASE_BODY(ti, component_ref, node_id)
         case TOK(TOK_COMPOUND_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, compound_expr, node_id)
         case TOK(TOK_COMPOUND_STMT):
            FIND_TREE_NODE_CASE_BODY(ti, compound_stmt, node_id)
         case TOK(TOK_COND_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, cond_expr, node_id)
          case TOK(TOK_GIMPLE_COND):
             FIND_TREE_NODE_CASE_BODY(ti, gimple_cond, node_id)
         case TOK(TOK_CONJ_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, conj_expr, node_id)
         case TOK(TOK_CONST_DECL):
            FIND_TREE_NODE_CASE_BODY(ti, const_decl, node_id)
         case TOK(TOK_CONSTRUCTOR):
            FIND_TREE_NODE_CASE_BODY(ti, constructor, node_id)
         case TOK(TOK_CONVERT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, convert_expr, node_id)
         case TOK(TOK_CTOR_INITIALIZER):
            FIND_TREE_NODE_CASE_BODY(ti, ctor_initializer, node_id)
         case TOK(TOK_DECL_STMT):
            FIND_TREE_NODE_CASE_BODY(ti, decl_stmt, node_id)
         case TOK(TOK_EH_FILTER_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, eh_filter_expr, node_id)
         case TOK(TOK_ENUMERAL_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, enumeral_type, node_id)
         case TOK(TOK_EQ_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, eq_expr, node_id)
         case TOK(TOK_ERROR_MARK):
            FIND_TREE_NODE_CASE_BODY(ti, error_mark, node_id)
         case TOK(TOK_EXACT_DIV_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, exact_div_expr, node_id)
         case TOK(TOK_EXC_PTR_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, exc_ptr_expr, node_id)
         case TOK(TOK_EXIT_BLOCK_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, exit_block_expr, node_id)
         case TOK(TOK_EXIT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, exit_expr, node_id)
         case TOK(TOK_EXPR_STMT):
            FIND_TREE_NODE_CASE_BODY(ti, expr_stmt, node_id)
         case TOK(TOK_FDESC_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, fdesc_expr, node_id)
         case TOK(TOK_FIELD_DECL):
            FIND_TREE_NODE_CASE_BODY(ti, field_decl, node_id)
         case TOK(TOK_FILTER_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, filter_expr, node_id)
         case TOK(TOK_FIX_CEIL_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, fix_ceil_expr, node_id)
         case TOK(TOK_FIX_FLOOR_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, fix_floor_expr, node_id)
         case TOK(TOK_FIX_ROUND_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, fix_round_expr, node_id)
         case TOK(TOK_FIX_TRUNC_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, fix_trunc_expr, node_id)
         case TOK(TOK_FLOAT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, float_expr, node_id)
         case TOK(TOK_FLOOR_DIV_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, floor_div_expr, node_id)
         case TOK(TOK_FLOOR_MOD_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, floor_mod_expr, node_id)
         case TOK(TOK_FUNCTION_DECL):
            FIND_TREE_NODE_CASE_BODY(ti, function_decl, node_id)
         case TOK(TOK_FUNCTION_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, function_type, node_id)
         case TOK(TOK_GE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, ge_expr, node_id)
         case TOK(TOK_GIMPLE_ASSIGN):
            FIND_TREE_NODE_CASE_BODY(ti, gimple_assign, node_id)
         case TOK(TOK_GIMPLE_GOTO):
            FIND_TREE_NODE_CASE_BODY(ti, gimple_goto, node_id)
         case TOK(TOK_GOTO_SUBROUTINE):
            FIND_TREE_NODE_CASE_BODY(ti, goto_subroutine, node_id)
         case TOK(TOK_GT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, gt_expr, node_id)
         case TOK(TOK_HANDLER):
            FIND_TREE_NODE_CASE_BODY(ti, handler, node_id)
         case TOK(TOK_IMAGPART_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, imagpart_expr, node_id)
         case TOK(TOK_INDIRECT_REF):
            FIND_TREE_NODE_CASE_BODY(ti, indirect_ref, node_id)
         case TOK(TOK_IN_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, in_expr, node_id)
         case TOK(TOK_INIT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, init_expr, node_id)
         case TOK(TOK_INTEGER_CST):
            FIND_TREE_NODE_CASE_BODY(ti, integer_cst, node_id)
         case TOK(TOK_INTEGER_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, integer_type, node_id)
         case TOK(TOK_LABEL_DECL):
            FIND_TREE_NODE_CASE_BODY(ti, label_decl, node_id)
         case TOK(TOK_LABELED_BLOCK_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, labeled_block_expr, node_id)
         case TOK(TOK_GIMPLE_LABEL):
            FIND_TREE_NODE_CASE_BODY(ti, gimple_label, node_id)
         case TOK(TOK_LANG_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, lang_type, node_id)
         case TOK(TOK_LE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, le_expr, node_id)
         case TOK(TOK_LOOP_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, loop_expr, node_id)
         case TOK(TOK_LROTATE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, lrotate_expr, node_id)
         case TOK(TOK_LSHIFT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, lshift_expr, node_id)
         case TOK(TOK_LT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, lt_expr, node_id)
         case TOK(TOK_LTGT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, ltgt_expr, node_id)
         case TOK(TOK_MAX_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, max_expr, node_id)
         case TOK(TOK_MEMORY_PARTITION_TAG):
            FIND_TREE_NODE_CASE_BODY(ti, memory_partition_tag, node_id)
         case TOK(TOK_METHOD_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, method_type, node_id)
         case TOK(TOK_MIN_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, min_expr, node_id)
         case TOK(TOK_MINUS_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, minus_expr, node_id)
         case TOK(TOK_MODIFY_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, modify_expr, node_id)
         case TOK(TOK_MODOP_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, modop_expr, node_id)
         case TOK(TOK_MULT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, mult_expr, node_id)
         case TOK(TOK_NAME_MEMORY_TAG):
            FIND_TREE_NODE_CASE_BODY(ti, name_memory_tag, node_id)
         case TOK(TOK_NAMESPACE_DECL):
            FIND_TREE_NODE_CASE_BODY(ti, namespace_decl, node_id)
         case TOK(TOK_NE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, ne_expr, node_id)
         case TOK(TOK_NEGATE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, negate_expr, node_id)
         case TOK(TOK_NEW_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, new_expr, node_id)
         case TOK(TOK_NON_LVALUE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, non_lvalue_expr, node_id)
         case TOK(TOK_NOP_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, nop_expr, node_id)
         case TOK(TOK_OBJ_TYPE_REF):
            FIND_TREE_NODE_CASE_BODY(ti, obj_type_ref, node_id)
         case TOK(TOK_OFFSET_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, offset_type, node_id)
         case TOK(TOK_ORDERED_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, ordered_expr, node_id)
         case TOK(TOK_OVERLOAD):
            FIND_TREE_NODE_CASE_BODY(ti, overload, node_id)
         case TOK(TOK_PARM_DECL):
            FIND_TREE_NODE_CASE_BODY(ti, parm_decl, node_id)
         case TOK(TOK_GIMPLE_PHI):
            FIND_TREE_NODE_CASE_BODY(ti, gimple_phi, node_id)
         case TOK(TOK_PLACEHOLDER_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, placeholder_expr, node_id)
         case TOK(TOK_PLUS_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, plus_expr, node_id)
         case TOK(TOK_POINTER_PLUS_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, pointer_plus_expr, node_id)
         case TOK(TOK_POINTER_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, pointer_type, node_id)
         case TOK(TOK_POSTDECREMENT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, postdecrement_expr, node_id)
         case TOK(TOK_POSTINCREMENT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, postincrement_expr, node_id)
         case TOK(TOK_PREDECREMENT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, predecrement_expr, node_id)
         case TOK(TOK_PREINCREMENT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, preincrement_expr, node_id)
         case TOK(TOK_QUAL_UNION_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, qual_union_type, node_id)
         case TOK(TOK_RANGE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, range_expr, node_id)
         case TOK(TOK_RDIV_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, rdiv_expr, node_id)
         case TOK(TOK_REAL_CST):
            FIND_TREE_NODE_CASE_BODY(ti, real_cst, node_id)
         case TOK(TOK_REALPART_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, realpart_expr, node_id)
         case TOK(TOK_REAL_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, real_type, node_id)
         case TOK(TOK_RECORD_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, record_type, node_id)
         case TOK(TOK_REFERENCE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, reference_expr, node_id)
         case TOK(TOK_REFERENCE_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, reference_type, node_id)
         case TOK(TOK_REINTERPRET_CAST_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, reinterpret_cast_expr, node_id)
         case TOK(TOK_RESULT_DECL):
            FIND_TREE_NODE_CASE_BODY(ti, result_decl, node_id)
         case TOK(TOK_GIMPLE_RESX):
            FIND_TREE_NODE_CASE_BODY(ti, gimple_resx, node_id)
         case TOK(TOK_GIMPLE_RETURN):
            FIND_TREE_NODE_CASE_BODY(ti, gimple_return, node_id)
         case TOK(TOK_RETURN_STMT):
            FIND_TREE_NODE_CASE_BODY(ti, return_stmt, node_id)
         case TOK(TOK_ROUND_DIV_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, round_div_expr, node_id)
         case TOK(TOK_ROUND_MOD_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, round_mod_expr, node_id)
         case TOK(TOK_RROTATE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, rrotate_expr, node_id)
         case TOK(TOK_RSHIFT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, rshift_expr, node_id)
         case TOK(TOK_SAVE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, save_expr, node_id)
         case TOK(TOK_SCOPE_REF):
            FIND_TREE_NODE_CASE_BODY(ti, scope_ref, node_id)
         case TOK(TOK_SET_LE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, set_le_expr, node_id)
         case TOK(TOK_SET_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, set_type, node_id)
         case TOK(TOK_SIZEOF_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, sizeof_expr, node_id)
         case TOK(TOK_SSA_NAME):
            FIND_TREE_NODE_CASE_BODY(ti, ssa_name, node_id)
         case TOK(TOK_STATEMENT_LIST):
            FIND_TREE_NODE_CASE_BODY(ti, statement_list, node_id)
         case TOK(TOK_STATIC_CAST_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, static_cast_expr, node_id)
         case TOK(TOK_STRING_CST):
            FIND_TREE_NODE_CASE_BODY(ti, string_cst, node_id)
         case TOK(TOK_STRUCT_FIELD_TAG):
            FIND_TREE_NODE_CASE_BODY(ti, struct_field_tag, node_id)
         case TOK(TOK_SYMBOL_MEMORY_TAG):
            FIND_TREE_NODE_CASE_BODY(ti, symbol_memory_tag, node_id)
         case TOK(TOK_GIMPLE_SWITCH):
            FIND_TREE_NODE_CASE_BODY(ti, gimple_switch, node_id)
         case TOK(TOK_TARGET_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, target_expr, node_id)
         case TOK(TOK_TARGET_MEM_REF):
            FIND_TREE_NODE_CASE_BODY(ti, target_mem_ref, node_id)
         case TOK(TOK_TARGET_MEM_REF461):
            FIND_TREE_NODE_CASE_BODY(ti, target_mem_ref461, node_id)
         case TOK(TOK_MEM_REF):
            FIND_TREE_NODE_CASE_BODY(ti, mem_ref, node_id)
         case TOK(TOK_TEMPLATE_DECL):
            FIND_TREE_NODE_CASE_BODY(ti, template_decl, node_id)
         case TOK(TOK_TEMPLATE_ID_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, template_id_expr, node_id)
         case TOK(TOK_TEMPLATE_PARM_INDEX):
            FIND_TREE_NODE_CASE_BODY(ti, template_parm_index, node_id)
         case TOK(TOK_TEMPLATE_TYPE_PARM):
            FIND_TREE_NODE_CASE_BODY(ti, template_type_parm, node_id)
         case TOK(TOK_THROW_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, throw_expr, node_id)
         case TOK(TOK_TRANSLATION_UNIT_DECL):
            FIND_TREE_NODE_CASE_BODY(ti, translation_unit_decl, node_id)
         case TOK(TOK_TREE_LIST):
            FIND_TREE_NODE_CASE_BODY(ti, tree_list, node_id)
         case TOK(TOK_TREE_VEC):
            FIND_TREE_NODE_CASE_BODY(ti, tree_vec, node_id)
         case TOK(TOK_TRUNC_DIV_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, trunc_div_expr, node_id)
         case TOK(TOK_TRUNC_MOD_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, trunc_mod_expr, node_id)
         case TOK(TOK_TRUTH_AND_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, truth_and_expr, node_id)
         case TOK(TOK_TRUTH_ANDIF_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, truth_andif_expr, node_id)
         case TOK(TOK_TRUTH_NOT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, truth_not_expr, node_id)
         case TOK(TOK_TRUTH_OR_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, truth_or_expr, node_id)
         case TOK(TOK_TRUTH_ORIF_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, truth_orif_expr, node_id)
         case TOK(TOK_TRUTH_XOR_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, truth_xor_expr, node_id)
         case TOK(TOK_TRY_BLOCK):
            FIND_TREE_NODE_CASE_BODY(ti, try_block, node_id)
         case TOK(TOK_TRY_CATCH_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, try_catch_expr, node_id)
         case TOK(TOK_TRY_FINALLY):
            FIND_TREE_NODE_CASE_BODY(ti, try_finally, node_id)
         case TOK(TOK_TYPE_DECL):
            FIND_TREE_NODE_CASE_BODY(ti, type_decl, node_id)
         case TOK(TOK_TYPENAME_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, typename_type, node_id)
         case TOK(TOK_UNEQ_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, uneq_expr, node_id)
         case TOK(TOK_UNGE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, unge_expr, node_id)
         case TOK(TOK_UNGT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, ungt_expr, node_id)
         case TOK(TOK_UNION_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, union_type, node_id)
         case TOK(TOK_UNLE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, unle_expr, node_id)
         case TOK(TOK_UNLT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, unlt_expr, node_id)
         case TOK(TOK_UNORDERED_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, unordered_expr, node_id)
         case TOK(TOK_UNSAVE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, unsave_expr, node_id)
         case TOK(TOK_VA_ARG_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, va_arg_expr, node_id)
         case TOK(TOK_VAR_DECL):
            FIND_TREE_NODE_CASE_BODY(ti, var_decl, node_id)
         case TOK(TOK_VEC_NEW_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, vec_new_expr, node_id)
         case TOK(TOK_VECTOR_CST):
            FIND_TREE_NODE_CASE_BODY(ti, vector_cst, node_id)
         case TOK(TOK_VECTOR_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, vector_type, node_id)
         case TOK(TOK_VIEW_CONVERT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, view_convert_expr, node_id)
         case TOK(TOK_GIMPLE_PREDICT):
            FIND_TREE_NODE_CASE_BODY(ti, gimple_predict, node_id)
         case TOK(TOK_VOID_TYPE):
            FIND_TREE_NODE_CASE_BODY(ti, void_type, node_id)
         case TOK(TOK_VTABLE_REF):
            FIND_TREE_NODE_CASE_BODY(ti, vtable_ref, node_id)
         case TOK(TOK_WITH_CLEANUP_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, with_cleanup_expr, node_id)
         case TOK(TOK_WITH_SIZE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, with_size_expr, node_id)
         case TOK(TOK_WHILE_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, while_expr, node_id)
         case TOK(TOK_WIDEN_SUM_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, widen_sum_expr, node_id)
         case TOK(TOK_WIDEN_MULT_EXPR):
            FIND_TREE_NODE_CASE_BODY(ti, widen_mult_expr, node_id)
         default:
            THROW_ERROR("tree_node_type node not yet supported. node type is s" + TI_getTokenName(tree_node_type));
      }
   }
   return 0;
}

void tree_manager::collapse_into(const unsigned int & funID, std::tr1::unordered_map<unsigned int, unsigned int> & stmt_to_bloc, const tree_nodeRef & tn, std::tr1::unordered_set<unsigned int> & removed_nodes)
{
   THROW_ASSERT(tn->get_kind() == tree_reindex_K, "Node is not a tree reindex");
   const unsigned int tree_node_index = GET_INDEX_NODE(tn);
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Collapsing into " + boost::lexical_cast<std::string>(tree_node_index) + " (" + std::string(GET_NODE(tn)->get_kind_text()) + ")");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->");
   if(debug_level >= DEBUG_LEVEL_VERY_PEDANTIC)
   {
      const std::string gimple_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "collapse_into_" + boost::lexical_cast<std::string>(collapse_into_counter) + "_before_" + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + ".gimple";
      collapse_into_counter++;
      std::ofstream gimple_file(gimple_file_name.c_str());
      PrintGimple(gimple_file, false);
      gimple_file.close();
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Writing file " + gimple_file_name);
   }
   tree_nodeRef curr_tn= GET_NODE(tn);
   switch (curr_tn->get_kind())
   {
      case gimple_assign_K:
      {
         stack.push_front(tn);
         gimple_assign * gm = GetPointer<gimple_assign>(curr_tn);
         ///If there is a conversion to a not built-in type, type has to be declared; but if no variable survives, type will be not declared by the backend
         null_deleter null_del;
         if(GET_NODE(gm->op1)->get_kind() != nop_expr_K or not tree_helper::HasToBeDeclared(tree_managerRef(this, null_del), GET_INDEX_NODE(GetPointer<nop_expr>(GET_NODE(gm->op1))->type)))
         {
            collapse_into(funID, stmt_to_bloc, gm->op1, removed_nodes);
         }
         stack.pop_front();
         break;
      }
      case gimple_cond_K:
      {
         stack.push_front(tn);
         gimple_cond* gc = GetPointer<gimple_cond>(curr_tn);
         collapse_into(funID, stmt_to_bloc, gc->op0, removed_nodes);
         stack.pop_front();
         break;
      }
      /* Unary expressions.  */
      case CASE_UNARY_EXPRESSION:
      {
         unary_expr * ue = GetPointer<unary_expr>(curr_tn);
         collapse_into(funID, stmt_to_bloc, ue->op, removed_nodes);
         break;
      }
      case CASE_BINARY_EXPRESSION:
      {
         binary_expr* be = GetPointer<binary_expr>(curr_tn);
         collapse_into(funID, stmt_to_bloc, be->op0, removed_nodes);
         collapse_into(funID, stmt_to_bloc, be->op1, removed_nodes);
         break;
      }
      /*ternary expressions*/
      case gimple_switch_K:
      {
         gimple_switch* se = GetPointer<gimple_switch>(curr_tn);
         collapse_into(funID, stmt_to_bloc, se->op0, removed_nodes);
         break;
      }
      case CASE_TERNARY_EXPRESSION:
      {
         ternary_expr * te = GetPointer<ternary_expr>(curr_tn);
         collapse_into(funID, stmt_to_bloc, te->op0, removed_nodes);
         collapse_into(funID, stmt_to_bloc, te->op1, removed_nodes);
         if(te->op2)
            collapse_into(funID, stmt_to_bloc, te->op2, removed_nodes);
         break;
      }
      case CASE_QUATERNARY_EXPRESSION:
      {
         quaternary_expr * qe = GetPointer<quaternary_expr>(curr_tn);
         collapse_into(funID, stmt_to_bloc, qe->op0, removed_nodes);
         collapse_into(funID, stmt_to_bloc, qe->op1, removed_nodes);
         if(qe->op2)
            collapse_into(funID, stmt_to_bloc, qe->op2, removed_nodes);
         if(qe->op3)
            collapse_into(funID, stmt_to_bloc, qe->op3, removed_nodes);
         break;
      }
      case var_decl_K:
      case result_decl_K:
      case parm_decl_K:
      case integer_cst_K:
      case real_cst_K:
      case string_cst_K:
      case vector_cst_K:
      case complex_cst_K:
      case field_decl_K:
      case exc_ptr_expr_K:
      case filter_expr_K:
      case label_decl_K:
      case gimple_label_K:
      case gimple_asm_K:
      case gimple_phi_K:
      case target_mem_ref_K:
      case target_mem_ref461_K:
      case gimple_nop_K:
      case function_decl_K:
      {
         break;
      }
      case ssa_name_K:
      {
         ssa_name * sn = GetPointer<ssa_name>(curr_tn);

         //if(sn->uses == 1 &&
         if(sn->def_stmts.size()==1)
         {
            gimple_assign * gm = GetPointer<gimple_assign>(GET_NODE(sn->def_stmts.front()));
            // Don't continue if declaration of ssa variable is not a gimple_assign
            if((!gm && GET_NODE(sn->def_stmts.front())->get_kind() == gimple_asm_K) || (!gm && GET_NODE(sn->def_stmts.front())->get_kind() == nop_expr_K) || (!gm && GET_NODE(sn->def_stmts.front())->get_kind() == gimple_nop_K))
               break;
            if(gm->virtual_operands and gm->virtual_operands->list_of_vdefs.size())
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---The gimple assignment where ssa name is defined has virtual def");
               break;
            }
            ///If there is a conversion to a not built-in type, type has to be declared; but if no variable survives, type will be not declared by the backend
            null_deleter null_del;
            if(GET_NODE(gm->op1)->get_kind() == nop_expr_K and tree_helper::HasToBeDeclared(tree_managerRef(this, null_del), GET_INDEX_NODE(GetPointer<nop_expr>(GET_NODE(gm->op1))->type)))
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---The gimple assignment where ssa name is defined has a non-builtin nop_expr");
               break;
            }

            if((Param->getOption<bool>(OPT_compare_models) or Param->getOption<bool>(OPT_normalize_models) or Param->getOption<bool>(OPT_compare_measure_regions)) and GET_NODE(gm->op1)->get_kind() == call_expr_K)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---The gimple assignment has call_expr on the right and we are target profiling");
               break;
            }
            collapse_into(funID, stmt_to_bloc, sn->def_stmts.front(), removed_nodes);
            THROW_ASSERT(gm, "ssa name " + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + " not defined in a gimple modify stmt, nor in gimple_asm, nor in a nop_expr");

            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Definition of " + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + ":");
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---" + boost::lexical_cast<std::string>(GET_INDEX_NODE(sn->def_stmts.front())) + " (" + std::string(GET_NODE(sn->def_stmts.front())->get_kind_text()) + ")");
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Uses of " + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + " - " + boost::lexical_cast<std::string>(sn->uses) + ":");
            std::vector<tree_nodeRef>::iterator uses_end = sn->use_stmts.end();
            for(std::vector<tree_nodeRef>::iterator uses_it = sn->use_stmts.begin(); uses_it != uses_end; uses_it++)
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->" + boost::lexical_cast<std::string>(GET_INDEX_NODE(*uses_it)) + ": " + (*uses_it)->ToString());
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
            }

            // Retrieve curr_block, the block which contains the conditional expression that originated the collapsing
            tree_nodeRef temp = get_tree_node_const(funID);
            function_decl * fd = GetPointer<function_decl>(temp);
            statement_list * sl = GetPointer<statement_list>(GET_NODE(fd->body));
            std::map<unsigned int, blocRef> & list_of_bloc = sl->list_of_bloc;
            unsigned int curr_block_index;
            blocRef curr_block;
            // Find the conditional expression in the stack
            std::deque<tree_nodeRef>::iterator stack_it;
            for(stack_it = stack.begin(); stack_it != stack.end(); stack_it++)
            {
               if(GET_NODE(*stack_it)->get_kind() == gimple_cond_K)
               {
                  curr_block_index = (stmt_to_bloc.find(GET_INDEX_NODE(*stack_it)))->second;
                  curr_block = list_of_bloc[curr_block_index];
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Found gimple_cond in block: " + boost::lexical_cast<std::string>((stmt_to_bloc.find(GET_INDEX_NODE(*stack_it)))->second));
                  break;
               }
            }
            THROW_ASSERT(curr_block, "Conditional expression not found in stack");

            // List of statements in curr_block
            std::vector<tree_nodeRef> curr_stmt_list = curr_block->list_of_stmt;

            // Retrieve def_block, the block which contains the definition of ssa_name (could be different from curr_block
            THROW_ASSERT(stmt_to_bloc.find(GET_INDEX_NODE(sn->def_stmts.front())) != stmt_to_bloc.end(), "Statement not found in stmt_to_bloc map: " + boost::lexical_cast<std::string>(GET_INDEX_NODE(sn->def_stmts.front())));
            blocRef def_block = list_of_bloc[(stmt_to_bloc.find(GET_INDEX_NODE(sn->def_stmts.front())))->second];
            THROW_ASSERT(def_block, "Definition block not found");

            // If the defined ssa variable is not only used in the conditional statement, copy the definition in the blocks
            // where it is used
            if(sn->uses != 1)
            {
               // This check is needed to avoid situations in which variables are modified internally to the conditional statement
               // An example is the ++ (or --) operator. In such case the collapsing procedure has to be stopped (break).
               // The control is carried out in the following way: if almost one instruction which is between the conditional 
               // expression and the current gimple_modify_statement (ssa_name definition) has virtual operands, then the
               // collapsing procedure stops.
               bool in_between = false;
               bool vdef_found = false;
               std::vector<tree_nodeRef>::iterator it, it_begin = curr_stmt_list.begin();
               std::vector<tree_nodeRef>::iterator it_end = curr_stmt_list.end();
               for(it = it_begin; it != it_end; it++)
               {
                  if(in_between && GET_NODE(*it)->get_kind() == gimple_assign_K)
                  {
                     gimple_assign * gms = GetPointer<gimple_assign>(GET_NODE(*it));
                     if(gms->virtual_operands)
                     {
                        vopsRef virtual_operands = gms->virtual_operands;
                        if(!(virtual_operands->list_of_vdefs.empty()))
                        {
                           INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Statement " + boost::lexical_cast<std::string>(GET_INDEX_NODE(*it)) + " contains vdefs");
                           vdef_found = true;
                        }
                     }
                  }
                  if(GET_INDEX_NODE(*it) == GET_INDEX_NODE(sn->def_stmts.front()))
                  {
                     in_between = true;
                  }
               }
               if(vdef_found) break;

               INDENT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "---None of uses of current ssa contains virtual defs so continuing in collapsing");

               // If the definition of ssa_name is elsewhere, there is no need of copying it
               if(curr_block == def_block)
               {
                  // Group the uses of ssa variables on the basis of the basic block they belong to
                  std::tr1::unordered_map<unsigned int, std::vector<tree_nodeRef> > copy_bloc_to_stmt;
                  std::vector<tree_nodeRef>::iterator uses_it, uses_it_end = sn->use_stmts.end();
                  for(uses_it = sn->use_stmts.begin(); uses_it != uses_it_end; uses_it++)
                  {

                     THROW_ASSERT(stmt_to_bloc.find(GET_INDEX_NODE(*uses_it)) != stmt_to_bloc.end(), boost::lexical_cast<std::string>(GET_INDEX_NODE(*uses_it)) + " is not in stmt_to_bloc");
                     unsigned int copy_block_index = (stmt_to_bloc.find(GET_INDEX_NODE(*uses_it)))->second;
                     copy_bloc_to_stmt[copy_block_index].push_back(*uses_it);

                  }

                  // Copy the definition statement into each block the ssa variable is used in
                  std::tr1::unordered_map<unsigned int, std::vector<tree_nodeRef> >::iterator copy_block_it;
                  for(copy_block_it = copy_bloc_to_stmt.begin(); copy_block_it != copy_bloc_to_stmt.end(); copy_block_it++)
                  {

                     blocRef copy_block = list_of_bloc[copy_block_it->first];
                     std::vector<tree_nodeRef> copy_block_use_stmts = copy_block_it->second;

                     if(copy_block != curr_block)
                     {

                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Copying statement " + boost::lexical_cast<std::string>(GET_INDEX_NODE(sn->def_stmts.front())) + " in block " + boost::lexical_cast<std::string>(copy_block_it->first) + " (!= " + boost::lexical_cast<std::string>((stmt_to_bloc.find(GET_INDEX_NODE(sn->def_stmts.front())))->second) + ")");

                        // Create a new ssa_node, identical to the one being analyzed, but with different version number
                        std::map<treeVocabularyTokenTypes::token_enum, std::string> IR_schema;
                        unsigned int version = get_next_vers();
                        IR_schema[TOK(TOK_VAR)] = boost::lexical_cast<std::string>(GET_INDEX_NODE(sn->var));
                        IR_schema[TOK(TOK_VERS)] = boost::lexical_cast<std::string>(version);
                        IR_schema[TOK(TOK_VOLATILE)] = boost::lexical_cast<std::string>(sn->volatile_flag);
                        unsigned int sn_index = new_tree_node_id();
                        create_tree_node(sn_index, TOK(TOK_SSA_NAME), IR_schema);
                        tree_nodeRef tree_reindexRef_sn = tree_nodeRef(new tree_reindex(sn_index, add(sn_index)));
                        ssa_name * new_sn = GetPointer<ssa_name>(GET_NODE(tree_reindexRef_sn));
                        new_sn->uses = copy_block_use_stmts.size();
                        new_sn->ptr_info = sn->ptr_info;

                        // Replace the occurrences of the considered ssa variable with the newly created ssa variable
                        std::vector<tree_nodeRef>::iterator copy_block_uses_it;
                        for(copy_block_uses_it = copy_block_use_stmts.begin(); copy_block_uses_it != copy_block_use_stmts.end(); copy_block_uses_it++)
                        {
                           replace_ssa_name(*copy_block_uses_it, tn, tree_reindexRef_sn, *copy_block_uses_it);
                           new_sn->use_stmts.push_back(*copy_block_uses_it);
                        }

                        // Create a new node, the copy of the definition of ssa_name
                        IR_schema.clear();
                        IR_schema[TOK(TOK_SRCP)] = "<built-in>:0:0";
                        IR_schema[TOK(TOK_OP0)] = boost::lexical_cast<std::string>(GET_INDEX_NODE(tree_reindexRef_sn));
                        IR_schema[TOK(TOK_OP1)] = boost::lexical_cast<std::string>(GET_INDEX_NODE(gm->op1));
                        if(gm->orig)
                           IR_schema[TOK(TOK_ORIG)] = boost::lexical_cast<std::string>(GET_INDEX_NODE(gm->orig));
                        unsigned int gm_index = new_tree_node_id();
                        create_tree_node(gm_index, TOK(TOK_GIMPLE_ASSIGN), IR_schema);
                        tree_nodeRef tree_reindexRef_gm = tree_nodeRef(new tree_reindex(gm_index, add(gm_index)));
                        gimple_assign * new_gm = GetPointer<gimple_assign>(GET_NODE(tree_reindexRef_gm));
#if HAVE_CODE_ESTIMATION_BUILT
                        new_gm->weight_information->instruction_size = gm->weight_information->instruction_size;
#if HAVE_RTL_BUILT
                        new_gm->weight_information->rtl_instruction_size = gm->weight_information->rtl_instruction_size;
#endif
                        new_gm->weight_information->recursive_weight = gm->weight_information->recursive_weight;
#endif
                        new_gm->virtual_operands = gm->virtual_operands;
                        // Need to control whether there are labels at the beginning of the block
                        // In such case, copy the statement after the labels
                        std::vector<tree_nodeRef> & copy_block_stmts = copy_block->list_of_stmt;
                        std::vector<tree_nodeRef>::iterator it_copy, it_copy_end = copy_block_stmts.end();
                        for(it_copy = copy_block_stmts.begin(); it_copy != it_copy_end; it_copy++)
                        {
                           if(GET_NODE(*it_copy)->get_kind() != gimple_label_K)
                           {
                              copy_block_stmts.insert(it_copy, tree_reindexRef_gm);
                              break;
                           }
                        }
                        // Update stmt_to_bloc map
                        stmt_to_bloc[GET_INDEX_NODE(tree_reindexRef_gm)] = copy_block_it->first;

                        new_sn->add_def(tree_reindexRef_gm);

                     }
                     else
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "copy_block = curr_block");
                  }
                  // Update usage info using uses_erase_temp map
                  std::tr1::unordered_map<ssa_name*, tree_nodeRef>::iterator uses_temp_it;
                  for(uses_temp_it = uses_erase_temp.begin(); uses_temp_it != uses_erase_temp.end(); uses_temp_it++)
                  {
                     ssa_name * sn_t = uses_temp_it->first;
                     tree_nodeRef stmt_t = uses_temp_it->second;
                     std::vector<tree_nodeRef>::iterator uses_sn_it, uses_sn_it_end = sn_t->use_stmts.end();
                     for(uses_sn_it = sn_t->use_stmts.begin(); uses_sn_it != uses_sn_it_end; uses_sn_it++)
                     {
                        if(GET_INDEX_NODE(*uses_sn_it) == GET_INDEX_NODE(stmt_t))
                        {
                           (sn_t->uses)--;
                           sn_t->use_stmts.erase(uses_sn_it);
                           break;
                        }
                     }
                  }
                  uses_erase_temp.clear();
               }
            }

            if(gm->virtual_operands)
            {
               vopsRef virtual_operands = gm->virtual_operands;
               // Target of collapsing is a conditional expression; we can't put into it expression with side effects
               if(virtual_operands->list_of_vdefs.size())
                  break;
               tree_nodeRef top = stack.front();
               if(GET_NODE(top)->get_kind() == gimple_cond_K)
               {
                  gimple_cond * gc = GetPointer<gimple_cond>(GET_NODE(top));
                  gc->virtual_operands = gm->virtual_operands;
               }
               else if (GET_NODE(top)->get_kind() == gimple_assign_K)
               {
                  gimple_assign * top_gm = GetPointer<gimple_assign>(GET_NODE(top));
                  top_gm->virtual_operands = gm->virtual_operands;
               }
               else
                  THROW_ERROR("Unsupported type of tree node during collapsing");
            }
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Substituting operand " + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + " with " + boost::lexical_cast<std::string>(GET_INDEX_NODE(gm->op1)));
            tree_nodeRef tn_old = tn;
            // Change the operand into every statement in the curr_block which uses the considered ssa_name
            if(def_block == curr_block)
            {
               std::vector<tree_nodeRef> use_stmts = sn->use_stmts;
               std::vector<tree_nodeRef>::iterator uses_it2, uses_it2_end = use_stmts.end();
               for(uses_it2 = use_stmts.begin(); uses_it2 != uses_it2_end; uses_it2++)
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Checking for operation " + boost::lexical_cast<std::string>(GET_INDEX_NODE(*uses_it2)));
                  THROW_ASSERT(stmt_to_bloc.find(GET_INDEX_NODE(*uses_it2)) != stmt_to_bloc.end(), "Statement not found in stmt_to_bloc map: " + boost::lexical_cast<std::string>(GET_INDEX_NODE(*uses_it2)));
                  unsigned int use_block_index = (stmt_to_bloc.find(GET_INDEX_NODE(*uses_it2)))->second;
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Operation belongs to BB" + boost::lexical_cast<std::string>(use_block_index));
                  blocRef use_block = list_of_bloc[use_block_index];
                  if(use_block == curr_block)
                  {
                     replace_ssa_name(*uses_it2, tn_old, gm->op1, *uses_it2);
                     // Update usage info using uses_erase_temp map
                     std::tr1::unordered_map<ssa_name*, tree_nodeRef>::iterator uses_temp_it;
                     for(uses_temp_it = uses_erase_temp.begin(); uses_temp_it != uses_erase_temp.end(); uses_temp_it++)
                     {
                        ssa_name * sn_t = uses_temp_it->first;
                        tree_nodeRef stmt_t = uses_temp_it->second;
                        std::vector<tree_nodeRef>::iterator uses_sn_it, uses_sn_it_end = sn_t->use_stmts.end();
                        for(uses_sn_it = sn_t->use_stmts.begin(); uses_sn_it != uses_sn_it_end; uses_sn_it++)
                        {
                           if(GET_INDEX_NODE(*uses_sn_it) == GET_INDEX_NODE(stmt_t))
                           {
                              (sn_t->uses)--;
                              sn_t->use_stmts.erase(uses_sn_it);
                              break;
                           }
                        }
                     }
                     uses_erase_temp.clear();
                     // Erase usage information about the definition being erased of the current ssa variable in every ssa variable used in it
                     erase_usage_info(sn->def_stmts.front(), sn->def_stmts.front());
                     // Insert usage information about the target statement of the substitution in every ssa variable used in the definition being erased
                     insert_usage_info(sn->def_stmts.front(), *uses_it2);
                  }
               }
            }
#if HAVE_CODE_ESTIMATION_BUILT
            THROW_ASSERT(GetPointer<WeightedNode>(GET_NODE(stack.front())), "Tree node in front of the stack is not weighted. Kind is " + std::string(GET_NODE(stack.front())->get_kind_text()));
            if(GetPointer<WeightedNode>(GET_NODE(gm->op1)))
            {
               const std::map<ComponentTypeConstRef, ProbabilityDistribution> & weights = GetPointer<WeightedNode>(GET_NODE(gm->op1))->weight_information->recursive_weight;
               std::map<ComponentTypeConstRef, ProbabilityDistribution>::const_iterator w, w_end = weights.end();
               for(w = weights.begin(); w != w_end; w++)
               {
                  GetPointer<WeightedNode>(GET_NODE(stack.front()))->weight_information->recursive_weight[w->first] += w->second;
               }
            }
#if HAVE_RTL_BUILT
            GetPointer<WeightedNode>(GET_NODE(stack.front()))->weight_information->rtl_instruction_size += gm->weight_information->rtl_instruction_size;
#endif
            GetPointer<WeightedNode>(GET_NODE(stack.front()))->weight_information->instruction_size += gm->weight_information->instruction_size;
#endif
            std::vector<tree_nodeRef> & list_of_stmt = curr_block->list_of_stmt;
            std::vector<tree_nodeRef>::iterator it2, it2_end = list_of_stmt.end();
            for(it2 = list_of_stmt.begin(); it2 != it2_end; it2++)
            {
               // Remove the definition statements contained in curr_block
               if(GET_INDEX_NODE(*it2) == GET_INDEX_NODE(sn->def_stmts.front()))
               {
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Removed statement " + boost::lexical_cast<std::string>(GET_INDEX_NODE(*it2)));
                  removed_nodes.insert(GET_INDEX_NODE(*it2));
                  list_of_stmt.erase(it2);
                  break;
               }
            }
         }
         else
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Multiple definitions for ssa_name " + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)));
         break;
      }

      case call_expr_K:
      {
         call_expr* ce = GetPointer<call_expr>(curr_tn);
         std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            collapse_into(funID, stmt_to_bloc, *arg, removed_nodes);
         }
         break;
      }
      case gimple_call_K:
      {
         gimple_call* ce = GetPointer<gimple_call>(curr_tn);
         std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            collapse_into(funID, stmt_to_bloc, *arg, removed_nodes);
         }
         break;
      }
      default:
         THROW_ERROR(std::string("Node not supported (") + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + std::string("): ") + curr_tn->get_kind_text());

   }
   if(debug_level >= DEBUG_LEVEL_VERY_PEDANTIC)
   {
      const std::string gimple_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "collapse_into_" + boost::lexical_cast<std::string>(collapse_into_counter) + "_after_" + boost::lexical_cast<std::string>(tree_node_index) + ".gimple";
      collapse_into_counter++;
      std::ofstream gimple_file(gimple_file_name.c_str());
      PrintGimple(gimple_file, false);
      gimple_file.close();
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Writing file " + gimple_file_name);
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Ended collapsing into " + boost::lexical_cast<std::string>(tree_node_index));
}

void tree_manager::replace_ssa_name(tree_nodeRef & tn, const tree_nodeRef & old_node, tree_nodeRef & new_node, tree_nodeRef & stmt)
{
   THROW_ASSERT(tn->get_kind() == tree_reindex_K, "Node is not a tree reindex");
   tree_nodeRef curr_tn= GET_NODE(tn);
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Replacing " + old_node->ToString() + "(" + boost::lexical_cast<std::string>(GET_INDEX_NODE(old_node)) + ") with " + new_node->ToString() + "(" + boost::lexical_cast<std::string>(GET_INDEX_NODE(new_node)) + ") starting from node " + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + ": " + tn->ToString() + ")");
   switch (curr_tn->get_kind())
   {
      case gimple_assign_K:
      {
         gimple_assign * gm = GetPointer<gimple_assign>(curr_tn);
         replace_ssa_name(gm->op0, old_node, new_node, stmt);
         replace_ssa_name(gm->op1, old_node, new_node, stmt);
         break;
      }
      case gimple_cond_K:
      {
         gimple_cond* gc = GetPointer<gimple_cond>(curr_tn);
         replace_ssa_name(gc->op0, old_node, new_node, stmt);
         break;
      }
      case CASE_UNARY_EXPRESSION:
      {
         unary_expr * ue = GetPointer<unary_expr>(curr_tn);
         replace_ssa_name(ue->op, old_node, new_node, stmt);
         break;
      }
      case CASE_BINARY_EXPRESSION:
      {
         binary_expr* be = GetPointer<binary_expr>(curr_tn);
         replace_ssa_name(be->op0, old_node, new_node, stmt);
         replace_ssa_name(be->op1, old_node, new_node, stmt);
         break;
      }
      case gimple_switch_K:
      {
         gimple_switch* se = GetPointer<gimple_switch>(curr_tn);
         replace_ssa_name(se->op0, old_node, new_node, stmt);
         break;
      }
      case CASE_TERNARY_EXPRESSION:
      {
         ternary_expr * te = GetPointer<ternary_expr>(curr_tn);
         replace_ssa_name(te->op0, old_node, new_node, stmt);
         replace_ssa_name(te->op1, old_node, new_node, stmt);
         if(te->op2)
            replace_ssa_name(te->op2, old_node, new_node, stmt);
         break;
      }
      case CASE_QUATERNARY_EXPRESSION:
      {
         quaternary_expr * qe = GetPointer<quaternary_expr>(curr_tn);
         replace_ssa_name(qe->op0, old_node, new_node, stmt);
         replace_ssa_name(qe->op1, old_node, new_node, stmt);
         if(qe->op2)
            replace_ssa_name(qe->op2, old_node, new_node, stmt);
         if(qe->op3)
            replace_ssa_name(qe->op3, old_node, new_node, stmt);
         break;
      }
      case var_decl_K:
      case result_decl_K:
      case parm_decl_K:
      case integer_cst_K:
      case real_cst_K:
      case string_cst_K:
      case vector_cst_K:
      case complex_cst_K:
      case field_decl_K:
      case exc_ptr_expr_K:
      case filter_expr_K:
      case label_decl_K:
      case gimple_label_K:
      case gimple_asm_K:
      case gimple_phi_K:
      case target_mem_ref_K:
      case target_mem_ref461_K:
      case function_decl_K:
      {
         break;
      }
      case ssa_name_K:
      {

         if(GET_INDEX_NODE(tn) == GET_INDEX_NODE(old_node))
         {
            ssa_name * sn = GetPointer<ssa_name>(curr_tn);
            uses_erase_temp[sn] = stmt;
            tn = new_node;
         }
         break;
      }
      case call_expr_K:
      {
         call_expr * ce = GetPointer<call_expr>(curr_tn);
         std::vector<tree_nodeRef> & args = ce->args;
         for(size_t arg_index = 0; arg_index < args.size(); arg_index++)
         {
            replace_ssa_name(args[arg_index], old_node, new_node, stmt);
         }
         break;
      }
      case gimple_call_K:
      {
         gimple_call * ce = GetPointer<gimple_call>(curr_tn);
         std::vector<tree_nodeRef> & args = ce->args;
         for(size_t arg_index = 0; arg_index < args.size(); arg_index++)
         {
            replace_ssa_name(args[arg_index], old_node, new_node, stmt);
         }
         break;
      }
      default:
         THROW_ERROR(std::string("Node not supported (") + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + std::string("): ") + curr_tn->get_kind_text());
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Replaced " + old_node->ToString() + "(" + boost::lexical_cast<std::string>(GET_INDEX_NODE(old_node)) + ") with " + new_node->ToString() + "(" + boost::lexical_cast<std::string>(GET_INDEX_NODE(new_node)) + ") New statement " + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + ": " + tn->ToString() + ")");
}

void tree_manager::erase_usage_info(tree_nodeRef & tn, tree_nodeRef & stmt)
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Erase usage into node " + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + " (" + GET_NODE(tn)->get_kind_text() + "). Statement: " + boost::lexical_cast<std::string>(GET_INDEX_NODE(stmt)) + " (" + GET_NODE(stmt)->get_kind_text() + ")");
   THROW_ASSERT(tn->get_kind() == tree_reindex_K, "Node is not a tree reindex");
   tree_nodeRef curr_tn= GET_NODE(tn);
   switch (curr_tn->get_kind())
   {
      case gimple_assign_K:
      {
         gimple_assign * gm = GetPointer<gimple_assign>(curr_tn);
         erase_usage_info(gm->op1, stmt);
         break;
      }
      case gimple_cond_K:
      {
         gimple_cond* gc = GetPointer<gimple_cond>(curr_tn);
         erase_usage_info(gc->op0, stmt);
         break;
      }
      case CASE_UNARY_EXPRESSION:
      {
         if(curr_tn->get_kind() == addr_expr_K)
         {
/*            if(already_visited.find(curr_tn) != already_visited.end())
            {
               break;
            }
            already_visited.insert(curr_tn);*/
         }
         unary_expr * ue = GetPointer<unary_expr>(curr_tn);
         erase_usage_info(ue->op, stmt);
         break;
      }
      case CASE_BINARY_EXPRESSION:
      {
         binary_expr* be = GetPointer<binary_expr>(curr_tn);
         erase_usage_info(be->op0, stmt);
         erase_usage_info(be->op1, stmt);
         break;
      }
      case gimple_switch_K:
      {
         gimple_switch* se = GetPointer<gimple_switch>(curr_tn);
         erase_usage_info(se->op0, stmt);
         break;
      }
      case CASE_TERNARY_EXPRESSION:
      {
         ternary_expr * te = GetPointer<ternary_expr>(curr_tn);
         erase_usage_info(te->op0, stmt);
         erase_usage_info(te->op1, stmt);
         if(te->op2)
            erase_usage_info(te->op2, stmt);
         break;
      }
      case CASE_QUATERNARY_EXPRESSION:
      {
         quaternary_expr * qe = GetPointer<quaternary_expr>(curr_tn);
         erase_usage_info(qe->op0, stmt);
         erase_usage_info(qe->op1, stmt);
         if(qe->op2)
            erase_usage_info(qe->op2, stmt);
         if(qe->op3)
            erase_usage_info(qe->op3, stmt);
         break;
      }
      case var_decl_K:
      case result_decl_K:
      case parm_decl_K:
      case integer_cst_K:
      case real_cst_K:
      case string_cst_K:
      case vector_cst_K:
      case complex_cst_K:
      case field_decl_K:
      case exc_ptr_expr_K:
      case filter_expr_K:
      case label_decl_K:
      case gimple_label_K:
      case gimple_asm_K:
      case gimple_phi_K:
      case target_mem_ref_K:
      case target_mem_ref461_K:
      case function_decl_K:
      {
         break;
      }
      case ssa_name_K:
      {
         ssa_name * sn = GetPointer<ssa_name>(curr_tn);
         std::vector<tree_nodeRef>::iterator uses_it = sn->use_stmts.begin();
         while(uses_it != sn->use_stmts.end())
         {
            if(GET_INDEX_NODE(*uses_it) == GET_INDEX_NODE(stmt))
            {
               (sn->uses)--;
               sn->use_stmts.erase(uses_it);
               uses_it = sn->use_stmts.begin();
            }
            else
            {
               uses_it++;
            }
         }
         break;
      }
      case call_expr_K:
      {
         call_expr* ce = GetPointer<call_expr>(curr_tn);
         std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            erase_usage_info(*arg, stmt);
         }
         break;
      }
      case gimple_call_K:
      {
         gimple_call* ce = GetPointer<gimple_call>(curr_tn);
         std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            erase_usage_info(*arg, stmt);
         }
         break;
      }
      default:
         THROW_ERROR(std::string("Node not supported (") + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + std::string("): ") + curr_tn->get_kind_text());
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Erased usage into node " + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + " (" + GET_NODE(tn)->get_kind_text() + "). Statement: " + boost::lexical_cast<std::string>(GET_INDEX_NODE(stmt)) + " (" + GET_NODE(stmt)->get_kind_text() + ")");
}

void tree_manager::insert_usage_info(tree_nodeRef & tn, tree_nodeRef & stmt)
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Insert usage info into node " + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + " (" + GET_NODE(tn)->get_kind_text() + "). Statement: " + boost::lexical_cast<std::string>(GET_INDEX_NODE(stmt)) + " (" + GET_NODE(stmt)->get_kind_text() + ")");
   THROW_ASSERT(tn->get_kind() == tree_reindex_K, "Node is not a tree reindex");
   tree_nodeRef curr_tn= GET_NODE(tn);
   switch (curr_tn->get_kind())
   {
      case gimple_assign_K:
      {
         gimple_assign * gm = GetPointer<gimple_assign>(curr_tn);
         insert_usage_info(gm->op1, stmt);
         break;
      }
      case gimple_cond_K:
      {
         gimple_cond* gc = GetPointer<gimple_cond>(curr_tn);
         insert_usage_info(gc->op0, stmt);
         break;
      }
      case CASE_UNARY_EXPRESSION:
      {
         if(curr_tn->get_kind() == addr_expr_K)
         {
/*            if(already_visited.find(curr_tn) != already_visited.end())
            {
               break;
            }
            already_visited.insert(curr_tn);*/
         }
         unary_expr * ue = GetPointer<unary_expr>(curr_tn);
         insert_usage_info(ue->op, stmt);
         break;
      }
      case CASE_BINARY_EXPRESSION:
      {
         binary_expr* be = GetPointer<binary_expr>(curr_tn);
         insert_usage_info(be->op0, stmt);
         insert_usage_info(be->op1, stmt);
         break;
      }
      case gimple_switch_K:
      {
         gimple_switch* se = GetPointer<gimple_switch>(curr_tn);
         insert_usage_info(se->op0, stmt);
         break;
      }
      case CASE_TERNARY_EXPRESSION:
      {
         ternary_expr * te = GetPointer<ternary_expr>(curr_tn);
         insert_usage_info(te->op0, stmt);
         insert_usage_info(te->op1, stmt);
         if(te->op2)
            insert_usage_info(te->op2, stmt);
         break;
      }
      case CASE_QUATERNARY_EXPRESSION:
      {
         quaternary_expr * qe = GetPointer<quaternary_expr>(curr_tn);
         insert_usage_info(qe->op0, stmt);
         insert_usage_info(qe->op1, stmt);
         if(qe->op2)
            insert_usage_info(qe->op2, stmt);
         if(qe->op3)
            insert_usage_info(qe->op3, stmt);
         break;
      }
      case var_decl_K:
      case result_decl_K:
      case parm_decl_K:
      case integer_cst_K:
      case real_cst_K:
      case string_cst_K:
      case vector_cst_K:
      case complex_cst_K:
      case field_decl_K:
      case exc_ptr_expr_K:
      case filter_expr_K:
      case label_decl_K:
      case gimple_label_K:
      case gimple_asm_K:
      case gimple_phi_K:
      case target_mem_ref_K:
      case target_mem_ref461_K:
      case function_decl_K:
      {
         break;
      }
      case ssa_name_K:
      {
         ssa_name * sn = GetPointer<ssa_name>(curr_tn);
         (sn->uses)++;
         sn->use_stmts.push_back(stmt);
         break;
      }
      case call_expr_K:
      {
         call_expr* ce = GetPointer<call_expr>(curr_tn);
         std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            insert_usage_info(*arg, stmt);
         }
         break;
      }
      case gimple_call_K:
      {
         gimple_call* ce = GetPointer<gimple_call>(curr_tn);
         std::vector<tree_nodeRef> & args = ce->args;
         std::vector<tree_nodeRef>::iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            insert_usage_info(*arg, stmt);
         }
         break;
      }
      default:
         THROW_ERROR(std::string("Node not supported (") + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + std::string("): ") + curr_tn->get_kind_text());
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Inserted usage info into node " + boost::lexical_cast<std::string>(GET_INDEX_NODE(tn)) + " (" + GET_NODE(tn)->get_kind_text() + "). Statement: " + boost::lexical_cast<std::string>(GET_INDEX_NODE(stmt)) + " (" + GET_NODE(stmt)->get_kind_text() + ")");
}

void tree_manager::merge_tree_managers(const tree_managerRef & TM_source)
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Starting merging of new tree_manager");
   /// a declaration is uniquely identified by the name and by the scope
   ///in case the decl_node has a mangle the associated identifier_node is unique
   ///a decl_node without name is not added to the symbol table
   ///a decl_node local to a function_decl (scpe is a function_decl) is not added to the symbol table
   ///a decl_node local to a type_node without name is not added to the symbol table
   ///a static decl_node is not added to the symbol table
   ///memory_tag, parm_decl, result_decl are not added to the symbol table
   ///declaration with type_node local to a function are not considered
   ///the key of the declaration symbol table is structured as "name--scope"
   ///the value of the declaration symbol table is the nodeID of the tree_node in the tree_manager
   std::tr1::unordered_map<std::string, unsigned int> global_decl_symbol_table;

   ///a type_node without name is not added to the symbol table
   ///a type_node local to a function_decl is not added to the symbol table
   ///the key of the type symbol table is structured as "name"
   ///the value of the type symbol table is the nodeID of the tree_node in the tree_manager
   std::tr1::unordered_map<std::string, unsigned int> global_type_symbol_table;

   ///this table is used to give a name to unqualified record or union
   std::tr1::unordered_map<unsigned int, std::string> global_type_unql_symbol_table;
   ///global static variable and function become global so we need some sort of uniquification
   std::tr1::unordered_set<std::string> static_symbol_table;
   std::tr1::unordered_set<std::string> static_function_header_symbol_table;
   null_deleter nullDel;
   tree_managerRef TM_this(this, nullDel);

   ///build the symbol tables of tree_node inheriting from decl_node and type_node
   std::string symbol_name;
   std::string symbol_scope;
   std::tr1::unordered_map<unsigned int, tree_nodeRef>::const_iterator ti_end = tree_nodes.end();
   for (std::tr1::unordered_map<unsigned int, tree_nodeRef>::const_iterator ti = tree_nodes.begin(); ti != ti_end; ++ti)
   {
      ///check for decl_node
      const tree_nodeRef & tn = ti->second;
      decl_node* dn = GetPointer<decl_node>(tn);
      if(dn)
      {
         if(check_for_decl(tn, TM_this, symbol_name, symbol_scope, ti->first, global_type_unql_symbol_table))
            continue;
         else
         {
            ///check for static
            if((GetPointer<function_decl>(tn) and GetPointer<function_decl>(tn)->static_flag) or (GetPointer<var_decl>(tn) and (GetPointer<var_decl>(tn)->static_flag or GetPointer<var_decl>(tn)->static_static_flag))              )
            {
               THROW_ASSERT((GetPointer<function_decl>(tn) and GetPointer<function_decl>(tn)->body) or (GetPointer<var_decl>(tn) and static_symbol_table.find(symbol_name+"-"+symbol_scope) == static_symbol_table.end()) or (not GetPointer<function_decl>(tn)->body and static_function_header_symbol_table.find(symbol_name+"-"+symbol_scope) == static_function_header_symbol_table.end()), "duplicated static symbol in the current tree_manager: "+symbol_name+"-"+symbol_scope+" " + boost::lexical_cast<std::string>(ti->first));
               if(GetPointer<function_decl>(tn) && !GetPointer<function_decl>(tn)->body)
                  static_function_header_symbol_table.insert(symbol_name+"-"+symbol_scope);
               else
                  static_symbol_table.insert(symbol_name+"-"+symbol_scope);
               continue;
            }
            ///check for function_decl undefined
            if(dn->get_kind() == function_decl_K && global_decl_symbol_table.find(symbol_name+"-"+symbol_scope) != global_decl_symbol_table.end())
            {
               if(GetPointer<function_decl>(tree_nodes.find(global_decl_symbol_table.find(symbol_name+"-"+symbol_scope)->second)->second)->body)
                  continue;
               if(GetPointer<function_decl>(tn)->undefined_flag)
                  continue;
               //else do overwrite
            } else if(dn->get_kind() != function_decl_K)
               THROW_ASSERT(global_decl_symbol_table.find(symbol_name+"-"+symbol_scope) == global_decl_symbol_table.end(), "duplicated symbol in global_decl_symbol_table: "+global_decl_symbol_table.find(symbol_name+"-"+symbol_scope)->first + " == " + boost::lexical_cast<std::string>(ti->first));

            global_decl_symbol_table[symbol_name+"-"+symbol_scope]=ti->first;
         }
      }
      else
      {
         if(check_for_type(tn, TM_this, symbol_name, symbol_scope, global_type_symbol_table, ti->first))
            continue;
         else
         {
            THROW_ASSERT(global_type_symbol_table.find(symbol_name) == global_type_symbol_table.end(), "duplicated symbol in global_type_symbol_table: "+global_type_symbol_table.find(symbol_name)->first);
            global_type_symbol_table[symbol_name]=ti->first;
            ///give a name to unql where possible
            if(tn->get_kind() == record_type_K && GetPointer<record_type>(tn)->unql && GetPointer<record_type>(tn)->qual <= 0)
            {
               global_type_symbol_table["u struct " + symbol_name]= GET_INDEX_NODE(GetPointer<record_type>(tn)->unql);
               global_type_unql_symbol_table[GET_INDEX_NODE(GetPointer<record_type>(tn)->unql)] = "u struct " + symbol_name;
            }
            else if(tn->get_kind() == union_type_K && GetPointer<union_type>(tn)->unql && GetPointer<union_type>(tn)->qual <= 0)
            {
               global_type_symbol_table["u union " + symbol_name]= GET_INDEX_NODE(GetPointer<union_type>(tn)->unql);
               global_type_unql_symbol_table[GET_INDEX_NODE(GetPointer<union_type>(tn)->unql)] = "u union " + symbol_name;
            }
         }
      }
   }

   std::tr1::unordered_map<unsigned int, std::string>::const_iterator gtust_i_end = global_type_unql_symbol_table.end();
   for(std::tr1::unordered_map<unsigned int, std::string>::const_iterator gtust_i = global_type_unql_symbol_table.begin(); gtust_i != gtust_i_end; gtust_i++)
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, boost::lexical_cast<std::string>(gtust_i->first) + "-u>" + gtust_i->second);
   std::tr1::unordered_map<std::string, unsigned int>::const_iterator gst_i_end = global_decl_symbol_table.end();
   for(std::tr1::unordered_map<std::string, unsigned int>::const_iterator gst_i = global_decl_symbol_table.begin(); gst_i != gst_i_end; gst_i++)
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, boost::lexical_cast<std::string>(gst_i->second) + "-decl>" + gst_i->first);

   gst_i_end = global_type_symbol_table.end();
   for(std::tr1::unordered_map<std::string, unsigned int>::const_iterator gst_i = global_type_symbol_table.begin(); gst_i != gst_i_end; gst_i++)
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, boost::lexical_cast<std::string>(gst_i->second) + "-type>" + gst_i->first);

   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Built table for this tree_manager");
   ///source static function declaration
   ///used to correctly rename declaration and definition of a function
   std::tr1::unordered_map<std::string, std::string> source_static_symbol_table;
   ///the key is the old index while the values is the new index
   std::tr1::unordered_map<unsigned int, unsigned int> remap;
   ///set of nodes that will be added to the current tree manager (this)
   std::tr1::unordered_set<unsigned int> not_yet_remapped;
   std::tr1::unordered_set<unsigned int> to_be_visited;
   tree_node_index_factory TNIF(remap, *this);

   ///remap tree_node from TM_source to this tree_manager
   ///first remap the types and then decl
   std::tr1::unordered_map<unsigned int, tree_nodeRef>::const_iterator ti_source_end = TM_source->tree_nodes.end();
   for (std::tr1::unordered_map<unsigned int, tree_nodeRef>::const_iterator ti = TM_source->tree_nodes.begin(); ti != ti_source_end; ++ti)
   {
      const tree_nodeRef & tn = ti->second;
      ///check for decl_node
      decl_node* dn = GetPointer<decl_node>(tn);
      if(!dn)
      {
         if(GetPointer<identifier_node>(tn))
         {
            identifier_node* id = GetPointer<identifier_node>(tn);
            unsigned int node_id = find_identifier_nodeID(id->strg);
            if(node_id)
               remap[ti->first] = node_id;
            else
               continue;
         }
         else if(check_for_type(tn, TM_source, symbol_name, symbol_scope, global_type_symbol_table, ti->first))
            continue;
         else
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, boost::lexical_cast<std::string>(ti->first) + "-ntype>" + symbol_name+"-"+symbol_scope);
            if(tn->get_kind() == record_type_K && GetPointer<record_type>(tn)->unql && GetPointer<record_type>(tn)->qual <= 0)
            {
               std::tr1::unordered_map<std::string, unsigned int>::const_iterator gst_it = global_type_symbol_table.find("u struct " + symbol_name);
               if(gst_it != global_type_symbol_table.end())
               {
                  remap[GET_INDEX_NODE(GetPointer<record_type>(tn)->unql)] = gst_it->second;
                  global_type_unql_symbol_table[GET_INDEX_NODE(GetPointer<record_type>(tn)->unql)] = "u struct " + symbol_name;
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, boost::lexical_cast<std::string>(GET_INDEX_NODE(GetPointer<record_type>(tn)->unql)) + "-nutype>" + symbol_name);
                  //std::cout << "remap unql: " + gst_it->first << " " << gst_it->second << std::endl;
               }
               else
               {
                  unsigned int new_tree_index;
                  if(remap.find(GET_INDEX_NODE(GetPointer<record_type>(tn)->unql)) == remap.end())
                  {
                     new_tree_index = new_tree_node_id(GET_INDEX_NODE(GetPointer<record_type>(tn)->unql));
                     not_yet_remapped.insert(GET_INDEX_NODE(GetPointer<record_type>(tn)->unql));
                     remap[GET_INDEX_NODE(GetPointer<record_type>(tn)->unql)] = new_tree_index;
                     to_be_visited.insert(GET_INDEX_NODE(GetPointer<record_type>(tn)->unql));
                  }
                  else
                  {
                     new_tree_index = remap[GET_INDEX_NODE(GetPointer<record_type>(tn)->unql)];
                  }
                  global_type_symbol_table["u struct " + symbol_name] = new_tree_index;
                  global_type_unql_symbol_table[GET_INDEX_NODE(GetPointer<record_type>(tn)->unql)] = "u struct " + symbol_name;
               }
            }
            else if(tn->get_kind() == union_type_K && GetPointer<union_type>(tn)->unql && GetPointer<union_type>(tn)->qual <= 0)
            {
               std::tr1::unordered_map<std::string, unsigned int>::const_iterator gst_it = global_type_symbol_table.find("u union " + symbol_name);
               if(gst_it != global_type_symbol_table.end())
               {
                  remap[GET_INDEX_NODE(GetPointer<union_type>(tn)->unql)] = gst_it->second;
                  global_type_unql_symbol_table[GET_INDEX_NODE(GetPointer<union_type>(tn)->unql)] = "u union " + symbol_name;
                  INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, boost::lexical_cast<std::string>(GET_INDEX_NODE(GetPointer<union_type>(tn)->unql)) + "-nutype>" + symbol_name);
                  //std::cout << "remap unql: " + gst_it->first << std::endl;
               }
               else
               {
                  unsigned int new_tree_index;
                  if(remap.find(GET_INDEX_NODE(GetPointer<union_type>(tn)->unql)) == remap.end())
                  {
                     new_tree_index = new_tree_node_id(GET_INDEX_NODE(GetPointer<union_type>(tn)->unql));
                     remap[GET_INDEX_NODE(GetPointer<union_type>(tn)->unql)] = new_tree_index;
                     not_yet_remapped.insert(GET_INDEX_NODE(GetPointer<union_type>(tn)->unql));
                     to_be_visited.insert(GET_INDEX_NODE(GetPointer<union_type>(tn)->unql));
                  }
                  else
                  {
                     new_tree_index = remap[GET_INDEX_NODE(GetPointer<union_type>(tn)->unql)];
                  }
                  global_type_symbol_table["u union " + symbol_name] = new_tree_index;
                  global_type_unql_symbol_table[GET_INDEX_NODE(GetPointer<union_type>(tn)->unql)] = "u union " + symbol_name;
               }
            }
            std::tr1::unordered_map<std::string, unsigned int>::const_iterator gst_it = global_type_symbol_table.find(symbol_name);
            if(gst_it == global_type_symbol_table.end())
            {
               if(remap.find(ti->first) == remap.end())
               {
                  unsigned int new_tree_index = new_tree_node_id(ti->first);
                  global_type_symbol_table[symbol_name] = remap[ti->first] = new_tree_index;
                  not_yet_remapped.insert(ti->first);
                  to_be_visited.insert(ti->first);
               }
               continue;
            }
            else
            {
               //record type
               if(tn->get_kind() == record_type_K)
               {
                  //present in this tree_manager
                  if(tree_nodes.find(gst_it->second) != tree_nodes.end())
                  {
                     if(GetPointer<record_type>(tree_nodes.find(gst_it->second)->second) && GetPointer<record_type>(tree_nodes.find(gst_it->second)->second)->list_of_flds.empty() && !GetPointer<record_type>(tn)->list_of_flds.empty())
                     {
                        not_yet_remapped.insert(ti->first); ///overwrite gst_it->second
                        to_be_visited.insert(ti->first);
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "RECORD TYPE: " + gst_it->first+ " " + boost::lexical_cast<std::string>(ti->first));
                     }
                  }
                  else
                  {
                     THROW_ASSERT(TM_source->tree_nodes.find(gst_it->second) != TM_source->tree_nodes.end(), "There is a symbol which is not present in this nor tree_manager nor the other");
                     if(GetPointer<record_type>(TM_source->tree_nodes.find(gst_it->second)->second)->list_of_flds.empty() && !GetPointer<record_type>(tn)->list_of_flds.empty())
                     {
                        not_yet_remapped.insert(ti->first); ///overwrite gst_it->second
                        to_be_visited.insert(ti->first);
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "RECORD TYPE: " + gst_it->first+ " " + boost::lexical_cast<std::string>(ti->first));
                     }
                  }
               }
               if(tn->get_kind() == union_type_K)
               {
                  //present in this tree_manager
                  if(tree_nodes.find(gst_it->second) != tree_nodes.end())
                  {
                     if(GetPointer<union_type>(tree_nodes.find(gst_it->second)->second) && GetPointer<union_type>(tree_nodes.find(gst_it->second)->second)->list_of_flds.empty() && !GetPointer<union_type>(tn)->list_of_flds.empty())
                     {
                        not_yet_remapped.insert(ti->first); ///overwrite gst_it->second
                        to_be_visited.insert(ti->first);
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "UNION TYPE: " + gst_it->first+ " " + boost::lexical_cast<std::string>(ti->first));
                     }
                  }
                  else
                  {
                     THROW_ASSERT(TM_source->tree_nodes.find(gst_it->second) != TM_source->tree_nodes.end(), "There is a symbol which is not present in this nor tree_manager nor the other");
                     if(GetPointer<union_type>(TM_source->tree_nodes.find(gst_it->second)->second)->list_of_flds.empty() && !GetPointer<union_type>(tn)->list_of_flds.empty())
                     {
                        not_yet_remapped.insert(ti->first); ///overwrite gst_it->second
                        to_be_visited.insert(ti->first);
                        INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "UNION TYPE: " + gst_it->first+ " " + boost::lexical_cast<std::string>(ti->first));
                     }
                  }
               }
               remap[ti->first] = gst_it->second;
            }
         }
      }
   }

   ///remap tree_node from TM_source to this tree_manager
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Remapping declaration nodes");
   ti_source_end = TM_source->tree_nodes.end();
   for (std::tr1::unordered_map<unsigned int, tree_nodeRef>::const_iterator ti = TM_source->tree_nodes.begin(); ti != ti_source_end; ++ti)
   {
      const tree_nodeRef & tn = ti->second;
      ///check for decl_node
      decl_node* dn = GetPointer<decl_node>(tn);
      if(dn)
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Examining declaration node " + boost::lexical_cast<std::string>(ti->first));
         if(check_for_decl(tn, TM_source, symbol_name, symbol_scope, ti->first, global_type_unql_symbol_table))
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Not suitable for symbol table");
            continue;
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, boost::lexical_cast<std::string>(ti->first) + "-ndecl>" + symbol_name+"-"+symbol_scope);
         ///check for static
         if((GetPointer<function_decl>(tn) and GetPointer<function_decl>(tn)->static_flag) or (GetPointer<var_decl>(tn) and (GetPointer<var_decl>(tn)->static_flag or GetPointer<var_decl>(tn)->static_static_flag)))
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Static");
            if(source_static_symbol_table.find(symbol_name) != source_static_symbol_table.end())
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Already present in source static symbol table");
               unsigned int node_id_source = TM_source->find_identifier_nodeID(source_static_symbol_table.find(symbol_name)->second);
               dn->name = tree_nodeRef(new tree_reindex(node_id_source, TM_source->add(node_id_source)));
            }
            else if(static_symbol_table.find(symbol_name+"-"+symbol_scope) != static_symbol_table.end() or static_function_header_symbol_table.find(symbol_name+"-"+symbol_scope) != static_function_header_symbol_table.end())
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Already present in destination static symbol table");
               ///static function or var_decl with a conflicting name
               ///so we fix the name...
               ///and we fix in the source tree_manager
               unsigned int counter = 0;
               unsigned int node_id_this, node_id_source;
               do
               {
                  counter++;
                  node_id_this = find_identifier_nodeID(symbol_name+boost::lexical_cast<std::string>(counter));
                  node_id_source = TM_source->find_identifier_nodeID(symbol_name+boost::lexical_cast<std::string>(counter));
               } while((node_id_this>0 || node_id_source>0));
               std::map<treeVocabularyTokenTypes::token_enum, std::string> tree_node_schema;
               tree_node_schema[TOK(TOK_STRG)]= symbol_name+boost::lexical_cast<std::string>(counter);
               node_id_source = TM_source->new_tree_node_id(ti->first);
               TM_source->create_tree_node(node_id_source, TOK(TOK_IDENTIFIER_NODE), tree_node_schema);
               dn->name = tree_nodeRef(new tree_reindex(node_id_source, TM_source->add(node_id_source)));
               source_static_symbol_table[symbol_name] = symbol_name+boost::lexical_cast<std::string>(counter);
            }
            else
            {
               if(not GetPointer<function_decl>(tn))
                  static_symbol_table.insert(symbol_name+"-"+symbol_scope);
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
               continue;
            }
         }
         else
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Not Static");
            std::tr1::unordered_map<std::string, unsigned int>::const_iterator gst_it =  global_decl_symbol_table.find(symbol_name+"-"+symbol_scope);
            if(gst_it == global_decl_symbol_table.end())
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Not yet in global decl symbol table");
               unsigned int new_index;
               new_index = new_tree_node_id(ti->first);
               global_decl_symbol_table[symbol_name+"-"+symbol_scope] = remap[ti->first] = new_index;
               not_yet_remapped.insert(ti->first);
               to_be_visited.insert(ti->first);
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
               continue;
            }
            else
            {
               INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Already in global decl symbol table");
               if
               (
                  (
                     tn->get_kind() == function_decl_K
                     && 
                     (
                        !GetPointer<function_decl>(tree_nodes.find(gst_it->second)->second) 
                        ||
                        !GetPointer<function_decl>(tree_nodes.find(gst_it->second)->second)->body
                     )
                     &&
                     GetPointer<function_decl>(tn)->body
                  )
                  ||
                  //In the first tree manager there is extern type variable; , in the second there is the definition
                  (
                     tn->get_kind() == var_decl_K
                     &&
                     GetPointer<var_decl>(tree_nodes.find(gst_it->second)->second)
                     &&
                     GetPointer<var_decl>(tree_nodes.find(gst_it->second)->second)->extern_flag
                     &&
                     !GetPointer<var_decl>(tn)->extern_flag
                  )
                  ||
                  //In the first tree manager there is a function_decl without srcp, in the second there is a funcion_decl with srcp
                  (
                     tn->get_kind() == function_decl_K
                     and
                     GetPointer<function_decl>(tree_nodes.find(gst_it->second)->second)
                     and
                     (GetPointer<function_decl>(tree_nodes.find(gst_it->second)->second))->include_name == "<built-in>"
                     and
                     (GetPointer<function_decl>(tn))->include_name.size()
                     and
                     (GetPointer<function_decl>(tn))->include_name != "<built-in>"
                  )
               )
               {
                  not_yet_remapped.insert(ti->first);///overwrite gst_it->second
                  to_be_visited.insert(ti->first);
               }
               remap[ti->first] = gst_it->second;
            }
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
      }
      else
      {
         ///already performed
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Remapped declaration nodes");
   ///tree node visitor
   tree_node_reached TNR(remap, not_yet_remapped, *this);
   std::tr1::unordered_set<unsigned int>::const_iterator it_to_be_visited_end = to_be_visited.end();
   for(std::tr1::unordered_set<unsigned int>::const_iterator it_to_be_visited = to_be_visited.begin(); it_to_be_visited_end != it_to_be_visited; it_to_be_visited++)
      TM_source->get_tree_node_const(*it_to_be_visited)->visit(&TNR);

   ///compute the vertexes reached from all function_decl of TM_source
   for(std::map<unsigned int, tree_nodeRef>::const_iterator it = TM_source->function_decl_nodes.begin(); it != TM_source->function_decl_nodes.end(); it++)
   {
      tree_nodeRef curr_tn = it->second;
      function_decl * fd = GetPointer<function_decl>(curr_tn);
      if(remap.find(it->first) == remap.end())
      {
         remap[it->first] = new_tree_node_id(it->first);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Function decl: old " + boost::lexical_cast<std::string>(it->first) + " - new " + boost::lexical_cast<std::string>(remap[it->first]));
         fd->visit(&TNR);
         not_yet_remapped.insert(it->first);
      }
      else
      {
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Function decl visited: " + boost::lexical_cast<std::string>(remap[it->first]) + "(" + boost::lexical_cast<std::string>(it->first) + ") - " + fd->get_function_name());
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Already remapped tree_node");
   std::tr1::unordered_map<unsigned int, unsigned int>::const_iterator it_remap_end = remap.end();
   for(std::tr1::unordered_map<unsigned int, unsigned int>::const_iterator it_remap = remap.begin(); it_remap_end != it_remap; it_remap++)
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Original " + boost::lexical_cast<std::string>(it_remap->first) + " New " + boost::lexical_cast<std::string>(it_remap->second));

   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Starting remapping remaining nodes");
   std::tr1::unordered_set<unsigned int>::const_iterator it_not_yet_remapped_end = not_yet_remapped.end();
   for(std::tr1::unordered_set<unsigned int>::const_iterator it_not_yet_remapped = not_yet_remapped.begin(); it_not_yet_remapped_end != it_not_yet_remapped; it_not_yet_remapped++)
   {
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Original " + boost::lexical_cast<std::string>(*it_not_yet_remapped) + " New " + boost::lexical_cast<std::string>(remap[*it_not_yet_remapped]));
      TNIF.create_tree_node(remap[*it_not_yet_remapped], TM_source->get_tree_node_const(*it_not_yet_remapped));
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Type is " + std::string(TM_source->get_tree_node_const(*it_not_yet_remapped)->get_kind_text()));
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "DONE");
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Ended merging of new tree_manager");
}

bool tree_manager::check_for_decl(const tree_nodeRef & tn, const tree_managerRef &TM_this, std::string & symbol_name, std::string & symbol_scope, unsigned int DEBUG_PARAMETER(node_id), const std::tr1::unordered_map<unsigned int, std::string> &global_type_unql_symbol_table)
{
   THROW_ASSERT(GetPointer<decl_node>(tn), "Node should be a declaration node");
   const decl_node * dn = GetPointer<decl_node>(tn);
   ///check for name
   if(!dn->name)
      return true;
   /// check for memory_tag, parm_decl, result_decl
   if(GetPointer<memory_tag>(tn) || dn->get_kind()==parm_decl_K || dn->get_kind()==result_decl_K)
      return true;
   ///check for scope
   if(dn->scpe && GET_NODE(dn->scpe)->get_kind() == function_decl_K)
      return true;
   if(dn->scpe && GetPointer<type_node>(GET_NODE(dn->scpe)) && (!GetPointer<type_node>(GET_NODE(dn->scpe))->name && global_type_unql_symbol_table.find(GET_INDEX_NODE(dn->scpe)) == global_type_unql_symbol_table.end()))
      return true;
   if(dn->mngl)
   {
      THROW_ASSERT(GET_NODE(dn->mngl)->get_kind() == identifier_node_K, "expected an identifier_node: " + boost::lexical_cast<std::string>(GET_NODE(dn->mngl)->get_kind_text()));
      symbol_name = GetPointer<identifier_node>(GET_NODE(dn->mngl))->strg;
   }
   else
   {
      THROW_ASSERT(GET_NODE(dn->name)->get_kind() == identifier_node_K, "expected an identifier_node: " + boost::lexical_cast<std::string>(GET_NODE(dn->name)->get_kind_text()) + " " + boost::lexical_cast<std::string>(node_id));
      symbol_name = GetPointer<identifier_node>(GET_NODE(dn->name))->strg;
   }
   if(dn->scpe && GetPointer<type_node>(GET_NODE(dn->scpe)) && GetPointer<type_node>(GET_NODE(dn->scpe))->name)
   {
      type_node* type = GetPointer<type_node>(GET_NODE(dn->scpe));
      THROW_ASSERT(type, "expected a type_node: "+std::string(GET_NODE(dn->scpe)->get_kind_text()));
      ///declaration with type_node local to a function are not considered
      if(type->scpe && GET_NODE(type->scpe)->get_kind() == function_decl_K)
         return true;
      unsigned int quals = type->qual;
      std::string type_name;
      if(GET_NODE(type->name)->get_kind() == identifier_node_K)
         type_name = GetPointer<identifier_node>(GET_NODE(type->name))->strg;
      else
         type_name = tree_helper::get_type_name(TM_this, GET_INDEX_NODE(type->name));
      if (type->name && (GET_NODE(type->name)->get_kind() == type_decl_K))
      {
         if(quals > 0)
            symbol_scope = tree_helper::return_qualifier_prefix(quals) + type_name;
         else
            symbol_scope = type_name;
      }
      else if (type->name)
         symbol_scope = "struct " + type_name;
   }
   else if(dn->scpe && GetPointer<type_node>(GET_NODE(dn->scpe)) && !GetPointer<type_node>(GET_NODE(dn->scpe))->name && global_type_unql_symbol_table.find(GET_INDEX_NODE(dn->scpe)) != global_type_unql_symbol_table.end())
   {
      symbol_scope = global_type_unql_symbol_table.find(GET_INDEX_NODE(dn->scpe))->second;
   }
   else
      symbol_scope = "";
   return false;
}

bool tree_manager::check_for_type(const tree_nodeRef & tn, const tree_managerRef &TM, std::string & symbol_name, std::string & , const std::tr1::unordered_map<std::string, unsigned int> &global_type_symbol_table, unsigned int node_id)
{
   PRINT_DBG_MEX(DEBUG_LEVEL_PARANOIC, debug_level, "Checking for type " + boost::lexical_cast<std::string>(node_id));
   type_node* type = GetPointer<type_node>(tn);
   if(!type || !type->name)///integer_type and real_type have some duplications
      return true;
   if(type->scpe && ((GET_NODE(type->scpe)->get_kind() == function_decl_K )))
      return true;
   unsigned int quals = type->qual;
   std::string type_name;
   if(GET_NODE(type->name)->get_kind() == identifier_node_K)
      type_name = GetPointer<identifier_node>(GET_NODE(type->name))->strg;
   else
      type_name = tree_helper::get_type_name(TM, GET_INDEX_NODE(type->name));
   if (type->name && (GET_NODE(type->name)->get_kind() == type_decl_K || type->unql))
   {
      if(quals > 0)
         symbol_name = tree_helper::return_qualifier_prefix(quals) + type_name;
      else
         symbol_name = type_name;
   }
   else if (type->name && tn->get_kind() == record_type_K)
      symbol_name = "struct " + type_name;
   else if (type->name && tn->get_kind() == union_type_K)
      symbol_name = "union " + type_name;
   else if (type->name && tn->get_kind() == enumeral_type_K)
      symbol_name = "enum " + type_name;
   else if(type->name)
      symbol_name = type_name;
   else
      return true;

   if(global_type_symbol_table.find(symbol_name) != global_type_symbol_table.end() and not tree_helper::HasToBeDeclared(TM, node_id))
      return true;

   return false;
}

unsigned int tree_manager::find_identifier_nodeID(const std::string &str) const
{
   std::tr1::unordered_map<std::string, unsigned int>::const_iterator it = identifiers_unique_table.find(str);
   if(it == identifiers_unique_table.end())
      return 0;
   else
      return it->second;
}

void tree_manager::add_identifier_node(unsigned int nodeID, const bool & DEBUG_PARAMETER(op))
{
   THROW_ASSERT(op, "improper use of add_identifier_node");
   identifiers_unique_table[STOK(TOK_OPERATOR)]=nodeID;
}

void tree_manager::get_functions(std::list<unsigned int> &functions_list) const 
{
   std::map<unsigned int, tree_nodeRef>::const_iterator beg, end;
   for(beg = function_decl_nodes.begin(), end = function_decl_nodes.end(); beg != end; beg++)
   {
      functions_list.push_back(beg->first);
   }
}

unsigned int tree_manager::get_next_vers()
{
   if(next_vers == 0){
      std::tr1::unordered_map<unsigned int, tree_nodeRef>::const_iterator it, it_end = tree_nodes.end();
      for(it = tree_nodes.begin(); it != it_end; it++)
      {
         if(it->second)
         {
            tree_nodeRef tn = it->second;
            if(tn->get_kind() == ssa_name_K)
            {
               ssa_name * sn = GetPointer<ssa_name>(tn);
               if(sn->vers > next_vers)
                  next_vers = sn->vers;
            }
         }
      }
   }
   return ++next_vers;
}

void tree_manager::add_goto()
{
   added_goto++;
}

unsigned int tree_manager::get_added_goto() const
{
   return added_goto;
}

void tree_manager::increment_removed_pointer_plus()
{
   removed_pointer_plus++;
}

unsigned int tree_manager::get_removed_pointer_plus() const
{
   return removed_pointer_plus;
}

void tree_manager::increment_removable_pointer_plus()
{
   removable_pointer_plus++;
}

unsigned int tree_manager::get_removable_pointer_plus() const
{
   return removable_pointer_plus;
}

void tree_manager::increment_unremoved_pointer_plus()
{
   unremoved_pointer_plus++;
}

unsigned int tree_manager::get_unremoved_pointer_plus() const
{
   return unremoved_pointer_plus;
}
