/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file var_pp_functor.hpp
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 */

#ifndef _VAR_PP_FUNCTOR_HPP
#define _VAR_PP_FUNCTOR_HPP

///STL include
#include "tr1/unordered_set"

///Utility include
#include "refcount.hpp"

#include "graph.hpp"

/**
 * @name Forward declarations.
 */
//@{
REF_FORWARD_DECL(Parameter);
CONSTREF_FORWARD_DECL(behavioral_helper);
REF_FORWARD_DECL(var_pp_functor);
//@}


/**
 * Base class functor used by prettyPrintVertex to print variables
*/
struct var_pp_functor
{
   /// Destructor
   virtual ~var_pp_functor() {}

   /**
    * This functor returns a string representing the variable (usually the name of the variable). This can be used both in variable declaration and in variable use.
    * The string returned depends on the type of manipulation performed by the backend layer.
    * @param var is the nodeid of the variable that should be analyzed.
    */
   virtual std::string operator() (unsigned int var) const = 0;

};

typedef refcount<var_pp_functor> var_pp_functorRef;

/**
 * Standard functor that returns the name of a variable
*/
struct std_var_pp_functor : public var_pp_functor
{
private:
   ///behavioral helper
   const behavioral_helperConstRef BH;

public:
   ///Constructor
   std_var_pp_functor(const behavioral_helperConstRef _BH): BH(_BH) {}

   ///Destructor
   ~std_var_pp_functor() {}

   /**
    * return the name of the variable.
    * @param var is the nodeid of the variable.
   */
   std::string operator() (unsigned int var) const;

};


/**
* Pointer version functor that returns the name of a variable with a star in front.
*/
struct pointer_var_pp_functor : public var_pp_functor
{
   /**
    * Constructor
    * @param BH is the behavioral helper.
    * @param vars is the reference to the set of variables for which a star should be returned along with their name.
    * @param add_restrict controls the addition to the parameters declarations of the __restrict__ keyword.
   */
   pointer_var_pp_functor(const behavioral_helperConstRef _BH, const std::tr1::unordered_set<unsigned int> vars, bool _add_restrict = false);

   /**
    * Destructor
    */
   ~pointer_var_pp_functor() {}

   /**
    * return the name of the variable with a star as a prefix.
    * @param var is the nodeid of the variable.
   */
   std::string operator() (unsigned int var) const;

private:
   ///reference to the set of variable that has to have a star in front when returned by operator()
   const std::tr1::unordered_set<unsigned int> pointer_based_variables;

   ///behavioral helper
   const behavioral_helperConstRef BH;

   ///standard functor used for print array variable
   const std_var_pp_functor std_functor;

   /// it controls the addition to the parameters declarations of the __restrict__ keyword.
   bool add_restrict;
};

/**
* Address version functor that returns the name of a variable with "&" in front.
*/
struct address_var_pp_functor : public var_pp_functor
{
public:
   /**
    * Constructor
    * @param BH is the behavioral helper.
    * @param vars is the reference to the set of variables for which a star should be returned along with their name.
    */
   address_var_pp_functor(const behavioral_helperConstRef _BH, const std::tr1::unordered_set<unsigned int> vars, const std::tr1::unordered_set<unsigned int> pointer_vars);

   ///Destructor
   ~address_var_pp_functor() {}

   /**
    * return the name of the variable with a star as a prefix.
    * @param var is the nodeid of the variable.
   */
   std::string operator() (unsigned int var) const;

   void setReadOnly()
   {
      readOnly = true;
   }
   void unsetReadOnly()
   {
      readOnly = false;
   }

private:
   ///reference to the set of variable that has to have a star in front when returned by operator()
   const std::tr1::unordered_set<unsigned int> addr_based_variables;

   ///reference to the set of variable that has to have a star in front when returned by operator()
   const std::tr1::unordered_set<unsigned int> pointer_based_variables;

   ///behavioral helper
   const behavioral_helperConstRef BH;

   bool readOnly;
};


#endif

