/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file fileIO.hpp
 * @brief utility function used to read files.
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10649 $
 * $Date: 2012-03-19 17:14:01 +0100 (Mon, 19 Mar 2012) $
 * Last modified by $Author: pilato $
 *
*/
#ifndef FILEIO_HPP
#define FILEIO_HPP

///STD include
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>

#include "gzstream.hpp"

///Parameter include
#include "Parameter.hpp"

///Utility include
#include <boost/filesystem.hpp>
#include "exceptions.hpp"
#include "refcount.hpp"

/**
 * RefCount type definition for the input stream object.
 */
typedef refcount<std::istream> fileIO_istreamRef;
typedef refcount<const std::istream> fileIO_istreamConstRef;

/**
 * RefCount type definition for the input stream object.
 */
typedef refcount<std::ostream> fileIO_ostreamRef;

/**
 * this function returns an istream compressed or not.
 * It first check for a compressed file, then search for the compressed version of the file and finally in case no compressed file is found it look for the plain text file.
 * this function is mainly based on the gzstream wrapper and on zlib library.
 * @param name is the file name.
 * @return the refcount to the istream
 */
inline fileIO_istreamRef fileIO_istream_open(const std::string name)
{
   fileIO_istreamRef res_file;
   res_file = fileIO_istreamRef(new igzstream((name).c_str()));
   if (res_file->fail())
   {
      res_file = fileIO_istreamRef(new igzstream((name + ".gz").c_str()));
      if (res_file->fail())
      {
         res_file = fileIO_istreamRef(new igzstream((name + ".Z").c_str()));
         if (res_file->fail())
         {
            res_file = fileIO_istreamRef(new std::ifstream(name.c_str()));
            if(res_file->fail())
            {
               THROW_ERROR("Error in opening " + name);
            }
         }
      }
   }
   return res_file;
}

/**
 * Create a fileIO_istreamRef starting from a string
 */
inline fileIO_istreamRef fileIO_istream_open_from_string(const std::string input)
{
   fileIO_istreamRef output;
   output = fileIO_istreamRef(new std::istringstream(input));
   return output;
}

/**
 * this function returns an ostream compressed or not.
 * this function is mainly based on the gzstream wrapper and on zlib library.
 * @param name is the file name.
 * @return the refcount to the ostream
 */
inline fileIO_ostreamRef fileIO_ostream_open(const std::string name)
{
   fileIO_ostreamRef res_file;
   res_file = fileIO_ostreamRef(new ogzstream((name).c_str()));
   return res_file;
}

#define EXEEXT std::string("")

/**
 * Copy a file to the standard output
*/
inline void CopyStdout(const std::string filename)
{
   FILE *filese;
   filese = fopen(filename.c_str(), "r");
   char buffer[255];
   size_t nBytes;
   while( (nBytes = fread(buffer,1,sizeof(buffer),filese)) > 0)
   {
      size_t wBytes = fwrite(buffer,1,nBytes,stdout);
      if (wBytes < nBytes) break;
   }
   fclose(filese);
}

inline
void rename_file(boost::filesystem::path from, boost::filesystem::path to)
{
#if BOOST_VERSION >= 104600
   ///FIXME: this part has not yet been tested
   boost::system::error_code return_error;
   boost::filesystem::rename(from, to, return_error);
   if(return_error.value())
   {
      boost::filesystem::remove(to);
      boost::filesystem::copy_file(from, to);
      boost::filesystem::remove(from);
   }
#else
   try
   {
      boost::filesystem::rename(from, to);
   }
   catch (const boost::filesystem::basic_filesystem_error<boost::filesystem::path>)
   {
      boost::filesystem::remove(to);
      boost::filesystem::copy_file(from, to);
      boost::filesystem::remove(from);
   }
#endif
};

/**
 * Return the filename (base + extension) without the path
 * @param file is the starting file
 * @return the file without the path
 */
inline std::string GetLeafFileName(boost::filesystem::path file)
{
#if BOOST_VERSION >= 104600
    return file.filename().string();
#else
    return file.leaf();
#endif
}

/**
 * Return the file without the extension
 * @param file is the starting file
 * @return the file without path and extension
 */
inline std::string GetBaseName(boost::filesystem::path file)
{
   std::string basename;
#if BOOST_VERSION >= 104600
   basename = file.stem().string();
#else
   basename = file.stem();
#endif
   return basename;
}

/**
 * Return the extenstion of the file
 * @param file is the starting file
 * @return the extension of the file
 */
inline std::string GetExtension(boost::filesystem::path file)
{
   std::string dot_extension;
#if BOOST_VERSION >= 104600
   dot_extension = file.extension().string();
#else
   dot_extension = file.extension();
#endif
   if(dot_extension.at(0) == '.')
      dot_extension = dot_extension.substr(1, dot_extension.size() - 1);
   return dot_extension;
}

inline std::string GetCurrentPath()
{
   return boost::filesystem::current_path().string();
}

/**
 * Copy file; if target already exist, overwrite
 * @param file_source is the file to be copied
 * @param file_target is the destination
 */
inline void CopyFile(boost::filesystem::path file_source, boost::filesystem::path file_target)
{
   if(file_source.string()=="-")
   {
      std::string line;
      std::ofstream new_file(file_target.string().c_str());
      while (std::cin)
      {
         std::getline(std::cin, line);
         new_file << line << std::endl;
      }
   }
   else
      boost::filesystem::copy_file(file_source, file_target, boost::filesystem::copy_option::overwrite_if_exists);
}

/**
 * System call forcing execution with bash
 * @param Param is the set of input parameters
 * @param system_command is the  to be executed
 * @param output is the file where output has to be saved
 * @return the value returned by the shell
 */
inline int PandaSystem(const ParameterConstRef Param, const std::string & system_command, std::string output = "")
{
   static size_t counter = 0;
   if(output == "")
      output = Param->getOption<std::string>(OPT_output_temporary_directory) + STR_CST_shell_output_file;
   const std::string script_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + STR_CST_shell_script + "_" + boost::lexical_cast<std::string>(counter++);
   std::ofstream script_file(script_file_name.c_str());
   script_file << "#!/bin/bash" << std::endl;
   script_file << system_command;
   if(Param->getOption<unsigned int>(OPT_output_level) >= OUTPUT_LEVEL_PEDANTIC)
   {
      script_file << " 2>&1 | tee " << output << std::endl;
   }
   else
   {
      script_file << " > " << output << " 2>&1 " << std::endl;
   }
   script_file << std::endl;
   script_file.close();
   std::string command = "bash -f " + script_file_name + "";
   return system(command.c_str());
}
#endif
