/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/

/**
 * @file flopoco_wrapper.cpp
 * @brief Implementation of the wrapper to FloPoCo for VHDL code generation.
 *
 * Implementation of the object used to invoke the FloPoCo framework
 * to generate VHDL code for floating-point functional units
 *
 * @author Daniele Mastrandrea <daniele.mastrandrea@mail.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Autoheader include
#include "config_SKIP_WARNING_SECTIONS.hpp"

#if SKIP_WARNING_SECTIONS
#pragma GCC diagnostic ignored "-Wshadow"
#pragma GCC diagnostic ignored "-Wconversion"
#pragma GCC diagnostic ignored "-Wsign-conversion"
#pragma GCC diagnostic ignored "-Wredundant-decls"
#pragma GCC diagnostic ignored "-Wunused-parameter"
#endif

/// Includes the class definition
#include "flopoco_wrapper.hpp"

/// Standard PandA include
#include "dbgPrintHelper.hpp"
#include "utility.hpp"
#include "simple_indent.hpp"
#include "exceptions.hpp"

#include "technology_builtin.hpp"

/// Standard include
#include <unistd.h>
#include <cerrno>

/// Streams include
#include <iostream>
#include <fstream>

/// STL include
#include <vector>
#include <tr1/unordered_map>
#include <tr1/unordered_set>

/// BOOST include
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>

/// FloPoCo include
#undef DEBUG
#include "Operator.hpp"
#include "Target.hpp"
#include "Targets/Spartan3.hpp"
#include "Targets/Virtex4.hpp"
#include "Targets/Virtex5.hpp"
#include "Targets/Virtex6.hpp"
#include "Targets/StratixII.hpp"
#include "Targets/StratixIII.hpp"
#include "Targets/StratixIV.hpp"
#include "FPMultiplier.hpp"
#include "FPAdderSinglePath.hpp"
#include "Fix2FP.hpp"
#include "FPExp.hpp"
#include "FPSqrtPoly.hpp"
#include "FPLog.hpp"
#include "FPDiv.hpp"
#include "FP2Fix.hpp"
#include "InputIEEE.hpp"
#include "OutputIEEE.hpp"
#include "FPAssign.hpp"
#include "FPgt_expr.hpp"
#include "FPlt_expr.hpp"
#include "FPge_expr.hpp"
#include "FPle_expr.hpp"

#define NO_PIPELINE

/// Redefinition of FloPoCo's global variables, declared in flopoco/src/main.cpp
/// None of the following is actually used inside the wrapper

#ifdef NEW_FLOPOCO
#define OPLIST oplist
#else
#define OPLIST flopoco::oplist
#endif

namespace flopoco
{
#ifndef NEW_FLOPOCO
   std::vector<Operator *> oplist;
#endif
   std::string filename = "flopoco.vhdl";
   std::string cl_name = "";
#ifndef NEW_FLOPOCO
   int verbose = 0;
#endif
   Target * target=0;
   int LongAccN=0;
   /* flex vars */
   int yyTheCycle;
   vector<pair<string, int> > theUseTable;
#ifndef NEW_FLOPOCO
   std::map<string, double> emptyDelayMap;
#endif
   bool combinatorialOperator;
#ifndef NEW_FLOPOCO
   int Operator::uid = 0; //init of the uid static member of Operator
#endif

}

void flopoco_wrapper::add_FU(const std::string & FU_type, const unsigned int FU_prec_in, const unsigned int FU_prec_out, const std::string & FU_name, const std::string & FU_target)
{
   // Get the number of bits for the number representation
   unsigned int n_mant_in, n_exp_in;
   unsigned int n_mant_out, n_exp_out;
   this->DECODE_BITS(FU_prec_in, n_mant_in, n_exp_in);
   this->DECODE_BITS(FU_prec_out, n_mant_out, n_exp_out);

   // Get the target architecture
   if ("Spartan-3" == FU_target)
      flopoco::target = new flopoco::Spartan3();
   else if ("Virtex-4" == FU_target)
      flopoco::target = new flopoco::Virtex4();
   else if ("Virtex-5" == FU_target)
      flopoco::target = new flopoco::Virtex5();
   else if ("Virtex-6" == FU_target)
      flopoco::target = new flopoco::Virtex6();
   else if ("Virtex-7" == FU_target) /// does not exist so we use Virtex 6 target
      flopoco::target = new flopoco::Virtex6();
   else if ("CycloneII" == FU_target) /// does not exist so we use StratixII target
      flopoco::target = new flopoco::StratixII();
   else if ("StratixII" == FU_target)
      flopoco::target = new flopoco::StratixII();
   else if ("StratixIII" == FU_target)
      flopoco::target = new flopoco::StratixIII();
   else if ("StratixIV" == FU_target)
      flopoco::target = new flopoco::StratixIV();
   else
      THROW_ASSERT(false, "Non supported target architecture.");

#ifdef NO_PIPELINE
   flopoco::target->setNotPipelined();
#else
   flopoco::target->setPipelined();
#endif

   double freq = DEFAULT_TARGET_FREQUENCY;
   ///set the target frequency
   flopoco::target->setFrequency(1e6*freq);

   flopoco::Operator * op;
   THROW_ASSERT(n_mant_in > 0 && n_exp_in > 0, "Unsupported significand and exponent values.");
   THROW_ASSERT(n_mant_out > 0 && n_exp_out > 0, "Unsupported significand and exponent values.");

   // Get the Functional Unit, sets correct name, then adds to resources
   type = flopoco_wrapper::UT_UNKNOWN;
   if ("FPAdder" == FU_type)
   {
      type = flopoco_wrapper::UT_ADD;
      op = new flopoco::FPAdderSinglePath(flopoco::target, (int)n_exp_in, (int)n_mant_in, (int)n_exp_in, (int)n_mant_in, (int)n_exp_out, (int)n_mant_out);
   }
   else if ("FPSub" == FU_type)
   {
      type = flopoco_wrapper::UT_SUB;
      op = new flopoco::FPAdderSinglePath(flopoco::target, (int)n_exp_in, (int)n_mant_in, (int)n_exp_in, (int)n_mant_in, (int)n_exp_out, (int)n_mant_out);
   }
   else if ("FPAddSub" == FU_type)
   {
      type = flopoco_wrapper::UT_ADDSUB;
      op = new flopoco::FPAdderSinglePath(flopoco::target, (int)n_exp_in, (int)n_mant_in, (int)n_exp_in, (int)n_mant_in, (int)n_exp_out, (int)n_mant_out);
   }
   else if ("FPMultiplier" == FU_type)
   {
      type = flopoco_wrapper::UT_MULT;
      op = new flopoco::FPMultiplier(flopoco::target, (int)n_exp_in, (int)n_mant_in, (int)n_exp_in, (int)n_mant_in, (int)n_exp_out, (int)n_mant_out);
   }
   else if ("FPDiv" == FU_type)
   {
      type = flopoco_wrapper::UT_DIV;
      op = new flopoco::FPDiv(flopoco::target, (int)n_exp_in, (int)n_mant_in);
   }
   else if ("FPExp" == FU_type)
   {
      type = flopoco_wrapper::UT_EXP;
      op = new flopoco::FPExp(flopoco::target, (int)n_exp_in, (int)n_mant_in, 0, 0);
   }
   else if ("FPSqrtPoly" == FU_type)
   {
      type = flopoco_wrapper::UT_SQRT;
      op = new flopoco::FPSqrtPoly(flopoco::target, (int)n_exp_in, (int)n_mant_in, 0, 3/*degree*/);
   }
   else if ("Fix2FP" == FU_type)
   {
      type = flopoco_wrapper::UT_FIX2FP;
      op = new flopoco::Fix2FP(flopoco::target, 0, (int)FU_prec_in-1, 1, (int)n_exp_out, (int)n_mant_out);
   }
   else if ("FP2Fix" == FU_type)
   {
      type = flopoco_wrapper::UT_FP2FIX;
      op = new flopoco::FP2Fix(flopoco::target, 0, (int)FU_prec_out-1, 1, (int)n_exp_in, (int)n_mant_in, true);
   }
   else if("FF_CONV" == FU_type)
   {
      type = flopoco_wrapper::UT_FF_CONV;
      op = new flopoco::FPAssign(flopoco::target, (int)n_exp_out, (int)n_mant_out);
   }
   else if("FPgt_expr" == FU_type)
   {
      type = flopoco_wrapper::UT_compare_expr;
      op = new flopoco::FPgt_expr(flopoco::target, (int)n_exp_in, (int)n_mant_in);
   }
   else if("FPlt_expr" == FU_type)
   {
      type = flopoco_wrapper::UT_compare_expr;
      op = new flopoco::FPlt_expr(flopoco::target, (int)n_exp_in, (int)n_mant_in);
   }
   else if("FPge_expr" == FU_type)
   {
      type = flopoco_wrapper::UT_compare_expr;
      op = new flopoco::FPge_expr(flopoco::target, (int)n_exp_in, (int)n_mant_in);
   }
   else if("FPle_expr" == FU_type)
   {
      type = flopoco_wrapper::UT_compare_expr;
      op = new flopoco::FPle_expr(flopoco::target, (int)n_exp_in, (int)n_mant_in);
   }
   else if("FPLog" == FU_type)
   {
      type = flopoco_wrapper::UT_LOG;
      op = new flopoco::FPLog(flopoco::target, (int)n_exp_in, (int)n_mant_in, 0);
   }
   else
   {
      THROW_ASSERT(false, "Not supported FU: " + FU_type);
   }
   OPLIST.push_back(op);
   std::string FU_name_stored = ENCODE_NAME(FU_name, FU_prec_in, FU_prec_out);
   op->changeName(WRAPPED_PREFIX + FU_name_stored);
   this->FUs[WRAPPED_PREFIX + FU_name_stored] = op;
   this->FU_to_prec.insert(make_pair(FU_name_stored, std::pair<unsigned int, unsigned int>(FU_prec_in, FU_prec_out)));

   flopoco::target->setNotPipelined();
   // Adds two additional Functional Units to perform conversion
   // from FloCoCo encoding to IEEE-754 number format and viceversa
   if(type != flopoco_wrapper::UT_FIX2FP)
   {
      if(type == flopoco_wrapper::UT_FF_CONV)
         op = new flopoco::InputIEEE(flopoco::target, (int)n_exp_in, (int)n_mant_in, (int)n_exp_out, (int)n_mant_out);
      else
         op = new flopoco::InputIEEE(flopoco::target, (int)n_exp_in, (int)n_mant_in, (int)n_exp_in, (int)n_mant_in);
      OPLIST.push_back(op);
      op->changeName(IN_WRAP_PREFIX + FU_name_stored);
      this->FUs[IN_WRAP_PREFIX + FU_name_stored] = op;
   }
   if(type != flopoco_wrapper::UT_FP2FIX && type != flopoco_wrapper::UT_compare_expr)
   {
      op = new flopoco::OutputIEEE(flopoco::target, (int)n_exp_out, (int)n_mant_out, (int)n_exp_out, (int)n_mant_out);
      OPLIST.push_back(op);
      op->changeName(OUT_WRAP_PREFIX + FU_name_stored);
      this->FUs[OUT_WRAP_PREFIX + FU_name_stored] = op;
   }
   //delete flopoco::target;
   //flopoco::target = 0;

}

flopoco::Operator * flopoco_wrapper::get_FU(std::string FU_name_stored) const
{
   std::tr1::unordered_map<std::string, flopoco::Operator*>::const_iterator op_found = FUs.find(FU_name_stored);
   THROW_ASSERT(op_found != FUs.end(), "Functional unit " + FU_name_stored + " not found (maybe not yet generated)");
   return op_found->second;
}

void flopoco_wrapper::outputWrapVHDL(std::string & FU_name_stored, std::ostream & os)
{
   this->outputHeaderVHDL(FU_name_stored, os);
   // Write library declaration
   PP(os, "library ieee;\n");
   PP(os, "use ieee.std_logic_1164.all;\n");
   PP(os, "\n");
   PP(os, "entity " + FU_name_stored + " is\n");
   PP.indent();
   PP(os, "generic(");
   std::tr1::unordered_map<std::string, std::pair<unsigned int,unsigned int> >::const_iterator FU_to_prec_it = this->FU_to_prec.find(FU_name_stored);
   // Write input port(s) generics
   const std::vector<std::string> p_in = this->get_ports(WRAPPED_PREFIX+FU_name_stored, 0, port_in, false);
   for (std::vector<std::string>::const_iterator p_in_it = p_in.begin(); p_in_it != p_in.end(); p_in_it++)
     PP(os, "BITSIZE_" + *p_in_it + ": integer := " + STR(FU_to_prec_it->second.first) + "; ");
   // Write output port(s) generics
   const std::vector<std::string> p_out = this->get_ports(WRAPPED_PREFIX+FU_name_stored, 0, port_out, false);
   for (std::vector<std::string>::const_iterator p_out_it = p_out.begin(); p_out_it != p_out.end(); p_out_it++)
      if (p_out_it + 1 != p_out.end())
         PP(os, "BITSIZE_" + *p_out_it + ": integer := " + STR(FU_to_prec_it->second.second) + ";");
      else
         PP(os, "BITSIZE_" + *p_out_it + ": integer := " + STR(FU_to_prec_it->second.second));
   PP(os, ");\n");
   PP(os, "port (\n");
   PP.indent();
   // Write port declaration for top entity
   this->outputPortDeclaration(WRAPPED_PREFIX, FU_name_stored, os, top);
   // Begin writing the architecture of top entity
   PP.deindent();
   PP(os, ");\n");
   PP.deindent();
   PP(os, "end entity;\n");
   PP(os, "\n");
   PP(os, "architecture arch of " + FU_name_stored + " is\n");
   PP.indent();
   // Write declaration for each of entity's components
   std::string prefixes[] = { WRAPPED_PREFIX, IN_WRAP_PREFIX, OUT_WRAP_PREFIX};
   for (unsigned int i = 0; i < sizeof(prefixes)/sizeof(prefixes[0]); i++)
   {
      if(!((type == flopoco_wrapper::UT_FIX2FP && prefixes[i] == IN_WRAP_PREFIX) ||
           ((type == flopoco_wrapper::UT_FP2FIX || type == flopoco_wrapper::UT_compare_expr)
            && prefixes[i] == OUT_WRAP_PREFIX)))
      {
         PP(os, "component " + prefixes[i] + FU_name_stored + "\n");
         PP.indent();
         PP(os, "port (\n");
         PP.indent();
         if (WRAPPED_PREFIX == prefixes[i])
            this->outputPortDeclaration(prefixes[i], FU_name_stored, os, wrapped);
         else if (IN_WRAP_PREFIX == prefixes[i])
            this->outputPortDeclaration(prefixes[i], FU_name_stored, os, in_wrap);
         else if (OUT_WRAP_PREFIX == prefixes[i])
            this->outputPortDeclaration(prefixes[i], FU_name_stored, os, out_wrap);
         else
            THROW_ASSERT(false, "Something wrong happened");
         PP.deindent();
         PP(os, ");\n");
         PP.deindent();
         PP(os, "end component;\n");
      }
   }
   // Write the declaration of needed signals
   this->outputSignals(FU_name_stored, os);
   // Begin behavioral description
   PP(os, "begin\n");
   PP.indent();
   // Write port mapping
   this->outputPortMap(FU_name_stored, os);
   // End with architecture
   PP.deindent();
   PP.deindent();
   PP(os, "end architecture;\n");
   PP(os, "\n");
}

void flopoco_wrapper::outputPortMap(std::string & FU_name_stored, std::ostream & os)
{
   std::string mapping;
   const std::vector<std::string> p_wrapped_in = this->get_ports(WRAPPED_PREFIX+FU_name_stored, 0, port_in, false);
   const std::vector<std::string> p_wrapped_out = this->get_ports(WRAPPED_PREFIX+FU_name_stored, 0, port_out, false);
   std::vector<std::string> p_in_wrap_in;
   std::vector<std::string> p_in_wrap_out;
   if(type != flopoco_wrapper::UT_FIX2FP)
   {
      p_in_wrap_in = this->get_ports(IN_WRAP_PREFIX+FU_name_stored, 0, port_in, false);
      p_in_wrap_out = this->get_ports(IN_WRAP_PREFIX+FU_name_stored, 0, port_out, false);
   }
   std::vector<std::string> p_out_wrap_in;
   std::vector<std::string> p_out_wrap_out;
   if(type != flopoco_wrapper::UT_FP2FIX && type != flopoco_wrapper::UT_compare_expr)
   {
      p_out_wrap_in = this->get_ports(OUT_WRAP_PREFIX+FU_name_stored, 0, port_in, false);
      p_out_wrap_out = this->get_ports(OUT_WRAP_PREFIX+FU_name_stored, 0, port_out, false);
   }

   std::tr1::unordered_map<std::string, std::pair<unsigned int,unsigned int> >::const_iterator FU_to_prec_it = this->FU_to_prec.find(FU_name_stored);
   unsigned int n_bits_in /*, n_bits_out*/;
   unsigned int prec_in /*, prec_out*/;
   prec_in = FU_to_prec_it->second.first;
//   prec_out = FU_to_prec_it->second.second;
   n_bits_in=prec_in;
//   n_bits_out=prec_out;

#ifndef NO_PIPELINE
  const std::string p_clock = this->get_port(clock);
   const std::string p_reset = this->get_port(reset);
#endif
   // Write mapping for wrapped component
   mapping = "";
   for (unsigned int i = 0; i < p_wrapped_in.size(); i++)
   {
      if(type == flopoco_wrapper::UT_ADDSUB && i == 1)
         mapping += p_wrapped_in.at(i) + " => wire_ADDSUB, ";
      else if(type == flopoco_wrapper::UT_SUB && i == 1)
         mapping += p_wrapped_in.at(i) + " => wire_SUB, ";
      else if(type == flopoco_wrapper::UT_FIX2FP)
         mapping += p_wrapped_in.at(i) + " => " + p_wrapped_in.at(i) + ", ";
      else
         mapping += p_wrapped_in.at(i) + " => wireIn" + STR(i+1) + ", ";
   }
   for (unsigned int j = 0; j < p_wrapped_out.size(); j++)
   {
      if(type == flopoco_wrapper::UT_FP2FIX)
         mapping += p_wrapped_out.at(j) + " => " + p_wrapped_out.at(j);
      else if(type == flopoco_wrapper::UT_compare_expr)
         mapping += p_wrapped_out.at(j) + " => " + p_wrapped_out.at(j) + "(0)";
      else
         mapping += p_wrapped_out.at(j) + " => wireOut" + STR(j+1);
   }
#ifndef NO_PIPELINE
   mapping +=  ", " + p_clock + "=> " + std::string(CLOCK_PORT_NAME) + ", " + p_reset + "=> " + std::string(RESET_PORT_NAME);
#endif
   if (type == flopoco_wrapper::UT_ADDSUB)
   {
      PP(os, "wire_ADDSUB <= wireIn2(" + STR(n_bits_in+1) + " downto " + STR(n_bits_in) + ") & (wireIn2(" +
         STR(n_bits_in-1) + ") xor sel_minus_expr) & wireIn2(" + STR(n_bits_in-2) + " downto 0);\n");
   }
   else if (type == flopoco_wrapper::UT_SUB)
   {
      PP(os, "wire_SUB <= wireIn2(" + STR(n_bits_in+1) + " downto " + STR(n_bits_in) + ") & (wireIn2(" +
         STR(n_bits_in-1) + ") xor '1') & wireIn2(" + STR(n_bits_in-2) + " downto 0);\n");
   } else if(type == flopoco_wrapper::UT_EXP || type == flopoco_wrapper::UT_SQRT || type == flopoco_wrapper::UT_LOG)
   {
      PP(os, std::string(DONE_PORT_NAME) + " <= '1';\n");
   }
   PP(os, "fu : " + STR(WRAPPED_PREFIX+FU_name_stored) + " port map (" + mapping + ");\n");
   // Write mapping for input converters
   if(type != flopoco_wrapper::UT_FIX2FP)
   {
      for (unsigned int i = 0; i < p_wrapped_in.size(); i++)
      {
         mapping = "";
         mapping += p_in_wrap_in.at(0) + "=>" + p_wrapped_in.at(i) + ", ";
         mapping += p_in_wrap_out.at(0) + "=>wireIn" + STR(i+1);
         PP(os, "in" + STR(i+1) + " : " + STR(IN_WRAP_PREFIX+FU_name_stored) + " port map (" + mapping + ");\n");
      }
   }
   // Write mapping for output converters
   if(type != flopoco_wrapper::UT_FP2FIX && type != flopoco_wrapper::UT_compare_expr)
   {
      for (unsigned int i = 0; i < p_wrapped_out.size(); i++)
      {
         mapping = "";
         mapping += p_out_wrap_in.at(0) + "=>wireOut" + STR(i+1) + ", ";
         mapping += p_out_wrap_out.at(0) + "=>" + p_wrapped_out.at(i);
         PP(os, "out" + STR(i+1) + " : " + STR(OUT_WRAP_PREFIX+FU_name_stored) + " port map (" + mapping + ");\n");
      }
   }
}

void flopoco_wrapper::outputSignals(std::string & FU_name_stored, std::ostream & os)
{
   std::string Signals = "";
   const std::vector<std::string> p_in = this->get_ports(WRAPPED_PREFIX+FU_name_stored, 0, port_in, false);
   const std::vector<std::string> p_out = this->get_ports(WRAPPED_PREFIX+FU_name_stored, 0, port_out, false);
   std::tr1::unordered_map<std::string, std::pair<unsigned int,unsigned int> >::const_iterator FU_to_prec_it = this->FU_to_prec.find(FU_name_stored);
   unsigned int n_bits_in, n_bits_out;
   unsigned int prec_in, prec_out;
   prec_in = FU_to_prec_it->second.first;
   prec_out = FU_to_prec_it->second.second;
   n_bits_in=prec_in;
   n_bits_out=prec_out;

   if (type == flopoco_wrapper::UT_ADDSUB)
      Signals += "wire_ADDSUB, ";
   else if (type == flopoco_wrapper::UT_SUB)
      Signals += "wire_SUB, ";
   else if (type== flopoco_wrapper::UT_FF_CONV)
      n_bits_in=prec_out;

   if(type != flopoco_wrapper::UT_FIX2FP)
   {
      size_t n_in_elements = p_in.size();
      for (unsigned int i = 0; i < n_in_elements; i++)
         Signals += "wireIn" + STR(i+1) + (i+1 != n_in_elements ? ", " : "");
      PP(os, "signal " + Signals + " : std_logic_vector(" + STR(n_bits_in+1)+ " downto 0);\n");
   }
   if(type != flopoco_wrapper::UT_FP2FIX && type != flopoco_wrapper::UT_compare_expr)
   {
      Signals = "";
      for (unsigned int j = 0; j < p_out.size(); j++)
         Signals += "wireOut" + STR(j+1) + (j+1 != p_out.size() ? ", " : "");
      PP(os, "signal " + Signals + " : std_logic_vector(" + STR(n_bits_out+1)+ " downto 0);\n");
   }
}

void flopoco_wrapper::outputPortDeclaration(std::string FU_prefix, std::string & FU_name_stored, std::ostream & os, component_type c_type)
{
   // Compute offsets for addition bits' handling
   int in_offset, out_offset;
   unsigned int n_bits_in, n_bits_out;
   unsigned int prec_in, prec_out;
   std::tr1::unordered_map<std::string, std::pair<unsigned int,unsigned int> >::const_iterator FU_to_prec_it = this->FU_to_prec.find(FU_name_stored);
   in_offset = out_offset = -1;
   n_bits_in = n_bits_out = 0;
   prec_in = FU_to_prec_it->second.first;
   prec_out = FU_to_prec_it->second.second;
   if (wrapped == c_type)
   {
      if(type != flopoco_wrapper::UT_FIX2FP)
         in_offset += FLOPOCO_ADDITIONAL_BITS;
      if(type != flopoco_wrapper::UT_FP2FIX && type != flopoco_wrapper::UT_compare_expr)
         out_offset += FLOPOCO_ADDITIONAL_BITS;
      if(type == flopoco_wrapper::UT_FF_CONV)
         n_bits_in = n_bits_out = prec_out;
      else if(type == flopoco_wrapper::UT_compare_expr)
      {
         n_bits_in=prec_in;
         n_bits_out=1;
      }
      else
      {
         n_bits_in=prec_in;
         n_bits_out=prec_out;
      }
   }
   else if (in_wrap == c_type)
   {
      out_offset += FLOPOCO_ADDITIONAL_BITS;
      if(type == flopoco_wrapper::UT_FF_CONV)
      {
         n_bits_in=prec_in;
         n_bits_out=prec_out;
      }
      else
         n_bits_in = n_bits_out = prec_in;
   }
   else if (out_wrap == c_type)
   {
      in_offset += FLOPOCO_ADDITIONAL_BITS;
      n_bits_in = n_bits_out = prec_out;
   }
#ifndef NO_PIPELINE
   // Write input port(s) declaration
   if(wrapped == c_type)
   {
      PP(os, this->get_port(clock) + " : in std_logic;\n");
      PP(os, this->get_port(reset) + " : in std_logic;\n");
   }
#endif
   const std::vector<std::string> p_in = this->get_ports(FU_prefix+FU_name_stored, 0, port_in, false);
   for (std::vector<std::string>::const_iterator p_in_it = p_in.begin(); p_in_it != p_in.end(); p_in_it++)
      if(top == c_type)
         PP(os, *p_in_it + " : in std_logic_vector(BITSIZE_" + *p_in_it + "-1 downto 0);\n");
      else if(n_bits_in+in_offset>0)
         PP(os, *p_in_it + " : in std_logic_vector(" + STR((int)n_bits_in+in_offset) + " downto 0);\n");
      else
         PP(os, *p_in_it + " : in std_logic;\n");
   // Write clock and reset ports declaration, only for top and wrapped entities
   if (top == c_type)
   {
      PP(os, std::string(CLOCK_PORT_NAME) + " : in std_logic;\n");
      PP(os, std::string(RESET_PORT_NAME) + " : in std_logic;\n");

      if (type == flopoco_wrapper::UT_ADDSUB)
      {
         PP(os, std::string("sel_plus_expr") + " : in std_logic;\n");
         PP(os, std::string("sel_minus_expr") + " : in std_logic;\n");
      }
      else if(type == flopoco_wrapper::UT_EXP || type == flopoco_wrapper::UT_SQRT || type == flopoco_wrapper::UT_LOG)
      {
         PP(os, std::string(START_PORT_NAME) + " : in std_logic;\n");
         PP(os, std::string(DONE_PORT_NAME) + " : out std_logic;\n");
      }
   }
   // Write output port(s) declaration
   const std::vector<std::string> p_out = this->get_ports(FU_prefix+FU_name_stored, 0, port_out, false);
   for (std::vector<std::string>::const_iterator p_out_it = p_out.begin(); p_out_it != p_out.end(); p_out_it++)
      if (p_out_it + 1 != p_out.end())
      {
         if(top == c_type)
            PP(os, *p_out_it + " : out std_logic_vector(BITSIZE_" + *p_out_it + "-1 downto 0);\n");
         else if(n_bits_out+out_offset>0)
            PP(os, *p_out_it + " : out std_logic_vector(" + STR((int)n_bits_out+out_offset) + " downto 0);\n");
         else
            PP(os, *p_out_it + " : out std_logic;\n");
      }
      else
      {
         if(top == c_type)
            PP(os, *p_out_it + " : out std_logic_vector(BITSIZE_" + *p_out_it + "-1 downto 0)\n");
         else if(n_bits_out+out_offset>0)
            PP(os, *p_out_it + " : out std_logic_vector(" + STR((int)n_bits_out+out_offset) + " downto 0)\n");
         else
            PP(os, *p_out_it + " : out std_logic\n");
      }
}

void flopoco_wrapper::outputHeaderVHDL(std::string & FU_name_stored, std::ostream & os) const
{
   os << "--------------------------------------------------------------------------------" << endl;
   os << "--                              " << FU_name_stored << endl;
   os << "-- Operator automatically generated by " << PACKAGE_NAME << " framework version " << PACKAGE_VERSION << endl;
   os << "-- assemblying operators generated by the Infinite Virtual Library FloPoCoLib" << endl;
   os << "-- which is distributed under the terms of the GNU Lesser General Public Licence" << endl;
   os << "-- Send any bug to: " << PACKAGE_BUGREPORT << endl;
   os << "--------------------------------------------------------------------------------" << endl;
}

int flopoco_wrapper::writeVHDL(const std::string & FU_name, const unsigned int FU_prec_in, const unsigned int FU_prec_out, std::string FU_file)
{
   std::string FU_name_stored = ENCODE_NAME(FU_name, FU_prec_in, FU_prec_out);
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, this->debug_level, "Writing VHDL code for unit " + FU_name_stored + " to file " + FU_file + FILE_EXT);
   if (this->FU_files.find(FU_name_stored) != this->FU_files.end())
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, this->debug_level, "- Code for unit has already been generated");
      return 1;
   }
   std::string filename = FU_file + FILE_EXT;
   std::ofstream file(filename.c_str());
   if (!file.is_open())
   {
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, this->debug_level, "- Something went wrong in file creation");
      return -1;
   }
   // Call FloPoCo method to generate VHDL for Functional Unit and Conversion Units
#ifndef NEW_FLOPOCO
   for(unsigned int i=0; i<OPLIST.size(); i++)
#endif
   {

      try {
#ifdef NEW_FLOPOCO
         flopoco::Operator::outputVHDLToFile(OPLIST, file);
#else
         OPLIST[i]->getFlopocoVHDLStream()->flush();
         /* second parse is only for sequential operators */
         if (OPLIST[i]->isSequential())
            OPLIST[i]->parse2();
         OPLIST[i]->outputVHDL(file);
#endif
      } catch (std::string s) {
         cerr << "Exception while generating '"
#ifndef NEW_FLOPOCO
              << OPLIST[i]->getName()
              << "': "
#endif
              << s <<endl;
      }
   }

   // Generate VHDL for Functional Unit wrapper entity
   this->outputWrapVHDL(FU_name_stored, file);
   // Mark the file as generated
   this->FU_files.insert(filename);
   file.close();
   OPLIST.clear();
   PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, this->debug_level, "- Successfully written to file!");
   return 0;
}

const std::vector<std::string> flopoco_wrapper::get_ports(std::string FU_name_stored, unsigned int DEBUG_PARAMETER(expected_ports), port_type type, bool DEBUG_PARAMETER(check_ports)) const
{
   std::vector<std::string> ports;
   flopoco::Operator * op = this->get_FU(FU_name_stored);
   for (int i = 0; i < op->getIOListSize(); i++)
   {
      flopoco::Signal * sig = op->getIOListSignal(i);
      if (type == port_in && sig->type() == flopoco::Signal::in)
         ports.push_back(sig->getName());
      else if (type == port_out && sig->type() == flopoco::Signal::out)
         ports.push_back(sig->getName());
   }
   THROW_ASSERT(!check_ports || expected_ports == ports.size(), "Expected a different number of " + (type == port_in ? std::string("input") : std::string("output")) + " ports");
   return ports;
}

const std::string flopoco_wrapper::get_port(port_type type) const
{
   if (type == reset)
      return "rst";
   else if (type == clock)
      return "clk";
   THROW_ASSERT(false, "Something went wrong!");
   return "";
}

void flopoco_wrapper::DECODE_FU(std::string & encoded_FU, std::string & FU_type, unsigned int & FU_prec_in, unsigned int & FU_prec_out) const
{
   std::vector<std::string> split;
   boost::algorithm::split(split, encoded_FU, boost::algorithm::is_any_of(" "));
   THROW_ASSERT(split.size() == 3, "Malformed FU");
   FU_type = split[0];
   FU_prec_in = boost::lexical_cast<unsigned int>(split[1]);
   FU_prec_out = boost::lexical_cast<unsigned int>(split[2]);
}

void flopoco_wrapper::DECODE_BITS(unsigned int FU_prec, unsigned int & n_mant, unsigned int & n_exp)
{
   switch (FU_prec)
   {
      case 16: // IEEE-754 half precision
      {
         n_exp = 5;
         n_mant = 10;
         break;
      }
      case 32: // IEEE-754 single precision
      {
         n_exp = 8;
         n_mant = 23;
         break;
      }
      case 64: // IEEE-754 double precision
      {
         n_exp = 11;
         n_mant = 52;
         break;
      }
      case 96: ///Intel 80 bit extended precision padded with 16bits
      {
         THROW_ERROR("Intel Extended Precision not currently supported");
         n_exp = 15;
         n_mant = 64;
         break;
      }
      case 128: // IEEE-754 quad precision
      {
         n_exp = 15;
         n_mant = 112;
         break;
      }
      default: // Linear growth for exponent number of bits
      {
         n_exp = 8 + ((unsigned int)3*(FU_prec-32)/32);
         n_mant = FU_prec - n_exp - 1;
         break;
      }
   }
}
