/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file ASICBackendFlow.cpp
 * @brief Implementation of the wrapper to ASIC tools
 *
 * Implementation of the methods to wrap synthesis tools for integrated circuits
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
///Header include
#include "ASICBackendFlow.hpp"

#include "SynthesisTool.hpp"
#include "DesignCompilerWrapper.hpp"

#include "target_device.hpp"
#include "area_model.hpp"
#include "time_model.hpp"
#include "technology_builtin.hpp"

#include "xml_script_command.hpp"
#include "xml_dom_parser.hpp"
#include "Parameter.hpp"
#include "fileIO.hpp"

ASICBackendFlow::ASICBackendFlow(const ParameterConstRef _Param, const std::string & _flow_name, const target_deviceRef _device) :
   BackendFlow(_Param, _flow_name, _device)
{
   PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, " .:: Creating ASIC Backend Flow ::.");

   default_data["Nangate"] =
      #include "Nangate.data"
         ;

   fileIO_istreamRef sname;
   if (Param->isOption(OPT_target_device_script))
   {
      std::string xml_file_path = Param->getOption<std::string>(OPT_target_device_script);
      if (!boost::filesystem::exists(xml_file_path)) THROW_ERROR("File \"" + xml_file_path + "\" does not exist!");
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Importing scripts from file: " + xml_file_path);
      sname = fileIO_istream_open(xml_file_path);
   }
   else
   {
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Importing default scripts for target technology: \"Nangate\"");
      sname = fileIO_istream_open_from_string(default_data["Nangate"]);
   }
   parse_flow(sname);
}

ASICBackendFlow::~ASICBackendFlow()
{

}

void ASICBackendFlow::create_sdc(const DesignParametersRef dp)
{
   std::string sdc_filename = out_dir + "/" + dp->component_name + ".sdc";
   std::ofstream SDC_file(sdc_filename.c_str());
   if (dp->parameter_values.find(PARAM_clk_name) != dp->parameter_values.end())
   {
      SDC_file << "create_clock " << dp->parameter_values[PARAM_clk_name] << " -period " << dp->parameter_values[PARAM_clk_period] << std::endl;
   }
   SDC_file.close();
   dp->parameter_values[PARAM_sdc_file] = sdc_filename;
}

void ASICBackendFlow::InitDesignParameters(const DesignParametersRef dp)
{
   create_sdc(dp);

   for (unsigned int i = 0; i < steps.size(); i++)
   {
      steps[i]->tool->EvaluateVariables(dp);
   }
}

void ASICBackendFlow::CheckSynthesisResults()
{
   PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Analyzing ASIC synthesis results");
   for(unsigned int s = 0; s < steps.size(); s++)
   {
      if (steps[s]->tool->get_tool_exec() == DESIGN_COMPILER_TOOL_ID)
      {
         area_modelRef area = GetPointer<DesignCompilerWrapper>(steps[s]->tool)->parse_area_reports();
         if (area) area_m = area;

         time_modelRef time = GetPointer<DesignCompilerWrapper>(steps[s]->tool)->parse_time_reports();
         if (time) time_m = time;
      }
   }
}

void ASICBackendFlow::WriteFlowConfiguration(std::ostream& /*script*/)
{
   ///nothing to do
}

