/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file XilinxBackendFlow.cpp
 * @brief Implementation of the wrapper to Xilinx tools
 *
 * Implementation of the methods to wrap synthesis tools by Xilinx
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Date: 2012-03-16 14:28:17 +0100 (Fri, 16 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
///Header include
#include "XilinxBackendFlow.hpp"

#include "config_HAVE_XILINX.hpp"
#include "config_XILINX_SETTINGS.hpp"

#include "target_device.hpp"
#include "area_model.hpp"
#include "clb_model.hpp"
#include "time_model.hpp"
#include "LUT_model.hpp"
#include "technology_builtin.hpp"
#include "map_wrapper.hpp"
#include "trce_wrapper.hpp"

#include "XilinxWrapper.hpp"

#include "xml_script_command.hpp"
#include "xml_dom_parser.hpp"
#include "Parameter.hpp"
#include "fileIO.hpp"

#define XST_NUMBER_OF_SLICE_REGISTERS           "XST_NUMBER_OF_SLICE_REGISTERS"
#define XST_NUMBER_OF_SLICE_LUTS                "XST_NUMBER_OF_SLICE_LUTS"
#define XST_NUMBER_OF_LUT_FLIP_FLOP_PAIRS_USED  "XST_NUMBER_OF_LUT_FLIP_FLOP_PAIRS_USED"
#define XST_NUMBER_OF_BONDED_IOBS               "XST_NUMBER_OF_BONDED_IOBS"

XilinxBackendFlow::XilinxBackendFlow(const ParameterConstRef _Param, const std::string& _flow_name, const target_deviceRef _device) :
   BackendFlow(_Param, _flow_name, _device)
{
   PRINT_OUT_MEX(OUTPUT_LEVEL_MINIMUM, output_level, " .:: Creating Xilinx Backend Flow ::.");
   boost::filesystem::create_directories(UCF_SUBDIR);

   default_data["Virtex-5"] =
      #include "Virtex-5.data"
         ;
   default_data["Virtex-6"] =
      #include "Virtex-6.data"
         ;
   default_data["Virtex-7"] =
      #include "Virtex-7.data"
         ;

   fileIO_istreamRef sname;
   if (Param->isOption(OPT_target_device_script))
   {
      std::string xml_file_path = Param->getOption<std::string>(OPT_target_device_script);
      if (!boost::filesystem::exists(xml_file_path)) THROW_ERROR("File \"" + xml_file_path + "\" does not exist!");
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Importing scripts from file: " + xml_file_path);
      sname = fileIO_istream_open(xml_file_path);
   }
   else
   {
      std::string device_string;
      if(device->has_parameter("family"))
         device_string = device->get_parameter<std::string>("family");
      else
         device_string = "Virtex-5";
      if (default_data.find(device_string) == default_data.end())
         THROW_ERROR("Device family \"" + device_string + "\" not supported!");
      PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Importing default scripts for target device family: " + device_string);
      sname = fileIO_istream_open_from_string(default_data[device_string]);
   }
   parse_flow(sname);
}

XilinxBackendFlow::~XilinxBackendFlow()
{

}

void XilinxBackendFlow::xparse_map_utilization(const std::string& fn)
{
   std::ifstream output_file(fn.c_str());
   if (output_file.is_open())
   {
      while(!output_file.eof())
      {
         std::string line;
         getline (output_file, line);
         if (line.size() and line.find("Number of occupied Slices:") != std::string::npos)
         {
            std::string tk = "Number of occupied Slices:";
            std::string token = line.substr(line.find(tk) + tk.size()+1, line.length());
            boost::algorithm::trim(token);
            token = token.substr(0, token.find(" "));
            boost::algorithm::trim(token);
            boost::replace_all(token, ",", "");
            area_m = area_model::create_model(device->get_target_technology(), Param);
            clb_model* clb_m = GetPointer<clb_model>(area_m);
            clb_m->set_resource_value(clb_model::SLICE_LUTS, boost::lexical_cast<unsigned int>(token));
         }
         if (line.size() and line.find("Number of LUT Flip Flop pairs used:") != std::string::npos)
         {
            std::string tk = "Number of LUT Flip Flop pairs used:";
            std::string token = line.substr(line.find(tk) + tk.size()+1, line.length());
            boost::algorithm::trim(token);
            token = token.substr(0, token.find(" "));
            boost::algorithm::trim(token);
            boost::replace_all(token, ",", "");
            area_m = area_model::create_model(device->get_target_technology(), Param);
            clb_model* clb_m = GetPointer<clb_model>(area_m);
            clb_m->set_resource_value(clb_model::LUT_FF_PAIRS, boost::lexical_cast<unsigned int>(token));
            clb_m->set_area_value(boost::lexical_cast<unsigned int>(token));
         }
      }
   }
}

void XilinxBackendFlow::xparse_xst_utilization(const std::string& fn)
{
   try
   {
      fileIO_istreamRef sname = fileIO_istream_open(fn);
      xml_dom_parser parser;
      //parser.set_validate();
      parser.parse_stream(*sname);
      if (parser)
      {
         //Walk the tree:
         const xml_element* node = parser.get_document()->get_root_node(); //deleted by DomParser.
         THROW_ASSERT(node->get_name() == "document", "Wrong root name: " + node->get_name());

         const xml_node::node_list list_int = node->get_children();
         for (xml_node::node_list::const_iterator iter_int = list_int.begin(); iter_int != list_int.end(); ++iter_int)
         {
            const xml_element* EnodeC = GetPointer<const xml_element>(*iter_int);
            if(!EnodeC) continue;

            if (EnodeC->get_name() == "application")
            {
               const xml_node::node_list list_sec = EnodeC->get_children();
               for (xml_node::node_list::const_iterator iter_sec = list_sec.begin(); iter_sec != list_sec.end(); ++iter_sec)
               {
                  const xml_element* nodeS = GetPointer<const xml_element>(*iter_sec);
                  if(!nodeS) continue;

                  if (nodeS->get_name() == "section")
                  {
                     std::string stringID;
                     if(CE_XVM(stringID, nodeS)) LOAD_XVM(stringID, nodeS);
                     if (stringID == "XST_DEVICE_UTILIZATION_SUMMARY")
                     {
                        const xml_node::node_list list_item = nodeS->get_children();
                        for (xml_node::node_list::const_iterator it_item = list_item.begin(); it_item != list_item.end(); ++it_item)
                        {
                           const xml_element* nodeIt = GetPointer<const xml_element>(*it_item);
                           if(!nodeIt or nodeIt->get_name() != "item") continue;

                           if(CE_XVM(stringID, nodeIt)) LOAD_XVM(stringID, nodeIt);

                           if (stringID != "XST_SELECTED_DEVICE")
                           {
                              std::string value;
                              if(CE_XVM(value, nodeIt))
                              {
                                 LOAD_XVM(value, nodeIt);
                                 boost::replace_all(value, ",", "");
                                 area_values[stringID] = boost::lexical_cast<unsigned int>(value);
                              }
                           }
                        }
                     }
                  }
               }
            }
         }

         area_m = area_model::create_model(device->get_target_technology(), Param);
         clb_model* clb_m = GetPointer<clb_model>(area_m);
         if (area_values.find(XST_NUMBER_OF_SLICE_REGISTERS) != area_values.end())
         {
            clb_m->set_resource_value(clb_model::SLICE_REGISTERS, area_values[XST_NUMBER_OF_SLICE_REGISTERS]);
         }
         if (area_values.find(XST_NUMBER_OF_SLICE_LUTS) != area_values.end())
         {
            clb_m->set_resource_value(clb_model::SLICE_LUTS, area_values[XST_NUMBER_OF_SLICE_LUTS]);
         }
         if (area_values.find(XST_NUMBER_OF_BONDED_IOBS) != area_values.end())
         {
            clb_m->set_resource_value(clb_model::BONDED_IOBS, area_values[XST_NUMBER_OF_BONDED_IOBS]);
         }
         /// setting the global area occupation as the number of LUT/FF pairs
         THROW_ASSERT(area_values.find(XST_NUMBER_OF_LUT_FLIP_FLOP_PAIRS_USED) != area_values.end(), "Missing LUT/FF pairs");
         clb_m->set_resource_value(clb_model::LUT_FF_PAIRS, area_values[XST_NUMBER_OF_LUT_FLIP_FLOP_PAIRS_USED]);
         area_m->set_area_value(area_values[XST_NUMBER_OF_LUT_FLIP_FLOP_PAIRS_USED]);

         return;
      }
   }
   catch (const char * msg)
   {
      std::cerr << msg << std::endl;
   }
   catch (const std::string & msg)
   {
      std::cerr << msg << std::endl;
   }
   catch (const std::exception& ex)
   {
      std::cout << "Exception caught: " << ex.what() << std::endl;
   }
   catch ( ... )
   {
      std::cerr << "unknown exception" << std::endl;
   }
   THROW_ERROR("Error during XST report parsing: " + fn);
}

void XilinxBackendFlow::parse_timing(const std::string& log_file)
{
   std::ifstream output_file(log_file.c_str());
   if (output_file.is_open())
   {
      while(!output_file.eof())
      {
         std::string line;
         getline (output_file, line);
         if (line.size() and line.find("Minimum period:") != std::string::npos)
         {
            if(line.find("No path found") != std::string::npos)
            {
               time_m = time_model::create_model(device->get_target_technology(), Param);
               LUT_model* lut_m = GetPointer<LUT_model>(time_m);
               lut_m->set_timing_value(LUT_model::COMBINATIONAL_DELAY, 0.0);
            }
            else
            {
               std::string token("Minimum period:");
               std::string tk = line.substr(line.find(token)+token.size()+1, line.size());
               boost::trim(tk);
               tk = tk.substr(0, tk.find_first_of(' '));
               boost::replace_all(tk, "ns", "");
               time_m = time_model::create_model(device->get_target_technology(), Param);
               LUT_model* lut_m = GetPointer<LUT_model>(time_m);
               lut_m->set_timing_value(LUT_model::COMBINATIONAL_DELAY, boost::lexical_cast<double>(tk));
            }
         }
         if (line.size() and line.find("WARNING:Xst") != std::string::npos)
            PRINT_OUT_MEX(OUTPUT_LEVEL_VERY_PEDANTIC, output_level, line);
      }
   }
}

void XilinxBackendFlow::xparse_timing(const std::string& fn, bool post)
{
   try
   {
      fileIO_istreamRef sname = fileIO_istream_open(fn);
      xml_dom_parser parser;
      //parser.set_validate();
      parser.parse_stream(*sname);
      if (parser)
      {
         //Walk the tree:
         const xml_element* node = parser.get_document()->get_root_node(); //deleted by DomParser.
         THROW_ASSERT(node->get_name() == "twReport", "Wrong root name: " + node->get_name());

         const xml_node::node_list list_int = node->get_children();
         for (xml_node::node_list::const_iterator iter_int = list_int.begin(); iter_int != list_int.end(); ++iter_int)
         {
            const xml_element* EnodeC = GetPointer<const xml_element>(*iter_int);
            if(!EnodeC) continue;
            if (EnodeC->get_name() == "twSum")
            {
               const xml_node::node_list list = EnodeC->get_children();
               for (xml_node::node_list::const_iterator iter = list.begin(); iter != list.end(); ++iter)
               {
                  const xml_element* Enode = GetPointer<const xml_element>(*iter);
                  if(!Enode) continue;
                  if (Enode->get_name() == "twStats")
                  {
                     const xml_node::node_list listS = Enode->get_children();
                     for (xml_node::node_list::const_iterator iterS = listS.begin(); iterS != listS.end(); ++iterS)
                     {
                        const xml_element* EnodeS = GetPointer<const xml_element>(*iterS);
                        if(!EnodeS) continue;
                        if (EnodeS->get_name() == "twMinPer")
                        {
                           const xml_text_node* child_text = EnodeS->get_child_text();
                           double period = boost::lexical_cast<double>(child_text->get_content());
                           time_m = time_model::create_model(device->get_target_technology(), Param);
                           LUT_model* lut_m = GetPointer<LUT_model>(time_m);
                           if (post)
                           {
                              lut_m->set_timing_value(LUT_model::MINIMUM_PERIOD_POST_PAR, period);
                           }
                           else
                           {
                              lut_m->set_timing_value(LUT_model::MINIMUM_PERIOD_POST_MAP, period);
                           }
                        }
                     }
                  }
               }
            }
         }
         return;
      }
   }
   catch (const char * msg)
   {
      std::cerr << msg << std::endl;
   }
   catch (const std::string & msg)
   {
      std::cerr << msg << std::endl;
   }
   catch (const std::exception& ex)
   {
      std::cout << "Exception caught: " << ex.what() << std::endl;
   }
   catch ( ... )
   {
      std::cerr << "unknown exception" << std::endl;
   }
   THROW_ERROR("Error during TRCE report parsing: " + fn);
}

void XilinxBackendFlow::CheckSynthesisResults()
{
   PRINT_OUT_MEX(OUTPUT_LEVEL_VERBOSE, output_level, "Analyzing Xilinx synthesis results");

   if (boost::filesystem::exists(actual_parameters->parameter_values[PARAM_map_report]))
   {
      xparse_map_utilization(actual_parameters->parameter_values[PARAM_map_report]);
   }
   else if (boost::filesystem::exists(actual_parameters->parameter_values[PARAM_xst_report]))
   {
      xparse_xst_utilization(actual_parameters->parameter_values[PARAM_xst_report]);
   }

   if (actual_parameters->parameter_values.find(PARAM_trce_report_post) != actual_parameters->parameter_values.end() &&
       boost::filesystem::exists(actual_parameters->parameter_values[PARAM_trce_report_post]))
   {
      xparse_timing(actual_parameters->parameter_values[PARAM_trce_report_post], true);
   }
   else if (actual_parameters->parameter_values.find(PARAM_trce_report_pre) != actual_parameters->parameter_values.end() &&
       boost::filesystem::exists(actual_parameters->parameter_values[PARAM_trce_report_pre]))
   {
      xparse_timing(actual_parameters->parameter_values[PARAM_trce_report_pre], false);
   }
   else if (actual_parameters->parameter_values.find(PARAM_xst_log_file) != actual_parameters->parameter_values.end() &&
       boost::filesystem::exists(actual_parameters->parameter_values[PARAM_xst_log_file]))
   {
      parse_timing(actual_parameters->parameter_values[PARAM_xst_log_file]);
   }
}

void XilinxBackendFlow::WriteFlowConfiguration(std::ostream& script)
{
   std::string setupscr = STR(XILINX_SETTINGS);
   if(setupscr.size() and setupscr != "0")
   {
      script << "#configuration" << std::endl;
      if(setupscr.find("export") == 0)
         script << setupscr + " >& /dev/null; ";
      else
         script << ". " << setupscr << " >& /dev/null; ";
      script << std::endl << std::endl;
   }
}

void XilinxBackendFlow::ExecuteSynthesis()
{
#if HAVE_XILINX
   BackendFlow::ExecuteSynthesis();
#else
   THROW_ERROR("Xilinx tools not configured; Execution of the synthesis flow is not possible");
#endif
}

void XilinxBackendFlow::InitDesignParameters(const DesignParametersRef dp)
{
   dp->parameter_values[PARAM_top_id] = dp->component_name;

   std::string ise_style;
   if (output_level >= OUTPUT_LEVEL_VERY_PEDANTIC)
      ise_style =  std::string(INTSTYLE_ISE);
   else
      ise_style =  std::string(INTSTYLE_SILENT);
   dp->parameter_values[PARAM_ise_style] = ise_style;

   dp->parameter_values[PARAM_clk_name] = CLOCK_PORT_NAME;
   dp->parameter_values[PARAM_clk_period] = Param->getOption<std::string>(OPT_clock_period);
   bool time_constrained = false;
   if(boost::lexical_cast<double>(dp->parameter_values[PARAM_clk_period]) != 0.0)
      time_constrained = true;
   dp->parameter_values[PARAM_time_constrained] = STR(time_constrained);

   ///determine if power optimization has to be performed
   bool xpwr_enabled = false;
   if(Param->isOption("power_optimization") && Param->getOption<bool>("power_optimization"))
      xpwr_enabled = true;
   dp->parameter_values[PARAM_power_optimization] = STR(xpwr_enabled);
   bool connect_iob = false;
   if(Param->isOption("connect_iob") && Param->getOption<bool>("connect_iob"))
      connect_iob = true;
   dp->parameter_values[PARAM_connect_iob] = STR(connect_iob);
   std::string device_name = device->get_parameter<std::string>("model");
   std::string package = device->get_parameter<std::string>("package");
   std::string speed_grade = device->get_parameter<std::string>("speed_grade");
   std::string device_string = device_name + package + speed_grade;
   dp->parameter_values[PARAM_target_device] = device_string;

   if(time_constrained)
   {
      create_cf(dp, true);
      create_cf(dp, false);
   }
   for (unsigned int i = 0; i < steps.size(); i++)
   {
      steps[i]->tool->EvaluateVariables(dp);
   }
}

void XilinxBackendFlow::create_cf(const DesignParametersRef dp, bool xst)
{
   std::string ucf_filename = UCF_SUBDIR + dp->component_name + (xst ? ".xcf" : ".ucf");
   std::ofstream UCF_file(ucf_filename.c_str());
   if (dp->parameter_values.find(PARAM_clk_name) != dp->parameter_values.end())
   {
      UCF_file << "NET \"" << dp->parameter_values[PARAM_clk_name] << "\" TNM_NET = " << dp->parameter_values[PARAM_clk_name] << ";" << std::endl;
      UCF_file << "TIMESPEC TS_" << dp->parameter_values[PARAM_clk_name] << " = PERIOD \"" << dp->parameter_values[PARAM_clk_name] << "\" " << dp->parameter_values[PARAM_clk_period] << " ns HIGH 50%;" << std::endl;
   }
   UCF_file.close();
   if (xst) dp->parameter_values[PARAM_xcf_file] = ucf_filename;
   else dp->parameter_values[PARAM_ucf_file] = ucf_filename;
}
