/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file gcc_wrapper.hpp
 * @brief Implementation of the wrapper to Gcc for C sources.
 *
 * Implementation of the methods for the object for invoke the GCC compiler from sources and create the dump.
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/

///Autoheader include
#include "config_ARM_CPP_EXE.hpp"
#include "config_ARM_EMPTY_PLUGIN.hpp"
#include "config_ARM_GCC_EXE.hpp"
#include "config_ARM_RTL_PLUGIN.hpp"
#include "config_ARM_SSA_PLUGIN.hpp"
#include "config_HAVE_ARM_COMPILER.hpp"
#include "config_HAVE_I386_COMPILER.hpp"
#include "config_HAVE_SPARC_COMPILER.hpp"
#include "config_HAVE_FROM_RTL_BUILT.hpp"
#include "config_I386_CPP_EXE.hpp"
#include "config_I386_EMPTY_PLUGIN.hpp"
#include "config_I386_GCC_EXE.hpp"
#include "config_I386_SSA_PLUGIN.hpp"
#include "config_SPARC_CPP_EXE.hpp"
#include "config_SPARC_ELF_CPP.hpp"
#include "config_SPARC_ELF_GCC.hpp"
#include "config_SPARC_EMPTY_PLUGIN.hpp"
#include "config_SPARC_GCC_EXE.hpp"
#include "config_SPARC_RTL_PLUGIN.hpp"
#include "config_SPARC_SSA_PLUGIN.hpp"

/// Header include
#include "gcc_wrapper.hpp"

/// BOOST include
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/filesystem/convenience.hpp>
#include <boost/filesystem/fstream.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/regex.hpp>

/// Frontend include
#include "Parameter.hpp"

///RTL include
#if HAVE_FROM_RTL_BUILT
#include "parse_rtl.hpp"
#endif
/// STD include
#include <cerrno>
#include <unistd.h>

///STL include
#include <tr1/unordered_set>

/// Tree includes
#include "parse_tree.hpp"
#include "tree_manager.hpp"

/// Utility include
#include "cpu_stats.hpp"
#include "dbgPrintHelper.hpp"
#include "exceptions.hpp"
#include "fileIO.hpp"
#include "global_enums.hpp"
#include "string_manipulation.hpp"
#include "treegcc_constants.hpp"
#include "utility.hpp"

/// XML includes used for writing and reading the configuration file
#include "xml_helper.hpp"
#include "polixml.hpp"
#include "xml_dom_parser.hpp"

std::string GccWrapper::current_gcc_version;

std::string GccWrapper::current_plugin_version;

GccWrapper::GccWrapper(const ParameterConstRef _Param, const CompilerTarget _compiler_target, const OptimizationSet _OS):
   Param(_Param),
   compiler_target(_compiler_target),
   OS(_OS),
   file_name_counter(0),
   output_level(_Param->getOption<int>(OPT_output_level)),
   debug_level(_Param->get_class_debug_level("GccWrapper"))
{
   InitializeGccParameters();
   if(Param->isOption(OPT_gcc_write_xml))
   {
      WriteXml(Param->getOption<std::string>(OPT_gcc_write_xml));
   }
}

// destructor
GccWrapper::~GccWrapper()
{
}

void GccWrapper::CompileFile(const std::string real_filename, std::string file_string, const std::string & parameters_line, bool empty_file)
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->compileFile: " + file_string);

   ///The gcc output
   const std::string gcc_output_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + STR_CST_gcc_output;

   std::string opt;// = " -O1";
   const Compiler compiler = GetCompiler();
   std::string command = compiler.gcc.string();
   command += " -D__NO_INLINE__ "; ///needed to avoid problem with glibc inlines

   ///Adding source code includes
   if(real_filename != "-" and real_filename != file_string)
   {
      std::tr1::unordered_set<std::string> source_files;
      source_files.insert(real_filename);
      command += AddSourceCodeIncludes(source_files) + " ";
   }
   command += " " + compiler.extra_options + " ";
   if(empty_file)
   {
      /*
       * If the file does not contain any function definition, compilation flow of gcc is not activated,
       * so with the hooks currently provided by the GCC it is not possible to execute the plugin in the
       * right moment; the workaround is adding a fake function to the source code
       */
      //If this is not a copy of the file, create it before append
      if(real_filename != "-" and real_filename != "/dev/null" and real_filename == file_string)
      {
         std::string new_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + "patched_" + boost::lexical_cast<std::string>(file_name_counter++) + ".c";
         file_string = new_file_name;
         if (boost::filesystem::exists(file_string))
            boost::filesystem::remove(file_string);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Copy file from " + real_filename + " to " + file_string);
         CopyFile(real_filename, file_string);
      }
      std::string append_command = "`echo -e \"\\nvoid __empty_function__(){}\" >> " + file_string + "`";
      int ret = PandaSystem(Param, append_command);
      if(IsError(ret))
      {
         THROW_ERROR("Error in appending empty function");
      }
      command += opt + " -c -fplugin=" + compiler.empty_plugin;
   }
   else if((Param->isOption(OPT_gcc_E) and Param->getOption<bool>(OPT_gcc_E)) or (Param->isOption(OPT_gcc_S) and Param->getOption<bool>(OPT_gcc_S)))
      command += opt;
   else
#if HAVE_FROM_RTL_BUILT
   if(Param->getOption<bool>(OPT_use_rtl))
   {
      command += opt + " -c -fplugin=" + compiler.rtl_plugin;
   }
   else
#endif
      command += opt + " -c -fplugin=" + compiler.ssa_plugin;
   if((Param->isOption(OPT_gcc_E) and Param->getOption<bool>(OPT_gcc_E)) or (Param->isOption(OPT_gcc_S) and Param->getOption<bool>(OPT_gcc_S)))
   {
      if(Param->isOption(OPT_output_file))
         command += " -o " + Param->getOption<std::string>(OPT_output_file);
   }
   else
      command += " -o " + std::string(STR_CST_gcc_obj_file);
   if(!(Param->getOption<bool>(OPT_compute_size_of)))
      command += " -D\"" + std::string(STR_CST_panda_sizeof) + "(arg)=" + STR_CST_string_sizeof + "(#arg)\"";
   command += " " + parameters_line;
   if(real_filename == "-" or real_filename == "/dev/null")
   {
      command += file_string;
   }
   else
   {
      boost::filesystem::path file_path(real_filename);
      std::string extension = GetExtension(file_path);
      /// assembler files are not allowed so in some cases we pass a C file renamed with extension .S
      if(extension == "S")
         command += "-x c ";
      command += "\"" + file_string + "\"";
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "---Invoke: " + command);
   int ret = PandaSystem(Param, command, gcc_output_file_name);
   if(!((Param->isOption(OPT_gcc_E) and Param->getOption<bool>(OPT_gcc_E)) or (Param->isOption(OPT_gcc_S) and Param->getOption<bool>(OPT_gcc_S))))
      std::remove(STR_CST_gcc_obj_file);
   if(IsError(ret))
   {
      PRINT_OUT_MEX(OUTPUT_LEVEL_NONE, 0, "Error in compilation");
      CopyStdout(gcc_output_file_name);
      THROW_ERROR_CODE(COMPILING_EC, "GCC returns an error during compilation "+boost::lexical_cast<std::string>(errno));
      THROW_ERROR("GCC returns an error during compilation "+boost::lexical_cast<std::string>(errno));
   }

   else
   {
      if(output_level >= OUTPUT_LEVEL_VERBOSE)
         CopyStdout(gcc_output_file_name);
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Compiled file " + file_string);
}

void GccWrapper::CreateTreeManager(tree_managerRef & TM, const std::tr1::unordered_set<std::string> & source_files)
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Invoking GCC");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->");
   const std::string output_temporary_directory = Param->getOption<std::string>(OPT_output_temporary_directory);
   if (source_files.size() == 0)
      THROW_ERROR("No files specified for parsing");

   // Set the optimization level for the compiler
   // Prepare the command line string for invoke the gcc compiler
   std::vector<std::string> compile_files;
   std::vector<std::string> original_compile_files;
   std::tr1::unordered_set<std::string>::const_iterator source_file, source_file_end = source_files.end();
   for(source_file = source_files.begin(); source_file != source_file_end; source_file++)
   {
      if (*source_file != "-" and !boost::filesystem::exists(*source_file))
         THROW_ERROR("File " + *source_file + " does not exist");

      boost::filesystem::path new_source;

      if(Param->getOption<bool>(OPT_compute_size_of))
      {
         new_source = *source_file;
      }
      else
      {
         ///We use sed to copy the original source code file to the temp directory replacing size_of with STR_CST_panda_sizeof
         std::string new_file_name = output_temporary_directory + "patched_" + boost::lexical_cast<std::string>(file_name_counter++) + ".c";
         new_source = new_file_name;
         if (boost::filesystem::exists(new_source))
            boost::filesystem::remove(new_source);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Renaming file from " + *source_file + " to " + new_source.string());
         std::string sed_command = "sed 's/sizeof/" + std::string(STR_CST_panda_sizeof) + "/g' " + *source_file;

         int ret = PandaSystem(Param, sed_command, new_source.string());
         if(IsError(ret))
         {
            THROW_ERROR("Error during the execution of sed on file " + *source_file);
         }
      }
      original_compile_files.push_back(*source_file);
      compile_files.push_back(new_source.string());
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Copied files into tmp directory");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Starting compilation of single files");
   for (unsigned int file = 0; file < compile_files.size(); file++)
   {
      std::string Name = compile_files[file] == "-" ? "-" : GetLeafFileName(compile_files[file]);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Compiling file " + Name);
      ///create obj
      CompileFile(original_compile_files[file], "\"" + compile_files[file] + "\"" , gcc_compiling_parameters);
      if(!Param->isOption(OPT_gcc_E) and !Param->isOption(OPT_gcc_S))
      {
         if(!(boost::filesystem::exists(boost::filesystem::path(Name + STR_CST_gcc_tree_suffix))))
         {
            THROW_WARNING("Raw not created for file " + Name);
            CompileFile(original_compile_files[file], compile_files[file], gcc_compiling_parameters, true);
            if(not(boost::filesystem::exists(boost::filesystem::path(Name + STR_CST_gcc_empty_suffix))))
               THROW_ERROR("Impossible to create raw file for " + Name);
            rename_file(Name + STR_CST_gcc_empty_suffix, Name + STR_CST_gcc_tree_suffix);
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Renaming " + Name + STR_CST_gcc_empty_suffix + " in " + Name + STR_CST_gcc_tree_suffix);
         }
         boost::filesystem::path obj = boost::filesystem::path(Name + STR_CST_gcc_tree_suffix);
         if (!TM)
            TM  = tree_managerRef(new tree_manager(Param));
         tree_managerRef TreeM = ParseTreeFile(Param, obj.string());
         boost::filesystem::path destination;
         if(Name == "-")
            destination = boost::filesystem::path(output_temporary_directory + "stdin-" + STR_CST_gcc_tree_suffix);
         else
            destination = boost::filesystem::path(output_temporary_directory + Name + STR_CST_gcc_tree_suffix);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Renaming " + obj.string() + " in " + destination.string());
         try
         {
            rename_file(obj, destination);
         }
         catch (const char * msg)
         {
            THROW_ERROR("Error in renaming file: " + std::string(msg));
         }
         catch (const std::string & msg)
         {
            THROW_ERROR("Error in renaming file: " + msg);
         }
         catch (const std::exception& ex)
         {
            THROW_ERROR("Error in renaming file : " + std::string(ex.what()));
         }
         catch ( ... )
         {
            THROW_ERROR("Error in renaming file unknown error");
         }

#if HAVE_FROM_RTL_BUILT
         if((Param->getOption<bool>(OPT_use_rtl)) and boost::filesystem::exists(boost::filesystem::path(Name + STR_CST_gcc_rtl_suffix)))
         {
            obj = boost::filesystem::path(Name + STR_CST_gcc_rtl_suffix);
            parse_rtl_File(obj.string(), TreeM, debug_level);
            rename_file(obj, boost::filesystem::path(output_temporary_directory + Name + STR_CST_gcc_rtl_suffix));
         }
#endif
         TM->merge_tree_managers(TreeM);
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "Ended compilation of single files");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Finished GCC invoking");
}

void GccWrapper::InitializeGccParameters()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Initializing gcc parameters");
   OptimizationSet optimization_set = OS;
   ///If we are compiling with sparc-elf-gcc
   if(compiler_target == CT_SPARC_ELF_GCC)
   {
      switch(Param->getOption<char>(OPT_gcc_opt_level))
      {
         case('0'):
            optimization_set = OS_O0;
            break;
         case('1'):
            optimization_set = OS_O1;
            break;
         case('2'):
            optimization_set = OS_O2;
            break;
         case('3'):
            optimization_set = OS_O3;
            break;
         case('s'):
            optimization_set = OS_Os;
            break;
         default:
         {
            THROW_ASSERT(false, "Optimization level not supported " + Param->getOption<std::string>(OPT_gcc_opt_level));
         }
      }
   }

   if(Param->isOption(OPT_gcc_read_xml))
   {
      ReadXml(Param->getOption<std::string>(OPT_gcc_read_xml));
   }
   else
   {
      switch(optimization_set)
      {
         case(OS_O0):
            gcc_compiling_parameters += " -O0 ";
            break;
         case(OS_O1):
            gcc_compiling_parameters += " -O1 ";
            break;
         case(OS_O2):
            gcc_compiling_parameters += " -O2 ";
            break;
         case(OS_O3):
            gcc_compiling_parameters += " -O3 ";
            break;
         case(OS_Os):
            gcc_compiling_parameters += " -Os ";
            break;
#if HAVE_BAMBU_BUILT
         case(OS_OBAMBU):
#endif
#if HAVE_TUCANO_BUILT
         case(OS_OTUCANO):
#endif
#if HAVE_ZEBU_BUILT
         case(OS_OZEBU):
#endif
            ///Filling optimizations map
            SetGccDefault();

            switch(OS)
            {
#if HAVE_BAMBU_BUILT
               case(OS_OBAMBU):
                  SetBambuDefault();
                  break;
#endif
#if HAVE_TUCANO_BUILT
               case(OS_OTUCANO):
                  break;
#endif
#if HAVE_ZEBU_BUILT
               case(OS_OZEBU):
                  SetZebuDefault();
                  break;
#endif
               default:
                  PANDA_UNREACHABLE();
            }
            ReadParameters();

            gcc_compiling_parameters += WriteOptimizationsString();

            break;
      }
   }

   ///Adding standard
   if(Param->isOption(OPT_gcc_standard))
   {
      std::string standard = Param->getOption<std::string>(OPT_gcc_standard);
      gcc_compiling_parameters += "--std=" + standard + " ";
   }

   if(Param->isOption(OPT_gcc_E) and Param->getOption<bool>(OPT_gcc_E))
         gcc_compiling_parameters += "-E ";
   if(Param->isOption(OPT_gcc_S) and Param->getOption<bool>(OPT_gcc_S))
         gcc_compiling_parameters += "-S ";
   ///Adding defines
   if(Param->isOption(OPT_gcc_defines))
   {
      const std::tr1::unordered_set<std::string> defines = Param->getOption<const std::tr1::unordered_set<std::string> >(OPT_gcc_defines);
      std::tr1::unordered_set<std::string>::const_iterator define, define_end = defines.end();
      for(define = defines.begin(); define != define_end; define++)
      {
         std::string escaped_string = *define;
         add_escape(escaped_string, "\"");
         gcc_compiling_parameters += "-D" + escaped_string + " ";
      }
   }

   ///Adding undefines
   if(Param->isOption(OPT_gcc_undefines))
   {
      const std::tr1::unordered_set<std::string> undefines = Param->getOption<const std::tr1::unordered_set<std::string> >(OPT_gcc_undefines);
      std::tr1::unordered_set<std::string>::const_iterator undefine, undefine_end = undefines.end();
      for(undefine = undefines.begin(); undefine != undefine_end; undefine++)
      {
         std::string escaped_string = *undefine;
         add_escape(escaped_string, "\"");
         gcc_compiling_parameters += "-U" + escaped_string + " ";
      }
   }

   ///Adding warnings
   if(Param->isOption(OPT_gcc_warnings))
   {
      const std::tr1::unordered_set<std::string> warnings = Param->getOption<const std::tr1::unordered_set<std::string> >(OPT_gcc_warnings);
      std::tr1::unordered_set<std::string>::const_iterator warning, warning_end = warnings.end();
      for(warning = warnings.begin(); warning != warning_end; warning++)
      {
         gcc_compiling_parameters += "-W" + *warning + " ";
      }
   }
#if HAVE_BAMBU_BUILT
   if(OS == OS_OBAMBU)
   {
      gcc_compiling_parameters += "-Wuninitialized ";
   }
#endif


   ///Adding includes
   if(Param->isOption(OPT_gcc_includes))
   {
      gcc_compiling_parameters += Param->getOption<std::string>(OPT_gcc_includes) + " ";
   }

   ///Adding libraries
   if(Param->isOption(OPT_gcc_libraries))
   {
      const std::tr1::unordered_set<std::string> libraries = Param->getOption<const std::tr1::unordered_set<std::string> >(OPT_gcc_libraries);
      std::tr1::unordered_set<std::string>::const_iterator library, library_end = libraries.end();
      for(library = libraries.begin(); library != library_end; library++)
      {
         gcc_linking_parameters += "-l" + *library + " ";
      }
   }

   ///Adding library directories
   if(Param->isOption(OPT_gcc_library_directories))
   {
      const std::tr1::unordered_set<std::string> library_directories = Param->getOption<const std::tr1::unordered_set<std::string> >(OPT_gcc_library_directories);
      std::tr1::unordered_set<std::string>::const_iterator library_directory, library_directory_end = library_directories.end();
      for(library_directory = library_directories.begin(); library_directory != library_directory_end; library_directory++)
      {
         gcc_linking_parameters += "-L" + *library_directory + " ";
      }
   }

   ///Adding no parse
   if(Param->isOption(OPT_no_parse_files))
   {
      const std::tr1::unordered_set<std::string> no_parse_files = Param->getOption<const std::tr1::unordered_set<std::string> >(OPT_no_parse_files);
      std::tr1::unordered_set<std::string>::const_iterator no_parse_file, no_parse_file_end = no_parse_files.end();
      for(no_parse_file = no_parse_files.begin(); no_parse_file != no_parse_file_end; no_parse_file++)
      {
         gcc_linking_parameters += *no_parse_file + " ";
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Initialized gcc parameters");
}

#if HAVE_ZEBU_BUILT
void GccWrapper::SetZebuDefault()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "-->Setting parameters for Zebu tool...");
   int opt_level;
   switch(Param->getOption<char>(OPT_gcc_opt_level))
   {
      case('s'):
         opt_level = 0;
         break;
      default:
      {
         opt_level = Param->getOption<int>(OPT_gcc_opt_level);
      }
   }

   INDENT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "-->Optimization level: " + boost::lexical_cast<std::string>(opt_level));

   if(Param->getOption<AnalysisLevel>(OPT_analysis_level) >= AL_COMPONENT)
   {
     optimization_flags["inline-functions-called-once"] = false;
     optimization_flags["inline-small-functions"] = false;
     optimization_flags["inline-functions"] = false;
   }

   ///parameters with enable
   // (e.g. optimization_flags["inline"] = true;)
   optimization_flags["ivopts"] = false; ///introduce target_memory_ref as gimple_assign operands. The default GCC value is true
   optimization_flags["tree-loop-im"] = false;///Execution error in 20040307-1.c example. The default GCC value is true
   optimization_flags["tree-loop-ivcanon"] = false; ///this is requested for rebuild while and for
   optimization_flags["tree-sink"] = true; ///this is requested for rebuild while and for

   ///builtin function;
   optimization_flags["builtin"] = false;

   //optimization_flags["openmp"] = true;

   ///parameters with values
   // (e.g. optimization_values["inline-limit"] = 10;)
//FIXME: to be replaced with plugin; deactivated since it does not work with sparc-elf-gcc
//   optimization_values["tree-parallelize-loops"]=1;///requires tree-loop-optimize
   //parameters_value["max-aliased-vops"]=1000;
/*
   parameters_value["avg-aliased-vops"]=1000000000;
   parameters_value["max-aliased-vops"]=1000000000;
   parameters_value["salias-max-implicit-fields"]=1000000000;
   parameters_value["salias-max-array-elements"]=1000000000;
   optimization_flags["strict-aliasing"] = true; ///help alias analysis
   optimization_flags["tree-salias"] = true;///help alias analysis
   optimization_flags["tree-reassoc"] = false;

   parameters_value["max-aliased-vops"]=1000;
   parameters_value["avg-aliased-vops"]=5;
   parameters_value["salias-max-implicit-fields"]=1000;
   parameters_value["salias-max-array-elements"]=1000;
   optimization_flags["strict-aliasing"] = true; ///help alias analysis
   optimization_flags["tree-salias"] = true;///help alias analysis
   optimization_flags["tree-reassoc"] = false;
*/

   if(opt_level >=1)
   {
      optimization_flags["guess-branch-probability"] = false;///error in declaration of structure used probably by profiling.
      optimization_flags["tree-ch"] = false;
      optimization_flags["tree-copy-prop"] = false; ///va_list undeclared
      optimization_flags["tree-dominator-opts"] = false; ///va_list undeclared
      optimization_flags["tree-sra"] = false; ///Introduces conversion from struct to scalar
      optimization_flags["rename-registers"]  = false; ///cross compilation problems
   }
   if(opt_level >=2)
   {
      optimization_flags["optimize-sibling-calls"] = false; ///Execution error on 20020406-1.c
      optimization_flags["tree-pre"] = false; ///some loop are incorrecty identified
      optimization_flags["tree-vrp"] = false; ///create several irriducible loops
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "<--");
   INDENT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "<--Set parameters for Zebu tool");
}
#endif

#if HAVE_BAMBU_BUILT
void GccWrapper::SetBambuDefault()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "-->Setting parameters for Bambu tool...");
   int opt_level;
   switch(Param->getOption<char>(OPT_gcc_opt_level))
   {
      case('s'):
         opt_level = 0;
         break;
      default:
      {
         opt_level = Param->getOption<int>(OPT_gcc_opt_level);
      }
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "---Optimization level: " + boost::lexical_cast<std::string>(opt_level));

   ///parameters with enable
   // (e.g. optimization_flags["inline"] = true;)

   ///builtin function;
   //optimization_flags["builtin"] = false;
   //optimization_flags["ipa-type-escape"] = true; /// no more supported starting from gcc 4.6.1

   if(opt_level >=1)
   {
      optimization_flags["guess-branch-probability"] = false;///error in declaration of structure used probably by profiling.
   }
   if(opt_level >=3)
   {
      optimization_flags["tree-slp-vectorize"] = false; /// vectorization not yet supported by HLS
      optimization_flags["tree-vectorize"] = false;
   }
   if(opt_level >=4)
   {
      optimization_flags["tree-loop-linear"] = true;
      optimization_flags["graphite-identity"] = true;
   }
   ///parameters with values
   // (e.g. optimization_values["inline-limit"] = 10;)
   optimization_values["tree-parallelize-loops"]=1;///requires tree-loop-optimize

   INDENT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "<--Set parameters for bambu tool");
}
#endif

void GccWrapper::SetGccDefault()
{
   INDENT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "-->Setting GCC defaults");
   int opt_level;
   switch(Param->getOption<char>(OPT_gcc_opt_level))
   {
      case('s'):
         opt_level = 0;
         break;
      default:
      {
         opt_level = Param->getOption<int>(OPT_gcc_opt_level);
      }
   }
   optimization_flags["stack-protector"] = false;//In Ubuntu 6.10 and later versions this option is enabled by default for C, C++, ObjC, ObjC++
   optimization_flags["rename-registers"] = false; ///cross compilation problems
   optimization_flags["trapping-math"] = true;///-fno-trapping-math compiles code assuming that floating-point operations cannot generate user-visible traps.  These traps include division by zero, overflow, underflow, inexact result and invalid operation.  This option implies -fno-signaling-nans.  Setting this option may allow faster code if one relies on "non-stop" IEEE arithmetic, for example. This option should never be turned on by any -O option since it can result in incorrect output for programs which depend on an exact implementation of IEEE or ISO rules/specifications for math functions.
   optimization_flags["signed-zeros"] = true;///-fno-signed-zeros allows optimizations for floating point arithmetic that ignore the signedness of zero. IEEE arithmetic specifies the behavior of distinct +0.0 and -0.0 values, which then prohibits simplification of expressions such as x+0.0 or 0.0*x (even with -ffinite-math-only).

   if(opt_level == 0)
   {
      gcc_compiling_parameters += " -O1 ";

      optimization_flags["align-functions"] = false;
      optimization_flags["align-jumps"] = false;
      optimization_flags["align-labels"] = false;
      optimization_flags["align-loops"] = true;
      optimization_flags["argument-alias"] = true;
      optimization_flags["argument-noalias"] = false;
      optimization_flags["argument-noalias-anything"] = false;
      optimization_flags["argument-noalias-global"] = false;
      optimization_flags["asynchronous-unwind-tables"] = true;
      optimization_flags["branch-count-reg"] = true;
      optimization_flags["branch-probabilities"] = false;
      optimization_flags["branch-target-load-optimize"] = false;
      optimization_flags["branch-target-load-optimize2"] = false;
      optimization_flags["btr-bb-exclusive"] = false;
      optimization_flags["caller-saves"] = false;
      optimization_flags["common"] = true;
      optimization_flags["conserve-stack"] = false;
      optimization_flags["cprop-registers"] = false;
      optimization_flags["crossjumping"] = false;
      optimization_flags["cse-follow-jumps"] = false;
      optimization_flags["cx-fortran-rules"] = false;
      optimization_flags["cx-limited-range"] = false;
      optimization_flags["data-sections"] = false;
      optimization_flags["dce"] = true;
      optimization_flags["defer-pop"] = false;
      optimization_flags["delayed-branch"] = false;
      optimization_flags["delete-null-pointer-checks"] = false;
      optimization_flags["dse"] = true;
      optimization_flags["early-inlining"] = true;
      optimization_flags["exceptions"] = false;
      optimization_flags["expensive-optimizations"] = false;
      optimization_flags["finite-math-only"] = false;
      optimization_flags["float-store"] = false;
      optimization_flags["forward-propagate"] = false;
      optimization_flags["gcse"] = false;
      optimization_flags["gcse-after-reload"] = false;
      optimization_flags["gcse-las"] = false;
      optimization_flags["gcse-lm"] = true;
      optimization_flags["gcse-sm"] = false;
      optimization_flags["graphite-identity"] = false;
      optimization_flags["guess-branch-probability"] = false;
      optimization_flags["if-conversion"] = false;
      optimization_flags["if-conversion2"] = false;
      optimization_flags["inline-functions"] = false;
      optimization_flags["inline-functions-called-once"] = true;
      optimization_flags["inline-small-functions"] = false;
      optimization_flags["ipa-cp"] = false;
      optimization_flags["ipa-cp-clone"] = false;
      optimization_flags["ipa-matrix-reorg"] = false;
      optimization_flags["ipa-pta"] = false;
      optimization_flags["ipa-pure-const"] = false;
      optimization_flags["ipa-reference"] = false;
      optimization_flags["ipa-sra"] = false;
      //optimization_flags["ipa-type-escape"] = false;
      optimization_flags["ivopts"] = true;
      optimization_flags["jump-tables"] = true;
      optimization_flags["loop-block"] = false;
      optimization_flags["loop-interchange"] = false;
      optimization_flags["loop-parallelize-all"] = false;
      optimization_flags["loop-parallelize-all"] = false;
      optimization_flags["loop-strip-mine"] = false;
      optimization_flags["merge-all-constants"] = false;
      optimization_flags["merge-constants"] = false;
      optimization_flags["modulo-sched"] = false;
      optimization_flags["move-loop-invariants"] = true;
      optimization_flags["non-call-exceptions"] = false;
      optimization_flags["omit-frame-pointer"] = false;
      optimization_flags["optimize-register-move"] = false;
      optimization_flags["optimize-sibling-calls"] = false;
      optimization_flags["pack-struct"] = false;
      optimization_flags["peel-loops"] = false;
      optimization_flags["peephole"] = true;
      optimization_flags["peephole2"] = false;
      optimization_flags["predictive-commoning"] = false;
      optimization_flags["prefetch-loop-arrays"] = false;
      optimization_flags["reg-struct-return"] = false;
      optimization_flags["regmove"] = false;
      optimization_flags["reorder-blocks"] = false;
      optimization_flags["reorder-blocks-and-partition"] = false;
      optimization_flags["reorder-functions"] = false;
      optimization_flags["rerun-cse-after-loop"] = true;
      optimization_flags["reschedule-modulo-scheduled-loops"] = false;
      optimization_flags["rounding-math"] = false;
      //optimization_flags["rtl-abstract-sequences"] = false;
      optimization_flags["sched-critical-path-heuristic"] = false;
      optimization_flags["sched-dep-count-heuristic"] = false;
      optimization_flags["sched-group-heuristic"] = false;
      optimization_flags["sched-interblock"] = false;
      optimization_flags["sched-last-insn-heuristic"] = false;
      optimization_flags["sched-pressure"] = false;
      optimization_flags["sched-rank-heuristic"] = false;
      optimization_flags["sched-spec"] = false;
      optimization_flags["sched-spec-insn-heuristic"] = false;
      optimization_flags["sched-spec-load"] = false;
      optimization_flags["sched-spec-load-dangerous"] = false;
      optimization_flags["sched-stalled-insns"] = false;
      optimization_flags["sched-stalled-insns-dep"] = true;
      optimization_flags["sched2-use-superblocks"] = false;
      optimization_flags["schedule-insns"] = false;
      optimization_flags["schedule-insns2"] = false;
      optimization_flags["section-anchors"] = false;
      optimization_flags["sel-sched-pipelining"] = false;
      optimization_flags["sel-sched-pipelining-outer-loops"] = false;
      optimization_flags["sel-sched-reschedule-pipelined"] = false;
      optimization_flags["selective-scheduling"] = false;
      optimization_flags["selective-scheduling2"] = false;
      optimization_flags["short-double"] = false;
      optimization_flags["short-enums"] = false;
      optimization_flags["short-wchar"] = false;
      optimization_flags["signaling-nans"] = false;
      optimization_flags["single-precision-constant"] = false;
      optimization_flags["split-ivs-in-unroller"] = true;
      optimization_flags["split-wide-types"] = false;
      optimization_flags["strict-aliasing"] = false;
      optimization_flags["thread-jumps"] = false;
      optimization_flags["toplevel-reorder"] = true;
      optimization_flags["trapv"] = false;
      optimization_flags["tree-builtin-call-dce"] = false;
      optimization_flags["tree-ccp"] = false;
      optimization_flags["tree-ch"] = false;
      optimization_flags["tree-copy-prop"] = false;
      optimization_flags["tree-copyrename"] = false;
      optimization_flags["tree-cselim"] = true;
      optimization_flags["tree-dce"] = false;
      optimization_flags["tree-dominator-opts"] = false;
      optimization_flags["tree-dse"] = false;
      optimization_flags["tree-forwprop"] = false;
      optimization_flags["tree-fre"] = false;
      optimization_flags["tree-loop-distribution"] = false;
      optimization_flags["tree-loop-im"] = true;
      optimization_flags["tree-loop-ivcanon"] = true;
      optimization_flags["tree-loop-linear"] = false;
      optimization_flags["tree-loop-optimize"] = true;
      optimization_flags["tree-lrs"] = false;
      optimization_flags["tree-phiprop"] = false;
      optimization_flags["tree-pre"] = false;
      optimization_flags["tree-pta"] = false;
      optimization_flags["tree-reassoc"] = false;
      optimization_flags["tree-scev-cprop"] = true;
      optimization_flags["tree-sink"] = false;
      optimization_flags["tree-slp-vectorize"] = false;
      optimization_flags["tree-sra"] = false;
      optimization_flags["tree-switch-conversion"] = false;
      optimization_flags["tree-ter"] = false;
      optimization_flags["tree-vect-loop-version"] = true;
      optimization_flags["tree-vectorize"] = false;
      optimization_flags["tree-vrp"] = false;
      //optimization_flags["unit-at-a-time"] = false;
      optimization_flags["unroll-all-loops"] = false;
      optimization_flags["unroll-loops"] = false;
      optimization_flags["unsafe-loop-optimizations"] = false;
      optimization_flags["unsafe-math-optimizations"] = false;
      optimization_flags["unswitch-loops"] = false;
      optimization_flags["unwind-tables"] = false;
      optimization_flags["var-tracking"] = false;
      optimization_flags["var-tracking-assignments"] = false;
      optimization_flags["var-tracking-assignments-toggle"] = false;
      optimization_flags["var-tracking-uninit"] = false;
      optimization_flags["variable-expansion-in-unroller"] = false;
      optimization_flags["vect-cost-model"] = true;
      optimization_flags["vpt"] = false;
      optimization_flags["web"] = true;
      optimization_flags["whole-program"] = false;
      optimization_flags["wrapv"] = false;
   }
   if(opt_level ==1)
      gcc_compiling_parameters += " -O1 ";
   if(opt_level >=1)
   {
#if 0
      optimization_flags["cprop-registers"] = true;
      optimization_flags["defer-pop"] = true;
      optimization_flags["guess-branch-probability"] = true;
      optimization_flags["if-conversion"] = true;
      optimization_flags["if-conversion2"] = true;
      optimization_flags["ipa-pure-const"] = true;
      optimization_flags["ipa-reference"] = true;
      optimization_flags["merge-constants"] = true;
      optimization_flags["omit-frame-pointer"] = true;
      optimization_flags["split-wide-types"] = true;
      optimization_flags["tree-builtin-call-dce"] = true;
      optimization_flags["tree-ccp"] = true;
      optimization_flags["tree-ch"] = true;
      optimization_flags["tree-copy-prop"] = true;
      optimization_flags["tree-copyrename"] = true;
      optimization_flags["tree-dce"] = true;
      optimization_flags["tree-dominator-opts"] = true;
      optimization_flags["tree-dse"] = true;
      optimization_flags["tree-forwprop"] = true;
      optimization_flags["tree-fre"] = true;
      optimization_flags["tree-phiprop"] = true;
      optimization_flags["tree-pta"] = true;
      optimization_flags["tree-reassoc"] = true;
      optimization_flags["tree-sink"] = true;
      optimization_flags["tree-sra"] = true;
      optimization_flags["tree-ter"] = true;
#endif
   }
   if(opt_level ==2)
      gcc_compiling_parameters += " -O2 ";
   if(opt_level >=2)
   {
#if 0
      optimization_flags["align-functions"] = true;
      optimization_flags["align-jumps"] = true;
      optimization_flags["align-labels"] = true;
      optimization_flags["align-loops"] = false;
      optimization_flags["caller-saves"] = true;
      optimization_flags["crossjumping"] = true;
      optimization_flags["cse-follow-jumps"] = true;
      optimization_flags["delete-null-pointer-checks"] = true;
      optimization_flags["expensive-optimizations"] = true;
      optimization_flags["forward-propagate"] = true;
      optimization_flags["gcse"] = true;
      optimization_flags["inline-small-functions"] = true;
      optimization_flags["optimize-register-move"] = true;
      optimization_flags["optimize-sibling-calls"] = true;
      optimization_flags["peephole2"] = true;
      optimization_flags["regmove"] = true;
      optimization_flags["reorder-blocks"] = true;
      optimization_flags["reorder-functions"] = true;
      optimization_flags["sched-critical-path-heuristic"] = true;
      optimization_flags["sched-dep-count-heuristic"] = true;
      optimization_flags["sched-group-heuristic"] = true;
      optimization_flags["sched-interblock"] = true;
      optimization_flags["schedule-insns"] = true;
      optimization_flags["schedule-insns2"] = true;
      optimization_flags["strict-aliasing"] = true;
      optimization_flags["thread-jumps"] = true;
      optimization_flags["tree-pre"] = true;
      optimization_flags["tree-switch-conversion"] = true;
      optimization_flags["tree-vrp"] = true;
      optimization_flags["ipa-sra"] = true;
#endif
   }
   if(opt_level >=3)
   {
      gcc_compiling_parameters += " -O3 ";
#if 0
      optimization_flags["gcse-after-reload"] = true;
      optimization_flags["inline-functions"] = true;
      optimization_flags["predictive-commoning"] = true;
      optimization_flags["tree-slp-vectorize"] = true;
      optimization_flags["tree-vectorize"] = true;
      optimization_flags["unswitch-loops"] = true;
      optimization_flags["ipa-cp-clone"] = true;
#endif
   }
   /// required by PandA
   optimization_flags["ipa-pure-const"] = true; ///needed to correctly manage global variables
   optimization_flags["tree-dce"] = true; ///needed to remove unnecessary computations
   //optimization_flags["unit-at-a-time"] = true; ///needed to correctly manage global variables
   INDENT_DBG_MEX(DEBUG_LEVEL_VERBOSE, debug_level, "<--Set GCC defaults");
}

GccWrapper::Compiler GccWrapper::GetCompiler() const
{
   Compiler compiler;
#ifndef NDEBUG
   CompilerTarget compatible_compilers = Param->getOption<CompilerTarget>(OPT_compatible_compilers);
#endif

   std::string gcc_extra_options;
   if(Param->isOption(OPT_gcc_extra_options))
      gcc_extra_options = Param->getOption<std::string>(OPT_gcc_extra_options);

   CompilerTarget preferred_compiler;
   if(compiler_target == CT_NO_GCC)
   {
      preferred_compiler = Param->getOption<CompilerTarget>(OPT_default_compiler);
   }
   else
   {
      THROW_ASSERT((compiler_target == CT_SPARC_ELF_GCC) or (compiler_target & compatible_compilers), "Required compiler is not among the compatible one: " + boost::lexical_cast<std::string>(compiler_target) + " vs " + boost::lexical_cast<std::string>(compatible_compilers) );
      preferred_compiler = compiler_target;
   }

#if HAVE_I386_COMPILER
   if(preferred_compiler & CT_I386_GCC)
   {
      compiler.gcc = I386_GCC_EXE;
      compiler.cpp = I386_CPP_EXE;
      compiler.extra_options = " -m32 -D_FORTIFY_SOURCE=0 " + gcc_extra_options;
      compiler.empty_plugin = I386_EMPTY_PLUGIN;
      compiler.ssa_plugin = I386_SSA_PLUGIN;
#if HAVE_FROM_RTL_BUILT
      compiler.rtl_plugin = "";
#endif
   }
#endif
#if HAVE_SPARC_COMPILER
   if(preferred_compiler & CT_SPARC_GCC)
   {
      compiler.gcc = SPARC_GCC_EXE;
      compiler.cpp = SPARC_CPP_EXE;
      compiler.extra_options = gcc_extra_options;
      compiler.empty_plugin = SPARC_EMPTY_PLUGIN;
      compiler.ssa_plugin = SPARC_SSA_PLUGIN;
#if HAVE_FROM_RTL_BUILT
      compiler.rtl_plugin = SPARC_RTL_PLUGIN;
#endif
   }
   if(preferred_compiler & CT_SPARC_ELF_GCC)
   {
      compiler.gcc = SPARC_ELF_GCC;
      compiler.cpp = SPARC_ELF_CPP;
      compiler.extra_options = gcc_extra_options;
   }
#endif
#if HAVE_ARM_COMPILER
   if(preferred_compiler & CT_ARM_GCC)
   {
      compiler.gcc = ARM_GCC_EXE;
      compiler.cpp = ARM_CPP_EXE;
      compiler.extra_options = gcc_extra_options + " -mlittle-endian -fsigned-char";
      compiler.empty_plugin = ARM_EMPTY_PLUGIN;
      compiler.ssa_plugin = ARM_SSA_PLUGIN;
#if HAVE_FROM_RTL_BUILT
      compiler.rtl_plugin = ARM_RTL_PLUGIN;
#endif
   }
#endif
   if(compiler.gcc == "")
   {
      THROW_ERROR("Not found any compatible compiler");
   }
   return compiler;
}

void GccWrapper::GetSystemIncludes(std::vector<std::string> & includes) const
{
   /// This string contains the path and name of the compiler to be invoked
   const std::string cpp = GetCompiler().cpp.string();

   std::string command = cpp + " -v  < /dev/null 2>&1 | grep -v -E \"(#|Configured with|Using built-in|Target|Thread model|gcc version|End of search list|ignoring nonexistent directory|cc1 -E -quiet|cc1.exe -E -quiet|COLLECT_GCC_OPTIONS|COMPILER_PATH|LIBRARY_PATH|COLLECT_GCC_OPTIONS|ignoring duplicate directory)\" | tr '\\n' ' ' | tr '\\r' ' '  | sed 's/\\\\/\\//g'";
   int ret = PandaSystem(Param, command, STR_CST_gcc_include);
   PRINT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "");
   if(IsError(ret))
   {
      util_print_cpu_stats(std::cerr);
      THROW_ERROR("Error in retrieving gcc system include. Error is " + boost::lexical_cast<std::string>(ret));
   }

   std::string list_of_dirs;

   std::ifstream includefile(STR_CST_gcc_include);
   if (includefile.is_open())
   {
      std::string line;
      while (!includefile.eof())
      {
         getline (includefile,line);
         if (line.size())
         {
            list_of_dirs += line;
         }
      }
   }
   else 
      THROW_ERROR("Error in retrieving gcc system include");

   std::remove(STR_CST_gcc_include);

   //Ok, now here there are the list of the system paths in which
   //the system includes are found
   boost::algorithm::split(includes, list_of_dirs, boost::algorithm::is_any_of(" "));
}

void GccWrapper::GetGccConfig() const
{
   QueryGccConfig("-v");
}

void GccWrapper::QueryGccConfig(std::string gcc_option) const
{
   const std::string gcc = GetCompiler().gcc.string();
   const std::string command = gcc + " " + gcc_option;
   const std::string output_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + STR_CST_shell_output_file;
   int ret = PandaSystem(Param, command, output_file_name);
   if(IsError(ret))
   {
      THROW_ERROR("Error in retrieving gcc configuration");
   }
   CopyStdout(output_file_name);
}

size_t GccWrapper::GetSourceCodeLines(const ParameterConstRef Param)
{
#ifndef NDEBUG
   const int debug_level = Param->get_class_debug_level("GccWrapper");
#endif
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Getting source code lines number");
   ///The output file
   const std::string output_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + STR_CST_shell_output_file;

   std::string command = "cat ";
   const std::tr1::unordered_set<std::string> source_files = Param->getOption<const std::tr1::unordered_set<std::string> >(OPT_input_file);
   std::tr1::unordered_set<std::string>::const_iterator source_file, source_file_end = source_files.end();
   for(source_file = source_files.begin(); source_file != source_file_end; source_file++)
   {
      boost::filesystem::path absolute_path = boost::filesystem::system_complete(*source_file);
      command += absolute_path.branch_path().string() + "/*\\.h ";
      command += *source_file + " ";
   }
   command += std::string(" 2> /dev/null | wc -l");
   int ret = PandaSystem(Param, command, output_file_name);
   if(IsError(ret))
   {
      THROW_ERROR("Error during execution of computing word lines");
   }

   std::ifstream output_file(output_file_name.c_str());
   if (output_file.is_open() and !output_file.eof())
   {
      std::string line;
      getline (output_file, line);
      INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--Got " + line);
      return boost::lexical_cast<size_t>(line);
   }
   else
   {
      THROW_ERROR("Error in opening " + output_file_name);
   }
   return 0;
}

void GccWrapper::CreateExecutable(const std::tr1::unordered_set<std::string> & file_names, const std::string & executable_name, const std::string & extra_gcc_options) const
{
   std::string file_names_string;
   std::tr1::unordered_set<std::string>::const_iterator file_name, file_name_end = file_names.end();
   for(file_name = file_names.begin(); file_name != file_name_end; file_name++)
   {
      file_names_string += *file_name + " ";
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Creating executable " + executable_name + " from " + file_names_string);
   std::string command;

   Compiler compiler = GetCompiler();
   command = compiler.gcc.string() + " ";

   command += file_names_string + " ";

   command += gcc_compiling_parameters + " " + AddSourceCodeIncludes(file_names) + " " + gcc_linking_parameters + " ";

   command += "-D__NO_INLINE__ "; ///needed to avoid problem with glibc inlines

   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Extra options are " + compiler.extra_options);
   command += compiler.extra_options + " " + extra_gcc_options + " ";

   command += "-o " + executable_name + " ";

   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---Compilation command is " + command);
   const std::string gcc_output_file_name = Param->getOption<std::string>(OPT_output_temporary_directory) + STR_CST_gcc_output;

   int ret = PandaSystem(Param, command, gcc_output_file_name);
   if(IsError(ret))
   {
      CopyStdout(gcc_output_file_name);
      THROW_ERROR_CODE(COMPILING_EC, "GCC returns an error during compilation " + boost::lexical_cast<std::string>(errno) + " - Command is " + command);
   }
   else
   {
      if(output_level >= OUTPUT_LEVEL_VERBOSE)
         CopyStdout(gcc_output_file_name);
   }
   INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
}

void GccWrapper::ReadParameters()
{
   if(Param->isOption(OPT_gcc_optimizations))
   {
      const std::tr1::unordered_set<std::string> parameters= Param->getOption<const std::tr1::unordered_set<std::string> >(OPT_gcc_optimizations);
      std::tr1::unordered_set<std::string>::const_iterator parameter, parameter_end = parameters.end();
      for (parameter = parameters.begin(); parameter != parameter_end; parameter++)
      {
         std::string param = *parameter;
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "-->Examining parameter " + param);
         size_t found  = param.find("dump-");
         found  = param.find("param ");
         size_t found_value  = param.find("=");
         if (found != std::string::npos and found == 0)
         {
            INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "---param" + param);
            std::string value = param.substr (found_value + std::string("=").size());
            std::string key = param.substr (found + std::string("param ").size());
            found_value  = key.find("=");
            key = key.substr (0, found_value);
            parameter_values[key] = boost::lexical_cast<int>(value);
            continue;
         }
         found  = param.find("no-");
         // if the token starts with "no-", the optimization has to be disabled
         if (found != std::string::npos and found == 0)
         {
            std::string str = param.substr (found + std::string("no-").size());
            optimization_flags[str] = false;
         }
         else
         {
            // otherwise it has to be enabled, provided it is not a optimization parameter value
            if (found_value == std::string::npos)
               optimization_flags[param] = true;
            else
            {
               std::string key = param.substr (0, found_value);
               std::string value = param.substr (found_value + std::string("=").size());
               optimization_values[key] = boost::lexical_cast<int>(value);
            }
         }
         INDENT_DBG_MEX(DEBUG_LEVEL_VERY_PEDANTIC, debug_level, "<--");
      }
   }
}

std::string GccWrapper::WriteOptimizationsString()
{
   std::string optimizations;
   ///Preparing optimizations string
   bool argument_noalias = optimization_flags.find("argument-noalias") != optimization_flags.end() and optimization_flags.find("argument-noalias")->second;
   bool argument_noalias_global = optimization_flags.find("argument-noalias-global") != optimization_flags.end() and optimization_flags.find("argument-noalias-global")->second;
   bool argument_noalias_anything = optimization_flags.find("argument-noalias-anything") != optimization_flags.end() and optimization_flags.find("argument-noalias-anything")->second;
   bool strict_aliasing = optimization_flags.find("strict-aliasing") != optimization_flags.end() and optimization_flags.find("strict-aliasing")->second;
   std::map<std::string, bool>::const_iterator it, it_end = optimization_flags.end();
   if(strict_aliasing)
      optimizations += std::string("-Wstrict-aliasing ");
   for (it = optimization_flags.begin(); it != it_end; it++)
   {
      /*argument aliasing should be treated in a different way*/
      if (it->first == "argument-alias" and (argument_noalias or argument_noalias_global or argument_noalias_anything))
         continue;
      if (it->first == "argument-noalias" and (argument_noalias_global or argument_noalias_anything))
         continue;
      else if (it->first == "argument-noalias" and (argument_noalias_global or argument_noalias_anything))
         continue;
      else if (it->first == "argument-noalias-global" and (argument_noalias_anything))
         continue;
      if (it->second)
         optimizations += std::string("-f") + it->first + " ";  // enable optimizations set to true
      else
         optimizations += std::string("-fno-") + it->first + " "; // disable optimizations set to false
   }
   std::map<std::string, int>::const_iterator it2, it2_end = optimization_values.end();
   for (it2 = optimization_values.begin(); it2 != it2_end; it2++)
   {
      optimizations += std::string("-f") + it2->first + "=" + boost::lexical_cast<std::string>(it2->second) + " ";
   }
   std::map<std::string, int>::const_iterator it3, it3_end = parameter_values.end();
   for(it3 = parameter_values.begin(); it3 != it3_end; it3++)
   {
      optimizations += "--param " + it3->first + "=" + boost::lexical_cast<std::string>(it3->second) + " ";
   }
   return optimizations;
}

void GccWrapper::ReadXml(const std::string file_name)
{
   try
   {
      const fileIO_istreamRef sname = fileIO_istream_open(file_name);
      xml_dom_parser parser;
      parser.parse_stream(*sname);
      if (parser)
      {
         const xml_element * root = parser.get_document()->get_root_node();

         const xml_node::node_list root_children = root->get_children();
         xml_node::node_list::const_iterator root_child, root_child_end = root_children.end();
         for(root_child = root_children.begin(); root_child != root_child_end; root_child++)
         {
            const xml_element * root_child_element = GetPointer<const xml_element>(*root_child);
            if(not root_child_element)
               continue;
            if(root_child_element->get_name() == STR_XML_gcc_optimizations)
            {
               const xml_node::node_list optimizations_children = root_child_element->get_children();
               xml_node::node_list::const_iterator optimizations_child, optimizations_child_end = optimizations_children.end();
               for(optimizations_child = optimizations_children.begin(); optimizations_child != optimizations_child_end; optimizations_child++)
               {
                  const xml_element * optimizations_child_element = GetPointer<const xml_element>(*optimizations_child);
                  if(not optimizations_child_element)
                     continue;
                  if(optimizations_child_element->get_name() == STR_XML_gcc_parameter_values)
                  {
                     const xml_node::node_list parameter_values_children = optimizations_child_element->get_children();
                     xml_node::node_list::const_iterator parameter_value, parameter_value_end = parameter_values_children.end();
                     for(parameter_value = parameter_values_children.begin(); parameter_value != parameter_value_end; parameter_value++)
                     {
                        const xml_element * parameter_value_element = GetPointer<const xml_element>(*parameter_value);
                        if(not parameter_value_element)
                           continue;
                        if(not (parameter_value_element->get_attribute(STR_XML_gcc_name) and parameter_value_element->get_attribute(STR_XML_gcc_value)))
                        {
                           THROW_ERROR("Parameter value node without name or value");
                        }
                        parameter_values[parameter_value_element->get_attribute(STR_XML_gcc_name)->get_value()] = boost::lexical_cast<int>(parameter_value_element->get_attribute(STR_XML_gcc_value));
                     }
                  }
                  else if(optimizations_child_element->get_name() == STR_XML_gcc_optimization_flags)
                  {
                     const xml_node::node_list optimization_flags_children = optimizations_child_element->get_children();
                     xml_node::node_list::const_iterator optimization_flag, optimization_flag_end = optimization_flags_children.end();
                     for(optimization_flag = optimization_flags_children.begin(); optimization_flag != optimization_flag_end; optimization_flag++)
                     {
                        const xml_element * optimization_flag_element = GetPointer<const xml_element>(*optimization_flag);
                        if(not optimization_flag_element)
                           continue;
                        if(not (optimization_flag_element->get_attribute(STR_XML_gcc_name) and optimization_flag_element->get_attribute(STR_XML_gcc_value)))
                        {
                           THROW_ERROR("Optimization flag node without name or value");
                        }
                        optimization_flags[optimization_flag_element->get_attribute(STR_XML_gcc_name)->get_value()] = boost::lexical_cast<bool>(optimization_flag_element->get_attribute(STR_XML_gcc_value));
                     }
                  }
                  else if(optimizations_child_element->get_name() == STR_XML_gcc_optimization_values)
                  {
                     const xml_node::node_list optimization_value_children = optimizations_child_element->get_children();
                     xml_node::node_list::const_iterator optimization_value, optimization_value_end = optimization_value_children.end();
                     for(optimization_value = optimization_value_children.begin(); optimization_value != optimization_value_end; optimization_value++)
                     {
                        const xml_element * optimization_value_element = GetPointer<const xml_element>(*optimization_value);
                        if(not optimization_value_element)
                           continue;
                        if(not (optimization_value_element->get_attribute(STR_XML_gcc_name) and optimization_value_element->get_attribute(STR_XML_gcc_value)))
                        {
                           THROW_ERROR("Optimization value node without name or value");
                        }
                        optimization_values[optimization_value_element->get_attribute(STR_XML_gcc_name)->get_value()] = boost::lexical_cast<int>(optimization_value_element->get_attribute(STR_XML_gcc_value));
                     }
                  }
               }
               gcc_compiling_parameters += WriteOptimizationsString();
            }
            else if(root_child_element->get_name() == STR_XML_gcc_standard)
            {
               const xml_element * standard = root_child_element;
               if(not standard->get_attribute(STR_XML_gcc_value))
               {
                  THROW_ERROR("Standard node without value");
               }
               gcc_compiling_parameters += "--std=" + standard->get_attribute(STR_XML_gcc_value)->get_value() + " ";
            }
            else if(root_child_element->get_name() == STR_XML_gcc_defines)
            {
               const xml_node::node_list defines = root_child_element->get_children();
               xml_node::node_list::const_iterator define, define_end = defines.end();
               for(define = defines.begin(); define != define_end; define++)
               {
                  const xml_element * define_element = GetPointer<const xml_element>(*define);
                  if(not define_element)
                     continue;
                  if(not define_element->get_attribute(STR_XML_gcc_value))
                  {
                     THROW_ERROR("Optimization value node without name or value");
                  }
                  gcc_compiling_parameters += "-D" + define_element->get_attribute(STR_XML_gcc_value)->get_value() + " ";
               }
            }
            else if(root_child_element->get_name() == STR_XML_gcc_undefines)
            {
               const xml_node::node_list undefines = root_child_element->get_children();
               xml_node::node_list::const_iterator undefine, undefine_end = undefines.end();
               for(undefine = undefines.begin(); undefine != undefine_end; undefine++)
               {
                  const xml_element * undefine_element = GetPointer<const xml_element>(*undefine);
                  if(not undefine_element)
                     continue;
                  if(not undefine_element->get_attribute(STR_XML_gcc_value))
                  {
                     THROW_ERROR("Optimization value node without name or value");
                  }
                  gcc_compiling_parameters += "-U" + undefine_element->get_attribute(STR_XML_gcc_value)->get_value() + " ";
               }
            }
            else if(root_child_element->get_name() == STR_XML_gcc_warnings)
            {
               const xml_node::node_list warnings = root_child_element->get_children();
               xml_node::node_list::const_iterator warning, warning_end = warnings.end();
               for(warning = warnings.begin(); warning != warning_end; warning++)
               {
                  const xml_element * warning_element = GetPointer<const xml_element>(*warning);
                  if(not warning_element)
                     continue;
                  if(not warning_element->get_attribute(STR_XML_gcc_value))
                  {
                     THROW_ERROR("Optimization value node without name or value");
                  }
                  gcc_compiling_parameters += "-W" + warning_element->get_attribute(STR_XML_gcc_value)->get_value() + " ";
               }
            }
            else if(root_child_element->get_name() == STR_XML_gcc_includes)
            {
               const xml_node::node_list includes = root_child_element->get_children();
               xml_node::node_list::const_iterator include, include_end = includes.end();
               for(include = includes.begin(); include != include_end; include++)
               {
                  const xml_element * include_element = GetPointer<const xml_element>(*include);
                  if(not include_element)
                     continue;
                  if(not include_element->get_attribute(STR_XML_gcc_value))
                  {
                     THROW_ERROR("Optimization value node without name or value");
                  }
                  gcc_compiling_parameters += "-I" + include_element->get_attribute(STR_XML_gcc_value)->get_value() + " ";
               }
            }
            else if(root_child_element->get_name() == STR_XML_gcc_libraries)
            {
               const xml_node::node_list libraries = root_child_element->get_children();
               xml_node::node_list::const_iterator library, library_end = libraries.end();
               for(library = libraries.begin(); library != library_end; library++)
               {
                  const xml_element * library_element = GetPointer<const xml_element>(*library);
                  if(not library_element)
                     continue;
                  if(not library_element->get_attribute(STR_XML_gcc_value))
                  {
                     THROW_ERROR("Optimization value node without name or value");
                  }
                  gcc_linking_parameters += "-l" + library_element->get_attribute(STR_XML_gcc_value)->get_value() + " ";
               }
            }
            else if(root_child_element->get_name() == STR_XML_gcc_library_directories)
            {
               const xml_node::node_list library_directories = root_child_element->get_children();
               xml_node::node_list::const_iterator library_directory, library_directory_end = library_directories.end();
               for(library_directory = library_directories.begin(); library_directory != library_directory_end; library_directory++)
               {
                  const xml_element * library_directory_element = GetPointer<const xml_element>(*library_directory);
                  if(not library_directory_element)
                     continue;
                  if(not library_directory_element->get_attribute(STR_XML_gcc_value))
                  {
                     THROW_ERROR("Optimization value node without name or value");
                  }
                  gcc_compiling_parameters += "-L" + library_directory_element->get_attribute(STR_XML_gcc_value)->get_value() + " ";
               }
            }
         }
      }
   }
   catch (const char * msg)
   {
      THROW_ERROR("Error " + std::string(msg) + " during reading of gcc configuration from " + file_name);
   }
   catch (const std::string & msg)
   {
      THROW_ERROR("Error " + msg + " during reading of gcc configuration from " + file_name);
   }
   catch (const std::exception& ex)
   {
      THROW_ERROR("Error " + std::string(ex.what()) + " during reading of gcc configuration from " + file_name);
   }
   catch ( ... )
   {
      THROW_ERROR("Unknown exception during reading of gcc configuration from " + file_name);
   }

}

void GccWrapper::WriteXml(const std::string file_name) const
{
   xml_document document;
   xml_element * root = document.create_root_node(STR_XML_gcc_root);
   xml_element * optimizations = root->add_child_element(STR_XML_gcc_optimizations);
   xml_element * parameter_values_xml = optimizations->add_child_element(STR_XML_gcc_parameter_values);
   std::map<std::string, int>::const_iterator parameter_value, parameter_value_end = this->parameter_values.end();
   for(parameter_value = this->parameter_values.begin(); parameter_value != parameter_value_end; parameter_value++)
   {
      xml_element * parameter_value_xml = parameter_values_xml->add_child_element(STR_XML_gcc_parameter_value);
      parameter_value_xml->set_attribute(STR_XML_gcc_name, parameter_value->first);
      parameter_value_xml->set_attribute(STR_XML_gcc_value, boost::lexical_cast<std::string>(parameter_value->second));
   }
   xml_element * optimization_flags_xml = optimizations->add_child_element(STR_XML_gcc_parameter_values);
   std::map<std::string, bool>::const_iterator optimization_flag, optimization_flag_end = this->optimization_flags.end();
   for(optimization_flag = this->optimization_flags.begin(); optimization_flag != optimization_flag_end; optimization_flag++)
   {
      xml_element * optimization_flag_xml = optimization_flags_xml->add_child_element(STR_XML_gcc_optimization_flag);
      optimization_flag_xml->set_attribute(STR_XML_gcc_name, optimization_flag->first);
      optimization_flag_xml->set_attribute(STR_XML_gcc_value, boost::lexical_cast<std::string>(optimization_flag->second));
   }
   xml_element * optimization_values_xml = optimizations->add_child_element(STR_XML_gcc_parameter_values);
   std::map<std::string, int>::const_iterator optimization_value, optimization_value_end = this->optimization_values.end();
   for(optimization_value = this->optimization_values.begin(); optimization_value != optimization_value_end; optimization_value++)
   {
      xml_element * optimization_value_xml = optimization_values_xml->add_child_element(STR_XML_gcc_optimization_value);
      optimization_value_xml->set_attribute(STR_XML_gcc_name, optimization_value->first);
      optimization_value_xml->set_attribute(STR_XML_gcc_value, boost::lexical_cast<std::string>(optimization_value->second));
   }
   document.write_to_file_formatted(file_name);
}

const std::string GccWrapper::AddSourceCodeIncludes(const std::tr1::unordered_set<std::string> & source_files) const
{
   std::string includes;
   ///Adding includes of original source code files
   std::tr1::unordered_set<std::string>::const_iterator source_file, source_file_end = source_files.end();
   for(source_file = source_files.begin(); source_file != source_file_end; source_file++)
   {
      boost::filesystem::path absolute_path = boost::filesystem::system_complete(*source_file);
      std::string new_path =  "-iquote " + absolute_path.branch_path().string() + " ";
      if(gcc_compiling_parameters.find(new_path) == std::string::npos and includes.find(new_path) == std::string::npos)
         includes += new_path;
   }
   return includes;
}

size_t GccWrapper::ConvertVersion(const std::string version)
{
   size_t ret_value = 0;
   std::vector<std::string> version_tokens;
   boost::algorithm::split(version_tokens, version, boost::algorithm::is_any_of("."));
   for(size_t index = version_tokens.size(); index > 0; index--)
   {
      const size_t shifter = static_cast<size_t>(pow(100, static_cast<double>(version_tokens.size() - index)));
      const size_t value = boost::lexical_cast<size_t>(version_tokens[index-1]);
      ret_value += (value * shifter);
   }
   return ret_value;
}

void GccWrapper::CheckGccCompatibleVersion(const std::string gcc_version, const std::string plugin_version)
{
   current_gcc_version = gcc_version;
   current_plugin_version = plugin_version;
   const size_t gcc_version_number = ConvertVersion(gcc_version);
   const size_t plugin_version_number = ConvertVersion(plugin_version);
   const size_t minimum_plugin_version_number = ConvertVersion(STR_CST_gcc_min_plugin_version);
   const size_t maximum_plugin_version_number = ConvertVersion(STR_CST_gcc_max_plugin_version);
   if(plugin_version_number < minimum_plugin_version_number or plugin_version_number > maximum_plugin_version_number)
   {
      THROW_ERROR("Too old plugin. At least plugin version " + std::string(STR_CST_gcc_min_plugin_version) + " is required");
   }
   if(std::string(STR_CST_gcc_supported_versions).find(gcc_version) != std::string::npos)
      return;
   if(gcc_version_number < ConvertVersion(STR_CST_gcc_first_not_supported))
   {
      THROW_WARNING("Gcc " + gcc_version + " has not been tested with tool");
      return;
   }
   THROW_ERROR("GCC " + gcc_version + " is not suppoted by this version of the tool"); 
}


