/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * @file gcc_wrapper.hpp
 * @brief Wrapper to Gcc for C sources.
 *
 * A object used to invoke the GCC compiler from sources and create the dump.
 *
 * @author Christian Pilato <pilato@elet.polimi.it>
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
 */
#ifndef GCC_WRAPPER_HPP
#define GCC_WRAPPER_HPP

///Autoheader include
#include "config_HAVE_FROM_RTL_BUILT.hpp"

///STD include
#include <iostream>
#include <map>
#include <set>
#include <tr1/unordered_map>
#include <tr1/unordered_set>
#include <vector>

///Utility include
#include <boost/filesystem/path.hpp>
#include "dbgPrintHelper.hpp"
#include "global_enums.hpp"
#include "refcount.hpp"

CONSTREF_FORWARD_DECL(Parameter);
REF_FORWARD_DECL(Parameter);
REF_FORWARD_DECL(tree_manager);
/**
 * @class GccWrapper
 * Main class for wrapping the GCC compiler. It allows an XML configuration file to be specified where the parameters
 * and the related values are stored.
 */
class GccWrapper
{
   private:
      ///Class storing information of a compiler
      class Compiler
      {
         public:
            ///The cpp executable
            boost::filesystem::path cpp;

            ///The gcc executable
            boost::filesystem::path gcc;

            ///The extra_options
            std::string extra_options;

            ///The plugin to dump empty gimple
            std::string empty_plugin;

            ///The plugin to dump ssa gimple
            std::string ssa_plugin;

#if HAVE_FROM_RTL_BUILT
            ///The plugin to dump gimple and rtl
            std::string rtl_plugin;
#endif
      };

      /// The set of input parameters
      const ParameterConstRef Param;

      ///The target compiler to be used
      const CompilerTarget compiler_target;

      ///The set of optimizations to be applied
      const OptimizationSet OS;

      ///Counter used to create unique file name for patched files
      unsigned int file_name_counter;

      ///The gcc parameters line for compiling a file
      std::string gcc_compiling_parameters;

      ///The gcc parameters line for creating an executable
      std::string gcc_linking_parameters;

      /// debug level
      int output_level;

      /// debug level
      int debug_level;

      ///The values of optimizations parameters
      std::map<std::string, int> parameter_values;

      ///The set of activated optimizations
      std::map<std::string, bool> optimization_flags;

      ///The value of parametric activate optimizations
      std::map<std::string, int> optimization_values;

      /**
       * Invoke gcc to compile file(s)
       * @param real_filename stores the original filename
       * @param file_string stores the files to be compiled
       * @param parameters_line are the parameters to be passed to gcc
       * @param empty_file tells if .001.tu tree has to be produced
       */
      void CompileFile(const std::string real_filename, std::string file_string, const std::string & parameters_line, bool empty_file = false);

      /**
       * Return the compiler for a given target
       * @return a structure containing information about compiler
       */
      Compiler GetCompiler() const;

      /**
       * Initialize the gcc parameters line
       * @param OS is the optimizations set to be considered
       * @return the string with the parameters
       */
      void InitializeGccParameters();

      /**
       * Analyze the command line options
       */
      void ReadParameters();

      /**
       * Set the default options for gcc
       */
      void SetGccDefault();

#if HAVE_BAMBU_BUILT
      /**
       * Set the default options for gcc in bambu
       */
      void SetBambuDefault();
#endif

#if HAVE_ZEBU_BUILT
      /**
       * Set the default options for gcc in zebu
       */
      void SetZebuDefault();
#endif

      /**
       * Write the string containing gcc optimization options
       * @return the string with optimization options to be passed to the gcc
       */
      std::string WriteOptimizationsString();

      /**
       * Add includes of source file directories
       * @param source_files are the source files to be considered
       * @return the string to be passed to gcc
       */
      const std::string AddSourceCodeIncludes(const std::tr1::unordered_set<std::string> & source_files) const;

      /**
       * Convert a string version to a number
       * @param version is the version to be converted
       * @return the corresponding number
       */
      static
         size_t ConvertVersion(const std::string version);

   public:
      ///The version of the gcc
      static
         std::string current_gcc_version;

      ///The version of the plugin
      static
         std::string current_plugin_version;

      /**
       * Constructor
       * @param Param is the set of parameters
       * @param compiler is the compiler to be used
       * @param OS is the optimization set
       */
      GccWrapper(const ParameterConstRef Param, const CompilerTarget compiler, const OptimizationSet OS);

      /**
       * Destructor
       */
      ~GccWrapper();

      /**
       * This function produces a single tree_manager starting from a list of source files
       * @param TM is where tree_manager will be stored
       * @param source_files are the source files to be compiled
       */
      void CreateTreeManager(tree_managerRef & TM, const std::tr1::unordered_set<std::string> & source_files);

      /**
       * Return the list of gcc system include
       * @param includes is where result will be stored
       */
      void GetSystemIncludes(std::vector<std::string> & includes) const;

      /**
       * Dump the gcc configuration
       */
      void GetGccConfig() const;

      /**
       * Function that print of stdout some useful information passing the given option
       */
      void QueryGccConfig(std::string gcc_option) const;

      /**
       * Return the total number of lines of the benchmark
       * @param Param is the set of input parameters
       * @return the total number of lines of the benchmark
       */
      static
         size_t GetSourceCodeLines(const ParameterConstRef Param);

      /**
       * Create an executable starting from source code
       * @param file_names is the list of string
       * @param executable_name is the name of the executable
       * @param extra_gcc_options is extra options to be used only for this compilation
       */
      void CreateExecutable(const std::tr1::unordered_set<std::string> & file_names, const std::string & executable_name, const std::string & extra_gcc_options) const;

      /**
       * Read gcc configuration from file
       * @param file_name is the name of the file
       */
      void ReadXml(const std::string file_name);

      /**
       * Write gcc configuration on file
       * @param file_name is the name of the file
       */
      void WriteXml(const std::string file_name) const;

      /**
       * Check if the combination gcc-plugin is supported; if not it throws error
       * @param gcc_version is the gcc version in form x.x.x
       * @param plugin_version is the plugin version in form x.x
       */
      static
         void CheckGccCompatibleVersion(const std::string gcc_version, const std::string plugin_version);
};

/// Refcount definition for the GccWrapper class
typedef refcount<GccWrapper> GccWrapperRef;
typedef refcount<const GccWrapper> GccWrapperConstRef;

#endif
