/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file FSM_NI_SSA_liveness.cpp
 * @brief liveness analysis exploiting the SSA form of the IR
 *
 * @author Fabrizio Ferrandi <ferrandi@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
#include "FSM_NI_SSA_liveness.hpp"
#include "liveness.hpp"
#include "memory.hpp"

#include "hls.hpp"
#include "hls_constraints.hpp"

#include "StateTransitionGraph.hpp"
#include "basic_block.hpp"
#include "op_node_info.hpp"

///Behavior include
#include "application_manager.hpp"
#include "behavioral_helper.hpp"
#include "function_behavior.hpp"

#include "tree_helper.hpp"

#include "Parameter.hpp"
#include "dbgPrintHelper.hpp"
#if 0
#include "Dominance.hpp"
#endif

FSM_NI_SSA_liveness::FSM_NI_SSA_liveness(const ParameterRef _Param, const hlsRef _HLS) :
   liveness_computer(_Param, _HLS)
{

}

FSM_NI_SSA_liveness::~FSM_NI_SSA_liveness()
{
}

void FSM_NI_SSA_liveness::exec()
{
   HLS->Rliv = livenessRef(new liveness(Param));
   const graph *astg = HLS->STG->get_astg();
   const graph *stg = HLS->STG->get_stg();
   const bb_graph * fbb = HLS->FB->get_bb_graph(function_behavior::FBB);
   const graphRef data = HLS->get_graph(function_behavior::DFG);

   /// Map between basic block node index and vertices
   std::tr1::unordered_map<unsigned int, vertex> bb_index_map;
   VertexIterator vit, vend;
   for(tie(vit, vend) = boost::vertices(*fbb); vit != vend; ++vit)
   {
      bb_index_map[CGET_BB_NODE_INFO(fbb, *vit)->get_bb_index()] = *vit;
   }

   /// vertex of the STG analyzed in reverse order
   std::deque<vertex> reverse_order_state_list;
   boost::topological_sort(*astg, std::back_inserter(reverse_order_state_list));
   const vertex bb_exit = fbb->GetConstBBGraphInfo()->exit_vertex;
   unsigned int prev_bb_index = CGET_BB_NODE_INFO(fbb, bb_exit)->get_bb_index();
   vertex exit_state = HLS->STG->get_exit_state();
   vertex entry_state = HLS->STG->get_entry_state();
   vertex prev_state = exit_state;
   bool second_state_of_BB = false;
   std::deque<vertex> dummy_states;

#if 0
   /// compute the dominator tree for the FSM
   refcount<dominance<graph> > stg_dominators = refcount<dominance<graph> >(new dominance<graph>(*stg, entry_state, exit_state, Param));
   stg_dominators->calculate_dominance_info(dominance<graph>::CDI_DOMINATORS);
   const std::tr1::unordered_map<vertex, vertex>& stg_dominator_map = stg_dominators->get_dominator_map();
   std::tr1::unordered_map<vertex, std::tr1::unordered_set<vertex> > out_edges;
   const std::tr1::unordered_map<vertex, vertex>::const_iterator sdm_it_end = stg_dominator_map.end();
   for(std::tr1::unordered_map<vertex, vertex>::const_iterator sdm_it = stg_dominator_map.begin(); sdm_it != sdm_it_end; ++sdm_it)
      out_edges[sdm_it->second].insert(sdm_it->first);

   std::tr1::unordered_set<vertex> visited_states;
   std::list<vertex> to_visit;
   to_visit.push_front(entry_state);
   while(!to_visit.empty())
   {
      vertex actual = to_visit.front();
      to_visit.pop_front();
      if(visited_states.find(actual)!= visited_states.end()) continue;
      visited_states.insert(actual);
      HLS->Rliv->add_support_state(actual);
      //std::cerr << "state " << HLS->STG->get_state_name(actual) << std::endl;
      if(out_edges.find(actual) == out_edges.end()) continue;
      const std::tr1::unordered_set<vertex> & out_vertices = out_edges.find(actual)->second;
      const std::tr1::unordered_set<vertex>::const_iterator ov_it_end = out_vertices.end();
      ///Examining successors
      for(std::tr1::unordered_set<vertex>::const_iterator ov_it = out_vertices.begin(); ov_it !=  ov_it_end; ++ov_it)
      {
         to_visit.push_back(*ov_it);
      }
   }
#endif

   std::deque<vertex>::const_iterator rosl_it_end = reverse_order_state_list.end();
   for(std::deque<vertex>::const_iterator rosl_it = reverse_order_state_list.begin(); rosl_it != rosl_it_end; ++rosl_it)
   {
      //std::cerr << "state: " << HLS->STG->get_state_name(*rosl_it) << std::endl;
      HLS->Rliv->add_name(*rosl_it, HLS->STG->get_state_name(*rosl_it));
      HLS->Rliv->add_support_state(*rosl_it);
      if(*rosl_it == exit_state) continue;
      const std::list<vertex>& ending_operations_cur = GET_ENDING_OPERATION_LIST(astg, *rosl_it);
      std::list<vertex>::const_reverse_iterator eoc_it_end = ending_operations_cur.rend();
      for(std::list<vertex>::const_reverse_iterator eoc_it = ending_operations_cur.rbegin(); eoc_it != eoc_it_end; ++eoc_it)
      {
         if(!(GET_SSA_WRITTEN(data,*eoc_it).empty()))
         {
            std::tr1::unordered_set<unsigned int>::const_iterator it_end = GET_SSA_WRITTEN(data,*eoc_it).end();
            for(std::tr1::unordered_set<unsigned int>::const_iterator it = GET_SSA_WRITTEN(data,*eoc_it).begin(); it != it_end; ++it)
               if(is_register_compatible(*it))
               {
                  HLS->Rliv->add_op_definition(*it, *eoc_it);
               }
         }
         HLS->Rliv->add_state_for_ending_op(*eoc_it, *rosl_it);
      }

      if(GET_DUMMY_STATE(astg, *rosl_it))
      {
         dummy_states.push_back(*rosl_it);
         continue;
      }
      vertex last_operation;
      if(GET_OPERATION_LIST(astg, *rosl_it).empty())
      {
         THROW_ASSERT(!GET_ENDING_OPERATION_LIST(astg, *rosl_it).empty(), "at least one operation has to be in execution in state: " + HLS->STG->get_state_name(*rosl_it));
         last_operation = *(GET_ENDING_OPERATION_LIST(astg, *rosl_it).rbegin());
      }
      else
      {
         last_operation = *(GET_OPERATION_LIST(astg, *rosl_it).rbegin());
      }
      unsigned int bb_index = GET_BB_INDEX(data, last_operation);

      if(prev_bb_index != bb_index)
      {
         std::set<unsigned int>::const_iterator lo_it_end = CGET_BB_NODE_INFO(fbb, bb_index_map[bb_index])->get_live_out().end();
         for(std::set<unsigned int>::const_iterator lo_it = CGET_BB_NODE_INFO(fbb, bb_index_map[bb_index])->get_live_out().begin(); lo_it != lo_it_end; ++lo_it)
            if(is_register_compatible(*lo_it))
               HLS->Rliv->set_live_out(*rosl_it, *lo_it);
         second_state_of_BB = true;
      }
      else
      {
         if(second_state_of_BB)
         {
            second_state_of_BB = false;
            //const std::map<unsigned int, unsigned int> & live_out_phi_defs = CGET_BB_NODE_INFO(fbb, bb_index_map[bb_index])->get_live_out_phi_defs();
            std::set<unsigned int>::const_iterator lo_it_end = HLS->Rliv->get_live_out(prev_state).end();
            for(std::set<unsigned int>::const_iterator lo_it = HLS->Rliv->get_live_out(prev_state).begin(); lo_it != lo_it_end; ++lo_it)
            {
               //if(live_out_phi_defs.find(*lo_it) == live_out_phi_defs.end())
               HLS->Rliv->set_live_out(*rosl_it, *lo_it);
               //else if(is_register_compatible(live_out_phi_defs.find(*lo_it)->second))
               //   HLS->Rliv->set_live_out(*rosl_it, live_out_phi_defs.find(*lo_it)->second);
            }
         }
         else
            HLS->Rliv->set_live_out(*rosl_it, HLS->Rliv->get_live_out(prev_state).begin(), HLS->Rliv->get_live_out(prev_state).end());
         const std::list<vertex>& executing_operations = GET_OPERATION_LIST(astg, prev_state);
         std::list<vertex>::const_reverse_iterator exo_it_end = executing_operations.rend();
         for(std::list<vertex>::const_reverse_iterator exo_it = executing_operations.rbegin(); exo_it != exo_it_end; ++exo_it)
         {
            std::list<unsigned int>::const_iterator sr_it_end = GET_SSA_READ(data,*exo_it).end();
            for(std::list<unsigned int>::const_iterator sr_it = GET_SSA_READ(data,*exo_it).begin(); sr_it != sr_it_end; ++sr_it)
            {
               if(is_register_compatible(*sr_it))
                  HLS->Rliv->set_live_out(*rosl_it, *sr_it);
            }
         }
         const std::list<vertex>& ending_operations = GET_ENDING_OPERATION_LIST(astg, prev_state);
         std::list<vertex>::const_reverse_iterator eo_it_end = ending_operations.rend();
         for(std::list<vertex>::const_reverse_iterator eo_it = ending_operations.rbegin(); eo_it != eo_it_end; ++eo_it)
         {
            if(!(GET_SSA_WRITTEN(data,*eo_it).empty()))
            {
               std::tr1::unordered_set<unsigned int>::const_iterator it_end = GET_SSA_WRITTEN(data,*eo_it).end();
               for(std::tr1::unordered_set<unsigned int>::const_iterator it = GET_SSA_WRITTEN(data,*eo_it).begin(); it != it_end; ++it)
                  if(is_register_compatible(*it))
                     HLS->Rliv->erase_el_live_out(*rosl_it, *it);
            }
         }
      }
      prev_state = *rosl_it;
      prev_bb_index = bb_index;
   }

   /// compute the live in of the state traversing the state list in topological order
   prev_state = entry_state;
   std::deque<vertex>::const_reverse_iterator osl_it_end = reverse_order_state_list.rend();
   for(std::deque<vertex>::const_reverse_iterator osl_it = reverse_order_state_list.rbegin(); osl_it != osl_it_end; ++osl_it)
   {
      if(GET_DUMMY_STATE(astg, *osl_it)) continue;
      vertex last_operation;
      if(GET_OPERATION_LIST(astg, *osl_it).empty())
      {
         THROW_ASSERT(!GET_ENDING_OPERATION_LIST(astg, *osl_it).empty(), "at least one operation has to be in execution in state: " + HLS->STG->get_state_name(*osl_it));
         last_operation = *(GET_ENDING_OPERATION_LIST(astg, *osl_it).rbegin());
      }
      else
      {
         last_operation = *(GET_OPERATION_LIST(astg, *osl_it).rbegin());
      }
      unsigned int bb_index = GET_BB_INDEX(data, last_operation);

      if(prev_bb_index != bb_index)
      {
         std::set<unsigned int>::const_iterator lo_it_end = CGET_BB_NODE_INFO(fbb, bb_index_map[bb_index])->get_live_in().end();
         for(std::set<unsigned int>::const_iterator lo_it = CGET_BB_NODE_INFO(fbb, bb_index_map[bb_index])->get_live_in().begin(); lo_it != lo_it_end; ++lo_it)
            if(is_register_compatible(*lo_it))
               HLS->Rliv->set_live_in(*osl_it, *lo_it);
      }
      else
         HLS->Rliv->set_live_in(*osl_it, HLS->Rliv->get_live_out(prev_state).begin(), HLS->Rliv->get_live_out(prev_state).end());
      prev_state = *osl_it;
      prev_bb_index = bb_index;

   }

   /// fix the live in/out of dummy states
   std::deque<vertex>::const_iterator ds_it_end = dummy_states.end();
   for(std::deque<vertex>::const_iterator ds_it = dummy_states.begin(); ds_it != ds_it_end; ++ds_it)
   {
      InEdgeIterator e_it, e_it_end;
      for(tie(e_it, e_it_end) = boost::in_edges(*ds_it, *astg); e_it != e_it_end; ++e_it)
      {
         vertex src_state = boost::source(*e_it, *astg);
         HLS->Rliv->set_live_out(*ds_it, HLS->Rliv->get_live_out(src_state).begin(), HLS->Rliv->get_live_out(src_state).end());
         HLS->Rliv->set_live_in(*ds_it, HLS->Rliv->get_live_out(src_state).begin(), HLS->Rliv->get_live_out(src_state).end());
         /// add all the uses of *ds_it to src_state
         const std::list<vertex>& executing_operations = GET_OPERATION_LIST(astg, *ds_it);
         std::list<vertex>::const_iterator eo_it_end = executing_operations.end();
         for(std::list<vertex>::const_iterator eo_it = executing_operations.begin(); eo_it != eo_it_end; ++eo_it)
         {
            std::list<unsigned int>::const_iterator sr_it_end = GET_SSA_READ(data,*eo_it).end();
            for(std::list<unsigned int>::const_iterator sr_it = GET_SSA_READ(data,*eo_it).begin(); sr_it != sr_it_end; ++sr_it)
            {
               if(is_register_compatible(*sr_it))
               {
                  HLS->Rliv->set_live_out(src_state, *sr_it);
                  HLS->Rliv->set_live_in(*ds_it, *sr_it);
               }
            }
         }
         /// remove all the variable written by the dummy state from its live in
         const std::list<vertex>& ending_operations_cur = GET_ENDING_OPERATION_LIST(astg, *ds_it);
         std::list<vertex>::const_iterator eoc_it_end = ending_operations_cur.end();
         for(std::list<vertex>::const_iterator eoc_it = ending_operations_cur.begin(); eoc_it != eoc_it_end; ++eoc_it)
         {
            if(!(GET_SSA_WRITTEN(data,*eoc_it).empty()))
            {
               std::tr1::unordered_set<unsigned int>::const_iterator sw_it_end = GET_SSA_WRITTEN(data,*eoc_it).end();
               for(std::tr1::unordered_set<unsigned int>::const_iterator sw_it = GET_SSA_WRITTEN(data,*eoc_it).begin(); sw_it != sw_it_end; ++sw_it)
                  if(is_register_compatible(*sw_it))
                  {
                     HLS->Rliv->erase_el_live_in(*ds_it, *sw_it);
                  }
            }
         }
      }
   }

   /// fix the liveness between pair of states given a use or a definition of a variable
   for(std::deque<vertex>::const_iterator rosl_it = reverse_order_state_list.begin(); rosl_it != rosl_it_end; ++rosl_it)
   {
      const std::list<vertex>& ending_operations_cur = GET_ENDING_OPERATION_LIST(astg, *rosl_it);
      const std::list<vertex>& running_operations_cur = GET_OPERATION_LIST(astg, *rosl_it);
      std::list<vertex>::const_iterator roc_it_end = running_operations_cur.end();
      for(std::list<vertex>::const_iterator roc_it = running_operations_cur.begin(); roc_it != roc_it_end; ++roc_it)
      {
         HLS->Rliv->add_state_for_running_op(*roc_it, *rosl_it);
         if((GET_TYPE(data, *roc_it) & TYPE_PHI) != 0)
         {
            std::vector<application_manager::io_binding_type> var_read = HLS->AppM->get_required_values(HLS->functionId, *roc_it);
            unsigned int in_index = 0;
            const std::vector<application_manager::io_binding_type>::const_iterator sr_it_end = var_read.end();
            for(std::vector<application_manager::io_binding_type>::const_iterator sr_it = var_read.begin(); sr_it != sr_it_end; ++sr_it,++in_index)
            {
               unsigned int tree_var = std::tr1::get<0>(*sr_it);
               if(tree_var == 0) continue;
               const std::list<unsigned int> & tree_node_operations = GET_NODEID(data, *roc_it);
               THROW_ASSERT(tree_node_operations.size() == 1, "unxpected condition");
               unsigned int bb_index = tree_helper::get_bb_edge(HLS->AppM->get_tree_manager(), tree_node_operations.front(), in_index);
               /// now we look for the last state with operations belonging to basic block bb_index
#ifndef NDEBUG
               bool found_state = false;
#endif
               InEdgeIterator ie, iend;
               for(tie(ie, iend) = boost::in_edges(*rosl_it, *stg); ie != iend; ie++)
               {
                  vertex src_state = boost::source(*ie, *stg);
                  vertex last_operation;
                  if(GET_OPERATION_LIST(stg, src_state).empty())
                  {
                     THROW_ASSERT(!GET_ENDING_OPERATION_LIST(stg, src_state).empty(), "at least one operation has to be in execution in state: " + HLS->STG->get_state_name(src_state));
                     last_operation = *(GET_ENDING_OPERATION_LIST(stg, src_state).rbegin());
                  }
                  else
                  {
                     last_operation = *(GET_OPERATION_LIST(stg, src_state).rbegin());
                  }
                  unsigned int state_bb_index = GET_BB_INDEX(data, last_operation);
                  if(state_bb_index == bb_index)
                  {
                     if(src_state == entry_state)
                        HLS->Rliv->add_state_in_for_var(tree_var, *roc_it, *rosl_it, NULL_VERTEX);
                     else
                        HLS->Rliv->add_state_in_for_var(tree_var, *roc_it, *rosl_it, src_state);
#ifndef NDEBUG
                     found_state = true;
#endif
                  }
               }
               THROW_ASSERT(found_state, "unexpected condition");
            }
         }
         else
         {
            OutEdgeIterator oe, oend;
            for(tie(oe, oend) = boost::out_edges(*rosl_it, *stg); oe != oend; oe++)
            {
               vertex tgt_state = boost::target(*oe, *stg);
               std::list<unsigned int>::const_iterator sr_it_end = GET_SSA_READ(data,*roc_it).end();
               for(std::list<unsigned int>::const_iterator sr_it = GET_SSA_READ(data,*roc_it).begin(); sr_it != sr_it_end; ++sr_it)
               {
                  if(is_register_compatible(*sr_it))
                     HLS->Rliv->add_state_in_for_var(*sr_it, *roc_it, *rosl_it, tgt_state);
                  else if(HLS->HLS_C->Rmem->has_base_address(*sr_it) ||
                       (tree_helper::is_ssa_name(HLS->AppM->get_tree_manager(), *sr_it) && tree_helper::is_parameter(HLS->AppM->get_tree_manager(), *sr_it)))
                  {
                     HLS->Rliv->add_state_in_for_var(*sr_it, *roc_it, *rosl_it, tgt_state);
                  }
                  else if(tree_helper::is_ssa_name(HLS->AppM->get_tree_manager(), *sr_it) && tree_helper::is_virtual(HLS->AppM->get_tree_manager(), *sr_it))
                  {
                     ;
                  }
                  else
                  {
                     THROW_ERROR("unexpected condition " +STR(*sr_it));
                  }
               }
            }
         }
      }

      std::list<vertex>::const_iterator eoc_it_end = ending_operations_cur.end();
      for(std::list<vertex>::const_iterator eoc_it = ending_operations_cur.begin(); eoc_it != eoc_it_end; ++eoc_it)
      {
         if((GET_TYPE(data, *eoc_it) & TYPE_PHI) != 0) continue;
         if(!(GET_SSA_WRITTEN(data,*eoc_it).empty()))
         {
            std::tr1::unordered_set<unsigned int>::const_iterator it_end = GET_SSA_WRITTEN(data,*eoc_it).end();
            for(std::tr1::unordered_set<unsigned int>::const_iterator it = GET_SSA_WRITTEN(data,*eoc_it).begin(); it != it_end; ++it)
               if(is_register_compatible(*it))
               {
                  OutEdgeIterator e_it, e_it_end;
                  for(tie(e_it, e_it_end) = boost::out_edges(*rosl_it, *stg); e_it != e_it_end; ++e_it)
                  {
                     vertex tgt_state = boost::target(*e_it, *stg);
                     if(HLS->Rliv->get_live_in(tgt_state).find(*it) != HLS->Rliv->get_live_in(tgt_state).end())
                        HLS->Rliv->add_state_out_for_var(*it, *eoc_it, *rosl_it, tgt_state);
                  }
               }
         }
      }
   }


#ifndef NDEBUG
   if(debug_level>= DEBUG_LEVEL_PEDANTIC)
   {
      /// print the analysis result
      const behavioral_helperRef & BH = HLS->FB->get_helper();
      PRINT_DBG_MEX(DEBUG_LEVEL_PEDANTIC, debug_level, "STG Liveness for function " + BH->get_function_name());
      VertexIterator vIt, vEnd;
      for(tie(vIt, vEnd) = boost::vertices(*astg); vIt != vEnd; vIt++)
      {
         PRINT_DBG_STRING(DEBUG_LEVEL_PEDANTIC, debug_level, "Live In for state " + HLS->STG->get_state_name(*vIt) + ": ");
         std::set<unsigned int>::const_iterator li_it_end = HLS->Rliv->get_live_in(*vIt).end();
         for(std::set<unsigned int>::const_iterator li_it = HLS->Rliv->get_live_in(*vIt).begin(); li_it != li_it_end; ++li_it)
         {
            PRINT_DBG_STRING(DEBUG_LEVEL_PEDANTIC, debug_level,  BH->PrintVariable(*li_it) + " ");
         }
         PRINT_DBG_STRING(DEBUG_LEVEL_PEDANTIC, debug_level, "\n");
         PRINT_DBG_STRING(DEBUG_LEVEL_PEDANTIC, debug_level, "Live Out for state " + HLS->STG->get_state_name(*vIt) + ": ");
         std::set<unsigned int>::const_iterator lo_it_end = HLS->Rliv->get_live_out(*vIt).end();
         for(std::set<unsigned int>::const_iterator lo_it = HLS->Rliv->get_live_out(*vIt).begin(); lo_it != lo_it_end; ++lo_it)
         {
            PRINT_DBG_STRING(DEBUG_LEVEL_PEDANTIC, debug_level,  BH->PrintVariable(*lo_it) + " ");
         }
         PRINT_DBG_STRING(DEBUG_LEVEL_PEDANTIC, debug_level, "\n");
      }
   }
#endif

}

std::string FSM_NI_SSA_liveness::get_kind_text() const
{
   return "FSM_NI_SSA_liveness";
}
