/*
 *
 *                   _/_/_/    _/_/   _/    _/ _/_/_/    _/_/
 *                  _/   _/ _/    _/ _/_/  _/ _/   _/ _/    _/
 *                 _/_/_/  _/_/_/_/ _/  _/_/ _/   _/ _/_/_/_/
 *                _/      _/    _/ _/    _/ _/   _/ _/    _/
 *               _/      _/    _/ _/    _/ _/_/_/  _/    _/
 *
 *             ***********************************************
 *                              PandA Project 
 *                     URL: http://panda.dei.polimi.it
 *                       Politecnico di Milano - DEI
 *                        System Architectures Group
 *             ***********************************************
 *              Copyright (c) 2004-2012 Politecnico di Milano
 *
 *   This file is part of the PandA framework.
 *
 *   The PandA framework is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
*/
/**
 * @file GimpleWriter.cpp
 * @brief tree node writer. This class exploiting the visitor design pattern write a tree node according to the gimple format (i.e. the format used by gcc in plain text dump of gimple)
 *
 * @author Marco Lattuada <lattuada@elet.polimi.it>
 * $Revision: 10608 $
 * $Date: 2012-03-12 10:46:42 +0100 (Mon, 12 Mar 2012) $
 * Last modified by $Author: ferrandi $
 *
*/
///Skipping warnings due to operator() redefinition
#pragma GCC diagnostic ignored "-Woverloaded-virtual"

///Header include
#include "gimple_writer.hpp"

///Behavior include
#include "basic_block.hpp"

///Tree include
#include "ext_tree_node.hpp"
#include "token_interface.hpp"
#include "tree_basic_block.hpp"
#include "tree_helper.hpp"
#include "tree_node.hpp"
#include "tree_reindex.hpp"

GimpleWriter::GimpleWriter(std::ostream & _os, const bool _use_uid) :
   os(_os),
   use_uid(_use_uid),
   current_node_index(0)
{
}

void GimpleWriter::operator()(const tree_node * , unsigned int & )
{
}

void GimpleWriter::operator()(const WeightedNode * obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const tree_reindex * obj, unsigned int & )
{
   current_node_index = obj->index;
//   obj->actual_tree_node->visit(this);
}

void GimpleWriter::operator()(const attr* , unsigned int & mask)
{
   mask = NO_VISIT;
}

void GimpleWriter::operator()(const srcp* , unsigned int & mask)
{
   mask = NO_VISIT;
}

void GimpleWriter::operator()(const vuse* , unsigned int & mask)
{
   mask = NO_VISIT;
}

void GimpleWriter::operator()(const vdef* , unsigned int & mask)
{
   mask = NO_VISIT;
}

void GimpleWriter::operator()(const vops* , unsigned int & mask)
{
   mask = NO_VISIT;
}

void GimpleWriter::operator()(const decl_node* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
   obj->srcp::visit(this);
}

void GimpleWriter::operator()(const expr_node* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->WeightedNode::visit(this);
   obj->srcp::visit(this);
}

void GimpleWriter::operator()(const expr_node_vops* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->WeightedNode::visit(this);
   obj->srcp::visit(this);
}

void GimpleWriter::operator()(const unary_expr* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   switch(obj->get_kind())
   {
      case addr_expr_K:
      {
         os << "&";
         break;
      }
      case float_expr_K:
      case nop_expr_K:
      case convert_expr_K:
      {
         os << "(";
         obj->type->visit(this);
         os << ") ";
         break;
      }
      case view_convert_expr_K:
      {
         os << "VIEW_CONVERT_EXPR<";
         obj->type->visit(this);
         os << ">(";
         obj->op->visit(this);
         os << ")";
         obj->expr_node::visit(this);
         return;
      }
      case imagpart_expr_K:
      {
         os << "IMAG_EXR<";
         obj->op->visit(this);
         os << ">";
         break;
      }
      case realpart_expr_K:
      {
         os << "REAL_EXR<";
         obj->op->visit(this);
         os << ">";
         break;
      }
      default:
      {
         const std::string op = tree_helper::op_symbol(obj);
         os << op;
         break;
      }
   }
   if(GET_NODE(obj->op)->get_kind() == function_decl_K)
      GetPointer<function_decl>(GET_NODE(obj->op))->name->visit(this);
   else
      obj->op->visit(this);
   obj->expr_node::visit(this);
}

void GimpleWriter::operator()(const binary_expr* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   switch(obj->get_kind())
   {
      case max_expr_K:
      {
         os << "MAX_EXPR <";
         obj->op0->visit(this);
         os << ", ";
         obj->op1->visit(this);
         os << ">";
         break;
      }
      case min_expr_K:
      {
         os << "MIN_EXPR <";
         obj->op0->visit(this);
         os << ", ";
         obj->op1->visit(this);
         os << ">";
         break;
      }
      case complex_expr_K:
      {
         os << "COMPLEX_EXPR <";
         obj->op0->visit(this);
         os << ", ";
         obj->op1->visit(this);
         os << ">";
         break;
      }
      default:
      {
         obj->op0->visit(this);
         const std::string op = tree_helper::op_symbol (obj);
         os << " " << op << " ";
         obj->op1->visit(this);
         break;
      }
   }
   obj->expr_node::visit(this);
}

void GimpleWriter::operator()(const ternary_expr* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   switch(obj->get_kind())
   {
      case bit_field_ref_K:
      {
         os << "BIT_FIELD_REF <";
         obj->op0->visit(this);
         os << ", ";
         obj->op1->visit(this);
         os << ", ";
         obj->op2->visit(this);
         os << ">";
         break;
      }
      case component_ref_K:
      {
         const indirect_ref * ir = GetPointer<indirect_ref>(GET_NODE(obj->op0));
         if(ir)
         {
            ir->op->visit(this);
            os << "->";
            obj->op1->visit(this);
         }
         else
         {
            obj->op0->visit(this);
            os << ".";
            obj->op1->visit(this);
         }
         break;
      }
      default:
      {
         break;
      }
   }
   obj->expr_node::visit(this);
}

void GimpleWriter::operator()(const quaternary_expr* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   if(obj->get_kind() == array_ref_K)
   {
      obj->op0->visit(this);
      os << "[";
      obj->op1->visit(this);
      os << "]";
   }
   obj->expr_node::visit(this);
}

void GimpleWriter::operator()(const type_node* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   if(obj->qual)
   {
      std::string replace = tree_helper::return_C_qualifiers(obj->qual, true);
      boost::replace_all(replace, "__restrict__ ", " restrict");
      os << replace;
   }
   if(obj->name)
   {
      obj->name->visit(this);
   }
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const memory_tag* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->decl_node::visit(this);
}

void GimpleWriter::operator()(const cst_node*, unsigned int &)
{
}

void GimpleWriter::operator()(const array_type* obj, unsigned int & mask)
{
   obj->elts->visit(this);
   os << "[";
   ///Computing size
   tree_nodeRef array_length = GET_NODE(obj->size);
   tree_nodeRef array_element = GET_NODE (obj->elts);
   if (array_length->get_kind() == integer_cst_K)
   {
      integer_cst *arr_ic = GetPointer<integer_cst>(array_length);
      type_node *tn = GetPointer<type_node>(array_element);
      integer_cst *eln_ic = GetPointer<integer_cst>(GET_NODE(tn->size));
      os << boost::lexical_cast<std::string>(tree_helper::get_integer_cst_value(arr_ic) / tree_helper::get_integer_cst_value(eln_ic));
   }

   os << "]";
   mask = NO_VISIT;
   obj->type_node::visit(this);
}

void GimpleWriter::operator()(const gimple_asm* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->expr_node_vops::visit(this);
}


void GimpleWriter::operator()(const baselink* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const gimple_bind* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->expr_node::visit(this);
}

void GimpleWriter::operator()(const binfo* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const block* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const call_expr* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   const addr_expr * ae = GetPointer<addr_expr>(GET_NODE(obj->fn));
   if(ae)
   {
      const function_decl * fd = GetPointer<function_decl>(GET_NODE(ae->op));
      if(fd)
      {
         fd->name->visit(this);
         os << " (";
         const std::vector<tree_nodeRef> & args = obj->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            if(arg != args.begin())
               os << ", ";
            (*arg)->visit(this);
         }
         os << ")";
      }
   }
}

void GimpleWriter::operator()(const gimple_call* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   const addr_expr * ae = GetPointer<addr_expr>(GET_NODE(obj->fn));
   if(ae)
   {
      const function_decl * fd = GetPointer<function_decl>(GET_NODE(ae->op));
      if(fd)
      {
         fd->name->visit(this);
         os << " (";
         const std::vector<tree_nodeRef> & args = obj->args;
         std::vector<tree_nodeRef>::const_iterator arg, arg_end = args.end();
         for(arg = args.begin(); arg != arg_end; arg++)
         {
            if(arg != args.begin())
               os << ", ";
            (*arg)->visit(this);
         }
         os << ")";
      }
   }

   obj->expr_node_vops::visit(this);
}

void GimpleWriter::operator()(const case_label_expr* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   if(obj->default_flag)
      os << "default: ";
   else
      os << "case ";
   if(obj->op0 and obj->op1)
   {
      obj->op0->visit(this);
      os << " ... ";
      obj->op1->visit(this);
   }
   else if(obj->op0)
   {
      obj->op0->visit(this);
      os << ": ";
   }
   if(obj->got)
      obj->got->visit(this);
   obj->expr_node::visit(this);
}

void GimpleWriter::operator()(const complex_cst* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->cst_node::visit(this);
}

void GimpleWriter::operator()(const complex_type* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->type_node::visit(this);
}

void GimpleWriter::operator()(const compound_stmt* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const gimple_cond* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   os << "if (";
   obj->op0->visit(this);
   os << ")";
   obj->expr_node_vops::visit(this);
}

void GimpleWriter::operator()(const const_decl* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->decl_node::visit(this);
}

void GimpleWriter::operator()(const constructor* obj, unsigned int & mask)
{
   os << "{";
   os << "}";
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const decl_stmt* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const enumeral_type* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->type_node::visit(this);
}

void GimpleWriter::operator()(const exit_block_expr* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->expr_node::visit(this);
}

void GimpleWriter::operator()(const expr_stmt* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const field_decl* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   if(obj->name)
      obj->name->visit(this);
   obj->decl_node::visit(this);
   obj->attr::visit(this);
}

void GimpleWriter::operator()(const function_decl* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   os << std::endl;
   os << ";; Function ";
   const tree_nodeRef name = obj->name;
   name->visit(this);
   os << " (";
   name->visit(this);
   os << ")" << std::endl << std::endl;
   name->visit(this);
   os << " (";
   const std::vector<tree_nodeRef> & list_of_args = obj->list_of_args;
   std::vector<tree_nodeRef>::const_iterator arg, arg_end = list_of_args.end();
   for(arg = list_of_args.begin(); arg != arg_end; arg++)
   {
      if(arg != list_of_args.begin())
         os << ", ";
      const parm_decl * pd = GetPointer<parm_decl>(GET_NODE(*arg));
      pd->type->visit(this);
      os << " ";
      pd->name->visit(this);
   }

   os << ")" << std::endl;
   if(obj->body)
   {
      os << "{" << std::endl;
      obj->body->visit(this);
      os << "}" << std::endl << std::endl << std::endl;
   }
   obj->decl_node::visit(this);
   obj->attr::visit(this);
}

void GimpleWriter::operator()(const function_type* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->type_node::visit(this);
}

void GimpleWriter::operator()(const gimple_assign* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->op0->visit(this);
   os << " = ";
   obj->op1->visit(this);
   obj->expr_node_vops::visit(this);
}

void GimpleWriter::operator()(const gimple_goto* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->expr_node_vops::visit(this);
}


void GimpleWriter::operator()(const handler* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const identifier_node* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   os << obj->strg;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const integer_cst* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   os << tree_helper::get_integer_cst_value(obj);
   obj->cst_node::visit(this);
}

void GimpleWriter::operator()(const integer_type* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->type_node::visit(this);
}

void GimpleWriter::operator()(const labeled_block_expr* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->expr_node::visit(this);
}

void GimpleWriter::operator()(const label_decl* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   if(obj->name)
      obj->name->visit(this);
   else
   {
      os << "<L" << boost::lexical_cast<std::string>(obj->uid) << ">";
   }
   obj->decl_node::visit(this);
}

void GimpleWriter::operator()(const gimple_label* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->op->visit(this);
   os << ":";
   obj->expr_node_vops::visit(this);
}

void GimpleWriter::operator()(const method_type* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->function_type::visit(this);
}

void GimpleWriter::operator()(const namespace_decl* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->decl_node::visit(this);
}

void GimpleWriter::operator()(const overload* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const parm_decl* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->name->visit(this);
   obj->decl_node::visit(this);
}

void GimpleWriter::operator()(const gimple_phi* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   if(not obj->virtual_flag)
   {
      os << "  # ";
      obj->res->visit(this);
      os << " = PHI <";
      const std::vector<std::pair< tree_nodeRef, unsigned int> > & list_of_def_edge = obj->list_of_def_edge;
      std::vector<std::pair< tree_nodeRef, unsigned int> >::const_iterator edge, edge_end = list_of_def_edge.end();
      for(edge = list_of_def_edge.begin(); edge != edge_end; edge++)
      {
         if(edge != list_of_def_edge.begin())
            os << ", ";
         edge->first->visit(this);
         os << "(" << edge->second << ")";
      }
      os << ">" << std::endl;
   }
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const pointer_type* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   if(not obj->name)
   {
      obj->ptd->visit(this);
      os << " *";
   }
   obj->type_node::visit(this);
}

void GimpleWriter::operator()(const real_cst* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   os << obj->valr;
   obj->cst_node::visit(this);
}

void GimpleWriter::operator()(const real_type* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->type_node::visit(this);
}

void GimpleWriter::operator()(const record_type* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   os << "struct ";
   obj->type_node::visit(this);
}

void GimpleWriter::operator()(const reference_type* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->type_node::visit(this);
}

void GimpleWriter::operator()(const result_decl* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->decl_node::visit(this);
}

void GimpleWriter::operator()(const gimple_return* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   os << "return";
   if(obj->op)
   {
      os << " ";
      obj->op->visit(this);
   }
   obj->expr_node_vops::visit(this);
}

void GimpleWriter::operator()(const return_stmt* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const scope_ref* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->expr_node::visit(this);
}

void GimpleWriter::operator()(const ssa_name* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->var->visit(this);
   os << "_" << obj->vers;
   if(obj->default_flag)
      os << "(D)";
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const statement_list* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   const std::map<unsigned int, blocRef> & list_of_block = obj->list_of_bloc;
   std::map<unsigned int, blocRef>::const_iterator block, block_end = list_of_block.end();
   for(block = list_of_block.begin(); block != block_end; block++)
   {
      if(block->first == BB_ENTRY or block->first == BB_EXIT)
      {
         continue;
      }
      if(block->second->list_of_stmt.empty() or GET_NODE(*(block->second->list_of_stmt.begin()))->get_kind() != gimple_label_K)
      {
         os << "<bb " << block->first << ">:" << std::endl;
      }
      const std::vector<tree_nodeRef> & list_of_phi = block->second->list_of_phi;
      std::vector<tree_nodeRef>::const_iterator phi, phi_end = list_of_phi.end();
      for(phi = list_of_phi.begin(); phi != phi_end; phi++)
      {
         (*phi)->visit(this);
      }
      const std::vector<tree_nodeRef> & list_of_stmt = block->second->list_of_stmt;
      std::vector<tree_nodeRef>::const_iterator stmt, stmt_end = list_of_stmt.end();
      for(stmt = list_of_stmt.begin(); stmt != stmt_end; stmt++)
      {
         if(not GetPointer<gimple_label>(GET_NODE(*stmt)))
            os << "  ";
         (*stmt)->visit(this);
         if(not GetPointer<gimple_cond>(GET_NODE(*stmt)) and not GetPointer<gimple_label>(GET_NODE(*stmt)) and not GetPointer<gimple_switch>(GET_NODE(*stmt)))
            os << ";";
         os << std::endl;
      }
      if(block->second->true_edge and block->second->false_edge)
      {
         os << "    goto <bb " << block->second->true_edge << ">";
         const blocRef next_true = list_of_block.find(block->second->true_edge)->second;
         if(next_true->list_of_stmt.size() and GET_NODE(*(next_true->list_of_stmt.begin()))->get_kind() == gimple_label_K)
         {
            const gimple_label * le = GetPointer<gimple_label>(GET_NODE(*(next_true->list_of_stmt.begin())));
            os << " (";
            le->op->visit(this);
            os << ")";
         }
         os << ";" << std::endl;
         os << "  else" << std::endl;
         os << "    goto <bb " << block->second->false_edge << ">";
         const blocRef next_false = list_of_block.find(block->second->false_edge)->second;
         if(next_false->list_of_stmt.size() and GET_NODE(*(next_false->list_of_stmt.begin()))->get_kind() == gimple_label_K)
         {
            const gimple_label * le = GetPointer<gimple_label>(GET_NODE(*(next_false->list_of_stmt.begin())));
            os << " (";
            le->op->visit(this);
            os << ")";
         }
         os << ";" << std::endl;
      }
      else if(block->second->list_of_succ.size() == 1)
      {
         const unsigned int succ_index = *(block->second->list_of_succ.begin());
         if(succ_index != block->second->number + 1 and succ_index != BB_EXIT)
         {
            os << "  goto <bb " << succ_index << ">";
            const blocRef next = list_of_block.find(succ_index)->second;
            if(next->list_of_stmt.size() and GET_NODE(*(next->list_of_stmt.begin()))->get_kind() == gimple_label_K)
            {
               const gimple_label * le = GetPointer<gimple_label>(GET_NODE(*(next->list_of_stmt.begin())));
               os << " (";
               le->op->visit(this);
               os << ")";
            }
            os << ";" << std::endl;
         }
      }
      os << std::endl;
   }
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const string_cst* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   os << "\"" << obj->strg << "\"";
   obj->cst_node::visit(this);
}

void GimpleWriter::operator()(const gimple_switch* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   os << "switch (";
   obj->op0->visit(this);
   os << ") <";
   obj->op1->visit(this);
   os << ">";
   obj->expr_node_vops::visit(this);
}

void GimpleWriter::operator()(const target_expr* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->expr_node::visit(this);
}

void GimpleWriter::operator()(const template_decl* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->decl_node::visit(this);
}

void GimpleWriter::operator()(const tree_list* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->valu->visit(this);
   if(obj->chan)
   {
      os << ", ";
      obj->chan->visit(this);
   }
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const tree_vec* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   std::vector<tree_nodeRef>::const_iterator op, op_end = obj->list_of_op.end();
   for(op = obj->list_of_op.begin(); op != op_end; op++)
   {
      if(op != obj->list_of_op.begin())
         os << ", ";
      (*op)->visit(this);
   }
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const try_block* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const type_decl* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->name->visit(this);
   obj->decl_node::visit(this);
}

void GimpleWriter::operator()(const union_type* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->type_node::visit(this);
}

void GimpleWriter::operator()(const var_decl* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   if(obj->name)
      obj->name->visit(this);
   else
   {
      if(this->use_uid)
      {
         os << "D." << obj->uid;
      }
      else
      {
         os << "internal_" << current_node_index;
      }
   }
   obj->decl_node::visit(this);
   obj->attr::visit(this);
}

void GimpleWriter::operator()(const vector_cst* obj, unsigned int & mask)
{
   os << "{ ";
   size_t vector_size = obj->list_of_valu.size();
   for (size_t i = 0; i < vector_size; i++)
   {
      obj->list_of_valu[i]->visit(this);
      if (i != (obj->list_of_valu).size() - 1)
         os << ", ";
   }
   os << " }";

   mask = NO_VISIT;
   obj->cst_node::visit(this);
}

void GimpleWriter::operator()(const vector_type* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   if(not obj->unql)
   {
      os << "vector ";
      obj->elts->visit(this);
      mask = NO_VISIT;
      obj->type_node::visit(this);
   }
   obj->type_node::visit(this);
}

void GimpleWriter::operator()(const target_mem_ref* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->WeightedNode::visit(this);
}

void GimpleWriter::operator()(const target_mem_ref461* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->WeightedNode::visit(this);
}

void GimpleWriter::operator()(const bloc*, unsigned int & mask)
{
   mask = NO_VISIT;
}

void GimpleWriter::operator()(const while_expr* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->expr_node_vops::visit(this);
}

void GimpleWriter::operator()(const for_expr* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->while_expr::visit(this);
}

void GimpleWriter::operator()(const null_node* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const pragma_node* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
   obj->srcp::visit(this);
}

void GimpleWriter::operator()(const omp_pragma* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const omp_parallel_pragma* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->omp_pragma::visit(this);
}

void GimpleWriter::operator()(const omp_sections_pragma* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->omp_pragma::visit(this);
}

void GimpleWriter::operator()(const omp_parallel_sections_pragma* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->omp_pragma::visit(this);
}

void GimpleWriter::operator()(const omp_section_pragma* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->omp_pragma::visit(this);
}

void GimpleWriter::operator()(const omp_for_pragma * obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->omp_pragma::visit(this);
}

void GimpleWriter::operator()(const map_pragma* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const call_hw_pragma * obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const call_point_hw_pragma * obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const issue_pragma* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const blackbox_pragma* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->issue_pragma::visit(this);
}

void GimpleWriter::operator()(const profiling_pragma* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->tree_node::visit(this);
}

void GimpleWriter::operator()(const statistical_profiling* obj, unsigned int & mask)
{
   mask = NO_VISIT;
   obj->profiling_pragma::visit(this);
}

